/*
 * Project:  Petruchio
 * Filename: Petruchio.java
 * Created:  04.12.2006
 * Author:   Tim Strazny
 */
package petruchio;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import petruchio.common.PropertyLoader;
import petruchio.interfaces.Compiler;
import petruchio.interfaces.CompilerProperty;


/**
 * The main class of the compiler. Here the command line arguments are read and
 * passed to the compiler which is instantiated here, too.
 * 
 * @author Tim Strazny
 */
public class Petruchio {

    /** Default configuration file */
    public static final String CONFIGURATION = "cfg/petruchio.cfg";
    
    static {
        System.setProperty("Logger", Petruchio.class.getName());
    }
    
    private static final PropertyLoader propertyLoader = new PropertyLoader();
    
    /** The logger for this class */
    static final Logger logger = Logger.getLogger(System.getProperty("Logger", Petruchio.class.getName()));
    
    static {
        Petruchio.logger.setLevel(Level.INFO);
    }
    
    
    /**
     * The properties this tool accepts.
     * 
     * @author Tim Strazny
     */
    public enum PetruchioProperty implements CompilerProperty {
            COMPILER("Compiler", null, 
                    petruchio.interfaces.Compiler.class,
                    false, false,
                    "petruchio.compiler.Pi2PN",
                    "The compiler to use.")
        ;

        private static Map<String, CompilerProperty> key2props;
        
        private final String key;
        private final Class<?> values;
        private final boolean acceptsList;
        private final boolean mayBeEmpty;
        private final String defaultValue;
        private final String description;
        private final String category;
        
        
        /**
         * Initializes the current property.
         * 
         * @param key The key of this property
         * @param category The category of this property
         * @param values The class of the value of this property
         * @param acceptsList <code>true</code> if this property accepts a list 
         * @param value The default value of this property
         * @param description This property's description
         */
        private PetruchioProperty(final String key, final String category, 
                final Class<?> values, final boolean acceptsList,
                final boolean mayBeEmpty, final String value, final String description) {
            this.key = key;
            this.values = values;
            this.acceptsList = acceptsList;
            this.mayBeEmpty = mayBeEmpty;
            this.defaultValue = value;
            this.description = description;
            this.category = category;
            PetruchioProperty.addPetruchioProperty(this);
        }
        
        
        /**
         * Adds the property to the static map that associates keys with properties
         * and checks for disjoint keys.
         *
         * @param p The property to add
         */
        private static void addPetruchioProperty(final CompilerProperty p) {
            if(PetruchioProperty.key2props == null) {
                PetruchioProperty.key2props = new HashMap<String, CompilerProperty>();
            }
            final CompilerProperty old = PetruchioProperty.key2props.put(p.getKey(), p);
            if(old != null) {
                PetruchioProperty.key2props.put(p.getKey(), old);
                throw new IllegalArgumentException("A CompilerProperty with key '"
                        + p.getKey() + "' has already been created.");
            }
        }
        
        
        /* (non-Javadoc)
         * @see petruchio.interfaces.CompilerProperty#getFromKey(java.lang.String)
         */
        @Override
        public petruchio.interfaces.CompilerProperty getFromKey(final String k) {
            return PetruchioProperty.key2props == null ? null : PetruchioProperty.key2props.get(k);
        }
        
        /* (non-Javadoc)
         * @see petruchio.interfaces.CompilerProperty#acceptsList()
         */
        @Override
        public boolean acceptsList() { return this.acceptsList; }
        
        /* (non-Javadoc)
         * @see petruchio.interfaces.CompilerProperty#mayBeEmpty()
         */
        @Override
        public boolean mayBeEmpty() { return this.mayBeEmpty; }
        
        /* (non-Javadoc)
         * @see petruchio.interfaces.CompilerProperty#getKey()
         */
        @Override
        public String getKey() { return this.key; }
        
        /* (non-Javadoc)
         * @see petruchio.interfaces.CompilerProperty#getValueClass()
         */
        @Override
        public Class<?> getValueClass() { return this.values; }
        
        /* (non-Javadoc)
         * @see petruchio.interfaces.CompilerProperty#getDefaultValue()
         */
        @Override
        public String getDefaultValue() { return this.defaultValue; }
        
        /* (non-Javadoc)
         * @see petruchio.interfaces.CompilerProperty#getDescription()
         */
        @Override
        public String getDescription() { return this.description; }
        
        /* (non-Javadoc)
         * @see java.lang.Enum#toString()
         */
        @Override
        public String toString() { return this.getKey(); }


        /* (non-Javadoc)
         * @see petruchio.interfaces.CompilerProperty#getCategory()
         */
        @Override
        public String getCategory() {
            return this.category;
        }
        
    }
    
    
    /**
     * Reads the command line arguments and starts the compilation process.
     *
     * @param args Command line arguments
     */
    public static void main(final String[] args) {
        final Properties argProps = new Properties();
        final Properties rtProps = new Properties();
        final List<String> in = new ArrayList<String>();
        final List<String> out = new ArrayList<String>();
        
        Petruchio.readArguments(args, in, out, rtProps, argProps);
        String cfg = argProps.getProperty("-c"); 
        if(cfg == null) {
            cfg = Petruchio.CONFIGURATION;
        }
        
        // load the configuration
        final String configFile = Petruchio.getAbsolutePath(cfg);
        final List<Properties> properties = new ArrayList<Properties>();
        
        try {
        	final Properties cfgProps = new Properties();
            cfgProps.load(new FileInputStream(configFile));
            properties.add(cfgProps);
            Petruchio.logger.fine("Loaded configuration from " + configFile);
            
        } catch(final FileNotFoundException e) {
            throw new RuntimeException("File '" + configFile
                            + "' does not exists.", e);
            
        } catch(final IOException e) {
            throw new RuntimeException("Unable to read from file '" 
                            + configFile + "'.", e);
        }
        
    	properties.add(0, rtProps);
        
    	final boolean print = argProps.getProperty("-hc") != null;
    	final boolean convert = argProps.getProperty("-r") != null;
    	final String statusString = argProps.getProperty("-s");
    	final int status = statusString != null ? Integer.parseInt(statusString) : -1;
    	
        Petruchio.startCompilation(properties, in, out, print, status, convert);
    }
    
    
    /**
     * Reads the command line arguments.
     *
     * @param args  The arguments to read
     * @param in    The names of the files to compile 
     * @param out   The names of the files to compile to
     * @param rtProps Additional properties to override the config file
     * @param argP Arguments that shall be passed directly to the Petruchio class
     */
    private static void readArguments(final String[] args,
                    final List<String> in, final List<String> out,
                    final Properties rtProps, final Properties argP) {
        for(int i = 0; i < args.length; ++i) {
            final String arg = args[i];
            
            if(arg.equals("-h")) {
                Petruchio.usage();
                System.exit(0);
                
            } else if(arg.equals("-hc")) {
                argP.setProperty("-hc", "true");
                
            } else if(arg.equals("-r")) {
                argP.setProperty("-r", "true");
                
            } else if(arg.equals("-q")) {
                Petruchio.logger.setLevel(Level.WARNING);
                
            } else if(arg.equals("-v")) {
                Petruchio.logger.setLevel(Level.FINE);
                
            } else if(arg.equals("-s")) {
                ++i;
                Petruchio.expectMoreArguments(i, 1, args, "-s");
                argP.setProperty("-s", args[i]);
                
            } else if(arg.equals("-c")) {
                ++i;
                Petruchio.expectMoreArguments(i, 1, args, "-c");
                argP.setProperty("-c", args[i]);
                
            } else if(arg.equals("-p")) {
                ++i;
                Petruchio.expectMoreArguments(i, 1, args, "-p");
                final int ind = args[i].indexOf('=');
                final int index = ind < 0 ? args[i].indexOf('#') : ind;
                if(index < 0) {
                    Petruchio.logger.warning("Error reading argument '-p "
                            + args[i] + "'. Malformed. Expecting '-p <property>=<value>'.\n"
                            + "Note that windows batch may handle '=' signs as separators. Use '#' instead.");
                } else {
                    final String key = args[i].substring(0, index);
                    final String value = args[i].substring(index+1);
                    rtProps.setProperty(key, value);
                    Petruchio.logger.fine("Added property: " + key + "=" + value);
                }
                
            } else if(arg.equals("-i")) {
                ++i;
                Petruchio.expectMoreArguments(i, 1, args, "-i");
                in.add(args[i]);
                
            } else if(arg.equals("-o")) {
                ++i;
                Petruchio.expectMoreArguments(i, 1, args, "-o");
                out.add(args[i]);
                
            } else {
                final String assist = arg.indexOf('=') > 0 ? "\nDid you mean -p " + arg : "";
                
                Petruchio.logger.warning("I don't know what to do with this argument: "
                                + arg + assist + "\nTry option '-h' for a list of options.");
                
            }
        }
    }
    
    
    /**
     * Creates a new compiler and starts the compilation of the input files
     *
     * @param props The properties to pass to the compiler
     * @param in    The names of the files to compile 
     * @param out   The names of the files to compile to
     */
    public static void startCompilation(final List<Properties> props,
                    final List<String> in, final List<String> out) {
        Petruchio.startCompilation(props, in, out, false, -1, false);
    }
    
    
    /**
     * Creates a new compiler and starts the compilation of the input files
     *
     * @param props The properties to pass to the compiler
     * @param in    The names of the files to compile 
     * @param out   The names of the files to compile to
     * @param print If <code>true</code> the compiler options will be printed
     * @param status Print the compiler status every x seconds
     * @param convert Convert Petri nets instead of compiling pi processes
     */
    public static void startCompilation(final List<Properties> props,
                    final List<String> in, final List<String> out,
                    final boolean print, final int status, final boolean convert) {
        
        // create the compiler
    	Petruchio.propertyLoader.setProperties(props);
        final Compiler compiler = Petruchio.propertyLoader.loadProperty(
        		Petruchio.PetruchioProperty.COMPILER);
        if(print) {
            System.out.println(Petruchio.printOptions(
                            compiler.getClass().getCanonicalName(),
                            compiler.getOptions()));
        }
        final StatusLogger statusLogger;
        if(status > 0) {
            statusLogger = new StatusLogger(compiler, status);
            statusLogger.start();
        } else {
            statusLogger = null;
        }
        Petruchio.checkDisjointOptions(compiler);
        Petruchio.checkValidOptionKeys(compiler, props);
        for(final Properties p : props) {
        	for(final CompilerProperty cp : Petruchio.getOptions()) {
        		p.remove(cp.getKey());
        	}
        }
        compiler.loadProperties(props);
        
        Petruchio.logger.info("Compiling " + in.size() + " files.");
        final long start = System.currentTimeMillis();
        try {
            for(int i = 0; i < in.size(); ++i) {
                final String fIn = in.get(i);
                String fOut;
                if(i >= out.size()) {
                    throw new RuntimeException("No output file name specified for "
                                    + "input file " + fIn);
                }
                fOut = out.get(i);
                if(fOut.equals("-")) {
                    fOut = fIn;
                }
                
                final String inputFile = Petruchio.getAbsolutePath(fIn);
                final String outputFile = Petruchio.getAbsolutePath(fOut);
                
                Petruchio.logger.info("Compiling " + fIn + " to " + fOut);
                // the compilation should be started here
                if(convert) {
                	compiler.convert(inputFile, outputFile);
                } else {
                	compiler.compile(inputFile, outputFile);
                }
            }
        } finally {
            if(statusLogger != null) {
                statusLogger.pleaseStop();
            }
            compiler.pleaseStop();
        }
        
        Petruchio.logger.info("Done. (" + Petruchio.getDuration(start) + ")");
    }
    
    
    /**
     * TODO: description
     *
     * @param start
     * @return
     */
    static String getDuration(final long start) {
        long dur = System.currentTimeMillis() - start;
        final StringBuilder b = new StringBuilder();
        final long s = dur % 60000;
        dur /= 60000;
        b.append(s/1000.0);
        b.append("s");
        if(dur == 0) {
            return b.toString();
        }
        
        final long m = dur % 60;
        dur /= 60;
        b.insert(0, "m ");
        b.insert(0, m);
        if(dur == 0) {
            return b.toString();
        }
        
        final long h = dur % 24;
        dur /= 24;
        b.insert(0, "h ");
        b.insert(0, h);
        if(dur == 0) {
            return b.toString();
        }
        
        final long d = dur % 7;
        dur /= 7;
        b.insert(0, "d ");
        b.insert(0, d);
        if(dur == 0) {
            return b.toString();
        }
        
        final long w = dur % 52;
        dur /= 52;
        b.insert(0, "w ");
        b.insert(0, w);
        if(dur == 0) {
            return b.toString();
        }
        
        final long y = dur;
        b.insert(0, "y ");
        b.insert(0, y);
        return b.toString();
    }
    
    
    /**
     * Throws an exception if there are less arguments than expected.
     *
     * @param nr    The current index
     * @param count The number of expected arguments
     * @param args  The list of arguments
     * @param op    The option's name
     */
    private static void expectMoreArguments(final int nr, final int count, final String[] args, final String op) {
        if(nr + count > args.length) {
            throw new IllegalArgumentException("Option '" + op + "' expects "
                            + count + " argument(s). Try '-h' for help.");
        }
    }
    
    
    /**
     * Prints the usage to System.out.
     */
    private static void usage() {
        final String usage = "usage: petruchio.Petruchio [options]\n"
            + "options:\n"
            + "  -h               print this text and exit\n"
            + "  -hc              print the compiler options\n"
            + "  -r               don't compile pi processes but convert Petri\n"
            + "                   nets and additionally reduce them. input files\n"
            + "                   are treated as Petri net files."
            + "  -c <filename>    configuration to use\n"
            + "  -s <seconds>     logs the compilation status every x seconds\n"
            + "  -q               quiet mode\n"
            + "  -v,              verbose mode\n"
            + "  -p <key>=<value> additional setup (overrides config)\n"
            + "                   use # instead of = for windows compatibility\n"
            + "  -i <filename>    translate this pi process file\n"
            + "  -o <filename>    and output this petri net file\n"
            + "                   write '-' as a shorthand for the input file\n"
            + "                   name with extension '.net'\n"
            + "\n"
            + Petruchio.printOptions(Petruchio.class.getCanonicalName(), Petruchio.getOptions());
        System.out.println(usage);
    }
    
    
    /**
     * Builds an absolute path from a given file and the current user directory.
     *
     * @param file A relative or absolute path to the file
     * @return The absolute path for the given file name
     */
    private static String getAbsolutePath(final String file) {
        final File f = new File(file);
        if(!f.isAbsolute()) {
            final File f2 = new File(System.getProperty("user.dir")
                            + File.separator + file);
            try {
                return f2.getCanonicalPath();
            } catch(final IOException e) {
                return f2.getAbsolutePath();
            }
        }
        return file;
    }
    
    
    /**
     * Returns the properties this class accepts and understands.
     *
     * @return The valid properties for this class.
     */
    public static CompilerProperty[] getOptions() {
        return Petruchio.PetruchioProperty.values();
    }
    
    
    /**
     * Checks whether the properties of this main class and the properties of
     * the loaded compiler have disjoint keys.
     * If there are properties in this class and the loaded compiler a warning
     * is emitted.  
     *
     * @param c The compiler with the properties to check.
     */
    private static void checkDisjointOptions(final Compiler c) {
        final CompilerProperty[] a = c.getOptions();
        if(a.length == 0) {
            return;
        }
        final CompilerProperty[] b = Petruchio.getOptions();
        final CompilerProperty pa0 = a[0];
        for(final CompilerProperty pb : b) {
            final CompilerProperty pa = pa0.getFromKey(pb.getKey());
            if(pa != null) {
                Petruchio.logger.warning("The main class petruchio.Petruchio "
                                + "and the loaded compiler " + c.getName()
                                + " (" + c.getClass().getCanonicalName() + ") "
                                + "accept properties that share the same key:"
                                + "\nKey: " + pa.getKey()
                                + "\nCompiler description:   " + pa.getDescription()
                                + "\nCompiler value class:   " + pa.getValueClass().getCanonicalName()
                                + "\nCompiler default value: " + pa.getDefaultValue()
                                + "\nMain description:       " + pb.getDescription()
                                + "\nMain value class:       " + pb.getValueClass().getCanonicalName()
                                + "\nMain default value:     " + pb.getDefaultValue()
                                + "\n=> This might be a problem.");
            }
        }
    }
    
    
    /**
     * Checks if all the given options are known by the main class or the
     * compiler.
     *
     * @param c The compiler with the properties to check.
     * @param props The properties to check the keys of.
     */
    public static void checkValidOptionKeys(final Compiler c, final List<Properties> props) {
        final CompilerProperty cp = (c.getOptions().length == 0) ? null : c.getOptions()[0];
        final CompilerProperty mp = (Petruchio.getOptions().length == 0) ? null : Petruchio.getOptions()[0];
        for(final Properties p : props) {
	        for(final Object k : p.keySet()) {
	            final String key = (String)k;
	            if((cp == null || cp.getFromKey(key) == null)
	                            && (mp == null || mp.getFromKey(key) == null)) {
	                Petruchio.logger.warning("The property is unknown to the main "
	                                + "class and the compiler: " + key);
	            }
	        }
        }
    }
    
    
    /**
     * Checks if all the given options are known by the main class or the
     * compiler.
     *
     * @param c The compiler with the properties to check.
     * @param p The properties to check the keys of.
     * @param excluded
     */
    public static void checkValidOptionKeys(final Compiler c, final Properties p,
    		final Collection<? extends CompilerProperty> excluded) {
        final CompilerProperty cp = (c.getOptions().length == 0) ? null : c.getOptions()[0];
        final CompilerProperty mp = (Petruchio.getOptions().length == 0) ? null : Petruchio.getOptions()[0];
        for(final Object k : p.keySet()) {
            final String key = (String)k;
            if((cp == null || cp.getFromKey(key) == null)
                            && (mp == null || mp.getFromKey(key) == null)) {
                Petruchio.logger.warning("The property is unknown to the main "
                                + "class and the compiler: " + key);
            } else {
            	for(final CompilerProperty cur : excluded) {
		            if(cur.getKey().equals(k)) {
		            	Petruchio.logger.warning("Property '" + key
		            			+ "' will not be used at this place.");
		            	break;
		            }
            	}
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param clazz
     * @param options
     * @return
     */
    public static String printOptions(final String clazz, final CompilerProperty[] options) {
        final StringBuilder b = new StringBuilder("Compiler properties of class ");
        b.append(clazz);
        b.append("\n");
        for(final CompilerProperty c : options) {
            b.append(c.getKey());
            b.append("\n    ");
            b.append(c.getDescription());
            if(!c.mayBeEmpty() || c.acceptsList()) {
                b.append("\n   ");
                if(!c.mayBeEmpty()) {
                    b.append(" Compulsory.");
                }
                if(c.acceptsList()) {
                    b.append(" Accepts a comma separated list of values.");
                }
            }
            b.append("\n    Accepts: ");
            b.append(c.getValueClass().getCanonicalName());
            b.append("\n    Default: ");
            b.append(c.getDefaultValue());
            b.append("\n\n");
        }
        return b.toString();
    }
    
    
    /**
     * <Description>
     * 
     * @author Tim Strazny
     */
    private static class StatusLogger extends Thread {
        
        private final Compiler compiler;
        private final long delay;
        private boolean pleaseStop;
        private long start;
        
        /**
         * @param compiler
         * @param delay
         */
        public StatusLogger(final Compiler compiler, final int delay) {
            this.compiler = compiler;
            this.delay = 1000L*delay;
            this.pleaseStop = false;
        }
        
        
        /**
         * TODO: document
         *
         */
        public void pleaseStop() {
            this.pleaseStop = true;
            this.interrupt();
        }
        
        
        /* (non-Javadoc)
         * @see java.lang.Thread#run()
         */
        @Override
        public void run() {
            this.start = System.currentTimeMillis();
            while(!this.pleaseStop) {
                try {
                    Thread.sleep(this.delay);
                } catch(final InterruptedException e) {
                    // ignore
                }
                if(!this.pleaseStop) {
                    Petruchio.logger.info("Status after " + Petruchio.getDuration(this.start) + ":\n"
                            + this.compiler.getStatus());
                }
            }
        }
        
        
    }
    
}