/*
 * Project:  Petruchio
 * Filename: PropertyLoader.java
 * Created:  22.08.2007
 * Author:   Tim Strazny
 */

package petruchio.common;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import petruchio.interfaces.CompilerProperty;
import petruchio.interfaces.Resettable;



/**
 * A loader for CompilerProperty-type properties that does all the casting and
 * stuff.
 * 
 * @author Tim Strazny
 */
public class PropertyLoader {
    
    private List<Properties> properties;
    
    static class Prop<T> {
        boolean upToDate = false;
        T obj = null;
    }
    
    private IdentityHashMap<CompilerProperty, Prop<?>> objects = new IdentityHashMap<CompilerProperty, Prop<?>>();
    
    private Logger logger;
    
    
    /**
     * Initializes the new instance.
     */
    public PropertyLoader() {
        this.properties = null;
        this.logger = Logger.getLogger(System.getProperty("Logger", this.getClass().getName()));
    }
    
    
    /**
     * TODO: Document
     *
     * @param <T>
     * @param prop
     * @return
     */
    @SuppressWarnings("unchecked")
    public <T> T getObject(CompilerProperty prop) {
        Prop<T> p = (Prop<T>)this.objects.get(prop);
        if(p == null) {
            p = new Prop<T>();
            this.objects.put(prop, p);
            if(this.logger.isLoggable(Level.FINEST)) {
                this.logger.finest("Loading " + prop.getKey());
            }
        }
        if(!p.upToDate) {
            p.upToDate = true;
            p.obj = this.loadProperty(prop, p.obj);
        }
        return p.obj;
    }
    
    
    /**
     * TODO: Document
     *
     */
    public void markObsolete() {
        for(final Prop<?> p : this.objects.values()) {
            p.upToDate = false;
        }
    }
    
    
    /**
     * Sets the hierarchy of properties object to use.
     * 
     * @param props The hierarchy of properties to use.
     */
    public void setProperties(final List<Properties> props) {
        this.markObsolete();
        this.properties = props;
    }
    
    
    /**
     * TODO: description
     *
     * @return
     */
    public List<Properties> getProperties() {
    	return this.properties;
    }
    
    
    /**
     * TODO: description
     *
     * @param key
     * @return
     */
    private String getProperty(final String key) {
    	if(this.properties == null) {
    		return null;
    	}
    	
    	for(final Properties props : this.properties) {
    		final String value = props.getProperty(key);
    		if(value != null) {
    			return value;
    		}
    	}
    	return null;
    }
    
    
    /**
     * Reads an String property from the compiler properties.
     *
     * @param key The property name
     * @param defaultValue Value to use if the property has not been set
     * @return The value of the property if it has been set, if it has not been
     *         set the default value will be returned. 
     */
    public String loadStringProperty(final CompilerProperty property) {
        String s = this.getProperty(property.getKey());
        if(s != null) {
        	s = s.trim();
        }
        final String value;
        if(((s == null) || (s.length() == 0)) && !property.mayBeEmpty()) {
            this.logger.warning("Property '" + property.getKey() + "' is not "
                    + "set correctly. It expects a value. Using default "
                    + "value '" + property.getDefaultValue() + "'.");
            value = property.getDefaultValue();
        } else {
            value = s;
        }
        return value;
    }
    

    /**
     * Tries to load a property.
     * 
     * @param <T> The type of the property to load.
     * @param property The property to load.
     * @return The property's value.
     */
    @SuppressWarnings("unchecked")
    public <T> T loadProperty(final CompilerProperty property) {
        if(property.acceptsList()) {
            return (T)this.loadList(property, null);
        }
        if(property.getValueClass().isEnum()) {
            return (T)this.loadEnum(property);
        } else if(PropertyLoader.getPrimitive(property.getValueClass()) != null) {
            return (T)this.loadPrimitive(property);
        } else {
            return (T)this.loadClass(property, null);
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param <T>
     * @param property
     * @param old
     * @return
     */
    @SuppressWarnings({ "unchecked", "cast" })
    public <T> T loadProperty(final CompilerProperty property, final T old) {
        if(property.acceptsList()) {
            return (T)this.loadList(property, (List<T>)old);
        }
        if(property.getValueClass().isEnum()) {
            return (T)this.loadEnum(property);
        } else if(PropertyLoader.getPrimitive(property.getValueClass()) != null) {
            return (T)this.loadPrimitive(property);
        } else {
            return (T)this.loadClass(property, old);
        }
    }
    
    
    /**
     * Tries to load, instantiate and cast the class that is associated with the 
     * given property key.
     *
     * @param <T> The type to cast to.
     * @param property The property key to look up.
     * @return An instance of the associated class.
     */
    @SuppressWarnings("cast")
    private <T> T loadClass(final CompilerProperty property, final T old) {
        final String className = this.loadStringProperty(property);
        if(property.mayBeEmpty() && ((className == null) || (className.length() == 0))) {
            return null;
        }
    	return (T)this.loadClass(className, property, old);
    }
    
    
    /**
     * Tries to load, instantiate and cast the class that is associated with the 
     * given property key.
     *
     * @param <T> The type to cast to.
     * @param property The property key to look up.
     * @return An instance of the associated class.
     */
    @SuppressWarnings("unchecked")
    private <T> T loadEnum(final CompilerProperty property) {
        final String className = this.loadStringProperty(property);
        if(property.mayBeEmpty() && ((className == null) || (className.length() == 0))) {
            return null;
        }
    	return (T)this.loadEnum(className, property);
    }
    
    
    /**
     * Tries to load, instantiate and cast the class that is associated with the 
     * given property key.
     *
     * @param <T> The type to cast to.
     * @param className The name of the class to be instantiated.
     * @param property The property key to look up.
     * @return An instance of the associated class.
     */
    @SuppressWarnings("unchecked")
    private <T> T loadClass(final String className, final CompilerProperty property,
    		final T old) {
    	if(old != null) {
    		if(old instanceof Resettable) {
    			((Resettable)old).reset();
    		}
    		if(old.getClass().getName().equals(className)) {
    			return old;
    		}
    	}
    	
        try {
            final Class<?> clazz = Thread.currentThread().getContextClassLoader().loadClass(className);
            final Constructor<?>[] c = clazz.getConstructors();
            Constructor<?> min = null;
            for(final Constructor<?> con : c) {
                if((min == null) || (con.getParameterTypes().length < min.getParameterTypes().length)) {
                    min = con;
                }
            }
            if(min == null) {
                throw new InternalError();
            }
            final T newInstance = (T)min.newInstance(new Object[min.getParameterTypes().length]);
            return newInstance;
            
        } catch(final NullPointerException e) {
            throw new RuntimeException("Property '" + property
                            + "' must be set.", e);
            
        } catch(final ClassNotFoundException e) {
            throw new RuntimeException("Unable to find class '" + className
                            + "' for property '" + property + "'.", e);
            
        } catch(final IllegalAccessException e) {
            throw new RuntimeException("Cannot access class '" + className
                            + "' for property '" + property + "'.", e);
            
        } catch(final InstantiationException e) {
            throw new RuntimeException("Unable to instantiate class '"
                            + className
                            + "' for property '" + property + "'.", e);
        } catch(final InvocationTargetException e) {
            throw new RuntimeException("Unable to instantiate class '"
                            + className
                            + "' for property '" + property + "'.", e);
            
        } catch(final ClassCastException e) {
            throw new RuntimeException("Class '" + className + "' for property '"
                            + property + "' does not implement the required "
                            + "interfaces and cannot be cast correctly.", e);
        }
    }
    
    
    /**
     * Tries to load, instantiate and cast the class that is associated with the 
     * given property key.
     *
     * @param <T> The type to cast to.
     * @param property The property key to look up.
     * @return An instance of the associated class.
     */
    @SuppressWarnings("unchecked")
    private <T> T loadPrimitive(final CompilerProperty property) {
        final String value = this.loadStringProperty(property);
        if(property.mayBeEmpty() && ((value == null) || (value.length() == 0))) {
            return null;
        }
        return (T)this.loadPrimitive(value, property);
    }
    
    
    /**
     * Returns the class corresponding to the primitive class.
     * 
     * @param c A class object
     * @return <code>null</code> if the class is no primitive and no String class,
     *         else the class corresponding to the primitive is returned.
     */
    private static Class<?> getPrimitive(final Class<?> c) {
        if(c == Boolean.class || c == Boolean.TYPE) {
            return Boolean.class;
        } else if(c == Integer.class || c == Integer.TYPE) {
            return Integer.class;
        } else if(c == String.class) {
            return String.class;
        } else if(c == Character.class || c == Character.TYPE) {
            return Character.class;
        } else if(c == Double.class || c == Double.TYPE) {
            return Double.class;
        } else if(c == Float.class || c == Float.TYPE) {
            return Float.class;
        } else if(c == Short.class || c == Short.TYPE) {
            return Short.class;
        } else if(c == Byte.class || c == Byte.TYPE) {
            return Byte.class;
        } else if(c == Long.class || c == Long.TYPE) {
            return Long.class;
        } else {
            return null;
        }
    }
    
    
    /**
     * Tries to load, instantiate and cast the class that is associated with the 
     * given property key.
     *
     * @param <T> The type to cast to.
     * @param property The property key to look up.
     * @return An instance of the associated class.
     */
    @SuppressWarnings("unchecked")
    private <T> T loadPrimitive(final String value, final CompilerProperty property) {
        try {
            final Constructor<?>[] c = PropertyLoader.getPrimitive(property.getValueClass()).getConstructors();
            Constructor<?> min = null;
            for(final Constructor<?> con : c) {
                if((con.getParameterTypes().length == 1)
                        && (con.getParameterTypes()[0] == String.class)) {
                    min = con;
                    break;
                }
            }
            if(min == null || min.getParameterTypes().length == 0) {
                throw new InternalError();
            }
            final Object[] ar = new Object[min.getParameterTypes().length];
            ar[0] = value;
            final T newInstance = (T)min.newInstance(ar);
            return newInstance;
            
        } catch(final NullPointerException e) {
            throw new RuntimeException("Property '" + property
                            + "' must be set.", e);
            
        } catch(final IllegalAccessException e) {
            throw new RuntimeException("Cannot access class '" + value
                            + "' for property '" + property + "'.", e);
            
        } catch(final InstantiationException e) {
            throw new RuntimeException("Unable to instantiate class '"
                            + value
                            + "' for property '" + property + "'.", e);
        } catch(final InvocationTargetException e) {
            throw new RuntimeException("Unable to instantiate class '"
                            + value
                            + "' for property '" + property + "'.", e);
            
        } catch(final ClassCastException e) {
            throw new RuntimeException("Class '" + value + "' for property '"
                            + property + "' does not implement the required "
                            + "interfaces and cannot be cast correctly.", e);
        }
    }
    
    
    /**
     * Tries to load, instantiate and cast the class that is associated with the 
     * given property key.
     *
     * @param <T> The type to cast to.
     * @param property The property key to look up.
     * @return An instance of the associated class.
     */
    @SuppressWarnings("unchecked")
    private <T> T loadEnum(final String className, final CompilerProperty property) {
        T[] values = null;
        try {
            values = (T[])property.getValueClass().getEnumConstants();
            for(final T value : values) {
                if(className.equalsIgnoreCase(((Enum<?>)value).name())) {
                    return value;
                }
            }
            throw new IllegalArgumentException("Enum not found: " + className);
            
        } catch(final IllegalArgumentException e) {
            throw new RuntimeException("Valid values are " + Arrays.toString(values), e);
            
        } catch(final NullPointerException e) {
            throw new RuntimeException("Property '" + property
                            + "' must be set.", e);
            
        } catch(final ClassCastException e) {
            throw new RuntimeException("Class '" + className + "' for property '"
                            + property + "' does not implement the required "
                            + "interfaces and cannot be cast correctly.", e);
        }
    }
    
    
    /**
     * Loads a list of classes or values.
     *
     * @param <T> The type of the list entries.
     * @param property The property to load.
     * @returns The list entries (values or instances).
     */
    @SuppressWarnings("unchecked")
    private <T> List<T> loadList(final CompilerProperty property, final List<T> old) {
        final List<T> c = new ArrayList<T>();
        final String list = this.loadStringProperty(property);
        if(list != null && list.length() != 0) {
            final boolean isEnum = property.getValueClass().isEnum();
            final boolean isPrimitive = PropertyLoader.getPrimitive(property.getValueClass()) != null;
            final String[] names = list.trim().split("[,;]\\s*");
            int i = 0;
            for(String s : names) {
                s = s.trim();
                if(s.length() == 0) {
                    continue;
                }
                if(isEnum) {
                    c.add((T)this.loadEnum(s, property));
                } else if(isPrimitive) {
                    c.add((T)this.loadPrimitive(s, property));
                } else {
                    c.add(this.loadClass(s, property,
                    		((old != null) && (i < old.size()) ? old.get(i++) : null)));
                }
            }
        }
        if(!property.mayBeEmpty() && c.isEmpty()) {
            throw new IllegalArgumentException("Property " + property
                            + " must be set to at least one valid value.");
        }
        return c;
    }
    
}
