/*
 * Project:  Petruchio
 * Filename: WeakCollection.java
 * Created:  12.12.2006
 * Author:   Tim Strazny
 * Remarks:  This class is based on class java.util.ArrayList.
 *           It has beend slightly modified to support weak references,
 *           but basically most of the code comes from Sun's ArrayList class.
 */
package petruchio.common;

import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.AbstractCollection;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * Resizable-array implementation of the <tt>List</tt> interface. Implements
 * all optional list operations, and permits all elements, including
 * <tt>null</tt>. In addition to implementing the <tt>List</tt> interface,
 * this class provides methods to manipulate the size of the array that is used
 * internally to store the list. (This class is roughly equivalent to
 * <tt>Vector</tt>, except that it is unsynchronized.)
 * <p>
 * 
 * The <tt>size</tt>, <tt>isEmpty</tt>, <tt>get</tt>, <tt>set</tt>,
 * <tt>iterator</tt>, and <tt>listIterator</tt> operations run in constant
 * time. The <tt>add</tt> operation runs in <i>amortized constant time</i>,
 * that is, adding n elements requires O(n) time. All of the other operations
 * run in linear time (roughly speaking). The constant factor is low compared to
 * that for the <tt>LinkedList</tt> implementation.
 * <p>
 * 
 * Each <tt>WeakCollection</tt> instance has a <i>capacity</i>. The capacity
 * is the size of the array used to store the elements in the list. It is always
 * at least as large as the list size. As elements are added to an
 * WeakCollection, its capacity grows automatically. The details of the growth
 * policy are not specified beyond the fact that adding an element has constant
 * amortized time cost.
 * <p>
 * 
 * An application can increase the capacity of an <tt>WeakCollection</tt>
 * instance before adding a large number of elements using the
 * <tt>ensureCapacity</tt> operation. This may reduce the amount of
 * incremental reallocation.
 * <p>
 * 
 * <strong>Note that this implementation is not synchronized.</strong> If
 * multiple threads access an <tt>WeakCollection</tt> instance concurrently,
 * and at least one of the threads modifies the list structurally, it <i>must</i>
 * be synchronized externally. (A structural modification is any operation that
 * adds or deletes one or more elements, or explicitly resizes the backing
 * array; merely setting the value of an element is not a structural
 * modification.) This is typically accomplished by synchronizing on some object
 * that naturally encapsulates the list. If no such object exists, the list
 * should be "wrapped" using the <tt>Collections.synchronizedList</tt> method.
 * This is best done at creation time, to prevent accidental unsynchronized
 * access to the list:
 * 
 * <pre>
 *   List list = Collections.synchronizedList(new WeakCollection(...));
 * </pre>
 * 
 * <p>
 * 
 * The iterators returned by this class's <tt>iterator</tt> and
 * <tt>listIterator</tt> methods are <i>fail-fast</i>: if list is
 * structurally modified at any time after the iterator is created, in any way
 * except through the iterator's own remove or add methods, the iterator will
 * throw a ConcurrentModificationException. Thus, in the face of concurrent
 * modification, the iterator fails quickly and cleanly, rather than risking
 * arbitrary, non-deterministic behavior at an undetermined time in the future.
 * <p>
 * 
 * Note that the fail-fast behavior of an iterator cannot be guaranteed as it
 * is, generally speaking, impossible to make any hard guarantees in the
 * presence of unsynchronized concurrent modification. Fail-fast iterators throw
 * <tt>ConcurrentModificationException</tt> on a best-effort basis. Therefore,
 * it would be wrong to write a program that depended on this exception for its
 * correctness: <i>the fail-fast behavior of iterators should be used only to
 * detect bugs.</i>
 * <p>
 * 
 * This class is a member of the <a href="{@docRoot}/../guide/collections/index.html">
 * Java Collections Framework</a>.
 * 
 * @author Josh Bloch
 * @author Neal Gafter
 * @version 1.50, 09/02/05
 * @see Collection
 * @see List
 * @see LinkedList
 * @see Vector
 * @see Collections#synchronizedList(List)
 * @since 1.2
 * @param <E>
 */

public class WeakCollection<E> extends AbstractCollection<E>
                    implements Cloneable, java.io.Serializable {

    private static final long serialVersionUID = 8683452581122892189L;

    /**
     * The array buffer into which the elements of the WeakCollection are
     * stored. The capacity of the WeakCollection is the length of this array
     * buffer.
     */
    private transient Entry<E>[] elementData;

    /**
     * The size of the WeakCollection (the number of elements it contains).
     * 
     * @serial
     */
    private int size;
    
    int modCount = 0;

    private transient final ReferenceQueue<E> queue = new ReferenceQueue<E>();


    /**
     * Constructs an empty list with the specified initial capacity.
     * 
     * @param initialCapacity the initial capacity of the list.
     * @exception IllegalArgumentException if the specified initial capacity is
     *                negative
     */
    @SuppressWarnings("unchecked")
    public WeakCollection(final int initialCapacity) {
        super();
        if (initialCapacity < 0) {
            throw new IllegalArgumentException("Illegal Capacity: "
                            + initialCapacity);
        }
        this.elementData = new Entry[initialCapacity];
    }


    /**
     * Constructs an empty list with an initial capacity of ten.
     */
    public WeakCollection() {
        this(10);
    }


    /**
     * Constructs a list containing the elements of the specified collection, in
     * the order they are returned by the collection's iterator. The
     * <tt>WeakCollection</tt> instance has an initial capacity of 110% the
     * size of the specified collection.
     * 
     * @param c the collection whose elements are to be placed into this list.
     * @throws NullPointerException if the specified collection is null.
     */
    @SuppressWarnings("unchecked")
    public WeakCollection(final Collection<? extends E> c) {
        this.size = c.size();
        // Allow 10% room for growth
        final int capacity = (int) Math.min((this.size * 110L) / 100, Integer.MAX_VALUE);
        this.elementData = c.toArray(new Entry[capacity]);
    }


    /**
     * Trims the capacity of this <tt>WeakCollection</tt> instance to be the
     * list's current size. An application can use this operation to minimize
     * the storage of an <tt>WeakCollection</tt> instance.
     */
    @SuppressWarnings("unchecked")
    public void trimToSize() {
        this.modCount++;
        this.expungeStaleEntries();
        final int oldCapacity = this.elementData.length;
        if(this.size < oldCapacity) {
            final Object oldData[] = this.elementData;
            this.elementData = new Entry[this.size];
            System.arraycopy(oldData, 0, this.elementData, 0, this.size);
        }
    }


    /**
     * Increases the capacity of this <tt>WeakCollection</tt> instance, if
     * necessary, to ensure that it can hold at least the number of elements
     * specified by the minimum capacity argument.
     * 
     * @param minCapacity the desired minimum capacity.
     */
    @SuppressWarnings("unchecked")
    public void ensureCapacity(final int minCapacity) {
        this.modCount++;
        final int oldCapacity = this.elementData.length;
        if (minCapacity > oldCapacity) {
            final Object oldData[] = this.elementData;
            int newCapacity = (oldCapacity * 3) / 2 + 1;
            if (newCapacity < minCapacity) {
                newCapacity = minCapacity;
            }
            this.elementData = new Entry[newCapacity];
            System.arraycopy(oldData, 0, this.elementData, 0, this.size);
        }
    }
    

    /**
     * TODO: document
     *
     */
    @SuppressWarnings("unchecked")
    private void expungeStaleEntries() {
        Entry<E> e;
        while ((e = (Entry<E>) this.queue.poll()) != null) {
            this.fastRemove(e.getIndex());
        }
    }


    /**
     * Returns the number of elements in this list.
     * 
     * @return the number of elements in this list.
     */
    @Override
    public int size() {
        if(this.size == 0) {
            return 0;
        }
        this.expungeStaleEntries();
        return this.size;
    }


    /**
     * Tests if this list has no elements.
     * 
     * @return <tt>true</tt> if this list has no elements; <tt>false</tt>
     *         otherwise.
     */
    @Override
    public boolean isEmpty() {
        return this.size() == 0;
    }


    /**
     * Returns <tt>true</tt> if this list contains the specified element.
     * 
     * @param elem element whose presence in this List is to be tested.
     * @return <code>true</code> if the specified element is present;
     *         <code>false</code> otherwise.
     */
    @Override
    public boolean contains(final Object elem) {
        return this.indexOf(elem) >= 0;
    }


    /**
     * Searches for the first occurence of the given argument, testing for
     * equality using the <tt>equals</tt> method.
     * 
     * @param elem an object.
     * @return the index of the first occurrence of the argument in this list;
     *         returns <tt>-1</tt> if the object is not found.
     * @see Object#equals(Object)
     */
    private int indexOf(final Object elem) {
        this.expungeStaleEntries();
        if (elem == null) {
            for (int i = 0; i < this.size; i++) {
                if (this.elementData[i].get() == null) {
                    return i;
                }
            }
        } else {
            for (int i = 0; i < this.size; i++) {
                if (elem.equals(this.elementData[i].get())) {
                    return i;
                }
            }
        }
        return -1;
    }


//    /**
//     * Returns the index of the last occurrence of the specified object in this
//     * list.
//     * 
//     * @param elem the desired element.
//     * @return the index of the last occurrence of the specified object in this
//     *         list; returns -1 if the object is not found.
//     */
//    private int lastIndexOf(final Object elem) {
//        this.expungeStaleEntries();
//        if (elem == null) {
//            for (int i = this.size - 1; i >= 0; i--) {
//                if (this.elementData[i].get() == null) {
//                    return i;
//                }
//            }
//        } else {
//            for (int i = this.size - 1; i >= 0; i--) {
//                if (elem.equals(this.elementData[i].get())) {
//                    return i;
//                }
//            }
//        }
//        return -1;
//    }


    /**
     * Returns a shallow copy of this <tt>WeakCollection</tt> instance. (The
     * elements themselves are not copied.)
     * 
     * @return a clone of this <tt>WeakCollection</tt> instance.
     */
    @Override
    public Object clone() {
        this.expungeStaleEntries();
        try {
            @SuppressWarnings("unchecked")
            final WeakCollection<E> v = (WeakCollection<E>) super.clone();
            @SuppressWarnings("unchecked")
            final Entry<E>[] e = new Entry[this.size]; 
            v.elementData = e;
            for (int i = 0; i < this.size; ++i) {
                v.elementData[i] = new Entry<E>(this.elementData[i].get(), i, v.queue);
            }
            v.modCount = 0;
            return v;
        } catch (final CloneNotSupportedException e) {
            // this shouldn't happen, since we are Cloneable
            throw new InternalError();
        }
    }


    /**
     * Returns an array containing all of the elements in this list in the
     * correct order.
     * 
     * @return an array containing all of the elements in this list in the
     *         correct order.
     */
    @Override
    public Object[] toArray() {
        this.expungeStaleEntries();
        final Object[] result = new Object[this.size];
        for (int i = 0; i < this.size; ++i) {
            result[i] = this.elementData[i].get();
        }
        return result;
    }


    /**
     * Returns an array containing all of the elements in this list in the
     * correct order; the runtime type of the returned array is that of the
     * specified array. If the list fits in the specified array, it is returned
     * therein. Otherwise, a new array is allocated with the runtime type of the
     * specified array and the size of this list.
     * <p>
     * 
     * If the list fits in the specified array with room to spare (i.e., the
     * array has more elements than the list), the element in the array
     * immediately following the end of the collection is set to <tt>null</tt>.
     * This is useful in determining the length of the list <i>only</i> if the
     * caller knows that the list does not contain any <tt>null</tt> elements.
     * 
     * @param a the array into which the elements of the list are to be stored,
     *            if it is big enough; otherwise, a new array of the same
     *            runtime type is allocated for this purpose.
     * @return an array containing the elements of the list.
     * @throws ArrayStoreException if the runtime type of a is not a supertype
     *             of the runtime type of every element in this list.
     */
    @Override
    @SuppressWarnings("unchecked")
    public <T> T[] toArray(final T[] b) {
        this.expungeStaleEntries();
        final T[] a;
        if (b.length < this.size) {
            a = (T[]) java.lang.reflect.Array.newInstance(
                            b.getClass().getComponentType(), this.size);
        } else {
            a = b;
        }
        for (int i = 0; i < this.size; i++) {
            a[i] = (T) this.elementData[i].get();
        }
        if (a.length > this.size) {
            a[this.size] = null;
        }
        return a;
    }


    // Positional Access Operations

    /**
     * Returns the element at the specified position in this list.
     * 
     * @param index index of element to return.
     * @return the element at the specified position in this list.
     * @throws IndexOutOfBoundsException if index is out of range <tt>(index
     *        &lt; 0 || index &gt;= size())</tt>.
     */
    public E get(final int index) {
        this.expungeStaleEntries();
        this.RangeCheck(index);
        return this.elementData[index].get();
    }


    /**
     * Replaces the element at the specified position in this list with the
     * specified element.
     * 
     * @param index index of element to replace.
     * @param element element to be stored at the specified position.
     * @return the element previously at the specified position.
     * @throws IndexOutOfBoundsException if index out of range
     *             <tt>(index &lt; 0 || index &gt;= size())</tt>.
     */
    public E set(final int index, final E element) {
        this.expungeStaleEntries();
        this.RangeCheck(index);
        
        final E oldValue = this.elementData[index].get();
        this.elementData[index].clear();

        this.elementData[index] = new Entry<E>(element, index, this.queue);
        return oldValue;
    }


    /**
     * Appends the specified element to the end of this list.
     * 
     * @param o element to be appended to this list.
     * @return <tt>true</tt> (as per the general contract of Collection.add).
     */
    @Override
    public boolean add(final E o) {
        this.expungeStaleEntries();
        this.ensureCapacity(this.size + 1); // Increments modCount!!
        this.elementData[this.size++] = new Entry<E>(o, this.size-1, this.queue);
        return true;
    }


    /**
     * Inserts the specified element at the specified position in this list.
     * Shifts the element currently at that position (if any) and any subsequent
     * elements to the right (adds one to their indices).
     * 
     * @param index index at which the specified element is to be inserted.
     * @param element element to be inserted.
     * @throws IndexOutOfBoundsException if index is out of range
     *             <tt>(index &lt; 0 || index &gt; size())</tt>.
     */
    public void add(final int index, final E element) {
        this.expungeStaleEntries();
        if (index > this.size || index < 0) {
            throw new IndexOutOfBoundsException("Index: " + index + ", Size: "
                            + this.size);
        }

        this.ensureCapacity(this.size + 1); // Increments modCount!!
        System.arraycopy(this.elementData, index, this.elementData, index + 1,
                        this.size - index);
        for(int i = index+1; i < this.size; i++) {
            this.elementData[i].setIndex(i);
        }
        this.elementData[index] = new Entry<E>(element, index, this.queue);
        this.size++;
    }


//    /**
//     * Removes the element at the specified position in this list. Shifts any
//     * subsequent elements to the left (subtracts one from their indices).
//     * 
//     * @param index the index of the element to removed.
//     * @return the element that was removed from the list.
//     * @throws IndexOutOfBoundsException if index out of range <tt>(index
//     *        &lt; 0 || index &gt;= size())</tt>.
//     */
//    private E remove(final int index) {
//        this.expungeStaleEntries();
//        RangeCheck(index);
//
//        this.modCount++;
//        final E oldValue = this.elementData[index].get();
//        this.elementData[index].clear();
//        this.elementData[index] = this.elementData[this.size - 1];
//        this.elementData[index].setIndex(index);
//        this.elementData[--this.size] = null; // Let gc do its work
//
//        return oldValue;
//    }


    /**
     * Removes a single instance of the specified element from this list, if it
     * is present (optional operation). More formally, removes an element
     * <tt>e</tt> such that <tt>(o==null ? e==null :
     * o.equals(e))</tt>, if
     * the list contains one or more such elements. Returns <tt>true</tt> if
     * the list contained the specified element (or equivalently, if the list
     * changed as a result of the call).
     * <p>
     * 
     * @param o element to be removed from this list, if present.
     * @return <tt>true</tt> if the list contained the specified element.
     */
    @Override
    public boolean remove(final Object o) {
        this.expungeStaleEntries();
        if (o == null) {
            for(int index = 0; index < this.size; index++)
                if(this.elementData[index].get() == null) {
                    fastRemove(index);
                    return true;
                }
        } else {
            for (int index = 0; index < this.size; index++)
                if (o.equals(this.elementData[index].get())) {
                    fastRemove(index);
                    return true;
                }
        }
        return false;
    }


    /*
     * Private remove method that skips bounds checking and does not return the
     * value removed.
     */
    void fastRemove(final int index) {
        this.modCount++;
        this.elementData[index].clear();
        this.elementData[index] = this.elementData[this.size-1];
        this.elementData[index].setIndex(index);
        this.elementData[--this.size] = null; // Let gc do its work
    }


    /**
     * Removes all of the elements from this list. The list will be empty after
     * this call returns.
     */
    @Override
    public void clear() {
        this.modCount++;

        while (this.queue.poll() != null) // empty the queue
            /* nothing */;
        
        // Let gc do its work
        for (int i = 0; i < this.size; i++) {
            this.elementData[i] = null;
        }
        
        while (this.queue.poll() != null) // empty the queue
            /* nothing */;

        this.size = 0;
    }


    /**
     * Appends all of the elements in the specified Collection to the end of
     * this list, in the order that they are returned by the specified
     * Collection's Iterator. The behavior of this operation is undefined if the
     * specified Collection is modified while the operation is in progress.
     * (This implies that the behavior of this call is undefined if the
     * specified Collection is this list, and this list is nonempty.)
     * 
     * @param c the elements to be inserted into this list.
     * @return <tt>true</tt> if this list changed as a result of the call.
     * @throws NullPointerException if the specified collection is null.
     */
    @Override
    @SuppressWarnings({ "unchecked", "null" })
    public boolean addAll(final Collection<? extends E> c) {
        this.expungeStaleEntries();
        
        final boolean instance = c instanceof WeakCollection<?>;
        final Object[] a;
        final Entry<? extends E>[] b;
        final int numNew;
        if(instance) {
            a = null;
            b = ((WeakCollection<E>)c).elementData;
            numNew = b.length;
        } else {
            a = c.toArray();
            numNew = a.length;
            b = null;
        }
        
        this.ensureCapacity(this.size + numNew); // Increments modCount
        for(int i = this.size; i < this.size+numNew; i++) {
            this.elementData[i] = instance ? new Entry<E>(b[i].get(), i, this.queue)
                                           : new Entry<E>((E)a[i], i, this.queue);
        }
        this.size += numNew;
        return numNew != 0;
            
    }


//    /**
//     * Inserts all of the elements in the specified Collection into this list,
//     * starting at the specified position. Shifts the element currently at that
//     * position (if any) and any subsequent elements to the right (increases
//     * their indices). The new elements will appear in the list in the order
//     * that they are returned by the specified Collection's iterator.
//     * 
//     * @param index index at which to insert first element from the specified
//     *            collection.
//     * @param c elements to be inserted into this list.
//     * @return <tt>true</tt> if this list changed as a result of the call.
//     * @throws IndexOutOfBoundsException if index out of range <tt>(index
//     *        &lt; 0 || index &gt; size())</tt>.
//     * @throws NullPointerException if the specified Collection is null.
//     */
//    @SuppressWarnings("unchecked")
//    private boolean addAll(final int index, final Collection<? extends E> c) {
//        this.expungeStaleEntries();
//        if (index > this.size || index < 0) {
//            throw new IndexOutOfBoundsException("Index: " + index + ", Size: "
//                            + this.size);
//        }
//
//        final boolean instance = c instanceof WeakCollection;
//        
//        final Object[] a;
//        final Entry<? extends E>[] b;
//        if(instance) {
//            a = null;
//            b = ((WeakCollection)c).elementData;
//        } else {
//            a = c.toArray();
//            b = null;
//        }
//        
//        final int numNew = a.length;
//        this.ensureCapacity(this.size + numNew); // Increments modCount
//
//        final int numMoved = this.size - index;
//        if (numMoved > 0) {
//            System.arraycopy(this.elementData, index, this.elementData,
//                            index + numNew, numMoved);
//            for(int i = index+numNew; i < this.size+numNew ; i++) {
//                this.elementData[i].setIndex(i);
//            }
//        }
//
//        for(int i = index; i < index+numNew; i++) {
//            this.elementData[i] = instance ? new Entry<E>(b[i].get(), i, this.queue)
//                                           : new Entry<E>((E)a[i], i, this.queue);
//        }
//        this.size += numNew;
//        return numNew != 0;
//    }


//    /**
//     * Removes from this List all of the elements whose index is between
//     * fromIndex, inclusive and toIndex, exclusive. Shifts any succeeding
//     * elements to the left (reduces their index). This call shortens the list
//     * by <tt>(toIndex - fromIndex)</tt> elements. (If
//     * <tt>toIndex==fromIndex</tt>, this operation has no effect.)
//     * 
//     * @param fromIndex index of first element to be removed.
//     * @param toIndex index after last element to be removed.
//     */
//    private void removeRange(final int fromIndex, final int toIndex) {
//        this.expungeStaleEntries();
//        this.modCount++;
//        final int numMoved = this.size - toIndex;
//        for(int i = fromIndex; i < toIndex; ++i) {
//            this.elementData[i].clear();
//        }
//        System.arraycopy(this.elementData, toIndex, this.elementData,
//                        fromIndex, numMoved);
//        for(int i = fromIndex; i < toIndex; ++i) {
//            this.elementData[i].setIndex(i);
//        }
//
//        // Let gc do its work
//        final int newSize = this.size - (toIndex - fromIndex);
//        while (this.size != newSize) {
//            --this.size;
//            this.elementData[this.size] = null;
//        }
//    }


    /**
     * Check if the given index is in range. If not, throw an appropriate
     * runtime exception. This method does *not* check if the index is negative:
     * It is always used immediately prior to an array access, which throws an
     * ArrayIndexOutOfBoundsException if index is negative.
     */
    private void RangeCheck(final int index) {
        if (index >= this.size) {
            throw new IndexOutOfBoundsException("Index: " + index + ", Size: "
                            + this.size);
        }
    }


    /**
     * Save the state of the <tt>WeakCollection</tt> instance to a stream
     * (that is, serialize it).
     * 
     * @serialData The length of the array backing the <tt>WeakCollection</tt>
     *             instance is emitted (int), followed by all of its elements
     *             (each an <tt>Object</tt>) in the proper order.
     */
    private void writeObject(final java.io.ObjectOutputStream s)
                    throws java.io.IOException {
        final int expectedModCount = this.modCount;
        // Write out element count, and any hidden stuff
        s.defaultWriteObject();

        // Write out array length
        s.writeInt(this.elementData.length);

        // Write out all elements in the proper order.
        for (int i = 0; i < this.size; i++)
            s.writeObject(this.elementData[i].get());

        if (this.modCount != expectedModCount) {
            throw new ConcurrentModificationException();
        }
    }


    /**
     * Reconstitute the <tt>WeakCollection</tt> instance from a stream (that
     * is, deserialize it).
     */
    @SuppressWarnings("unchecked")
    private void readObject(final java.io.ObjectInputStream s)
                    throws java.io.IOException, ClassNotFoundException {
        // Read in size, and any hidden stuff
        s.defaultReadObject();

        // Read in array length and allocate array
        int arrayLength = s.readInt();
        final Object[] a = this.elementData = new Entry[arrayLength];

        // Read in all elements in the proper order.
        for (int i = 0; i < this.size; i++) {
            a[i] = new Entry<E>((E)s.readObject(), i, this.queue);
        }
    }


    private static class Entry<E> extends WeakReference<E> {

        private int index;


        public Entry(final E e, final int index, final ReferenceQueue<E> queue) {
            super(e, queue);
            this.index = index;
        }


        public int getIndex() {
            return this.index;
        }


        public void setIndex(final int index) {
            this.index = index;
        }
    }


    /**
     * Returns an iterator over the elements in this list in proper
     * sequence. <p>
     *
     * This implementation returns a straightforward implementation of the
     * iterator interface, relying on the backing list's <tt>size()</tt>,
     * <tt>get(int)</tt>, and <tt>remove(int)</tt> methods.<p>
     *
     * Note that the iterator returned by this method will throw an
     * <tt>UnsupportedOperationException</tt> in response to its
     * <tt>remove</tt> method unless the list's <tt>remove(int)</tt> method is
     * overridden.<p>
     *
     * This implementation can be made to throw runtime exceptions in the face
     * of concurrent modification, as described in the specification for the
     * (protected) <tt>modCount</tt> field.
     *
     * @return an iterator over the elements in this list in proper sequence.
     * 
     * @see #modCount
     */
    @Override
    public Iterator<E> iterator() {
        this.expungeStaleEntries();
        return new Itr();
    }
    
    
    private class Itr implements Iterator<E> {

        /**
         * Index of element to be returned by subsequent call to next.
         */
        int cursor = 0;

        /**
         * Index of element returned by most recent call to next or previous.
         * Reset to -1 if this element is deleted by a call to remove.
         */
        int lastRet = -1;

        /**
         * The modCount value that the iterator believes that the backing List
         * should have. If this expectation is violated, the iterator has
         * detected concurrent modification.
         */
        int expectedModCount = WeakCollection.this.modCount;

        
        Itr() { /* nothing */ }
        

        @Override
        public boolean hasNext() {
            return this.cursor != size();
        }


        @Override
        public E next() {
            this.checkForComodification();
            try {
                final E next = get(this.cursor);
                this.lastRet = this.cursor++;
                return next;
            } catch (final IndexOutOfBoundsException e) {
                this.checkForComodification();
                throw new NoSuchElementException();
            }
        }


        @Override
        public void remove() {
            if (this.lastRet == -1)
                throw new IllegalStateException();
            this.checkForComodification();

            try {
                WeakCollection.this.fastRemove(this.lastRet);
                if (this.lastRet < this.cursor)
                    this.cursor--;
                this.lastRet = -1;
                this.expectedModCount = WeakCollection.this.modCount;
            } catch (final IndexOutOfBoundsException e) {
                throw new ConcurrentModificationException();
            }
        }


        final void checkForComodification() {
            if (WeakCollection.this.modCount != this.expectedModCount) {
                throw new ConcurrentModificationException();
            }
        }
    }

}
