/*
 * Project:  Petruchio
 * Filename: WeakIdentityHashMap.java
 * Created:  11.12.2006
 * Author:   Tim Strazny
 * Remarks:  This class is based on class java.util.WeakHashMap.
 *           It has beend slightly modified to support use identity mapping,
 *           but basically most of the code comes from Sun's WeakHashMap class.
 */
package petruchio.common;


import java.lang.ref.ReferenceQueue;
import java.lang.ref.SoftReference;
import java.lang.ref.WeakReference;
import java.util.AbstractCollection;
import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;


/**
 * A hashtable-based <tt>Map</tt> implementation with <em>weak keys</em> and <em>soft values</em>.
 * An entry in a <tt>WeakIdentityHashMap</tt> will automatically be removed when
 * its key is no longer in ordinary use.  More precisely, the presence of a
 * mapping for a given key will not prevent the key from being discarded by the
 * garbage collector, that is, made finalizable, finalized, and then reclaimed.
 * When a key has been discarded its entry is effectively removed from the map,
 * so this class behaves somewhat differently from other <tt>Map</tt>
 * implementations.
 *
 * <p> Both null values and the null key are supported. This class has
 * performance characteristics similar to those of the <tt>HashMap</tt>
 * class, and has the same efficiency parameters of <em>initial capacity</em>
 * and <em>load factor</em>.
 *
 * <p> Like most collection classes, this class is not synchronized.
 * A synchronized <tt>WeakIdentityHashMap</tt> may be constructed using the
 * {@link Collections#synchronizedMap Collections.synchronizedMap}
 * method.
 *
 * <p> This class is intended primarily for use with key objects whose
 * <tt>equals</tt> methods test for object identity using the
 * <tt>==</tt> operator.  Once such a key is discarded it can never be
 * recreated, so it is impossible to do a lookup of that key in a
 * <tt>WeakIdentityHashMap</tt> at some later time and be surprised that its entry
 * has been removed.  This class will work perfectly well with key objects
 * whose <tt>equals</tt> methods are not based upon object identity, such
 * as <tt>String</tt> instances.  With such recreatable key objects,
 * however, the automatic removal of <tt>WeakIdentityHashMap</tt> entries whose
 * keys have been discarded may prove to be confusing.
 *
 * <p> The behavior of the <tt>WeakIdentityHashMap</tt> class depends in part upon
 * the actions of the garbage collector, so several familiar (though not
 * required) <tt>Map</tt> invariants do not hold for this class.  Because
 * the garbage collector may discard keys at any time, a
 * <tt>WeakIdentityHashMap</tt> may behave as though an unknown thread is silently
 * removing entries.  In particular, even if you synchronize on a
 * <tt>WeakIdentityHashMap</tt> instance and invoke none of its mutator methods, it
 * is possible for the <tt>size</tt> method to return smaller values over
 * time, for the <tt>isEmpty</tt> method to return <tt>false</tt> and
 * then <tt>true</tt>, for the <tt>containsKey</tt> method to return
 * <tt>true</tt> and later <tt>false</tt> for a given key, for the
 * <tt>get</tt> method to return a value for a given key but later return
 * <tt>null</tt>, for the <tt>put</tt> method to return
 * <tt>null</tt> and the <tt>remove</tt> method to return
 * <tt>false</tt> for a key that previously appeared to be in the map, and
 * for successive examinations of the key set, the value collection, and
 * the entry set to yield successively smaller numbers of elements.
 *
 * <p> Each key object in a <tt>WeakIdentityHashMap</tt> is stored indirectly as
 * the referent of a weak reference.  Therefore a key will automatically be
 * removed only after the weak references to it, both inside and outside of the
 * map, have been cleared by the garbage collector.
 *
 * <p> <strong>Implementation note:</strong> The value objects in a
 * <tt>WeakIdentityHashMap</tt> are held by ordinary strong references.  Thus care
 * should be taken to ensure that value objects do not strongly refer to their
 * own keys, either directly or indirectly, since that will prevent the keys
 * from being discarded.  Note that a value object may refer indirectly to its
 * key via the <tt>WeakIdentityHashMap</tt> itself; that is, a value object may
 * strongly refer to some other key object whose associated value object, in
 * turn, strongly refers to the key of the first value object.  One way
 * to deal with this is to wrap values themselves within
 * <tt>WeakReferences</tt> before
 * inserting, as in: <tt>m.put(key, new WeakReference(value))</tt>,
 * and then unwrapping upon each <tt>get</tt>.
 *
 * <p>The iterators returned by the <tt>iterator</tt> method of the collections
 * returned by all of this class's "collection view methods" are
 * <i>fail-fast</i>: if the map is structurally modified at any time after the
 * iterator is created, in any way except through the iterator's own
 * <tt>remove</tt> method, the iterator will throw a {@link
 * ConcurrentModificationException}.  Thus, in the face of concurrent
 * modification, the iterator fails quickly and cleanly, rather than risking
 * arbitrary, non-deterministic behavior at an undetermined time in the future.
 *
 * <p>Note that the fail-fast behavior of an iterator cannot be guaranteed
 * as it is, generally speaking, impossible to make any hard guarantees in the
 * presence of unsynchronized concurrent modification.  Fail-fast iterators
 * throw <tt>ConcurrentModificationException</tt> on a best-effort basis.
 * Therefore, it would be wrong to write a program that depended on this
 * exception for its correctness:  <i>the fail-fast behavior of iterators
 * should be used only to detect bugs.</i>
 *
 * <p>This class is a member of the
 * <a href="{@docRoot}/../technotes/guides/collections/index.html">
 * Java Collections Framework</a>.
 *
 * @param <K> the type of keys maintained by this map
 * @param <V> the type of mapped values
 *
 * @version 1.38, 04/24/06
 * @author      Doug Lea
 * @author      Josh Bloch
 * @author  Mark Reinhold
 * @since   1.2
 * @see     java.util.HashMap
 * @see     java.lang.ref.WeakReference
 */
public class WeakIdentityHashMap<K,V> extends AbstractMap<K,V> {

    /**
     * The default initial capacity -- MUST be a power of two.
     */
    private static final int DEFAULT_INITIAL_CAPACITY = 16;

    /**
     * The maximum capacity, used if a higher value is implicitly specified
     * by either of the constructors with arguments.
     * MUST be a power of two <= 1<<30.
     */
    private static final int MAXIMUM_CAPACITY = 1 << 30;

    /**
     * The load fast used when none specified in constructor.
     */
    private static final float DEFAULT_LOAD_FACTOR = 0.75f;

    /**
     * The table, resized as necessary. Length MUST Always be a power of two.
     */
    Entry<K,V>[] table;

    /**
     * The number of key-value mappings contained in this weak hash map.
     */
    private int size;

    /**
     * The next size value at which to resize (capacity * load factor).
     */
    private int threshold;

    /**
     * The load factor for the hash table.
     */
    private final float loadFactor;

    /**
     * Reference queue for cleared WeakEntries
     */
    private final ReferenceQueue<K> queue = new ReferenceQueue<K>();
    
    /**
     * Reference queue for cleared SoftEntries
     */
    private final ReferenceQueue<V> vqueue = new ReferenceQueue<V>();
    
    transient volatile Set<K>        keySet = null;
    transient volatile Collection<V> values = null;

    /**
     * The number of times this WeakIdentityHashMap has been structurally modified.
     * Structural modifications are those that change the number of
     * mappings in the map or otherwise modify its internal structure
     * (e.g., rehash).  This field is used to make iterators on
     * Collection-views of the map fail-fast.
     *
     * @see ConcurrentModificationException
     */
    volatile int modCount;

    /**
     * Constructs a new, empty <tt>WeakIdentityHashMap</tt> with the given initial
     * capacity and the given load factor.
     *
     * @param  initialCapacity The initial capacity of the <tt>WeakIdentityHashMap</tt>
     * @param  loadFactor      The load factor of the <tt>WeakIdentityHashMap</tt>
     * @throws IllegalArgumentException if the initial capacity is negative,
     *         or if the load factor is nonpositive.
     */
    @SuppressWarnings("unchecked")
    public WeakIdentityHashMap(final int initialCapacity, final float loadFactor) {
        if (initialCapacity < 0) {
            throw new IllegalArgumentException("Illegal Initial Capacity: "+
                                               initialCapacity);
        }
        final int cap;
        if (initialCapacity > MAXIMUM_CAPACITY) {
            cap = MAXIMUM_CAPACITY;
        } else {
            cap = initialCapacity;
        }
        

        if (loadFactor <= 0 || Float.isNaN(loadFactor)) {
            throw new IllegalArgumentException("Illegal Load factor: "+
                                               loadFactor);
        }
        int capacity = 1;
        while (capacity < cap) {
            capacity <<= 1;
        }
        this.table = new Entry[capacity];
        this.loadFactor = loadFactor;
        this.threshold = (int)(capacity * loadFactor);
    }

    /**
     * Constructs a new, empty <tt>WeakIdentityHashMap</tt> with the given initial
     * capacity and the default load factor (0.75).
     *
     * @param  initialCapacity The initial capacity of the <tt>WeakIdentityHashMap</tt>
     * @throws IllegalArgumentException if the initial capacity is negative
     */
    public WeakIdentityHashMap(final int initialCapacity) {
        this(initialCapacity, DEFAULT_LOAD_FACTOR);
    }

    /**
     * Constructs a new, empty <tt>WeakIdentityHashMap</tt> with the default initial
     * capacity (16) and load factor (0.75).
     */
    @SuppressWarnings("unchecked")
    public WeakIdentityHashMap() {
        this.loadFactor = DEFAULT_LOAD_FACTOR;
        this.threshold = DEFAULT_INITIAL_CAPACITY;
        this.table = new Entry[DEFAULT_INITIAL_CAPACITY];
    }

    /**
     * Constructs a new <tt>WeakIdentityHashMap</tt> with the same mappings as the
     * specified map.  The <tt>WeakIdentityHashMap</tt> is created with the default
     * load factor (0.75) and an initial capacity sufficient to hold the
     * mappings in the specified map.
     *
     * @param   m the map whose mappings are to be placed in this map
     * @throws  NullPointerException if the specified map is null
     * @since   1.3
     */
    public WeakIdentityHashMap(final Map<? extends K, ? extends V> m) {
        this(Math.max((int) (m.size() / DEFAULT_LOAD_FACTOR) + 1, 16),
             DEFAULT_LOAD_FACTOR);
        this.putAll(m);
    }

    // internal utilities

    /**
     * Value representing null keys inside tables.
     */
    private static final Object NULL_KEY = new Object();

    
    /**
     * Returns index for Object x.
     */
    private static int hash(final Object x, final int length) {
        final int h = System.identityHashCode(x);
        // Multiply by -127, and left-shift to use least bit as part of hash
        return ((h << 1) - (h << 8)) & (length - 1);
    }
    
    /**
     * Use NULL_KEY for key if it is null.
     */
    private static Object maskNull(final Object key) {
        return (key == null ? NULL_KEY : key);
    }

    /**
     * Returns internal representation of null key back to caller as null.
     */
    @SuppressWarnings("unchecked")
    static <K> K unmaskNull(final Object key) {
        return (K) (key == NULL_KEY ? null : key);
    }

    /**
     * Checks identity
     */
    static boolean eq(final Object x, final Object y) {
        return x == y; // || x.equals(y);
    }

    /**
     * Returns index for hash code h.
     */
    static int indexFor(final int h, final int length) {
        return h & (length-1);
    }
    
    /**
     * Returns a shallow copy of this <tt>AbstractMap</tt> instance: the keys
     * and values themselves are not cloned.
     *
     * @return a shallow copy of this map
     */
    @Override
    @SuppressWarnings("unchecked")
    protected Object clone() {
        try {
            WeakIdentityHashMap<K,V> result = (WeakIdentityHashMap<K,V>)super.clone();
            result.keySet = null;
            result.values = null;
            return result;
        } catch (CloneNotSupportedException e) {
            throw new InternalError();
        }
    }


    /**
     * Expunges stale entries from the table.
     */
    @SuppressWarnings({ "unchecked", "null" })
    private void expungeStaleEntries() {
        Entry<K,V> e;
        SoftValue<K, V> v;
        while ( (e = (Entry<K,V>) this.queue.poll()) != null
                || ((v = (SoftValue<K,V>) this.vqueue.poll()) != null && (e = v.e) != null)) {
            int h = e.hash;
            int i = indexFor(h, this.table.length);

            Entry<K,V> prev = this.table[i];
            Entry<K,V> p = prev;
            while (p != null) {
                final Entry<K,V> next = p.next;
                if (p == e) {
                    if (prev == e) {
                        this.table[i] = next;
                    } else {
                        prev.next = next;
                    }
                    e.next = null;  // Help GC
                    e.value = null; //  "   "
                    this.size--;
                    break;
                }
                prev = p;
                p = next;
            }
        }
    }

    /**
     * Returns the table after first expunging stale entries.
     */
    private Entry<K,V>[] getTable() {
        this.expungeStaleEntries();
        return this.table;
    }

    /**
     * Returns the number of key-value mappings in this map.
     * This result is a snapshot, and may not reflect unprocessed
     * entries that will be removed before next attempted access
     * because they are no longer referenced.
     */
    @Override
    public int size() {
        if (this.size == 0)
            return 0;
        this.expungeStaleEntries();
        return this.size;
    }

    /**
     * Returns <tt>true</tt> if this map contains no key-value mappings.
     * This result is a snapshot, and may not reflect unprocessed
     * entries that will be removed before next attempted access
     * because they are no longer referenced.
     */
    @Override
    public boolean isEmpty() {
        return this.size() == 0;
    }

    /**
     * Returns the value to which the specified key is mapped,
     * or {@code null} if this map contains no mapping for the key.
     *
     * <p>More formally, if this map contains a mapping from a key
     * {@code k} to a value {@code v} such that {@code (key==null ? k==null :
     * key.equals(k))}, then this method returns {@code v}; otherwise
     * it returns {@code null}.  (There can be at most one such mapping.)
     *
     * <p>A return value of {@code null} does not <i>necessarily</i>
     * indicate that the map contains no mapping for the key; it's also
     * possible that the map explicitly maps the key to {@code null}.
     * The {@link #containsKey containsKey} operation may be used to
     * distinguish these two cases.
     *
     * @see #put(Object, Object)
     */
    @Override
    public V get(final Object key) {
        final Object k = WeakIdentityHashMap.maskNull(key);
        final Entry<K,V>[] tab = this.getTable();
        final int h = WeakIdentityHashMap.hash(k, tab.length);
        final int index = WeakIdentityHashMap.indexFor(h, tab.length);
        Entry<K,V> e = tab[index];
        while (e != null) {
            if (e.hash == h && eq(k, e.get())) {
                return e.value.get();
            }
            e = e.next;
        }
        return null;
    }

    /**
     * Returns <tt>true</tt> if this map contains a mapping for the
     * specified key.
     *
     * @param  key   The key whose presence in this map is to be tested
     * @return <tt>true</tt> if there is a mapping for <tt>key</tt>;
     *         <tt>false</tt> otherwise
     */
    @Override
    public boolean containsKey(Object key) {
        return this.getEntry(key) != null;
    }

    /**
     * Returns the entry associated with the specified key in this map.
     * Returns null if the map contains no mapping for this key.
     */
    Entry<K,V> getEntry(final Object key) {
        final Object k = WeakIdentityHashMap.maskNull(key);
        final Entry<K, V>[] tab = this.getTable();
        final int h = WeakIdentityHashMap.hash(k, tab.length);
        final int index = WeakIdentityHashMap.indexFor(h, tab.length);
        Entry<K,V> e = tab[index];
        while (e != null && !(e.hash == h && eq(k, e.get()))) {
            e = e.next;
        }
        return e;
    }

    /**
     * Associates the specified value with the specified key in this map.
     * If the map previously contained a mapping for this key, the old
     * value is replaced.
     *
     * @param key key with which the specified value is to be associated.
     * @param value value to be associated with the specified key.
     * @return the previous value associated with <tt>key</tt>, or
     *         <tt>null</tt> if there was no mapping for <tt>key</tt>.
     *         (A <tt>null</tt> return can also indicate that the map
     *         previously associated <tt>null</tt> with <tt>key</tt>.)
     */
    @Override
    @SuppressWarnings("unchecked")
    public V put(final K key, final V value) {
        final K k = (K) WeakIdentityHashMap.maskNull(key);
        final Entry<K,V>[] tab = this.getTable();
        final int h = WeakIdentityHashMap.hash(k, tab.length);
        final int i = WeakIdentityHashMap.indexFor(h, tab.length);

        for (Entry<K,V> e = tab[i]; e != null; e = e.next) {
            if (h == e.hash && eq(k, e.get())) {
                final V oldValue = e.value.get();
                if (value != oldValue) {
                    e.value = new SoftValue<K, V>(value, this.vqueue, e);
                }
                return oldValue;
            }
        }

        this.modCount++;
        final Entry<K,V> e = tab[i];
        tab[i] = new Entry<K,V>(k, value, this.queue, this.vqueue, h, e);
        if (++this.size >= this.threshold) {
            resize(tab.length * 2);
        }
        return null;
    }

    /**
     * Rehashes the contents of this map into a new array with a
     * larger capacity.  This method is called automatically when the
     * number of keys in this map reaches its threshold.
     *
     * If current capacity is MAXIMUM_CAPACITY, this method does not
     * resize the map, but sets threshold to Integer.MAX_VALUE.
     * This has the effect of preventing future calls.
     *
     * @param newCapacity the new capacity, MUST be a power of two;
     *        must be greater than current capacity unless current
     *        capacity is MAXIMUM_CAPACITY (in which case value
     *        is irrelevant).
     */
    @SuppressWarnings("unchecked")
    void resize(final int newCapacity) {
        final Entry<K, V>[] oldTable = getTable();
        final int oldCapacity = oldTable.length;
        if (oldCapacity == WeakIdentityHashMap.MAXIMUM_CAPACITY) {
            this.threshold = Integer.MAX_VALUE;
            return;
        }

        Entry<K, V>[] newTable = new Entry[newCapacity];
        this.transfer(oldTable, newTable);
        this.table = newTable;

        /*
         * If ignoring null elements and processing ref queue caused massive
         * shrinkage, then restore old table.  This should be rare, but avoids
         * unbounded expansion of garbage-filled tables.
         */
        if (this.size >= this.threshold / 2) {
            this.threshold = (int)(newCapacity * this.loadFactor);
        } else {
            this.expungeStaleEntries();
            this.transfer(newTable, oldTable);
            this.table = oldTable;
        }
    }

    /** Transfers all entries from src to dest tables */
    private void transfer(final Entry<K, V>[] src, final Entry<K, V>[] dest) {
        for (int j = 0; j < src.length; ++j) {
            Entry<K,V> e = src[j];
            src[j] = null;
            while (e != null) {
                final Entry<K,V> next = e.next;
                final Object key = e.get();
                if (key == null) {
                    e.next = null;  // Help GC
                    e.value = null; //  "   "
                    this.size--;
                } else {
                    final int i = WeakIdentityHashMap.indexFor(e.hash, dest.length);
                    e.next = dest[i];
                    dest[i] = e;
                }
                e = next;
            }
        }
    }

    /**
     * Copies all of the mappings from the specified map to this map.
     * These mappings will replace any mappings that this map had for any
     * of the keys currently in the specified map.
     *
     * @param m mappings to be stored in this map.
     * @throws  NullPointerException if the specified map is null.
     */
    @Override
    public void putAll(final Map<? extends K, ? extends V> m) {
        final int numKeysToBeAdded = m.size();
        if (numKeysToBeAdded == 0) {
            return;
        }

        /*
         * Expand the map if the map if the number of mappings to be added
         * is greater than or equal to threshold.  This is conservative; the
         * obvious condition is (m.size() + size) >= threshold, but this
         * condition could result in a map with twice the appropriate capacity,
         * if the keys to be added overlap with the keys already in this map.
         * By using the conservative calculation, we subject ourself
         * to at most one extra resize.
         */
        if (numKeysToBeAdded > this.threshold) {
            int targetCapacity = (int)(numKeysToBeAdded / this.loadFactor + 1);
            if (targetCapacity > WeakIdentityHashMap.MAXIMUM_CAPACITY) {
                targetCapacity = WeakIdentityHashMap.MAXIMUM_CAPACITY;
            }
            int newCapacity = this.table.length;
            while (newCapacity < targetCapacity) {
                newCapacity <<= 1;
            }
            if (newCapacity > this.table.length) {
                this.resize(newCapacity);
            }
        }

        for (final Map.Entry<? extends K, ? extends V> e : m.entrySet()) {
            this.put(e.getKey(), e.getValue());
        }
    }

    /**
     * Removes the mapping for a key from this weak hash map if it is present.
     * More formally, if this map contains a mapping from key <tt>k</tt> to
     * value <tt>v</tt> such that <code>(key==null ?  k==null :
     * key.equals(k))</code>, that mapping is removed.  (The map can contain
     * at most one such mapping.)
     *
     * <p>Returns the value to which this map previously associated the key,
     * or <tt>null</tt> if the map contained no mapping for the key.  A
     * return value of <tt>null</tt> does not <i>necessarily</i> indicate
     * that the map contained no mapping for the key; it's also possible
     * that the map explicitly mapped the key to <tt>null</tt>.
     *
     * <p>The map will not contain a mapping for the specified key once the
     * call returns.
     *
     * @param key key whose mapping is to be removed from the map
     * @return the previous value associated with <tt>key</tt>, or
     *         <tt>null</tt> if there was no mapping for <tt>key</tt>
     */
    @Override
    public V remove(final Object key) {
        final Object k = WeakIdentityHashMap.maskNull(key);
        final Entry<K, V>[] tab = this.getTable();
        final int h = WeakIdentityHashMap.hash(k, tab.length);
        final int i = WeakIdentityHashMap.indexFor(h, tab.length);
        Entry<K,V> prev = tab[i];
        Entry<K,V> e = prev;

        while (e != null) {
            final Entry<K,V> next = e.next;
            if (h == e.hash && eq(k, e.get())) {
                this.modCount++;
                this.size--;
                if (prev == e) {
                    tab[i] = next;
                } else {
                    prev.next = next;
                }
                return e.value.get();
            }
            prev = e;
            e = next;
        }

        return null;
    }



    /** Special version of remove needed by Entry set */
    @SuppressWarnings("unchecked")
    Entry<K,V> removeMapping(final Object o) {
        if (!(o instanceof Map.Entry<?,?>)) {
            return null;
        }
        final Entry<K, V>[] tab = getTable();
        final Map.Entry<K, V> entry = (Map.Entry<K,V>)o;
        final Object k = WeakIdentityHashMap.maskNull(entry.getKey());
        int h = WeakIdentityHashMap.hash(k, tab.length);
        int i = WeakIdentityHashMap.indexFor(h, tab.length);
        Entry<K,V> prev = tab[i];
        Entry<K,V> e = prev;

        while (e != null) {
            Entry<K,V> next = e.next;
            if (h == e.hash && e.equals(entry)) {
                this.modCount++;
                this.size--;
                if (prev == e) {
                    tab[i] = next;
                } else {
                    prev.next = next;
                }
                return e;
            }
            prev = e;
            e = next;
        }

        return null;
    }

    /**
     * Removes all of the mappings from this map.
     * The map will be empty after this call returns.
     */
    @Override
    public void clear() {
        // clear out ref queue. We don't need to expunge entries
        // since table is getting cleared.
        while (this.queue.poll() != null || this.vqueue.poll() != null) // empty the queue
            /* nothing */;

        this.modCount++;
        final Entry<K, V>[] tab = this.table;
        for (int i = 0; i < tab.length; ++i) {
            tab[i] = null;
        }
        this.size = 0;

        // Allocation of array may have caused GC, which may have caused
        // additional entries to go stale.  Removing these entries from the
        // reference queue will make them eligible for reclamation.
        while (this.queue.poll() != null || this.vqueue.poll() != null) // empty the queue
            /* nothing */;
    }

    /**
     * Returns <tt>true</tt> if this map maps one or more keys to the
     * specified value.
     *
     * @param value value whose presence in this map is to be tested
     * @return <tt>true</tt> if this map maps one or more keys to the
     *         specified value
     */
    @Override
    public boolean containsValue(final Object value) {
        if (value==null) {
                return this.containsNullValue();
        }
    
        final Entry<K, V>[] tab = this.getTable();
        for (int i = tab.length ; i-- > 0 ;) {
            for (Entry<K, V> e = tab[i] ; e != null ; e = e.next) {
                if (WeakIdentityHashMap.eq(value, e.value)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Special-case code for containsValue with null argument
     */
    private boolean containsNullValue() {
        final Entry<K, V>[] tab = this.getTable();
        for (int i = tab.length ; i-- > 0 ;) {
            for (Entry<K, V> e = tab[i] ; e != null ; e = e.next) {
                if (e.value==null) {
                    return true;
                }
            }
        }
        return false;
    }
    
    private static class SoftValue<K, V> extends SoftReference<V> {
        final Entry<K, V> e;
        
        /**
         * @param referent
         * @param q
         */
        public SoftValue(final V referent, final ReferenceQueue<? super V> q,
                final Entry<K, V> e) {
            super(referent, q);
            this.e = e;
        }
        
//        Entry<K, V> getEntry() {
//            return this.e;
//        }
        
    }

    /**
     * The entries in this hash table extend WeakReference, using its main ref
     * field as the key.
     */
    private static class Entry<K,V> extends WeakReference<K> implements Map.Entry<K,V> {
        SoftValue<K, V> value;
        final int hash;
        Entry<K,V> next;
        ReferenceQueue<V> vqueue;

        /**
         * Creates new entry.
         */
        Entry(final K key, final V value, final ReferenceQueue<K> queue,
                final ReferenceQueue<V> vqueue, final int hash, final Entry<K,V> next) {
            super(key, queue);
            this.vqueue = vqueue;
            this.value = new SoftValue<K, V>(value, vqueue, this);
            this.hash  = hash;
            this.next  = next;
        }

        @Override
        public K getKey() {
            return WeakIdentityHashMap.<K>unmaskNull(get());
        }

        @Override
        public V getValue() {
            return this.value.get();
        }

        @Override
        public V setValue(final V newValue) {
            V oldValue = this.value.get();
            this.value = new SoftValue<K, V>(newValue, this.vqueue, this);
            return oldValue;
        }

        @Override
        public boolean equals(final Object o) {
            if (!(o instanceof Map.Entry<?, ?>)) {
                return false;
            }
            final Map.Entry<?, ?> e = (Map.Entry<?, ?>)o;
            final Object k1 = getKey();
            final Object k2 = e.getKey();
            if (k1 == k2 || (k1 != null && k1.equals(k2))) {
                final Object v1 = getValue();
                final Object v2 = e.getValue();
                if ((v1 == v2) || (v1 != null && WeakIdentityHashMap.eq(v1, v2))) {
                    return true;
                }
            }
            return false;
        }

        @Override
        public int hashCode() {
            final Object k = getKey();
            final Object v = getValue();
            return  ((k==null ? 0 : k.hashCode()) ^
                     (v==null ? 0 : v.hashCode()));
        }

        @Override
        public String toString() {
            return getKey() + "=" + getValue();
        }
    }

    private abstract class HashIterator<T> implements Iterator<T> {
        int index;
        Entry<K,V> entry = null;
        Entry<K,V> lastReturned = null;
        int expectedModCount = WeakIdentityHashMap.this.modCount;

        /**
         * Strong reference needed to avoid disappearance of key
         * between hasNext and next
         */
        Object nextKey = null;

        /**
         * Strong reference needed to avoid disappearance of key
         * between nextEntry() and any use of the entry
         */
        Object currentKey = null;

        HashIterator() {
            this.index = (size() != 0 ? WeakIdentityHashMap.this.table.length : 0);
        }

        @Override
        public boolean hasNext() {
            final Entry<K, V>[] t = WeakIdentityHashMap.this.table;

            while (this.nextKey == null) {
                Entry<K,V> e = this.entry;
                int i = this.index;
                while ((e == null) && (i > 0)) {
                    e = t[--i];
                }
                this.entry = e;
                this.index = i;
                if (e == null) {
                    this.currentKey = null;
                    return false;
                }
                this.nextKey = e.get(); // hold on to key in strong ref
                if (this.nextKey == null) {
                    this.entry = this.entry.next;
                }
            }
            return true;
        }

        /** The common parts of next() across different types of iterators */
        protected Entry<K,V> nextEntry() {
            if (WeakIdentityHashMap.this.modCount != this.expectedModCount) {
                throw new ConcurrentModificationException();
            }
            if ((this.nextKey == null) && !this.hasNext()) {
                throw new NoSuchElementException();
            }

            this.lastReturned = this.entry;
            this.entry = this.entry.next;
            this.currentKey = this.nextKey;
            this.nextKey = null;
            return this.lastReturned;
        }

        @Override
        public void remove() {
            if (this.lastReturned == null) {
                throw new IllegalStateException();
            }
            if (WeakIdentityHashMap.this.modCount != this.expectedModCount) {
                throw new ConcurrentModificationException();
            }

            WeakIdentityHashMap.this.remove(this.currentKey);
            this.expectedModCount = WeakIdentityHashMap.this.modCount;
            this.lastReturned = null;
            this.currentKey = null;
        }

    }

    class ValueIterator extends HashIterator<V> {
        @Override
        public V next() {
            return nextEntry().value.get();
        }
    }

    class KeyIterator extends HashIterator<K> {
        @Override
        public K next() {
            return nextEntry().getKey();
        }
    }

    class EntryIterator extends HashIterator<Map.Entry<K,V>> {
        @Override
        public Map.Entry<K,V> next() {
            return nextEntry();
        }
    }

    // Views

    private transient Set<Map.Entry<K,V>> entrySet = null;

    /**
     * Returns a {@link Set} view of the keys contained in this map.
     * The set is backed by the map, so changes to the map are
     * reflected in the set, and vice-versa.  If the map is modified
     * while an iteration over the set is in progress (except through
     * the iterator's own <tt>remove</tt> operation), the results of
     * the iteration are undefined.  The set supports element removal,
     * which removes the corresponding mapping from the map, via the
     * <tt>Iterator.remove</tt>, <tt>Set.remove</tt>,
     * <tt>removeAll</tt>, <tt>retainAll</tt>, and <tt>clear</tt>
     * operations.  It does not support the <tt>add</tt> or <tt>addAll</tt>
     * operations.
     */
    @Override
    public Set<K> keySet() {
        Set<K> ks = this.keySet;
        return (ks != null ? ks : (this.keySet = new KeySet()));
    }

    class KeySet extends AbstractSet<K> {
        @Override
        public Iterator<K> iterator() {
            return new KeyIterator();
        }

        @Override
        public int size() {
            return WeakIdentityHashMap.this.size();
        }

        @Override
        public boolean contains(Object o) {
            return containsKey(o);
        }

        @Override
        public boolean remove(Object o) {
            if (containsKey(o)) {
                WeakIdentityHashMap.this.remove(o);
                return true;
            }
            return false;
        }

        @Override
        public void clear() {
            WeakIdentityHashMap.this.clear();
        }

        @Override
        public Object[] toArray() {
            final Collection<K> c = new ArrayList<K>(size());
            for (Iterator<K> i = iterator(); i.hasNext(); ) {
                c.add(i.next());
            }
            return c.toArray();
        }

        @Override
        public <T> T[] toArray(final T[] a) {
            final Collection<K> c = new ArrayList<K>(size());
            for (Iterator<K> i = iterator(); i.hasNext(); ) {
                c.add(i.next());
            }
            return c.toArray(a);
        }
    }

    /**
     * Returns a {@link Collection} view of the values contained in this map.
     * The collection is backed by the map, so changes to the map are
     * reflected in the collection, and vice-versa.  If the map is
     * modified while an iteration over the collection is in progress
     * (except through the iterator's own <tt>remove</tt> operation),
     * the results of the iteration are undefined.  The collection
     * supports element removal, which removes the corresponding
     * mapping from the map, via the <tt>Iterator.remove</tt>,
     * <tt>Collection.remove</tt>, <tt>removeAll</tt>,
     * <tt>retainAll</tt> and <tt>clear</tt> operations.  It does not
     * support the <tt>add</tt> or <tt>addAll</tt> operations.
     */
    @Override
    public Collection<V> values() {
        final Collection<V> vs = this.values;
        return (vs != null ?  vs : (this.values = new Values()));
    }

    class Values extends AbstractCollection<V> {
        @Override
        public Iterator<V> iterator() {
            return new ValueIterator();
        }

        @Override
        public int size() {
            return WeakIdentityHashMap.this.size();
        }

        @Override
        public boolean contains(final Object o) {
            return containsValue(o);
        }

        @Override
        public void clear() {
            WeakIdentityHashMap.this.clear();
        }

        @Override
        public Object[] toArray() {
            final Collection<V> c = new ArrayList<V>(size());
            for (Iterator<V> i = iterator(); i.hasNext(); ) {
                c.add(i.next());
            }
            return c.toArray();
        }

        @Override
        public <T> T[] toArray(T[] a) {
            final Collection<V> c = new ArrayList<V>(size());
            for (Iterator<V> i = iterator(); i.hasNext(); ) {
                c.add(i.next());
            }
            return c.toArray(a);
        }
    }

    /**
     * Returns a {@link Set} view of the mappings contained in this map.
     * The set is backed by the map, so changes to the map are
     * reflected in the set, and vice-versa.  If the map is modified
     * while an iteration over the set is in progress (except through
     * the iterator's own <tt>remove</tt> operation, or through the
     * <tt>setValue</tt> operation on a map entry returned by the
     * iterator) the results of the iteration are undefined.  The set
     * supports element removal, which removes the corresponding
     * mapping from the map, via the <tt>Iterator.remove</tt>,
     * <tt>Set.remove</tt>, <tt>removeAll</tt>, <tt>retainAll</tt> and
     * <tt>clear</tt> operations.  It does not support the
     * <tt>add</tt> or <tt>addAll</tt> operations.
     */
    @Override
    public Set<Map.Entry<K,V>> entrySet() {
        final Set<Map.Entry<K,V>> es = WeakIdentityHashMap.this.entrySet;
        return es != null ? es : (WeakIdentityHashMap.this.entrySet = new EntrySet());
    }

    class EntrySet extends AbstractSet<Map.Entry<K,V>> {
        @Override
        public Iterator<Map.Entry<K,V>> iterator() {
            return new EntryIterator();
        }

        @Override
        public boolean contains(final Object o) {
            if (!(o instanceof Map.Entry<?, ?>)) {
                return false;
            }
            final Map.Entry<?, ?> e = (Map.Entry<?, ?>)o;
            final Entry<K, V> candidate = getEntry(e.getKey());
            return candidate != null && candidate.equals(e);
        }

        @Override
        public boolean remove(final Object o) {
            return removeMapping(o) != null;
        }
        
        @Override
        public int size() {
            return WeakIdentityHashMap.this.size();
        }

        @Override
        public void clear() {
            WeakIdentityHashMap.this.clear();
        }

        @Override
        public Object[] toArray() {
            final Collection<Map.Entry<K,V>> c = new ArrayList<Map.Entry<K,V>>(size());
            for (Iterator<Map.Entry<K,V>> i = iterator(); i.hasNext(); ) {
                c.add(new SimpleEntry<K,V>(i.next()));
            }
            return c.toArray();
        }

        @Override
        public <T> T[] toArray(final T[] a) {
            final Collection<Map.Entry<K,V>> c = new ArrayList<Map.Entry<K,V>>(size());
            for (Iterator<Map.Entry<K,V>> i = iterator(); i.hasNext(); ) {
                c.add(new SimpleEntry<K,V>(i.next()));
            }
            return c.toArray(a);
        }
    }
    
    /**
     * This should be made public as soon as possible.  It greatly simplifies
     * the task of implementing Map.
     */
    static class SimpleEntry<K,V> implements Map.Entry<K,V> {
        K key;
        V value;
    
        public SimpleEntry(K key, V value) {
            this.key   = key;
            this.value = value;
        }
    
        public SimpleEntry(final Map.Entry<K,V> e) {
            this.key   = e.getKey();
            this.value = e.getValue();
        }
    
        @Override
        public K getKey() {
            return this.key;
        }
    
        @Override
        public V getValue() {
            return this.value;
        }
    
        @Override
        public V setValue(final V value) {
            final V oldValue = this.value;
            this.value = value;
            return oldValue;
        }
    
        @Override
        public boolean equals(Object o) {
            if (!(o instanceof Map.Entry<?, ?>)) {
                return false;
            }
            final Map.Entry<?, ?> e = (Map.Entry<?, ?>)o;
            return SimpleEntry.eq(this.key, e.getKey())
                && SimpleEntry.eq(this.value, e.getValue());
        }
    
        @Override
        public int hashCode() {
            return ((this.key   == null)   ? 0 :   this.key.hashCode()) ^
                   ((this.value == null)   ? 0 : this.value.hashCode());
        }
    
        @Override
        public String toString() {
            return this.key + "=" + this.value;
        }

        private static boolean eq(final Object o1, final Object o2) {
            return (o1 == null ? o2 == null : o1.equals(o2));
        }
    }
}