/*
 * Project:  Petruchio
 * Filename: WeakWeakIdentityHashSet.java
 * Created:  02.03.2007
 * Author:   Tim Strazny
 */
package petruchio.common;

import java.util.AbstractSet;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.Iterator;


/**
 * <Description>
 * 
 * @param <E>
 * @author Tim Strazny
 */
public class WeakIdentityHashSet<E> extends AbstractSet<E> implements Cloneable, java.io.Serializable {

    /**
     * 
     */
    private static final long serialVersionUID = -8237892338136333564L;

    private transient WeakIdentityHashMap<E, Object> map;

    // Dummy value to associate with an Object in the backing Map
    private static final Object PRESENT = new Object();


    /**
     * Constructs a new, empty set; the backing <tt>HashMap</tt> instance has
     * default initial capacity (16) and load factor (0.75).
     */
    public WeakIdentityHashSet() {
        this.map = new WeakIdentityHashMap<E, Object>();
    }


    /**
     * Constructs a new set containing the elements in the specified collection.
     * The <tt>HashMap</tt> is created with default load factor (0.75) and an
     * initial capacity sufficient to contain the elements in the specified
     * collection.
     * 
     * @param c the collection whose elements are to be placed into this set
     * @throws NullPointerException if the specified collection is null
     */
    public WeakIdentityHashSet(Collection<? extends E> c) {
        this.map = new WeakIdentityHashMap<E, Object>(Math.max(
                        (int) (c.size() / .75f) + 1, 16));
        addAll(c);
    }


    /**
     * Constructs a new, empty set; the backing <tt>HashMap</tt> instance has
     * the specified initial capacity and default load factor (0.75).
     * 
     * @param initialCapacity the initial capacity of the hash table
     * @throws IllegalArgumentException if the initial capacity is less than
     *             zero
     */
    public WeakIdentityHashSet(int initialCapacity) {
        this.map = new WeakIdentityHashMap<E, Object>(initialCapacity);
    }


    /**
     * Returns an iterator over the elements in this set. The elements are
     * returned in no particular order.
     * 
     * @return an Iterator over the elements in this set
     * @see ConcurrentModificationException
     */
    @Override
    public Iterator<E> iterator() {
        return this.map.keySet().iterator();
    }


    /**
     * Returns the number of elements in this set (its cardinality).
     * 
     * @return the number of elements in this set (its cardinality)
     */
    @Override
    public int size() {
        return this.map.size();
    }


    /**
     * Returns <tt>true</tt> if this set contains no elements.
     * 
     * @return <tt>true</tt> if this set contains no elements
     */
    @Override
    public boolean isEmpty() {
        return this.map.isEmpty();
    }


    /**
     * Returns <tt>true</tt> if this set contains the specified element. More
     * formally, returns <tt>true</tt> if and only if this set contains an
     * element <tt>e</tt> such that
     * <tt>(o==null&nbsp;?&nbsp;e==null&nbsp;:&nbsp;o.equals(e))</tt>.
     * 
     * @param o element whose presence in this set is to be tested
     * @return <tt>true</tt> if this set contains the specified element
     */
    @Override
    public boolean contains(Object o) {
        return this.map.containsKey(o);
    }


    /**
     * Adds the specified element to this set if it is not already present. More
     * formally, adds the specified element <tt>e</tt> to this set if this set
     * contains no element <tt>e2</tt> such that
     * <tt>(e==null&nbsp;?&nbsp;e2==null&nbsp;:&nbsp;e.equals(e2))</tt>. If
     * this set already contains the element, the call leaves the set unchanged
     * and returns <tt>false</tt>.
     * 
     * @param e element to be added to this set
     * @return <tt>true</tt> if this set did not already contain the specified
     *         element
     */
    @Override
    public boolean add(E e) {
        return this.map.put(e, PRESENT) == null;
    }


    /**
     * Removes the specified element from this set if it is present. More
     * formally, removes an element <tt>e</tt> such that
     * <tt>(o==null&nbsp;?&nbsp;e==null&nbsp;:&nbsp;o.equals(e))</tt>, if
     * this set contains such an element. Returns <tt>true</tt> if this set
     * contained the element (or equivalently, if this set changed as a result
     * of the call). (This set will not contain the element once the call
     * returns.)
     * 
     * @param o object to be removed from this set, if present
     * @return <tt>true</tt> if the set contained the specified element
     */
    @Override
    public boolean remove(Object o) {
        return this.map.remove(o) == PRESENT;
    }


    /**
     * Removes all of the elements from this set. The set will be empty after
     * this call returns.
     */
    @Override
    public void clear() {
        this.map.clear();
    }


    /**
     * Returns a shallow copy of this <tt>HashSet</tt> instance: the elements
     * themselves are not cloned.
     * 
     * @return a shallow copy of this set
     */
    @Override
    @SuppressWarnings("unchecked")
    public Object clone() {
        try {
            WeakIdentityHashSet<E> newSet = (WeakIdentityHashSet<E>) super.clone();
            newSet.map = (WeakIdentityHashMap<E, Object>) this.map.clone();
            return newSet;
        } catch (CloneNotSupportedException e) {
            throw new InternalError();
        }
    }


    /**
     * Save the state of this <tt>HashSet</tt> instance to a stream (that is,
     * serialize it).
     * 
     * @serialData The capacity of the backing <tt>HashMap</tt> instance
     *             (int), and its load factor (float) are emitted, followed by
     *             the size of the set (the number of elements it contains)
     *             (int), followed by all of its elements (each an Object) in no
     *             particular order.
     */
    private void writeObject(java.io.ObjectOutputStream s)
                    throws java.io.IOException {
        // Write out any hidden serialization magic
        s.defaultWriteObject();

        // Write out size
        s.writeInt(this.map.size());

        // Write out all elements in the proper order.
        for (Iterator<E> i = this.map.keySet().iterator(); i.hasNext();)
            s.writeObject(i.next());
    }


    /**
     * Reconstitute the <tt>HashSet</tt> instance from a stream (that is,
     * deserialize it).
     */
    @SuppressWarnings("unchecked")
    private void readObject(java.io.ObjectInputStream s)
                    throws java.io.IOException, ClassNotFoundException {
        // Read in any hidden serialization magic
        s.defaultReadObject();

        // Read in size
        int size = s.readInt();
        this.map = new WeakIdentityHashMap<E, Object>(size);

        // Read in all elements in the proper order.
        for (int i = 0; i < size; i++) {
            E e = (E) s.readObject();
            this.map.put(e, PRESENT);
        }
    }

}