package petruchio.common.rationals;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.RoundingMode;

public class BigRational extends Number implements Rational<BigRational> {

    private static final long serialVersionUID = 8603651643464656341L;
    private static final int MAX_CONSTANT = 32;
    private static final BigRational[] posConst = new BigRational[MAX_CONSTANT+1];

    public static final BigRational MINUS_ONE = new BigRational(-1L, 1L);
    public static final BigRational ZERO = posConst[0] = new BigRational(0L, 1L);
    public static final BigRational ONE = posConst[1] = new BigRational(1L, 1L);
    public static final BigRational TEN = posConst[10] = new BigRational(10L, 1L);

    static {
        for(int i = 0; i <= MAX_CONSTANT; ++i) {
            if(posConst[i] == null) {
                posConst[i] = new BigRational(i, 1L);
            }
        }
    }
    
    private final BigInteger numerator;
    private final BigInteger denominator;
    private int hash = 0;
    
    private BigRational(final BigInteger numerator, final BigInteger denominator) {
        this.numerator = numerator;
        this.denominator = denominator;
    }
    
    private BigRational(final BigInteger value) {
        this(value, BigInteger.ONE);
    }
    
    private BigRational(final long numerator, final long denominator) {
        this(BigInteger.valueOf(numerator), BigInteger.valueOf(denominator));
    }
    
    private BigRational(final long value) {
        this(BigInteger.valueOf(value));
    }
    
    public static BigRational valueOf(final BigInteger i) {
        if(i.equals(BigInteger.ONE)) {
            return ONE;
        } else if(i.equals(BigInteger.ZERO)) {
            return ZERO;
        } else if(i.equals(BigInteger.TEN)) {
            return TEN;
        } else {
            return new BigRational(i);
        }
    }
    
    public static BigRational valueOf(BigInteger numerator, BigInteger denominator) {
        if(denominator.equals(BigInteger.ZERO)) {
            throw new ArithmeticException("Division by zero!");
        }
        if(numerator.equals(BigInteger.ZERO)) {
            return ZERO;
        }
        if(denominator.equals(BigInteger.ONE)) {
            return BigRational.valueOf(numerator);
        }
        if(denominator.signum() < 0) {
            numerator = numerator.negate();
            denominator = denominator.negate();
        }
        final BigInteger gcd = numerator.gcd(denominator);
        if(!gcd.equals(BigInteger.ONE)) {
            numerator = numerator.divide(gcd);
            denominator = denominator.divide(gcd);
        }
        return new BigRational(numerator, denominator);
    }
    
    public static BigRational valueOf(long numerator, long denominator) {
        if(denominator == 0L) {
            throw new ArithmeticException("Division by zero!");
        }
        if(denominator == 1L) {
            return valueOf(numerator);
        }
        if(numerator == 0L) {
            return ZERO;
        }
        if(denominator < 0L) {
            numerator = -numerator;
            denominator = -denominator;
        }
        final long gcd = gcd(numerator, denominator);
        if(gcd != 1L) {
            numerator /= gcd;
            denominator /= gcd;
        }
        return new BigRational(numerator, denominator);
    }
    
    public static BigRational valueOf(final long value) {
        if(value == -1L) {
            return MINUS_ONE;
        } else if(value == 0L) {
            return ZERO;
        } else if(value == 1L) {
            return ONE;
        } else if(value == 10L) {
            return TEN;
        } else if(value > 0 && value <= MAX_CONSTANT) {
            return posConst[(int) value];
        } else {
            return new BigRational(value);
        }
    }
    
    @Override
    public int signum() {
        return this.numerator.signum();
    }
    
    @Override
    public int hashCode() {
        if(this.hash == 0) {
            this.hash = this.numerator.hashCode()*31 + this.denominator.hashCode();
        }
        return this.hash;
    }
    
    @Override
    public boolean equals(final Object o) {
        if(o == this) {
            return true;
        }
        if(!(o instanceof BigRational)) {
            return false;
        }
        final BigRational r = (BigRational)o;
        return this.numerator.equals(r.numerator) && this.denominator.equals(r.denominator);
    }
    
    public boolean equals(final BigRational r) {
        return r == this || (this.numerator.equals(r.numerator) && this.denominator.equals(r.denominator));
    }
    
    /* (non-Javadoc)
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    @Override
    public int compareTo(final Rational<BigRational> o) {
        return this.compareTo((BigRational)o);
    }
    
    public int compareTo(final BigRational o) {
        if(this.denominator.equals(o.denominator)) {
            return this.numerator.compareTo(o.numerator);
        }
        return this.numerator.multiply(o.denominator).compareTo(o.numerator.multiply(this.denominator));
    }
    
    public int compareTo(final BigInteger o) {
        if(this.denominator.equals(BigInteger.ONE)) {
            return this.numerator.compareTo(o);
        }
        return this.numerator.compareTo(o.multiply(this.denominator));
    }
    
    @Override
    public int compareTo(final int o) {
        return this.numerator.compareTo(this.denominator.multiply(BigInteger.valueOf(o)));
    }
    
    @Override
    public BigRational add(final Rational<BigRational> summand) {
        return this.add((BigRational)summand);
    }
    
    @Override
    public BigRational subtract(final Rational<BigRational> subtrahend) {
        return this.subtract((BigRational)subtrahend);
    }
    
    @Override
    public BigRational multiply(final Rational<BigRational> factor) {
        return this.multiply((BigRational)factor);
    }
    
    @Override
    public BigRational divide(final Rational<BigRational> divisor) {
        return this.divide((BigRational)divisor);
    }
    
    @Override
    public BigRational add(final int summand) {
        return this.add(BigInteger.valueOf(summand));
    }
    
    @Override
    public BigRational subtract(final int subtrahend) {
        return this.subtract(BigInteger.valueOf(subtrahend));
    }
    
    @Override
    public BigRational multiply(final int factor) {
        return this.multiply(BigInteger.valueOf(factor));
    }
    
    @Override
    public BigRational divide(final int divisor) {
        return this.divide(BigInteger.valueOf(divisor));
    }
    
    @Override
    public BigRational negate() {
        return BigRational.valueOf(this.numerator.negate(), this.denominator);
    }
    
    public BigRational add(final BigRational summand) {
        if(this == ZERO) {
            return summand;
        } else if(summand == ZERO) {
            return this;
        }
        return BigRational.valueOf(this.numerator.multiply(summand.denominator).add(this.denominator.multiply(summand.numerator)),
                            this.denominator.multiply(summand.denominator));
    }
    
    public BigRational subtract(final BigRational subtrahend) {
        if(this == ZERO) {
            return subtrahend.negate();
        } else if(subtrahend == ZERO) {
            return this;
        } 
        return BigRational.valueOf(this.numerator.multiply(subtrahend.denominator).subtract(this.denominator.multiply(subtrahend.numerator)),
                            this.denominator.multiply(subtrahend.denominator));
    }
    
    public BigRational multiply(final BigRational factor) {
        if(this == ONE) {
            return factor;
        } else if(factor == ONE) {
            return this;
        } else if(factor == ZERO || this == ZERO) {
            return ZERO;
        } else if(this == MINUS_ONE) {
            return factor.negate();
        } else if(factor == MINUS_ONE) {
            return this.negate();
        }
        return BigRational.valueOf(this.numerator.multiply(factor.numerator), this.denominator.multiply(factor.denominator));
    }
    
    public BigRational divide(final BigRational divisor) {
        if(divisor == ONE) {
            return this;
        } else if(this == ZERO) {
            return ZERO;
        } else if(divisor == ZERO) {
            throw new ArithmeticException("Division by zero!");
        } else if(divisor == MINUS_ONE) {
            return this.negate();
        }
        return BigRational.valueOf(this.numerator.multiply(divisor.denominator), this.denominator.multiply(divisor.numerator));
    }
    
    @Override
    public BigRational add(final BigInteger summand) {
        return BigRational.valueOf(this.numerator.add(this.denominator.multiply(summand)), this.denominator);
    }
    
    @Override
    public BigRational subtract(final BigInteger subtrahend) {
        return BigRational.valueOf(this.numerator.subtract(this.denominator.multiply(subtrahend)), this.denominator);
    }
    
    @Override
    public BigRational multiply(final BigInteger factor) {
        return BigRational.valueOf(this.numerator.multiply(factor), this.denominator);
    }
    
    @Override
    public BigRational divide(final BigInteger divisor) {
        return BigRational.valueOf(this.numerator, this.denominator.multiply(divisor));
    }
    
    public BigDecimal toDecimal() {
        return new BigDecimal(this.numerator).divide(new BigDecimal(this.denominator));
    }
    
    public BigDecimal toDecimal(final RoundingMode roundingMode) {
        return new BigDecimal(this.numerator).divide(new BigDecimal(this.denominator), roundingMode);
    }
    
    public BigDecimal toDecimal(final int scale, final RoundingMode roundingMode) {
        return new BigDecimal(this.numerator).divide(new BigDecimal(this.denominator), scale, roundingMode);
    }
    
    private static long gcd(long a, long b) {
        a = Math.abs(a);
        b = Math.abs(b);

        // ensure b > a
        if (b <= a) {
            final long tmp = b;
            b = a;
            a = tmp;
        }

        if (a != 0) {
            for (long tmp; (b %= a) != 0L;) {
                tmp = b;
                b = a;
                a = tmp;
            }
            return a;
        } else if (b != 0L) {
            return b;
        } else {
            return 0L;
        }
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        if(this.denominator.equals(BigInteger.ONE)) {
            return this.numerator.toString();
        }
        return "(" + this.numerator.toString() + "/" + this.denominator.toString() + ")";
    }

    // implementing abstract methods from java.lang.Number
    @Override public int    intValue()    { return this.numerator.divide(this.denominator).intValue(); }
    @Override public long   longValue()   { return this.numerator.divide(this.denominator).longValue(); }
    @Override public float  floatValue()  { return this.numerator.floatValue() / this.denominator.floatValue(); }
    @Override public double doubleValue() { return this.numerator.doubleValue() / this.denominator.doubleValue(); }

    @Override public boolean isZero() { return this.numerator.equals(BigInteger.ZERO); }
    @Override public boolean isPositive() { return this.numerator.compareTo(BigInteger.ZERO) > 0; }
    @Override public boolean isNegative() { return this.numerator.compareTo(BigInteger.ZERO) < 0; }
    @Override public boolean isNonPositive() { return this.numerator.compareTo(BigInteger.ZERO) <= 0; }
    @Override public boolean isNonNegative() { return this.numerator.compareTo(BigInteger.ZERO) >= 0; }
    
    @Override public BigInteger numerator() { return this.numerator; }
    @Override public BigInteger denominator() { return this.denominator; }

    @Override
    public Rational<BigRational> ceil() {
        // ceil(a/b) == (a+b-1) DIV b
        return BigRational.valueOf(this.numerator.add(this.denominator).subtract(BigInteger.ONE).divide(this.denominator));
    }

    @Override
    public Rational<BigRational> floor() {
        // floor(a/b) == a DIV b
        return BigRational.valueOf(this.numerator.divide(this.denominator));
    }
    
}
