package petruchio.common.rationals;

import java.math.BigInteger;

public class IntRational extends Number implements Rational<IntRational> {

    private static final long serialVersionUID = 8603651643464656341L;
    private static final int MAX_CONSTANT = 32;
    private static final IntRational[] posConst = new IntRational[MAX_CONSTANT+1];
    
    public static final BigInteger MAX_VALUE = BigInteger.valueOf(Integer.MAX_VALUE);
    public static final IntRational MINUS_ONE = new IntRational(-1);
    public static final IntRational ZERO = posConst[0] = new IntRational(0);
    public static final IntRational ONE = posConst[1] = new IntRational(1);
    public static final IntRational TEN = posConst[10] = new IntRational(10);

    static {
        for(int i = 0; i <= MAX_CONSTANT; ++i) {
            if(posConst[i] == null) {
                posConst[i] = new IntRational(i, 1);
            }
        }
    }
    
    private final int numerator;
    private final int denominator;
    
    private IntRational(final int numerator, final int denominator) {
        this.numerator = numerator;
        this.denominator = denominator;
    }
    
    private IntRational(final int value) {
        this(value, 1);
    }
    
    public static IntRational valueOf(final BigInteger value) {
        if(value.compareTo(MAX_VALUE) > 0) {
            throw new ArithmeticException("Value to big.");
        }
        return valueOf(value.intValue());
    }
    
    public static IntRational valueOf(final int value) {
        switch(value) {
            case -1: return MINUS_ONE;
            case  0: return ZERO;
            case  1: return ONE;
            case 10: return TEN;
            default : 
                if(value > 0 && value <= MAX_CONSTANT) {
                    return posConst[value];
                }
                return new IntRational(value);
        }
    }
    
    public static IntRational valueOf(int numerator, int denominator) {
        if(denominator == 0) {
            throw new ArithmeticException("Division by zero!");
        }
        if(denominator == 1) {
            return valueOf(numerator);
        }
        if(numerator == 0) {
            return ZERO;
        }
        if(denominator < 0) {
            numerator = -numerator;
            denominator = -denominator;
        }
        final int gcd = gcd(numerator, denominator);
        if(gcd != 1) {
            numerator /= gcd;
            denominator /= gcd;
        }
        return new IntRational(numerator, denominator);
    }
    
    @Override
    public int signum() {
        if(this.numerator == 0) {
            return 0;
        } else if(this.numerator < 0) {
            return -1;
        } else {
            return 1;
        }
    }
    
    @Override
    public int hashCode() {
        return this.numerator*31 + this.denominator;
    }
    
    @Override
    public boolean equals(final Object o) {
        if(o == this) {
            return true;
        }
        if(!(o instanceof IntRational)) {
            return false;
        }
        final IntRational r = (IntRational)o;
        return this.numerator == r.numerator && this.denominator == r.denominator;
    }
    
    public boolean equals(final IntRational r) {
        return r == this || (this.numerator == r.numerator && this.denominator == r.denominator);
    }
    
    /* (non-Javadoc)
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    @Override
    public int compareTo(final Rational<IntRational> o) {
        return this.compareTo((IntRational)o);
    }
    
    public int compareTo(final IntRational o) {
        if(this.denominator == o.denominator) {
            return this.numerator-o.numerator;
        }
        return this.numerator*o.denominator-o.numerator*this.denominator;
    }
    
    @Override
    public int compareTo(final int o) {
        return this.numerator-o*this.denominator;
    }
    
    @Override
    public IntRational add(final Rational<IntRational> summand) {
        return this.add((IntRational)summand);
    }
    
    @Override
    public IntRational subtract(final Rational<IntRational> subtrahend) {
        return this.subtract((IntRational)subtrahend);
    }
    
    @Override
    public IntRational multiply(final Rational<IntRational> factor) {
        return this.multiply((IntRational)factor);
    }
    
    @Override
    public IntRational divide(final Rational<IntRational> divisor) {
        return this.divide((IntRational)divisor);
    }
    
    @Override
    public IntRational add(final BigInteger summand) {
        return this.add(summand.intValue());
    }
    
    @Override
    public IntRational subtract(final BigInteger subtrahend) {
        return this.subtract(subtrahend.intValue());
    }
    
    @Override
    public IntRational multiply(final BigInteger factor) {
        return this.multiply(factor.intValue());
    }
    
    @Override
    public IntRational divide(final BigInteger divisor) {
        return this.divide(divisor.intValue());
    }
    
    @Override
    public IntRational negate() {
        if(this.numerator == Integer.MIN_VALUE) {
            throw new ArithmeticException("Overflow!");
        }
        return new IntRational(-this.numerator);
    }
    
    public IntRational add(final IntRational summand) {
        if(this == ZERO) {
            return summand;
        } else if(summand == ZERO) {
            return this;
        }
        return IntRational.valueOf(this.numerator*summand.denominator+this.denominator*summand.numerator,
                            this.denominator*summand.denominator);
    }
    
    public IntRational subtract(final IntRational subtrahend) {
        if(this == ZERO) {
            return subtrahend.negate();
        } else if(subtrahend == ZERO) {
            return this;
        } 
        return IntRational.valueOf(this.numerator*subtrahend.denominator-this.denominator*subtrahend.numerator,
                            this.denominator*subtrahend.denominator);
    }
    
    public IntRational multiply(final IntRational factor) {
        if(this == ONE) {
            return factor;
        } else if(factor == ONE) {
            return this;
        } else if(factor == ZERO || this == ZERO) {
            return ZERO;
        } else if(this == MINUS_ONE) {
            return factor.negate();
        } else if(factor == MINUS_ONE) {
            return this.negate();
        }
        return IntRational.valueOf(this.numerator*factor.numerator, this.denominator*factor.denominator);
    }
    
    public IntRational divide(final IntRational divisor) {
        if(divisor == ONE) {
            return this;
        } else if(this == ZERO) {
            return ZERO;
        } else if(divisor == ZERO) {
            throw new ArithmeticException("Division by zero!");
        } else if(divisor == MINUS_ONE) {
            return this.negate();
        }
        return IntRational.valueOf(this.numerator*divisor.denominator, this.denominator*divisor.numerator);
    }
    
    @Override
    public IntRational add(final int summand) {
        if(this == ZERO) {
            return valueOf(summand);
        } else if(summand == 0) {
            return this;
        }
        return IntRational.valueOf(this.numerator+this.denominator*summand, this.denominator);
    }
    
    @Override
    public IntRational subtract(final int subtrahend) {
        if(this == ZERO) {
            return valueOf(-subtrahend);
        } else if(subtrahend == 0) {
            return this;
        } 
        return IntRational.valueOf(this.numerator-this.denominator*subtrahend, this.denominator);
    }
    
    @Override
    public IntRational multiply(final int factor) {
        if(this == ONE) {
            return valueOf(factor);
        } else if(factor == 1) {
            return this;
        } else if(factor == 0 || this == ZERO) {
            return ZERO;
        } else if(this == MINUS_ONE) {
            return valueOf(-factor);
        } else if(factor == -1) {
            return this.negate();
        }
        return IntRational.valueOf(this.numerator*factor, this.denominator);
    }
    
    @Override
    public BigInteger numerator() {
        return BigInteger.valueOf(this.numerator);
    }

    @Override
    public BigInteger denominator() {
        return BigInteger.valueOf(this.denominator);
    }
    
    @Override
    public IntRational divide(final int divisor) {
        if(divisor == 1) {
            return this;
        } else if(this == ZERO) {
            return ZERO;
        } else if(divisor == 0) {
            throw new ArithmeticException("Division by zero!");
        } else if(divisor == -1) {
            return this.negate();
        }
        return IntRational.valueOf(this.numerator, this.denominator*divisor);
    }
    
    private static int gcd(int a, int b) {
        a = Math.abs(a);
        b = Math.abs(b);

        // ensure b > a
        if (b <= a) {
            final int tmp = b;
            b = a;
            a = tmp;
        }

        if (a != 0) {
            for (int tmp; (b %= a) != 0;) {
                tmp = b;
                b = a;
                a = tmp;
            }
            return a;
        } else if (b != 0) {
            return b;
        } else {
            return 0;
        }
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        if(this.denominator == 1) {
            return String.valueOf(this.numerator);
        }
        return "(" + this.numerator + "/" + this.denominator + ")";
    }

    // implementing abstract methods from java.lang.Number
    @Override public int    intValue()    { return this.numerator/this.denominator; }
    @Override public long   longValue()   { return (long)this.numerator/this.denominator; }
    @Override public float  floatValue()  { return (float)this.numerator / this.denominator; }
    @Override public double doubleValue() { return (double)this.numerator / this.denominator; }
    
    @Override public boolean isZero() { return this.numerator == 0; }
    @Override public boolean isPositive() { return this.numerator > 0; }
    @Override public boolean isNegative() { return this.numerator < 0; }
    @Override public boolean isNonPositive() { return this.numerator <= 0; }
    @Override public boolean isNonNegative() { return this.numerator >= 0; }
    
    @Override
    public Rational<IntRational> ceil() {
        // ceil(a/b) == (a+b-1) DIV b
        return IntRational.valueOf((this.numerator + this.denominator - 1) / this.denominator);
    }

    @Override
    public Rational<IntRational> floor() {
        // floor(a/b) == a DIV b
        return IntRational.valueOf(this.numerator / this.denominator);
    }
    
}
