package petruchio.common.rationals;

import java.math.BigInteger;

public class LongRational extends Number implements Rational<LongRational> {

    private static final long serialVersionUID = 8603651643464656341L;
    private static final int MAX_CONSTANT = 32;
    private static final LongRational[] posConst = new LongRational[MAX_CONSTANT+1];
    
    public static final BigInteger MAX_VALUE = BigInteger.valueOf(Long.MAX_VALUE);
    public static final LongRational MINUS_ONE = new LongRational(-1L, 1L);
    public static final LongRational ZERO = posConst[0] = new LongRational(0L, 1L);
    public static final LongRational ONE = posConst[1] = new LongRational(1L, 1L);
    public static final LongRational TEN = posConst[10] = new LongRational(10L, 1L);

    static {
        for(int i = 0; i <= MAX_CONSTANT; ++i) {
            if(posConst[i] == null) {
                posConst[i] = new LongRational(i, 1L);
            }
        }
    }
    
    private final long numerator;
    private final long denominator;
    
    private LongRational(final long numerator, final long denominator) {
        this.numerator = numerator;
        this.denominator = denominator;
    }
    
    private LongRational(final long value) {
        this(value, 1L);
    }
    
    @Override
    public BigInteger numerator() {
        return BigInteger.valueOf(this.numerator);
    }

    @Override
    public BigInteger denominator() {
        return BigInteger.valueOf(this.denominator);
    }
    
    public static LongRational valueOf(long numerator, long denominator) {
        if(denominator == 0L) {
            throw new ArithmeticException("Division by zero!");
        }
        if(denominator == 1L) {
            return valueOf(numerator);
        }
        if(numerator == 0L) {
            return ZERO;
        }
        if(denominator < 0L) {
            numerator = -numerator;
            denominator = -denominator;
        }
        final long gcd = gcd(numerator, denominator);
        if(gcd != 1L) {
            numerator /= gcd;
            denominator /= gcd;
        }
        return new LongRational(numerator, denominator);
    }
    
    public static LongRational valueOf(final BigInteger value) {
        if(value.compareTo(MAX_VALUE) > 0) {
            throw new ArithmeticException("Value to big.");
        }
        return valueOf(value.longValue());
    }
    
    public static LongRational valueOf(final long value) {
        if(value == -1L) {
            return MINUS_ONE;
        } else if(value == 0L) {
            return ZERO;
        } else if(value == 1L) {
            return ONE;
        } else if(value == 10L) {
            return TEN;
        } else if(value > 0 && value <= MAX_CONSTANT) {
            return posConst[(int) value];
        } else {
            return new LongRational(value);
        }
    }
    
    @Override
    public int signum() {
        if(this.numerator == 0L) {
            return 0;
        } else if(this.numerator < 0L) {
            return -1;
        } else {
            return 1;
        }
    }
    
    @Override
    public int hashCode() {
        return (int)(this.numerator*31L + this.denominator);
    }
    
    @Override
    public boolean equals(final Object o) {
        if(o == this) {
            return true;
        }
        if(!(o instanceof LongRational)) {
            return false;
        }
        final LongRational r = (LongRational)o;
        return this.numerator == r.numerator && this.denominator == r.denominator;
    }
    
    public boolean equals(final LongRational r) {
        return r == this || (this.numerator == r.numerator && this.denominator == r.denominator);
    }
    
    /* (non-Javadoc)
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    @Override
    public int compareTo(final Rational<LongRational> o) {
        return this.compareTo((LongRational)o);
    }
    
    public int compareTo(final LongRational o) {
        if(this.denominator == o.denominator) {
            if(this.numerator == o.numerator) {
                return 0;
            }
            return this.numerator > o.numerator ? 1 : -1;
        }
        final long a = this.numerator*o.denominator;
        final long b = o.numerator*this.denominator;
        if(a == b) {
            return 0;
        }
        return a > b ? 1 : -1;
    }
    
    @Override
    public int compareTo(final int o) {
        return (int)(this.numerator-o*this.denominator);
    }
    
    @Override
    public LongRational add(final BigInteger summand) {
        return this.add(summand.longValue());
    }
    
    @Override
    public LongRational subtract(final BigInteger subtrahend) {
        return this.subtract(subtrahend.longValue());
    }
    
    @Override
    public LongRational multiply(final BigInteger factor) {
        return this.multiply(factor.longValue());
    }
    
    @Override
    public LongRational divide(final BigInteger divisor) {
        return this.divide(divisor.longValue());
    }
    
    @Override
    public LongRational add(final int summand) {
        return this.add((long)summand);
    }
    
    @Override
    public LongRational subtract(final int subtrahend) {
        return this.subtract((long)subtrahend);
    }
    
    @Override
    public LongRational multiply(final int factor) {
        return this.multiply((long)factor);
    }
    
    @Override
    public LongRational divide(final int divisor) {
        return this.divide((long)divisor);
    }
    
    @Override
    public LongRational add(final Rational<LongRational> summand) {
        return this.add((LongRational)summand);
    }
    
    @Override
    public LongRational subtract(final Rational<LongRational> subtrahend) {
        return this.subtract((LongRational)subtrahend);
    }
    
    @Override
    public LongRational multiply(final Rational<LongRational> factor) {
        return this.multiply((LongRational)factor);
    }
    
    @Override
    public LongRational divide(final Rational<LongRational> divisor) {
        return this.divide((LongRational)divisor);
    }
    
    @Override
    public LongRational negate() {
        return LongRational.valueOf(-this.numerator);
    }
    
    public LongRational add(final LongRational summand) {
        if(this == ZERO) {
            return summand;
        } else if(summand == ZERO) {
            return this;
        }
        return LongRational.valueOf(this.numerator*summand.denominator+this.denominator*summand.numerator,
                            this.denominator*summand.denominator);
    }
    
    public LongRational subtract(final LongRational subtrahend) {
        if(this == ZERO) {
            return subtrahend.negate();
        } else if(subtrahend == ZERO) {
            return this;
        } 
        return LongRational.valueOf(this.numerator*subtrahend.denominator-this.denominator*subtrahend.numerator,
                            this.denominator*subtrahend.denominator);
    }
    
    public LongRational multiply(final LongRational factor) {
        if(this == ONE) {
            return factor;
        } else if(factor == ONE) {
            return this;
        } else if(factor == ZERO || this == ZERO) {
            return ZERO;
        } else if(this == MINUS_ONE) {
            return factor.negate();
        } else if(factor == MINUS_ONE) {
            return this.negate();
        }
        return LongRational.valueOf(this.numerator*factor.numerator, this.denominator*factor.denominator);
    }
    
    public LongRational divide(final LongRational divisor) {
        if(divisor == ONE) {
            return this;
        } else if(this == ZERO) {
            return ZERO;
        } else if(divisor == ZERO) {
            throw new ArithmeticException("Division by zero!");
        } else if(divisor == MINUS_ONE) {
            return this.negate();
        }
        return LongRational.valueOf(this.numerator*divisor.denominator, this.denominator*divisor.numerator);
    }
    
    public LongRational add(final long summand) {
        if(this == ZERO) {
            return valueOf(summand);
        } else if(summand == 0L) {
            return this;
        }
        return LongRational.valueOf(this.numerator+this.denominator*summand, this.denominator);
    }
    
    public LongRational subtract(final long subtrahend) {
        if(this == ZERO) {
            return valueOf(-subtrahend);
        } else if(subtrahend == 0L) {
            return this;
        } 
        return LongRational.valueOf(this.numerator-this.denominator*subtrahend, this.denominator);
    }
    
    public LongRational multiply(final long factor) {
        if(this == ONE) {
            return valueOf(factor);
        } else if(factor == 1L) {
            return this;
        } else if(factor == 0L || this == ZERO) {
            return ZERO;
        } else if(this == MINUS_ONE) {
            return valueOf(-factor);
        } else if(factor == -1L) {
            return this.negate();
        }
        return LongRational.valueOf(this.numerator*factor, this.denominator);
    }
    
    public LongRational divide(final long divisor) {
        if(divisor == 1L) {
            return this;
        } else if(this == ZERO) {
            return ZERO;
        } else if(divisor == 0L) {
            throw new ArithmeticException("Division by zero!");
        } else if(divisor == -1L) {
            return this.negate();
        }
        return LongRational.valueOf(this.numerator, this.denominator*divisor);
    }
    
    private static long gcd(long a, long b) {
        a = Math.abs(a);
        b = Math.abs(b);

        // ensure b > a
        if (b <= a) {
            final long tmp = b;
            b = a;
            a = tmp;
        }

        if (a != 0) {
            for (long tmp; (b %= a) != 0L;) {
                tmp = b;
                b = a;
                a = tmp;
            }
            return a;
        } else if (b != 0L) {
            return b;
        } else {
            return 0L;
        }
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        if(this.denominator == 1L) {
            return String.valueOf(this.numerator);
        }
        return "(" + this.numerator + "/" + this.denominator + ")";
    }

    // implementing abstract methods from java.lang.Number
    @Override public int    intValue()    { return (int)(this.numerator/this.denominator); }
    @Override public long   longValue()   { return this.numerator/this.denominator; }
    @Override public float  floatValue()  { return (float)this.numerator / this.denominator; }
    @Override public double doubleValue() { return (double)this.numerator / this.denominator; }

    @Override public boolean isZero() { return this.numerator == 0L; }
    @Override public boolean isPositive() { return this.numerator > 0L; }
    @Override public boolean isNegative() { return this.numerator < 0L; }
    @Override public boolean isNonPositive() { return this.numerator <= 0L; }
    @Override public boolean isNonNegative() { return this.numerator >= 0L; }
    
    
    @Override
    public Rational<LongRational> ceil() {
        // ceil(a/b) == (a+b-1) DIV b
        return LongRational.valueOf((this.numerator + this.denominator - 1L) / this.denominator);
    }

    @Override
    public Rational<LongRational> floor() {
        // floor(a/b) == a DIV b
        return LongRational.valueOf(this.numerator / this.denominator);
    }
    
}

