/*
 * File:    CommunicatorCache.java
 * Project: petruchio
 * Author:  Tim Strazny
 * Created: 15.03.2009
 */
package petruchio.compiler;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import petruchio.common.IdentityHashSet;
import petruchio.compiler.ProcessData.NameCard;
import petruchio.compiler.reactions.SingleReactionTask;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.algorithms.ReactionFinder;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.InputAction;
import petruchio.interfaces.pi.OutputAction;
import petruchio.interfaces.pi.Process;

/**
 * @author Tim Strazny
 */
public class CommunicatorCache implements petruchio.interfaces.CommunicatorCache<ProcessData>, SelfDescribing {

    private final Map<NameCard, Collection<SingleReactionTask<ProcessData>>> communicatorsIn = new HashMap<NameCard, Collection<SingleReactionTask<ProcessData>>>();
    private final Map<NameCard, Collection<SingleReactionTask<ProcessData>>> communicatorsOut = new HashMap<NameCard, Collection<SingleReactionTask<ProcessData>>>();
    private final Map<Place, Process<ProcessData>> place2proc = Collections.synchronizedMap(new IdentityHashMap<Place, Process<ProcessData>>());
    private final Map<Place, Process<ProcessData>> place2procUnmod = Collections.unmodifiableMap(this.place2proc);
    private ReactionFinder<ProcessData> reactionFinder = null;
    
    @Override
    public synchronized void addAll(final Collection<SingleReactionTask<ProcessData>> p) {
        for(final SingleReactionTask<ProcessData> a : p) {
            this.add(a);
        }
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CommunicatorCache#add(petruchio.compiler.reactions.SingleReactionTask)
     */
    @Override
    public synchronized Collection<ActionPrefix> add(final SingleReactionTask<ProcessData> p) {
        final boolean update = CommonTasks.getData(p.getProcess()).getCommunicators() == null;
        final Collection<ActionPrefix> actions = this.getCommunicationPrefixes(p.getProcess());
        if(update) {
            p.getPlace().setCouldCommunicate(!actions.isEmpty());
            for(final ActionPrefix ap : actions) {
                if(ap instanceof InputAction) {
                    final InputAction a = (InputAction)ap;
                    final NameCard n = new NameCard(a.getChannel(), a.getNames().size());
                    Collection<SingleReactionTask<ProcessData>> col = this.communicatorsIn.get(n);
                    if(col == null) {
                        col = new ArrayList<SingleReactionTask<ProcessData>>();
                        this.communicatorsIn.put(n, col);
                    }
                    col.add(p);
                } else if(ap instanceof OutputAction) {
                    final OutputAction a = (OutputAction)ap;
                    final NameCard n = new NameCard(a.getChannel(), a.getNames().size());
                    Collection<SingleReactionTask<ProcessData>> col = this.communicatorsOut.get(n);
                    if(col == null) {
                        col = new ArrayList<SingleReactionTask<ProcessData>>();
                        this.communicatorsOut.put(n, col);
                    }
                    col.add(p);
                }
            }
        }
        return actions;
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public synchronized void reset() {
        this.communicatorsIn.clear();
        this.communicatorsOut.clear();
        this.place2proc.clear();
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "Keeps track of the input and output actions processes have and makes finding communicating pairs easy.";
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.CommunicatorCache#getReceiversOn(petruchio.compiler.ProcessData.NameCard)
     */
    @Override
    public synchronized Collection<SingleReactionTask<ProcessData>> getReceiversOn(final NameCard channel) {
        return this.communicatorsIn.get(channel);
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.CommunicatorCache#getSendersOn(petruchio.compiler.ProcessData.NameCard)
     */
    @Override
    public synchronized Collection<SingleReactionTask<ProcessData>> getSendersOn(final NameCard channel) {
        return this.communicatorsOut.get(channel);
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.CommunicatorCache#setReactionFinder(petruchio.interfaces.algorithms.ReactionFinder)
     */
    @Override
    public synchronized void setReactionFinder(final ReactionFinder<ProcessData> reactionFinder) {
       this.reactionFinder = reactionFinder;
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.CommunicatorCache#canCommunicate(petruchio.compiler.reactions.SingleReactionTask, petruchio.compiler.reactions.SingleReactionTask)
     */
    @Override
    public synchronized boolean canCommunicate(final SingleReactionTask<ProcessData> a, final SingleReactionTask<ProcessData> b) {
        if(!a.getPlace().couldCommunicate() || !b.getPlace().couldCommunicate()) {
            return false;
        }
        final Collection<ActionPrefix> actionsA = this.getCommunicationPrefixes(a.getProcess());
        final Collection<ActionPrefix> actionsB = this.getCommunicationPrefixes(b.getProcess());
        for(final ActionPrefix apA : actionsA) {
            if(apA instanceof InputAction) {
                final InputAction in = (InputAction)apA;
                for(final ActionPrefix apB : actionsB) {
                    if(apB instanceof OutputAction) {
                        final OutputAction out = (OutputAction)apB;
                        if(in.getNames().size() == out.getNames().size() && in.getChannel().equals(out.getChannel())) {
                            return true;
                        }
                    }
                }
            } else if(apA instanceof OutputAction) {
                final OutputAction out = (OutputAction)apA;
                for(final ActionPrefix apB : actionsB) {
                    if(apB instanceof InputAction) {
                        final InputAction in = (InputAction)apB;
                        if(in.getNames().size() == out.getNames().size() && in.getChannel().equals(out.getChannel())) {
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CommunicatorCache#canCommunicate(petruchio.compiler.reactions.SingleReactionTask, petruchio.compiler.reactions.SingleReactionTask)
     */
    @Override
    public synchronized boolean canCommunicate(final Place a, final Place b) {
        if(!a.couldCommunicate() || !b.couldCommunicate()) {
            return false;
        }
        final Collection<ActionPrefix> actionsA = this.getCommunicationPrefixes(a);
        final Collection<ActionPrefix> actionsB = this.getCommunicationPrefixes(b);
        for(final ActionPrefix apA : actionsA) {
            if(apA instanceof InputAction) {
                final InputAction in = (InputAction)apA;
                for(final ActionPrefix apB : actionsB) {
                    if(apB instanceof OutputAction) {
                        final OutputAction out = (OutputAction)apB;
                        if(in.getNames().size() == out.getNames().size() && in.getChannel().equals(out.getChannel())) {
                            return true;
                        }
                    }
                }
            } else if(apA instanceof OutputAction) {
                final OutputAction out = (OutputAction)apA;
                for(final ActionPrefix apB : actionsB) {
                    if(apB instanceof InputAction) {
                        final InputAction in = (InputAction)apB;
                        if(in.getNames().size() == out.getNames().size() && in.getChannel().equals(out.getChannel())) {
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }

    /**
     * @param p
     * @return
     */
    private synchronized Collection<ActionPrefix> getCommunicationPrefixes(final Place p) {
        return this.getCommunicationPrefixes(this.place2proc.get(p));
    }
    
    /**
     * @param p
     * @return
     */
    private synchronized Collection<ActionPrefix> getCommunicationPrefixes(final Process<ProcessData> p) {
        final ProcessData d = CommonTasks.getData(p);
        if(d.getCommunicators() == null) {
            d.setCommunicators(this.reactionFinder.getAvailableCommunicationsViaFreeNames(p));
        }
        return d.getCommunicators();
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.CommunicatorCache#getCommunicators(petruchio.interfaces.pi.Process)
     */
    @Override
    public synchronized Collection<SingleReactionTask<ProcessData>> getCommunicators(final Process<ProcessData> p) {
        return this.getCommunicators(this.getCommunicationPrefixes(p));
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CommunicatorCache#getCommunicators(petruchio.interfaces.petrinet.Place)
     */
    @Override
    public synchronized Collection<SingleReactionTask<ProcessData>> getCommunicators(final Place p) {
        return this.getCommunicators(this.getCommunicationPrefixes(p));
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.CommunicatorCache#getCommunicators(java.util.Collection)
     */
    @Override
    public synchronized Collection<SingleReactionTask<ProcessData>> getCommunicators(final Collection<ActionPrefix> actions) {
        final Set<SingleReactionTask<ProcessData>> procs = new IdentityHashSet<SingleReactionTask<ProcessData>>(0);
        for(final ActionPrefix ap : actions) {
            if(ap instanceof InputAction) {
                final InputAction a = (InputAction)ap;
                final NameCard n = new NameCard(a.getChannel(), a.getNames().size());
                final Collection<SingleReactionTask<ProcessData>> r = this.getSendersOn(n);
                if(r != null) {
                    procs.addAll(r);
                }
            } else if(ap instanceof OutputAction) {
                final OutputAction a = (OutputAction)ap;
                final NameCard n = new NameCard(a.getChannel(), a.getNames().size());
                final Collection<SingleReactionTask<ProcessData>> r = this.getReceiversOn(n);
                if(r != null) {
                    procs.addAll(r);
                }
            }
        }
        return procs;
    }

    @Override
    public Process<ProcessData> getProcess(final Place p) {
        return this.place2proc.get(p);
    }

    @Override
    public void associateWithProcess(final Place p, final Process<ProcessData> proc) {
        this.place2proc.put(p, proc);
    }

    @Override
    public Set<Entry<Place, Process<ProcessData>>> placeProcAssociations() {
        return this.place2proc.entrySet();
    }

    @Override
    public Map<Place, Process<ProcessData>> placeProcMap() {
        return this.place2procUnmod;
    }

}
