/*
 * Project:  Petruchio
 * Filename: CompilerProperty.java
 * Created:  23.05.2007
 * Author:   Tim Strazny
 */
package petruchio.compiler;

import java.util.HashMap;
import java.util.Map;
import java.util.Arrays;


/**
 * The compiler properties accepted by the Petruchio Pi2PN tool.
 * 
 * @author Tim Strazny
 */
public enum CompilerProperty implements petruchio.interfaces.CompilerProperty {
    
    // algorithms
    PROCESS_CREATOR("ProcessCreator", "Algorithms",
                    petruchio.interfaces.pi.ProcessCreator.class,
                    false, false,
                    "petruchio.pi.ProcessCreator",
                    "The fully qualified name of the pi process creator class that shall be used."),
    PROCESS_MANAGER("ProcessManager", "Algorithms",
                    petruchio.interfaces.pi.ProcessManager.class,
                    false, false,
                    "petruchio.pi.ProcessManager",
                    "The fully qualified name of the process manager class that shall be used."),
    RESTRICTED_FORM("RestrictedForm", "Algorithms",
                    petruchio.interfaces.algorithms.RestrictedForm.class,
                    false, false,
                    "petruchio.compiler.Restrictor",
                    "The fully qualified name of the restricted form builder class that shall be used."),
    STRUCTURAL_CONGRUENCE("StructuralCongruence", "Algorithms",
                    petruchio.interfaces.algorithms.StructuralCongruence.class,
                    false, false,
                    "petruchio.compiler.struct.DefaultStructChecker",
                    "The fully qualified name of the structural congruence checker class that shall be used."),
    REACTION_FINDER("ReactionFinder", "Algorithms",
                    petruchio.interfaces.algorithms.ReactionFinder.class,
                    false, false,
                    "petruchio.compiler.ReactionFinder",
                    "The fully qualified name of the reaction finder class that shall be used."),
    COVERABILITY("Coverability", "Algorithms",
                    petruchio.interfaces.algorithms.Coverability.class,
                    false, false,
                    "petruchio.cov.Backward",//"petruchio.cov.Coverability",
                    "The fully qualified name of the coverability checker class that shall be used."),
    STRUCTURAL_BOUNDEDNESS_CHECKER("StructuralBoundednessCheckers", "Algorithms",
                    petruchio.interfaces.algorithms.StructuralBoundedness.class,
                    true, true,
                    "", //"petruchio.compiler.sbcheckers.SimpleUnboundedBreadthChecker",
                    "The fully qualified names of the structural boundedness checker classes that shall be used."),
    PETRI_NET("PetriNet", "Algorithms",
                    petruchio.interfaces.petrinet.PetriNet.class,
                    false, false,
                    "petruchio.pn.PetriNet",
                    "The fully qualified name of the Petri net class that shall be used."),
    COMMUNICATOR_CACHE("CommunicatorCache", "Algorithms",
                    petruchio.interfaces.CommunicatorCache.class,
                    false, false,
                    "petruchio.compiler.CommunicatorCache",
                    "The fully qualified name of the class that shall be used to cache which processes can communicate."),
    // IO
    PARSER("PiParser", "Input/Output",
                    petruchio.interfaces.pi.PiParser.class,
                    false, true,
                    "petruchio.pi.readers.cup.PiParser",
                    "The fully qualified name of the pi parser class that shall be used."),
    PI_WRITERS("PiWriters", "Input/Output",
                    petruchio.interfaces.pi.PiWriter.class,
                    true, true,
                    "",
                    "The fully qualified name of the pi writer class that shall be used to output the pi process definitions."),
    LOCAL_CONFIG_PARSER("LocalConfigurationParser", "Input/Output",
		    petruchio.interfaces.LocalConfigurationParser.class,
		    false, true,
		    "petruchio.pi.readers.LocalConfigParser",
		    "The fully qualified name of the parser class that shall be used to read the local configuration from a pi file."),
    PN_READER("PetriNetReader", "Input/Output",
                    petruchio.interfaces.petrinet.PetriNetReader.class,
                    false, true,
                    "petruchio.pn.writers.PEPReader",
                    "The fully qualified name of the Petri net reader that shall be used when converting Petri nets."),
    PN_WRITERS("PetriNetWriters", "Input/Output",
                    petruchio.interfaces.petrinet.PetriNetWriter.class,
                    true, true,
                    "petruchio.pn.writers.PEPWriter",
                    "The fully qualified name of the Petri net writers that shall be used."),

    // Reductions
    PARTIAL_ORDER_REDUCTION("PartialOrderReduction", "Reductions",
                    Boolean.TYPE,
                    false, false,
                    "true",
                    "If set to true, methods of partial order reduction are applied during the compilation."),
    PN_REDUCTORS("PetriNetReductors", "Reductions",
                    petruchio.interfaces.petrinet.PetriNetReductor.class,
                    true, true,
                    "", //"petruchio.pn.reductors.AlternativesReduction",
                    "A comma separated list of Petri net reductors that shall be applied after the compilation has finished."),
    PROCESS_MODIFIERS("ProcessModifiers", "Reductions",
                    petruchio.interfaces.pi.ProcessModifier.class,
                    true, true,
                    "",
                    "A comma separated list of pi process modificators that will modify the processes before compilation."),
    PN_REDUCTOR_IGNORER("PetriNetReductorIgnorer", "Reductions",
                    petruchio.interfaces.petrinet.PetriNetReductorIgnorer.class,
                    false, true,
                    "",
                    "The fully qualified name of the class that tells the reductors which elements to ignore."),
    PN_REDUCTOR_IGNORER_INPUT("PetriNetReductorIgnorerInput", "Reductions",
                    String.class,
                    false, true,
                    "",
                    "The input for the PetriNetReductorIgnore object."),
    PN_REDUCTOR_IGNORER_INPUT_TYPE("PetriNetReductorIgnorerInputType", "Reductions",
                    petruchio.interfaces.petrinet.PetriNetReductorIgnorerInput.class,
                    false, true,
                    "petruchio.pn.rdiginput.StringInput",
                    "The input type converter for the PetriNetReductorIgnore object."),
    REPEAT_REDUCTIONS("RepeatReductions", "Reductions",
                    Boolean.TYPE,
                    false, false,
                    "true",
                    "If set to true, the reductions are iterated to reduce the net even further."),
                    
                
    // logging
    LOG_LEVEL("LogLevel", "Logging",
                    String.class,
                    false, false,
                    "INFO",
                    "The level of logging. Values are: OFF, SEVERE, WARNING, INFO, CONFIG, FINE, FINER, FINEST, ALL"),
    USE_PARENT_LOGGERS("UseParentLoggers", "Logging",
                    Boolean.TYPE,
                    false, false,
                    "false",
                    "If set to true, the log output will be passed to the logger's parent."),
    LOG_HANDLER("LogHandler", "Logging",
                    java.util.logging.Handler.class,
                    false, false,
                    "java.util.logging.ConsoleHandler",
                    "The fully qualified name of the log handler class that shall be used."),
    LOG_ERROR_MANAGER("LogErrorManager", "Logging",
                    java.util.logging.ErrorManager.class,
                    false, false,
                    "java.util.logging.ErrorManager",
                    "The fully qualified name of the error manager the log handler shall use."),
    LOG_FORMATTER("LogFormatter", "Logging",
                    java.util.logging.Formatter.class,
                    false, false,
                    "petruchio.common.SimplerFormatter",
                    "The fully qualified name of the formatter the log handler shall use."),
    LOG_FILTER("LogFilter", "Logging",
                    java.util.logging.Filter.class,
                    false, true,
                    "",
                    "The fully qualified name of the filter the log handler shall use."),
    LOG_ENCODING("LogEncoding", "Logging",
                    String.class,
                    false, true,
                    "",
                    "The name of the encoding the log handler shall use."),
    PARSER_ERROR_HANDLERS("PiParserErrorHandlers", "Logging",
                    petruchio.interfaces.ParserErrorHandler.class,
                    true, true,
                    "petruchio.pi.readers.ConsoleErrorHandler",
                    "The fully qualified name of the pi parser error handler class that shall be used."),
                    
    // compilation
    SEMANTICS("Semantics", "Compilation",
                    petruchio.compiler.Semantics.class,
                    true, false,
                    "STRUCTURAL",
                    "A comma separated list of the semantics that shall be used. Values are: " + Arrays.toString(Semantics.values())),
    MAXIMUM_REACTION_COUNT("MaximumReactionCount", "Compilation",
                    Integer.TYPE,
                    false, false,
                    "-1",
                    "The maximal number of reactions that shall be compiled. Set it to a value < 0 to disable."),
    ENVIRONMENT_PROCESSES("EnvironmentProcesses", "Compilation",
                    String.class,
                    true, true,
                    "",
                    "Identificators of 'environment' processes."),
    REACHABLE_ONLY("ReachableOnly", "Compilation",
                    Boolean.TYPE,
                    false, false,
                    "true",
                    "If set to false, communications of places that cannot be marked at the same time will be compiled (terminates only if process is absolutely structurally stationary."),
    IGNORE_COMMUNICATIONS("IgnoreCommunications", "Compilation",
                    Boolean.TYPE,
                    false, false,
                    "false",
                    "If set to true, no communications via free names are compiled."),
    KEEP_GLOBAL_RESTRICTIONS("KeepRestrictionsOnPublicNames", "Compilation",
                    Boolean.TYPE,
                    false, false,
                    "true",
                    "If set to false, the restrictions in the system process that are shared by all sub processes are removed."),
    MAIN_PROCESS("MainProcess", "Compilation",
                    String.class,
                    false, true,
                    "SYSTEM",
                    "The name that identifies the main process ID if the 'main' keyword isn't used."),

    // Optimizations
    MINIMUM_TOKEN_COUNT_FOR_ACCELERATION("MinimumTokenCountForAcceleration", "Optimizations",
                    Integer.TYPE,
                    false, false,
                    "1",
                    "Optional optimization for the coverability checker. Only test new nodes for covering if it contains a place with at least this many tokens. Set to 1 to disable or <= 0 to never test for covering."),
    MAXIMUM_TOKEN_COUNT_FOR_ACCELERATION("MaximumTokenCountForAcceleration", "Optimizations",
                    Integer.TYPE,
                    false, false,
                    "-1",
                    "Optional optimization for the coverability checker. Introduces an omega as soon, as the token count of a place gets greater than this value. Set to < 0 to disable."),
    MARK_UNUSED_PLACES("MarkUnusedPlaces", "Optimizations",
                    Boolean.TYPE,
                    false, false,
                    "false",
                    "If set to true, a simple optimization for the coverability checker may be performed."),

    // Additional Information
    COMPUTE_BOUNDS("ComputeBounds", "Additional Information",
                    petruchio.interfaces.CompilerProperty.ExtendedBoolean.class,
                    false, false,
                    "TRUE_IF_FOR_FREE",
                    "If enabled, the places will be annotated with their specific boundedness."),
    CHECK_FOR_DEADLOCKS("CheckForDeadlocks", "Additional Information",
                    petruchio.interfaces.CompilerProperty.ExtendedBoolean.class,
                    false, false,
                    "TRUE_IF_FOR_FREE",
                    "If enabled, the net is checked for some kinds of deadlocks."),
    IGNORE_LAYOUT("IgnoreLayout", "Additional Information",
                    Boolean.TYPE,
                    false, false,
                    "false",
                    "If set to true, no layouting is performed."),

    
    // misc
    RETRIEVER("Retriever", "Miscellaneous",
                    petruchio.interfaces.Retriever.class,
                    false, true,
                    "petruchio.compiler.Retriever",
                    "Translates a Petri net marking to a process and a firing sequence to a reaction sequence."),
    MAIN_PRIORITY("MainPriority", "Miscellaneous",
                    Integer.TYPE,
                    false, false,
                    "4",
                    "The priority that shall be used for the main compiler thread."),
    COVERABILITY_PRIORITY("CoverabilityPriority", "Miscellaneous",
                    Integer.TYPE,
                    false, false,
                    "5",
                    "The priority that shall be used for the coverability checker thread."),
    USE_THREADS("UseThreads", "Miscellaneous",
                    Boolean.TYPE,
                    false, false,
                    "true",
                    "If set to false, the main compiler thread will wait for the coverability checker thread to finish before moving on."),
    TERMINATION_PATIENCE("TerminationPatience", "Miscellaneous",
                    Integer.TYPE,
                    false, false,
                    "5000",
                    "How many ms the compiler should wait after pleaseStop() has been called before stopping the coverability checker thread forcefully."),
    PN_OFFSET_X("PetriNetOffsetX", "Miscellaneous",
                    Integer.TYPE,
                    false, false,
                    "50",
                    "The maximal offset of Petri net elements in x-direction."),
    PN_OFFSET_Y("PetriNetOffsetY", "Miscellaneous",
                    Integer.TYPE,
                    false, false,
                    "50",
                    "The maximal offset of Petri net elements in y-direction.");
    
    private static Map<String, petruchio.interfaces.CompilerProperty> key2props;
    
    private final String key;
    private final Class<?> values;
    private final boolean acceptsList;
    private final boolean mayBeEmpty;
    private final String defaultValue;
    private final String description;
    private final String category;
    
    
    /**
     * Initializes the current property.
     * 
     * @param key The key of this property
     * @param values The class of the value of this property
     * @param acceptsList <code>true</code> if this property accepts a list
     * @param mayBeEmpty <code>true</code> if this property accepts an empty value
     * @param value The default value of this property
     * @param description This property's description
     * @param category This property's category
     */
    private CompilerProperty(final String key, final String category,
            final Class<?> values, final boolean acceptsList,
            final boolean mayBeEmpty, final String value, final String description) {
        this.key = key;
        this.values = values;
        this.acceptsList = acceptsList;
        this.mayBeEmpty = mayBeEmpty;
        this.defaultValue = value;
        this.description = description;
        this.category = category;
        CompilerProperty.addCompilerProperty(this);
    }
    
    
    /**
     * Adds the property to the static map that associates keys with properties
     * and checks for disjoint keys.
     *
     * @param p The property to add
     */
    private static void addCompilerProperty(final petruchio.interfaces.CompilerProperty p) {
        if(CompilerProperty.key2props == null) {
            CompilerProperty.key2props = new HashMap<String, petruchio.interfaces.CompilerProperty>();
        }
        final petruchio.interfaces.CompilerProperty old = CompilerProperty.key2props.put(p.getKey(), p);
        if(old != null) {
            CompilerProperty.key2props.put(p.getKey(), old);
            throw new IllegalArgumentException("A CompilerProperty with key '"
                    + p.getKey() + "' has already been created.");
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CompilerProperty#getFromKey(java.lang.String)
     */
    @Override
    public petruchio.interfaces.CompilerProperty getFromKey(final String k) {
        return CompilerProperty.key2props == null ? null : CompilerProperty.key2props.get(k);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CompilerProperty#getKey()
     */
    @Override
    public String getKey() {
        return this.key;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CompilerProperty#getValueClass()
     */
    @Override
    public Class<?> getValueClass() {
        return this.values;
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CompilerProperty#acceptsList()
     */
    @Override
    public boolean acceptsList() {
        return this.acceptsList;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CompilerProperty#mayBeEmpty()
     */
    @Override
    public boolean mayBeEmpty() {
        return this.mayBeEmpty;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CompilerProperty#getDefaultValue()
     */
    @Override
    public String getDefaultValue() {
        return this.defaultValue;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CompilerProperty#getDescription()
     */
    @Override
    public String getDescription() {
        return this.description;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.CompilerProperty#getCategory()
     */
    @Override
    public String getCategory() {
        return this.category;
    }
    
    
    /* (non-Javadoc)
     * @see java.lang.Enum#toString()
     */
    @Override
    public String toString() {
        return this.getKey();
    }
    
}
