/*
 * Project:  Petruchio
 * Filename: Pi2PN.java
 * Created:  01.12.2006
 * Author:   Tim Strazny
 */
package petruchio.compiler;

import java.io.File;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Deque;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Queue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.ErrorManager;
import java.util.logging.Filter;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.regex.Pattern;

import petruchio.Petruchio;
import petruchio.common.IdentityHashSet;
import petruchio.common.PropertyLoader;
import petruchio.compiler.ProcessData.NameCard;
import petruchio.compiler.reactions.CommunicationReactionTask;
import petruchio.compiler.reactions.ReactionTask;
import petruchio.compiler.reactions.SingleReactionTask;
import petruchio.interfaces.Compiler;
import petruchio.interfaces.LocalConfigurationParser;
import petruchio.interfaces.ParserErrorHandler;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.CommunicatorCache;
import petruchio.interfaces.pi.PiParser;
import petruchio.interfaces.algorithms.CommunicationReaction;
import petruchio.interfaces.algorithms.Coverability;
import petruchio.interfaces.algorithms.Reaction;
import petruchio.interfaces.algorithms.ReactionFinder;
import petruchio.interfaces.algorithms.RestrictedForm;
import petruchio.interfaces.algorithms.SingleReaction;
import petruchio.interfaces.algorithms.StructuralCongruence;
import petruchio.interfaces.algorithms.StructuralBoundedness;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetCovWriter;
import petruchio.interfaces.petrinet.PetriNetReader;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.PetriNetReductorIgnorer;
import petruchio.interfaces.petrinet.PetriNetReductorIgnorerInput;
import petruchio.interfaces.petrinet.PetriNetStreamReader;
import petruchio.interfaces.petrinet.PetriNetStreamWriter;
import petruchio.interfaces.petrinet.PetriNetWriter;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.PlacePair;
import petruchio.interfaces.petrinet.Transition;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.Guard;
import petruchio.interfaces.pi.InputAction;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.OutputAction;
import petruchio.interfaces.pi.Parameters;
import petruchio.interfaces.pi.PiWriter;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessDefinition;
import petruchio.interfaces.pi.ProcessID;
import petruchio.interfaces.pi.ProcessManager;
import petruchio.interfaces.pi.ProcessModifier;
import petruchio.interfaces.pi.ProcessReference;


/**
 * A compiler from pi calculus processes to Petri nets that supports Meyer's
 * structural and concurrency semantics, as well as Strazny's combinations of
 * these.
 *
 * @param <V> The data type of the processes.
 * @author Tim Strazny
 */
public class Pi2PN<V> implements SelfDescribing, Compiler {
    
    private static final char MAGIC = ' ';
    
    private static final String ALTERNATIVE_TRANSITION_SEP = " & ";
    
    private static final Collection<CompilerProperty> EXCLUDE_FOR_LOCAL_CONFIG = new ArrayList<CompilerProperty>();
    
    private static final Collection<CompilerProperty> CHANGES_SEMANTICS = new ArrayList<CompilerProperty>();
    
    static {
        if(Compiler.NAME_CHARS.indexOf(Pi2PN.MAGIC) >= 0) {
            throw new InternalError("I need a different magic character!");
        }
        
        Pi2PN.EXCLUDE_FOR_LOCAL_CONFIG.add(CompilerProperty.LOCAL_CONFIG_PARSER);
        Pi2PN.EXCLUDE_FOR_LOCAL_CONFIG.add(CompilerProperty.PARSER_ERROR_HANDLERS);
        Pi2PN.EXCLUDE_FOR_LOCAL_CONFIG.add(CompilerProperty.LOG_HANDLER);
        Pi2PN.EXCLUDE_FOR_LOCAL_CONFIG.add(CompilerProperty.LOG_ENCODING);
        Pi2PN.EXCLUDE_FOR_LOCAL_CONFIG.add(CompilerProperty.LOG_ERROR_MANAGER);
        Pi2PN.EXCLUDE_FOR_LOCAL_CONFIG.add(CompilerProperty.LOG_FILTER);
        Pi2PN.EXCLUDE_FOR_LOCAL_CONFIG.add(CompilerProperty.LOG_FORMATTER);
        Pi2PN.EXCLUDE_FOR_LOCAL_CONFIG.add(CompilerProperty.LOG_LEVEL);
        
        Pi2PN.CHANGES_SEMANTICS.add(CompilerProperty.IGNORE_COMMUNICATIONS);
        Pi2PN.CHANGES_SEMANTICS.add(CompilerProperty.REACHABLE_ONLY);
        Pi2PN.CHANGES_SEMANTICS.add(CompilerProperty.KEEP_GLOBAL_RESTRICTIONS);
        Pi2PN.CHANGES_SEMANTICS.add(CompilerProperty.MAXIMUM_REACTION_COUNT);
        Pi2PN.CHANGES_SEMANTICS.add(CompilerProperty.PROCESS_MODIFIERS);
    }
    
    
    /** Logger */
    private Logger logger;
    
    /** Compiler options */
    private PropertyLoader propertyLoader;
    
    /** ProcessCreator object to be used during compilation */
    private ProcessCreator pbuilder;
    
    /** ProcessManager object to be used during compilation */
    private ProcessManager<V> pmanager;
    
    /** Structural congruence checker to be used during compilation */
    private StructuralCongruence<V> structChecker;
    
    /** Restrictor to be used during compilation */
    private RestrictedForm<V> restrictor;
    
    /** ReactionFinder finder to be used during compilation */
    private ReactionFinder<V> reactionFinder;
    
    /** Coverability graph to check which fragments can communicate */
    private Coverability cov;
    
    private Collection<PetriNetWriter> pnWriters;
    
    private Collection<PiWriter<V>> piWriters;
    
    private Collection<Name> envProcs;
    
    private CommunicatorCache<V> communicators;
    
    private PetriNetReader pnReader;
    
    private PetriNetReductorIgnorer pnReductorIgnorer;
    
    private PetriNetReductorIgnorerInput pnReductorIgnorerInput;
    
    private Collection<PetriNetReductor> pnReductors;
    
    private Collection<ProcessModifier<V>> processModifiers;
    
    private LocalConfigurationParser localConfigParser;
    
    private PiParser parser;
    
    private Collection<StructuralBoundedness<V>> structBoundCheckers;
    
    private volatile Collection<ParserErrorHandler> parserErrorHandlers;
    
    /** The names that have been used in the processes of the net */
    private Collection<Name> namesUsed;
    
    private Collection<ProcessID> excludedForMarking;
    
    /** Communications to check */
    private BlockingQueue<PlacePair> comm;
    
    /** The id of the main process */
    private ProcessID mainProcessID;
    
    /** The maximal reaction count */
    private int maximumReactionCount;
    
    /** X distance in Petri net */
    private int offsetX;
    
    /** Y distance in Petri net */
    private int offsetY;
    
    /** The semantics to use */
    private Collection<Semantics> semantics;
    
    private PetriNet pNet;
    
    private Collection<Name> unboundNames;
    
    private volatile boolean pleaseStop;
    
    private long compilationStarted;
    
    /** */
    private boolean reachableOnly;
    
    /** */
    private boolean ignoreCommunications;
    
    /** */
    private boolean doLayout;
    
    private boolean markUnusedPlaces;
    
    /** */
    private boolean keepGlobalRestrictions;
    
    /** */
    private boolean useThreads;
    
    /** */
    private CompilerProperty.ExtendedBoolean computeBounds;
    
    private CompilerProperty.ExtendedBoolean checkDeadlocks;
    
    private boolean partialOrderReduction;
    
    private boolean repeatReductions;
    
    /** */
    private int patience;
    
    private int minimumTokens;
    private int maximumTokens;
    
    /** */
    private Thread covThread;
    
    
    /**
     * Initializes the new instance.
     */
    public Pi2PN() {
        this.logger = Logger.getLogger(System.getProperty("Logger", this.getClass().getName()));
        this.propertyLoader = new PropertyLoader();
        this.excludedForMarking = new IdentityHashSet<ProcessID>();
        this.namesUsed = new HashSet<Name>(0);
    	this.unboundNames = new HashSet<Name>(0);
    	this.comm = new LinkedBlockingQueue<PlacePair>();
    	this.envProcs = new HashSet<Name>(0);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "The default compiler.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Compiler#getName()
     */
    @Override
    public String getName() {
        return "Pi2PN";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Compiler#getOptions()
     */
    @Override
    public CompilerProperty[] getOptions() {
        return CompilerProperty.values();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Compiler#pleaseStop()
     */
    @Override
    public void pleaseStop() {
        this.pleaseStop = true;
        this.pleaseStopCov();
    }
    
    
    public void pleaseStopCov() {
        if(this.cov != null) {
            this.cov.pleaseTerminate();
        }
        if(!this.hasTerminated()) {
            this.covThread.interrupt();
        }
        
        final long waiting = System.currentTimeMillis();
        final int delay = Math.max(Math.min(this.patience/10, 1000), 100);
        try {
            while(!this.hasTerminated() && (System.currentTimeMillis() - waiting < this.patience)) {
                Thread.sleep(delay);
            }
        } catch(final InterruptedException e) {
            // nothing
        }
//        if(!this.hasTerminated()) {
//            this.stopForcefully();
//            this.logger.info("The compiler did not terminate in "
//                    + this.patience + " ms. It was stopped forcefully.");
//        }
        this.clearCaches();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Compiler#loadProperties(java.util.List)
     */
    @Override
    public synchronized void loadProperties(final List<Properties> props) {
    	Petruchio.checkValidOptionKeys(this, props);
    	this.propertyLoader.setProperties(props);
        
        this.loadLogger();
        
        this.namesUsed.clear();
        this.unboundNames.clear();
        this.comm.clear();

        this.localConfigParser = this.loadProperty(CompilerProperty.LOCAL_CONFIG_PARSER, this.localConfigParser);
        
        this.parserErrorHandlers = this.loadProperty(CompilerProperty.PARSER_ERROR_HANDLERS, this.parserErrorHandlers);
        for(final ParserErrorHandler h : this.parserErrorHandlers) {
            this.localConfigParser.addErrorHandler(h);
        }
        
    }
    
    
    /**
     * 
     *
     * @param props
     */
    private void loadCompilationProperties(final List<Properties> props) {
    	this.propertyLoader.setProperties(props);
    	this.loadCompilationProperties();
    }

    
    /**
     * 
     *
     */
    @SuppressWarnings("boxing")
    private void loadCompilationProperties() {
    	final int priority = this.readPriority((Integer)this.loadProperty(CompilerProperty.MAIN_PRIORITY));
        final int covPriority = this.readPriority((Integer)this.loadProperty(CompilerProperty.COVERABILITY_PRIORITY));
        Thread.currentThread().setPriority(priority);
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Thread priority set to " + Thread.currentThread().getPriority());
        }
        
        this.maximumReactionCount = (Integer)this.loadProperty(CompilerProperty.MAXIMUM_REACTION_COUNT);
        this.patience = (Integer)this.loadProperty(CompilerProperty.TERMINATION_PATIENCE);
        this.offsetX = (Integer)this.loadProperty(CompilerProperty.PN_OFFSET_X);
        this.offsetY = (Integer)this.loadProperty(CompilerProperty.PN_OFFSET_Y);
        this.reachableOnly = (Boolean)this.loadProperty(CompilerProperty.REACHABLE_ONLY);
        this.markUnusedPlaces = (Boolean)this.loadProperty(CompilerProperty.MARK_UNUSED_PLACES);
        this.ignoreCommunications = (Boolean)this.loadProperty(CompilerProperty.IGNORE_COMMUNICATIONS);
        this.minimumTokens = (Integer)this.loadProperty(CompilerProperty.MINIMUM_TOKEN_COUNT_FOR_ACCELERATION);
        this.maximumTokens = (Integer)this.loadProperty(CompilerProperty.MAXIMUM_TOKEN_COUNT_FOR_ACCELERATION);
        this.doLayout = !(Boolean)this.loadProperty(CompilerProperty.IGNORE_LAYOUT);
        this.keepGlobalRestrictions = (Boolean)this.loadProperty(CompilerProperty.KEEP_GLOBAL_RESTRICTIONS);
        this.useThreads = (Boolean)this.loadProperty(CompilerProperty.USE_THREADS);
        this.partialOrderReduction = (Boolean)this.loadProperty(CompilerProperty.PARTIAL_ORDER_REDUCTION);
        this.repeatReductions = (Boolean)this.loadProperty(CompilerProperty.REPEAT_REDUCTIONS);
        this.computeBounds = this.loadProperty(CompilerProperty.COMPUTE_BOUNDS);
        this.checkDeadlocks = this.loadProperty(CompilerProperty.CHECK_FOR_DEADLOCKS);
        
        this.pbuilder = this.loadProperty(CompilerProperty.PROCESS_CREATOR, this.pbuilder);
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Process creator has been loaded.");
        }
        
        this.pmanager = this.loadProperty(CompilerProperty.PROCESS_MANAGER, this.pmanager);
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Process manager has been loaded.");
        }
        
        this.structChecker = this.loadProperty(CompilerProperty.STRUCTURAL_CONGRUENCE, this.structChecker);
        this.structChecker.setProcessCreator(this.pbuilder);
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Structural congruence checker has been loaded.");
        }
        
        this.reactionFinder = this.loadProperty(CompilerProperty.REACTION_FINDER, this.reactionFinder);
        this.reactionFinder.setProcessCreator(this.pbuilder);
        this.reactionFinder.setProcessManager(this.pmanager);
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("ReactionFinder finder has been loaded.");
        }
        
        this.restrictor = this.loadProperty(CompilerProperty.RESTRICTED_FORM, this.restrictor);
        this.restrictor.setProcessCreator(this.pbuilder);
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Restricted form builder has been loaded.");
        }
        
        this.communicators = this.loadProperty(CompilerProperty.COMMUNICATOR_CACHE, this.communicators);
        this.communicators.setReactionFinder(this.reactionFinder);
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Communicator cache has been loaded.");
        }
        
        final Coverability oldCov = this.cov;
        this.cov = this.loadProperty(CompilerProperty.COVERABILITY, this.cov);
        
        if(!this.cov.supportsThreading()) {
            this.useThreads = false;
        }
        
        this.cov.setCheckMarkable(true);
        this.cov.setCommunications(this.comm);
        this.cov.setCommunicatorCache(this.communicators);
        final int availableProcessors = Runtime.getRuntime().availableProcessors() - (this.useThreads ? 1 : 0); // main process
        
        try {
            this.cov.setMinTokenCount(this.minimumTokens);
        } catch(final UnsupportedOperationException e) {
            // ignore
        }
        try {
            this.cov.setMaxTokenCount(this.maximumTokens);
        } catch(final UnsupportedOperationException e) {
            // ignore
        }
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Coverability checker has been loaded.");
        }
        
        this.structBoundCheckers = this.loadProperty(CompilerProperty.STRUCTURAL_BOUNDEDNESS_CHECKER, this.structBoundCheckers);
        for(final StructuralBoundedness<V> checker : this.structBoundCheckers) {
            checker.setReactionFinder(this.reactionFinder);
            checker.setStructuralCongruence(this.structChecker);
        }
        
        this.parser = this.loadProperty(CompilerProperty.PARSER, this.parser);
        if(this.parser != null) {
            this.parser.setProcessCreator(this.pbuilder);
            for(final ParserErrorHandler h : this.parserErrorHandlers) {
                this.parser.addErrorHandler(h);
            }
        }
        this.piWriters = this.loadProperty(CompilerProperty.PI_WRITERS, this.piWriters);
        
        final Collection<String> ep = this.loadProperty(CompilerProperty.ENVIRONMENT_PROCESSES);
        this.envProcs.clear();
        if(ep != null && !ep.isEmpty()) {
            for(final String s : ep) {
                this.envProcs.add(this.pbuilder.newName(s));
            }
        }
        
        this.semantics = this.loadProperty(CompilerProperty.SEMANTICS);
        this.semantics = EnumSet.copyOf(this.semantics);
        
        this.pnReader = this.loadProperty(CompilerProperty.PN_READER, this.pnReader);
        if(this.pnReader != null) {
            for(final ParserErrorHandler h : this.parserErrorHandlers) {
                this.pnReader.addErrorHandler(h);
            }
        }
        this.pnWriters = this.loadProperty(CompilerProperty.PN_WRITERS, this.pnWriters);
        this.pnReductors = this.loadProperty(CompilerProperty.PN_REDUCTORS, this.pnReductors);
        this.pnReductorIgnorer = this.loadProperty(CompilerProperty.PN_REDUCTOR_IGNORER, this.pnReductorIgnorer);
        this.processModifiers = this.loadProperty(CompilerProperty.PROCESS_MODIFIERS, this.processModifiers);
        for(final ProcessModifier<?> pm : this.processModifiers) {
            pm.setProcessCreator(this.pbuilder);
        }
        
        final String curPNReductorIgnorerInput = this.loadProperty(CompilerProperty.PN_REDUCTOR_IGNORER_INPUT);
        this.pnReductorIgnorerInput = this.loadProperty(CompilerProperty.PN_REDUCTOR_IGNORER_INPUT_TYPE, this.pnReductorIgnorerInput);
        if(this.pnReductorIgnorer != null) {
            this.pnReductorIgnorer.setInput(this.pnReductorIgnorerInput.getInput(curPNReductorIgnorerInput));
        }
        
        if((oldCov == this.cov) && !this.hasTerminated()) {
        	this.cov.pleaseStopComputation();
        	this.cov.awaitComputationEnd();
        	this.cov.clearCaches();
        } else {
	        if(this.covThread != null) {
	            this.pleaseStopCov();
	        }
	        this.cov.clearCaches();
	    	this.covThread = new Thread(this.cov, "Coverability Checker");
	    	this.covThread.start();
        }
        this.cov.setThreadNumber(availableProcessors);
        
    	this.covThread.setPriority(covPriority);
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Coverability checker thread started with priority "
                            + this.covThread.getPriority());
        }
        
        
        if(this.ignoreCommunications) {
            this.reachableOnly = false;
        }
        
        if(!this.reachableOnly) {
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Coverability checker deactivated.");
            }
            this.cov.setCheckMarkable(false);
            this.cov.setCommunications(null);
            
            if(this.computeBounds != CompilerProperty.ExtendedBoolean.FALSE) {
                if((this.computeBounds == CompilerProperty.ExtendedBoolean.TRUE)
                                && this.logger.isLoggable(Level.WARNING)) {
                    this.logger.warning("Will not compute bounds when coverability checker is deactivated.");
                }
                this.computeBounds = CompilerProperty.ExtendedBoolean.FALSE;
            }
            if(this.checkDeadlocks != CompilerProperty.ExtendedBoolean.FALSE) {
                if((this.checkDeadlocks == CompilerProperty.ExtendedBoolean.TRUE)
                                && this.logger.isLoggable(Level.WARNING)) {
                    this.logger.warning("Will not check for deadlocks when coverability checker is deactivated.");
                }
                this.checkDeadlocks = CompilerProperty.ExtendedBoolean.FALSE;
            }
        }
    }
    
    
    /**
     * Loads a property using the {@link petruchio.common.PropertyLoader}-object.
     * 
     * @param <T> The class of the value returned.
     * @param property The property to load.
     * @return The property's value.
     */
    @SuppressWarnings("unchecked")
    private <T> T loadProperty(final CompilerProperty property) {
        return (T)this.propertyLoader.loadProperty(property);
    }
    
    
    /**
     * 
     *
     * @param <T>
     * @param property
     * @param old
     * @return
     */
    private <T> T loadProperty(final CompilerProperty property, final T old) {
        return this.propertyLoader.loadProperty(property, old);
    }
    
    
    /**
     * Processes a given thread priority to be within then valid range.
     *
     * @param pri An integer
     * @return An integer in the range of valid thread priorities that is
     *            closest to the given integer.
     */
    private int readPriority(final int pri) {
        if(pri < Thread.MIN_PRIORITY) {
            this.logger.warning("Thread priority must be between "
                            + Thread.MIN_PRIORITY + " and "
                            + Thread.currentThread().getThreadGroup().getMaxPriority()
                            + ". " + pri + " is too low. Using minimum.");
            return Thread.MIN_PRIORITY;
        } else if (pri > Thread.currentThread().getThreadGroup().getMaxPriority()) {
            this.logger.warning("Thread priority must be between "
                            + Thread.MIN_PRIORITY + " and "
                            + Thread.currentThread().getThreadGroup().getMaxPriority()
                            + ". " + pri + " is too high. Using minimum.");
            return Thread.currentThread().getThreadGroup().getMaxPriority();
        } else {
            return pri;
        }
    }
    
    
    /**
     * Loads the user specified log handler and a logger.
     */
    private void loadLogger() {
        // create user specified log handler
        final Handler logHandler = this.loadProperty(CompilerProperty.LOG_HANDLER);
        
        final Boolean useParentLoggers = this.loadProperty(CompilerProperty.USE_PARENT_LOGGERS);
        this.logger.setUseParentHandlers(useParentLoggers.booleanValue());
        boolean handlerAdded = false;
        final Class<?> logHandlerClass = logHandler.getClass();
        for(final Handler h : this.logger.getHandlers()) {
        	if(h.getClass() == logHandlerClass) {
        		handlerAdded = true;
        		break;
        	}
        }
        if(!handlerAdded) {
        	this.logger.addHandler(logHandler);
        }

        final String encoding = this.loadProperty(CompilerProperty.LOG_ENCODING);
        try {
            logHandler.setEncoding((encoding == null || encoding.length() == 0) ? null : encoding);
        } catch(final UnsupportedEncodingException e) {
            try {
                logHandler.setEncoding(null);
            } catch(final UnsupportedEncodingException e2) {
                // ignore
            }
            this.logger.warning("Unable to use encoding " + encoding
                            + " for log handler "
                            + logHandler.getClass().getCanonicalName()
                            + ". Using default platform encoding instead.");
        }
        
        final Formatter formatter = this.loadProperty(CompilerProperty.LOG_FORMATTER, logHandler.getFormatter());
        logHandler.setFormatter(formatter);
        
        final ErrorManager errorManager = this.loadProperty(CompilerProperty.LOG_ERROR_MANAGER, logHandler.getErrorManager());
        logHandler.setErrorManager(errorManager);
        
        final Filter filter = this.loadProperty(CompilerProperty.LOG_FILTER, logHandler.getFilter());
        logHandler.setFilter(filter);
        
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Log handler and logger have been loaded.");
        }
        
        // user defined logger level
        final String loggerLevel = this.loadProperty(CompilerProperty.LOG_LEVEL);
        if(loggerLevel != null) {
            final Level level = Level.parse(loggerLevel);
            logHandler.setLevel(level);
            this.logger.setLevel(level);
            if(this.logger.isLoggable(Level.FINEST)) {
                this.logger.finest("Log level has been set to: " + level);
            }
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
    	this.pleaseStop();
    	this.clearCaches();

    	this.compilationStarted = 0;
    	this.pNet = null;
    	this.checkDeadlocks = null;
    	this.computeBounds = null;
    	this.cov = null;
    	this.covThread = null;
    	this.markUnusedPlaces = false;
    	this.ignoreCommunications = false;
    	this.doLayout = true;
    	this.keepGlobalRestrictions = true;
    	this.localConfigParser = null;
    	this.mainProcessID = null;
    	this.maximumReactionCount = -1;
    	this.offsetX = 0;
    	this.offsetY = 0;
    	this.minimumTokens = 1;
    	this.maximumTokens = -1;
    	this.parser = null;
    	this.parserErrorHandlers = null;
    	this.partialOrderReduction = true;
    	this.patience = 0;
    	this.pbuilder = null;
    	this.pmanager = null;
    	this.pnReductors = null;
    	this.processModifiers = null;
    	this.pnReductorIgnorer = null;
    	this.pnReductorIgnorerInput = null;
    	this.pnWriters = null;
    	this.reachableOnly = true;
    	this.reactionFinder = null;
    	this.repeatReductions = true;
    	this.restrictor = null;
    	this.semantics = null;
    	this.structBoundCheckers = null;
    	this.structChecker = null;
    	this.useThreads = true;
    	this.pleaseStop = false;
    }
    
    
    /**
     * 
     *
     */
    private void clearCaches() {
        this.smallReset();
        if(this.pbuilder != null) {
        	this.pbuilder.clearCaches();
        }
        if(this.pmanager != null) {
        	this.pmanager.clearCaches();
        }
        if(this.restrictor != null) {
        	this.restrictor.clearCaches();
        }
        if(this.reactionFinder != null) {
        	this.reactionFinder.clearCaches();
        }
        if(this.structChecker != null) {
        	this.structChecker.clearCaches();
        }
    }
    
    
    /**
     * Resets the compiler to compile the same input file with a different
     * semantics.
     */
    private void smallReset() {
    	if(this.cov != null) {
    		this.cov.clearCaches();
    	}
		this.comm.clear();
    	this.namesUsed.clear();
    	this.unboundNames.clear();
        this.excludedForMarking.clear();
        if(this.structBoundCheckers != null) {
	        for(final StructuralBoundedness<V> checker : this.structBoundCheckers) {
	            checker.clearCaches();
	        }
        }
    }
    
    
    /**
     * 
     *
     * @return
     */
    private boolean hasTerminated() {
        return (this.covThread == null) || (this.cov == null)
                || this.cov.hasTerminated() || !this.covThread.isAlive();
    }
    
//    
//    /**
//     * Stops the coverability checker thread forcefully.
//     */
//    private void stopForcefully() {
//        try {
//            this.covThread.stop();
//        } catch(final Exception e) {
//            this.logger.warning("An exception was thrown while stopping "
//                    + "forcefully:\n" + e);
//        }
//    }
//    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Compiler#compile(java.lang.String, java.lang.String)
     */
    @Override
    public synchronized void compile(final String inputFile, final String outputFile) {
        this.pleaseStop = false;
    	final List<Properties> originalProperties = this.propertyLoader.getProperties();
        try {
            this.clearCaches();
            
            // load processes
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Parsing file '" + inputFile + "'.");
            }
            this.parseLocalConfiguration(inputFile);
            this.parse(inputFile);
            this.loadMainProcessID();
            this.backupProcesses();
            
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Preprocessing parsed processes.");
            }
            final String preprocessMsg = this.preprocess(null);
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Preprocessing done: " + preprocessMsg);
            }
            
            this.writePiFiles(this.pmanager, outputFile);
            
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Parsing done: "
                            + this.pmanager.getProcessIDs().size()
                            + " processes");
            }
    
            this.getProcessesFromBackup();
            
        } catch(final Throwable e) {
            if(this.cov != null) {
            	this.cov.pleaseTerminate();
            }
            if(this.logger.isLoggable(Level.SEVERE)) {
                this.logger.severe("An error has occured!");
            }
            
            if(e instanceof RuntimeException) {
                throw (RuntimeException)e;
            } else if(e instanceof Error) {
                throw (Error)e;
            } else {
                throw new RuntimeException(e);
            }
        }
        
        if(!this.pnWriters.isEmpty()) {
            for(final Semantics sem : this.semantics) {
                if(this.pleaseStop) {
                    break;
                }
                final String suffix = (this.semantics.size() > 1)
                                                   ? "_" + sem.getSuffix()
                                                   : "";
                this.smallReset();
                this.getProcessesFromBackup();
                this.compile(sem, inputFile, outputFile + suffix);
            }
        }
        
        this.mainProcessID = null;
            
        this.propertyLoader.setProperties(originalProperties);
    }
    
    
    /**
     * TODO: document
     *
     */
    private void backupProcesses() {
        this.pmanager.backup();
    }
    
    
    /**
     * TODO: document
     *
     */
    private void getProcessesFromBackup() {
        this.pmanager.loadBackup();
    }
    
    
    /**
     * TODO: document
     *
     * @param sem        The semantics to use.
     * @param inputFile  The file with pi process definitions to read.
     *                   The file name has to be absolute.
     * @param outputFile The generated Petri net file to write.
     *                   The file name has to be absolute.
     */
    private void compile(final Semantics sem, final String inputFile, final String outputFile) {
        if(this.pleaseStop) {
            return;
        }
        
        if(sem.equals(Semantics.SEQ_STR) && !this.reachableOnly) {
            this.logger.warning("Cannot compile with " + sem + " and "
                            + CompilerProperty.REACHABLE_ONLY
                            + " set to false.");
            return;
        }
        
        if(sem.equals(Semantics.STR_SEQ) && (this.structBoundCheckers == null)) {
            this.logger.warning("Cannot compile with " + sem + " when no "
                            + "structural boundedness checker is available.");
            return;
        }
        
        if(this.logger.isLoggable(Level.FINE)) {
            this.logger.fine("Using " + sem.getName());
        }
        if(this.logger.isLoggable(Level.FINER)) {
            this.logger.finer("Preprocessing parsed processes.");
        }
        final String preprocessMsg = this.preprocess(sem);
        if(this.logger.isLoggable(Level.FINER)) {
            this.logger.finer("Preprocessing done.");
        }
              
        // petri net builder and writer
        try {
            this.pNet = this.loadProperty(CompilerProperty.PETRI_NET);
        } catch(final RuntimeException e) {
            this.pleaseStop();
            throw e;
        }
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Petri net class has been loaded.");
        }
        
        this.pNet.setName(inputFile.substring(inputFile.lastIndexOf(File.separator)+1, inputFile.length()));
        
        final StringBuilder header = new StringBuilder("-- generated by the petruchio tool (Tim Strazny 2007) --"
                + "\ncompiler " + this.getName() + " (" + this.getClass().getCanonicalName() + ")"
                + "\nfile " + outputFile
                + "\ncompiled from " + inputFile
                + "\nusing " + sem.getName() + " (" + sem.getAuthor() + ")"
                + "\nmain process: " + this.mainProcessID);
        
        for(final CompilerProperty critical : Pi2PN.CHANGES_SEMANTICS) {
        	final String value = this.propertyLoader.loadStringProperty(critical);
        	if(!critical.getDefaultValue().equals(value)) {
        		header.append("\n::");
        		header.append(critical.getKey());
        		header.append(" set to ");
        		header.append(value);
        		header.append("::");
        	}
        }
        header.append(preprocessMsg);
        header.append("\ncompilation startet " + (new java.util.Date()).toString());
        
        this.pNet.setHeaderComment(header.toString());
        
        ////////////////////////////////////////////////////////////////////////

        this.compilationStarted = System.currentTimeMillis();
        
        boolean constructionAborted = true;
        String msg = "\nTHE COMPILATION OF THIS NET HAS BEEN ABORTED"
                     + "\nTHIS NET HAS BEEN GENERATED PARTLY ONLY";
        
        if(this.partialOrderReduction && this.logger.isLoggable(Level.INFO)) {
            this.logger.info("Using partial order reduction on process level.");
        }
        
        try {
            this.translate(sem, this.pNet, this.maximumReactionCount);
            msg = "";
            constructionAborted = false;
            
        } catch(final Throwable e) {
            if(this.logger.isLoggable(Level.SEVERE)) {
                this.logger.severe("An error has occured! If this is a memory "
                        + "problem you might want to set "
                        + CompilerProperty.REACHABLE_ONLY
                        + " to false.");
            }
            
            if(e instanceof RuntimeException) {
                throw (RuntimeException)e;
            } else if(e instanceof Error) {
                throw (Error)e;
            } else {
                throw new RuntimeException(e);
            }
            
        } finally {
            this.clearCaches();
//            boolean keepCov = false;
//            if(this.cov != null) {
////                for(final PetriNetWriter w : this.pnWriters) {
////                    if(w instanceof PetriNetCovWriter) {
////                        keepCov = true;
////                        break;
////                    }
////                }
//                if(keepCov) {
//                    this.cov.pleaseTerminate();
//                    this.cov = null;
//                }
//            }
            
            // output the petri net
            this.pNet.setHeaderComment(this.pNet.getHeaderComment()
                            + msg
                            + "\ncompilation ended " + (new java.util.Date()).toString());
            
            if(this.doLayout) {
            	this.fixLayout(this.pNet);
            }
            
            if(this.logger.isLoggable(Level.FINE)) {
                this.logger.fine("Writing generated petri net.");
            }
            for(final Place p : this.pNet.getPlaces()) {
                p.setMeaning(p.getName() + petruchio.interfaces.Retriever.MEANING_SEP + p.getMeaning());
            }
            for(final Transition t : this.pNet.getTransitions()) {
                t.setMeaning(t.getName() + petruchio.interfaces.Retriever.MEANING_SEP + t.getMeaning());
            }
            this.writeNetFiles(this.pNet, outputFile, constructionAborted);
            if(this.logger.isLoggable(Level.FINE)) {
                this.logger.fine("Writing petri net done.");
            }
            
//            if(this.cov != null) {
//                this.cov.pleaseTerminate();
//                this.cov = null;
//            }
            this.pNet.reset();
        }
        
    }
    
    
    private void fixLayout(final PetriNet net) {
		final int addX = this.offsetX / 20;
		final int addY = this.offsetY / 20;
		Collection<Point> cords = new HashSet<Point>();
		int minX = Integer.MAX_VALUE;
        int minY = Integer.MAX_VALUE;
        for(final Place p : net.getPlaces()) {
            boolean done = false;
            do {
                final Point po = new Point(p.getX(), p.getY());
                if(!cords.add(po)) {
                    p.setX(p.getX()+addX);
                    p.setY(p.getY()+addY);
                } else {
                    done = true;
                }
            } while(!done);
            
            if(p.getX() < minX) {
                minX = p.getX();
            }
            if(p.getY() < minY) {
                minY = p.getY();
            }
        }
        for(final Transition t : net.getTransitions()) {
            boolean done = false;
            do {
                final Point po = new Point(t.getX(), t.getY());
                if(!cords.add(po)) {
                    t.setX(t.getX()+addX);
                    t.setY(t.getY()+addY);
                } else {
                    done = true;
                }
            } while(!done);
            
            if(t.getX() < minX) {
                minX = t.getX();
            }
            if(t.getY() < minY) {
                minY = t.getY();
            }
        }
        cords = null;
        
        minX -= this.offsetX;
        minY -= this.offsetY;
        
        for(final Transition t : net.getTransitions()) {
        	t.setX(t.getX()-minX);
        	t.setY(t.getY()-minY);
        }
        for(final Place p : net.getPlaces()) {
        	p.setX(p.getX()-minX);
        	p.setY(p.getY()-minY);
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param pm
     * @param outputFile
     */
    private void writePiFiles(final ProcessManager<V> pm, final String outputFile) {
        final long start = System.currentTimeMillis();
        for(final PiWriter<V> writer : this.piWriters) {
            final String file = outputFile + "."  + writer.getDefaultExtension();
            if(this.logger.isLoggable(Level.FINE)) {
                this.logger.fine("Writing net to: " + file);
            }
            OutputStream out = null;
            try {
                // create a convenient PrintWriter that's associated with the file
                out = new FileOutputStream(file);
                // now let the file be written
                writer.write("", "", pm, out);
                
            } catch(final FileNotFoundException fnfe) {
                // uh. trouble. file not writeable
                throw new RuntimeException("Unable to write to file: " + file, fnfe);
                
            } finally {
                try {
                    // always close streams after use.
                    if(out != null) {
                        out.close();
                    }
                } catch(final IOException e) {
                    // ignore
                }
            }
        }
        if(!this.piWriters.isEmpty() && this.logger.isLoggable(Level.INFO)) {
            this.logger.info("Wrote Pi files in " + Pi2PN.getDuration(start));
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param net
     * @param outputFile
     */
    private void writeNetFiles(final PetriNet net, final String outputFile,
            final boolean constructionAborted) {
    	final long start = System.currentTimeMillis();
    	
        for(final PetriNetWriter writer : this.pnWriters) {
            final String file = outputFile + "."  + writer.getDefaultExtension();
            if(this.logger.isLoggable(Level.FINE)) {
        		this.logger.fine("Writing net to: " + file);
            }
            OutputStream out = null;
            try {
                // create a convenient PrintWriter that's associated with the file
                out = new FileOutputStream(file);
                // now let the file be written
                if(!constructionAborted || !(writer instanceof PetriNetCovWriter)) {
                    writer.write(net, out);
                }
                
            } catch(final FileNotFoundException fnfe) {
                // uh. trouble. file not writeable
                throw new RuntimeException("Unable to write to file: " + file, fnfe);
                
            } finally {
                try {
                    // always close streams after use.
                    if(out != null) {
                        out.close();
                    }
                } catch(final IOException e) {
                    // ignore
                }
            }
        }
        if(!this.pnWriters.isEmpty() && this.logger.isLoggable(Level.INFO)) {
			this.logger.info("Wrote net in " + Pi2PN.getDuration(start));
        }
    }
    
    
    /**
     * TODO: document
     * 
     * @param name
     * @return
     */
    private boolean isMagicName(final Name name) {
        return name.toString().charAt(0) == Pi2PN.MAGIC;
    }
    
    
    /**
     * TODO: document
     * 
     * @param name
     * @return
     */
    private Name makeMagicName(final Name name) {
        return this.isMagicName(name) ? name : this.pbuilder.newName(Pi2PN.MAGIC + name.toString());
    }
    
    
    /**
     * @return
     */
    private boolean containsCommunicationViaFreeName() {
        final Collection<ProcessID> seen = new HashSet<ProcessID>();
        final Queue<ProcessID> todo = new LinkedList<ProcessID>();
        todo.add(this.mainProcessID);
        seen.add(this.mainProcessID);
        
        while(!todo.isEmpty()) {
            final ProcessID pid = todo.remove();
            
            final Process<V> p = this.pmanager.getProcess(pid);
            if(p == null) {
                continue;
            }
            
            if(this.reactionFinder.containsCommunicationViaFreeName(p)) {
                return true;
            }
            
            for(final ProcessID id : this.pmanager.references(p)) {
                if(seen.add(id)) {
                    todo.add(id);
                }
            }
        }
        
        return false;
    }
    
    
    private void waitForCovResults() {
        this.cov.awaitResults();
    }
    
    
    private void initializeTasks(final Collection<ReactionTask<V>> tasks, final boolean containsFreeCommunications,
            final Collection<SingleReactionTask<V>> structures) {
        for(final SingleReactionTask<V> a : structures) {
//        for(final Collection<SingleReactionTask<V>> col : structuresMap.values()) for(final SingleReactionTask<V> a : col) {
            tasks.add(a);
            final Collection<ActionPrefix> actions = containsFreeCommunications ? this.communicators.add(a) : null;
            
            this.createCommWithoutCov(a, actions, tasks, containsFreeCommunications);
        }
    }
    
    
    private void createCommWithoutCov(final SingleReactionTask<V> a, final Collection<ActionPrefix> actions,
            final Collection<ReactionTask<V>> tasks, final boolean containsFreeCommunications) {
        if(!this.reachableOnly && containsFreeCommunications && !this.ignoreCommunications) {
            final Collection<SingleReactionTask<V>> seen = new IdentityHashSet<SingleReactionTask<V>>();
            for(final ActionPrefix ap : actions) {
                final Collection<SingleReactionTask<V>> comms;
                if(ap instanceof InputAction) {
                    final InputAction in = (InputAction)ap;
                    comms = this.communicators.getSendersOn(new NameCard(in.getChannel(), in.getNames().size()));
                } else if(ap instanceof OutputAction) {
                    final OutputAction out = (OutputAction)ap;
                    comms = this.communicators.getReceiversOn(new NameCard(out.getChannel(), out.getNames().size()));
                } else {
                    comms = null;
                }
                if(comms != null) {
                    for(final SingleReactionTask<V> b : comms) {
                        if(seen.add(b)) {
                            final ReactionTask<V> task
                                    = new CommunicationReactionTask<V>(a.getProcess(),
                                            a.getPlace(), b.getProcess(), b.getPlace());
                            tasks.add(task);
                        }
                    }
                }
            }
        }
    }
    
    
    private int currentX;
    
    private CompilationState processReactionTask(final ReactionTask<V> task, final Semantics sem, CompilationState state,
            final List<SingleReactionTask<V>> structures, final Collection<ReactionTask<V>> tasks, final boolean containsFreeCommunications,
            final PetriNet net, final Map<Transition, Collection<Name>> nameCreators, final int maxReactionCount) {
        final List<Reaction<V>> reactions = new ArrayList<Reaction<V>>(2);
        final Map<Place, Integer> placesBefore = new IdentityHashMap<Place, Integer>(2);
        
        int placeX = 0;
        int placeY = 0;
        int transY = 0;
        
        switch(task.getType()) {
            case SINGLE:
                final SingleReactionTask<V> single = (SingleReactionTask<V>)task;
                placesBefore.put(single.getPlace(), Integer.valueOf(1));
                this.getReactions(single.getProcess(), reactions);
                
                if(this.doLayout) {
                        placeX = single.getPlace().getX();
                        placeY = single.getPlace().getY() + (this.offsetY << 1);
                        transY = placeY - this.offsetY;
                }
                break;
            case COMMUNICATION:
                final CommunicationReactionTask<V> comr = (CommunicationReactionTask<V>)task;
                if(comr.getPlaceA() == comr.getPlaceB()) {
                    placesBefore.put(comr.getPlaceA(), Integer.valueOf(2));
                } else {
                    placesBefore.put(comr.getPlaceA(), Integer.valueOf(1));
                    placesBefore.put(comr.getPlaceB(), Integer.valueOf(1));
                }
                this.getReactionsBetween(comr.getProcessA(), comr.getProcessB(), reactions);
                
                if(this.doLayout) {
                        placeX = (comr.getPlaceA().getX() + comr.getPlaceB().getX()) >> 1;
                        placeY = Math.max(comr.getPlaceA().getY(), comr.getPlaceB().getY()) + (this.offsetY << 1);
                        transY = placeY - this.offsetY;
                }
                break;
            default:
                throw new InternalError("Unknown reaction task: " + task);
        }
        
        int partialOrderReduced = -1;
        if(this.partialOrderReduction) {
            boolean hasToCheck = false;
            for(final Reaction<V> reaction : reactions) {
                if(reaction.isSequentiallyIndependent() == Reaction.Independent.TRUE) {
                    hasToCheck = true;
                    break;
                }
            }
            if(hasToCheck) {
                final Map<ProcessComposition<V>, Integer> choiceReactions = new IdentityHashMap<ProcessComposition<V>, Integer>();
                for(final Reaction<V> reaction : reactions) {
                    for(final ProcessComposition<V> c : reaction.getChoices()) {
                        final Integer j = choiceReactions.put(c, Integer.valueOf(1));
                        if(j != null) {
                            choiceReactions.put(c, Integer.valueOf(j.intValue()+1));
                        }
                    }
                }
                Collection<Reaction<V>> backup = null;
                for(int i = reactions.size()-1; i >= 0; --i) {
                    final Reaction<V> reaction = reactions.get(i);
                    if((reaction.isSequentiallyIndependent() != Reaction.Independent.TRUE)
                            || !Pi2PN.choiceSeqIndep(choiceReactions, reaction.getChoices())) {
                        reactions.remove(i);
                        ++partialOrderReduced;
                        this.partialOrderUsed = true;
                        if(backup == null) {
                            backup = new ArrayList<Reaction<V>>(1);
                        }
                        backup.add(reaction);
                    }
                }
                if(reactions.isEmpty() && backup != null) {
                    reactions.addAll(backup);
                }
            }
        }
        
        this.currentX = placeX;
        this.namesCreatedReactions = false;
        for(final Reaction<V> reaction : reactions) {
            if(this.pleaseStop) {
                break;
            }
            
            state = this.processReaction(reaction, task, state, sem, structures, net, tasks, containsFreeCommunications,
                    placesBefore, nameCreators, maxReactionCount, placeX, placeY, transY);
            
            if(state == CompilationState.ITERATE) {
                return state;
            }
            
        }
        this.checkForUnusedPlaces(reactions, this.communicators, structures, sem, this.namesCreatedReactions);
        
        if(partialOrderReduced > 0) {
            for(final Place p : placesBefore.keySet()) {
                p.setNote((p.getNote().length() == 0 ? "" : ", ")
                                + partialOrderReduced
                                + " reaction(s) ignored (partial order reduction)");
            }
        }
        
        return state;
    }
    
    private boolean namesCreatedReactions;
    
    
    private CompilationState processReaction(final Reaction<V> reaction, final ReactionTask<V> task, CompilationState state,
            final Semantics sem, final List<SingleReactionTask<V>> structures,
            final PetriNet net, final Collection<ReactionTask<V>> tasks, final boolean containsFreeCommunications,
            final Map<Place, Integer> placesBefore, final Map<Transition, Collection<Name>> nameCreators, final int maxReactionCount,
            final int placeX, final int placeY, final int transY) {
        final Map<Place, Integer> placesAfter = new IdentityHashMap<Place, Integer>(2);
        
        final Process<V> result = reaction.getReactedProcess();
        
        final Deque<Process<V>> decomposition = new LinkedList<Process<V>>();
        this.decompose(result, decomposition);

        boolean placeExisted = false;
        boolean exists = true;
//            boolean checkPlacesAfterSeq = false;
        Collection<Name> namesCreated = null;
        while(!decomposition.isEmpty()) {
            if(this.pleaseStop) {
                break;
            }
            
            Process<V> frag = decomposition.removeLast();
            if(this.reactionFinder.isBareNull(frag)) {
                // null processes are being ignored
                continue;
            }
            
            if(sem.equals(Semantics.SEQUENTIAL)) {
                if(!frag.getRestrictions().isEmpty()) {
                    if(namesCreated == null) {
                        namesCreated = new HashSet<Name>(0);
                    }
                    final List<Name> s = new ArrayList<Name>(0);
                    final List<Name> r = new ArrayList<Name>(0);
                    for(final Name n : frag.getRestrictions()) {
                        if(this.namesUsed.contains(n)) {
                            s.add(n);
                            final Name nName = this.reactionFinder.newName(n, this.namesUsed, this.reactionFinder.boundNames(frag));
                            r.add(nName);
                            this.namesUsed.add(nName);
                        }
                    }
                    if(!s.isEmpty()) {
                        frag = this.reactionFinder.replaceNames(frag, s, r);
                    }
                    namesCreated.addAll(frag.getRestrictions());
                    this.namesUsed.addAll(frag.getRestrictions());
                    frag = this.reactionFinder.copyTopLevel(frag);
                    frag.getRestrictions().clear();
                    this.decompose(frag, decomposition);
                    continue;
                }
//                this.namesUsed.addAll(this.reactionFinder.boundNames(frag));
                this.namesUsed.addAll(this.reactionFinder.freeNames(frag));
                
            } else if(sem.equals(Semantics.SEQ_STR) || sem.equals(Semantics.STR_SEQ)) {
                if(!frag.getRestrictions().isEmpty()) {
                    final Collection<Name> names = new ArrayList<Name>(frag.getRestrictions().size());
                    final List<Name> s = new ArrayList<Name>(0);
                    final List<Name> r = new ArrayList<Name>(0);
                    for(final Name name : frag.getRestrictions()) {
                        final boolean magic = this.isMagicName(name);
                        if((sem.equals(Semantics.SEQ_STR) && !magic)
                                || (sem.equals(Semantics.STR_SEQ) && magic)) {
                            if(namesCreated == null) {
                                namesCreated = new HashSet<Name>(0);
                            }
                            if(this.namesUsed.contains(name)) {
                                s.add(name);
                                final Name nName = this.reactionFinder.newName(name, this.namesUsed, this.reactionFinder.boundNames(frag));
                                r.add(nName);
                                namesCreated.add(nName);
                                this.namesUsed.add(nName);
                            } else {
                                namesCreated.add(name);
                                this.namesUsed.add(name);
                            }
                        } else {
                            names.add(name);
                        }
                    }
                    if(!s.isEmpty()) {
                        frag = this.reactionFinder.replaceNames(frag, s, r);
                    }
                    if(names.isEmpty()) {
                        frag = this.reactionFinder.copyTopLevel(frag);
                        frag.getRestrictions().clear();
                        this.decompose(frag, decomposition);
                        continue;
                    } else if(names.size() != frag.getRestrictions().size()){
                        frag = this.reactionFinder.copyTopLevel(frag);
                        frag.getRestrictions().clear();
                        frag.getRestrictions().addAll(names);
                    }
                }
//                this.namesUsed.addAll(this.reactionFinder.boundNames(frag));
                this.namesUsed.addAll(this.reactionFinder.freeNames(frag));
                
            }
            
            final int idx = this.binarySearch(structures, frag);
//            SingleReactionTask<V> p = this.find(structuresMap, frag);
            
            final boolean add = (idx < 0) // (p == null)
                || (!this.reachableOnly && ((sem.equals(Semantics.SEQUENTIAL) && !this.pmanager.createsNames(frag).isEmpty())
                                         || (sem.equals(Semantics.STR_SEQ) && this.containsMagicName(this.pmanager.createsNames(frag)))));
            
            final SingleReactionTask<V> p;
            if(add) {
                final Place place = net.addPlace(this.nextPlaceName(net), 0);
                place.setMeaning(frag.toString());
                if(((frag.getType() == Process.Type.REFERENCE)
                        && !this.pmanager.containsProcessID(((ProcessReference<V>)frag).getProcessID()))
                        || this.isEnvironmentProcess(frag)) {
                    place.setFinal(true);
                }
                this.communicators.associateWithProcess(place, frag);
                
                if(this.doLayout) {
                    place.setX(this.currentX);
                    place.setY(placeY);
                    this.currentX = this.currentX < placeX
                                     ? 2*placeX+this.offsetX-this.currentX
                                     : 2*placeX-this.offsetX-this.currentX;
                }
                
                p = this.add(idx, structures, frag, place);
                
                exists = false;
                
                final Collection<ActionPrefix> actions = containsFreeCommunications ? this.communicators.add(p) : null;
                
                tasks.add(p);
                this.createCommWithoutCov(p, actions, tasks, containsFreeCommunications);

                if(this.logger.isLoggable(Level.FINER)) {
                    this.logger.finer("Created place for fragment " + frag);
                }
            } else {
                p = structures.get(idx);
                placeExisted = true;
            }
            
            final Integer i = placesAfter.put(p.getPlace(), Integer.valueOf(1));
            if(i != null) {
                placesAfter.put(p.getPlace(), Integer.valueOf(i.intValue()+1));
            }
        }
        
        final Collection<Transition> between;
        if(exists && (between = net.getTransitionsBetween(placesBefore, placesAfter)) != null && !between.isEmpty()) {
            // the transitions "between" stand also for this reaction
            final String r = this.reactionToString(reaction);
            for(final Transition t : between) {
                    reaction.setTransition(t);
                t.setMeaning(t.getMeaning() + Pi2PN.ALTERNATIVE_TRANSITION_SEP + r);
                    
                if(namesCreated != null) {
                    final Collection<Name> c = nameCreators.get(t);
                    if(c == null) {
                        nameCreators.put(t, namesCreated);
                    } else {
                        c.addAll(namesCreated);
                    }
                }
            }
            
        } else {
            // create the new transition
            final Transition syncTrans = net.addTransition(this.nextTransitionName(net));
            syncTrans.setMeaning(this.reactionToString(reaction));
            reaction.setTransition(syncTrans);
            ++this.reactionCount;

            if(namesCreated != null) {
                nameCreators.put(syncTrans, namesCreated);
            }
            
            if(this.doLayout) {
                int bx = 0;
                for(final Map.Entry<Place, Integer> e : placesBefore.entrySet()) {
                    final Place place = e.getKey();
                    net.addArc(place, syncTrans, e.getValue().intValue());
                    bx += place.getX();
                }
                if(bx != 0) {
                    bx /= placesBefore.size();
                }
                
                int ax = 0;
                for(final Map.Entry<Place, Integer> e : placesAfter.entrySet()) {
                    final Place place = e.getKey();
                    net.addArc(syncTrans, place, e.getValue().intValue());
                    ax += place.getX();
                }
        
                if(ax != 0) {
                    ax /= placesAfter.size();
                }
                
                if(ax == 0 || bx == 0) {
                    ax = ax + bx;
                } else {
                    ax = (ax+bx) >> 1;
                }
                
                syncTrans.setX(ax);
                syncTrans.setY(transY);
            } else {
                for(final Map.Entry<Place, Integer> e : placesBefore.entrySet()) {
                    final Place place = e.getKey();
                    net.addArc(place, syncTrans, e.getValue().intValue());
                }
                
                for(final Map.Entry<Place, Integer> e : placesAfter.entrySet()) {
                    final Place place = e.getKey();
                    net.addArc(syncTrans, place, e.getValue().intValue());
                }
            }
            
            if(placesBefore.size() != 1) {
                ++this.communicationReactionCount;
            }
            
            if((state == CompilationState.RUNNING)
                    && (maxReactionCount >= 0)
                    && (this.reactionCount >= maxReactionCount)) {
                state = CompilationState.REACTION_COUNT_REACHED;
                if(this.reachableOnly && containsFreeCommunications) {
                    this.cov.pleaseStopComputation();
                }
            } else {
                if(this.reachableOnly && containsFreeCommunications) {
                    this.cov.addTransition(syncTrans);
                    
                    if(!this.useThreads) {
                        this.cov.awaitComputationEnd();
                        while(!this.comm.isEmpty()) {
                            final PlacePair pair = this.comm.remove();
                            final Process<V> p1 = this.communicators.getProcess(pair.getPlaceA());
                            final Process<V> p2 = this.communicators.getProcess(pair.getPlaceB());
                            tasks.add(new CommunicationReactionTask<V>(p1, pair.getPlaceA(),
                                            p2, pair.getPlaceB()));
                        }
                    }
                }
                
                Collection<Name> infiniteBreadthWith = null;
                if(sem.equals(Semantics.STRUCTURAL)
                        && (task.getType() == ReactionTask.Type.SINGLE)
                        && ((infiniteBreadthWith = this.causesStructuralUnboundedness(syncTrans, this.communicators, placeExisted)) != null)) {
                    if(this.logger.isLoggable(Level.INFO)) {
                        this.logger.log(Level.INFO, "Infinite breadth with "
                                + "names " + infiniteBreadthWith);
                    }
                    
                    state = CompilationState.INFINITE_NET;
                    
                } else if(sem.equals(Semantics.STR_SEQ)
                        && (task.getType() == ReactionTask.Type.SINGLE)) {
                    final Collection<Name> names = this.causesStructuralUnboundedness(syncTrans, this.communicators, placeExisted);
                    if(names != null) {
                        if(this.logger.isLoggable(Level.FINE)) {
                            this.logger.fine("Restriction on " + names
                                    + " can grow unboundedly in breadth. Starting next run.");
                        }
                        state = CompilationState.ITERATE;
                        this.prepareNextRun(sem, net, names);
                        return state;
                    }
                }
            }
            
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Created transition between\n"
                                + placesBefore + "\nand\n" + placesAfter);
            }
        }
        
        if(namesCreated != null) {
            this.namesCreatedReactions = true;
        }
        
        return state;
    }
    
    
    int reactionCount;
    int communicationReactionCount;
    boolean partialOrderUsed;
    
    private boolean translate(final Semantics sem, final PetriNet net,
                    final int maxReactionCount) {

        final long startTime = System.currentTimeMillis(); 
        CompilationState state = CompilationState.RUNNING;
        // how many reactions have been translated
        
        if(sem.equals(Semantics.SEQUENTIAL)) {
            this.partialOrderReduction = false;
        }
        
        final boolean containsFreeCommunications = this.containsCommunicationViaFreeName(); 
        
        translationLoop:
        do {
            
            final Process<V> system = this.getSystemProcess();
            
            this.reactionCount = 0;
            this.communicationReactionCount = 0;
            this.partialOrderUsed = false;
            
            // processes that are not structurally congruent
            final List<SingleReactionTask<V>> structures = new ArrayList<SingleReactionTask<V>>();
//            final Map<V, Collection<SingleReactionTask<V>>> structuresMap = new HashMap<V, Collection<SingleReactionTask<V>>>();
//            final Map<Place, Process<V>> place2proc = Collections.synchronizedMap(new IdentityHashMap<Place, Process<V>>());
            
            final Deque<ReactionTask<V>> tasks = new LinkedList<ReactionTask<V>>();
            
//            if(containsFreeCommunications) {
                this.communicators.reset();
//            }
            
            // initialize the net: one place per fragment
            final Collection<Place> startState = new ArrayList<Place>();
            this.initializeNet(sem, startState, system, this.communicators, structures, net);
//            this.initializeNet(sem, startState, system, place2proc, structuresMap, net);
            
            if(containsFreeCommunications) {
                this.communicators.addAll(structures);
            }
            
            if(this.reachableOnly && containsFreeCommunications) {
                this.cov.setCheckMarkable(true);
                this.cov.setStartState(startState);
            }

            this.initializeTasks(tasks, containsFreeCommunications, structures);
            
    //        final Collection<Place> placesAfterSeq = new ArrayList<Place>();
            final Map<Transition, Collection<Name>> nameCreators = new IdentityHashMap<Transition, Collection<Name>>(0);
            
            state = CompilationState.RUNNING;
            
            // as long as there are places to check for "sequential" reactions:
            // handle those
            while((state == CompilationState.RUNNING)
                            && (!tasks.isEmpty() || !this.comm.isEmpty() || this.cov.isWorking())) {
                
                if(this.pleaseStop) {
                    state = CompilationState.STOPPED;
                    break;
                }
                
                if(this.useThreads && this.reachableOnly && containsFreeCommunications && tasks.isEmpty()) {
                    this.waitForCovResults();
                }
                
                this.collectCommunicationTasks(tasks);
                
                if(tasks.isEmpty()) {
                    continue;
                }
                
                state = this.processReactionTask(tasks.removeLast(), sem, state, structures, tasks, containsFreeCommunications, net, nameCreators, maxReactionCount);
                
                if(tasks.isEmpty()) {
                    this.cov.compilerIsWaiting();
                    if(!this.useThreads) {
                        this.cov.awaitComputationEnd();
                    }
                }

                if(state == CompilationState.ITERATE) {
                    continue translationLoop;
                }
                
                if(this.pleaseStop) {
                    state = CompilationState.STOPPED;
                    break;
                }
            }
            
            if(this.pleaseStop) {
                state = CompilationState.STOPPED;
                break translationLoop;
            }
            
            if(this.reachableOnly && containsFreeCommunications && this.useThreads && this.hasTerminated()) {
                if(state != CompilationState.STOPPED) {
                    state = CompilationState.COVERABILITY_DOWN;
                }
            } else {
                state = this.handleSequential(sem, state, net, nameCreators, startState);
                if(state == CompilationState.ITERATE) {
                    continue translationLoop;
                }
            
                if((state != CompilationState.ITERATE) && this.reachableOnly
                		&& !sem.equals(Semantics.STRUCTURAL)) {
                    this.removeDeadParts(net);
                }
            }
            
        } while(state == CompilationState.ITERATE);
        
        
        if(this.partialOrderUsed) {
            net.setHeaderComment(net.getHeaderComment()
                    + "\npartial order reduction has been used during compilation.");
        }
        
        if(state == CompilationState.RUNNING || state == CompilationState.REACTION_COUNT_REACHED) {
            
            boolean knownBounded;
            if((this.computeBounds == CompilerProperty.ExtendedBoolean.TRUE)
                    || (this.computeBounds == CompilerProperty.ExtendedBoolean.TRUE_IF_FOR_FREE)) {
    //            if(!covUpToDate) {
    //                this.updateCov(net);
    //                covUpToDate = true;
    //            }
                try {
                    final Map<Place, Integer> bounds = this.cov.getBounds(net.getPlaces());
                    knownBounded = this.writeBoundedness(net, bounds, bounds.get(null)) >= 0;
                } catch(final UnsupportedOperationException e) {
                    if((this.computeBounds != CompilerProperty.ExtendedBoolean.TRUE_IF_FOR_FREE)
                            && this.logger.isLoggable(Level.WARNING)) {
                        this.logger.warning("Unable to compute boundedness for "
                                        + "places. Current coverability checker "
                                        + "does not support this operation.");
                    }
                    knownBounded = false;
                }
            } else {
                knownBounded = false;
            }
            
            boolean covUpToDate = !this.reduce(net);
            
            if((this.checkDeadlocks == CompilerProperty.ExtendedBoolean.TRUE)
                    || (covUpToDate && this.checkDeadlocks == CompilerProperty.ExtendedBoolean.TRUE_IF_FOR_FREE)) {
                if(!covUpToDate) {
                    this.updateCov(net);
                    covUpToDate = true;
                }
                try {
                    final Collection<String> deadlocks = this.cov.getDeadlocks(knownBounded);
                    if(deadlocks == null) {
                        net.setHeaderComment(net.getHeaderComment()
                                        + "\ndeadlocks: no");
                    } else if(deadlocks.isEmpty()) {
                        net.setHeaderComment(net.getHeaderComment()
                                    + "\ndeadlocks: don't know");
                    } else {
                        final StringBuilder b = new StringBuilder();
                        for(final String lock : deadlocks) {
                            b.append("\n  ");
                            b.append(lock);
                        }
                        if(knownBounded) {
                            net.setHeaderComment(net.getHeaderComment()
                                    + "\ndeadlocks: yes, the following" + b.toString());
                        } else {
                            net.setHeaderComment(net.getHeaderComment()
                                    + "\ndeadlocks: yes, and there may be more" + b.toString());
                        }
                    }
                } catch(final UnsupportedOperationException e) {
                    if((this.checkDeadlocks != CompilerProperty.ExtendedBoolean.TRUE_IF_FOR_FREE)
                            && this.logger.isLoggable(Level.WARNING)) {
                        this.logger.warning("Unable to compute possible deadlocks. "
                                        + "Current coverability checker "
                                        + "does not support this operation.");
                    }
                }
            }
        }
        

        if(state == CompilationState.RUNNING) {
            state = CompilationState.DONE;
        }
        
        if(this.reachableOnly && containsFreeCommunications) {
            net.setHeaderComment(net.getHeaderComment() + "\nlast coverability "
                            + "checker size: " + this.cov.size());
        }
        
        if(this.reachableOnly && containsFreeCommunications && this.useThreads && this.hasTerminated()) {
            if(state != CompilationState.STOPPED) {
                state = CompilationState.COVERABILITY_DOWN;
            }
        } else {
            // wait for cov
            this.cov.pleaseStopComputation();
        }
        
        switch(state) {
            case RUNNING:
                throw new InternalError("Unexpected compiler state!");
            case DONE:
                if(this.logger.isLoggable(Level.INFO)) {
                    this.logger.info("Translated " + this.reactionCount
                                    + " reactions (" +  this.communicationReactionCount
                                    + " com via fn) with " + sem.getName() + " in"
                                    + " " + Pi2PN.getDuration(startTime) + ".");
                }
                break;
            case STOPPED:
                net.setHeaderComment(net.getHeaderComment()
                        + "\ncompilation ended prematurely as the compilation "
                        + "process has been stopped.");
                if(this.logger.isLoggable(Level.INFO)) {
                    this.logger.info("Compilation process has been stopped.");
                }
                break;
            case REACTION_COUNT_REACHED:
                net.setHeaderComment(net.getHeaderComment()
                        + "\ncompilation ended prematurely as a maximum "
                        + "reaction count of " + maxReactionCount
                        + " has been reached");
                if(this.logger.isLoggable(Level.INFO)) {
                    this.logger.info("Maximum reaction count (" + maxReactionCount
                                    + ") has been reached.\nTranslation stopped.");
                }
                break;
            case INFINITE_NET:
                if(this.logger.isLoggable(Level.INFO)) {
                    this.logger.info("The petri net would be "
                            + "infinite with " + sem.getName());
                }
                net.setHeaderComment(net.getHeaderComment()
                        + "\ncompilation ended prematurely as "
                        + "the semantics would lead to an infinite net");
                break;
            case COVERABILITY_DOWN:
                if(this.logger.isLoggable(Level.WARNING)) {
                    this.logger.warning("The coverability checker stopped prematurely.");
                }
                net.setHeaderComment(net.getHeaderComment()
                        + "\ncompilation ended prematurely as "
                        + "the coverability checker caused an error");
                break;
            default:
                throw new InternalError("Unknown compiler state: " + state);
        }
        
        if(this.reachableOnly && containsFreeCommunications && this.logger.isLoggable(Level.FINER)) {
            this.logger.finer("Coverability checker size: " + this.cov.size());
        }

        return state == CompilationState.DONE;
    }
    
    
    private void collectCommunicationTasks(final Collection<ReactionTask<V>> tasks) {
        while(!this.comm.isEmpty()) {
            final PlacePair pair = this.comm.remove();
            final Process<V> p1 = this.communicators.getProcess(pair.getPlaceA());
            if(this.reactionFinder.freeNames(p1).isEmpty()) {
                continue;
            }
            final Process<V> p2 = this.communicators.getProcess(pair.getPlaceB());
            if(this.reactionFinder.freeNames(p2).isEmpty()) {
                continue;
            }
            tasks.add(new CommunicationReactionTask<V>(p1, pair.getPlaceA(),
                            p2, pair.getPlaceB()));
        }
    }
    
    
    private CompilationState handleSequential(final Semantics sem, final CompilationState state,
            final PetriNet net, final Map<Transition, Collection<Name>> nameCreators,
            final Collection<Place> startState) {
        if(this.reachableOnly && (sem.equals(Semantics.SEQUENTIAL) || sem.equals(Semantics.SEQ_STR)
                || sem.equals(Semantics.STR_SEQ))
        && !nameCreators.isEmpty() && (state == CompilationState.RUNNING)) {
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Built folded net. Waiting for coverability checker.");
            }
            
            this.cov.awaitComputationEnd();
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Computing firing bounds of name creating transitions " + nameCreators.keySet());
            }
            
            Transition unboundedNameTrans;
            boolean containsBoundedNames;
            do {
                this.unboundNames.clear();
                unboundedNameTrans = null;
                containsBoundedNames = false;
                final Map<Transition, Integer> multipliers = this.cov.getFiringBounds(nameCreators.keySet());
                for(final Map.Entry<Transition, Collection<Name>> e : nameCreators.entrySet()) {
                    final Transition trans = e.getKey();
                    final Integer multiplier = multipliers.get(trans);
                    if(multiplier == null) {
                        throw new InternalError("The coverability checker did not "
                                + "compute the firing bounds for transition " + trans);
                    } else if(multiplier.intValue() < 0) {
                        unboundedNameTrans = trans;
                        if(sem.equals(Semantics.SEQ_STR)) {
                            this.unboundNames.addAll(e.getValue());
                        }
                    } else if(multiplier.intValue() > 1) {
                        containsBoundedNames = true;
                        if(this.logger.isLoggable(Level.FINER)) {
                            this.logger.finer("Name creating transition " + trans
                                            + " can fire " + multiplier + " times.");
                        }
                    }
                }
                if(containsBoundedNames) {
                    this.unfoldSequential(net, startState, this.communicators, nameCreators, multipliers);
                    this.updateCov(net);
                }
            } while(containsBoundedNames);
            
            if(sem.equals(Semantics.SEQ_STR) && !this.unboundNames.isEmpty()) {
                if(this.logger.isLoggable(Level.FINE)) {
                    this.logger.fine("Names " + this.unboundNames
                            + " can be created unboundedly. Starting next run.");
                }
                this.prepareNextRun(sem, net, this.unboundNames);
                return CompilationState.ITERATE;
            }
            
            if((state == CompilationState.RUNNING) && (unboundedNameTrans != null)) {
                if(this.logger.isLoggable(Level.INFO)) {
                    this.logger.info("Name creating transition " + unboundedNameTrans
                                    + " can fire unboundedly. Names "
                                    + nameCreators.get(unboundedNameTrans)
                                    + " would be created unboundedly often.");
                }
                return CompilationState.INFINITE_NET;
            }
        }
        return state;
    }
    
    
    /**
     * TODO: document
     *
     * @param net
     */
    private void updateCov(final PetriNet net) {
        final Collection<Place> startState = new ArrayList<Place>();
        for(final Place p : net.getPlaces()) {
            for(int i = p.getMarking(); i > 0; --i) {
                startState.add(p);
            }
        }
        this.cov.setCheckMarkable(false);
        this.cov.compute(startState, net.getTransitions());
    }
    
    
    /**
     * TODO: Document
     *
     * @param p
     * @return
     */
    private boolean isEnvironmentProcess(final Process<V> p) {
        final Collection<Name> ids = this.reactionFinder.getProcessIDs(p);
        return (!ids.isEmpty() && this.envProcs.containsAll(ids))
            || ((p.getType() == Process.Type.REFERENCE)
                    && (this.reactionFinder.getNullProcessCount(p) == 0)
                    && this.envProcs.contains(((ProcessReference<V>)p).getProcessID().getID()));
    }
    
    
    /**
     * 
     *
     * @param t
     * @param place2proc
     * @param placeExisted
     * @return
     */
    private Collection<Name> causesStructuralUnboundedness(final Transition t,
    		final CommunicatorCache<V> place2proc, final boolean placeExisted) {
    	Collection<Name> names = null;
    	for(final StructuralBoundedness<V> checker : this.structBoundCheckers) {
    		final Collection<Name> curNames = checker.causesStructuralUnboundedness(
    				t, place2proc.placeProcMap(), placeExisted);
    		if(curNames != null) {
    			if(names == null) {
    				names = new HashSet<Name>(0);
    			}
    			names.addAll(curNames);
    		}
    	}
    	return names;
    }
    
    
    /**
     * TODO: document
     *
     * @param net
     */
    private void removeDeadParts(final PetriNet net) {
        try {
        	this.cov.awaitComputationEnd();
            final Collection<Place> places = new IdentityHashSet<Place>(net.getPlaces());
            final Collection<Transition> trans = new IdentityHashSet<Transition>(net.getTransitions());
            this.cov.keepDeadElements(places, trans);
            if(this.logger.isLoggable(Level.FINER)) {
            	this.logger.finer("Removing dead elements\nUnmarkable places: "
            			+ places + "\nDead transitions: " + trans);
            }
            for(final Place p : places) {
                net.removePlace(p);
            }
            for(final Transition t : trans) {
                net.removeTransition(t);
            }
        } catch(final UnsupportedOperationException e) {
            if(this.logger.isLoggable(Level.FINE)) {
                this.logger.warning("Unable to compute dead net elements. "
                                + "Current coverability checker "
                                + "does not support this operation.");
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param sem
     * @param net
     * @param names
     */
    private void prepareNextRun(final Semantics sem, final PetriNet net,
            final Collection<Name> names) {

        final List<Name> search = new ArrayList<Name>(names);
        this.smallReset();
        
        this.makeMagic(search);
        
        final String comment = net.getHeaderComment();
        net.reset();
        net.setHeaderComment(comment);
    }
    
    
    /**
     * TODO: document
     *
     * @param m
     * @param c
     * @return
     */
    private static boolean choiceSeqIndep(final Map<?, Integer> m, final Collection<?> c) {
        for(final Object o : c) {
            Integer cur = m.get(o);
            if((cur != null) && (cur.intValue() > 1)) {
                return false;
            }
        }
        return true;
    }
    
    
    /**
     * TODO: document
     *
     * @param search
     */
    private void makeMagic(final List<Name> search) {
        final List<Name> replace = new ArrayList<Name>(search.size());
        for(final Name n : search) {
            final Name magic = this.makeMagicName(n);
            replace.add(magic);
            
        }
        for(final ProcessID pid : this.pmanager.getProcessIDs()) {
            final Process<V> proc = this.pmanager.getProcess(pid);
            final Process<V> nproc = this.reactionFinder.replaceNames(proc, search, replace);
            this.pmanager.updateProcess(pid, nproc);
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param net
     * @param startState
     * @param place2proc
     * @param nameCreators
     * @param multipliers
     */
    private void unfoldSequential(final PetriNet net,
            final Collection<Place> startState,
            final CommunicatorCache<V> place2proc,
            final Map<Transition, Collection<Name>> nameCreators,
            final Map<Transition, Integer> multipliers) {
        final Map<Transition, Collection<Transition>> copies = new IdentityHashMap<Transition, Collection<Transition>>();
        final Map<Transition, Transition> cpy = new IdentityHashMap<Transition, Transition>();
        for(final Transition t : multipliers.keySet()) {
            final Collection<Transition> c = new ArrayList<Transition>();
            c.add(t);
            copies.put(t, c);
            cpy.put(t, t);
        }
        final Deque<Transition> todo = new LinkedList<Transition>();
        todo.addAll(multipliers.keySet());
        while(!todo.isEmpty()) {
            final Transition trans = todo.removeLast();
            final int multiplier = multipliers.get(trans).intValue();
            if(multiplier > 1) {
                final Collection<Name> names = nameCreators.get(trans);
                this.duplicate(net, multiplier-1, nameCreators, copies, cpy, multipliers, todo, place2proc, trans, names);
            }
        }
        for(final Map.Entry<Transition, Integer> e : multipliers.entrySet()) {
            final Transition trans = e.getKey();
            final int multiplier = e.getValue().intValue();
            final Collection<Transition> theCopies = copies.get(trans);
            
            if(multiplier <= 1) {
                continue;
            }
                
            for(final Transition t : theCopies) {
                final StringBuilder b = new StringBuilder();
                b.append("new ");
                final Collection<Name> names = nameCreators.get(t);
                for(final Iterator<Name> it2 = names.iterator(); it2.hasNext();) {
                    final Name name = it2.next();
                    b.append(name);
                    if(it2.hasNext()) {
                        b.append(", ");
                    }
                }
                
                final Place bound = net.addPlace(this.nextPlaceName(net), 1);
                bound.setMeaning("(" + b.toString() + ")");
                startState.add(bound);
                if(this.doLayout) {
	                bound.setX(trans.getX());
	                bound.setY(trans.getY()-(this.offsetY >> 1));
                }
                bound.setBound(1);
                bound.setCouldCommunicate(false);
                net.addArc(bound, t, 1);
            }
        }
    }
    
    
    /**
     * TODO: document
     * 
     * @param net
     * @param maxTokenCount
     * @return
     */
    private int writeBoundedness(final PetriNet net, final Map<Place, Integer> bounds, final Integer maxTokenCount) {
        int max = 0;
        boolean unknown = false;
        for(final Map.Entry<Place, Integer> e : bounds.entrySet()) {
            final Place p = e.getKey();
            if(p == null) {
                continue;
            }
            final int b = e.getValue().intValue();
            if(b == Place.BOUND_UNBOUNDED) {
                max = -1;
                break;
            } else if(b == Place.BOUND_UNKNOWN) {
                unknown = true;
                break;
            } else if(b > max) {
                max = b;
            }
        }
        if(unknown) {
            // no output
        } else if(max >= 0) {
            net.setHeaderComment(net.getHeaderComment() + "\nbounded: " + max);
        } else if(max < 0) {
            net.setHeaderComment(net.getHeaderComment() + "\nbounded: no");
        }
        if((maxTokenCount != null) && (maxTokenCount.intValue() > 0)) {
            net.setHeaderComment(net.getHeaderComment()
                            + "\nmaximum token count: " + maxTokenCount);
        }
        return max;
    }

    
    /**
     * TODO: document
     *
     * @param net
     * @param multiplier
     * @param nameCreators
     * @param copies
     * @param cpy
     * @param multipliers
     * @param todo
     * @param place2proc
     * @param start
     * @param names
     */
    private void duplicate(final PetriNet net, final int multiplier,
            final Map<Transition, Collection<Name>> nameCreators,
            final Map<Transition, Collection<Transition>> copies,
            final Map<Transition, Transition> cpy,
            final Map<Transition, Integer> multipliers,
            final Collection<Transition> todo,
            final CommunicatorCache<V> place2proc,
            final Transition start, final Collection<Name> names) {
        
        final int[] x = new int[multiplier];
        final List<Name> oldNames = new ArrayList<Name>(names);
        @SuppressWarnings("unchecked")
        final List<Name>[] newNames = new List[multiplier];
        final List<Pattern> patterns = new ArrayList<Pattern>();
        @SuppressWarnings("unchecked")
        final Map<Place, Place>[] places = new Map[multiplier];
        @SuppressWarnings("unchecked")
        final Map<Transition, Transition>[] transitions = new Map[multiplier];
        
        for(int k = 0; k < multiplier; ++k) {
            newNames[k] = new ArrayList<Name>(names.size());
            for(final Name name : oldNames) {
                if(k == 0) {
                    patterns.add(Pattern.compile("(?<![" + Compiler.NAME_CHARS
                                    + "])" + name + "(?![" + Compiler.NAME_CHARS
                                    + "])"));
                }
                
                final Name newName = this.reactionFinder.newName(name, this.namesUsed, null);
                newNames[k].add(newName);
                this.namesUsed.add(newName);
            }
            
            transitions[k] = new IdentityHashMap<Transition, Transition>();
            places[k] = new IdentityHashMap<Place, Place>();
            
            final Transition copy = net.addTransition(this.nextTransitionName(net));
            copy.setMeaning(start.getMeaning());
            
            final Transition theRealOne = cpy.get(start);
            cpy.put(copy, theRealOne);
            copies.get(theRealOne).add(copy);
            transitions[k].put(start, copy);
            nameCreators.put(copy, newNames[k]);
            
            // set up transition
            if(this.doLayout) {
            	x[k] = (k+1)*(this.offsetX >> 1);
	            copy.setX(start.getX() + x[k]);
	            copy.setY(start.getY());
            }
        }
        
        final Map<Place, Collection<Transition>> ptArcs = new IdentityHashMap<Place, Collection<Transition>>();
        final Map<Place, Collection<Transition>> tpArcs = new IdentityHashMap<Place, Collection<Transition>>();
        final Deque<Transition> original = new LinkedList<Transition>();
        original.add(start);
        
        while(!original.isEmpty()) {
            final Transition originalT = original.removeLast();
            if(start != originalT) {
                // duplicate map entries if it's a transition that has to be multiplied
                final Collection<Name> moreNames = nameCreators.get(originalT);
                if(moreNames != null) {
                    final Collection<Name> copyNames = new ArrayList<Name>(moreNames.size());
                    for(final Name name : moreNames) {
                        final Name newName = this.reactionFinder.newName(name, this.namesUsed, null);
                        oldNames.add(name);
                        copyNames.add(newName);
                        patterns.add(Pattern.compile("(?<![" + Compiler.NAME_CHARS
                                        + "])" + name + "(?![" + Compiler.NAME_CHARS
                                        + "])"));
                        for(int k = 0; k < multiplier; ++k) {
                            newNames[k].add(newName);
                            this.namesUsed.add(newName);
                        }
                    }
                
                    final Transition theRealOne = cpy.get(originalT);
                    final Collection<Transition> copyV = copies.get(theRealOne);
                    final Integer i = multipliers.get(originalT);
                    for(int k = 0; k < multiplier; ++k) {
                        final Transition cloneT = transitions[k].get(originalT);
                        nameCreators.put(cloneT, copyNames);
                        copyV.add(cloneT);
                        cpy.put(cloneT, theRealOne);
                        multipliers.put(cloneT, i);
                        todo.add(cloneT);
                    }
                }
            }
            
            for(final PTArc arc : originalT.getInput()) {
                final Place originalP = arc.getSource();
                if(!places[0].containsKey(originalP)) {
                    Collection<Transition> arcs = ptArcs.get(originalP);
                    if(arcs == null) {
                        arcs = new ArrayList<Transition>();
                        ptArcs.put(originalP, arcs);
                    }
                    for(int i = 0; i < arc.getWeight(); ++i) {
                        arcs.add(originalT);
                    }
                }
            }
            
            for(final TPArc arc : originalT.getOutput()) {
                final Place originalP = arc.getTarget();
                Place cloneP = places[0].get(originalP);
                if(cloneP != null) {
                    for(int k = 0; k < multiplier; ++k) {
                        cloneP = places[k].get(originalP);
                        final Transition cloneT = transitions[k].get(originalT);
                        net.addArc(cloneT, cloneP, arc.getWeight());
                    }
                } else {
                    final Process<V> proc = place2proc.getProcess(originalP);
                    if(proc == null) {
                        throw new InternalError("" + originalP);
                    }
                    if(this.containsAFreeName(proc, oldNames)) {
                        originalP.setBound(1);
                        for(int k = 0; k < multiplier; ++k) {
                            final Transition cloneT = transitions[k].get(originalT);
                            final Process<V> newProc = this.reactionFinder.replaceNames(proc, oldNames, newNames[k]);
                            cloneP = net.addPlace(this.nextPlaceName(net), 0);
                            cloneP.setMeaning(newProc.toString());
                            if(this.doLayout) {
	                            cloneP.setX(originalP.getX() + x[k]);
	                            cloneP.setY(originalP.getY());
                            }
                            cloneP.setBound(originalP.getBound());
                            cloneP.setCouldCommunicate(originalP.couldCommunicate());
                            places[k].put(originalP, cloneP);
                            net.addArc(cloneT, cloneP, arc.getWeight());
                            
                            place2proc.associateWithProcess(cloneP, newProc);
                        }
                            
                        for(final PTArc arc2 : originalP.getOutput()) {
                            final Transition t = arc2.getTarget();
                            Transition copyT = transitions[0].get(t);
                            if(copyT == null) {
                                original.add(t);
                                for(int k = 0; k < multiplier; ++k) {
                                    String newName = t.getMeaning();
                                    for(int i = oldNames.size()-1; i >= 0; --i) {
                                        newName = patterns.get(i).matcher(newName).replaceAll(newNames[k].get(i).toString());
                                    }
                                    copyT = net.addTransition(this.nextTransitionName(net));
                                    copyT.setMeaning(newName);
                                    transitions[k].put(t, copyT);
                                    if(this.doLayout) {
	                                    copyT.setX(t.getX() + x[k]);
	                                    copyT.setY(t.getY());
                                    }
                                    cloneP = places[k].get(originalP);
                                    net.addArc(cloneP, copyT, arc2.getWeight());
                                }
                            } else {
                                for(int k = 0; k < multiplier; ++k) {
                                    cloneP = places[k].get(originalP);
                                    copyT = transitions[k].get(t);
                                    net.addArc(cloneP, copyT, arc2.getWeight());
                                }
                            }
                        }
                    } else {
                        Collection<Transition> arcs = tpArcs.get(originalP);
                        if(arcs == null) {
                            arcs = new ArrayList<Transition>();
                            tpArcs.put(originalP, arcs);
                        }
                        for(int i = 0; i < arc.getWeight(); ++i) {
                            arcs.add(originalT);
                        }
                    }
                }
            }
        }
        
        for(final Map.Entry<Place, Collection<Transition>> e : ptArcs.entrySet()) {
            final Place p = e.getKey();
            final Collection<Transition> c = e.getValue();
            final Place theP = places[0].get(p);
            if(theP == null) {
                for(final Transition t : c) {
                    for(int k = 0; k < multiplier; ++k) {
                        net.addArc(p, transitions[k].get(t), 1);
                    }
                }
            } else {
                for(final Transition t : c) {
                    for(int k = 0; k < multiplier; ++k) {
                        net.addArc(places[k].get(p), transitions[k].get(t), 1);
                    }
                }
            }
        }
        
        for(final Map.Entry<Place, Collection<Transition>> e : tpArcs.entrySet()) {
            final Place p = e.getKey();
            final Collection<Transition> c = e.getValue();
            final Collection<Transition> c2 = ptArcs.get(p);
            final Place theP = places[0].get(p);
            if(theP == null) {
                for(final Transition t : c) {
                    if((c2 == null) || !Pi2PN.containsThis(c2, t)) {
                        for(int k = 0; k < multiplier; ++k) {
                            net.addArc(transitions[k].get(t), p, 1);
                        }
                    }
                }
            } else {
                for(final Transition t : c) {
                    if((c2 == null) || !Pi2PN.containsThis(c2, t)) {
                        for(int k = 0; k < multiplier; ++k) {
                            net.addArc(transitions[k].get(t), places[k].get(p), 1);
                        }
                    }
                }
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param names
     * @return
     */
    private boolean containsAFreeName(final Process<V> p, final Collection<Name> names) {
        final Collection<Name> fn = this.reactionFinder.freeNames(p);
        for(final Name name : names) {
            if(fn.contains(name)) {
                return true;
            }
        }
        return false;
    }
    
    
    /**
     * TODO: document
     *
     * @param <T>
     * @param c
     * @param t
     * @return
     */
    private static <T> boolean containsThis(final Collection<T> c, final T t) {
        for(final T x : c) {
            if(x == t) {
                return true;
            }
        }
        return false;
    }
    
    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @param reactions
     */
    private void getReactionsBetween(final Process<V> p1,
            final Process<V> p2,
            final Collection<Reaction<V>> reactions) {
        
        long start = System.currentTimeMillis();
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Getting reactions between " + p1 + " and " + p2);
        }
        
        // get all reactions
        final Collection<Reaction<V>> restrictedReactions = new ArrayList<Reaction<V>>();
        final Collection<Reaction<V>> reactionsToBeRestricted = this.reactionFinder.getReactionsBetween(p1, p2);
        
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Getting reactions took " + Pi2PN.getDuration(start));
        }
        start = System.currentTimeMillis();
        
        
        for(final Reaction<V> r : reactionsToBeRestricted) {
            try {
                final Process<V> restrictedTop = this.restrictor.restrict(r.getReactedProcess());
                r.setReactedProcess(restrictedTop);

                restrictedReactions.add(r);
            } catch(final RuntimeException e) {
                throw new RuntimeException("Reaction: " + p1 + " | " + p2 + "\n=> " + r, e);
            }
        }
        
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Restricting reactions took " + Pi2PN.getDuration(start));
        }
        
        reactions.addAll(restrictedReactions);
        if(restrictedReactions.isEmpty()) {
            if(this.logger.isLoggable(Level.FINEST)) {
                this.logger.finest("No reactions of interest in process " + p1 + " | " + p2);
            }
        } else {
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Process\n" + p1 + " | " + p2
                                + "\nhas these possible reactions:\n"
                                + restrictedReactions);
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param reactions
     */
    private void getReactions(final Process<V> p, final Collection<Reaction<V>> reactions) {
        long start = 0;
        if(this.logger.isLoggable(Level.FINEST)) {
            start = System.currentTimeMillis();
            this.logger.finest("Getting reactions of " + p);
        }
        
        // get all reactions
        final Collection<Reaction<V>> restrictedReactions = new ArrayList<Reaction<V>>();
        final Collection<Reaction<V>> reactionsToBeRestricted = this.reactionFinder.getReactions(p);
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Getting reactions took " + Pi2PN.getDuration(start));
            start = System.currentTimeMillis();
        }
        
        for(final Reaction<V> r : reactionsToBeRestricted) {
            try {
                final Process<V> restrictedTop = this.restrictor.restrict(r.getReactedProcess());
                r.setReactedProcess(restrictedTop);

                restrictedReactions.add(r);
            } catch(final RuntimeException e) {
                throw new RuntimeException("Reaction: " + p + "\n=> " + r, e);
            }
        }
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Restricting reactions took " + Pi2PN.getDuration(start));
        }
        
        reactions.addAll(restrictedReactions);
        if(restrictedReactions.isEmpty()) {
            if(this.logger.isLoggable(Level.FINEST)) {
                this.logger.finest("No reactions of interest in process " + p);
            }
        } else {
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Process\n" + p
                                + "\nhas these possible reactions:\n"
                                + restrictedReactions);
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param sem
     * @param startState
     * @param proc
     * @param place2proc
     * @param structures
     * @param net
     */
    private void initializeNet(final Semantics sem,
                    final Collection<Place> startState,
                    final Process<V> proc,
                    final CommunicatorCache<V> place2proc,
                    final List<SingleReactionTask<V>> structures,
//                    final Map<V, Collection<SingleReactionTask<V>>> structuresMap,
                    final PetriNet net) {
        final Deque<Process<V>> decomposition = new LinkedList<Process<V>>();
        this.decompose(proc, decomposition);
        int x = 0;
        while(!decomposition.isEmpty()) {
            Process<V> p = decomposition.removeLast();
            
            if(this.reactionFinder.isBareNull(p)) {
                // null processes are being ignored
                continue;
            }
            if(sem.equals(Semantics.SEQUENTIAL)) {
                this.namesUsed.addAll(this.reactionFinder.boundNames(p));
                this.namesUsed.addAll(this.reactionFinder.freeNames(p));
                if(!p.getRestrictions().isEmpty()) {
                    // the restricted names of the initial processes may be kept
                    // unchanged. of course, the restrictions have to be removed
                    // p = this.reactionFinder.changeNames(p, p.getNames(), this.namesUsed);
                    // this.namesUsed.addAll(p.getRestrictions());
                    p = this.reactionFinder.copyTopLevel(p);
                    p.getRestrictions().clear();
                    this.decompose(p, decomposition);
                    continue;
                }
                
            } else if(sem.equals(Semantics.SEQ_STR) || sem.equals(Semantics.STR_SEQ)) {
                this.namesUsed.addAll(this.reactionFinder.boundNames(p));
                this.namesUsed.addAll(this.reactionFinder.freeNames(p));
                if(!p.getRestrictions().isEmpty()) {
                    final Collection<Name> names = new ArrayList<Name>(0);
                    for(final Name name : p.getRestrictions()) {
                        if((sem.equals(Semantics.SEQ_STR) && this.isMagicName(name))
                                || (sem.equals(Semantics.STR_SEQ) && !this.isMagicName(name))) {
                            names.add(name);
                        }
                    }
                    if(names.isEmpty()) {
                        p = this.reactionFinder.copyTopLevel(p);
                        p.getRestrictions().clear();
                        this.decompose(p, decomposition);
                        continue;
                    } else if(p.getRestrictions().size() != names.size()) {
                        p = this.reactionFinder.copyTopLevel(p);
                        p.getRestrictions().clear();
                        p.getRestrictions().addAll(names);
                    }
                }
            }
            
//            final SingleReactionTask<V> t = this.find(structuresMap, p);
            
            final int idx = this.binarySearch(structures, p);
            if(idx < 0) {
//            if(t == null) {
                final Place place = net.addPlace(this.nextPlaceName(net), 1);
                place.setMeaning(p.toString());
                place2proc.associateWithProcess(place, p);
                startState.add(place);
                
                if(this.doLayout) {
	                place.setX(x + (this.offsetX << 2));
	                place.setY(this.offsetY);
	                x += this.offsetX;
                }
//                this.add(structuresMap, p, place);
                this.add(idx, structures, p, place);
                
            } else {
                final SingleReactionTask<V> t = structures.get(idx);
                startState.add(t.getPlace());
                t.getPlace().addToken();
            }
            
        }
    }
    
    /**
     * @param idx
     * @param list
     * @param p
     * @param place
     */
    private SingleReactionTask<V> add(final int idx, final List<SingleReactionTask<V>> list, final Process<V> p, final Place place) {
        final SingleReactionTask<V> t = new SingleReactionTask<V>(p, place);
        list.add(-(idx+1), t);
        return t;
    }
    
    
//    /**
//     * @param map
//     * @param p
//     * @param place
//     */
//    private SingleReactionTask<V> add(final Map<V, Collection<SingleReactionTask<V>>> map, final Process<V> p, final Place place) {
//        Collection<SingleReactionTask<V>> col = map.get(p.getData());
//        if(col == null) {
//            col = new ArrayList<SingleReactionTask<V>>();
//            map.put(p.getData(), col);
//        }
//        final SingleReactionTask<V> t = new SingleReactionTask<V>(p, place);
//        col.add(t);
//        return t;
//    }
//    
//    
//    /**
//     * @param map
//     * @param p
//     * @return
//     */
//    private SingleReactionTask<V> find(final Map<V, Collection<SingleReactionTask<V>>> map, final Process<V> p) {
//        final Collection<SingleReactionTask<V>> col = map.get(p.getData());
//        if(col == null) {
//            return null;
//        }
//        if(col.size() > 1) System.out.println(col.size());
//        for(final SingleReactionTask<V> t : col) {
//            if(this.structChecker.structurallyCongruent(p, t.getProcess())) {
//                return t;
//            }
//        }
//        return null;
//    }
    
    
    /**
     * TODO: document
     *
     * @param list
     * @param key
     * @return
     */
    private int binarySearch(final List<SingleReactionTask<V>> list,
                    final Process<V> key) {
        int low = 0;
        int high = list.size()-1;
        
        while (low <= high) {
            final int mid = (low + high) >>> 1;
            final SingleReactionTask<V> midVal = list.get(mid);
            final int cmp = this.structChecker.getComparator().compare(midVal.getProcess(), key);

            if (cmp < 0) {
                low = mid + 1;
            } else if (cmp != 0) {
                high = mid - 1;
            } else {
                if(this.structChecker.structurallyCongruent(midVal.getProcess(), key)) {
                    return mid;
                }
                
                boolean equal = true;
                int i = mid-1;
                while(equal && (i >= low)) {
                    final SingleReactionTask<V> m = list.get(i);
                    equal = this.structChecker.getComparator().compare(m.getProcess(), key) == 0;
                    if(equal && this.structChecker.structurallyCongruent(m.getProcess(), key)) {
                        return i;
                    }
                    --i;
                }
                
                equal = true;
                i = mid+1;
                while(equal && (i <= high)) {
                    final SingleReactionTask<V> m = list.get(i);
                    equal = this.structChecker.getComparator().compare(m.getProcess(), key) == 0;
                    if(equal && this.structChecker.structurallyCongruent(m.getProcess(), key)) {
                        return i;
                    }
                    ++i;
                }

                return -(mid+1); // key not found
            }
        }
        return -(low + 1);  // key not found
    }
    
       
    /**
     * TODO: document
     *
     * @param <T>
     * @param proc
     * @param decomposition
     */
    private <T> void decompose(final Process<T> proc, final Collection<Process<T>> decomposition) {
        final boolean onlyOnePlace;
        boolean empty = true;
        for(final ActionPrefix a : proc.getActionPrefixes()) {
            if(a.getType() == ActionPrefix.Type.GUARD) {
                if(!((Guard)a).isSatisfied()) {
                    empty = false;
                    break;
                }
                empty = false;
                break;
            }
            empty = false;
            break;
        }
        if(!proc.getRestrictions().isEmpty()) {
            onlyOnePlace = true;
        } else if(!empty) {
            onlyOnePlace = true;

        } else if(proc.getType() == Process.Type.COMPOSITION) {
            final ProcessComposition<T> comp = (ProcessComposition<T>) proc;
            if(comp.getOperator() == ProcessComposition.Operator.PARALLEL) {
                onlyOnePlace = false;
                for(final Process<T> p : comp.getProcesses()) {
                    this.decompose(p, decomposition);
                }

            } else {
                onlyOnePlace = true;
            }
        } else {
            onlyOnePlace = true;

        }

        if(onlyOnePlace) {
            decomposition.add(proc);
        }
    }
    

    /**
     * TODO: document
     */
    private void loadMainProcessID() {
        if(this.mainProcessID == null) {
            final String sMain = this.loadProperty(CompilerProperty.MAIN_PROCESS);
            if(sMain == null || sMain.length() == 0) {
                throw new IllegalArgumentException("No main process declared!");
            }
            final Name mainProcessName = this.pbuilder.newName(sMain);
            if(this.logger.isLoggable(Level.FINER)) {
                this.logger.finer("Main process id set to: " + mainProcessName);
            }
            this.mainProcessID = this.pbuilder.newProcessID(mainProcessName);
            final ProcessDefinition<?> d = this.pmanager.getProcessDefinition(this.mainProcessID);
            if(d != null) {
                d.setMain(true);
            }
        }
        
        if(this.logger.isLoggable(Level.FINEST)) {
            this.logger.finest("Loading main process '" + this.mainProcessID + "'.");
        }
    }
    
    
    /**
     * TODO: document
     *
     * @return
     */
    private Process<V> getSystemProcess() {
        final Process<V> main = this.pmanager.getProcess(this.mainProcessID);
        if(main == null) {
            throw new RuntimeException("Could not find main process '"
                            + this.mainProcessID + "'.");
        }
        return main;
    }
    
    
    /**
     * 
     *
     * @param file
     */
    private void parseLocalConfiguration(final String file) {
    	if(this.localConfigParser == null) {
    		this.loadCompilationProperties();
    		return;
    	}
    	InputStream stream = null;
        try {
	    	try {
	            stream = new FileInputStream(file);
	        } catch(final FileNotFoundException e) {
	            throw new RuntimeException("Unable to open file '" + file + "'.", e);
	        }
	        
	        try {
                this.localConfigParser.setInput(stream);
            } catch(final RuntimeException e) {
                throw new RuntimeException("Unable to initialise local "
                		+ "configuration parser for file '" + file + "'.", e);
            }
            
	        
	        // try to load the local file's options
	        try {
	        	final Properties p = this.localConfigParser.getLocalConfiguration();
	        	Petruchio.checkValidOptionKeys(this, p, Pi2PN.EXCLUDE_FOR_LOCAL_CONFIG);
	        	final List<Properties> props = new ArrayList<Properties>(this.propertyLoader.getProperties());
	        	props.add(1, p);
	        	this.loadCompilationProperties(props);
	        	if(!p.isEmpty() && this.logger.isLoggable(Level.CONFIG)) {
	        		this.logger.config("Local configuration loaded: " + p);
	        	}
	        } catch(final RuntimeException e) {
	        	throw new RuntimeException("Unable to parse the local configuration"
	                    + " from file '" + file + "'.", e);
	        }
        } finally {
        	if(stream != null) {
        		try {
        			stream.close();
        		} catch(final IOException e) {
        			// ignore
        		}
        	}
        }
    }
    
    
    /**
     * Reads the input file and parses the process terms into the process
     * manager.
     *
     * @param file The absolute path to the file containing pi process definitions
     */
    private void parse(final String file) {
        if(this.parser == null) {
            throw new IllegalArgumentException("Unable to compile since pi parser has not been set.");
        }
        InputStream stream = null;
        try {
            try {
                stream = new FileInputStream(file);
            } catch(final FileNotFoundException e) {
                throw new RuntimeException("Unable to open file '" + file + "'.", e);
            }
            
            try {
                this.parser.setInput(stream);
            } catch(final RuntimeException e) {
                throw new RuntimeException("Unable to initialise parser for file '"
                                + file + "'.", e);
            }
            
            // try to read the process definitions from the input file
            try {
                boolean parse = true;
                do {
                    // read the next definition
                    final ProcessDefinition<V> p = this.parser.nextProcessDefinition();
                    
                    // the grammar returns 'null' if the end of file has been reached
                    if(p != null) {
                        // now add the process definition to the process manager
                        this.pmanager.addProcessDefinition(p);
                        if(p.isMain()) {
                            if(this.logger.isLoggable(Level.FINER)) {
                                this.logger.finer("Parsed main process definition:\n" + p);
                            }
                            if(this.mainProcessID != null) {
                                if(this.logger.isLoggable(Level.INFO)) {
                                    this.logger.info("Main process id has already been set to "
                                                + this.mainProcessID.getID() + ".\n"
                                                + "This will be used.");
                                }
                            } else {
                                this.mainProcessID = p.getProcessID();
                            }
                        } else {
                            if(this.logger.isLoggable(Level.FINER)) {
                                this.logger.finer("Parsed process definition:\n" + p);
                            }
                        }
                    } else {
                        // end of file has been reached - stop the parsing
                        parse = false;
                    }
                } while(parse);
                
            } catch(final RuntimeException e) {
                throw new RuntimeException("Unable to parse all the pi process "
                                + "definitions from file '" + file + "'.", e);
                
            }
            
        } finally {
            if(stream != null) {
                try {
                    stream.close();
                } catch(final Exception e) {
                    // ignore
                }
            }
        }
    }
    
    
    /**
     * Translates all the process definitions in the process manager into
     * restricted forms with unique bound names.
     * 
     * @param sem The semantics to use.
     */
    private void restrictAll(final Semantics sem) {
        final Map<ProcessID, Process<V>> buffer =
            new HashMap<ProcessID, Process<V>>(this.pmanager.getProcessIDs().size());
        
        this.namesUsed.clear();
        
        // iterate over all key-value pairs and restrict the process (value) 
        for(final ProcessID pid : this.pmanager.getProcessIDs()) {
            // get the process
            Process<V> p = this.pmanager.getProcess(pid);
            final Parameters formalParams = this.pmanager.getFormalParameters(pid);
            
            if(this.logger.isLoggable(Level.FINE)) {
                this.logger.fine("Preprocessing: " + pid
                                + (formalParams.isEmpty() ? "" : ("(" + formalParams + ")"))
                                + " := " + p);
            }
            
            if(!this.keepGlobalRestrictions && pid.equals(this.mainProcessID)) {
                while(true) {
                    if(!p.getActionPrefixes().isEmpty()) {
                        break;
                    }
                    if(p.getType() == Process.Type.PREFIX) {
                        final PrefixProcess<V> p2 = (PrefixProcess<V>)p;
                        p = p2.getProcess();
                    } else if(p.getType() == Process.Type.COMPOSITION) {
                        final ProcessComposition<V> p2 = (ProcessComposition<V>)p;
                        if(p2.getProcesses().size() == 1) {
                            p = p2.getProcesses().get(0);
                        } else {
                            break;
                        }
                    } else {
                        break;
                    }
                }
                if(!p.getRestrictions().isEmpty()) {
                    p.getRestrictions().clear();
                    p.changed();
                }
                if(this.logger.isLoggable(Level.FINE)) {
                    this.logger.fine("Removed public names from main process: " + p);
                }
            }

            // alpha convert double bound names
            final Process<V> pUnique = this.reactionFinder.uniqueBoundNames(p, this.namesUsed);
            
            if(sem.equals(Semantics.SEQ_STR)) {
//                this.namesUsed.addAll(this.reactionFinder.boundNames(pUnique));
                this.namesUsed.addAll(this.reactionFinder.freeNames(pUnique));
            }
            
            if(this.logger.isLoggable(Level.FINE)) {
                this.logger.fine("Unique:  " + pUnique);
            }
            
            // alpha convert double bound names
            final Process<V> pLifted = this.reactionFinder.liftRestrictions(pUnique);
            
            if(this.logger.isLoggable(Level.FINE)) {
                this.logger.fine("Lifted:  " + pLifted);
            }
            
            // bring to restricted form
            final Process<V> pRestricted = this.restrictor.restrict(pLifted);

            if(this.logger.isLoggable(Level.FINE)) {
                this.logger.fine("Restricted: " + pRestricted);
            }
            
            // save restriction to buffer
            buffer.put(pid, pRestricted);
            
            if(this.logger.isLoggable(Level.FINE) &&
                    (this.keepGlobalRestrictions || !pid.equals(this.mainProcessID))) {
                final Collection<Name> n = new ArrayList<Name>(this.reactionFinder.freeNames(pRestricted));
                n.removeAll(formalParams.getParameters());
                if(!n.isEmpty()) {
                    this.logger.fine("The parameters of process definition "
                                    + pid + " do not contain all the free "
                                    + "names of that process.\nThey will be "
                                    + "treated as public names: " + n);
                }
            }
            
        }
        // write back
        for(final Map.Entry<ProcessID, Process<V>> e : buffer.entrySet()) {
            this.pmanager.updateProcess(e.getKey(), e.getValue());
        }
    }
    
    
    private String preprocess(final Semantics sem) {
        final StringBuffer output = new StringBuffer();
        if(sem != null) {
            this.restrictAll(sem);
        }
        for(final ProcessModifier<V> pm : this.processModifiers) {
            if(this.logger.isLoggable(Level.FINE)) {
                this.logger.fine("Modifying processes with " + pm.getName());
            }
            final String message = pm.modify(this.pmanager, this.mainProcessID);
            output.append("\n:: Process modifier ");
            output.append(pm.getName());
            if(message != null) {
                output.append(" returned message '");
                output.append(message);
                output.append("' ::");
            } else {
                output.append(" returned no message ::");
            }
            if(sem != null) {
                this.restrictAll(sem);
            }
        }
        return output.toString();
    }
    
    
    /**
     * TODO: document
     *
     * @param reactions
     * @param place2proc
     * @param structures
     * @param sem
     * @param namesCreated
     */
    private void checkForUnusedPlaces(final Collection<Reaction<V>> reactions,
    		final CommunicatorCache<V> place2proc,
//    		final Map<V, Collection<SingleReactionTask<V>>> structuresMap,
    		final List<SingleReactionTask<V>> structures,
    		final Semantics sem, final boolean namesCreated) {
    	if(!this.reachableOnly || !this.markUnusedPlaces) {
    		return;
    	}
    	try {
	    	for(final Reaction<V> reaction : reactions) {
	    		if(!sem.equals(Semantics.STRUCTURAL) && namesCreated) {
	    			return;
	    		}
		    	if(reaction.getType() != Reaction.Type.SINGLE) {
		    		continue;
		        }
		    	final SingleReaction<V> r = (SingleReaction<V>)reaction;
		    	if(r.getStepType() != SingleReaction.StepType.RECURSE) {
		    		continue;
		    	}
		    	final ProcessID pid = r.getProcessID();
		    	if(this.excludedForMarking.contains(pid)) {
		    	    return;
		    	}
		    	for(final Map.Entry<Place, Process<V>> e : place2proc.placeProcAssociations()) {
		    		final Place pl = e.getKey();
		    		final Process<V> pr = e.getValue();
		    		if((pl.getOutput().isEmpty() || pl.couldCommunicate())
		    				&& this.pmanager.references(pr).contains(pid)) {
		    		    if(pl.couldCommunicate()) {
		    		        this.excludedForMarking.add(pid);
		    		    }
		    			return;
		    		}
		    	}
		    	for(final PTArc arc : reaction.getTransition().getInput()) {
		    		this.cov.markUnused(arc.getSource());
//		    		place2proc.remove(arc.getSource());
//		    		for(final Iterator<SingleReactionTask<V>> it = structures.iterator(); it.hasNext();) {
//		    		    final SingleReactionTask<V> srt = it.next();
//		    		    if(srt.getPlace() == arc.getSource()) {
//		    		        it.remove();
//		    		        break;
//		    		    }
//		    		}
		    	}
		    	return;
	    	}
    	} catch(final UnsupportedOperationException e) {
    		// ignore
    	}
    }
    
    
    /**
     * TODO: document
     *
     * @param names
     * @return
     */
    private boolean containsMagicName(final Collection<Name> names) {
        for(final Name name : names) {
            if(this.isMagicName(name)) {
                return true;
            }
        }
        return false;
    }
    
    
    /**
     * TODO: document
     *
     * @param net
     * @return
     */
    private boolean reduce(final PetriNet net) {
        final PetriNetReductor[] reductors = new PetriNetReductor[this.pnReductors.size()];
        this.pnReductors.toArray(reductors);
        
        final List<int[]> results = new ArrayList<int[]>();
        Collection<Place> ignorePlace = new ArrayList<Place>(0);
        Collection<Transition> ignoreTrans = new ArrayList<Transition>(0);
        
        boolean changed;
        int lastChange = -1;
        do {
            changed = false;
            final int[] result = new int[reductors.length];
            for(int i = 0; i < reductors.length; ++i) {
                if(lastChange != i) {
                    if(this.pnReductorIgnorer != null) {
                    	ignorePlace = this.pnReductorIgnorer.getPlacesToIgnore(net);
                    	ignoreTrans = this.pnReductorIgnorer.getTransitionsToIgnore(net);
                    }
                    final int reduced = reductors[i].reduce(net,
                			ignorePlace, ignoreTrans);
                    if(reduced != 0) {
                        changed = true;
                        lastChange = i;
                        result[i] = reduced;
                    }
                }
            }
            if(changed) {
                results.add(result);
            }
        } while(changed && this.repeatReductions);
        
        int sum = 0;
        if(!results.isEmpty()) {
            final StringBuilder b = new StringBuilder();
            if(results.size() != 1) {
                b.append("\nreductions");
            }
            boolean summation = false;
            int run = 1;
            for(final int[] result : results) {
                if(results.size() != 1) {
                    b.append("\n  run ");
                    b.append(run);
                }
                for(int i = 0; i < result.length; ++i) {
                    if(result[i] != 0) {
                        b.append("\n");
                        if(results.size() != 1) {
                            b.append("   ");
                        }
                        if(sum != 0) {
                            summation = true;
                        }
                        sum += result[i];
                        b.append(result[i]);
                        b.append(" elements removed by ");
                        b.append(reductors[i].getName());
                    }
                }
                ++run;
            }
            if(summation) {
                b.append("\n");
                b.append(sum);
                b.append(" elements removed by reductions");
            }
            if((ignorePlace != null) && !ignorePlace.isEmpty()) {
            	b.append("\nplaces excluded from reductions: ");
            	b.append(ignorePlace);
            }
            if((ignoreTrans != null) && !ignoreTrans.isEmpty()) {
            	b.append("\ntransitions excluded from reductions: ");
            	b.append(ignoreTrans);
            }
            net.setHeaderComment(net.getHeaderComment() + b.toString());
        }
        if(!this.pnReductors.isEmpty() && this.logger.isLoggable(Level.INFO)) {
            this.logger.info("Reduced net by " + sum + " elements.");
        }
        
        return lastChange != -1;
    }
    
    
    /**
     * 
     *
     * @param net
     * @return
     */
    private String nextTransitionName(final PetriNet net) {
    	int i = net.getTransitions().size();
    	String name;
    	do {
    		name = "t" + (++i);
    	} while(!net.getTransitions(name).isEmpty());
    	return name;
    }
    
    
    /**
     * 
     *
     * @param net
     * @return
     */
    private String nextPlaceName(final PetriNet net) {
    	int i = net.getPlaces().size();
    	String name;
    	do {
    		name = "p" + (++i);
    	} while(!net.getPlaces(name).isEmpty());
    	return name;
    }
    
    
    /**
     * TODO: document
     *
     * @param <T>
     * @param r
     * @return
     */
    private <T> String reactionToString(final Reaction<T> r) {
        switch(r.getType()) {
            case SINGLE: {
                final SingleReaction<T> reaction = (SingleReaction<T>)r;
                switch(reaction.getStepType()) {
                    case TAU:
                        return "tau";
                        
                    case RECURSE:
                        final StringBuilder buf = new StringBuilder();
                        final Iterator<Name> b = reaction.getNamesBefore().iterator();
                        final Iterator<Name> a = reaction.getNamesAfter().iterator();
                        buf.append(reaction.getProcessID());
                        buf.append("(");
                        while(b.hasNext()) {
                            buf.append(a.next());
                            buf.append("/");
                            buf.append(b.next());
                            if(b.hasNext()) {
                                buf.append(",");
                            }
                        }
                        buf.append(")");
                        return buf.toString();
                    default:
                        throw new InternalError("Unknown type of reaction: " + r);
                }
            }
            case COMMUNICATION: {
                final CommunicationReaction<T> reaction = (CommunicationReaction<T>)r;
                final StringBuilder b = new StringBuilder();
                b.append(reaction.getChannel());
                b.append("<");
                for(final Iterator<Name> it = reaction.getNamesAfter().iterator(); it.hasNext();) {
                    b.append(it.next());
                    if(it.hasNext()) {
                        b.append(",");
                    }
                }
                b.append(">|");
                b.append(reaction.getChannel());
                b.append("(");
                for(final Iterator<Name> it = reaction.getNamesBefore().iterator(); it.hasNext();) {
                    b.append(it.next());
                    if(it.hasNext()) {
                        b.append(",");
                    }
                }
                b.append(")");
                return b.toString();
            }
            default:
                throw new InternalError("Unexpected type of reaction: " + r);
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Compiler#getStatus()
     */
    @Override
    public String getStatus() {
        final StringBuilder b = new StringBuilder();
        b.append("Petri net status: ");
        if(this.pNet != null) {
            b.append(this.pNet.getPlaces().size());
            b.append(" P, ");
            b.append(this.pNet.getTransitions().size());
            b.append(" T, ");
            b.append(this.pNet.getPTArcCount());
            b.append(" PT, ");
            b.append(this.pNet.getTPArcCount());
            b.append(" TP (");
            b.append(1000*this.pNet.getTransitions().size() / (System.currentTimeMillis() - this.compilationStarted));
            b.append(" T/s)");
        } else {
            b.append("not available");
        }
        if(this.cov != null && this.reachableOnly) {
            b.append("\nReach: ");
            b.append(this.cov.size());
            b.append(" N (");
            b.append(1000*this.cov.size() / (System.currentTimeMillis() - this.compilationStarted));
            b.append(" N/s), ");
        } else {
            b.append("\n");
        }
        b.append("Mem: ");
        b.append(Pi2PN.memoryUsed());
        return b.toString();
    }
    
    
    /**
     * TODO: Document
     *
     * @return
     */
    private static String memoryUsed() {
        long mem = Runtime.getRuntime().totalMemory()-Runtime.getRuntime().freeMemory();
        final StringBuilder b = new StringBuilder();
        long v = mem % 1024;
        if(v != 0) {
            b.insert(0, "b");
            b.insert(0, v);
        }
        mem >>= 10;
        v = mem % 1024;
        if(v != 0) {
            b.insert(0, "kb ");
            b.insert(0, v);
        }
        mem >>= 10;
        v = mem % 1024;
        if(v != 0) {
            b.insert(0, "mb ");
            b.insert(0, v);
        }
        mem >>= 10;
        if(mem != 0) {
            b.insert(0, "gb ");
            b.insert(0, mem);
        }
        return b.toString();
    }


	/* (non-Javadoc)
	 * @see petruchio.interfaces.Compiler#convert(java.lang.String, java.lang.String)
	 */
    @Override
    public void convert(final String inputFile, final String outputFile) {
        this.pleaseStop = false;
        this.parseLocalConfiguration(inputFile);
        try {
            if (this.pnReader == null) {
                if (this.parser == null) {
                    throw new IllegalArgumentException("The Pi parser and the "
                            + "Petri net reader may not both be null.");
                }
                long start = System.currentTimeMillis();
                this.parse(inputFile);
                if (this.logger.isLoggable(Level.INFO)) {
                    this.logger.info("Read Pi file in "
                            + Pi2PN.getDuration(start));
                }
                
                if(this.logger.isLoggable(Level.FINER)) {
                    this.logger.finer("Preprocessing parsed processes.");
                }
                final String preprocessMsg = this.preprocess(null);
                if(this.logger.isLoggable(Level.FINER)) {
                    this.logger.finer("Preprocessing done: " + preprocessMsg);
                }
                
                this.writePiFiles(this.pmanager, outputFile);
            } else if (this.pnReductors.isEmpty()
                    && (this.pnReader instanceof PetriNetStreamReader)
                    && this.pnStreamWritersOnly()) {
                final PetriNetStreamReader sr = (PetriNetStreamReader) this.pnReader;
                final long start = System.currentTimeMillis();
                for (final PetriNetWriter w : this.pnWriters) {
                    final PetriNetStreamWriter sw = (PetriNetStreamWriter) w;
                    sw.setOutput(new FileOutputStream(outputFile));
                    sr.setInput(new FileInputStream(inputFile));
                    sr.readTo(sw);
                }
                if (this.logger.isLoggable(Level.INFO)) {
                    this.logger.info("Converted net in "
                            + Pi2PN.getDuration(start));
                }
            } else {
                long start = System.currentTimeMillis();
                this.pnReader.setInput(new FileInputStream(inputFile));
                final PetriNet net = this.pnReader.read();
                if (this.logger.isLoggable(Level.INFO)) {
                    this.logger.info("Read net in " + Pi2PN.getDuration(start));
                }
                start = System.currentTimeMillis();
                if (!this.pnReductors.isEmpty()) {
                    this.reduce(net);
                    if (this.logger.isLoggable(Level.INFO)) {
                        this.logger.info("Reduced net in "
                                + Pi2PN.getDuration(start));
                    }
                }
                this.writeNetFiles(net, outputFile, false);
            }
        } catch (final FileNotFoundException e) {
            throw new RuntimeException("File not found!", e);
        }

    }
	
	
	/**
	 * 
	 *
	 * @return
	 */
	private boolean pnStreamWritersOnly() {
	    if(this.pnWriters == null) {
	        return false;
	    }
	    for(final PetriNetWriter w : this.pnWriters) {
	        if(!(w instanceof PetriNetStreamWriter)) {
	            return false;
	        }
	    }
	    return true;
	}
	
	
	/**
     * 
     *
     * @param start
     * @return
     */
    private static String getDuration(final long start) {
        long dur = System.currentTimeMillis() - start;
        final StringBuilder b = new StringBuilder();
        final long s = dur % 60000;
        dur /= 60000;
        b.append(s/1000.0);
        b.append("s");
        if(dur == 0) {
            return b.toString();
        }
        
        final long m = dur % 60;
        dur /= 60;
        b.insert(0, "m ");
        b.insert(0, m);
        if(dur == 0) {
            return b.toString();
        }
        
        final long h = dur % 24;
        dur /= 24;
        b.insert(0, "h ");
        b.insert(0, h);
        if(dur == 0) {
            return b.toString();
        }
        
        final long d = dur % 7;
        dur /= 7;
        b.insert(0, "d ");
        b.insert(0, d);
        if(dur == 0) {
            return b.toString();
        }
        
        final long w = dur % 52;
        dur /= 52;
        b.insert(0, "w ");
        b.insert(0, w);
        if(dur == 0) {
            return b.toString();
        }
        
        final long y = dur;
        b.insert(0, "y ");
        b.insert(0, y);
        return b.toString();
    }
    
}