/*
 * Project:  Petruchio
 * Filename: ReactionFinder.java
 * Created:  01.12.2006
 * Author:   Tim Strazny
 */
package petruchio.compiler;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import petruchio.common.WeakIdentityHashMap;
import petruchio.compiler.ProcessData.NameCard;
import petruchio.compiler.reactions.CommunicationReaction;
import petruchio.compiler.reactions.SingleReaction;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.algorithms.Reaction;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.Guard;
import petruchio.interfaces.pi.InputAction;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.OutputAction;
import petruchio.interfaces.pi.Parameters;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessID;
import petruchio.interfaces.pi.ProcessManager;
import petruchio.interfaces.pi.ProcessReference;

/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class ReactionFinder implements SelfDescribing, petruchio.interfaces.algorithms.ReactionFinder<ProcessData> {
    
    /** */
    private static final boolean TOP_SPLIT_PREPROCESS = false;

    /** */
    private static final Pattern STRIP_NAME = Pattern.compile("^(.*?)(_(\\d+))?$");
    
    /** */
    private static final int STRIPPED_NAME_GROUP = 1;
    
    private boolean simpleCaching = false;
    
//    /** */
//    private static final int STRIPPED_INDEX_GROUP = 3;
    
    /** The ProcessCreator to use. */
    private ProcessCreator pbuilder = null;
    
    /** The ProcessManager that contains all the processes. */
    private ProcessManager<ProcessData> pmanager = null;
    
    private final Map<Process<ProcessData>, Collection<Reaction<ProcessData>>> reactionCache;
    
    
    /**
     * 
     */
    public ReactionFinder() {
        this(false);
    }
    
    
    /**
     * @param simpleCaching
     */
    ReactionFinder(final boolean simpleCaching) {
        this.simpleCaching = simpleCaching;
        this.reactionCache = !this.simpleCaching ? null
                : new WeakIdentityHashMap<Process<ProcessData>, Collection<Reaction<ProcessData>>>();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "Default algorithm to efficiently calculate all possible "
            + "reactions of a process. This class gives information on partial "
            + "order reduction for the main compiler class.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#setProcessManager(petruchio.interfaces.pi.ProcessManager)
     */
    @Override
    public void setProcessManager(final ProcessManager<ProcessData> pmanager) {
        if(this.pmanager != pmanager) {
            this.clearCaches();
            this.pmanager = pmanager;
        }
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
    	this.clearCaches();
    	if(this.simpleCaching) {
    	    this.reactionCache.clear();
    	}
    	this.pbuilder = null;
    	this.pmanager = null;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#clearCaches()
     */
    @Override
    public void clearCaches() {
    	// no internal caches...
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#setProcessCreator(petruchio.interfaces.pi.ProcessCreator)
     */
    @Override
    public void setProcessCreator(final ProcessCreator pbuilder) {
        this.pbuilder = pbuilder;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#possibleReactions(java.lang.Process)
     */
    @Override
    public Collection<Reaction<ProcessData>> getReactions(final Process<ProcessData> p) {
        if(this.simpleCaching) {
            // look at the cache
            Collection<Reaction<ProcessData>> theReactions = this.reactionCache.get(p);
            // cache miss?
            if(theReactions == null) {
                // just create one new collection
                theReactions = new ArrayList<Reaction<ProcessData>>(0);
                // and pass it through all the (recursive) method calls
                this.getReactions(p, p, new ArrayList<Integer>(0), theReactions, true);
                this.reactionCache.put(p, theReactions);
            }
            // return reactions to the caller
            return theReactions;
        }
        // just create one new collection
        final Collection<Reaction<ProcessData>> theReactions = new ArrayList<Reaction<ProcessData>>(0);
        // and pass it through all the (recursive) method calls
        this.getReactions(p, p, new ArrayList<Integer>(0), theReactions, true);
        // return reactions to the caller
        return theReactions;
    }
    
    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @return
     */
    private Process<ProcessData> composeParallel(final Process<ProcessData> p1, final Process<ProcessData> p2) {
        if(CommonTasks.isBareNull(p1)) {
            return p2;
        } else if(CommonTasks.isBareNull(p2)) {
            return p1;
        }
        
        final ProcessComposition<ProcessData> comp = this.pbuilder.newProcessComposition();
        comp.setOperator(ProcessComposition.Operator.PARALLEL);
        comp.addProcess(p1);
        comp.addProcess(this.makeDisjoint(p1, p2));
        return comp;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#newName(java.lang.String, java.util.Collection, java.util.Collection)
     */
    @Override
    public Name newName(final Name name, final Collection<Name> n, Collection<Name> n2) {
        return this.newName(name, n, n2, null, null, null);
    }
    
    
    /**
     * TODO: document
     *
     * @param name
     * @param n1 not null
     * @param n2 not null
     * @param n3 not null if n4 is not null
     * @param n4 not null if n5 is not null
     * @param n5 
     * @return
     */
    private Name newName(final Name name, final Collection<Name> n1,
                    final Collection<Name> n2, final Collection<Name> n3,
                    final Collection<Name> n4, final Collection<Name> n5) {
        final Matcher matcher = ReactionFinder.STRIP_NAME.matcher(name.toString());
        if(!matcher.find()) {
            throw new InternalError("This was unexpected. The regular "
                            + "expression to strip the suffix seems to be broken!");
        }
        final String strippedName = matcher.group(ReactionFinder.STRIPPED_NAME_GROUP);
//        final String index = matcher.group(ReactionFinder.STRIPPED_INDEX_GROUP);
        Name n = this.pbuilder.newName(strippedName);
//        int i = (index != null) ? Integer.parseInt(index) : 0;
        int i = 0;
        while(n1.contains(n)
                    || ( ((n2 != null) && n2.contains(n))
                      || ( ((n3 != null) && n3.contains(n))
                        || ( ((n4 != null) && n4.contains(n))
                          || ((n5 != null) && n5.contains(n)))))) {
            n = this.pbuilder.newName(strippedName + "_" + (++i));
        }
        return n;
    }
    
    
    /**
     * @param p
     * @param fragment
     * @param indices
     * @param theReactions
     * @param handleCommunications
     */
    private void getReactions(final Process<ProcessData> p,
            final Process<ProcessData> fragment,
            final List<Integer> indices,
            final Collection<Reaction<ProcessData>> theReactions,
            final boolean handleCommunications) {
        if(this.simpleCaching) {
            Collection<Reaction<ProcessData>> newReactions = this.reactionCache.get(fragment);
            // cache miss?
            if(newReactions == null) {
                // just create one new collection
                newReactions = new ArrayList<Reaction<ProcessData>>(0);
                // and pass it through all the (recursive) method calls
                this.getReactionsWithoutCache(p, fragment, indices, newReactions, handleCommunications);
                this.reactionCache.put(p, newReactions);
            }
            // return reactions to the caller
            theReactions.addAll(newReactions);
        } else {
            this.getReactionsWithoutCache(p, fragment, indices, theReactions, handleCommunications);
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param fragment
     * @param indices
     * @param theReactions
     * @param handleCommunications
     */
    private void getReactionsWithoutCache(final Process<ProcessData> p,
                    final Process<ProcessData> fragment,
                    final List<Integer> indices,
                    final Collection<Reaction<ProcessData>> theReactions,
                    final boolean handleCommunications) {
        
        if(fragment.getActionPrefixes().isEmpty()) {
            // no action prefixes
            
            switch(fragment.getType()) {
                case NULL:
                    // no action prefixes and it's the null process => no reaction
                    break;
                case PREFIX:
                    this.getReactions(p,
                            ((PrefixProcess<ProcessData>)fragment).getProcess(),
                            indices, theReactions, handleCommunications);
                    break;
                case REFERENCE:
                    this.getReactionsReference(p, (ProcessReference<ProcessData>)fragment, indices, theReactions);
                    break;
                case COMPOSITION:
                    // a composition of multiple processes
                    this.getReactionsComposition(p,
                            (ProcessComposition<ProcessData>)fragment,
                            indices, theReactions, handleCommunications);
                    break;
                default:
                    throw new InternalError("Unexpected type of process: " + fragment);
            }
            
        } else {
            // there are action prefixes
            this.getReactionsActionPrefix(p, fragment, indices, theReactions);
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param fragment
     * @param indices
     * @param theReactions
     */
    private void getReactionsComposition(final Process<ProcessData> p,
                    final ProcessComposition<ProcessData> fragment,
                    final List<Integer> indices,
                    final Collection<Reaction<ProcessData>> theReactions,
                    final boolean handleCommunications) {
        if(fragment.getOperator() == ProcessComposition.Operator.PARALLEL) {
            this.getReactionsCompositionParallel(p, fragment, indices,
                            theReactions, handleCommunications);
            
        } else if(fragment.getOperator() == ProcessComposition.Operator.CHOICE) {
            this.getReactionsCompositionChoice(p, fragment, indices, theReactions);
            
        } else if(fragment.getOperator() == ProcessComposition.Operator.SEQUENCE) {
            this.getReactionsCompositionSequence(p, fragment, indices,
                            theReactions, handleCommunications);
            
        } else {
            throw new InternalError("Unexpected operator: " + fragment.getOperator());
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param fragment
     * @param localReactions
     * @param handleCommunication
     */
    private void getReactionsCompositionParallel(
                    final Process<ProcessData> p,
                    final ProcessComposition<ProcessData> fragment,
                    final List<Integer> indices,
                    final Collection<Reaction<ProcessData>> theReactions,
                    final boolean handleCommunication) {
        
        if(handleCommunication) {
            this.getReactionsCompositionParallelWithCommuncation(
                            p, fragment, indices, theReactions);
        } else {
            this.getReactionsCompositionParallelWithoutCommunication(
                            p, fragment, indices, theReactions);
        }
        
    }
    
    /**
     * TODO: document
     *
     * @param fragment
     * @param localReactions
     */
    private void getReactionsCompositionParallelWithoutCommunication(
                    final Process<ProcessData> p,
                    final ProcessComposition<ProcessData> fragment,
                    final List<Integer> indices,
                    final Collection<Reaction<ProcessData>> theReactions) {
        CommonTasks.ensureOrdering(fragment);
        final Collection<Reaction<ProcessData>> newReactions = new ArrayList<Reaction<ProcessData>>();
        int i = 0;
        for(final Process<ProcessData> sub : fragment.getProcesses()) {
            indices.add(Integer.valueOf(i));
            this.getReactions(p, sub, indices, newReactions, false);
            for(final Reaction<ProcessData> r : newReactions) {
                if(r.isSequentiallyIndependent() == Reaction.Independent.NOT_SET) {
                    r.setSequentiallyIndependent(Reaction.Independent.TRUE);
                }
            }
            theReactions.addAll(newReactions);
            newReactions.clear();
            indices.remove(indices.size()-1);
            ++i;
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param fragment
     * @param localReactions
     */
    private void getReactionsCompositionParallelWithCommuncation(
                    final Process<ProcessData> p,
                    final ProcessComposition<ProcessData> fragment,
                    final List<Integer> indices,
                    final Collection<Reaction<ProcessData>> theReactions) {
        this.getReactionsCompositionParallelWithoutCommunication(p, fragment,
                        indices, theReactions);
        this.getReactionsCommuncation(p, fragment, theReactions, false);
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param fragment
     * @param theReactions
     * @param useTopSplitPoint
     */
    @SuppressWarnings("unused")
	private void getReactionsCommuncation(
                    final Process<ProcessData> p,
                    final ProcessComposition<ProcessData> fragment,
                    final Collection<Reaction<ProcessData>> theReactions,
                    final boolean useTopSplitPoint) {
        
        // all the names that could be used as channels to communicate
        final Collection<NameCard> names;
        
        if(ReactionFinder.TOP_SPLIT_PREPROCESS && useTopSplitPoint) {
            // this is accurate but quite slow.
            names = new HashSet<NameCard>(0);
            CommonTasks.ensureOrdering(fragment);
            for(int i = fragment.getProcesses().size()-1; i >= 0; --i) {
                final Collection<NameCard> inI = CommonTasks.inputChannels(fragment.getProcesses().get(i));
                final Collection<NameCard> outI = CommonTasks.inputChannels(fragment.getProcesses().get(i));
                for(int j = fragment.getProcesses().size()-1; j >= 0; --j) {
                    if(i != j) {
                        for(final NameCard c : CommonTasks.outputChannels(fragment.getProcesses().get(j))) {
                            if(inI.contains(c)) {
                                names.add(c);
                            }
                        }
                        for(final NameCard c : CommonTasks.inputChannels(fragment.getProcesses().get(j))) {
                            if(outI.contains(c)) {
                                names.add(c);
                            }
                        }
                    }
                }
            }
        } else {
            // this overapproximates much if useTopSplitPoint but is quite fast
            // if !useTopSplitPoint this should be used anyway!
            names = new HashSet<NameCard>(0);
            final Collection<NameCard> out = CommonTasks.outputChannels(p);
            for(final NameCard c : CommonTasks.inputChannels(p)) {
                if(out.contains(c)) {
                    names.add(c);
                }
            }
        }
                
        if(names.isEmpty()) {
            return;
        }
        
        final Map<Name, Integer> nameCount = new HashMap<Name, Integer>();
        for(final NameCard channelCard : names) {
            nameCount.put(channelCard.getName(), Integer.valueOf(0));
        }
        
        final Map<NameCard, Collection<IDProcess<ProcessData>>> in = new HashMap<NameCard, Collection<IDProcess<ProcessData>>>(0);
        final Map<NameCard, Collection<IDProcess<ProcessData>>> out = new HashMap<NameCard, Collection<IDProcess<ProcessData>>>(0);
        final Map<Process<ProcessData>, ProcessComposition<ProcessData>> choices = new IdentityHashMap<Process<ProcessData>, ProcessComposition<ProcessData>>(0);
        final Map<Process<ProcessData>, List<Process<ProcessData>>> paths = new IdentityHashMap<Process<ProcessData>, List<Process<ProcessData>>>(0);
        final List<Integer> indices = new ArrayList<Integer>(0);
        
        this.collectCommunications(fragment, names, in, out, choices, nameCount, indices, paths, new ArrayList<Process<ProcessData>>());
        
        final List<Process<ProcessData>> search = new ArrayList<Process<ProcessData>>(2);
        final List<Process<ProcessData>> replace = new ArrayList<Process<ProcessData>>(2);
        final List<Name> nameSearch = new ArrayList<Name>(2);
        final List<Name> nameReplace = new ArrayList<Name>(2);
        final Collection<Name> originalRestrictions = new ArrayList<Name>(0);
        // this list will contain the restrictions that have to be lifted
        final List<Name> restrictionsToAdd = new ArrayList<Name>();
        // this collection contains the restrictions that possibly have to be lifted
        final Collection<Name> restrictions = new ArrayList<Name>();
        final List<Name> newSearch = new ArrayList<Name>();
        final List<Name> newReplace = new ArrayList<Name>();
        
        for(final Map.Entry<NameCard, Collection<IDProcess<ProcessData>>> e : in.entrySet()) {
            final NameCard channelCard = e.getKey();
            final Name channel = channelCard.getName();
            if(useTopSplitPoint && this.boundNames(p).contains(channel)) {
                continue;
            }
            final Collection<IDProcess<ProcessData>> input = e.getValue();
            final Collection<IDProcess<ProcessData>> output = out.get(channelCard);
            
            if(output == null) {
                continue;
            }
            
            // iterate over all combinations of process communications
            for(final IDProcess<ProcessData> ipOut : output) {
                final Process<ProcessData> pOut = ipOut.getProcess();
                final ProcessComposition<ProcessData> outChoice = choices.get(pOut);
                final Process<ProcessData> pOut2Repl = (outChoice == null) ? pOut : outChoice;
                
                ActionPrefix prefixOut = null;
                for(int i = 0; i < pOut.getActionPrefixes().size(); ++i) {
                    prefixOut = pOut.getActionPrefixes().get(i);
                    if((prefixOut.getType() != ActionPrefix.Type.GUARD)
                            || !((Guard)prefixOut).isSatisfied()) {
                        break;
                    }
                }
                final OutputAction actionOut = (OutputAction)prefixOut;
                
                // path from the top to pOut
                final List<Integer> outPath = ipOut.getIndices();
                final List<Process<ProcessData>> theOutPath = paths.get(pOut);
                
                originalRestrictions.addAll(actionOut.getNames().getParameters());
                originalRestrictions.retainAll(this.boundNames(p));
                
                for(final IDProcess<ProcessData> ipIn : input) {
                    final Process<ProcessData> pIn = ipIn.getProcess();
                    final ProcessComposition<ProcessData> inChoice = choices.get(pIn);

                    // path from the top to pIn
                    final List<Integer> inPath = ipIn.getIndices();
                    if(useTopSplitPoint && (outPath.get(0) == inPath.get(0))) {
                        continue;
                    }
                    
                    Process<ProcessData> splitPoint = fragment;
                    final List<Integer> splitPointIndex = inPath;
                    {
                        boolean found = false;
                        splitPointSearch:
                        for(int i = 0; i < inPath.size(); ++i) {
                            if(i >= outPath.size()) {
                                break;
                            }
                            while(splitPoint.getType() == Process.Type.PREFIX) {
                                final PrefixProcess<ProcessData> pref = (PrefixProcess<ProcessData>)splitPoint;
                                splitPoint = pref.getProcess();
                            }
                            final Integer idxIn = inPath.get(i);
                            final Integer idxOut = outPath.get(i);
                            if(idxIn != idxOut) {
                                found = true;
                                break;
                            }
                            switch(splitPoint.getType()) {
                                case NULL:
                                case REFERENCE:
                                case PREFIX:
                                    break splitPointSearch;
                                case COMPOSITION:
                                    final ProcessComposition<ProcessData> comp = (ProcessComposition<ProcessData>)splitPoint;
                                    splitPoint = comp.getProcesses().get(idxIn.intValue());
                                    break;
                                default:
                                    throw new InternalError("Unknown type of "
                                                    + "process: " + splitPoint);
                            }
                        }
                        if(!found) {
                            continue;
                        }
                    }
                    
                    // no conflict
                    
                    restrictions.addAll(originalRestrictions);
                    
                    // walk from pOut up to the split point and collect the restrictions
                    // of the names that are sent
                    collectRestrictions:
                    for(int i = theOutPath.size()-1; i >= 0; --i) {
                        final Process<ProcessData> node = theOutPath.get(i);
                        if(node == splitPoint) {
                            // no need to lift restrictions higher than splitpoint
                            break collectRestrictions;
                        }
                        Process<ProcessData> curNode = node;
                        for(final Iterator<Name> it = restrictions.iterator(); it.hasNext(); ) {
                            final Name nameSent = it.next();
                            
                            if(((curNode == node) && curNode.getRestrictions().contains(nameSent))
                                            || ((curNode != node) && curNode.getRestrictions().remove(nameSent))) {
                                if(curNode == node) {
                                    curNode = CommonTasks.replaceProcess(curNode, search, replace, this.pbuilder);
                                    if(curNode == node) {
                                        curNode = CommonTasks.copyTopLevel(node, this.pbuilder);
                                    }
                                    
                                    curNode.getRestrictions().remove(nameSent);
                                    curNode.changed();
                                    this.SRAdd(search, replace, node, curNode);
                                }
                                restrictionsToAdd.add(nameSent);

                                it.remove();
                                if(restrictions.isEmpty()) {
                                    // all restrictions have been added
                                    break collectRestrictions;
                                }
                            }
                        }
                    }
                    
                    
                    final Process<ProcessData> replOut = CommonTasks.copyTopLevel(
                            CommonTasks.replaceProcess(pOut, search, replace, this.pbuilder),
                            this.pbuilder);
                    // remove the FIRST action prefix (and satisfied guards)
                    while(prefixOut != replOut.getActionPrefixes().remove(0))
                        /* nothing */;
                    
                    // if there is a choice, we have to keep the restrictions of that
                    // choice process (if they are not lifted)
                    if(outChoice != null) {
                        for(final Name n : outChoice.getRestrictions()) {
                            if(!restrictionsToAdd.contains(n)) {
                                replOut.addRestriction(n);
                            }
                        }
                    }
                    replOut.changed();
                    
                    // add reacted sub process
                    this.SRAdd(search, replace, pOut2Repl, replOut);
                    
                    final Process<ProcessData> pIn2Repl = (inChoice == null) ? pIn : inChoice;
                    final Process<ProcessData> pInCopy = CommonTasks.copyTopLevel(pIn, this.pbuilder);
                    // if there is a choice, we have to keep the restrictions of that
                    // choice process
                    if(inChoice != null) {
                        pInCopy.addRestrictions(inChoice.getRestrictions());
                    }
                    
                    // now remove the first action (and satisfied guards)
                    ActionPrefix prefixIn = null;
                    do {
                        prefixIn = pInCopy.getActionPrefixes().remove(0);
                    } while((prefixIn.getType() == ActionPrefix.Type.GUARD)
                            && ((Guard)prefixIn).isSatisfied());
                    pInCopy.changed();
                    
                    final InputAction actionIn = (InputAction)prefixIn;
                    
                    nameSearch.addAll(actionIn.getNames().getParameters());
                    nameReplace.addAll(actionOut.getNames().getParameters());
                    
                    final Process<ProcessData> replIn = this.replaceFreeNames(
                                    pInCopy, nameSearch, nameReplace);
                    
                    this.SRAdd(search, replace, pIn2Repl, replIn);
                    
                    final List<Name> namesReceived = new ArrayList<Name>(actionOut.getNames().getParameters());
                    
                    final Process<ProcessData> replaced;
                    if(restrictionsToAdd.isEmpty()) {
                        replaced = CommonTasks.replaceProcess(p, search, replace, this.pbuilder);
                        
                    } else {
                        final Process<ProcessData> splitPointRepl = CommonTasks.replaceProcess(
                                        splitPoint, search, replace, this.pbuilder);
                        final Collection<Name> bn = this.boundNames(splitPointRepl);
                        
                        for(int i = 0; i < restrictionsToAdd.size(); ++i) {
                            Name name = restrictionsToAdd.get(i);
                            final int index = nameSearch.indexOf(name);
                            if(index >= 0) {
                                name = nameReplace.get(index);
                                restrictionsToAdd.set(i, name);
                            }
                            
                            if(bn.contains(name)) {
                                final Collection<Name> fn = this.freeNames(splitPointRepl);
                                final Name n = this.newName(name, bn, fn, restrictionsToAdd, null, null);
                                newSearch.add(name);
                                newReplace.add(n);
                            }
                        }
                        
                        for(int i = 0; i < restrictionsToAdd.size(); ++i) {
                            Name name = restrictionsToAdd.get(i);
                            final int index = newSearch.indexOf(name);
                            if(index >= 0) {
                                name = newReplace.get(index);
                                restrictionsToAdd.set(i, name);
                            }
                        }
                        
                        for(int i = 0; i < namesReceived.size(); ++i) {
                            Name name = namesReceived.get(i);
                            int index = nameSearch.indexOf(name);
                            if(index >= 0) {
                                name = nameReplace.get(index);
                            }
                            
                            index = newSearch.indexOf(name);
                            if(index >= 0) {
                                name = newReplace.get(index);
                            }
                            
                            namesReceived.set(i, name);
                        }
                        
                        final Process<ProcessData> s;
                        if(!newSearch.isEmpty()) {
                            s = this.replaceNames(splitPointRepl, newSearch, newReplace);
                        } else {
                            s = splitPointRepl;
                        }
                        s.addRestrictions(restrictionsToAdd);
                        s.changed();
                        replaced = CommonTasks.replaceProcess(p, splitPoint, s,
                                        splitPointIndex, 0, this.pbuilder);
                        
                        newSearch.clear();
                        newReplace.clear();
                    } // end of if(restrictionsToAdd.isEmpty()) { ... } else
                    
                    final CommunicationReaction<ProcessData> cr = new CommunicationReaction<ProcessData>();
                    cr.setReactedProcess(replaced);
                    cr.setChannel(channel);
                    cr.setNamesBefore(actionIn.getNames().getParameters());
                    cr.setNamesAfter(namesReceived);
                    
                    if(this.boundNames(p).contains(channel)
                            && (nameCount.get(channel).intValue() == 2)) {
                        cr.setSequentiallyIndependent(Reaction.Independent.TRUE);
                    } else {
                        cr.setSequentiallyIndependent(Reaction.Independent.FALSE);
                    }
                    
                    if(inChoice != null) {
                        cr.addChoice(inChoice);
                    }
                    if(outChoice != null) {
                        cr.addChoice(outChoice);
                    }
                    
                    // add the reaction
                    theReactions.add(cr);

                    // clear the data
                    nameSearch.clear();
                    nameReplace.clear();
                    search.clear();
                    replace.clear();
                    restrictions.clear();
                    restrictionsToAdd.clear();
                }
            }
        }
    }
    
    
    private void SRAdd(final List<Process<ProcessData>> search,
            final List<Process<ProcessData>> replace,
            final Process<ProcessData> pS, final Process<ProcessData> pR) {
        if(!search.isEmpty()) {
            final List<Process<ProcessData>> s = new ArrayList<Process<ProcessData>>(search);
            final List<Process<ProcessData>> r = new ArrayList<Process<ProcessData>>(replace);
            CommonTasks.replaceProcess(pR, s, r, this.pbuilder);
            
            boolean found = false;
            for(int i = search.size()-1; i >= 0; --i) {
                if(search.get(i) == pS) {
                    replace.set(i, pR);
                    found = true;
                    break;
                }
            }
            if(!found) {
                for(int i = replace.size()-1; i >= 0; --i) {
                    s.clear();
                    r.clear();
                    s.add(pS);
                    r.add(pR);
                    replace.set(i, CommonTasks.replaceProcess(replace.get(i), s, r, this.pbuilder));
                }
            }
        }
        search.add(pS);
        replace.add(pR);
    }
    
    
    /**
     * TODO: document
     *
     * @param fragment
     * @param in
     * @param out
     * @param choicePaths
     * @param paths
     * @param path
     * @param choicePath
     */
    private void collectCommunications(final Process<ProcessData> p,
                    final Collection<NameCard> names,
                    final Map<NameCard, Collection<IDProcess<ProcessData>>> in,
                    final Map<NameCard, Collection<IDProcess<ProcessData>>> out,
                    final Map<Process<ProcessData>, ProcessComposition<ProcessData>> choices,
                    final Map<Name, Integer> nameCount,
                    final List<Integer> indices,
                    final Map<Process<ProcessData>, List<Process<ProcessData>>> paths,
                    final List<Process<ProcessData>> path) {
        if(p.getType() != Process.Type.COMPOSITION) {
            return;
        }
        final ProcessComposition<ProcessData> fragment = (ProcessComposition<ProcessData>)p;
        
        // these are the child processes
        // not needed: CommonTasks.ensureOrdering(fragment);
        final Collection<Process<ProcessData>> frags = fragment.getProcesses();
        final boolean isChoice = (fragment.getOperator() == ProcessComposition.Operator.CHOICE);
        path.add(fragment);

        int i = 0;
        for(final Process<ProcessData> f : frags) {
            indices.add(Integer.valueOf(i));
            
            ActionPrefix first = null;
            for(final ActionPrefix a : f.getActionPrefixes()) {
                if(a.getType() == ActionPrefix.Type.GUARD) {
                    final Guard g = (Guard)a;
                    if(!g.isSatisfied()) {
                        first = a;
                        break;
                    }
                } else {
                    first = a;
                    break;
                }
            }
            
            if(first != null) {
                final Collection<Name> fn = this.freeNames(f);
                for(final Map.Entry<Name, Integer> e : nameCount.entrySet()) {
                    final Name name = e.getKey();
                    if(fn.contains(name)) {
                        e.setValue(Integer.valueOf(e.getValue().intValue()+1));
                    }
                }
                
                // there are prefixes
                path.add(f);
                
                switch(first.getType()) {
                    case INTERNAL:
                    case GUARD:
                        // ignore. handled elsewhere
                        break;
                    case INPUT: {
                        final InputAction ap = (InputAction)first;
                        final NameCard channel = new NameCard(ap.getChannel(), ap.getNames().size());
                        if(names.contains(channel)) {
                            Collection<IDProcess<ProcessData>> inColl = in.get(channel);
                            if(inColl == null) {
                                inColl = new ArrayList<IDProcess<ProcessData>>(1);
                                in.put(channel, inColl);
                            }
                            inColl.add(new IDProcess<ProcessData>(f, new ArrayList<Integer>(indices)));
                            if(isChoice) {
                                final ProcessComposition<ProcessData> old = choices.put(f, fragment);
                                if(old != null) {
                                    choices.put(f, old);
                                }
                            }
                            paths.put(f, new ArrayList<Process<ProcessData>>(path));
                        }
                        break;
                    }
                    case OUTPUT: {
                        final OutputAction ap = (OutputAction)first;
                        final NameCard channel = new NameCard(ap.getChannel(), ap.getNames().size());
                        if(names.contains(channel)) {
                            Collection<IDProcess<ProcessData>> outColl = out.get(channel);
                            if(outColl == null) {
                                outColl = new ArrayList<IDProcess<ProcessData>>(1);
                                out.put(channel, outColl);
                            }
                            outColl.add(new IDProcess<ProcessData>(f, new ArrayList<Integer>(indices)));
                            if(isChoice) {
                                final ProcessComposition<ProcessData> old = choices.put(f, fragment);
                                if(old != null) {
                                    choices.put(f, old);
                                }
                            }
                            paths.put(f, new ArrayList<Process<ProcessData>>(path));
                        }
                        break;
                    }
                    default:
                        throw new InternalError("Unexpected type of prefix: " + first);
                }
                
                path.remove(path.size()-1);
                
                
            } else {
                // prefixes are empty.
                switch(f.getType()) {
                    case NULL:
                    case REFERENCE:
                        final Collection<Name> fn = this.freeNames(f);
                        for(final Map.Entry<Name, Integer> e : nameCount.entrySet()) {
                            final Name name = e.getKey();
                            if(fn.contains(name)) {
                                e.setValue(Integer.valueOf(e.getValue().intValue()+1));
                            }
                        }
                        // nothing else to do
                        break;
                    case PREFIX:
                        this.collectCommunications(
                                ((PrefixProcess<ProcessData>)f).getProcess(),
                                names, in, out, choices, nameCount, 
                                indices, paths, path);
                        break;
                    case COMPOSITION:
                        this.collectCommunications(f, names, in, out, choices,
                                        nameCount, indices, paths, path);
                        break;
                    default:
                        throw new InternalError("Unexpected type of process: " + f);       
                }
            }
            
            if(fragment.getOperator() == ProcessComposition.Operator.SEQUENCE) {
                break;
            }
            
            indices.remove(indices.size()-1);
            ++i;
        }
        
        path.remove(path.size()-1);
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param fragment
     * @param theReactions
     */
    private void getReactionsCompositionSequence(
                    final Process<ProcessData> p,
                    final ProcessComposition<ProcessData> fragment,
                    final List<Integer> indices,
                    final Collection<Reaction<ProcessData>> theReactions,
                    final boolean handleCommunications) {
        
        Process<ProcessData> proc = null;
        for(Process<ProcessData> sub : fragment.getProcesses()) {
            if(!CommonTasks.isBareNull(sub)) {
                proc = sub;
                break;
            }
        }
        
        if(proc == null) {
            return;
        }
        
        indices.add(Integer.valueOf(0));
        final Collection<Reaction<ProcessData>> childReactions = new ArrayList<Reaction<ProcessData>>();
        this.getReactions(proc, proc, new ArrayList<Integer>(), childReactions, handleCommunications);
        
        for(final Reaction<ProcessData> childReaction : childReactions) {
            Process<ProcessData> repl = CommonTasks.replaceProcess(p, proc, childReaction.getReactedProcess(), indices, 0, this.pbuilder);
            childReaction.setReactedProcess(repl);
        }
        theReactions.addAll(childReactions);
        
        indices.remove(indices.size()-1);
    }
    
    
    /**
     * TODO: document
     * @param p
     * @param fragment
     * @param indices
     * @param theReactions
     */
    private void getReactionsCompositionChoice(
                    final Process<ProcessData> p,
                    final ProcessComposition<ProcessData> fragment,
                    final List<Integer> indices,
                    final Collection<Reaction<ProcessData>> theReactions) {
        // these are the child processes
        final Collection<Process<ProcessData>> frags = fragment.getProcesses();
        final Collection<Reaction<ProcessData>> childReactions = new ArrayList<Reaction<ProcessData>>();
        CommonTasks.ensureOrdering(fragment);
        // lets iterate over them and look at all the possible reactions
        int i = 0;
        for(final Process<ProcessData> f : frags) {
            indices.add(Integer.valueOf(i));
            this.getReactions(f, f, indices, childReactions, false);
            for(final Reaction<ProcessData> childReaction : childReactions) {
                childReaction.getReactedProcess().addRestrictions(fragment.getRestrictions());
                Process<ProcessData> repl = CommonTasks.replaceProcess(p, fragment, childReaction.getReactedProcess(), indices, 0, this.pbuilder);
                
                if(repl != p) {
                    repl.addRestrictions(p.getRestrictions());
                }
                childReaction.setReactedProcess(repl);
                childReaction.setSequentiallyIndependent(Reaction.Independent.FALSE);
                childReaction.addChoice(fragment);
            }
            theReactions.addAll(childReactions);
            childReactions.clear();
            indices.remove(indices.size()-1);
            ++i;
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param fragment
     * @param theReactions
     */
    private void getReactionsReference(
                    final Process<ProcessData> p,
                    final ProcessReference<ProcessData> fragment,
                    final List<Integer> indices,
                    final Collection<Reaction<ProcessData>> theReactions) {
        final ProcessID pid = fragment.getProcessID();
        if(!this.pmanager.containsProcessID(pid)) {
            // cannot resolve reference
            return;
        }
        
        final Parameters parameters = fragment.getParameters();
        final Parameters formalParameters = this.pmanager.getFormalParameters(pid);
        if(parameters.getParameters().size() != formalParameters.getParameters().size()) {
            throw new RuntimeException("Process reference " + fragment + " has "
                            + parameters.getParameters().size() + " parameters "
                            + " but it's supposed to have "
                            + formalParameters.getParameters().size() + ".");
        }
        final Process<ProcessData> original = this.pmanager.getProcess(pid);

        final Collection<Name> bn = this.boundNames(p);
        final Collection<Name> fn = this.freeNames(p);
        final Collection<Name> bnN = this.boundNames(original);
        final Collection<Name> fnN = this.freeNames(original);
        
        // part I : alpha convert bound names in P that collide with a free name in Q
        final List<Name> searchP = new ArrayList<Name>(0);
        final List<Name> replP = new ArrayList<Name>(0);
        for(final Name n : fnN) {
            if(bn.contains(n) && !formalParameters.getParameters().contains(n)) {
                searchP.add(n);
                replP.add(this.newName(n, bnN, fnN, bn, fn, replP));
            }
        }
        
        // alpha convert bound names in Q that collide with bound or free names in P
        final List<Name> searchQ = new ArrayList<Name>(0);
        final List<Name> replQ = new ArrayList<Name>(0);
        for(final Name n : bnN) {
            if(bn.contains(n) || fn.contains(n)) {
                searchQ.add(n);
                replQ.add(this.newName(n, bnN, fnN, bn, fn, replQ));
            }
        }
        // replacing is done later on
        
        final int size = searchQ.size();
        searchQ.addAll(formalParameters.getParameters());
        for(final Name n : parameters.getParameters()) {
            final int index = searchP.indexOf(n);
            if(index >= 0) {
                replQ.add(replP.get(index));
            } else {
                replQ.add(n);
            }
        }
        Process<ProcessData> q = this.replaceNames(original, searchQ, replQ);
        
        // part II : alpha convert restrictions for Q (see part I)
        if(q == original) {
            q = this.copyTopLevel(original);
        }
        for(final Name n : fragment.getRestrictions()) {
            final int index = searchP.indexOf(n);
            if(index >= 0) {
                q.addRestriction(replP.get(index));
            } else {
                q.addRestriction(n);
            }
        }
        q.changed();
        
        // insert the process Q into
        Process<ProcessData> r = null;
        try {
            r = CommonTasks.replaceProcess(p, fragment, q, indices, 0, this.pbuilder);
        } catch(final Exception e) {
            throw new InternalError(p + " -- " + indices + " -- " + fragment);
        }
        r = this.replaceNames(r, q, searchP, replP);
        
        final SingleReaction<ProcessData> sr = new SingleReaction<ProcessData>();
        sr.setReactedProcess(r);
        sr.setStepType(SingleReaction.StepType.RECURSE);
        sr.setSequentiallyIndependent(Reaction.Independent.TRUE);
        sr.setProcessID(pid);
        sr.setNamesBefore(formalParameters.getParameters());
        for(int i = 0; i < size; ++i) {
            replQ.remove(0);
        }
        sr.setNamesAfter(replQ);
        theReactions.add(sr);
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param fragment
     * @param indices
     * @param theReactions
     */
    private void getReactionsActionPrefix(
                    final Process<ProcessData> p,
                    final Process<ProcessData> fragment,
                    final List<Integer> indices,
                    final Collection<Reaction<ProcessData>> theReactions) {
        final ActionPrefix first = fragment.getActionPrefixes().get(0);
        switch(first.getType()) {
            case INTERNAL: {
                final Process<ProcessData> repl = CommonTasks.copyTopLevel(fragment, this.pbuilder);
                repl.getActionPrefixes().remove(0);
                final Process<ProcessData> r = CommonTasks.replaceProcess(p, fragment, repl, indices, 0, this.pbuilder);
                final SingleReaction<ProcessData> sr = new SingleReaction<ProcessData>();
                sr.setReactedProcess(r);
                sr.setStepType(SingleReaction.StepType.TAU);
                sr.setSequentiallyIndependent(Reaction.Independent.TRUE);
                theReactions.add(sr);
                break;
            }
            case INPUT:
            case OUTPUT:
                // this is handled someplace else
                break;
            case GUARD:
                final Guard g = (Guard)first;
                if(g.isSatisfied()) {
                    final Process<ProcessData> copy = CommonTasks.copyTopLevel(fragment, this.pbuilder);
                    copy.getActionPrefixes().remove(0);
                    final Process<ProcessData> nu = CommonTasks.replaceProcess(p, fragment, copy, indices, 0, this.pbuilder);
                    theReactions.addAll(this.getReactions(nu));
                }
                break;
            default:                
                throw new InternalError("Unexpected type of action prefix: " + first);
        }
    }
        
    
    /**
     * TODO: document
     *
     * @param p
     * @param search
     * @param replace
     * @return
     */
    private Process<ProcessData> replaceFreeNames(final Process<ProcessData> p,
                    final List<Name> search, final List<Name> replace) {
        return this.replaceFreeNames(p, null, search, replace);
    }
    
    /**
     * TODO: document
     *
     * @param p
     * @param exclude
     * @param search
     * @param replace
     * @return
     */
    private Process<ProcessData> replaceFreeNames(final Process<ProcessData> p,
                    final Process<ProcessData> exclude,
                    final List<Name> search, final List<Name> replace) {
        if(p == exclude) {
            return p;
        }
        
        final Collection<Name> fn = this.freeNames(p); 
        final Collection<Name> bn = this.boundNames(p);
        
        for(int i = 0; i < replace.size(); ++i) {
            final Name name = replace.get(i);
            if(bn.contains(name)) {
                final Name confRepl = this.newName(name, search, replace, bn, fn, null);
                replace.set(i, confRepl);
                for(int j = i+1; j < replace.size(); ++j) {
                    if(replace.get(j).equals(name)) {
                        replace.set(j, confRepl);
                    }
                }
            }
        }
                
        if(search.isEmpty()) {
            // no names to replace, short cut!
            return p;
        }
        return this.replaceNames(p, search, replace);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#replaceNames(petruchio.interfaces.pi.Process, java.util.List, java.util.List)
     */
    @Override
    public Process<ProcessData> replaceNames(final Process<ProcessData> p,
                    final List<Name> search, final List<Name> replace) {
        return this.replaceNames(p, null, search, replace, true, 0, false);
    }
    
    
    /**
     * @param p
     * @param exclude
     * @param search
     * @param replace
     * @return
     */
    private Process<ProcessData> replaceNames(final Process<ProcessData> p,
            final Process<ProcessData> exclude,
            final List<Name> search, final List<Name> replace) {
        return this.replaceNames(p, exclude, search, replace, true, 0, false);
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param search
     * @param replace
     * @param restrictions
     * @param firstPrefix
     * @return
     */
    private <T> Process<T> replaceNames(final Process<T> p,
                    final Process<T> exclude,
                    final List<Name> search, final List<Name> replace,
                    final boolean restrictions, final int firstPrefix,
                    final boolean ignoreChannelOfFirstPrefix) {
        if(p == exclude) {
            return p;
        }
        
        final Process<T> replaced;
        
        switch(p.getType()) {
            case NULL:
                replaced = this.pbuilder.nullProcess();
                break;
            case PREFIX:
                final PrefixProcess<T> prefix = (PrefixProcess<T>)p;
                final Process<T> repl = this.replaceNames(prefix.getProcess(),
                                exclude, search, replace, restrictions, 0, false);
                replaced = this.pbuilder.newPrefixProcess(repl);
                break;
            case REFERENCE: {
                final ProcessReference<T> ref = (ProcessReference<T>)p;
                
                boolean useNewParameters = false;
                final Parameters params = this.pbuilder.newParameterList();
                for(final Name param : ref.getParameters().getParameters()) {
                    final int index = search.indexOf(param);
                    if(index >= 0) {
                        useNewParameters = true;
                        params.add(replace.get(index));
                    } else {
                        params.add(param);
                    }
                }
                final ProcessReference<T> p2;
                final ProcessID pid = ref.getProcessID();
                p2 = this.pbuilder.newProcessReference(pid,
                                        useNewParameters ? params
                                                         : ref.getParameters());
                replaced = p2;
                break;
            }
            case COMPOSITION: {
                final ProcessComposition<T> comp = (ProcessComposition<T>)p;
                final ProcessComposition<T> p2 = this.pbuilder.newProcessComposition();
                p2.setOperator(comp.getOperator());
                
                for(final Process<T> child : comp.getProcesses()) {
                    // recurse!
                    final Process<T> replacedChild = this.replaceNames(child,
                                    exclude, search, replace, restrictions, 0, false);
                    p2.addProcess(replacedChild);
                }

                replaced = p2;
                break;
            }
            default:
                throw new InternalError("Unexpected type of process: " + p);
        }

        if(restrictions) {
            for(final Name name : p.getRestrictions()) {
                final int index = search.indexOf(name);
                if(index >= 0) {
                    final Name replacement = replace.get(index);
                    replaced.addRestriction(replacement);
                } else {
                    replaced.addRestriction(name);
                }
            }
        } else {
            replaced.addRestrictions(p.getRestrictions());
        }
            
        replaced.addActionPrefixes(p.getActionPrefixes());
        this.replaceInActionPrefixes(replaced, search, replace, firstPrefix, ignoreChannelOfFirstPrefix);
        return replaced;
    }
    
    
    /**
     * TODO: document
     *
     * @param <T>
     * @param p
     * @param search
     * @param replace
     * @return
     */
    private <T> Process<T> replaceNames(final Process<T> p, final Name search,
                    final Name replace) {
        return this.replaceNames(p, search, replace, true, 0);
    }
    
    
    /**
     * TODO: document
     *
     * @param <T>
     * @param p
     * @param search
     * @param replace
     * @param restrictions
     * @param firstPrefix
     * @return
     */
    private <T> Process<T> replaceNames(final Process<T> p, final Name search,
            final Name replace, final boolean restrictions, final int firstPrefix) {
        final Process<T> replaced;

        switch(p.getType()) {
            case NULL:
                replaced = this.pbuilder.nullProcess();
                break;
            case PREFIX:
                final PrefixProcess<T> prefix = (PrefixProcess<T>)p;
                final Process<T> repl = this.replaceNames(prefix.getProcess(), search, replace);
                replaced = this.pbuilder.newPrefixProcess(repl);
                break;
            case REFERENCE: {
                final ProcessReference<T> ref = (ProcessReference<T>) p;

                boolean useNewParameters = false;
                final Parameters params = this.pbuilder.newParameterList();
                for(final Name param : ref.getParameters().getParameters()) {
                    if(param.equals(search)) {
                        useNewParameters = true;
                        params.add(replace);
                    } else {
                        params.add(param);
                    }
                }
                final ProcessReference<T> p2;
                final ProcessID pid = ref.getProcessID();
                p2 = this.pbuilder.newProcessReference(pid,
                        useNewParameters ? params : ref.getParameters());

                replaced = p2;
            }
            break;
            case COMPOSITION: {
                final ProcessComposition<T> comp = (ProcessComposition<T>) p;
                
                final ProcessComposition<T> p2 = this.pbuilder.newProcessComposition();
                p2.setOperator(comp.getOperator());

                for(final Process<T> child : comp.getProcesses()) {
                    // recurse!
                    final Process<T> replacedChild = this.replaceNames(child, search,
                            replace);
                    p2.addProcess(replacedChild);
                }

                replaced = p2;
            }
            break;
            default:
                throw new InternalError("Unexpected type of process: " + p);                
        }

        if(restrictions) {
            for(final Name name : p.getRestrictions()) {
                if(name.equals(search)) {
                    replaced.addRestriction(replace);
                } else {
                    replaced.addRestriction(name);
                }
            }
            
        } else {
            replaced.addRestrictions(p.getRestrictions());
            
        }

        replaced.addActionPrefixes(p.getActionPrefixes());
        this.replaceInActionPrefixes(replaced, search, replace, firstPrefix);
        return replaced;
    }

    
    /**
     * TODO: document
     * Use with care! This may modify the process!
     *
     * @param p
     * @param search
     * @param replace
     * @param <T>
     */
    private <T> void replaceInActionPrefixes(final Process<T> p,
                    final List<Name> search, final List<Name> replace,
                    final int firstPrefix, final boolean ignoreChannelOfFirstPrefix) {
        
        for(int i = p.getActionPrefixes().size()-1; i >= firstPrefix; --i) {
            final ActionPrefix ap = p.getActionPrefixes().get(i);
            final ActionPrefix newAP;
            
            switch(ap.getType()) {
                case INTERNAL:
                    // ignore
                    newAP = ap;
                    break;
                case INPUT: {
                    final InputAction in = (InputAction)ap;
                    final int oIndex = (ignoreChannelOfFirstPrefix && i == firstPrefix) ? -1
                            : search.indexOf(in.getChannel());
                    final Parameters newParams = this.pbuilder.newParameterList();
                    boolean paramsChanged = false;
                    for(final Name s : in.getNames().getParameters()) {
                        final int sIndex = search.indexOf(s);
                        if(sIndex >= 0) {
                            newParams.add(replace.get(sIndex));
                            paramsChanged = true;
                        } else {
                            newParams.add(s);
                        }
                    }
                    if((oIndex >= 0) || paramsChanged) {
                        final Name channel = (oIndex >= 0) ? replace.get(oIndex) : in.getChannel();
                        final Parameters names = paramsChanged ? newParams : in.getNames();
                        newAP = this.pbuilder.newInputAction(channel, names);
                    } else {
                        newAP = ap;
                    }
                    break;
                }
                case OUTPUT: {
                    final OutputAction out = (OutputAction)ap;
                    final int oIndex = search.indexOf(out.getChannel());
                    final Parameters newParams = this.pbuilder.newParameterList();
                    boolean paramsChanged = false;
                    for(final Name s : out.getNames().getParameters()) {
                        final int sIndex = search.indexOf(s);
                        if(sIndex >= 0) {
                            newParams.add(replace.get(sIndex));
                            paramsChanged = true;
                        } else {
                            newParams.add(s);
                        }
                    }
                    if((oIndex >= 0) || paramsChanged) {
                        final Name channel = (oIndex >= 0) ? replace.get(oIndex) : out.getChannel();
                        final Parameters names = paramsChanged ? newParams : out.getNames();
                        newAP = this.pbuilder.newOutputAction(channel, names);
                    } else {
                        newAP = ap;
                    }
                    break;
                }
                case GUARD: {
                    final Guard g = (Guard)ap;
                    final int lIndex = search.indexOf(g.leftOperand());
                    final int rIndex = search.indexOf(g.rightOperand());
                    if((lIndex >= 0) || (rIndex >= 0)) {
                        final Name lOp = (lIndex >= 0) ? replace.get(lIndex)
                                                         : g.leftOperand();
                        final Name rOp = (rIndex >= 0) ? replace.get(rIndex)
                                                         : g.rightOperand();
                        final Guard newGuard = this.pbuilder.newGuard(lOp, rOp);
                        
                        // delete the guard iff it is satisfied
                        newAP = newGuard.isSatisfied() ? null : newGuard;
                    } else {
                        newAP = ap;
                    }
                    break;
                }
                default:
                    throw new InternalError("Unexpected type of prefix: " + ap);                    
            }
                
            if(newAP != null) {
                p.getActionPrefixes().set(i, newAP);
                p.changed();
            } else {
                p.getActionPrefixes().remove(i);
                p.changed();
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param search
     * @param replace
     * @param <T>
     */
    private <T> void replaceInActionPrefixes(final Process<T> p,
            final Name search, final Name replace, final int firstPrefix) {
        for(int i = p.getActionPrefixes().size() - 1; i >= firstPrefix; --i) {
            final ActionPrefix ap = p.getActionPrefixes().get(i);
            final ActionPrefix newAP;

            switch(ap.getType()) {
                case INTERNAL:
                    // ignore
                    newAP = ap;
                    break;
                case INPUT: {
                    final InputAction in = (InputAction) ap;
                    final boolean o = in.getChannel().equals(search);
                    final Parameters newParams = this.pbuilder.newParameterList();
                    boolean paramsChanged = false;
                    for(final Name s : in.getNames().getParameters()) {
                        final boolean sa = search.equals(s);
                        if(sa) {
                            newParams.add(replace);
                            paramsChanged = true;
                        } else {
                            newParams.add(s);
                        }
                    }
                    
                    if(o || paramsChanged) {
                        final Name channel = o ? replace : in.getChannel();
                        final Parameters names = paramsChanged ? newParams : in.getNames();
                        newAP = this.pbuilder.newInputAction(channel, names);
                    } else {
                        newAP = ap;
                    }
                    break;
                }
                case OUTPUT: {
                    final OutputAction out = (OutputAction) ap;
                    final boolean o = out.getChannel().equals(search);
                    final Parameters newParams = this.pbuilder.newParameterList();
                    boolean paramsChanged = false;
                    for(final Name s : out.getNames().getParameters()) {
                        final boolean sa = search.equals(s);
                        if(sa) {
                            newParams.add(replace);
                            paramsChanged = true;
                        } else {
                            newParams.add(s);
                        }
                    }
                    
                    if(o || paramsChanged) {
                        final Name channel = o ? replace : out.getChannel();
                        final Parameters names = paramsChanged ? newParams : out.getNames();
                        newAP = this.pbuilder.newOutputAction(channel, names);
                    } else {
                        newAP = ap;
                    }
                    break;
                }
                case GUARD: {
                    final Guard g = (Guard) ap;
                    final boolean l = g.leftOperand().equals(search);
                    final boolean r = g.rightOperand().equals(search);
                    if(l || r) {
                        final Name lOp = l ? replace : g.leftOperand();
                        final Name rOp = r ? replace : g.rightOperand();
                        final Guard newGuard = this.pbuilder.newGuard(lOp, rOp);

                        // delete the guard iff it is satisfied
                        newAP = newGuard.isSatisfied() ? null : newGuard;
                    } else {
                        newAP = ap;
                    }
                    break;
                }
                default:
                    throw new InternalError("Unexpected type of prefix: " + ap);
            }

            if(newAP != null) {
                p.getActionPrefixes().set(i, newAP);
                p.changed();
            } else {
                p.getActionPrefixes().remove(i);
                p.changed();
            }
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#copyTopLevel(petruchio.interfaces.pi.Process)
     */
    @Override
    public Process<ProcessData> copyTopLevel(final Process<ProcessData> p) {
        return CommonTasks.copyTopLevel(p, this.pbuilder);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#freeNames(petruchio.interfaces.pi.Process)
     */
    @Override
    public Collection<Name> freeNames(final Process<ProcessData> p) {
        return CommonTasks.freeNames(p);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#boundNames(petruchio.interfaces.pi.Process)
     */
    @Override
    public Collection<Name> boundNames(final Process<ProcessData> p) {
        return CommonTasks.boundNames(p);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#uniqueBoundNames(petruchio.interfaces.pi.Process, java.util.Collection)
     */
    @Override
    public Process<ProcessData> uniqueBoundNames(final Process<ProcessData> p,
                    final Collection<Name> excludedNames) {
        final Collection<Name> fn = new ArrayList<Name>(0);
        this.collectFreeNames(p, fn);
        return this.uniqueBoundNames(p, fn, new ArrayList<Name>(0), excludedNames);
    }
    
    
    /**
     * @param p
     * @param fn
     * @param bn
     * @param excludedNames
     * @return
     */
    private Process<ProcessData> uniqueBoundNames(final Process<ProcessData> theProcess,
                    final Collection<Name> fn, final Collection<Name> bn,
                    final Collection<Name> excludedNames) {
        
        Process<ProcessData> p = theProcess;
        final List<Name> res = new ArrayList<Name>(0);
        final List<Name> search = new ArrayList<Name>(0);
        final List<Name> replace = new ArrayList<Name>(0);
        for(final Iterator<Name> it = p.getRestrictions().iterator(); it.hasNext();) {
            final Name name = it.next();
            if(bn.contains(name) || fn.contains(name)) {
                // replace
                final Name newName = this.newName(name, fn, bn, p.getRestrictions(), excludedNames, null);
                search.add(name);
                replace.add(newName);
                res.add(newName);
                bn.add(newName);
                
            } else if(res.contains(name)) {
                // remove one instance of the double restricted name
            	p.changed();
                it.remove();
                
            } else {
                res.add(name);
                bn.add(name);
            }
        }
        if(!search.isEmpty()) {
            p = this.replaceNames(p, search, replace);
        }
        
                
        for(int i = 0; i < p.getActionPrefixes().size(); ++i) {
            final ActionPrefix a = p.getActionPrefixes().get(i);
            
            if(a.getType() == ActionPrefix.Type.INPUT) {
                final InputAction in = (InputAction)a;
                
                List<Name> s = null;
                List<Name> r = null;
                for(final Name name : in.getNames().getParameters()) {
                    // handle bound name
                    if(bn.contains(name) || fn.contains(name)) {
                        if(s == null || r == null) {
                            s = new ArrayList<Name>(1);
                            r = new ArrayList<Name>(1);
                        }
                        
                        // it's been bound later on
                        final Name newName = this.newName(name, fn, bn, excludedNames, null, null);
                        s.add(name);
                        r.add(newName);
                        bn.add(newName);
                        
                    } else {
                        bn.add(name);
                    }
                }
                
                if(s != null) {
                    p = this.replaceNames(p, null, s, r, true, i, true);
                }
            }
        }
        
        if(p.getType() == Process.Type.PREFIX) {
            final PrefixProcess<ProcessData> prefix = (PrefixProcess<ProcessData>)p;
            final Process<ProcessData> sub = prefix.getProcess();
            final Process<ProcessData> sub2 = this.uniqueBoundNames(sub, fn, bn, excludedNames);
            if(sub != sub2) {
                final Process<ProcessData> newP = this.pbuilder.newPrefixProcess(sub2);
                newP.addActionPrefixes(p.getActionPrefixes());
                newP.addRestrictions(p.getRestrictions());
                p = newP;
            }
            
        } else if(p.getType() == Process.Type.COMPOSITION) {
            final ProcessComposition<ProcessData> comp = (ProcessComposition<ProcessData>)p;
            final List<Process<ProcessData>> l = new ArrayList<Process<ProcessData>>(comp.getProcesses().size());
            boolean changed = false;
            for(final Process<ProcessData> sub : comp.getProcesses()) {
                final Process<ProcessData> r = this.uniqueBoundNames(sub, fn, bn, excludedNames);
                l.add(r);
                
                if(r != sub) {
                    changed = true;
                }
            }
            
            if(changed) {
                final ProcessComposition<ProcessData> c = this.pbuilder.newProcessComposition();
                c.setOperator(comp.getOperator());
                c.addProcesses(l);
                c.addActionPrefixes(p.getActionPrefixes());
                c.addRestrictions(p.getRestrictions());
                p = c;
                
            }
            
        }
        
        return p;
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param fn
     */
    private void collectFreeNames(final Process<ProcessData> p,
                    final Collection<Name> fn) {
        
        // this goes bottom up
        
        switch(p.getType()) {
            case NULL:
                break;
            case PREFIX:
                this.collectFreeNames(((PrefixProcess<ProcessData>)p).getProcess(), fn);
                break;
            case REFERENCE:
                fn.addAll(((ProcessReference<ProcessData>)p).getParameters().getParameters());
                break;
            case COMPOSITION:
                // not needed CommonTasks.ensureOrdering(p);
                for(final Process<ProcessData> sub : ((ProcessComposition<ProcessData>)p).getProcesses()) {
                    this.collectFreeNames(sub, fn);
                }
                break;
            default:
                throw new InternalError("Unknown type of process: " + p);
        }
        
        for(int i = p.getActionPrefixes().size()-1; i >= 0; --i) {
            final ActionPrefix a = p.getActionPrefixes().get(i);
            
            switch(a.getType()) {
                case INTERNAL:
                    // ignore
                    break;
                case INPUT: {
                    final InputAction in = (InputAction)a;
                    if(!fn.contains(in.getChannel())){
                        fn.add(in.getChannel());
                    }
                    fn.removeAll(in.getNames().getParameters());
                    break;
                }
                case OUTPUT: {
                    final OutputAction out = (OutputAction)a;
                    if(!fn.contains(out.getChannel())){
                        fn.add(out.getChannel());
                    }
                    for(final Name n : out.getNames().getParameters()) {
                        if(!fn.contains(n)){
                            fn.add(n);
                        }
                    }
                    break;
                }
                case GUARD: {
                    final Guard g = (Guard)a;
                    if(!fn.contains(g.leftOperand())){
                        fn.add(g.leftOperand());
                    }
                    if(!fn.contains(g.rightOperand())){
                        fn.add(g.rightOperand());
                    }
                    break;
                }
                default:
                    throw new InternalError("Unknown type of prefix: " + a);
            }
        }
        
        fn.removeAll(p.getRestrictions());
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#liftRestrictions(petruchio.interfaces.pi.Process)
     */
    @Override
    public Process<ProcessData> liftRestrictions(final Process<ProcessData> p) {
        return CommonTasks.liftRestrictions(p, this.pbuilder);
    }

   

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#getReactionsBetween(petruchio.interfaces.pi.Process, petruchio.interfaces.pi.Process)
     */
    @Override
    public Collection<Reaction<ProcessData>> getReactionsBetween(
                    final Process<ProcessData> p1, final Process<ProcessData> p2) {
        final Process<ProcessData> comp = this.composeParallel(p1, p2);
        final Collection<Reaction<ProcessData>> theReactions = new ArrayList<Reaction<ProcessData>>();
        if((comp != p1) && (comp != p2)) {
            this.getReactionsCommuncation(comp,
                            (ProcessComposition<ProcessData>)comp, theReactions, true);
        }
        return theReactions;
    }
    
    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @return
     */
    private Process<ProcessData> makeDisjoint(final Process<ProcessData> p1,
                    final Process<ProcessData> p2) {
        if(CommonTasks.isBareNull(p1) || CommonTasks.isBareNull(p2)) {
            return p2;
        }
        
        final Collection<Name> fn1 = this.freeNames(p1); 
        final Collection<Name> bn1 = this.boundNames(p1);
        final Collection<Name> fn2 = this.freeNames(p2); 
        final Collection<Name> bn2 = this.boundNames(p2);
        
        final List<Name> search = new ArrayList<Name>(0);
        final List<Name> replace = new ArrayList<Name>(0);
        for(final Name name : bn2) {
            if(bn1.contains(name) || fn1.contains(name)) {
                final Name repl = this.newName(name, bn1, fn1, bn2, replace, fn2);
                if(!repl.equals(name)) {
                    search.add(name);
                    replace.add(repl);
                }
            }
        }
        
        final Process<ProcessData> p2New = search.isEmpty() ? p2 : this.replaceNames(p2, search, replace);
        
        return p2New;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#isBareNull(petruchio.interfaces.pi.Process)
     */
    @Override
    public boolean isBareNull(final Process<ProcessData> p) {
        return CommonTasks.isBareNull(p);
    }
    
    
    /**
     * <Description>
     * 
     * @author Tim Strazny
     */
    private static final class IDProcess<V> {
        private final Process<V> proc;
        private final List<Integer> indices;
        
        /**
         * @param proc
         * @param indices
         */
        public IDProcess(final Process<V> proc,
                        final List<Integer> indices) {
            this.proc = proc;
            this.indices = indices;
        }
        
        /**
         * TODO: document
         *
         * @return
         */
        public Process<V> getProcess() {
            return this.proc;
        }
        
        
        /**
         * TODO: document
         *
         * @return
         */
        public List<Integer> getIndices() {
            return this.indices;
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#containsCommunicationViaFreeName(petruchio.interfaces.pi.Process)
     */
    @Override
    public boolean containsCommunicationViaFreeName(final Process<ProcessData> p) {
        final Collection<Name> fn = this.freeNames(p);
        if(fn.isEmpty()) {
            return false;
        }
        return this.couldCommunicateVia(p, fn);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#getAvailableCommunicationsViaFreeNames(petruchio.interfaces.pi.Process)
     */
    @Override
    public Collection<ActionPrefix> getAvailableCommunicationsViaFreeNames(final Process<ProcessData> p) {
        final Collection<ActionPrefix> actions = new ArrayList<ActionPrefix>(0);
    	final Collection<Name> fn = this.freeNames(p);
    	if(fn.isEmpty()) {
    	    return actions;
    	}
        this.couldCommunicateVia(p, fn, actions);
        return actions;
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @param names
     * @return
     */
    private void couldCommunicateVia(final Process<?> p,
                    final Collection<Name> names, final Collection<ActionPrefix> actions) {
        ActionPrefix first = null;
        for(final ActionPrefix a : p.getActionPrefixes()) {
            if(a.getType() == ActionPrefix.Type.GUARD) {
                final Guard g = (Guard)a;
                if(!g.isSatisfied()) {
                    first = a;
                    break;
                }
            } else {
                first = a;
                break;
            }
        }
        
        if(first != null) {
            switch(first.getType()) {
                case INTERNAL:
                case GUARD:
                    break;
                case INPUT: {
                    final InputAction ap = (InputAction)first;
                    if(names.contains(ap.getChannel())) {
                        actions.add(ap);
                    }
                    break;
                }
                case OUTPUT: {
                    final OutputAction ap = (OutputAction)first;
                    if(names.contains(ap.getChannel())) {
                        actions.add(ap);
                    }
                    break;
                }
                default:
                    throw new InternalError("Unexpected type of prefix: " + first);
            }
            
        } else {
            // prefixes are empty.
            switch(p.getType()) {
                case NULL:
                case REFERENCE:
                    break;
                case PREFIX:
                    this.couldCommunicateVia(((PrefixProcess<?>)p).getProcess(), names, actions);
                    break;
                case COMPOSITION:
                    final ProcessComposition<?> c = ((ProcessComposition<?>)p);
                    for(final Process<?> f : c.getProcesses()) {
                        this.couldCommunicateVia(f, names, actions);
                        if(c.getOperator() == ProcessComposition.Operator.SEQUENCE) {
                            break;
                        }
                    }
                    break;
                default:
                    throw new InternalError("Unexpected type of process: " + p);       
            }
        }
    }
    
    
    private boolean couldCommunicateVia(final Process<?> p, final Collection<Name> names) {
        for (final ActionPrefix a : p.getActionPrefixes()) {
            switch (a.getType()) {
                case INTERNAL:
                case GUARD:
                    break;
                case INPUT: {
                    final InputAction ap = (InputAction) a;
                    if (names.contains(ap.getChannel())) {
                        return true;
                    }
                    break;
                }
                case OUTPUT: {
                    final OutputAction ap = (OutputAction) a;
                    if (names.contains(ap.getChannel())) {
                        return true;
                    }
                    break;
                }
                default:
                    throw new InternalError("Unexpected type of prefix: " + a);
            }
        }
        switch (p.getType()) {
            case NULL:
            case REFERENCE:
                break;
            case PREFIX:
                return this.couldCommunicateVia(((PrefixProcess<?>) p).getProcess(), names);
            case COMPOSITION:
                final ProcessComposition<?> c = ((ProcessComposition<?>) p);
                for (final Process<?> f : c.getProcesses()) {
                    if(this.couldCommunicateVia(f, names)) {
                        return true;
                    }
                }
                break;
            default:
                throw new InternalError("Unexpected type of process: " + p);
        }
        return false;
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#getProcessIDs(petruchio.interfaces.pi.Process)
     */
    @Override
    public Collection<Name> getProcessIDs(final Process<ProcessData> p) {
        return CommonTasks.getData(p).getRefProcIDs();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.ReactionFinder#getNullProcessCount(petruchio.interfaces.pi.Process)
     */
    @Override
    public int getNullProcessCount(final Process<ProcessData> p) {
        return CommonTasks.getData(p).getNullProcCount();
    }

}