/*
 * Project:  Petruchio
 * Filename: Restrictor.java
 * Created:  01.12.2006
 * Author:   Tim Strazny
 */
package petruchio.compiler;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import petruchio.common.WeakIdentityHashMap;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.algorithms.RestrictedForm;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.Guard;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.NullProcess;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessReference;

/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class Restrictor implements SelfDescribing, RestrictedForm<ProcessData>{

    /** The ProcessCreator to use. */
    private ProcessCreator pbuilder = null;
    
    /** A mapping from process to process in restricted form. */
    private final Map<Process<ProcessData>, Process<ProcessData>> restricted;
    
    
    /** 
     * Initialize the new instance by creating the needed objects.
     */
    public Restrictor() {
        this.restricted = new WeakIdentityHashMap<Process<ProcessData>, Process<ProcessData>>();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "The default algorithm to bring a process to (strict) restricted form."; 
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
    	this.clearCaches();
        this.pbuilder = null;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.RestrictedForm#clearCaches()
     */
    @Override
    public void clearCaches() {
    	this.restricted.clear();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.RestrictedForm#setProcessCreator(petruchio.interfaces.pi.ProcessCreator)
     */
    @Override
    public void setProcessCreator(final ProcessCreator pbuilder) {
        this.pbuilder = pbuilder;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.RestrictedForm#restrict(java.lang.Process)
     */
    @Override
    public Process<ProcessData> restrict(final Process<ProcessData> p) {
        final Process<ProcessData> lookUp = this.restricted.get(p);
        
        if(lookUp == null) {
            // look up failed. build the restricted form
            Process<ProcessData> r;
            
            switch(p.getType()) {
                case NULL:
                    r = this.restrictNullProcess((NullProcess<ProcessData>)p);
                    break;
                case PREFIX:
                    r = this.restrictPrefixProcess((PrefixProcess<ProcessData>)p);
                    break;
                case COMPOSITION:
                    r = this.restrictComposition((ProcessComposition<ProcessData>)p);
                    break;
                case REFERENCE:
                    r = this.restrictReference((ProcessReference<ProcessData>)p);
                    break;
                default:
                    throw new InternalError("Unexpected type of process: " + p);
            }
            
            this.minimizeRestrictions(p);
            
            // save the restricted form to the hash table
            this.restricted.put(r, r);
            // (no mapping from p to r needed since r replaces p which is discarded)
            
            return r;
            
        }
        // look up successful
        return lookUp;
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     */
    private void minimizeRestrictions(final Process<ProcessData> p) {
        final Collection<Name> res = new HashSet<Name>(p.getRestrictions());
        if(res.retainAll(this.freeNamesWOR(p))) {
            p.changed();
        }
        p.getRestrictions().clear();
        p.addRestrictions(res);
    }
    
    
    /**
     * Restricts a null process.
     *
     * @param p A null process.
     * @return  The null process in restricted form.
     */
    private NullProcess<ProcessData> restrictNullProcess(final NullProcess<ProcessData> p) {
        // the null process with some action prefixes and restrictions
        return p;
    }
    
    
    /**
     * Restricts a prefix process
     *
     * @param p A prefix process
     * @return  The process in restricted form
     */
    private Process<ProcessData> restrictPrefixProcess(final PrefixProcess<ProcessData> p) {
        final Process<ProcessData> r;
        final Process<ProcessData> sub = p.getProcess();
        final Process<ProcessData> repl = this.restrict(sub);
        if(p.getActionPrefixes().isEmpty() || repl.getRestrictions().isEmpty()) {
            r = (repl == sub) ? CommonTasks.copyTopLevel(repl, this.pbuilder) : repl;
            r.addActionPrefixes(p.getActionPrefixes());
            r.addRestrictions(p.getRestrictions());
        } else if(sub == repl) {
            r = p;
        } else {
            r = this.pbuilder.newPrefixProcess(repl);
            r.addActionPrefixes(p.getActionPrefixes());
            r.addRestrictions(p.getRestrictions());
        }
        return r;
    }
    
    
    /**
     * Restricts a process reference.
     *
     * @param p A process reference.
     * @return  The process reference in restricted form.
     */
    private ProcessReference<ProcessData> restrictReference(final ProcessReference<ProcessData> p) {
        return p;
    }
    
    
    /**
     * Restricts a composition of processes.
     *
     * @param p A composition of processes.
     * @return  The composition in restricted form.
     */
    private Process<ProcessData> restrictComposition(final ProcessComposition<ProcessData> p) {
        Process<ProcessData> r;
        
        if(p.getOperator().allowsScopeExtrusion()) {
            r = this.restrictCompositionWithExtrusion(p);
            
        } else {
            r = this.restrictCompositionNoExtrusion(p);
            
        }
        
        
        boolean done = false;
        while(!done) {
            switch(r.getType()) {
                case PREFIX:
                    if(r.getRestrictions().isEmpty() && r.getActionPrefixes().isEmpty()) {
                        final PrefixProcess<ProcessData> prefix = (PrefixProcess<ProcessData>)r;
                        r = prefix.getProcess();
                    } else {
                        done = true;
                    }
                    break;
                case COMPOSITION:
                    final ProcessComposition<ProcessData> comp = (ProcessComposition<ProcessData>)r;
                    if(comp.getProcesses().isEmpty()) {
                        // empty composition => null process
                        final NullProcess<ProcessData> n = this.pbuilder.nullProcess();
                        n.addActionPrefixes(comp.getActionPrefixes());
                        r = n;
                        
                    } else if(comp.getProcesses().size() == 1) {
                        // composition with just one child process
                        final Process<ProcessData> child = comp.getProcesses().iterator().next();
                        final Process<ProcessData> copy = CommonTasks.copyTopLevel(child, this.pbuilder);
                        copy.addActionPrefixes(comp.getActionPrefixes());
                        copy.addRestrictions(comp.getRestrictions());
                        r = copy;
                        
                    } else {
                        // composition with more than one child process
                        r = comp;
                        done = true;
                    }
                    break;
                default:
                    done = true;
            }
        }
        
        return r;
    }
    
    
    /**
     * Restricts a composition of processes.
     *
     * @param p A composition of processes.
     * @return  The composition in restricted form.
     */
    private Process<ProcessData> restrictCompositionWithExtrusion(final ProcessComposition<ProcessData> p) {
        // new process composition
        ProcessComposition<ProcessData> r = this.pbuilder.newProcessComposition();
        
        this.copyActionPrefixes(p, r);
        r.setOperator(p.getOperator());
        
        if(r.getActionPrefixes().isEmpty()) {
            // there are no action prefixes so we can use the rule of
            // scope extrusion. this is gonna be complex.

            // for each bound name n we create a collection of child processes
            // c where n is not in freeNames(c).
            final Map<Name, List<Process<ProcessData>>> free = new HashMap<Name, List<Process<ProcessData>>>();
            
            // initialize the list of the sub-processes with all the sub-processes of p
            final List<Process<ProcessData>> processes = new ArrayList<Process<ProcessData>>(p.getProcesses());
            
            final Collection<Name> restrictions = new ArrayList<Name>(p.getRestrictions());
            
            // now, for each child process, we calculate the free names
            // and check which bound names are in the free names.
            for(int i = 0; i < processes.size(); i++) {
                final Process<ProcessData> sub = processes.get(i);
                if(sub.getType() == Process.Type.COMPOSITION) {
                    final ProcessComposition<ProcessData> comp = (ProcessComposition<ProcessData>)sub;
                    if(comp.getActionPrefixes().isEmpty()
                                    && (comp.getOperator() == p.getOperator())) {
                        processes.remove(i);
                        processes.addAll(i, comp.getProcesses());
                        restrictions.addAll(comp.getRestrictions());
                        --i;
                        continue;
                    }
                }
                
                final Collection<Name> subFN = this.freeNames(sub);
                for(final Name name : restrictions) {
                    if(subFN.contains(name)) {
                        // this name is bound by the restriction in p
                        // add this sub-process to the collection of processes
                        // that use that name
                        List<Process<ProcessData>> children = free.get(name);
                        if(children == null) {
                            children = new ArrayList<Process<ProcessData>>();
                            free.put(name, children);
                        } else if(!p.getOperator().isCommutative()) {
                            final Process<ProcessData> last = children.get(children.size()-1);
                            final int lastIndex = CommonTasks.indexOfThis(processes, last);
                            for(int j = lastIndex+1; j < i; ++j) {
                                children.add(processes.get(j));
                            }
                        }
                        children.add(sub);
                    }
                }
            }
            
            // find a process collection with minimal but positive number of processes
            List<Process<ProcessData>> minProcesses = null;
            Name minName = null;
            
            for(final Map.Entry<Name, List<Process<ProcessData>>> e : free.entrySet()) {
                final List<Process<ProcessData>> procs = e.getValue();
                if(procs == null) {
                    continue;
                }
                if((minProcesses == null) || (procs.size() < minProcesses.size())) {
                    minProcesses = procs;
                    minName = e.getKey();
                }
            }
            
            // something will be changed since we identified a non-empty minmal process collection
            // that does not contain all the processes
            boolean changed = ((minProcesses != null) && (minProcesses.size() != processes.size())) || (minProcesses==null && free.isEmpty());
            
            while((minProcesses != null) && (minProcesses.size() != processes.size())) {
                /* O.K. What do we have now?
                 * => for each restriction we have a collection of all the child
                 *    processes, which "use" that restriction.
                 *
                 * what is to do now?
                 * => create a new process composition containing the processes
                 *      - a composition of the minimal collection with the restriction
                 *      - all the other processes
                 *    we will recurse until the rule of scope extrusion can no
                 *    longer be applied.
                 */
                
                // we identified a minimal collection of processes minProcesses
                // that "use" the name minName.
                // these processes we put together and add a restriction.
                final ProcessComposition<ProcessData> coll = this.pbuilder.newProcessComposition();
                coll.setOperator(p.getOperator());
                coll.addRestriction(minName);
                coll.addProcesses(minProcesses);
                
                // remove the processes from the list
                final int firstIndex = CommonTasks.removeAllOfThese(processes, minProcesses);
                assert firstIndex >= 0 : "Does not make sense. Ask author.";
                
                // and add the new composition
                if(firstIndex >= processes.size()) {
                    processes.add(coll);
                } else {
                    processes.add(firstIndex, coll);
                }
                
                // that name is bound, so we may forget it at this level
                free.remove(minName);
                
                // now look for a new minimal collection
                // AND update the free map (remove minProcesses, add coll, where apropriate)
                List<Process<ProcessData>> newMinProcesses = null;
                Name newMinName = null;
                for(final Map.Entry<Name, List<Process<ProcessData>>> e : free.entrySet()) {
                    final List<Process<ProcessData>> procs = e.getValue();
                    if(procs == null) {
                        continue;
                    }
                    // if a collection contains a process in minProcesses,
                    // we remove that process and have to add coll lateron
                    int findex = -1;
                    int index = 0;
                    for(final Process<ProcessData> proc : minProcesses) {
                        final int cindex = CommonTasks.indexOfThis(index, procs, proc);
                        if(cindex >= 0) {
                            if(findex < 0) {
                                findex = cindex;
                            }
                            procs.remove(cindex);
                            index = cindex;
                            if(procs.isEmpty()) {
                                break;
                            }
                        }
                    }
                    if(findex >= 0) {
                        if(findex >= procs.size()) {
                            procs.add(coll);
                        } else {
                            procs.add(findex, coll);
                        }
                    } else if(procs.isEmpty()) {
                        e.setValue(null);
                        continue;
                    }
                    
                    // we iterate over the free map, so we may identify a
                    // minimal collection for the next iteration of the restriction
                    if((newMinProcesses == null) || (procs.size() < newMinProcesses.size())) {
                        newMinProcesses = procs;
                        newMinName = e.getKey();
                    }
                }
                // done updating free map and identifying a minimal collection
                minProcesses = newMinProcesses;
                minName = newMinName;
                
            } // end of while
            
            // we cannot use the rule of scope extrusion anymore
            
            // restrict all the child processes
            for(final Process<ProcessData> sub : processes) {
                final Process<ProcessData> subRestricted = this.restrict(sub);
                
                // and add them to the new process composition
                if(!CommonTasks.isBareNull(subRestricted)) {
                    // 0-processes without action prefixes will not be added
                    if(subRestricted.getType() == Process.Type.COMPOSITION) {
                        final ProcessComposition<ProcessData> comp = (ProcessComposition<ProcessData>)subRestricted;
                        if((comp.getOperator() == p.getOperator())
                                        && comp.getActionPrefixes().isEmpty()
                                        && comp.getRestrictions().isEmpty()) {
                            
                            r.addProcesses(comp.getProcesses());
                            changed = true;
                        } else {
                            r.addProcess(subRestricted);
                        }
                    } else {
                        r.addProcess(subRestricted);
                    }
                    if(sub != subRestricted) {
                        changed = true;
                    }
                } else {
                    changed = true;
                }
            }

            for(final Map.Entry<Name, List<Process<ProcessData>>> e : free.entrySet()) {
                final List<Process<ProcessData>> procs = e.getValue();
                if(procs == null) {
                    changed = true;
                } else {
                    r.addRestriction(e.getKey());
                }
            }
            
            if(changed) {
                return r;
            }
            return p;
                
        }
        boolean changed = false;
        
        // now restrict all the child processes
        CommonTasks.ensureOrdering(p);
        for(final Process<ProcessData> sub : p.getProcesses()) {
            final Process<ProcessData> subRestricted = this.restrict(sub);
            
            // and add them to the new process composition
            if(!CommonTasks.isBareNull(subRestricted)) {
                // 0-processes without action prefixes will not be added
                if(subRestricted.getType() == Process.Type.COMPOSITION) {
                    final ProcessComposition<ProcessData> comp = (ProcessComposition<ProcessData>)subRestricted;
                    if((comp.getOperator() == p.getOperator())
                                    && comp.getActionPrefixes().isEmpty()
                                    && comp.getRestrictions().isEmpty()) {
                        
                        r.addProcesses(comp.getProcesses());
                        changed = true;
                    } else {
                        r.addProcess(subRestricted);
                    }
                } else {
                    r.addProcess(subRestricted);
                }
                
                if(sub != subRestricted) {
                    changed = true;
                }
            } else {
                changed = true;
            }
        }
        
        if(changed) {
            // keep the restrictions
            r.addRestrictions(p.getRestrictions());
            return r;
        }
        return p;
    }
    
    
    /**
     * Restricts a choice composition of processes.
     *
     * @param p A choice composition of processes.
     * @return  The composition in restricted form.
     */
    private ProcessComposition<ProcessData> restrictCompositionNoExtrusion(final ProcessComposition<ProcessData> p) {
        // the restriction of this process is the sum of the restricted
        // child processes.
        
        // new process composition
        final ProcessComposition<ProcessData> r = this.pbuilder.newProcessComposition();
        
        boolean changed = false;
        boolean choice = p.getOperator() == ProcessComposition.Operator.CHOICE;
        
        CommonTasks.ensureOrdering(p);
        // now restrict all the child processes
        for(final Process<ProcessData> sub : p.getProcesses()) {
            final Process<ProcessData> subRestricted = this.restrict(sub);
            
            if(CommonTasks.isBareNull(subRestricted)) {
                // do not add 0
                changed = true;
            } else {
            	if(choice && (subRestricted.getType() == Process.Type.COMPOSITION)
            			&& !subRestricted.hasRealPrefixes()) {
            		throw new IllegalArgumentException("A choice composition "
            				+ "with another composition without prefix as an "
            				+ "operand is not supported.\nChoice: " + p
            				+ "\nOperand: " + sub
            				+ "\n(Restricted: " + subRestricted + ")");
            	}
            	
                // and add them to the new process composition
                r.addProcess(subRestricted);
                
                if(sub != subRestricted) {
                    changed = true;
                }
            }
        }
        
        if(changed) {
            // the same operator
            r.setOperator(p.getOperator());
            this.copyActionPrefixes(p, r);
            // and the same restrictions
            r.addRestrictions(p.getRestrictions());
            
            return r;
        }
        return p;
    }
    
    
    /**
     * TODO: document
     *
     * @param from
     * @param to
     * @param <T>
     */
    private <T> void copyActionPrefixes(final Process<T> from, final Process<T> to) {
        to.addActionPrefixes(from.getActionPrefixes());
        for(final Iterator<ActionPrefix> it = to.getActionPrefixes().iterator(); it.hasNext();) {
            final ActionPrefix a = it.next();
            if(a.getType() == ActionPrefix.Type.GUARD) {
                final Guard g = (Guard)a;
                if(g.isSatisfied()) {
                    it.remove();
                    to.changed();
                }
            }
        }
    }
    

    /**
     * Return the set of all names in the process that are not bound (the 
     * restrictions of this process are not counted but the restrictions of the
     * child processes!).
     *
     * @param p A process.
     * @return  The set of free names (see above)
     */
    private Collection<Name> freeNamesWOR(final Process<ProcessData> p) {
        return CommonTasks.freeNamesWOR(p);
    }
    
    
    private Collection<Name> freeNames(final Process<ProcessData> p) {
        return CommonTasks.freeNames(p);
    }
    
}
