/*
 * Project: Petruchio
 * File:    Retriever.java
 * Author:  Tim Strazny
 * Created: 15.11.2007
 */
package petruchio.compiler;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;

import petruchio.compiler.reactions.CommunicationReaction;
import petruchio.compiler.reactions.SingleReaction;
import petruchio.interfaces.ParserErrorHandler;
import petruchio.interfaces.algorithms.Reaction;
import petruchio.interfaces.algorithms.ReactionFinder;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReader;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.PiParser;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessDefinition;


/**
 * TODO: Document
 * 
 * @author Tim Strazny
 */
public class Retriever<E> implements petruchio.interfaces.Retriever<E> {
    
    private PiParser parser;
    
    private PetriNet net;
    
    private ProcessCreator pbuilder;
    
    private ReactionFinder<E> finder;
    
    private Collection<ParserErrorHandler> errorHandlers;
    
    private Map<Place, Process<E>> cache;

    
    /**
     * Well - it initializes the new instance.
     */
    public Retriever() {
        this.parser = null;
        this.net = null;
        this.pbuilder = null;
        this.finder = null;
        this.errorHandlers = new ArrayList<ParserErrorHandler>(0);
        this.cache = new IdentityHashMap<Place, Process<E>>(0);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Retriever#setParser(petruchio.interfaces.pi.PiParser)
     */
    @Override
    public void setParser(final PiParser parser) {
        this.parser = parser;
        if(this.parser != null) {
            this.parser.setProcessCreator(this.pbuilder);
            for(final ParserErrorHandler handler : this.errorHandlers) {
                this.parser.addErrorHandler(handler);
            }
        }
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.Retriever#addErrorHandler(petruchio.interfaces.ParserErrorHandler)
     */
    @Override
    public void addErrorHandler(final ParserErrorHandler handler) {
        this.errorHandlers.add(handler);
        if(this.parser != null) {
            this.parser.addErrorHandler(handler);
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Retriever#setReactionFinder(ReactionFinder)
     */
    @Override
    public void setReactionFinder(final ReactionFinder<E> finder) {
        this.finder = finder;
        if(this.finder != null) {
            this.finder.setProcessCreator(this.pbuilder);
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Retriever#retrieve(java.util.Map)
     */
    @Override
    public Process<E> retrieve(final Map<Place, Integer> marking) {
        if(marking == null || marking.isEmpty()) {
            return this.pbuilder.nullProcess();
        }
        
        final List<Process<E>> processes = new ArrayList<Process<E>>(marking.size());
        for(final Map.Entry<Place, Integer> e : marking.entrySet()) {
            final Process<E> p = this.getProcess(e.getKey());
            for(int i = e.getValue().intValue(); i > 0; --i) {
                processes.add(p);
            }
        }
        return this.compose(processes);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Retriever#retrieve()
     */
    @Override
    public Process<E> retrieve() {
        if(this.net == null) {
            return this.pbuilder.nullProcess();
        }
        
        final List<Process<E>> processes = new ArrayList<Process<E>>(1);
        for(final Place place : this.net.getPlaces()) {
            if(place.getMarking() == 0) {
                continue;
            }
            final Process<E> p = this.getProcess(place);
            for(int i = place.getMarking(); i > 0; --i) {
                processes.add(p);
            }
        }
        return this.compose(processes);
    }
    
    
    /**
     * Composes a bunch of processes in parallel and takes care of name clashes.
     *
     * @param <V>
     * @param processes
     * @return
     */
    private Process<E> compose(final List<Process<E>> processes) {
        if(processes.size() == 1) {
            return processes.iterator().next();
        }
        ProcessComposition<E> c = this.pbuilder.newProcessComposition();
        c.setOperator(ProcessComposition.Operator.PARALLEL);
        c.addProcesses(processes);
        return this.finder.uniqueBoundNames(c, null);
    }
    
    
    /**
     * Retrieves the process associated with a place.
     *
     * @param <V>
     * @param p
     * @return
     */
    private Process<E> getProcess(final Place p) {
        Process<E> proc = this.cache.get(p);
        if(proc == null) {
            final String prc = p.getMeaning();
            final int start = prc.indexOf(petruchio.interfaces.Retriever.MEANING_SEP);
            proc = this.parseProcess(prc.substring(start < 0 ? 0 : start+petruchio.interfaces.Retriever.MEANING_SEP.length()));
            this.cache.put(p, proc);
        }
        return proc;
    }
    
    
    /**
     * Parses a single process.
     *
     * @param <V>
     * @param prc
     * @return
     */
    private Process<E> parseProcess(final String prc) {
        try {
            final String proc = this.parser.getDummyProcessDefinitionPrefix()
                    + prc
                    + this.parser.getDummyProcessDefinitionSuffix();
            this.parser.setInput(new ByteArrayInputStream(proc.getBytes()));
            final ProcessDefinition<E> def = this.parser.nextProcessDefinition();
            return def.getProcess();
        } catch(final RuntimeException e) {
            throw new RuntimeException("Error parsing process: " + prc, e);
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Retriever#retrieve(java.util.List)
     */
    @Override
    public List<Step<E>> retrieve(final List<Transition> sequence)
            throws UnsupportedOperationException {
        final List<Step<E>> steps = new ArrayList<Step<E>>(sequence == null ? 1 : sequence.size()+1);
        
        final Map<Place, Integer> marking = new IdentityHashMap<Place, Integer>(this.net.getPlaces().size());
        for(final Place place : this.net.getPlaces()) {
            if(place.getMarking() != 0) {
                marking.put(place, Integer.valueOf(place.getMarking()));
                place.setMarking(0);
            }
        }
        
        Process<E> currentProcess = this.retrieve(marking);
        
        // initial process
        final SingleReaction<E> r = new SingleReaction<E>();
        r.setReactedProcess(currentProcess);
        r.setStepType(null);
        r.setNamesAfter(null);
        r.setNamesBefore(null);
        r.setProcessID(null);
        r.setSequentiallyIndependent(null);
        r.setTransition(null);
        final Step<E> init = new Step<E>(r, new IdentityHashMap<Place, Integer>(marking));
        steps.add(init);
        
        if(sequence == null || sequence.isEmpty()) {
            return steps;
        }
        
        for(final Transition t : sequence) {
            this.fire(t, marking);
            final Process<E> newProcess = this.retrieve(marking);
            
            final Reaction<E> reaction = this.getReaction(t, currentProcess, newProcess);
            final Step<E> step = new Step<E>(reaction, new IdentityHashMap<Place, Integer>(marking));
            steps.add(step);
        }
        
        for(final Map.Entry<Place, Integer> e : marking.entrySet()) {
            e.getKey().setMarking(e.getValue().intValue());
        }
        
        return steps;
    }
    
    
    /**
     * Still bogus.
     *
     * @param t
     * @param from
     * @param to
     * @return
     */
    private Reaction<E> getReaction(final Transition t, final Process<E> from,
            final Process<E> to) {
        final String xm = t.getMeaning();
        final int msep = xm.indexOf(petruchio.interfaces.Retriever.MEANING_SEP);
        final String m = xm.substring(msep < 0 ? 0 : msep+petruchio.interfaces.Retriever.MEANING_SEP.length());
        final Reaction<E> r;
        if(m.equals("t") || m.equals("tau")) {
            final SingleReaction<E> s = new SingleReaction<E>();
            s.setStepType(SingleReaction.StepType.TAU);
            r = s;
            
        } else if(!m.contains("|")) {
            final SingleReaction<E> s = new SingleReaction<E>();
            s.setStepType(SingleReaction.StepType.RECURSE);
            final int end = m.indexOf(')');
            final int start = m.lastIndexOf('(');
            
            final String pid = start < 0 ? m : m.substring(0, start);
            s.setProcessID(this.pbuilder.newProcessID(this.pbuilder.newName(pid.trim())));
            
            if(end > start + 1) {
                final String m2 = m.substring(start+1, end);
                final String[] n = m2.split(",\\s*");
                final List<Name> namesBefore = new ArrayList<Name>(n.length);
                final List<Name> namesAfter = new ArrayList<Name>(n.length);
                for(final String pair : n) {
                    if(pair.length() == 0) {
                        continue;
                    }
                    final int split = pair.indexOf('/');
                    final String after = pair.substring(0, split);
                    final String before = pair.substring(split+1);
                    namesAfter.add(this.pbuilder.newName(after.trim()));
                    namesBefore.add(this.pbuilder.newName(before.trim()));
                }
                s.setNamesAfter(namesAfter);
                s.setNamesBefore(namesBefore);
            } else {
                s.setNamesAfter(new ArrayList<Name>());
                s.setNamesBefore(new ArrayList<Name>());
            }
            
            r = s;
            
        } else {
            final CommunicationReaction<E> c = new CommunicationReaction<E>();
            final int beginRead = m.indexOf('(');
            final int endRead = m.lastIndexOf(')');
            final int beginSend = m.indexOf('<');
            final int endSend = m.lastIndexOf('>');
            
            final String channel = beginRead < 0 ? m : m.substring(0, Math.min(beginRead, beginSend));
            c.setChannel(this.pbuilder.newName(channel));
            
            if(endRead > beginRead + 1) {
                final String m2 = m.substring(beginRead+1, endRead);
                final String[] n = m2.split(",\\s*");
                final List<Name> names = new ArrayList<Name>(n.length);
                for(final String name : n) {
                    if(name.length() == 0) {
                        continue;
                    }
                    names.add(this.pbuilder.newName(name.trim()));
                }
                c.setNamesBefore(names);
            } else {
                c.setNamesBefore(new ArrayList<Name>());
            }
            
            if(endSend > beginSend + 1) {
                final String m2 = m.substring(beginSend+1, endSend);
                final String[] n = m2.split(",\\s*");
                final List<Name> names = new ArrayList<Name>(n.length);
                for(final String name : n) {
                    if(name.length() == 0) {
                        continue;
                    }
                    names.add(this.pbuilder.newName(name.trim()));
                }
                c.setNamesAfter(names);
            } else {
                c.setNamesAfter(new ArrayList<Name>());
            }
            
            r = c;
        }
        
        r.setTransition(t);
        r.setReactedProcess(to);
        return r;
    }
    
    
    /**
     * Simulates the firing of a transition.
     *
     * @param t
     * @param m
     */
    private void fire(final Transition t, final Map<Place, Integer> m) {
        for(final PTArc in : t.getInput()) {
            final Place pl = in.getSource();
            final Integer old = m.get(pl);
            if(old == null || old.intValue() < in.getWeight()) {
                throw new RuntimeException("Simulated firing of a transition that is not enabled: " + pl + " -> " + t + ": " + old +  " < " + in.getWeight()); 
            } else if(old.intValue() == in.getWeight()) {
                m.remove(pl);
            } else {
                m.put(pl, Integer.valueOf(old.intValue() - in.getWeight()));
            }
        }
        for(final TPArc out : t.getOutput()) {
            final Place pl = out.getTarget();
            final Integer old = m.put(pl, Integer.valueOf(out.getWeight()));
            if(old != null) {
                m.put(pl, Integer.valueOf(old.intValue() + out.getWeight()));
            }
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Retriever#setPetriNet(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public void setPetriNet(final PetriNet net) {
        this.net = net;
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Retriever#setProcessCreator(petruchio.interfaces.pi.ProcessCreator)
     */
    @Override
    public void setProcessCreator(final ProcessCreator processBuilder) {
        this.pbuilder = processBuilder;
        if(this.parser != null) {
            this.parser.setProcessCreator(this.pbuilder);
        }
        if(this.finder != null) {
            this.finder.setProcessCreator(this.pbuilder);
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        this.net = null;
        this.pbuilder = null;
        this.parser = null;
        this.finder = null;
        this.errorHandlers.clear();
        this.cache.clear();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "Default implementation of the retriever interface. Allows to "
            + "retrieve a reaction sequence of processes from a Petri net and "
            + "an enabled sequence of transitions.";
    }

    
    /**
     * The main method!
     *
     * @param args
     * @throws ClassNotFoundException 
     * @throws IllegalAccessException 
     * @throws InstantiationException 
     */
    public static void main(final String[] args) throws FileNotFoundException,
            InstantiationException, IllegalAccessException, ClassNotFoundException {
        final PetriNetReader reader = (PetriNetReader)Class.forName(args[0]).newInstance();
        final ParserErrorHandler handler = new petruchio.pi.readers.ConsoleErrorHandler();
        reader.addErrorHandler(handler);
        reader.setInput(new FileInputStream(Retriever.getAbsolutePath(args[1])));
        final PetriNet net = reader.read();
        final List<Transition> sequence = new ArrayList<Transition>();
        for(int i = 2; i < args.length; ++i) {
            final String[] trans = args[i].split(",\\s*");
            for(final String tr : trans) {
                sequence.add(net.getTransitions(tr).iterator().next());
            }
        }
        
        final ProcessCreator builder = new petruchio.pi.ProcessCreator();
        final ReactionFinder<ProcessData> finder = new petruchio.compiler.ReactionFinder();
        final PiParser parser = new petruchio.pi.readers.cup.PiParser();
        final Retriever<ProcessData> retriever = new Retriever<ProcessData>();
        retriever.setParser(parser);
        retriever.setProcessCreator(builder);
        retriever.setReactionFinder(finder);
        retriever.addErrorHandler(handler);
        retriever.setPetriNet(net);
        final List<Step<ProcessData>> seq = retriever.retrieve(sequence);
        for(final Step<?> s : seq) {
        	final Reaction<?> r = s.getReaction();
            if(r.getTransition() != null) {
                System.out.println("(" + r.getTransition().getName() + "): " + r);
            } else {
                System.out.println("(init): " + r.getReactedProcess());
            }
            System.out.println();
        }
        System.out.println("Target state of Petri net:");
        for(final Place p : net.getPlaces()) {
            if(p.getMarking() != 0) {
                System.out.println(p.getName() + " : " + p.getMarking());
            }
        }
    }
    
    
    /**
     * Builds an absolute path from a given file and the current user directory.
     *
     * @param file A relative or absolute path to the file
     * @return The absolute path for the given file name
     */
    private static String getAbsolutePath(final String file) {
        final File f = new File(file);
        if(!f.isAbsolute()) {
            final File f2 = new File(System.getProperty("user.dir")
                            + File.separator + file);
            try {
                return f2.getCanonicalPath();
            } catch(final IOException e) {
                return f2.getAbsolutePath();
            }
        }
        return file;
    }
    
}