/*
 * Project:  Petruchio
 * Filename: Semantics.java
 * Created:  18.04.2007
 * Author:   Tim Strazny
 */
package petruchio.compiler;

import java.util.HashMap;
import java.util.Map;

import petruchio.interfaces.SelfDescribing;


/**
 * Enum that holds the different semantics.
 * A semantics has a name, a suffix that is appended to the output file name
 * when compiling with multiple semantics and an author.
 * A semantics can be obtained via {@link #valueOf(String)} or via
 * {@link #getSemanticsFromSuffix(String)}.
 * 
 * @author Tim Strazny
 */
public enum Semantics implements SelfDescribing {
    STRUCTURAL("structural semantics", "str",
            "R. Meyer 2007",
            "Minimal communication structures are translated into places and reactions between them into transitions."),
    SEQUENTIAL("sequential semantics", "seq",
            "R.M. Amadio, C. Meyssonnier 2002, R. Meyer, T. Strazny 2007",
            "Sequential processes are translated into places and whenever a restriction is encountered it is resolved into a unique free name. Transitions are formed from the reactions between places."),
    SEQ_STR("combined (seq/struct) semantics", "seq_str",
            "R.M. Amadio, C. Meyssonnier 2002, R. Meyer, T. Strazny 2007 / R. Meyer 2007",
            "The sequential semantics is applied as possible and when a cause of infinitely many places is identified, a next run is performed and it will be handled using the structural semantics."),
    STR_SEQ("combined (struct/seq) semantics", "str_seq",
            "R. Meyer 2007 / R.M. Amadio, C. Meyssonnier 2002, R. Meyer, T. Strazny 2007",
            "The structural semantics is applied as possible and when a cause of infinitely many places is identified, a next run is performed and it will be handled using the sequential semantics.");
    
    /** A mapping from suffix to semantics */
    private static Map<String, Semantics> SUFFIXES;
    
    
    /** The semantics' name */
    private String name;
    
    /** The semantics' author(s) */
    private String author;
    
    /** The file name suffix for this semantics */
    private String suffix;
    
    private String description;
    
    
    /**
     * Initializes the new semantics.
     * 
     * @param name   The name of this semantics
     * @param suffix The file name suffix for this semantics
     * @param author The author of this semantics
     * @param descr  A short description
     */
    Semantics(final String name, final String suffix, final String author, final String descr) {
        this.name = name;
        this.author = author;
        this.description = descr;
        this.setSuffix(suffix);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return this.description;
    }
    
    
    /**
     * Sets the suffix of this semantics and adds it to the mapping.
     *
     * @param suffix The file name suffix of this semantics.
     * @see #getSemanticsFromSuffix(String) 
     */
    private void setSuffix(final String suffix) {
        if(Semantics.SUFFIXES == null) {
            Semantics.SUFFIXES = new HashMap<String, Semantics>();
        } else if(Semantics.SUFFIXES.put(suffix, this) != null) {
            throw new InternalError("Will not add two semantics with the same "
                            + "suffix: " + suffix);
        }
        this.suffix = suffix;
    }
    
    
    /**
     * This semantics' file name suffix
     *
     * @return The file name suffix of this semantics
     * @see #getSemanticsFromSuffix(String)
     */
    public String getSuffix() {
        return this.suffix;
    }
    
    
    /**
     * The name of this semantics
     *
     * @return The name of this semantics
     */
    public String getName() {
        return this.name;
    }
    
    
    /**
     * The author(s) of this semantics
     *
     * @return The author(s) of this semantics
     */
    public String getAuthor() {
        return this.author;
    }
    
    
    /**
     * Returns the semantics with the given suffix or <code>null</code> if there
     * is no semantics with that suffix.
     *
     * @param suffix A file name suffix
     * @return       The semantics with that file name suffix
     */
    public static Semantics getSemanticsFromSuffix(final String suffix) {
        return Semantics.SUFFIXES.get(suffix);
    }
    
}