/*
 * Project:  Petruchio
 * Filename: StructBoundChecker.java
 * Created:  24.08.2007
 * Author:   Tim Strazny
 */
package petruchio.compiler.sbcheckers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Stack;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.algorithms.ReactionFinder;
import petruchio.interfaces.algorithms.StructuralCongruence;
import petruchio.interfaces.algorithms.StructuralBoundedness;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;


/**
 * <Description>
 * 
 * @param <V> The data type of the processes
 * @author Tim Strazny
 */
public class SimpleUnboundedBreadthChecker<V> implements SelfDescribing, StructuralBoundedness<V> {
    
    private final Map<Transition, Map<Process<V>, Integer>> structInf = new IdentityHashMap<Transition, Map<Process<V>, Integer>>();
    
    private StructuralCongruence<V> structChecker = null;
    
    private ReactionFinder<V> reactionFinder = null;
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A test for identifying simple situations and causes when a process becomes not structurally stationary.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        this.clearCaches();
        this.structChecker = null;
        this.reactionFinder = null;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.StructuralBoundedness#clearCaches()
     */
    @Override
    public void clearCaches() {
    	this.structInf.clear();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.StructuralInfinityChecker#setStructuralCongruence(petruchio.interfaces.algorithms.StructuralCongruence)
     */
    @Override
    public void setStructuralCongruence(final StructuralCongruence<V> structCong) {
        this.structChecker = structCong;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.StructuralInfinityChecker#setReactionFinder(petruchio.interfaces.algorithms.ReactionFinder)
     */
    @Override
    public void setReactionFinder(final ReactionFinder<V> reactionFinder) {
        this.reactionFinder = reactionFinder;
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.StructuralBoundedness#causesStructuralUnboundedness(petruchio.interfaces.petrinet.Transition, java.util.Map, boolean)
     */
    @Override
    public Collection<Name> causesStructuralUnboundedness(
                    final Transition syncTrans,
                    final Map<Place, Process<V>> place2proc,
                    final boolean update) {
        Map<Process<V>, Integer> procs = this.structInf.get(syncTrans);
        if(procs == null) {
            procs = new IdentityHashMap<Process<V>, Integer>();
            this.structInf.put(syncTrans, procs);

            final Place p = syncTrans.getInput().iterator().next().getSource();
            Process<V> proc = place2proc.get(p);
            final Integer oldBreadth = procs.put(proc, Integer.valueOf(1));
            if((oldBreadth != null) && (oldBreadth.intValue() < 1)) {
                procs.put(proc, oldBreadth);
            }

            next_in:
            for(final TPArc in : p.getInput()) {
                final Transition t = in.getSource();
                if((t.getInput().size() == 1) && !t.getName().matches(">\\s*\\|")) { // TODO: Further improve
                    for(final PTArc arc : t.getInput()) {
                        if(arc.getWeight() != 1) {
                            continue next_in;
                        }
                    }
                    final Map<Process<V>, Integer> procs2 = this.structInf.get(t);
                    procs.putAll(procs2);
                }
            }
        }

        if(update) {
            final Collection<Name> names = this.checkStructInfinityUpdate(syncTrans, place2proc);
            return (names != null) ? names : this.checkStructInfinity(syncTrans, place2proc);
        }
        return this.checkStructInfinity(syncTrans, place2proc);
    }
    
    
    /**
     * TODO: document
     *
     * @param syncTrans
     * @param place2proc
     * @return
     */
    private Collection<Name> checkStructInfinityUpdate(final Transition syncTrans,
            final Map<Place, Process<V>> place2proc) {
        final Map<Process<V>, Integer> procs = this.structInf.get(syncTrans);
        final Stack<Transition> todo = new Stack<Transition>();
        todo.push(syncTrans);
        while(!todo.isEmpty()) {
            final Transition pre = todo.pop();
            for(final TPArc out: pre.getOutput()) {
                final Place p = out.getTarget();
                for(final PTArc out2: p.getOutput()) {
                    final Transition post = out2.getTarget();
                    if((post != syncTrans) && (post.getInput().size() == 1) && !post.getName().contains(">|")) { // TODO: Improve
                        final Map<Process<V>, Integer> procs2 = this.structInf.get(post);
                        boolean goOn = false;
                        for(final Map.Entry<Process<V>, Integer> e : procs2.entrySet()) {
                            final Integer oldBreadth = procs.put(e.getKey(), e.getValue());
                            if((oldBreadth != null) && (oldBreadth.intValue() < e.getValue().intValue())) {
                                procs.put(e.getKey(), oldBreadth);
                            } else if((oldBreadth == null) || (oldBreadth.intValue() > e.getValue().intValue())) {
                                goOn = true;
                            }
                        }
                        if(goOn) {
                            todo.push(post);
                            
                            final Collection<Name> names = this.checkStructInfinity(post, place2proc); 
                            if(names != null) {
                                return names;
                            }
                        }
                    }
                }
            }
        }
        
        return null;
    }
    
    
    /**
     * TODO: document
     *
     * @param syncTrans
     * @param place2proc
     * @return
     */
    private Collection<Name> checkStructInfinity(final Transition syncTrans,
            final Map<Place, Process<V>> place2proc) {
        final Stack<Process<V>> todo = new Stack<Process<V>>();
        final Stack<Integer> breadths = new Stack<Integer>();
        final Map<Process<V>, Integer> procs = this.structInf.get(syncTrans);
        
        for(final TPArc out : syncTrans.getOutput()) {
            final Place p = out.getTarget();
            Process<V> proc = place2proc.get(p);
            if(!proc.getActionPrefixes().isEmpty()) {
                continue;
            }
            if(proc.getType() == Process.Type.COMPOSITION) {
                final ProcessComposition<V> comp = (ProcessComposition<V>)proc;
                if(comp.getOperator() == ProcessComposition.Operator.PARALLEL) {
                    final Collection<Name> res = new ArrayList<Name>(comp.getRestrictions());
                    todo.addAll(comp.getProcesses());
                    for(int i = 0; i < comp.getProcesses().size(); ++i) {
                        breadths.add(Integer.valueOf(comp.getProcesses().size()));
                    }
                    while(!todo.isEmpty()) {
                        proc = todo.pop();
                        int breadth = breadths.pop().intValue();
                        if(proc.getType() == Process.Type.COMPOSITION) {
                            final ProcessComposition<V> comp2 = (ProcessComposition<V>)proc;
                            if(comp2.getOperator() == ProcessComposition.Operator.PARALLEL) {
                                res.addAll(comp2.getRestrictions());
                                int nbreadth = breadth + comp2.getProcesses().size() - 1;
                                for(int i = 0; i < comp2.getProcesses().size(); ++i) {
                                    breadths.add(Integer.valueOf(nbreadth));
                                }
                                todo.addAll(comp2.getProcesses());
                                continue;
                            }
                        }
                        
                        final Collection<Name> n = new ArrayList<Name>(this.reactionFinder.boundNames(proc));
                        n.addAll(this.reactionFinder.freeNames(proc));
                        proc = this.reactionFinder.copyTopLevel(proc);
                        proc.getRestrictions().addAll(res);
                        proc.getRestrictions().retainAll(n);
                        for(final Map.Entry<Process<V>, Integer> e : procs.entrySet()) {
                            if((breadth > e.getValue().intValue())
                                    && this.structChecker.structurallyCongruent(e.getKey(), proc)) {
                                return proc.getRestrictions();
                            }
                        }
                    }
                }
            }
        }
        return null;
    }

}
