/*
 * Project:  Petruchio
 * Filename: DefaultStructChecker.java
 * Created:  01.12.2006
 * Author:   Tim Strazny
 */
package petruchio.compiler.struct;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import petruchio.common.WeakIdentityHashMap;
import petruchio.compiler.CommonTasks;
import petruchio.compiler.ProcessData;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.algorithms.StructuralCongruence;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.Guard;
import petruchio.interfaces.pi.InputAction;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.NullProcess;
import petruchio.interfaces.pi.OutputAction;
import petruchio.interfaces.pi.Parameters;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessID;
import petruchio.interfaces.pi.ProcessReference;

/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class DefaultStructChecker implements SelfDescribing, StructuralCongruence<ProcessData> {
    
    /** */
    private final Map<ProcessComposition<ProcessData>, ProcessComposition<ProcessData>> flattened;
    
    /** The ProcessCreator to use. */
    private ProcessCreator pbuilder;
    
    
    /**
     * 
     */
    public DefaultStructChecker() {
        this.flattened = new WeakIdentityHashMap<ProcessComposition<ProcessData>, ProcessComposition<ProcessData>>();
        this.pbuilder = null;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "Default implementation to efficiently check structural congruence between processes.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
    	this.clearCaches();
        this.pbuilder = null;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.StructuralCongruence#clearCaches()
     */
    @Override
    public void clearCaches() {
    	this.flattened.clear();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.StructuralCongruence#setProcessCreator(petruchio.interfaces.pi.ProcessCreator)
     */
    @Override
    public void setProcessCreator(final ProcessCreator pbuilder) {
        this.pbuilder = pbuilder;
    }
    
    
    /**
     * @param p1
     * @param p2
     * @param alpha
     * @return
     */
    private boolean structurallyCongruent(final Process<ProcessData> p1,
            final Process<ProcessData> p2, final Collection<Name> bn1,
            final Collection<Name> bn2, final Map<Name, Name> alpha1,
            final Map<Name, Name> alpha2) {

        // a process is structurally congruent to itself. yes it is.
        if(p1 == p2) {
            return true;
        }
        
        if(!CommonTasks.equalIndexNoFN(p1, p2)) {
            if(debug) {
                System.out.println("index!");
            }
            return false;
        }
        
        final ProcessData p1d;
        final Map<Name, Name> curAlpha1;
        final Map<Name, Name> curAlpha2;
        if(alpha1 == null) {
            p1d = CommonTasks.getData(p1);
            
            final Boolean cong = p1d.getStructCong().get(p2); 
            if(cong != null) {
                return cong.booleanValue();
            }
            
            curAlpha1 = new HashMap<Name, Name>(0);
            curAlpha2 = new HashMap<Name, Name>(0);
            
        } else {
            p1d = null;
            curAlpha1 = alpha1;
            curAlpha2 = alpha2;
        }
        
        
        // look up failed.
        
        boolean congruent;
        
        // distinguish different comparisons
        if(this.actionPrefixesDifferent(p1, p2, bn1, bn2, curAlpha1, curAlpha2)) {
            if(debug) {
                System.out.println("prefixes!");
            }
            congruent = false;
        } else if(p1.getType() != p2.getType()) {
            if(debug) {
                System.out.println("type!");
            }
            congruent = false;
        } else {
            switch(p1.getType()) {
                case NULL:
                    congruent = this.structurallyCongruentNullProcess(
                                    (NullProcess<ProcessData>)p1,
                                    (NullProcess<ProcessData>)p2,
                                    bn1, bn2, curAlpha1, curAlpha2);
                    if(debug && !congruent) {
                        System.out.println(p1.getType() + " bn " + bn1 + " / " + bn2 + " alpha " + curAlpha1 + " / " + curAlpha2);
                    }
                    break;
                    
                case PREFIX:
                    congruent = this.structurallyCongruentPrefix(
                                    (PrefixProcess<ProcessData>)p1,
                                    (PrefixProcess<ProcessData>)p2,
                                    bn1, bn2, curAlpha1, curAlpha2);
                    if(debug && !congruent) {
                        System.out.println(p1.getType() + " bn " + bn1 + " / " + bn2 + " alpha " + curAlpha1 + " / " + curAlpha2);
                    }
                    break;
                    
                case REFERENCE:
                    congruent = this.structurallyCongruentReference(
                                    (ProcessReference<ProcessData>)p1,
                                    (ProcessReference<ProcessData>)p2,
                                    bn1, bn2, curAlpha1, curAlpha2);
                    if(debug && !congruent) {
                        System.out.println(p1.getType() + " bn " + bn1 + " / " + bn2 + " alpha " + curAlpha1 + " / " + curAlpha2);
                    }
                    break;
                    
                case COMPOSITION:
                    congruent = this.structurallyCongruentComposition(
                                    (ProcessComposition<ProcessData>)p1,
                                    (ProcessComposition<ProcessData>)p2,
                                    bn1, bn2, curAlpha1, curAlpha2);
                    if(debug && !congruent) {
                        System.out.println(p1.getType() + " bn " + bn1 + " / " + bn2 + " alpha " + curAlpha1 + " / " + curAlpha2);
                    }
                    break;
                    
                default:
                    throw new InternalError("Unexpected type of process: " + p1);
            }
        }
        
        if(congruent) {
            final Collection<Name> names = new HashSet<Name>(curAlpha1.values().size());
            for(final Name n : curAlpha1.values()) {
                if((bn2.contains(n) || bn1.contains(n)) && !names.add(n)) {
                    congruent = false;
                    break;
                }
            }
            if(debug && !congruent) {
                System.out.println("alpha 1!");
            }
            if(congruent) {
                names.clear();
                for(final Name n : curAlpha2.values()) {
                    if((bn2.contains(n) || bn1.contains(n)) && !names.add(n)) {
                        congruent = false;
                        break;
                    }
                }
                if(debug && !congruent) {
                    System.out.println("alpha 2!");
                }
            }
        }
        
        if(p1d != null) {
            final ProcessData p2d = CommonTasks.getData(p2);
            // store result
            p1d.getStructCong().put(p2, Boolean.valueOf(congruent));
            p2d.getStructCong().put(p1, Boolean.valueOf(congruent));
        }
        
        return congruent;
    }
    
    
    private static final boolean debug = false;
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.StructuralCongruence#structurallyCongruent(java.lang.Process, java.lang.Process)
     */
    @Override
    public boolean structurallyCongruent(final Process<ProcessData> p1, final Process<ProcessData> p2) {
//        debug = p1.toString().contains("r_1.tau") && p2.toString().contains("r.tau");
//        if(debug) {
//            System.out.println("checking");
//        }
        final boolean result = CommonTasks.compareNameList(CommonTasks.freeNames(p1), CommonTasks.freeNames(p2)) == 0
            && this.structurallyCongruent(p1, p2, this.boundNames(p1), this.boundNames(p2), null, null);
//        debug = false;
        return result;
    }
    
    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @param bn1
     * @param bn2
     * @param alpha1
     * @param alpha2
     * @return
     */
    private boolean structurallyCongruentNullProcess(
                    final NullProcess<ProcessData> p1,
                    final NullProcess<ProcessData> p2,
                    final Collection<Name> bn1,
                    final Collection<Name> bn2,
                    final Map<Name, Name> alpha1,
                    final Map<Name, Name> alpha2) {
        return true;
//        // can p1 be alpha converted to p2?
//        return this.unifyRestrictions(p1.getRestrictions(),
//                        p2.getRestrictions(), bn1, bn2, alpha1, alpha2);
    }
    
    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @param alpha
     * @return
     */
    private boolean structurallyCongruentPrefix(
                    final PrefixProcess<ProcessData> p1,
                    final PrefixProcess<ProcessData> p2,
                    final Collection<Name> bn1,
                    final Collection<Name> bn2,
                    final Map<Name, Name> alpha1,
                    final Map<Name, Name> alpha2) {
        return this.structurallyCongruent(p1.getProcess(), p2.getProcess(),
                        bn1, bn2, alpha1, alpha2);
//        // can p1 be alpha converted to p2?
//        return this.unifyRestrictions(p1.getRestrictions(),
//                        p2.getRestrictions(), bn1, bn2, alpha1, alpha2);
    }

    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @return
     */
    private boolean structurallyCongruentReference(
                    final ProcessReference<ProcessData> p1,
                    final ProcessReference<ProcessData> p2,
                    final Collection<Name> bn1,
                    final Collection<Name> bn2,
                    final Map<Name, Name> alpha1,
                    final Map<Name, Name> alpha2) {
        final ProcessID pid1 = p1.getProcessID();
        final ProcessID pid2 = p2.getProcessID();
        if(!pid1.getID().equals(pid2.getID())) {
            if(debug) {
                System.out.println("id!");
            }
            return false;
        }
        if(p1.getRestrictions().size() != p2.getRestrictions().size()) {
            if(debug) {
                System.out.println("restrictions!");
            }
            return false;
        }
        
        final Parameters params1 = p1.getParameters();
        final Parameters params2 = p2.getParameters();
        if(params1.getParameters().size() != params2.getParameters().size()) {
            if(debug) {
                System.out.println("parameters!");
            }
            return false;
        }
        for(int i = params1.getParameters().size()-1; i >= 0; --i) {
            final Name param1 = params1.getParameters().get(i);
            final Name param2 = params2.getParameters().get(i);
            if(!this.addToAlpha(alpha1, alpha2, bn1, bn2, param1, param2)) {
                if(debug) {
                    System.out.println("alpha!");
                }
                return false;
            }
        }
        return true;
        
//        return this.unifyRestrictions(p1.getRestrictions(),
//                        p2.getRestrictions(), bn1, bn2, alpha1, alpha2);
    }
    
    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @return
     */
    private boolean structurallyCongruentComposition(
                    final ProcessComposition<ProcessData> p1,
                    final ProcessComposition<ProcessData> p2,
                    final Collection<Name> bn1,
                    final Collection<Name> bn2,
                    final Map<Name, Name> alpha1,
                    final Map<Name, Name> alpha2) {
        if(p1.getOperator() != p2.getOperator()) {
            return false;
            
        } else if(p1.getOperator() == ProcessComposition.Operator.PARALLEL) {
            return this.structurallyCongruentCompositionParallel(p1, p2, bn1, bn2, alpha1, alpha2);
            
        } else if(p1.getOperator() == ProcessComposition.Operator.CHOICE) {
            return this.structurallyCongruentCompositionChoice(p1, p2, bn1, bn2, alpha1, alpha2);
            
        } else if(p1.getOperator() == ProcessComposition.Operator.SEQUENCE) {
            return this.structurallyCongruentCompositionSequence(p1, p2, bn1, bn2, alpha1, alpha2);
            
        } else {
            throw new InternalError("Unexpected operator: " + p1.getOperator());
            
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @param alpha1
     * @param alpha2
     * @return
     */
    private boolean structurallyCongruentCompositionParallel(
                    final ProcessComposition<ProcessData> p1,
                    final ProcessComposition<ProcessData> p2,
                    final Collection<Name> bn1,
                    final Collection<Name> bn2,
                    final Map<Name, Name> alpha1,
                    final Map<Name, Name> alpha2) {
        final ProcessComposition<ProcessData> p1Flat = this.flatten(p1);
        final ProcessComposition<ProcessData> p2Flat = this.flatten(p2);
        final int processCount = p1Flat.getProcesses().size();
        if(p2Flat.getProcesses().size() != processCount) {
            if(debug) {
                System.out.println("proc count");
            }
            return false;
        }
        CommonTasks.ensureOrdering(p1Flat);
        CommonTasks.ensureOrdering(p2Flat);
        final boolean congruent = this.processCollectionsStructurallyCongruent(
                        p1Flat.getProcesses(), 0,
                        p2Flat.getProcesses(), new boolean[processCount],
                        bn1, bn2, alpha1, alpha2);
//               && this.unifyRestrictions(p1Flat.getRestrictions(),
//                               p2Flat.getRestrictions(), bn1, bn2, alpha1, alpha2);
        return congruent;
    }
    
    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @param alpha1
     * @param alpha2
     * @return
     */
    private boolean structurallyCongruentCompositionChoice(
                    final ProcessComposition<ProcessData> p1,
                    final ProcessComposition<ProcessData> p2,
                    final Collection<Name> bn1,
                    final Collection<Name> bn2,
                    final Map<Name, Name> alpha1,
                    final Map<Name, Name> alpha2) {
        final int processCount = p1.getProcesses().size();
        if(p2.getProcesses().size() != processCount) {
            return false;
        }
        
        CommonTasks.ensureOrdering(p1);
        CommonTasks.ensureOrdering(p2);
        
        return this.processCollectionsStructurallyCongruent(
                        p1.getProcesses(), 0,
                        p2.getProcesses(), new boolean[processCount],
                        bn1, bn2, alpha1, alpha2);
//                && this.unifyRestrictions(p1.getRestrictions(),
//                                p2.getRestrictions(), bn1, bn2, alpha1, alpha2);
    }
    
    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @param alpha1
     * @param alpha2
     * @return
     */
    private boolean structurallyCongruentCompositionSequence(
                    final ProcessComposition<ProcessData> p1,
                    final ProcessComposition<ProcessData> p2,
                    final Collection<Name> bn1,
                    final Collection<Name> bn2,
                    final Map<Name, Name> alpha1,
                    final Map<Name, Name> alpha2) {
        final ProcessComposition<ProcessData> p1Flat = this.flatten(p1);
        final ProcessComposition<ProcessData> p2Flat = this.flatten(p2);
        final int processCount = p1Flat.getProcesses().size();
        if(p2Flat.getProcesses().size() != processCount) {
            return false;
        }
        CommonTasks.ensureOrdering(p1Flat);
        CommonTasks.ensureOrdering(p2Flat);
        
        for(int i = 0; i < processCount; ++i) {
            final Process<ProcessData> a = p1Flat.getProcesses().get(i);
            final Process<ProcessData> b = p2Flat.getProcesses().get(i);
            
            if(!this.structurallyCongruent(a, b, bn1, bn2, alpha1, alpha2)) {
                return false;
            }
        }
        return true;
//        return this.unifyRestrictions(p1Flat.getRestrictions(),
//                                p2Flat.getRestrictions(), bn1, bn2, alpha1, alpha2);
    }
    
    
    /**
     * @param l1
     * @param i
     * @param l2
     * @param used
     * @return
     */
    private boolean processCollectionsStructurallyCongruent(
            final List<Process<ProcessData>> l1, final int i,
            final List<Process<ProcessData>> l2,
            final boolean[] used, final Collection<Name> bn1,
            final Collection<Name> bn2,
            final Map<Name, Name> alpha1, final Map<Name, Name> alpha2) {

        if(i >= used.length) {
            return true;
        }
        
        final Process<ProcessData> sub1 = l1.get(i);
        for(int k = 0; k < used.length; ++k) {
            if(!used[k]) {
                final Process<ProcessData> sub2 = l2.get(k);
                final int comparison = CommonTasks.PCOMPARATORNOFN.compare(sub1, sub2);
                if(comparison == 0) {
                    final Map<Name, Name> copyAlpha1 = new HashMap<Name, Name>(alpha1);
                    final Map<Name, Name> copyAlpha2 = new HashMap<Name, Name>(alpha2);
                    if(this.structurallyCongruent(sub1, sub2, bn1, bn2, copyAlpha1, copyAlpha2)) {
                        used[k] = true;
                        if(this.processCollectionsStructurallyCongruent(l1, i+1, l2,
                                        used, bn1, bn2, copyAlpha1, copyAlpha2)) {
                            alpha1.putAll(copyAlpha1);
                            alpha2.putAll(copyAlpha2);
                            return true;
                        }
                        used[k] = false;
                    }
                } else if(comparison < 0) {
                    break;
                }
            }
        }
        
        return false;
    }
    

//    /**
//     * TODO: document
//     *
//     * @param r1
//     * @param r2
//     * @param alpha
//     * @return
//     */
//    private boolean unifyRestrictions(final Collection<String> r1,
//                    final Collection<String> r2, final Collection<String> bn1,
//                    final Collection<String> bn2, final Map<String, String> alpha1,
//                    final Map<String, String> alpha2) {
//        if(r1.size() != r2.size()) {
//            return false;
//        }
//        if(r1.isEmpty()) {
//            return true;
//        }
//        
//        // "unify" the restrictions
//        final Collection<String> restrAlpha = new ArrayList<String>();
//        final List<String> leftOvers1 = new ArrayList<String>();
//        for(final String n : r1) {
//            final String repl = alpha1.get(n);
//            final String name = (repl == null) ? n : repl;
//            restrAlpha.add(name);
//            if(!r2.contains(name)) {
//                leftOvers1.add(name);
//            }
//        }
//        if(leftOvers1.isEmpty()) {
//            return true;
//        }
//        final List<String> leftOvers2 = new ArrayList<String>();
//        for(final String name : r2) {
//            if(!restrAlpha.contains(name)) {
//                leftOvers2.add(name);
//            }
//        }
//        
//        // TODO bogus
//        for(int i = leftOvers1.size()-1; i >= 0; --i) {
//            final String n1 = leftOvers1.get(i);
//            final String n2 = leftOvers2.get(i);
//            if(!this.addToAlpha(alpha1, alpha2, bn1, bn2, n1, n2)) {
//                return false;
//            }
//        }
//        return true;
//    }
    
    
    /**
     * TODO: document
     *
     * @param p1
     * @param p2
     * @return
     */
    private boolean actionPrefixesDifferent(final Process<ProcessData> p1,
                    final Process<ProcessData> p2, final Collection<Name> bn1,
                    final Collection<Name> bn2, final Map<Name, Name> alpha1,
                    final Map<Name, Name> alpha2) {
        final List<ActionPrefix> prefixes1;
        final List<ActionPrefix> prefixes2;
        
        boolean containsGuard = false;
        for(final ActionPrefix a : p1.getActionPrefixes()) {
            if(a.getType() == ActionPrefix.Type.GUARD) {
                containsGuard = true;
                break;
            }
        }
        if(containsGuard) {
            prefixes1 = new ArrayList<ActionPrefix>(p1.getActionPrefixes());
            for(final Iterator<ActionPrefix> it = prefixes1.iterator(); it.hasNext();) {
                final ActionPrefix a = it.next();
                if(a.getType() == ActionPrefix.Type.GUARD) {
                    final Guard g = (Guard)a;
                    if(g.isSatisfied()) {
                        it.remove();
                    }
                }
            }
        } else {
            prefixes1 = p1.getActionPrefixes();
        }
        
        containsGuard = false;
        for(final ActionPrefix a : p2.getActionPrefixes()) {
            if(a.getType() == ActionPrefix.Type.GUARD) {
                containsGuard = true;
                break;
            }
        }
        
        if(containsGuard) {
            prefixes2 = new ArrayList<ActionPrefix>(p2.getActionPrefixes());
            for(final Iterator<ActionPrefix> it = prefixes2.iterator(); it.hasNext();) {
                final ActionPrefix a = it.next();
                if(a.getType() == ActionPrefix.Type.GUARD) {
                    final Guard g = (Guard)a;
                    if(g.isSatisfied()) {
                        it.remove();
                    }
                }
            }
        } else {
            prefixes2 = p2.getActionPrefixes();
        }
        
        if(prefixes1.size() != prefixes2.size()) {
            return true;
        }
        
        for(int i = prefixes1.size()-1; i >= 0; i--) {
            final ActionPrefix ap1 = prefixes1.get(i);
            final ActionPrefix ap2 = prefixes2.get(i);
            if(ap1.getType() != ap2.getType()) {
                return true;
            }
            switch(ap1.getType()) {
                case INTERNAL:
                    break;
                case INPUT: {
                    final InputAction in1 = (InputAction)ap1;
                    final InputAction in2 = (InputAction)ap2;
                    if(in1.getNames().size() != in2.getNames().size()) {
                        return true;
                    }
                    
                    if(!this.addToAlpha(alpha1, alpha2, bn1, bn2, in1.getChannel(), in2.getChannel())) {
                        return true;
                    }
                    for(int j = 0; j < in1.getNames().size(); ++j) {
                        final Name n1 = in1.getNames().getParameters().get(j);
                        final Name n2 = in2.getNames().getParameters().get(j);
                        if(!this.addToAlpha(alpha1, alpha2, bn1, bn2, n1, n2)) {
                            return true;
                        }
                    }
                    break;
                }
                case OUTPUT: {
                    final OutputAction out1 = (OutputAction)ap1;
                    final OutputAction out2 = (OutputAction)ap2;
                    if(out1.getNames().size() != out2.getNames().size()) {
                        return true;
                    }
                    
                    if(!this.addToAlpha(alpha1, alpha2, bn1, bn2, out1.getChannel(), out2.getChannel())) {
                        return true;
                    }
                    for(int j = 0; j < out1.getNames().size(); ++j) {
                        final Name n1 = out1.getNames().getParameters().get(j);
                        final Name n2 = out2.getNames().getParameters().get(j);
                        if(!this.addToAlpha(alpha1, alpha2, bn1, bn2, n1, n2)) {
                            return true;
                        }
                    }
                    break;
                }
                case GUARD: {
                    final Guard g1 = (Guard)ap1;
                    final Guard g2 = (Guard)ap2;
                    // [g1l = g1r], [g2l = g2r]
                    
                    final Map<Name, Name> copyAlpha1 = new HashMap<Name, Name>(alpha1);
                    final Map<Name, Name> copyAlpha2 = new HashMap<Name, Name>(alpha2);
                    if(!this.addToAlpha(copyAlpha1, copyAlpha2, bn1, bn2, g1.leftOperand(), g2.leftOperand())
                                    || !this.addToAlpha(copyAlpha1, copyAlpha2, bn1, bn2, g1.rightOperand(), g2.rightOperand())) {
                        if(!this.addToAlpha(alpha1, alpha2, bn1, bn2, g1.leftOperand(), g2.rightOperand())
                                    || !this.addToAlpha(alpha1, alpha2, bn1, bn2, g1.rightOperand(), g2.leftOperand())) {
                            return true;
                        }
                    } else {
                        alpha1.putAll(copyAlpha1);
                        alpha2.putAll(copyAlpha2);
                    }
                    break;
                }
                default:
                    throw new InternalError("Unknown type of prefix: " + ap1);
            }
        }
        
        return false;
    }

    
    /**
     * TODO: document
     *
     * @param alpha
     * @param bn1
     * @param bn2
     * @param a
     * @param b
     * @return
     */
    private boolean addToAlpha(final Map<Name, Name> alpha1,
                    final Map<Name, Name> alpha2,
                    final Collection<Name> bn1, final Collection<Name> bn2,
                    final Name a, final Name b) {
        final boolean b1 = bn1.contains(a);
        final boolean b2 = bn2.contains(b);
        if(!b1 && !b2) {
            return a.equals(b);
        } else if(b1 && b2) {
            if(this.addToAlpha(alpha1, alpha2, a, b)) {
                return true;
            }
            return this.addToAlpha(alpha2, alpha1, b, a);
        } else {
            return false;
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param alpha
     * @param a
     * @param b
     * @return
     */
    private boolean addToAlpha(final Map<Name, Name> alpha1,
                    final Map<Name, Name> alpha2,
                    final Name a, final Name b) {
        final Name l = alpha2.get(b);
        final Name nub = (l == null) ? b : l;
        final Name oldValue = alpha1.put(a, nub);
        if((oldValue != null) && !nub.equals(oldValue)) {
            alpha1.put(a, oldValue);
            return false;
        }
        return true;
    }
    
    
    /**
     * @param p
     * @return
     */
    private ProcessComposition<ProcessData> flatten(final ProcessComposition<ProcessData> p) {
        ProcessComposition<ProcessData> flat = this.flattened.get(p);
        
        if(flat == null) {
            flat = CommonTasks.flatten(p, this.pbuilder);
            this.flattened.put(p, flat);
        }
        
        return flat;
    }
    
    
    /**
     * TODO: document
     *
     * @param p
     * @return
     */
    private Collection<Name> boundNames(final Process<ProcessData> p) {
        return CommonTasks.getData(p).getBn();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.StructuralCongruence#getComparator()
     */
    @Override
    public Comparator<Process<ProcessData>> getComparator() {
        return CommonTasks.PCOMPARATOR;
    }
    
}