/*
 * Project:  Petruchio
 * Filename: GraphIsomorphismStructChecker.java
 * Created:  13.01.2010
 * Author:   Tim Strazny
 */
package petruchio.compiler.struct;

import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;

import petruchio.common.WeakIdentityHashMap;
import petruchio.compiler.CommonTasks;
import petruchio.compiler.ProcessData;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.algorithms.StructuralCongruence;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.Guard;
import petruchio.interfaces.pi.InputAction;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.OutputAction;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessID;
import petruchio.interfaces.pi.ProcessReference;

/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public abstract class GraphIsomorphismStructChecker implements SelfDescribing, StructuralCongruence<ProcessData> {

    private final Map<Process<ProcessData>, LDigraph<String>> ldgCache = new WeakIdentityHashMap<Process<ProcessData>, LDigraph<String>>();
    
    private static final String COLOR_SEND = "(snd)";
    private static final String COLOR_RECEIVE = "(rec)";
    private static final String COLOR_GUARD = "(grd)";
    private static final String COLOR_PREFIX = "(pref)";
    private static final String COLOR_CHOICE = "(choice)";
    private static final String COLOR_PARALLEL = "(par)";
    private static final String COLOR_SEQUENCE = "(seq)";
    private static final String COLOR_TAU = "(tau)";
    private static final String COLOR_STOP = "(stop)";
    private static final String COLOR_RHO_V = "(rhov)";
    private static final String COLOR_NU_V = "(nuv)";
    private static final String COLOR_NU = "(nu)";
    
    private static final LDigraph<String> STOP_GRAPH = new LDigraph<String>(COLOR_STOP);
    private static final LDigraph<String> TAU_GRAPH = new LDigraph<String>(COLOR_TAU);
    
    private final Map<ProcessComposition<ProcessData>, ProcessComposition<ProcessData>> flattened
        = new WeakIdentityHashMap<ProcessComposition<ProcessData>, ProcessComposition<ProcessData>>();
    private ProcessCreator pbuilder = null;
    
    /**
     * @param a
     * @param b
     * @return
     */
    public abstract <T extends Comparable<T>> boolean existsIsomorphism(LDigraph<T> a, LDigraph<T> b);
    
    @Override
    public String getDescription() {
        return "Implementation to efficiently check structural congruence between processes using reduction to a graph isomorphism problem.";
    }

    @Override
    public Comparator<Process<ProcessData>> getComparator() {
        return CommonTasks.PCOMPARATOR;
    }
    
    @Override
    public void clearCaches() {
        this.ldgCache.clear();
        this.flattened.clear();
    }
    
    @Override
    public void setProcessCreator(final ProcessCreator pbuilder) {
        this.pbuilder = pbuilder;
    }

    @Override
    public boolean structurallyCongruent(final Process<ProcessData> p1, final Process<ProcessData> p2) {
        // a process is structurally congruent to itself. yes it is.
        if(p1 == p2) {
            return true;
        }
        
        // this checks fn1.equal(fn2) as well
        if(!CommonTasks.equalIndex(p1, p2)) {
            return false;
        }
        
        final LDigraph<String> g1 = this.getLDigraph(p1);
        final LDigraph<String> g2 = this.getLDigraph(p2);
        return (g1.getSize() == g2.getSize()) && this.existsIsomorphism(g1, g2);
    }

    @Override
    public void reset() {
        this.clearCaches();
        this.pbuilder = null;
    }
    
    private ProcessComposition<ProcessData> flatten(final ProcessComposition<ProcessData> p) {
        ProcessComposition<ProcessData> flat = this.flattened.get(p);
        
        if(flat == null) {
            flat = CommonTasks.flatten(p, this.pbuilder);
            this.flattened.put(p, flat);
        }
        
        return flat;
    }
    
    /**
     * @param p
     * @return
     */
    private LDigraph<String> getLDigraph(final Process<ProcessData> p) {
        LDigraph<String> ldag = this.ldgCache.get(p);
        if(ldag == null) {
            ldag = this.buildLDigraph(p, new HashMap<Name, LDigraph<String>>());
            this.ldgCache.put(p, ldag);
        }
        return ldag;
    }
    
    private static String getOperandColor(final int nth) {
        return "(" + nth + ")";
    }
    
    private static String getReferenceColor(final ProcessID pid) {
        return "((" + pid + "))";
    }
    
    private static LDigraph<String> getName(final Name name, final Map<Name, LDigraph<String>> restrictions) {
        final LDigraph<String> n = restrictions.get(name);
        return n == null ? new LDigraph<String>("" + name) : n;
    }
    
    /**
     * @param p
     * @return
     */
    private LDigraph<String> buildLDigraph(final Process<ProcessData> proc, final Map<Name, LDigraph<String>> restrictions) {
        final Process<ProcessData> p;
        if(proc.getType() == Process.Type.COMPOSITION) {
            final ProcessComposition<ProcessData> comp = (ProcessComposition<ProcessData>)proc;
            switch(comp.getOperator()) {
                case SEQUENCE:
                case PARALLEL:
                    p = flatten(comp);
                    break;
                default:
                    p = proc;
                    break;
            }
        } else {
            p = proc;
        }
        
        LDigraph<String> g = null;
        int size = 0;
        if(!p.getRestrictions().isEmpty()) {
            ++size;
            g = new LDigraph<String>(COLOR_NU);
            for(final Name n : p.getRestrictions()) {
                ++size;
                final LDigraph<String> r = new LDigraph<String>(COLOR_NU_V);
                g.addChild(r);
                
                restrictions.put(n, r);
            }
        }
        
        if(!p.getActionPrefixes().isEmpty()) {
            ++size;
            LDigraph<String> gc = new LDigraph<String>(COLOR_PREFIX);
            if(g != null) {
                g.addChild(gc);
            } else {
                g = gc;
            }
            int i = 1;
            for(final ActionPrefix a : p.getActionPrefixes()) {
                ++size;
                final LDigraph<String> c = new LDigraph<String>(getOperandColor(i));
                gc.addChild(c);
                switch(a.getType()) {
                    case INTERNAL:
                        c.addChild(TAU_GRAPH);
                        break;
                        
                    case OUTPUT: {
                        final OutputAction out = (OutputAction)a;
                        ++size;
                        final LDigraph<String> send = new LDigraph<String>(COLOR_SEND);
                        c.addChild(send);
                        ++size;
                        final LDigraph<String> chan = new LDigraph<String>(getOperandColor(1));
                        send.addChild(chan);
                        chan.addChild(getName(out.getChannel(), restrictions));
                        int j = 2;
                        for(final Name n : out.getNames().getParameters()) {
                            ++size;
                            final LDigraph<String> param = new LDigraph<String>(getOperandColor(j));
                            send.addChild(param);
                            param.addChild(getName(n, restrictions));
                        }
                        break;
                    }
                    
                    case INPUT: {
                        final InputAction in = (InputAction)a;
                        ++size;
                        final LDigraph<String> receive = new LDigraph<String>(COLOR_RECEIVE);
                        c.addChild(receive);
                        ++size;
                        final LDigraph<String> chan = new LDigraph<String>(getOperandColor(1));
                        receive.addChild(chan);
                        chan.addChild(getName(in.getChannel(), restrictions));
                        int j = 2;
                        for(final Name n : in.getNames().getParameters()) {
                            ++size;
                            final LDigraph<String> param = new LDigraph<String>(getOperandColor(j));
                            receive.addChild(param);
                            ++size;
                            final LDigraph<String> rec = new LDigraph<String>(COLOR_RHO_V);
                            param.addChild(rec);
                            
                            restrictions.put(n, rec);
                        }
                        break;
                    }
                    
                    case GUARD: {
                        final Guard guard = (Guard)a;
                        ++size;
                        final LDigraph<String> gg = new LDigraph<String>(COLOR_GUARD);
                        gg.addChild(getName(guard.leftOperand(), restrictions));
                        gg.addChild(getName(guard.rightOperand(), restrictions));
                        c.addChild(gg);
                        break;
                    }
                    
                    default:
                        throw new InternalError("Unknown action: " + a);
                }
                ++i;
            }
        }
        
        final LDigraph<String> gc;
        switch(p.getType()) {
            case NULL:
                gc = null;
                //gc = STOP_GRAPH;
                break;
                
            case PREFIX: {
                final PrefixProcess<ProcessData> pp = (PrefixProcess<ProcessData>)p;
                ++size;
                gc = new LDigraph<String>(COLOR_PREFIX);
                final LDigraph<String> l = this.buildLDigraph(pp.getProcess(), restrictions);
                size += l.getSize();
                gc.addChild(l);
                break;
            }
                
            case REFERENCE: {
                final ProcessReference<ProcessData> r = (ProcessReference<ProcessData>)p;
                ++size;
                gc = new LDigraph<String>(getReferenceColor(r.getProcessID()));
                int i = 1;
                for(final Name n : r.getParameters().getParameters()) {
                    ++size;
                    final LDigraph<String> c = new LDigraph<String>(getOperandColor(i));
                    c.addChild(getName(n, restrictions));
                    gc.addChild(c);
                    ++i;
                }
                break;
            }
            
            case COMPOSITION: {
                final ProcessComposition<ProcessData> pc = (ProcessComposition<ProcessData>)p;
                switch(pc.getOperator()) {
                    case CHOICE:
                        gc = new LDigraph<String>(COLOR_CHOICE);
                        break;
                    case PARALLEL:
                        gc = new LDigraph<String>(COLOR_PARALLEL);
                        break;
                    case SEQUENCE:
                        gc = new LDigraph<String>(COLOR_SEQUENCE);
                        break;
                    default:
                        throw new InternalError("Unknown process composition operator: " + pc);
                }
                if(pc.getOperator().isCommutative()) {
                    for(final Process<ProcessData> subp : pc.getProcesses()) {
                        final LDigraph<String> l = this.buildLDigraph(subp, restrictions);
                        size += l.getSize();
                        gc.addChild(l);
                    }
                } else {
                    int i = 1;
                    for(final Process<ProcessData> subp : pc.getProcesses()) {
                        ++size;
                        final LDigraph<String> c = new LDigraph<String>(getOperandColor(i));
                        final LDigraph<String> l = this.buildLDigraph(subp, restrictions);
                        size += l.getSize();
                        c.addChild(l);
                        gc.addChild(c);
                        ++i;
                    }
                }
                break;
            }
                
            default:
                throw new InternalError("Unexpected type of process: " + p);
        }
        
        if(g != null && gc != null) {
            g.addChild(gc);
        } else {
            g = gc;
        }
         
        if(g != null) {
            g.setSize(size);
            return g;
        }
        return STOP_GRAPH;
    }

}
