/*
 * Project:  Petruchio
 * Filename: NautyStructChecker.java
 * Created:  13.01.2010
 * Author:   Tim Strazny
 */
package petruchio.compiler.struct;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;
import java.util.SortedMap;
import java.util.TreeMap;

import petruchio.common.IdentityHashSet;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class NautyStructChecker extends GraphIsomorphismStructChecker {

    public static final String DREADNAUT_PROPERTY = "gi.nauty.dreadnaut";
    private static final String DEFAULT_DREADNAUT = "/home/user/critter/space/opt/nauty24/dreadnaut";
    private final Dreadnaut dreadnaut;
    
    public NautyStructChecker() {
        final String dreadnautExec = System.getProperty(DREADNAUT_PROPERTY, DEFAULT_DREADNAUT);
        this.dreadnaut = new Dreadnaut(dreadnautExec);
    }
    
    @Override
    public String getDescription() {
        return "Implementation to efficiently check structural congruence between processes using reduction to a graph isomorphism problem and calling the nauty tool.";
    }

    @Override
    public <T extends Comparable<T>> boolean existsIsomorphism(final LDigraph<T> a, final LDigraph<T> b) {
        this.dreadnaut.getReady();
        // enter graph a
        this.dreadnaut.firstGraph(a);
        // enter graph b
        this.dreadnaut.secondGraph(b);
        // read result
        boolean result = this.dreadnaut.checkIsomorphism();
        // clean up
        this.dreadnaut.finish();
        return result;
    }

    
    private static class Dreadnaut {
        final String cmd;
        Process proc = null;
        BufferedWriter out = null;
        File output = null;
//        long magicNumber = 0;
        
        Dreadnaut(final String executable) {
            this.cmd = executable;
        }
        
        private void open() {
            try {
                this.output = File.createTempFile("petruchio_dreadnaut_out", null);
                this.output.deleteOnExit();
                
                this.proc = Runtime.getRuntime().exec(this.cmd);
                this.out = new BufferedWriter(new OutputStreamWriter(this.proc.getOutputStream()));
                
                // redirect dreadnaut output to file
                this.enter(">> " + this.output.getCanonicalPath());
            } catch (final IOException e) {
                throw new RuntimeException("Unable to start dreadnaut. Have you correctly set property '"
                        + DREADNAUT_PROPERTY + "'? Current value is '" + this.cmd + "'", e);
            }
            this.enter("-a"); // writeautoms = FALSE
            this.enter("-m"); // writemarkers = FALSE 
            this.enter("d"); // digraph = TRUE
        }
        
        void getReady() {
            this.open();
        }
        
        <T extends Comparable<T>> void firstGraph(final LDigraph<T> g) {
            this.enter(g);
        }
        
        <T extends Comparable<T>> void secondGraph(final LDigraph<T> g) {
            this.enter("c"); // getcanon = TRUE
            this.enter("x"); // execute nauty, results in h
            this.enter("@"); // copy h to h'
            this.enter(g);
        }
        
        void enter(final String s) {
            this.enter(s, true);
        }
        
        /**
         * @param s
         * @param skipOutput
         */
        void enter(final String s, final boolean skipOutput) {
            try {
                this.out.write(s);
                this.out.write('\r');
                this.out.flush();
//                System.out.println("::: " + s);
            } catch(final IOException e) {
                throw new RuntimeException(e);
            }
        }
        
        <T extends Comparable<T>> void enter(final LDigraph<T> g) {
            this.enter("n=" + g.getSize());
            this.enter("g"); // read graph g
            
            // attention: partions have to be in right order!
            final SortedMap<T, Collection<LDigraph<T>>> partition = new TreeMap<T, Collection<LDigraph<T>>>();
            final Queue<LDigraph<T>> todo = new LinkedList<LDigraph<T>>();
            final Collection<LDigraph<T>> processed = new IdentityHashSet<LDigraph<T>>();
            processed.add(g);
            g.setIndex(0);
            todo.add(g);
            
            int i = 1;
            while(!todo.isEmpty()) {
                final LDigraph<T> cur = todo.remove();
                
                // save labelling to partition
                Collection<LDigraph<T>> part = partition.get(cur.getLabel());
                if(part == null) {
                    part = new ArrayList<LDigraph<T>>();
                    partition.put(cur.getLabel(), part);
                }
                part.add(cur);
                
                // process current vertex
                this.enter(cur.getIndex() + ":"); // make cur the current vertex
                for(final LDigraph<T> next : cur) {
                    if(processed.add(next)) { // next wasn't processed yet
                        next.setIndex(i++);
                        todo.add(next);
                    }
                    this.enter("" + next.getIndex()); // add edge from cur to next
                }
            }
            this.enter("."); // stop (reading graph)
            
            // compile partition
            final StringBuilder pi = new StringBuilder("[");
            for(final Iterator<Collection<LDigraph<T>>> it = partition.values().iterator(); it.hasNext(); ) {
                final Collection<LDigraph<T>> c = it.next();
                for(final Iterator<LDigraph<T>> it2 = c.iterator(); it2.hasNext(); ) {
                    pi.append(it2.next().getIndex());
                    if(it2.hasNext()) {
                        pi.append(",");
                    }
                }
                if(it.hasNext()) {
                    pi.append("|");
                }
            }
            pi.append("]");
            this.enter("f=" + pi); // specify initial partition
        }
        
        boolean checkIsomorphism() {
//            this.magicNumber++;
            this.enter("x"); // execute nauty, results in h
//            this.enter("\"petruchio magic number " + this.magicNumber + "\""); // mark output
            this.enter("#", false); // compare labelled graphs h and h' 
            this.enter("q"); // quit
            // dreadnaut will query nauty
            // nauty will answer either "h and h' are identical."
            // or "h and h' are different."
            // or "h and h' have different sizes."
            try {
                this.proc.waitFor(); // wait for termination
            } catch(final InterruptedException e) {
                Thread.currentThread().interrupt();
            }
            // read results
            try {
                final BufferedReader in = new BufferedReader(new FileReader(this.output));
                String result;
//                do {
//                    result = in.readLine();
//                } while(result != null && !result.contains("petruchio magic number " + this.magicNumber));
                do {
                    result = in.readLine();
                } while(result != null && !result.contains("h and h'"));
                try { in.close(); } catch(final IOException e) { /* ignored */ }
                
                if(result == null) {
                    throw new InternalError("End of output without answer from nauty.");
                } else if(result.contains("different")) {
                    return false;
                } else if(result.contains("identical")) {
                    return true;
                } else {
                    throw new InternalError("Unrecognised answer from nauty: " + result);
                }
            } catch(final IOException e) {
                throw new RuntimeException(e);
            }
        }
        
        void finish() {
            this.close();
        }
        
        private void close() {
            if(this.proc != null) {
                this.proc.destroy();
            }
            if(!this.output.delete()) { System.out.println("Unable to delete file " + this.output); }
            try { this.out.close(); } catch(final IOException e) { /* ignored */ }
        }
    }
    
}
