package petruchio.cov;

import java.util.AbstractCollection;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;

import petruchio.common.IdentityHashSet;
import petruchio.compiler.reactions.SingleReactionTask;
import petruchio.interfaces.CommunicatorCache;
import petruchio.interfaces.algorithms.Coverability;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.PlacePair;
import petruchio.interfaces.petrinet.Transition;
import petruchio.interfaces.petrinet.TPArc;

/**
 * @author critter
 *
 * @param <V>
 */
@ThreadSafe
public abstract class AbstractMultiThreadedCoverability implements Coverability {

    private static final Transition NO_MORE_TRANSITIONS = new Transition() {
        @Override public String getName() { /* nothing */ return null; }
        @Override public void setName(String name) { /* nothing */ }
        @Override public String getMeaning() { /* nothing */ return null; }
        @Override public void setMeaning(String meaning) { /* nothing */ }
        @Override public int getX() { /* nothing */ return 0; }
        @Override public int getY() { /* nothing */ return 0; }
        @Override public void setX(int x) { /* nothing */ }
        @Override public void setY(int y) { /* nothing */ }
        @Override public Collection<PTArc> getInput() { /* nothing */ return null; }
        @Override public Collection<TPArc> getOutput() { /* nothing */ return null; }
    };
    
    private Lock transitionsLock = new ReentrantLock();
    private Condition transitionPresent = this.transitionsLock.newCondition();
    private Lock resultsLock = new ReentrantLock();
    private Condition resultsPresent = this.resultsLock.newCondition();
    private Condition computationDone = this.resultsLock.newCondition();
    
    private BlockingQueue<Transition> newTransitions = new LinkedBlockingQueue<Transition>();
    
    private BlockingQueue<PlacePair> communications = null;
    
    @GuardedBy("executorLock")
    private ExecutorService executor = Executors.newSingleThreadExecutor();

    private int nThreads = 0;
    private int openExecutorTasks = 0;
    private final Lock executorLock = new ReentrantLock();
    private final Condition noOpenTasks = this.executorLock.newCondition();
    private final Condition newOpenTasks = this.executorLock.newCondition();
    
    @GuardedBy("itself")
    private final Collection<Place> knownPlaces = new IdentityHashSet<Place>();
    @GuardedBy("immutable")
    private final Collection<Place> knownPlaces_unmodifiable = Collections.unmodifiableCollection(this.knownPlaces);
    @GuardedBy("knownPlaces")
    private SimpleList<Place> knownFriendlyPlaces = null;
//    @GuardedBy("itself")
//    private final Collection<Place> allPlaces = new IdentityHashSet<Place>();
    
    @GuardedBy("itself")
    private final Collection<Transition> knownTransitions = new IdentityHashSet<Transition>();
    @GuardedBy("immutable")
    private final Collection<Transition> knownTransitions_unmodifiable = Collections.unmodifiableCollection(this.knownTransitions);
    @GuardedBy("knownTransitions")
    private SimpleList<Transition> knownFriendlyTransitions = null;
//    @GuardedBy("itself")
//    private final Collection<Transition> allTransitions = new IdentityHashSet<Transition>();
    
    private volatile boolean isWorking = false;
    private volatile boolean stopRequest = false;
    private volatile boolean terminationRequest = false;
    private volatile boolean hasTerminated = false;
    
    /**
     * The communicator cache object which delivers possible communication partners
     * to places. The object is fed by the compiler.
     */
    private volatile CommunicatorCache<? extends Object> communicators = null;
    
    /**
     * @param newThreads
     */
    @Override
    public void setThreadNumber(final int newThreads) {
        final int newNumThreads = newThreads < 1 ? 1 : newThreads;
        this.executorLock.lock();
        try {
            if(this.executor != null && newNumThreads != this.nThreads) {
                this.executor.shutdown();
                this.awaitNoOpenTasks();
            }
            if(this.executor == null || this.executor.isShutdown()) {
                if(newNumThreads <= 1) {
                    this.nThreads = 1;
                    this.executor = Executors.newSingleThreadExecutor();
                } else {
                    this.nThreads = newNumThreads;
                    this.executor = Executors.newFixedThreadPool(newNumThreads);
                }
            }
        } finally {
            this.executorLock.unlock();
        }
    }
    
    /**
     * @param command
     */
    final void dispatchRunnableTask(final Runnable command) {
        if(this.nThreads > 1) {
            final Runnable wrapper = new Runnable() {
                @Override
                public void run() {
                    try {
                        command.run();
                    } finally {
                        AbstractMultiThreadedCoverability.this.taskTerminated();
                    }
                }
            };
            this.executorLock.lock();
            try {
                ++this.openExecutorTasks;
                this.executor.execute(wrapper);
                this.signalNewOpenTasks();
            } finally {
                this.executorLock.unlock();
            }
        } else {
            command.run();
        }
    }
    
    /**
     * @param command
     */
    final <V> Future<V> dispatchCallableTask(final Callable<V> command) {
        final Callable<V> wrapper = new Callable<V>() {
            @Override
            public V call() {
                try {
                    return command.call();
                } catch (final Exception e) {
                    throw AbstractMultiThreadedCoverability.launderThrowable(e);
                } finally {
                    AbstractMultiThreadedCoverability.this.taskTerminated();
                }
            }
        };
        this.executorLock.lock();
        try {
            ++this.openExecutorTasks;
            final Future<V> result = this.executor.submit(wrapper);
            this.signalNewOpenTasks();
            return result;
        } finally {
            this.executorLock.unlock();
        }
    }
    
    /**
     * @param t
     * @return
     */
    public static RuntimeException launderThrowable(final Throwable t) {
        if(t instanceof RuntimeException) {
            return (RuntimeException)t;
        } else if(t instanceof Error) {
            throw (Error)t;
        }
        throw new IllegalStateException("Not checked", t);
    }
    
    
    /**
     * 
     */
    final void taskTerminated() {
        this.executorLock.lock();
        try {
            if(--this.openExecutorTasks == 0) {
                this.signalNoOpenTasks();
            }
        } finally {
            this.executorLock.unlock();
        }
    }
    
    /**
     * 
     */
    public final void awaitNoOpenTasks() {
        if(this.openExecutorTasks != 0) {
            this.executorLock.lock();
            try {
                while(this.openExecutorTasks != 0) {
                    this.noOpenTasks.await();
                }
            } catch(final InterruptedException e) {
                Thread.currentThread().interrupt();
            } finally {
                this.executorLock.unlock();
            }
        }
    }
    
    /**
     * 
     */
    private void signalNoOpenTasks() {
        this.executorLock.lock();
        try {
            this.noOpenTasks.signalAll();
        } finally {
            this.executorLock.unlock();
        }
    }
    
    /**
     * 
     */
    public final void awaitNewOpenTasks() {
        if(this.openExecutorTasks == 0) {
            this.executorLock.lock();
            try {
                while(this.openExecutorTasks == 0) {
                    this.newOpenTasks.await();
                }
            } catch(final InterruptedException e) {
                Thread.currentThread().interrupt();
            } finally {
                this.executorLock.unlock();
            }
        }
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#compilerIsWaiting()
     */
    @Override
    public void compilerIsWaiting() {
        this.addTransition(AbstractMultiThreadedCoverability.NO_MORE_TRANSITIONS);
    }
    
    /**
     * 
     */
    private void signalNewOpenTasks() {
        this.executorLock.lock();
        try {
            this.newOpenTasks.signalAll();
        } finally {
            this.executorLock.unlock();
        }
    }
    
    /**
     * @param pair
     */
    public final void addPlacePair(final PlacePair pair) {
        assert pair != null;
        
        if(this.communications == null) {
            return;
        }
        
        try {
            this.communications.put(pair);
            this.signalResultsPresent();
        } catch (final InterruptedException e) {
            Thread.currentThread().interrupt();
        }
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public final void reset() {
        this.isWorking = false;
        this.stopRequest = false;
        this.newTransitions.clear();
        this.terminationRequest = false;
        synchronized(this.knownPlaces) {
            this.knownPlaces.clear();   
            if(this.knownFriendlyPlaces != null) {
                for(final Place p : this.knownFriendlyPlaces) {
                    ((petruchio.pn.Place)p).known = false;
                }
                this.knownFriendlyPlaces = null;
            }
        }
        synchronized(this.knownTransitions) {
            this.knownTransitions.clear();
            if(this.knownFriendlyTransitions != null) {
                for(final Transition t : this.knownFriendlyTransitions) {
                    ((petruchio.pn.Transition)t).known = false;
                }
                this.knownFriendlyTransitions = null;
            }
        }
//        synchronized(this.allPlaces) {
//            this.allPlaces.clear();            
//        }
//        synchronized(this.allTransitions) {
//            this.allTransitions.clear();
//        }
        
        this.resultsLock.lock();
        try {
            this.doReset();
        } finally {
            this.resultsLock.unlock();
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#addTransition(petruchio.interfaces.petrinet.Transition)
     */
    @Override
    public final void addTransition(final Transition t) {
        assert t != null;
        
        synchronized(this.newTransitions) {
            if(this.stopRequest) {
                return;
            }
            this.isWorking = true;
            try {
                this.newTransitions.put(t);
            } catch (final InterruptedException e) {
                Thread.currentThread().interrupt();
            }
        }
        this.signalNewTransition();
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#hasTerminated()
     */
    @Override
    public final boolean hasTerminated() {
        return this.hasTerminated;
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#isWorking()
     */
    @Override
    public final boolean isWorking() {
        return this.isWorking;
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#pleaseStopComputation()
     */
    @Override
    public final void pleaseStopComputation() {
        synchronized(this.newTransitions) {
            this.stopRequest = true;
            this.newTransitions.clear();
        }
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#pleaseTerminate()
     */
    @Override
    public final void pleaseTerminate() {
        this.terminationRequest = true;
        this.pleaseStopComputation();
        this.signalNewTransition();
        this.terminationCleanUp();
        this.executorLock.lock();
        try {
            if(this.executor != null) {
                this.executor.shutdownNow();
            }
        } finally {
            this.executorLock.unlock();
        }
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#setCommunications(java.util.concurrent.BlockingQueue)
     */
    @Override
    public final void setCommunications(final BlockingQueue<PlacePair> c) {
        this.communications = c;
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#supportsThreading()
     */
    @Override
    public boolean supportsThreading() {
        return true;
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#waitForComputationEnd()
     */
    @Override
    public final void awaitComputationEnd() {
        if(!this.stopRequest && this.isWorking) {
            this.resultsLock.lock();
            try {
                while(!this.stopRequest && this.isWorking) {
                    this.computationDone.await();
                }
            } catch(final InterruptedException e) {
                Thread.currentThread().interrupt();
            } finally {
                this.resultsLock.unlock();
            }
        }
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#waitForResults()
     */
    @Override
    public final void awaitResults() {
        if(!this.stopRequest && this.isWorking) {
            this.resultsLock.lock();
            try {
                while(!this.stopRequest && this.isWorking && this.communications.isEmpty()) {
                    this.resultsPresent.await();
                }
            } catch(final InterruptedException e) {
                Thread.currentThread().interrupt();
            } finally {
                this.resultsLock.unlock();
            }
        }
    }
    
    
    /**
     * 
     */
    private void awaitNewTransition() {
        if(!this.terminationRequest && this.newTransitions.isEmpty()) {
            this.transitionsLock.lock();
            try {
                while(!this.terminationRequest && this.newTransitions.isEmpty()) {
                    this.transitionPresent.await();
                }
            } catch(final InterruptedException e) {
                Thread.currentThread().interrupt();
            } finally {
                this.transitionsLock.unlock();
            }
        }
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run()
     */
    @Override
    public final void run() {
        this.hasTerminated = false;
        try {
            while(!this.terminationRequest) {
                this.awaitNewTransition();
                if(this.terminationRequest) {
                    break;
                }
                final Transition t = this.newTransitions.take();
                if(t == AbstractMultiThreadedCoverability.NO_MORE_TRANSITIONS) {
                    this.processCollectedTransitions();
                } else {
                    if(this.supportsThreading()) {
                        synchronized(this.knownPlaces) {
                            for(final TPArc a : t.getOutput()) {
                                final Place p = a.getTarget();
                                if(p instanceof petruchio.pn.Place) {
                                    if(!((petruchio.pn.Place)p).known) {
                                        ((petruchio.pn.Place)p).known = true;
                                        this.knownFriendlyPlaces = new SimpleList<Place>(p, this.knownFriendlyPlaces);
                                    }
                                } else {
                                    this.knownPlaces.add(a.getTarget());
                                }
                            }
                        }
                        synchronized(this.knownTransitions) {
                            if(t instanceof petruchio.pn.Transition) {
                                if(!((petruchio.pn.Transition)t).known) {
                                    ((petruchio.pn.Transition)t).known = true;
                                    this.knownFriendlyTransitions = new SimpleList<Transition>(t, this.knownFriendlyTransitions);
                                }
                            } else {
                                this.knownTransitions.add(t);
                            }
                        }
                    }
                    this.processTransition(t);
                }
                if(this.stopRequest || this.newTransitions.isEmpty()) {
                    this.isWorking = false;
                    this.signalComputationDone();
                }
            }
            this.isWorking = false;
            this.signalComputationDone();
        } catch (final InterruptedException e) {
            this.hasTerminated = true;
            Thread.currentThread().interrupt();
        } catch(final Throwable e) {
            this.hasTerminated = true;
            e.printStackTrace();
            this.signalComputationDone();
        } finally {
            this.reset();
        }
    }
    
    /**
     * @return
     */
    public Collection<Transition> getAccessibleTransitions() {
        return new JointCollection<Transition>(this.knownTransitions_unmodifiable, this.knownFriendlyTransitions);
    }
    
    public Collection<Place> getAccessiblePlaces() {
        return new JointCollection<Place>(this.knownPlaces_unmodifiable, this.knownFriendlyPlaces);
    }
    
    
    static class JointCollection<E> extends AbstractCollection<E> {
        private final Collection<E> colA;
        private final SimpleList<E> colB;
        public JointCollection(final Collection<E> colA, final SimpleList<E> colB) {
            this.colA = colA;
            this.colB = colB;
        }

        @Override
        public int size() {
            return (this.colA == null ? 0 : this.colA.size()) + (this.colB == null ? 0 : this.colB.getLength());
        }

        @Override
        public Iterator<E> iterator() {
            return new JointIterator<E>(this.colA, this.colB);
        }

    }
    
    static class MetaIterator<E> implements Iterator<E> {
        Iterator<? extends Iterable<E>> it;
        Iterator<E> current;
        public <T extends Iterable<? extends Iterable<E>>> MetaIterator(final T col) {
            this.it = col.iterator();
            this.current = null;
        }
        
        @Override
        public boolean hasNext() {
            if(this.current != null && this.current.hasNext()) {
                return true;
            }
            while(this.it.hasNext()) {
                this.current = this.it.next().iterator();
                if(this.current.hasNext()) {
                    return true;
                }
            }
            return false;
        }

        @Override
        public E next() {
            return this.current.next();
        }

        @Override
        public void remove() {
            this.current.remove();
        }
    }
    
    static class JointIterator<E> implements Iterator<E> {
        Iterator<E> it;
        Iterable<E> second;
        boolean first;
        public <T extends Iterable<E>> JointIterator(final T colA, final T colB) {
            this.it = colA.iterator();
            this.second = colB;
            this.first = true;
        }
        
        @Override
        public boolean hasNext() {
            if(this.it.hasNext()) {
                return true;
            }
            if(this.first) {
                this.first = false;
                if(this.second == null) {
                    return false;
                }
                this.it = this.second.iterator();
                this.second = null;
            }
            return this.it.hasNext();
        }

        @Override
        public E next() {
            return this.it.next();
        }

        @Override
        public void remove() {
            this.it.remove();
        }
    }
    
    public boolean hasMoreTransitionsToProcess() {
        return !this.newTransitions.isEmpty();
    }
    
    /**
     * 
     */
    private void signalComputationDone() {
        this.resultsLock.lock();
        try {
            this.computationDone.signalAll();
        } finally {
            this.resultsLock.unlock();
        }
        this.signalResultsPresent();
    }
    
    
    /**
     * 
     */
    private void signalResultsPresent() {
        this.resultsLock.lock();
        try {
            this.resultsPresent.signalAll();
        } finally {
            this.resultsLock.unlock();
        }
    }
    
    
    /**
     * 
     */
    private void signalNewTransition() {
        this.transitionsLock.lock();
        try {
            this.transitionPresent.signalAll();
        } finally {
            this.transitionsLock.unlock();
        }
    }
    
    
    /**
     * @param t
     */
    abstract void processTransition(Transition t);
    
    abstract void processCollectedTransitions();
    
    abstract void doReset();
    
    abstract void doClearCaches();
    
    abstract void terminationCleanUp();
    
    abstract void doSetStartState(Collection<Place> startState);
    
    @Override
    public synchronized final void clearCaches() {
        this.pleaseStopComputation();
        this.awaitComputationEnd();
        
        this.doClearCaches();
        
        synchronized(this.knownPlaces) {
            this.knownPlaces.clear();
            if(this.knownFriendlyPlaces != null) {
                for(final Place p : this.knownFriendlyPlaces) {
                    ((petruchio.pn.Place)p).known = false;
                }
                this.knownFriendlyPlaces = null;
            }
        }
        synchronized(this.knownTransitions) {
            this.knownTransitions.clear();
            if(this.knownFriendlyTransitions != null) {
                for(final Transition t : this.knownFriendlyTransitions) {
                    ((petruchio.pn.Transition)t).known = false;
                }
                this.knownFriendlyTransitions = null;
            }
        }
        synchronized(this.newTransitions) {
            this.isWorking = false;
            this.stopRequest = false;
            this.newTransitions.clear();
        }
        this.terminationRequest = false;
        this.hasTerminated = false;
    }
    
    /**
     * @return
     */
    final boolean stopRequested() {
//        return this.supportsThreading() && (this.stopRequest || this.terminationRequest);
        return this.stopRequest || this.terminationRequest;
    }

    
    /**
     * @param p
     * @return
     */
    final boolean placeAccessible(final Place p) {
        if(!this.supportsThreading()) {
            return true;
        }
        if(p instanceof petruchio.pn.Place) {
            return ((petruchio.pn.Place)p).known;
        }
        synchronized(this.knownPlaces) {
            return this.knownPlaces.contains(p);
        }
    }
    
    
    /**
     * @param p
     * @return
     */
    final boolean transitionAccessible(final Transition t) {
        if(!this.supportsThreading()) {
            return true;
        }
        if(t instanceof petruchio.pn.Transition) {
            return ((petruchio.pn.Transition)t).known;
        }
        synchronized(this.knownTransitions) {
            return this.knownTransitions.contains(t);
        }
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#setCommunicatorCache(petruchio.interfaces.CommunicatorCache)
     */
    @Override
    public synchronized void setCommunicatorCache(final CommunicatorCache<? extends Object> communicators) {
        this.communicators = communicators;
    }
    
    /**
     * @param p
     * @return
     */
    Collection<Place> getCommunicators(final Place p) {
        final Collection<Place> c = new ArrayList<Place>();
        for(final SingleReactionTask<?> t : this.communicators.getCommunicators(p)) {
            c.add(t.getPlace());
        }
        return c;
    }
    
    @Override
    public final void setStartState(final Collection<Place> startState) {
        synchronized(this.knownPlaces) {
            for(final Place p : startState) {
                if(p instanceof petruchio.pn.Place) {
                    if(!((petruchio.pn.Place)p).known) {
                        ((petruchio.pn.Place)p).known = true;
                        this.knownFriendlyPlaces = new SimpleList<Place>(p, this.knownFriendlyPlaces);
                    }
                } else {
                    this.knownPlaces.add(p);
                }
            }
        }
        this.doSetStartState(startState);
    }
    
    /**
     * @param p1
     * @param p2
     * @return
     */
    boolean canCommunicate(final Place p1, final Place p2) {
        return this.communicators.canCommunicate(p1, p2);
    }
    
    public int countKnownTransitions() {
        return (this.knownFriendlyTransitions == null ? 0 : this.knownFriendlyTransitions.getLength()) + this.knownTransitions.size();
    }
    
    public int countKnownPlaces() {
        return (this.knownFriendlyPlaces == null ? 0 : this.knownFriendlyPlaces.getLength()) + this.knownPlaces.size();
    }
    
}
