/*
 * Project:  Petruchio
 * Filename: Coverability.java
 * Created:  18.02.2007
 * Author:   Tim Strazny
 */
package petruchio.cov;

import java.io.OutputStream;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.Stack;

import petruchio.common.IdentityHashSet;
import petruchio.common.WeakIdentityHashMap;
import petruchio.interfaces.CommunicatorCache;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PlacePair;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;

/**
 * On-the-fly coverability graph computation.
 * 
 * @author Tim Strazny
 */
public class Coverability extends AbstractMultiThreadedCoverability implements SelfDescribing {

    private static final StructuredNode REPLACES_NODES = StructuredNode.newNode();
	
    private static final Transition NEXT_UNUSED_PLACE = new Transition() {
        @Override
        public int getX() { throw new InternalError(); }
        @Override
        public int getY() { throw new InternalError(); }
        @Override
        public void setX(int x) { throw new InternalError(); }
        @Override
        public void setY(int y) { throw new InternalError(); }
        @Override
        public void setName(String name) { throw new InternalError(); }
        @Override
        public String getName() { throw new InternalError(); }
        @Override
        public Collection<PTArc> getInput() { throw new InternalError(); }
        @Override
        public Collection<TPArc> getOutput() { throw new InternalError(); }
		@Override
        public String getMeaning() { throw new InternalError(); }
		@Override
        public void setMeaning(String meaning) { throw new InternalError(); }
    };
    
    
    private final TransitionComparator transitionComparator = new TransitionComparator();
        
    /** Dynamically growing table where the 'markable' pairs are filled in. */
    private final Map<Place, Set<Place>> markable;

    /** The transitions known by this class. */
    private final List<Transition> transitions;

    /** Cache which maps a transition to its pre set as a node. */
    private final Map<Transition, StructuredNode> inStructuredNodes;

    /** Cache which maps a transition to its post set as a node. */
    private final Map<Transition, StructuredNode> outStructuredNodes;
    
    /** StructuredNodes that have been created and have to be processes and added the graph. */
    private final LinkedList<StructuredNode> newStructuredNodes;
    
    private final Collection<StructuredNode> newCovered;
    
    private final Queue<Place> unusedPlaces;
    
    private final Map<Place, Integer> boundsRemovedPlaces;
    
    /** The nodes in the coverability graph. */
    private final Map<StructuredNode, StructuredNode> nodes;
    
    /** The start node. */
    private StructuredNode root;

    /** A re-used collection */
    private final Collection<StructuredNode> visited;
    
    private final Collection<PlacePair> doNotCheck;

    /** */
    private volatile boolean checkMarkable;
    
    private volatile int minTokens;
    
    private volatile int maxTokens;
    
    /**
     * Initializes the new instance.
     */
    public Coverability() {
        this.checkMarkable = false;
        this.minTokens = 1;
        this.maxTokens = -1;
        
        this.doNotCheck = new IdentityHashSet<PlacePair>();
        this.root = StructuredNode.newNode();
        this.visited = new IdentityHashSet<StructuredNode>();
        this.markable = new IdentityHashMap<Place, Set<Place>>();
        this.inStructuredNodes = new WeakIdentityHashMap<Transition, StructuredNode>();
        this.outStructuredNodes = new WeakIdentityHashMap<Transition, StructuredNode>();
        this.transitions = new LinkedList<Transition>();
        this.newStructuredNodes = new LinkedList<StructuredNode>();
        this.newCovered = new IdentityHashSet<StructuredNode>();
        this.nodes = new HashMap<StructuredNode, StructuredNode>();
        this.nodes.put(this.root, this.root);
        this.boundsRemovedPlaces = new IdentityHashMap<Place, Integer>();
        this.unusedPlaces = new LinkedList<Place>();
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "The default implementation of a coverability checker using a reduced coverability graph.";
    }
    
        
    /* (non-Javadoc)
     * @see petruchio.cov.AbstractMultiThreadedCoverability#doReset()
     */
    @Override
    synchronized void doReset() {
        this.clearCaches();
        this.checkMarkable = false;
        this.minTokens = 1;
        this.maxTokens = -1;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#doNotCheck(petruchio.interfaces.petrinet.Place, petruchio.interfaces.petrinet.Place)
     */
    public void doNotCheck(final Place p1, final Place p2) {
        this.doNotCheck.add(new petruchio.cov.PlacePair(p1, p2));
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#clearCaches()
     */
    @Override
    void doClearCaches() {
        for(final StructuredNode n : this.nodes.values()) {
            n.clear();
        }
        this.nodes.clear();
        this.root = StructuredNode.newNode();
        this.nodes.put(this.root, this.root);
        
        this.setCheckMarkable(false);

        this.doNotCheck.clear();
        this.markable.clear();
        this.transitions.clear();
        this.newStructuredNodes.clear();
        this.newCovered.clear();
        this.transitionComparator.reset();
        this.inStructuredNodes.clear();
        this.outStructuredNodes.clear();
        this.visited.clear();
        this.unusedPlaces.clear();
        this.boundsRemovedPlaces.clear();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#size()
     */
    @Override
    public int size() {
    	synchronized(this.nodes) {
//    		System.out.println(this.root);
//    		for(final StructuredNode n : this.nodes.values()) {
//    			System.out.println(n.isCovered() + " " + n.getPredecessors().size() + " " + n);
//    			System.out.println("    " + n.getPredecessors().keySet());
//    		}
    		return this.nodes.size();
    	}
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#markUnused(petruchio.interfaces.petrinet.Place)
     */
    @Override
    public void markUnused(final Place p) {
        synchronized(this.unusedPlaces) {
            this.unusedPlaces.add(p);
            this.addTransition(Coverability.NEXT_UNUSED_PLACE);
        }
    }
    
    
    /**
     * TODO: document
     *
     */
    private void removeNextUnusedPlace() {
        final Place p;
        synchronized(this.unusedPlaces) {
            p = this.unusedPlaces.remove();
        }
        
//        if(this.boundsRemovedPlaces.containsKey(p)) {
//            return;
//        }
        
        // remove p from markable set
        this.markable.remove(p);
        for(final Collection<Place> c : this.markable.values()) {
            c.remove(p);
        }
        
//        final int oldSize = this.size();
        final int mark = this.removePlace(p);
        this.boundsRemovedPlaces.put(p, Integer.valueOf(mark));
//        if(this.size() != oldSize) {
//            System.out.println("size " + (this.size() - oldSize) + " ("
//                            + this.size() + ") removed " + p);
//        }
    }


    /**
     * TODO: document
     * 
     * @param n
     * @param p
     * @param curVisited
     */
    private int removePlace(final Place p) {
        final LinkedList<StructuredNode> nds = new LinkedList<StructuredNode>(this.nodes.values());
        int max = 0;
        synchronized (this.nodes) {
        	while(!nds.isEmpty()) {
        		final StructuredNode n = nds.remove();
                if ((n.getMarking(p) != null) && (this.nodes.remove(n) != null)) {
                    int mark = n.remove(p);
                    if ((max >= 0) && ((mark < 0) || (mark > max))) {
                        max = mark;
                    }
                    final StructuredNode equal = this.nodes.get(n);
                    if (equal == null) {
                        this.nodes.put(n, n);

                    } else {
                        // node exists
                        if (this.root == n) {
                            this.root = equal;
                        }
                        this.copyArcs(n, equal);
                        if (n.isCovered()) {
                            equal.setIsCovered(true);
                        }

                        for (final Map.Entry<Transition, Collection<StructuredNode>> e : n.getPredecessors().entrySet()) {
                            for (final StructuredNode pre : e.getValue()) {
                                pre.removeSuccessor(e.getKey());
                            }
                        }
                        for (final Map.Entry<Transition, StructuredNode> e : n.getSuccessors().entrySet()) {
                            e.getValue().removePredecessor(e.getKey(), n);
                        }
                        n.clear();
                    }
                }
            }
        }

        return max;
    }
    
    
    /**
     * TODO: document
     *
     * @param a
     * @param b
     */
    private void copyArcs(final StructuredNode a, final StructuredNode b) {
        for(final Map.Entry<Transition, Collection<StructuredNode>> e : a.getPredecessors().entrySet()) {
            for(final StructuredNode pre : e.getValue()) {
                pre.addSuccessor(e.getKey(), b);
            }
        }
        
        for(final Map.Entry<Transition, StructuredNode> e : a.getSuccessors().entrySet()) {
            if(!b.getSuccessors().containsKey(e.getKey())) {
                b.addSuccessor(e.getKey(), e.getValue());
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param rem
     * @param fromStructuredNodes
     */
    private void removeStructuredNodes(final LinkedList<StructuredNode> remove) {
        synchronized(this.nodes) {
	        while(!remove.isEmpty()) {
	        	if(this.stopRequested()) {
	        		return;
	        	}
	        	
	            final StructuredNode n = remove.remove();

	            // remove node from node set
	            if(n == this.root) {
	                continue;
	            }
	            
	            this.nodes.remove(n);
	            
//	            System.err.println("removed: " + n);
	
	            // remove node from its predecessors
	            for (final Map.Entry<Transition, Collection<StructuredNode>> e : n.getPredecessors().entrySet()) {
	            	final Transition t = e.getKey();
	                for (final StructuredNode p : e.getValue()) {
	                    p.removeSuccessor(t);
	                }
	            }
	
	            // remove node from its successors
	            for (final Map.Entry<Transition, StructuredNode> e : n.getSuccessors().entrySet()) {
	                final StructuredNode s = e.getValue();
	                s.removePredecessor(e.getKey(), n);
	                
	                if((s == n) || (s == this.root)) {
	                	continue;
	                }
	
	                boolean hasPredecessors = false;
	                checkPredecessors:
	                for(final Collection<StructuredNode> c : s.getPredecessors().values()) {
	                	for(final StructuredNode c2 : c) {
	                		if((c2 != s) && (c2 != n)) {
	                			hasPredecessors = true;
	                			break checkPredecessors;
	                		}
	                	}
	                }
	                
	                // if that successor has no more predecessors (and is not the
	                // root node) it isn't reachable and may be removed.
	                if (!hasPredecessors) {
	                    remove.add(s);
	                }
	            }
	
	            // clear the node's memory
	            n.clear();
	        }
        }
    }
    
    
    /**
     * Tries to mark the pair of places as 'markable'.
     * 
     * @param p1 A place.
     * @param p2 A place.
     * @return   <code>true</code> iff the pair has not been marked before.
     */
    private boolean addMarkable(final Place p1, final Place p2) {
        final Set<Place> l1 = this.markable.get(p1);
        final Set<Place> l2 = this.markable.get(p2);
        if(l1 == null) {
            if(l2 == null) {
                final Set<Place> l = new IdentityHashSet<Place>();
                this.markable.put(p1, l);
                l.add(p2);
                return true;
            }
            return l2.add(p1);
        }
        if(l2 == null) {
            return l1.add(p2);
        } else if(l1.contains(p2)){
            return false;
        } else {
            return l2.add(p1);
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.cov.AbstractMultiThreadedCoverability#processTransition(petruchio.interfaces.petrinet.Transition)
     */
    @Override
    void processTransition(final Transition t) {
        if(t == Coverability.NEXT_UNUSED_PLACE) {
            this.removeNextUnusedPlace();
            return;
        }

        this.preprocessTransition(t);

        // check all the nodes with the new transition
        for(final StructuredNode n : this.nodes.values()) {
            if(this.stopRequested()) {
                break;
            }
            if(!n.isCovered()) {
                this.checkTransition(n, t);
            }
        }
        this.processNewCovered();
        
        // process new nodes
        this.processNewStructuredNodes();
    }
    
    
    /**
     * TODO: document
     *
     * @param t
     */
    private void preprocessTransition(final Transition t) {
        // add to collection of transitions
        this.addTransitionSorted(t);
//        this.transitions.add(t);
    }
    
    
    /**
     * TODO: document
     *
     * @param t
     */
    private void addTransitionSorted(final Transition t) {
        final int index = Collections.binarySearch(this.transitions, t, this.transitionComparator);
        if(index < 0) {
            this.transitions.add(-(index+1), t);
        } else {
            this.transitions.add(index, t);
        }
    }
    
    /**
     * Successively adds the new nodes to the coverability graph and ands
     * further nodes if a transition is enabled. As a pair of places is
     * recognized as 'markable' it will be added to the passed collection. 
     */
    private void processNewStructuredNodes() {
        while(!this.stopRequested() && !this.newStructuredNodes.isEmpty()) {
            final StructuredNode node = this.newStructuredNodes.removeLast();
            if(node.isCovered() || ((this.root != node) && node.getPredecessors().isEmpty())) {
            	node.clear();
            	continue;
            }
            
            if(this.root != node) {
                this.nodes.put(node, node);
            }
//            System.out.println("Added node: " + node);
//            System.out.println("In " + node.getPredecessors());
            
            if(this.checkMarkable) {
                this.checkIfStructuredNodeMarksPlacePair(node);
            }

            for (final Transition t : this.transitions) {
                if(node.isCovered() || this.stopRequested()) {
                    break;
                }
                if(!node.getSuccessors().containsKey(t)) {
                	this.checkTransition(node, t);
                }
                this.processNewCovered();
            }
        }
    }
    
    
    /**
     * Sets all pairs of places that are marked by the given node, as well as
     * places that are marked with more than one token, as 'markable'. As a pair
     * of places is recognized as 'markable' it will be added to the passed
     * collection.
     *
     * @param node The node to check.
     */
    private void checkIfStructuredNodeMarksPlacePair(final StructuredNode node) {
        final List<Marking> m = node.getMarking();
        for(int i = m.size() - 1; i >= 0; --i) {
            final Marking a = m.get(i);
            final Place p1 = a.getPlace();
            if(!p1.couldCommunicate()) {
                continue;
            }
            if (a.isOmega() || (a.getMarking() > 1) && this.canCommunicate(p1, p1)) {
                if(this.addMarkable(p1, p1)) {
                    final PlacePair p = new petruchio.cov.PlacePair(p1, p1);
                    if(!this.doNotCheck.contains(p)) {
                        this.addPlacePair(p);
                    }
                }
            }
            boolean anotherPlace = false;
            for (int j = i - 1; j >= 0; --j) {
                final Place p2 = m.get(j).getPlace();
//                if(!p1.couldCommunicate() || !p2.couldCommunicate()) {
                if(!this.canCommunicate(p1, p2)) {
                    continue;
                }
                anotherPlace = true;
                if(this.addMarkable(p1, p2)) {
                    final PlacePair p = new petruchio.cov.PlacePair(p1, p2);
                    if(!this.doNotCheck.contains(p)) {
                        this.addPlacePair(p);
                    }
                }
            }
            if(!anotherPlace) {
                break;
            }
        }
    }
    
    
    /**
     * @param t
     * @return
     */
    private StructuredNode getInStructuredNode(final Transition t) {
    	StructuredNode inStructuredNode = this.inStructuredNodes.get(t);
    	if(inStructuredNode == null) {
            // create pre set node of transition
            inStructuredNode = StructuredNode.newNode();
            for(final PTArc arc : t.getInput()) {
                inStructuredNode.add(arc.getSource(), arc.getWeight());
            }
            this.inStructuredNodes.put(t, inStructuredNode);
    	}
    	return inStructuredNode;
    }
    
    
    /**
     * @param t
     * @return
     */
    private StructuredNode getOutStructuredNode(final Transition t) {
    	StructuredNode outStructuredNode = this.outStructuredNodes.get(t);
    	if(outStructuredNode == null) {
            // create post set node of transition
            outStructuredNode = StructuredNode.newNode();
            for(final TPArc arc : t.getOutput()) {
            	outStructuredNode.add(arc.getTarget(), arc.getWeight());
            }
            this.outStructuredNodes.put(t, outStructuredNode);
    	}
    	return outStructuredNode;
    }
    

    /**
     * If the given node enables the given transition the coverability graph
     * is extended. A new node is created using the petri net firing rule and
     * then it is checked, whether it is covered by a node in the graph or if it
     * covers some of the nodes in the graph or if it is just new. If it is not
     * covered, the maximum of the covered nodes is calculated and added to the
     * 'new nodes'.
     * 
     * @param n The node
     * @param t The transition
     */
    private void checkTransition(final StructuredNode n, final Transition t) {
        final StructuredNode inStructuredNode = this.getInStructuredNode(t);
        if(n.enables(inStructuredNode)) {
            final StructuredNode outStructuredNode = StructuredNode.copyNode(n);
            outStructuredNode.add(this.getOutStructuredNode(t));
            outStructuredNode.subtract(inStructuredNode);

            StructuredNode out = this.nodes.get(outStructuredNode);
            if(out == null) {
                if(this.allowAcceleration(outStructuredNode, this.minTokens)) {
                    out = this.accelerate(outStructuredNode, n);
                }
                if(out == null) {
                    outStructuredNode.setIsCovered(false);
                    this.newStructuredNodes.addLast(outStructuredNode);
//                    if(outStructuredNode.getTokens() == 0) {
//                    	System.out.println(n + "\n  --> " + outStructuredNode + "\n" + t);
//                    }
                    out = outStructuredNode;
                } else if(out != Coverability.REPLACES_NODES) {
                    outStructuredNode.clear();
                }
            } else {
                outStructuredNode.clear();
            }
            
            if(out == Coverability.REPLACES_NODES) {
            	this.newStructuredNodes.addLast(outStructuredNode);
            } else {
	            // connect n via t with out
	            final StructuredNode oldSucc = n.addSuccessor(t, out);
	            out.addPredecessor(t, n);
	            if(oldSucc != null) {
	            	oldSucc.removePredecessor(t, n);
	            	boolean hasPredecessors = false;
	                checkPredecessors:
	                for(final Collection<StructuredNode> c : oldSucc.getPredecessors().values()) {
	                	for(final StructuredNode c2 : c) {
	                		if(c2 != oldSucc) {
	                			hasPredecessors = true;
	                			break checkPredecessors;
	                		}
	                	}
	                }
	                if (!hasPredecessors) {
	                	final LinkedList<StructuredNode> l = new LinkedList<StructuredNode>();
		            	l.add(oldSucc);
		            	this.removeStructuredNodes(l);
	                }
	            }
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param n
     * @param tokens
     * @return
     */
    private boolean allowAcceleration(final StructuredNode n, final int tokens) {
        if(tokens <= 0 || ((n.getOmegas() == 0) && (n.getTokens() < tokens))) {
            return false;
        }
        if((tokens == 1) || (n.getMarking().size() == 1)) {
            return true;
        }
        for(final Marking m : n.getMarking()) {
            if(m.isOmega() || (m.getMarking() >= tokens)) {
                return true;
            }
        }
        return false;
    }

    
    /**
     * TODO: document
     *
     * @param n
     * @param start
     * @return
     */
    private StructuredNode accelerate(final StructuredNode n, final StructuredNode start) {
        final LinkedList<StructuredNode> toTreat = new LinkedList<StructuredNode>();
        final Collection<StructuredNode> toCheck = new IdentityHashSet<StructuredNode>();
        final Collection<StructuredNode> locCov = new IdentityHashSet<StructuredNode>();
        toTreat.add(start);
        
//        System.out.println(n);
        
        boolean changed = false;
        int omegas = n.getOmegas();
        while(!toTreat.isEmpty()) {
            if(this.stopRequested()) {
                break;
            }
            
            final StructuredNode cur = toTreat.remove();
            
            if(locCov.contains(cur) || !toCheck.add(cur)) {
                continue;
            }
            
            final Marking.Comparison cmp = n.covMaxCompareTo(cur);
            switch (cmp) {
                case EQUAL:
                	break;
                case GREATER:
//                    System.out.println("> " + cur);
                    toCheck.remove(cur); // new covered nodes don't have to be checked again
                    locCov.add(cur);
                    if(!cur.isCovered()) {
                        cur.setIsCovered(true);
                        this.newCovered.add(cur);
                    }
                    changed = true;
                    break;
                case LESS:
                case UNKNOWN:
                    break;
                case INVALID:
                default:
                    throw new InternalError("Unknown comparison result: " + cmp);
            }
            
            for(final Collection<StructuredNode> prev : cur.getPredecessors().values()) {
            	toTreat.addAll(prev);
            }
        }
        
        // max tokens
        if(this.maxTokens >= 0 && n.getTokens() > this.maxTokens) {
            Collection<Place> nu = null;
            for(final Marking m : n.getMarking()) {
                if(m.getMarking() > this.maxTokens) {
                    if(nu == null) {
                        nu = new ArrayList<Place>();
                    }
                    nu.add(m.getPlace());
                }
            }
            if(nu != null) {
                for(final Place p : nu) {
                    n.addOmega(p);
                }
            }
        }
        
        accel:
        while(omegas != n.getOmegas()) {
            omegas = n.getOmegas();
            for(final Iterator<StructuredNode> it = toCheck.iterator(); it.hasNext();) {
                if(this.stopRequested()) {
                    break accel;
                }
                final StructuredNode cur = it.next();
                final Marking.Comparison cmp = n.covMaxCompareTo(cur);
                switch (cmp) {
                    case EQUAL:
                    	break;
                    case GREATER:
//                        System.out.println("> " + cur);
                        if(!cur.isCovered()) {
                            cur.setIsCovered(true);
                            this.newCovered.add(cur);
                        }
                        locCov.add(cur);
                        it.remove(); // new covered nodes don't have to be checked again
                        changed = true;
                        continue accel;
                    case LESS:
                    case UNKNOWN:
                        break;
                    case INVALID:
                    default:
                        throw new InternalError("Unknown comparison result: " + cmp);
                }
            }
        }
        
//        for(final StructuredNode covered : locCov) {
//        	if(covered == this.root) {
//        		this.root = n;
//        	}
//        	for(final Map.Entry<Transition, Collection<StructuredNode>> e : covered.getPredecessors().entrySet()) {
//        		final Transition t = e.getKey();
//        		for(final StructuredNode prev : e.getValue()) {
//        			if(!prev.isCovered()) {
//	        			n.addPredecessor(t, prev);
//	        			prev.addSuccessor(t, n);
//        			}
//        		}
//        	}
////        	for(final Map.Entry<Transition, StructuredNode> e : covered.getSuccessors().entrySet()) {
////        		final StructuredNode succ = e.getValue();
////        		if(!succ.isCovered() && (succ != n) && (succ != this.root)) {
////    				succ.setIsCovered(true);
////    				this.newCovered.add(succ);
////        		}
////        	}
//        	
//        }
//        
//        final StructuredNode result = locCov.isEmpty() ? null : Coverability.REPLACES_NODES;
//        
//        return result;
        return changed ? this.nodes.get(n) : null;
    }
    
    
    /**
     * TODO: description
     *
     */
    private void processNewCovered() {
//    	if(this.newCovered.isEmpty()) {
//    		return;
//    	}
//        this.removeStructuredNodes(new LinkedList<StructuredNode>(this.newCovered));
        this.newCovered.clear();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#getFiringBounds(java.util.Collection)
     */
    @Override
    public synchronized Map<Transition, Integer> getFiringBounds(final Collection<Transition> trans) {
        final Collection<Transition> check = new IdentityHashSet<Transition>(trans);
        final Stack<StructuredNode> order = new Stack<StructuredNode>();
        final Stack<Transition> orderTrans = new Stack<Transition>();
        final Map<StructuredNode, Map<Transition, Integer>> counters = new IdentityHashMap<StructuredNode, Map<Transition, Integer>>();
        
        this.visited.clear();
        this.computeFiringBounds(check, null, null, this.root, order, orderTrans, this.visited, counters);
        this.visited.clear();
        orderTrans.clear();
        order.clear();
        
        Map<Transition, Integer> firingBounds = counters.get(this.root);
        counters.clear();
        
        if(firingBounds == null) {
            firingBounds = new IdentityHashMap<Transition, Integer>();
        }
        if(check.isEmpty()) {
            for(final Transition t : trans) {
                firingBounds.put(t, Integer.valueOf(-1));
            }
        }
        for(final Transition t : trans) {
            final Integer i = firingBounds.get(t);
            if(i == null) {
                firingBounds.put(t, Integer.valueOf(0));
            }
        }
        
        return firingBounds;
    }
    
    
    /**
     * TODO: document
     *
     * @param cur
     * @param order
     * @param curVisited
     * @param current
     */
    private synchronized void computeFiringBounds(final Collection<Transition> trans,
                    final StructuredNode from, final Transition with, final StructuredNode cur,
                    final Stack<StructuredNode> order, final Stack<Transition> orderTrans,
                    final Collection<StructuredNode> curVisited,
                    final Map<StructuredNode, Map<Transition, Integer>> counters) {

        final Collection<Transition> looped;
        final boolean isLoop = (with != null) && curVisited.contains(cur);
        
        if(isLoop) {
            looped = new LinkedList<Transition>();
            // a loop!
            for(int i = order.size()-1; i >= 0; --i) {
                Transition t = orderTrans.get(i);
                if(trans.remove(t)) {
                    looped.add(t);
                }
                
                if(order.get(i) == cur) {
                    break;
                }
            }
        } else {
            if(!counters.containsKey(cur)) {
                order.push(cur);
                curVisited.add(cur);
        
                
                for(final Map.Entry<Transition, StructuredNode> e : cur.getSuccessors().entrySet()) {
                    final Transition t = e.getKey();
                    final StructuredNode next = e.getValue();
                    
                    orderTrans.push(t);
                    this.computeFiringBounds(trans, cur, t, next, order, orderTrans, curVisited, counters);
                    if(trans.isEmpty()) {
                        return;
                    }
                    orderTrans.pop();
                }
                
                curVisited.remove(cur);
                order.pop();
            }
            
            looped = null;
            
        }
        
        if(with == null) {
            return;
        }
        
        boolean add = trans.contains(with);
        Map<Transition, Integer> counterFrom = counters.get(from);
        Map<Transition, Integer> counterCur = counters.get(cur);
        if(counterCur == null) {
            counterCur = new IdentityHashMap<Transition, Integer>();
            counters.put(cur, counterCur);
        }
        
        boolean removeVisitedStructuredNodeCounter = !isLoop && (cur.getPredecessors().size() == 1);
        if(!isLoop && !removeVisitedStructuredNodeCounter) {
            removeVisitedStructuredNodeCounter = true;
            for(final Collection<StructuredNode> prev : cur.getPredecessors().values()) {
                for(final StructuredNode p : prev) {
                    if((p != from) && !counters.containsKey(p)) {
                        removeVisitedStructuredNodeCounter = false;
                        break;
                    }
                }
            }
        }
        if(removeVisitedStructuredNodeCounter) {
            counters.remove(cur);
        }
        
        boolean copy = !add && (!isLoop || (looped != null && looped.isEmpty())) && (counterFrom == null) && (from != cur);
        if(copy) {
            for(final Transition t : trans) {
                if((t != with) && from.getSuccessors().containsKey(t)) {
                    copy = false;
                    break;
                }
            }
        }
        if(copy) {
            counterFrom = counterCur;
            counters.put(from, counterFrom);
        } else if(counterFrom == null) {
            if(from == cur) {
                counterFrom = counterCur;
            } else {
                counterFrom = new IdentityHashMap<Transition, Integer>(counterCur);
                counters.put(from, counterFrom);
            }
        }
        
        if(from != cur) {
            for(final Map.Entry<Transition, Integer> e : counterCur.entrySet()) {
                final Transition t = e.getKey();
                int countCur = e.getValue().intValue();
                if(add && (with == t)) {
                    if(countCur >= 0) {
                        ++countCur;
                    }
                    add = false;
                }
                final Integer countFrom = counterFrom.get(t);
                if((countFrom == null) || ((countFrom.intValue() >= 0)
                                && ((countCur < 0) || (countCur > countFrom.intValue())))) {
                    counterFrom.put(t, Integer.valueOf(countCur));
                }
            }
        }
        
        if(add) {
            final Integer count = counterFrom.get(with);
            if(count == null) {
                counterFrom.put(with, Integer.valueOf(1));
            }
        }
        
        if(isLoop && looped != null) {
            for(final Transition t : looped) {
                counterFrom.put(t, Integer.valueOf(-1));
            }
        }
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#getBounds(java.util.Collection)
     */
    @Override
    public synchronized Map<Place, Integer> getBounds(final Collection<Place> places) {
        this.awaitComputationEnd();
        final Collection<Place> check = new IdentityHashSet<Place>(places);
        final Map<Place, Integer> bounds = new IdentityHashMap<Place, Integer>(places.size());
        for(final Place p : places) {
            final Integer i = this.boundsRemovedPlaces.get(p);
            bounds.put(p, (i == null) ? Integer.valueOf(Place.BOUND_UNKNOWN) : i);
        }
        int maxTokenCount = 0;
        for(final StructuredNode n : this.nodes.values()) {
            if(n.isCovered()) {
                continue;
            }
            if(maxTokenCount >= 0) {
                if(n.getOmegas() != 0) {
                    maxTokenCount = -1;
                } else if(n.getTokens() > maxTokenCount) {
                    maxTokenCount = n.getTokens();
                }
            }
            for(final Marking m : n.getMarking()) {
                final Place p = m.getPlace();
                if(!check.contains(p)) {
                    continue;
                }
                final int curMax = bounds.get(p).intValue();
                if(curMax >= 0) {
                    if(m.isOmega()) {
                        bounds.put(p, Integer.valueOf(Place.BOUND_UNBOUNDED));
                        check.remove(p);
                        if(check.isEmpty()) {
                            break;
                        }
                    } else if(m.getMarking() > curMax) {
                        bounds.put(p, Integer.valueOf(m.getMarking()));
                    }
                }
            }
        }
        bounds.put(null, Integer.valueOf(maxTokenCount));
        return bounds;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#getDeadlocks(boolean)
     */
    @Override
    public synchronized Collection<String> getDeadlocks(final boolean assumeBoundedness) {
        final Collection<String> deadlocks = new LinkedList<String>();
        boolean omegaNotNeeded = true;
        for(final StructuredNode n : this.nodes.values()) {
        	if(n.getSuccessors().isEmpty()) {
                final StringBuffer b = new StringBuffer();
                for(final Iterator<Marking> it = n.getMarking().iterator(); it.hasNext();) {
                    final Marking m = it.next();
                    if(m.isOmega()) {
                        b.append("\\omega");
                        b.append("*");
                    } else if(m.getMarking() != 1){
                        b.append(m.getMarking());
                        b.append("*");
                    }
                    b.append("[");
                    b.append(m.getPlace());
                    b.append("]");
                    if(it.hasNext()) {
                        b.append(" + ");
                    }
                }
                if(b.length() == 0) {
                	b.append("--empty marking--");
                }
                deadlocks.add(b.toString());
            } else if(!assumeBoundedness && omegaNotNeeded && deadlocks.isEmpty()) {
                boolean all = true;
                for(final Transition t : n.getSuccessors().keySet()) {
                    boolean needsOmega = false;
                    for(final PTArc arc : t.getInput()) {
                        if(n.getMarking(arc.getSource()).isOmega()) {
                            needsOmega = true;
                            break;
                        }
                    }
                    if(!needsOmega) {
                        all = false;
                        break;
                    }
                }
                if(all) {
                    omegaNotNeeded = false;
                }
            }
        }
        
        return deadlocks.isEmpty() && (assumeBoundedness || omegaNotNeeded) ? null : deadlocks;
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#compute(java.util.Collection, java.util.Collection)
     */
    @Override
    public synchronized void compute(final Collection<Place> state,
                    final Collection<Transition> trans)
                throws UnsupportedOperationException {
        this.clearCaches();
        this.setStartState(state);
        for(final Transition t : trans) {
            this.preprocessTransition(t);
        }
        this.newStructuredNodes.addLast(this.root);
        this.processNewStructuredNodes();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#doSetStartState(java.util.Collection)
     */
    @Override
    public synchronized void doSetStartState(final Collection<Place> state) {
        for(final Place p : state) {
            this.root.add(p);
        }
        if(this.checkMarkable) {
            this.checkIfStructuredNodeMarksPlacePair(this.root);
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#keepDeadElements(java.util.Collection, java.util.Collection)
     */
    @Override
    public synchronized void keepDeadElements(final Collection<Place> pl, final Collection<Transition> tr) {
        for(final Map.Entry<Place, Integer> e : this.boundsRemovedPlaces.entrySet()) {
            if(e.getValue().intValue() != 0) {
                pl.remove(e.getKey());
            }
        }
        for(final StructuredNode n : this.nodes.values()) {
            for(final Marking m : n.getMarking()) {
                pl.remove(m.getPlace());
            }
            tr.removeAll(n.getSuccessors().keySet());
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#setCheckMarkable(boolean)
     */
    @Override
    public synchronized void setCheckMarkable(final boolean checkMarkable) {
        this.checkMarkable = checkMarkable;
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#setMinTokenCount(int)
     */
    @Override
    public void setMinTokenCount(final int tokens) {
        this.minTokens = tokens;
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#setMaxTokenCount(int)
     */
    @Override
    public void setMaxTokenCount(final int tokens) {
        this.maxTokens = tokens;
    }
    
    
    public void printTo(final OutputStream o) {
        final PrintWriter out = new PrintWriter(o);
        out.print("digraph Cov {\n"
                + "  node [shape=\"ellipse\"];\n");
        int nodeCount = 0;
        final List<String> elems = new ArrayList<String>();
        for(final StructuredNode n : this.nodes.values()) {
            final int id = System.identityHashCode(n);
            out.print("  ");
            out.print(id);
            out.print(" [label=\"");
            out.print(++nodeCount);
            if(this.root == n) {
                out.print("\", shape=\"box");
            }
            out.print("\", tip=\"");
            
            for(final Marking m : n.getMarking()) {
                elems.add(m.getPlace().getName() + "=" + (m.isOmega() ? "w" : String.valueOf(m.getMarking())));
            }
            Collections.sort(elems);
            final StringBuffer b = new StringBuffer();
            for(final Iterator<String> it = elems.iterator(); it.hasNext();) {
                final String elem = it.next();
                b.append(elem);
                if(it.hasNext()) {
                    b.append(", ");
                }
            }
            elems.clear();
            out.print(b);
            out.print("\"];\n");
            
            if(this.root == n) {
                out.print("  nil [shape=\"plaintext\", style=\"invis\"];\n  nil -> ");
                out.print(id);
                out.print(";\n");
            }
        }
        int transCount = 0;
        final Map<StructuredNode, List<String>> succ = new IdentityHashMap<StructuredNode, List<String>>();
        for(final StructuredNode n : this.nodes.values()) {
            final String id = String.valueOf(System.identityHashCode(n));
            for(final Map.Entry<Transition, StructuredNode> e : n.getSuccessors().entrySet()) {
                ++transCount;
                List<String> label = succ.get(e.getValue());
                if(label == null) {
                    label = new ArrayList<String>(1);
                    succ.put(e.getValue(), label);
                }
                label.add(e.getKey().getName());
            }
            for(final Map.Entry<StructuredNode, List<String>> e : succ.entrySet()) {
                Collections.sort(e.getValue());
                final StringBuffer b = new StringBuffer();
                for(final Iterator<String> it = e.getValue().iterator(); it.hasNext();) {
                    final String elem = it.next();
                    b.append(elem);
                    if(it.hasNext()) {
                        b.append(", ");
                    }
                }
                
                out.print("  ");
                out.print(id);
                out.print(" -> ");
                out.print(System.identityHashCode(this.nodes.get(e.getKey())));
                out.print(" [label=\"");
                out.print(b);
                out.print("\"];\n");
            }
            succ.clear();
        }
        
        out.print("}\n// ");
        out.print(nodeCount);
        out.print(" nodes, ");
        out.print(transCount);
        out.print(" transitions");
        out.close();
    }

    /* (non-Javadoc)
     * @see petruchio.cov.AbstractMultiThreadedCoverability#terminationCleanUp()
     */
    @Override
    void terminationCleanUp() {
        // nothing to do
    }


    @Override
    public void setCommunicatorCache(CommunicatorCache<?> communicators) {
        // TODO Auto-generated method stub
        
    }


    @Override
    void processCollectedTransitions() {
        // TODO Auto-generated method stub
        
    }
    
}