package petruchio.cov;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.logging.Level;

import ch.javasoft.math.BigFraction;
import ch.javasoft.metabolic.efm.config.Arithmetic;
import ch.javasoft.polco.adapter.PolcoAdapter;
import ch.javasoft.polco.adapter.Options;
import ch.javasoft.xml.config.XmlConfigException;

import petruchio.common.IdentityHashSet;
import petruchio.common.rationals.Rational;
import petruchio.common.rationals.IntRational;
import petruchio.common.rationals.LongRational;
import petruchio.common.rationals.BigRational;
import petruchio.interfaces.petrinet.ExtendedTransition;
import petruchio.interfaces.petrinet.InhibitorArc;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.ResetArc;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.TransferArc;
import petruchio.interfaces.petrinet.Transition;
import petruchio.pn.Converter;

public abstract class JRTBackward<R> {
    private static final boolean SORT_CONTRIBUTORS = true;
    private static final boolean ACCELERATE = true;
    private static final boolean ACCELERATE_EXTENDED = true;
    private static final boolean PARTIAL_ORDER_REDUCE = true;
    private static final boolean SIMPLE_PARTIAL_ORDER_REDUCE = true;
    private static final boolean PROCESS_INVARIANTS = true;
    private static final boolean CLUSTER_BOUNDS = true;
    private static final boolean PRUNE_BY_INVARIANTS = true;
    private static final boolean PRUNE_BY_BOUNDS = true;
    
    // how the program calls itself
    private static final String NAME = "JRTBackward";
    // every X iterations a short info on the current state is printed
    private static final long INFO_EVERY = 10000L;
    
    public static void main(final String[] args) {
//        testPolco();
        
        try {
            if(args.length == 0) {
                throw new IllegalArgumentException("No input file given.");
            }
            // for every given filename
            for(final String filename : args) {
                // load the Petri net (using the Petruchio tool)
                final PetriNet net = Converter.loadNet(filename);
                System.out.println(NAME + " loaded file '" + filename + "'.");
                final Collection<Map<Place, Integer>> targets = Converter.loadTargetMarkings(filename, net);
                System.out.println("Target markings: " + Backward.formatTargetLegacy(targets));
                // remember current cpu time used this process (to measure its cpu runtime)
                final long startTime = Backward.getMeasuringTime();
                // instantiate the coverability checker with a specific rational type
                // change to BigRational for unlimited precision or LongRational for long precision or IntRational for standard precision
                // run the coverability checker
                final SimpleList<Transition> seq = check(newBigRationalInstance(), net, targets);
//                final SimpleList<Transition> seq = check(newLongRationalInstance(), net, targets);
//                final SimpleList<Transition> seq = check(newIntRationalInstance(), net, targets);
                // print cpu time used
                System.out.println(NAME + " took " + Backward.formatTime(Backward.getMeasuringTime()-startTime) + ".");
                // distinguish resulting cases
                if(seq == null) {
                    // no trace --> target states aren't coverable
                    System.out.println(NAME + " concludes safe.");
                } else if(seq.getValue() == null) {
                    // empty trace --> initial state covers a target state
                    System.out.println(NAME + " concludes unsafe: Initial state covers some target marking(s).");
                    final Map<Place, Integer> initialMarking = new IdentityHashMap<Place, Integer>();
                    for(final Place p : net.getPlaces()) {
                        if(p.getMarking() > 0) {
                            initialMarking.put(p, Integer.valueOf(p.getMarking()));
                        }
                    }
                    System.out.println("Initial marking: " + Backward.formatMarking(initialMarking));
                } else {
                    // non-empty trace --> a target state is covered when following the trace from the initial marking
                    // pretty printing by Petruchio
                    System.out.println(NAME + " concludes unsafe with trace: " + Backward.formatTrace(seq));
                    final Map<Place, Integer> initialMarking = new IdentityHashMap<Place, Integer>();
                    for(final Place p : net.getPlaces()) {
                        if(p.getMarking() > 0) {
                            initialMarking.put(p, Integer.valueOf(p.getMarking()));
                        }
                        ((petruchio.pn.Place)p).data = null;
                    }
                    for(final Transition t : net.getTransitions()) {
                        ((petruchio.pn.Transition)t).data = null;
                    }
                    final Backward bw = new Backward();
                    System.out.println("Reached marking: " + Backward.formatMarking(bw.fire(initialMarking, seq)));
                }
                // an empty line (to improve readability when passing multiple filenames)
                System.out.println();
            }
        } catch(final RuntimeException e) {
            System.out.println("Usage: java " + JRTBackward.class.getCanonicalName() + " <filename1> [<filename2> ...]");
            System.out.flush();
            e.printStackTrace();
            return;
        }
    }
    
    private static <R> SimpleList<Transition> check(final JRTBackward<R> checker, final PetriNet net, final Collection<Map<Place, Integer>> targets) {
        return checker.coveringSequence(net, checker.constructInitial(net), checker.convert(targets));
    }
    
    public static final JRTBackward<BigRational> newBigRationalInstance() {
        return new JRTBackward<BigRational>() {
            @Override
            BigRational rational(final int numerator, final int denominator) {
                return BigRational.valueOf(numerator, denominator);
            }
            @Override
            BigRational rational(final BigInteger numerator, final BigInteger denominator) {
                return BigRational.valueOf(numerator, denominator);
            }
        };
    }
    
    public static final JRTBackward<LongRational> newLongRationalInstance() {
        return new JRTBackward<LongRational>() {
            @Override
            LongRational rational(final int numerator, final int denominator) {
                return LongRational.valueOf(numerator, denominator);
            }
            @Override
            LongRational rational(final BigInteger numerator, final BigInteger denominator) {
                return LongRational.valueOf(numerator.longValue(), denominator.longValue());
            }
        };
    }
    
    public static final JRTBackward<IntRational> newIntRationalInstance() {
        return new JRTBackward<IntRational>() {
            @Override
            IntRational rational(final int numerator, final int denominator) {
                return IntRational.valueOf(numerator, denominator);
            }
            @Override
            IntRational rational(final BigInteger numerator, final BigInteger denominator) {
                return IntRational.valueOf(numerator.intValue(), denominator.intValue());
            }
        };
    }
    
    private static final Comparator<Transition> TRANSITION_COMPARATOR = new Comparator<Transition>() {
        @Override
        public int compare(final Transition o1, final Transition o2) {
            return System.identityHashCode(o1) - System.identityHashCode(o2);
        }
    };
    
    public static void testPolco() {
        try {
            final Options options = new Options();
            options.setArithmetic(Arithmetic.fractional, Arithmetic.fractional, Arithmetic.fractional);
            final PolcoAdapter p = new PolcoAdapter(options);
            final int[][] eq = null;
            final int[][] iq = {
                    {  -6,  3,  1 },
                    {  -8,  1,  2 },
                    { -15,  5,  3 },
                    {   0,  1,  0 },
                    {   0,  0,  1 },
            };
            final double[][] r = p.getDoubleRays(eq, iq);
            for(final double[] x : r) {
                final double f = x[0];
                if(f < 0) {
                    continue;
                }
                final double[] y = new double[x.length-1];
                if(f > 0) {
                    for(int i = 0; i < y.length; ++i) {
                        y[i] = x[i+1]/f;
                    }
                    System.out.println(java.util.Arrays.toString(y) + " a vertex.");
                    continue;
                }
                System.out.println(java.util.Arrays.toString(y) + " a ray.");
            }
        } catch (final XmlConfigException e) {
            e.printStackTrace();
        }
        
        System.exit(0);
    }
    
    private final Map<Place, Collection<Constraint>> supportingInvariants = new IdentityHashMap<Place, Collection<Constraint>>();
    private final Collection<Constraint> invariants = new IdentityHashSet<Constraint>();
    private final PolcoAdapter polco;
    private BigFraction[][] nonnegative;
    
    public JRTBackward() {
        final Options options = new Options();
        options.setArithmetic(Arithmetic.fractional, Arithmetic.fractional, Arithmetic.fractional);
        options.setLoglevel(Level.WARNING);
        try {
            this.polco = new PolcoAdapter(options);
        } catch (final XmlConfigException e) {
            throw new RuntimeException("Unable to instantiate PolcoAdapter.", e);
        }
    }
    
    public DownwardClosed constructInitial(final PetriNet net) {
        final DownwardClosed init = new DownwardClosed();
        for(final Place p : net.getPlaces()) {
            if(p.getMarking() > 0) {
                init.put(p, rational(p.getMarking()));
            }
        }
        return init;
    }
    
    public Collection<UpwardClosed> convert(final Collection<Map<Place, Integer>> targets) {
        final Collection<UpwardClosed> ucTargets = new ArrayList<UpwardClosed>();
        for(final Map<Place, Integer> target : targets) {
            ucTargets.add(new UpwardClosed(target));
        }
        return ucTargets;
    }

    private Place[] pidx = null;
    
    public SimpleList<Transition> coveringSequence(final PetriNet net, final DownwardClosed initial, final Collection<UpwardClosed> theTarget) {
        // intersection of target and initial non-empty?
        for(final UpwardClosed m : theTarget) {
            if(initial.nonemptyIntersection(m)) {
                return new SimpleList<Transition>(null); // trivially coverable
            }
        }
        
        // check net type and abort with an exception if net isn't supported
        // create data objects for fast look-ups 
        this.initializeFor(net);
        
        // just that
        this.findInvariants(net);
        
        // intersect targets and invariants
        final Collection<UpwardClosed> target = this.pruneByInvariants(theTarget);
        
        // intersection of target and over-approximation by invariants empty?
        if(target.isEmpty()) {
            System.out.println(NAME + " explored 0 entries.");
            return null; // trivially not coverable
        }
        
        {
            this.pidx = new Place[net.getPlaces().size()+1];
            int idx = 1;
            for(final Place p : net.getPlaces()) {
                this.getPlaceData(p).index = idx;
                this.pidx[idx++] = p;
            }
            // generate constraints: for all p \in P : p >= 0
            // this is basically [0 E]
            // it will be used by constraintsToVertices(...)
            this.nonnegative = new BigFraction[net.getPlaces().size()][this.pidx.length];
            for(int i = 0; i < this.pidx.length-1; ++i) {
                for(int j = this.nonnegative[i].length-1; j >= 0; --j) {
                    this.nonnegative[i][j] = j == i+1 ? BigFraction.ONE : BigFraction.ZERO;
                }
            }
        }
        
        // collection of markings to explore // add all target markings to todo list
        final Queue<UpwardClosed> todo = new LinkedList<UpwardClosed>(target);
        
        // set of minimal markings already explored
        final MinimalStates explored = new MinimalStates();
        // add target markings to explored markings and remove non-minimal elements from todo list
        todo.removeAll(explored.addAll(target));
        
        // count how many elements are explored
        long iteration = 0L;
        
//        System.out.println(NAME + " enters main loop.");
        
        // main loop -- until todo is empty
        do {
            // pick marking to explore next
            final UpwardClosed cur = todo.remove();
            ++iteration;
            
            // collect transitions that might have contributed to the current markings
            final Iterable<Transition> contributors = cur.contributors();
            
            // for each possible contributor generate the pre-image of the current markings
            for(final Transition t : contributors) {
                // compute the pre-image of cur under transition t
                final UpwardClosed pre = this.fireBackward(cur, t);
                // check invariants
                if(pre == null || this.violates(pre)) {
                    // we know that pre isn't coverable: proceed with next transition (if any).
                    continue;
                }
                final Collection<UpwardClosed> removed = explored.add(pre);
                // check if pre is already contained in explored
                if(removed == null) {
                    // pre isn't "new" proceed with next transition (if any).
                    continue;
                }
                // if pre is smaller than some element in explored we might be able to accelerate
                final Collection<UpwardClosed> accelerated = removed.isEmpty() ? null : this.accelerate(pre, removed);
                if(accelerated == null) {
                    // check if the pre-image has a non-empty intersection with the initial states 
                    if(initial.nonemptyIntersection(pre)) {
                        // success! a covering trace has been found!
                        System.out.println(NAME + " explored " + iteration + " entries.");
                        return pre.getPath();
                    }
                    // still work to do: add pre to todo collection
                    todo.add(pre);
                } else {
                    for(final UpwardClosed accPre : accelerated) {
                        explored.add(accPre);
                        // check if the pre-image has a non-empty intersection with the initial states 
                        if(initial.nonemptyIntersection(accPre)) {
                            // success! a covering trace has been found!
                            System.out.println(NAME + " explored " + iteration + " entries.");
                            return accPre.getPath();
                        }
                        // still work to do: add pre to todo collection
                        todo.add(accPre);
                    }
                }
            }
            
            // some info on the current state of the search
            if(INFO_EVERY > 0 && iteration % INFO_EVERY == 0L) {
                System.out.println(todo.size() + "/" + explored.size() + ", " + cur);
            }
            
            // exit loop when the whole backward search space has been explored
        } while(!todo.isEmpty());
        System.out.println(NAME + " explored " + iteration + " entries.");
        
        // whole backward search space has been explored --> not coverable
        return null;
    }
    
    private void findInvariants(final PetriNet net) {
        final Place[] pindices = new Place[net.getPlaces().size()];
        final Matrix invs = (new InvariantAnalysis()).findInvariants(net, pindices);
        for(int i = invs.getColumns()-1; i >= 0; --i) {
            if(!invs.columnExists(i)) {
                continue;
            }
            final Constraint inv = new Constraint();
            boolean isSemipositive = true;
            boolean allnegative = true;
            for(int p = pindices.length-1; p >= 0; --p) {
                final BigInteger coeff = invs.get(p, i);
                final int signum = coeff.signum();
                if(allnegative && signum > 0) {
                    allnegative = false;
                }
                if(!allnegative && signum < 0) {
                    isSemipositive = false;
                    break;
                }
                if(signum != 0) {
                    final Place place = pindices[p];
                    final Rational<R> theCoeff = rational(allnegative ? coeff.negate() : coeff);
                    inv.setCoefficient(place, theCoeff);
                    if(place.getMarking() > 0) {
                        inv.setConstantPart(inv.getConstantPart().add(rational(place.getMarking()).multiply(theCoeff)));
                    }
                }
            }
            if(!isSemipositive) {
                continue;
            }
            this.invariants.add(inv);
            for(final Map.Entry<Place, Rational<R>> e : inv.coefficients.entrySet()) {
                final Place p = e.getKey();
                Collection<Constraint> supportingInvs = this.supportingInvariants.get(p);
                if(supportingInvs == null) {
                    supportingInvs = new IdentityHashSet<Constraint>();
                    this.supportingInvariants.put(p, supportingInvs);
                }
                supportingInvs.add(inv);
                
                final int b = inv.constantPart.divide(e.getValue()).intValue();
                if(!p.boundKnown() || b < p.getBound()) {
                    p.setBound(b);
                }
            }
        }
        if(PROCESS_INVARIANTS) {
            this.propagateBounds(net);
        }
        // TODO
    }
    
    private void propagateBounds(final PetriNet net) {
        final Map<Transition, Integer> transitionBounds = new IdentityHashMap<Transition, Integer>();
        final LinkedList<Transition> checkOutput = new LinkedList<Transition>();
        final Queue<Place> todo = new LinkedList<Place>();
        final Collection<Place> bounded = new ArrayList<Place>();
        for(final Place p : net.getPlaces()) {
            if(p.boundKnown()) {
                todo.add(p);
                bounded.add(p);
            }
        }
        while(!todo.isEmpty()) {
            final Place p = todo.remove();
            for(final TPArc a : p.getInput()) {
                final Transition t = a.getSource();
                
                if(a instanceof TransferArc) {
                    continue;
                }
                boolean takesTooMuch = false;
                for(final PTArc b : p.getOutput()) {
                    if(b.getTarget() == a.getSource() && b.getWeight() >= a.getWeight()) {
                        takesTooMuch = true;
                        break;
                    }
                }
                if(takesTooMuch) {
                    continue;
                }
                
                final int transitionFiringBound;
                final int newTransitionFiringBound = p.getBound() / a.getWeight();
                final Integer oldTransitionFiringBound = transitionBounds.put(t, Integer.valueOf(newTransitionFiringBound));
                if(oldTransitionFiringBound == null || newTransitionFiringBound < oldTransitionFiringBound.intValue()) {
                    transitionFiringBound = newTransitionFiringBound;
                    checkOutput.add(t);
                } else {
                    transitionBounds.put(t, oldTransitionFiringBound);
                    transitionFiringBound = oldTransitionFiringBound.intValue();
                }
                
                int inputsize = t.getInput().size();
                if(a instanceof TransferArc) {
                    boolean containsinh = false;
                    for(final PTArc b : t.getInput()) {
                        if(b instanceof ResetArc) {
                            --inputsize;
                        } else if(b instanceof InhibitorArc) {
                            containsinh = true;
                            break;
                        }
                    }
                    if(containsinh) {
                        continue;
                    }
                }
                
                if(inputsize == 1) {
                    for(final PTArc b : t.getInput()) {
                        if(b instanceof ResetArc) {
                            continue;
                        }
                        final int bound = (transitionFiringBound+1) * b.getWeight() - 1;
                        final Place p2 = b.getSource();
                        if(!p2.boundKnown()) {
                            p2.setBound(bound);
                            bounded.add(p2);
                            todo.add(p2);
                        } else if(p2.getBound() > bound) {
                            p2.setBound(bound);
                            todo.add(p2);
                        }
                    }
                }
            }
            
            out_arcs:
            for(final PTArc a : p.getOutput()) {
                final Transition t = a.getTarget();
                final boolean extended = t instanceof ExtendedTransition;
                if(extended && a instanceof InhibitorArc) {
                    continue;
                }
                int putBack = 0;
                for(final TPArc b : t.getOutput()) {
                    if(b.getTarget() == p) {
                        if(extended && b instanceof TransferArc) {
                            continue out_arcs;
                        }
                        putBack = b.getWeight();
                        break;
                    }
                }
                if(putBack >= a.getWeight()) {
                    break;
                }
                final int newTransitionFiringBound = firingBound(p.getBound(), a.getWeight(), putBack);
                final Integer oldTransitionFiringBound = transitionBounds.put(t, Integer.valueOf(newTransitionFiringBound));
                if(oldTransitionFiringBound == null || newTransitionFiringBound < oldTransitionFiringBound.intValue()) {
                    checkOutput.add(t);
                } else {
                    transitionBounds.put(t, oldTransitionFiringBound);
                }
            }
            
            while(!checkOutput.isEmpty()) {
                final Transition t = checkOutput.remove();
                final boolean extended = t instanceof ExtendedTransition;
                final int transitionFiringBound = transitionBounds.get(t).intValue();
                for(final TPArc b : t.getOutput()) {
                    final Place p2 = b.getTarget();
                    if(p2.isParametric() || p2 == p || p2.getInput().size() != 1) {
                        continue;
                    }
                    int transferb = 0;
                    if(extended && b instanceof TransferArc) {
                        for(final Map.Entry<Place, Integer> e : ((TransferArc)b).getTransfer().entrySet()) {
                            final Place p3 = e.getKey();
                            if(!p3.boundKnown()) {
                                transferb = p3.getBound();
                                break;
                            }
                            transferb += p3.getBound()*e.getValue().intValue();
                        }
                    }
                    if(transferb >= 0) {
                        final int bound = transferb + transitionFiringBound * b.getWeight();
                        if(!p2.boundKnown()) {
                            p2.setBound(bound);
                            bounded.add(p2);
                            todo.add(p2);
                        } else {
                            p2.setBound(p2.getBound() + bound);
                            todo.add(p2);
                        }
                    }
                }
            }
        }
        
        if(CLUSTER_BOUNDS) {
            final Map<Place, Constraint> cc = new IdentityHashMap<Place, Constraint>();
            for(final Place p : bounded) {
                if(!p.boundKnown() || p.getBound() <= 0) {
                    continue;
                }
                
                final int b = p.getBound();
                Constraint c = cc.get(p);
                if(c == null) {
                    c = new Constraint(rational(b));
                    c.setCoefficient(p, rational(1));
                    cc.put(p, c);
                }
                
                for(final PTArc pt : p.getOutput()) {
                    final Transition t = pt.getTarget(); 
                    if(t.getOutput().size() == 1 && t.getInput().size() == 1 && !(t instanceof ExtendedTransition)) {
                        for(final TPArc tp : t.getOutput()) {
                            final Place p2 = tp.getTarget();
                            if(p2.boundKnown() && p2.getBound() == b) {
                                final Rational<R> coeff = rational(pt.getWeight(), tp.getWeight());
                                final Constraint c2 = cc.get(p2);
                                if(c2 != null) {
                                    c.addLeft(c2, coeff);
                                    for(final Place pp : c2.coefficients.keySet()) {
                                        cc.put(pp, c);
                                    }
                                } else {
                                    c.setCoefficient(p2, coeff);
                                    cc.put(p2, c);
                                }
                            }
                        }
                    }
                }
            }
            
            for(final Map.Entry<Place, Constraint> e : cc.entrySet()) {
                if(e.getValue().coefficients.size() <= 1) {
                    continue;
                }
                Collection<Constraint> supportingInvs = this.supportingInvariants.get(e.getKey());
                if(supportingInvs == null) {
                    supportingInvs = new IdentityHashSet<Constraint>();
                    this.supportingInvariants.put(e.getKey(), supportingInvs);
                }
                supportingInvs.add(e.getValue());
                this.invariants.add(e.getValue());
            }
        }
        
    }
    
    private static int firingBound(final int m, final int in, final int out) {
        return (m-out)/(in-out);
    }
    
    private Collection<UpwardClosed> pruneByInvariants(final Collection<UpwardClosed> m) {
        if(!PRUNE_BY_BOUNDS && !PRUNE_BY_INVARIANTS) {
            return m;
        }
        final Collection<UpwardClosed> r = new ArrayList<UpwardClosed>();
        for(final UpwardClosed n : m) {
            if(!this.violates(n)) {
                r.add(n);
            }
        }
        // TODO
//        if(m.size()-r.size() != 0) {
//            System.out.println(NAME + " removed " + (m.size()-r.size()) + " inconsistent target states -- remaining: " + r.size());
//        }
        return r;
    }
    
    private void initializeFor(final PetriNet net) {
//        boolean warningIssued = false;
        for(final Transition t : net.getTransitions()) {
            // check for unsupported net elements
            if(t instanceof ExtendedTransition) {
//                if(!warningIssued) {
//                    System.out.println("WARNING: Not working for extended Petri nets, yet.");
//                    warningIssued = true;
//                }
                for(final PTArc in : t.getInput()) {
                    if(in instanceof InhibitorArc) {
                        throw new IllegalArgumentException("Inhibitor arcs aren't supported.");
                    }
                }
            }
            
            // initialize data for fast lookup
            final TransitionData data = new TransitionData();
            ((petruchio.pn.Transition)t).data = data;
            final boolean ext = t instanceof ExtendedTransition;
            Set<Place> resets = null;
            for(final PTArc in : t.getInput()) {
                if(ext && in instanceof InhibitorArc) {
                    throw new IllegalArgumentException("Inhibitor arcs aren't supported.");
                } else if(ext && in instanceof ResetArc) {
                    if(in.getWeight() > 0) {
                        if(resets == null) {
                            resets = new IdentityHashSet<Place>();
                        }
                        resets.add(in.getSource());
                        data.out.put(in.getSource(), rational(in.getWeight()));
                    }
                } else {
                    data.in.put(in.getSource(), rational(in.getWeight()));
                }
            }
            for(final TPArc out : t.getOutput()) {
                final Rational<R> w = rational(out.getWeight());
                final Rational<R> o = data.out.put(out.getTarget(), w);
                if(o != null) {
                    data.out.put(out.getTarget(), o.add(w));
                }
            }
            if(ext) {
                data.change = new Weights();
                data.change.w.putAll(data.out.w);
                for(final Map.Entry<Place, Rational<R>> e : data.in.w.entrySet()) {
                    final Rational<R> o = data.change.put(e.getKey(), e.getValue().negate());
                    if(o != null) {
                        final Rational<R> diff = o.subtract(e.getValue());
                        if(diff.isZero()) {
                            data.change.remove(e.getKey());
                        } else {
                            data.change.put(e.getKey(), diff);
                        }
                    }
                }
                data.transfer = new IdentityHashMap<Place, Weights>();
                data.columns = new IdentityHashSet<Place>();
                if(resets != null) {
                    data.columns.addAll(resets);
                }
                for(final TPArc out : t.getOutput()) {
                    if(out instanceof TransferArc) {
                        final TransferArc trans = (TransferArc)out;
                        final Weights tr = new Weights(trans.getTransfer());
                        data.transfer.put(out.getTarget(), tr);
                        data.columns.addAll(tr.w.keySet());
                    }
                }
            }
        }
        for(final Place p : net.getPlaces()) {
            ((petruchio.pn.Place)p).data = new PlaceData();
        }
    }
    
    private boolean violates(final UpwardClosed m) {
        for(final Vertex v : m.vertices) {
            if(!this.violates(v)) {
                return false;
            }
        }
        return true;
    }
    
    private boolean violates(final Vertex v) {
        if(PRUNE_BY_BOUNDS) {
            for(final Map.Entry<Place, Rational<R>> e : v.supp.entrySet()) {
                if(e.getKey().boundKnown() && e.getValue().compareTo(e.getKey().getBound()) > 0) {
                    return true;
                }
            }
        }
        if(PRUNE_BY_INVARIANTS) {
            final Set<Constraint> checked = new IdentityHashSet<Constraint>();
            for(final Map.Entry<Place, Rational<R>> e : v.supp.entrySet()) {
                final Collection<Constraint> supportedInvariants = this.supportingInvariants.get(e.getKey());
                if(supportedInvariants == null) {
                    continue;
                }
                for(final Constraint c : supportedInvariants) {
                    if(checked.add(c) && !c.satisfiedLEQ(v)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }
    
    private Collection<UpwardClosed> accelerate(final UpwardClosed pre, final Collection<UpwardClosed> preContains) {
        if(!ACCELERATE) {
            return null;
        }
        Collection<UpwardClosed> accelerated = null;
        for(final UpwardClosed accelerationCandidate : preContains) {
            final SimpleList<Transition> trace = this.getAccelerationTrace(pre, accelerationCandidate);
            if(trace == null) {
                continue;
            }
            int ctr = -1;
            UpwardClosed lastAcc;
            UpwardClosed acc = pre;
            do {
                ++ctr;
                lastAcc = acc;
                acc = acc.copy();
                for(SimpleList<Transition> cur = trace; cur != null; cur = cur.getNext()) {
                    final Transition t = cur.getValue();
                    if(t != null) {
                        acc = this.fireBackwardNoCopy(acc, t);
                        if(acc == null) {
                            throw new RuntimeException();
                        }
                    }
                }
            } while(acc.containsNonEqual(lastAcc));
            if(lastAcc != pre) {
                if(accelerated == null) {
                    accelerated = new ArrayList<UpwardClosed>();
                }
                accelerated.add(lastAcc);
//                System.out.println(pre + " --> " + lastAcc + " with (" + trace + ")^" + ctr);
            }
        }
        // TODO
        return accelerated;
    }
    
    private SimpleList<Transition> getAccelerationTrace(final UpwardClosed pre, final UpwardClosed candidate) {
        if(pre.getPath() == null) {
            return null;
        }
        final int candidateLen = candidate.getPath() == null ? 0 : candidate.getPath().getLength();
        if(pre.getPath().getLength() <= candidateLen) {
            return null;
        }
        // |pre.getPath()| > |candidate.getPath()|
        SimpleList<Transition> fromCandidateToPre = null;
        SimpleList<Transition> prePathPart = pre.getPath();
        for(int i = pre.getPath().getLength()-candidateLen; i > 0; --i) {
            final Transition t = prePathPart.getValue();
            if(!ACCELERATE_EXTENDED && t instanceof ExtendedTransition) {
                return null;
            }
            fromCandidateToPre = SimpleList.prepend(t, fromCandidateToPre);
            prePathPart = prePathPart.getNext();
        }
        SimpleList<Transition> candPathPart = candidate.getPath();
        while(prePathPart != null) {
            if(prePathPart.getValue() != candPathPart.getValue()) {
                return null;
            }
            prePathPart = prePathPart.getNext();
            candPathPart = candPathPart.getNext();
        }
        return fromCandidateToPre;
    }
    
    @SuppressWarnings("unchecked")
    TransitionData getTransitionData(final Transition t) {
        return (TransitionData)((petruchio.pn.Transition)t).data;
    }
    
    @SuppressWarnings("unchecked")
    PlaceData getPlaceData(final Place p) {
        return (PlaceData)((petruchio.pn.Place)p).data;
    }
    
    private UpwardClosed fireBackward(final UpwardClosed s, final Transition t) {
        if(t instanceof ExtendedTransition) {
            return this.fireBackwardTransfer(s, (ExtendedTransition)t);
        }
        return this.fireBackwardCommon(s.copy(), t);
    }
    
    private UpwardClosed fireBackwardNoCopy(final UpwardClosed s, final Transition t) {
        if(t instanceof ExtendedTransition) {
            return this.fireBackwardTransfer(s, (ExtendedTransition)t);
        }
        return this.fireBackwardCommon(s, t);
    }
    
    private UpwardClosed fireBackwardCommon(final UpwardClosed s, final Transition t) {
        final TransitionData d = getTransitionData(t);
        int mult = 1;
        if(SIMPLE_PARTIAL_ORDER_REDUCE && d.out.w.size() == 1) { // t* = { p }
            final Map.Entry<Place, Rational<R>> e = d.out.w.entrySet().iterator().next();
            if(e.getKey().getInput().size() == 1) { // *p = { t }
                final Rational<R> max = s.getMaximal(e.getKey());
                if(max != null) {
                    mult = max.subtract(rational(e.getKey().getMarking())).divide(e.getValue()).intValue();
                }
            }
        }
        for(int i = 0; i < mult; ++i) {
            s.semisubtract(d.out);
            s.add(d.in);
            s.path = SimpleList.prepend(t, s.path);
        }
//        System.out.println(s.constraints + " (" + s.vertices.size() + ")" + " under " + mult + "*" + t +  " -> " + result.constraints + " with " + result.vertices);
        return s;
    }
    
    private UpwardClosed fireBackwardTransfer(final UpwardClosed s, final ExtendedTransition t) {
        final Collection<Constraint> product = this.matrixMultiply(s.constraints, t);
        if(product == null) {
            return null;
        }
        final TransitionData d = getTransitionData(t);
        for(final Map.Entry<Place, Rational<R>> e : d.in.w.entrySet()) {
            product.add(new Constraint(e.getKey(), e.getValue()));
        }
        final UpwardClosed result = new UpwardClosed(product);
        if(result.isEmpty()) {
            return null;
        }
        result.path = SimpleList.prepend(t, s.path);
//        System.out.println(s.constraints + " (" + s.vertices.size() + ")" + " under " + t +  " -> " + result.constraints + " with " + result.vertices + "\nwith " + d.transfer);
        return result;
    }
    
    
    private Collection<Constraint> matrixMultiply(final Collection<Constraint> constraints, final ExtendedTransition t) {
        // matrix multiplication of CONSTRAINTS * ACTION(t)
        final Collection<Constraint> result = new ArrayList<Constraint>(constraints.size());
        final TransitionData d = getTransitionData(t);
        final Rational<R> zero = rational(0);
        for(final Constraint constraintRow : constraints) {
            final Constraint prod = new Constraint();
            boolean allNonNegative = true;
            final Set<Place> identity = new IdentityHashSet<Place>(constraintRow.coefficients.keySet());
            identity.removeAll(d.columns);
            for(final Place coefficientFor : d.columns) {
                Rational<R> coefficient = zero;
                for(final Map.Entry<Place, Rational<R>> constraintColumn : constraintRow.coefficients.entrySet()) {
                    if(coefficientFor == constraintColumn && identity.contains(coefficientFor)) {
                        coefficient = coefficient.add(constraintColumn.getValue());
                        continue;
                    }
                    final Weights actionRow = d.transfer.get(constraintColumn.getKey());
                    if(actionRow != null) {
                        final Rational<R> co = actionRow.get(coefficientFor);
                        if(co != null) {
                            coefficient = coefficient.add(constraintColumn.getValue().multiply(co));
                        }
                    }
                }
                if(!coefficient.isZero()) {
                    if(allNonNegative && coefficient.isNegative()) {
                        allNonNegative = false;
                    }
                    prod.setCoefficient(coefficientFor, coefficient);
                }
            }
            Rational<R> coefficient = zero;
            for(final Place p : identity) {
                final Rational<R> coeff = constraintRow.coefficients.get(p);
                if(coeff != null) {
                    coefficient = coefficient.add(coeff);
                }
                if(!coefficient.isZero()) {
                    final Rational<R> old = prod.setCoefficient(p, coefficient);
                    if(old != null) {
                        coefficient = coefficient.add(old);
                        prod.setCoefficient(p, coefficient);
                    }
                    if(allNonNegative && coefficient.isNegative()) {
                        allNonNegative = false;
                    }
                }
            }
            coefficient = null;
            
            Rational<R> rightHandSide = constraintRow.getConstantPart();
            for(final Map.Entry<Place, Rational<R>> e : constraintRow.coefficients.entrySet()) {
                final Rational<R> change = d.change.get(e.getKey());
                if(change != null) {
                    rightHandSide = rightHandSide.subtract(e.getValue().multiply(change));
                }
            }
            if(prod.coefficients.isEmpty() && rightHandSide.isPositive()) {
                return null; // 0 >= 1
            }
            if(!allNonNegative || rightHandSide.isPositive()) {
                prod.setConstantPart(rightHandSide);
                result.add(prod);
            } // else all coefficients are >= 0 and right hand side is <= 0, which is trivially satisfied
        }
        return result;
        // TODO
    }
    
    
    Collection<Vertex> constraintsToVertices(final Collection<Constraint> constraints) {
        final Collection<Vertex> result = new ArrayList<Vertex>();
        final BigFraction[][] iq = new BigFraction[constraints.size()+this.pidx.length-1][this.pidx.length];
        for(int i = this.nonnegative.length-1; i >= 0; --i) {
            System.arraycopy(this.nonnegative[i], 0, iq[i], 0, this.nonnegative[i].length);
        }
        int i = this.nonnegative.length;
        // generate actual constraints
        for(final Constraint c : constraints) {
            iq[i][0] = toBigFraction(c.getConstantPart().negate());
            for(final Map.Entry<Place, Rational<R>> e : c.coefficients.entrySet()) {
                iq[i][this.getPlaceData(e.getKey()).index] = toBigFraction(e.getValue());
            }
            for(int j = iq[i].length-1; j >= 1; --j) {
                if(iq[i][j] == null) {
                    iq[i][j] = BigFraction.ZERO;
                }
            }
            ++i;
        }
        final BigFraction[][] vertices = this.polco.getBigFractionRays(null, iq);
        for(final BigFraction[] vertex : vertices) {
            final BigFraction f = vertex[0];
            if(f.isNonPositive()) { // not a vertex
                continue;
            }
            final Vertex v = new Vertex();
            for(int j = vertex.length-1; j > 0; --j) {
                final BigFraction c = vertex[j];
                if(c.isNonZero()) {
                    v.setValue(this.pidx[j], rational(c.divide(f))); // TODO .ceil() ?
                }
            }
            result.add(v);
        }
        if(result.isEmpty()) {
            result.add(new Vertex());
        }
//        System.out.println(constraints + " --> " + result);
        return result;
    }
    
    private static BigFraction toBigFraction(final Rational<?> r) {
        return BigFraction.valueOf(r.numerator(), r.denominator());
    }
    
    class TransitionData {
        Weights in = new Weights();
        Weights out = new Weights();
        Weights change = null;
        Map<Place, Weights> transfer = null;
        Set<Place> columns = null;
        int index = -1;
    }
    
    class PlaceData {
        int index = -1;
    }
    
    Rational<R> rational(final BigFraction frac) {
        return rational(frac.getNumerator(), frac.getDenominator());
    }
    Rational<R> rational(final int value) {
        return rational(value, 1);
    }
    Rational<R> rational(final BigInteger value) {
        return rational(value, BigInteger.ONE);
    }
    abstract Rational<R> rational(final int numerator, final int denominator);
    abstract Rational<R> rational(final BigInteger numerator, final BigInteger denominator);
    
    class DownwardClosed {
        final Map<Place, Rational<R>> vertices = new IdentityHashMap<Place, Rational<R>>();
        
        public Rational<R> put(final Place p, final Rational<R> i) {
            return this.vertices.put(p, i);
        }
        
        public boolean nonemptyIntersection(final UpwardClosed uc) {
            for(final Vertex v : uc.vertices) {
                boolean vertexCovered = true;
                for(final Map.Entry<Place, Rational<R>> e : v.supp.entrySet()) {
                    final Rational<R> r = this.vertices.get(e.getKey());
                    if(r == null || r.compareTo(e.getValue()) < 0) {
                        vertexCovered = false;
                        break;
                    }
                }
                if(vertexCovered) {
                    return true;
                }
            }
            
            return false;
        }
        
        @Override
        public String toString() {
            return this.vertices.toString();
        }
    }
    
    class Weights {
        final Map<Place, Rational<R>> w = new IdentityHashMap<Place, Rational<R>>();
        
        public Weights() {
            // empty constructor
        }
        
        public Rational<R> remove(final Place key) {
            return this.w.remove(key);
        }

        public Rational<R> get(final Place key) {
            return this.w.get(key);
        }

        public Weights(final Map<Place, Integer> weights) {
            for(final Map.Entry<Place, Integer> e : weights.entrySet()) {
                this.w.put(e.getKey(), rational(e.getValue().intValue()));
            }
        }
        
        public Rational<R> put(final Place p, final Rational<R> i) {
            return this.w.put(p, i);
        }
        
        @Override
        public String toString() {
            return this.w.toString();
        }
    }
    
    /**
     * @author Tim
     */
    class Vertex {
        final Map<Place, Rational<R>> supp = new IdentityHashMap<Place, Rational<R>>();
        Rational<R> sum = rational(0);
        
        /**
         * Creates an empty vertex.
         */
        public Vertex() {
            // nothing
        }
        
        public Vertex(final Map<Place, Integer> s) {
            for(final Map.Entry<Place, Integer> e : s.entrySet()) {
                if(e.getValue().intValue() != 0) {
                    final Rational<R> r = rational(e.getValue());
                    this.sum = this.sum.add(r);
                    this.supp.put(e.getKey(), r);
                }
            }
        }
        
        /**
         * Creates a deep copy of the given vertex.
         * 
         * @param v The vertex to copy
         */
        public Vertex(final Vertex v) {
            this.sum = v.sum;
            this.supp.putAll(v.supp);
        }
        
        public int size() {
            return this.supp.size();
        }
        
        public Rational<R> getValue(final Place p) {
            return this.supp.get(p);
        }
        
        public Rational<R> setValue(final Place p, final Rational<R> value) {
            return this.supp.put(p, value);
        }
        
        /**
         * @return A new vertex object which is a deep copy of this.
         */
        public Vertex copy() {
            return new Vertex(this);
        }
        
        public Rational<R> sum() {
            return this.sum;
        }
        
        public void add(final Weights w) {
            for(final Map.Entry<Place, Rational<R>> e : w.w.entrySet()) {
                final Rational<R> r = e.getValue();
                final Rational<R> i = this.supp.put(e.getKey(), r);
                if(i != null) {
                    this.supp.put(e.getKey(), i.add(e.getValue()));
                }
                this.sum = this.sum.add(r);
            }
        }
        
        public void semisubtract(final Weights w) {
            for(final Iterator<Map.Entry<Place, Rational<R>>> it = this.supp.entrySet().iterator(); it.hasNext(); ) {
                final Map.Entry<Place, Rational<R>> e = it.next();
                final Rational<R> i = w.w.get(e.getKey());
                if(i != null) {
                    if(e.getValue().compareTo(i) <= 0) {
                        this.sum = this.sum.subtract(e.getValue());
                        it.remove();
                    } else {
                        e.setValue(e.getValue().subtract(i));
                        this.sum = this.sum.subtract(i);
                    }
                }
            }
        }
        
        @Override
        public String toString() {
            return this.supp.toString();
        }
    }
    
    class Constraint {
        final Map<Place, Rational<R>> coefficients = new IdentityHashMap<Place, Rational<R>>();
        Rational<R> constantPart;
        
        public Constraint(final Rational<R> constantPart) {
            this.constantPart = constantPart;
        }
        
        public Constraint(final Place p, final Rational<R> constantPart) {
            this.setCoefficient(p, rational(1));
            this.constantPart = constantPart;
        }
        
        public Constraint() {
            this(rational(0));
        }
        
        public Constraint(final Constraint c) {
            this(c.constantPart);
            this.coefficients.putAll(c.coefficients);
        }
        
        public boolean isSimple() {
            return this.coefficients.size() == 1;
        }
        
        public Constraint copy() {
            return new Constraint(this);
        }
        
        public Rational<R> getConstantPart() {
            return this.constantPart;
        }
        
        public void setConstantPart(final Rational<R> r) {
            this.constantPart = r;
        }
        
        public Rational<R> setCoefficient(final Place p, final Rational<R> coefficient) {
            return this.coefficients.put(p, coefficient);
        }
        
        public boolean satisfiedGEQ(final Vertex v) {
            if(this.coefficients.isEmpty()) {
                return true;
            }
            Rational<R> result = null;
            for(final Map.Entry<Place, Rational<R>> e : this.coefficients.entrySet()) {
                final Rational<R> value = v.getValue(e.getKey());
                if(value != null) {
                    result = result == null ? e.getValue().multiply(value) : result.add(e.getValue().multiply(value));
                    if(result.compareTo(this.constantPart) >= 0) {
                        return true;
                    }
                }
            }
            return false;
        }
        
        public boolean satisfiedLEQ(final Vertex v) {
            Rational<R> result = null;
            for(final Map.Entry<Place, Rational<R>> e : this.coefficients.entrySet()) {
                final Rational<R> value = v.getValue(e.getKey());
                if(value != null) {
                    result = result == null ? e.getValue().multiply(value) : result.add(e.getValue().multiply(value));
                    if(result.compareTo(this.constantPart) > 0) {
                        return false;
                    }
                }
            }
            return true;
        }
        
        public void add(final Weights w) {
           for(final Map.Entry<Place, Rational<R>> e : w.w.entrySet()) {
               final Rational<R> coeff = this.coefficients.get(e.getKey());
               if(coeff != null) {
                   this.constantPart = this.constantPart.add(e.getValue().divide(coeff));
               }
           }
        }
        
        public void add(final Constraint c, final Rational<R> factor) {
            this.setConstantPart(this.getConstantPart().add(c.getConstantPart()));
            for(final Map.Entry<Place, Rational<R>> e : c.coefficients.entrySet()) {
                final Rational<R> summand = e.getValue().multiply(factor);
                final Rational<R> coeff = this.coefficients.put(e.getKey(), summand);
                if(coeff != null) {
                    this.coefficients.put(e.getKey(), e.getValue().add(summand));
                }
            }
        }
        
        public void add(final Constraint c) {
            this.setConstantPart(this.getConstantPart().add(c.getConstantPart()));
            for(final Map.Entry<Place, Rational<R>> e : c.coefficients.entrySet()) {
                final Rational<R> coeff = this.coefficients.put(e.getKey(), e.getValue());
                if(coeff != null) {
                    this.coefficients.put(e.getKey(), e.getValue().add(coeff));
                }
            }
        }
        
        public void addLeft(final Constraint c, final Rational<R> factor) {
            for(final Map.Entry<Place, Rational<R>> e : c.coefficients.entrySet()) {
                final Rational<R> summand = e.getValue().multiply(factor);
                final Rational<R> coeff = this.coefficients.put(e.getKey(), summand);
                if(coeff != null) {
                    this.coefficients.put(e.getKey(), summand.add(coeff));
                }
            }
        }
        
        public void semisubtract(final Weights w) {
            for(final Map.Entry<Place, Rational<R>> e : w.w.entrySet()) {
                final Rational<R> coeff = this.coefficients.get(e.getKey());
                if(coeff != null) {
                    this.constantPart = this.constantPart.subtract(e.getValue().divide(coeff));
                    if(this.isSimple() && this.constantPart.signum() <= 0) {
                        this.constantPart = rational(0);
                        return;
                    }
                }
            }
        }
        
        public boolean equals(final Constraint c) {
            if(c.coefficients.size() != this.coefficients.size()) {
                return false;
            }
            for(final Map.Entry<Place, Rational<R>> e : this.coefficients.entrySet()) {
                final Rational<R> otherCoeff = c.coefficients.get(e.getKey());
                if(otherCoeff == null || !e.getValue().divide(this.constantPart).equals(otherCoeff.divide(c.constantPart))) {
                    return false;
                }
            }
            return true;
        }
        
        @Override
        public String toString() {
            return this.coefficients + " ~ " + this.constantPart;
        }
        
    }
    
    /**
     * @author Tim
     */
    class UpwardClosed {
        final Collection<Vertex> vertices = new ArrayList<Vertex>();
        final Collection<Constraint> constraints = new ArrayList<Constraint>();
        int maxSupp = -1;
        int minSupp = -1;
        Rational<R> minToken = null;
        Rational<R> maxToken = null;
        SimpleList<Transition> path = null;
        
        /**
         * Creates an empty upward closed set.
         */
        public UpwardClosed() {
            // nothing
        }
        
        public UpwardClosed(final Map<Place, Integer> s) {
            final Vertex v = new Vertex();
            for(final Map.Entry<Place, Integer> e : s.entrySet()) {
                final Rational<R> r = rational(e.getValue());
                v.setValue(e.getKey(), r);
                final Constraint c = new Constraint(r);
                c.setCoefficient(e.getKey(), rational(1));
                this.constraints.add(c);
            }
            this.vertices.add(v);
            this.maxSupp = s.size();
            this.minSupp = s.size();
        }
        
        public UpwardClosed(final Collection<Constraint> constraints) {
            this.constraints.addAll(constraints);
            boolean allSimple = true;
            final Vertex vert = new Vertex();
            for(final Constraint c : constraints) {
                if(!c.isSimple()) {
                    allSimple = false;
                    break;
                } 
                final Map.Entry<Place, Rational<R>> e = c.coefficients.entrySet().iterator().next();
                final Rational<R> r = c.getConstantPart().divide(e.getValue()); // TODO .ceil() ?
                final Rational<R> old = vert.setValue(e.getKey(), r);
                if(old != null && old.compareTo(r) > 0) {
                    vert.setValue(e.getKey(), old);
                }
            }
            if(allSimple) {
                this.vertices.add(vert);
            } else {
//                System.out.print('.');
                for(final Vertex v : JRTBackward.this.constraintsToVertices(constraints)) {
                    this.vertices.add(v);
                    if(this.maxSupp == -1 || v.size() > this.maxSupp) {
                        this.maxSupp = v.size();
                    }
                    if(this.minSupp == -1 || v.size() < this.minSupp) {
                        this.minSupp = v.size();
                    }
                }
            }
            this.removeRedundantConstraints(); // TODO ?
        }
        
        /**
         * Creates a deep copy of the given upward closed set.
         * @param s The upward closed set to copy
         */
        public UpwardClosed(final UpwardClosed s) {
            this.path = s.path;
            this.maxSupp = s.maxSupp;
            this.minSupp = s.minSupp;
            this.minToken = s.minToken;
            this.maxToken = s.maxToken;
            for(final Vertex v : s.vertices) {
                this.vertices.add(v.copy());
            }
            for(final Constraint c : s.constraints) {
                this.constraints.add(c.copy());
            }
        }
        
        private void removeRedundantConstraints() {
            // TODO
        }
        
        public Rational<R> getMaximal(final Place p) {
            Rational<R> max = null;
            for(final Vertex r : this.vertices) {
                final Rational<R> cur = r.getValue(p);
                if(cur != null && (max == null || cur.compareTo(max) > 0)) {
                    max = cur;
                }
            }
            return max;
        }
        
        public int getMaxSupport() {
            if(this.maxSupp == -1) {
                for(final Vertex v : this.vertices) {
                    if(v.supp.size() > this.maxSupp) {
                        this.maxSupp = v.supp.size();
                    }
                    if(v.supp.size() < this.minSupp) {
                        this.minSupp = v.supp.size();
                    }
                }
            }
            return this.maxSupp;
        }
        
        public int getMinSupport() {
            if(this.minSupp == -1) {
                this.getMaxSupport();
            }
            return this.minSupp;
        }
        
        public Rational<R> getMaxTokens() {
            if(this.maxToken == null) {
                for(final Vertex v : this.vertices) {
                    if(this.maxToken == null || v.sum().compareTo(this.maxToken) > 0) {
                        this.maxToken = v.sum();
                    }
                    if(this.minToken == null || v.sum().compareTo(this.minToken) < 0) {
                        this.minToken = v.sum();
                    }
                }
            }
            return this.maxToken;
        }
        
        public Rational<R> getMinTokens() {
            if(this.minToken == null) {
                this.getMaxTokens();
            }
            return this.minToken;
        }
        
        public boolean contains(final UpwardClosed m) {
            if(this.getMinSupport() > m.getMaxSupport() || this.getMinTokens().compareTo(m.getMaxTokens()) > 0) {
                return false;
            }
            for(final Vertex v : m.vertices) {
                if(!this.contains(v)) {
                    return false;
                }
            }
            
            return true;
        }
        
        public boolean containsNonEqual(final UpwardClosed m) {
            return this.contains(m) && !m.contains(this);
        }
        
        public boolean contains(final Vertex v) {
            for(final Vertex w : this.vertices) {
                if(w.supp.size() > v.supp.size() || w.sum().compareTo(v.sum()) > 0) {
                    continue;
                }
                boolean contains = true;
                for(final Map.Entry<Place, Rational<R>> e : w.supp.entrySet()) {
                    final Rational<R> r = v.supp.get(e.getKey());
                    if(r == null || r.compareTo(e.getValue()) < 0) {
                        contains = false;
                        break;
                    }
                }
                if(contains) {
                    return true;
                }
            }
            if(this.vertices.size() == 1) {
                return false;
            }
            for(final Constraint c : this.constraints) {
                if(!c.satisfiedGEQ(v)) {
                    return false;
                }
            }
            return true;
        }
        
        public boolean isEmpty() {
            return this.vertices.isEmpty();
        }

        public boolean contained(final Collection<UpwardClosed> m) {
            for(final UpwardClosed uc : m) {
                if(uc.contains(this)) {
                    return true;
                }
            }
            return false;
        }
        
        public boolean contains(final Collection<UpwardClosed> m) {
            for(final UpwardClosed uc : m ) {
                if(!this.contains(uc)) {
                    return false;
                }
            }
            return true;
        }
        
        public Iterable<Transition> contributors() {
            final Set<Transition> contributors = new IdentityHashSet<Transition>();
            
            for(final Vertex v : this.vertices) {
                for(final Map.Entry<Place, Rational<R>> e : v.supp.entrySet()) {
                    final Place p = e.getKey();
                    if(PARTIAL_ORDER_REDUCE) {
                        if(p.getInput().size() == 1 && e.getValue().compareTo(p.getMarking()) > 0) { // check for partial order
                            final Transition t = p.getInput().iterator().next().getSource();
                            boolean partialOrder = true;
                            for(final TPArc a : t.getOutput()) {
                                if(a.getTarget() != p) {
                                    final Rational<R> r = v.supp.get(a.getTarget());
                                    if(r == null || r.compareTo(a.getWeight()) < 0 || a.getSource() instanceof ExtendedTransition) {
                                        partialOrder = false;
                                        break;
                                    }
                                }
                            }
                            if(partialOrder) {
                                contributors.clear();
                                contributors.add(t);
                                return contributors;
                            }
                        }
                    }
                    
                    for(final TPArc in : p.getInput()) {
                        contributors.add(in.getSource());
                    }
                }
            }
            
            if(SORT_CONTRIBUTORS && contributors.size() > 1) {
                final List<Transition> sorted = new ArrayList<Transition>(contributors);
                Collections.sort(sorted, TRANSITION_COMPARATOR);
                return sorted;
            }
            
            return contributors;
        }
        
//        public void retain(final UpwardClosed uc) {
//        }
        
        public void semisubtract(final Weights w) {
            this.maxSupp = -1;
            this.minSupp = -1;
            this.minToken = null;
            this.maxToken = null;
            for(final Vertex v : this.vertices) {
                v.semisubtract(w);
            }
            for(final Iterator<Constraint> it = this.constraints.iterator(); it.hasNext(); ) {
                final Constraint c = it.next();
                c.semisubtract(w);
                if(c.getConstantPart().isNonPositive()) {
                    it.remove();
                }
            }
        }
        
        public void add(final Weights w) {
            this.maxSupp = -1;
            this.minSupp = -1;
            this.minToken = null;
            this.maxToken = null;
            for(final Vertex v : this.vertices) {
                v.add(w);
            }
            final Set<Place> todo = new IdentityHashSet<Place>(w.w.keySet());
            for(final Constraint c : this.constraints) {
                if(c.coefficients.size() == 1) {
                    final Map.Entry<Place, Rational<R>> e = c.coefficients.entrySet().iterator().next();
                    final Rational<R> r = w.get(e.getKey());
                    if(r != null) {
                        todo.remove(e.getKey());
                        c.setConstantPart(c.getConstantPart().add(r.multiply(e.getValue())));
                    }
                }
            }
            for(final Place p : todo) {
                this.constraints.add(new Constraint(p, w.get(p)));
            }
        }
        
        public SimpleList<Transition> getPath() {
            return this.path;
        }
        
        public UpwardClosed copy() {
            return new UpwardClosed(this);
        }
        
        @Override
        public String toString() {
            return this.vertices.toString();
        }
        
    }
    
    class MinimalStates {
        final List<UpwardClosed> states = new LinkedList<UpwardClosed>();

        public int size() {
            return this.states.size();
        }
        
        public Collection<UpwardClosed> addAll(final Iterable<UpwardClosed> markings) {
            final Collection<UpwardClosed> removed = new ArrayList<UpwardClosed>();
            for(final UpwardClosed s : markings) {
                final Collection<UpwardClosed> rm = this.add(s);
                if(rm != null) {
                    removed.addAll(rm);
                }
            }
            return removed;
        }
        
        /**
         * @param newStates The upward closed set of states to add.
         * @return <code>null</code> iff the given states haven't been added as they are already contained in this set.
         * The collection of states (possibly empty) that have been removed as they were contained in the given states.
         */
        public Collection<UpwardClosed> add(final UpwardClosed newStates) {
            Collection<UpwardClosed> removed = null;
            // for now a simplistic implementation: walk trough all minimal elements and check whether they're <= or >=
            for(final Iterator<UpwardClosed> it = this.states.iterator(); it.hasNext(); ) {
                final UpwardClosed s = it.next();
                if(removed == null && s.contains(newStates)) {
                    return null;
                }
                if(removed == null ? newStates.contains(s) : newStates.containsNonEqual(s)) {
                    if(removed == null) {
                        removed = new ArrayList<UpwardClosed>();
                    }
                    removed.add(s);
                    it.remove();
                }
            }
            this.states.add(newStates);
            if(removed == null) {
                removed = new ArrayList<UpwardClosed>();
            }
            return removed;
        }
        
        @Override
        public String toString() {
            return this.states.toString();
        }
    }
    
    class InvariantAnalysis {
        
        public Matrix findInvariants(final PetriNet net, final Place[] pindices) {
            final Matrix incidence = generateExtendedTransposedIncidenceMatrix(net, pindices);
            
            final int columns = incidence.getColumns();
            
            final Queue<MatrixOperation> ops = solve(incidence);
            
            incidence.kill(); // remove internal array to free the memory
            
            // create solution matrix
            final Matrix invariants = incidence.identity(columns, columns);
            while(!ops.isEmpty()) {
                ops.remove().perform(invariants);
            }
            return invariants;
        }
        
        private Matrix generateTransposedIncidenceMatrix(final PetriNet net, final Place[] pindices) {
            final Matrix incidence = new Matrix(net.getTransitions().size(), net.getPlaces().size());
            
            int tctr = 0;
            for(final Transition t : net.getTransitions()) {
                final TransitionData td = getTransitionData(t);
                td.index = tctr++;
            }
            
            int pctr = 0;
            for(final Place p : net.getPlaces()) {
                final PlaceData pd = getPlaceData(p);
                pindices[pctr] = p;
                pd.index = pctr++;

                for(final TPArc in : p.getInput()) {
                    final TransitionData td = getTransitionData(in.getSource());
                    incidence.set(td.index, pd.index, BigInteger.valueOf(in.getWeight()));
                }
                for(final PTArc out : p.getOutput()) {
                    final TransitionData td = getTransitionData(out.getTarget());
                    incidence.subtract(td.index, pd.index, BigInteger.valueOf(out.getWeight()));
                }
            }
            
            return incidence;
        }
        
        private Matrix generateExtendedTransposedIncidenceMatrix(final PetriNet net, final Place[] pindices) {
            final Matrix incidenceTr = new Matrix(net.getPlaces().size(), net.getTransitions().size());
            
            int tctr = 0;
            for(final Transition t : net.getTransitions()) {
                final TransitionData td = getTransitionData(t);
                td.index = tctr++;
            }
            
            class Tmp {
                final Place p1;
                final Integer t;
                final Place p2;
                final int h;
                public Tmp(final Place p1, final Integer t, final Place p2) {
                    this.p1 = p1; this.p2 = p2; this.t = t;
                    this.h = (System.identityHashCode(p1)*31 + System.identityHashCode(p2))*31 + t.hashCode();
                }
                @Override public int hashCode() { return this.h; }
                @Override public boolean equals(final Object o) {
                    return o != null && o.getClass() == this.getClass() && this.equals((Tmp)o);
                }
                public boolean equals(final Tmp o) {
                    return o != null && o.p1 == this.p1 && o.p2 == this.p2 && this.t.equals(o.t);
                }
            }
            
            final Map<Tmp, Integer> transIndices = new HashMap<Tmp, Integer>();
            
            int pctr = 0;
            for(final Place p : net.getPlaces()) {
                final PlaceData pd = getPlaceData(p);
                pd.index = pctr++;
                pindices[pd.index] = p;

                for(final TPArc in : p.getInput()) {
                    final TransitionData td = getTransitionData(in.getSource());
                    incidenceTr.set(pd.index, td.index, BigInteger.valueOf(in.getWeight()));
                    if(in instanceof TransferArc) {
                        for(final Map.Entry<Place, Integer> e : ((TransferArc)in).getTransfer().entrySet()) {
                            if(e.getKey() == p) {
                                continue;
                            }
                            final Tmp pair = new Tmp(e.getKey(), e.getValue(), p);
                            Integer idx = transIndices.get(pair);
                            if(idx == null) {
                                idx = Integer.valueOf(incidenceTr.addColumn());
                                transIndices.put(pair, idx);
                            }
                            incidenceTr.set(pd.index, idx.intValue(), BigInteger.valueOf(e.getValue().intValue()));
                        }
                    }
                }
                for(final PTArc out : p.getOutput()) {
                    final TransitionData td = getTransitionData(out.getTarget());
                    incidenceTr.subtract(pd.index, td.index, BigInteger.valueOf(out.getWeight()));
                    if(out instanceof ResetArc) {
                        boolean matchingTransferFound = false;
                        for(final TPArc in : out.getTarget().getOutput()) {
                            if(in instanceof TransferArc) {
                                final Integer trans = ((TransferArc)in).getTransfer().get(p);
                                if(trans != null) {
                                    final Tmp pair = new Tmp(p, trans, in.getTarget());
                                    Integer idx = transIndices.get(pair);
                                    if(idx == null) {
                                        idx = Integer.valueOf(incidenceTr.addColumn());
                                        transIndices.put(pair, idx);
                                    }
                                    incidenceTr.subtract(pd.index, idx.intValue(), BigInteger.ONE);
                                    matchingTransferFound = true;
                                }
                            }                        
                        }
                        if(!matchingTransferFound) {
                            final int idx = incidenceTr.addColumn();
                            incidenceTr.subtract(pd.index, idx, BigInteger.ONE);
                        }
                    }
                }
            }
            
            incidenceTr.transpose();
            return incidenceTr;
        }
        
        private Queue<MatrixOperation> solve(final Matrix incidence) {
            final Queue<MatrixOperation> ops = new LinkedList<MatrixOperation>();
            final int[] pPlus = new int[incidence.getColumns()];
            final int[] pMinus = new int[incidence.getColumns()];
            
            int nonZeroRow;
            do {
                boolean cardinalityUsed = false;
                nonZeroRow = -1;

                for (int h = 0; h < incidence.getRows() && !cardinalityUsed && incidence.getColumns() != 0; ++h) {
                    
                    // pPlus and pMinus are large enough. in phase 1, the matrices don't grow.
                    getIndices(incidence, h, pPlus, pMinus);
                    
                    boolean pPlusIsSingleton;

                    if ((pPlus[0] == -1) ^ (pMinus[0] == -1)) { // case 1.1.a
                        
                        cardinalityUsed = true;
                        nonZeroRow = h;

                        int[] pPlusMinus = pPlus[0] == -1 ? pMinus : pPlus;

                        for (int j = 0; j < pPlusMinus.length && pPlusMinus[j] != -1; ++j) {
                            incidence.killColumn(pPlusMinus[j]);
                            
                            ops.add(new MatrixOperation(pPlusMinus[j]));
                        }
                        
                    } else if ((pPlusIsSingleton = (pPlus[0] != -1 && (pPlus.length == 1 || pPlus[1] == -1)))
                            || ((pMinus[0] != -1 && (pMinus.length == 1 || pMinus[1] == -1)))) { // case 1.1.b

                        cardinalityUsed = true;
                        nonZeroRow = h;

                        final int[] pPlusMinus;
                        final int[] pMinusPlus;
                        if (pPlusIsSingleton) {
                            pPlusMinus = pPlus;
                            pMinusPlus = pMinus;
                        } else {
                            pPlusMinus = pMinus;
                            pMinusPlus = pPlus;
                        }

                        final int k = pPlusMinus[0];
                        final BigInteger chk = incidence.get(h, k).abs();

                        for (int cur = 0; cur < pMinusPlus.length && pMinusPlus[cur] != -1; ++cur) {
                            
                            final int j = pMinusPlus[cur];
                            final BigInteger chj = incidence.get(h, j).abs();
                            
                            for (int i = 0; i < incidence.getRows(); ++i) {
                                incidence.set(i, j, chj.multiply(incidence.get(i, k)).add(chk.multiply(incidence.get(i, j))));
                            }
                            
                            ops.add(new MatrixOperation(chj, k, chk, j));
                        }

                        incidence.killColumn(k);
                        
                        ops.add(new MatrixOperation(k));
                        
                        
                    } else if (pPlus[0] != -1 || pMinus[0] != -1) {
                        nonZeroRow = h;
                        
                    }

                } 

                if (!cardinalityUsed && nonZeroRow != -1) { // case 1.1.c
                    final int h = nonZeroRow;
                    int k = 0;
                    while (!incidence.columnExists(k) || incidence.get(h, k).equals(BigInteger.ZERO)) {
                        ++k;
                    }
                    final BigInteger chk = incidence.get(h, k);

                    for (int j = 0; j < incidence.getColumns(); ++j) {
                        if (j == k || !incidence.columnExists(j)) {
                            continue;
                        }
                        final BigInteger chj = incidence.get(h, j);
                        if(chj.equals(BigInteger.ZERO)) {
                            continue;
                        }

                        final BigInteger alpha = chj.signum() != chk.signum() && chk.signum() != 0 ? chj.abs() : chj.abs().negate();
                        final BigInteger beta = chk.abs();

                        for (int i = 0; i < incidence.getRows(); ++i) {
                            incidence.set(i, j, alpha.multiply(incidence.get(i, k)).add(beta.multiply(incidence.get(i, j))));
                        }
                        
                        ops.add(new MatrixOperation(alpha, k, beta, j));
                    }

                    incidence.killColumn(k);
                    ops.add(new MatrixOperation(k));
                    
                }

            } while(nonZeroRow != -1);
            
            return ops;
        }

        
        private void getIndices(final Matrix A, final int rowNo, final int[] positivesArray,
                final int[] negativesArray) {
            final int n = A.getColumns();
            int negcount = 0;
            int poscount = 0;

            for (int i = 0; i < n; i++) {
                if(A.columnExists(i)) {
                    final int sig = A.get(rowNo, i).signum();
                    if (sig < 0) {
                        negativesArray[negcount++] = i;
                    } else if (sig > 0) {
                        positivesArray[poscount++] = i;
                    }
                }
            }
            if (negcount < n) {
                negativesArray[negcount] = -1;
            }
            if (poscount < n) {
                positivesArray[poscount] = -1;
            }

        }
        
//        private BigInteger[] ensureSize(final BigInteger[] a, final int size) {
//            // the content of the array is irrelevant!
//            return a.length >= size ? a : new BigInteger[size+10];
//        }
//        
//        private BigInteger gcd(final Matrix A, final int column) {
//            BigInteger gcd = A.get(0, column);
//            for (int i = 1; i < A.getRows(); i++) {
//                if (!A.get(i, column).equals(BigInteger.ZERO) || gcd.equals(BigInteger.ZERO)) {
//                    gcd = gcd.gcd(A.get(i, column));
//                }
//            }
//            return gcd;
//        }
//        
//        private void removeRedundant(final Matrix matrix) {
//            int[] nonmin = matrix.findNonMinimal();
//            for(int i = 0; i < nonmin.length && nonmin[i] != -1; ++i) {
//                matrix.killColumn(nonmin[i]);
//            }
//        }
        
        private class MatrixOperation {
            final int kill;
            final BigInteger alpha;
            final int alphaColumn;
            final BigInteger beta;
            final int betaColumn;
            
            MatrixOperation(final int kill) {
                this.kill = kill;
                this.alpha = BigInteger.ZERO;
                this.beta = BigInteger.ZERO;
                this.alphaColumn = -1;
                this.betaColumn = -1;
            }
            
            MatrixOperation(final BigInteger alpha, final int alphaColumn, final BigInteger beta, final int betaColumn) {
                this.kill = -1;
                this.alpha = alpha;
                this.beta = beta;
                this.alphaColumn = alphaColumn;
                this.betaColumn = betaColumn;
            }
            
            void perform(final Matrix matrix) {
                if(this.kill >= 0) {
                    matrix.killColumn(this.kill);
                } else {
                    for (int i = 0; i < matrix.getRows(); ++i) {
                        matrix.set(i, this.betaColumn, this.alpha.multiply(matrix.get(i, this.alphaColumn)).add(this.beta.multiply(matrix.get(i, this.betaColumn))));
                    }
                }
            }
        }
    }
    
    class Matrix {

        // number of spare columns created when up-sizing matrix
        private static final int INCREMENT = 10;
        
        // column-major order
        private BigInteger[][] content;
        
        // current number of rows
        private int rows;
        
        // current number of columns (some might be nulled)
        private int columns;
        
        // stack for indices of nulled columns 
        private int[] freeColumns = new int[10];
        
        // current size of the stack
        private int numFreeColumns = 0;

        // private constructor, use static factory methods instead
        Matrix(final int rows, final int columns) {
            this.rows = rows;
            this.columns = columns;
            this.content = new BigInteger[this.columns][this.rows];
        }
        
        
        /**
         * Creates a matrix from the given row-major array.
         * All rows are needed to have the same number of columns.
         * 
         * @param array A row-major array.
         * @return The matrix representation of the array.
         */
        public Matrix fromRowMajorArray(final BigInteger[][] array) {
            final int rows = array.length;
            if(rows == 0) {
                throw new IllegalArgumentException("Array must have positive row dimension.");
            }
            final int columns = array[0].length;
            if(columns == 0) {
                throw new IllegalArgumentException("Array must have positive column dimension.");
            }
            final Matrix m = new Matrix(rows, columns);
            for(int i = 0; i < rows; ++i) {
                if(array[i].length != columns) {
                    throw new IllegalArgumentException("All rows of array must have the same length.");
                }
                for(int j = 0; j < columns; ++j) {
                    m.set(j, i, array[i][j]);
                }
            }
            return m;
        }
        
        
        /**
         * Creates a matrix from the given column-major array.
         * All columns are needed to have the same number of rows.
         * 
         * @param array A column-major array.
         * @return The matrix representation of the array.
         */
        public Matrix fromColumnMajorArray(final BigInteger[][] array) {
            final int columns = array.length;
            if(columns == 0) {
                throw new IllegalArgumentException("Array must have positive column dimension.");
            }
            final int rows = array[0].length;
            if(rows == 0) {
                throw new IllegalArgumentException("Array must have positive row dimension.");
            }
            final Matrix m = new Matrix(rows, columns);
            for(int i = 0; i < columns; ++i) {
                if(array[i].length != rows) {
                    throw new IllegalArgumentException("All columns of array must have the same length.");
                }
                System.arraycopy(array[i], 0, m.content[i], 0, rows);
            }
            return m;
        }
        
        
        /**
         * Returns an instance of the specified size initialized with zero elements.
         * 
         * @param rows Number of rows.
         * @param columns Number of columns.
         * @return A rows x columns PNMatrix of zeros.
         */
        public Matrix getInstance(final int rows, final int columns) {
            return new Matrix(rows, columns);
        }
        
        
        /**
         * Returns an identity matrix of the specified size.
         * 
         * @param rows Number of rows.
         * @param columns Number of columns.
         * @return A rows x columns identity PNMatrix where the diagonal contains
         *         1 elements and all other elements are 0.
         */
        public Matrix identity(final int rows, final int columns) {
            final Matrix id = new Matrix(rows, columns);
            for (int j = Math.min(rows, columns) - 1; j >= 0; --j) {
                id.set(j, j, BigInteger.ONE);
            }
            return id;
        }
        
        
        /**
         * Creates a row-major array representation of this matrix.
         * 
         * @return Row-major array representation of this matrix.
         */
        public BigInteger[][] toRowMajorArray() {
            final BigInteger[][] ar = new BigInteger[this.rows][this.columns-this.numFreeColumns];
            for(int j = 0, k = 0; j < this.columns; ++j) {
                if(this.columnExists(j)) {
                    for(int i = 0; i < this.rows; ++i) {
                        ar[i][k] = this.get(i, j);
                    }
                    ++k;
                }
            }
            return ar;
        }
        
        
        /**
         * Creates a column-major array representation of this matrix.
         * 
         * @return Column-major array representation of this matrix.
         */
        public BigInteger[][] toColumnMajorArray() {
            final BigInteger[][] ar = new BigInteger[this.columns-this.numFreeColumns][this.rows];
            for(int j = 0; j < this.columns; ++j) {
                if(this.columnExists(j)) {
                    System.arraycopy(this.content[j], 0, ar[j], 0, this.rows);
                }
            }
            return ar;
        }
        
        
        /**
         * Clears the internal data.
         * <b>The matrix must not be used after this call.</b>
         * This method exists for memory performance reasons only.
         */
        public void kill() {
            this.rows = -1;
            this.columns = -1;
            this.content = null;
            this.freeColumns = null;
            this.numFreeColumns = -1;
        }
        
        
        /**
         * Removes the column with the given index from the matrix.
         * This does not change the matrix' column dimension as it is performed in-place for performance reasons.
         * The column is marked as inaccessible and {@link #columnExists(int)} returns <code>false</code>.
         * 
         * @param column A column index.
         * @see #columnExists(int)
         * @see #addColumn()
         */
        public void killColumn(final int column) {
            if(this.columnExists(column)) {
                this.content[column] = null;
                this.ensureSizeFreeColumns(1);
                this.freeColumns[this.numFreeColumns++] = column;
            }
        }
        
        
        // ensure the stack for free column indices is large enough
        private void ensureSizeFreeColumns(final int increment) {
            if(this.freeColumns.length <= this.numFreeColumns + increment-1) {
                final int[] temp = new int[this.freeColumns.length+increment+10];
                System.arraycopy(this.freeColumns, 0, temp, 0, this.freeColumns.length);
                this.freeColumns = temp;
            }
        }
        
        
        /**
         * Adds a column to the matrix and returns the new columns index.
         * 
         * If there are spare columns left, this does not change the matrix'
         * column dimension and it is performed in-place.
         * The column is accessible and {@link #columnExists(int)} returns <code>true</code>.
         * 
         * @return The index of the column created.
         * @see #killColumn(int)
         * @see #columnExists(int)
         */
        public int addColumn() {
            if(this.numFreeColumns != 0) {
                final int column = this.freeColumns[--this.numFreeColumns];
                this.content[column] = new BigInteger[this.getRows()];
                return column;
            }
            final BigInteger[][] temp = new BigInteger[this.columns + INCREMENT+1][];
            final int column = this.columns;
            System.arraycopy(this.content, 0, temp, 0, this.columns);
            temp[this.columns] = new BigInteger[this.rows];
            this.content = temp;
            this.columns = this.content.length;
            
            this.ensureSizeFreeColumns(INCREMENT);
            for(int i = 0; i < INCREMENT; ++i) {
                this.freeColumns[i] = column + i + 1;
            }
            this.numFreeColumns = INCREMENT;
            
            return column;
        }

        
        /**
         * Checks whether the column with the given index exists in the matrix.
         * 
         * @param column A column index.
         * @return <code>true</code> if the column may be accessed.
         */
        public boolean columnExists(final int column) {
            return this.content[column] != null;
        }
        
        
        /**
         * Retrieves the specified element from the matrix.
         * 
         * @param row A row index.
         * @param column A column index.
         * @return The matrix' element at the specified position.
         * @see #columnExists(int)
         */
        public BigInteger get(final int row, final int column) {
            final BigInteger value = this.content[column][row];
            return value == null ? this.content[column][row] = BigInteger.ZERO : value;
        }
        
        
        /**
         * Sets the value of the specified element to <code>s</code>.
         * 
         * @param row A row index.
         * @param column A column index.
         * @param s The new value.
         * @see #columnExists(int)
         */
        public void set(final int row, final int column, final BigInteger s) {
            this.content[column][row] = s;
        }

        
        @Override
        public int hashCode() {
            return Arrays.hashCode(content);
        }


        @Override
        public boolean equals(final Object obj) {
            if(this == obj)
                return true;
            if(obj == null)
                return false;
            if(getClass() != obj.getClass())
                return false;
            Matrix other = (Matrix)obj;
            if(columns != other.columns)
                return false;
            if(!Arrays.equals(content, other.content))
                return false;
            if(rows != other.rows)
                return false;
            return true;
        }


        /**
         * Returns the number of columns in the matrix.
         * Note that not every column has to be accessible.
         * 
         * @return The number of columns in the matrix.
         * @see #columnExists(int)
         */
        public int getColumns() {
            return this.columns;
        }
        
        
        /**
         * Returns the number of rows in the matrix.
         * 
         * @return Number of rows in the matrix.
         */
        public int getRows() {
            return this.rows;
        }

        
        /**
         * Subtracts value <code>s</code> from the element at the specified index.
         * 
         * @param row A row index.
         * @param column A column index.
         * @param s The value to subtract.
         * @see #columnExists(int)
         */
        public void subtract(final int row, final int column, final BigInteger s) {
            this.set(row, column, this.get(row, column).subtract(s));
        }
        
        
        /**
         * Divides a column by the value <code>s</code>.
         * 
         * @param column A column index.
         * @param s The value to divide by.
         * @see #columnExists(int)
         */
        public void divide(final int column, final BigInteger s) {
            for (int i = 0; i < this.rows; i++) {
                this.set(i, column, this.get(i, column).divide(s));
            }
        }

        
        /**
         * Returns an array of columns with non-minimal support.
         * The array may be longer than the actual number of non-minimal support columns.
         * In this case, the indices are succeeded by a -1 entry in the array.
         * 
         * @return An array of column indices with non-minimal support.
         */
        public int[] findNonMinimal() {
            int[] indices = new int[10];
            int count = 0;
            for (int i = 0; i < this.columns; i++) {
                if(!this.columnExists(i)) {
                    continue;
                }
                for (int j = i + 1; j < this.columns; j++) {
                    if(!this.columnExists(j)) {
                        continue;
                    }
                    int r = this.getGEQ(i, j);
                    if (r != -1) {
                        if(indices.length == count) {
                            final int[] tmp = new int[count + 10];
                            System.arraycopy(indices, 0, tmp, 0, count);
                            indices = tmp;
                        }
                        indices[count++] = r;
                    }
                }
            }
            if(count < indices.length) {
                indices[count] = -1;
            }

            return indices;
        }

        
        /**
         * Compares two columns of this matrix element-wise and returns the index of a column
         * which is greater than or equal to the other (or -1) if they are incomparable. 
         * 
         * @param colA A column index.
         * @param colB A second column index.
         * @return The index of the greater (or equal) column or -1 if they are incomparable.
         */
        public int getGEQ(final int colA, final int colB) {
            boolean aGEQb = true;
            boolean bGEQa = true;
            for (int i = 0; i < this.rows; i++) {
                final boolean a = !this.get(i, colA).equals(BigInteger.ZERO);
                final boolean b = !this.get(i, colB).equals(BigInteger.ZERO);
                if (aGEQb && !a && b) {
                    if (!bGEQa) {
                        return -1;
                    }
                    aGEQb = false;
                } else if (bGEQa && a && !b) {
                    if (!aGEQb) {
                        return -1;
                    }
                    bGEQa = false;
                }
            }

            return aGEQb ? colA : colB;
        }
        
        /* (non-Javadoc)
         * @see petruchio.interfaces.petrinet.Matrix#transpose()
         */
        public void transpose() {
            this.content = this.toRowMajorArray();
            this.rows = this.columns-this.numFreeColumns;
            this.columns = this.content.length;
            this.numFreeColumns = 0;
            this.freeColumns = new int[10];
        }
        
        @Override
        public String toString() {
            final StringBuilder s = new StringBuilder();
            s.append(this.rows);
            s.append('x');
            s.append(this.columns-this.numFreeColumns);
//            s.append('\n');
//            for(int i = 0; i < this.rows; ++i) {
//                for(int j = 0; j < this.columns; ++j) {
//                    if(!this.columnExists(j)) {
//                        continue;
//                    }
//                    s.append(this.get(i, j));
//                    s.append(' ');
//                }
//                s.append('\n');
//            }
//            s.append(" hash:");
//            s.append(this.hashCode());
            return s.toString();
        }
    }
    
}
