/*
 * Project:  Petruchio
 * Filename: TransitionComparator.java
 * Created:  21.09.2007
 * Author:   Tim Strazny
 */
package petruchio.cov;

import java.io.Serializable;
import java.util.Comparator;
import java.util.IdentityHashMap;
import java.util.Map;

import petruchio.interfaces.Resettable;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;

/**
 * This comparator implements an order on transitions in respect of the
 * transition's token balance. By the term &quot;produce&quot; the
 * difference of tokens created minus tokens consumed is meant.
 * If transition T1 produces more tokens than another transition T2,
 * then T1 would be smaller than T2. Using this order when processing
 * transitions on new nodes in a coverability graph may lead to a
 * smaller graph as using no particular order on transitions.
 *
 * @author Tim Strazny
 */
public class TransitionComparator implements Comparator<Transition>, Serializable, Resettable {
	
	/**
     * 
     */
    private static final long serialVersionUID = 9004333409733521525L;
    private final Map<Transition, Integer> balance = new IdentityHashMap<Transition, Integer>();
	
	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		this.balance.clear();
	}
	
	
	/**
	 * This comparator implements an order on transitions in respect of the
	 * transition's token balance. By the term &quot;produce&quot; the
	 * difference of tokens created minus tokens consumed is meant.
	 * If transition T1 produces more tokens than another transition T2,
	 * then T1 would be smaller than T2. Using this order when processing
	 * transitions on new nodes in a coverability graph may lead to a
	 * smaller graph as using no particular order on transitions.
	 *
	 * @param o1 A transition
	 * @param o2 A transition
	 * @return <code>&lt;0</code> if <code>o1</code> produces more tokens
	 *            than <code>o2</code>.
	 *         <code>0</code> if <code>o1</code> produces the same number of
	 *            tokens as <code>o2</code>.
	 *         <code>&gt;0</code> else, i.e. <code>o2</code> produces more
	 *            tokens than <code>o1</code>.
	 */
    @Override
    public int compare(final Transition o1, final Transition o2) {
        if(o1 == o2) {
            return 0;
        }
        
        final int b1 = this.getBalance(o1);
        final int b2 = this.getBalance(o2);
        if(b1 == b2) {
        	return 0;
        } else if(b1 < b2) {
        	return -1;
        } else {
        	return 1;
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param t
     * @return
     */
    private int getBalance(final Transition t) {
    	Integer b = this.balance.get(t);
        if(b == null) {
            int sum = 0;
            for(final PTArc arc : t.getInput()) {
                sum -= arc.getWeight();
            }
            for(final TPArc arc : t.getOutput()) {
                sum += arc.getWeight();
            }
            b = Integer.valueOf(sum);
            this.balance.put(t, b);
        }
        return b.intValue();
    }
    
}