/*
 * Project:  Petruchio
 * Filename: CompilerProperty.java
 * Created:  23.05.2007
 * Author:   Tim Strazny
 */
package petruchio.interfaces;


/**
 * A described property that is accepted and understood by another class. 
 * 
 * @author Tim Strazny
 */
public interface CompilerProperty {
    
    /**
     * An extended boolean to enable or disable costly tasks depending of the costs. 
     * 
     * @author Tim Strazny
     */
    public enum ExtendedBoolean {
        /** true */
        TRUE,
        
        /** false */
        FALSE,
        
        /** true iff it is for free */
        TRUE_IF_FOR_FREE;
    }
    
    
    /**
     * Returns the key of this property.
     * Via this key the property may be set, f.e.
     * <code>ExampleProperty=example.Value</code>.
     * 
     * @return The property's key
     */
    public String getKey();
    
    
    /**
     * Gives the class of values this property expects.
     * F.e. if the class is {@link java.lang.Integer} this property expects an
     * integer value. Some properties may indicate in their description
     * ({@link #getDescription()}) that they accept comma separated lists of
     * such values. In the above example <code>1,-3,2,9</code> would then be a
     * valid right hand side of an assignment to that property.
     * The method {@link #acceptsList()} provides this information, too.
     *
     * @return The class of values that are accepted 
     * @see #acceptsList()
     */
    public Class<?> getValueClass();
   

    /**
     * Returns whether this property accepts a comma separated list of
     * instances of the value class.
     *
     * @return <code>true</code> if this property accepts a comma separated
     *            list of instances of the value class.
     * @see #getValueClass()
     */
    public boolean acceptsList();
    
    
    /**
     * Returns whether this property may be left unset and no default value will
     * be inserted instead.
     *
     * @return <code>true</code> if this property accepts an empty value
     */
    public boolean mayBeEmpty();
 
    
    /**
     * The default value of the property.
     * This may be used if the property is not set or to do some sort of
     * initialization.
     * 
     * @return The property's default value
     */
    public String getDefaultValue();
    
    
    /**
     * Gives a short description of this property.
     * 
     * @return A short description of this property.
     */
    public String getDescription();
    
    
    /**
     * Returns the property to which the key belongs or <code>null</code> if the
     * key is unknown.
     * 
     * @param key The key to look up 
     * @return The property to which the key belongs or <code>null</code> if the
     *         key is unknown.
     */
    public CompilerProperty getFromKey(String key);
    
    
    /**
     * Returns the category this property belongs to or <code>null</code> if
     * this property belongs to no category.
     *
     * @return The category this property belongs to or <code>null</code> if
     *         this property belongs to no category.
     */
    public String getCategory();
    
}
