/*
 * Project: Petruchio
 * File:    Retriever.java
 * Author:  Tim Strazny
 * Created: 15.11.2007
 */
package petruchio.interfaces;

import java.util.List;
import java.util.Map;

import petruchio.interfaces.algorithms.Reaction;
import petruchio.interfaces.algorithms.ReactionFinder;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;
import petruchio.interfaces.pi.PiParser;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessCreator;


/**
 * Allows to retrieve a reaction sequence of processes from a Petri net and an
 * enabled sequence of transitions.
 * 
 * @author Tim Strazny
 */
public interface Retriever<E> extends Resettable, SelfDescribing {
    
    public static final String MEANING_SEP = "::";
    
    public class Step<E> {
    	private Reaction<E> reaction;
    	private Map<Place, Integer> state;
    	
    	public Step(final Reaction<E> reaction, final Map<Place, Integer> state) {
    		this.reaction = reaction;
    		this.state = state;
    	}
		public Reaction<E> getReaction() {
			return this.reaction;
		}
		public void setReaction(Reaction<E> reaction) {
			this.reaction = reaction;
		}
		public Map<Place, Integer> getState() {
			return this.state;
		}
		public void setState(Map<Place, Integer> state) {
			this.state = state;
		}
    }
    
    /**
     * Adds an error handler to be used to report errors.
     *
     * @param handler An error handler to use during the parsing
     */
    public void addErrorHandler(ParserErrorHandler handler);
    
    
    /**
     * Sets the parser class to use.
     *
     * @param parser A parser.
     */
    public void setParser(PiParser parser);
    
    
    /**
     * Sets the process creator class to use.
     *
     * @param processBuilder A process creator
     */
    public void setProcessCreator(ProcessCreator processBuilder);
    
    
    /**
     * Sets the reaction finder class to use.
     *
     * @param reactionFinder A reaction finder.
     */
    public void setReactionFinder(ReactionFinder<E> reactionFinder);
    
    
    /**
     * Sets the Petri net that serves as a part of the input.
     *
     * @param net A Petri net.
     */
    public void setPetriNet(PetriNet net);
    
    
    /**
     * Retrieves the process that corresponds to the initial marking of the
     * Petri net.
     *
     * @return A process for the Petri net's initial marking.
     */
    public Process<E> retrieve();
    
    
    /**
     * Retrieves the process that corresponds to the given marking of the
     * Petri net.
     *
     * @param marking A marking of the Petri net.
     * @return A process for the given Petri net's marking.
     */
    public Process<E> retrieve(Map<Place, Integer> marking);
    
    
    /**
     * Retrieves a reaction sequence of processes from the given Petri net and
     * a given (enabled) sequence of transitions.
     * The first reaction in the list will contain the process corresponding to
     * the Petri net's initial marking as the reacted process and all other
     * fields may be null.
     * Additionally the Petri net will be set to the reached marking!
     *
     * @param sequence An enabled sequence of transitions.
     * @return A sequence of process reactions that corresponds to the sequence
     *         of transitions.
     * @throws UnsupportedOperationException If this method is not supported.
     */
    public List<Step<E>> retrieve(List<Transition> sequence) throws UnsupportedOperationException;
    
}