/*
 * Project:  Petruchio
 * Filename: Reachability.java
 * Created:  21.04.2007
 * Author:   Tim Strazny
 */
package petruchio.interfaces.algorithms;

import java.util.Collection;

import java.util.Map;
import java.util.concurrent.BlockingQueue;

import petruchio.interfaces.CommunicatorCache;
import petruchio.interfaces.Resettable;
import petruchio.interfaces.petrinet.PlacePair;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;


/**
 * A class that implements this interface provides routines to check the
 * reachability of certain Petri net states.
 * The computation should be performed step wise: transitions will be added to
 * the internal graph structure as the petruchio tool generates the Petri net.
 * The computation may be swapped to another thread. Methods are provided to
 * control the needed synchronization.
 * 
 * @author Tim Strazny
 */
public interface Coverability extends Runnable, Resettable {
    
    /**
     * Checks whether this reachability checker has work to do.
     *
     * @return <code>true</code> if the reachability checker is currently
     *           working or has work packages (new transitions) to process.
     */
    public boolean isWorking();
    
    
    /**
     * Clears internal caches and resets the coverability graph.
     */
    public void clearCaches();
    
    
    /**
     * Allow the coverability checker to spawn additional tasks.  
     * 
     * @param numThreads The number of threads the coverability checker may use.
     */
    public void setThreadNumber(final int numThreads);
    
    
    /**
     * If this implementation supports being run in parallel to the main
     * compiler thread, this method should return <code>true</code> and methods
     * {@link #pleaseStopComputation()}, {@link #pleaseTerminate()},
     * {@link #hasTerminated()}, {@link #awaitComputationEnd()},
     * {@link #awaitResults()}, as well as {@link #addTransition(Transition)}
     * and {@link #markUnused(Place)} must be implemented with extra care!
     *
     * @return <code>true</code> iff this implementation may be run in parallel
     *         to the main compiler thread. 
     */
    public boolean supportsThreading();
    
    
    /**
     * Returns <code>true</code> if the reachability checker thread has
     * terminated with an error. A call to {@link #reset()} enables the
     * reachability checker to be run again.
     *
     * @return <code>true</code> if the run method has been exited with an error.
     */
    public boolean hasTerminated();
    
    
    /**
     * Sets the minimal number of tokens required on a place of a marking to
     * allow it being tested for covering another marking. Setting it to
     * <code>1</code> leads to default behavior. A value less than or equal to
     * <code>0</code> stands for &quot;never test covering&quot;.  
     *
     * @param tokens The minimal number of tokens as specified above.
     * @throws UnsupportedOperationException Thrown if this operation is not
     *            supported by the reachability checker implementation.
     */
    public void setMinTokenCount(int tokens)
            throws UnsupportedOperationException;
    
    /**
     * Sets the maximal number of tokens required on a place of a marking to
     * allow it being tested for covering another marking. Setting it to a value
     * less than <code>0</code> leads to default behavior. A value greater equal to
     * <code>0</code> will introduce an omega if a place contains more tokens than
     * the value.
     *
     * @param tokens The maximal number of tokens as specified above.
     * @throws UnsupportedOperationException Thrown if this operation is not
     *            supported by the reachability checker implementation.
     */
    public void setMaxTokenCount(int tokens)
            throws UnsupportedOperationException;
    
    
    /**
     * Returns a measure for the size of the reachability checker. For example
     * the number of nodes in a coverability graph.
     *
     * @return The size of the checker
     */
    public int size();
    
    
    /**
     * Resets the reachability checker and performs the computation for a
     * complete Petri net.
     *
     * @param root A multiset of places that represents the initial marking of
     *            the net.
     * @param trans A set of transitions that have to be processed.
     */
    public void compute(Collection<Place> root, Collection<Transition> trans)
        throws UnsupportedOperationException;
    
    
    /**
     * Tells the reachability checker to stop its computations and stop all the
     * threads created.
     * The reachability checker is not needed anymore and it may terminate
     * completely.
     */
    public void pleaseTerminate();
    
    
    /**
     * Tells the reachability checker to stop its current computations.
     * As the reachability checker may be needed later on, the threads should
     * not be halted.
     */
    public void pleaseStopComputation();
    
    
    /**
     * Returns when possible communication partners are found and added to the
     * communication collection or all work is done.
     */
    public void awaitResults();
    
    
    /**
     * Returns when all work is done.
     */
    public void awaitComputationEnd();
    
    
    /**
     * Signals the coverability checker, that currently there are no more transitions that will be added, so the
     * compiler is just waiting for the coverability checker. 
     */
    public void compilerIsWaiting();
    
    /**
     * Sets the initial marking for the Petri net.
     * Note that all places, that are added to the net later on have to be in
     * post sets of transition's that are added via
     * {@link #addTransition(Transition)} and must not be marked under the
     * initial marking.
     *
     * @param state A multiset of places that represents the initial marking of
     *            the net.
     */
    public void setStartState(Collection<Place> state);
    
    
    /**
     * Adds a transition to this reachability checker.
     * As a pair of places is recognized as 'markable' it will be added to the
     * passed collection {@link #setCommunications(Collection)}.
     * If there are new places adjacent to the transition, they form a sub set
     * of the transition's post set.
     * All places that are in the transition's post set and not in it's pre set
     * have to be unmarked under the initial marking.
     * 
     * @param t A transition
     */
    public void addTransition(Transition t);
    
    
    /**
     * Tells the reachability checker to which collection possible communication
     * partners shall be added.
     *
     * @param c The collection to add possible communication partners to
     */
    public void setCommunications(BlockingQueue<PlacePair> c);
    
    
    /**
     * Enable or disable the generation of possible communication partners.
     *
     * @param b Enable or disable the generation of possible communication
     *            partners.
     */
    public void setCheckMarkable(boolean b);
    
    
    /**
     * Calculates the smallest upper bounds for the number of fire steps for the
     * given set of transitions.
     * If a transition can fire unboundedly often, a negative value is returned.
     *
     * @param transitions A set of transitions
     * @return A mapping from a transition to its smallest upper bound for the
     *            number of firing steps or a negative value if the transition
     *            can fire unboundedly often.
     */
    public Map<Transition, Integer> getFiringBounds(Collection<Transition> transitions);
    
    
    /**
     * Calculates the smallest upper bounds (boundedness) for the number of
     * tokens each place in the given set could be marked with.
     * If a place can be marked with unboundedly many tokens, a negative value
     * is returned.
     *
     * @param places A set of places
     * @return A mapping from a place to its smallest upper bound for the
     *            number of tokens it may be marked withor a negative value if
     *            the place can be marked with unboundedly many tokens.
     * @throws UnsupportedOperationException Thrown if this operation is not
     *            supported by the reachability checker implementation.
     */
    public Map<Place, Integer> getBounds(Collection<Place> places)
            throws UnsupportedOperationException;
    
    
    /**
     * Tells the reachability checker that the specific place will not be marked
     * again. The reachability checker may then remove the place from the states
     * it has computed, when all states that mark this place have been explored.
     * This could lead to a smaller size of the explored state space.
     *
     * @param p
     * @throws UnsupportedOperationException Thrown if this operation is not
     *            supported by the reachability checker implementation.
     */
    public void markUnused(Place p)
            throws UnsupportedOperationException;

    
    /**
     * Computes the set of (generalized) deadlocks for the current reachability
     * checker state space.
     *
     * @param assumeBoundedness Allow or disallow the operation to assume that
     *            the net is bounded.
     * @return A set of String representations of (possibly generalized)
     *            deadlocks.
     * @throws UnsupportedOperationException Thrown if this operation is not
     *            supported by the reachability checker implementation.
     */
    public Collection<String> getDeadlocks(boolean assumeBoundedness)
            throws UnsupportedOperationException;
    
    
    /**
     * Given a set of places and a set of transitions, all places that can be
     * marked by the net are removed from the first set and all transitions that
     * can fire are removed from the second set. This leaves only the places,
     * that can never be marked and the transitions that are dead in the sets.
     *
     * @param p A set of places to check
     * @param t A set of transitions to check
     * @throws UnsupportedOperationException Thrown if this operation is not
     *            supported by the reachability checker implementation.
     */
    public void keepDeadElements(Collection<Place> p, Collection<Transition> t)
            throws UnsupportedOperationException;
    
    
    /**
     * @param communicators
     */
    public void setCommunicatorCache(CommunicatorCache<?> communicators);
    
    
    /**
     * @param p1
     * @param p2
     */
    public void doNotCheck(Place p1, Place p2);
    
}