/*
 * Project:  Petruchio
 * Filename: Reaction.java
 * Created:  15.12.2006
 * Author:   Tim Strazny
 */
package petruchio.interfaces.algorithms;

import java.util.Collection;

import petruchio.interfaces.petrinet.Transition;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;

/**
 * A reaction in terms of the pi-calculus.
 * 
 * @param <E> The data type of the processes
 * @author Tim Strazny
 */
public interface Reaction<E> {

    /**
     * The type of a reaction.
     * An enum is used to avoid the expensive instanceof-operator.
     * 
     * @author Tim Strazny
     */
    public enum Type {
        /** A tau reaction or a reference. */
        SINGLE,
        
        /** The communications of two parallel processes. */
        COMMUNICATION
    }
    
    
    /**
     * 
     * 
     * @author Tim Strazny
     */
    public enum Independent {
        /** The reaction is sequentially independent. */
        TRUE,
        
        /** The reaction is sequentially dependent. */
        FALSE,
        
        /** It is unknown whether the reaction is sequentially independent. */
        NOT_SET
    }
    
    
    /**
     * Returns the type of this reaction.
     *
     * @return The reactions' type
     */
    public Reaction.Type getType();
    
    
    /**
     * The process that emerged from the reaction.
     *
     * @return The reacted process
     */
    public Process<E> getReactedProcess();
    
    
    /**
     * Sets the transition this reaction is represented by.
     * @param t A transition
     */
    public void setTransition(Transition t);
    
    
    /**
     * Returns the transition that represents this reaction.
     *
     * @return The transition that represents this reaction.
     */
    public Transition getTransition();
    
    
    /**
     * Sets the process that emerged from the reaction.
     *
     * @param reactedProcess The reacted Process
     */
    public void setReactedProcess(Process<E> reactedProcess);
    

    /**
     * Set the information whether this reaction is sequentially independent
     * from others.
     *
     * @param sequentiallyIndependent Whether this reaction is sequentially independent
     */
    public void setSequentiallyIndependent(Independent sequentiallyIndependent);
    
    
    /**
     * Returns whether this reaction is sequentially independent from others.
     *
     * @return Whether this reaction is sequentially independent from others.
     */
    public Independent isSequentiallyIndependent();
    
    
    /**
     * Allows to add choice compositions that have reacted during this reaction.
     *
     * @param choice A choice composition that has reacted
     */
    public void addChoice(ProcessComposition<E> choice);
    
    
    /**
     * Returns the choice compositions that have reacted during this reaction.
     *
     * @return The choice compositions that have reacted during this reaction.
     */
    public Collection<ProcessComposition<E>> getChoices();
    
}