/*
 * Project:  Petruchio
 * Filename: ReactionFinder.java
 * Created:  30.11.2006
 * Author:   Tim Strazny
 */
package petruchio.interfaces.algorithms;

import java.util.Collection;
import java.util.List;

import petruchio.interfaces.Resettable;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessManager;


/**
 * A class that implements this interface provides method to compute one step
 * reactions of processes and one step communications of two processes via free
 * names. Additional methods are provided to simplify the work with pi calculus
 * processes. 
 * 
 * @param <E> The data type of processes.
 * @author Tim Strazny
 */
public interface ReactionFinder<E> extends Resettable {
    
    /**
     * Computes all the pi calculus reactions the given process can do in one
     * step and returns them.
     *
     * @param p A process in strict restricted form, unique bound names and
     *            disjoint sets of free and bound names
     * @return The collection of all the reactions the given process can perform
     *            in one step.
     */
    public Collection<Reaction<E>> getReactions(Process<E> p);
    
    
    /**
     * Computes all the pi calculus communication reactions the two given
     * process can do in one step and returns them.
     * The communications have to involve both processes and thus use free names
     * as channels.
     *
     * @param p1 A process in strict restricted form, unique bound names and
     *            disjoint sets of free and bound names
     * @param p2 A process in strict restricted form, unique bound names and
     *            disjoint sets of free and bound names
     * @return The collection of communication reactions that involve both
     *            processes (one step).
     */
    public Collection<Reaction<E>> getReactionsBetween(Process<E> p1, Process<E> p2);

    
    /**
     * Sets the process manager to be used that contains associations between
     * process identifiers, formal parameters and processes.
     * This is needed to resolve reference reactions. 
     *
     * @param pmanager The process manager to use
     */
    public void setProcessManager(ProcessManager<E> pmanager);
    
    
    /**
     * Clears internal caches.
     */
    public void clearCaches();
    
    
    /**
     * Sets the process creator to be used when processes have to be created.
     *
     * @param pbuilder The process creator to use
     */
    public void setProcessCreator(ProcessCreator pbuilder);
    
    
    /**
     * Given a process and a set of excluded names, this method creates a
     * process that is structurally congruent to the original process but the
     * sets of free and bound names are disjoint and bound names are unique.
     * This means, that if a name is free in one sub processes, it is never been
     * bound in any sub process and if a name is bound in one sub process it is
     * never been bound in any other sub process. 
     * Whenever a new name is introduced by alpha conversion, it is ensured that
     * it's not in the set of excluded names.
     * 
     * @param p A process
     * @param excludedNames A set of names
     * @return A process that is structurally congruent to the given process
     *            where the sets of free and bound names are disjoint and the
     *            bound names are unique.
     */
    public Process<E> uniqueBoundNames(Process<E> p, Collection<Name> excludedNames);
    
    
    /**
     * Uses the rule of scope extrusion to move restrictions up in the syntax
     * tree as far as possible.
     * A kind of least restricted form is produced which is structurally
     * congruent to the original process.
     *
     * @param p A process with unique bound names and disjoint sets of free and
     *            bound names
     * @return A structurally congruent process to the given process with unique
     *            bound names, disjoint sets of free and bound names and in
     *            least restricted form.
     */
    public Process<E> liftRestrictions(Process<E> p);
    
    
    /**
     * Returns the set of process IDs referenced by this process.
     *
     * @param p A process.
     * @return The set of process IDs referenced by this process.
     */
    public Collection<Name> getProcessIDs(Process<E> p);
    

    /**
     * Returns the number of null process occurrences within the given process.
     *
     * @param p A process.
     * @return The number of null process occurrences withing the process.
     */
    public int getNullProcessCount(Process<E> p);
    
    
    /**
     * Returns a fresh name that is not contained in the sets of excluded names.
     *
     * @param name The original name (it may be used as a starting point for the
     *            search for a new name).
     * @param excluded A collection of excluded names (or <code>null</code>)
     * @param moreExcluded Another collection of excluded names (or
     *            <code>null</code>)
     * @return A name that (may have the original name as a prefix and) is not
     *            contained in the collections of excluded names.
     */
    public Name newName(Name name, Collection<Name> excluded, Collection<Name> moreExcluded);
    
    
    /**
     * Returns a process that matches the given process where the names in
     * <code>search</code> have been replaced by the names in
     * <code>replace</code>. The two lists must have the same number of elements
     * and the <code>i</code>th name in list <code>search</code> is replaced by
     * the <code>i</code>th name in list <code>replace</code>. 
     *
     * @param p A process
     * @param search A list of names to be searched for
     * @param replace A list of names to replace the names searched for 
     * @return
     */
    public Process<E> replaceNames(Process<E> p, List<Name> search, 
                    List<Name> replace);

    
    /**
     * Checks if the given process could possibly communicate via a free name.
     * I.e. if a process could be constructed so that a call to
     * {@link #getReactionsBetween(Process, Process)} returns a non empty
     * collection of communication reactions. 
     *
     * @param p A process in strict restricted form, unique bound names and
     *            disjoint sets of free and bound names
     * @return <code>true</code> if the given process has a sequential part with
     *            an input or output prefix that uses a free name as its channel.
     */
    public Collection<ActionPrefix> getAvailableCommunicationsViaFreeNames(Process<E> p);
    
    
    /**
     * @param p
     * @return
     */
    public boolean containsCommunicationViaFreeName(Process<E> p);
    
    
    /**
     * Returns the set of all names in the process that are bound.
     *
     * @param p  A process
     * @return   The set of bound names of the given process.
     */
    public Collection<Name> boundNames(Process<E> p);
    
    
    /**
     * Returns the set of all names in the process that are free.
     *
     * @param p  A process
     * @return   The set of free names of the given process.
     */
    public Collection<Name> freeNames(Process<E> p);
    
    
    /**
     * Copies a process.
     * At least the very first (top level) process object is copied. The sub
     * processes may be reused.
     *
     * @param p A process
     * @return A copy of the process where at least the top level process is
     *            a fresh instance.
     */
    public Process<E> copyTopLevel(Process<E> p);
    
    
    /**
     * Checks whether the given process is structurally congruent to the
     * <code>0</code>-process.
     *
     * @param p A process in strict restricted form, unique bound names and
     *            disjoint sets of free and bound names
     * @return <code>true</code> if the process is structurally congruent to the
     *            <code>0</code>-process.
     */
    public boolean isBareNull(Process<E> p);
    
}