/*
 * Project:  Petruchio
 * Filename: StructuralBoundednessChecker.java
 * Created:  24.08.2007
 * Author:   Tim Strazny
 */
package petruchio.interfaces.algorithms;

import java.util.Collection;
import java.util.Map;

import petruchio.interfaces.Resettable;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;


/**
 * A class implementing this interface provides methods for testing whether a
 * transition that is added to a Petri net causes the structural unboundedness
 * in the terms of Meyer's structural semantics.
 * 
 * @param <V> The data type of the processes.
 * @author Tim Strazny
 */
public interface StructuralBoundedness<V> extends Resettable {
    
    /**
     * Sets the structural congruence checker to use.
     *
     * @param structCong A structural congruence checker
     */
    public void setStructuralCongruence(StructuralCongruence<V> structCong);
    
    
    /**
     * Empties internal caches.
     */
    public void clearCaches();
    
    
    /**
     * Sets the reaction finder to use.
     *
     * @param reactionFinder A reaction finder
     */
    public void setReactionFinder(ReactionFinder<V> reactionFinder);
    
    
    /**
     * Checks if the given transition causes structural unboundedness and thus
     * would lead to an infinite structural semantics. The restrictions that
     * cause the unboundedness are returned or <code>null</code> if no
     * structural unboundedness could be identified.
     *
     * @param trans A transition that is added to the net.
     * @param place2proc A mapping from net places to corresponding processes.
     * @param update <code>true</code> if exists a place in the transition's
     *            post set that has been created before the transition (and thus
     *            is in the post sets of other transitions).
     * @return <code>null</code> is returned if no case of structural
     *            unboundedness is identified. If structural unboundedness is
     *            identified, a set of restricted names that cause the
     *            unboundedness.
     */
    public Collection<Name> causesStructuralUnboundedness(Transition trans,
                    Map<Place, Process<V>> place2proc, boolean update);
    
}