/*
 * Project:  Petruchio
 * Filename: StructCongChecker.java
 * Created:  30.11.2006
 * Author:   Tim Strazny
 */
package petruchio.interfaces.algorithms;

import java.util.Comparator;

import petruchio.interfaces.Resettable;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessCreator;


/**
 * A class that implements this interface has to be able to check whether two
 * processes are structurally congruent.
 *
 * @param <E> The data type of processes.
 * @author Tim Strazny
 */
public interface StructuralCongruence<E> extends Resettable {
    
    /**
     * Sets the process creator to be used when processes have to be created.
     *
     * @param pbuilder The process creator to use
     */
    public void setProcessCreator(ProcessCreator pbuilder);
    
    
    /**
     * Clears internal caches.
     */
    public void clearCaches();
    
    
    /**
     * Checks if two given processes are structurally congruent.
     * The processes have to be in strict restricted form and the bound and free
     * names have to be disjoint, as well as all the bound names have to be
     * unique, to simplify the work this method has to accomplish.
     *
     * @param p1 A process in strict restricted form, unique bound names and
     *            disjoint sets of free and bound names
     * @param p2 A process in strict restricted form, unique bound names and
     *            disjoint sets of free and bound names
     * @return <code>true</code> if the two processes are structurally congruent.
     */
    public boolean structurallyCongruent(Process<E> p1, Process<E> p2);

    
    /**
     * Returns a comparator on processes to allow the proper use of binary
     * searches in lists of processes.
     *
     * @return A comparator on processes.
     */
    public Comparator<Process<E>> getComparator();
    
}