/*
 * Project:  Petruchio
 * Filename: Matrix.java
 * Created:  02.06.2009
 * Author:   Tim Strazny
 */
package petruchio.interfaces.petrinet;

/**
 * @author Tim Strazny
 *
 */
public interface LongMatrix {
    
    /**
     * Creates a row-major array representation of this matrix.
     * 
     * @return Row-major array representation of this matrix.
     */
    public long[][] toRowMajorArray();
    
    /**
     * Creates a column-major array representation of this matrix.
     * 
     * @return Column-major array representation of this matrix.
     */
    public long[][] toColumnMajorArray();
    
    /**
     * Clears the internal data.
     * <b>The matrix must not be used after this call.</b>
     * This method exists for memory performance reasons only.
     */
    public void kill();
    
    
    /**
     * Removes the column with the given index from the matrix.
     * This does not change the matrix' column dimension as it is performed in-place for performance reasons.
     * The column is marked as inaccessible and {@link #columnExists(int)} returns <code>false</code>.
     * 
     * @param column A column index.
     * @see #columnExists(int)
     * @see #addColumn()
     */
    public void killColumn(int column);
    
    
    /**
     * Adds a column to the matrix and returns the new columns index.
     * 
     * If there are spare columns left, this does not change the matrix'
     * column dimension and it is performed in-place.
     * The column is accessible and {@link #columnExists(int)} returns <code>true</code>.
     * 
     * @return The index of the column created.
     * @see #killColumn(int)
     * @see #columnExists(int)
     */
    public int addColumn();
    
    /**
     * Checks whether the column with the given index exists in the matrix.
     * 
     * @param column A column index.
     * @return <code>true</code> if the column may be accessed.
     */
    public boolean columnExists(int column);
    
    /**
     * Retrieves the specified element from the matrix.
     * 
     * @param row A row index.
     * @param column A column index.
     * @return The matrix' element at the specified position.
     * @see #columnExists(int)
     */
    public long get(int row, int column);
    
    /**
     * Sets the value of the specified element to <code>s</code>.
     * 
     * @param row A row index.
     * @param column A column index.
     * @param s The new value.
     * @see #columnExists(int)
     */
    public void set(int row, int column, long s);
    
    /**
     * Returns the number of columns in the matrix.
     * Note that not every column has to be accessible.
     * 
     * @return The number of columns in the matrix.
     * @see #columnExists(int)
     */
    public int getColumns();
    
    /**
     * Returns the number of rows in the matrix.
     * 
     * @return Number of rows in the matrix.
     */
    public int getRows();
    
    /**
     * Subtracts value <code>s</code> from the element at the specified index.
     * 
     * @param row A row index.
     * @param column A column index.
     * @param s The value to subtract.
     * @see #columnExists(int)
     */
    public void subtract(int row, int column, long s);
    
    /**
     * Divides a column by the value <code>s</code>.
     * 
     * @param column A column index.
     * @param s The value to divide by.
     * @see #columnExists(int)
     */
    public void divide(int column, long s);
    
    /**
     * Returns an array of columns with non-minimal support.
     * The array may be longer than the actual number of non-minimal support columns.
     * In this case, the indices are succeeded by a -1 entry in the array.
     * 
     * @return An array of column indices with non-minimal support.
     */
    public int[] findNonMinimal();
    
    /**
     * Compares two columns of this matrix element-wise and returns the index of a column
     * which is greater than or equal to the other (or -1) if they are incomparable. 
     * 
     * @param colA A column index.
     * @param colB A second column index.
     * @return The index of the greater (or equal) column or -1 if they are incomparable.
     */
    public int getGEQ(int colA, int colB);
    
    /**
     * Transposes the matrix.
     */
    public void transpose();

}
