/*
 * Project:  Petruchio
 * Filename: PetriNet.java
 * Created:  01.12.2006
 * Author:   Tim Strazny
 */
package petruchio.interfaces.petrinet;

import java.util.Collection;
import java.util.Map;

import petruchio.interfaces.Resettable;


/**
 * A simple P/T Petri net with arc weights.
 * 
 * @author Tim Strazny
 */
public interface PetriNet extends Resettable {
    
    /**
     * Creates a new place, adds it to the net and returns it.
     *
     * @param name The name of the new place
     * @param marking The initial marking of the place
     * @return The newly created place.
     */
    public Place addPlace(String name, int marking);
    
    
    /**
     * Creates a new transition, adds it to the net and returns it.
     *
     * @param name The name of the new transition
     * @return The newly created transition.
     */
    public Transition addTransition(String name);
    
    
    /**
     * Adds an arc from a place to a transition with a specified weight.
     * If there already exists an arc with the same source and target, the
     * weight is added to the existing arc.
     *
     * @param p A place
     * @param t A transition
     * @param weight The arc weight
     * @return The arc created
     */
    public PTArc addArc(Place p, Transition t, int weight);
    
    
    /**
     * Adds an arc from a transition to a place with a specified weight.
     * If there already exists an arc with the same source and target, the
     * weight is added to the existing arc.
     *
     * @param t A transition
     * @param p A place
     * @param weight The arc weight
     * @return The arc created
     */
    public TPArc addArc(Transition t, Place p, int weight);
    
    
    /**
     * 
     *
     * @param in A mapping of places to integers that represent the input arcs
     *            and weights
     * @param out A mapping of places to integers that represent the output arcs
     *            and weights
     * @return A set of transitions with the specified input and output arcs and
     *            weights.
     */
    public Collection<Transition> getTransitionsBetween(Map<Place, Integer> in, Map<Place, Integer> out);
    
    
    /**
     * Sets the header comment for this net.
     *
     * @param header The new header comment
     */
    public void setHeaderComment(String header);
    
    
    /**
     * Returns <code>true</code> if this net has a header comment.
     *
     * @return <code>true</code> if this net has a header comment
     */
    public boolean hasHeaderComment();
    
    
    /**
     * Returns the header comment of this net.
     *
     * @return This net's header comment
     */
    public String getHeaderComment();
    
    
    /**
     * Returns the set of places in this net.
     *
     * @return This net's set of places
     */
    public Collection<Place> getPlaces();
    
    
    /**
     * Returns the set of places in this net that carry the given names.
     *
     * @param names The names
     * @return This net's set of places with the given names.
     */
    public Collection<Place> getPlaces(Collection<String> names);
    
    
    /**
     * Returns the set of places in this net that carry the given name.
     * 
     * @param name The name
     * @return This net's set of places with the given name.
     */
    public Collection<Place> getPlaces(String name);
    
    
    /**
     * Returns the set of transitions in this net.
     *
     * @return This net's set of transitions
     */
    public Collection<Transition> getTransitions();
    
    
    /**
     * Returns the set of transitions in this net that carry the given names.
     * 
     * @param names The names
     * @return This net's set of transitions with the given names
     */
    public Collection<Transition> getTransitions(Collection<String> names);
    
    
    /**
     * Returns the set of transitions in this net that carry the given name.
     *
     * @param name The name
     * @return This net's set of transitions with the given name
     */
    public Collection<Transition> getTransitions(String name);
    
    
    /**
     * Removes a place and all its adjacent arcs from the net.
     *
     * @param p The place to remove
     * @return <code>true</code> if the removal changed the net.
     */
    public boolean removePlace(Place p);
    
    
    /**
     * Removes a transition and all its adjacent arcs from the net.
     *
     * @param t The transition to remove
     * @return <code>true</code> if the removal changed the net.
     */
    public boolean removeTransition(Transition t);
    
    
    /**
     * Removes a P/T arc from the net.
     *
     * @param p The source place of the arc to remove
     * @param t The target transition of the arc to remove
     * @return <code>true</code> if the removal changed the net.
     */
    public boolean removeArc(Place p, Transition t);
    
    
    /**
     * Removes a T/P arc from the net.
     *
     * @param t The source transition of the arc to remove
     * @param p The target place of the arc to remove
     * @return <code>true</code> if the removal changed the net.
     */
    public boolean removeArc(Transition t, Place p);
    
    
    /**
     * Removes a P/T arc from the net.
     *
     * @param arc The arc to remove
     * @return <code>true</code> if the removal changed the net.
     */
    public boolean removeArc(PTArc arc);
    
    
    /**
     * Removes a T/P arc from the net.
     *
     * @param arc The arc to remove
     * @return <code>true</code> if the removal changed the net.
     */
    public boolean removeArc(TPArc arc);
    
    
    /**
     * Sets the name of this net.
     *
     * @param name The new name for this net
     */
    public void setName(String name);
    
    
    /**
     * Returns the name of this net.
     *
     * @return This net's name
     */
    public String getName();
    
    
    /**
     * Returns the number of P/T arcs in this net (not weights).
     *
     * @return The number of P/T arcs
     */
    public int getPTArcCount();
    
    
    /**
     * Returns the number of T/P arcs in this net (not weights).
     *
     * @return The number of T/P arcs
     */
    public int getTPArcCount();
    
    
    /**
     * Clears the Petri net and resets it to the empty net.
     * Only the net elements (places, transitions, arcs) will be removed.
     * Name and additional information will be kept.
     */
    @Override
    public void reset();
    
}