/*
 * Project:  Petruchio
 * Filename: PetriNetReader.java
 * Created:  05.09.2007
 * Author:   Tim Strazny
 */
package petruchio.interfaces.petrinet;


import java.io.InputStream;

import petruchio.interfaces.ParserErrorHandler;
import petruchio.interfaces.Resettable;

/**
 * A class implementing this interface is capable of building a Petri net from
 * a given InputStream that delivers a net representation the class understands.
 * 
 * An implementing class may provide a <code>public</code> method named
 * {@link #METHOD_READ_TARGETS} accepting a {@link java.lang.String} and a
 * {@link petruchio.interfaces.petrinet.PetriNet}, returning a {@link java.util.Collection}
 * of {@link java.util.Map} instances, each associating {@link petruchio.interfaces.petrinet.Place}
 * objects with {@link java.lang.Integer} objects to read target markings.
 * 
 * An implementing class may provide a <code>public</code> method named
 * {@link #METHOD_READ_INVARIANTS} accepting a {@link java.lang.String} and a
 * {@link petruchio.interfaces.petrinet.PetriNet}, returning a {@link java.util.Collection}
 * of {@link java.util.Map} instances, each associating {@link petruchio.interfaces.petrinet.Place}
 * objects with {@link java.lang.Integer} objects to read place invariants.
 * 
 * @author Tim Strazny
 */
public interface PetriNetReader extends Resettable {

    public static final String METHOD_READ_TARGETS = "loadTargetMarkings";
    @SuppressWarnings("rawtypes")
    public static final Class[] METHOD_READ_TARGETS_PARAM_TYPES = { java.lang.String.class, petruchio.interfaces.petrinet.PetriNet.class };
    
    public static final String METHOD_READ_INVARIANTS = "loadInvariants";
    @SuppressWarnings("rawtypes")
    public static final Class[] METHOD_READ_INVARIANTS_PARAM_TYPES = { java.lang.String.class, petruchio.interfaces.petrinet.PetriNet.class };
    
    /**
     * Returns a comma separated list of the default file name extension for 
     * this Petri net format.
     *
     * @return The file name extensions for this Petri net format
     */
    public String getDefaultExtensions();
    
    
    /**
     * Builds a Petri net from the given representation.
     *
     * @return The Petri net that corresponds to the representation.
     */
    public PetriNet read();
    
    /**
     * Builds a Petri net from the given representation.
     * Used to specify the exact implementation of the PetriNet interface.
     */
    public void read(PetriNet net);
    
    
    /**
     * Sets the input for the reader.
     *
     * @param in The input stream to the Petri net representation
     */
    public void setInput(InputStream in);
    
    
    /**
     * Adds an error handler to report errors to.
     *
     * @param handler The error handler to add.
     */
    public void addErrorHandler(ParserErrorHandler handler);

}