/*
 * Project:  Petruchio
 * Filename: Place.java
 * Created:  01.12.2006
 * Author:   Tim Strazny
 */
package petruchio.interfaces.petrinet;

import java.util.Collection;


/**
 * A Petri net place.
 * 
 * @author Tim Strazny
 */
public interface Place extends NetElement {
    
    public static final int BOUND_UNBOUNDED = Integer.MAX_VALUE;
    public static final int BOUND_UNKNOWN   = -1;

    /**
     * Returns the initial token count of this place.
     *
     * @return The initial marking of this place
     */
    public int getMarking();
    
    
    /**
     * Adds a token to the initial token count of this place.
     */
    public void addToken();
    
    
    /**
     * Adds tokens to the initial token count of this place.
     * 
     * @param tokens The number of tokens to add.
     */
    public void addTokens(int tokens);
    
    
    /**
     * Sets the initial token count of this place.
     * 
     * @param tokens The number of tokens.
     */
    public void setMarking(int tokens);
    
    
    /**
     * If the place is marked as final, this method returns true.
     *
     * @return <code>true</code> iff the place is marked as final.
     */
    public boolean isFinal();
    
    /**
     * If the place is marked as parametric, this method returns true.
     * 
     * @return <code>true</code> iff the place is marked as parametric.
     */
    public boolean isParametric();
    
    /**
     * Sets or removes the "parametric" mark for this place
     *
     * @param isParametric The new return value of {@link #isParametric()}.
     */
    public void setParametric(boolean isParametric);
    
    
    /**
     * Sets or removes the "final" mark for this place
     *
     * @param isFinal The new return value of {@link #isFinal()}.
     */
    public void setFinal(boolean isFinal);
    
    
    /**
     * Returns the name of this place.
     *
     * @return This place's name
     */
    public String getName();
    
    
    /**
     * Returns the meaning of this place.
     *
     * @return The meaning of this place.
     */
    public String getMeaning();
    
    
    /**
     * Sets the meaning of this place.
     *
     * @param meaning The new meaning of this place.
     */
    public void setMeaning(String meaning);
    
    
    /**
     * Sets the name of this place.
     *
     * @param name The new name for this place
     */
    public void setName(String name);

    
    /**
     * Returns the x coordinate of this place's position.
     *
     * @return The x coordinate of this place
     */
    public int getX();
    
    
    /**
     * Returns the y coordinate of this place's position.
     *
     * @return The y coordinate of this place
     */
    public int getY();
    
    
    /**
     * Sets the x coordinate of this place's position.
     *
     * @param x The new x coordinate of this place
     */
    public void setX(int x);
    
    
    /**
     * Sets the y coordinate of this place's position.
     *
     * @param y The new y coordinate of this place
     */
    public void setY(int y);
 
    
    /**
     * Returns the smallest <code>k</code> for which this place is
     * <code>k</code>-bounded.
     * If this place is unbounded, a <code>BOUND_UNBOUNDED</code> is returned.
     * If this <code>k</code> is unknown, <code>BOUND_UNKNOWN</code> is returned.
     * 
     * @return A <code>k</code> for which this place is <code>k</code>-bounded.
     */
    public int getBound();
    
    
    /**
     * Sets the smallest <code>k</code> for which this place is
     * <code>k</code>-bounded.
     * If this place is unbounded, <code>BOUND_UNBOUNDED</code> should be used.
     * If this <code>k</code> is unknown, <code>BOUND_UNKNOWN</code> should be used.
     * 
     * @param bound A <code>k</code> for which this place is <code>k</code>-bounded.
     */
    public void setBound(int bound);
    
    /**
     * @return
     */
    public boolean boundKnown();
    
    
    /**
     * Returns this place's input arcs.
     * 
     * @return The input arcs of this place
     */
    public Collection<TPArc> getInput();
    
    
    /**
     * Returns this place's output arcs.
     * 
     * @return The output arcs of this place
     */
    public Collection<PTArc> getOutput();
    
    
    /**
     * Sets a note for this place.
     *
     * @param note The note for this place 
     */
    public void setNote(String note);
    
    
    /**
     * Returns the specified note for this place.
     *
     * @return The note for this place
     */
    public String getNote();
    
    
    /**
     * Returns <code>true</code> if the pi calculus process this place
     * stands for contains free names and could communicate. It could
     * communicate over free names and it has to be tracked, whether this place
     * is markable at the same time as other places that contain these free
     * names.
     *
     * @return <code>true</code> if the this place's pi calculus process could
     *         communicate via free names
     */
    public boolean couldCommunicate();
    
    
    /**
     * Sets whether this place's pi calculus process could communicate via free
     * names. If this is unknown, do not set it to <code>false</code>.
     *
     * @param couldCommunicate <code>true</code> whether this pi calculus could
     *            communicate via free names.
     */
    public void setCouldCommunicate(boolean couldCommunicate);
    
}