/*
 * Project:  Petruchio
 * Filename: PiParser.java
 * Created:  09.07.2007
 * Author:   Tim Strazny
 */
package petruchio.interfaces.pi;

import java.io.InputStream;

import petruchio.interfaces.ParserErrorHandler;
import petruchio.interfaces.Resettable;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessDefinition;


/**
 * A parser for pi calculus processes. 
 * 
 * @author Tim Strazny
 */
public interface PiParser extends Resettable {
    
    /**
     * Returns a comma separated list of the default file name extension for 
     * this pi format.
     *
     * @return The file name extensions for this pi format
     */
    public String getDefaultExtensions();
    
    
    /**
     * Sets the process creator that shall be used to build the syntax trees of
     * the processes during parsing.
     *
     * @param pbuilder The process creator to use during the parsing
     */
    public void setProcessCreator(ProcessCreator pbuilder);
    
    
    /**
     * Returns a string that can be added in front of a process to get a valid
     * process definition this class can parse.
     *
     * @return A string that can be added in front of a process to get a valid
     *         process definition this class can parse.
     */
    public String getDummyProcessDefinitionPrefix();
    
    
    /**
     * Returns a string that can be added after a process to get a valid
     * process definition this class can parse.
     *
     * @return A string that can be added after a process to get a valid
     *         process definition this class can parse.
     */
    public String getDummyProcessDefinitionSuffix();
    
    
    /**
     * Adds an error handler to be used to report errors.
     *
     * @param handler An error handler to use during the parsing
     */
    public void addErrorHandler(ParserErrorHandler handler);
    
    
    /**
     * Sets the input stream for the next parsing task.
     *
     * @param input The new input stream
     */
    public void setInput(InputStream input);
    
    
    /**
     * The parsing. Repetitive calls to this method deliver all the process
     * definitions the input set by {@link #setInput(InputStream)} provides.
     * When there are no more process definitions to ask for, the method has to
     * return <code>null</code>. If the <code>null</code> value has been
     * returned no further calls to this method are made until a new stream is
     * set as new input.  
     *
     * @param <V> The data type of the process(defined by
     *            {@link #setProcessCreator(ProcessCreator)})
     * @return The next process definition or <code>null</code> if there are no
     *            more definitions.
     */
    public <V> ProcessDefinition<V> nextProcessDefinition();
    
}