/*
 * Project:  Petruchio
 * Filename: Process.java
 * Created:  30.11.2006
 * Author:   Tim Strazny
 */
package petruchio.interfaces.pi;

import java.util.Collection;
import java.util.List;


/**
 * A pi calculus process.
 * An instance of a process of this type represents a pi calculus process
 * &quot;&lt;restrictions&gt;.&lt;action prefixes&gt;.&gt;null, reference, next
 * process or composition&gt;&quot; which type has to be more specified.
 * 
 * @param <E> The data type for this process.
 * @author Tim Strazny
 */
public interface Process<E> {
    
    /**
     * Process types.
     * An enum is used to avoid the expensive instanceof-operator.
     * 
     * @author Tim Strazny
     */
    public static enum Type {
        /** This process is a prefix to a null process. */
        NULL,
        
        /** This process is a prefix to another process. */
        PREFIX,
        
        /** This process is a prefix to a reference process. */
        REFERENCE,
        
        /** This process is a prefix to a composition of processes. */
        COMPOSITION
    }
    
    
    /**
     * Returns the type of this process.
     * 
     * @return The type of this process
     */
    public Type getType();
    
    
    /**
     * Adds an action prefix as the new first action prefix.
     *
     * @param ap The new first action prefix
     */
    public void addActionPrefix(ActionPrefix ap);
    
    
    /**
     * Adds a restriction to this process.
     *
     * @param name The newly restricted name
     */
    public void addRestriction(Name name);
    
    
    /**
     * Adds a list of action prefixes as the beginning of the action prefixes
     * for this process.
     *
     * @param prefixes The new first action prefixes
     */
    public void addActionPrefixes(List<ActionPrefix> prefixes);
    
    
    /**
     * Adds a set of restrictions to this process.
     *
     * @param names The restrictions to add
     */
    public void addRestrictions(Collection<Name> names);
      
    
    /**
     * Returns a String representation of this process.
     *
     * @return A String representation of this process
     */
    @Override
    public String toString();
    
    
    /**
     * Returns the list of action prefixes of this process.
     *
     * @return The list of action prefixes of this process
     */
    public List<ActionPrefix> getActionPrefixes();
    
    
    /**
     * Returns the set of restrictions of this process. 
     *
     * @return The set of restrictions of this process
     */
    public Collection<Name> getRestrictions();
    
    
    /**
     * Tells the process that it has been changed and the data may not be up to
     * date.
     */
    public void changed();
    
    
    /**
     * Sets the data object for this process.
     *
     * @param data The data for this process.
     */
    public void setData(E data);
    
    
    /**
     * Returns the data of this process.
     *
     * @return The data of this process
     */
    public E getData();

    
    /**
     * Checks whether this process has real action prefixes as input, output or
     * tau prefixes (non guard prefixes).
     *
     * @return <code>true</code> if this process has a prefix that is no guard.
     */
    public boolean hasRealPrefixes();
    
}