/*
 * Project:  Petruchio
 * Filename: ProcessComposition.java
 * Created:  30.11.2006
 * Author:   Tim Strazny
 */
package petruchio.interfaces.pi;

import java.util.List;

/**
 * A composition of arbitrarily many processes.
 * 
 * @param <E> The data type of the process.
 * @author Tim Strazny
 */
public interface ProcessComposition<E> extends Process<E> {
    
    /**
     * The compositional operators.
     * 
     * @author Tim Strazny
     */
    public enum Operator {
        /** The choice or alternative operator. */
        CHOICE(true, false),
        
        /** The parallel operator. */
        PARALLEL(true, true),
        
        /** The sequence operator. */
        SEQUENCE(false, true);

        private final boolean isCommutative;
        
        private final boolean allowsScopeExtrusion;
        
        
        /**
         * Initializes the operator.
         * 
         * @param isCommutative <code>true</code> if the composition operator is
         *            commutative.
         * @param allowsScopeExtrusion <code>true</code> if the composition
         *            operator allows scope extrusion.
         */
        private Operator(final boolean isCommutative,
                        final boolean allowsScopeExtrusion) {
            this.isCommutative = isCommutative;
            this.allowsScopeExtrusion = allowsScopeExtrusion;
        }
        
        
        /**
         * Returns whether this operator allows reordering of its operands.
         *
         * @return <code>true</code> if this operator is commutative
         */
        public boolean isCommutative() {
            return this.isCommutative;
        }
        
        
        /**
         * Returns whether this operator allows extraction of restrictions.
         *
         * @return <code>true</code> if this operator allows scope extrusion
         */
        public boolean allowsScopeExtrusion() {
            return this.allowsScopeExtrusion;
        }
    }
    
    
    /**
     * Adds the process at the end of this composition.
     *
     * @param p The process to append
     */
    public void addProcess(Process<E> p);
    
    
    /**
     * Adds the processes at the end of this composition.
     *
     * @param p The processes to append
     */
    public void addProcesses(List<Process<E>> p);
    
    
    /**
     * Sets the operator of this composition.
     *
     * @param op The new operator
     */
    public void setOperator(ProcessComposition.Operator op);
    
    
    /**
     * Returns the operator of this composition.
     *
     * @return This composition's operator
     */
    public ProcessComposition.Operator getOperator();
    
    
    /**
     * Returns the list of the operand processes.
     *
     * @return The list of the operand processes
     */
    public List<Process<E>> getProcesses();
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.Process#toString()
     */
    @Override
    public String toString();
    
}