/*
 * Project:  Petruchio
 * Filename: ProcessBuilder.java
 * Created:  01.12.2006
 * Author:   Tim Strazny
 */
package petruchio.interfaces.pi;

import petruchio.interfaces.Resettable;


/**
 * A class that implements this interface can be passed to a pi calculus process
 * parser to build the syntax trees of the parsed processes.
 * It creates instances of all types of processes and works as a wrapper class
 * to all the process constructors.
 * 
 * @author Tim Strazny
 */
public interface ProcessCreator extends Resettable {

	/**
     * Clears internal caches.
     */
    public void clearCaches();
	
	
    /**
     * Returns a fresh instance of a process definition.
     *
     * @param <T> The data type of the returned process. 
     * @param pid A process identifier
     * @param p A process
     * @param params Formal parameters
     * @param isMain Sets whether the process definition should be marked as the
     *           main/system process definition.
     * @return A fresh instance of a process definition that associates the
     *           given process identifier with the formal parameters and the
     *           process.
     */
    public <T> ProcessDefinition<T> newProcessDefinition(ProcessID pid,
                    Process<T> p, Parameters params, boolean isMain);
    
    
    /**
     * Returns a fresh instance of a process identifier.
     *
     * @param name A name
     * @return A fresh instance of a process identifier that represents the
     *            given name.
     */
    public ProcessID newProcessID(Name name);
    
    
    /**
     * Returns a fresh instance of a null process.
     *
     * @param <T> The data type of the returned process
     * @return A fresh instance of a null process
     */
    public <T> NullProcess<T> nullProcess(); 
    
    
    /**
     * Returns a fresh instance of a process reference.
     *
     * @param <T> The data type of the returned process
     * @param pid A process identifier
     * @param params A parameter list
     * @return A fresh instance of a process reference to the given process
     *            identifier with the given parameters.
     */
    public <T> ProcessReference<T> newProcessReference(ProcessID pid, Parameters params);
    
    
    /**
     * Returns a fresh instance of a process composition.
     *
     * @param <T> The data type of the returned process
     * @return A fresh instance of a process composition.
     */
    public <T> ProcessComposition<T> newProcessComposition();
    
    
    /**
     * Returns a fresh instance of a prefix process.
     *
     * @param <T> The data type of the returned process
     * @param p A process
     * @return A fresh instance of a prefix process that is prefix to the given
     *            process.
     */
    public <T> PrefixProcess<T> newPrefixProcess(Process<T> p);
    
    
    /**
     * Returns a fresh instance of an input action.
     *
     * @param channel A channel name
     * @param names A list of parameters to be received
     * @return A fresh instance of an input action that receives the given list
     *            of parameters via the given channel.
     */
    public InputAction newInputAction(Name channel, Parameters names);
    
    
    /**
     * Returns a fresh instance of an output action.
     *
     * @param channel A channel name
     * @param names A list of parameters to be sent
     * @return A fresh instance of an output action that sends the given list of
     *            parameters via the given channel.
     */
    public OutputAction newOutputAction(Name channel, Parameters names);
    
    
    /**
     * Returns an instance of an internal, unobservable action.
     *
     * @return An instance of an internal, unobservable action.
     */
    public InternalAction newInternalAction();
    
    
    /**
     * Returns a fresh and empty instance of a parameter list.
     *
     * @return A fresh and empty instance of a parameter list.
     */
    public Parameters newParameterList();
    
    
    /**
     * Returns a fresh instance of a guard prefix.
     *
     * @param left A name
     * @param right A name
     * @return A fresh instance of a guard prefix that represents the equality
     *            expression of the left and the right name.
     */
    public Guard newGuard(Name left, Name right);
    
    
    /**
     * TODO: Document
     *
     * @param name
     * @return
     */
    public Name newName(String name);
    
}