/*
 * Project:  Petruchio
 * Filename: ProcessManager.java
 * Created:  30.11.2006
 * Author:   Tim Strazny
 */
package petruchio.interfaces.pi;

import java.util.Collection;
import java.util.Set;

import petruchio.interfaces.Resettable;

/**
 * A manager for associations between process identifiers and processes and
 * formal parameters.
 * 
 * @param <E> The data type of the managed processes. 
 * @author Tim Strazny
 */
public interface ProcessManager<E>  extends Resettable {
    
	/**
     * Clears internal caches.
     */
    public void clearCaches();
	
	
    /**
     * Stores a backup of the associations to allow temporary changes to the
     * associations.
     */
    public void backup();
    
    
    /**
     * Loads the backup of the associations and overwrites all the temporary
     * changes made to the associations.
     */
    public void loadBackup();
    
    
    /**
     * Returns the set of process identifiers to which associations are stored
     * in this manager.
     *
     * @return The set of process identifiers this manager knows
     */
    public Set<ProcessID> getProcessIDs();
    
    
    /**
     * Checks whether there exists an association to the given process
     * identifier.
     *
     * @param pid The process identifier to look up
     * @return <code>true</code> if an association to the given identifier
     *            exists in this manager.
     */
    public boolean containsProcessID(ProcessID pid);
    
    
    /**
     * Returns the process identifier to a given process if there exists an
     * association from an identifier to that process.
     * If there is no such association, <code>null</code> is returned. 
     *
     * @param p The process to look up
     * @return A process identifier that is associated with the process.
     */
    public ProcessID getProcessID(Process<E> p);
    
    
    /**
     * Returns the process that is associated with the given process identifier.
     * If there is no such association, <code>null</code> is returned.
     *
     * @param pid The process identifier to look up
     * @return The process that is associated with the process identifier.
     */
    public Process<E> getProcess(ProcessID pid);
    
    
    /**
     * Returns the process definition that is associated with the given process
     * identifier.
     * If there is no such association, <code>null</code> is returned.
     *
     * @param pid The process identifier to look up
     * @return The process that is associated with the process identifier.
     */
    public ProcessDefinition<E> getProcessDefinition(ProcessID pid);
    
    
    /**
     * Returns the formal parameters that is associated with the given process
     * identifier.
     * If there is no such association, <code>null</code> is returned.
     *
     * @param pid The process identifier to look up
     * @return The formal parameters that are associated with the process
     *            identifier.
     */
    public Parameters getFormalParameters(ProcessID pid);
    
    
    /**
     * Adds an association from the given process identifier to the formal
     * parameters and the given process.
     * If there already exists an association from that process identifier, an
     * exception could be thrown.
     *
     * @param def A process definition
     */
    public void addProcessDefinition(ProcessDefinition<E> def);
    
    
    /**
     * Returns the set of process definitions managed by this class.
     *
     * @return The set of process definitions managed by this class.
     */
    public Collection<ProcessDefinition<E>> getProcessDefinitions();

    
    /**
     * Removes the association that has the given process identifier as its key.
     * The formerly associated process is returned.
     * If there was no association with the process identifier,
     * <code>null</code> is returned.
     *
     * @param pid A process identifier
     * @return The process that was associated with the process identifier.
     */
    public Process<E> removeProcess(ProcessID pid);
    
    
    /**
     * Replaces an association from a process identifier to a process with a new
     * association.
     *
     * @param pid A process identifier that is already in an association
     * @param p A process to be associated with the process identifier
     */
    public void updateProcess(ProcessID pid, Process<E> p);
    
    
    /**
     * Replaces an association from a process identifier to a process with a new
     * association.
     *
     * @param p A process definition
     */
    public void updateProcess(ProcessDefinition<E> p);
    
    
    /**
     * Returns the set of names that could be created by the process via
     * restriction. A name <code>n</code> could be created, if this or any
     * process referenced by this process contains a restriction on
     * <code>n</code>.
     * 
     * @param p A process
     * @return A set of names as specified above.
     */
    public Collection<Name> createsNames(Process<?> p);
    
    
    /**
     * Returns the set of process ids that could be referenced by the process.
     * A process <code>n</code> could be referenced, if this or any
     * process referenced by this process contains a reference to
     * <code>n</code>.
     * 
     * @param p A process
     * @return A set of references as specified above.
     */
    public Collection<ProcessID> references(Process<?> p);
    
}