/*
 * Project:  Petruchio
 * Filename: LTLScopeParser.java
 * Created:  25.09.2007
 * Author:   Tim Strazny
 */
package petruchio.ltl;

import java.io.BufferedReader;
import java.io.Reader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;


/**
 * Quickly runs over an LTL formula given by {@link #setInput(InputStream)} and
 * reads out all the places in atomic propositions <code>(place, marking)</code>
 * to return them via {@link #getPlacesToIgnore()}. 
 *
 * @author Tim Strazny
 */
public class LTLScopeIgnorer implements SelfDescribing, petruchio.interfaces.petrinet.PetriNetReductorIgnorer {
	
	private static final int EOF = -1;
	
	private InputStream in = null;
	
	private Collection<String> scope = null;
	
	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.SelfDescribing#getDescription()
	 */
	@Override
    public String getDescription() {
	    return "Parses a ltl formula for pairs like (p, 1) and delivers p to the reductors to be ignored.";
	}
	
	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductorIgnore#setInput(java.io.InputStream)
	 */
	@Override
    public void setInput(final InputStream in) {
		this.reset();
		this.in = in;
	}


	/* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		this.scope = null;
		if(this.in != null) {
			try {
				this.in.close();
			} catch(final IOException e) {
				// ignore
			}
			this.in = null;
		}
	}
	
	
	/**
	 * The input stream parsing for place names in atomic propositions.
	 */
	private void parse() {
		this.scope = new HashSet<String>();
		
		try {
			final Reader reader = new BufferedReader(new InputStreamReader(this.in));
			final StringBuilder name = new StringBuilder();
			
			boolean first = false;
			int ch = reader.read();
			while(ch != LTLScopeIgnorer.EOF) {
				if(ch == '(') {
					first = true;
					name.delete(0, name.length());
					
				} else if(ch == ')') {
					name.delete(0, name.length());
					
				} else if(ch == ',') {
					first = false;
					this.scope.add(name.toString());
					name.delete(0, name.length());
					
				} else if(first && !Character.isWhitespace(ch)) {
					name.append((char)ch);
					
				}
				
				ch = reader.read();
			}
			
			//System.out.println(this.scope);
			
		} catch(final IOException e) {
			throw new RuntimeException("Unable to parse the scope of the LTL formula.", e);
			
		} finally {
			if(this.in != null) {
				try {
					this.in.close();
				} catch(final IOException e) {
					// ignore
				}
			}
		}
	}
	

	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductorIgnore#getPlacesToIgnore(petruchio.interfaces.petrinet.PetriNet)
	 */
	@Override
    public Collection<Place> getPlacesToIgnore(final PetriNet net) {
		if(this.in == null) {
			return new ArrayList<Place>();
		}
		if(this.scope == null) {
			this.parse();
		}
		return net.getPlaces(this.scope);
	}


	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductorIgnore#getTransitionsToIgnore(petruchio.interfaces.petrinet.PetriNet)
	 */
	@Override
    public Collection<Transition> getTransitionsToIgnore(final PetriNet net) {
		return new ArrayList<Transition>();
		
		// maybe for later use.
//		if(this.in == null) {
//			return new ArrayList<Transition>();
//		}
//		if(this.scope == null) {
//			this.parse();
//		}
//		return net.getTransitions(this.scope);
	}

}