/*
 * Project:  Petruchio
 * Filename: ProcessManager.java
 * Created:  01.12.2006
 * Author:   Tim Strazny
 */
package petruchio.pi;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.pi.ProcessDefinition;
import petruchio.interfaces.pi.Parameters;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessID;
import petruchio.interfaces.pi.ProcessReference;


/**
 * <Description>
 * 
 * @param <E>
 * @author Tim Strazny
 */
public class ProcessManager<E> implements SelfDescribing, petruchio.interfaces.pi.ProcessManager<E> {

    private final Map<ProcessID, ProcessDefinition<E>> definitions;
    
    private final Map<ProcessID, Collection<Name>> createsNamesRef;
    
    private final Map<ProcessID, Collection<ProcessID>> referencesRef;
    
    private final Map<Process<E>, ProcessID> reverseDefinitions;
    
    private ProcessManager<E> backupManager;
    
    
    /**
     * 
     */
    public ProcessManager() {
        this.createsNamesRef = new WeakHashMap<ProcessID, Collection<Name>>(0);
        this.referencesRef = new WeakHashMap<ProcessID, Collection<ProcessID>>(0);
        this.definitions = new HashMap<ProcessID, ProcessDefinition<E>>(0);
        this.reverseDefinitions = new IdentityHashMap<Process<E>, ProcessID>(0);
        this.backupManager = null;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "The default process manager that holds pi calculus process definitions in memory.";
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#backup()
     */
    @Override
    public void backup() {
        if(this.backupManager == null) {
            this.backupManager = new ProcessManager<E>();
        }
        this.backupManager.clearCaches();
        this.backupManager.definitions.putAll(this.definitions);
        this.backupManager.createsNamesRef.putAll(this.createsNamesRef);
        this.backupManager.referencesRef.putAll(this.referencesRef);
        this.backupManager.reverseDefinitions.putAll(this.reverseDefinitions);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#loadBackup()
     */
    @Override
    public void loadBackup() {
        this.clearCaches();
        this.definitions.putAll(this.backupManager.definitions);
        this.createsNamesRef.putAll(this.backupManager.createsNamesRef);
        this.referencesRef.putAll(this.backupManager.referencesRef);
        this.reverseDefinitions.putAll(this.backupManager.reverseDefinitions);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#clearCaches()
     */
    @Override
    public void clearCaches() {
    	this.createsNamesRef.clear();
        this.definitions.clear();
        this.reverseDefinitions.clear();
        this.referencesRef.clear();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        this.clearCaches();
        this.backupManager = null;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#addProcess(petruchio.interfaces.pi.ProcessID, petruchio.interfaces.pi.Process)
     */
    @Override
    public void addProcessDefinition(final ProcessDefinition<E> def) {
        if(this.definitions.containsKey(def.getProcessID())) {
            throw new DuplicateProcessIDException(def.getProcessID());
        } else if(def.getProcess() == null) {
            throw new IllegalArgumentException("Will not add null process. (ID "
                            + def.getProcessID() + ")");
        } else {
            this.definitions.put(def.getProcessID(), def);
            this.reverseDefinitions.put(def.getProcess(), def.getProcessID());
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#getProcess(petruchio.interfaces.pi.ProcessID)
     */
    @Override
    public Process<E> getProcess(final ProcessID pid) {
        final ProcessDefinition<E> p = this.definitions.get(pid);
        if(p == null) {
            return null;
        }
        return p.getProcess();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#getFormalParameters(petruchio.interfaces.pi.ProcessID)
     */
    @Override
    public Parameters getFormalParameters(final ProcessID pid) {
        final ProcessDefinition<E> p = this.definitions.get(pid);
        if(p == null) {
            return null;
        }
        return p.getParameters();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#getProcessIDs()
     */
    @Override
    public Set<ProcessID> getProcessIDs() {
        return this.definitions.keySet();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#removeProcess(petruchio.interfaces.pi.ProcessID)
     */
    @Override
    public Process<E> removeProcess(final ProcessID pid) {
        final ProcessDefinition<E> p = this.definitions.remove(pid);
        if(p == null) {
            return null;
        }
        return p.getProcess();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#updateProcess(petruchio.interfaces.pi.ProcessID, petruchio.interfaces.pi.Process)
     */
    @Override
    public void updateProcess(final ProcessID pid, final Process<E> p) {
        if(p == null) {
            throw new IllegalArgumentException("Will not add null process. (ID "
                            + pid + ")");
        }
        ProcessDefinition<E> d = this.definitions.get(pid);
        if(d == null) {
            d = new petruchio.pi.ProcessDefinition<E>(pid, p, new petruchio.pi.Parameters(), false);
        } else {
            d = new petruchio.pi.ProcessDefinition<E>(pid, p, d.getParameters(), d.isMain());
        }
        this.definitions.put(pid, d);
        this.reverseDefinitions.put(p, pid);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#updateProcess(petruchio.interfaces.pi.ProcessDefinition)
     */
    @Override
    public void updateProcess(final ProcessDefinition<E> p) {
        if(p == null) {
            throw new IllegalArgumentException("Will not add null process definition.");
        } else if(p.getProcessID() == null) {
            throw new IllegalArgumentException("Will not add null process id.");
        } else if(p.getProcess() == null) {
            throw new IllegalArgumentException("Will not add null process.");
        } else {
            this.definitions.put(p.getProcessID(), p);
            this.reverseDefinitions.put(p.getProcess(), p.getProcessID());
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#getProcessID(petruchio.interfaces.pi.Process)
     */
    @Override
    public ProcessID getProcessID(final Process<E> p) {
//        for(final Map.Entry<ProcessID, Process> e : this.definitions.entrySet()) {
//            final Process proc = e.getValue();
//            if(proc == p) {
//                final ProcessID pid = e.getKey();
//                return pid;
//            }
//        }
//        return null;
        
        return this.reverseDefinitions.get(p);
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#createsNames(petruchio.interfaces.pi.Process)
     */
    @Override
    public Collection<Name> createsNames(final Process<?> p) {
        final Collection<Name> names = new ArrayList<Name>(0);
        this.createsNames(p, names, null);
        return names;
    }
    
    
    /**
     * @param p
     * @param exclude
     * @return
     */
    private void createsNames(final Process<?> p, final Collection<Name> names, final Collection<ProcessID> toExclude) {
        if(!p.getRestrictions().isEmpty()) {
            names.addAll(p.getRestrictions());
        }
        Collection<ProcessID> exclude = toExclude;
        switch(p.getType()) {
            case REFERENCE:
                final ProcessReference<?> ref = (ProcessReference<?>)p;
                final ProcessDefinition<?> refd = this.definitions.get(ref.getProcessID());
                if(refd != null) {
                    Collection<Name> b = this.createsNamesRef.get(ref.getProcessID());
                    if(b == null) {
                        if(exclude == null) {
                            exclude = new HashSet<ProcessID>();
                            exclude.add(ref.getProcessID());
                        } else if(!exclude.add(ref.getProcessID())) {
                            return;
                        }
                        b = new ArrayList<Name>(0);
                        this.createsNames(refd.getProcess(), b, exclude);
                        this.createsNamesRef.put(ref.getProcessID(), b);
                    }
                    names.addAll(b);
                }
                break;
            case PREFIX:
                this.createsNames(((PrefixProcess<?>)p).getProcess(), names, exclude);
                break;
            case COMPOSITION:
                for(final Process<?> s : ((ProcessComposition<?>)p).getProcesses()) {
                    this.createsNames(s, names, exclude);
                }
                break;
            case NULL:
                break;
            default:
                throw new InternalError("Unknown process type: " + p.getType());
        }
    }
    
    
    @Override
    public Collection<ProcessID> references(final Process<?> p) {
        final Collection<ProcessID> pids = new ArrayList<ProcessID>();
        this.references(p, pids, null);
        return pids;
    }
    
    
    /**
     * @param p
     * @param exclude
     * @return
     */
    private void references(final Process<?> p, final Collection<ProcessID> pids, final Collection<ProcessID> toExclude) {
        Collection<ProcessID> exclude = toExclude;
        switch(p.getType()) {
            case REFERENCE:
                final ProcessReference<?> ref = (ProcessReference<?>)p;
                final ProcessDefinition<?> refd = this.definitions.get(ref.getProcessID());
                pids.add(ref.getProcessID());
                if(refd != null) {
                    Collection<ProcessID> b = this.referencesRef.get(ref.getProcessID());
                    if(b == null) {
                        if(exclude == null) {
                            exclude = new HashSet<ProcessID>();
                            exclude.add(ref.getProcessID());
                        } else if(!exclude.add(ref.getProcessID())) {
                            return;
                        }
                        b = new ArrayList<ProcessID>();
                        this.references(refd.getProcess(), b, exclude);
                        this.referencesRef.put(ref.getProcessID(), b);
                    }
                    pids.addAll(b);
                }
                break;
            case PREFIX:
                this.references(((PrefixProcess<?>)p).getProcess(), pids, exclude);
                break;
            case COMPOSITION:
                for(final Process<?> s : ((ProcessComposition<?>)p).getProcesses()) {
                    this.references(s, pids, exclude);
                }
                break;
            case NULL:
                break;
            default:
                throw new InternalError("Unknown process type: " + p.getType());
        }
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#containsProcessID(petruchio.interfaces.pi.ProcessID)
     */
    @Override
    public boolean containsProcessID(final ProcessID pid) {
        return this.definitions.containsKey(pid);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#getProcessDefinitions()
     */
    @Override
    public Collection<ProcessDefinition<E>> getProcessDefinitions() {
        return this.definitions.values();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessManager#getProcessDefinition(petruchio.interfaces.pi.ProcessID)
     */
    @Override
    public ProcessDefinition<E> getProcessDefinition(final ProcessID pid) {
        return this.definitions.get(pid);
    }
    
}