/*
 * Project: Petruchio
 * File:    ProcessMetrics.java
 * Author:  Tim Strazny
 * Created: 03.01.2008
 */
package petruchio.pi.modifiers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import petruchio.common.WeakIdentityHashMap;
import petruchio.compiler.CommonTasks;
import petruchio.compiler.ProcessData;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.InputAction;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.OutputAction;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessDefinition;
import petruchio.interfaces.pi.ProcessID;
import petruchio.interfaces.pi.ProcessManager;
import petruchio.interfaces.pi.ProcessModifier;
import petruchio.interfaces.pi.ProcessReference;


/**
 * TODO: Document
 * 
 * @author Tim Strazny
 */
public class ProcessMetrics implements ProcessModifier<ProcessData>, SelfDescribing {

    private final Map<Process<?>, Integer> sizes = new WeakIdentityHashMap<Process<?>, Integer>(0);
    private final Map<ProcessID, List<Name>> defs = new HashMap<ProcessID, List<Name>>(0);
    private final Map<Name, List<Name>> deadReferences = new HashMap<Name, List<Name>>(0);
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessModifier#getName()
     */
    @Override
    public String getName() {
        return "process metrics";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        this.sizes.clear();
        this.defs.clear();
        this.deadReferences.clear();
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessModifier#modify(petruchio.interfaces.pi.ProcessManager, petruchio.interfaces.pi.ProcessID)
     */
    @Override
    public String modify(ProcessManager<ProcessData> pm, ProcessID main) {
        this.reset();
        final Map<Name, Collection<Name>> free = new HashMap<Name, Collection<Name>>();
        boolean done;
        do {
            done = true;
            for(final ProcessDefinition<ProcessData> pd : pm.getProcessDefinitions()) {
                if(this.collectDef(pd, free)) {
                    done = false;
                }
            }
        } while(!done);
        for(final ProcessDefinition<ProcessData> pd : pm.getProcessDefinitions()) {
            this.propagateDef(pd, free);
        }
        
        int size = 0;
        final Collection<ProcessDefinition<ProcessData>> checked = new HashSet<ProcessDefinition<ProcessData>>();
        final LinkedList<ProcessDefinition<ProcessData>> todo = new LinkedList<ProcessDefinition<ProcessData>>();
        todo.add(pm.getProcessDefinition(main));
        while(!todo.isEmpty()) {
            final ProcessDefinition<ProcessData> pd = todo.remove();
            if(!checked.add(pd)) {
                continue;
            }
            size += this.measure(pd, pm, todo);
        }
        for(final Map.Entry<Name, List<Name>> e : this.deadReferences.entrySet()) {
            size += 2 + e.getValue().size();
        }
        this.reset();
        return "Process size: " + size;
    }
    
    
    private boolean collectDef(final ProcessDefinition<ProcessData> pd, final Map<Name, Collection<Name>> free) {
        Collection<Name> fn = free.get(pd.getProcessID().getID());
        boolean result = false;
        if(fn == null) {
            fn = new HashSet<Name>(CommonTasks.freeNames(pd.getProcess()));
            fn.removeAll(pd.getParameters().getParameters());
            free.put(pd.getProcessID().getID(), fn);
            result = true;
        }
        for(final Name id : CommonTasks.getData(pd.getProcess()).getRefProcIDs()) {
            Collection<Name> idfn = free.get(id);
            if(idfn != null && fn.addAll(idfn)) {
                result = true;
            }
        }
        return result;
    }
    
    
    private boolean propagateDef(final ProcessDefinition<ProcessData> pd, final Map<Name, Collection<Name>> free) {
        List<Name> params = pd.getParameters().getParameters();
        boolean result = false;
        if(!free.isEmpty()) {
            params = new ArrayList<Name>(free.get(pd.getProcessID().getID()));
            params.removeAll(pd.getParameters().getParameters());
            params.removeAll(CommonTasks.boundNames(pd.getProcess()));
            Collections.sort(params);
            params.addAll(0, pd.getParameters().getParameters());
            result = true;
        }
        this.defs.put(pd.getProcessID(), params);
        return result;
    }
    
    
    /**
     * @param p
     * @param pm
     * @param todo
     * @return
     */
    public int measure(final ProcessDefinition<?> p, final ProcessManager<ProcessData> pm,
            final Collection<ProcessDefinition<ProcessData>> todo) {
        final Integer size = this.sizes.get(p.getProcess());
        int s;
        if(size == null) {
            s = this.calc(p.getProcess(), pm, todo);
        } else {
            s = size.intValue();
        }
        return s + 1 + this.defs.get(p.getProcessID()).size();
    }
    
    
    /**
     * @param p
     * @param pm
     * @param todo
     * @return
     */
    public int measure(final Process<?> p, final ProcessManager<ProcessData> pm,
            final Collection<ProcessDefinition<ProcessData>> todo) {
        final Integer size = this.sizes.get(p);
        if(size == null) {
            return this.calc(p, pm, todo);
        }
        return size.intValue();
    }
    
    
    /**
     * TODO: Document
     *
     * @param p
     * @return
     */
    private int calc(final Process<?> p, final ProcessManager<ProcessData> pm,
            final Collection<ProcessDefinition<ProcessData>> todo) {
        int size = p.getRestrictions().size();
        for(final ActionPrefix a : p.getActionPrefixes()) {
            switch(a.getType()) {
                case INTERNAL: {
                    size += 1;
                    break;
                }
                case GUARD: {
                    size += 2;
                    break;
                }
                case INPUT: {
                    final InputAction in = (InputAction)a;
                    size += 1 + in.getNames().size();
                    break;
                }
                case OUTPUT: {
                    final OutputAction out = (OutputAction)a;
                    size += 1 + out.getNames().size();
                    break;
                }
                default:
                    throw new InternalError("Unknown type of action prefix: " + a);
            }
        }
        switch(p.getType()) {
            case NULL: {
                size += 1;
                break;
            }
            case REFERENCE: {
                final ProcessReference<?> c = (ProcessReference<?>)p;
                size += 1 + c.getParameters().size();
                final ProcessDefinition<ProcessData> def = pm.getProcessDefinition(c.getProcessID());
                if(def == null) {
                    this.deadReferences.put(c.getProcessID().getID(), c.getParameters().getParameters());
                } else if(todo != null) {
                    todo.add(def);
                }
                break;
            }
            case PREFIX: {
                final PrefixProcess<?> c = (PrefixProcess<?>)p;
                size += this.measure(c.getProcess(), pm, todo);
                break;
            }
            case COMPOSITION: {
                final ProcessComposition<?> c = (ProcessComposition<?>)p;
                size += 2*c.getProcesses().size() - 1;
                for(final Process<?> s : c.getProcesses()) {
                    size += this.measure(s, pm, todo);
                }
                break;
            }
            default:
                throw new InternalError("Unknown type of process: " + p);
        }
        this.sizes.put(p, Integer.valueOf(size));
        return size;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "Does not modify a process but outputs its number of actions";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessModifier#setProcessCreator(petruchio.interfaces.pi.ProcessCreator)
     */
    @Override
    public void setProcessCreator(ProcessCreator pbuilder) {
        // nothing to do
    }

}