/*
 * Project: Petruchio
 * File:    SafeProcess.java
 * Author:  Tim Strazny
 * Created: 07.12.2007
 */
package petruchio.pi.modifiers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import petruchio.compiler.CommonTasks;
import petruchio.compiler.ProcessData;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessDefinition;
import petruchio.interfaces.pi.ProcessID;
import petruchio.interfaces.pi.ProcessManager;
import petruchio.interfaces.pi.ProcessReference;


/**
 * TODO: Document
 * 
 * @author Tim Strazny
 */
public class SafeProcess implements SelfDescribing, petruchio.interfaces.pi.ProcessModifier<ProcessData> {

    private static final String ZERO_ID = "ZERO";
    
    private static final Pattern STRIP_NAME = Pattern.compile("^(.*?)(_(\\d+))?$");
    
    private static final int STRIPPED_NAME_GROUP = 1;
    
    private Name zeroID = null;
    
    private ProcessCreator pbuilder = null;
    
    private Collection<Name> possiblyAutoConcurrent = new HashSet<Name>();
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        this.zeroID = null;
        this.pbuilder = null;
        this.possiblyAutoConcurrent.clear();
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessModifier#modify(petruchio.interfaces.pi.ProcessManager, petruchio.interfaces.pi.ProcessID)
     */
    @Override
    public String modify(final ProcessManager<ProcessData> pm, final ProcessID main) {
        this.preprocess(pm, main);
        
        Process<ProcessData> p = pm.getProcess(main);
        final Collection<Name> allIDs = new HashSet<Name>(pm.getProcessIDs().size());
        for(final ProcessID id : pm.getProcessIDs()) {
            allIDs.add(id.getID());
        }
        allIDs.addAll(CommonTasks.getData(p).getRefProcIDs());
        final Collection<Name> ids = new HashSet<Name>(allIDs.size());
        final Map<Name, Name> renids = new HashMap<Name, Name>(allIDs.size());
        renids.put(main.getID(), main.getID());
        final Collection<ProcessID> observed = new HashSet<ProcessID>();
        observed.add(main);
        p = this.makeSafe(pm, allIDs, p, ids, renids);
        pm.updateProcess(main, p);
        return null;
    }
    
    
    private void preprocess(final ProcessManager<ProcessData> pm, final ProcessID main) {
        final Map<Name, Collection<Name>> orbits = new HashMap<Name, Collection<Name>>(pm.getProcessDefinitions().size());
        final Collection<Name> criticalOrbit = new ArrayList<Name>(this.getOrbit(main, orbits, pm));
        final Collection<ProcessID> observed = new HashSet<ProcessID>();
        for(final Iterator<Name> it = criticalOrbit.iterator(); it.hasNext();) {
            final Process<ProcessData> cur = pm.getProcess(this.pbuilder.newProcessID(it.next()));
            observed.clear();
            if(this.isSequential(pm, cur)) {
                it.remove();
            }
        }
        this.possiblyAutoConcurrent.clear();
        for(final Name crit : criticalOrbit) {
            final Collection<Name> orbit = this.getOrbit(this.pbuilder.newProcessID(crit), orbits, pm);
            if(orbit.contains(crit)) {
                this.possiblyAutoConcurrent.add(crit);
            }
        }
        
        if(!this.possiblyAutoConcurrent.isEmpty()) {
            System.out.println("The following processes seem to be auto-concurrent.");
            System.out.println(this.possiblyAutoConcurrent);
            System.out.println("The processes (and the compiled net) might not be safe.");
        }
    }
    
    
    private Collection<Name> getOrbit(final ProcessID id, final Map<Name, Collection<Name>> orbits,
            final ProcessManager<ProcessData> pm) {
        Collection<Name> orbit = orbits.get(id.getID());
        if(orbit == null) {
            orbit = new HashSet<Name>(CommonTasks.getData(pm.getProcess(id)).getRefProcIDs());
            final LinkedList<Name> todo = new LinkedList<Name>(orbit);
            while(!todo.isEmpty()) {
                final Name n = todo.remove();
                if(n == id.getID()) {
                    continue;
                }
                final Collection<Name> orb = orbits.get(n);
                if(orb != null) {
                    orbit.addAll(orb);
                } else {
                    final Process<ProcessData> proc = pm.getProcess(this.pbuilder.newProcessID(n));
                    if(proc == null) {
                        continue;
                    }
                    final Collection<Name> ident = CommonTasks.getData(proc).getRefProcIDs();
                    for(final Name name : ident) {
                        if(orbit.add(name)) {
                            todo.add(name);
                        }
                    }
                    if(ident.contains(id.getID())) {
                        orbits.put(n, orbit);
                    }
                }
            }
            orbits.put(id.getID(), orbit);
        }
//        System.out.println(id + " : " + orbit);
        return orbit;
    }
    
    
    /**
     * TODO: Document
     *
     * @param pm
     * @param id
     * @param ids
     * @param allIDs
     */
    private Process<ProcessData> makeSafe(final ProcessManager<ProcessData> pm,
            final Collection<Name> allIDs,
            final Process<ProcessData> p, final Collection<Name> ids,
            final Map<Name, Name> renids) {
//        System.out.println(p + ": " + p.getType());
        switch(p.getType()) {
            case REFERENCE: {
                final ProcessReference<ProcessData> ref = (ProcessReference<ProcessData>)p;
                final Name curID = ref.getProcessID().getID();
//                System.out.println(curID + "\n" + ids + "\n" + renids);
                if(renids.containsKey(curID)) {
                    final Name nid = renids.get(curID);
                    if(!curID.equals(nid)) {
                        ProcessID id = this.pbuilder.newProcessID(nid);
                        final ProcessReference<ProcessData> newP = this.pbuilder.newProcessReference(id, ref.getParameters());
                        newP.addActionPrefixes(p.getActionPrefixes());
                        newP.addRestrictions(p.getRestrictions());
//                        System.out.println("converted " + curID + " to " + nid);
                        return newP;
                    }
                    return p;
                    
                } else if(ids.contains(curID) && !this.possiblyAutoConcurrent.contains(curID)) {
                    final Name newID = this.newName(curID, allIDs);
                    allIDs.add(newID);
                    ids.add(newID);
                    renids.put(curID, newID);
                    final ProcessReference<ProcessData> newP = this.pbuilder.newProcessReference(this.pbuilder.newProcessID(newID), ref.getParameters());
                    newP.addActionPrefixes(p.getActionPrefixes());
                    newP.addRestrictions(p.getRestrictions());
                    
//                    System.out.println("Renamed: " + curID + " to " + newID);
                    ProcessDefinition<ProcessData> def = pm.getProcessDefinition(ref.getProcessID());
                    if(def != null) {
                        final Process<ProcessData> proc = this.makeSafe(pm,
                                allIDs, def.getProcess(), ids, renids);
                        final ProcessDefinition<ProcessData> newDef =
                            this.pbuilder.newProcessDefinition(this.pbuilder.newProcessID(newID), proc, def.getParameters(), false);
                        pm.addProcessDefinition(newDef);
                    }
                    return newP;
                } else {
//                    System.out.println("Not Renamed: " + curID);
                    renids.put(curID, curID);
                    if(ids.add(curID)) {
                        final ProcessDefinition<ProcessData> def = pm.getProcessDefinition(ref.getProcessID());
                        if(def != null) {
                            final Process<ProcessData> proc = this.makeSafe(pm,
                                    allIDs, def.getProcess(), ids, renids);
                            pm.updateProcess(ref.getProcessID(), proc);
                        }
                    }
                    
                    return p;
                }
            }
            
            case COMPOSITION: {
                final ProcessComposition<ProcessData> comp = (ProcessComposition<ProcessData>)p;
                final boolean isPar = comp.getOperator() == ProcessComposition.Operator.PARALLEL;
                final List<Process<ProcessData>> l = new ArrayList<Process<ProcessData>>(comp.getProcesses().size());
                boolean changed = false;
                for(final Process<ProcessData> sub : comp.getProcesses()) {
                    final Process<ProcessData> subReplaced = this.makeSafe(pm,
                            allIDs, sub, ids, renids);
                    if(subReplaced != sub) {
                        changed = true;
                    }
                    l.add(subReplaced);
                    
                    if(isPar) {
                        renids.clear();
                    }
                }
                if(changed) {
                    final ProcessComposition<ProcessData> c = this.pbuilder.newProcessComposition();
                    c.setOperator(comp.getOperator());
                    c.addActionPrefixes(comp.getActionPrefixes());
                    c.addRestrictions(comp.getRestrictions());
                    c.addProcesses(l);
                    c.changed();
                    return c;
                }
                return p;
            }
                
            case PREFIX: {
                final PrefixProcess<ProcessData> prefix = (PrefixProcess<ProcessData>)p;
                final Process<ProcessData> r = this.makeSafe(pm, allIDs,
                        prefix.getProcess(), ids, renids);
                if(r != prefix.getProcess()) {
                    final PrefixProcess<ProcessData> rep = this.pbuilder.newPrefixProcess(r);
                    rep.addActionPrefixes(prefix.getActionPrefixes());
                    rep.addRestrictions(prefix.getRestrictions());
                    rep.changed();
                    return rep;
                }
                return p;
            }
            
            case NULL: {
                final Name curID = this.zeroID;
                if(renids.containsKey(curID)) {
                    final Name nid = renids.get(curID);
                    ProcessID id = this.pbuilder.newProcessID(nid);
                    final ProcessReference<ProcessData> newP = this.pbuilder.newProcessReference(id, this.pbuilder.newParameterList());
                    newP.addActionPrefixes(p.getActionPrefixes());
                    newP.addRestrictions(p.getRestrictions());
//                    System.out.println("changed: 0 to " + nid);
                    return newP;
                    
                }
                final Name newID = this.newName(curID, allIDs);
                allIDs.add(newID);
                ids.add(newID);
                renids.put(curID, newID);
                final ProcessReference<ProcessData> newP = this.pbuilder.newProcessReference(this.pbuilder.newProcessID(newID), this.pbuilder.newParameterList());
                newP.addActionPrefixes(p.getActionPrefixes());
                newP.addRestrictions(p.getRestrictions());
                
//                    System.out.println("Renamed: 0 to " + newID);
                
                final Process<ProcessData> proc = this.pbuilder.nullProcess();
                final ProcessDefinition<ProcessData> newDef =
                    this.pbuilder.newProcessDefinition(this.pbuilder.newProcessID(newID),
                            proc,
                            this.pbuilder.newParameterList(), false);
                pm.addProcessDefinition(newDef);
                return newP;
            }
            
            default:
                throw new InternalError("Unexpected type of process: " + p);
        }
    }
    
    
    private boolean isSequential(final ProcessManager<?> pm, final Process<?> p) {
        if(p == null) {
            return true;
        }
        switch(p.getType()) {
            case REFERENCE: {
                return true;
            }
            
            case COMPOSITION: {
                final ProcessComposition<?> comp = (ProcessComposition<?>)p;
                if(comp.getOperator() == ProcessComposition.Operator.PARALLEL) {
                    return false;
                }
                for(final Process<?> sub : comp.getProcesses()) {
                    if(!this.isSequential(pm, sub)) {
                        return false;
                    }
                }
                return true;
            }
                
            case PREFIX: {
                final PrefixProcess<?> prefix = (PrefixProcess<?>)p;
                return this.isSequential(pm, prefix.getProcess());
            }
            
            case NULL: {
                return true;
            }
            
            default:
                throw new InternalError("Unexpected type of process: " + p);
        }
    }
    
    
    private Name newName(final Name name, final Collection<Name> n1) {
        final Matcher matcher = SafeProcess.STRIP_NAME.matcher(name
                .toString());
        if (!matcher.find()) {
            throw new InternalError("This was unexpected. The regular "
                    + "expression to strip the suffix seems to be broken!");
        }
        final String strippedName = matcher
                .group(SafeProcess.STRIPPED_NAME_GROUP);
        // final String index =
        // matcher.group(ReactionFinder.STRIPPED_INDEX_GROUP);
        Name n = this.pbuilder.newName(strippedName);
        // int i = (index != null) ? Integer.parseInt(index) : 0;
        int i = 0;
        while(n1.contains(n)) {
            n = this.pbuilder.newName(strippedName + "_" + (++i));
        }
        return n;
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "Default implementation of an algorithm that translates a given "
            + "process into a safe process.";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.SafeProcess#setProcessCreator(petruchio.interfaces.pi.ProcessCreator)
     */
    @Override
    public void setProcessCreator(final ProcessCreator pbuilder) {
        this.pbuilder = pbuilder;
        this.zeroID = this.pbuilder.newName(SafeProcess.ZERO_ID);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.ProcessModifier#getName()
     */
    @Override
    public String getName() {
        return "generate safe processes";
    }

}