/*
 * Project:  Petruchio
 * Filename: LocalConfigParser.java
 * Created:  21.09.2007
 * Author:   Tim Strazny
 */
package petruchio.pi.readers;

import java.io.BufferedReader;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Properties;

import petruchio.interfaces.LocalConfigurationParser;
import petruchio.interfaces.ParserErrorHandler;
import petruchio.interfaces.SelfDescribing;

/**
 * TODO: description 
 *
 * @author Tim Strazny
 */
public class LocalConfigParser implements SelfDescribing, LocalConfigurationParser {
	
	private static final char[] KEY = "options".toCharArray();
	
	private static final int EOF = -1;
	
	private final Collection<ParserErrorHandler> errorHandlers;
	private Reader in;
	private int line;
	private int column;
	
	private int buffer;
	
	
	/**
	 * TODO: description
	 *
	 */
	public LocalConfigParser() {
		this.line = 1;
		this.column = 0;
		this.buffer = -1;
		this.in = null;
		this.errorHandlers = new ArrayList<ParserErrorHandler>();
	}
	
	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.SelfDescribing#getDescription()
	 */
	@Override
    public String getDescription() {
	    return "Parses a local configuration from a pi definition file if the "
	    + "options keyword occurrs before any other input except for comments."
	    + " Java style comments are permitted both outside and inside the options.";
	}
	
	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		this.line = 1;
		this.column = 0;
		this.buffer = -1;
		this.in = null;
		this.errorHandlers.clear();
	}
	

	/* (non-Javadoc)
	 * @see petruchio.interfaces.pi.LocalConfigurationParser#addErrorHandler(petruchio.interfaces.pi.PiParserErrorHandler)
	 */
	@Override
    public void addErrorHandler(final ParserErrorHandler handler) {
		this.errorHandlers.add(handler);
	}

	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.pi.LocalConfigurationParser#setInput(java.io.InputStream)
	 */
	@Override
    public void setInput(final InputStream input) {
		this.line = 1;
		this.column = 0;
		this.buffer = -1;
		this.in = new BufferedReader(new InputStreamReader(input));
	}


	/* (non-Javadoc)
	 * @see petruchio.interfaces.pi.LocalConfigurationParser#getLocalConfiguration()
	 */
	@Override
    public Properties getLocalConfiguration() {
		final Properties props = new Properties();
		
		int ch = -1;
		try {
			ch = this.skip();
			
			// match 'options'
			boolean match = true;
			int i = 0;
			while(i < LocalConfigParser.KEY.length) {
				if(ch != LocalConfigParser.KEY[i++]) {
					match = false;
					break;
				}
				ch = this.read();
			}
			if(!match) {
				return props;
			}
			
			ch = this.skip();
			
			if(ch != '{') {
				return props;
			}
			
			ch = this.skip();
			
			while(ch != '}') {
				
				// now read the properties!
				final StringBuilder key = new StringBuilder();
				while(ch != '=') {
					key.append((char)ch);
					if(ch == '"') {
						this.readString(key);
					}
					ch = this.skip();
				}
				
				ch = this.skip();
				
				final StringBuilder value = new StringBuilder();
				while(ch != ';') {
					value.append((char)ch);
					if(ch == '"') {
						this.readString(value);
					}
					ch = this.skip();
				}
				
				props.setProperty(key.toString().replaceAll("\"", ""),
						value.toString().replaceAll("\"", ""));
				
				ch = this.skip();
			}

			
		} catch(final IOException e) {
			this.reportError(this.line, this.column, "" + ch, e.getLocalizedMessage());
		}
		
		return props;
	}
	
	
	/**
	 * TODO: description
	 *
	 * @param n
	 */
	private void readString(final Appendable n) throws IOException {
		int ch;
		do {
			ch = this.read();
			n.append((char)ch);
		} while(ch != '"');
	}
	
	
	/**
	 * TODO: description
	 *
	 * @return
	 * @throws IOException
	 */
	private int read() throws IOException {
		++this.column;
		int ch;
		if(this.buffer != -1) {
			ch = this.buffer;
			this.buffer = -1;
		} else {
			ch = this.in.read();
			if(ch == '\n') {
				this.nextLine();
			}
		}
		if(ch == LocalConfigParser.EOF) {
			throw new EOFException("Input stream ended prematurely");
		}
		return ch;
	}
	
	
	/**
	 * TODO: description
	 *
	 * @return
	 * @throws IOException
	 */
	private void unread(final int ch) throws IOException {
		--this.column;
		this.buffer = ch; 
	}
	
	
	/**
	 * TODO: description
	 *
	 * @return
	 * @throws IOException
	 */
	private int skip() throws IOException {
		int ch;
		boolean comment;
		do {
			ch = this.skipWhitespaces();
			boolean jstyle = ch == '/';
			comment = jstyle || ch == '('; // support for mwb comments. 
			if(comment) {
				// decide kind of comment
				ch = this.read();
				if(jstyle && ch == '/') {
					// skip line comment
					do {
						ch = this.read();
					} while(ch != '\n');
				} else if(ch == '*') {
					// skip block comment
					int last;
					do {
						last = ch;
						ch = this.read();
					} while((last != '*') || (jstyle && ch != '/') || (!jstyle && ch != ')'));
				} else {
					this.unread(ch);
					ch = jstyle ? '/' : '(';
					comment = false;
				}
			}
		} while(comment);
		
		return ch;
	}
	
	
	/**
	 * TODO: description
	 *
	 * @return
	 * @throws IOException
	 */
	private int skipWhitespaces() throws IOException {
		int ch;
		do {
			ch = this.read();
		} while(Character.isWhitespace(ch));
		return ch;
	}
	
	
	/**
	 * TODO: description
	 *
	 */
	private void nextLine() {
		this.column = 1;
		++this.line;
	}
	
	
	/**
	 * TODO: description
	 *
	 * @param ln
	 * @param col
	 * @param token
	 * @param msg
	 */
	private void reportError(final int ln, final int col,
			final String token, final String msg) {
		for(final ParserErrorHandler h : this.errorHandlers) {
			h.reportError(ln, col, token, msg);
		}
	}
	
}