/*
 * Project:  Petruchio
 * Filename: CUPParser.java
 * Created:  18.08.2007
 * Author:   Tim Strazny
 */
package petruchio.pi.readers.cup;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Properties;

import petruchio.interfaces.LocalConfigurationParser;
import petruchio.interfaces.ParserErrorHandler;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.pi.PiParser;
import petruchio.interfaces.pi.ProcessCreator;
import petruchio.interfaces.pi.ProcessDefinition;
import petruchio.pi.readers.cup.CUPLexer;


/**
 * An abstract parser to connect a cup generated parser with the petruchio tool.
 * 
 * @author Tim Strazny
 */
public abstract class CUPParser implements SelfDescribing, PiParser, LocalConfigurationParser {

    private Collection<ProcessDefinition<?>> definitions;
    private Iterator<ProcessDefinition<?>> procIt;
    private ProcessCreator pbuilder;
    private CUPLexer lexer;
    private Collection<ParserErrorHandler> errorHandlers;
    private Properties properties;
    
    /**
     * Initializes the new parser instance.
     */
    public CUPParser() {
    	this.properties = null;
        this.definitions = null;
        this.procIt = null;
        this.pbuilder = null;
        this.lexer = null;
        this.errorHandlers = new ArrayList<ParserErrorHandler>();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        this.definitions = null;
        this.procIt = null;
        this.lexer = null;
        this.properties = null;
        this.pbuilder = null;
        this.errorHandlers.clear();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiParser#addErrorHandler(petruchio.interfaces.pi.PiParserErrorHandler)
     */
    @Override
    public void addErrorHandler(final ParserErrorHandler handler) {
        this.errorHandlers.add(handler);
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.LocalConfigurationParser#getLocalConfiguration()
     */
    @Override
    public Properties getLocalConfiguration() {
    	this.ensureAvailable();
    	return this.properties;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiParser#nextProcessDefinition()
     */
    @Override
    @SuppressWarnings("unchecked")
    public <V> ProcessDefinition<V> nextProcessDefinition() {
        this.ensureAvailable();
        if(this.procIt.hasNext()) {
            return (ProcessDefinition<V>)this.procIt.next();
        }
        this.definitions = null;
        this.procIt = null;
        return null;
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiParser#setInput(java.io.InputStream)
     */
    @Override
    public void setInput(final InputStream input) {
        final Reader reader = new InputStreamReader(input);
        
        this.definitions = null;
        this.procIt = null;
        
        if(this.lexer == null) {
            this.lexer = this.getLexer(reader);
        } else {
            this.lexer.yyreset(reader);
        }
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiParser#setProcessCreator(petruchio.interfaces.pi.ProcessCreator)
     */
    @Override
    public void setProcessCreator(final ProcessCreator pbuilder) {
        this.pbuilder = pbuilder;
    }
    
    
    /**
     * TODO: description
     *
     */
    private void ensureAvailable() {
    	if(this.definitions == null) {
            this.definitions = this.parseProcessDefinitions(this.lexer,
                            this.errorHandlers, this.pbuilder);
            this.procIt = this.definitions.iterator();
            this.properties = this.getProperties();
            if(this.properties == null) {
            	this.properties = new Properties();
            }
        }
    }
    
    
    /**
     * Returns a fresh instance of the lexer to use. The lexer uses the given
     * reader as its input.
     *
     * @param reader The lexer's input
     * @return A fresh instance of the lexer to use which reads the input using
     *         the given reader.
     */
    public abstract CUPLexer getLexer(Reader reader);
    
    
    /**
     * Parses the the input given by the lexer and uses the given ProcessCreator
     * to build all the syntax trees for the process definitions.
     * The local compiler properties have to be generated as well.
     *
     * @param lexer The lexer to use
     * @param handler The error handler to use
     * @param pbuilder The ProcessCreator to use during parsing
     * @return The parsed process definitions from the lexer's input 
     */
    public abstract Collection<ProcessDefinition<?>> parseProcessDefinitions(
                    CUPLexer lexer, Collection<ParserErrorHandler> handlers,
                    ProcessCreator pbuilder);
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public abstract String getDescription();
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiParser#getDefaultExtensions()
     */
    @Override
    public abstract String getDefaultExtensions();

    
    /**
     * TODO: Document
     *
     * @return
     */
    public abstract Properties getProperties();

}