/*
 * Project: Petruchio
 * File:    DotWriter.java
 * Author:  Tim Strazny
 * Created: 21.02.2008
 */
package petruchio.pi.writers;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import petruchio.compiler.CommonTasks;
import petruchio.compiler.ProcessData;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.PiWriter;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessDefinition;
import petruchio.interfaces.pi.ProcessManager;


/**
 * TODO: Document
 * 
 * @author Tim Strazny
 */
public class DOTWriter implements PiWriter<ProcessData>, SelfDescribing {

    public static final int MAX_LENGTH = 20;
    
    private static final String PROCESS_PREFIX = "P";
    
    private static final boolean CREATE_CLUSTERS = false;
    private static final boolean PROCESS_FREE_NAMES = false;
    
    private int procCounter = 0;
    private int defCounter = 0;
    private int nameCounter = 0;
    private int fnCounter = 0;
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A pi calculus process writer, that outputs the process graph in graphviz's DOT language.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "dot";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiWriter#write(java.lang.String, java.lang.String, petruchio.interfaces.pi.ProcessManager, java.io.OutputStream)
     */
    @Override
    public void write(String comment, String header, ProcessManager<ProcessData> pm,
            OutputStream output) {
        final PrintWriter out = new PrintWriter(output);
        
        this.writeHeaderComment(comment, out);
        
        this.procCounter = 0;
        this.defCounter = 0;
        this.nameCounter = 0;
        this.fnCounter = 0;
        
        out.print("graph \"");
        out.print(this.encode("process"));
        out.print("\" {\n"
                + "  node [shape=\"plaintext\", "
                + "fontsize=\"10\", fixedsize=\"false\", width=\".3\", "
                + "height=\".3\", label=\"\", style=\"filled\", "
                + "fillcolor=\"white\"];\n");

        final Map<Name, String> free = new HashMap<Name, String>();
        final SortedMap<Integer, String> legend = new TreeMap<Integer, String>();
        for(final ProcessDefinition<ProcessData> pd : pm.getProcessDefinitions()) {
            this.writeProcessDefinition(pd, free, out, legend);
        }
        for(final Map.Entry<Name, String> e : free.entrySet()) {
            out.print("  \"");
            out.print(e.getValue());
            out.print("\" [shape=\"ellipse\", label=\"");
            out.print(e.getKey());
            out.print("\"];\n");
        }
        if(!legend.isEmpty()) {
            // grappa does not support HTML-style labels
//            out.print("  legend [shape=\"plaintext\", label=<<TABLE BORDER=\"1\">");
//            for(final Map.Entry<Integer, String> e : legend.entrySet()) {
//                out.print("<TR><TD>");
//                out.print(DOTWriter.PROCESS_PREFIX);
//                out.print(e.getKey());
//                out.print("</TD><TD>");
//                for(final char ch : this.breakLongLines(e.getValue()).toCharArray()) {
//                    switch(ch) {
//                        case '\n': out.print("<BR/>"); break;
//                        case '\"': out.print("&quot;"); break;
//                        case '<' : out.print("&lt;"); break;
//                        case '>' : out.print("&gt;"); break;
//                        default  : out.print(ch);
//                    }
//                }
//                out.print("</TD></TR>");
//            }
//            out.print("</TABLE>>];\n");
            
            out.print("  legend [shape=\"record\", rankdir=\"TB\", label=\"{");
            for(final Iterator<Map.Entry<Integer, String>> it = legend.entrySet().iterator(); it.hasNext();) {
                final Map.Entry<Integer, String> e = it.next();
                out.print("{");
                out.print(DOTWriter.PROCESS_PREFIX);
                out.print(e.getKey());
                out.print("|");
                for(final char ch : this.breakLongLines(e.getValue()).toCharArray()) {
                    switch(ch) {
                        case '\n': out.print("\\n"); break;
                        case '\"': out.print("\\\""); break;
                        case ' ' : out.print("\\ "); break;
                        case '|' : out.print("\\|"); break;
                        case '{' : out.print("\\{"); break;
                        case '}' : out.print("\\}"); break;
                        case '<' : out.print("\\<"); break;
                        case '>' : out.print("\\>"); break;
                        default  : out.print(ch);
                    }
                }
                out.print("}");
                if(it.hasNext()) {
                    out.print("|");
                }
            }
            out.print("}\"];\n");
        }
        
        out.print("}");
        
        out.close();
    }

    
    private void writeProcessDefinition(final ProcessDefinition<ProcessData> pd,
            final Map<Name, String> free, final PrintWriter out, final Map<Integer, String> legend) {
        if(CREATE_CLUSTERS) {
            out.print("  subgraph \"cluster_");
        } else {
            out.print("  subgraph \"");
        }
        final String id = "def" + ++this.defCounter;
        out.print(id);
        out.print("\" {\n    label=\"");
        out.print(this.encode(pd.getProcessID().toString()));
        if(!pd.getParameters().isEmpty()) {
            out.print("(");
            out.print(this.encode(pd.getParameters().toString()));
            out.print(")");
        }
        out.print("\";\n");
        final Map<Name, String> names = new HashMap<Name, String>(pd.getParameters().size());
        final Collection<Name> fn = CommonTasks.freeNames(pd.getProcess());
        final Collection<Name> toCheck = new HashSet<Name>(fn);
        for(final Name n : pd.getParameters().getParameters()) {
            toCheck.remove(n);
            final String nName = this.encode(n.toString());
            final String nID = "def" + this.defCounter + "n" + ++this.nameCounter;
            names.put(n, nID);
            out.print("    \"");
            out.print(nID);
            out.print("\" [shape=\"octagon\", label=\"");
            out.print(nName);
            out.print("\"];\n");
        }
        for(final Name n : toCheck) {
            String nID = free.get(n);
            if(nID == null) {
                nID = "fn" + ++this.fnCounter;
                free.put(n, nID);
            }
            if(PROCESS_FREE_NAMES) {
                names.put(n, nID);
            }
        }
        this.writeProcess(pd.getProcess(), names, out, legend);
        out.print("  }\n");
    }
    
    
    private void writeProcess(final Process<ProcessData> p, final Map<Name, String> names,
            final PrintWriter out, final Map<Integer, String> legend) {
        if(!p.getActionPrefixes().isEmpty() || p.getType() != Process.Type.COMPOSITION
                || ((ProcessComposition<ProcessData>)p).getOperator() != ProcessComposition.Operator.PARALLEL) {
            final String s = p.toString();
            final String broken = this.breakLongLines(s);
            final String string = this.encode(broken);
            final String id = "proc" + ++this.procCounter;
            out.print("    \"");
            out.print(id);
            boolean connected = false;
            final Collection<Name> fn = CommonTasks.freeNames(p);
            for(final Map.Entry<Name, String> e : names.entrySet()) {
                final Name n = e.getKey();
                if(fn.contains(n)) {
                    connected = true;
                    break;
                }
            }
            if(broken.equals(s) || !connected) {
                out.print("\" [label = \"");
            } else {
                out.print("\" [label = \"");
                out.print(DOTWriter.PROCESS_PREFIX);
                out.print(this.procCounter);
                out.print("\", tip = \"");
                legend.put(this.procCounter, this.encode(s));
            }
            out.print(string);
            out.print("\"];\n");
            if(connected) {
                this.connect(id, p, names, out);
            }
        } else {
            final ProcessComposition<ProcessData> c = (ProcessComposition<ProcessData>)p;
            final Map<Name, String> subNames = new HashMap<Name, String>(names);
            for(final Name n : c.getRestrictions()) {
                final String nName = this.encode(n.toString());
                final String nID = "def" + this.defCounter + "proc" + this.procCounter + "n" + ++this.nameCounter;
                subNames.put(n, nID);
                out.print("    \"");
                out.print(nID);
                out.print("\" [shape = \"box\", label = \"");
                out.print(nName);
                out.print("\"];\n");
            }
            for(final Process<ProcessData> sub : c.getProcesses()) {
                this.writeProcess(sub, subNames, out, legend);
            }
        }
    }
    
    
    private boolean connect(final String id, final Process<ProcessData> p,
            final Map<Name, String> names, final PrintWriter out) {
        if(names == null) {
            return false;
        }
        boolean connected = false;
        final Collection<Name> fn = CommonTasks.freeNames(p);
        for(final Map.Entry<Name, String> e : names.entrySet()) {
            final Name n = e.getKey();
            if(fn.contains(n)) {
                out.print("    \"");
                out.print(e.getValue());
                out.print("\" -- \"");
                out.print(id);
                out.print("\";\n");
                connected = true;
            }
        }
        return connected;
    }
    
    
    private void writeHeaderComment(final String comment, final PrintWriter out) {
        if(comment == null) {
            return;
        }
        out.print("/*\n");
        final String[] lines = comment.split("\r\n|[\n\r\u2028\u2029\u0085]");
        for(final String line : lines) {
            out.print(" * ");
            out.print(line);
            out.print("\n");
        }
        out.print(" */\n\n");
    }
    
    
    private String encode(final String s) {
        final StringBuilder b = new StringBuilder();
        for(final char ch : s.toCharArray()) {
            switch(ch) {
                case '\"': b.append("\\\""); break;
                case '\n': b.append("\\n"); break;
                default: b.append(ch);
            }
        }
        return b.toString();
    }
    
    
    private String breakLongLines(final String s) {
        if(s.length() < DOTWriter.MAX_LENGTH) {
            return s;
        }
        final int split = Math.min(DOTWriter.MAX_LENGTH, s.length()/2);
        final StringBuilder b = new StringBuilder();
        for(int start = 0; start < s.length(); start += split) {
            b.append(s.substring(start, Math.min(s.length(), start+split)));
            if(s.length() > start + split) {
                char ch = s.charAt(start+split);
                while(ch != '.' && ch != ' ' && ch != '+' && ch != ';' && ch != '|'
                        && s.length() > start + split) {
                    b.append(ch);
                    ++start;
                    if(s.length() > start + split) {
                        ch = s.charAt(start+split);
                    }
                }
                if(s.length() > start + split) {
                    b.append("\n");
                }
            }
        }
        return b.toString();
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        // nothing to do
    }

}
