/*
 * Project: Petruchio
 * File:    HALPiWriter.java
 * Author:  Tim Strazny
 * Created: 07.01.2008
 */
package petruchio.pi.writers;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import petruchio.compiler.CommonTasks;
import petruchio.compiler.LexProcessComparator;
import petruchio.compiler.ProcessData;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.InputAction;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.OutputAction;
import petruchio.interfaces.pi.Parameters;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessDefinition;
import petruchio.interfaces.pi.ProcessID;
import petruchio.interfaces.pi.ProcessManager;
import petruchio.interfaces.pi.ProcessReference;
import petruchio.pi.Guard;


/**
 * TODO: Document
 * 
 * @author Tim Strazny
 */
public class HALPiWriter implements SelfDescribing, petruchio.interfaces.pi.PiWriter<ProcessData> {

    private final Map<Name, String> names = new HashMap<Name, String>(0);
    private final Map<Name, String> ids = new HashMap<Name, String>(0);
    private final Collection<String> theNames = new HashSet<String>(0);
    private final Collection<String> theIDs = new HashSet<String>(0);
    private final Map<Name, List<Name>> deadReferences = new HashMap<Name, List<Name>>(0);
    private final Map<ProcessID, List<Name>> defs = new HashMap<ProcessID, List<Name>>(0);
    private boolean isMonadic = true;
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "Writes pi calculus process definitions in the HAL tool's syntax.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "halpi";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiWriter#write(java.lang.String, java.lang.String, petruchio.interfaces.pi.ProcessManager, java.io.OutputStream)
     */
    @Override
    public void write(final String comment, final String header,
            final ProcessManager<ProcessData> pm, final OutputStream output) {
        this.reset();
        final PrintWriter out = new PrintWriter(output);
        if(header.length() != 0) {
            out.println(header);
        }
        if(comment.length() != 0) {
            this.writeComment(comment, out);
        }
        final Map<Name, Collection<Name>> free = new HashMap<Name, Collection<Name>>();
        boolean done;
        do {
            done = true;
            for(final ProcessDefinition<ProcessData> pd : pm.getProcessDefinitions()) {
                if(this.collectDef(pd, free)) {
                    done = false;
                }
            }
        } while(!done);
        for(final ProcessDefinition<ProcessData> pd : pm.getProcessDefinitions()) {
            this.propagateDef(pd, free);
        }
        for(final ProcessDefinition<?> pd : pm.getProcessDefinitions()) {
            this.writeProcessDefinition(pm, pd, out);
        }
        for(final Map.Entry<Name, List<Name>> n : this.deadReferences.entrySet()) {
            final String id = this.getID(n.getKey());
            out.println();
            out.print("define ");
            out.print(id);
            out.print("(");
            final List<Name> thenames = n.getValue();
            if(!thenames.isEmpty()) {
                this.writeParameters(thenames, out);
            }
            out.println(") = nil");
        }
        
        out.println();
        
        for(final Name id : this.deadReferences.keySet()) {
            out.print("build ");
            out.println(this.getID(id));
        }
        for(final ProcessDefinition<?> pd : pm.getProcessDefinitions()) {
            out.print("build ");
            out.println(this.getID(pd.getProcessID().getID()));
        }
        out.close();
        if(!this.isMonadic) {
            System.out.println("WARNING: There is a process that makes use of the polyadic "
                    + "extension of the Pi-Calculus. HAL's syntax only supports the monadic "
                    + "version. Please correct your code and re-export to HAL.");
        }
        this.reset();
    }
    
    
    private boolean collectDef(final ProcessDefinition<ProcessData> pd, final Map<Name, Collection<Name>> free) {
        Collection<Name> fn = free.get(pd.getProcessID().getID());
        boolean result = false;
        if(fn == null) {
            fn = new HashSet<Name>(CommonTasks.freeNames(pd.getProcess()));
            fn.removeAll(pd.getParameters().getParameters());
            free.put(pd.getProcessID().getID(), fn);
            result = true;
        }
        for(final Name id : CommonTasks.getData(pd.getProcess()).getRefProcIDs()) {
            Collection<Name> idfn = free.get(id);
            if(idfn != null && fn.addAll(idfn)) {
                result = true;
            }
        }
        return result;
    }
    
    
    private boolean propagateDef(final ProcessDefinition<ProcessData> pd, final Map<Name, Collection<Name>> free) {
        List<Name> params = pd.getParameters().getParameters();
        boolean result = false;
        if(!free.isEmpty()) {
            params = new ArrayList<Name>(free.get(pd.getProcessID().getID()));
            params.removeAll(pd.getParameters().getParameters());
            params.removeAll(CommonTasks.boundNames(pd.getProcess()));
            Collections.sort(params);
            params.addAll(0, pd.getParameters().getParameters());
            result = true;
        }
        this.defs.put(pd.getProcessID(), params);
        return result;
    }
    
    
    /**
     * TODO: Document
     *
     * @param n
     * @param names
     * @return
     */
    private String getName(final Name name) {
        String n = this.names.get(name);
        if(n == null) {
            n = name.toString();
            final StringBuilder b = new StringBuilder();
            final char[] nar = n.toCharArray();
            int i;
            if(Character.isLetter(nar[0])) {
                b.append(Character.toLowerCase(nar[0]));
                i = 1;
            } else {
                b.append('n');
                i = 0;
            }
            
            while(i < nar.length) {
                final char ch = nar[i];
                if(ch != '_' && ch != '$' && ch != '\'' && (ch < 'a' || ch > 'z')
                        && (ch < 'A' || ch > 'Z') && (ch < '0' || ch > '9')) {
                    b.append('_');
                } else {
                    b.append(ch);
                }
                ++i;
            }
            n = b.toString();
            i = 0;
            while(this.theNames.contains(n)) {
                n = b.toString() + "_" + i;
                ++i;
            }
            this.names.put(name, n);
            this.theNames.add(n);
        }
        return n;
    }
    
    
    /**
     * TODO: Document
     *
     * @param id
     * @param names
     * @return
     */
    private String getID(final Name id) {
        String n = this.ids.get(id);
        if(n == null) {
            n = id.toString();
            final StringBuilder b = new StringBuilder();
            final char[] nar = n.toCharArray();
            int i;
            if(Character.isLetter(nar[0])) {
                b.append(nar[0]);
//                b.append(Character.toLowerCase(nar[0]));
                i = 1;
            } else {
                b.append('n');
                i = 0;
            }
            
            while(i < nar.length) {
                final char ch = nar[i];
                if(ch != '_' && ch != '$' && ch != '\'' && (ch < 'a' || ch > 'z')
                        && (ch < 'A' || ch > 'Z') && (ch < '0' || ch > '9')) {
                    b.append('_');
                } else {
                    b.append(ch);
                }
                ++i;
            }
            n = b.toString();
            i = 0;
            while(this.theIDs.contains(n)) {
                n = b.toString() + "_" + i;
                ++i;
            }
            this.ids.put(id, n);
            this.theIDs.add(n);
        }
        return n;
    }
    

    /**
     * TODO: Document
     *
     * @param main
     * @param pd
     * @param out
     */
    private void writeProcessDefinition(final ProcessManager<?> pm,
            final ProcessDefinition<?> pd, final PrintWriter out) {
        out.println();
        this.writeProcessID(pd, out);
        out.println(" =");
        out.print("  ");
        this.writeProcess(pm, "  ", pd.getProcess(), false, out);
    }
    
    
    /**
     * TODO: Document
     *
     * @param indent
     * @param p
     * @param out
     */
    @SuppressWarnings("unchecked")
    private void writeProcess(final ProcessManager<?> pm,
            final String indent, final Process<?> p,
            final boolean underPrefix, final PrintWriter out) {
        this.writeRestrictions(p, out);
        this.writePrefixes(p, out);
        final boolean curUnderPrefix = underPrefix || !p.getActionPrefixes().isEmpty() || !p.getRestrictions().isEmpty();
        switch(p.getType()) {
            case NULL: {
                out.println("nil");
                break;
            }
            case PREFIX: {
                final PrefixProcess<?> a = (PrefixProcess<?>)p;
                this.writeProcess(pm, indent, a.getProcess(), curUnderPrefix, out);
                break;
            }
            case REFERENCE: {
                final ProcessReference<?> a = (ProcessReference<?>)p;
                this.writeReference(pm, a.getProcessID(), a.getParameters(), out);
                out.println();
                break;
            }
            case COMPOSITION: {
                final ProcessComposition<?> a = (ProcessComposition<?>)p;
                final String op;
                switch(a.getOperator()) {
                    case CHOICE: op = "+"; break;
                    case PARALLEL: op = "|"; break;
                    case SEQUENCE: op = ";"; break;
                    default: throw new InternalError("Unknown operator: " + a.getOperator());
                }
                final StringBuilder b = new StringBuilder(indent);
                for(int i = op.length(); i >= 0; --i) {
                    b.append(' ');
                }
                final String indent2 = b.toString();
                final List<?> procs;
                if(a.getProcesses().size() != 1) {
                    if(curUnderPrefix) {
                        out.println();
                        out.print(indent);
                    }
                    out.print("( ");
                    
                    if(a.getOperator() != ProcessComposition.Operator.SEQUENCE) {
                        procs = new ArrayList<Process<?>>(a.getProcesses());
                        Collections.sort((List<Process<?>>)procs, LexProcessComparator.getInstance());
                    } else {
                        procs = a.getProcesses();
                    }
                } else {
                    procs = a.getProcesses();
                }
                for(final Iterator<?> it = procs.iterator(); it.hasNext(); ) {
                    final Process<?> s = (Process<?>)it.next();
                    this.writeProcess(pm, indent2, s, a.getProcesses().size() == 1 && curUnderPrefix, out);
                    if(it.hasNext()) {
                        out.print(indent);
                        out.print(op);
                        out.print(" ");
                    }
                }
                if(a.getProcesses().size() != 1) {
                    out.print(indent);
                    out.println(")");
                }
                break;
            }
            default:
                throw new InternalError("Unknown process type: " + p.getType());
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param p
     * @param out
     */
    private void writeRestrictions(final Process<?> p, final PrintWriter out) {
        if(!p.getRestrictions().isEmpty()) {
            for(final Iterator<Name> it = p.getRestrictions().iterator(); it.hasNext();) {
                out.print("(");
                out.print(this.getName(it.next()));
                out.print(")");
            }
        }
    }
    
    /**
     * TODO: Document
     *
     * @param p
     * @param out
     */
    private void writePrefixes(final Process<?> p, final PrintWriter out) {
        if(!p.getActionPrefixes().isEmpty()) {
            for(final ActionPrefix ap : p.getActionPrefixes()) {
                switch(ap.getType()) {
                    case INTERNAL: {
                        out.print("tau.");
                        break;
                    }
                    case INPUT: {
                        final InputAction a = (InputAction)ap;
                        out.print(this.getName(a.getChannel()));
                        out.print("?(");
                        this.writeParameters(a.getNames().getParameters(), out);
                        out.print(").");
                        break;
                    }
                    case OUTPUT: {
                        final OutputAction a = (OutputAction)ap;
                        out.print(this.getName(a.getChannel()));
                        out.print("!");
                        this.writeParameters(a.getNames().getParameters(), out);
                        out.print(".");
                        break;
                    }
                    case GUARD: {
                        final Guard a = (Guard)ap;
                        out.print("[");
                        out.print(this.getName(a.leftOperand()));
                        out.print("=");
                        out.print(this.getName(a.rightOperand()));
                        out.print("]");
                        break;
                    }
                    default:
                        throw new InternalError("Unknown type of action prefix: " + ap.getType());
                }
            }
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param main
     * @param pd
     * @param out
     */
    private void writeProcessID(final ProcessDefinition<?> pd, final PrintWriter out) {
//        if(pd.isMain()) {
//            out.println("main");
//        }
        out.print("define ");
        final String id = this.getID(pd.getProcessID().getID());
        out.print(id);
        final List<Name> params = this.defs.get(pd.getProcessID());
        out.print("(");
        if(!params.isEmpty()) {
            this.writeParameters(params, out);
        }
        out.print(")");
    }
    
    
    /**
     * TODO: Document
     *
     * @param pid
     * @param params
     * @param out
     */
    private void writeReference(final ProcessManager<?> pm,
            final ProcessID pid, final Parameters theParams,
            final PrintWriter out) {
        final ProcessDefinition<?> pd = pm.getProcessDefinition(pid);
        final List<Name> params;
        if(pd == null) {
            this.deadReferences.put(pid.getID(), theParams.getParameters());
            params = null;
        } else {
            params = new ArrayList<Name>(this.defs.get(pd.getProcessID()));
            for(int i = theParams.size()-1; i >= 0; --i) {
                params.set(i, theParams.getParameters().get(i));
            }
        }
         
        final String id = this.getID(pid.getID());
        out.print(id);
        out.print("(");
        if(params != null && !params.isEmpty()) {
            this.writeParameters(params, out);
        }
        out.print(")");
    }
    
    
    /**
     * TODO: Document
     *
     * @param params
     * @param out
     */
    private void writeParameters(final List<Name> params, final PrintWriter out) {
        if(params.size() != 1) {
            this.isMonadic = false;
        }
        for(final Iterator<Name> it = params.iterator(); it.hasNext(); ) {
            out.print(this.getName(it.next()));
            if(it.hasNext()) {
                out.print(", ");
            }
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param comment
     * @parm out
     */
    private void writeComment(final String comment, final PrintWriter out) {
        out.println("/*");
        final String[] lines = comment.split("\r\n|[\n\r\u2028\u2029\u0085]");
        for(final String line : lines) {
            out.print(" * ");
            out.println(line);
        }
        out.println(" */");
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        this.isMonadic = true;
        this.names.clear();
        this.ids.clear();
        this.theIDs.clear();
        this.theNames.clear();
        this.deadReferences.clear();
        this.defs.clear();
    }
    
}