/*
 * Project: Petruchio
 * File:    LaTeXWriter.java
 * Author:  Tim Strazny
 * Created: 17.10.2007
 */
package petruchio.pi.writers;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import petruchio.compiler.LexProcessComparator;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.InputAction;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.OutputAction;
import petruchio.interfaces.pi.Parameters;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessDefinition;
import petruchio.interfaces.pi.ProcessID;
import petruchio.interfaces.pi.ProcessManager;
import petruchio.interfaces.pi.ProcessReference;
import petruchio.pi.Guard;


/**
 * TODO: Document
 * 
 * @author Tim Strazny
 */
public class LaTeXWriter<V> implements SelfDescribing, petruchio.interfaces.pi.PiWriter<V> {

    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "Writes the pi calculus process definitions as LaTeX code - with redefinable user commands.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "tex";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiWriter#write(java.lang.String, java.lang.String, petruchio.interfaces.pi.ProcessManager, java.io.OutputStream)
     */
    @Override
    public void write(final String comment, final String header,
            final ProcessManager<V> pm, final OutputStream output) {
        final PrintWriter out = new PrintWriter(output);
        
        out.println("\\documentclass[a4paper]{article}");
        out.println("\\newcommand{\\mainProc}{\\ensuremath{\\mathbf{main}\\;}}");
        out.println("\\newcommand{\\referenceProc}[2]{\\ensuremath{\\mathit{#1}\\lfloor{}#2\\rfloor{}}}");
        out.println("\\newcommand{\\referenceEmptyProc}[1]{\\ensuremath{\\mathit{#1}\\lfloor{}\\rfloor{}}}");
        out.println("\\newcommand{\\nullProc}{\\ensuremath{\\mathbf{0}}}");
        out.println("\\newcommand{\\defProc}[3]{\\ensuremath{\\mathit{#1}(#2) := #3;\\\\}}");
        out.println("\\newcommand{\\defEmptyProc}[2]{\\ensuremath{\\mathit{#1} := #2;\\\\}}");
        out.println("\\newcommand{\\choiceOp}{\\ensuremath{+}}");
        out.println("\\newcommand{\\parallelOp}{\\ensuremath{\\mid}}");
        out.println("\\newcommand{\\sequenceOp}{\\ensuremath{;}}");
        out.println("\\newcommand{\\restrictOp}{\\ensuremath{\\nu}}");
        out.println("\\newcommand{\\internalAction}{\\ensuremath{\\tau}}");
        out.println("\\newcommand{\\guardPrefix}[2]{\\ensuremath{[#1=#2]}}");
        out.println("\\newcommand{\\inputAction}[2]{\\ensuremath{#1(#2)}}");
        out.println("\\newcommand{\\inputEmptyAction}[1]{\\ensuremath{#1}}");
        out.println("\\newcommand{\\outputAction}[2]{\\ensuremath{\\overline{#1}\\langle{}#2\\rangle{}}}");
        out.println("\\newcommand{\\outputEmptyAction}[1]{\\ensuremath{\\overline{#1}}}");
        out.println("\\newcommand{\\printName}[1]{\\ensuremath{\\mathit{#1}}}");
        out.println("\\begin{document}");
        
        if(header.length() != 0) {
            out.println(header);
        }
        if(comment.length() != 0) {
            this.writeComment(comment, out);
        }
        for(final ProcessDefinition<?> pd : pm.getProcessDefinitions()) {
            this.writeProcessDefinition(pd, out);
        }
        out.println("\\end{document}");
        out.close();
    }
    
    
    /**
     * TODO: Document
     *
     * @param pd
     * @param out
     */
    private void writeProcessDefinition(final ProcessDefinition<?> pd, final PrintWriter out) {
        out.println();
        this.writeProcessID(pd, out);
        out.println();
        out.print("{  ");
        this.writeProcess("  ", pd.getProcess(), false, out);
        out.println("}");
    }
    
    
    /**
     * TODO: Document
     *
     * @param indent
     * @param p
     * @param out
     */
    @SuppressWarnings("unchecked")
    private void writeProcess(final String indent, final Process<?> p,
            final boolean underPrefix, final PrintWriter out) {
        this.writeRestrictions(p, out);
        this.writePrefixes(p, out);
        final boolean curUnderPrefix = underPrefix || !p.getActionPrefixes().isEmpty() || !p.getRestrictions().isEmpty();
        switch(p.getType()) {
            case NULL: {
                out.println("\\nullProc{}");
                break;
            }
            case PREFIX: {
                final PrefixProcess<?> a = (PrefixProcess<?>)p;
                this.writeProcess(indent, a.getProcess(), true, out);
                break;
            }
            case REFERENCE: {
                final ProcessReference<?> a = (ProcessReference<?>)p;
                this.writeReference(a.getProcessID(), a.getParameters(), out);
                out.println();
                break;
            }
            case COMPOSITION: {
                final ProcessComposition<?> a = (ProcessComposition<?>)p;
                final String op;
                switch(a.getOperator()) {
                    case CHOICE: op = "\\choiceOp{}"; break;
                    case PARALLEL: op = "\\parallelOp{}"; break;
                    case SEQUENCE: op = "\\sequenceOp{}"; break;
                    default: throw new InternalError("Unknown operator: " + a.getOperator());
                }
                String indent2 = indent;
                final List<?> procs;
                if(a.getProcesses().size() != 1) {
                    if(curUnderPrefix) {
                        out.println();
                        out.print(indent);
                    }
                    out.print("( ");
                    for(int i = op.length()-2; i >= 0; --i) {
                        out.print(' ');
                    }
                    
                    final StringBuilder b = new StringBuilder(indent);
                    for(int i = op.length(); i >= 0; --i) {
                        b.append(' ');
                    }
                    indent2 = b.toString();
                    
                    if(a.getOperator() != ProcessComposition.Operator.SEQUENCE) {
                        procs = new ArrayList<Process<?>>(a.getProcesses());
                        Collections.sort((List<Process<?>>)procs, LexProcessComparator.getInstance());
                    } else {
                        procs = a.getProcesses();
                    }
                } else {
                    procs = a.getProcesses();
                }
                for(final Iterator<?> it = procs.iterator(); it.hasNext(); ) {
                    final Process<?> s = (Process<?>)it.next();
                    this.writeProcess(indent2, s, a.getProcesses().size() == 1 && curUnderPrefix, out);
                    if(it.hasNext()) {
                        out.print(indent);
                        out.print(op);
                        out.print(" ");
                    }
                }
                if(a.getProcesses().size() != 1) {
                    out.print(indent);
                    out.println(")");
                }
                break;
            }
            default:
                throw new InternalError("Unknown process type: " + p.getType());
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param p
     * @param out
     */
    private void writeRestrictions(final Process<?> p, final PrintWriter out) {
        if(!p.getRestrictions().isEmpty()) {
            out.print("\\restrictOp{} ");
            for(final Iterator<Name> it = p.getRestrictions().iterator(); it.hasNext();) {
                out.print("\\printName{");
                out.print(this.encode(it.next()));
                out.print("}");
                if(it.hasNext()) {
                    out.print(", ");
                }
            }
            out.print(".");
        }
    }
    
    /**
     * TODO: Document
     *
     * @param p
     * @param out
     */
    private void writePrefixes(final Process<?> p, final PrintWriter out) {
        if(!p.getActionPrefixes().isEmpty()) {
            for(final ActionPrefix ap : p.getActionPrefixes()) {
                switch(ap.getType()) {
                    case INTERNAL: {
                        out.print("\\internalAction{}");
                        break;
                    }
                    case INPUT: {
                        final InputAction a = (InputAction)ap;
                        if(a.getNames().isEmpty()) {
                            out.print("\\inputEmptyAction{\\printName{");
                            out.print(this.encode(a.getChannel()));
                            out.print("}}");
                        } else {
                            out.print("\\inputAction{\\printName{");
                            out.print(this.encode(a.getChannel()));
                            out.print("}}{");
                            this.writeParameters(a.getNames(), out);
                            out.print("}");
                        }
                        break;
                    }
                    case OUTPUT: {
                        final OutputAction a = (OutputAction)ap;
                        if(a.getNames().isEmpty()) {
                            out.print("\\outputEmptyAction{\\printName{");
                            out.print(this.encode(a.getChannel()));
                            out.print("}}");
                        } else {
                            out.print("\\outputAction{\\printName{");
                            out.print(this.encode(a.getChannel()));
                            out.print("}}{");
                            this.writeParameters(a.getNames(), out);
                            out.print("}");
                        }
                        break;
                    }
                    case GUARD: {
                        final Guard a = (Guard)ap;
                        out.print("\\guardPrefix{\\printName{");
                        out.print(this.encode(a.leftOperand()));
                        out.print("}}{\\printName{");
                        out.print(this.encode(a.rightOperand()));
                        out.print("}}");
                        break;
                    }
                    default:
                        throw new InternalError("Unknown type of action prefix: " + ap.getType());
                }
                out.print(".");
            }
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param pd
     * @param out
     */
    private void writeProcessID(final ProcessDefinition<?> pd,
            final PrintWriter out) {
        if(pd.isMain()) {
            out.println("\\mainProc{}");
        }
        if(pd.getParameters().isEmpty()) {
            out.print("\\defEmptyProc{\\printName{");
            out.print(this.encode(pd.getProcessID().getID()));
            out.print("}}");
        } else {
            out.print("\\defProc{\\printName{");
            out.print(this.encode(pd.getProcessID().getID()));
            out.print("}}{");
            this.writeParameters(pd.getParameters(), out);
            out.print("}");
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param pid
     * @param params
     * @param out
     */
    private void writeReference(final ProcessID pid, final Parameters params,
            final PrintWriter out) {
        if(params.isEmpty()) {
            out.print("\\referenceEmptyProc{\\printName{");
            out.print(this.encode(pid.getID()));
            out.print("}}");
        } else {
            out.print("\\referenceProc{\\printName{");
            out.print(this.encode(pid.getID()));
            out.print("}}{");
            this.writeParameters(params, out);
            out.print("}");
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param params
     * @param out
     */
    private void writeParameters(final Parameters params, final PrintWriter out) {
        for(final Iterator<Name> it = params.getParameters().iterator(); it.hasNext(); ) {
            out.print("\\printName{");
            out.print(this.encode(it.next()));
            out.print("}");
            if(it.hasNext()) {
                out.print(", ");
            }
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param comment
     * @parm out
     */
    private void writeComment(final String comment, final PrintWriter out) {
        final String[] lines = comment.split("\r\n|[\n\r\u2028\u2029\u0085]");
        for(final String line : lines) {
            out.print("% ");
            out.println(line);
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param s
     * @return
     */
    private String encode(final Name s) {
        final StringBuilder b = new StringBuilder();
        for(final char ch : s.toString().toCharArray()) {
            switch(ch) {
                case '$': b.append("\\$"); break;
                case '%': b.append("\\%"); break;
                case '\\': b.append("\\backslash"); break;
                case '{': b.append("\\{"); break;
                case '}': b.append("\\}"); break;
                case '_': b.append("\\_"); break;
                default: b.append(ch);
            }
        }
        return b.toString();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        // nothing to do
    }
    
}
