/*
 * Project: Petruchio
 * File:    PiWriter.java
 * Author:  Tim Strazny
 * Created: 17.10.2007
 */
package petruchio.pi.writers;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import petruchio.compiler.LexProcessComparator;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.pi.ActionPrefix;
import petruchio.interfaces.pi.InputAction;
import petruchio.interfaces.pi.Name;
import petruchio.interfaces.pi.OutputAction;
import petruchio.interfaces.pi.Parameters;
import petruchio.interfaces.pi.PrefixProcess;
import petruchio.interfaces.pi.Process;
import petruchio.interfaces.pi.ProcessComposition;
import petruchio.interfaces.pi.ProcessDefinition;
import petruchio.interfaces.pi.ProcessID;
import petruchio.interfaces.pi.ProcessManager;
import petruchio.interfaces.pi.ProcessReference;
import petruchio.pi.Guard;


/**
 * TODO: Document
 * 
 * @author Tim Strazny
 */
public class PiWriter<V> implements SelfDescribing, petruchio.interfaces.pi.PiWriter<V> {

    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "Writes the pi process definitions in a simple pi format with indentation.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "pi";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.pi.PiWriter#write(java.lang.String, java.lang.String, petruchio.interfaces.pi.ProcessManager, java.io.OutputStream)
     */
    @Override
    public void write(final String comment, final String header,
            final ProcessManager<V> pm, final OutputStream output) {
        final PrintWriter out = new PrintWriter(output);
        if(header.length() != 0) {
            out.println(header);
        }
        if(comment.length() != 0) {
            this.writeComment(comment, out);
        }
        for(final ProcessDefinition<?> pd : pm.getProcessDefinitions()) {
            this.writeProcessDefinition(pd, out);
        }
        out.close();
    }
    
    
    /**
     * TODO: Document
     *
     * @param pd
     * @param out
     */
    private void writeProcessDefinition(final ProcessDefinition<?> pd, final PrintWriter out) {
        out.println();
        this.writeProcessID(pd, out);
        out.println(" :=");
        out.print("  ");
        this.writeProcess("  ", pd.getProcess(), out);
        out.println("  ;");
    }
    
    
    /**
     * TODO: Document
     *
     * @param indent
     * @param p
     * @param out
     */
    @SuppressWarnings("unchecked")
    private void writeProcess(final String indent, final Process<?> p, final PrintWriter out) {
        this.writeRestrictions(indent, p, out);
        this.writePrefixes(indent, p, out);
        switch(p.getType()) {
            case NULL: {
                out.println("0");
                break;
            }
            case PREFIX: {
                final PrefixProcess<?> a = (PrefixProcess<?>)p;
                this.writeProcess(indent, a.getProcess(), out);
                break;
            }
            case REFERENCE: {
                final ProcessReference<?> a = (ProcessReference<?>)p;
                this.writeReference(a.getProcessID(), a.getParameters(), out);
                out.println();
                break;
            }
            case COMPOSITION: {
                final ProcessComposition<?> a = (ProcessComposition<?>)p;
                final String op;
                switch(a.getOperator()) {
                    case CHOICE: op = "+"; break;
                    case PARALLEL: op = "|"; break;
                    case SEQUENCE: op = ";"; break;
                    default: throw new InternalError("Unknown operator: " + a.getOperator());
                }
                final StringBuilder b = new StringBuilder(indent);
                for(int i = op.length(); i >= 0; --i) {
                    b.append(' ');
                }
                final String indent2 = b.toString();
                final List<?> procs;
                if(a.getProcesses().size() != 1) {
                    out.print("( ");
                    
                    if(a.getOperator() != ProcessComposition.Operator.SEQUENCE) {
                        procs = new ArrayList<Process<?>>(a.getProcesses());
                        Collections.sort((List<Process<?>>)procs, LexProcessComparator.getInstance());
                    } else {
                        procs = a.getProcesses();
                    }
                } else {
                    procs = a.getProcesses();
                }
                
                for(final Iterator<?> it = procs.iterator(); it.hasNext(); ) {
                    final Process<?> s = (Process<?>)it.next();
                    this.writeProcess(indent2, s, out);
                    if(it.hasNext()) {
                        out.print(indent);
                        out.print(op);
                        out.print(" ");
                    }
                }
                if(a.getProcesses().size() != 1) {
                    out.print(indent);
                    out.println(")");
                }
                break;
            }
            default:
                throw new InternalError("Unknown process type: " + p.getType());
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param p
     * @param out
     */
    private void writeRestrictions(final String indent, final Process<?> p, final PrintWriter out) {
        if(!p.getRestrictions().isEmpty()) {
            out.print("new ");
            for(final Iterator<Name> it = p.getRestrictions().iterator(); it.hasNext();) {
                out.print(it.next());
                if(it.hasNext()) {
                    out.print(", ");
                }
            }
            out.println(".");
            out.print(indent);
        }
    }
    
    /**
     * TODO: Document
     *
     * @param p
     * @param out
     */
    private void writePrefixes(final String indent,
            final Process<?> p, final PrintWriter out) {
        if(!p.getActionPrefixes().isEmpty()) {
            for(final ActionPrefix ap : p.getActionPrefixes()) {
                switch(ap.getType()) {
                    case INTERNAL: {
                        out.print("tau");
                        break;
                    }
                    case INPUT: {
                        final InputAction a = (InputAction)ap;
                        out.print(a.getChannel());
                        out.print("(");
                        this.writeParameters(a.getNames(), out);
                        out.print(")");
                        break;
                    }
                    case OUTPUT: {
                        final OutputAction a = (OutputAction)ap;
                        out.print(a.getChannel());
                        out.print("<");
                        this.writeParameters(a.getNames(), out);
                        out.print(">");
                        break;
                    }
                    case GUARD: {
                        final Guard a = (Guard)ap;
                        out.print("[");
                        out.print(a.leftOperand());
                        out.print("=");
                        out.print(a.rightOperand());
                        out.print("]");
                        break;
                    }
                    default:
                        throw new InternalError("Unknown type of action prefix: " + ap.getType());
                }
                out.println(".");
                out.print(indent);
            }
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param pd
     * @param out
     */
    private void writeProcessID(final ProcessDefinition<?> pd,
            final PrintWriter out) {
        if(pd.isMain()) {
            out.println("main");
        }
        this.writeReference(pd.getProcessID(), pd.getParameters(), out);
    }
    
    
    /**
     * TODO: Document
     *
     * @param pid
     * @param params
     * @param out
     */
    private void writeReference(final ProcessID pid, final Parameters params,
            final PrintWriter out) {
        out.print(pid.getID());
        if(!params.isEmpty()) {
            out.print("(");
            this.writeParameters(params, out);
            out.print(")");
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param params
     * @param out
     */
    private void writeParameters(final Parameters params, final PrintWriter out) {
        for(final Iterator<Name> it = params.getParameters().iterator(); it.hasNext(); ) {
            out.print(it.next());
            if(it.hasNext()) {
                out.print(", ");
            }
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param comment
     * @parm out
     */
    private void writeComment(final String comment, final PrintWriter out) {
        out.println("/*");
        final String[] lines = comment.split("\r\n|[\n\r\u2028\u2029\u0085]");
        for(final String line : lines) {
            out.print(" * ");
            out.println(line);
        }
        out.println(" */");
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        // nothing to do
    }

}