package petruchio.pn;

import java.math.BigInteger;
import java.util.LinkedList;
import java.util.Queue;

import petruchio.interfaces.petrinet.BigIntegerMatrix;
import petruchio.interfaces.petrinet.IntMatrix;
import petruchio.interfaces.petrinet.Place;

/**
 * @author Tim Strazny
 */
public class BigIntegerInvariantAnalysis {

    /**
     * Returns a matrix where the columns form minimal-support non-negative integer solutions
     * of the integer linear equation system defined by given matrix.
     * 
     * @param incidence A matrix (which will be altered).
     * @return Matrix where columns form minimal-support non-negative integer solutions.
     */
    public static BigIntegerMatrix findInvariants(final BigIntegerMatrix incidence) {
        return findInvariants(incidence, true, 0, null);
    }
    
    
    /**
     * Returns a matrix where the columns form minimal-support (possibly non-negative) integer solutions
     * of the integer linear equation system defined by given matrix.
     * 
     * @param incidence A matrix (which will be altered).
     * @param semipositive Whether the solutions shall be non-negative.
     * @return Matrix where columns form minimal-support (possibly non-negative) integer solutions.
     */
    public static BigIntegerMatrix findInvariants(final BigIntegerMatrix incidence, final boolean semipositive) {
        return findInvariants(incidence, semipositive, -1, null);
    }
    
    /**
     * Returns a matrix where the columns form minimal-support (possibly non-negative) integer solutions
     * of the integer linear equation system defined by given matrix.
     * 
     * @param incidence A matrix (which will be altered).
     * @param semipositive Whether the solutions shall be non-negative.
     * @param count Number of rounds when attempting to make solutions non-negative.
     * @param indices 
     * @return Matrix where columns form minimal-support (possibly non-negative) integer solutions.
     */
    public static IntMatrix findInvariants(final IntMatrix incidence, final boolean semipositive,
            final int count, final Place[] indices) {
        final BigIntegerMatrix lresult = findInvariants(BigIntegerPNMatrix.fromColumnMajorArray(incidence.toColumnMajorArray()), semipositive, count, indices);
        final IntMatrix result = IntPNMatrix.fromMatrix(lresult);
        lresult.kill();
        return result;
    }
    
    
    /**
     * Returns a matrix where the columns form minimal-support (possibly non-negative) integer solutions
     * of the integer linear equation system defined by given matrix.
     * 
     * @param incidence A matrix (which will be altered).
     * @param semipositive Whether the solutions shall be non-negative.
     * @param count Number of rounds when attempting to make solutions non-negative.
     * @param indices 
     * @return Matrix where columns form minimal-support (possibly non-negative) integer solutions.
     */
    public static BigIntegerMatrix findInvariants(final BigIntegerMatrix incidence, final boolean semipositive,
            final int count, final Place[] indices) {
        final int columns = incidence.getColumns();
        
        final Queue<MatrixOperation> ops = solve(incidence);
        
        incidence.kill(); // remove internal array to free the memory
        
        // create solution matrix
        final BigIntegerMatrix invariants = BigIntegerPNMatrix.identity(columns, columns);
        while(!ops.isEmpty()) {
            ops.remove().perform(invariants);
        }

        ensureSemipositive(invariants, !semipositive, count, indices);
        
        if(semipositive) {
            removeRedundant(invariants);
        }
        
        return invariants;
    }
    
    private static class MatrixOperation {
        final int kill;
        final BigInteger alpha;
        final int alphaColumn;
        final BigInteger beta;
        final int betaColumn;
        
        MatrixOperation(final int kill) {
            this.kill = kill;
            this.alpha = BigInteger.ZERO;
            this.beta = BigInteger.ZERO;
            this.alphaColumn = -1;
            this.betaColumn = -1;
        }
        
        MatrixOperation(final BigInteger alpha, final int alphaColumn, final BigInteger beta, final int betaColumn) {
            this.kill = -1;
            this.alpha = alpha;
            this.beta = beta;
            this.alphaColumn = alphaColumn;
            this.betaColumn = betaColumn;
        }
        
        void perform(final BigIntegerMatrix matrix) {
            if(this.kill >= 0) {
                matrix.killColumn(this.kill);
            } else {
                for (int i = 0; i < matrix.getRows(); ++i) {
                    matrix.set(i, this.betaColumn, this.alpha.multiply(matrix.get(i, this.alphaColumn)).add(this.beta.multiply(matrix.get(i, this.betaColumn))));
                }
            }
        }
    }
    
    
    private static Queue<MatrixOperation> solve(final BigIntegerMatrix incidence) {
        final Queue<MatrixOperation> ops = new LinkedList<MatrixOperation>();
        final int[] pPlus = new int[incidence.getColumns()];
        final int[] pMinus = new int[incidence.getColumns()];
        
        int nonZeroRow;
        do {
            boolean cardinalityUsed = false;
            nonZeroRow = -1;

            for (int h = 0; h < incidence.getRows() && !cardinalityUsed && incidence.getColumns() != 0; ++h) {
                
                // pPlus and pMinus are large enough. in phase 1, the matrices don't grow.
                getIndices(incidence, h, pPlus, pMinus);
                
                boolean pPlusIsSingleton;

                if ((pPlus[0] == -1) ^ (pMinus[0] == -1)) { // case 1.1.a
                    
                    cardinalityUsed = true;
                    nonZeroRow = h;

                    int[] pPlusMinus = pPlus[0] == -1 ? pMinus : pPlus;

                    for (int j = 0; j < pPlusMinus.length && pPlusMinus[j] != -1; ++j) {
                        incidence.killColumn(pPlusMinus[j]);
                        
                        ops.add(new MatrixOperation(pPlusMinus[j]));
                    }
                    
                } else if ((pPlusIsSingleton = (pPlus[0] != -1 && (pPlus.length == 1 || pPlus[1] == -1)))
                        || ((pMinus[0] != -1 && (pMinus.length == 1 || pMinus[1] == -1)))) { // case 1.1.b

                    cardinalityUsed = true;
                    nonZeroRow = h;

                    final int[] pPlusMinus;
                    final int[] pMinusPlus;
                    if (pPlusIsSingleton) {
                        pPlusMinus = pPlus;
                        pMinusPlus = pMinus;
                    } else {
                        pPlusMinus = pMinus;
                        pMinusPlus = pPlus;
                    }

                    final int k = pPlusMinus[0];
                    final BigInteger chk = incidence.get(h, k).abs();

                    for (int cur = 0; cur < pMinusPlus.length && pMinusPlus[cur] != -1; ++cur) {
                        
                        final int j = pMinusPlus[cur];
                        final BigInteger chj = incidence.get(h, j).abs();
                        
                        for (int i = 0; i < incidence.getRows(); ++i) {
                            incidence.set(i, j, chj.multiply(incidence.get(i, k)).add(chk.multiply(incidence.get(i, j))));
                        }
                        
                        ops.add(new MatrixOperation(chj, k, chk, j));
                    }

                    incidence.killColumn(k);
                    
                    ops.add(new MatrixOperation(k));
                    
                    
                } else if (pPlus[0] != -1 || pMinus[0] != -1) {
                    nonZeroRow = h;
                    
                }

            } 

            if (!cardinalityUsed && nonZeroRow != -1) { // case 1.1.c
                final int h = nonZeroRow;
                int k = 0;
                while (!incidence.columnExists(k) || isZero(incidence.get(h, k))) {
                    ++k;
                }
                final BigInteger chk = incidence.get(h, k);

                for (int j = 0; j < incidence.getColumns(); ++j) {
                    if (j == k || !incidence.columnExists(j)) {
                        continue;
                    }
                    final BigInteger chj = incidence.get(h, j);
                    if(isZero(chj)) {
                        continue;
                    }

                    final BigInteger alpha = (chj.signum()*chk.signum() < 0) ? chj.abs() : (isNegative(chj) ? chj : chj.negate());
                    final BigInteger beta = chk.abs();

                    for (int i = 0; i < incidence.getRows(); ++i) {
                        incidence.set(i, j, alpha.multiply(incidence.get(i, k)).add(beta.multiply(incidence.get(i, j))));
                    }
                    
                    ops.add(new MatrixOperation(alpha, k, beta, j));
                }

                incidence.killColumn(k);
                ops.add(new MatrixOperation(k));
                
            }

        } while(nonZeroRow != -1);
        
        return ops;
    }
    
    private static boolean isZero(final BigInteger i) {
        return i.signum() == 0;
    }
    
    private static boolean isNegative(final BigInteger i) {
        return i.signum() < 0;
    }

    
    private static void getIndices(final BigIntegerMatrix A, final int rowNo, final int[] positivesArray,
            final int[] negativesArray) {
        final int n = A.getColumns();
        int negcount = 0;
        int poscount = 0;

        for (int i = 0; i < n; i++) {
            if(A.columnExists(i)) {
                final BigInteger val = A.get(rowNo, i);
                final int comp = val.signum();
                if (comp < 0) {
                    negativesArray[negcount++] = i;
                } else if (comp > 0) {
                    positivesArray[poscount++] = i;
                }
            }
        }
        if (negcount < n) {
            negativesArray[negcount] = -1;
        }
        if (poscount < n) {
            positivesArray[poscount] = -1;
        }

    }
    
//    private static void ensureSemipositive(final PNMatrix matrix) {
//        ensureSemipositive(matrix, false, -1, null);
//    }
    
    private static void ensureSemipositive(final BigIntegerMatrix matrix, final boolean correctSinglesOnly, final int count, final Place[] indices) {
        if(count == 0) {
            return;
        }
        
        int[] pPlus = new int[matrix.getColumns()];
        int[] pMinus = new int[matrix.getColumns()];
        int nextH = findNextH(matrix, correctSinglesOnly, indices);
        int ctr = 0;
        
        while (nextH != -1 && (count < 0 || ctr++ < count)) {
            final int h = nextH;
            nextH = -1;
            
            pPlus = ensureSize(pPlus, matrix.getColumns());
            pMinus = ensureSize(pMinus, matrix.getColumns());
            getIndices(matrix, h, pPlus, pMinus);
            
            for (int curj = 0; curj < pPlus.length && pPlus[curj] != -1; ++curj) {
                for (int curk = 0; curk < pMinus.length && pMinus[curk] != -1; ++curk) {

                    final int j = pPlus[curj];
                    final int k = pMinus[curk];

                    final BigInteger a = matrix.get(h, k).negate();
                    final BigInteger b = matrix.get(h, j);

                    final int newColumn = matrix.addColumn();
                    for (int i = 0; i < matrix.getRows(); ++i) {
                        final BigInteger v = a.multiply(matrix.get(i, j)).add(b.multiply(matrix.get(i, k)));
                        matrix.set(i, newColumn, v);
                        if(!correctSinglesOnly && isNegative(v)) {
                            nextH = i;
                        }
                    }

                    BigInteger gcd = gcd(matrix, newColumn);

                    // divide all the col elements by their gcd if gcd > 1
                    if (gcd.compareTo(BigInteger.ONE) > 1) {
                        matrix.divide(newColumn, gcd);
                    }
                }
            }

            for (int i = 0; i < pMinus.length && pMinus[i] != 0; ++i) {
                matrix.killColumn(pMinus[i] - 1);
            }
            
            if(nextH == -1) {
                nextH = findNextH(matrix, correctSinglesOnly, indices);
            }
        }
    }
    
    private static int findNextH(final BigIntegerMatrix matrix, final boolean correctSinglesOnly, final Place[] indices) {
        for (int j = 0; j < matrix.getColumns(); j++) {
            if(matrix.columnExists(j)) {
                int nextH = -1;
                for (int i = 0; i < matrix.getRows(); i++) {            
                    if (isNegative(matrix.get(i, j)) && (!correctSinglesOnly || !indices[j].boundKnown())) {
                        if(correctSinglesOnly) {
                            if(nextH != -1) {
                                nextH = -1;
                                break;
                            }
                            nextH = i;
                        } else {
                            return i;
                        }
                    }
                }
                if(nextH != -1) {
                    return nextH;
                }
            }
        }
        return -1;
    }
    
    private static int[] ensureSize(final int[] a, final int size) {
        // the content of the array is irrelevant!
        return a.length >= size ? a : new int[size+10];
    }
    
    private static BigInteger gcd(final BigIntegerMatrix A, final int column) {
        BigInteger gcd = A.get(0, column);
        for (int i = 1; i < A.getRows(); i++) {
            if (!isZero(A.get(i, column)) || !isZero(gcd)) {
                gcd = gcd.gcd(A.get(i, column));
            }
        }
        return gcd;
    }
    
    private static void removeRedundant(final BigIntegerMatrix matrix) {
        int[] nonmin = matrix.findNonMinimal();
        for(int i = 0; i < nonmin.length && nonmin[i] != -1; ++i) {
            matrix.killColumn(nonmin[i]);
        }
    }
    
}