/*
 * Project:  Petruchio
 * Filename: Matrix.java
 * Created:  02.06.2009
 * Author:   Tim Strazny
 */
package petruchio.pn;

import java.math.BigInteger;
import java.util.Arrays;

import petruchio.interfaces.petrinet.BigIntegerMatrix;

/**
 * A matrix class for fast column operations.
 * 
 * @author Tim Strazny
 */
public class BigIntegerPNMatrix implements BigIntegerMatrix {

    // number of spare columns created when up-sizing matrix
    private static final int INCREMENT = 10;
    
    // column-major order
    private BigInteger[][] content;
    
    // current number of rows
    private int rows;
    
    // current number of columns (some might be nulled)
    private int columns;
    
    // stack for indices of nulled columns 
    private int[] freeColumns = new int[10];
    
    // current size of the stack
    private int numFreeColumns = 0;

    // private constructor, use static factory methods instead
    private BigIntegerPNMatrix(final int rows, final int columns) {
        this.rows = rows;
        this.columns = columns;
        this.content = new BigInteger[this.columns][this.rows];
    }
    
    
    /**
     * Creates a matrix from the given row-major array.
     * All rows are needed to have the same number of columns.
     * 
     * @param array A row-major array.
     * @return The matrix representation of the array.
     */
    public static BigIntegerPNMatrix fromRowMajorArray(final int[][] array) {
        final int rows = array.length;
        if(rows == 0) {
            throw new IllegalArgumentException("Array must have positive row dimension.");
        }
        final int columns = array[0].length;
        if(columns == 0) {
            throw new IllegalArgumentException("Array must have positive column dimension.");
        }
        final BigIntegerPNMatrix m = new BigIntegerPNMatrix(rows, columns);
        for(int i = 0; i < rows; ++i) {
            if(array[i].length != columns) {
                throw new IllegalArgumentException("All rows of array must have the same length.");
            }
            for(int j = 0; j < columns; ++j) {
                m.set(j, i, array[i][j]);
            }
        }
        return m;
    }
    
    
    /**
     * Creates a matrix from the given row-major array.
     * All rows are needed to have the same number of columns.
     * 
     * @param array A row-major array.
     * @return The matrix representation of the array.
     */
    public static BigIntegerPNMatrix fromRowMajorArray(final long[][] array) {
        final int rows = array.length;
        if(rows == 0) {
            throw new IllegalArgumentException("Array must have positive row dimension.");
        }
        final int columns = array[0].length;
        if(columns == 0) {
            throw new IllegalArgumentException("Array must have positive column dimension.");
        }
        final BigIntegerPNMatrix m = new BigIntegerPNMatrix(rows, columns);
        for(int i = 0; i < rows; ++i) {
            if(array[i].length != columns) {
                throw new IllegalArgumentException("All rows of array must have the same length.");
            }
            for(int j = 0; j < columns; ++j) {
                m.set(j, i, array[i][j]);
            }
        }
        return m;
    }
    
    /**
     * Creates a matrix from the given row-major array.
     * All rows are needed to have the same number of columns.
     * 
     * @param array A row-major array.
     * @return The matrix representation of the array.
     */
    public static BigIntegerPNMatrix fromRowMajorArray(final BigInteger[][] array) {
        final int rows = array.length;
        if(rows == 0) {
            throw new IllegalArgumentException("Array must have positive row dimension.");
        }
        final int columns = array[0].length;
        if(columns == 0) {
            throw new IllegalArgumentException("Array must have positive column dimension.");
        }
        final BigIntegerPNMatrix m = new BigIntegerPNMatrix(rows, columns);
        for(int i = 0; i < rows; ++i) {
            if(array[i].length != columns) {
                throw new IllegalArgumentException("All rows of array must have the same length.");
            }
            for(int j = 0; j < columns; ++j) {
                m.set(j, i, array[i][j]);
            }
        }
        return m;
    }
    
    
    /**
     * Creates a matrix from the given column-major array.
     * All columns are needed to have the same number of rows.
     * 
     * @param array A column-major array.
     * @return The matrix representation of the array.
     */
    public static BigIntegerPNMatrix fromColumnMajorArray(final int[][] array) {
        final int columns = array.length;
        if(columns == 0) {
            throw new IllegalArgumentException("Array must have positive column dimension.");
        }
        final int rows = array[0].length;
        if(rows == 0) {
            throw new IllegalArgumentException("Array must have positive row dimension.");
        }
        final BigIntegerPNMatrix m = new BigIntegerPNMatrix(rows, columns);
        for(int i = 0; i < columns; ++i) {
            if(array[i].length != rows) {
                throw new IllegalArgumentException("All columns of array must have the same length.");
            }
            for(int j = 0; j < rows; ++j) {
                m.set(j, i, array[i][j]);
            }
        }
        return m;
    }
    
    /**
     * Creates a matrix from the given column-major array.
     * All columns are needed to have the same number of rows.
     * 
     * @param array A column-major array.
     * @return The matrix representation of the array.
     */
    public static BigIntegerPNMatrix fromColumnMajorArray(final long[][] array) {
        final int columns = array.length;
        if(columns == 0) {
            throw new IllegalArgumentException("Array must have positive column dimension.");
        }
        final int rows = array[0].length;
        if(rows == 0) {
            throw new IllegalArgumentException("Array must have positive row dimension.");
        }
        final BigIntegerPNMatrix m = new BigIntegerPNMatrix(rows, columns);
        for(int i = 0; i < columns; ++i) {
            if(array[i].length != rows) {
                throw new IllegalArgumentException("All columns of array must have the same length.");
            }
            for(int j = 0; j < rows; ++j) {
                m.set(j, i, array[i][j]);
            }
        }
        return m;
    }
    
    /**
     * Creates a matrix from the given column-major array.
     * All columns are needed to have the same number of rows.
     * 
     * @param array A column-major array.
     * @return The matrix representation of the array.
     */
    public static BigIntegerPNMatrix fromColumnMajorArray(final BigInteger[][] array) {
        final int columns = array.length;
        if(columns == 0) {
            throw new IllegalArgumentException("Array must have positive column dimension.");
        }
        final int rows = array[0].length;
        if(rows == 0) {
            throw new IllegalArgumentException("Array must have positive row dimension.");
        }
        final BigIntegerPNMatrix m = new BigIntegerPNMatrix(rows, columns);
        for(int i = 0; i < columns; ++i) {
            if(array[i].length != rows) {
                throw new IllegalArgumentException("All columns of array must have the same length.");
            }
            System.arraycopy(array[i], 0, m.content[i], 0, rows);
        }
        return m;
    }
    
    
    /**
     * Returns an instance of the specified size initialized with zero elements.
     * 
     * @param rows Number of rows.
     * @param columns Number of columns.
     * @return A rows x columns PNMatrix of zeros.
     */
    public static BigIntegerPNMatrix getInstance(final int rows, final int columns) {
        return new BigIntegerPNMatrix(rows, columns);
    }
    
    
    /**
     * Returns an identity matrix of the specified size.
     * 
     * @param rows Number of rows.
     * @param columns Number of columns.
     * @return A rows x columns identity PNMatrix where the diagonal contains
     *         1 elements and all other elements are 0.
     */
    public static BigIntegerPNMatrix identity(final int rows, final int columns) {
        final BigIntegerPNMatrix id = new BigIntegerPNMatrix(rows, columns);
        for (int j = Math.min(rows, columns) - 1; j >= 0; --j) {
            id.set(j, j, 1);
        }
        return id;
    }
    
    
    /**
     * Creates a row-major array representation of this matrix.
     * 
     * @return Row-major array representation of this matrix.
     */
    @Override
    public BigInteger[][] toRowMajorArray() {
        final BigInteger[][] ar = new BigInteger[this.rows][this.columns-this.numFreeColumns];
        for(int j = 0, k = 0; j < this.columns; ++j) {
            if(this.columnExists(j)) {
                for(int i = 0; i < this.rows; ++i) {
                    ar[i][k] = this.get(i, j);
                }
                ++k;
            }
        }
        return ar;
    }
    
    
    /**
     * Creates a column-major array representation of this matrix.
     * 
     * @return Column-major array representation of this matrix.
     */
    @Override
    public BigInteger[][] toColumnMajorArray() {
        final BigInteger[][] ar = new BigInteger[this.columns-this.numFreeColumns][this.rows];
        for(int j = 0, i = 0; j < this.columns; ++j) {
            if(this.columnExists(j)) {
                System.arraycopy(this.content[j], 0, ar[i], 0, this.rows);
                ++i;
            }
        }
        return ar;
    }
    
    
    /**
     * Clears the internal data.
     * <b>The matrix must not be used after this call.</b>
     * This method exists for memory performance reasons only.
     */
    @Override
    public void kill() {
        this.rows = -1;
        this.columns = -1;
        this.content = null;
        this.freeColumns = null;
        this.numFreeColumns = -1;
    }
    
    
    /**
     * Removes the column with the given index from the matrix.
     * This does not change the matrix' column dimension as it is performed in-place for performance reasons.
     * The column is marked as inaccessible and {@link #columnExists(int)} returns <code>false</code>.
     * 
     * @param column A column index.
     * @see #columnExists(int)
     * @see #addColumn()
     */
    @Override
    public void killColumn(final int column) {
        if(this.columnExists(column)) {
            this.content[column] = null;
            this.ensureSizeFreeColumns(1);
            this.freeColumns[this.numFreeColumns++] = column;
        }
    }
    
    
    // ensure the stack for free column indices is large enough
    private void ensureSizeFreeColumns(final int increment) {
        if(this.freeColumns.length <= this.numFreeColumns + increment-1) {
            final int[] temp = new int[this.freeColumns.length+increment+10];
            System.arraycopy(this.freeColumns, 0, temp, 0, this.freeColumns.length);
            this.freeColumns = temp;
        }
    }
    
    
    /**
     * Adds a column to the matrix and returns the new columns index.
     * 
     * If there are spare columns left, this does not change the matrix'
     * column dimension and it is performed in-place.
     * The column is accessible and {@link #columnExists(int)} returns <code>true</code>.
     * 
     * @return The index of the column created.
     * @see #killColumn(int)
     * @see #columnExists(int)
     */
    @Override
    public int addColumn() {
        if(this.numFreeColumns != 0) {
            final int column = this.freeColumns[--this.numFreeColumns];
            this.content[column] = new BigInteger[this.getRows()];
            return column;
        }
        final BigInteger[][] temp = new BigInteger[this.columns + INCREMENT+1][];
        final int column = this.columns;
        System.arraycopy(this.content, 0, temp, 0, this.columns);
        temp[this.columns] = new BigInteger[this.rows];
        this.content = temp;
        this.columns = this.content.length;
        
        this.ensureSizeFreeColumns(INCREMENT);
        for(int i = 0; i < INCREMENT; ++i) {
            this.freeColumns[i] = column + i + 1;
        }
        this.numFreeColumns = INCREMENT;
        
        return column;
    }

    
    /**
     * Checks whether the column with the given index exists in the matrix.
     * 
     * @param column A column index.
     * @return <code>true</code> if the column may be accessed.
     */
    @Override
    public boolean columnExists(final int column) {
        return this.content[column] != null;
    }
    
    
    /**
     * Retrieves the specified element from the matrix.
     * 
     * @param row A row index.
     * @param column A column index.
     * @return The matrix' element at the specified position.
     * @see #columnExists(int)
     */
    @Override
    public BigInteger get(final int row, final int column) {
        final BigInteger b = this.content[column][row];
        return b == null ? BigInteger.ZERO : b;
    }
    
    
    /**
     * Sets the value of the specified element to <code>s</code>.
     * 
     * @param row A row index.
     * @param column A column index.
     * @param s The new value.
     * @see #columnExists(int)
     */
    @Override
    public void set(final int row, final int column, final BigInteger s) {
        this.content[column][row] = s;
    }
    
    /**
     * Sets the value of the specified element to <code>s</code>.
     * 
     * @param row A row index.
     * @param column A column index.
     * @param s The new value.
     * @see #columnExists(int)
     */
    public void set(final int row, final int column, final long s) {
        this.content[column][row] = BigInteger.valueOf(s);
    }
    
    /**
     * Sets the value of the specified element to <code>s</code>.
     * 
     * @param row A row index.
     * @param column A column index.
     * @param s The new value.
     * @see #columnExists(int)
     */
    public void set(final int row, final int column, final int s) {
        this.content[column][row] = BigInteger.valueOf(s);
    }

    
    @Override
    public int hashCode() {
        return Arrays.hashCode(content);
    }


    @Override
    public boolean equals(Object obj) {
        if(this == obj)
            return true;
        if(obj == null)
            return false;
        if(getClass() != obj.getClass())
            return false;
        BigIntegerPNMatrix other = (BigIntegerPNMatrix)obj;
        if(columns != other.columns)
            return false;
        if(!Arrays.equals(content, other.content))
            return false;
        if(rows != other.rows)
            return false;
        return true;
    }


    /**
     * Returns the number of columns in the matrix.
     * Note that not every column has to be accessible.
     * 
     * @return The number of columns in the matrix.
     * @see #columnExists(int)
     */
    @Override
    public int getColumns() {
        return this.columns;
    }
    
    
    /**
     * Returns the number of rows in the matrix.
     * 
     * @return Number of rows in the matrix.
     */
    @Override
    public int getRows() {
        return this.rows;
    }

    
    /**
     * Subtracts value <code>s</code> from the element at the specified index.
     * 
     * @param row A row index.
     * @param column A column index.
     * @param s The value to subtract.
     * @see #columnExists(int)
     */
    @Override
    public void subtract(final int row, final int column, final BigInteger s) {
        this.content[column][row] = this.content[column][row].subtract(s);
    }
    
    /**
     * Subtracts value <code>s</code> from the element at the specified index.
     * 
     * @param row A row index.
     * @param column A column index.
     * @param s The value to subtract.
     * @see #columnExists(int)
     */
    public void subtract(final int row, final int column, final long s) {
        this.content[column][row] = this.content[column][row].subtract(BigInteger.valueOf(s));
    }
    
    /**
     * Subtracts value <code>s</code> from the element at the specified index.
     * 
     * @param row A row index.
     * @param column A column index.
     * @param s The value to subtract.
     * @see #columnExists(int)
     */
    public void subtract(final int row, final int column, final int s) {
        this.content[column][row] = this.content[column][row].subtract(BigInteger.valueOf(s));
    }
    
    
    /**
     * Divides a column by the value <code>s</code>.
     * 
     * @param column A column index.
     * @param s The value to divide by.
     * @see #columnExists(int)
     */
    @Override
    public void divide(final int column, final BigInteger s) {
        for (int i = 0; i < this.rows; i++) {
            this.content[column][i] = this.content[column][i].divide(s);
        }
    }
    
    /**
     * Divides a column by the value <code>s</code>.
     * 
     * @param column A column index.
     * @param s The value to divide by.
     * @see #columnExists(int)
     */
    public void divide(final int column, final long s) {
        for (int i = 0; i < this.rows; i++) {
            this.content[column][i] = this.content[column][i].divide(BigInteger.valueOf(s));
        }
    }
    
    /**
     * Divides a column by the value <code>s</code>.
     * 
     * @param column A column index.
     * @param s The value to divide by.
     * @see #columnExists(int)
     */
    public void divide(final int column, final int s) {
        for (int i = 0; i < this.rows; i++) {
            this.content[column][i] = this.content[column][i].divide(BigInteger.valueOf(s));
        }
    }

    
    /**
     * Returns an array of columns with non-minimal support.
     * The array may be longer than the actual number of non-minimal support columns.
     * In this case, the indices are succeeded by a -1 entry in the array.
     * 
     * @return An array of column indices with non-minimal support.
     */
    @Override
    public int[] findNonMinimal() {
        int[] indices = new int[10];
        int count = 0;
        for (int i = 0; i < this.columns; i++) {
            if(!this.columnExists(i)) {
                continue;
            }
            for (int j = i + 1; j < this.columns; j++) {
                if(!this.columnExists(j)) {
                    continue;
                }
                final int r = this.getGEQ(i, j);
                if (r != -1) {
                    if(indices.length == count) {
                        final int[] tmp = new int[count + 10];
                        System.arraycopy(indices, 0, tmp, 0, count);
                        indices = tmp;
                    }
                    indices[count++] = r;
                }
            }
        }
        if(count < indices.length) {
            indices[count] = -1;
        }

        return indices;
    }

    
    private static boolean isZero(final BigInteger i) {
        return i.signum() == 0;
    }
    
    
    /**
     * Compares two columns of this matrix element-wise and returns the index of a column
     * which is greater than or equal to the other (or -1) if they are incomparable. 
     * 
     * @param colA A column index.
     * @param colB A second column index.
     * @return The index of the greater (or equal) column or -1 if they are incomparable.
     */
    @Override
    public int getGEQ(final int colA, final int colB) {
        boolean aGEQb = true;
        boolean bGEQa = true;
        for (int i = 0; i < this.rows; i++) {
            final boolean a = !isZero(this.get(i, colA));
            final boolean b = !isZero(this.get(i, colB));
            if (aGEQb && !a && b) {
                if (!bGEQa) {
                    return -1;
                }
                aGEQb = false;
            } else if (bGEQa && a && !b) {
                if (!aGEQb) {
                    return -1;
                }
                bGEQa = false;
            }
        }

        return aGEQb ? colA : colB;
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.Matrix#transpose()
     */
    @Override
    public void transpose() {
        this.content = this.toRowMajorArray();
        this.rows = this.columns-this.numFreeColumns;
        this.columns = this.content.length;
        this.numFreeColumns = 0;
        this.freeColumns = new int[10];
    }
    
    @Override
    public String toString() {
        final StringBuilder s = new StringBuilder();
        s.append(this.rows);
        s.append('x');
        s.append(this.columns-this.numFreeColumns);
//        s.append('\n');
//        for(int i = 0; i < this.rows; ++i) {
//            for(int j = 0; j < this.columns; ++j) {
//                if(!this.columnExists(j)) {
//                    continue;
//                }
//                s.append(this.get(i, j));
//                s.append(' ');
//            }
//            s.append('\n');
//        }
//        s.append(" hash:");
//        s.append(this.hashCode());
        return s.toString();
    }

}