/*
 * Project:  Petruchio
 * Filename: PNConverter.java
 * Created:  05.09.2007
 * Author:   Tim Strazny
 */
package petruchio.pn;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReader;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.PetriNetStreamReader;
import petruchio.interfaces.petrinet.PetriNetStreamWriter;
import petruchio.interfaces.petrinet.PetriNetWriter;
import petruchio.interfaces.petrinet.Place;

/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class Converter {

    private static final Map<String, String> readers = new HashMap<String, String>();

    private static final Map<String, String> writers = new HashMap<String, String>();

    static {
        Converter.registerReader("xml", "petruchio.pn.readers.PNMLReader");
        Converter.registerReader("pnml", "petruchio.pn.readers.PNMLReader");
        Converter.registerReader("ll_net", "petruchio.pn.readers.PEPReader");
        Converter.registerReader("spec", "petruchio.pn.readers.CMReader");
        Converter.registerReader("pnt", "petruchio.pn.readers.INAReader");
        Converter.registerReader("lola", "petruchio.pn.readers.LoLAReader");

        Converter.registerWriter("xml", "petruchio.pn.writers.PNMLWriter");
        Converter.registerWriter("pnml", "petruchio.pn.writers.PNMLWriter");
        Converter.registerWriter("pn.tex", "petruchio.pn.writers.LaTeXWriter");
        Converter.registerWriter("tex", "petruchio.pn.writers.TikZWriter");
        Converter.registerWriter("pnt", "petruchio.pn.writers.INAWriter");
        Converter.registerWriter("ll_net", "petruchio.pn.writers.PEPWriter");
        Converter.registerWriter("spec", "petruchio.pn.writers.CMWriter");
        Converter.registerWriter("dot", "petruchio.pn.writers.DOTWriter");
        Converter.registerWriter("lola", "petruchio.pn.writers.LoLAWriter");
        Converter.registerWriter("pm", "petruchio.pn.writers.SPINWriter");
        Converter.registerWriter("dve", "petruchio.pn.writers.DVEWriter");
        Converter.registerWriter("svg", "petruchio.pn.writers.SVGWriter");
        Converter.registerWriter("armc", "petruchio.pn.writers.ARMCWriter");
        Converter.registerWriter("bpp", "petruchio.pn.writers.BPPWriter");
    }

    /**
     * TODO: Document
     * 
     * @return
     */
    public static Map<String, String> getReaders() {
        return Collections.unmodifiableMap(Converter.readers);
    }

    /**
     * TODO: Document
     * 
     * @return
     */
    public static Map<String, String> getWriters() {
        return Collections.unmodifiableMap(Converter.writers);
    }

    /**
     * @param ext
     * @param clazz
     * @return
     */
    public static String registerReader(final String ext, final String clazz) {
        return Converter.readers.put(ext, clazz);
    }

    /**
     * @param ext
     * @param clazz
     * @return
     */
    public static String registerWriter(final String ext, final String clazz) {
        return Converter.writers.put(ext, clazz);
    }
    
    /**
     * @param filename
     * @return
     */
    public static String getReaderName(final String filename) {
        final int idx = filename.lastIndexOf('.');
        return Converter.readers.get(filename.substring(idx + 1).toLowerCase());
    }
    
    /**
     * @param filename
     * @return
     */
    public static String getWriterName(final String filename) {
        final int idx = filename.lastIndexOf('.');
        return Converter.writers.get(filename.substring(idx + 1).toLowerCase());
    }
    
    /**
     * @param filename
     * @return
     */
    public static PetriNet loadNet(final String filename) {
        final String reader = getReaderName(filename);
        if(reader == null) {
            throw new IllegalArgumentException("PetriNetReader for file '" + filename + "' unknown.");
        }
        try {
            final PetriNetReader r = (PetriNetReader) Class.forName(reader).newInstance();
            r.setInput(new FileInputStream(filename));
            return r.read();
        } catch (final InstantiationException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (final IllegalAccessException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (final ClassNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (final FileNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        throw new IllegalArgumentException("Unable to load '" + filename + "'.");
    }
    
    /**
     * @param filename
     * @param net
     * @return
     */
    @SuppressWarnings("unchecked")
    public static Collection<Map<Place, Integer>> loadTargetMarkings(final String filename, final PetriNet net) {
        final String reader = getReaderName(filename);
        if(reader == null) {
            throw new IllegalArgumentException("PetriNetReader for file '" + filename + "' unknown.");
        }
        Exception cause = null;
        try {
            final Class<?> readerClass = Class.forName(reader);
            final Method loadTarget = readerClass.getMethod(PetriNetReader.METHOD_READ_TARGETS,
                    PetriNetReader.METHOD_READ_TARGETS_PARAM_TYPES);
            final Object[] args = { filename, net };
            final Object returned = loadTarget.invoke(readerClass.newInstance(), args);
            return (Collection<Map<Place, Integer>>) returned;
        } catch(final InvocationTargetException e) {
            cause = new RuntimeException("PetriNetReader '" + reader + "' was unable to load target markings from file '" + filename + "'.", e.getCause());
        } catch(final ClassCastException e) {
            cause = new RuntimeException("PetriNetReader '" + reader + "' cannot be cast to appropriate class.", e.getCause());
        } catch(final NoSuchMethodException e) {
            cause = new RuntimeException("PetriNetReader '" + reader + "' does not provide applicable " + PetriNetReader.METHOD_READ_TARGETS + " method. (NoSuchMethod)");
        } catch (final Exception e) {
            cause = e;
        }
        throw new IllegalArgumentException("Unable to load target markings from '" + filename + "'.", cause);
    }
    
    /**
     * @param filename
     * @param net
     * @return
     */
    @SuppressWarnings("unchecked")
    public static Collection<Map<Place, Integer>> loadInvariants(final String filename, final PetriNet net) {
        final String reader = getReaderName(filename);
        if(reader == null) {
            throw new IllegalArgumentException("PetriNetReader for file '" + filename + "' unknown.");
        }
        Exception cause = null;
        try {
            final Class<?> readerClass = Class.forName(reader);
            final Method loadTarget = readerClass.getMethod(PetriNetReader.METHOD_READ_INVARIANTS,
                    PetriNetReader.METHOD_READ_INVARIANTS_PARAM_TYPES);
            final Object[] args = { filename, net };
            final Object returned = loadTarget.invoke(readerClass.newInstance(), args);
            return (Collection<Map<Place, Integer>>) returned;
        } catch(final InvocationTargetException e) {
            cause = new RuntimeException("PetriNetReader '" + reader + "' was unable to load invariants from file '" + filename + "'.", e.getCause());
        } catch(final ClassCastException e) {
            cause = new RuntimeException("PetriNetReader '" + reader + "' cannot be cast to appropriate class.", e.getCause());
        } catch(final NoSuchMethodException e) {
            cause = new RuntimeException("PetriNetReader '" + reader + "' does not provide applicable " + PetriNetReader.METHOD_READ_INVARIANTS + " method. (NoSuchMethod)");
        } catch (final Exception e) {
            cause = e;
        }
        throw new IllegalArgumentException("Unable to load invariants from '" + filename + "'.", cause);
    }
    
    /**
     * @param filename
     * @return
     */
    public static void writeNet(final PetriNet net, final String filename) {
        final String writer = getWriterName(filename);
        if(writer == null) {
            throw new IllegalArgumentException("PetriNetWriter for file '" + filename + "' unknown.");
        }
        try {
            final PetriNetWriter r = (PetriNetWriter) Class.forName(writer).newInstance();
            r.write(net, new FileOutputStream(filename));
        } catch (InstantiationException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            throw new IllegalArgumentException("Unable to write '" + filename + "'.");
        } catch (IllegalAccessException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            throw new IllegalArgumentException("Unable to write '" + filename + "'.");
        } catch (ClassNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            throw new IllegalArgumentException("Unable to write '" + filename + "'.");
        } catch (FileNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            throw new IllegalArgumentException("Unable to write '" + filename + "'.");
        }
    }

    /**
     * TODO: document
     * 
     * @param args
     * @throws ClassNotFoundException
     * @throws IllegalAccessException
     * @throws InstantiationException
     * @throws FileNotFoundException
     */
    public static void main(String[] args) throws FileNotFoundException,
            InstantiationException, IllegalAccessException,
            ClassNotFoundException {
        if (args.length < 4 && args.length != 2) {
            System.out
                    .println("usage: java petruchio.pn.Converter <reader-class> "
                            + "<input-file> <writer-class> <output-file> [PetriNetReductors...]\n"
                            + "or:     java petruchio.pn.Converter <input-file> <output-file>");
        } else if (args.length == 2) {
            final String input = Converter.getAbsolutePath(args[0]);
            final String output = Converter.getAbsolutePath(args[1]);
            final String reader;
            if ((reader = getReaderName(args[0])) == null) {
                System.err.println("Unable to identify reader class for file: "
                        + args[0]);
                return;
            }
            final String writer;
            if ((writer = getWriterName(args[1])) == null) {
                System.err.println("Unable to identify writer class for file: "
                        + args[1]);
                return;
            }
            Converter.convert(reader, input, writer, output,
                    new ArrayList<PetriNetReductor>());
        } else {
            final String reader = args[0];
            final String input = Converter.getAbsolutePath(args[1]);
            final String writer = args[2];
            final String output = Converter.getAbsolutePath(args[3]);
            final List<PetriNetReductor> reductors = new ArrayList<PetriNetReductor>();
            for (int i = 4; i < args.length; ++i) {
                final PetriNetReductor reductor = (PetriNetReductor) Class
                        .forName(args[i]).newInstance();
                reductors.add(reductor);
            }
            Converter.convert(reader, input, writer, output, reductors);
        }
    }

    /**
     * Builds an absolute path from a given file and the current user directory.
     * 
     * @param file
     *            A relative or absolute path to the file
     * @return The absolute path for the given file name
     */
    private static String getAbsolutePath(final String file) {
        final File f = new File(file);
        if (!f.isAbsolute()) {
            final File f2 = new File(System.getProperty("user.dir")
                    + File.separator + file);
            try {
                return f2.getCanonicalPath();
            } catch (final IOException e) {
                return f2.getAbsolutePath();
            }
        }
        return file;
    }

    /**
     * TODO: document
     * 
     * @param reader
     * @param input
     * @param writer
     * @param output
     * @throws ClassNotFoundException
     * @throws IllegalAccessException
     * @throws InstantiationException
     * @throws FileNotFoundException
     */
    public static void convert(final String reader, final String input,
            final String writer, final String output,
            final List<PetriNetReductor> reductors)
            throws InstantiationException, IllegalAccessException,
            ClassNotFoundException, FileNotFoundException {

        final PetriNetReader r = (PetriNetReader) Class.forName(reader)
                .newInstance();
        final PetriNetWriter w = (PetriNetWriter) Class.forName(writer)
                .newInstance();
        r.setInput(new FileInputStream(input));
        if (reductors.isEmpty() && (r instanceof PetriNetStreamReader)
                && (w instanceof PetriNetStreamWriter)) {
            final long start = System.currentTimeMillis();
            final PetriNetStreamReader sr = (PetriNetStreamReader) r;
            final PetriNetStreamWriter sw = (PetriNetStreamWriter) w;
            sw.setOutput(new FileOutputStream(output));
            sr.readTo(sw);
            System.out.println("Converted net in "
                    + Converter.getDuration(start));
        } else {
            long start = System.currentTimeMillis();
            final PetriNet net = r.read();
            System.out.println("Read net in " + Converter.getDuration(start));
            if (!reductors.isEmpty()) {
                start = System.currentTimeMillis();
                Converter.reduce(net, reductors);
                System.out.println("Reduced net in "
                        + Converter.getDuration(start));
            }
            start = System.currentTimeMillis();
            w.write(net, new FileOutputStream(output));
            System.out.println("Wrote net in " + Converter.getDuration(start));
        }
    }

    /**
     * TODO: document
     * 
     * @param net
     * @param reductors
     */
    public static void reduce(final PetriNet net,
            final List<PetriNetReductor> reductorsList) {
        final PetriNetReductor[] reductors = new PetriNetReductor[reductorsList
                .size()];
        reductorsList.toArray(reductors);

        final List<int[]> results = new ArrayList<int[]>();

        boolean changed;
        int lastChange = -1;
        do {
            changed = false;
            final int[] result = new int[reductors.length];
            for (int i = 0; i < reductors.length; ++i) {
                if (lastChange != i) {
                    final int reduced = reductors[i].reduce(net);
                    if (reduced != 0) {
                        changed = true;
                        lastChange = i;
                        result[i] = reduced;
                    }
                }
            }
            if (changed) {
                results.add(result);
            }
        } while (changed);

        int sum = 0;
        if (!results.isEmpty()) {
            final StringBuilder b = new StringBuilder();
            if (results.size() != 1) {
                b.append("\nreductions");
            }
            boolean summation = false;
            int run = 1;
            for (final int[] result : results) {
                if (results.size() != 1) {
                    b.append("\n  run ");
                    b.append(run);
                }
                for (int i = 0; i < result.length; ++i) {
                    if (result[i] != 0) {
                        b.append("\n");
                        if (results.size() != 1) {
                            b.append("   ");
                        }
                        if (sum != 0) {
                            summation = true;
                        }
                        sum += result[i];
                        b.append(result[i]);
                        b.append(" elements removed by ");
                        b.append(reductors[i].getName());
                    }
                }
                ++run;
            }
            if (summation) {
                b.append("\n");
                b.append(sum);
                b.append(" elements removed by reductions");
            }
            net.setHeaderComment(net.getHeaderComment() + b.toString());
        }
        System.out.println("reduced by " + sum + " elements");
    }

    /**
     * @param start
     * @return
     */
    private static String getDuration(final long start) {
        long dur = System.currentTimeMillis() - start;
        final StringBuilder b = new StringBuilder();
        final long s = dur % 60000;
        dur /= 60000;
        b.append(s / 1000.0);
        b.append("s");
        if (dur == 0) {
            return b.toString();
        }

        final long m = dur % 60;
        dur /= 60;
        b.insert(0, "m ");
        b.insert(0, m);
        if (dur == 0) {
            return b.toString();
        }

        final long h = dur % 24;
        dur /= 24;
        b.insert(0, "h ");
        b.insert(0, h);
        if (dur == 0) {
            return b.toString();
        }

        final long d = dur % 7;
        dur /= 7;
        b.insert(0, "d ");
        b.insert(0, d);
        if (dur == 0) {
            return b.toString();
        }

        final long w = dur % 52;
        dur /= 52;
        b.insert(0, "w ");
        b.insert(0, w);
        if (dur == 0) {
            return b.toString();
        }

        final long y = dur;
        b.insert(0, "y ");
        b.insert(0, y);
        return b.toString();
    }

}
