package petruchio.pn;

import java.util.LinkedList;
import java.util.Queue;

import petruchio.interfaces.petrinet.IntMatrix;
import petruchio.interfaces.petrinet.Place;

/**
 * @author Tim Strazny
 */
public class IntInvariantAnalysis {

    /**
     * Returns a matrix where the columns form minimal-support non-negative integer solutions
     * of the integer linear equation system defined by given matrix.
     * 
     * @param incidence A matrix (which will be altered).
     * @return Matrix where columns form minimal-support non-negative integer solutions.
     */
    public static IntMatrix findInvariants(final IntMatrix incidence) {
        return findInvariants(incidence, true, 0, null);
    }
    
    
    /**
     * Returns a matrix where the columns form minimal-support (possibly non-negative) integer solutions
     * of the integer linear equation system defined by given matrix.
     * 
     * @param incidence A matrix (which will be altered).
     * @param semipositive Whether the solutions shall be non-negative.
     * @return Matrix where columns form minimal-support (possibly non-negative) integer solutions.
     */
    public static IntMatrix findInvariants(final IntMatrix incidence, final boolean semipositive) {
        return findInvariants(incidence, semipositive, -1, null);
    }
    
    
    /**
     * Returns a matrix where the columns form minimal-support (possibly non-negative) integer solutions
     * of the integer linear equation system defined by given matrix.
     * 
     * @param incidence A matrix (which will be altered).
     * @param semipositive Whether the solutions shall be non-negative.
     * @param count Number of rounds when attempting to make solutions non-negative.
     * @param indices 
     * @return Matrix where columns form minimal-support (possibly non-negative) integer solutions.
     */
    public static IntMatrix findInvariants(final IntMatrix incidence, final boolean semipositive,
            final int count, final Place[] indices) {
        final int columns = incidence.getColumns();
        
        final Queue<MatrixOperation> ops = solve(incidence);
        
        incidence.kill(); // remove internal array to free the memory
        
        // create solution matrix
        final IntMatrix invariants = IntPNMatrix.identity(columns, columns);
        while(!ops.isEmpty()) {
            ops.remove().perform(invariants);
        }

        ensureSemipositive(invariants, !semipositive, count, indices);
        
        if(semipositive) {
            removeRedundant(invariants);
        }
        
        return invariants;
    }
    
    private static class MatrixOperation {
        final int kill;
        final int alpha;
        final int alphaColumn;
        final int beta;
        final int betaColumn;
        
        MatrixOperation(final int kill) {
            this.kill = kill;
            this.alpha = 0;
            this.beta = 0;
            this.alphaColumn = -1;
            this.betaColumn = -1;
        }
        
        MatrixOperation(final int alpha, final int alphaColumn, final int beta, final int betaColumn) {
            this.kill = -1;
            this.alpha = alpha;
            this.beta = beta;
            this.alphaColumn = alphaColumn;
            this.betaColumn = betaColumn;
        }
        
        void perform(final IntMatrix matrix) {
            if(this.kill >= 0) {
                matrix.killColumn(this.kill);
            } else {
                for (int i = 0; i < matrix.getRows(); ++i) {
                    matrix.set(i, this.betaColumn, checkedADD(checkedMUL(this.alpha, matrix.get(i, this.alphaColumn)), checkedMUL(this.beta, matrix.get(i, this.betaColumn))));
                }
            }
        }
    }
    
    static int checkedABS(final int a) {
        if(a >= 0) {
            return a;
        } 
        if(a == Integer.MIN_VALUE) {
            throw new ArithmeticException("Overflow: abs(" + a + ") = " + -(long)a);
        }
        return -a;
    }
    
    
    static int checkedNEG(final int a) {
        if(a >= 0) {
            return -a;
        } 
        if(a == Integer.MIN_VALUE) {
            throw new ArithmeticException("Overflow: -(" + a + ") = " + -(long)a);
        }
        return -a;
    }
    
    
    static int signum(final int a) {
        return a > 0 ? 1 : (a == 0 ? 0 : -1);
    }
    
    
    static int checkedADD(final int a, final int b) {
        final int r = a + b;
        // ((a^r)&(b^r))<0) is a shorthand for:
        // (a<0 && b<0 && r>=0) ||
        // (a>0 && b>0 && r<=0) )
        if (((a ^ r) & (b ^ r)) < 0) {
            throw new ArithmeticException("Overflow: " + a + " + " + b + " = " + ((long)a + (long)b));
        }
        return r;
    }
    
    static int checkedMUL(final int a, final int b) {
        if(a == 0 || b == 0) {
            return 0;
        }
        // do and check
        final long r = (long)a * (long)b;
        if(r > Integer.MAX_VALUE || r < Integer.MIN_VALUE) {
            throw new ArithmeticException("Overflow: " + a + " * " + b + " = " + r);
        }
        return (int)r;
    }
    
    
    private static Queue<MatrixOperation> solve(final IntMatrix incidence) {
        final Queue<MatrixOperation> ops = new LinkedList<MatrixOperation>();
        final int[] pPlus = new int[incidence.getColumns()];
        final int[] pMinus = new int[incidence.getColumns()];
        
        int nonZeroRow;
        do {
            boolean cardinalityUsed = false;
            nonZeroRow = -1;

            for (int h = 0; h < incidence.getRows() && !cardinalityUsed && incidence.getColumns() != 0; ++h) {
                
                // pPlus and pMinus are large enough. in phase 1, the matrices don't grow.
                getIndices(incidence, h, pPlus, pMinus);
                
                boolean pPlusIsSingleton;

                if ((pPlus[0] == -1) ^ (pMinus[0] == -1)) { // case 1.1.a
                    
                    cardinalityUsed = true;
                    nonZeroRow = h;

                    int[] pPlusMinus = pPlus[0] == -1 ? pMinus : pPlus;

                    for (int j = 0; j < pPlusMinus.length && pPlusMinus[j] != -1; ++j) {
                        incidence.killColumn(pPlusMinus[j]);
                        
                        ops.add(new MatrixOperation(pPlusMinus[j]));
                    }
                    
                } else if ((pPlusIsSingleton = (pPlus[0] != -1 && (pPlus.length == 1 || pPlus[1] == -1)))
                        || ((pMinus[0] != -1 && (pMinus.length == 1 || pMinus[1] == -1)))) { // case 1.1.b

                    cardinalityUsed = true;
                    nonZeroRow = h;

                    final int[] pPlusMinus;
                    final int[] pMinusPlus;
                    if (pPlusIsSingleton) {
                        pPlusMinus = pPlus;
                        pMinusPlus = pMinus;
                    } else {
                        pPlusMinus = pMinus;
                        pMinusPlus = pPlus;
                    }

                    final int k = pPlusMinus[0];
                    final int chk = checkedABS(incidence.get(h, k));

                    for (int cur = 0; cur < pMinusPlus.length && pMinusPlus[cur] != -1; ++cur) {
                        
                        final int j = pMinusPlus[cur];
                        final int chj = checkedABS(incidence.get(h, j));
                        
                        for (int i = 0; i < incidence.getRows(); ++i) {
                            incidence.set(i, j, checkedADD(checkedMUL(chj, incidence.get(i, k)), checkedMUL(chk, incidence.get(i, j))));
                        }
                        
                        ops.add(new MatrixOperation(chj, k, chk, j));
                    }

                    incidence.killColumn(k);
                    
                    ops.add(new MatrixOperation(k));
                    
                    
                } else if (pPlus[0] != -1 || pMinus[0] != -1) {
                    nonZeroRow = h;
                    
                }

            } 

            if (!cardinalityUsed && nonZeroRow != -1) { // case 1.1.c
                final int h = nonZeroRow;
                int k = 0;
                while (!incidence.columnExists(k) || incidence.get(h, k) == 0) {
                    ++k;
                }
                final int chk = incidence.get(h, k);

                for (int j = 0; j < incidence.getColumns(); ++j) {
                    if (j == k || !incidence.columnExists(j)) {
                        continue;
                    }
                    final int chj = incidence.get(h, j);
                    if(chj == 0) {
                        continue;
                    }

                    final int alpha = (signum(chj)*signum(chk) < 0) ? checkedABS(chj) : -checkedABS(chj);
                    final int beta = checkedABS(chk);

                    for (int i = 0; i < incidence.getRows(); ++i) {
                        incidence.set(i, j, checkedADD(checkedMUL(alpha, incidence.get(i, k)), checkedMUL(beta, incidence.get(i, j))));
                    }
                    
                    ops.add(new MatrixOperation(alpha, k, beta, j));
                }

                incidence.killColumn(k);
                ops.add(new MatrixOperation(k));
                
            }

        } while(nonZeroRow != -1);
        
        return ops;
    }

    
    private static void getIndices(final IntMatrix A, final int rowNo, final int[] positivesArray,
            final int[] negativesArray) {
        final int n = A.getColumns();
        int negcount = 0;
        int poscount = 0;

        for (int i = 0; i < n; i++) {
            if(A.columnExists(i)) {
                final int val = A.get(rowNo, i);
                if (val < 0) {
                    negativesArray[negcount++] = i;
                } else if (val > 0) {
                    positivesArray[poscount++] = i;
                }
            }
        }
        if (negcount < n) {
            negativesArray[negcount] = -1;
        }
        if (poscount < n) {
            positivesArray[poscount] = -1;
        }

    }
    
//    private static void ensureSemipositive(final PNMatrix matrix) {
//        ensureSemipositive(matrix, false, -1, null);
//    }
    
    private static void ensureSemipositive(final IntMatrix matrix, final boolean correctSinglesOnly, final int count, final Place[] indices) {
        if(count == 0) {
            return;
        }
        
        int[] pPlus = new int[matrix.getColumns()];
        int[] pMinus = new int[matrix.getColumns()];
        int nextH = findNextH(matrix, correctSinglesOnly, indices);
        int ctr = 0;
        
        while (nextH != -1 && (count < 0 || ctr++ < count)) {
            final int h = nextH;
            nextH = -1;
            
            pPlus = ensureSize(pPlus, matrix.getColumns());
            pMinus = ensureSize(pMinus, matrix.getColumns());
            getIndices(matrix, h, pPlus, pMinus);
            
            for (int curj = 0; curj < pPlus.length && pPlus[curj] != -1; ++curj) {
                for (int curk = 0; curk < pMinus.length && pMinus[curk] != -1; ++curk) {

                    final int j = pPlus[curj];
                    final int k = pMinus[curk];

                    final int a = checkedNEG(matrix.get(h, k));
                    final int b = matrix.get(h, j);

                    final int newColumn = matrix.addColumn();
                    for (int i = 0; i < matrix.getRows(); ++i) {
                        final int v = checkedADD(checkedMUL(a, matrix.get(i, j)), checkedMUL(b, matrix.get(i, k)));
                        matrix.set(i, newColumn, v);
                        if(!correctSinglesOnly && v < 0) {
                            nextH = i;
                        }
                    }

                    int gcd = gcd(matrix, newColumn);

                    // divide all the col elements by their gcd if gcd > 1
                    if (gcd > 1) {
                        matrix.divide(newColumn, gcd);
                    }
                }
            }

            for (int i = 0; i < pMinus.length && pMinus[i] != 0; ++i) {
                matrix.killColumn(pMinus[i] - 1);
            }
            
            if(nextH == -1) {
                nextH = findNextH(matrix, correctSinglesOnly, indices);
            }
        }
    }
    
    private static int findNextH(final IntMatrix matrix, final boolean correctSinglesOnly, final Place[] indices) {
        for (int j = 0; j < matrix.getColumns(); j++) {
            if(matrix.columnExists(j)) {
                int nextH = -1;
                for (int i = 0; i < matrix.getRows(); i++) {            
                    if (matrix.get(i, j) < 0 && (!correctSinglesOnly || !indices[j].boundKnown())) {
                        if(correctSinglesOnly) {
                            if(nextH != -1) {
                                nextH = -1;
                                break;
                            }
                            nextH = i;
                        } else {
                            return i;
                        }
                    }
                }
                if(nextH != -1) {
                    return nextH;
                }
            }
        }
        return -1;
    }
    
    private static int[] ensureSize(final int[] a, final int size) {
        // the content of the array is irrelevant!
        return a.length >= size ? a : new int[size+10];
    }
    
    private static int gcd(final IntMatrix A, final int column) {
        int gcd = A.get(0, column);
        for (int i = 1; i < A.getRows(); i++) {
            if (A.get(i, column) != 0 || gcd != 0) {
                gcd = gcd2(gcd, A.get(i, column));
            }
        }
        return gcd;
    }

    private static int gcd2(final int valueA, final int valueB) {
        int a = Math.abs(valueA);
        int b = Math.abs(valueB);

        // ensure b > a
        if (b <= a) {
            final int tmp = b;
            b = a;
            a = tmp;
        }

        if (a != 0) {
            for (int tmp; (b %= a) != 0;) {
                tmp = b;
                b = a;
                a = tmp;
            }
            return a;
        } else if (b != 0) {
            return b;
        } else {
            return 0;
        }
    }
    
    
    private static void removeRedundant(final IntMatrix matrix) {
        int[] nonmin = matrix.findNonMinimal();
        for(int i = 0; i < nonmin.length && nonmin[i] != -1; ++i) {
            matrix.killColumn(nonmin[i]);
        }
    }
    
}