/*
 * Project:  Petruchio
 * Filename: ListIgnorer.java
 * Created:  25.09.2007
 * Author:   Tim Strazny
 */
package petruchio.pn;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReductorIgnorer;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;

/**
 * Ignores a list of places and/or transitions. 
 *
 * @author Tim Strazny
 */
public class ListIgnorer implements SelfDescribing, PetriNetReductorIgnorer {
	
	private static final int EOF = -1;
	
	private InputStream in = null;
	
	private Collection<String> names = null;
	
	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.SelfDescribing#getDescription()
	 */
	@Override
    public String getDescription() {
	    return "Parses a comma separated list of Petri net elements to be ignored by the reductors.";
	}
	

	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductorIgnore#setInput(java.io.InputStream)
	 */
	@Override
    public void setInput(final InputStream in) {
		this.reset();
		this.in = in;
	}


	/* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		this.names = null;
		if(this.in != null) {
			try {
				this.in.close();
			} catch(final IOException e) {
				// ignore
			}
			this.in = null;
		}
	}
	
	
	/**
	 * The input stream parsing for place/transition names.
	 */
	private void parse() {
		this.names = new HashSet<String>();
		
		try {
			final Reader reader = new BufferedReader(new InputStreamReader(this.in));
			final StringBuilder name = new StringBuilder();
			
			int ch = reader.read();
			while(ch != ListIgnorer.EOF) {
				if(ch == ',') {
					this.names.add(name.toString());
					name.delete(0, name.length());
					
				} else if(!Character.isWhitespace(ch)) {
					name.append((char)ch);
					
				}
				
				ch = reader.read();
			}
			
			this.names.add(name.toString());
			
//			System.out.println(this.names);
			
		} catch(final IOException e) {
			throw new RuntimeException("Unable to parse the scope of the LTL formula.", e);
			
		} finally {
			if(this.in != null) {
				try {
					this.in.close();
				} catch(final IOException e) {
					// ignore
				}
			}
		}
	}
	

	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductorIgnore#getPlacesToIgnore(petruchio.interfaces.petrinet.PetriNet)
	 */
	@Override
    public Collection<Place> getPlacesToIgnore(final PetriNet net) {
		if(this.in == null) {
			return new ArrayList<Place>();
		}
		if(this.names == null) {
			this.parse();
		}
		return net.getPlaces(this.names);
	}


	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductorIgnore#getTransitionsToIgnore(petruchio.interfaces.petrinet.PetriNet)
	 */
	@Override
    public Collection<Transition> getTransitionsToIgnore(final PetriNet net) {
		if(this.in == null) {
			return new ArrayList<Transition>();
		}
		if(this.names == null) {
			this.parse();
		}
		return net.getTransitions(this.names);
	}
	
}