/*
 * Project:  Petruchio
 * Filename: PetriNet.java
 * Created:  04.12.2006
 * Author:   Tim Strazny
 */
package petruchio.pn;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;
import petruchio.interfaces.petrinet.ExtendedTransition;

/**
 * A simple Petri net model.
 * Extend this class and implement {@link #writeNet(PrintWriter)} to write the
 * net in a specified format.
 * 
 * @author Tim Strazny
 */
public class PetriNet implements SelfDescribing, petruchio.interfaces.petrinet.PetriNet {

    /** All the net's places */
    private final Collection<Place> places;
    
    /** All the net's transitions */
    private final Collection<Transition> transitions;
    
    private int ptArcCount;
    
    private int tpArcCount;
    
    private String name;
    
    /** The header comment of the net */
    private String headerComment;
    
    
    /**
     * Initialize the new instance by creating the needed collections.
     */
    public PetriNet() {
        this.places = new IdentityHashSet<Place>(0);
        this.transitions = new IdentityHashSet<Transition>(0);
        this.headerComment = null;
        this.ptArcCount = 0;
        this.tpArcCount = 0;
        this.name = "unnamed";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A simple, yet fast implementation of the petruchio.petrinet."
            + "PetriNet interface and all associated classes.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#addArc(petruchio.interfaces.petrinet.Place, petruchio.interfaces.petrinet.Transition, int)
     */
    @Override
    public PTArc addArc(final Place p, final Transition t, final int weight) {
        if(weight < 0) {
            throw new IllegalArgumentException("Illegal arc weight.");
        }
        if(weight == 0) {
            return null;
        }
        PTArc existing = null;
        for(final PTArc arc : p.getOutput()) {
            if(arc.getTarget() == t && !(arc instanceof InhibitorArc)) {
                existing = arc;
                break;
            }
        }
        if(existing != null) {
            ((petruchio.pn.PTArc)existing).addWeight(weight);
            return existing;
        }
        ++this.ptArcCount;
        final PTArc arc = new petruchio.pn.PTArc(p, t, weight);
        ((petruchio.pn.Place)p).addOutput(arc);
        ((petruchio.pn.Transition)t).addInput(arc);
        return arc;
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#addArc(petruchio.interfaces.petrinet.Transition, petruchio.interfaces.petrinet.Place, int)
     */
    @Override
    public TPArc addArc(final Transition t, final Place p, final int weight) {
        if(weight < 0) {
            throw new IllegalArgumentException("Illegal arc weight.");
        }
        if(weight == 0) {
            return null;
        }
        TPArc existing = null;
        for(final TPArc arc : t.getOutput()) {
            if(arc.getTarget() == p) {
                existing = arc;
                break;
            }
        }
        if(existing != null) {
            ((petruchio.pn.TPArc)existing).addWeight(weight);
            return existing;
        }
        ++this.tpArcCount;
        final TPArc arc = new petruchio.pn.TPArc(t, p, weight);
        ((petruchio.pn.Place)p).addInput(arc);
        ((petruchio.pn.Transition)t).addOutput(arc);
        return arc;
    }
    
    
    public ResetArc addResetArc(final Place p, final ExtendedTransition t) {
        return this.addResetArc(p, t, 0);
    }
    
    public ResetArc addResetArc(final Place p, final ExtendedTransition t, final int weight) {
        PTArc existing = null;
        for(final PTArc arc : p.getOutput()) {
            if(arc.getTarget() == t && !(arc instanceof InhibitorArc)) {
                existing = arc;
                break;
            }
        }
        if(!(existing instanceof petruchio.pn.ResetArc)) {
            ((petruchio.pn.Place)p).removeOutput(existing);
            ((petruchio.pn.Transition)t).removeInput(existing);
            existing = null;
        }
        if(existing != null) {
            ((petruchio.pn.ResetArc)existing).addWeight(weight);
            return (petruchio.pn.ResetArc)existing;
        }
        ++this.ptArcCount;
        final ResetArc arc = new petruchio.pn.ResetArc(p, t, weight);
        ((petruchio.pn.Place)p).addOutput(arc);
        ((petruchio.pn.Transition)t).addInput(arc);
        return arc;
    }
    
    public InhibitorArc addInhibitorArc(final Place p, final ExtendedTransition t) {
        final int weight = 0;
        PTArc existing = null;
        for(final PTArc arc : p.getOutput()) {
            if(arc.getTarget() == t && arc instanceof InhibitorArc) {
                existing = arc;
                break;
            }
        }
        if(existing != null) {
            ((petruchio.pn.InhibitorArc)existing).addWeight(weight);
            return (petruchio.pn.InhibitorArc)existing;
        }
        ++this.ptArcCount;
        final InhibitorArc arc = new petruchio.pn.InhibitorArc(p, t, weight);
        ((petruchio.pn.Place)p).addOutput(arc);
        ((petruchio.pn.Transition)t).addInput(arc);
        return arc;
    }
    
    public TransferArc addTranserArc(final ExtendedTransition t, final Place p) {
        return this.addTransferArc(t, p, 0);
    }
    
    public TransferArc addTransferArc(final ExtendedTransition t, final Place p, final int weight) {
        TPArc existing = null;
        for(final TPArc arc : p.getInput()) {
            if(arc.getTarget() == t) {
                existing = arc;
                break;
            }
        }
        if(!(existing instanceof petruchio.pn.TransferArc)) {
            ((petruchio.pn.Place)p).removeInput(existing);
            ((petruchio.pn.Transition)t).removeOutput(existing);
            existing = null;
        }
        if(existing != null) {
            ((petruchio.pn.TransferArc)existing).addWeight(weight);
            return (petruchio.pn.TransferArc)existing;
        }
        ++this.ptArcCount;
        final TransferArc arc = new petruchio.pn.TransferArc(t, p, weight);
        ((petruchio.pn.Place)p).addInput(arc);
        ((petruchio.pn.Transition)t).addOutput(arc);
        return arc;
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#addPlace(java.lang.String, int)
     */
    @Override
    public Place addPlace(final String pname, final int marking) {
        final Place place = new petruchio.pn.Place(pname, marking);
        this.places.add(place);
        return place;
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#addTransition(java.lang.String)
     */
    @Override
    public Transition addTransition(final String tname) {
        final Transition trans = new petruchio.pn.Transition(tname);
        this.transitions.add(trans);
        return trans;
    }
    
    
    public ExtendedTransition addExtendedTransition(final String tname) {
        final ExtendedTransition trans = new petruchio.pn.ExtendedTransition(tname);
        this.transitions.add(trans);
        return trans;
    }

    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getPlaces()
     */
    @Override
    public Collection<Place> getPlaces() {
        return this.places;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getTransitions()
     */
    @Override
    public Collection<Transition> getTransitions() {
        return this.transitions;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#containsTransitionBetween(petruchio.interfaces.petrinet.Place, petruchio.interfaces.petrinet.Place)
     */
    public boolean containsTransitionBetween(final Place a, final Place b) {
        for(final PTArc out : ((petruchio.pn.Place)a).getOutput()) {
            final Transition t = out.getTarget();
            if(PetriNet.containsArcFrom(((petruchio.pn.Place)b).getInput(), t)) {
                return true;
            }
        }
        return false;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getTransitionsBetween(java.util.Map, java.util.Map)
     */
    @Override
    public Collection<Transition> getTransitionsBetween(final Map<Place, Integer> in,
                    final Map<Place, Integer> out) {
        Collection<Transition> trans = null;
        for(final Map.Entry<Place, Integer> e : in.entrySet()) {
            final Place p = e.getKey();
            final int count = e.getValue();
            if(trans == null) {
                trans = new LinkedList<Transition>();
                for(final PTArc arc : p.getOutput()) {
                    if(arc.getWeight() == count) {
                        trans.add(arc.getTarget());
                    }
                }
            } else {
                for(final Iterator<Transition> it = trans.iterator(); it.hasNext(); ) {
                    final Transition t = it.next();
                    boolean found = false;
                    for(final PTArc arc : p.getOutput()) {
                        if((arc.getTarget() == t) && (arc.getWeight() == count)) {
                            found = true;
                            break;
                        }
                    }
                    if(!found) {
                        it.remove();
                    }
                }
            }
            if(trans.isEmpty()) {
                return trans;
            }
        }
        for(final Map.Entry<Place, Integer> e : out.entrySet()) {
            final Place p = e.getKey();
            final int count = e.getValue();
            if(trans == null) {
                trans = new LinkedList<Transition>();
                for(final TPArc arc : p.getInput()) {
                    if(arc.getWeight() == count) {
                        trans.add(arc.getSource());
                    }
                }
            } else {
                for(final Iterator<Transition> it = trans.iterator(); it.hasNext(); ) {
                    final Transition t = it.next();
                    boolean found = false;
                    for(final TPArc arc : p.getInput()) {
                        if((arc.getSource() == t) && (arc.getWeight() == count)) {
                            found = true;
                            break;
                        }
                    }
                    if(!found) {
                        it.remove();
                    }
                }
            }
            if(trans.isEmpty()) {
                return trans;
            }
        }
        return trans;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#setHeaderComment(java.lang.String)
     */
    @Override
    public void setHeaderComment(final String header) {
        this.headerComment = header == null ? null : header.trim();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#hasHeaderComment()
     */
    @Override
    public boolean hasHeaderComment() {
        return (this.headerComment != null) && !this.headerComment.isEmpty();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getHeaderComment()
     */
    @Override
    public String getHeaderComment() {
        return this.headerComment;
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#reset()
     */
    @Override
    public void reset() {
        this.headerComment = null;
        for(final Place p : this.places) {
            ((petruchio.pn.Place)p).clear();
        }
        for(final Transition t : this.transitions) {
            ((petruchio.pn.Transition)t).clear();
        }
        this.places.clear();
        this.transitions.clear();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#removeArc(petruchio.interfaces.petrinet.PTArc)
     */
    @Override
    public boolean removeArc(final PTArc arc) {
    	boolean changed = false;
        if(PetriNet.removePTArc(arc.getSource().getOutput(), arc)) {
        	changed = true;
        }
        if(PetriNet.removePTArc(arc.getTarget().getInput(), arc)) {
        	changed = true;
        }
        if(changed) {
        	--this.ptArcCount;
        }
        return changed; 
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#removeArc(petruchio.interfaces.petrinet.TPArc)
     */
    @Override
    public boolean removeArc(final TPArc arc) {
    	boolean changed = false;
        if(PetriNet.removeTPArc(arc.getSource().getOutput(), arc)) {
        	changed = true;
        }
        if(PetriNet.removeTPArc(arc.getTarget().getInput(), arc)) {
        	changed = true;
        }
        if(changed) {
        	--this.tpArcCount;
        }
        return changed;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#removeArc(petruchio.interfaces.petrinet.Place, petruchio.interfaces.petrinet.Transition)
     */
    @Override
    public boolean removeArc(final Place p, final Transition t) {
        final PTArc arc = new petruchio.pn.PTArc(p, t, 1);
        return this.removeArc(arc);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#removeArc(petruchio.interfaces.petrinet.Transition, petruchio.interfaces.petrinet.Place)
     */
    @Override
    public boolean removeArc(final Transition t, final Place p) {
        final TPArc arc = new petruchio.pn.TPArc(t, p, 1);
        return this.removeArc(arc);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#removePlace(petruchio.interfaces.petrinet.Place)
     */
    @Override
    public boolean removePlace(final Place p) {
        if(!p.getInput().isEmpty()) {
            final Collection<TPArc> tr = new ArrayList<TPArc>(p.getInput());
            p.getInput().clear();
            for(final TPArc arc : tr) {
                this.removeArc(arc);
            }
        }
        
        if(!p.getOutput().isEmpty()) {
            final Collection<PTArc> tr = new ArrayList<PTArc>(p.getOutput());
            p.getOutput().clear();
            for(final PTArc arc : tr) {
                this.removeArc(arc);
            }
        }

        ((petruchio.pn.Place)p).clear();
        return this.places.remove(p); 
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#removeTransition(petruchio.interfaces.petrinet.Transition)
     */
    @Override
    public boolean removeTransition(final Transition t) {
    	if(!t.getInput().isEmpty()) {
            final Collection<PTArc> pl = new ArrayList<PTArc>(t.getInput());
            t.getInput().clear();
            for(final PTArc arc : pl) {
                this.removeArc(arc);
            }
        }
    	
    	if(!t.getOutput().isEmpty()) {
            final Collection<TPArc> pl = new ArrayList<TPArc>(t.getOutput());
            t.getOutput().clear();
            for(final TPArc arc : pl) {
                this.removeArc(arc);
            }
        }
    	
        ((petruchio.pn.Transition)t).clear();
        return this.transitions.remove(t);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getName()
     */
    @Override
    public String getName() {
        return this.name;
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#setName(java.lang.String)
     */
    @Override
    public void setName(String name) {
        this.name = name;
    }

    
    /**
     * TODO: document
     *
     * @param arcs
     * @param arc
     * @return
     */
    private static boolean removePTArc(final Collection<PTArc> arcs, final PTArc arc) {
        for(final Iterator<PTArc> it = arcs.iterator(); it.hasNext();) {
            final PTArc cur = it.next();
            if((cur.getSource() == arc.getSource()) && (cur.getTarget() == arc.getTarget())) {
                it.remove();
                return true;
            }
        }
        return false;
    }
    
    
    /**
     * TODO: document
     *
     * @param arcs
     * @param arc
     * @return
     */
    private static boolean removeTPArc(final Collection<TPArc> arcs, final TPArc arc) {
        for(final Iterator<TPArc> it = arcs.iterator(); it.hasNext();) {
            final TPArc cur = it.next();
            if((cur.getSource() == arc.getSource()) && (cur.getTarget() == arc.getTarget())) {
                it.remove();
                return true;
            }
        }
        return false;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getPTArcCount()
     */
    @Override
    public int getPTArcCount() {
        return this.ptArcCount;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getTPArcCount()
     */
    @Override
    public int getTPArcCount() {
        return this.tpArcCount;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getPlaces(java.util.Collection)
     */
    @Override
    public Collection<Place> getPlaces(final Collection<String> names) {
    	final Collection<Place> result = new ArrayList<Place>();
    	for(final Place p : this.places) {
    		if(names.contains(p.getName())) {
    			result.add(p);
    		}
    	}
    	return result;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getPlaces(java.lang.String)
     */
    @Override
    public Collection<Place> getPlaces(final String thename) {
    	final Collection<Place> result = new ArrayList<Place>();
    	for(final Place p : this.places) {
    		if(thename.equals(p.getName())) {
    			result.add(p);
    		}
    	}
    	return result;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getTransitions(java.util.Collection)
     */
    @Override
    public Collection<Transition> getTransitions(final Collection<String> names) {
    	final Collection<Transition> result = new ArrayList<Transition>();
    	for(final Transition t : this.transitions) {
    		if(names.contains(t.getName())) {
    			result.add(t);
    		}
    	}
    	return result;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNet#getTransitions(java.lang.String)
     */
    @Override
    public Collection<Transition> getTransitions(final String thename) {
    	final Collection<Transition> result = new ArrayList<Transition>();
    	for(final Transition t : this.transitions) {
    		if(thename.equals(t.getName())) {
    			result.add(t);
    		}
    	}
    	return result;
    }
    
    
    /**
     * TODO: document
     *
     * @param arcs
     * @param t
     * @return
     */
    private static boolean containsArcFrom(final Collection<TPArc> arcs, final Transition t) {
        for(final TPArc arc : arcs) {
            if(arc.getSource() == t) {
                return true;
            }
        }
        return false;
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        final StringBuilder s = new StringBuilder();
        for(final Place p : this.places) {
            s.append(p);
            s.append(", ");
        }
        s.append('\n');
        for(final Transition t : this.transitions) {
            s.append(t);
            s.append(", ");
        }
        s.append('\n');
        for(final Place p : this.places) {
            for(final PTArc a : p.getOutput()) {
                s.append(a);
                s.append(", ");
            }
            for(final TPArc a : p.getInput()) {
                s.append(a);
                s.append(", ");
            }
        }
        return s.toString();
    }

}