/*
 * Project:  Petruchio
 * Filename: TransferArc.java
 * Created:  18.03.2010
 * Author:   Tim Strazny
 */
package petruchio.pn;

import java.util.Collection;
import java.util.Collections;
import java.util.IdentityHashMap;
import java.util.Map;

import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;

public class TransferArc extends TPArc implements
        petruchio.interfaces.petrinet.TransferArc {

    private final Map<Place, Integer> multipliers = new IdentityHashMap<Place, Integer>();
    private final Map<Place, Integer> multipliers_unmod = Collections.unmodifiableMap(this.multipliers);
    private int bound = 0;
    
    /**
     * Initializes an empty transfer arc with an additional constant.
     * 
     * @param t The source transition.
     * @param p The target place.
     * @param weight Additional constant to the transfer arc.
     */
    public TransferArc(final Transition t, final Place p, final int weight) {
        super(t, p, weight);
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.TransferArc#setMultiplier(petruchio.interfaces.petrinet.Place, int)
     */
    @Override
    public void setMultiplier(final Place p, final int m) {
        if(m == 0) {
            this.removeMultiplier(p);
        } else {
            this.multipliers.put(p, Integer.valueOf(m));
        }
    }
    
    public void setMultiplier(final Place p, final Integer m) {
        if(m.intValue() == 0) {
            this.removeMultiplier(p);
        } else {
            this.multipliers.put(p, m);
        }
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.TransferArc#addMultiplier(petruchio.interfaces.petrinet.Place, int)
     */
    @Override
    public void addMultiplier(final Place p, final int m) {
        final Integer i = this.multipliers.put(p, Integer.valueOf(m));
        if(i != null) {
            final int m2 = i.intValue() + m;
            if(m2 == 0) {
                this.removeMultiplier(p);
            } else if(i.intValue() != m2) {
                this.multipliers.put(p, Integer.valueOf(m2));
            }
        }
    }
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.TransferArc#getMultiplier(petruchio.interfaces.petrinet.Place)
     */
    @Override
    public int getMultiplier(final Place p) {
        final Integer m = this.multipliers.get(p);
        return m == null ? 0 : m.intValue();
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.TransferArc#getTransfer()
     */
    @Override
    public Map<Place, Integer> getTransfer() {
        return this.multipliers_unmod;
    }
    
    /* (non-Javadoc)
     * @see petruchio.pn.TPArc#toString()
     */
    @Override
    public String toString() {
        final StringBuilder s = new StringBuilder();
        s.append(this.getSource());
        s.append(" =");
        for(final Map.Entry<Place, Integer> e : this.multipliers.entrySet()) {
            s.append(e.getValue());
            s.append('*');
            s.append(e.getKey());
            s.append(" + ");
        }
        s.append(this.getWeight());
        s.append("=> ");
        s.append(this.getTarget());
        return s.toString();
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.TransferArc#setMultipliers(java.util.Map)
     */
    @Override
    public void setMultipliers(final Map<Place, Integer> m) {
        for(final Map.Entry<Place, Integer> e : m.entrySet()) {
            this.setMultiplier(e.getKey(), e.getValue());
        }
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.TransferArc#removeMultipliers(java.util.Collection)
     */
    @Override
    public void removeMultipliers(final Collection<Place> p) {
        for(final Place place : p) {
            this.removeMultiplier(place);
        }
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.TransferArc#removeMultiplier(petruchio.interfaces.petrinet.Place)
     */
    @Override
    public void removeMultiplier(final Place p) {
        this.multipliers.remove(p);
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.TransferArc#getBound()
     */
    @Override
    public int getBound() {
        return this.bound;
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.TransferArc#setBound(int)
     */
    @Override
    public void setBound(final int bound) {
        this.bound = bound;
    }

}
