/*
 * Project:  Petruchio
 * Filename: INAReader.java
 * Created:  05.09.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.readers;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Scanner;

import petruchio.interfaces.LocalConfigurationParser;
import petruchio.interfaces.ParserErrorHandler;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReader;
import petruchio.interfaces.petrinet.PetriNetStreamReader;
import petruchio.interfaces.petrinet.PetriNetStreamWriter;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;
import petruchio.pi.readers.LocalConfigParser;


/**
 * Reads a low level INA Petri net file and builds the corresponding Petri net.
 * 
 * @author Tim Strazny
 */
public class INAReader implements SelfDescribing, PetriNetReader,
        PetriNetStreamReader, LocalConfigurationParser {
    
    private InputStream in = null;
    private Collection<ParserErrorHandler> errorHandlers = new ArrayList<ParserErrorHandler>();
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net reader for files in the format of the INA tool.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#getDefaultExtension()
     */
    @Override
    public String getDefaultExtensions() {
        return "pnt";
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        this.errorHandlers.clear();
        if (this.in != null) {
            try {
                this.in.close();
            } catch (final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#read(java.io.InputStream)
     */
    @Override
    public PetriNet read() {
        final PetriNet net = new petruchio.pn.PetriNet();
        this.read(net);
        return net;
    }
        
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#read(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public void read(final PetriNet net) {
        try {
            final Scanner s = new Scanner(this.in).useDelimiter("[\\s&&[^\n\r\u2028\u2029\u0085]]+");
            
            final Map<String, Place> pl = new HashMap<String, Place>();
            final Map<String, Transition> tr = new HashMap<String, Transition>();
            
            this.readStructure(s, net, pl, tr);
            
            this.readPlaces(s, net, pl);
            
            this.readTransitions(s, net, tr);
            
            this.readComment(s, net);
            
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
        } finally {
            try {
                this.in.close();
            } catch(final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param e
     */
    private void reportError(final Exception e) {
        for(final ParserErrorHandler handler : this.errorHandlers) {
            handler.reportError(e.getLocalizedMessage());
        }
        if(this.errorHandlers.isEmpty()) {
            e.printStackTrace();
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param net
     * @param pl
     * @param tr
     */
    private void readStructure(final Scanner s, final PetriNet net,
                    final Map<String, Place> pl, final Map<String, Transition> tr) {
        s.skip("(?i)\\s*P\\s+M\\s+PRE\\s*,\\s*POST\\s+NETZ\\s+\\d+\\s*:\\s*");
        net.setName(s.next().trim());
        
        while(!s.hasNext("(?s)\\s*@.*")) {
            s.skip("\\s*(\\S+)");
            final String pNr = s.match().group(1);
//            System.out.println("p: " + pNr);
            
            s.skip("\\s*(\\d+)\\s*");
            final int marking = Integer.parseInt(s.match().group(1));
//            System.out.println("  " + marking);
            
            Place p = pl.get(pNr);
            if(p == null) {
                p = net.addPlace(pNr, marking);
                pl.put(pNr, p);
            } else {
                p.setMarking(marking);
            }
            
            while(!s.hasNext("(?ms)\\s*(?:,|$).*")) {
                s.skip("\\s*(\\w+)");
                final String tNr = s.match().group(1);
                final int weight;
                if(s.hasNext("\\s*:")) {
                    s.skip("\\s*:\\s*(\\d+)");
                    weight = Integer.parseInt(s.match().group(1));
                } else {
                    weight = 1;
                }
                
//                System.out.println("  in " + tNr + ":" + weight);
                
                Transition t = tr.get(tNr);
                if(t == null) {
                    t = net.addTransition(tNr);
                    tr.put(tNr, t);
                }
                
                net.addArc(t, p, weight);
            }
            if(s.hasNext("(?s)\\s*,.*")) {
                s.skip("\\s*,");
                while(!s.hasNext("(?ms)\\s*$.*")) {
                    s.skip("\\s*(\\w+)");
                    final String tNr = s.match().group(1);
                    final int weight;
                    if(s.hasNext("\\s*:")) {
                        s.skip("\\s*:\\s*(\\d+)");
                        weight = Integer.parseInt(s.match().group(1));
                    } else {
                        weight = 1;
                    }
                    
//                    System.out.println("  out " + tNr + ":" + weight);
                    
                    Transition t = tr.get(tNr);
                    if(t == null) {
                        t = net.addTransition(tNr);
                        tr.put(tNr, t);
                    }
                    
                    net.addArc(p, t, weight);
                }
            }
        }
        s.skip("\\s*@");
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param net
     * @param pl
     */
    private void readPlaces(final Scanner s, final PetriNet net,
                    final Map<String, Place> pl) {
        s.skip("(?i)\\s*place\\s+nr\\.\\s*name\\s+capacity\\s+time\\s*");
        while(!s.hasNext("(?s)\\s*@.*")) {
            s.skip("\\s*(\\w+)");
            final String pNr = s.match().group(1);
            s.skip("\\s*:\\s*(\\S+)");
            final String name = s.match().group(1);
            Place p = pl.get(pNr);
            if(p == null) {
                p = net.addPlace(name, 0);
                pl.put(pNr, p);
            } else {
                p.setName(name);
            }
//            System.out.println(pNr + " = " + name);
            s.skip("(?m).*$");
        }
        s.skip("\\s*@");
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param net
     * @param tr
     */
    private void readTransitions(final Scanner s, final PetriNet net,
                    final Map<String, Transition> tr) {
        s.skip("(?i)\\s*trans\\s+nr\\.\\s*name\\s+priority\\s+time\\s*");
        while(!s.hasNext("(?s)\\s*@.*")) {
            s.skip("\\s*(\\w+)");
            final String tNr = s.match().group(1);
            s.skip("\\s*:\\s*(\\S+)");
            final String name = s.match().group(1);
            Transition t = tr.get(tNr);
            if(t == null) {
                t = net.addTransition(name);
                tr.put(tNr, t);
            } else {
                t.setName(name);
            }
//            System.out.println(tNr + " = " + name);
            s.skip("(?m).*$");
        }
        s.skip("\\s*@");
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param net
     */
    private void readComment(final Scanner s, final PetriNet net) {
        net.setHeaderComment(s.skip("(?s)\\s*(.*)\\s*").match().group(1));
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#setInput(java.io.InputStream)
     */
    @Override
    public void setInput(final InputStream in) {
        this.in = in;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#addErrorHandler(ParserErrorHandler)
     */
    @Override
    public void addErrorHandler(final ParserErrorHandler handler) {
        this.errorHandlers.add(handler);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.LocalConfigurationParser#getLocalConfiguration()
     */
    @Override
    public Properties getLocalConfiguration() {
        try {
            final Scanner s = new Scanner(this.in);
            final String cmt = s.skip("(?s)\\s*(.*)\\s*").match().group(1);
            final LocalConfigurationParser p = new LocalConfigParser();
            p.setInput(new ByteArrayInputStream(cmt.getBytes()));
            final Properties props = p.getLocalConfiguration();
            p.reset();
            return props;
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamConverter#convertTo(petruchio.interfaces.petrinet.PetriNetStreamWriter)
     */
    @Override
    public void readTo(final PetriNetStreamWriter writer) {
        try {
            final Scanner s = new Scanner(this.in).useDelimiter("[\\s&&[^\n\r\u2028\u2029\u0085]]+");
            
            final Map<String, Place> pl = new HashMap<String, Place>();
            
            s.skip("(?i)\\s*P\\s+M\\s+PRE\\s*,\\s*POST\\s+NETZ\\s+\\d+\\s*:\\s*");
            
            writer.beginNet(s.next("[^\\s]+"));
            
            while(!s.hasNext("(?s)\\s*@.*")) {
                s.skip("\\s*(\\S+)");
                final String pNr = "p" + s.match().group(1);
//                System.out.println("p: " + pNr);
                
                s.skip("\\s*(\\d+)\\s*");
                final int marking = Integer.parseInt(s.match().group(1));
//                System.out.println("  " + marking);
                
                Place p = pl.get(pNr);
                if(p == null) {
                    p = new petruchio.pn.Place(pNr, marking);
                    pl.put(pNr, p);
                } else {
                    p.setMarking(marking);
                }
                
                while(!s.hasNext("(?ms)\\s*(?:,|$).*")) {
                    s.skip("\\s*(\\w+)");
                    final String tNr = "t" + s.match().group(1);
                    final int weight;
                    if(s.hasNext("\\s*:")) {
                        s.skip("\\s*:\\s*(\\d+)");
                        weight = Integer.parseInt(s.match().group(1));
                    } else {
                        weight = 1;
                    }
                    
                    writer.writeTPArc(tNr, pNr, weight);
                }
                if(s.hasNext("(?s)\\s*,.*")) {
                    s.skip("\\s*,");
                    while(!s.hasNext("(?ms)\\s*$.*")) {
                        s.skip("\\s*(\\w+)");
                        final String tNr = "t" + s.match().group(1);
                        final int weight;
                        if(s.hasNext("\\s*:")) {
                            s.skip("\\s*:\\s*(\\d+)");
                            weight = Integer.parseInt(s.match().group(1));
                        } else {
                            weight = 1;
                        }
                        
                        writer.writePTArc(pNr, tNr, weight);
                    }
                }
            }
            s.skip("\\s*@");
            
            
            
            s.skip("(?i)\\s*place\\s+nr\\.\\s*name\\s+capacity\\s+time\\s*");
            while(!s.hasNext("(?s)\\s*@.*")) {
                s.skip("\\s*(\\w+)");
                final String pNr = "p" + s.match().group(1);
                s.skip("\\s*:\\s*(\\S+)");
                final String name = s.match().group(1);
                Place p = pl.get(pNr);
                if(p == null) {
                    p = new petruchio.pn.Place(pNr, 0);
                    pl.put(pNr, p);
                }
                p.setMeaning(name);
                writer.writePlace(p);
                
                s.skip("(?m).*$");
            }
            s.skip("\\s*@");
            
            
            s.skip("(?i)\\s*trans\\s+nr\\.\\s*name\\s+priority\\s+time\\s*");
            while(!s.hasNext("(?s)\\s*@.*")) {
                s.skip("\\s*(\\w+)");
                final String tNr = "t" + s.match().group(1);
                s.skip("\\s*:\\s*(\\S+)");
                final String name = s.match().group(1);
                final Transition t = new petruchio.pn.Transition(tNr);
                t.setMeaning(name);
                writer.writeTransition(t);
                s.skip("(?m).*$");
            }
            s.skip("\\s*@");
            
            
            writer.writeHeaderComment(s.skip("(?s)\\s*(.*)\\s*").match().group(1));
            
            writer.endNet();
            
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
        } finally {
            try {
                this.in.close();
            } catch(final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }

}