/*
 * Project: Petruchio
 * File:    LoLAReader.java
 * Author:  Tim Strazny
 * Created: 11.10.2007
 */
package petruchio.pn.readers;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Scanner;

import petruchio.interfaces.LocalConfigurationParser;
import petruchio.interfaces.ParserErrorHandler;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReader;
import petruchio.interfaces.petrinet.PetriNetStreamReader;
import petruchio.interfaces.petrinet.PetriNetStreamWriter;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;
import petruchio.pi.readers.LocalConfigParser;


/**
 * TODO: Document
 * 
 * @author Tim Strazny
 */
public class LoLAReader implements SelfDescribing, PetriNetReader,
        PetriNetStreamReader, LocalConfigurationParser {

    private InputStream in = null;
    private Collection<ParserErrorHandler> errorHandlers = new ArrayList<ParserErrorHandler>();
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net reader for files in the format of the LoLA tool.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#addErrorHandler(petruchio.interfaces.ParserErrorHandler)
     */
    @Override
    public void addErrorHandler(final ParserErrorHandler handler) {
        this.errorHandlers.add(handler);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#getDefaultExtension()
     */
    @Override
    public String getDefaultExtensions() {
        return "lola";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#read(java.io.InputStream)
     */
    @Override
    public PetriNet read() {
        final PetriNet net = new petruchio.pn.PetriNet();
        this.read(net);
        return net;
    }
        
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#read(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public void read(final PetriNet net) {
        try {
            final Scanner s = new Scanner(this.in).useDelimiter("\\s+|\\{.*?\\}|(?=[;,:])");
            
            final Map<String, Place> pl = new HashMap<String, Place>();
            
            net.setHeaderComment(this.readHeaderComment(s));
            
            if(s.hasNext("(?i)PLACE")) {
                s.next();
                while(!s.hasNext(";")) {
                    final String name = s.next();
                    final Place p = net.addPlace(name, 0);
                    pl.put(name, p);
                    while(s.hasNext(",")) {
                        s.next();
                    }
                }
                s.next();
            }
            if(s.hasNext("(?i)MARKING")) {
                s.next();
                while(!s.hasNext(";")) {
                    final String name = s.next();
                    final String n = s.next();
                    if(!n.equals(":")) {
                        throw new RuntimeException("Syntax error: expecting ':' but found '" + n + "'.");
                    }
                    final Place p = pl.get(name);
                    final int marking = s.nextInt();
                    p.setMarking(marking);
                    
                    while(s.hasNext(",")) {
                        s.next();
                    }
                }
                s.next();
            }
            
            while(s.hasNext("(?i)TRANSITION")) {
                s.next();
                final String name = s.next();
                final Transition t = net.addTransition(name);
                if(s.hasNext("(?i)CONSUME")) {
                    s.next();
                    while(!s.hasNext(";")) {
                        final String pName = s.next();
                        final String n = s.next();
                        if(!n.equals(":")) {
                            throw new RuntimeException("Syntax error: expecting ':' but found '" + n + "'.");
                        }
                        final Place p = pl.get(pName);
                        final int weight = s.nextInt();
                        net.addArc(p, t, weight);
                        
                        while(s.hasNext(",")) {
                            s.next();
                        }
                    }
                    s.next();
                }
                if(s.hasNext("(?i)PRODUCE")) {
                    s.next();
                    while(!s.hasNext(";")) {
                        final String pName = s.next();
                        final String n = s.next();
                        if(!n.equals(":")) {
                            throw new RuntimeException("Syntax error: expecting ':' but found '" + n + "'.");
                        }
                        final Place p = pl.get(pName);
                        final int weight = s.nextInt();
                        net.addArc(t, p, weight);
                        
                        while(s.hasNext(",")) {
                            s.next();
                        }
                    }
                    s.next();
                }
            }
            
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
            
        } finally {
            try {
                this.in.close();
            } catch(final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param s
     * @return
     */
    private String readHeaderComment(final Scanner s) {
        return s.skip("(?s)(\\s*\\{.*?\\})*").match().group().replaceAll("[ \t]*\\{ ?|\\s*\\}", "").trim();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#setInput(java.io.InputStream)
     */
    @Override
    public void setInput(final InputStream in) {
        this.in = in;
    }
    
    
    /**
     * TODO: description
     *
     * @param e
     */
    private void reportError(final Exception e) {
        for(final ParserErrorHandler handler : this.errorHandlers) {
            handler.reportError(e.getLocalizedMessage());
        }
        if(this.errorHandlers.isEmpty()) {
            e.printStackTrace();
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        this.errorHandlers.clear();
        if(this.in != null) {
            try {
                this.in.close();
            } catch(final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamReader#readTo(petruchio.interfaces.petrinet.PetriNetStreamWriter)
     */
    @Override
    public void readTo(final PetriNetStreamWriter writer) {
        try {
            writer.beginNet("noname");
            
            final Scanner s = new Scanner(this.in).useDelimiter("\\s+|\\{.*?\\}|(?=[:;,])");
            
            final Map<String, Place> pl = new HashMap<String, Place>();
            
            writer.writeHeaderComment(this.readHeaderComment(s));
            
            if(s.hasNext("(?i)PLACE")) {
                s.next();
                while(!s.hasNext(";")) {
                    final String name = s.next();
                    final Place p = new petruchio.pn.Place(name, 0);
                    pl.put(name, p);
                    while(s.hasNext(",")) {
                        s.next();
                    }
                }
                s.next();
            }
            if(s.hasNext("(?i)MARKING")) {
                s.next();
                while(!s.hasNext(";")) {
                    final String name = s.next();
                    final String n = s.next();
                    if(!n.equals(":")) {
                        throw new RuntimeException("Syntax error: expecting ':' but found '" + n + "'.");
                    }
                    final Place p = pl.remove(name);
                    final int marking = s.nextInt();
                    p.setMarking(marking);
                    writer.writePlace(p);
                    
                    while(s.hasNext(",")) {
                        s.next();
                    }
                }
                s.next();
            }
            for(final Place p : pl.values()) {
                writer.writePlace(p);
            }
            pl.clear();
            
            
            while(s.hasNext("(?i)TRANSITION")) {
                s.next();
                final String tName = s.next();
                final Transition t = new petruchio.pn.Transition(tName);
                writer.writeTransition(t);
                
                if(s.hasNext("(?i)CONSUME")) {
                    s.next();
                    while(!s.hasNext(";")) {
                        final String pName = s.next();
                        final String n = s.next();
                        if(!n.equals(":")) {
                            throw new RuntimeException("Syntax error: expecting ':' but found '" + n + "'.");
                        }
                        final int weight = s.nextInt();
                        writer.writePTArc(pName, tName, weight);
                        
                        while(s.hasNext(",")) {
                            s.next();
                        }
                    }
                    s.next();
                }
                if(s.hasNext("(?i)PRODUCE")) {
                    s.next();
                    while(!s.hasNext(";")) {
                        final String pName = s.next();
                        final String n = s.next();
                        if(!n.equals(":")) {
                            throw new RuntimeException("Syntax error: expecting ':' but found '" + n + "'.");
                        }
                        final int weight = s.nextInt();
                        writer.writeTPArc(tName, pName, weight);
                        
                        while(s.hasNext(",")) {
                            s.next();
                        }
                    }
                    s.next();
                }
            }
            
            writer.endNet();
            
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
            
        } finally {
            try {
                this.in.close();
            } catch(final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.LocalConfigurationParser#getLocalConfiguration()
     */
    @Override
    public Properties getLocalConfiguration() {
        try {
            final Scanner s = new Scanner(this.in);
            final String cmt = this.readHeaderComment(s);
            final LocalConfigurationParser p = new LocalConfigParser();
            p.setInput(new ByteArrayInputStream(cmt.getBytes()));
            final Properties props = p.getLocalConfiguration();
            p.reset();
            return props;
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
        }
    }

}
