/*
 * Project:  Petruchio
 * Filename: PEPReader.java
 * Created:  05.09.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.readers;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Scanner;
import java.util.regex.MatchResult;
import java.util.regex.Pattern;

import petruchio.interfaces.LocalConfigurationParser;
import petruchio.interfaces.ParserErrorHandler;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReader;
import petruchio.interfaces.petrinet.PetriNetStreamReader;
import petruchio.interfaces.petrinet.PetriNetStreamWriter;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;
import petruchio.pi.readers.LocalConfigParser;


/**
 * Reads a low level PEP Petri net file and builds the corresponding Petri net.
 * 
 * @author Tim Strazny
 */
public class PEPReader implements SelfDescribing, PetriNetReader,
        PetriNetStreamReader, LocalConfigurationParser {
    
    private static enum BlockType {
        PLACES("PL"),
        TRANSITIONS("TR"),
        PT_ARCS("PT"),
        TP_ARCS("TP"),
        DONT_CARE("..");
        
        private final Pattern pattern;
        
        /**
         * TODO: document
         * 
         * @param p
         */
        private BlockType(final String p) {
            this.pattern = Pattern.compile("(?m)\\s*" + p + "\\s*$.*");
        }

        
        /**
         * TODO: document
         *
         * @param s
         * @return
         */
        public boolean match(final Scanner s) {
            if(s.hasNext(this.pattern)) {
                s.skip(this.pattern);
                return true;
            }
            return false;
        }
    } 
    
    
    private InputStream in = null;
    private Collection<ParserErrorHandler> errorHandlers = new ArrayList<ParserErrorHandler>();
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net reader for files in the low level net format of the PEP tool.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#getDefaultExtension()
     */
    @Override
    public String getDefaultExtensions() {
        return "ll_net";
    }


    /* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
	    this.errorHandlers.clear();
		if(this.in != null) {
		    try {
		        this.in.close();
		    } catch(final IOException e) {
		        // ignore
		    }
            this.in = null;
		}
	}
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#read(java.io.InputStream)
     */
    @Override
    public PetriNet read() {
        final PetriNet net = new petruchio.pn.PetriNet();
        this.read(net);
        return net;
    }
        
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#read(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public void read(final PetriNet net) {
        final Scanner s = new Scanner(this.in).useDelimiter("\\s+");
        try {
            final Map<String, Place> pl = new HashMap<String, Place>();
            final Map<String, Transition> tr = new HashMap<String, Transition>();
    
            this.readComment(s, net);
            
            BlockType currentBlock = PEPReader.BlockType.DONT_CARE;
            do {
                s.skip("\\s*");
                
                if(PEPReader.BlockType.PLACES.match(s)) {
                    currentBlock = PEPReader.BlockType.PLACES;
                    
                } else if(PEPReader.BlockType.TRANSITIONS.match(s)) {
                    currentBlock = PEPReader.BlockType.TRANSITIONS;
                    
                } else if(PEPReader.BlockType.PT_ARCS.match(s)) {
                    currentBlock = PEPReader.BlockType.PT_ARCS;
                    
                } else if(PEPReader.BlockType.TP_ARCS.match(s)) {
                    currentBlock = PEPReader.BlockType.TP_ARCS;
                    
                } else if(PEPReader.BlockType.DONT_CARE.match(s)) {
                    currentBlock = PEPReader.BlockType.DONT_CARE;
                    
                } else if(s.skip("%?").match().group().length() != 0) {
                    this.skipLine(s);
                    
                } else {
                    switch(currentBlock) {
                        case PLACES:
                            this.readPlace(s, net, pl);
                            break;
                        case TRANSITIONS:
                            this.readTransition(s, net, tr);
                            break;
                        case PT_ARCS:
                            this.readPTArc(s, net, pl, tr);
                            break;
                        case TP_ARCS:
                            this.readTPArc(s, net, pl, tr);
                            break;
                        case DONT_CARE:
                            this.skipLine(s);
                            break;
                        default:
                            throw new InternalError("Unknown block type: " + currentBlock);
                    }
                }
            } while(s.hasNext());
            
        } catch(final RuntimeException e) {
//            System.out.println(this.skipLine(s));
            this.reportError(e);
            throw e;
        } finally {
            try {
                this.in.close();
            } catch(final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param e
     */
    private void reportError(final Exception e) {
        for(final ParserErrorHandler handler : this.errorHandlers) {
            handler.reportError(e.getLocalizedMessage());
        }
        if(this.errorHandlers.isEmpty()) {
            e.printStackTrace();
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param net
     * @param pl
     */
    private void readPlace(final Scanner s, final PetriNet net,
                    final Map<String, Place> pl) {
        s.skip("\\s*(\\d+)?\\s*(?:\"((?:[^\"]|\\\\\")*)\")?\\s*(\\d+)\\s*@\\s*(\\d+)");
        final MatchResult r = s.match();
        String pNr = r.group(1);
        if(pNr == null) {
            pNr = String.valueOf(pl.size()+1);
        }
        String name = this.decode(r.group(2));
        
        final int x = Integer.parseInt(r.group(3));
        final int y = Integer.parseInt(r.group(4));
        
        Place p = pl.get(pNr);
        if(p == null) {
            p = net.addPlace(name, 0);
            pl.put(pNr, p);
        } else {
            p.setName(name);
        }
        
        p.setX(x);
        p.setY(y);
        
//        System.out.println(pNr + " \"" + name + "\" " + x + " @ " + y);
        
        this.skipLine(s);
        final String l = s.match().group();
        
        final String m = this.decode(this.getValue(l, 'b'));
        if(m.length() != 0) {
            p.setMeaning(m);
        }
        
        final String mark = this.decode(this.getValue(l, 'M'));
        if(mark.length() != 0) {
            p.setMarking(Integer.parseInt(mark));
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @return
     */
    private String getValue(final String s, final char key) {
        final char[] c = s.toCharArray();
        final StringBuilder b = new StringBuilder();
        boolean escape = false;
        boolean quote = false;
        boolean number = false;
        boolean found = false;
        for(int i = 0; i < c.length; ++i) {
            if(quote) {
                if(!escape && (c[i] == '\"')) {
                    quote = !quote;
                    if(found) {
                        break;
                    }
                } else if(found) {
                    b.append(c[i]);
                }
                escape = (c[i] == '\\');
            } else if(c[i] == key) {
                found = true;
            } else if(c[i] == '\"') {
                quote = true;
            } else if(found && (Character.isDigit(c[i]) || (c[i] == '@'))) {
                number = true;
                b.append(c[i]);
            } else if(number && (c[i] != '@') && !Character.isDigit(c[i])) {
                break;
            }
        }
        return b.toString();
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param net
     * @param tr
     */
    private void readTransition(final Scanner s, final PetriNet net,
                    final Map<String, Transition> tr) {
        s.skip("\\s*(\\w+)?\\s*(?:\"((?:[^\"]|\\\\\")*)\")?\\s*(\\d+)\\s*@\\s*(\\d+)");
        final MatchResult r = s.match();
        String tNr = r.group(1);
        if(tNr == null) {
            tNr = String.valueOf(tr.size()+1);
        }
        String name = this.decode(r.group(2));
        
        final int x = Integer.parseInt(r.group(3));
        final int y = Integer.parseInt(r.group(4));
        
        Transition t = tr.get(tNr);
        if(t == null) {
            t = net.addTransition(name);
            tr.put(tNr, t);
        } else {
            t.setName(name);
        }
        
        t.setX(x);
        t.setY(y);
        
//        System.out.println(tNr + " \"" + name + "\" " + x + " @ " + y);
        
        this.skipLine(s);
        final String m = this.decode(this.getValue(s.match().group(), 'b'));
        if(m.length() != 0) {
            t.setMeaning(m);
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @return
     */
    private String decode(final String s) {
        return s.replaceAll("\\\\\"", "\"").replaceAll("\\\\\\\\", "\\");
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param net
     * @param pl
     * @param tr
     */
    private void readPTArc(final Scanner s, final PetriNet net,
                    final Map<String, Place> pl, final Map<String, Transition> tr) {
        s.skip("\\s*(\\d+)\\s*>\\s*(\\d+)");
        final MatchResult r = s.match();
        final String pNr = r.group(1);
        final String tNr = r.group(2);
        
        Place p = pl.get(pNr);
        if(p == null) {
            p = net.addPlace("P" + pNr, 0);
            pl.put(pNr, p);
        }
        
        Transition t = tr.get(tNr);
        if(t == null) {
            t = net.addTransition("T" + tNr);
            tr.put(tNr, t);
        }
        
        this.skipLine(s);
        final String w = this.getValue(s.match().group(), 'w');
        final int weight = (w.length() == 0) ? 1 : Integer.parseInt(w);
        
        net.addArc(p, t, weight);
        
//        System.out.println(pNr + " > " + tNr + " w " + weight);
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param net
     * @param pl
     * @param tr
     */
    private void readTPArc(final Scanner s, final PetriNet net,
                    final Map<String, Place> pl, final Map<String, Transition> tr) {
        s.skip("\\s*(\\d+)\\s*<\\s*(\\d+)");
        final MatchResult r = s.match();
        final String tNr = r.group(1);
        final String pNr = r.group(2);
        
        Place p = pl.get(pNr);
        if(p == null) {
            p = net.addPlace("P" + pNr, 0);
            pl.put(pNr, p);
        }
        
        Transition t = tr.get(tNr);
        if(t == null) {
            t = net.addTransition("T" + tNr);
            tr.put(tNr, t);
        }
        
        this.skipLine(s);
        final String w = this.getValue(s.match().group(), 'w');
        final int weight = (w.length() == 0) ? 1 : Integer.parseInt(w);
        net.addArc(t, p, weight);
        
//        System.out.println(tNr + " < " + pNr + " w " + weight);
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     */
    private String skipLine(final Scanner s) {
        return s.skip("[^\n\r\u2028\u2029\u0085]*\\s*").match().group();
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param net
     */
    private void readComment(final Scanner s, final PetriNet net) {
        s.skip("(?m)(\\s*%.*$)*");
        net.setHeaderComment(s.match().group().trim().replaceAll("(?m)^\\s*%\\s", ""));
        
//        System.out.println(net.getHeaderComment());
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#addErrorHandler(ParserErrorHandler)
     */
    @Override
    public void addErrorHandler(final ParserErrorHandler handler) {
        this.errorHandlers.add(handler);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.LocalConfigurationParser#getLocalConfiguration()
     */
    @Override
    public Properties getLocalConfiguration() {
        try {
            final Scanner s = new Scanner(this.in);
            final String cmt = s.skip("(?m)(\\s*%.*$)*").match().group().trim().replaceAll("(?m)^\\s*%\\s", "");
            final LocalConfigurationParser p = new LocalConfigParser();
            p.setInput(new ByteArrayInputStream(cmt.getBytes()));
            final Properties props = p.getLocalConfiguration();
            p.reset();
            return props;
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#setInput(java.io.InputStream)
     */
    @Override
    public void setInput(final InputStream in) {
        this.in = in;        
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamReader#readTo(petruchio.interfaces.petrinet.PetriNetStreamWriter)
     */
    @Override
    public void readTo(final PetriNetStreamWriter writer) {
        try {
            final Scanner s = new Scanner(this.in).useDelimiter("\\s+");
            writer.beginNet("noname");
            
            String[] pl = new String[100];
            String[] tr = new String[100];
    
            writer.writeHeaderComment(s.skip("(?m)(\\s*%.*$)*").match().group().trim().replaceAll("(?m)^\\s*%\\s", ""));
            
            BlockType currentBlock = PEPReader.BlockType.DONT_CARE;
            do {
                s.skip("\\s*");
                
                if(PEPReader.BlockType.PLACES.match(s)) {
                    currentBlock = PEPReader.BlockType.PLACES;
                    
                } else if(PEPReader.BlockType.TRANSITIONS.match(s)) {
                    currentBlock = PEPReader.BlockType.TRANSITIONS;
                    
                } else if(PEPReader.BlockType.PT_ARCS.match(s)) {
                    currentBlock = PEPReader.BlockType.PT_ARCS;
                    
                } else if(PEPReader.BlockType.TP_ARCS.match(s)) {
                    currentBlock = PEPReader.BlockType.TP_ARCS;
                    
                } else if(PEPReader.BlockType.DONT_CARE.match(s)) {
                    currentBlock = PEPReader.BlockType.DONT_CARE;
                    
                } else if(s.skip("%?").match().group().length() != 0) {
                    this.skipLine(s);
                    
                } else {
                    switch(currentBlock) {
                        case PLACES: {
                            s.skip("\\s*(\\w+)\\s*(?:\"((?:[^\"]|\\\\\")*)\")?\\s*(\\d+)\\s*@\\s*(\\d+)");
                            final MatchResult r = s.match();
                            final int pNr = Integer.parseInt(r.group(1));
                            String name = this.decode(r.group(2));
                            
                            final int x = Integer.parseInt(r.group(3));
                            final int y = Integer.parseInt(r.group(4));
                            
                            if(pNr >= pl.length) {
                                final String[] b = new String[pl.length+100];
                                System.arraycopy(pl, 0, b, 0, pl.length);
                                pl = b;
                            }
                            pl[pNr] = name;
                            final Place p = new petruchio.pn.Place(name, 0);
                            
                            p.setX(x);
                            p.setY(y);
                            
                            this.skipLine(s);
                            final String l = s.match().group();
                            
                            final String m = this.decode(this.getValue(l, 'b'));
                            if(m.length() != 0) {
                                p.setMeaning(m);
                            }
                            
                            final String mark = this.decode(this.getValue(l, 'M'));
                            if(mark.length() != 0) {
                                p.setMarking(Integer.parseInt(mark));
                            }
                            writer.writePlace(p);
                            break; }
                        case TRANSITIONS: {
                            s.skip("\\s*(\\w+)\\s*(?:\"((?:[^\"]|\\\\\")*)\")?\\s*(\\d+)\\s*@\\s*(\\d+)");
                            final MatchResult r = s.match();
                            final int tNr = Integer.parseInt(r.group(1));
                            String name = this.decode(r.group(2));
                            
                            final int x = Integer.parseInt(r.group(3));
                            final int y = Integer.parseInt(r.group(4));
                            
                            if(tNr >= tr.length) {
                                final String[] b = new String[tr.length+100];
                                System.arraycopy(tr, 0, b, 0, tr.length);
                                tr = b;
                            }
                            tr[tNr] = name;
                            final Transition t = new petruchio.pn.Transition(name);
                            
                            t.setX(x);
                            t.setY(y);
                            
                            this.skipLine(s);
                            final String m = this.decode(this.getValue(s.match().group(), 'b'));
                            if(m.length() != 0) {
                                t.setMeaning(m);
                            }
                            writer.writeTransition(t);
                            break; }
                        case PT_ARCS: {
                            s.skip("\\s*(\\d+)\\s*>\\s*(\\d+)");
                            final MatchResult r = s.match();
                            final int pNr = Integer.parseInt(r.group(1));
                            final int tNr = Integer.parseInt(r.group(2));
                            
                            this.skipLine(s);
                            final String w = this.getValue(s.match().group(), 'w');
                            final int weight = (w.length() == 0) ? 1 : Integer.parseInt(w);
                            
                            writer.writePTArc(pl[pNr], tr[tNr], weight);
                            break; }
                        case TP_ARCS: {
                            
                            s.skip("\\s*(\\d+)\\s*<\\s*(\\d+)");
                            final MatchResult r = s.match();
                            final int tNr = Integer.parseInt(r.group(1));
                            final int pNr = Integer.parseInt(r.group(2));
                            
                            this.skipLine(s);
                            final String w = this.getValue(s.match().group(), 'w');
                            final int weight = (w.length() == 0) ? 1 : Integer.parseInt(w);
                            
                            writer.writeTPArc(tr[tNr], pl[pNr], weight);
                            break; }
                        case DONT_CARE:
                            this.skipLine(s);
                            break;
                        default:
                            throw new InternalError("Unknown block type: " + currentBlock);
                    }
                }
            } while(s.hasNext());
            
            
            writer.endNet();
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
            
        } finally {
            try {
                this.in.close();
            } catch(final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }

}