/*
 * Project:  Petruchio
 * Filename: PNMLReader.java
 * Created:  05.09.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.readers;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Scanner;
import java.util.Stack;

//import javax.xml.parsers.SAXParser;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xml.sax.ext.DefaultHandler2;
import org.xml.sax.helpers.XMLReaderFactory;

import petruchio.interfaces.LocalConfigurationParser;
import petruchio.interfaces.ParserErrorHandler;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReader;
import petruchio.interfaces.petrinet.PetriNetStreamReader;
import petruchio.interfaces.petrinet.PetriNetStreamWriter;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;
import petruchio.pi.readers.LocalConfigParser;


/**
 * Reads a low level PNML Petri net file and builds the corresponding Petri net.
 * 
 * @author Tim Strazny
 */
public class PNMLReader implements SelfDescribing, PetriNetReader,
        PetriNetStreamReader, LocalConfigurationParser {
    
    private InputStream in = null;
    private Collection<ParserErrorHandler> errorHandlers = new ArrayList<ParserErrorHandler>();
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net reader for files in the PNML format.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#getDefaultExtension()
     */
    @Override
    public String getDefaultExtensions() {
        return "pnml, xml";
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        this.errorHandlers.clear();
        if (this.in != null) {
            try {
                this.in.close();
            } catch (final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#read(java.io.InputStream)
     */
    @Override
    public PetriNet read() {
        final PetriNet net = new petruchio.pn.PetriNet();
        this.read(net);
        return net;
    }
        
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#read(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public void read(final PetriNet net) {
        try {
            final PNMLContentHandler handler = new PNMLContentHandler(net);

            final XMLReader parser = XMLReaderFactory.createXMLReader();
            
            parser.setContentHandler(handler);
            parser.setProperty("http://xml.org/sax/properties/lexical-handler", handler);
            
            parser.parse(new InputSource(this.in));
            
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
        } catch(final IOException e) {
            this.reportError(e);
            throw new RuntimeException(e);
        } catch(final SAXException e) {
            this.reportError(e);
            throw new RuntimeException(e);
        } finally {
            try {
                this.in.close();
            } catch(final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param e
     */
    private void reportError(final Exception e) {
        for(final ParserErrorHandler handler : this.errorHandlers) {
            handler.reportError(e.getLocalizedMessage());
        }
        if(this.errorHandlers.isEmpty()) {
            e.printStackTrace();
        }
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#addErrorHandler(ParserErrorHandler)
     */
    @Override
    public void addErrorHandler(final ParserErrorHandler handler) {
        this.errorHandlers.add(handler);
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReader#setInput(java.io.InputStream)
     */
    @Override
    public void setInput(final InputStream in) {
        this.in = in;
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.LocalConfigurationParser#getLocalConfiguration()
     */
    @Override
    public Properties getLocalConfiguration() {
        try {
            final Scanner s = new Scanner(this.in);
            s.skip("(?s).*?<pnml\\b.*?>.*?<net\\b.*?>\\s*(?:<name\\b.*?>.*?</name>)?(<!--)?");
            if(s.match().group(1) == null) {
                return new Properties();
            }
            final String cmt = s.skip("(?s)(.*?)-->").match().group().replaceAll("-->", "").replaceAll("(?m)^\\s*-\\s", "").trim();
            final LocalConfigurationParser p = new LocalConfigParser();
            p.setInput(new ByteArrayInputStream(cmt.getBytes()));
            final Properties props = p.getLocalConfiguration();
            p.reset();
            return props;
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamReader#readTo(petruchio.interfaces.petrinet.PetriNetStreamWriter)
     */
    @Override
    public void readTo(final PetriNetStreamWriter writer) {
        try {
            final PNMLStreamContentHandler handler = new PNMLStreamContentHandler(writer);

            final XMLReader parser = XMLReaderFactory.createXMLReader();
            
            parser.setContentHandler(handler);
            parser.setProperty("http://xml.org/sax/properties/lexical-handler", handler);
            
            parser.parse(new InputSource(this.in));
            
        } catch(final RuntimeException e) {
            this.reportError(e);
            throw e;
        } catch(final IOException e) {
            this.reportError(e);
            throw new RuntimeException(e);
        } catch(final SAXException e) {
            this.reportError(e);
            throw new RuntimeException(e);
        } finally {
            try {
                this.in.close();
            } catch(final IOException e) {
                // ignore
            }
            this.in = null;
        }
    }

    
    /**
     * <Description>
     * 
     * @author Tim Strazny
     */
    static class PNMLContentHandler extends DefaultHandler2 {

        private Stack<String> state;
        
        private final Map<String, Place> pl;
        
        private final Map<String, Transition> tr;
        
        private final PetriNet net;
        
        private Place curPlace;
        
        private Transition curTransition;
        
        private PTArc curPTArc;
        
        private TPArc curTPArc;
        
        private StringBuilder lastChars;
        
        private boolean done;
        
        
        /**
         * 
         */
        public PNMLContentHandler(final PetriNet net) {
            this.pl = new HashMap<String, Place>();
            this.tr = new HashMap<String, Transition>();
            this.net = net;
            this.state = new Stack<String>();
            this.curPlace = null;
            this.curTransition = null;
            this.curPTArc = null;
            this.curTPArc = null;
            this.lastChars = new StringBuilder();
            this.done = false;
        }
        
        
        /**
         * TODO: document
         *
         * @return
         */
        public PetriNet getNet() {
            return this.net;
        }
        
        
        /* (non-Javadoc)
         * @see org.xml.sax.ContentHandler#characters(char[], int, int)
         */
        @Override
        public void characters(char[] ch, int start, int length)
                        throws SAXException {
            if(this.done) {
                return;
            }
            final int lvl = this.state.size();
            if(lvl < 4 || lvl > 5) {
                return;
            }
            
            if((lvl >= 1) && this.state.get(0).equals("pnml")) {
                if((lvl >= 2) && this.state.get(1).equals("net")) {
                    if((lvl >= 3) && this.state.get(2).equals("name")) {
                        if((lvl == 4) && this.isValue(this.state.get(3))) {
                            this.lastChars.append(ch, start, length);
                            this.net.setName(this.lastChars.toString());
                            return;
                        }
                        
                    } else if((lvl >= 3) && this.state.get(2).equals("place")) {
                        if((lvl >= 4) && this.state.get(3).equals("initialMarking")) {
                            if((lvl == 5) && this.isValue(this.state.get(4))) {
                                this.lastChars.append(ch, start, length);
                                this.curPlace.setMarking(Integer.parseInt(this.lastChars.toString()));
                                return;
                            }
                        } else if((lvl >= 4) && this.state.get(3).equals("name")) {
                            if((lvl == 5) && this.isValue(this.state.get(4))) {
                                this.lastChars.append(ch, start, length);
                                this.curPlace.setMeaning(this.lastChars.toString());
                                return;
                            }
                        }
                        
                    } else if((lvl >= 3) && this.state.get(2).equals("transition")) {
                        if((lvl >= 4) && this.state.get(3).equals("name")) {
                            if((lvl == 5) && this.isValue(this.state.get(4))) {
                                this.lastChars.append(ch, start, length);
                                this.curTransition.setMeaning(this.lastChars.toString());
                                return;
                            }
                        }
                        
                    } else if((lvl >= 3) && this.state.get(2).equals("arc")) {
                        if((lvl >= 4) && this.state.get(3).equals("inscription")) {
                            if((lvl == 5) && this.isValue(this.state.get(4))) {
                                this.lastChars.append(ch, start, length);
                                if(this.curPTArc != null) {
                                    this.curPTArc.setWeight(Integer.parseInt(this.lastChars.toString()));
                                }
                                if(this.curTPArc != null) {
                                    this.curTPArc.setWeight(Integer.parseInt(this.lastChars.toString()));
                                }
                                return;
                            }
                        }
                    }
                }
            }
        }
        
        
        /* (non-Javadoc)
         * @see org.xml.sax.ext.DefaultHandler2#comment(char[], int, int)
         */
        @Override
        public void comment(char[] ch, int start, int length)
                        throws SAXException {
            if(this.done) {
                return;
            }
            
            final int lvl = this.state.size();
            if(lvl != 2) {
                return;
            }

            if((lvl >= 1) && this.state.get(0).equals("pnml")) {
                if((lvl == 2) && this.state.get(1).equals("net")) {
                    if(!this.net.hasHeaderComment()) {
                        this.lastChars.append(ch, start, length);
                        this.net.setHeaderComment(this.lastChars.toString().trim().replaceAll("(?m)^\\s*--?\\s", ""));
                    }
                }
            }
        }
        
        
        /* (non-Javadoc)
         * @see org.xml.sax.ContentHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
         */
        @Override
        public void startElement(String uri, String localName, String name,
                        Attributes atts) throws SAXException {
            if(this.done) {
                return;
            }
            this.lastChars.delete(0, this.lastChars.length());
            
            this.state.push(localName);
            
            final int lvl = this.state.size();
            
            if((lvl == 3) && localName.equals("place")) {
                if(this.curPlace == null) {
                    this.curPlace = this.getPlace(atts.getValue("", "id"));
                }
                
            } else if((lvl == 3) && localName.equals("transition")) {
                if(this.curTransition == null) {
                    this.curTransition = this.getTransition(atts.getValue("", "id"));
                }
                
            } else if((lvl == 3) && localName.equals("arc")) {
                final String source = atts.getValue("", "source");
                final String target = atts.getValue("", "target");
                final Place testP = this.pl.get(source);
                if(testP == null) {
                    if(this.curTPArc == null) {
                        this.curTPArc = this.net.addArc(this.getTransition(source), this.getPlace(target), 1);
                    }
                } else {
                    if(this.curPTArc == null) {
                        this.curPTArc = this.net.addArc(testP, this.getTransition(target), 1);
                    }
                }
                
            } else if((lvl == 5) && localName.equals("position")) {
                if(this.curPlace != null) {
                    this.curPlace.setX((int)Double.parseDouble(atts.getValue("", "x")));
                    this.curPlace.setY((int)Double.parseDouble(atts.getValue("", "y")));
                }
                if(this.curTransition != null) {
                    this.curTransition.setX((int)Double.parseDouble(atts.getValue("", "x")));
                    this.curTransition.setY((int)Double.parseDouble(atts.getValue("", "y")));
                }
            }
        }
        
        
        /**
         * TODO: document
         *
         * @param localName
         * @return
         */
        private boolean isValue(final String localName) {
            return localName.equals("value") || localName.equals("text");
        }
        
        
        /**
         * TODO: document
         *
         * @param id
         * @return
         */
        private Place getPlace(final String id) {
            Place p = this.pl.get(id);
            if(p == null) {
                p = this.net.addPlace(id, 0);
                this.pl.put(id, p);
            }            
            return p;
        }
        
        
        /**
         * TODO: document
         *
         * @param id
         * @return
         */
        private Transition getTransition(final String id) {
            Transition t = this.tr.get(id);
            if(t == null) {
                t = this.net.addTransition(id);
                this.tr.put(id, t);
            }            
            return t;
        }


        /* (non-Javadoc)
         * @see org.xml.sax.ContentHandler#endElement(java.lang.String, java.lang.String, java.lang.String)
         */
        @Override
        public void endElement(String uri, String localName, String name)
                        throws SAXException {
            if(this.done) {
                return;
            }
            
            this.lastChars.delete(0, this.lastChars.length());
            
            final int lvl = this.state.size();
            final String pop = this.state.pop();
            
            if((lvl == 2) && pop.equals("net")) {
                this.done = true;
                this.state.clear();
                
            } if((lvl == 3) && pop.equals("place")) {
                this.curPlace = null;
                
            } if((lvl == 3) && pop.equals("transition")) {
                this.curTransition = null;
                
            } if((lvl == 3) && pop.equals("arc")) {
                this.curPTArc = null;
                this.curTPArc = null;
                
            }
        }
        
    }
    
    
    static class PNMLStreamContentHandler extends DefaultHandler2 {

        private final PetriNetStreamWriter writer;
        
        private Stack<String> state;
        
        private final Collection<String> pl;
        
        private Place curPlace;
        
        private Transition curTransition;
        
        private String curSource;
        
        private String curTarget;
        
        private boolean curPT;
        
        private int curWeight;
        
        private StringBuilder lastChars;
        
        private boolean done;
        
        private boolean headerWritten;
        
        
        /**
         * 
         */
        public PNMLStreamContentHandler(final PetriNetStreamWriter writer) {
            this.writer = writer;
            this.pl = new HashSet<String>();
            this.state = new Stack<String>();
            this.curPlace = null;
            this.curTransition = null;
            this.curSource = null;
            this.curTarget = null;
            this.curPT = false;
            this.curWeight = 1;
            this.lastChars = new StringBuilder();
            this.done = false;
            this.headerWritten = false;
        }
        
        
        /* (non-Javadoc)
         * @see org.xml.sax.ContentHandler#characters(char[], int, int)
         */
        @Override
        public void characters(char[] ch, int start, int length)
                        throws SAXException {
            if(this.done) {
                return;
            }
            final int lvl = this.state.size();
            if(lvl < 4 || lvl > 5) {
                return;
            }
            
            if((lvl >= 1) && this.state.get(0).equals("pnml")) {
                if((lvl >= 2) && this.state.get(1).equals("net")) {
                    if((lvl >= 3) && this.state.get(2).equals("name")) {
//                        if((lvl == 4) && this.isValue(this.state.get(3))) {
//                            this.lastChars.append(ch, start, length);
//                            this.writer.beginNet(this.lastChars.toString());
//                            return;
//                        }
                        return;
                        
                    } else if((lvl >= 3) && this.state.get(2).equals("place")) {
                        if((lvl >= 4) && this.state.get(3).equals("initialMarking")) {
                            if((lvl == 5) && this.isValue(this.state.get(4))) {
                                this.lastChars.append(ch, start, length);
                                this.curPlace.setMarking(Integer.parseInt(this.lastChars.toString()));
                                return;
                            }
                        } else if((lvl >= 4) && this.state.get(3).equals("name")) {
                            if((lvl == 5) && this.isValue(this.state.get(4))) {
                                this.lastChars.append(ch, start, length);
                                this.curPlace.setMeaning(this.lastChars.toString());
                                return;
                            }
                        }
                        
                    } else if((lvl >= 3) && this.state.get(2).equals("transition")) {
                        if((lvl >= 4) && this.state.get(3).equals("name")) {
                            if((lvl == 5) && this.isValue(this.state.get(4))) {
                                this.lastChars.append(ch, start, length);
                                this.curTransition.setMeaning(this.lastChars.toString());
                                return;
                            }
                        }
                        
                    } else if((lvl >= 3) && this.state.get(2).equals("arc")) {
                        if((lvl >= 4) && this.state.get(3).equals("inscription")) {
                            if((lvl == 5) && this.isValue(this.state.get(4))) {
                                this.lastChars.append(ch, start, length);
                                this.curWeight = Integer.parseInt(this.lastChars.toString());
                                return;
                            }
                        }
                    }
                }
            }
        }
        
        
        /* (non-Javadoc)
         * @see org.xml.sax.ext.DefaultHandler2#comment(char[], int, int)
         */
        @Override
        public void comment(char[] ch, int start, int length)
                        throws SAXException {
            if(this.done) {
                return;
            }
            
            final int lvl = this.state.size();
            if(lvl != 2) {
                return;
            }

            if((lvl >= 1) && this.state.get(0).equals("pnml")) {
                if((lvl == 2) && this.state.get(1).equals("net")) {
                    if(!this.headerWritten) {
                        this.lastChars.append(ch, start, length);
                        this.writer.writeHeaderComment(this.lastChars.toString().trim().replaceAll("(?m)^\\s*--?\\s", "").trim());
                        this.headerWritten = true;
                    }
                }
            }
        }
        
        
        /* (non-Javadoc)
         * @see org.xml.sax.ContentHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
         */
        @Override
        public void startElement(String uri, String localName, String name,
                        Attributes atts) throws SAXException {
            if(this.done) {
                return;
            }
            this.lastChars.delete(0, this.lastChars.length());
            
            this.state.push(localName);
            
            final int lvl = this.state.size();
            
            if((lvl == 2) && localName.equals("net")) {
                this.writer.beginNet(atts.getValue("", "id"));
            
            } else if((lvl == 3) && localName.equals("place")) {
                if(this.curPlace == null) {
                    this.curPlace = new petruchio.pn.Place(atts.getValue("", "id"), 0);
                    this.pl.add(this.curPlace.getName());
                }
                
            } else if((lvl == 3) && localName.equals("transition")) {
                if(this.curTransition == null) {
                    this.curTransition = new petruchio.pn.Transition(atts.getValue("", "id"));
                }
                
            } else if((lvl == 3) && localName.equals("arc")) {
                this.curSource = atts.getValue("", "source");
                this.curTarget = atts.getValue("", "target");
                this.curPT = this.pl.contains(this.curSource);
                
            } else if((lvl == 5) && localName.equals("position")) {
                if(this.curPlace != null) {
                    this.curPlace.setX((int)Double.parseDouble(atts.getValue("", "x")));
                    this.curPlace.setY((int)Double.parseDouble(atts.getValue("", "y")));
                }
                if(this.curTransition != null) {
                    this.curTransition.setX((int)Double.parseDouble(atts.getValue("", "x")));
                    this.curTransition.setY((int)Double.parseDouble(atts.getValue("", "y")));
                }
            }
        }
        
        
        /**
         * TODO: document
         *
         * @param localName
         * @return
         */
        private boolean isValue(final String localName) {
            return localName.equals("value") || localName.equals("text");
        }


        /* (non-Javadoc)
         * @see org.xml.sax.ContentHandler#endElement(java.lang.String, java.lang.String, java.lang.String)
         */
        @Override
        public void endElement(String uri, String localName, String name)
                        throws SAXException {
            if(this.done) {
                return;
            }
            
            this.lastChars.delete(0, this.lastChars.length());
            
            final int lvl = this.state.size();
            final String pop = this.state.pop();
            
            if((lvl == 2) && pop.equals("net")) {
                this.done = true;
                this.writer.endNet();
                this.state.clear();
                
            } if((lvl == 3) && pop.equals("place")) {
                this.writer.writePlace(this.curPlace);
                this.curPlace = null;
                
            } if((lvl == 3) && pop.equals("transition")) {
                this.writer.writeTransition(this.curTransition);
                this.curTransition = null;
                
            } if((lvl == 3) && pop.equals("arc")) {
                if(this.curPT) {
                    this.writer.writePTArc(this.curSource, this.curTarget, this.curWeight);
                } else {
                    this.writer.writeTPArc(this.curSource, this.curTarget, this.curWeight);
                }
                this.curSource = null;
                this.curTarget = null;
                this.curWeight = 1;
                
            }
        }
        
    }

}