/*
 * Project:  Petruchio
 * Filename: AggressiveSequenceReduction.java
 * Created:  09.08.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.reductors;

import java.util.ArrayList;
import java.util.Collection;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;
import petruchio.interfaces.petrinet.TPArc;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class AggressiveSequenceReduction implements PetriNetReductor {
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#getDescription()
     */
    @Override
    public String getDescription() {
        return "Iff there is a sequence P1-T, where P1*={T} and *T={P1}, then the output arcs from T are added to each transition in *P1 and P1, T are removed.";
    }
    
    
    /* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		// nothing
	}


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#getName()
     */
    @Override
    public String getName() {
        return "aggressive sequence reduction";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public int reduce(final PetriNet net) {
    	return this.reduce(net, new ArrayList<Place>(), new ArrayList<Transition>());
    }
    	
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet, java.util.Collection, java.util.Collection)
     */
    @Override
    public int reduce(final PetriNet net, final Collection<Place> ignorePlaces,
    		final Collection<Transition> ignoreTransitions) {
    	final Collection<Place> ignoreP = new IdentityHashSet<Place>(ignorePlaces);
    	final Collection<Transition> ignoreT = new IdentityHashSet<Transition>(ignoreTransitions);
    	
    	final Collection<Place> places = new ArrayList<Place>(net.getPlaces());
        int removed = 0;
        
        next_place:
        for(final Place p : places) {
            if((p.getOutput().size() != 1) || p.getInput().isEmpty() || ignoreP.contains(p)) {
                continue next_place;
            }
            final PTArc theArc = p.getOutput().iterator().next();
            final Transition t = theArc.getTarget();
            
            if((t.getInput().size() != 1) || ignoreT.contains(t)) {
                continue next_place;
            }
            
            final int weight = theArc.getWeight();
            
            boolean aggressive = false;
            for(final TPArc outArc : t.getOutput()) {
                final Place out = outArc.getTarget();
                if((out == p) || (outArc.getWeight() % weight != 0)) {
                    continue next_place;
                }
                if(!aggressive) {
                    for(final TPArc inArc: out.getInput()) {
                        final Transition pin = inArc.getSource(); 
                        if(pin == t) {
                            aggressive = true;
                            break;
                        }
                    }
                }
            }
            
            for(final TPArc outArc : t.getOutput()) {
                final Place out = outArc.getTarget();
                final int w = outArc.getWeight() / weight;
                if(p.getMarking() != 0) {
                    out.addTokens(w*p.getMarking());
                }
                if(!aggressive) {
                    out.setMeaning(p.getMeaning() + " --> " + out.getMeaning());
                }
            }
            
            for(final TPArc inArc : p.getInput()) {
                final Transition in = inArc.getSource();
                for(final TPArc outArc : t.getOutput()) {
                    final Place out = outArc.getTarget();
                    net.addArc(in, out, inArc.getWeight()*outArc.getWeight() / weight);
                }
                if(!aggressive) {
                    in.setMeaning(in.getMeaning() + " --> " + t.getMeaning());
                }
            }
            
            if(net.removePlace(p)) {
            	++removed;
            }
            if(net.removeTransition(t)) {
            	++removed;
            }
        }
        
        return removed;
    }

}