/*
 * Project:  Petruchio
 * Filename: AlternativesReduction.java
 * Created:  09.08.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.reductors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class AlternativesReduction implements PetriNetReductor {
    
    private final List<Transition> trans = new ArrayList<Transition>();
    private final List<Map<Place, Integer>> in = new ArrayList<Map<Place, Integer>>();
    private final List<Map<Place, Integer>> out = new ArrayList<Map<Place, Integer>>();

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#getDescription()
     */
    @Override
    public String getDescription() {
        return "Iff there are two transitions with the same input and output arcs, one of them is removed.";
    }
    
    
    /* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		this.trans.clear();
		this.in.clear();
		this.out.clear();
	}


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#getName()
     */
    @Override
    public String getName() {
        return "alternatives reduction";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#reduce(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public synchronized int reduce(final PetriNet net) {
    	return this.reduce(net, null, new ArrayList<Transition>());
    }
    	
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet, java.util.Collection, java.util.Collection)
     */
    @Override
    public int reduce(final PetriNet net, final Collection<Place> ignorePlaces,
    		final Collection<Transition> ignoreTransitions) {
    	final Collection<Transition> ignoreT = new IdentityHashSet<Transition>(ignoreTransitions);
    	
        int removed = 0;
        for(final Place p : net.getPlaces()) {
            if(p.getOutput().size() >= 2) {
                for(final PTArc arc : p.getOutput()) {
                    this.trans.add(arc.getTarget());
                }
                removed += this.reduce(net, ignoreT);
                this.trans.clear();
            }
            if(p.getInput().size() >= 2) {
                for(final TPArc arc : p.getInput()) {
                    this.trans.add(arc.getSource());
                }
                removed += this.reduce(net, ignoreT);
                this.trans.clear();
            }
        }
        return removed;
    }
    
    
    private int reduce(final PetriNet net, final Collection<Transition> ignoreT) {
        int removed = 0;
        this.in.clear();
        this.out.clear();
        for(final Transition t : this.trans) {
            final Map<Place, Integer> inm = new IdentityHashMap<Place, Integer>();
            final Map<Place, Integer> outm = new IdentityHashMap<Place, Integer>();
            this.in.add(inm);
            this.out.add(outm);
            for(final PTArc inArc : t.getInput()) {
                inm.put(inArc.getSource(), inArc.getWeight());
            }
            for(final TPArc outArc : t.getOutput()) {
                outm.put(outArc.getTarget(), outArc.getWeight());
            }
        }
        for(int i = this.trans.size()-2; i >= 0; --i) {
            final Transition ti = this.trans.get(i);
            final Map<Place, Integer> ini = this.in.get(i);
            final Map<Place, Integer> outi = this.out.get(i);
            for(int j = this.trans.size()-1; j > i; --j) {
                final Transition tj = this.trans.get(j);
                if((ti == tj) || ignoreT.contains(tj)) {
                    continue;
                }
                final Map<Place, Integer> inj = this.in.get(j);
                final Map<Place, Integer> outj = this.out.get(j);
                if(ini.equals(inj) && outi.equals(outj)) {
                    ti.setX((ti.getX() + tj.getX()) >> 1);
                    ti.setY((ti.getY() + tj.getY()) >> 1);
                    ti.setName("(" + ti.getName() + ") & (" + tj.getName() + ")");
                    ti.setMeaning("(" + ti.getMeaning() + ") & (" + tj.getMeaning() + ")");
                    this.trans.remove(j);
                    if(net.removeTransition(tj)) {
                        this.trans.remove(tj);
                    }
                }
            }
        }
        this.trans.clear();
        this.in.clear();
        this.out.clear();
        return removed;
    }

}