/*
 * Project:  Petruchio
 * Filename: CoveringReduction.java
 * Created:  21.08.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.reductors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class CoveringReduction implements PetriNetReductor {
    
    private final List<TPArc> trans = new ArrayList<TPArc>();
    private final List<Map<Place, Integer>> in = new ArrayList<Map<Place, Integer>>();
    private final List<Map<Place, Integer>> out = new ArrayList<Map<Place, Integer>>();
    
    
    /* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		this.trans.clear();
		this.in.clear();
		this.out.clear();
	}
	
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#getDescription()
     */
    @Override
    public String getDescription() {
        return "If pre and post sets of transitions in respect of the arc "
            + "weights are interpreted as multisets, a transition T2 is removed"
            + " if there exists a transition T1 with *T1 <= *T2 and T1* >= T2*."
            + " Transition T1 then consumes less or equally many tokens as T2 "
            + "and produces more or equally many tokens as T2. Thus a state "
            + "reachable from state X by firing T2 is covered by a state "
            + "reachable from state X by firing T1.";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#getName()
     */
    @Override
    public String getName() {
        return "covering reduction";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public synchronized int reduce(final PetriNet net) {
    	return this.reduce(net, null, new ArrayList<Transition>());
    }
    	
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet, java.util.Collection, java.util.Collection)
     */
    @Override
    public int reduce(final PetriNet net, final Collection<Place> ignorePlaces,
    		final Collection<Transition> ignoreTransitions) {
    	final Collection<Transition> ignoreT = new IdentityHashSet<Transition>(ignoreTransitions);
    	
        int removed = 0;
        for(final Place p : net.getPlaces()) {
            if(p.getInput().size() < 2) {
                continue;
            }
            this.trans.clear();
            this.trans.addAll(p.getInput());
            this.in.clear();
            this.out.clear();
            for(final TPArc arc : this.trans) {
                final Transition t = arc.getSource();
                final Map<Place, Integer> inm = new IdentityHashMap<Place, Integer>();
                final Map<Place, Integer> outm = new IdentityHashMap<Place, Integer>();
                this.in.add(inm);
                this.out.add(outm);
                for(final PTArc inArc : t.getInput()) {
                    inm.put(inArc.getSource(), inArc.getWeight());
                }
                for(final TPArc outArc : t.getOutput()) {
                    outm.put(outArc.getTarget(), outArc.getWeight());
                }
            }
            for(int i = this.trans.size()-2; i >= 0; --i) {
                final Transition ti = this.trans.get(i).getSource();
                final Map<Place, Integer> ini = this.in.get(i);
                final Map<Place, Integer> outi = this.out.get(i);
                trans_j:
                for(int j = this.trans.size()-1; j > i; --j) {
                    final Transition tj = this.trans.get(j).getSource();
                    if((ti == tj) || ignoreT.contains(tj)) {
                        continue trans_j;
                    }
                    // ensure that *Ti <=_multiset *Tj
                    if(!CoveringReduction.leq(ini, this.in.get(j))) {
                        continue trans_j;
                    }
                    // ensure that Ti* >=_multiset Tj*
                    // (this test might return later than the test before,
                    //  since Ti, Tj are both elements in *P)
                    if(!CoveringReduction.leq(this.out.get(j), outi)) {
                        continue trans_j;
                    }
                    ti.setX((ti.getX() + tj.getX()) >> 1);
                    ti.setY((ti.getY() + tj.getY()) >> 1);
                    this.trans.remove(j);
                    if(net.removeTransition(tj)) {
                    	++removed;
                    }
                }
            }
            this.trans.clear();
            this.in.clear();
            this.out.clear();
        }
        return removed;
    }
    
    
    /**
     * TODO: document
     *
     * @param a
     * @param b
     * @return
     */
    private static boolean leq(final Map<?, Integer> a, final Map<?, Integer> b) {
        for(final Map.Entry<?, Integer> e : a.entrySet()) {
            final Object curP = e.getKey();
            final Integer va = b.get(curP);
            if((va == null) || (e.getValue() > va)) {
                return false;
            }
        }
        return true;
    }

}