/*
 * Project:  Petruchio
 * Filename: DeadPartRemover.java
 * Created:  17.09.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.reductors;

import java.util.ArrayList;
import java.util.Collection;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.algorithms.Coverability;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;

/**
 * This reduction builds a coverability graph or set for the Petri net and uses
 * it to compute the dead transitions and places that are never marked, which
 * then are removed.
 * 
 * @author Tim Strazny
 */
public class DeadElementRemover implements PetriNetReductor {

	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductor#getDescription()
	 */
	@Override
    public String getDescription() {
		return "Removes every dead transition and every place that cannot be marked.";
	}
	
	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		// nothing
	}

	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductor#getName()
	 */
	@Override
    public String getName() {
		return "dead part remover";
	}

	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet)
	 */
	@Override
    public int reduce(final PetriNet net) {
		return this.reduce(net, new ArrayList<Place>(), new ArrayList<Transition>());
    }
    	
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet, java.util.Collection, java.util.Collection)
     */
    @Override
    public int reduce(final PetriNet net, final Collection<Place> ignorePlaces,
    		final Collection<Transition> ignoreTransitions) {
		final Coverability cov = new petruchio.cov.Coverability();
		cov.setCheckMarkable(false);
		final Collection<Place> startState = new ArrayList<Place>();
        for(final Place p : net.getPlaces()) {
            for(int i = p.getMarking(); i > 0; --i) {
                startState.add(p);
            }
        }
        cov.compute(startState, net.getTransitions());
        
        final Collection<Place> deadPlaces = new IdentityHashSet<Place>(net.getPlaces());
        deadPlaces.removeAll(ignorePlaces);
        final Collection<Transition> deadTrans = new IdentityHashSet<Transition>(net.getTransitions());
        deadTrans.removeAll(ignoreTransitions);
        for(final Transition t : deadTrans) {
	        for(final PTArc arc : t.getInput()) {
	        	deadPlaces.remove(arc.getSource());
	    	}
        }
        cov.keepDeadElements(deadPlaces, deadTrans);
        int removed = 0;
        for(final Place p : deadPlaces) {
        	if(net.removePlace(p)) {
        		++removed;
        	}
        }
        for(final Transition t : deadTrans) {
        	if(net.removeTransition(t)) {
        		++removed;
        	}
        }
		return removed;
	}

}