/*
 * Project:  Petruchio
 * Filename: DependencyReduction.java
 * Created:  25.09.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.reductors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.Map;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;

/**
 * This dependency reduction is a special case and prefix of the omega reduction. 
 * The omega reduction as proposed in &quot;Slicing unbounded Petri nets&quot;
 * by A. Rakow, R. Meyer and T. Strazny, 2007, University of Oldenburg.
 * It keeps LTL-x properties.
 *
 * @author Tim Strazny
 */
public class DependencyReduction implements PetriNetReductor {

	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductor#getDescription()
	 */
	@Override
    public String getDescription() {
		return "The a special case of the omega reduction as proposed in"
			+ "\"Slicing unbounded Petri nets\""
			+ " by A. Rakow, R. Meyer and T. Strazny, 2007, University of Oldenburg."
			+ " It keeps LTL-x properties.";
	}


	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductor#getName()
	 */
	@Override
    public String getName() {
		return "dependency reduction";
	}
	
	
	/* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		// no caches
	}


	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet)
	 */
	@Override
    public int reduce(final PetriNet net) {
		return 0;
	}


	/* (non-Javadoc)
	 * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet, java.util.Collection, java.util.Collection)
	 */
	@Override
    public int reduce(final PetriNet net, final Collection<Place> ignorePlaces,
			final Collection<Transition> ignoreTransitions) {
		if(ignorePlaces.isEmpty() && ignoreTransitions.isEmpty()) {
			return 0;
		}
		
		final Collection<Place> ignoreP = new IdentityHashSet<Place>(ignorePlaces);
		final Collection<Transition> ignoreT = new IdentityHashSet<Transition>(ignoreTransitions);
		
		final Collection<Place> marked = new IdentityHashSet<Place>();
		for(final Place p : net.getPlaces()) {
			if(p.getMarking() != 0) {
				marked.add(p);
			}
		}
		
		int removed = 0;
		
		int change = -1;
		int lastChange;
		do {
			lastChange = change;
			if(change != 0) {
				int oldR = removed;
				removed += this.removeOutput(net, marked, ignoreP, ignoreT);
				if(oldR != removed) {
					change = 0;
				}
			}
			if(change != 1) {
				int oldR = removed;
				removed += this.removeAlwaysMarked(net, marked, ignoreP, ignoreT);
				if(oldR != removed) {
					change = 1;
				}
			}
			if(change != 2) {
				int oldR = removed;
				removed += this.removeSimpleUnbounded(net, marked, ignoreP, ignoreT);
				if(oldR != removed) {
					change = 2;
				}
			}
			if(change != 3) {
                int oldR = removed;
                removed += this.removeIsolated(net, ignoreP, ignoreT);
                if(oldR != removed) {
                    change = 3;
                }
            }
			
		} while(lastChange != change);

		return removed;
	}
	
	
	/**
	 * TODO: description
	 *
	 * @param net
	 * @param ignorePlaces
	 * @param ignoreTransitions
	 * @param marked
	 * @return
	 */
	private int removeOutput(final PetriNet net, final Collection<Place> marked,
            final Collection<Place> ignorePlaces, final Collection<Transition> ignoreTransitions) {

        final LinkedList<Place> toTreatP = new LinkedList<Place>(ignorePlaces);
        final LinkedList<Transition> toTreatT = new LinkedList<Transition>(ignoreTransitions);
        
        final Collection<Place> ignoreP = new IdentityHashSet<Place>();
        final Collection<Transition> ignoreT = new IdentityHashSet<Transition>();
        final Map<Transition, Integer> in = new IdentityHashMap<Transition, Integer>();
        
        while(!toTreatP.isEmpty() || !toTreatT.isEmpty()) {
            while(!toTreatP.isEmpty()) {
                final Place p = toTreatP.remove();
                if(!ignoreP.add(p)) {
                    continue;
                }
                for(final TPArc arc : p.getInput()) {
                    in.put(arc.getSource(), arc.getWeight());
                }
                boolean removeOutput = true;
                for(final PTArc arc : p.getOutput()) {
                    final Integer w = in.remove(arc.getTarget());
                    if(w == null || w < arc.getWeight()) {
                        removeOutput = false;
                        break;
                    }
                }
                if(removeOutput) {
                    toTreatT.addAll(in.keySet());
                } else {
                    for(final TPArc arc : p.getInput()) {
                        toTreatT.add(arc.getSource());
                    }
                    for(final PTArc arc : p.getOutput()) {
                        toTreatT.add(arc.getTarget());
                    }
                }
            }
            while(!toTreatT.isEmpty()) {
                final Transition t = toTreatT.remove();
                if(!ignoreT.add(t)) {
                    continue;
                }
                for(final PTArc arc : t.getInput()) {
                    toTreatP.add(arc.getSource());
                }
            }
        }
        

        int removed = 0;
        final LinkedList<Place> rmP = new LinkedList<Place>(net.getPlaces());
        while(!rmP.isEmpty()) {
            final Place p = rmP.remove();
            if(!ignoreP.contains(p) && net.removePlace(p)) {
                ++removed;
                marked.remove(p);
            }
        }
        final LinkedList<Transition> rmT = new LinkedList<Transition>(net.getTransitions());
        while(!rmT.isEmpty()) {
            final Transition t = rmT.remove();
            if(!ignoreT.contains(t) && net.removeTransition(t)) {
                ++removed;
            }
        }
        return removed;
    }
	
	
	/**
	 * TODO: description
	 *
	 * @param net
	 * @param marked
	 * @param ignorePlaces
	 * @param ignoreTransitions
	 * @return
	 */
	private int removeSimpleUnbounded(final PetriNet net, final Collection<Place> marked,
			final Collection<Place> ignoreP, final Collection<Transition> ignoreT) {
		final LinkedList<Transition> toTreat = new LinkedList<Transition>();
		int removed = 0;
		
		for(final Transition t : net.getTransitions()) {
			if(t.getInput().isEmpty()) {
				toTreat.add(t);
			}
		}
		
		final LinkedList<Place> rm = new LinkedList<Place>();
		while(!toTreat.isEmpty()) {
			final Transition t = toTreat.remove();
			for(final TPArc arc : t.getOutput()) {
				final Place p = arc.getTarget();
				if(!ignoreP.contains(p)) {
					for(final PTArc arc2 : p.getOutput()) {
						final Transition t2 = arc2.getTarget();
						if(t2.getInput().size() == 1) {
							toTreat.add(t2);
						}
					}
					rm.add(p);
				}
			}
			while(!rm.isEmpty()) {
				final Place p = rm.remove();
				if(net.removePlace(p)) {
					marked.remove(p);
					++removed;
				}
			}
			// remove transition if isolated
			if(t.getOutput().isEmpty() && !ignoreT.contains(t)) {
				if(net.removeTransition(t)) {
					++removed;
				}
			}
		}
		
		return removed;
	}
	
	
	/**
	 * TODO: description
	 *
	 * @param net
	 * @param marked Set containing all the marked places that are not in
	 *         ignoreP. Will be updated on removal.
	 * @param ignoreP
	 * @param ignoreT
	 * @return
	 */
	private int removeAlwaysMarked(final PetriNet net, final Collection<Place> marked,
			final Collection<Place> ignoreP, final Collection<Transition> ignoreT) {
		if(marked.isEmpty()) {
			return 0;
		}
		
		int removed = 0;
		
		final Map<Transition, Integer> in = new IdentityHashMap<Transition, Integer>(0);
		final Collection<Transition> produces = new IdentityHashSet<Transition>(0);
		final Collection<Transition> tests = new IdentityHashSet<Transition>(0);
		final Collection<Transition> disabled = new IdentityHashSet<Transition>(0);
        final LinkedList<Place> rm = new LinkedList<Place>();
        for (final Place p : marked) {
            for (final TPArc arc : p.getInput()) {
                in.put(arc.getSource(), arc.getWeight());
            }
            boolean alwaysMarked = true;
            for (final PTArc arc : p.getOutput()) {
                final Integer inW = in.get(arc.getTarget());
                if(inW == null || inW < arc.getWeight()) {
                    alwaysMarked = false;
                } else if(inW == arc.getWeight()) {
                    tests.add(arc.getTarget());
                } else if(inW > arc.getWeight()) {
                    produces.add(arc.getTarget());
                }
                if(arc.getWeight() > p.getMarking()) {
                    disabled.add(arc.getTarget());
                }
            }
            in.clear();
            if(alwaysMarked) {
                if(ignoreP.contains(p)) {
                    final Collection<TPArc> tparcs = new ArrayList<TPArc>(p.getInput());
                    for(final TPArc arc : tparcs) {
                        net.removeArc(arc);
                    }
                    final Collection<PTArc> ptarcs = new ArrayList<PTArc>(p.getOutput());
                    for(final PTArc arc : ptarcs) {
                        net.removeArc(arc);
                    }
                } else {
                    rm.add(p);
                }
            } else if(disabled.containsAll(produces)) {
                for(final Transition t : disabled) {
                    if(net.removeTransition(t)) {
                        ++removed;
                    }
                    tests.remove(t);
                }
                for(final Transition t : tests) {
                    net.removeArc(t, p);
                    net.removeArc(p, t);
                }
                if(p.getInput().isEmpty() && p.getOutput().isEmpty() && ignoreP.contains(p)) {
                    rm.add(p);
                }
            }
            tests.clear();
            produces.clear();
            disabled.clear();
        }
		while(!rm.isEmpty()) {
			final Place p = rm.remove();
			if(net.removePlace(p)) {
				++removed;
			}
			marked.remove(p);
		}
		return removed;
	}

	
	/**
     * TODO: description
     *
     * @param net
     * @param ignoreP
     * @param ignoreT
     * @return
     */
    private int removeIsolated(final PetriNet net, final Collection<Place> ignoreP,
            final Collection<Transition> ignoreT) {
        final LinkedList<Place> rmP = new LinkedList<Place>();
        final LinkedList<Transition> rmT = new LinkedList<Transition>();
        for(final Place p : net.getPlaces()) {
            if(ignoreP.contains(p)) {
                continue;
            }
            if(p.getInput().isEmpty() && p.getOutput().isEmpty()) {
                rmP.add(p);
            }
        }
        for(final Transition t : net.getTransitions()) {
            if(ignoreT.contains(t)) {
                continue;
            }
            if(t.getInput().isEmpty() && t.getOutput().isEmpty()) {
                rmT.add(t);
            }
        }
        int removed = 0;
        while(!rmP.isEmpty()) {
            if(net.removePlace(rmP.remove())) {
                ++removed;
            }
        }
        while(!rmT.isEmpty()) {
            if(net.removeTransition(rmT.remove())) {
                ++removed;
            }
        }
        return removed;
    }
}