/*
 * Project: Petruchio
 * File:    OmegaFReduction.java
 * Author:  Tim Strazny
 * Created: 10.10.2007
 */
package petruchio.pn.reductors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;

/**
 * An omega reduction as proposed in &quot;Slicing unbounded Petri nets&quot;
 * by A. Rakow, R. Meyer and T. Strazny, 2007, University of Oldenburg.
 * 
 * @author Tim Strazny
 */
public class OmegaFReduction implements PetriNetReductor {

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#getDescription()
     */
    @Override
    public String getDescription() {
        return "An omega reduction for falsification (backward search) as proposed in \"Slicing "
                + "unbounded Petri nets\" by A. Rakow and T. Strazny,"
                + " 2007, University of Oldenburg. It keeps LTL-x properties.";
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.petrinet.PetriNetReductor#getName()
     */
    @Override
    public String getName() {
        return "omega reduction for falsification (using backward search)";
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        // nothing to do
    }

    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public int reduce(final PetriNet net) {
        return 0;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet,
     *      java.util.Collection, java.util.Collection)
     */
    @Override
    @SuppressWarnings("unused")
	public int reduce(final PetriNet net, final Collection<Place> ignorePlaces,
            final Collection<Transition> ignoreTransitions) {
        
        if(true) throw new UnsupportedOperationException("Not implemented yet!");
        
        // test for arc weights == 1
        for(final Place p : net.getPlaces()) {
            for(final TPArc arc : p.getInput()) {
                if(arc.getWeight() != 1) {
                    return 0;
                }
            }
            for(final PTArc arc : p.getOutput()) {
                if(arc.getWeight() != 1) {
                    return 0;
                }
            }
        }
        
        final Collection<Place> ignoreP = new IdentityHashSet<Place>(ignorePlaces);
        final Collection<Transition> ignoreT = new IdentityHashSet<Transition>(ignoreTransitions);

        int removed = 0;

        int change = -1;
        int lastChange;
        do {
            lastChange = change;
            // this next reduction is performed unconditionally.
            // it does not have to produce its own fix-point as it is complex
            // and performing simpler reductions before the next run might be a
            // good idea
            int oldR = removed;
            removed += this.omegaFReduce(net, ignoreP, ignoreT);
            if (oldR != removed) {
                // if something changed: iterate!!!
                lastChange = -1;
                change = 1;
            }
        } while(lastChange != change);

        return removed;
    }
    

    /**
     * TODO: description
     * 
     * @param net
     * @param ignoreP
     * @param ignoreT
     * @return
     */
    private int omegaFReduce(final PetriNet net, final Collection<Place> ignoreP,
            final Collection<Transition> ignoreT) {
        
        if(ignoreP.isEmpty() && ignoreT.isEmpty()) {
            return 0;
        }

        final LinkedList<Place> qc = new LinkedList<Place>();
        final Collection<Place> q = new IdentityHashSet<Place>();
        final Collection<Place> p_ = new IdentityHashSet<Place>(ignoreP);
        final Collection<Transition> t_ = new IdentityHashSet<Transition>(ignoreT);
        final Collection<Transition> t_pi = new LinkedHashSet<Transition>();
        Map<Place, Integer> m = new IdentityHashMap<Place, Integer>();
        
        Map<Place, Integer> m0 = new IdentityHashMap<Place, Integer>();
        for(final Place p : net.getPlaces()) {
            if(p.getMarking() != 0) {
                m0.put(p, p.getMarking());
            }
        }
        
        for(final Place p : p_) {
            for(final PTArc arc : p.getOutput()) {
                t_.add(arc.getTarget());
            }
            for(final TPArc arc : p.getInput()) {
                t_.add(arc.getSource());
            }
        }
        
        for(final Transition t : t_) {
            for(final PTArc arc : t.getInput()) {
                qc.add(arc.getSource());
            }
        }
        
        while(!qc.isEmpty()) {
            final Place p = qc.remove();
            if(p_.contains(p) || q.contains(p)) {
                continue;
            }
            
            boolean init_seq;
            final Integer mark = m0.get(p);
            if(mark != null && mark != 0) {
                if(mark == 1) {
                    m0.remove(p);
                } else {
                    m0.put(p, mark-1);
                }
                q.add(p);
                init_seq = true;
            } else {
                boolean found = false;
                for(final TPArc arc : p.getInput()) {
                    final Transition t = arc.getSource();
                    t_pi.clear();
                    final Map<Place, Integer> m_ = this.findPC(t_, t, m0, m0, t_pi);
                    if(m_ != null) {
                        m0 = m_;
                        found = true;
                        break;
                    }
                }
                if(found) {
                    q.add(p);
                    init_seq = true;
                } else {
                    init_seq = false;
                }
            } 
            
            if(init_seq) {
                m.clear();
                for(final PTArc out : p.getOutput()) {
                    final Transition t = out.getTarget();
                    for(final PTArc in2 : t.getInput()) {
                        final Integer o = m.put(in2.getSource(), in2.getWeight());
                        if(o != null && o > in2.getWeight()) {
                            m.put(in2.getSource(), o);
                        }
                    }
                }
                
                t_pi.clear();
                m = this.findMGM(m, p, p, t_pi);
                final Map<Place, Integer> m_old = m;
                
                boolean add = false;
                if(m.containsKey(p)) {
                    final Integer newMark = m.get(p) - 1;
                    if(newMark > 0) {
                        m.put(p, newMark);
                    } else {
                        m.remove(p);
                    }
                    add = true;
                } else {
                    boolean found = false;
                    for(final TPArc arc : p.getInput()) {
                        final Transition t = arc.getSource();
                        t_pi.clear();
                        final Map<Place, Integer> m_ = this.findPC(t_, t, m, m0, t_pi);
                        if(m_ != null) {
                            m = m_;
                            found = true;
                            break;
                        }
                    }
                    if(found) {
                        add = true;
                    }
                }
                
                if(add) {
                    q.add(p);
                    for(final Place p2 : net.getPlaces()) {
                        final Integer m_V = m_old.get(p2);
                        if(m_V == null) {
                            continue;
                        }
                        final Integer mV = m.get(p2);
                        final int m_v = m_V;
                        final int mv = mV == null ? 0 : mV;
                        if(mv >= m_v) {
                            continue;
                        }
//                        final int np2 = m_v - mv;
                        // TODO t_.add(tq);
                        // TODO
                        // TODO
                        // TODO
                        // TODO
                        // TODO
                        // TODO
                    }
                    continue;
                }
            }
            
            p_.add(p);
            for (final TPArc in : p.getInput()) {
                final Transition t = in.getSource();
                t_.add(t);
                for (final PTArc in2 : t.getInput()) {
                    qc.add(in2.getSource());
                }
            }
            for (final PTArc out : p.getOutput()) {
                final Transition t = out.getTarget();
                t_.add(t);
                for (final PTArc in : t.getInput()) {
                    qc.add(in.getSource());
                }
            }
        }
        
        int removed = 0;
        final LinkedList<Place> rmP = new LinkedList<Place>();
        for(final Place p : net.getPlaces()) {
            if(!p_.contains(p)) {
                rmP.add(p);
            }
        }
        while(!rmP.isEmpty()) {
            if(net.removePlace(rmP.remove())) {
                ++removed;
            }
        }
        
        final LinkedList<Transition> rmT = new LinkedList<Transition>();
        for(final Transition t : net.getTransitions()) {
            if(!t_.contains(t)) {
                rmT.add(t);
            }
        }
        while(!rmT.isEmpty()) {
            if(net.removeTransition(rmT.remove())) {
                ++removed;
            }
        }
        
        return removed;
    }
    
    
    /**
     * TODO: Document
     *
     * @param m
     * @param p
     * @param q
     * @param t_pi
     * @return
     */
    private Map<Place, Integer> findMGM(final Map<Place, Integer> m,
            final Place p, final Place q, final Collection<Transition> t_pi) {
        final Transition t;
        Map<Place, Integer> m_ = m;
        if(p.getOutput().size() == 1 && (t = p.getOutput().iterator().next().getTarget()).getInput().size() == 1) {
            if((p == q && !t_pi.isEmpty()) || t_pi.contains(t)) {
                return m_;
            }
            final Map<Place, Integer> m1 = new IdentityHashMap<Place, Integer>(m);
            this.fire(t, m1);
            m_ = new IdentityHashMap<Place, Integer>(m1.size());
            t_pi.add(t);
            for(final TPArc out : t.getOutput()) {
                final Place p_ = out.getTarget();
                final Map<Place, Integer> m2 = this.findMGM(m1, p_, q, t_pi);
                for(final Map.Entry<Place, Integer> e : m2.entrySet()) {
                    final Integer old = m_.put(e.getKey(), e.getValue());
                    if(old != null) {
                        m_.put(e.getKey(), old + e.getValue());
                    }
                }
            }
            t_pi.remove(t);
        }
        return m_;
    }
    
    
    /**
     * TODO: Document
     *
     * @param t_
     * @param t
     * @param m
     * @param t_pi
     * @return
     */
    private Map<Place, Integer> findPC(final Collection<Transition> t_,
            final Transition t, final Map<Place, Integer> m, final Map<Place, Integer> m0,
            final Collection<Transition> t_pi) {
        
        if(m.isEmpty() || t_.contains(t)) {
            return null;
        }
        final Collection<Place> in = new IdentityHashSet<Place>();
        for(final PTArc arc : t.getInput()) {
            in.add(arc.getSource());
        }
        for(final TPArc arc : t.getOutput()) {
            if(in.contains(arc.getTarget())) {
                return null;
            }
        }
        
        if(t_pi.contains(t)) {
            final Collection<Place> pc = new IdentityHashSet<Place>(0);
            final Collection<Transition> tc = new IdentityHashSet<Transition>(0);
            this.onCycle(t, new ArrayList<Transition>(t_pi), pc, tc);
            
            if(this.isProducingTrap(pc, tc)) {
                for(final Transition c : tc) {
                    boolean enabled = true;
                    for(final PTArc arc : c.getInput()) {
                        final Place p = arc.getSource();
                        if(!pc.contains(p)) {
                            continue;
                        }
                        final Integer mark = m0.get(p);
                        if(mark == null || mark < arc.getWeight()) {
                            enabled = false;
                            break;
                        }
                    }
                    return enabled ? m : null;
                }
            }
            return null;
            
        }
        t_pi.add(t);
        Map<Place, Integer> cur_m = m;
        for(final PTArc arc : t.getInput()) {
            final Place p = arc.getSource();
            final Integer mark = cur_m.get(p);
            if(mark != null && mark >= arc.getWeight()) {
                if(mark == arc.getWeight()) {
                    cur_m.remove(p);
                } else {
                    cur_m.put(p, mark-arc.getWeight());
                }
            } else {
                for(final TPArc arc2 : p.getInput()) {
                    final Transition t2 = arc2.getSource();
                    final Map<Place, Integer> m_ = this.findPC(t_, t2, cur_m, m0, t_pi);
                    if(m_ != null) {
                        cur_m = m_;
                        break;
                    }
                }
            }
        }
        t_pi.remove(t);
        return cur_m;
    }
    
    
    /**
     * TODO: Document
     *
     * @param places
     * @param transitions
     * @return
     */
    private boolean isProducingTrap(final Collection<Place> places,
            final Collection<Transition> transitions) {
        // test whether it produces token if it's a trap
        boolean produces = false;
        for(final Transition t : transitions) {
            int balance = 0;
            for(final PTArc arc : t.getInput()) {
                if(!places.contains(arc.getSource())) {
                    continue;
                }
                balance -= arc.getWeight();
            }
            for(final TPArc arc : t.getOutput()) {
                if(!places.contains(arc.getTarget())) {
                    continue;
                }
                balance += arc.getWeight();
            }
            if(balance > 0) {
                produces = true;
                break;
            }
        }
        if(!produces) {
            return false;
        }
        
        // test whether or not it's a trap
        // TODO consider only t in the transitions-set?
        final Collection<Transition> post = new IdentityHashSet<Transition>();
        for(final Place p : places) {
            for(final PTArc arc : p.getOutput()) {
                post.add(arc.getTarget());
            }
        }
        for(final Place p : places) {
            for(final TPArc arc : p.getInput()) {
                if(!post.contains(arc.getSource())) {
                    return false; // not a trap
                }
            }
        }
        return true;
    }
    
    
    /**
     * TODO: Document
     *
     * @param t
     * @param path
     * @param pc
     * @param tc
     */
    private void onCycle(final Transition t, final List<Transition> path,
            final Collection<Place> pc, final Collection<Transition> tc) {
        final Collection<Place> toAdd = new ArrayList<Place>(0);
        final Collection<Place> pre = new IdentityHashSet<Place>(0);
        
        for(int i = path.size()-1; i >= 0; --i) {
            final Transition cur = path.get(i);
            for(final TPArc arc : cur.getOutput()) {
                toAdd.add(arc.getTarget());
            }
            final Transition next = path.get((cur == t) ? path.size()-1 : i-1);
            for(final PTArc arc : next.getInput()) {
                pre.add(arc.getSource());
            }
            toAdd.retainAll(pre);
            pc.addAll(toAdd);
            tc.add(cur);
            if(cur == t) {
                break;
            }
        }
    }
    
    
    /**
     * TODO: Document
     *
     * @param t
     * @param m
     */
    private void fire(final Transition t, final Map<Place, Integer> m) {
        for(final TPArc out : t.getOutput()) {
            final Place pl = out.getTarget();
            final Integer old = m.put(pl, out.getWeight());
            if(old != null) {
                m.put(pl, old + out.getWeight());
            }
        }
        for(final PTArc in : t.getInput()) {
            final Place pl = in.getSource();
            final Integer old = m.get(pl);
            if(old == null || old < in.getWeight()) {
                throw new InternalError("Simulated firing of a transitin that is not enabled: " + pl + " -> " + t + ": " + old +  " < " + in.getWeight()); 
            } else if(old == in.getWeight()) {
                m.remove(pl);
            } else {
                m.put(pl, old - in.getWeight());
            }
        }
    }
    
}