/*
 * Project:  Petruchio
 * Filename: OmegaFReductionBPP.java
 * Created:  25.09.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.reductors;

import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.Map;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;

/**
 * An omega reduction as proposed in &quot;Slicing unbounded Petri nets&quot;
 * by A. Rakow, R. Meyer and T. Strazny, 2007, University of BPPenburg. It keeps
 * LTL-x properties for falsification.
 * 
 * @author Tim Strazny
 */
public class OmegaFReductionBPP implements PetriNetReductor {

    final Map<Object, SCC> SCCs = new IdentityHashMap<Object, SCC>();
    

    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.petrinet.PetriNetReductor#getDescription()
     */
    @Override
    public String getDescription() {
        return "An omega reduction for falsification as proposed in \"Slicing "
                + "unbounded Petri nets\" by A. Rakow, R. Meyer and T. Strazny,"
                + " 2007, University of BPPenburg. It keeps LTL-x properties.";
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.petrinet.PetriNetReductor#getName()
     */
    @Override
    public String getName() {
        return "omega reduction for falsification (using BPP sub nets)";
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        for(final SCC scc : this.SCCs.values()) {
            scc.clear();
        }
        this.SCCs.clear();
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public int reduce(final PetriNet net) {
        return 0;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet,
     *      java.util.Collection, java.util.Collection)
     */
    @Override
    public int reduce(final PetriNet net, final Collection<Place> ignorePlaces,
            final Collection<Transition> ignoreTransitions) {
        if (ignorePlaces.isEmpty() && ignoreTransitions.isEmpty()) {
            return 0;
        }
        
        final Collection<Place> ignoreP = new IdentityHashSet<Place>(ignorePlaces);
        final Collection<Transition> ignoreT = new IdentityHashSet<Transition>(ignoreTransitions);

        int removed = 0;

        int change = -1;
        int lastChange;
        do {
            lastChange = change;
            // this next reduction is performed unconditionally.
            // it does not have to produce its own fix-point as it is complex
            // and performing simpler reductions before the next run might be a
            // good idea
            {
                int oldR = removed;
                removed += this.omegaReduce(net, ignoreP, ignoreT);
                if (oldR != removed) {
                    // if something changed: iterate!!!
                    lastChange = -1;
                    change = 1;
                }
            }
        } while(lastChange != change);

        return removed;
    }
         

    /**
     * TODO: description
     * 
     * @param net
     * @param ignoreP
     * @param ignoreT
     * @return
     */
    private int omegaReduce(final PetriNet net, final Collection<Place> ignoreP,
            final Collection<Transition> ignoreT) {

        final Collection<Place> analyzed = new IdentityHashSet<Place>();
        int removed = 0;
        for (final Place p : net.getPlaces()) {
            if(p.getMarking() == 0 || ignoreP.contains(p)) {
                continue;
            }
            removed = this.buildAndRemoveN2(net, p, analyzed, ignoreP, ignoreT);
            if (removed != 0) {
                return removed;
            }
        }
        
        return removed;
    }


    /**
     * TODO: description
     *
     * @param net
     * @param start
     * @param analyzed
     * @return
     */
    private int buildAndRemoveN2(final PetriNet net, final Place start,
            final Collection<Place> analyzed,
            final Collection<Place> ignoreP, final Collection<Transition> ignoreT) {
        if(analyzed.contains(start)) {
            return 0;
        }

        final Collection<Place> places = new IdentityHashSet<Place>();
        final Collection<Transition> transitions = new IdentityHashSet<Transition>();
        final LinkedList<Place> toTreat = new LinkedList<Place>();
        final LinkedList<Transition> path = new LinkedList<Transition>();
        final Collection<Place> pathSet = new IdentityHashSet<Place>();
        final Collection<SCC> g = new IdentityHashSet<SCC>();
        final Collection<Place> cycle = new IdentityHashSet<Place>();
        places.add(start);
        toTreat.add(start);
        new SCC(start);
        analyzed.add(start);
        
        final int marking = start.getMarking();
        boolean finished = false;
        
//        System.out.println("ENTERING MAIN LOOP");
        while (!toTreat.isEmpty() && !finished) {
            final Place p = toTreat.getFirst();
//            System.out.println("exploring place " + p);

            Transition t = null;
            for (final PTArc arc : p.getOutput()) {
                final Transition candidate = arc.getTarget();
                if ((candidate.getInput().size() == 1) // only one place in pre set
                        // simple arc to place in pre set
                        && (candidate.getInput().iterator().next().getWeight() == 1)
                        // check/add transition (yes, this adds the candidate)
                        && transitions.add(candidate)) {
                    t = candidate;
                    break;
                }
            }
            pathSet.add(p);
            
            if (t == null) {
                // no more output arcs of this place to check
                pathSet.remove(p);
                toTreat.removeFirst();
//                System.out.println("done with " + p + ", " + toTreat);
                continue;
            }
            
            path.add(t);
            
//            System.out.println("exploring trans " + t);
            
            SCC sccP = this.SCCs.get(p);
            final SCC sccT = new SCC(t, sccP);
            
//            System.out.println("new descendant for " + sccP);
            
            boolean produces = false;
            for (final TPArc arc : t.getOutput()) {
                final Place target = arc.getTarget();
                if(places.add(target)) {
//                    System.out.println("adding " + target);
                    new SCC(target, sccT);
                    toTreat.addFirst(target);
                    // TODO: is this correct?
                    if(target.getMarking() != 0) {
                        analyzed.add(target);
                    }
                } else if(pathSet.contains(target)) {
                    if(!cycle.isEmpty() || (arc.getWeight() > 1)) {
                        produces = true;
                    }
                    cycle.add(target);
                    this.SCCs.get(p).addTransPredecessor(sccT);
//                    System.out.println("cycle with " + target);
                } else {
                    final SCC succ = this.SCCs.get(p);
                    succ.addTransPredecessor(sccT);
//                    System.out.println("marking " + target);
                }
            }
            
            if(!cycle.isEmpty()) {
//                System.out.println("cycle detected: " + path + ", " + cycle);
                int i = path.size()-1;
                do {
                    final Transition pt = path.get(i);
                    final SCC nextSCC = this.SCCs.get(pt);
//                    System.out.println("added " + nextSCC);
                    sccT.merge(nextSCC);
                    --i;
                    for(final PTArc arc : pt.getInput()) {
                        final SCC nextSCC2 = this.SCCs.get(arc.getSource()); 
//                        System.out.println("added " + nextSCC2);
                        sccT.merge(nextSCC2);
                        cycle.remove(arc.getSource());
                    }
                } while(i >= 0 && !cycle.isEmpty());
                sccT.setProduces(produces);
//                System.out.println("built scc " + sccT);
            }
//            System.out.println("succ: " + sccT.getPlaceSuccessors());
//            System.out.println("osucc: " + sccT.getTransSuccessors());
            
            g.addAll(this.SCCs.values());
            final Collection<SCC> minSCCs = this.getMinimalSCCs(start);
//            System.out.println("minimal SCCs: " + minSCCs);
            final SCC[] min = minSCCs.toArray(new SCC[minSCCs.size()]);
            if(min.length <= marking) {
//                System.out.println("building Q" + g);
                if(this.buildQ(min, g, places, transitions, ignoreP, ignoreT)) {
                    finished = true;
                }
            } else {
                final SCC[] use = new SCC[marking];
                final int[] used = new int[min.length];
                if(this.enumerateSCCsAndBuildQ(min, use, used, 0, g, places,
                        transitions, ignoreP, ignoreT)) {
                    finished = true;
                }
            }
            g.clear();
        }

        for(final SCC scc : this.SCCs.values()) {
            scc.clear();
        }
        this.SCCs.clear();
        
        int removed = 0;
        if(finished) {
            for(final Place p : places) {
                if(net.removePlace(p)) {
                    ++removed;
                }
            }
            for(final Transition t : transitions) {
                if(net.removeTransition(t)) {
                    ++removed;
                }
            }
        }

        return removed;
    }
    
    
    /**
     * TODO: description
     *
     * @param cfs
     * @param use
     * @param used
     * @param pos
     * @param g
     * @param places
     * @param transitions
     * @return
     */
    private boolean enumerateSCCsAndBuildQ(final SCC[] cfs, final SCC[] use,
            final int[] used, final int pos, final Collection<SCC> g,
            final Collection<Place> places, final Collection<Transition> transitions,
            final Collection<Place> ignoreP, final Collection<Transition> ignoreT) {
        if(pos >= cfs.length) {
            return this.buildQ(cfs, g, places, transitions, ignoreP, ignoreT);
        }
        if(pos == 0) {
            for(int i = 0; i < use.length; ++i) {
                used[pos] = i;
                cfs[pos] = use[i];
                this.enumerateSCCsAndBuildQ(cfs, use, used, pos+1, g, places,
                        transitions, ignoreP, ignoreT);
            }
        } else {
            for(int i = used[pos-1]+1; i < use.length; ++i) {
                used[pos] = i;
                cfs[pos] = use[i];
                this.enumerateSCCsAndBuildQ(cfs, use, used, pos+1, g, places,
                        transitions, ignoreP, ignoreT);
            }
        }
        return false;
    }
    
    
    /**
     * TODO: description
     *
     * @param cfs
     * @param g
     * @param places
     * @param transitions
     * @return
     */
    private boolean buildQ(final SCC[] cfs, final Collection<SCC> g,
            final Collection<Place> places, final Collection<Transition> transitions,
            final Collection<Place> ignoreP, final Collection<Transition> ignoreT) {
        final Collection<Place> q = new IdentityHashSet<Place>();
        
        for(final SCC c : cfs) {
            for(final SCC c2 : g) {
//                System.out.print(c + " <=? " + c2);
                if(c.leq(c2)) {
//                    System.out.println("yes");
                    q.addAll(c2.getPlaces());
                } else {
//                    System.out.println("no");
                }
            }
            if(c.produces()) {
                q.addAll(c.getPlaces());
            }
        }
        
//        System.out.println("" + places + transitions + q);
        if(this.checkArcs(places, transitions, q, ignoreP, ignoreT)) {
            places.addAll(q);
            return true;
        }
        return false;
    }
    
    
    /**
     * TODO: description
     *
     * @param start
     * @return
     */
    private Collection<SCC> getMinimalSCCs(final Place start) {
        final Collection<SCC> min = new IdentityHashSet<SCC>();
        final Collection<SCC> visited = new IdentityHashSet<SCC>();
        final LinkedList<SCC> stack = new LinkedList<SCC>();
        stack.addFirst(this.SCCs.get(start));
        while(!stack.isEmpty()) {
            final SCC root = stack.removeFirst();
            if(!visited.add(root)) {
                continue;
            } else if(root.produces()) {
                min.add(root);
            } else {
                if(root.producesOutput()) {
                    min.add(root);
                } else {
                    stack.addAll(root.getTransSuccessors());
                }
                stack.addAll(root.getPlaceSuccessors());
            }
        }
        return min;
    }


    /**
     * TODO: description
     * 
     * @param places
     * @param transitions
     * @param q
     * @return
     */
    private boolean checkArcs(final Collection<Place> places,
            final Collection<Transition> transitions, final Collection<Place> q,
            final Collection<Place> ignoreP, final Collection<Transition> ignoreT) {
        for (final Place p : places) {
            if(ignoreP.contains(p)) {
                return false;
            }
            if (q.contains(p)) {
                continue;
            }
            for (final PTArc arc : p.getOutput()) {
                if (!transitions.contains(arc.getTarget())) {
                    return false;
                }
            }
        }
        for (final Transition t : transitions) {
            if(ignoreT.contains(t)) {
                return false;
            }
            for (final PTArc arc : t.getInput()) {
                if (!places.contains(arc.getSource())) {
                    return false;
                }
            }
            for (final TPArc arc : t.getOutput()) {
                if (!places.contains(arc.getTarget())) {
                    return false;
                }
            }
        }
        return true;
    }


//    /**
//     * TODO: description
//     * 
//     * @param net
//     * @param ignorePlaces
//     * @param ignoreTransitions
//     * @param marked
//     * @return
//     */
//    private int removeOutput(final PetriNet net, final Collection<Place> marked,
//            final Collection<Place> ignorePlaces, final Collection<Transition> ignoreTransitions) {
//
//        final LinkedList<Place> toTreatP = new LinkedList<Place>(ignorePlaces);
//        final LinkedList<Transition> toTreatT = new LinkedList<Transition>(ignoreTransitions);
//        
//        final Collection<Place> ignoreP = new IdentityHashSet<Place>();
//        final Collection<Transition> ignoreT = new IdentityHashSet<Transition>();
//        final Map<Transition, Integer> in = new IdentityHashMap<Transition, Integer>();
//        
//        while(!toTreatP.isEmpty() || !toTreatT.isEmpty()) {
//            while(!toTreatP.isEmpty()) {
//                final Place p = toTreatP.remove();
//                if(!ignoreP.add(p)) {
//                    continue;
//                }
//                for(final TPArc arc : p.getInput()) {
//                    in.put(arc.getSource(), arc.getWeight());
//                }
//                boolean removeOutput = true;
//                for(final PTArc arc : p.getOutput()) {
//                    final Integer w = in.remove(arc.getTarget());
//                    if(w == null || w < arc.getWeight()) {
//                        removeOutput = false;
//                        break;
//                    }
//                }
//                if(removeOutput) {
//                    toTreatT.addAll(in.keySet());
//                } else {
//                    for(final TPArc arc : p.getInput()) {
//                        toTreatT.add(arc.getSource());
//                    }
//                    for(final PTArc arc : p.getOutput()) {
//                        toTreatT.add(arc.getTarget());
//                    }
//                }
//            }
//            while(!toTreatT.isEmpty()) {
//                final Transition t = toTreatT.remove();
//                if(!ignoreT.add(t)) {
//                    continue;
//                }
//                for(final PTArc arc : t.getInput()) {
//                    toTreatP.add(arc.getSource());
//                }
//            }
//        }
//        
//
//        int removed = 0;
//        final LinkedList<Place> rmP = new LinkedList<Place>(net.getPlaces());
//        while(!rmP.isEmpty()) {
//            final Place p = rmP.remove();
//            if(!ignoreP.contains(p) && net.removePlace(p)) {
//                ++removed;
//                marked.remove(p);
//            }
//        }
//        final LinkedList<Transition> rmT = new LinkedList<Transition>(net.getTransitions());
//        while(!rmT.isEmpty()) {
//            final Transition t = rmT.remove();
//            if(!ignoreT.contains(t) && net.removeTransition(t)) {
//                ++removed;
//            }
//        }
//        return removed;
//    }
//
//
//    /**
//     * TODO: description
//     * 
//     * @param net
//     * @param marked
//     * @param ignorePlaces
//     * @param ignoreTransitions
//     * @return
//     */
//    private int removeSimpleUnbounded(final PetriNet net,
//            final Collection<Place> marked, final Collection<Place> ignoreP,
//            final Collection<Transition> ignoreT) {
//        final LinkedList<Transition> toTreat = new LinkedList<Transition>();
//        int removed = 0;
//
//        for (final Transition t : net.getTransitions()) {
//            if (t.getInput().isEmpty()) {
//                toTreat.add(t);
//            }
//        }
//
//        final LinkedList<Place> rm = new LinkedList<Place>();
//        while (!toTreat.isEmpty()) {
//            final Transition t = toTreat.remove();
//            for (final TPArc arc : t.getOutput()) {
//                final Place p = arc.getTarget();
//                if (!ignoreP.contains(p)) {
//                    for (final PTArc arc2 : p.getOutput()) {
//                        final Transition t2 = arc2.getTarget();
//                        if (t2.getInput().size() == 1) {
//                            toTreat.add(t2);
//                        }
//                    }
//                    rm.add(p);
//                }
//            }
//            while (!rm.isEmpty()) {
//                final Place p = rm.remove();
//                if (net.removePlace(p)) {
//                    marked.remove(p);
//                    ++removed;
//                }
//            }
//            // remove transition if isolated
//            if (t.getOutput().isEmpty() && !ignoreT.contains(t)) {
//                if (net.removeTransition(t)) {
//                    ++removed;
//                }
//            }
//        }
//
//        return removed;
//    }
//
//
//    /**
//     * TODO: description
//     *
//     * @param net
//     * @param marked Set containing all the marked places that are not in
//     *         ignoreP. Will be updated on removal.
//     * @param ignoreP
//     * @param ignoreT
//     * @return
//     */
//    private int removeAlwaysMarked(final PetriNet net,
//            final Collection<Place> marked, final Collection<Place> ignoreP,
//            final Collection<Transition> ignoreT) {
//        if (marked.isEmpty()) {
//            return 0;
//        }
//
//        final Map<Transition, Integer> in = new IdentityHashMap<Transition, Integer>();
//        final LinkedList<Place> rm = new LinkedList<Place>();
//        for (final Place p : marked) {
//            for (final TPArc arc : p.getInput()) {
//                in.put(arc.getSource(), arc.getWeight());
//            }
//            boolean remove = true;
//            for (final PTArc arc : p.getOutput()) {
//                final Integer inW = in.get(arc.getTarget());
//                if(inW == null || inW < arc.getWeight()) {
//                    remove = false;
//                    break;
//                }
//            }
//            in.clear();
//            if(remove) {
//                if(ignoreP.contains(p)) {
//                    final Collection<TPArc> tparcs = new ArrayList<TPArc>(p.getInput());
//                    for(final TPArc arc : tparcs) {
//                        net.removeArc(arc);
//                    }
//                    final Collection<PTArc> ptarcs = new ArrayList<PTArc>(p.getOutput());
//                    for(final PTArc arc : ptarcs) {
//                        net.removeArc(arc);
//                    }
//                } else {
//                    rm.add(p);
//                }
//            }
//        }
//        int removed = 0;
//        while (!rm.isEmpty()) {
//            final Place p = rm.remove();
//            if (net.removePlace(p)) {
//                ++removed;
//            }
//            marked.remove(p);
//        }
//        return removed;
//    }

    
    /**
     * TODO: description 
     *
     * @author Tim Strazny
     */
    private class SCC {
        private final Collection<Place> places = new IdentityHashSet<Place>();
        private final Collection<Transition> transitions = new IdentityHashSet<Transition>();
        private final Collection<SCC> predecessors = new IdentityHashSet<SCC>();
        private final Collection<SCC> transPredecessors = new IdentityHashSet<SCC>();
        private final Collection<SCC> successors = new IdentityHashSet<SCC>();
        private final Collection<SCC> transSuccessors = new IdentityHashSet<SCC>();
        private boolean produces = false;
        
        
        public SCC(final Place p) {
            this.places.add(p);
            OmegaFReductionBPP.this.SCCs.put(p, this);
        }
        
        
        public SCC(final Transition t, final SCC pre) {
            this.transitions.add(t);
            this.predecessors.add(pre);
            pre.successors.add(this);
            OmegaFReductionBPP.this.SCCs.put(t, this);
        }
        
        
        public SCC(final Place p, final SCC pre) {
            this.places.add(p);
            this.addTransPredecessor(pre);
            OmegaFReductionBPP.this.SCCs.put(p, this);
        }
        
        
        public void addTransPredecessor(final SCC scc) {
            this.transPredecessors.add(scc);
            scc.transSuccessors.add(this);
        }
        

        public boolean produces() {
            return this.produces;
        }
        
        
        public boolean producesOutput() {
            return !this.transSuccessors.isEmpty()
                    && (this.places.size() + this.transitions.size() > 1);
        }
        
        
        public void setProduces(final boolean p) {
            this.produces = p;
        }

        
        public Collection<SCC> getPlaceSuccessors() {
            return this.successors;
        }
        
        
        public Collection<SCC> getTransSuccessors() {
            return this.transSuccessors;
        }
        
        
        public Collection<Place> getPlaces() {
            return this.places;
        }
        
        
        public boolean leq(final SCC scc) {
            if(this.produces()) {
                return this.hasDescendant(scc);
            } else if(this.producesOutput()) {
                return this.hasOutputDescendant(scc);
            } else {
                return false;
            }
        }
        
        
        private boolean hasDescendant(final SCC scc) {
            final LinkedList<SCC> stack = new LinkedList<SCC>();
            final Collection<SCC> visited = new IdentityHashSet<SCC>();
            stack.add(this);
            while(!stack.isEmpty()) {
                final SCC root = stack.removeFirst();
                if(!visited.add(root)) {
                    continue;
                }
                if(root.successors.contains(scc) || root.transSuccessors.contains(scc)) {
                    return true;
                }
                stack.addAll(root.successors);
                stack.addAll(root.transSuccessors);
            }
            return false;
        }
        
        
        private boolean hasOutputDescendant(final SCC scc) {
            if(this.transSuccessors.contains(scc)) {
                return true;
            }
            final LinkedList<SCC> stack = new LinkedList<SCC>();
            final Collection<SCC> visited = new IdentityHashSet<SCC>();
            for(final SCC succ : this.transSuccessors) {
                stack.add(succ);
                while(!stack.isEmpty()) {
                    final SCC root = stack.removeFirst();
                    if(!visited.add(root)) {
                        continue;
                    }
                    if(root.successors.contains(scc) || root.transSuccessors.contains(scc)) {
                        return true;
                    }
                    stack.addAll(root.successors);
                    stack.addAll(root.transSuccessors);
                }
            }
            return false;
        }
        
        
//        public boolean contains(final Place p) {
//            return this.places.contains(p);
//        }
        
        
        public void clear() {
            this.places.clear();
            this.transitions.clear();
            this.predecessors.clear();
            this.transPredecessors.clear();
            this.successors.clear();
            this.transSuccessors.clear();
            this.produces = false;
        }
        
        
        public void merge(final SCC scc) {
            if(scc == this) {
                return;
            }
            for(final Place p : scc.places) {
                OmegaFReductionBPP.this.SCCs.put(p, this);
                this.places.add(p);
            }
            for(final Transition t : scc.transitions) {
                OmegaFReductionBPP.this.SCCs.put(t, this);
                this.transitions.add(t);
            }
            for(final SCC next : scc.successors) {
                next.predecessors.remove(scc);
                if(next != this) {
                    next.predecessors.add(this);
                    this.successors.add(next);
                }
            }
            for(final SCC next : scc.transSuccessors) {
                next.transPredecessors.remove(scc);
                if(next != this) {
                    next.transPredecessors.add(this);
                    this.transSuccessors.add(next);
                } else {
                    this.produces = true;
                }
            }
            for(final SCC prev : scc.predecessors) {
                prev.successors.remove(scc);
                if(prev != this) {
                    prev.successors.add(this);
                    this.predecessors.add(prev);
                }
            }
            for(final SCC prev : scc.transPredecessors) {
                prev.transSuccessors.remove(scc);
                if(prev != this) {
                    prev.transSuccessors.add(this);
                    this.transPredecessors.add(prev);
                }
            }
            this.produces |= scc.produces;
            scc.clear();
        }
        
        
        @Override
        public String toString() {
            return this.places.toString() + " + " + this.transitions.toString()
                    + (this.produces ? "p" : (!this.producesOutput() ? "o" : ""));
        }
    }

}