/*
 * Project: Petruchio
 * File:    OmegaVReduction.java
 * Author:  Tim Strazny
 * Created: 10.10.2007
 */
package petruchio.pn.reductors;

import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.Map;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;

/**
 * An omega reduction as proposed in &quot;Slicing unbounded Petri nets&quot;
 * by A. Rakow, R. Meyer and T. Strazny, 2007, University of Oldenburg.
 * 
 * @author Tim Strazny
 */
public class OmegaVReduction implements PetriNetReductor {

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#getDescription()
     */
    @Override
    public String getDescription() {
        return "An omega reduction for verification (backward search) as proposed in \"Slicing "
                + "unbounded Petri nets\" by A. Rakow and T. Strazny,"
                + " 2007, University of Oldenburg. It keeps LTL-x properties.";
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.petrinet.PetriNetReductor#getName()
     */
    @Override
    public String getName() {
        return "omega reduction for verification (using backward search)";
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        // nothing to do
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public int reduce(final PetriNet net) {
        return 0;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet,
     *      java.util.Collection, java.util.Collection)
     */
    @Override
    public int reduce(final PetriNet net, final Collection<Place> ignorePlaces,
            final Collection<Transition> ignoreTransitions) {
        
        // test for arc weights == 1
        for(final Place p : net.getPlaces()) {
            for(final TPArc arc : p.getInput()) {
                if(arc.getWeight() != 1) {
                    return 0;
                }
            }
            for(final PTArc arc : p.getOutput()) {
                if(arc.getWeight() != 1) {
                    return 0;
                }
            }
        }
        
        final Collection<Place> ignoreP = new IdentityHashSet<Place>(ignorePlaces);
        final Collection<Transition> ignoreT = new IdentityHashSet<Transition>(ignoreTransitions);

        int removed = 0;

        int change = -1;
        int lastChange;
        do {
            lastChange = change;
            // this next reduction is performed unconditionally.
            // it does not have to produce its own fix-point as it is complex
            // and performing simpler reductions before the next run might be a
            // good idea
            int oldR = removed;
            removed += this.omegaVReduce(net, ignoreP, ignoreT);
            if (oldR != removed) {
                // if something changed: iterate!!!
                lastChange = -1;
                change = 1;
            }
        } while(lastChange != change);

        return removed;
    }
            
            
    /**
     * TODO: description
     * 
     * @param net
     * @param ignoreP
     * @param ignoreT
     * @return
     */
    private int omegaVReduce(final PetriNet net, final Collection<Place> ignoreP,
            final Collection<Transition> ignoreT) {
        
        // no scope? don't reduce.
        if(ignoreP.isEmpty() && ignoreT.isEmpty()) {
            return 0;
        }

        // data that is used later in the algorithm
        final LinkedList<Transition> sigma = new LinkedList<Transition>();
        
        // build support of the initial marking
        final Map<Place, Integer> m0 = new IdentityHashMap<Place, Integer>();
        for(final Place p : net.getPlaces()) {
            if(p.getMarking() != 0) {
                m0.put(p, p.getMarking());
            }
        }
        
        // THE ALGORITHM STARTS HERE
        
        // add the pre and post sets of places in the scope of the formula to T'
        // the transition scope has already been added to T'
        final Collection<Transition> t_prime = new IdentityHashSet<Transition>(ignoreT);
        for(final Place p : ignoreP) {
            for(final PTArc arc : p.getOutput()) {
                t_prime.add(arc.getTarget());
            }
            for(final TPArc arc : p.getInput()) {
                t_prime.add(arc.getSource());
            }
        }
        
        // C_temp = pre set of T'
        final LinkedList<Place> c_temp = new LinkedList<Place>();
        for(final Transition t : t_prime) {
            for(final PTArc arc : t.getInput()) {
                c_temp.add(arc.getSource());
            }
        }
        
        // P' = emptyset
        // this is changed since places in ignoreP must not be deleted!
        final Collection<Place> p_prime = new IdentityHashSet<Place>(ignoreP);
        
        // C = emptyset
        final Collection<Place> c = new IdentityHashSet<Place>();
        
        // this loop differs from the original algorithm only that candidates
        // are removed from C_temp if they are in P' or C.
        // exists q in C_temp setminus (P' union C)
        while(!c_temp.isEmpty()) {
            final Place q = c_temp.remove();
            if(p_prime.contains(q) || c.contains(q)) {
                continue;
            }
            // now: q in C_temp setminus (P' union C)
            
            if(this.onb(q) && this.syn(q)) {
                if(this.isInitMarkable(q, sigma, t_prime)) {
                    final Map<Place, Integer> m_shift = this.shiftToken(q, t_prime);
                    if(this.isMarkable(q, m_shift, sigma, t_prime)) {
                        c.add(q);
                        continue;
                    }
                }
            }
            
            p_prime.add(q);
            // T' = T' union *q union q*
            // C_temp = C_temp union *(*q)) union *(q*)
            for(final TPArc in : q.getInput()) {
                final Transition t = in.getSource();
                t_prime.add(t);
                for(final PTArc in2 : t.getInput()) {
                    c_temp.add(in2.getSource());
                }
            }
            for(final PTArc out : q.getOutput()) {
                final Transition t = out.getTarget();
                t_prime.add(t);
                for(final PTArc in : t.getInput()) {
                    c_temp.add(in.getSource());
                }
            }
        }
        
        // now reduce net to P', T' by removing all places not in P' and removing
        // all transitions not in T'. (this is done below...)
        
        // MAIN ALGORITHM ENDS HERE
        
        // here goes the removal of selected elements.
        
        // we count how many elements are removed as demanded by the interface. 
        int removed = 0;
        final LinkedList<Place> rmP = new LinkedList<Place>();
        for(final Place p : net.getPlaces()) {
            if(!p_prime.contains(p)) {
                rmP.add(p);
            }
        }
        while(!rmP.isEmpty()) {
            if(net.removePlace(rmP.remove())) {
                ++removed;
            }
        }
        
        final LinkedList<Transition> rmT = new LinkedList<Transition>();
        for(final Transition t : net.getTransitions()) {
            if(!t_prime.contains(t)) {
                rmT.add(t);
            }
        }
        while(!rmT.isEmpty()) {
            if(net.removeTransition(rmT.remove())) {
                ++removed;
            }
        }
        
        return removed;
    }
    
    
    /**
     * TODO: Document
     *
     * @param q
     * @param sigma
     * @param t_prime
     * @return
     */
    private boolean isInitMarkable(final Place q,
            final LinkedList<Transition> sigma, final Collection<Transition> t_prime) {
        // direct access to the initial Petri net marking!
        // if M_0(q) >= 1
        if(q.getMarking() > 0) {
            return true;
        }
        // or exists t in *p : findFs(...)
        for(final TPArc arc : q.getInput()) {
            final Transition t = arc.getSource();
            
            // call findInitFs
            if(this.findInitFs(t, sigma, t_prime)) {
                return true;
            }
        }
        return false;
    }
    
    
    private boolean isMarkable(final Place q, final Map<Place, Integer> marking,
            final LinkedList<Transition> sigma, final Collection<Transition> t_prime) {
        // do not use initial Petri net marking but given map
        
        // if M(q) >= 1
        if(marking.containsKey(q)) {
            return true;
        }
        // or exists t in *p : findFs(...)
        for(final TPArc arc : q.getInput()) {
            final Transition t = arc.getSource();
            
            // call findFs
            if(this.findFs(t, marking, sigma, t_prime)) {
                return true;
            }
        }
        return false;
    }
    
    
    private boolean findInitFs(final Transition t, 
            final LinkedList<Transition> sigma, final Collection<Transition> t_prime) {
        // if t in T' or t in transitions(sigma) or not onb(*t) then return false
        if(t_prime.contains(t) || sigma.contains(t) || !this.onb(t.getInput())) {
            return false;
        }

        // return for all p in *t : isInitMarkable(...)
        boolean allMarkable = true;
        sigma.addFirst(t);
        for(final PTArc arc : t.getInput()) {
            final Place p = arc.getSource();
            if(!this.isInitMarkable(p, sigma, t_prime)) {
                allMarkable = false;
                break;
            }
        }
        sigma.removeFirst();
        return allMarkable;
    }
    
    
    private boolean findFs(final Transition t, final Map<Place, Integer> marking,
            final LinkedList<Transition> sigma, final Collection<Transition> t_prime) {
        // if t in T' or not onb(*t) then return false
        if(t_prime.contains(t) || !this.onb(t.getInput())) {
            return false;
        }

        // if t in transitions(sigma)
        if(sigma.contains(t)) {
            // get cycle transitions and cycle places
            final Collection<Transition> t_cycle = new IdentityHashSet<Transition>(0);
            final Collection<Place> p_cycle = new IdentityHashSet<Place>(0);
            this.getCycle(t, sigma, t_cycle, p_cycle);
            
            // if exists t' in T_cycle :
            //   for all p in *t intersect P_cycle :
            //     M_0(p) > 0 then return true
            for(final Transition tc : t_cycle) {
                boolean allMarked = true;
                for(final PTArc arc : tc.getInput()) {
                    final Place p = arc.getSource();
                    if(!p_cycle.contains(p)) {
                        continue;
                    }
                    if(p.getMarking() == 0) {
                        allMarked = false;
                        break;
                    }
                }
                if(allMarked) {
                    return true;
                }
            }
            
            // if exists t' in T_cycle :
            //   for all p in *t intersect P_cycle :
            //     exists t'' in *p setminus T_cycle :
            //       findInitFs(t'', epsilon, T') then return true
            LinkedList<Transition> sigma_prime = null;
            for(final Transition tc : t_cycle) {
                boolean allMarkable = true;
                for(final PTArc arc : tc.getInput()) {
                    final Place p = arc.getSource();
                    if(!p_cycle.contains(p)) {
                        continue;
                    }
                    boolean isMarkable = false;
                    for(final TPArc arc2 : p.getInput()) {
                        final Transition t2 = arc2.getSource();
                        if(t_cycle.contains(t2)) {
                            continue;
                        }
                        if(sigma_prime == null) {
                            sigma_prime = new LinkedList<Transition>();
                        }
                        if(!this.findInitFs(t2, sigma_prime, t_prime)) {
                            isMarkable = true;
                            break;
                        }
                    }
                    if(!isMarkable) {
                        allMarkable = false;
                        break;
                    }
                }
                if(allMarkable) {
                    return true;
                }
            }
            return false;
        }

        // return for all p in *t : isMarkable(...)
        boolean allMarkable = true;
        sigma.addFirst(t);
        for(final PTArc arc : t.getInput()) {
            final Place p = arc.getSource();
            if(!this.isMarkable(p, marking, sigma, t_prime)) {
                allMarkable = false;
                break;
            }
        }
        sigma.removeFirst();
        return allMarkable;
    }
    
    
    private Map<Place, Integer> shiftToken(final Place q, final Collection<Transition> t_prime) {
        // M = enabling(q*)
        final Map<Place, Integer> m = new IdentityHashMap<Place, Integer>(q.getOutput().size());
        for(final PTArc out : q.getOutput()) {
            final Transition t = out.getTarget();
            for(final PTArc in : t.getInput()) {
                final Place p = in.getSource();
                m.put(p, 1);
            }
        }
        
        final LinkedList<Transition> t_done = new LinkedList<Transition>(t_prime);
        next_candidate:
        while(!t_done.isEmpty()) {
            final Transition t = t_done.remove();
            for(final PTArc arc : t.getInput()) {
                final Place p = arc.getSource();
                if(!this.onb(p)) {
                    continue next_candidate;
                }
            }
            if(!this.isEnabled(m, t)) {
                continue next_candidate;
            }
            
            this.fire(t, m);
        }
        
        return m;
    }
    
    
    private boolean isEnabled(final Map<Place, Integer> marking, final Transition t) {
        for(final PTArc arc : t.getInput()) {
            final Place p = arc.getSource();
            final Integer m = marking.get(p);
            if((m == null) || (arc.getWeight() > m)) {
                return false;
            }
        }
        return true;
    }
    
    
    /**
     * @param t
     * @param sigma
     * @param t_cycle
     * @param p_cycle
     */
    private void getCycle(final Transition t, final LinkedList<Transition> sigma,
            final Collection<Transition> t_cycle, final Collection<Place> p_cycle) {
        final Collection<Place> p_out = new IdentityHashSet<Place>();
        
        for(final Transition current : sigma) {
            t_cycle.add(current);
            for(final PTArc arc : current.getInput()) {
                final Place p = arc.getSource();
                p_cycle.add(p);
            }
            for(final TPArc arc : current.getOutput()) {
                final Place p = arc.getTarget();
                p_out.add(p);
            }
            
            if(t_cycle.contains(current)) {
                break;
            }
        }
        
        p_cycle.retainAll(p_out);
    }
    
    
    /**
     * TODO: Document
     *
     * @param p
     * @return
     */
    private boolean onb(final Place p) {
        return p.getOutput().size() <= 1;
    }
    
    
    /**
     * TODO: Document
     *
     * @param in
     * @return
     */
    private boolean onb(final Collection<PTArc> in) {
        for(final PTArc arc : in) {
            if(!this.onb(arc.getSource())) {
                return false;
            }
        }
        return true;
    }
    
    
    /**
     * TODO: Document
     *
     * @param p
     * @return
     */
    private boolean syn(final Place p) {
        for(final PTArc arc : p.getOutput()) {
            final Transition t = arc.getTarget();
            if(t.getInput().size() <= 1) {
                return false;
            }
        }
        return true;
    }
    
    
    /**
     * TODO: Document
     *
     * @param t
     * @param m
     */
    private void fire(final Transition t, final Map<Place, Integer> m) {
        for(final TPArc out : t.getOutput()) {
            final Place pl = out.getTarget();
            final Integer old = m.put(pl, out.getWeight());
            if(old != null) {
                m.put(pl, old + out.getWeight());
            }
        }
        for(final PTArc in : t.getInput()) {
            final Place pl = in.getSource();
            final Integer old = m.get(pl);
            if(old == null || old < in.getWeight()) {
                throw new InternalError("Simulated firing of a transition that "
                        + "is not enabled: " + pl + " -> " + t + ": " + old
                        + " < " + in.getWeight()); 
            } else if(old == in.getWeight()) {
                m.remove(pl);
            } else {
                m.put(pl, old - in.getWeight());
            }
        }
    }
    
}