/*
 * Project:  Petruchio
 * Filename: PreAgglomeration.java
 * Created:  09.08.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.reductors;

import java.util.ArrayList;
import java.util.Collection;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class PostAgglomeration implements PetriNetReductor {
    
    private final Collection<Transition> trans2remove = new ArrayList<Transition>();

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#getDescription()
     */
    @Override
    public String getDescription() {
        return "Post-Agglomeration as proposed in G. Berthelot. Checking Properties of Nets Using Transformations. 1985. (Modified for non-safe nets.)";
    }

    
    /* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		this.trans2remove.clear();
	}
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#getName()
     */
    @Override
    public String getName() {
        return "post-agglomeration";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#reduce(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public synchronized int reduce(final PetriNet net) {
    	return this.reduce(net, new ArrayList<Place>(), new ArrayList<Transition>());
    }
    	
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet, java.util.Collection, java.util.Collection)
     */
    @Override
    public int reduce(final PetriNet net, final Collection<Place> ignorePlaces,
    		final Collection<Transition> ignoreTransitions) {
    	final Collection<Place> ignoreP = new IdentityHashSet<Place>(ignorePlaces);
    	final Collection<Transition> ignoreT = new IdentityHashSet<Transition>(ignoreTransitions);
    	
    	final Collection<Place> places = new ArrayList<Place>(net.getPlaces());
        int removed = 0;
        
        next_place:
        for(final Place p : places) {
            if((p.getMarking() != 0) || p.getInput().isEmpty()
                            || p.getOutput().isEmpty()
                            || ignoreP.contains(p)) {
                continue next_place;
            }
            for(final PTArc out : p.getOutput()) {
                for(final PTArc in : out.getTarget().getInput()) {
                    if(in.getSource() != p) {
                        continue next_place;
                    }
                }
                for(final TPArc out2 : out.getTarget().getOutput()) {
                    if(out2.getTarget() == p) {
                        continue next_place;
                    }
                }
            }
            
            for(final TPArc in : p.getInput()) {
                for(final PTArc out : p.getOutput()) {
                    if(out.getWeight() % in.getWeight() != 0) {
                        continue next_place;
                    }
                }
            }
            
            this.trans2remove.clear();
            for(final TPArc in : p.getInput()) {
                final Transition tIn = in.getSource();
                if(ignoreT.contains(tIn)) {
                	continue;
                }
                this.trans2remove.add(tIn);
                for(final PTArc out : p.getOutput()) {
                    final Transition tOut = out.getTarget();
                    if(ignoreT.contains(tOut)) {
                    	continue;
                    }
                    this.trans2remove.add(tOut);
                    final int fires = out.getWeight() / in.getWeight();
                    
                    final Transition tNew = net.addTransition(tIn.getName()
                                    + " --> " + tOut.getName());
                    tNew.setMeaning(tIn.getMeaning() + " --> " + tOut.getMeaning());
                    tNew.setX((tIn.getX() + tOut.getX()) >> 1);
                    tNew.setY((tIn.getY() + tOut.getY()) >> 1);
                    --removed;
                    
                    for(final TPArc tOut_ : tOut.getOutput()) {
                        net.addArc(tNew, tOut_.getTarget(), tOut_.getWeight());
                    }
                    for(final PTArc _tIn : tIn.getInput()) {
                        net.addArc(_tIn.getSource(), tNew, fires*_tIn.getWeight());
                    }
                    for(final TPArc tIn_ : tIn.getOutput()) {
                        if(tIn_.getTarget() != p) {
                            net.addArc(tNew, tIn_.getTarget(), fires*tIn_.getWeight());
                        }
                    }
                }
            }
            
            if(net.removePlace(p)) {
            	++removed;
            }
            for(final Transition t : this.trans2remove) {
                if(net.removeTransition(t)) {
                	++removed;
                }
            }
            this.trans2remove.clear();
        }
        return removed;
    }
}