/*
 * Project:  Petruchio
 * Filename: SequenceReduction.java
 * Created:  09.08.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.reductors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.petrinet.PetriNetReductor;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.Transition;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class SequenceReduction implements PetriNetReductor {
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#getDescription()
     */
    @Override
    public String getDescription() {
        return "Iff there is a sequence P1-T-P2, where P1*=*P2={T}, *T={P1} and"
        + " T*={P2}, then the output arc from T is added to each transition in "
        + "*P1 if it is not empty and P1, T are removed or the output arcs from"
        + " P2 are added to P1 and T, P2 are removed.";
    }

    
    /* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		// nothing
	}
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#getName()
     */
    @Override
    public String getName() {
        return "sequence reduction";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PNReductor#reduce(petruchio.interfaces.petrinet.PetriNet)
     */
    @Override
    public int reduce(final PetriNet net) {
    	return this.reduce(net, new ArrayList<Place>(), new ArrayList<Transition>());
    }
    	
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetReductor#reduce(petruchio.interfaces.petrinet.PetriNet, java.util.Collection, java.util.Collection)
     */
    @Override
    public int reduce(final PetriNet net, final Collection<Place> ignorePlaces,
    		final Collection<Transition> ignoreTransitions) {
    	final Collection<Place> ignoreP = new IdentityHashSet<Place>(ignorePlaces);
    	final Collection<Transition> ignoreT = new IdentityHashSet<Transition>(ignoreTransitions);
    	
    	final LinkedList<Place> places = new LinkedList<Place>(net.getPlaces());
        int removed = 0;
        
        next_place:
        while(!places.isEmpty()) {
            final Place p = places.remove();
            if((p.getOutput().size() != 1) || (!p.getInput().isEmpty() && ignoreP.contains(p))) {
                continue next_place;
            }
            final PTArc theInArc = p.getOutput().iterator().next();
            final Transition t = theInArc.getTarget();
            
            if((t.getInput().size() != 1) || (t.getOutput().size() != 1)
            		|| ignoreT.contains(t)) {
                continue next_place;
            }
            
            final TPArc theOutArc = t.getOutput().iterator().next();
            final Place p2 = theOutArc.getTarget();
            
            if((p == p2) || (p2.getInput().size() != 1)
                        || (p.getInput().isEmpty() && ((p2.getMarking() != 0) || ignoreP.contains(p2)))) {
                continue next_place;
            }
            
            final int weightIn = theInArc.getWeight();
            final int weightOut = theOutArc.getWeight();
            
            if(weightOut % weightIn != 0) {
                continue next_place;
            }
            final int w = weightOut / weightIn;
            
            if(p.getInput().isEmpty()) {
                p.setMeaning(p.getMeaning() + " --> " + p2.getMeaning());
                
                for(final PTArc outArc : p2.getOutput()) {
                    final Transition out = outArc.getTarget();
                    net.addArc(p, out, w*outArc.getWeight());
                    out.setMeaning(out.getMeaning() + " --> " + t.getMeaning());
                }
                
                if(net.removePlace(p2)) {
                    ++removed;
                }
                if(net.removeTransition(t)) {
                    ++removed;
                }
                places.add(p);
            } else {
                if(p.getMarking() != 0) {
                    p2.addTokens(w*p.getMarking());
                }
                p2.setMeaning(p.getMeaning() + " --> " + p2.getMeaning());
                
                for(final TPArc inArc : p.getInput()) {
                    final Transition in = inArc.getSource();
                    net.addArc(in, p2, w*inArc.getWeight());
                    in.setMeaning(in.getMeaning() + " --> " + t.getMeaning());
                }
                
                if(net.removePlace(p)) {
                	++removed;
                }
                if(net.removeTransition(t)) {
                	++removed;
                }
            }
        }
        return removed;
    }

}