/*
 * Project:  Petruchio
 * Filename: CMWriter.java
 * Created:  04.10.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.writers;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.Map;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetWriter;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;

/**
 * TODO: description 
 *
 * @author Tim Strazny
 */
public class CMWriter implements SelfDescribing, PetriNetWriter {

    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net writer for files in a simple counter machine format "
            + "as used by G. Geeraerts, J.-F. Raskin, L. Van Begin on "
            + "http://www.ulb.ac.be/di/ssd/ggeeraer/eec/";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "spec";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#write(petruchio.interfaces.petrinet.PetriNet, java.io.OutputStream)
     */
    @Override
    public void write(final PetriNet net, final OutputStream output) {
        final PrintWriter out = new PrintWriter(output);
        
        if(net.hasHeaderComment()) {
            final String[] lines = net.getHeaderComment().split("\r\n|[\n\r\u2028\u2029\u0085]");
            for(final String line : lines) {
                out.print("# ");
                out.print(line);
                out.print("\n");
            }
        }
        out.print("# ");
        out.print(net.getPlaces().size());
        out.print(" places, ");
        out.print(net.getTransitions().size());
        out.print(" transitions, ");
        out.print(net.getPTArcCount());
        out.print(" pt-arcs, ");
        out.print(net.getTPArcCount());
        out.print(" tp-arcs\n");
        out.print("# mapping of names to places:\n");
        
        final Map<Place, String> pl = new IdentityHashMap<Place, String>();
        {
            final Collection<String> plNames = new HashSet<String>();
            int i = 0;
            for(final Place p : net.getPlaces()) {
                boolean uncool = false;
                for(final char ch : p.getName().toCharArray()) {
                    if(!Character.isJavaIdentifierPart(ch)) {
                        uncool = true;
                        break;
                    }
                }
                String name;
                if(uncool || !plNames.add(p.getName())) {
                    do {
                        name = "x" + i++;
                    } while(!plNames.add(name));
                    
                    out.print("# ");
                    out.print(name);
                    out.print(" := ");
                    out.print(p.getName());
                    out.print("\n");
                } else {
                    name = p.getName();
                }
                pl.put(p, name);
            }
            if(i == 0) {
                out.print("# no illegal names found\n");
            }
        }
        
        out.print("\nvars\n\t");
        for(final Iterator<Place> it = net.getPlaces().iterator(); it.hasNext(); ) {
            final Place p = it.next();
            out.print(pl.get(p));
            if(it.hasNext()) {
                out.print(" ");
            }
        }
        out.print("\n\nrules\n");
        final Map<Place, Integer> change = new IdentityHashMap<Place, Integer>();
        for(final Transition t : net.getTransitions()) {
            out.print('\t');
            for(final Iterator<PTArc> it = t.getInput().iterator(); it.hasNext(); ) {
                final PTArc arc = it.next();
                out.print(pl.get(arc.getSource()));
                out.print(" >= ");
                out.print(arc.getWeight());
                if(it.hasNext()) {
                    out.print(", ");
                }
                
                final Integer old = change.put(arc.getSource(), -arc.getWeight());
                if(old != null) {
                    change.put(arc.getSource(), old-arc.getWeight());
                }
            }
            for(final TPArc arc : t.getOutput()) {
                final Integer old = change.put(arc.getTarget(), arc.getWeight());
                if(old != null) {
                    change.put(arc.getTarget(), old+arc.getWeight());
                }
            }
            out.print(" ->\n");
            boolean first = true;
            for(final Iterator<Map.Entry<Place, Integer>> it = change.entrySet().iterator(); it.hasNext(); ) {
                final Map.Entry<Place, Integer> e = it.next();
                final Place p = e.getKey();
                final Integer c = e.getValue();
                if((c == null) || (c == 0)) {
                    continue;
                }
                if(!first) {
                    out.print(",\n");
                } else {
                    first = false;
                }
                out.print("\t\t");
                out.print(pl.get(p));
                out.print("' = ");
                out.print(pl.get(p));
                if(c > 0) {
                    out.print('+');
                }
                out.print(c);
            }
            out.print(";\n\n");
            
            change.clear();
        }
        out.print("init\n\t");
        for(final Iterator<Place> it = net.getPlaces().iterator(); it.hasNext(); ) {
            final Place p = it.next();
            out.print(pl.get(p));
            out.print(" = ");
            out.print(p.getMarking());
            if(it.hasNext()) {
                out.print(", ");
            }
        }
        out.print("\n\ntarget\n\ninvariants\n\n");
        out.close();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        // no caches
    }

}