/*
 * Project:  Petruchio
 * Filename: DOTWriter.java
 * Created:  08.10.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.writers;

import java.io.OutputStream;
import java.io.PrintWriter;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetStreamWriter;
import petruchio.interfaces.petrinet.PetriNetWriter;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;

/**
 * TODO: description 
 *
 * @author Tim Strazny
 */
public class DOTWriter implements SelfDescribing, PetriNetWriter, PetriNetStreamWriter {

    private PrintWriter theout = null;
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net writer, that outputs the net in graphviz's DOT language.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "dot";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#write(petruchio.interfaces.petrinet.PetriNet, java.io.OutputStream)
     */
    @Override
    public void write(final PetriNet net, final OutputStream output) {
        final PrintWriter out = new PrintWriter(output);
        
        this.writeHeaderComment(net, out);
        
        out.print("digraph \"");
        out.print(this.encode(net.getName()));
        out.print("\" {\n"
                + "  node [fontsize=\"8\", fixedsize=\"true\", width=\".3\", height=\".3\", label=\"\", style=\"filled\", fillcolor=\"white\"];\n");
        
        this.writePlaces(net, out);
        out.print("\n");
        this.writeTransitions(net, out);
        out.print("\n");
        this.writePTArcs(net, out);
        out.print("\n");
        this.writeTPArcs(net, out);
        
        out.print("}");
        
        out.close();
    }
    
    
    /**
     * TODO: description
     *
     * @param s
     * @return
     */
    private String encode(final String s) {
        final StringBuilder b = new StringBuilder();
        for(final char ch : s.toCharArray()) {
            switch(ch) {
                case '\"': b.append("\\\""); break;
                case '\n': b.append("\\n"); break;
                default: b.append(ch);
            }
        }
        return b.toString();
    }
    
    
    
    /**
     * TODO: description
     *
     * @param net
     * @param theout
     */
    private void writeHeaderComment(final PetriNet net, final PrintWriter out) {
        out.print("/*\n");
        if(net.hasHeaderComment()) {
            final String[] lines = net.getHeaderComment().split("\r\n|[\n\r\u2028\u2029\u0085]");
            for(final String line : lines) {
                out.print(" * ");
                out.print(line);
                out.print("\n");
            }
        }
        out.print(" * ");
        out.print(net.getPlaces().size());
        out.print(" places, ");
        out.print(net.getTransitions().size());
        out.print(" transitions, ");
        out.print(net.getPTArcCount());
        out.print(" pt-arcs, ");
        out.print(net.getTPArcCount());
        out.print(" tp-arcs\n */\n\n");
    }
    
    
    /**
     * TODO: description
     *
     * @param net
     * @param theout
     */
    private void writePlaces(final PetriNet net, final PrintWriter out) {
        for(final Place p : net.getPlaces()) {
            this.writePlace(p, out);
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param p
     * @param theout
     */
    private void writePlace(final Place p, final PrintWriter out) {
        final String n = this.encode(p.getName());
        out.print("  \"");
        out.print(n);
        out.print("\" [shape=\"circle");
        if(p.getX() != 0 || p.getY() != 0) {
            out.print("\", pos=\"");
            out.print(p.getX());
            out.print(", ");
            out.print(p.getY());
        }
        if(p.getMarking() != 0) {
            out.print("\", label=\"");
            out.print(p.getMarking());
        }
        out.print("\"];");
        if(p.getNote() != null && p.getNote().length() != 0) {
            out.print("    // ");
            out.print(p.getNote());
        }
        out.print("\n");
        
        if(p.getMeaning().length() != 0) {
	        out.print("  \"");
	        out.print(n);
	        out.print("\" -> \"");
	        out.print(n);
	        out.print("\" [color=\"transparent\", labeldistance=\"1.6\", labelangle=\"45\", taillabel=\"");
	        out.print(this.encode(p.getMeaning()));
	        out.print("\"];\n");
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param net
     * @param theout
     */
    private void writeTransitions(final PetriNet net, final PrintWriter out) {
        for(final Transition t : net.getTransitions()) {
            this.writeTransition(t, out);
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param t
     * @param theout
     */
    private void writeTransition(final Transition t, final PrintWriter out) {
        final String n = this.encode(t.getName());
        out.print("  \"");
        out.print(n);
        out.print("\" [shape=\"box");
        if(t.getX() != 0 || t.getY() != 0) {
            out.print("\", pos=\"");
            out.print(t.getX());
            out.print(", ");
            out.print(t.getY());
        }
        out.print("\"];\n");
        
        if(t.getMeaning().length() != 0) {
	        out.print("  \"");
	        out.print(n);
	        out.print("\" -> \"");
	        out.print(n);
	        out.print("\" [color=\"transparent\", labeldistance=\"1.6\", labelangle=\"45\", taillabel=\"");
	        out.print(this.encode(t.getMeaning()));
	        out.print("\"];\n");
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param net
     * @param theout
     */
    private void writePTArcs(final PetriNet net, final PrintWriter out) {
        for(final Place p : net.getPlaces()) {
            for(final PTArc arc : p.getOutput()) {
                this.writeArc(arc.getSource().getName(),
                        arc.getTarget().getName(), arc.getWeight(), out);
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param net
     * @param theout
     */
    private void writeTPArcs(final PetriNet net, final PrintWriter out) {
        for(final Transition t : net.getTransitions()) {
            for(final TPArc arc : t.getOutput()) {
                this.writeArc(arc.getSource().getName(),
                        arc.getTarget().getName(), arc.getWeight(), out);
            }
        }
    }

    
    /**
     * TODO: description
     *
     * @param from
     * @param to
     * @param weight
     * @param theout
     */
    private void writeArc(final String from, final String to, final int weight, final PrintWriter out) {
        out.print("  \"");
        out.print(this.encode(from));
        out.print("\" -> \"");
        out.print(this.encode(to));
        out.print("\"");
        if(weight != 1) {
            out.print(" [label=\"");
            out.print(weight);
            out.print("\"]");
        }
        out.print(";\n");
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        if(this.theout != null) {
            this.theout.close();
            this.theout = null;
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#beginNet(java.lang.String)
     */
    @Override
    public void beginNet(final String name) {
        this.theout.print("digraph \"");
        this.theout.print(this.encode(name));
        this.theout.print("\" {\n  "
                + "node [fontsize=\"8\", fixedsize=\"true\", width=\".3\", height=\".3\", label=\"\", style=\"filled\", fillcolor=\"white\"];\n");
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#endNet()
     */
    @Override
    public void endNet() {
        this.theout.print("}");
        this.theout.close();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#setOutput(java.io.OutputStream)
     */
    @Override
    public void setOutput(OutputStream output) {
        this.theout = new PrintWriter(output);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeHeaderComment(java.lang.String)
     */
    @Override
    public void writeHeaderComment(final String comment) {
        this.theout.print("/*\n");
        final String[] lines = comment.split("\r\n|[\n\r\u2028\u2029\u0085]");
        for(final String line : lines) {
            this.theout.print(" * ");
            this.theout.print(line);
            this.theout.print("\n");
        }
        this.theout.print(" */\n");
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writePTArc(java.lang.String, java.lang.String, int)
     */
    @Override
    public void writePTArc(final String name, final String name2, final int weight) {
        this.writeArc(name, name2, weight, this.theout);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writePlace(petruchio.interfaces.petrinet.Place)
     */
    @Override
    public void writePlace(final Place p) {
        this.writePlace(p, this.theout);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeTPArc(java.lang.String, java.lang.String, int)
     */
    @Override
    public void writeTPArc(final String name, final String name2, final int weight) {
        this.writeArc(name, name2, weight, this.theout);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeTransition(petruchio.interfaces.petrinet.Transition)
     */
    @Override
    public void writeTransition(final Transition t) {
        this.writeTransition(t, this.theout);
    }

}