/*
 * Project: Petruchio
 * File:    DVEWriter.java
 * Author:  Tim Strazny
 * Created: 15.10.2007
 */
package petruchio.pn.writers;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.Map;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetWriter;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;


/**
 * TODO: Document
 * 
 * @author Tim Strazny
 */
public class DVEWriter implements SelfDescribing, PetriNetWriter {

    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net writer, that outputs in the model checking tool DiVinE's DVE format.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "dve";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#write(petruchio.interfaces.petrinet.PetriNet, java.io.OutputStream)
     */
    @Override
    public void write(final PetriNet net, final OutputStream output) {
        final PrintWriter out = new PrintWriter(output);
        
        this.writeHeaderComment(net, out);
        
        for(final Place p : net.getPlaces()) {
            out.print("\n/* ");
            out.print(p.getMeaning());
            out.print(" */\n");
            out.print("int ");
            out.print(p.getName());
            out.print("=");
            out.print(p.getMarking());
            out.print(";\n");
        }
        
        out.print("\nprocess ");
        out.print(net.getName());
        out.print(" {\nstate run;\ninit run;\ntrans\n");
        
        final Map<Place, Integer> in = new IdentityHashMap<Place, Integer>();
        for(final Iterator<Transition> tIt = net.getTransitions().iterator(); tIt.hasNext(); ) {
            final Transition t = tIt.next();
            out.print("\n  /* ");
            out.print(t.getName());
            out.print(" := ");
            out.print(t.getMeaning());
            out.print(" */\n  run -> run {");
            if(!t.getInput().isEmpty()) {
                out.print("\n    guard  ");
                for(final Iterator<PTArc> it = t.getInput().iterator(); it.hasNext();) {
                    final PTArc arc = it.next();
                    in.put(arc.getSource(), arc.getWeight());
                    out.print(arc.getSource().getName());
                    out.print(">=");
                    out.print(arc.getWeight());
                    if(it.hasNext()) {
                        out.print(" and ");
                    }
                }
                out.print(";");
            }
            boolean changes = false;
            for(final TPArc arc : t.getOutput()) {
                final Integer minus = in.remove(arc.getTarget());
                final int change = arc.getWeight() - (minus == null ? 0 : minus);
                if(change != 0) {
                    if(changes) {
                        out.print(",\n           ");
                    } else {
                        out.print("\n    effect ");
                    }
                    out.print(arc.getTarget().getName());
                    out.print("=");
                    out.print(arc.getTarget().getName());
                    out.print(change > 0 ? "+" : "");
                    out.print(change);
                    changes = true;
                }
            }
            if(changes || !in.isEmpty()) {
                for(final Map.Entry<Place, Integer> e : in.entrySet()) {
                    if(changes) {
                        out.print(",\n           ");
                    } else {
                        out.print("\n    effect ");
                    }
                    out.print(e.getKey().getName());
                    out.print("=");
                    out.print(e.getKey().getName());
                    out.print("-");
                    out.print(e.getValue());
                    changes = true;
                }
                in.clear();
            }
            if(changes) {
                out.print("; }");
            } else {
                out.print(" }");
            }
            
            if(tIt.hasNext()) {
                out.print(",");
            } else {
                out.print(";");
            }
            out.print("\n");
        }
        
        out.print("\n}\nsystem sync;");
        out.close();
    }

    
    /**
     * TODO: Document
     *
     * @param net
     * @param out
     */
    private void writeHeaderComment(final PetriNet net, final PrintWriter out) {
        out.print("/*\n");
        if(net.hasHeaderComment()) {
            final String[] lines = net.getHeaderComment().split("\r\n|[\n\r\u2028\u2029\u0085]");
            for(final String line : lines) {
                out.print(" * ");
                out.print(line);
                out.print("\n");
            }
        }
        out.print(" * ");
        out.print(net.getPlaces().size());
        out.print(" places, ");
        out.print(net.getTransitions().size());
        out.print(" transitions, ");
        out.print(net.getPTArcCount());
        out.print(" pt-arcs, ");
        out.print(net.getTPArcCount());
        out.print(" tp-arcs\n */\n");
    }
    

    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        //( no caches
    }

}