/*
 * Project:  Petruchio
 * Filename: INAWriter.java
 * Created:  04.09.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.writers;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.Map;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetWriter;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;


/**
 * Writes the low level Petri net in INA format.
 * 
 * @author Tim Strazny
 */
public class INAWriter implements SelfDescribing, PetriNetWriter {

    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net writer, which outputs in the format accepted by the INA tool.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.pn.PetriNetWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "pnt";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#write(petruchio.interfaces.petrinet.PetriNet, java.io.OutputStream)
     */
    @Override
    public synchronized void write(final PetriNet net, final OutputStream outStream) {
        final PrintWriter out = new PrintWriter(outStream);
        
        final Map<Place, String> pl = new IdentityHashMap<Place, String>();
        int i = 0;
        for(final Place p : net.getPlaces()) {
            pl.put(p, "" + i);
            ++i;
        }
        
        final Map<Transition, String> tr = new IdentityHashMap<Transition, String>();
        i = 0;
        for(final Transition t : net.getTransitions()) {
            tr.put(t, "" + i);
            ++i;
        }

        this.writeStructure(net, out, pl, tr);
        
        this.writePlaces(net, out, pl);
        
        this.writeTransitions(net, out, tr);
        
        this.writeComment(net, pl, tr, out);
        
        out.close();
    }
    
    
    /**
     * TODO: document
     *
     * @param net
     * @param pl
     * @param tr
     * @param out
     */
    private void writeComment(final PetriNet net, final Map<Place, String> pl,
                    final Map<Transition, String> tr, final PrintWriter out) {
        
        if(net.hasHeaderComment()) {
            final String[] lines = net.getHeaderComment().split("\r\n|[\n\r\u2028\u2029\u0085]");
            for(final String line : lines) {
                out.println(line);
            }
        }
        out.print(net.getPlaces().size());
        out.print(" places, ");
        out.print(net.getTransitions().size());
        out.print(" transitions, ");
        out.print(net.getPTArcCount());
        out.print(" pt-arcs, ");
        out.print(net.getTPArcCount());
        out.println(" tp-arcs");
        
        out.println();
        
        out.println("Net name: " + net.getName());
        
        out.println();
        
        out.println("Mapping: PLACES with long names, meanings or notes");
        for(final Map.Entry<Place, String> e : pl.entrySet()) {
            final boolean valid = this.validName(e.getKey().getName());
            final Place p = e.getKey();
            if(!valid || (p.getNote().length() != 0) || (p.getMeaning().length() != 0)) {
                final String n = valid ? p.getName() : "p" + p;
                out.println(n + " := " + p.getName());
                final String meaning = p.getMeaning();
                if(meaning.length() != 0) {
                    out.println("    Meaning(" + n + "): " + meaning);
                }
                final String note = INAWriter.getNote(p);
                if(note != null) {
                    out.println("    Note(" + n + "): " + note);
                }
                out.println();
            }
        }
        
        out.println();
        
        out.println("Mapping: TRANSITIONS with long names");
        for(final Map.Entry<Transition, String> e : tr.entrySet()) {
        	final Transition t = e.getKey();
            if(!this.validName(t.getName()) || (t.getMeaning().length() != 0)) {
                out.println("t" + t + " := " + t.toString());
                final String meaning = t.getMeaning();
                if(meaning.length() != 0) {
                    out.println("    Meaning(t" + t + "): " + meaning);
                }
                out.println();
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param name
     * @return
     */
    private boolean validName(final String name) {
        if((name.length() == 0) || (name.length() > 16)) {
            return false;
        }
        for(char c : name.toCharArray()) {
            if(!Character.isJavaIdentifierPart(c)) {
                return false;
            }
        }
        return true;
    }
    
    
    /**
     * TODO: document
     *
     * @param out
     * @param pl
     * @param tr
     */
    private void writeStructure(final PetriNet net, final PrintWriter out,
                    final Map<Place, String> pl, final Map<Transition, String> tr) {
        out.print("P   M   PRE,POST  NETZ 1:");
        out.println(INAWriter.enlargeBack(net.getName(), 16, true));
        for(final Map.Entry<Place, String> e : pl.entrySet()) {
            final Place p = e.getKey();
            out.print(INAWriter.enlargeFront(e.getValue(), 3, false));
            out.print(" ");
            out.print(INAWriter.enlargeBack(""+p.getMarking(), 4, false));
            out.print(" ");
            for(final Iterator<TPArc> it = p.getInput().iterator(); it.hasNext(); ) {
                final TPArc arc = it.next();
                out.print(tr.get(arc.getSource()));
                if(arc.getWeight() != 1) {
                    out.print(": ");
                    out.print(arc.getWeight());
                }
                if(it.hasNext()) {
                    out.print(" ");
                }
            }
            if(!p.getOutput().isEmpty()) {
                out.print(", ");
                for(final Iterator<PTArc> it = p.getOutput().iterator(); it.hasNext(); ) {
                    final PTArc arc = it.next();
                    out.print(tr.get(arc.getTarget()));
                    if(arc.getWeight() != 1) {
                        out.print(": ");
                        out.print(arc.getWeight());
                    }
                    if(it.hasNext()) {
                        out.print(" ");
                    }
                }
            }
            out.println();
        }
        out.println("@");
    }
    
    
    /**
     * TODO: document
     *
     * @param net
     * @param out
     * @param pl
     */
    private void writePlaces(final PetriNet net, final PrintWriter out,
                    final Map<Place, String> pl) {
        out.println("place nr.             name capacity time");
        for(final Map.Entry<Place, String> e : pl.entrySet()) {
            out.print(INAWriter.enlargeFront(e.getValue(), 8, false));
            out.print(": ");
            if(this.validName(e.getKey().getName())) {
                out.print(INAWriter.enlargeBack(e.getKey().getName(), 16, false));
            } else {
                out.print(INAWriter.enlargeBack("p" + e.getValue(), 16, false));
            }
            out.println("       oo    0");
        }
        out.println("@");
    }
    
    
    /**
     * TODO: document
     *
     * @param net
     * @param out
     * @param tr
     */
    private void writeTransitions(final PetriNet net, final PrintWriter out,
                    final Map<Transition, String> tr) {
        out.println("trans nr.             name priority time");
        for(final Map.Entry<Transition, String> e : tr.entrySet()) {
            out.print(INAWriter.enlargeFront(e.getValue(), 8, false));
            out.print(": ");
            if(this.validName(e.getKey().getName())) {
                out.print(INAWriter.enlargeBack(e.getKey().getName(), 16, false));
            } else {
                out.print(INAWriter.enlargeBack("t" + e.getValue(), 16, false));
            }
            out.println("        0    0");
        }
        out.println("@");
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param l
     * @param cut
     * @return
     */
    public static String enlargeBack(final String s, final int l, final boolean cut) {
        final StringBuilder n = new StringBuilder(l);
        if(cut && (s.length() > l)) {
            return s.substring(0, l-1);
        }
        
        n.append(s);
        for(int i = 0; i < s.length(); ++i) {
            if(!Character.isJavaIdentifierPart(n.charAt(i))) {
                n.setCharAt(i, '_');
            }
        }
        
        for(int i = l-s.length(); i > 0; --i) {
            n.append(' ');
        }
        return n.toString();
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @param l
     * @param cut
     * @return
     */
    public static String enlargeFront(final String s, final int l, final boolean cut) {
        final StringBuilder n = new StringBuilder(l);
        if(cut && (s.length() > l)) {
            return s.substring(0, l-1);
        }
        
        for(int i = l-s.length(); i > 0; --i) {
            n.append(' ');
        }
        n.append(s);
        for(int i = 1; i <= s.length(); ++i) {
            if(!Character.isJavaIdentifierPart(n.charAt(n.length()-i))) {
                n.setCharAt(i, '_');
            }
        }
        return n.toString();
    }
    
    
    /**
     * TODO: document
     * 
     * @param p
     * @return
     */
    public static String getNote(final Place p) {
        final StringBuilder note = new StringBuilder();
        
        if(p.getBound() == Place.BOUND_UNBOUNDED) {
            note.append("unbounded");
        } else if(p.getBound() != Place.BOUND_UNKNOWN) {
            note.append(p.getBound());
            note.append("-bounded");
        }
        
        if((p.getNote().length() != 0) && (p.getBound() != Place.BOUND_UNKNOWN)) {
            note.append(", ");
            note.append(p.getNote());
        }
        
        if(note.length() != 0) {
            return note.toString();
        }
        return null;
    }
    
    
    /* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
		// nothing
	}
    
}