/*
 * Project:  Petruchio
 * Filename: PEPWriter.java
 * Created:  09.12.2006
 * Author:   Tim Strazny
 */
package petruchio.pn.writers;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.File;
import java.util.IdentityHashMap;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetWriter;
import petruchio.interfaces.petrinet.PetriNetStreamWriter;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.Transition;


/**
 * Writes the low level Petri net in PEP format.
 * Note that only \n newline characters are used to workaround the PEP loading
 * routines and the specific block order (and no text blocks) as well as no
 * empty lines between blocks to allow the mckit to load the file. 
 * 
 * @author Tim Strazny
 */
public class PEPWriter implements SelfDescribing, PetriNetWriter, PetriNetStreamWriter {

    private PrintWriter ptOut;
    private PrintWriter tpOut;
    private PrintWriter tOut;
    private PrintWriter pOut;
    private OutputStream theout;
    
    private File ptOutTmp;
    private File tpOutTmp;
    private File tOutTmp;
    private File pOutTmp;
    
    private int plC;
    private int trC;
    private Map<String, Integer> thepl = new HashMap<String, Integer>();
    private Map<String, Integer> thetr = new HashMap<String, Integer>();
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net writer that outputs in the PEP tool's low level net format.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.pn.PetriNetWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "ll_net";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#write(petruchio.interfaces.petrinet.PetriNet, java.io.OutputStream)
     */
    @Override
    public synchronized void write(final PetriNet net, final OutputStream outStream) {
        final PrintWriter out = new PrintWriter(outStream);
        
        this.writeHeaderComment(net, out);
        this.writeHeader(out);
        
        // the mckit does not understand TX blocks
//        theout.print("\nTX\n");
//        theout.print("N 20@10 \"");
//        theout.print(net.getName());
//        theout.print("\"\n");
        
        final Map<Place, Integer> pl = new IdentityHashMap<Place, Integer>();
        this.writePlaces(net, out, pl);
        
        final Map<Transition, Integer> tr = new IdentityHashMap<Transition, Integer>();
        this.writeTransitions(net, out, tr);
        
        // the mckit needs exactly this order of blocks
        this.writeTPArcs(net, out, tr, pl);
        
        this.writePTArcs(net, out, pl, tr);
        
        out.close();
    }
    
    
    /**
     * TODO: document
     *
     * @param theout
     */
    private void writeHeaderComment(final PetriNet net, final PrintWriter out) {
        if(net.hasHeaderComment()) {
            final String[] lines = net.getHeaderComment().split("\r\n|[\n\r\u2028\u2029\u0085]");
            for(final String line : lines) {
                out.print("% ");
                out.print(line);
                out.print("\n");
            }
        }
        out.print("% ");
        out.print(net.getPlaces().size());
        out.print(" places, ");
        out.print(net.getTransitions().size());
        out.print(" transitions, ");
        out.print(net.getPTArcCount());
        out.print(" pt-arcs, ");
        out.print(net.getTPArcCount());
        out.print(" tp-arcs\n");
        //theout.print("\n"); // mckit does not like empty lines
    }
    
    
    /**
     * Writes a low level Petri net PEP header.
     *
     * @param theout The PrintWriter to write to.
     */
    private void writeHeader(final PrintWriter out) {
        out.print("PEP\nPTNet\nFORMAT_N\n");
    }
    
    
    /**
     * TODO: document
     *
     * @param theout
     * @param thepl
     */
    private void writePlaces(final PetriNet net, final PrintWriter out,
                    final Map<Place, Integer> pl) {
        if(!net.getPlaces().isEmpty()) {
            //theout.print("\n"); // mckit does not like empty lines
            out.print("PL\n");
        
            // index counter
            int i = 0;
            for(final Place p : net.getPlaces()) {
                ++i;
                pl.put(p, i);
                
                this.writePlace(p, i, out);
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param p
     * @param pNr
     * @param theout
     */
    private void writePlace(final Place p, final int pNr, final PrintWriter out) {
        final String note = PEPWriter.getNote(p);
        if(note != null) {
            out.print("%");
            out.print(note);
            out.print("\n");
        }
        out.print(pNr);
        out.print("\"");
        out.print(this.encode(p.getName()));
        out.print("\"");
        out.print(p.getX());
        out.print("@");
        out.print(p.getY());
        if(p.getMarking() != 0) {
            out.print("m");
            out.print(p.getMarking());
            out.print("M");
            out.print(p.getMarking());
        }
        if(p.getMarking() != 0) {
            out.print("e");
        }
        if(p.isFinal()) {
            out.print("x");
        }
        if(p.getMeaning() != null) {
            out.print("b\"");
            out.print(this.encode(p.getMeaning()));
            out.print("\"");
        }
        out.print("a15@-15\n");
    }
    
    
    /**
     * @param p
     * @return
     */
    private static String getNote(final Place p) {
        final StringBuilder note = new StringBuilder();
        
        if(p.getBound() == Place.BOUND_UNBOUNDED) {
            note.append("unbounded");
        } else if(p.getBound() != Place.BOUND_UNKNOWN) {
            note.append(p.getBound());
            note.append("-bounded");
        }
        
        if((p.getNote().length() != 0) && (p.getBound() != Place.BOUND_UNKNOWN)) {
            note.append(", ");
            note.append(p.getNote());
        }
        
        if(note.length() != 0) {
            return note.toString();
        }
        return null;
    }
    
    
    /**
     * TODO: document
     *
     * @param theout
     * @param thetr
     */
    private void writeTransitions(final PetriNet net, final PrintWriter out,
                    final Map<Transition, Integer> tr) {
        if(!net.getTransitions().isEmpty()) {
            //theout.print("\n"); // mckit does not like empty lines
            out.print("TR\n");
            
            // index counter
            int i = 0;
            for(final Transition t : net.getTransitions()) {
                ++i;
                tr.put(t, i);
                
                this.writeTransition(t, i, out);
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param t
     * @param tNr
     * @param theout
     */
    private void writeTransition(final Transition t, final int tNr, final PrintWriter out) {
        out.print(tNr);
        out.print("\"");
        out.print(this.encode(t.getName()));
        out.print("\"");
        out.print(t.getX());
        out.print("@");
        out.print(t.getY());
        if(t.getMeaning() != null) {
            out.print("b\"");
            out.print(this.encode(t.getMeaning()));
            out.print("\"");
        }
        out.print("a15@-15");
        out.print("\n");
    }
    
    
    /**
     * TODO: document
     *
     * @param s
     * @return
     */
    private String encode(final String s) {
        return s.replaceAll("\\\\", "\\\\").replaceAll("\"", "\\\"");
    }
    
    
    /**
     * TODO: document
     *
     * @param theout
     * @param thepl
     * @param thetr
     */
    private void writePTArcs(final PetriNet net, final PrintWriter out,
                    final Map<Place, Integer> pl, final Map<Transition, Integer> tr) {
        if(net.getPTArcCount() != 0) {
            //theout.print("\n"); // mckit does not like empty lines
            out.print("PT\n");
            
            for(final Place p : net.getPlaces()) {
                for(final PTArc arc : p.getOutput()) {
                    final Transition t = arc.getTarget();
                    
                    final int pid = pl.get(p);
                    final int tid = tr.get(t);
                    
                    this.writePTArc(pid, tid, arc.getWeight(), out);
                }
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param pid
     * @param tid
     * @param w
     * @param theout
     */
    private void writePTArc(final int pid, final int tid, final int w, final PrintWriter out) {
        out.print(pid);
        out.print(">");
        out.print(tid);
        
        if(w > 1) {
            out.print(" w");
            out.print(w);
        }
        
        out.print("\n");
    }
    
    
    /**
     * TODO: description
     *
     * @param pid
     * @param tid
     * @param w
     * @param theout
     */
    private void writeTPArc(final int pid, final int tid, final int w, final PrintWriter out) {
        out.print(pid);
        out.print("<");
        out.print(tid);
        
        if(w > 1) {
            out.print(" w");
            out.print(w);
        }
        
        out.print("\n");
    }
    
    
    /**
     * TODO: document
     *
     * @param theout
     * @param thetr
     * @param thepl
     */
    private void writeTPArcs(final PetriNet net, final PrintWriter out,
                    final Map<Transition, Integer> tr, final Map<Place, Integer> pl) {
        if(net.getTPArcCount() != 0) {
            //theout.print("\n"); // mckit does not like empty lines
            out.print("TP\n");
            
            for(final Transition t : net.getTransitions()) {
                for(final TPArc arc : t.getOutput()) {
                    final Place p = arc.getTarget();
                    
                    final int tid = tr.get(t);
                    final int pid = pl.get(p);
                    
                    this.writeTPArc(tid, pid, arc.getWeight(), out);
                }
            }
        }
    }
    
    
    /* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
	    this.thepl.clear();
	    this.thetr.clear();
	    
	    if(this.theout != null) {
	        try {
	            this.theout.close();
	        } catch(final IOException e) {
	            // ignore
	        }
    		this.theout = null;
	    }
		
	    if(this.pOut != null) {
	        this.pOut.close();
	        this.pOut = null;
	    }
	    if(this.tOut != null) {
            this.tOut.close();
            this.tOut = null;
        }
	    if(this.ptOut != null) {
            this.ptOut.close();
            this.ptOut = null;
        }
	    if(this.tpOut != null) {
            this.tpOut.close();
            this.tpOut = null;
        }
		this.pOutTmp = null;
		this.tOutTmp = null;
		this.ptOutTmp = null;
		this.tpOutTmp = null;
	}


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#beginNet(java.lang.String)
     */
    @Override
    public void beginNet(final String name) {
        try {
            final Random r = new Random();
            int i;
            
            i = 0;
            do {
                this.pOutTmp = File.createTempFile("pepstreamwriter_pl" + r.nextInt(Integer.MAX_VALUE), null);
            } while(!this.pOutTmp.canWrite() && ++i < 100);
            if(i >= 100) {
                throw new RuntimeException("Unable to create temporary file. But tried 100 times. Last attempt: " + this.pOutTmp);
            }
            this.pOutTmp.deleteOnExit();
            
            i = 0;
            do {
                this.tOutTmp = File.createTempFile("pepstreamwriter_tr" + r.nextInt(Integer.MAX_VALUE), null);
            } while(!this.tOutTmp.canWrite() && ++i < 100);
            if(i >= 100) {
                throw new RuntimeException("Unable to create temporary file. But tried 100 times. Last attempt: " + this.tOutTmp);
            }
            this.tOutTmp.deleteOnExit();
            
            i = 0;
            do {
                this.ptOutTmp = File.createTempFile("pepstreamwriter_pt" + r.nextInt(Integer.MAX_VALUE), null);
            } while(!this.ptOutTmp.canWrite() && ++i < 100);
            if(i >= 100) {
                throw new RuntimeException("Unable to create temporary file. But tried 100 times. Last attempt: " + this.ptOutTmp);
            }
            this.ptOutTmp.deleteOnExit();
            
            i = 0;
            do {
                this.tpOutTmp = File.createTempFile("pepstreamwriter_tp" + r.nextInt(Integer.MAX_VALUE), null);
            } while(!this.tpOutTmp.canWrite() && ++i < 100);
            this.tpOutTmp.deleteOnExit();
            if(i >= 100) {
                throw new RuntimeException("Unable to create temporary file. But tried 100 times. Last attempt: " + this.tpOutTmp);
            }
            
            this.pOut = new PrintWriter(this.pOutTmp);
            this.tOut = new PrintWriter(this.tOutTmp);
            this.ptOut = new PrintWriter(this.ptOutTmp);
            this.tpOut = new PrintWriter(this.tpOutTmp);
            
            this.pOut.print("PL\n");
            this.tOut.print("TR\n");
            this.ptOut.print("PT\n");
            this.tpOut.print("TP\n");
        } catch(final FileNotFoundException e) {
            // ignore
        } catch (final IOException e) {
            // ignore
        }
        
        this.plC = 1;
        this.trC = 1;
        this.thepl.clear();
        this.thetr.clear();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#endNet()
     */
    @Override
    public void endNet() {
        this.pOut.close();
        this.tOut.close();
        this.ptOut.close();
        this.tpOut.close();

        try {
            this.theout.write("PEP\nPTNet\nFORMAT_N\n".getBytes());
            
            byte[] buf = new byte[8192];
            int len;
            InputStream in = new FileInputStream(this.pOutTmp);
            while ((len = in.read(buf)) > 0) {
                this.theout.write(buf, 0, len);
            }
            in.close();
            
            in = new FileInputStream(this.tOutTmp);
            while ((len = in.read(buf)) > 0) {
                this.theout.write(buf, 0, len);
            }
            in.close();
            
            in = new FileInputStream(this.tpOutTmp);
            while ((len = in.read(buf)) > 0) {
                this.theout.write(buf, 0, len);
            }
            in.close();
            
            in = new FileInputStream(this.ptOutTmp);
            while ((len = in.read(buf)) > 0) {
                this.theout.write(buf, 0, len);
            }
            in.close();
            
            try {
                if(!this.pOutTmp.delete()) {
                    System.err.println("Unable to delete temporary file " + this.pOutTmp);
                }
                if(!this.tOutTmp.delete()) {
                    System.err.println("Unable to delete temporary file " + this.tOutTmp);
                }
                if(!this.tpOutTmp.delete()) {
                    System.err.println("Unable to delete temporary file " + this.tpOutTmp);
                }
                if(!this.ptOutTmp.delete()) {
                    System.err.println("Unable to delete temporary file " + this.ptOutTmp);
                }
            } finally {
                this.theout.close();
            }
            
        } catch (final FileNotFoundException e) {
            // ignore
            
        } catch (final IOException e) {
            // ignore
        }
        this.thepl.clear();
        this.thetr.clear();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#setOutput(java.io.OutputStream)
     */
    @Override
    public void setOutput(final OutputStream output) {
        this.theout = output;
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeHeaderComment(java.lang.String)
     */
    @Override
    public void writeHeaderComment(final String comment) {
        final String[] lines = comment.split("\r\n|[\n\r\u2028\u2029\u0085]");
        try {
            for(final String line : lines) {
                this.theout.write("% ".getBytes());
                this.theout.write(line.getBytes());
                this.theout.write('\n');
            }
        } catch(final IOException e) {
            // ignore
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writePTArc(java.lang.String, java.lang.String, int)
     */
    @Override
    public void writePTArc(final String name, final String name2, final int weight) {
        this.writePTArc(this.getPlaceID(name), this.getTransitionID(name2), weight, this.ptOut);
    }
    
    
    private int getPlaceID(final String name) {
        final Integer id = this.thepl.get(name);
        if(id == null) {
            final int newID = this.plC++;
            this.thepl.put(name, Integer.valueOf(newID));
            return newID;
        }
        return id.intValue();
    }
    
    private int getTransitionID(final String name) {
        final Integer id = this.thetr.get(name);
        if(id == null) {
            final int newID = this.trC++;
            this.thetr.put(name, Integer.valueOf(newID));
            return newID;
        }
        return id.intValue();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writePlace(petruchio.interfaces.petrinet.Place)
     */
    @Override
    public void writePlace(final Place p) {
        this.writePlace(p, this.getPlaceID(p.getName()), this.pOut);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeTPArc(java.lang.String, java.lang.String, int)
     */
    @Override
    public void writeTPArc(final String name, final String name2, final int weight) {
        this.writeTPArc(this.getTransitionID(name), this.getPlaceID(name2), weight, this.tpOut);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeTransition(petruchio.interfaces.petrinet.Transition)
     */
    @Override
    public void writeTransition(final Transition t) {
        this.writeTransition(t, this.getTransitionID(t.getName()), this.tOut);
    }

}