/*
 * Project:  Petruchio
 * Filename: PNMLWriter.java
 * Created:  31.07.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.writers;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Random;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PetriNetStreamWriter;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.PetriNetWriter;
import petruchio.interfaces.petrinet.Transition;
import petruchio.interfaces.petrinet.PetriNet;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class PNMLWriter implements SelfDescribing, PetriNetWriter, PetriNetStreamWriter {

    private PrintWriter aOut = null;
    private PrintWriter theout = null;
    
    private File aOutTmp = null;
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net writer that outputs in the PNML format.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.pn.PetriNetWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "pnml";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#write(petruchio.interfaces.petrinet.PetriNet, java.io.OutputStream)
     */
    @Override
    public synchronized void write(final PetriNet net, final OutputStream outStream) {
        final PrintWriter out = new PrintWriter(outStream);
        
        out.println("<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>");
        out.println("<pnml>");
        out.println("<net id=\"n1\" type=\"PTNet\">");

        this.writeHeaderComment(net, out);
        
        out.print("  <name><value>");
        out.print(PNMLWriter.encode(net.getName()));
        out.println("</value></name>");
        
        final Collection<String> ids = new HashSet<String>();
        
        final Map<Place, String> pl = new IdentityHashMap<Place, String>();
        this.writePlaces(net, out, pl, ids);
        
        final Map<Transition, String> tr = new IdentityHashMap<Transition, String>();
        this.writeTransitions(net, out, tr, ids);
        
        this.writePTArcs(net, out, pl, tr);
        
        this.writeTPArcs(net, out, tr, pl);
        
        out.println("</net>");
        out.print("</pnml>");
        
        out.close();
    }
    
    
    /**
     * TODO: document
     *
     * @param theout
     */
    private void writeHeaderComment(final PetriNet net, final PrintWriter out) {
        out.println("  <!--");
        if(net.hasHeaderComment()) {
            final String[] lines = PNMLWriter.encodeComment(net.getHeaderComment()).split("\r\n|[\n\r\u2028\u2029\u0085]");
            for(final String line : lines) {
                out.print("    - ");
                out.println(line);
            }
        }
        out.print("    - ");
        out.print(net.getPlaces().size());
        out.print(" places, ");
        out.print(net.getTransitions().size());
        out.print(" transitions, ");
        out.print(net.getPTArcCount());
        out.print(" pt-arcs, ");
        out.print(net.getTPArcCount());
        out.println(" tp-arcs");
        out.println("    -->");
    }

    
    /**
     * TODO: document
     *
     * @param theout
     * @param pl
     */
    private void writePlaces(final PetriNet net, final PrintWriter out,
                    final Map<Place, String> pl, final Collection<String> ids) {
        if(!net.getPlaces().isEmpty()) {
            out.println();
        
            // index counter
            int i = 0;
            for(final Place p : net.getPlaces()) {
            	String id = PNMLWriter.encode(p.getName());
            	while(!ids.add(id.toLowerCase())) {
	                ++i;
	                id = "p" + i;
            	}
            	pl.put(p, id);
                
                this.writePlace(p, id, out);
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param p
     * @param id
     * @param theout
     */
    private void writePlace(final Place p, final String id, final PrintWriter out) {
        out.print("  <place id=\"");
        out.print(id);
        out.println("\">");
        
        final String note = PNMLWriter.getNote(p);
        if(note != null) {
            out.print("    <!-- ");
            out.print(note);
            out.println(" -->");
        }
        
        out.print("    <graphics><position page=\"1\" x=\"");
        out.print(p.getX());
        out.print("\" y=\"");
        out.print(p.getY());
        out.println("\" /></graphics>");
        
//        if(p.getMarking() != 0) { // pipe3 needs explicit 0-markings
            out.println("    <initialMarking>");
            out.print("      <value>");
            out.print(p.getMarking());
            out.println("</value>");
            out.println("      <graphics><offset page=\"1\" x=\"0\" y=\"0\" /></graphics>");
            out.println("    </initialMarking>");
            out.println("    <marking>");
            out.print("      <value>");
            out.print(p.getMarking());
            out.println("</value>");
            out.println("      <graphics><offset page=\"1\" x=\"0\" y=\"0\" /></graphics>");
            out.println("    </marking>");
//        }
        
        out.println("    <name>");
        out.println("      <graphics><offset page=\"1\" x=\"15\" y=\"-15\" /></graphics>");
        out.print("      <value>");
        final String mean = PNMLWriter.encode(p.getMeaning());
        if(mean.isEmpty()) {
            out.print(id);
        } else {
            out.print(mean);
        }
        out.println("</value>");
        out.println("    </name>");
        
        out.println("  </place>");
    }
    
    
    /**
     * @param p
     * @return
     */
    private static String getNote(final Place p) {
        final StringBuilder note = new StringBuilder();
        
        if(p.getBound() == Place.BOUND_UNBOUNDED) {
            note.append("unbounded");
        } else if(p.getBound() != Place.BOUND_UNKNOWN) {
            note.append(p.getBound());
            note.append("-bounded");
        }
        
        if((p.getNote().length() != 0) && (p.getBound() != Place.BOUND_UNKNOWN)) {
            note.append(", ");
            note.append(p.getNote());
        }
        
        if(note.length() != 0) {
            return PNMLWriter.encodeComment(note.toString());
        }
        return null;
    }
    
    
    /**
     * TODO: document
     *
     * @param theout
     * @param tr
     */
    private void writeTransitions(final PetriNet net, final PrintWriter out,
                    final Map<Transition, String> tr, final Collection<String> ids) {
        if(!net.getTransitions().isEmpty()) {

            out.println();
            
            // index counter
            int i = 0;
            for(final Transition t : net.getTransitions()) {
            	String id = PNMLWriter.encode(t.getName());
            	while(!ids.add(id.toLowerCase())) {
	                ++i;
	                id = "t" + i;
            	}
            	tr.put(t, id);
                
                this.writeTransition(t, id, out);
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param t
     * @param id
     * @param theout
     */
    private void writeTransition(final Transition t, final String id, final PrintWriter out) {
        out.print("  <transition id=\"");
        out.print(id);
        out.println("\">");
        
        out.print("    <graphics><position page=\"1\" x=\"");
        out.print(t.getX());
        out.print("\" y=\"");
        out.print(t.getY());
        out.println("\" /></graphics>");
        
        out.println("    <name>");
        out.println("      <graphics><offset page=\"1\" x=\"15\" y=\"-15\" /></graphics>");
        out.print("      <value>");
        final String mean = PNMLWriter.encode(t.getMeaning());
        if(mean.isEmpty()) {
            out.print(id);
        } else {
            out.print(mean);
        }
        out.println("</value>");
        out.println("    </name>");
        
        out.println("  </transition>");
    }
    
    
    /**
     * TODO: document
     *
     * @param theout
     * @param pl
     * @param tr
     */
    private void writePTArcs(final PetriNet net, final PrintWriter out,
                    final Map<Place, String> pl, final Map<Transition, String> tr) {
        if(net.getPTArcCount() != 0) {
            out.println();
            
            int i = 0;
            for(final Place p : net.getPlaces()) {
                for(final PTArc arc : p.getOutput()) { 
                    ++i;
                    
                    final Transition t = arc.getTarget();
                    
                    final String pid = pl.get(p);
                    final String tid = tr.get(t);
                    
                    this.writeArc(pid, tid, arc.getWeight(), "pt" + i, out);
                }
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param theout
     * @param tr
     * @param pl
     */
    private void writeTPArcs(final PetriNet net, final PrintWriter out,
                    final Map<Transition, String> tr, final Map<Place, String> pl) {
        if(net.getTPArcCount() != 0) {
            
            int i = 0;
            for(final Transition t : net.getTransitions()) {
                for(final TPArc arc : t.getOutput()) {
                    ++i;
                    
                    final Place p = arc.getTarget();
                    
                    final String pid = pl.get(p);
                    final String tid = tr.get(t);
                    
                    this.writeArc(tid, pid, arc.getWeight(), "tp" + i, out);
                }
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param from
     * @param to
     * @param weight
     * @param id
     * @param theout
     */
    private void writeArc(final String from, final String to, final int weight,
            final String id, final PrintWriter out) {
        out.print("  <arc id=\"");
        out.print(id);
        out.print("\" source=\"");
        out.print(from);
        out.print("\" target=\"");
        out.print(to);
        
        if(weight <= 1) {
            out.println("\" />");
        } else {
            out.println("\">");
            out.print("    <inscription><value>");
            out.print(weight);
            out.println("</value></inscription>");
            out.println("  </arc>");
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param txt
     * @return
     */
    private static String encode(final String txt) {
        final StringBuilder b = new StringBuilder();
        for(final char c : txt.toCharArray()) {
            switch(c) {
                case '<': b.append("&lt;"); break;
                case '>': b.append("&gt;"); break;
                case '"': b.append("&quot;"); break;
                case '&': b.append("&amp;"); break;
                default: b.append(c); break;
            }
        }
        return b.toString();
    }
    
    
    /**
     * TODO: document
     *
     * @param txt
     * @return
     */
    private static String encodeComment(final String txt) {
        final StringBuilder b = new StringBuilder();
        boolean lastMinus = false;
        for(final char c : txt.toCharArray()) {
            switch(c) {
                case '<': b.append("&lt;"); lastMinus = false; break;
                case '>': b.append("&gt;"); lastMinus = false; break;
                case '"': b.append("&quot;"); lastMinus = false; break;
                case '&': b.append("&amp;"); lastMinus = false; break;
                case '-':
                    if(lastMinus) {
                        b.append(".-");
                        lastMinus = false;
                    } else {
                        b.append('-');
                        lastMinus = true;
                    }
                    break;
                default: b.append(c); lastMinus = false; break;
            }
        }
        return b.toString();
    }


	/* (non-Javadoc)
	 * @see petruchio.interfaces.Resettable#reset()
	 */
	@Override
    public void reset() {
	    if(this.theout != null) {
            this.theout.close();
            this.theout = null;
        }
	    
	    if(this.aOut != null) {
            this.aOut.close();
            this.aOut = null;
        }
	}


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#beginNet(java.lang.String)
     */
    @Override
    public void beginNet(final String name) {
        this.theout.println("<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>");
        this.theout.println("<pnml>");
        this.theout.println("<net id=\"n1\" type=\"PTNet\">");
        this.theout.print("  <name><value>");
        this.theout.print(PNMLWriter.encode(name));
        this.theout.println("</value></name>");
        
        try {
            final Random r = new Random();
            int i = 0;
            
            do {
                this.aOutTmp = File.createTempFile("pnmlstreamwriter_arcs" + r.nextInt(Integer.MAX_VALUE), null);
            } while(!this.aOutTmp.canWrite() && ++i < 100);
            if(i >= 100) {
                throw new RuntimeException("Unable to create temporary file. But tried 100 times. Last attempt: " + this.aOutTmp);
            }
            
            this.aOutTmp.deleteOnExit();
            this.aOut = new PrintWriter(this.aOutTmp);
        } catch(final Exception e) {
            e.printStackTrace();
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#endNet()
     */
    @Override
    public void endNet() {
        this.aOut.close();
        try {
            byte[] buf = new byte[8192];
            int len;
            InputStream in = new FileInputStream(this.aOutTmp);
            while ((len = in.read(buf)) > 0) {
                this.theout.print(new String(buf, 0, len));
            }
            in.close();
        } catch(final Exception e) {
            e.printStackTrace();
        }
        
        try {
            if(!this.aOutTmp.delete()) {
                System.err.println("Unable to delete temporary file " + this.aOutTmp);
            }
        } finally {
            this.theout.println("</net>");
            this.theout.print("</pnml>");
            this.theout.close();
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#setOutput(java.io.OutputStream)
     */
    @Override
    public void setOutput(final OutputStream output) {
        this.theout = new PrintWriter(output);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeHeaderComment(java.lang.String)
     */
    @Override
    public void writeHeaderComment(final String comment) {
        this.theout.println("  <!--");
        final String[] lines = PNMLWriter.encodeComment(comment).split("\r\n|[\n\r\u2028\u2029\u0085]");
        for(final String line : lines) {
            this.theout.print("    - ");
            this.theout.println(line);
        }
        this.theout.println("    -->");
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writePTArc(java.lang.String, java.lang.String, int)
     */
    @Override
    public void writePTArc(final String name, final String name2, final int weight) {
        this.writeArc(name, name2, weight, name+name2, this.aOut);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writePlace(petruchio.interfaces.petrinet.Place)
     */
    @Override
    public void writePlace(final Place p) {
        this.writePlace(p, p.getName(), this.theout);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeTPArc(java.lang.String, java.lang.String, int)
     */
    @Override
    public void writeTPArc(final String name, final String name2, final int weight) {
        this.writeArc(name, name2, weight, name+name2, this.aOut);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeTransition(petruchio.interfaces.petrinet.Transition)
     */
    @Override
    public void writeTransition(final Transition t) {
        this.writeTransition(t, t.getName(), this.theout);
    }
    
}