/*
 * Project:  Petruchio
 * Filename: SVGSimWriter.java
 * Created:  15.10.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.writers;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Map;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.PetriNetWriter;
import petruchio.interfaces.petrinet.Transition;
import petruchio.interfaces.petrinet.PetriNet;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class SVGSimWriter implements SelfDescribing, PetriNetWriter {

    private static final float size = 25f;
    private static final float halfSize = size/2f;
    private static final float textOffset = 2f*size/3f;
    private static final float tokenOffsetY = size/5f;
    private static final String visible = "inline";
    private static final String invisible = "none";
    private static final String enabled = "lemonchiffon";
    private static final String disabled = "white";
    private static final String fontsize = "10pt";
    private static final String tokenPrefix = "token of ";
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net writer that outputs in the scalable vector graphics"
            + " (SVG) format with simulation capatibilities using ECMAScript.";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.pn.PetriNetWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "svg";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#write(petruchio.interfaces.petrinet.PetriNet, java.io.OutputStream)
     */
    @Override
    public synchronized void write(final PetriNet net, final OutputStream outStream) {
        final PrintWriter out = new PrintWriter(outStream);
        
        this.beginNet(net, out);
        this.writeHeaderComment(net, out);

        this.writePTArcs(net, out);
        
        this.writeTPArcs(net, out);
        
        final Collection<String> ids = new HashSet<String>();
        
        final Map<Place, String> pl = new IdentityHashMap<Place, String>();
        this.writePlaces(net, out, pl, ids);
        
        final Map<Transition, String> tr = new IdentityHashMap<Transition, String>();
        this.writeTransitions(net, out, tr, ids);
        
        this.endNet(out);
        
        out.close();
    }
    
    
    /**
     * TODO: Document
     *
     * @param net
     * @param out
     */
    private void beginNet(final PetriNet net, final PrintWriter out) {
        int minX = Integer.MAX_VALUE;
        int minY = Integer.MAX_VALUE;
        int maxX = Integer.MIN_VALUE;
        int maxY = Integer.MIN_VALUE;
        for(final Place p : net.getPlaces()) {
            if(p.getX() < minX) {
                minX = p.getX();
            }
            if(p.getY() < minY) {
                minY = p.getY();
            }
            if(p.getX() > maxX) {
                maxX = p.getX();
            }
            if(p.getY() > maxY) {
                maxY = p.getY();
            }
        }
        for(final Transition t : net.getTransitions()) {
            if(t.getX() < minX) {
                minX = t.getX();
            }
            if(t.getY() < minY) {
                minY = t.getY();
            }
            if(t.getX() > maxX) {
                maxX = t.getX();
            }
            if(t.getY() > maxY) {
                maxY = t.getY();
            }
        }
        out.println("<?xml version=\"1.0\" encoding=\"iso-8859-1\" standalone='no'?>");
        out.println("<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\"");
        out.println("          \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">");
        out.println("<svg xmlns=\"http://www.w3.org/2000/svg\"");
        out.println("     xmlns:ev=\"http://www.w3.org/2001/xml-events\"");
        out.println("     version=\"1.1\" baseProfile=\"full\"");
        out.print("     width=\"100%\" height=\"100%\" viewBox=\"");
        out.print(minX);
        out.print(" ");
        out.print(minY);
        out.print(" ");
        out.print(maxX);
        out.print(" ");
        out.print(maxY);
        out.println("\">");
        out.print("  <title>");
        out.print(SVGSimWriter.encode(net.getName()));
        out.println("</title>");
        out.println("  <defs>");
        out.println("<marker id=\"arrowHead\" viewBox=\"0 0 10 10\" refX=\"10\" "
                + "refY=\"5\" markerUnits=\"strokeWidth\" markerWidth=\"8\" "
                + "markerHeight=\"15\" orient=\"auto\"><path d=\"M 0,0 l 10,5 l"
                + " -10,5 z\" /></marker>");
        out.println("    <script type=\"text/ecmascript\"><![CDATA[");
        out.println("var locked = false;");
        out.println("var tin = new Array();");
        out.println("var tchange = new Array();");
        out.println("var pout = new Array();");
        final Map<Place, Integer> in = new IdentityHashMap<Place, Integer>();
        for(final Transition t : net.getTransitions()) {
            out.println("tin['" + t.getName() + "'] = new Array()");
            int i = 0;
            for(final PTArc arc : t.getInput()) {
                out.print("tin['" + t.getName() + "'][" + i + "] = new Object(); ");
                out.print("tin['" + t.getName() + "'][" + i + "].pid = '" + SVGSimWriter.tokenPrefix + arc.getSource().getName() + "'; ");
                out.println("tin['" + t.getName() + "'][" + i + "].weight = " + arc.getWeight() + ";");
                ++i;
                in.put(arc.getSource(), arc.getWeight());
            }
            out.println("tchange['" + t.getName() + "'] = new Array()");            
            i = 0;
            for(final TPArc arc : t.getOutput()) {
                final Integer consume = in.remove(arc.getTarget());
                final int w = (consume != null) ? arc.getWeight()-consume : arc.getWeight();
                if(w != 0) {
                    out.print("tchange['" + t.getName() + "'][" + i + "] = new Object(); ");
                    out.print("tchange['" + t.getName() + "'][" + i + "].pid = '" + SVGSimWriter.tokenPrefix + arc.getTarget().getName() + "'; ");
                    out.println("tchange['" + t.getName() + "'][" + i + "].weight = " + w + ";");
                    ++i;
                }
            }
            for(final Map.Entry<Place, Integer> e : in.entrySet()) {
                out.print("tchange['" + t.getName() + "'][" + i + "] = new Object(); ");
                out.print("tchange['" + t.getName() + "'][" + i + "].pid = '" + SVGSimWriter.tokenPrefix + e.getKey().getName() + "'; ");
                out.println("tchange['" + t.getName() + "'][" + i + "].weight = " + (-e.getValue()) + ";");
                ++i;
            }
            in.clear();
        }
        for(final Place p : net.getPlaces()) {
            final String pid = SVGSimWriter.tokenPrefix + p.getName();
            out.println("pout['" + pid + "'] = new Array()");
            int i = 0;
            for(final PTArc arc : p.getOutput()) {
                out.print("pout['" + pid + "'][" + i + "] = new Object(); ");
                out.print("pout['" + pid + "'][" + i + "].tid = '" + arc.getTarget().getName() + "'; ");
                out.println("pout['" + pid + "'][" + i + "].weight = " + arc.getWeight() + ";");
                ++i;
            }
        }
        out.println("function fire(evt) {");
        out.println("  var trans = evt.target;");
        out.println("  if(!locked && trans.getAttributeNS(null, 'fill') == '" + SVGSimWriter.enabled + "') {");
        out.println("    locked = true;");
        out.println("    change(trans);");
        out.println("    updateEnabledness(trans);");
        out.println("    locked = false;");
        out.println("  }");
        out.println("}");
        out.println("function isEnabled(trans) {");
        out.println("  var tid = trans.getAttributeNS(null, 'id');");
        out.println("  var doc = trans.ownerDocument;");
        out.println("  for(var i = 0; i < tin[tid].length; ++i) {");
        out.println("    var arc = tin[tid][i];");
        out.println("    var token = doc.getElementById(arc.pid);");
        out.println("    if(arc.weight > parseInt(token.firstChild.data)) {");
        out.println("      return false;");
        out.println("    }");
        out.println("  }");
        out.println("  return true;");
        out.println("}");
        out.println("function change(trans) {");
        out.println("  var tid = trans.getAttributeNS(null, 'id');");
        out.println("  var doc = trans.ownerDocument;");
        out.println("  for(var i = 0; i < tchange[tid].length; ++i) {");
        out.println("    var arc = tchange[tid][i];");
        out.println("    var token = doc.getElementById(arc.pid);");
        out.println("    var before = parseInt(token.firstChild.data);");
        out.println("    var after = before+arc.weight;");
        out.println("    token.firstChild.data = after;");
        out.println("    if(before != 0 && after == 0) {");
        out.println("      token.setAttributeNS(null, 'display', '" + SVGSimWriter.invisible + "');");
        out.println("    } else if(before == 0 && after != 0) {");
        out.println("      token.setAttributeNS(null, 'display', '" + SVGSimWriter.visible + "');");
        out.println("    }");
        out.println("  }");
        out.println("}");
        out.println("function updateEnabledness(trans) {");
        out.println("  var tid = trans.getAttributeNS(null, 'id');");
        out.println("  var doc = trans.ownerDocument;");
        out.println("  for(var i = 0; i < tchange[tid].length; ++i) {");
        out.println("    var arc = tchange[tid][i];");
        out.println("    var token = doc.getElementById(arc.pid);");
        out.println("    var m = parseInt(token.firstChild.data);");
        out.println("    for(var j = 0; j < pout[arc.pid].length; ++j) {");
        out.println("      var out = pout[arc.pid][j];");
        out.println("      var t = doc.getElementById(out.tid)");
        out.println("      if(isEnabled(t)) {");
        out.println("        t.setAttributeNS(null, 'fill', '" + SVGSimWriter.enabled + "');");
        out.println("      } else {");
        out.println("        t.setAttributeNS(null, 'fill', '" + SVGSimWriter.disabled + "');");
        out.println("      }");
        out.println("    }");
        out.println("  }");
        out.println("}");
        out.println("    ]]></script>");
        out.println("  </defs>");
    }
    
    
    /**
     * TODO: Document
     *
     * @param out
     */
    private void endNet(final PrintWriter out) {
        out.println("</svg>");
    }
    
    
    /**
     * TODO: document
     *
     * @param out
     */
    private void writeHeaderComment(final PetriNet net, final PrintWriter out) {
        if(net.hasHeaderComment()) {
            out.println("  <!--");
            final String[] lines = SVGSimWriter.encodeComment(net.getHeaderComment()).split("\r\n|[\n\r\u2028\u2029\u0085]");
            for(final String line : lines) {
                out.print("    - ");
                out.println(line);
            }
        } else {
            out.println("  <!--");
        }
        out.print("    - ");
        out.print(net.getPlaces().size());
        out.print(" places, ");
        out.print(net.getTransitions().size());
        out.print(" transitions, ");
        out.print(net.getPTArcCount());
        out.print(" pt-arcs, ");
        out.print(net.getTPArcCount());
        out.println(" tp-arcs");
        out.println("    -->");
    }

    
    /**
     * TODO: document
     *
     * @param out
     * @param pl
     */
    private void writePlaces(final PetriNet net, final PrintWriter out,
                    final Map<Place, String> pl, final Collection<String> ids) {
        if(!net.getPlaces().isEmpty()) {
            out.println();
        
            // index counter
            int i = 0;
            for(final Place p : net.getPlaces()) {
                String id = SVGSimWriter.encode(p.getName());
                while(!ids.add(id.toLowerCase())) {
                    ++i;
                    id = "p" + i;
                }
                pl.put(p, id);
                
                this.writePlace(p, id, out);
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param p
     * @param id
     * @param out
     */
    private void writePlace(final Place p, final String id, final PrintWriter out) {
        out.println();
        final String note = SVGSimWriter.getNote(p);
        if(note != null) {
            out.print("    <!-- ");
            out.print(note);
            out.println(" -->");
        }
        out.print("  <circle fill=\"white\" stroke=\"black\" id=\"");
        out.print(id);
        out.print("\" r=\"");
        out.print(SVGSimWriter.halfSize);
        out.print("\" cx=\"");
        out.print(p.getX());
        out.print("\" cy=\"");
        out.print(p.getY());
        out.println("\" />");

        out.print("  <text text-anchor=\"middle\" font-size=\"" + SVGSimWriter.fontsize + "\" id=\"");
        out.print(SVGSimWriter.tokenPrefix);
        out.print(id);
        out.print("\" x=\"");
        out.print(p.getX());
        out.print("\" y=\"");
        out.print(p.getY() + SVGSimWriter.tokenOffsetY);
        out.print("\" display=\"");
        if(p.getMarking() == 0) {
            out.print(SVGSimWriter.invisible);
        } else {
            out.print(SVGSimWriter.visible);
        }
        out.print("\">");
        out.print(p.getMarking());
        out.println("</text>");
        
        out.print("  <text font-size=\"" + SVGSimWriter.fontsize + "\" x=\"");
        out.print(p.getX()+SVGSimWriter.textOffset);
        out.print("\" y=\"");
        out.print(p.getY()+SVGSimWriter.textOffset);
        out.print("\">");
        out.print(SVGSimWriter.encode(p.getMeaning()));
        out.println("</text>");
    }
    
    
    /**
     * @param p
     * @return
     */
    private static String getNote(final Place p) {
        final StringBuilder note = new StringBuilder();
        
        if(p.getBound() == Place.BOUND_UNBOUNDED) {
            note.append("unbounded");
        } else if(p.getBound() != Place.BOUND_UNKNOWN) {
            note.append(p.getBound());
            note.append("-bounded");
        }
        
        if((p.getNote().length() != 0) && (p.getBound() != Place.BOUND_UNKNOWN)) {
            note.append(", ");
            note.append(p.getNote());
        }
        
        if(note.length() != 0) {
            return SVGSimWriter.encodeComment(note.toString());
        }
        return null;
    }
    
    
    /**
     * TODO: document
     *
     * @param out
     * @param tr
     */
    private void writeTransitions(final PetriNet net, final PrintWriter out,
                    final Map<Transition, String> tr, final Collection<String> ids) {
        if(!net.getTransitions().isEmpty()) {

            out.println();
            
            // index counter
            int i = 0;
            for(final Transition t : net.getTransitions()) {
                String id = SVGSimWriter.encode(t.getName());
                while(!ids.add(id.toLowerCase())) {
                    ++i;
                    id = "t" + i;
                }
                tr.put(t, id);
                
                this.writeTransition(t, id, out);
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param t
     * @param id
     * @param out
     */
    private void writeTransition(final Transition t, final String id, final PrintWriter out) {
        out.println();
        out.print("  <rect stroke=\"black\" id=\"");
        out.print(id);
        out.print("\" x=\"");
        out.print(t.getX()-SVGSimWriter.halfSize);
        out.print("\" y=\"");
        out.print(t.getY()-SVGSimWriter.halfSize);
        out.print("\" width=\"");
        out.print(SVGSimWriter.size);
        out.print("\" height=\"");
        out.print(SVGSimWriter.size);
        out.print("\" fill=\"");
        if(this.isEnabled(t)) {
            out.print(SVGSimWriter.enabled);
        } else {
            out.print(SVGSimWriter.disabled);
        }
        out.println("\" onclick=\"fire(evt)\" />");
        
        out.print("  <text font-size=\"" + SVGSimWriter.fontsize + "\" x=\"");
        out.print(t.getX()+SVGSimWriter.textOffset);
        out.print("\" y=\"");
        out.print(t.getY()+SVGSimWriter.textOffset);
        out.print("\">");
        out.print(SVGSimWriter.encode(t.getMeaning()));
        out.println("</text>");
    }
    
    
    /**
     * TODO: Document
     *
     * @param t
     * @return
     */
    private boolean isEnabled(final Transition t) {
        for(final PTArc arc : t.getInput()) {
            if(arc.getWeight() > arc.getSource().getMarking()) {
                return false;
            }
        }
        return true;
    }
    
    
    /**
     * TODO: document
     *
     * @param out
     */
    private void writePTArcs(final PetriNet net, final PrintWriter out) {
        if(net.getPTArcCount() != 0) {
            out.println();
            
            int i = 0;
            for(final Place p : net.getPlaces()) {
                for(final PTArc arc : p.getOutput()) { 
                    ++i;
                    
                    final Transition t = arc.getTarget();
                    
                    this.writeArc(true, p.getX(), p.getY(), t.getX(), t.getY(),
                            arc.getWeight(), "pt" + i, out);
                }
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param out
     * @param tr
     * @param pl
     */
    private void writeTPArcs(final PetriNet net, final PrintWriter out) {
        if(net.getTPArcCount() != 0) {
            
            int i = 0;
            for(final Transition t : net.getTransitions()) {
                for(final TPArc arc : t.getOutput()) {
                    ++i;
                    
                    final Place p = arc.getTarget();
                    this.writeArc(false, t.getX(), t.getY(), p.getX(), p.getY(),
                            arc.getWeight(), "tp" + i, out);
                }
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param pt
     * @param x1
     * @param y1
     * @param x2
     * @param y2
     * @param weight
     * @param id
     * @param out
     */
    private void writeArc(final boolean pt, final int x1, final int y1,
            final int x2, final int y2, final int weight, final String id,
            final PrintWriter out) {
        
        final int dx = x2-x1;
        final int dy = y2-y1;
        final float l = (float)Math.sqrt(dx*dx+dy*dy);
        
        final float xx1;
        final float yy1;
        final float xx2;
        final float yy2;
        
        if(pt) {
            xx1 = x1 + dx*SVGSimWriter.halfSize/l;
            yy1 = y1 + dy*SVGSimWriter.halfSize/l;
            
            if(dx == 0) {
                xx2 = x2;
                yy2 = y2 - Math.signum(dy)*SVGSimWriter.halfSize;
            } else {
                double m = 1d*dy/dx;
                if(Math.abs(m) > 1) { m = 1/m; }
                final float dist = SVGSimWriter.halfSize/(float)Math.cos(Math.atan(m));
                xx2 = x2 - dx*dist/l;
                yy2 = y2 - dy*dist/l;
            }
        } else {
            if(dx == 0) {
                xx1 = x1;
                yy1 = y1 + Math.signum(dy)*SVGSimWriter.halfSize;
            } else {
                double m = 1d*dy/dx;
                if(Math.abs(m) > 1) { m = 1/m; }
                final float dist = SVGSimWriter.halfSize/(float)Math.cos(Math.atan(m));
                xx1 = x1 + dx*dist/l;
                yy1 = y1 + dy*dist/l;
            }
            
            xx2 = x2 - dx*SVGSimWriter.halfSize/l;
            yy2 = y2 - dy*SVGSimWriter.halfSize/l;
        }
        
        
        out.print("  <line stroke=\"black\" marker-end=\"url(#arrowHead)\" id=\"");
        out.print(id);
        out.print("\" x1=\"");
        out.print(xx1);
        out.print("\" y1=\"");
        out.print(yy1);
        out.print("\" x2=\"");
        out.print(xx2);
        out.print("\" y2=\"");
        out.print(yy2);
        out.print("\" />");
        
        if(weight != 1) {
            out.print("  <text id=\"weight ");
            out.print(id);
            out.print("\" x=\"");
            out.print((x1+x2)/2f);
            out.print("\" y=\"");
            out.print((x1+x2)/2f);
            out.print("\">");
            out.print(weight);
            out.println("</text>");
        }
        out.println();
    }
    
    
    /**
     * TODO: document
     *
     * @param txt
     * @return
     */
    private static String encode(final String txt) {
        final StringBuilder b = new StringBuilder();
        for(final char c : txt.toCharArray()) {
            switch(c) {
                case '<': b.append("&lt;"); break;
                case '>': b.append("&gt;"); break;
                case '"': b.append("&quot;"); break;
                case '&': b.append("&amp;"); break;
                default: b.append(c); break;
            }
        }
        return b.toString();
    }
    
    
    /**
     * TODO: document
     *
     * @param txt
     * @return
     */
    private static String encodeComment(final String txt) {
        final StringBuilder b = new StringBuilder();
        boolean lastMinus = false;
        for(final char c : txt.toCharArray()) {
            switch(c) {
                case '<': b.append("&lt;"); lastMinus = false; break;
                case '>': b.append("&gt;"); lastMinus = false; break;
                case '"': b.append("&quot;"); lastMinus = false; break;
                case '&': b.append("&amp;"); lastMinus = false; break;
                case '-':
                    if(lastMinus) {
                        b.append(".-");
                        lastMinus = false;
                    } else {
                        b.append('-');
                        lastMinus = true;
                    }
                    break;
                default: b.append(c); lastMinus = false; break;
            }
        }
        return b.toString();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        // no caches
    }
    
}