/*
 * Project:  Petruchio
 * Filename: LaTeXWriter.java
 * Created:  08.10.2007
 * Author:   Tim Strazny
 */
package petruchio.pn.writers;

import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileInputStream;

import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.PetriNetWriter;
import petruchio.interfaces.petrinet.PetriNetStreamWriter;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.Transition;

/**
 * TODO: description 
 *
 * @author Tim Strazny
 */
public class TikZWriter implements SelfDescribing, PetriNetWriter, PetriNetStreamWriter {
    
    private static final float pageWidth = 15f;
    private static final float pageHeight = 23f;
    private OutputStream theoutput;
    private PrintWriter theout;
    private File tmp;
    private PrintWriter tmpOut; 
    private File tmpArcs;
    private PrintWriter tmpArcsOut;
    private int minX;
    private int minY;
    private int maxX;
    private int maxY;
    private boolean hasArcs;

    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "A Petri net writer, that outputs the net in LaTeX using the PGF/TikZ package by Till Tantau."
            + " It can be found at http://pgf.sf.net/";
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#getDefaultExtension()
     */
    @Override
    public String getDefaultExtension() {
        return "tex";
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetWriter#write(petruchio.interfaces.petrinet.PetriNet, java.io.OutputStream)
     */
    @Override
    public void write(final PetriNet net, final OutputStream output) {
        final PrintWriter out = new PrintWriter(output);
        
        this.minX = Integer.MAX_VALUE;
        this.minY = Integer.MAX_VALUE;
        this.maxX = Integer.MIN_VALUE;
        this.maxY = Integer.MIN_VALUE;
        for(final Place p : net.getPlaces()) {
            final int x = p.getX();
            if(x > this.maxX) {
                this.maxX = p.getX();
            }
            if(x < this.minX) {
                this.minX = x;
            }
            final int y = -p.getY();
            if(y > this.maxY) {
                this.maxY = y;
            }
            if(y < this.minY) {
                this.minY = y;
            }
        }
        
        for(final Transition t : net.getTransitions()) {
            final int x = t.getX();
            if(x > this.maxX) {
                this.maxX = t.getX();
            }
            if(x < this.minX) {
                this.minX = x;
            }
            final int y = -t.getY();
            if(y > this.maxY) {
                this.maxY = y;
            }
            if(y < this.minY) {
                this.minY = y;
            }
        }
        final float x = TikZWriter.pageWidth / (this.maxX-this.minX == 0 ? 1 : this.maxX-this.minX); 
        final float y = TikZWriter.pageHeight / (this.maxY-this.minY == 0 ? 1 : this.maxY-this.minY);
        
        this.writeHeaderComment(net, out);
        
        final float unit = Math.min(1, Math.min(x, y));
        out.print("\\documentclass[a4paper]{article}\n\\usepackage{tikz}\n"
                + "\\usepackage{geometry}\n\\usetikzlibrary{petri}\n\\pagestyle{empty}\n\\begin{document}"
                + "\n\\begin{figure}\n\\centering\n\\begin{tikzpicture}[->,p/.style={place,fill=white,minimum size="+25*unit+"cm},"
                + "t/.style={transition,fill=white,minimum size="+25*unit+"cm},auto,x="+unit+"cm,y="+unit+"cm,f/.style={gray},b/.style{lightgray},"
                + "every token/.style={minimum size="+5*unit+"cm}]\n");
        
        this.writePlaces(net, out);
        
        out.print("\n");
        
        this.writeTransitions(net, out);
        
        if(net.getTPArcCount() != 0 || net.getPTArcCount() != 0) {
            out.print("\\begin{pgfonlayer}{background}\n\\draw");
            if(net.getPTArcCount() != 0) {
                this.writePTArcs(net, out);
            }
        
            if(net.getTPArcCount() != 0) {
                this.writeTPArcs(net, out);
            }
            out.print(";\n\\end{pgfonlayer}\n");
        }
        out.print("\\end{tikzpicture}\n\\end{figure}\n\\end{document}");
        out.close();
    }
    
    
    /**
     * TODO: document
     *
     * @param theout
     */
    private void writeHeaderComment(final PetriNet net, final PrintWriter out) {
        if(net.hasHeaderComment()) {
            final String[] lines = net.getHeaderComment().split("\r\n|[\n\r\u2028\u2029\u0085]");
            for(final String line : lines) {
                out.print("% ");
                out.print(line);
                out.print("\n");
            }
        }
        out.print("% ");
        out.print(net.getPlaces().size());
        out.print(" places, ");
        out.print(net.getTransitions().size());
        out.print(" transitions, ");
        out.print(net.getPTArcCount());
        out.print(" pt-arcs, ");
        out.print(net.getTPArcCount());
        out.print(" tp-arcs\n\n");
    }
    
    
    /**
     * TODO: description
     *
     * @param net
     * @param theout
     */
    private void writePlaces(final PetriNet net, final PrintWriter out) {
        for(final Place p : net.getPlaces()) {
            this.writePlace(p, out);
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param p
     * @param theout
     */
    private void writePlace(final Place p, final PrintWriter out) {
        out.print("\\node[p");
        if(p.getMarking() != 0) {
            out.print(", tokens=");
            out.print(p.getMarking());
        }
        final String m = this.encode(p.getMeaning());
        if(!m.isEmpty()) {
            out.print(", label={above:");
            out.print(m);
            out.print("}");
        }
        out.print("] (");
        out.print(this.encode(p.getName()));
        out.print(") at (");
        out.print(p.getX());
        out.print(", ");
        out.print(-p.getY());
        out.print("){};");
        if(p.getNote() != null && p.getNote().length() != 0) {
            out.print("  % ");
            out.print(p.getNote());
        }
        out.print("\n");
    }
    
    
    /**
     * TODO: description
     *
     * @param net
     * @param theout
     */
    private void writeTransitions(final PetriNet net, final PrintWriter out) {
        for(final Transition t : net.getTransitions()) {
            this.writeTransition(t, out);
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param t
     * @param theout
     */
    private void writeTransition(final Transition t, final PrintWriter out) {
        out.print("\\node[t");
        final String m = this.encode(t.getMeaning());
        if(!m.isEmpty()) {
            out.print(", label={above:");
            out.print(m);
            out.print("}");
        }
        out.print("] (");
        out.print(this.encode(t.getName()));
        out.print(") at (");
        out.print(t.getX());
        out.print(", ");
        out.print(-t.getY());
        out.print("){};\n");
    }
    
    
    /**
     * TODO: description
     *
     * @param net
     * @param theout
     */
    private void writePTArcs(final PetriNet net, final PrintWriter out) {
        for(final Place p : net.getPlaces()) {
            for(final PTArc arc : p.getOutput()) {
                this.writeArc(arc.getSource().getName(), arc.getTarget().getName(), arc.getWeight(), out);
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param net
     * @param theout
     */
    private void writeTPArcs(final PetriNet net, final PrintWriter out) {
        for(final Transition t : net.getTransitions()) {
            for(final TPArc arc : t.getOutput()) {
                this.writeArc(arc.getSource().getName(), arc.getTarget().getName(), arc.getWeight(), out);
            }
        }
    }
    
    
    /**
     * TODO: description
     *
     * @param from
     * @param to
     * @param w
     * @param theout
     */
    private void writeArc(final String from, final String to, final int w, final PrintWriter out) {
        out.print(" (");
        out.print(this.encode(from));
        out.print(") edge ");
        if(w != 1) {
            out.print("node {");
            out.print(w);
            out.print("} ");
        }
        out.print("(");
        out.print(this.encode(to));
        out.print(")");
    }
    
    
    /**
     * TODO: description
     *
     * @param s
     * @return
     */
    private String encode(final String s) {
        final StringBuilder b = new StringBuilder(s.length());
        for(final char ch : s.toCharArray()) {
            switch(ch) {
                case '%': b.append("\\%"); break;
                case '$': b.append("\\$"); break;
                case '_': b.append("\\_"); break;
                case '{': b.append("\\{"); break;
                case '}': b.append("\\}"); break;
                case '<': b.append("\\langle{}"); break;
                case '>': b.append("\\rangle{}"); break;
                case '\n': b.append(" "); break;
                default: b.append(ch);
            }
        }
        return b.toString();
    }
    


    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    public void reset() {
        if(this.theout != null) {
            this.theout.close();
            this.theout = null;
        }
        if(this.tmpOut != null) {
            this.tmpOut.close();
            this.tmpOut = null;
        }
        if(this.tmpArcsOut != null) {
            this.tmpArcsOut.close();
            this.tmpArcsOut = null;
        }
        this.theoutput = null;
        this.tmpArcs = null;
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#beginNet(java.lang.String)
     */
    @Override
    public void beginNet(final String name) {
        this.minX = Integer.MAX_VALUE;
        this.minY = Integer.MAX_VALUE;
        this.maxX = Integer.MIN_VALUE;
        this.maxY = Integer.MIN_VALUE;
        this.hasArcs = false;
        try {
            this.tmp = File.createTempFile("latexstreamwriter_", null);
            this.tmpArcs = File.createTempFile("latexstreamwriterarcs_", null);
            this.tmp.deleteOnExit();
            this.tmpOut = new PrintWriter(this.tmp);
            this.tmpArcs.deleteOnExit();
            this.tmpArcsOut = new PrintWriter(this.tmpArcs);
        } catch(final IOException e) {
            // ignore;
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#endNet()
     */
    @Override
    public void endNet() {
        this.tmpOut.close();
        this.tmpOut = null;
        this.tmpArcsOut.close();
        this.tmpArcsOut = null;
        
        final float x = TikZWriter.pageWidth / (this.maxX-this.minX == 0 ? 1 : this.maxX-this.minX); 
        final float y = TikZWriter.pageHeight / (this.maxY-this.minY == 0 ? 1 : this.maxY-this.minY);
        
        final float unit = Math.min(1, Math.min(x, y));
        this.theout.print("\\documentclass[a4paper]{article}\n\\usepackage{tikz}\n"
                + "\\usepackage{geometry}\n\\usetikzlibrary{petri}\n\\pagestyle{empty}\n\\begin{document}"
                + "\n\\begin{figure}\n\\centering\n\\begin{tikzpicture}[->,p/.style={place,fill=white,minimum size="+25*unit+"cm},"
                + "t/.style={transition,fill=white,minimum size="+25*unit+"cm},auto,x="+unit+"cm,y="+unit+"cm,f/.style={gray},b/.style{lightgray},"
                + "every token/.style={minimum size="+5*unit+"cm}]\n");
        
        this.theout.flush();
        
        try {
            byte[] buf = new byte[8192];
            int len;
            final InputStream in = new FileInputStream(this.tmp);
            while ((len = in.read(buf)) > 0) {
                this.theoutput.write(buf, 0, len);
            }
            in.close();
            
            if(!this.tmp.delete()) {
                System.err.println("Unable to delete temporary file " + this.tmp);
            }
            
        } catch(final FileNotFoundException e) {
            // ignore
            
        } catch(final IOException e) {
            // ignore
            
        }
        
        if(this.hasArcs) {
            this.theout.print("\\begin{pgfonlayer}{background}\n\\draw");
            this.theout.flush();
            try {
                byte[] buf = new byte[8192];
                int len;
                final InputStream in = new FileInputStream(this.tmpArcs);
                while ((len = in.read(buf)) > 0) {
                    this.theoutput.write(buf, 0, len);
                }
                in.close();
                
                if(!this.tmpArcs.delete()) {
                    System.err.println("Unable to delete temporary file " + this.tmpArcs);
                }
                
            } catch(final FileNotFoundException e) {
                // ignore
                
            } catch(final IOException e) {
                // ignore
                
            }
            this.theout.print(";\n\\end{pgfonlayer}\n");
        }
        
        this.theout.print("\\end{tikzpicture}\n\\end{figure}\n\\end{document}");
        this.theout.close();
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#setOutput(java.io.OutputStream)
     */
    @Override
    public void setOutput(final OutputStream output) {
        this.theoutput = output;
        this.theout = new PrintWriter(output);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeHeaderComment(java.lang.String)
     */
    @Override
    public void writeHeaderComment(final String comment) {
        final String[] lines = comment.split("\r\n|[\n\r\u2028\u2029\u0085]");
        for(final String line : lines) {
            this.theout.print("% ");
            this.theout.print(line);
            this.theout.print("\n");
        }
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writePTArc(java.lang.String, java.lang.String, int)
     */
    @Override
    public void writePTArc(final String name, final String name2, final int weight) {
        this.hasArcs = true;
        this.writeArc(name, name2, weight, this.tmpArcsOut);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writePlace(petruchio.interfaces.petrinet.Place)
     */
    @Override
    public void writePlace(final Place p) {
        final int x = p.getX();
        if(x > this.maxX) {
            this.maxX = p.getX();
        }
        if(x < this.minX) {
            this.minX = x;
        }
        final int y = -p.getY();
        if(y > this.maxY) {
            this.maxY = y;
        }
        if(y < this.minY) {
            this.minY = y;
        }
        this.writePlace(p, this.tmpOut);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeTPArc(java.lang.String, java.lang.String, int)
     */
    @Override
    public void writeTPArc(final String name, final String name2, final int weight) {
        this.hasArcs = true;
        this.writeArc(name, name2, weight, this.tmpArcsOut);
    }


    /* (non-Javadoc)
     * @see petruchio.interfaces.petrinet.PetriNetStreamWriter#writeTransition(petruchio.interfaces.petrinet.Transition)
     */
    @Override
    public void writeTransition(final Transition t) {
        final int x = t.getX();
        if(x > this.maxX) {
            this.maxX = t.getX();
        }
        if(x < this.minX) {
            this.minX = x;
        }
        final int y = -t.getY();
        if(y > this.maxY) {
            this.maxY = y;
        }
        if(y < this.minY) {
            this.minY = y;
        }
        this.writeTransition(t, this.tmpOut);
    }

}
