/*
 * Project:  UML2PN
 * Filename: MultiSet.java
 * Created:  02.03.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.petrinettool;

/**
 * Represents a petri net multi-set.
 * To add new ties like "{mytie(x, y, z)+, receive()-}" you would do the
 * following<br>
 * myset.addSendTie("mytie", "x, y, z");<br>
 * myset.addReceiveTie("receive", null);</code><br>
 * 
 * @author Tim Strazny
 * @see IPetriNetTool
 * @see IPlace
 * @see IPetriNet
 * @see ITransition
 */
public interface IMultiSet {

    /**
     * Creates a new sending tie and adds it to the MultiSet.
     * The label String is empty (or <code>null</code>) or it has to follow the
     * form (also see {@link IPetriNet#createArc(IPlace, ITransition, String)} or
     * {@link IPetriNet#createArc(IPlace, ITransition, String)})<br>
     * <code>
     * LABEL ::= &lt;ELEM&gt; {"," &lt;ELEM&gt;}<br>
     * ELEM &nbsp;::= &lt;SYMB&gt; | &lt;TUPLE&gt;<br>
     * TUPLE ::= "(" &lt;ELEM&gt; {"," &lt;ELEM&gt;} ")"<br>
     * SYMB &nbsp;::= &lt;VALUE&gt; | &lt;VARIABLE NAME&gt;<br>
     * VALUE ::= "'" &lt;STRING LITERAL&gt; "'" | &lt;NATURAL NUMBER&gt;
     * </code><br>
     * If the label is <code>null</code> or the empty String a single unique
     * variable will be used as a default. Notice that String literals are
     * enclosed in single quotes like 'dot', 'true'.
     * 
     * @param tie The tie symbol.
     * @param label See above.
     * @see #addReceiveTie(String, String)
     */
    public abstract void addSendTie(String tie, String label);
    
    
    /**
     * Creates a new receiving tie and adds it to the MultiSet.
     * The label String is empty (or <code>null</code>) or it has to follow the
     * form (also see {@link IPetriNet#createArc(IPlace, ITransition, String)} or
     * {@link IPetriNet#createArc(IPlace, ITransition, String)})<br>
     * <code>
     * LABEL ::= &lt;ELEM&gt; {"," &lt;ELEM&gt;}<br>
     * ELEM &nbsp;::= &lt;SYMB&gt; | &lt;TUPLE&gt;<br>
     * TUPLE ::= "(" &lt;ELEM&gt; {"," &lt;ELEM&gt;} ")"<br>
     * SYMB &nbsp;::= &lt;VALUE&gt; | &lt;VARIABLE NAME&gt;<br>
     * VALUE ::= "'" &lt;STRING LITERAL&gt; "'" | &lt;NATURAL NUMBER&gt;
     * </code><br>
     * If the label is <code>null</code> or the empty String a single unique
     * variable will be used as a default. Notice that String literals are
     * enclosed in single quotes like 'dot', 'true'.
     * 
     * @param tie The tie symbol.
     * @param label See above.
     * @see #addSendTie(String, String)
     */
    public abstract void addReceiveTie(String tie, String label);
    
    
    /**
     * Creates a new not conjugated action symbol and adds it to the MultiSet.
     * The label String is empty (or <code>null</code>) or it has to follow the
     * form (also see {@link IPetriNet#createArc(IPlace, ITransition, String)} or
     * {@link IPetriNet#createArc(IPlace, ITransition, String)})<br>
     * <code>
     * LABEL ::= &lt;SYMB&gt; {"," &lt;SYMB&gt;}<br>
     * SYMB &nbsp;::= &lt;VALUE&gt; | &lt;VARIABLE NAME&gt;<br>
     * VALUE ::= "'" &lt;STRING LITERAL&gt; "'" | &lt;NATURAL NUMBER&gt;
     * </code><br>
     * If the label is <code>null</code> the empty String will be used as a
     * default. Notice that String literals are enclosed in single quotes like
     * 'dot', 'true'.
     * 
     * @param action The action symbol.
     * @param label See above.
     * @see #addReceiveTie(String, String)
     */
    public abstract void addActionSymbol(String action, String label);
    
    
    /**
     * Creates a new conjugated action symbol and adds it to the MultiSet.
     * The label String is empty (or <code>null</code>) or it has to follow the
     * form (also see {@link IPetriNet#createArc(IPlace, ITransition, String)} or
     * {@link IPetriNet#createArc(IPlace, ITransition, String)})<br>
     * <code>
     * LABEL ::= &lt;SYMB&gt; {"," &lt;SYMB&gt;}<br>
     * SYMB &nbsp;::= &lt;VALUE&gt; | &lt;VARIABLE NAME&gt;<br>
     * VALUE ::= "'" &lt;STRING LITERAL&gt; "'" | &lt;NATURAL NUMBER&gt;
     * </code><br>
     * If the label is <code>null</code> the empty String will be used as a
     * default. Notice that String literals are enclosed in single quotes like
     * 'dot', 'true'.
     * 
     * @param action The action symbol.
     * @param label See above.
     * @see #addReceiveTie(String, String)
     */
    public abstract void addConjugatedActionSymbol(String action, String label);
    
    
    /**
     * Copies all the actions from the given set to <code>this</code> set.
     *
     * @param actions The multi set of actions to add.
     */
    public abstract void addSet(IMultiSet actions);
    
    
    /**
     * Returns a deep copy of this MultiSet.
     * 
     * @return A deep copy of this MultiSet.
     */
    public abstract Object clone();
}