/*
 * Project:  UML2PN
 * Filename: PetriNet.java
 * Created:  02.03.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.petrinettool;

import java.util.Collection;

/**
 * Representation of a petri net.
 *  
 * @author Tim Strazny
 * @see IPetriNetTool
 * @see IMultiSet
 * @see IPlace
 * @see ITransition
 */
public interface IPetriNet extends Cloneable {

    /* (non-Javadoc)
     * @see java.lang.Object#clone()
     */
    public abstract Object clone();


    /**
     * Returns a new place after adding it to the petri net.
     * 
     * The markings String is empty (or <code>null</code>) or it has to follow
     * the form<br>
     * <code>
     * MARKING ::= &lt;ELEM&gt; {"," &lt;ELEM&gt;}<br>
     * ELEM &nbsp;::= &lt;VALUE&gt; | &lt;TUPLE&gt;<br>
     * TUPLE ::= "(" &lt;ELEM&gt; {"," &lt;ELEM&gt;} ")"<br>
     * VALUE ::= "'" &lt;STRING LITERAL&gt; "'" | &lt;NATURAL NUMBER&gt;
     * </code><br>
     * If the marking is <code>null</code> the empty String will be used as a
     * default. Notice that String literals are enclosed in single quotes like
     * 'dot', 'true'.
     * 
     * @param name The (unique) name of the place.
     * @param status The place's status.
     * @param type What type of elements the place may hold.
     * @param initialMarking The initial marking. See above for syntax.           
     * @return The newly added Place object.
     * @see #createPlace(IPlaceStatus, IPlaceType, String)
     */
    public abstract IPlace createPlace(String name, IPlaceStatus status,
                    IPlaceType type, String initialMarking);
    
    public abstract IPlace createPlace(String name, IPlaceStatus status,
                    IPlaceType type, String initialMarking, String currentMarking);
    
    public abstract IPlace createPlace(String name, IPlaceStatus status,
                    IPlaceType type, boolean allowAnyInput,
                    String initialMarking, String currentMarking);
    
    
    /**
     * Returns a new place after adding it to the petri net.
     * The place will have some unique name.
     * 
     * The markings String is empty (or <code>null</code>) or it has to follow
     * the form<br>
     * <code>
     * MARKING ::= &lt;ELEM&gt; {"," &lt;ELEM&gt;}<br>
     * ELEM &nbsp;::= &lt;VALUE&gt; | &lt;TUPLE&gt;<br>
     * TUPLE ::= "(" &lt;ELEM&gt; {"," &lt;ELEM&gt;} ")"<br>
     * VALUE ::= "'" &lt;STRING LITERAL&gt; "'" | &lt;NATURAL NUMBER&gt;
     * </code><br>
     * If the marking is <code>null</code> the empty String will be used as a
     * default. Notice that String literals are enclosed in single quotes like
     * 'dot', 'true'.
     * 
     * @param status The place's status.
     * @param type What type of elements the place may hold. 
     * @param initialMarking The initial marking. See above for syntax.           
     * @return The newly added Place object.
     * @see #createPlace(String, IPlaceStatus, IPlaceType, String)
     */
    public abstract IPlace createPlace(IPlaceStatus status, IPlaceType type,
                    String initialMarking);
    
    public abstract IPlace createPlace(IPlaceStatus status, IPlaceType type,
                    String initialMarking, String currentMarking);


    /**
     * Returns a new transition after adding it to the petri net.
     * The guard String is empty (or <code>null</code>) or it has to follow
     * the form<br>
     * <code>
     * GUARD ::= "'true'" | &lt;AND&gt; | &lt;NOT&gt; | &lt;OR&gt; | &lt;COMP&gt;<br>
     * NOT &nbsp;&nbsp;::= "not" &lt;GUARD&gt;<br>
     * AND &nbsp;&nbsp;::= "(" &lt;GUARD&gt; " and "  &lt;GUARD&gt; ")"<br>  
     * OR &nbsp;&nbsp;&nbsp;::= "(" &lt;GUARD&gt; " or " &lt;GUARD&gt; ")"<br>
     * COMP ::= "(" &lt;EXPR&gt; &lt;COP&gt; &lt;EXPR&gt; ")"<br>
     * COP ::= "=" | ">" | "<" | ">=" | "<="<br>
     * EXPR &nbsp;::= &lt;SYMB&gt; | &lt;SYMB&gt; &lt;OP&gt; &lt;SYMB&gt;<br>
     * OP &nbsp;&nbsp;&nbsp;::= "+" | "-" | "*" | "/" | "%"<br>
     * SYMB &nbsp;::= &lt;VALUE&gt; | &lt;VARIABLE NAME&gt;<br>
     * VALUE ::= "'" &lt;STRING LITERAL&gt; "'" | &lt;NATURAL NUMBER&gt;
     * </code><br>
     * If the guard is <code>null</code> or the empty String 'true' will be used
     * as a default. Notice that String literals are enclosed in single quotes
     * like 'dot', 'true'.
     * 
     * @param name The (unique) name of the transition.
     * @param actionEvents The MultiSet object will be copied so that you may
     *                  re-use it. If it is <code>null</code> it is assumed to
     *                  be an empty set.
     * @param guard See above.
     * @return The newly added Transition object.
     * @see #createTransition(IMultiSet, String)
     */
    public abstract ITransition createTransition(String name,
                    IMultiSet actionEvents, String guard);


    /**
     * Returns a new transition after adding it to the petri net.
     * The transition will have some unique name.
     * The guard String is empty (or <code>null</code>) or it has to follow
     * the form<br>
     * <code>
     * GUARD ::= "'true'" | &lt;AND&gt; | &lt;NOT&gt; | &lt;OR&gt; | &lt;COMP&gt;<br>
     * NOT &nbsp;&nbsp;::= "not" &lt;GUARD&gt;<br>
     * AND &nbsp;&nbsp;::= "(" &lt;GUARD&gt; " and "  &lt;GUARD&gt; ")"<br>  
     * OR &nbsp;&nbsp;&nbsp;::= "(" &lt;GUARD&gt; " or " &lt;GUARD&gt; ")"<br>
     * COMP ::= "(" &lt;EXPR&gt; &lt;COP&gt; &lt;EXPR&gt; ")"<br>
     * COP ::= "=" | ">" | "<" | ">=" | "<="<br>
     * EXPR &nbsp;::= &lt;SYMB&gt; | &lt;SYMB&gt; &lt;OP&gt; &lt;SYMB&gt;<br>
     * OP &nbsp;&nbsp;&nbsp;::= "+" | "-" | "*" | "/" | "%"<br>
     * SYMB &nbsp;::= &lt;VALUE&gt; | &lt;VARIABLE NAME&gt;<br>
     * VALUE ::= "'" &lt;STRING LITERAL&gt; "'" | &lt;NATURAL NUMBER&gt;
     * </code><br>
     * If the guard is <code>null</code> or the empty String 'true' will be used
     * as a default. Notice that String literals are enclosed in single quotes
     * like 'dot', 'true'.
     * 
     * @param actionEvents The MultiSet object will be copied so that you may
     *                  re-use it. If it is <code>null</code> it is assumed to
     *                  be an empty set.
     * @param guard See above.
     * @return The newly added Transition object.
     * @see #createTransition(String, IMultiSet, String)
     */
    public abstract ITransition createTransition(IMultiSet actionEvents,
                    String guard);
    
    
    /**
     * Creates a new labelled arc from a place to a transition.
     * The label String is empty (or <code>null</code>) or it has to follow the
     * form<br>
     * <code>
     * LABEL ::= &lt;ELEM&gt; {"," &lt;ELEM&gt;}<br>
     * ELEM &nbsp;::= &lt;SYMB&gt; | &lt;TUPLE&gt;<br>
     * TUPLE ::= "(" &lt;ELEM&gt; {"," &lt;ELEM&gt;} ")"<br>
     * SYMB &nbsp;::= &lt;VALUE&gt; | &lt;VARIABLE NAME&gt;<br>
     * VALUE ::= "'" &lt;STRING LITERAL&gt; "'" | &lt;NATURAL NUMBER&gt;
     * </code><br>
     * If the label is <code>null</code> or the empty String a single unique
     * variable will be used as a default. Notice that String literals are
     * enclosed in single quotes like 'dot', 'true'.
     * 
     * @param from The place the new arc originates from.
     * @param to The destination transition.
     * @param label A list of variables and tuples of variables as
     *          described above.
     * @see #createArc(ITransition, IPlace, String)
     */
    public abstract void createArc(IPlace from, ITransition to, String label);
    

    /**
     * Creates a new labelled arc from a transition to a place.
     * The label String is empty (or <code>null</code>) or it has to follow the
     * form<br>
     * <code>
     * LABEL ::= &lt;ELEM&gt; {"," &lt;ELEM&gt;}<br>
     * ELEM &nbsp;::= &lt;SYMB&gt; | &lt;TUPLE&gt;<br>
     * TUPLE ::= "(" &lt;ELEM&gt; {"," &lt;ELEM&gt;} ")"<br>
     * SYMB &nbsp;::= &lt;VALUE&gt; | &lt;VARIABLE NAME&gt;<br>
     * VALUE ::= "'" &lt;STRING LITERAL&gt; "'" | &lt;NATURAL NUMBER&gt;
     * </code><br>
     * If the label is <code>null</code> or the empty String a single unique
     * variable will be used as a default. Notice that String literals are
     * enclosed in single quotes like 'dot', 'true'.
     * 
     * @param from The transition the new arc originates from.
     * @param to The destination place.
     * @param label A list of variables and tuples of variables as
     *          described above.
     * @see #createArc(IPlace, ITransition, String)
     */
    public abstract void createArc(ITransition from, IPlace to, String label);
    
    public abstract void createInhibitorArc(ITransition trans, IPlace place, String label);
    
    public abstract void createResetArc(ITransition trans, IPlace place, String label);
    
    public abstract void createReadArc(ITransition trans, IPlace place, String label);


    /**
     * Copies a whole net into <code>this</code> net.
     * All the places, transitions and arcs are copied with it.
     * After performing the operation this net contains (at least) two separate
     * nets as no new arcs are created.
     * The nets may not have transitions or places with same names otherwise
     * they will be renamed.
     * 
     * @param netB The net to insert.
     *          The net will not be altered by this operation.
     * @see #mergePlaces(IPlace, IPlace)
     * @see #concatenate(IPetriNet, int)
     * @see #tieAll(String)
     */
    public abstract void insert(IPetriNet netB);
    
    
    /**
     * Works just like {@link #insert(IPetriNet)} but adds offsetX and offsetY
     * to the position of each inserted element.
     *
     * @param netB The net to insert.
     *          The net will not be altered by this operation.
     * @param offsetX The offset in X direction
     * @param offsetY The offset in Y direction
     * @see #insert(IPetriNet)
     */
    public abstract void insert(IPetriNet netB, int offsetX, int offsetY);


    /**
     * Creates a new place by merging the two given places.
     * One of the places has to have status {@link IPlace#STATUS_ENTRY}
     * and the other place has to have status {@link IPlace#STATUS_EXIT} the new
     * place will have an internal status, that each implementation may define
     * itself. The markings are added and a unique name is given.<br>
     * The old places are removed after merging and may not be used
     * afterwards.
     * 
     * @param placeA A place.
     * @param placeB Another place.
     * @return The new place that was created.
     * @see #insert(IPetriNet)
     * @see #concatenate(IPetriNet, int)
     * @see #tieAll(String)
     */
    public abstract IPlace mergePlaces(IPlace placeA, IPlace placeB);

    
    /**
     * Performs tie operations on all tie symbols that match the regular
     * expression in the petri net.
     * The two transitions with the tie-symbol and the conjugated tie-symbol (or
     * sending tie and receiving tie) are searched and between them a new place
     * is created with the given type. An arc is created from the transition
     * with the sending tie to the newly created place and labeled with the
     * label that was attached to the sending tie symbol. Analogously an second
     * arc is created from the place to the transition with the receiving tie
     * and labeled with the label that was attached to the receiving tie.<br>
     * The tie symbols are removed from the transitions afterwards. The place's
     * initial marking is empty and it's status is some kind of internal that
     * the specific implementation may define.<br>
     * There may be exactly two Transitions with the tie symbol: one sending and
     * one receiving.
     * 
     * @param regexp 
     * @see #mergePlaces(IPlace, IPlace)
     * @see #insert(IPetriNet)
     * @see #concatenate(IPetriNet, int)
     * @see #tieAll(String)
     */
    public abstract void tieAll(String regexp);

    
    /**
     * Performs a concatenation where <code>count</code> exit places of
     * <code>this</code> net's are merged with <code>net</code>'s entry places.
     * The net will be copied into <code>this</code> net, then
     * <code>count</code> randomly chosen exit places (from <code>this</code>)
     * and <code>count</code> randomly chosen entry places
     * (from <code>net</code>) are merged in pairs.
     * Normally <code>count</code> would be 1 and each of the nets would contain
     * exactly one entry/exit place. Entry/exit places that are left over become
     * internal places.
     * The nets may not have transitions or places with same names otherwise
     * they will be renamed.
     * 
     * @param net The net to concatenate.
     *          The net will not be altered by this operation. 
     * @param count The number of places to be merged. It must be less or equal
     *              to the number of exit/entry places present!
     * @see #mergePlaces(IPlace, IPlace)
     * @see #insert(IPetriNet)
     * @see #tieAll(String)
     */
    public abstract void concatenate(IPetriNet net, int count);

  
    /**
     * Removes a place from the net.
     * All arcs connected to the place are removed as well.
     * After this method has been called with <code>Place p</code>,
     * <code>p</code> may not be used again in any way. 
     * 
     * @param place The place to remove.
     * @see #createPlace(String, IPlaceStatus, IPlaceType, String)
     * @see #createPlace(IPlaceStatus, IPlaceType, String)
     * @see #getPlaceByName(String)
     * @see #removeTransition(ITransition)
     */
    public abstract void removePlace(IPlace place);


    /**
     * Removes a transition from the net.
     * All arcs connected to the transition are removed as well.
     * After this method has been called with <code>Transition t</code>,
     * <code>t</code> may not be used again in any way.
     * 
     * @param transition The transition to remove.
     * @see #createTransition(IMultiSet, String)
     * @see #createTransition(String, IMultiSet, String)
     * @see #getTransitionByName(String)
     * @see #removePlace(IPlace)
     */
    public abstract void removeTransition(ITransition transition);

    /**
     * Returns the place with the given name.
     * 
     * @param name The name of the places to get.
     * @return The place with the name.
     * @see #getTransitionByName(String)
     */
    public abstract IPlace getPlaceByName(String name);


    /**
     * Returns The transition with the given name.
     * 
     * @param name The name of the transition to get.
     * @return The transition with the name.
     * @see #getPlaceByName(String)
     */
    public abstract ITransition getTransitionByName(String name);


    /**
     * Returns the PNML representation of the petri net with all its elements.
     * 
     * @return The PNML representation of the petri net.
     * @see #toPEP()
     */
    public abstract String toPNML();
    
    
    /**
     * Returns the PEP-format representation of the petri net with all its elements.
     * 
     * @return The PEP-format representation of the petri net.
     * @see #toPNML()
     */
    public abstract String toPEP();  
    

    /**
     * Returns an array containing all the places from the net that are marked
     * as entry places (i.e. with status value {@link IPlace#STATUS_ENTRY}).
     * 
     * @return All contained entry places.
     * @see #getExitPlaces()
     * @see #getPlaceByName(String) 
     */
    public abstract IPlace[] getEntryPlaces();
    
    
    /**
     * Returns an array containing all the places from the net that are marked
     * as exit places (i.e. with status value {@link IPlace#STATUS_EXIT}).
     * 
     * @return All contained exit places.
     * @see #getEntryPlaces()
     * @see #getPlaceByName(String) 
     */
    public abstract IPlace[] getExitPlaces();

    
    /**
     * Returns the largest y coordinate of all the elements in the net.
     *
     * @return The largest y coordinate of all the elements in the net. 
     */
    public abstract int getMinY();
    
    
    /**
     * Returns the smallest y coordinate of all the elements in the net.
     *
     * @return The smallest y coordinate of all the elements in the net. 
     */
    public abstract int getMaxY();

    
    /**
     * Returns the smallest x coordinate of all the elements in the net.
     *
     * @return The smallest x coordinate of all the elements in the net. 
     */
    public abstract int getMinX();
    
    
    /**
     * Returns the largest x coordinate of all the elements in the net.
     *
     * @return The largest x coordinate of all the elements in the net. 
     */
    public abstract int getMaxX();
    
    
    /**
     * Translates all the net's elements by the specified values.
     *
     * @param xOffset The value to add to each element's x coordinate 
     * @param yOffset The value to add to each element's y coordinate
     */
    public abstract void translate(int xOffset, int yOffset);
    
    
    public abstract void optimize();
    
    
    public abstract void setIsLowLevelNet(boolean value);
    
    public abstract boolean isLowLevelNet();

    public abstract boolean usesTime();
    
    public abstract boolean usesCapacity();
    
    public abstract boolean usesPriority();
    
    public abstract void addFunction(String function);
    
    public abstract Collection<String> getFunctions();
}