/*
 * Project:  UML2PN
 * Filename: petruchio.sim.petrinettool.java
 * Created:  02.03.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.petrinettool;

import java.util.regex.Pattern;

/**
 * The place to start when using or implementing this interface package.
 * To start working with this interface, instantiate a concrete implementation 
 * of petruchio.sim.petrinettool like
 * <code>PetriNetTool tool = new myimplementation.MyPetriNetTool()</code>.
 * PetriNets and MultiSets should not be instantiated directly -- use
 * {@link #createMultiSet()} and {@link #createPetriNet()} instead.
 * All relevant methods to manipulate petri nets are contained in
 * {@link IMultiSet}, {@link IPetriNet}, {@link IPlace} and {@link ITransition}.
 * Please call {@link #close()} to allow the tool to do some clean up when you
 * don't need it any more.
 * 
 * @author Tim Strazny
 */
public interface IPetriNetTool {
   
    public static final char STRING_SEPARATOR = '\"';
    
    public static final String TRUE = "true";
    public static final String FALSE = "false";
    public static final String DOT = "dot";
    
    public static final String LOGOP = "!=|<>|!|&|&&|\\||\\|\\||\\b(or|and|not)\\b";
    public static final String OP = "[=<>+\\-*/%,;#]|==|:=|" + LOGOP;
    public static final Pattern OPERATORS = Pattern.compile(OP);
    public static final Pattern GUARDVARS =
        // (\(|<OP>|^) \s* ([a-zA-Z_]\w*)  \s* (\)|<OP>|$)
        // ( or <op>       <variable name>      <op> or )
        //   or start      capture group 2       or end
        Pattern.compile("(^|(?<=\\(|"+OP+"))\\s*"      // some unimportant prefix
                      + "((?!"+TRUE+"|"+FALSE+"|"+DOT+")[a-zA-Z_.][a-zA-Z0-9._]*)" // the variable
                      + "\\s*($|(?=\\)|"+OP+"))");     // some unimportant suffix
    
    public void setAnyType(IPlaceType type);
    
    public IPlaceType getAnyType();
    
    public void setTieType(String symbol, IPlaceType type);
    
    public void clearTieTypes();
    
    
    /**
     * Creates a new PetriNet for the implemented interface.
     * 
     * @return A fresh PetriNet object.
     * @see #createMultiSet()
     */
    public IPetriNet createPetriNet();
    
    
    /**
     * Creates a new MultiSet for the implemented interface.
     * 
     * @return A fresh MuliSet object.
     * @see #createPetriNet()
     */
    public IMultiSet createMultiSet();
    
    
    /**
     * Returns whether the {@link IPetriNet#toPEP()} method may be called.
     * 
     * @return <code>true</code> if the positions of elements are inserted into
     *      the {@link IPetriNet#toPEP()}- and {@link IPetriNet#toPNML()}-output. 
     */
    public abstract boolean writesPositions();
    
    
//    /**
//     * To a given guard, the method extracts all the variable names and returns
//     * them as an array. The guard follows the syntax described in
//     * {@link PetriNet#createTransition(String, MultiSet, String)}.
//     * For example:<br>
//     * <code>
//     * GUARD = (not (((A=B+1) and (C=D-10)) or ((7*E=8*F) and (G/3=H))) or
//     *         ("true" and (A=H))<br>
//     * VARS &nbsp;= ["A", "B", "C", "D", "E", "F", "G", "H"]<br>
//     * </code>
//     * 
//     * @param guard The guard to scan for variable names.
//     * @return An array containing all the variable names from the guard.
//     * @see PetriNet#createTransition(String, MultiSet, String)
//     * @see PetriNet#createTransition(MultiSet, String)
//     */
//    public String[] getVariablesFromGuard(String guard);
    
    
    /**
     * Perform all the needed clean up operations.
     */
    public void close();
         
}