/*
 * Project:  UML2PN
 * Filename: Place.java
 * Created:  02.03.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.petrinettool;

/**
 * Represents a petri net place.
 * To create a new place call
 * {@link IPetriNet#createPlace(String, IPlaceStatus, IPlaceType, String)} or
 * {@link IPetriNet#createPlace(IPlaceStatus, IPlaceType, String)}.
 * There are three kinds of states allowed: {@link #STATUS_ENTRY},
 * {@link #STATUS_EXIT}, {@link #STATUS_INTERNAL}. These are important when
 * performing a {@link IPetriNet#concatenate(IPetriNet, int)} operation. Take
 * notice that ENTRY and EXIT places may only be of type
 * {@link #TYPE_BLACK_TOKEN}.
 *
 * @author Tim Strazny
 * @see IPetriNetTool
 * @see IMultiSet
 * @see IPetriNet
 * @see ITransition
 */
public interface IPlace {
    
    public static final int UNBOUNDED = -1;
    
    /**
     * Status code for an entry place.
     * Entry places may only be of type {@link #TYPE_BLACK_TOKEN}.
     * See {@link IPetriNet#concatenate(IPetriNet, int)} for details.
     * 
     * @see #STATUS_EXIT
     * @see #STATUS_INTERNAL
     * @see IPetriNet#createPlace(String, IPlaceStatus, IPlaceType, String)
     * @see IPetriNet#createPlace(IPlaceStatus, IPlaceType, String)
     * @see IPetriNet#concatenate(IPetriNet, int)
     */
    public static final IPlaceStatus STATUS_ENTRY = petruchio.sim.petrinettool.status.Entry.ENTRY;
    
    
    /**
     * Status code for an internal place.
     * 
     * @see #STATUS_ENTRY
     * @see #STATUS_EXIT 
     * @see IPetriNet#createPlace(String, IPlaceStatus, IPlaceType, String)
     * @see IPetriNet#createPlace(IPlaceStatus, IPlaceType, String)
     */
    public static final IPlaceStatus STATUS_INTERNAL = petruchio.sim.petrinettool.status.Internal.INTERNAL;

    
    /**
     * Status code for an exit place.
     * Exit places may only be of type {@link #TYPE_BLACK_TOKEN}.
     * See {@link IPetriNet#concatenate(IPetriNet, int)} for details.
     * 
     * @see #STATUS_ENTRY
     * @see #STATUS_INTERNAL
     * @see IPetriNet#createPlace(String, IPlaceStatus, IPlaceType, String)
     * @see IPetriNet#createPlace(IPlaceStatus, IPlaceType, String)
     * @see IPetriNet#concatenate(IPetriNet, int)
     */
    public static final IPlaceStatus STATUS_EXIT = petruchio.sim.petrinettool.status.Exit.EXIT;
    

    public static final String STRING_ANY = petruchio.sim.petrinettool.type.Any.STRING_ANY;
    
    
    public static final IPlaceType TYPE_ANY = petruchio.sim.petrinettool.type.Any.ANY;
    
    
    /**
     * Type code for a place that may hold (simple) black tokens. 
     * 
     * @see IPetriNet#createPlace(String, IPlaceStatus, IPlaceType, String)
     * @see IPetriNet#createPlace(IPlaceStatus, IPlaceType, String)
     */
    public static final IPlaceType TYPE_BLACK_TOKEN = petruchio.sim.petrinettool.type.BlackToken.BLACK_TOKEN;
    
    
    /**
     * Type code for a place that may boolean values (tt, ff) 
     * 
     * @see IPetriNet#createPlace(String, IPlaceStatus, IPlaceType, String)
     * @see IPetriNet#createPlace(IPlaceStatus, IPlaceType, String)
     */
    public static final IPlaceType TYPE_BOOLEAN = petruchio.sim.petrinettool.type.Boolean.BOOLEAN;
    
    
    /**
     * Assigns the new name to the place.
     * The name must be unique.
     * 
     * @param name The new (unique) name of the place.
     */
    public abstract void rename(String name);
    
    
    /**
     * Returns an array of all the input transitions of this place
     * (the pre-set).
     * 
     * @return All the input transitions of this place. 
     * @see #getSuccessorTransitions()
     */
    public abstract ITransition[] getPredecessorTransitions();
    
    
    /**
     * Returns an array of all the output transitions of this place
     * (the post-set).
     * 
     * @return All the output transitions of this place.
     * @see #getPredecessorTransitions()
     */
    public abstract ITransition[] getSuccessorTransitions();
    
    
    /**
     * Returns the status value of this place.
     * Notice that other than the &quot;usual&quot; values may be returned as
     * some implementations may define their own additional status values.
     * 
     * @return The status value.
     * @see #STATUS_ENTRY
     * @see #STATUS_EXIT
     * @see #STATUS_INTERNAL
     */
    public abstract IPlaceStatus getStatus();
    
    
    /**
     * Sets the status of this Place to the new value.
     * Notice that you may not override the restrictions mentioned in
     * {@link IPetriNet#createPlace(String, IPlaceStatus, IPlaceType, String)}.
     * 
     * @param status The new status value.
     * @see #STATUS_ENTRY
     * @see #STATUS_EXIT
     * @see #STATUS_INTERNAL
     */
    public abstract void setStatus(IPlaceStatus status);
 
    
    /**
     * Returns the x coordinate of this place's position.
     *
     * @return The x coordinate of this place's position.
     * @see #getPosY()
     * @see #setPosX(int)
     * @see #setPosY(int)
     * @see #setPos(int, int) 
     */
    public abstract int getPosX();
    
    
    /**
     * Returns the y coordinate of this place's position.
     *
     * @return The y coordinate of this place's position.
     * @see #getPosX()
     * @see #setPosX(int)
     * @see #setPosY(int)
     * @see #setPos(int, int) 
     */
    public abstract int getPosY();
    
    
    /**
     * Sets the x coordinate of this place's position.
     *
     * @param posX The new x coordinate of this place's position.
     * @see #getPosX()
     * @see #getPosY()
     * @see #setPosX(int)
     * @see #setPosY(int)
     * @see #setPos(int, int) 
     */
    public abstract void setPosX(int posX);
    
    
    /**
     * Sets the y coordinate of this place's position.
     *
     * @param posY The new y coordinate of this place's position.
     * @see #getPosX()
     * @see #getPosY()
     * @see #setPosX(int)
     * @see #setPos(int, int) 
     */
    public abstract void setPosY(int posY);
    
    
    /**
     * Sets the x coordinate of this place's position.
     *
     * @param posX The new x coordinate of this place's position.
     * @param posY The new y coordinate of this place's position.
     * @see #getPosX()
     * @see #getPosY()
     * @see #setPosX(int)
     * @see #setPosY(int)
     */
    public abstract void setPos(int posX, int posY);

    
    public abstract void setMeaning(String meaning);
    
    public abstract String getMeaning();
   
    /**
     * Sets the place's capacity. Values less than 1 mean unbounded. 
     *
     * @param capacity
     */
    public abstract void setCapacity(int capacity);
    
    /**
     * Returns the place's capacity. Values less than 1 mean unbounded. 
     * @return the capacity
     */
    public abstract int getCapacity();
    
    
    public abstract boolean infiniteCapacity();
    
    
    public abstract boolean allowAnyInput();
    
    
    public abstract void setAllowAnyInput(boolean value);
}