/*
 * Project:  UML2PN
 * Filename: Transition.java
 * Created:  02.03.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.petrinettool;

/**
 * Represents a petri net transition.
 * Transitions may be created by using the
 * {@link IPetriNet#createTransition(String, IMultiSet, String)} or the
 * {@link IPetriNet#createTransition(IMultiSet, String)} method. See those for
 * more information.
 *
 * @author Tim Strazny
 * @see IPetriNetTool
 * @see IMultiSet
 * @see IPetriNet
 * @see IPlace
 */
public interface ITransition {
    
    /**
     * Returns an array of all the input places of this transition
     * (the pre-set).
     * 
     * @return All the input places of this transition.
     * @see #getSuccessorPlaces()
     */
    public abstract IPlace[] getPredecessorPlaces();
    
    
    /**
     * Returns an array of all the output places of this transition
     * (the post-set).
     * 
     * @return All the output places of this transition.
     * @see #getPredecessorPlaces()
     */
    public abstract IPlace[] getSuccessorPlaces();
    
    
    
    /**
     * Returns the multiset of actions that is attached to this transition.
     * 
     * @return The MultiSet object attached to this transition.
     */
    public abstract IMultiSet getActions();
    
    
    /**
     * Returns the x coordinate of this transition's position.
     *
     * @return The x coordinate of this transition's position.
     * @see #getPosY()
     * @see #setPosX(int)
     * @see #setPosY(int)
     * @see #setPos(int, int) 
     */
    public abstract int getPosX();
    
    
    /**
     * Returns the y coordinate of this transition's position.
     *
     * @return The y coordinate of this transition's position.
     * @see #getPosX()
     * @see #setPosX(int)
     * @see #setPosY(int)
     * @see #setPos(int, int) 
     */
    public abstract int getPosY();
    
    
    /**
     * Sets the x coordinate of this transition's position.
     *
     * @param posX The new x coordinate of this transition's position.
     * @see #getPosX()
     * @see #getPosY()
     * @see #setPosX(int)
     * @see #setPosY(int)
     * @see #setPos(int, int) 
     */
    public abstract void setPosX(int posX);
    
    
    /**
     * Sets the y coordinate of this transition's position.
     *
     * @param posY The new y coordinate of this transition's position.
     * @see #getPosX()
     * @see #getPosY()
     * @see #setPosX(int)
     * @see #setPos(int, int) 
     */
    public abstract void setPosY(int posY);
    
    
    /**
     * Sets the x coordinate of this place's position.
     *
     * @param posX The new x coordinate of this transition's position.
     * @param posY The new y coordinate of this transition's position.
     * @see #getPosX()
     * @see #getPosY()
     * @see #setPosX(int)
     * @see #setPosY(int)
     */
    public abstract void setPos(int posX, int posY);
    
    
    public abstract boolean isImmediate();

    
    public abstract void setMeaning(String meaning);
    
    
    public abstract String getMeaning();
    
    
    public abstract void setEarliestFiring(String deltatime);
    
    
    public abstract void setLatestFiring(String deltatime);
    
    
    public abstract String getEarliestFiring();
    
    
    public abstract String getLatestFiring();
    
    
    public abstract void setPriority(String expression);
    
    
    public abstract void setRate(String expression);
    
    
    public abstract void setWeight(String expression);
    
    
    public abstract void setGeneralizedRate(String expression);
    
    
    public abstract void setGeneralizedWeight(String expression);
    
    
    public abstract void setServers(String expression);
    
    
    public abstract void setAgeMemory(String expression);
    
    
    public abstract String getPriority();
    
    
    public abstract String getServers();
    
    
    public abstract String getWeight();
    
    
    public abstract String getRate();
    

    public abstract String getGeneralizedWeight();
    
    
    public abstract String getGeneralizedRate();
    
    
    public abstract String getAgeMemory();
}