/*
 * Project:  P-UMLaut
 * Filename: PEPReader.java
 * Created:  02.08.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.petrinettool;

import java.util.Map;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;

import petruchio.sim.petrinettool.type.IntegerRange;
import petruchio.sim.petrinettool.type.Set;
import petruchio.sim.petrinettool.type.SingleValue;
import petruchio.sim.petrinettool.type.Tuple;

public class PEPReader implements PetriNetReader {
    public static enum Section {
        UNKNOWN("", true),
        PLACES("PL"),
        TRANSITIONS("TR"),
        PLACE2TRANS("PT"),
        TRANS2PLACE("TP"),
        INHIBITOR_ARCS("IH"),
        RESET_ARCS("RS"),
        READ_ARCS("RD"),
        FUNCTIONS("FN"),
        TRANSITIONS_PHANTOM("PTR", true),
        PLACE2TRANS_PHANTOM("PPT", true),
        TRANS2PLACE_PHANTOM("PTP", true),
        BLOCKS("BL", true),
        HEADER("PEP", true);
        
        private static Map<String, Section> strings;
        
        private static boolean contains(final String key) {
            if(Section.strings == null) {
                Section.strings = new HashMap<String, Section>();
                return false;
            }
            return Section.strings.containsKey(key);
        }

        private static void add(final String key, final Section sec) {
            if(Section.strings == null) {
                Section.strings = new HashMap<String, Section>();
            }
            Section.strings.put(key, sec);
        }

        public static Section getSection(final String key) {
            final Section s = Section.strings.get(key);
            return s == null ? UNKNOWN : s;
        }
        
        
        private final String key;
        
        private final boolean ignore;
        
        Section(final String key) {
            this(key, false);
        }
        
        Section(final String key, final boolean ignore) {
            if(Section.contains(key)) {
                throw new IllegalArgumentException("There already exists an "
                                + "option with key '" + key + "'.");
            }
            this.key = key;
            this.ignore = ignore;
            Section.add(key, this);
        }
        
        
        public String getKey() {
            return this.key;
        }
        
        public boolean ignore() {
            return this.ignore;
        }
        
        @Override
        public String toString() {
            return this.key;
        }
    };
    
    public static final String TRANS_GUARD = "g";
    public static final String TRANS_MEANING = "b";
    public static final String TRANS_EARLIEST = "E";
    public static final String TRANS_LATEST = "L";
    public static final String TRANS_RATE = "r";
    public static final String TRANS_WEIGHT = "w";
    public static final String TRANS_GENERALIZED_RATE = "R";
    public static final String TRANS_GENERALIZED_WEIGHT = "W";
    public static final String TRANS_PRIORITY = "P";
    public static final String TRANS_SERVERS = "s";
    public static final String TRANS_AGE_MEMORY = "a";
    
    public static final String PLACE_ENTRY = "e";
    public static final String PLACE_EXIT = "x";
    public static final String PLACE_MEANING = "b";
    public static final String PLACE_TYPE = "Z";
    public static final String PLACE_INITIAL_MARKING = "z";
    public static final String PLACE_CURRENT_MARKING = "y";
    public static final String PLACE_CAPACITY = "k";
    public static final String PLACE_INITIAL_COUNT = "M";
    public static final String PLACE_CURRENT_COUNT = "m";
    public static final String PLACE_ANYTYPE = "*";
    
    public static final String ARC_WEIGHT = "w";
    public static final String ARC_INSCRIPTION = "p";
    
    public static final Pattern REM_QUOT = Pattern.compile("^\\s*\"\\s*(.*)\\s*\"\\s*$");
    public static final String REP_QUOT = "$1";
    
    public static final Pattern REM_BRACK = Pattern.compile("^\\s*\\{\\s*(.*)\\s*\\}\\s*$");
    public static final String REP_BRACK = "$1";
    
    public static final Pattern KEY_VALUE = Pattern.compile("(\\S)(\"(.*?)[^\\\\]\"|\"\"|\\d*(\\.\\d+|))");
    public static final int KEY = 1;
    public static final int VALUE = 2;
    
    public static final Pattern COMMENT = Pattern.compile("\\s*%.*");
    
    public static final Pattern PLACE = Pattern.compile("^\\s*(\\d*)\\s*(\"\\s*(.*?)\\s*\"|)\\s*((\\d+)@\\s*(\\d+)|)\\s*([ex]|)");
    public static final int PL_NR = 1;
    public static final int PL_NAME = 3;
    public static final int PL_X = 5;
    public static final int PL_Y = 6;
    public static final int PL_STATUS = 7;
    
    public static final Pattern TRANSITION = Pattern.compile("^\\s*(\\d*)\\s*(\"\\s*(.*?)\\s*\"|)\\s*((\\d+)@\\s*(\\d+)|)");
    public static final int TR_NR = 1;
    public static final int TR_NAME = 3;
    public static final int TR_X = 5;
    public static final int TR_Y = 6;
    public static final int TR_GUARD = 7;
    
    public static final Pattern PT_ARC = Pattern.compile("^\\s*(\\d+|\".*?\")\\s*>\\s*(\\d+|\".*?\")");
    public static final int PT_PLACE = 1;
    public static final int PT_TRANS = 2;
    
    public static final Pattern TP_ARC = Pattern.compile("^\\s*(\\d+|\".*?\")\\s*<\\s*(\\d+|\".*?\")");
    public static final int TP_PLACE = 2;
    public static final int TP_TRANS = 1;
    
    
    public static final Pattern TYPE_SET = Pattern.compile("\\s*\\{\\s*\\}\\s*");
    public static final Pattern TYPE_ANY = Pattern.compile("\\s*(\\{\\s*[*]\\s*\\}|[*])\\s*");
    public static final Pattern TYPE_DOT = Pattern.compile("\\s*(\\{\\s*dot\\s*\\}|dot)\\s*");
    public static final Pattern TYPE_BOOLEAN = Pattern.compile("\\s*(\\{\\s*\\s*(tt\\s*,\\s*ff|ff\\s*,\\s*tt)\\s*\\}|(tt\\s*,\\s*ff|ff\\s*,\\s*tt))\\s*");
    public static final Pattern TYPE_INT = Pattern.compile("\\s*\\{\\s*\\s*-?\\d+\\s*\\.\\.\\s*-?\\d+\\s*\\}\\s*");
    public static final Pattern TYPE_TUPLE = Pattern.compile("\\s*X\\s*\\{");
    
    
    private IPetriNet net;
    private IPetriNetTool tool;
    private final Map<String, ITransition> transitions;
    private final Map<String, IPlace> places;
    private PEPReader.Section part;
    private int transitionCount;
    private int placeCount;
    
    public PEPReader() {
        this.transitionCount = 0;
        this.placeCount = 0;
        this.transitions = new HashMap<String, ITransition>();
        this.places = new HashMap<String, IPlace>();
        this.part = PEPReader.Section.UNKNOWN;
        this.net = null;
        this.tool = null;
    }
    
    
    public synchronized IPetriNet readFromString(final IPetriNetTool theTool,
                    final String content) {
        this.tool = theTool;
        this.net = this.tool.createPetriNet();
        
        int startOfLine = 0;
        int endOfLine;
        String line = ""; 
        do {
            endOfLine = content.indexOf('\n', startOfLine);
            if(endOfLine < 0) {
                endOfLine = content.length();
            }
            line += content.substring(startOfLine, endOfLine);
            if(line.endsWith("\\")) {
                line = line.substring(0, line.length()-1);
            } else {
                try {
                    this.parseLine(line);
                    line = "";
                } catch(final RuntimeException e) {
                    throw new RuntimeException("Error reading line \"" + line + "\".", e);
                }
            }
            startOfLine = endOfLine + 1;
        } while(endOfLine < content.length());
        
        this.transitions.clear();
        this.places.clear();
        this.part = PEPReader.Section.UNKNOWN;
        this.tool = null;
        final IPetriNet theNet = this.net;
        this.net = null;
        return theNet;
    }
    
    
    public synchronized IPetriNet readFromFile(final IPetriNetTool theTool,
                    String filename) {
        File f = new File(filename);
        if(!f.isAbsolute()) {
            f = new File(System.getProperty("user.dir") + File.separator + filename);
        }
        try {
            filename = f.getCanonicalPath();
        } catch(IOException e) {
            filename = f.getAbsolutePath();
        }
        this.tool = theTool;
        this.net = this.tool.createPetriNet();
        this.net.setIsLowLevelNet(true);
        try {
            final BufferedReader in = new BufferedReader(new InputStreamReader(
                            new FileInputStream(filename)));
            String line = "";
            String curline;
            while((curline = in.readLine()) != null) {
                line += curline;
                if(line.endsWith("\\")) {
                    line = line.substring(0, line.length()-1);
                } else {
                    try {
                        this.parseLine(line);
                        line = "";
                    } catch(final RuntimeException e) {
                        throw new RuntimeException("Error reading line \"" + line + "\". Line type: "
                                + this.part, e);
                    }
                }
            }
            in.close();
        } catch(final IOException e) {
            throw new RuntimeException("Unable to read from file " + filename
                            + "\n" + e.getMessage());
        }
        this.transitions.clear();
        this.places.clear();
        this.part = PEPReader.Section.UNKNOWN;
        this.tool = null;
        final IPetriNet theNet = this.net;
        this.net = null;
        
        return theNet;
    }
    
    
    private void parseLine(String line) {
        line = line.trim();
        if(line.equals("PTNet")) {
            return;
        } else if(PEPReader.COMMENT.matcher(line).matches()) {
            // ignore comment
            return;
        } else {
            final PEPReader.Section p = PEPReader.Section.getSection(line);
            if(p != PEPReader.Section.UNKNOWN) {
                this.part = p;
                return;
            }
        }
        
        if(line.length() == 0) {
            return;
        }
        
        if(!this.part.ignore()) {
            switch(this.part) {
                case PLACES:
                    this.parsePlace(line);
                    break;
                case TRANSITIONS:
                    this.parseTransition(line);
                    break;
                case FUNCTIONS:
                    this.parseFunction(line);
                    break;
                case PLACE2TRANS:
                    this.parsePTArc(line);
                    break;
                case TRANS2PLACE:
                    this.parseTPArc(line);
                    break;
                case INHIBITOR_ARCS:
                    this.parseInhibitorArc(line);
                    break;
                case READ_ARCS:
                    this.parseReadArc(line);
                    break;
                case RESET_ARCS:
                    this.parseResetArc(line);
                    break;
                default:
                    System.err.println("Ignoring line: " + line);
                    break;
            }
        }
    }
    
    
    private void parseFunction(final String line) {
        this.net.addFunction(PEPReader.unescapeStrings(line));
    }
    
    
    private void parsePlace(final String line) {
        // <nr>"<name>"<x>@<y>[e|x][y"{<current>}"]Z"<type>"[z"{<initial>}"]
        final Matcher p = PEPReader.PLACE.matcher(line);
        if(!p.lookingAt()) {
            System.err.println("Unable to parse place: " + line);
            return;
        }
        String nr = p.group(PEPReader.PL_NR);
        String name = p.group(PEPReader.PL_NAME);
        String x = p.group(PEPReader.PL_X);
        String y = p.group(PEPReader.PL_Y);
        String status = p.group(PEPReader.PL_STATUS);
        x = (x == null || x.length() == 0) ? "0" : x;
        y = (y == null || y.length() == 0) ? "0" : y;
        
        if(name == null) {
            name = "";
        }
        
        if((nr == null) || (nr.length() == 0)) {
            nr = "" + ++this.placeCount;
        }
        
        if(status == null) {
            status = "";
        }
        
        if(name.length() == 0) {
            if(nr.length() == 0) {
                throw new RuntimeException("Unable to read the place's name in "
                                + "line: " + line);
            } else {
                name = "P" + nr;
            }
        }
        
        if(nr.length() == 0) {
            nr = name;
        }
        
        IPlaceStatus theStatus;
        if(status.equals(PEPReader.PLACE_ENTRY)) {
            theStatus = IPlace.STATUS_ENTRY;
        } else if(status.equals(PEPReader.PLACE_EXIT)) {
            theStatus = IPlace.STATUS_EXIT;
        } else {
            theStatus = IPlace.STATUS_INTERNAL;
        }
        
        int start = p.end();
        final Matcher kv = PEPReader.KEY_VALUE.matcher(line);
        final Matcher c = PEPReader.COMMENT.matcher(line);
        
        String type = null;
        String currentMarking = null;
        String initialMarking = null;
        String meaning = null;
        int capacity = IPlace.UNBOUNDED;
        boolean allowAnyInput = false;
        if(kv.find(start) && !(c.find(start) && (c.start() <= kv.start()))) {
            do {
                final String key = kv.group(PEPReader.KEY);
                String value = kv.group(PEPReader.VALUE);
                if(key.equals(PEPReader.PLACE_CURRENT_MARKING)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    value = PEPReader.REM_BRACK.matcher(value).replaceFirst(PEPReader.REP_BRACK);
                    currentMarking = PEPReader.unescapeStrings(value);
                } else if(key.equals(PEPReader.PLACE_INITIAL_MARKING)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    value = PEPReader.REM_BRACK.matcher(value).replaceFirst(PEPReader.REP_BRACK);
                    initialMarking = PEPReader.unescapeStrings(value);
                } else if(key.equals(PEPReader.PLACE_TYPE)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    type = value;
                } else if(key.equals(PEPReader.PLACE_MEANING)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    meaning = value;
                } else if(key.equals(PEPReader.PLACE_CAPACITY)) {
                    capacity = Integer.parseInt(value);
                } else if(key.equals(PEPReader.PLACE_ANYTYPE)) {
                    allowAnyInput = true;
                } else if(key.equals(PEPReader.PLACE_INITIAL_COUNT)){
                    final int count = Integer.parseInt(value);
                    initialMarking = "";
                    for(int i = 0; i < count; i++) {
                        initialMarking += IPetriNetTool.DOT;
                        if(i < count-1) {
                            initialMarking += ",";
                        }
                    }
                } else if(key.equals(PEPReader.PLACE_CURRENT_COUNT)){
                    int count = Integer.parseInt(value);
                    currentMarking = "";
                    for(int i = 0; i < count; i++) {
                        currentMarking += "dot";
                        if(i < count-1) {
                            currentMarking += ",";
                        }
                    }
                } else {
                    // ignored
                }
                start = kv.end();
            } while(kv.find() && !(c.find(start) && (c.start() <= kv.start())));
        }
        
        final IPlaceType theType = type == null ? null : PEPReader.parseType(type);
        
        if((theType != null) && !IPlace.TYPE_BLACK_TOKEN.equals(theType)) {
            this.net.setIsLowLevelNet(false);
        }
        final IPlace pl = this.net.createPlace(name, theStatus, theType, allowAnyInput, initialMarking, currentMarking);
        pl.setPos(Integer.parseInt(x), Integer.parseInt(y));
        pl.setMeaning(meaning);
        pl.setCapacity(capacity);
        this.places.put(nr, pl);
    }
    
    
    private void parseTransition(final String line) {
        final Matcher t = PEPReader.TRANSITION.matcher(line);
        if(!t.lookingAt()) {
            System.err.println("Unable to parse transition: " + line);
            return;
        }
        String nr = t.group(PEPReader.TR_NR);
        String name = t.group(PEPReader.TR_NAME);
        String x = t.group(PEPReader.TR_X);
        String y = t.group(PEPReader.TR_Y);
        x = (x == null || x.length() == 0) ? "0" : x;
        y = (y == null || y.length() == 0) ? "0" : y;
        
        if(name == null) {
            name = "";
        }
        
        if((nr == null) || (nr.length() == 0)) {
            nr = "" + ++this.transitionCount;
        }
        
        if(name.length() == 0) {
            if(nr.length() == 0) {
                throw new RuntimeException("Unable to read the transition's name.");
            } else {
                name = "T" + nr;
            }
        }
        
        if(nr.length() == 0) {
            nr = name;
        }
        
        int start = t.end();
        final Matcher kv = PEPReader.KEY_VALUE.matcher(line);
        final Matcher c = PEPReader.COMMENT.matcher(line);
        
        String guard = null;
        String meaning = null;
        String earliest = null;
        String latest = null;
        String rate = null;
        String weight = null;
        String grate = null;
        String gweight = null;
        String priority = null;
        String servers = null;
        String ageMemory = null;
        
        if(kv.find(start) && !(c.find(start) && (c.start() <= kv.start()))) {
            do {
                final String key = kv.group(PEPReader.KEY);
                String value = kv.group(PEPReader.VALUE);
//                System.out.println(key + " : " + value);
                if(key.equals(PEPReader.TRANS_GUARD)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    guard = PEPReader.unescapeStrings(value);
//                    System.out.println(guard);
                } else if(key.equals(PEPReader.TRANS_MEANING)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    meaning = value;
                } else if(key.equals(PEPReader.TRANS_EARLIEST)) {
                    earliest = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    earliest = PEPReader.unescapeStrings(earliest);
                } else if(key.equals(PEPReader.TRANS_LATEST)) {
                    latest = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    latest = PEPReader.unescapeStrings(latest);
                } else if(key.equals(PEPReader.TRANS_RATE)) {
                    rate = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    rate = PEPReader.unescapeStrings(rate);
                } else if(key.equals(PEPReader.TRANS_WEIGHT)) {
                    weight = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    weight = PEPReader.unescapeStrings(weight);
                } else if(key.equals(PEPReader.TRANS_GENERALIZED_RATE)) {
                    grate = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    grate = PEPReader.unescapeStrings(grate);
                } else if(key.equals(PEPReader.TRANS_GENERALIZED_WEIGHT)) {
                    gweight = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    gweight = PEPReader.unescapeStrings(gweight);
                } else if(key.equals(PEPReader.TRANS_SERVERS)) {
                    servers = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    servers = PEPReader.unescapeStrings(servers);
                } else if(key.equals(PEPReader.TRANS_AGE_MEMORY)) {
                    ageMemory = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    ageMemory = PEPReader.unescapeStrings(ageMemory);
                } else if(key.equals(PEPReader.TRANS_PRIORITY)) {
                    priority = PEPReader.REM_QUOT.matcher(value).replaceFirst(PEPReader.REP_QUOT);
                    priority = PEPReader.unescapeStrings(priority);
                } else {
                    // ignored
                }
                start = kv.end();
            } while(kv.find() && !(c.find(start) && (c.start() <= kv.start())));
        }
        
        if(guard != null) {
            // this could be interpreted as #
            guard = guard.replaceAll("!=", "! =");
        }
        final ITransition tr = this.net.createTransition(name, null, guard);
        tr.setPos(Integer.parseInt(x), Integer.parseInt(y));
        if(meaning != null) {
            tr.setMeaning(meaning);
        }
        if(earliest != null) {
            tr.setEarliestFiring(earliest);
        }
        if(latest != null) {
            tr.setLatestFiring(latest);
        }
        if(rate != null) {
            tr.setRate(rate);
        }
        if(weight != null) {
            tr.setWeight(weight);
        }
        if(grate != null) {
            tr.setGeneralizedRate(grate);
        }
        if(gweight != null) {
            tr.setGeneralizedWeight(gweight);
        }
        if(priority != null) {
            tr.setPriority(priority);
        }
        if(servers != null) {
            tr.setServers(servers);
        }
        if(ageMemory != null) {
            tr.setAgeMemory(ageMemory);
        }
        this.transitions.put(nr, tr);
    }
    
    
    private void parseInhibitorArc(final String line) {
        final Matcher m = PEPReader.PT_ARC.matcher(line);
        if(!m.lookingAt()) {
            System.err.println("Unable to parse inhibitor arc: " + line);
            return;
        }
        String fromID = m.group(PEPReader.PT_PLACE);
        String toID = m.group(PEPReader.PT_TRANS);
        if(fromID.startsWith("\"")) {
            fromID = fromID.substring(1, fromID.length()-1).trim();
        }
        if(toID.startsWith("\"")) {
            toID = toID.substring(1, toID.length()-1).trim();
        }

        int start = m.end();
        final Matcher kv = PEPReader.KEY_VALUE.matcher(line);
        
        String label = null;
        if(kv.find(start)) {
            do {
                start = kv.end();
                final String key = kv.group(PEPReader.KEY);
                String value = kv.group(PEPReader.VALUE);
                if(key.equals(PEPReader.ARC_INSCRIPTION)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceAll(PEPReader.REP_QUOT);
                    label = PEPReader.unescapeStrings(value);
                } else if(key.equals(PEPReader.ARC_WEIGHT)) {
                    final int count = Integer.parseInt(value);
                    final StringBuffer l = new StringBuffer();
                    for(int i = 0; i < count; i++) {
                        l.append(IPetriNetTool.DOT);
                        if(i < count-1) {
                            l.append(",");
                        }
                    }
                    label = l.toString();
                }
            } while(m.find());
        }
        
        final IPlace place = this.places.get(fromID);
        final ITransition trans = this.transitions.get(toID);
        
        if(trans == null) {
            throw new RuntimeException("Couldn't find transition \"" + toID + "\".");
        } else if(place == null) {
            throw new RuntimeException("Couldn't find place \"" + fromID + "\".");
        }
        
        this.net.createInhibitorArc(trans, place, label);
    }
    
    
    private void parseReadArc(final String line) {
        final Matcher m = PEPReader.PT_ARC.matcher(line);
        if(!m.lookingAt()) {
            System.err.println("Unable to parse inhibitor arc: " + line);
            return;
        }
        String fromID = m.group(PEPReader.PT_PLACE);
        String toID = m.group(PEPReader.PT_TRANS);
        if(fromID.startsWith("\"")) {
            fromID = fromID.substring(1, fromID.length()-1).trim();
        }
        if(toID.startsWith("\"")) {
            toID = toID.substring(1, toID.length()-1).trim();
        }

        int start = m.end();
        final Matcher kv = PEPReader.KEY_VALUE.matcher(line);
        
        String label = null;
        if(kv.find(start)) {
            do {
                start = kv.end();
                final String key = kv.group(PEPReader.KEY);
                String value = kv.group(PEPReader.VALUE);
                if(key.equals(PEPReader.ARC_INSCRIPTION)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceAll(PEPReader.REP_QUOT);
                    label = PEPReader.unescapeStrings(value);
                } else if(key.equals(PEPReader.ARC_WEIGHT)) {
                    final int count = Integer.parseInt(value);
                    final StringBuffer l = new StringBuffer();
                    for(int i = 0; i < count; i++) {
                        l.append(IPetriNetTool.DOT);
                        if(i < count-1) {
                            l.append(",");
                        }
                    }
                    label = l.toString();
                }
            } while(m.find());
        }
        
        final IPlace place = this.places.get(fromID);
        final ITransition trans = this.transitions.get(toID);
        
        if(trans == null) {
            throw new RuntimeException("Couldn't find transition \"" + toID + "\".");
        } else if(place == null) {
            throw new RuntimeException("Couldn't find place \"" + fromID + "\".");
        }
        
        this.net.createReadArc(trans, place, label);
    }
    
    
    private void parseResetArc(final String line) {
        final Matcher m = PEPReader.PT_ARC.matcher(line);
        if(!m.lookingAt()) {
            System.err.println("Unable to parse reset arc: " + line);
            return;
        }
        String fromID = m.group(PEPReader.PT_PLACE);
        String toID = m.group(PEPReader.PT_TRANS);
        if(fromID.startsWith("\"")) {
            fromID = fromID.substring(1, fromID.length()-1).trim();
        }
        if(toID.startsWith("\"")) {
            toID = toID.substring(1, toID.length()-1).trim();
        }

        int start = m.end();
        final Matcher kv = PEPReader.KEY_VALUE.matcher(line);
        
        String label = null;
        if(kv.find(start)) {
            do {
                start = kv.end();
                final String key = kv.group(PEPReader.KEY);
                String value = kv.group(PEPReader.VALUE);
                if(key.equals(PEPReader.ARC_INSCRIPTION)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceAll(PEPReader.REP_QUOT);
                    label = PEPReader.unescapeStrings(value);
                } else if(key.equals(PEPReader.ARC_WEIGHT)) {
                    final int count = Integer.parseInt(value);
                    final StringBuffer l = new StringBuffer();
                    for(int i = 0; i < count; i++) {
                        l.append(IPetriNetTool.DOT);
                        if(i < count-1) {
                            l.append(",");
                        }
                    }
                    label = l.toString();
                }
            } while(m.find());
        }
        
        final IPlace place = this.places.get(fromID);
        final ITransition trans = this.transitions.get(toID);
        
        if(trans == null) {
            throw new RuntimeException("Couldn't find transition \"" + toID + "\".");
        } else if(place == null) {
            throw new RuntimeException("Couldn't find place \"" + fromID + "\".");
        }
        
        this.net.createResetArc(trans, place, label);
    }
    
    
    private void parsePTArc(final String line) {
        final Matcher m = PEPReader.PT_ARC.matcher(line);
        if(!m.lookingAt()) {
            System.err.println("Unable to parse PT arc: " + line);
            return;
        }
        String fromID = m.group(PEPReader.PT_PLACE);
        String toID = m.group(PEPReader.PT_TRANS);
        if(fromID.startsWith("\"")) {
            fromID = fromID.substring(1, fromID.length()-1).trim();
        }
        if(toID.startsWith("\"")) {
            toID = toID.substring(1, toID.length()-1).trim();
        }

        int start = m.end();
        final Matcher kv = PEPReader.KEY_VALUE.matcher(line);
        
        String label = null;
        if(kv.find(start)) {
            do {
                start = kv.end();
                final String key = kv.group(PEPReader.KEY);
                String value = kv.group(PEPReader.VALUE);
                if(key.equals(PEPReader.ARC_INSCRIPTION)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceAll(PEPReader.REP_QUOT);
                    label = PEPReader.unescapeStrings(value);
                } else if(key.equals(PEPReader.ARC_WEIGHT)) {
                    final int count = Integer.parseInt(value);
                    final StringBuffer l = new StringBuffer();
                    for(int i = 0; i < count; i++) {
                        l.append(IPetriNetTool.DOT);
                        if(i < count-1) {
                            l.append(",");
                        }
                    }
                    label = l.toString();
                }
            } while(m.find());
        }
        
        final IPlace place = this.places.get(fromID);
        final ITransition trans = this.transitions.get(toID);
        
        if(trans == null) {
            throw new RuntimeException("Couldn't find transition \"" + toID + "\".");
        } else if(place == null) {
            throw new RuntimeException("Couldn't find place \"" + fromID + "\".");
        }
        
        this.net.createArc(place, trans, label);
    }
    
    
    private void parseTPArc(final String line) {
        final Matcher m = PEPReader.TP_ARC.matcher(line);
        if(!m.lookingAt()) {
            System.err.println("Unable to parse TP arc: " + line);
            return;
        }
        String fromID = m.group(PEPReader.TP_TRANS);
        String toID = m.group(PEPReader.TP_PLACE);
        if(fromID.startsWith("\"")) {
            fromID = fromID.substring(1, fromID.length()-1).trim();
        }
        if(toID.startsWith("\"")) {
            toID = toID.substring(1, toID.length()-1).trim();
        }

        int start = m.end();
        final Matcher kv = PEPReader.KEY_VALUE.matcher(line);
        
        String label = null;
        
        if(kv.find(start)) {
            do {
                final String key = kv.group(PEPReader.KEY);
                String value = kv.group(PEPReader.VALUE);
                if(key.equals(PEPReader.ARC_INSCRIPTION)) {
                    value = PEPReader.REM_QUOT.matcher(value).replaceAll(PEPReader.REP_QUOT);
                    label = PEPReader.unescapeStrings(value);
                } else if(key.equals(PEPReader.ARC_WEIGHT)) {
                    final int count = Integer.parseInt(value);
                    final StringBuffer l = new StringBuffer();
                    for(int i = 0; i < count; i++) {
                        l.append(IPetriNetTool.DOT);
                        if(i < count-1) {
                            l.append(",");
                        }
                    }
                    label = l.toString();
                }
            } while(kv.find());
        }
        
        final ITransition trans = this.transitions.get(fromID);
        final IPlace place = this.places.get(toID);
        
        if(trans == null) {
            throw new RuntimeException("Couldn't find transition \"" + fromID + "\".");
        } else if(place == null) {
            throw new RuntimeException("Couldn't find place \"" + toID + "\".");
        }
        
        this.net.createArc(trans, place, label);
    }
    
    
    public static IPlaceType parseType(String text) {
//        System.out.println("parse type " + text);
        if(PEPReader.TYPE_SET.matcher(text).matches()) {
            return new Set();
        } else if(PEPReader.TYPE_ANY.matcher(text).matches()) {
            return IPlace.TYPE_ANY;
        } else if(PEPReader.TYPE_DOT.matcher(text).matches()) {
            return IPlace.TYPE_BLACK_TOKEN;
        } else if(PEPReader.TYPE_BOOLEAN.matcher(text).matches()) {
            return IPlace.TYPE_BOOLEAN;
        } else if(PEPReader.TYPE_INT.matcher(text).matches()) {
            text = text.trim();
            final int idx = text.indexOf("..");
            final int from = Integer.parseInt(text.substring(1, idx).trim());
            final int to = Integer.parseInt(text.substring(idx+2, text.length()-1).trim());
            return new IntegerRange(from, to);
        } else if(!PEPReader.TYPE_TUPLE.matcher(text).find()) {
            if(text.indexOf(",") < 0) {
                //System.out.println("parse single value: " + text);
                text = PEPReader.REM_BRACK.matcher(text).replaceFirst(PEPReader.REP_BRACK);
                return new SingleValue(PEPReader.unescapeStrings(text).trim());
            } else {
                return PEPReader.parseSet(text);
            }
        }
        
        boolean isSet = true;
        int parent = 0;
        boolean string = false;
        final char[] chars = text.toCharArray();
        for(int i = 0; i < chars.length; i++) {
            if((chars[i] == '\"') && (i > 0) && (chars[i-1] == '\\')) {
                string = !string;
            } else if(chars[i] == '{') {
                ++parent;
            } else if(chars[i] == '}') {
                --parent;
            } else if(Character.isWhitespace(chars[i])) {
                continue;
            }
            if((parent == 0) && !string && (i < chars.length - 1)) {
                isSet = false;
                break;
            }
        }
        
        final IPlaceType type;
        if(isSet) {
            type = PEPReader.parseSet(text);
        } else {
            type = PEPReader.parseTuple(text);
        }
        
//        System.out.println("--> " + type);
        return type;
    }
    
    
    public static String unescapeStrings(final String str) {
        return str.replaceAll("\\\\\"", "\"").replaceAll("\\\\\\\\", "\\\\");
    }
    
    
    public static Tuple parseTuple(final String text) {
        final Tuple tuple = new Tuple();
        final StringBuffer token = new StringBuffer();
        final char[] chars = text.toCharArray();
        int i = 0;
        int parentheses = 0;
        boolean string = false;
        while(i < chars.length) {
            switch(chars[i]) {
                case '\"':
                    if((i > 0) && (chars[i-1] == '\\')) {
                        string = !string;
                    }
                    break;
                case '{':
                    if((parentheses == 0) && !string) {
                        token.delete(0, token.length());
                        token.append("{");
                    }
                    ++parentheses;
                    break;
                case '}':
                    --parentheses;
                    if((parentheses == 0) && !string) {
                        token.append("}");
                        IPlaceType type = PEPReader.parseType(token.toString());
                        tuple.addType(type);
                    }
                    break;
                default:
                    token.append(chars[i]);
                    break;
            }
            i++;
        }
//        System.out.println("parse tuple " + text + " -> " + tuple);
        return tuple;
    }
    
    
    public static Set parseSet(String text) {
//        System.out.println("parse set " + text);
        text = PEPReader.REM_BRACK.matcher(text).replaceFirst(PEPReader.REP_BRACK);
        final Set set = new Set();
        final StringBuffer token = new StringBuffer();
        final char[] chars = text.toCharArray();
        int i = 0;
        int parentheses = 0;
        boolean string = false;
        while(i < chars.length) {
            switch(chars[i]) {
                case '\"':
                    if((i > 0) && (chars[i-1] == '\\')) {
                        string = !string;
                    }
                    token.append(chars[i]);
                    break;
                case '{':
                    token.append(chars[i]);
                    ++parentheses;
                    break;
                case '}':
                    token.append(chars[i]);
                    --parentheses;
                    break;
                case ',':
                    if((parentheses == 0) && !string) {
                        String theToken = token.toString();
                        if(!theToken.startsWith("{")) {
                            theToken = "{" + theToken + "}";
                        }
                        final IPlaceType type = PEPReader.parseType(theToken);
                        set.addType(type);
                        token.delete(0, token.length());
                    }
                    break;
                default:
                    token.append(chars[i]);
                    break;
            }
            i++;
        }
        String theToken = token.toString();
        if(!theToken.startsWith("{")) {
            theToken = "{" + theToken + "}";
        }
        final IPlaceType type = PEPReader.parseType(theToken);
        set.addType(type);
        return set;
    }
    
    
    public static void writeToFile(final String net, final String filename) {
        try {
            final File file = new File(filename);
            final FileWriter outputstream = new FileWriter(file);
            final PrintWriter output = new PrintWriter(outputstream);
            output.print(net);
            output.close();
        } catch (final IOException ioex) {
            throw new RuntimeException(ioex);
        }
    }
    
    
    public static void main(final String[] args) {
        final PEPReader reader = new PEPReader();
        final IPetriNet net = reader.readFromFile(new petruchio.sim.pnmodel.PNModel(), args[0]);
        PEPReader.writeToFile(net.toPEP(), args[1]);
    }
}
