/*
 * Project:  P-UMLaut
 * Filename: PNMLReader.java
 * Created:  01.02.2006
 * Author:   Tim Strazny
 */
package petruchio.sim.petrinettool;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import org.dom4j.Node;
import org.dom4j.Document;
import org.dom4j.io.SAXReader;


public class PNMLReader implements PetriNetReader {
    private static SAXReader reader = new SAXReader(); 
    private Document document = null;
    private IPetriNetTool tool = null;
    private final Collection<String> functions = new ArrayList<String>();
    private IPetriNet net = null;

    public synchronized IPetriNet readFromFile(final IPetriNetTool theTool,
                    String filename) {
        File f = new File(filename);
        if(!f.isAbsolute()) {
            f = new File(System.getProperty("user.dir") + File.separator + filename);
        }
        try {
            filename = f.getCanonicalPath();
        } catch(IOException e) {
            filename = f.getAbsolutePath();
        }
        
        try {
            this.document = PNMLReader.reader.read(filename);
        } catch (final Exception e) {
            throw new RuntimeException("Unable to read pnml from file " + filename
                    + "\n" + e.getMessage(), e);
        }
        
        return this.buildNet(theTool);
    }


    public synchronized IPetriNet readFromString(final IPetriNetTool theTool,
                    final String content) {
        try {
            final ByteArrayInputStream in = new ByteArrayInputStream(content.getBytes());
            this.document = PNMLReader.reader.read(in);
            in.close();
        } catch (final Exception e) {
            throw new RuntimeException("Unable to read pnml from string.\n"
                            + e.getMessage(), e);
        }

        return this.buildNet(theTool);
    }
    
    
    private IPetriNet buildNet(final IPetriNetTool theTool) {
        this.tool = theTool;
        this.net = this.tool.createPetriNet();
        final String id = this.document.valueOf("/pnml/net/@id");
        this.scanFunctions(id);
        this.scanPlaces(id);
        this.scanTransitions(id);
        this.scanArcs(id);
        
        final IPetriNet theNet = this.net;
        this.net = null;
        this.tool = null;
        return theNet;
    }
    
    
    @SuppressWarnings("unchecked")
    private void scanFunctions(final String id) {
        this.functions.clear();
        final List<Node> fns = this.document.selectNodes("/pnml/net[@id='" + id + "']/page/declaration/structure/declare[@type='function']");
        for(final Node func : fns) {
            final StringBuffer fname = new StringBuffer();
            fname.append(func.valueOf("ci"));
            fname.append("(");
            final List<Node> params = func.selectNodes("fn/bvar/ci");
            for(final Node n : params) {
                fname.append(n.getStringValue());
            }
            fname.append(")");
            final List nodes = func.selectNodes("fn/*");
            Node fnkt = null;
            for(final Iterator it = nodes.iterator(); it.hasNext();) {
                final Node n = (Node)it.next();
                if(!n.getName().equals("bvar")) {
                    fnkt = n;
                    break;
                }
            }
            final String body = this.scanExpression(fnkt);
            final String fnamestring = fname.toString();
            this.functions.add(fnamestring);
            this.net.addFunction(fnamestring + "=" + body);
        }
    }
    
    
    @SuppressWarnings("unchecked")
    private void scanPlaces(final String id) {
        final List<Node> places = this.document.selectNodes("/pnml/net[@id='" + id + "']/page/place");
        for(final Node place : places) {
            final String name = place.valueOf("name/text");
            IPlaceStatus status = null;
            final String statusString = place.valueOf("status/text");
            if(statusString.equals("entry")) {
                status = IPlace.STATUS_ENTRY;
            } else if(statusString.equals("exit")) {
                status = IPlace.STATUS_EXIT;
            } else {
                status = IPlace.STATUS_INTERNAL;
            }
            final IPlaceType type = this.scanPlaceType(place.selectSingleNode("type/structure/*"));
            final String init = this.scanSet(place.selectSingleNode("hlinitialMarking/structure/set"));
            final String current = this.scanSet(place.selectSingleNode("hlcurrentMarking/structure/set"));
            final IPlace p = this.net.createPlace(name, status, type, init, current);
            p.setMeaning(place.valueOf("meaning/text"));
            
            final String sc = place.valueOf("capacity/text");
            if(sc.length() != 0) {
                p.setCapacity(Integer.parseInt(sc));
            }
            
            final String sx = place.valueOf("graphics/position/@x");
            final String sy = place.valueOf("graphics/position/@y");
            final int x = sx.length() != 0 ? Integer.parseInt(sx) : 0; 
            final int y = sy.length() != 0 ? Integer.parseInt(sy) : 0;
            p.setPos(x, y);
        }
    }
    
    
    @SuppressWarnings("unchecked")
    private void scanTransitions(final String id) {
        final List<Node> transitions = this.document.selectNodes("/pnml/net[@id='" + id + "']/page/transition");
        for(final Node trans : transitions) {
            final String name = trans.valueOf("name/text");
            final String guard = this.scanExpression(trans.selectSingleNode("condition/structure/*"));
            final IMultiSet actions = this.scanActions(trans.selectSingleNode("actions/text"));
            
            final ITransition t = this.net.createTransition(name, actions, guard);
            
            final String priority = this.scanExpression(trans.selectSingleNode("priority/structure/*"));
            final String sx = trans.valueOf("graphics/position/@x");
            final String sy = trans.valueOf("graphics/position/@y");
            final int x = sx.length() != 0 ? Integer.parseInt(sx) : 0; 
            final int y = sy.length() != 0 ? Integer.parseInt(sy) : 0;
            t.setPos(x, y);
            t.setMeaning(trans.valueOf("meaning/text"));
            t.setPriority(priority);
            
            final List<Node> interval = trans.selectNodes("timeConstraint/structure/interval/*");
            if(!interval.isEmpty()) {
                t.setEarliestFiring(this.scanExpression(interval.get(0)));
                t.setLatestFiring(this.scanExpression(interval.get(1)));
            }
            final String rate = this.scanExpression(trans.selectSingleNode("rate/structure/*"));
            final String weight = this.scanExpression(trans.selectSingleNode("weight/structure/*"));
            final String grate = this.scanExpression(trans.selectSingleNode("generalizedRate/structure/*"));
            final String gweight = this.scanExpression(trans.selectSingleNode("generalizedWeight/structure/*"));
            final String servers = this.scanExpression(trans.selectSingleNode("servers/structure/*"));
            final String ageMemory = this.scanExpression(trans.selectSingleNode("ageMemory/structure/*"));
            t.setRate(rate);
            t.setWeight(weight);
            t.setGeneralizedRate(grate);
            t.setGeneralizedWeight(gweight);
            t.setServers(servers);
            t.setAgeMemory(ageMemory);
        }
    }
    
    
    @SuppressWarnings("unchecked")
    private void scanArcs(final String id) {
        final List<Node> arcs = this.document.selectNodes("/pnml/net[@id='" + id + "']/page/arc");
        for(final Node arc : arcs) {
            final String label = this.scanSet(arc.selectSingleNode("hlinscription/structure/set"));
            final String src = arc.valueOf("@source");
            final String tar = arc.valueOf("@target");
            final String type = arc.valueOf("@type");
            final IPlace sourcePlace = this.net.getPlaceByName(src);
            final IPlace place;
            final boolean transIsSource = sourcePlace == null;
            if(transIsSource) {
                place = this.net.getPlaceByName(tar);
                final ITransition trans = this.net.getTransitionByName(src);
                this.net.createArc(trans, place, label);
            } else {
                place = sourcePlace;
                final ITransition trans = this.net.getTransitionByName(tar);
                if((type == null) || type.equals("pt")) {
                    this.net.createArc(place, trans, label);
                } else if(type.equals("inhibitor")) {
                    this.net.createInhibitorArc(trans, place, label);
                } else if(type.equals("read")) {
                    this.net.createReadArc(trans, place, label);
                } else if(type.equals("reset")) {
                    this.net.createResetArc(trans, place, label);
                } else {
                    System.err.println("Ignoring arc with unknown type: " + type); 
                }
            }
        }
    }
    
    
    private IMultiSet scanActions(final Node node) {
        final IMultiSet set = this.tool.createMultiSet();
        final String txt = node.getStringValue();
        int i = 0;
        do {
            final int send = txt.indexOf('!', i);
            final int rec = txt.indexOf('!', i);
            if((send < 0) && (rec < 0)) {
                break;
            }
            final boolean isSending = (rec < 0) || (send < rec);
            final int tie = txt.indexOf("tie", i);
            final int action = txt.indexOf("action", i);
            final boolean isTie = (action < 0) || (tie < action);
            final String curAction = txt.substring(txt.indexOf('(', i) + 1, isSending ? send : rec);
            int parentheseses = 1;
            int x = (isSending ? send : rec) + 2;
            final StringBuffer label = new StringBuffer();
            while((parentheseses != 0)) {
                switch(txt.charAt(x)) {
                    case '(' : ++parentheseses; break;
                    case ')' : --parentheseses; break;
                }
                if(parentheseses != 0) {
                    label.append(txt.charAt(x));
                }
                ++x;
            }
            
            i = x+2;
            
            if(isTie) {
                if(isSending) {
                    set.addSendTie(curAction, label.toString());
                } else {
                    set.addReceiveTie(curAction, label.toString());
                }
            } else {
                if(isSending) {
                    set.addActionSymbol(curAction, label.toString());
                } else {
                    set.addConjugatedActionSymbol(curAction, label.toString());
                }
            }
        } while(i < txt.length());
        return set;
    }
    
    
    @SuppressWarnings("unchecked")
    private IPlaceType scanPlaceType(final Node node) {
        if(node == null) {
            return IPlace.TYPE_ANY;
        } else if(node.getName().equals("symbol") && (node.getStringValue().equals("dot"))) {
            return IPlace.TYPE_BLACK_TOKEN;
        } else if(node.getName().equals("boolean")) {
            return IPlace.TYPE_BOOLEAN;
        } else if(node.getName().equals("any")) {
            return IPlace.TYPE_ANY;
        } else if(node.getName().equals("symbol")) {
            final String val;
            if(node.valueOf("@type").equals("string")) {
                val = "\"" + node.getStringValue() + "\"";
            } else {
                val = node.getStringValue();
            }
            return new petruchio.sim.petrinettool.type.SingleValue(val);
        } else if(node.getName().equals("value")) {
            return new petruchio.sim.petrinettool.type.SingleValue(node.getStringValue());
        } else if(node.getName().equals("set")) {
            if(node.selectSingleNode("bvar") != null) {
                try {
                    final List l = ((Node)node.selectNodes("condition/apply/apply").get(1)).selectNodes("cn");
                    final int from = Integer.parseInt(((Node)l.get(0)).getStringValue());
                    final int to = Integer.parseInt(((Node)l.get(1)).getStringValue());
                    return new petruchio.sim.petrinettool.type.IntegerRange(from, to);
                } catch(Exception e) {
                    throw new RuntimeException("I thought this would be an "
                                    + "integer range, but I cannot translate "
                                    + "this into a proper IPlaceType:\n"
                                    + node.asXML());
                }
            } else {
                petruchio.sim.petrinettool.type.Set set = new petruchio.sim.petrinettool.type.Set();
                final List children = node.selectNodes("*");
                for(final Iterator it = children.iterator(); it.hasNext();) {
                    final Node child = (Node)it.next();
                    set.addType(this.scanPlaceType(child));
                }
                return set;
            }
        } else if(node.getName().equals("apply")
                        && (node.selectSingleNode("union") != null)) {
            petruchio.sim.petrinettool.type.Set set = new petruchio.sim.petrinettool.type.Set();
            final List<Node> children = node.selectNodes("*");
            boolean first = true;
            for(final Node child : children) {
                if(first) {
                    first = false;
                } else {
                    set.addType(this.scanPlaceType(child));
                }
            }
            return set;
        } else if(node.getName().equals("apply")
                        && (node.selectSingleNode("cartesianproduct") != null)) {
            petruchio.sim.petrinettool.type.Tuple tuple = new petruchio.sim.petrinettool.type.Tuple();
            final List<Node> children = node.selectNodes("*");
            boolean first = true;
            for(final Node child : children) {
                if(first) {
                    first = false;
                } else {
                    tuple.addType(this.scanPlaceType(child));
                }
            }
            return tuple;
        } else { 
            throw new RuntimeException("I cannot translate this into a proper "
                    + "IPlaceType:\n" + node.asXML());
        }
    }
    
    
    @SuppressWarnings("unchecked")
    private String scanSet(final Node node) {
        if(node == null) {
            return "";
        }
        final StringBuffer buf = new StringBuffer();
        final List<Node> children = node.selectNodes("*");
        for(final Iterator<Node> it = children.iterator(); it.hasNext();) {
            final Node child = it.next();
            buf.append(this.scanExpression(child));
            if(it.hasNext()) {
                buf.append(", ");
            }
        }
        return buf.toString();
    }
    
    
    @SuppressWarnings("unchecked")
    private String scanExpression(final Node node) {
        if(node == null) {
            return "";
        } else if(node.getName().equals("cn") || node.getName().equals("ci")) {
            if(node.valueOf("@type").equals("string")) {
                return "\"" + node.getStringValue() + "\"";
            } else {
                return node.getStringValue();
            }
        } else if(node.getName().equals("symbol")) {
            return node.getStringValue();
        } else if(node.getName().equals("true")) {
            return "tt";
        } else if(node.getName().equals("false")) {
            return "ff";
        } 
        
        final List<Node> children;
        final String pnmlOp;
        boolean ignoreFirst = true;
        if(node.getName().equals("vector")) {
            children = node.selectNodes("*");
            pnmlOp = "vector";
            ignoreFirst = false;
        } else {
            children = node.selectNodes("*");
            pnmlOp = children.get(0).getName();
        }
        //System.out.println(node.asXML());
        
        boolean isFunction = false;
        boolean isUnary = false;
        final String op;
        if(pnmlOp.equals("plus")) {
            op = "+";
        } else if(pnmlOp.equals("minus")) {
            op = "-";
        } else if(pnmlOp.equals("times")) {
            op = "*";
        } else if(pnmlOp.equals("divide")) {
            op = "/";
        } else if(pnmlOp.equals("rem")) {
            op = "%";
        } else if(pnmlOp.equals("power")) {
            op = "**";
        } else if(pnmlOp.equals("eq")) {
            op = "=";
        } else if(pnmlOp.equals("neq")) {
            op = "#";
        } else if(pnmlOp.equals("lt")) {
            op = "<";
        } else if(pnmlOp.equals("gt")) {
            op = ">";
        } else if(pnmlOp.equals("leq")) {
            op = "<=";
        } else if(pnmlOp.equals("geq")) {
            op = ">=";
        } else if(pnmlOp.equals("and")) {
            op = "&";
        } else if(pnmlOp.equals("or")) {
            op = "|";
        } else if(pnmlOp.equals("not")) {
            op = "!";
            isUnary = true;
        } else {
            isFunction = true;
            if(pnmlOp.equals("ln")) {
                op = "logn";
            } else if(pnmlOp.equals("vector")) {
                op = ""; // empty for simple tuple
            } else if(pnmlOp.equals("arcsin")) {
                op = "asin";
            } else if(pnmlOp.equals("arccos")) {
                op = "acos";
            } else if(pnmlOp.equals("arctan")) {
                op = "atan";
            } else if(pnmlOp.equals("ceiling")) {
                op = "ceil";
            } else if(pnmlOp.equals("abs") || pnmlOp.equals("floor") || pnmlOp.equals("exp") || pnmlOp.equals("min")
                    || pnmlOp.equals("log") || pnmlOp.equals("sin") || pnmlOp.equals("cos") || pnmlOp.equals("tan")
                    || pnmlOp.equals("cot") || pnmlOp.equals("sinh") || pnmlOp.equals("cosh") || pnmlOp.equals("tanh")
                    || pnmlOp.equals("coth") || pnmlOp.equals("max") || pnmlOp.equals("root")) {
                // don't change op
                op = pnmlOp;
            } else if(pnmlOp.equals("fn") && this.functions.contains(node.valueOf("ci"))) {
                op = node.valueOf("ci");
            } else {
                throw new RuntimeException("I cannot translate MathML 2.0 expression \"" + node.asXML() + "\".");
            }
        }
        
        
        final StringBuffer str = new StringBuffer();
        if(isFunction) {
            str.append(op);
            str.append("(");
        } else { 
            str.append("(");
            if(isUnary) {
                str.append(op);
            }
        }
        for(final Iterator<Node> it = children.iterator(); it.hasNext(); ) {
            if(ignoreFirst) { // skip operator
                ignoreFirst = false;
                it.next();
                continue;
            }
            final Node child = it.next();
            str.append(this.scanExpression(child));
            if(it.hasNext()) {
                if(isFunction) {
                    str.append(", ");
                } else {
                    str.append(op);
                }
            }
        }
        str.append(")");
        
        return str.toString();
    }
    
}