/*
 * Project:  pg
 * Filename: NaturalRange.java
 * Created:  Jun 9, 2005
 * Author:   Tim Strazny
 */
package petruchio.sim.petrinettool.type;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;

import petruchio.sim.petrinettool.IPlaceType;


/**
 * Represents a type of natural numbers in a specified range between a lower
 * bound and an upper bound like {0, .., 113} which is a sub set of the natural
 * numbers.
 *
 * @author Tim Strazny
 */
public class IntegerRange implements IPlaceType {

    /** Lower bound */
    private final int from;
    
    /** Upper bound */
    private final int to;

    private final Collection<String> values;
    
    
    /**
     * Initialises this NaturalRange with the given boundaries.
     * 
     * @param from The lower bound.
     * @param to The upper bound.
     */
    public IntegerRange(final int from, final int to) {
        if(from > to) {
            throw new IllegalArgumentException("Lower bound may not be greater than upper bound.");
        }
        this.from = from;
        this.to = to;
        final Collection<String> temp = new ArrayList<String>();
        for(int i = from; i <= to; i++) {
            temp.add(String.valueOf(i));
        }
        this.values = Collections.unmodifiableCollection(temp);
    }
    
    
    public int getLowerBound() {
        return this.from;
    }
    
    
    public int getUpperBound() {
        return this.to;
    }
    
    
    /**
     * Returns the name of the natural range type. It returns &quot;Natural
     * Number { <i>from</i>, ..., <i>to</i> }&quot; where <i>from,to</i> are the
     * type's boundaries.
     * 
     * @return A String representing the type.
     */
    public String getName() {
        return "Integer(" + this.from + ".." + this.to + ")";
    }


    /**
     * Returns this type's PNML representation which is <code>&lt;int
     * from=&quot;<i>from</i>&quot; to=&quot;<i>to</i>&quot; /&gt;</code>.
     * 
     * @return The PNML representation of this type.
     */
    public String getPNML() {
        return    "<set>\n"
                +   "<bvar><ci>x</ci></bvar>\n"
                +   "<condition>\n"
                +     "<apply>\n"
                +       "<and />\n"
                +       "<apply>\n"
                +         "<in />\n"
                +         "<ci>x</ci>\n"
                +         "<integers />\n"
                +       "</apply>\n"
                +       "<apply>\n"
                +         "<leq />\n"
                +         "<cn>" + this.from + "</cn>\n"
                +         "<ci>x</ci>\n"
                +         "<cn>" + this.to + "</cn>\n"
                +       "</apply>\n"
                +     "</apply>\n"
                +   "</condition>\n"
                + "</set>\n";
    }


    /**
     * Returns this type's PEP representation which is
     * <code><i>from</i>..<i>to</i></code>.
     * 
     * @return The PEP representation of this type.
     */
    public String getPEP() {
        return this.from + ".." + this.to;
    }

    
    /**
     * See {@link #getName()}.
     * @return A String representing this object.
     * @see #getName()
     */
    public String toString() {
        return this.getName();
    }
    
    
    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(final Object o) {
        return (o instanceof IntegerRange) && this.equals((IntegerRange)o);
    }
    
    
    /**
     * Returns <code>true</code> if and only if this and the given NaturalRange
     * objects represent the same range of natural numbers.
     *
     * @param nat A second NaturalRange to compare to.
     * @return <code>true</code> if the boundaries of the given NaturalRange
     *          are equal to the boundaries of this NaturalRange.
     */
    public boolean equals(final IntegerRange nat) {
        return (nat == this) || ((nat != null) && (nat.from == this.from) && (nat.to == this.to));
    }
    
    
    public boolean isValue(final String value) {
        try {
            final int v = Integer.parseInt(value);
            return (this.from <= v) && (v <= this.to); 
        } catch(NumberFormatException e) {
            return false;
        }
    }
    
    
    public String getRandomValue() {
        return String.valueOf(this.from + (int)(Math.random()*(this.to-this.from+1)));
    }


    public Collection<String> getValues() {
        return this.values;
    }
}
