/*
 * Project:  PG-UML
 * Filename: PEPWriter.java
 * Created:  03.06.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Iterator;
import java.util.Collection;
import java.util.Map;

import petruchio.sim.petrinettool.IPetriNet;
import petruchio.sim.petrinettool.PEPReader;
import petruchio.sim.petrinettool.PetriNetWriter;

import petruchio.sim.pnmodel.net.PetriNet;
import petruchio.sim.pnmodel.net.Transition;
import petruchio.sim.pnmodel.net.Place;
import petruchio.sim.pnmodel.net.Arc;
import petruchio.sim.pnmodel.net.Value;
import petruchio.sim.pnmodel.net.ValueList;
import petruchio.sim.pnmodel.util.Pool;

/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class PEPWriter implements PetriNetWriter {
    
    public static final int MIN_X = 30;
    public static final int MIN_Y = 30;
    
    
    public String writeString(final IPetriNet inet) {
        final PetriNet net;
        try {
            net = (PetriNet)inet;
        } catch(ClassCastException e) {
            throw new RuntimeException("This PEPWriter can only handle petruchio.sim.pnmodel.PetriNets.");
        }
        final StringBuffer pep = new StringBuffer();
        if(net.isLowLevelNet()) {
            pep.append("PEP\n");
            pep.append("PTNet\n");
            pep.append("FORMAT_N\n");
            pep.append("DPL s7n10@-9t2\n");
            pep.append("DTR s7n10@-9t2\n");
            pep.append("DPT w1t1\n");
        } else {
            pep.append("PEP\n");
            pep.append("HLNet\n");
            pep.append("FORMAT_N\n");
            pep.append("DPL s7n10@-9t2\n");
            pep.append("DTR s7n10@-9t2\n");
            pep.append("DPT w1t1\n");
        }
            
        final Map<String, Integer> place_id = Pool.getPool().getMap();
        final Map<String, Integer> trans_id = Pool.getPool().getMap();
        if(!net.getFunctions().isEmpty()) {
            pep.append("FN\n");
            for(final Iterator<String> it = net.getFunctions().iterator(); it.hasNext();) {
                pep.append(it.next() + "\n");
            }
        }
        
        pep.append("PL\n");
        final Iterator<Place> placeit = net.getPlaceIterator();
        for(int i = 1; placeit.hasNext(); i++) {
            final Place place = placeit.next();
            final String name = place.getName();
            place_id.put(name, i);
            pep.append(PEPWriter.placeToPEP(net, place, i));
        }
        
        pep.append("TR\n");
        final Iterator<Transition> transit = net.getTransitionIterator();
        for(int i = 1; transit.hasNext(); i++) {
            final Transition trans = transit.next();
            trans_id.put(trans.getName(), i);
            pep.append(PEPWriter.transToPEP(net, trans, i));
        }
        
        pep.append(PEPWriter.arcsToPEP(net, net.getArcs(), place_id, trans_id));
        return pep.toString();
    }
    
    
    public void writeFile(final IPetriNet inet, final String filename) {
        final PetriNet net;
        try {
            net = (PetriNet)inet;
        } catch(final ClassCastException e) {
            throw new RuntimeException("This PEPWriter can only handle petruchio.sim.pnmodel.PetriNets.");
        }
        try {
            final File file = new File(filename);
            final FileWriter outputstream = new FileWriter(file);
            final PrintWriter output = new PrintWriter(outputstream);
            
            if(net.isLowLevelNet()) {
                output.println(PEPReader.Section.HEADER.getKey());
                output.println("PTNet");
                output.println("FORMAT_N");
                output.println("DPL s7n10@-9t2");
                output.println("DTR s7n10@-9t2");
                output.println("DPT w1t1");
            } else {
                output.println(PEPReader.Section.HEADER.getKey());
                output.println("HLNet");
                output.println("FORMAT_N");
                output.println("DPL s7n10@-9t2");
                output.println("DTR s7n10@-9t2");
                output.println("DPT w1t1");
            }
                
            final Map<String, Integer> place_id = Pool.getPool().getMap();
            final Map<String, Integer> trans_id = Pool.getPool().getMap();
            
            output.println(PEPReader.Section.PLACES.getKey());
            final Iterator<Place> placeit = net.getPlaceIterator();
            for(int i = 1; placeit.hasNext(); i++) {
                final Place place = placeit.next();
                final String name = place.getName();
                place_id.put(name, new Integer(i));
                output.print(PEPWriter.placeToPEP(net, place, i));
            }
            
            output.println(PEPReader.Section.TRANSITIONS.getKey());
            final Iterator<Transition> transit = net.getTransitionIterator();
            for(int i = 1; transit.hasNext(); i++) {
                final Transition trans = transit.next();
                trans_id.put(trans.getName(), new Integer(i));
                output.print(PEPWriter.transToPEP(net, trans, i));
            }
            
            output.print(PEPWriter.arcsToPEP(net, net.getArcs(), place_id, trans_id));
            
            output.close();
        } catch (final IOException ioex) {
            throw new RuntimeException(ioex);
        }
    }

    
    public static String escapeStrings(final String str) {
        return str.replaceAll("\\\\", "\\\\").replaceAll("\"", "\\\"");
    }
    
    
    public static StringBuffer placeToPEP(final PetriNet net, final Place place,
                    final int idx) {
        final StringBuffer pep = new StringBuffer();
        final String status = place.getStatus().getPEP();
        final String meaning;
        if(place.getMeaning() != null && place.getMeaning().length() > 0) {
            meaning = PEPReader.PLACE_MEANING + "\"" + place.getMeaning() + "\"";
        } else {
            meaning = "";
        }
        
        final String capacity;
        if(!place.infiniteCapacity()) {
            capacity = PEPReader.PLACE_CAPACITY + place.getCapacity();
        } else {
            capacity = "";
        }
        
        final int x = (place.getPosX() >= PEPWriter.MIN_X) ? place.getPosX()
                        : PEPWriter.MIN_X; 
        final int y = (place.getPosY() >= PEPWriter.MIN_Y) ? place.getPosY()
                        : PEPWriter.MIN_Y;
        
        String name = place.getName();
        int i = 0;
        while(net.getTransitionByName(name) != null) {
            name = place.getName() + "_" + (i++);
        }
        
        final String curmarking;
        final String marking;
        final String type;
        if(net.isLowLevelNet()) {
            int imarking = place.getInitialMarkingSet().size();
            int icurmarking = place.getCurrentMarkingSet().size();
            if(icurmarking > 0) {
                curmarking = PEPReader.PLACE_INITIAL_COUNT + icurmarking;
            } else {
                curmarking = "";
            }
            if(imarking > 0) {
                marking = PEPReader.PLACE_CURRENT_COUNT + imarking;
            } else {
                marking = "";
            }
            
            type = "";
            
        } else {
            type = PEPReader.PLACE_TYPE + "\"" + place.getType().getPEP() + "\"";
            
            String smarking = PEPWriter.escapeStrings(place.getInitialMarking());
            String scurmarking = PEPWriter.escapeStrings(place.getCurrentMarking());
            if(scurmarking.length() > 0) {
                curmarking = PEPReader.PLACE_CURRENT_MARKING + "\"{" + scurmarking + "}\"";
            } else {
                curmarking = "";
            }
            if(smarking.length() > 0) {
                marking = PEPReader.PLACE_INITIAL_MARKING + "\"{" + smarking + "}\"";
            } else {
                marking = "";
            }
        }
        pep.append(idx);
        pep.append("\"");
        pep.append(name);
        pep.append("\"");
        pep.append(x);
        pep.append("@");
        pep.append(y);
        pep.append(status);
        pep.append(capacity);
        pep.append(meaning);
        pep.append(type);
        pep.append(curmarking);
        pep.append(marking);
        pep.append("\n");
        return pep;
    }
    
    
    public static String transToPEP(final PetriNet net, final Transition trans,
                    final int idx) {
        final StringBuffer pep = new StringBuffer();
        final int x = (trans.getPosX() >= PEPWriter.MIN_X) ? trans.getPosX()
                                                     : PEPWriter.MIN_X; 
        final int y = (trans.getPosY() >= PEPWriter.MIN_Y) ? trans.getPosY()
                                                     : PEPWriter.MIN_Y;
        final String name = trans.getName();

        pep.append(idx);
        pep.append("\"");
        pep.append(name);
        pep.append("\"");
        pep.append(x);
        pep.append("@");
        pep.append(y);
        if(trans.getMeaning() != null && trans.getMeaning().length() > 0) {
            pep.append(PEPReader.TRANS_MEANING);
            pep.append("\"");
            pep.append(trans.getMeaning());
            pep.append("\"");
        }
        if(!net.isLowLevelNet()) {
            pep.append(PEPReader.TRANS_GUARD);
            pep.append("\"");
            pep.append(PEPWriter.escapeStrings(trans.getGuard()));
            pep.append("\"");
        }
        if(trans.hasServers()) {
            pep.append(PEPReader.TRANS_SERVERS);
            pep.append("\"");
            pep.append(trans.getServers());
            pep.append("\"");
        }
        if(trans.hasPriority()) {
            pep.append(PEPReader.TRANS_PRIORITY);
            pep.append("\"");
            pep.append(trans.getPriority());
            pep.append("\"");
        }
        if(trans.hasEarliestFiring()) {
            pep.append(PEPReader.TRANS_EARLIEST);
            pep.append("\"");
            pep.append(trans.getEarliestFiring());
            pep.append("\"");
        }
        if(trans.hasLatestFiring()) {
            pep.append(PEPReader.TRANS_LATEST);
            pep.append("\"");
            pep.append(trans.getLatestFiring());
            pep.append("\"");
        }
        if(trans.hasRate()) {
            pep.append(PEPReader.TRANS_RATE);
            pep.append("\"");
            pep.append(trans.getRate());
            pep.append("\"");
        }
        if(trans.hasWeight()) {
            pep.append(PEPReader.TRANS_WEIGHT);
            pep.append("\"");
            pep.append(trans.getWeight());
            pep.append("\"");
        }
        if(trans.hasGeneralizedRate()) {
            pep.append(PEPReader.TRANS_GENERALIZED_RATE);
            pep.append("\"");
            pep.append(trans.getGeneralizedRate());
            pep.append("\"");
        }
        if(trans.hasGeneralizedWeight()) {
            pep.append(PEPReader.TRANS_GENERALIZED_WEIGHT);
            pep.append("\"");
            pep.append(trans.getGeneralizedWeight());
            pep.append("\"");
        }
        if(trans.hasAgeMemory()) {
            pep.append(PEPReader.TRANS_AGE_MEMORY);
            pep.append("\"");
            pep.append(trans.getAgeMemory());
            pep.append("\"");
        }
        pep.append("\n");
        return pep.toString();
    }
    
    
    public static String constraintsToString(final Map<String, ValueList> constraints) {
        if((constraints == null) || constraints.isEmpty()) {
            return "";
        } else {
            final StringBuffer str = new StringBuffer();
            for(final Iterator<Map.Entry<String, ValueList>> it = constraints.entrySet().iterator(); it.hasNext(); ) {
                final Map.Entry<String, ValueList> entry = it.next();
                final String var = entry.getKey();
                final ValueList list = entry.getValue();
                if(Value.containsAny(list)) {
                    continue;
                }
                str.append(var);
                str.append("{");
                for(final Iterator<Value> it2 = list.iterator(); it2.hasNext(); ) {
                    final Value val = it2.next();
                    str.append(val.getNetValue());
                    if(it2.hasNext()) {
                        str.append(",");
                    }
                }
                str.append("}");
                if(it.hasNext()) {
                    str.append("&");
                }
            }
            if(str.length() != 0) {
                str.append(":");
            }
            return str.toString();
        }
    }
    
    
    // TODO
    public static String arcsToPEP(final PetriNet net, final Collection<Arc> arcs,
                    final Map<String, Integer> place_id, final Map<String, Integer> trans_id) {
        final StringBuffer tp = new StringBuffer();
        final StringBuffer pt = new StringBuffer();
        final StringBuffer ih = new StringBuffer();
        final StringBuffer rd = new StringBuffer();
        final StringBuffer rs = new StringBuffer();
        
        for(final Arc arc : arcs) {
            switch(arc.getType()) {
                case TP: {
                    if(tp.length() == 0) {
                        tp.append(PEPReader.Section.TRANS2PLACE.getKey());
                        tp.append("\n");
                    }
                    final String label;
                    if(net.isLowLevelNet()) {
                        final int ilabel = arc.getLabel().size();
                        label = ilabel > 1 ? "w" + ilabel : "";
                    } else {
                        label = "p\"" + PEPWriter.escapeStrings(arc.getLabel().toString()) + "\"";
                    }
                    final int fromID = trans_id.get(arc.getTransition().getName());
                    final int toID = place_id.get(arc.getPlace().getName());
                    tp.append(fromID);
                    tp.append("<");
                    tp.append(toID);
                    tp.append(label);
                    tp.append("\n");
                } break;
                case PT: {
                    if(pt.length() == 0) {
                        pt.append(PEPReader.Section.PLACE2TRANS.getKey());
                        pt.append("\n");
                    }
                    final String label;
                    if(net.isLowLevelNet()) {
                        final int ilabel = arc.getLabel().size();
                        label = ilabel > 1 ? "w" + ilabel : "";
                    } else {
                        label = "p\"" + PEPWriter.escapeStrings(arc.getLabel().toString()) + "\"";
                    }
                    final int fromID = place_id.get(arc.getPlace().getName());
                    final int toID = trans_id.get(arc.getTransition().getName());
                    pt.append(fromID);
                    pt.append(">");
                    pt.append(toID);
                    pt.append(label);
                    pt.append("\n");
                } break;
                case INHIBITOR: {
                    if(ih.length() == 0) {
                        ih.append(PEPReader.Section.INHIBITOR_ARCS.getKey());
                        ih.append("\n");
                    }
                    final String label;
                    if(net.isLowLevelNet()) {
                        label = "";
                    } else {
                        label = "p\"" + PEPWriter.escapeStrings(arc.getLabel().toString()) + "\"";
                    }
                    final int fromID = place_id.get(arc.getPlace().getName());
                    final int toID = trans_id.get(arc.getTransition().getName());
                    pt.append(fromID);
                    pt.append(">");
                    pt.append(toID);
                    pt.append(label);
                    pt.append("\n");
                } break;
                case READ: {
                    if(rd.length() == 0) {
                        rd.append(PEPReader.Section.READ_ARCS.getKey());
                        rd.append("\n");
                    }
                    final String label;
                    if(net.isLowLevelNet()) {
                        label = "";
                    } else {
                        label = "p\"" + PEPWriter.escapeStrings(arc.getLabel().toString()) + "\"";
                    }
                    final int fromID = place_id.get(arc.getPlace().getName());
                    final int toID = trans_id.get(arc.getTransition().getName());
                    pt.append(fromID);
                    pt.append(">");
                    pt.append(toID);
                    pt.append(label);
                    pt.append("\n");
                } break;
                case RESET: {
                    if(rs.length() == 0) {
                        rs.append(PEPReader.Section.RESET_ARCS.getKey());
                        rs.append("\n");
                    }
                    final String label;
                    if(net.isLowLevelNet()) {
                        label = "";
                    } else {
                        label = "p\"" + PEPWriter.escapeStrings(arc.getLabel().toString()) + "\"";
                    }
                    final int fromID = place_id.get(arc.getPlace().getName());
                    final int toID = trans_id.get(arc.getTransition().getName());
                    pt.append(fromID);
                    pt.append(">");
                    pt.append(toID);
                    pt.append(label);
                    pt.append("\n");
                } break;
            }
        }
        
        return pt.toString() + tp.toString() + ih.toString() + rd.toString() + rs.toString();
    }

    
    public static void writeToFile(final String net, final String filename) {
        try {
            final File file = new File(filename);
            final FileWriter outputstream = new FileWriter(file);
            final PrintWriter output = new PrintWriter(outputstream);
            output.print(net);
            output.close();
        } catch (final IOException ioex) {
            throw new RuntimeException(ioex);
        }
    }
}
