/*
 * Project:  PG-UML
 * Filename: PNMLWriter.java
 * Created:  17.06.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel;

import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Iterator;
import java.util.Map;

import petruchio.sim.petrinettool.IPetriNet;
import petruchio.sim.petrinettool.PetriNetWriter;
import petruchio.sim.pnmodel.util.Function;
import petruchio.sim.pnmodel.util.Node;
import petruchio.sim.pnmodel.util.Pool;
import petruchio.sim.pnmodel.util.RPNComputer;
import petruchio.sim.pnmodel.util.Token;
import petruchio.sim.pnmodel.util.TreeComputer;
import petruchio.sim.pnmodel.net.PetriNet;
import petruchio.sim.pnmodel.net.Transition;
import petruchio.sim.pnmodel.net.Place;
import petruchio.sim.pnmodel.net.Arc;
import petruchio.sim.pnmodel.net.Value;
import petruchio.sim.pnmodel.net.Tuple;
import petruchio.sim.pnmodel.net.MultiSet;

/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class PNMLWriter implements PetriNetWriter {
    public String writeString(final IPetriNet originalNet) {
        final PetriNet net = (PetriNet)originalNet.clone();
        
        final StringBuffer pnml = new StringBuffer();
        pnml.append("<?xml version=\"1.0\" ?>\n<pnml>\n<net id=\"" + net.getName() + "\">\n<page>\n");
        
        final Map<String, Function> functions = PNMLWriter.createFunctions(net);
        
        pnml.append(PNMLWriter.functionsToPNML(functions));
        
        for(final Iterator<Place> it = net.getPlaceIterator(); it.hasNext(); ) {
            final Place place = it.next();
            pnml.append(PNMLWriter.placeToPNML(place));
        }
        
        for(final Iterator<Transition> it = net.getTransitionIterator(); it.hasNext(); ) {
            final Transition trans = it.next();
            pnml.append(PNMLWriter.transToPNML(trans, functions));
        }
        
        for(final Iterator<Arc> it = net.getArcIterator(); it.hasNext(); ) {
            final Arc arc = it.next();
            pnml.append(PNMLWriter.arcToPNML(arc, functions));
        }
        
        pnml.append("</page>\n</net>\n</pnml>");
        return pnml.toString();
    }
    
    
    public void writeFile(final IPetriNet originalNet, final String filename) {
        final PetriNet net = (PetriNet)originalNet.clone();
        
        try {
            final File file = new File(filename);
            final FileWriter outputstream = new FileWriter(file);
            final PrintWriter output = new PrintWriter(outputstream);
            output.print("<?xml version=\"1.0\" ?>\n<pnml>\n<net id=\"" + net.getName() + "\">\n<page>\n");
        
            final Map<String, Function> functions = PNMLWriter.createFunctions(net);
            
            output.print(PNMLWriter.functionsToPNML(functions));
            
            for(final Iterator<Place> it = net.getPlaceIterator(); it.hasNext(); ) {
                final Place place = it.next();
                output.print(PNMLWriter.placeToPNML(place));
            }
            
            for(final Iterator<Transition> it = net.getTransitionIterator(); it.hasNext(); ) {
                final Transition trans = it.next();
                output.print(PNMLWriter.transToPNML(trans, functions));
            }
            
            for(final Iterator<Arc> it = net.getArcIterator(); it.hasNext(); ) {
                final Arc arc = it.next();
                output.print(PNMLWriter.arcToPNML(arc, functions));
            }
            
            output.print("</page>\n</net>\n</pnml>");
            output.close();
            outputstream.close();
        } catch (final Exception ex) {
            throw new RuntimeException(ex);
        }
    }
    
    
    private static String functionsToPNML(final Map<String, Function> functions) {
        final StringBuffer str = new StringBuffer();
        if(!functions.isEmpty()) {
            str.append("<declaration>\n<structure>\n");
            for(final Function f : functions.values()) {
                str.append("<declare type=\"function\" nargs=\"" + f.getParameterCount()
                        + "\">\n<ci>" + PNMLWriter.htmlEncode(f.getName()) + "</ci>\n<fn>\n");
                for(final Iterator<Value> pit = f.getParameters().iterator(); pit.hasNext(); ) {
                    str.append("<bvar><ci>");
                    str.append(pit);
                    str.append("</ci></bvar>\n");
                }
                str.append(PNMLWriter.nodeToPNML(f.getTree(), functions));
                str.append("</fn>\n</declare>\n");
            }
            str.append("</structure>\n</declaration>\n");
        }
        return str.toString();
    }
    
    
    private static Map<String, Function> createFunctions(final PetriNet net) {
        final Map<String, Function> functions = Pool.getPool().getMap();
        for(final String function : net.getFunctions()) {
            final Function func = new Function(function);
            if(functions.containsKey(func.getName())) {
                System.err.println("Warning: multiple definitions of function "
                                + func.getName() + "\nThe function will be "
                                + "overwritten.\nOld: "
                                + functions.get(func.getName())
                                + "\nNew: " + func);
            }
            functions.put(func.getName(), func);
        }
        return functions;
    }

    
    public static String placeToPNML(final Place place) {
        final String type_value = place.getType().getPNML();
        final String status_value = place.getStatus().getPNML();
        final String name = PNModel.htmlEncode(place.getName());
        final String curmarking;
        final String initmarking;
        if(place.getCurrentMarkingSet().isEmpty()) {
            curmarking = "";
        } else {
            final StringBuffer s = new StringBuffer(); 
            s.append("<hlcurrentMarking>\n<structure>\n<set type=\"multiset\">\n");
            for(final Iterator<Value> it = place.getCurrentMarkingSet().iterator(); it.hasNext();) {
                s.append(PNMLWriter.valueToPNML(it.next()));
            }
            s.append("</set>\n</structure>\n</hlcurrentMarking>\n");
            curmarking = s.toString();
        }
        if(place.getInitialMarkingSet().isEmpty()) {
            initmarking = "";
        } else {
            final StringBuffer s = new StringBuffer(); 
            s.append("<hlinitialMarking>\n<structure>\n<set type=\"multiset\">\n");;
            for(final Iterator<Value> it = place.getInitialMarkingSet().iterator(); it.hasNext();) {
                s.append(PNMLWriter.valueToPNML(it.next()));
            }
            s.append("</set>\n</structure>\n</hlinitialMarking>\n");
            initmarking = s.toString();
        }
        
        return "<place id=\"" + name + "\">\n"
                + "<graphics><position x=\"" + place.getPosX() + "\" y=\"" + place.getPosY()
                            + "\" /></graphics>\n"
                + "<name><text>" + name + "</text></name>\n"
                + "<type><structure>" + type_value + "</structure></type>\n"
                + "<status><text>" + status_value + "</text></status>\n"
                + (place.getCapacity() < 0 ? "" : "<capacity><text>" + place.getCapacity()
                            + "</text></capacity>\n")
                + (place.getMeaning().length() == 0 ? "" : "<meaning><text>"
                            + PNMLWriter.htmlEncode(place.getMeaning()) + "</text></meaning>\n")
                + (initmarking.length() == 0 ? "" : initmarking)
                + ( curmarking.length() == 0 ? "" : curmarking)
                + "</place>\n";
    }


    public static String transToPNML(final Transition trans, final Map<String, Function> functions) {
        final String name = PNMLWriter.htmlEncode(trans.getName());
        final String guard = PNMLWriter.expressionToPNML(trans.getGuardTree(), functions);
        final String actions = PNMLWriter.htmlEncode(trans.getActions().toString());
        
        final String time_weight;
        if(trans.isImmediate()) {
            if(trans.hasWeight()) {
                time_weight = "<weight>\n<structure>\n"
                    + PNMLWriter.nodeToPNML(trans.getWeightTree(), functions)
                    + "</structure>\n</weight>\n";
            } else if(trans.hasGeneralizedWeight()) {
                time_weight = "<generalizedWeight>\n<structure>\n"
                    + PNMLWriter.nodeToPNML(trans.getGeneralizedWeightTree(), functions)
                    + "</structure>\n</generalizedWeight>\n";
            } else {
                time_weight = "";
            }
        } else if(trans.hasRate()) {
            time_weight = "<rate>\n<structure>\n"
                        + PNMLWriter.nodeToPNML(trans.getRateTree(), functions)
                        + "</structure>\n</rate>\n";
        } else if(trans.hasGeneralizedRate()) {
            time_weight = "<generalizedRate>\n<structure>\n"
                        + PNMLWriter.nodeToPNML(trans.getGeneralizedRateTree(), functions)
                        + "</structure>\n</generalizedRate>\n";
        } else if(trans.hasEarliestFiring() || trans.hasLatestFiring()){
            time_weight = "<timeConstraint>\n<structure>\n<interval>\n"
                        + PNMLWriter.nodeToPNML(trans.getEarliestFiringTree(), functions)
                        + PNMLWriter.nodeToPNML(trans.getLatestFiringTree(), functions)
                        + "</interval>\n</structure>\n</timeConstraint>\n";
        } else {
            time_weight = "";
        }
        
        final String priority;
        if(!trans.hasPriority()) {
            priority = "";
        } else {
            priority = "<priority>\n<structure>\n"
                    + PNMLWriter.expressionToPNML(trans.getPriorityTree(), functions)
                    + "</structure>\n</priority>\n";
        }
        
        final String servers;
        if(!trans.hasServers()) {
            servers = "";
        } else {
            servers = "<servers>\n<structure>\n"
                    + PNMLWriter.expressionToPNML(trans.getServersTree(), functions)
                    + "</structure>\n</servers>\n";
        }
        
        final String ageMemory;
        if(!trans.hasAgeMemory()) {
            ageMemory = "";
        } else {
            ageMemory = "<ageMemory>\n<structure>\n"
                    + PNMLWriter.expressionToPNML(trans.getAgeMemoryTree(), functions)
                    + "</ageMemory>\n</servers>\n";
        }
        
        return "<transition id=\"" + name + "\">\n"
                + "<graphics><position x=\"" + trans.getPosX() + "\" y=\"" + trans.getPosY()
                        + "\" /></graphics>\n"
                + "<name><text>" + name + "</text></name>\n"
                + "<condition>\n" + guard + "</condition>\n"
                + (trans.getMeaning().length() == 0 ? "" : "<meaning><text>"
                        + PNMLWriter.htmlEncode(trans.getMeaning()) + "</text></meaning>\n")
                + (((MultiSet) trans.getActions()).isEmpty() ? "" : "<actions><text>" + actions
                            + "</text></actions>\n")
                + time_weight
                + priority
                + servers
                + ageMemory
                + "</transition>\n";
    }
    
    public static String tupleToPNML(final Tuple t) {
        StringBuffer str = new StringBuffer();
        str.append("<vector>\n");
        for(int i = 0; i < t.size(); i++) {
            str.append(PNMLWriter.valueToPNML(t.getValue(i)));
        }
        str.append("</vector>\n");
        return str.toString();
    }
    
    
    public static String valueToPNML(final Value v) {
        if(v.isBoolean()) {
            return "<" + v.getBoolean() + " />\n";
        } else if(v.isDot()) {
            return "<symbol>dot</symbol>\n";
        } else if(v.isNumber()) {
            if(Double.POSITIVE_INFINITY == v.getNumber()) {
                return "<infinity />\n";
            } else if(Double.NEGATIVE_INFINITY == v.getNumber()) {
                return "<apply><minus /><infinity /></apply>\n";
            } else if(Double.isNaN(v.getNumber())) {
                return "<notanumber />\n";
            } else {
                return "<cn>" + v.getNetValue() + "</cn>\n";
            }
        } else if(v.isVariable()) {
            return "<ci>" + PNMLWriter.htmlEncode(v.getVariable()) + "</ci>\n";
        } else if(v.isString()) {
            return "<symbol type=\"string\">" + PNMLWriter.htmlEncode(v.getString()) + "</symbol>\n";
        } else if(v.isTuple()) {
            return PNMLWriter.tupleToPNML(v.getTuple());
        } else {
            return "";
        }
    }
    
    public static String nodeToPNML(final Node node, final Map<String, Function> functions) {
        final Token o = node.getToken();
        final Token f = node.getParent() == null ? null : node.getParent().getToken();
        if(o.isValue()) {
            return PNMLWriter.valueToPNML(o.getValue());
        } else if((f != null) && o.equals(f) && o.isOperator()) {
            
            final StringBuffer str = new StringBuffer();
            for(int i = 0; i < node.getChildren().size(); i++) {
                str.append(PNMLWriter.nodeToPNML(node.getChildren().get(i), functions));
            }
            return str.toString();
            
        } else {
            final StringBuffer str = new StringBuffer();
            boolean special = false;
            String begin = "<apply>\n";
            String end = "</apply>\n";
            
            if(o.isOperator()) {
                final Token.Operator op = o.getOperator();
                if(op == Token.Operator.PLUS) {
                    begin += "<plus />\n";
                    
                } else if(op == Token.Operator.UNARY_MINUS) {
                    begin += "<minus />\n";
                    
                } else if(op == Token.Operator.MINUS) {
                    begin += "<minus />\n";
                    
                } else if(op == Token.Operator.TIMES) {
                    begin += "<times />\n";
                    
                } else if(op == Token.Operator.DIVIDE) {
                    begin += "<divide />\n";
                    
                } else if(op == Token.Operator.MODULO) {
                    begin += "<rem />\n";
                    
                } else if(op == Token.Operator.POWER) {
                    begin += "<power />\n";
                    
                } else if(op == Token.Operator.EQUAL) {
                    begin += "<eq />\n";
                    
                } else if(op == Token.Operator.NOT_EQUAL) {
                    begin += "<neq />\n";
                    
                } else if(op == Token.Operator.LESS) {
                    begin += "<lt />\n";
                    
                } else if(op == Token.Operator.LESS_EQUAL) {
                    begin += "<leq />\n";
                    
                } else if(op == Token.Operator.GREATER) {
                    begin += "<gt />\n";
                    
                } else if(op == Token.Operator.GREATER_EQUAL) {
                    begin += "<geq />\n";
                    
                } else if(op == Token.Operator.AND) {
                    begin += "<and />\n";
                    
                } else if(op == Token.Operator.OR) {
                    begin += "<or />\n";
                    
                } else if(op == Token.Operator.NOT) {
                    begin += "<not />\n";
                    
                } else if(op == Token.Operator.STRING_CONCATENATION) {
                    begin += "<plus />\n";
                    
                } else if(op == Token.Operator.SEPARATOR) {
                    begin = "<vector>\n";
                    end = "</vector>\n";
                } else {
                    throw new RuntimeException("I cannot translate \"" + node + "\" to MathML 2.0.");
                }
            
            } else if(o.isFunction()) {
                final String op = o.getFunction();
                
                if(op.equals("logn")) {
                    begin += "<ln />\n";
                    
                } else if(op.equals("sqrt")) {
                    begin += "<root />\n";
                    
                } else if(op.equals("cbrt")) {
                    begin += "<root />\n<degree><cn>3</cn></degree>\n";
                    
                } else if(op.equals("root")) {
                    str.append(begin);
                    str.append("<root />\n<degree><cn>");
                    str.append(PNMLWriter.nodeToPNML(node.getChildren().get(0), functions));
                    str.append("</cn></degree>\n");
                    str.append(PNMLWriter.nodeToPNML(node.getChildren().get(1), functions));
                    str.append(end);
                    
                } else if(op.equals("asin")) {
                    begin += "<arcsin />\n";
                    
                } else if(op.equals("acos")) {
                    begin += "<arccos />\n";
                    
                } else if(op.equals("atan")) {
                    begin += "<arctan />\n";
                    
                } else if(op.equals("ceil")) {
                    begin += "<ceiling />\n";
                    
                } else if(op.equals("round")) {
                    begin += "<floor />\n<apply>\n<plus />\n<cn>0.5</cn>\n";
                    end = "</apply>\n" + end;
                    
                } else if(op.equals("deg")) {
                    begin += "<divide />\n<apply>\n<times />\n<cn>180</cn>\n";
                    end = "</apply>\n<pi />\n" + end;
                    
                } else if(op.equals("rad")) {
                    begin += "<divide />\n<apply>\n<times />\n<pi />\n";
                    end = "</apply>\n<cn>180</cn>\n" + end;
                    
                } else if(op.equals("e")) {
                    begin = "<e />\n";
                    end = "";
                    
                } else if(op.equals("pi")) {
                    begin = "<pi />\n";
                    end = "";
                    
                } else if(op.equals("log10")) {
                    begin += "<log />\n";
                    end = "<cn>10</cn>\n" + end;
                    
                } else if(op.equals("dim")) {
                    begin += "<card />\n<apply>\n<selector />\n";
                    end = "</apply>\n" + end;
                    
                } else if(op.equals("proj")) {
                    special = true;
                    begin += "<selector />\n";
                    str.append(begin);
                    str.append(PNMLWriter.nodeToPNML(node.getChildren().get(0), functions));
                    str.append("<apply>\n<plus />\n<cn>1</cn>\n");
                    str.append(PNMLWriter.nodeToPNML(node.getChildren().get(1), functions));
                    str.append("</apply>");
                    str.append(end);
                    
                } else if(op.equals("prod") || op.equals("sum")) {
                    begin += "<" + o + " />\n<bvar><ci>x</ci></bvar><condition>\n<apply>\n<in /><ci>x</ci>\n<set>\n";
                    end = "</set>\n</apply>\n</condition>\n<apply>\n<ci type=\"function\">f</ci>\n<ci>x</ci>\n</apply>\n" + end;
                    special = node.getChildren().size() == 1;
                    if(special) {
                        str.append(begin);
                        str.append(PNMLWriter.nodeAsList(node.getChildren().get(0), functions));
                        str.append(end);
                    }
                    
                } else if(op.equals("forall") || op.equals("exists")) {
                    special = true;
                    begin += "<" + o + " />\n";
                    str.append(begin);
                    final String var = node.getChildren().get(0).getToken().getValue().getVariable();
                    str.append("<bvar>");
                    str.append(var);
                    str.append("</bvar>\n");
                    if(node.getChildren().size() == 3) {
                        str.append("<condition>\n<apply>\n<in />\n<ci>");
                        str.append(var);
                        str.append("</ci>\n<set>\n");
                        str.append(PNMLWriter.nodeAsList(node, functions));
                        str.append("</set>\n</apply>\n</condition>");
                        
                    } else if(node.getChildren().size() == 5) {
                        final String from = PNMLWriter.nodeToPNML(node.getChildren().get(1), functions);
                        str.append("<condition>\n<apply>\n<in />\n<ci>");
                        str.append(var);
                        str.append("</ci>\n<set>\n");
                        str.append("<bvar><ci>x</ci></bvar>\n<condition>\n<apply>\n<and />\n<apply>\n<in />\n");
                        str.append("<ci>x</ci>\n<integer />\n</apply>\n<apply>\n<leq />\n");
                        str.append(from);
                        str.append("<apply>\n<plus /><apply>\n<times /><ci>x</ci>\n");
                        str.append(PNMLWriter.nodeToPNML(node.getChildren().get(3), functions));
                        str.append("</apply>\n");
                        str.append(from);
                        str.append("</apply>");
                        str.append(PNMLWriter.nodeToPNML(node.getChildren().get(2), functions));
                        str.append("</apply>\n</condition>\n</set>\n</apply>\n</condition>");
                    }
                    
                    final Node cond = node.getChildren().get(node.getChildren().size()-1);
                    str.append(PNMLWriter.nodeToPNML(cond, functions));
                    str.append(end);
                    
                } else if(op.equals("abs") || op.equals("floor") || op.equals("exp") || op.equals("min")
                        || op.equals("log") || op.equals("sin") || op.equals("cos") || op.equals("tan")
                        || op.equals("cot") || op.equals("sinh") || op.equals("cosh") || op.equals("tanh")
                        || op.equals("coth") || op.equals("max")) {
                    begin += "<" + o + " />\n";
                    
                } else if(functions.containsKey(Function.getFunctionName(op, node.getChildren().size()))) {
                    begin += "<fn><ci>" + PNMLWriter.htmlEncode(Function.getFunctionName(op, node.getChildren().size())) + "</ci></fn>"; 
                    
                } else {
                    throw new RuntimeException("I cannot translate \"" + node + "\" to MathML 2.0.");
                }
            }
            
            if(!special) {
                str.append(begin);
                for(int i = 0; i < node.getChildren().size(); i++) {
                    str.append(PNMLWriter.nodeToPNML(node.getChildren().get(i), functions));
                }
                str.append(end);
            }
            
            return str.toString();
        }
    }
    
    public static String nodeAsList(final Node n, final Map<String, Function> functions) {
        final StringBuffer str = new StringBuffer();
        final Token v = n.getToken();
        if(v.isOperator() && v.getOperator() == Token.Operator.SEPARATOR) {
            for(int i = 0; i < n.getChildren().size(); i++) {
                str.append(PNMLWriter.nodeToPNML(n.getChildren().get(i), functions));
            }
        } else if(v.isValue()) {
            final Value val = v.getValue();
            if(val.isTuple()) {
                final Tuple tup = val.getTuple();
                for(int i = 0; i < tup.size(); i++) {
                    str.append(PNMLWriter.valueToPNML(tup.getValue(i)));
                }
            } else {
                str.append(PNMLWriter.valueToPNML(val));
            }
            
        } else {
            throw new RuntimeException("Bad syntax.");
        }
        return str.toString();
    }
    
    public static String expressionToPNML(final Node expr, final Map<String, Function> functions) {
        final StringBuffer str = new StringBuffer();
        str.append("<text>");
        str.append(PNMLWriter.htmlEncode(PEPWriter.constraintsToString(expr.getConstraints()) + expr.toString()));
        str.append("</text>\n");
        str.append("<structure>\n");
        str.append(PNMLWriter.nodeToPNML(expr, functions));
        str.append("</structure>\n");
        return str.toString();
    }


    public static String htmlEncode(final String str) {
        final StringBuffer buf = new StringBuffer();
        for (int i = 0; i < str.length(); i++) {
            char c = str.charAt(i);
            switch (c) {
                case '&':
                    buf.append("&amp;");
                    break;
                case '<':
                    buf.append("&lt;");
                    break;
                case '>':
                    buf.append("&gt;");
                    break;
                case '"':
                    buf.append("&quot;");
                    break;
                default:
                    buf.append(c);
            }
        }

        return buf.toString();
    }


    public static String arcToPNML(final Arc arc, final Map<String, Function> functions) {
        final String source;
        final String target;
        if (arc.transIsSource()) {
            source = PNMLWriter.htmlEncode(arc.getTransition().getName());
            target = PNMLWriter.htmlEncode(arc.getPlace().getName());
        } else {
            source = PNMLWriter.htmlEncode(arc.getPlace().getName());
            target = PNMLWriter.htmlEncode(arc.getTransition().getName());
        }

        final String label;
        if(arc.getLabel().size() == 0) {
            label = "";
        } else {
            final StringBuffer s = new StringBuffer();
            s.append("<set type=\"multiset\">\n");
            for(final Iterator<Value> it = arc.getLabel().getValues().iterator(); it.hasNext(); ) {
                s.append(PNMLWriter.valueToPNML(it.next()));
            }
            for(final Iterator<Value> it = arc.getLabel().getVariables().iterator(); it.hasNext(); ) {
                s.append(PNMLWriter.valueToPNML(it.next()));
            }
            for(final Iterator<Value> it = arc.getLabel().getTuples().iterator(); it.hasNext(); ) {
                s.append(PNMLWriter.valueToPNML(it.next()));
            }
            for(final Iterator<String> it = arc.getLabel().getExpressions().iterator(); it.hasNext(); ) {
                s.append(PNMLWriter.nodeToPNML(TreeComputer.rpnToTreeWithDeletion(
                            RPNComputer.infixToRPN(it.next())), functions));
            }
            s.append("</set>\n");
            label = s.toString();
        }
        
        final String type;
        switch(arc.getType()) {
            case PT:
                type = "pt";
                break;
            case TP:
                type = "tp";
                break;
            case INHIBITOR:
                type = "inhibitor";
                break;
            case READ:
                type = "read";
                break;
            case RESET:
                type = "reset";
                break;
            default:
                throw new RuntimeException("Unknown arc type.");
        }

        return "<arc source=\"" + source + "\" target=\"" + target
                + "\" type=\"" + type + "\">\n<hlinscription>\n<text>\n"
                + arc.getLabel().toString() + "\n</text>\n<structure>\n" + label
                + "</structure>\n</hlinscription>\n</arc>\n";
    }
}
