/*
 * Project:  UML
 * Filename: PNModel.java
 * Created:  24.04.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel;

import petruchio.sim.pnmodel.util.Node;
import petruchio.sim.pnmodel.util.Pool;
import petruchio.sim.pnmodel.util.RPN;
import petruchio.sim.pnmodel.util.RPNComputer;
import petruchio.sim.pnmodel.util.Token;
import petruchio.sim.pnmodel.util.TreeComputer;
import petruchio.sim.pnmodel.util.StringList;
import petruchio.sim.pnmodel.net.Value;
import petruchio.sim.pnmodel.net.PetriNet;
import petruchio.sim.pnmodel.net.MultiSet;
import petruchio.sim.pnmodel.net.ValueList;

import java.util.Map;
import java.util.List;
import java.util.Collection;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import petruchio.sim.petrinettool.IPlace;
import petruchio.sim.petrinettool.IPlaceType;
import petruchio.sim.petrinettool.IPetriNetTool;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class PNModel implements IPetriNetTool {

    public static final Pattern CONTAINS_FUNCTION = Pattern.compile(".*\\w+\\s*\\(.*");
    public static final Pattern ALLOWED_LITERALS = Pattern.compile(IPetriNetTool.TRUE + "|" + IPetriNetTool.FALSE + "|" + IPetriNetTool.DOT);
    public static final Pattern STRING_LITERAL = Pattern.compile(IPetriNetTool.STRING_SEPARATOR + ".+?" + IPetriNetTool.STRING_SEPARATOR);
    public static final Pattern SIMPLIFY_TUPLE = Pattern.compile("(?<=[,\\(])[^,\\(\\)]*?(?=[\\),])");
    public static final String WORD = "a-zA-Z0-9'._\"";
    public static final String MY_NONWORD = "[^" + WORD + "]";
    public static final String SPACES =  " +(?="+MY_NONWORD+")|(?<="+MY_NONWORD+") +";
    public static final Pattern REMOVE = Pattern.compile(SPACES);
    public static final Pattern WHITESPACES = Pattern.compile("\\s+");
    public static final Pattern DOT = Pattern.compile("([^a-zA-Z0-9_.])\\.");
    public static final Pattern FALSE = Pattern.compile("\\s*ff\\s*");
    public static final Pattern NUMBER = Pattern.compile("\\s*-?(\\d+\\.?\\d*|\\.\\d+)\\s*");
    public static final Pattern VALUE = Pattern.compile("\\s*(ff|tt|"
            + IPetriNetTool.STRING_SEPARATOR+".+"
            + IPetriNetTool.STRING_SEPARATOR + "|dot|-?(\\d+\\.?\\d*|\\.\\d+))\\s*");
    public static final String DOT_REPL = "$1_";

    static final Pattern[] SINGLE_REPLACE = 
        {
                    Pattern.compile("Math\\."),
                    Pattern.compile("\\band\\b|&&|&", Pattern.CASE_INSENSITIVE),      // AND
                    Pattern.compile("\\bor\\b|\\|\\||\\|", Pattern.CASE_INSENSITIVE), // OR
                    Pattern.compile("\\bnot\\b", Pattern.CASE_INSENSITIVE),           // NOT
                    Pattern.compile(IPetriNetTool.STRING_SEPARATOR + "true" + IPetriNetTool.STRING_SEPARATOR + "|\\btrue\\b", Pattern.CASE_INSENSITIVE),   // TRUE
                    Pattern.compile(IPetriNetTool.STRING_SEPARATOR + "false" + IPetriNetTool.STRING_SEPARATOR + "|\\bfalse\\b", Pattern.CASE_INSENSITIVE), // FALSE
                    Pattern.compile("\t"),   // TABS
                    Pattern.compile("\n\n"), // DOUBLE BREAKS
//                    Pattern.compile("\\bX"), // BIG X
//                    Pattern.compile("\\bU"), // BIG U
                    Pattern.compile(":="),   // ASSIGN
                    Pattern.compile("=="),   // EQUAL
                    //Pattern.compile("([^=&\\|!\\(\\)]+)(!=|<>)([^=&\\|!\\(\\)]+)") // NOT EQUAL
                    Pattern.compile("!=|<>") // NOT EQUAL
        };
    
    static final String[] SINGLE_REPL =
        {
                    "",
                    Token.Operator.AND.toString(),
                    Token.Operator.OR.toString(),
                    Token.Operator.NOT.toString(),
                    "tt",
                    "ff",
                    " ",
                    "\n",
//                    "x",
//                    "u",
                    Token.Operator.EQUAL.toString(),
                    Token.Operator.EQUAL.toString(),
                    //"($1#$3)" // "!($1=$3)"
                    Token.Operator.NOT_EQUAL.toString()
        };
    
    private IPlaceType anyType = IPlace.TYPE_BLACK_TOKEN;
    
    private Map<String, IPlaceType> tieTypes;
    
    
    public PNModel() {
        this.tieTypes = Pool.getPool().getMap();
    }
    
    
    public IPlaceType getTypeOf(final String symbol) {
        return this.tieTypes.get(symbol);
    }
    
    
    public void clearTieTypes() {
        this.tieTypes.clear();
    }
    
    
    public void setTieType(final String symbol, final IPlaceType type) {
        if(type == null) {
            throw new RuntimeException("Cannot set type for tie place with "
                            + "symbol \"" + symbol + "\" to null.");
        }
        this.tieTypes.put(symbol, type);
    }
    
    
    public void setAnyType(final IPlaceType type) {
        this.anyType = type;
    }
    
    
    public IPlaceType getAnyType() {
        return this.anyType;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.sim.petrinettool.PetriNetTool#createPetriNet()
     */
    public PetriNet createPetriNet() {
        return new PetriNet(this);
    }


    /* (non-Javadoc)
     * @see petruchio.sim.petrinettool.PetriNetTool#createEmptyMultiSet()
     */
    public MultiSet createMultiSet() {
        return new MultiSet();
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.sim.petrinettool.PetriNetTool#getVariablesFromGuard(String)
     */
    public String[] getVariablesFromGuard(final String guard) {
        List<String> vars = Pool.getPool().getList();
        
        final Matcher varMatcher = IPetriNetTool.GUARDVARS.matcher(guard);
        while(varMatcher.find()) {
            final String var = guard.substring(varMatcher.start(1), varMatcher.end(1));
            if(!vars.contains(var)) {
                vars.add(var);
            }
        }
        
        String[] strings = new String[vars.size()];
        vars.toArray(strings);
        return strings;
    }

    
    /* (non-Javadoc)
     * @see petruchio.sim.petrinettool.PetriNetTool#writesPositions()
     */
    public boolean writesPositions() {
        return true;
    }
    

    /* (non-Javadoc)
     * @see petruchio.sim.petrinettool.PetriNetTool#close()
     */
    public void close() { /* nothing to clean up */ }

    
    /**
     * Concatenates all the elements from the array and putting the glue
     * between them.
     * <code>join({"Hell", "world", "something."}, " of a ")</code> returns
     * <code>"Hell of a world of a something."</code>
     * 
     * @param pieces The String elements to concatenate.
     * @param glue The String to put between each pair of elements.
     * @return The concatenated string.
     */
    public static String join(final String[] pieces, final String glue) {
        if(pieces == null) {
            return "";
        }
        
        final StringBuffer buf = new StringBuffer();
        for(int i = 0; i < pieces.length; i++) {
            buf.append(pieces[i]);
            if(i < pieces.length-1) {
                buf.append(glue);
            }
        }
        
        return buf.toString();
    }
    
    
    /**
     * Returns an ArrayList that contains all the elements from list
     * <code>a</code> that are not elements in list <code>b</code>.
     * {@link #containsThisObject(ArrayList, Object)} is used.
     * 
     * @param <T> 
     * @param a An ArrayList.
     * @param b Another ArrayList.
     * @return An ArrayList with elements from a, that are not in b.
     */
    public static <T> List<T> setMinus(final List<T> a, final List<?> b) {
        List<T> result = Pool.getPool().getList();
        for(int i = 0; i < a.size(); i++) {
            T o = a.get(i);
            if(!PNModel.containsThisObject(b, o)) {
                result.add(o);
            }
        }
        return result;
    }
    
    
    /**
     * TODO
     * 
     * @param str
     * @param character
     * @return Number of occurrences of character in str.
     */
    public static int countChars(final String str, final char character) {
        int count = -1;
        int start = -1;
        do {
            ++count;
            ++start;
            start = str.indexOf(character, start);
        } while(start >= 0);
        return count;
    }
    
    
    /**
     * TODO: document
     *
     * @param list
     * @return Number of separate thingies. :-|
     */
    public static int countParts(final String list) {
        int variableCount = PNModel.countChars(list, ',');
        if((variableCount == 0) && (list.length() > 0)) {
            variableCount = 1;
        }
        return variableCount;
    }

    
    /**
     * TODO: document
     *
     * @param text
     * @param values
     * @param variables
     * @param tuples
     * @param expressions
     */
    public static void parse(final String text, final Collection<Value> values,
                    final Collection<Value> variables,
                    final Collection<Value> tuples,
                    final Collection<String> expressions) {
        PNModel.parse(text, values, variables, tuples, expressions, false);
    }
    
    public static void parse(final String text, final Collection<Value> values,
                    final Collection<Value> variables,
                    final Collection<Value> tuples,
                    final Collection<String> expressions, boolean strings) {
        char[] txt = text.toCharArray();
        String part;
        int pos = 0;
        boolean isExpression = false;
        while(pos < txt.length) {
            int start = pos;
            int par1 = 0; // ()
            int par2 = 0; // {}
            int par3 = 0; // []
            do {
                switch(txt[pos]) {
                    case '(' : ++par1; break;
                    case ')' : --par1; break;
                    case '{' : ++par2; break;
                    case '}' : --par2; break;
                    case '[' : ++par3; break;
                    case ']' : --par3; break;
                    case '+' : case '-' :
                        if(pos > 0) {
                            switch(txt[pos-1]) {
                                case '(' : case '[' : case '{' :
                                case '*' : case '/' : case '%' :
                                case '&' : case '|' : case '!' :
                                case '=' : case '#' : case '<' :
                                case '>' : case ',' : case ';' :
                                    // unary operator
                                    break;
                                default :
                                    // binary operator
                                    isExpression = true;
                                    break;
                            }
                        }
                        break;
                    case ':' : case '*' : case '/' :
                    case '%' : case '&' : case '|' : case '!' : case '=' :
                    case '#' : case '<' : case '>' :
                        isExpression = true;
                        break;
                    default  : break; 
                }
                ++pos;
            } while((pos < txt.length) && ((par1 + par2 + par3 > 0) || (txt[pos] != ',')));
            part = text.substring(start, pos);
            isExpression = isExpression || PNModel.CONTAINS_FUNCTION.matcher(part).find();
            ++pos;
            
            if(part.equals(IPlace.STRING_ANY)) {
                values.add(Value.V_ANY);
            } else if(isExpression) {
//                System.out.println("EXP: " + part);
                expressions.add(part);
            } else {
                final Value v;
                if(strings) {
                    v = Value.getValueNoVars(part);
                } else {
                    v = Value.getValue(part);
                }
                    
                if(v.isTuple()) {
                    tuples.add(v);
                    //System.out.println("TUP: " + part);
                } else if(v.isVariable()) {
                    variables.add(v);
                    //System.out.println("VAR: " + part);
                } else {
                    values.add(v);
                    //System.out.println("VAL: " + part);
                }
            }
        }
    }
    
    
    /**
     * TODO: document
     *
     * @param str
     * @return true iff the str represents a value
     */
    public static boolean isValue(final String str) {
        return (str != null) && (str.length() > 0) 
                && PNModel.VALUE.matcher(str).matches();
    }
    
    
    /**
     * TODO: document
     *
     * @param str
     * @return true iff the str represents a number
     */
    public static boolean isNumber(final String str) {
        return (str != null) && (str.length() > 0)
                && PNModel.NUMBER.matcher(str).matches();
    }
    
    
    /**
     * TODO: document
     *
     * @param guard
     * @return true iff the str represents true
     */
    public static boolean isTrue(final String guard) {
        return guard.equals(IPetriNetTool.TRUE);
    }
    
    
    /**
     * Removes the specific object in the ArrayList and returns
     * <code>true</code> if the ArrayList contained that object,
     * <code>false</code> if otherwise.
     * The difference to {@link ArrayList#remove(Object)} is that this method
     * uses <code>==</code> where {@link ArrayList#remove(Object)} uses
     * <code>.equals(Object)</code>.
     * 
     * @param list The list to be searched.
     * @param obj The object to remove.
     * @return <code>true</code> if the Object was successfully removed.
     * @see #indexOfThisObject(ArrayList, Object)
     * @see #containsThisObject(ArrayList, Object)
     */
    public static boolean removeThisObject(final List<?> list, final Object obj) {
        if(list == null) {
            return false;
        }
        int index;
        boolean removed = false;
        while((index = PNModel.indexOfThisObject(list, obj)) >= 0) {
            list.remove(index);
            removed = true;
        }
        return removed;
    }
    
    
    /**
     * Searches for the specific object in the ArrayList and returns
     * <code>true</code> if the ArrayList contains that object,
     * <code>false</code> otherwise.
     * The difference to {@link ArrayList#contains(Object)} is that this method
     * uses <code>==</code> where {@link ArrayList#contains(Object)} uses
     * <code>.equals(Object)</code>.
     * 
     * @param set The collection to be searched.
     * @param obj The object to search for.
     * @return <code>true</code> if the Object is an element of the list.
     * @see #indexOfThisObject(ArrayList, Object)
     * @see #removeThisObject(ArrayList, Object)
     */
    public static boolean containsThisObject(final Collection<?> set, final Object obj) {
        for(Object o : set) {
            if(obj == o) {
                return true;
            }
        }
        return false;
    }
    
    
    /**
     * Searches for the specific object in the ArrayList and returns its
     * index or -1 if the ArrayList doesn't contain that object.
     * The difference to {@link ArrayList#indexOf(Object)} is that this method
     * uses <code>==</code> where {@link ArrayList#indexOf(Object)} uses
     * <code>.equals(Object)</code>.
     * 
     * @param list The list to be searched.
     * @param obj The object to search for.
     * @return The index of the object in the ArrayList or -1 if the list
     *          doesn't contain the object.
     * @see #containsThisObject(ArrayList, Object)
     * @see #removeThisObject(ArrayList, Object)
     */
    public static int indexOfThisObject(final List<?> list, final Object obj) {
        for(int i = 0; i < list.size(); i++) {
            if(obj == list.get(i)) {
                return i;
            }
        }
        return -1;
    }
    
    
    public static String htmlEncode(final String str) {
        final StringBuffer buf = new StringBuffer();
        for(int i = 0; i < str.length(); i++) {
            char c = str.charAt(i);
            switch(c) {
                case '&': buf.append("&amp;"); break;
                case '<': buf.append("&lt;"); break;
                case '>': buf.append("&gt;"); break;
                default : buf.append(c);
            }
        }
        
        return buf.toString();
    }

    
    public static int getIndexOfConstraintThingy(final String guard) {
        boolean isString = false;
        for(int i = 0; i < guard.length(); i++) {
            final Token.Operator op = Token.Operator.getOperator("" + guard.charAt(i));
            if(op == Token.Operator.STRING_SEPARATOR
                            && ((i < 1) || op != Token.Operator.ESCAPE)) {
                isString = !isString;
            } else if(!isString && op == Token.Operator.CONSTRAINT_SEPARATOR) {
                return i+1;
            }
        }
        return -1;
    }
    
    
    public static String simplifyGuard(String guard) {
        final Map<String, String> strings = Pool.getPool().getMap();
        guard = PNModel.removeStringLiterals(guard, strings);
        final int pos = PNModel.getIndexOfConstraintThingy(guard);
        final String constraints;
        if(pos >= 0) {
            constraints = guard.substring(0, pos);
            guard = guard.substring(pos);
        } else {
            constraints = "";
        }
        for(int i = 0; i < PNModel.SINGLE_REPLACE.length; i++) {
            guard = PNModel.SINGLE_REPLACE[i].matcher(guard).replaceAll(
                            PNModel.SINGLE_REPL[i]);
        }
        guard = PNModel.reinsertStringLiterals(guard, strings);
        final RPN rpn = RPNComputer.infixToRPN(guard);
        final Node tree = TreeComputer.rpnToTreeWithDeletion(rpn);
        TreeComputer.optimize(tree, null);
        guard = tree.toString();
        if(pos >= 0) {
            guard = constraints + " " + guard;
        }
//        System.out.println(guard);
        return guard;
    }
    
    
    public static String correctGuard(String guard) {
        final Map<String, String> strings = Pool.getPool().getMap();
        guard = PNModel.removeStringLiterals(guard, strings);
        final int pos = PNModel.getIndexOfConstraintThingy(guard);
        final String constraints;
        if(pos >= 0) {
            constraints = guard.substring(0, pos);
            guard = guard.substring(pos);
        } else {
            constraints = "";
        }
        for(int i = 0; i < PNModel.SINGLE_REPLACE.length; i++) {
            guard = PNModel.SINGLE_REPLACE[i].matcher(guard).replaceAll(
                            PNModel.SINGLE_REPL[i]);
        }
        guard = PNModel.reinsertStringLiterals(guard, strings);
        if(pos >= 0) {
            guard = constraints + " " + guard;
        }
//        System.out.println(guard);
        return guard;
    }
    
    
    public static String cleanup(String pep) {
        final Map<String, String> strings = Pool.getPool().getMap();
        pep = PNModel.removeStringLiterals(pep, strings);
        
        for(int i = 0; i < PNModel.SINGLE_REPLACE.length; i++) {
            pep = PNModel.SINGLE_REPLACE[i].matcher(pep).replaceAll(
                            PNModel.SINGLE_REPL[i]);
        }
        pep = PNModel.REMOVE.matcher(pep).replaceAll("");
        
        pep = PNModel.reinsertStringLiterals(pep, strings);
        
        
        return pep;
    }
    

    public static String removeStringLiterals(String str, final Map<String, String> strings) {
        strings.clear();
        final Matcher match = PNModel.STRING_LITERAL.matcher("");
        int counter = 1;
        int start = 0;
        while(match.reset(str).find(start)) {
            final String literal = str.substring(match.start(), match.end());
            if(PNModel.ALLOWED_LITERALS.matcher(literal).matches()) {
                start = match.end();
            } else {
                String replacement = strings.get(literal);
                if(replacement == null) {
                    replacement = "\"" + counter + "\"";
                    strings.put(literal, replacement);
                }
                str = PNModel.replaceAll(str, literal, replacement);
                start = match.start() + replacement.length();
                counter++;
            }
        }
        return str;
    }
    
    
    public static String reinsertStringLiterals(String str, final Map<String, String> strings) {
        for(final Map.Entry<String, String> e : strings.entrySet()) {
            str = PNModel.replaceAll(str, e.getValue(), e.getKey());
        }
        strings.clear();
        return str;
    }
    
    
    public static String escape(String s) {
        s = PNModel.replaceAll(s, "\\", "\\\\");
        s = PNModel.replaceAll(s, "\n", "\\n");
        s = PNModel.replaceAll(s, "\t", "\\t");
        return s;
    }
    
    
    public static String unescape(String s) {
        s = PNModel.replaceAll(s, "\\t", "\t");
        s = PNModel.replaceAll(s, "\\n", "\n");
        s = PNModel.replaceAll(s, "\\\\", "\\");
        return s;
    }
    
    
    public static String replaceAll(final String s, final String key, final String replace) {
        final StringBuffer r = new StringBuffer();
        int from = 0;
        int to = s.indexOf(key, from);
        while((to < s.length()) && (to >= 0)) {
            r.append(s.substring(from, to));
            r.append(replace);
            to += key.length();
            from = to;
            to = s.indexOf(key, to);
        }
        r.append(s.substring(from));
        return r.toString();
    }
    
    public static boolean setEquals(final ValueList a, final ValueList b) {
        return a.equals(b);
    }
    

    public static boolean setEquals(final StringList a, final StringList b) {
        return a.equals(b);
    }

    
    public static <T> boolean setEquals(final Collection<T> a, final List<T> b) {
        if(a.size() != b.size()) {
            return false;
        }
        boolean[] used = new boolean[b.size()];
        for(final T o : a) {
            boolean matched = false;
            for(int k = 0; k < b.size(); k++) {
                if(!used[k] && o.equals(b.get(k))) {
                    used[k] = true;
                    matched = true;
                    break;
                }
            }
            if(!matched) {
                return false;
            }
        }
        return true;
    }
    
    
    public static void deleteDublicates(final ValueList list) {
        list.deleteDublicates();
    }
    
    
    public static int countOccurrences(final ValueList c, Value o) {
        return c.count(o);
    }
    
    public static <T> int countOccurrences(final Collection<T> c, T o) {
        int count = 0;
        for(final T t : c) {
            if(t.equals(o)) {
                ++count;
            }
        }
        return count;
    }
}
