/*
 * Project:  UML
 * Filename: Arc.java
 * Created:  24.04.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.net;

/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class Arc {  
    public static enum Type {
        RESET(new Label(Value.V_ANY)),
        INHIBITOR(new Label(Value.V_ANY)),
        READ(new Label(Value.V_DOT)),
        PT(new Label(Value.V_DOT)),
        TP(new Label(Value.V_DOT));
        
        private final Label defaultLabel;
        
        Type(final Label defaultLabel) {
            this.defaultLabel = defaultLabel;
        }
        
        public Label getDefaultLabel() {
            return new Label(this.defaultLabel);
        }
    };
    
    private final Transition trans;
    private final Place place;
    private final Type type;
    private Label label;

    public static Arc getArc(final Place from, final Transition to, final String label) {
        return Arc.getArc(from, to, Arc.getLabel(label));
    }
    
    public static Arc getArc(final Transition from, final Place to, final String label) {
        return Arc.getArc(from, to, Arc.getLabel(label));
    }
    
    public static Arc getArc(final Transition trans, final Place place, final Label label, final Arc.Type type) {
        return new Arc(trans, place, type, label);
    }
    
    public static Arc getArc(final Place from, final Transition to, final Tuple label) {
        return Arc.getArc(from, to, Arc.getLabel(label));
    }
    
    public static Arc getArc(final Transition from, final Place to, final Tuple label) {
        return Arc.getArc(from, to, Arc.getLabel(label));
    }
    
    public static Arc getArc(final Place from, final Transition to, final Label label) {
        return Arc.getArc(to, from, label, Arc.Type.PT);
    }
    
    public static Arc getArc(final Transition from, final Place to, final Label label) {
        return Arc.getArc(from, to, label, Arc.Type.TP);
    }
    
    public static Arc getResetArc(final Place from, final Transition to, final Label label) {
        return Arc.getArc(to, from, label, Arc.Type.RESET);
    }
    
    public static Arc getInhibitorArc(final Place from, final Transition to, final Label label) {
        return Arc.getArc(to, from, label, Arc.Type.INHIBITOR);
    }
    
    private Arc(final Transition trans, final Place place, final Type type, final Label label) {
        this.trans = trans;
        this.place = place;
        this.type = type;
        this.setLabel(label);
        this.typeCheck(true);
    }

    
    public boolean transIsSource() {
        return this.getType() == Arc.Type.TP;
    }
    
    
    public Arc.Type getType() {
        return this.type;
    }
    
    
    public boolean isSimpleArc() {
        return (this.getType() == Arc.Type.TP) || (this.getType() == Arc.Type.PT);
    }
    
    
    public boolean isReadArc() {
        return this.getType() == Arc.Type.READ;
    }
    
    
    public boolean isResetArc() {
        return this.getType() == Arc.Type.RESET;
    }
    
    
    public boolean isInhibitorArc() {
        return this.getType() == Arc.Type.INHIBITOR;
    }
    
    
    public boolean typeCheck(final boolean out) {
        if(!this.isSimpleArc() && !this.isReadArc()) {
            return true;
        }
        for(final Value value : this.getLabel().getValues()) {
            if(!this.getPlace().getType().isValue(value.getNetValue())) {
                if(out && this.transIsSource()) {
                    System.err.println("Warning: Token " + value
                            + " cannot be taken from\n" + this.place + ".\nIt "
                            + "is incompatible with this place's type.\n"
                            + this.trans + " will be dead. This error was "
                            + "caused by\n" + this + "\n");
                } else if(out) {
                    System.err.println("Warning: Token " + value
                            + " cannot be added to\n" + this.place + ".\nIt is "
                            + "incompatible with this place's type.\n"
                            + this.trans + " will be dead. This error was "
                            + "caused by\n" + this + "\n");
                }
                return false;
            }
        }
        return true;
    }
    
    public static Label getLabel(final String label) {
        return (label == null) || (label.length() == 0)? null : new Label(label);
    }
    
    
    public static Label getLabel(final Tuple label) {
        return (label == null) || (label.size() == 0) ? null : new Label(label);
    }
    
    
    public void setLabel(final String label) {
        this.setLabel(Arc.getLabel(label));
    }
    
    public void setLabel(final Label label) {
        if((label == null) || label.isEmpty()) {
            this.label = new Label(this.getType().getDefaultLabel());
        } else {
            this.label = label;
        }
    }
    
    
    public void setLabel(final Tuple label) {
        this.setLabel(Arc.getLabel(label));
    }
    
    
    public Transition getTransition() {
        return this.trans;
    }
    
    
    public Place getPlace() {
        return this.place;
    }
    
    
    /**
     * @return Returns the label.
     */
    public Label getLabel() {
        if(this.label == null) {
            this.label = this.getType().getDefaultLabel();
        }
        return this.label;
    }
    
    
    public Object getSource() {
        if(this.transIsSource()) {
            return this.getTransition();
        } else {
            return this.getPlace();
        }
    }
    
    
    public Object getTarget() {
        if(this.transIsSource()) {
            return this.getPlace();
        } else {
            return this.getTransition();
        }
    }
    
    
    public void clear() {
    }
    
    
    @Override
    public int hashCode() {
        return (this.getPlace().hashCode() ^ this.getTransition().hashCode()
                ^ this.getLabel().hashCode()) * (this.transIsSource() ? 1 : -1);
    }
    
    
    @Override
    public boolean equals(final Object o) {
        return (o instanceof Arc) && this.equals((Arc)o);
    }
    
    
    public boolean equals(final Arc arc) {
        return (arc == this) || ((arc != null)
                        && (this.getType() == arc.getType())
                        && (this.getTransition() == arc.getTransition())
                        && (this.getPlace() == arc.getPlace())
                        && (this.getLabel().equals(arc.getLabel())));
    }
    
    
    @Override
    public String toString() {
        final String source;
        final String target;
        if(this.transIsSource()) {
            source = "Transition(" + this.getTransition().getName() + ")";
            target = "Place(" + this.getPlace().getName() + ")";
        } else {
            source = "Place(" + this.getPlace().getName() + ")";
            target = "Transition(" + this.getTransition().getName() + ")";
        }
        
        return "Arc(\n" 
        + "  source = " + source + ",\n"
        + "  target = " + target + ",\n"
        + "  type   = " + this.getType() + ",\n"
        + "  label  = " + this.getLabel() + "\n"
        + ")";
    }
}
