/*
 * Project:  UML
 * Filename: PetriNet.java
 * Created:  24.04.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.net;

import java.util.Collection;
import java.util.Map;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import petruchio.sim.petrinettool.IPetriNet;
import petruchio.sim.petrinettool.IMultiSet;
import petruchio.sim.petrinettool.IPlace;
import petruchio.sim.petrinettool.IPlaceType;
import petruchio.sim.petrinettool.IPlaceStatus;
import petruchio.sim.petrinettool.ITransition;

import petruchio.sim.pnmodel.PNModel;
import petruchio.sim.pnmodel.PEPWriter;
import petruchio.sim.pnmodel.PNMLWriter;
import petruchio.sim.pnmodel.util.Node;
import petruchio.sim.pnmodel.util.Pool;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class PetriNet implements IPetriNet {

    public static final int SYNCHRONIZATION_LIMIT = 1000;

    private static int counter = 0;

    private String netName;

    private List<Arc> arcs;

    private Map<String, Place> places;

    private Map<String, Transition> transitions;
    
    private List<IPlace> entryPlaces;
    
    private List<IPlace> exitPlaces;
    
    private List<String> functions;
    
    private int minX;
    
    private int maxX;
    
    private int minY;
    
    private int maxY;
    
    private boolean isWidthCorrect;
    
    private boolean isHeightCorrect;
    
    private boolean isLowLevelNet;
    
    private boolean checkPlaceValues;
    
    private int elemCount; 
    
    private PNModel tool;
    

    public PetriNet(final PNModel tool) {
        this.tool = tool;
        this.netName = "_net" + (++PetriNet.counter);
        this.arcs = Pool.getPool().getList();
        this.places = Pool.getPool().getMap();
        this.transitions = Pool.getPool().getMap();
        this.entryPlaces = Pool.getPool().getList();
        this.exitPlaces = Pool.getPool().getList();
        this.functions = Pool.getPool().getList();
        this.isHeightCorrect = true;
        this.isWidthCorrect = true;
        this.isLowLevelNet = false;
        this.elemCount = 0;
        this.minX = 0;
        this.minY = 0;
        this.maxX = 0;
        this.maxY = 0;
    }

    
//    private void scanPlaces() {
//        this.entryPlaces.clear();
//        this.exitPlaces.clear();
//        Iterator it = this.getPlaceIterator();
//        while(it.hasNext()) {
//            Place p = (Place)it.next();
//            this.addToStatusList(p, p.getStatus());
//        }
//    }
    
    
    public boolean usesTime() {
        for(final Transition trans : this.getTransitionCollection()) {
            if(!trans.isImmediate()) {
                return true;
            }
        }
        return false;
    }
    
    
    public boolean usesPriority() {
        for(final Transition trans : this.getTransitionCollection()) {
            if(trans.hasPriority()) {
                return true;
            }
        }
        return false;
    }
    
    
    public boolean usesWeights() {
        for(final Transition trans : this.getTransitionCollection()) {
            if(trans.hasWeight()) {
                return true;
            }
        }
        return false;
    }
    
    
    public boolean hasInhibitorArcs() {
        for(final Arc arc : this.arcs) {
            if(arc.isInhibitorArc()) {
                return true;
            }
        }
        return false;
    }
    
    
    public boolean hasResetArcs() {
        for(final Arc arc : this.arcs) {
            if(arc.isResetArc()) {
                return true;
            }
        }
        return false;
    }
    
    
    public boolean hasReadArcs() {
        for(final Arc arc : this.arcs) {
            if(arc.isReadArc()) {
                return true;
            }
        }
        return false;
    }
    
    
    public boolean usesCapacity() {
        for(final Place place : this.places.values()) {
            if(!place.infiniteCapacity()) {
                return true;
            }
        }
        return false;
    }
    

    public boolean isLowLevelNet() {
        return this.isLowLevelNet;
    }
    
    
    public void setIsLowLevelNet(boolean value) {
        this.isLowLevelNet = value;
    }
    
    
    public String getName() {
    	return this.netName;
    }
    

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#clone()
     */
    public PetriNet clone() {
        PetriNet net;
        try {
            net = (PetriNet)super.clone();
        } catch(final CloneNotSupportedException e) {
            throw new InternalError();
        }
        net.tool = this.tool;
        net.isLowLevelNet = this.isLowLevelNet;
        net.netName = "_net" + (++PetriNet.counter);
        net.arcs = Pool.getPool().getList();
        net.places = Pool.getPool().getMap();
        net.transitions = Pool.getPool().getMap();
        net.entryPlaces = Pool.getPool().getList();
        net.exitPlaces = Pool.getPool().getList();
        net.functions = Pool.getPool().getList();
        net.isHeightCorrect = true;
        net.isWidthCorrect = true;
        net.isLowLevelNet = false;
        net.elemCount = 0;
        net.minX = 0;
        net.minY = 0;
        net.maxX = 0;
        net.maxY = 0;
        for(final Place p : this.getPlaces()) {
            final IPlace newPlace = net.createPlace(p.getName(), p.getStatus(), p.getType(),
                            p.getInitialMarking(), p.getCurrentMarking());
            newPlace.setPos(p.getPosX(), p.getPosY());
        }
        
        for(final Transition t : this.getTransitions()) {
            final ITransition newTrans = net.createTransition(t.getName(),
                            (IMultiSet) t.getActions().clone(), t.getGuard());
            
            newTrans.setPos(t.getPosX(), t.getPosY());
        }

        for(final Arc arc : this.getArcs()) {
            final IPlace p = net.getPlaceByName(arc.getPlace().getName());
            final ITransition t = net.getTransitionByName(arc.getTransition().getName());
            if (arc.transIsSource()) {
                net.createArcNoCheck(t, p, arc.getLabel());
            } else {
                net.createArcNoCheck(p, t, arc.getLabel());
            }
        }
        return net;
    }
    

    public Place createPlace(final String name, final IPlaceStatus status,
                    final IPlaceType type, final String initmarking) {
        return this.createPlace(name, status, type, initmarking, initmarking);
    }
    
    
    public Place createPlace(final String name, final IPlaceStatus status,
                    final IPlaceType type, final String initmarking,
                    final String curmarking) {
        return this.createPlace(name, status, type, false, initmarking, curmarking);
    }
    
    
    public Place createPlace(final String name, final IPlaceStatus status,
                    final IPlaceType type, final boolean allowAnyInput,
                    final String initmarking, final String curmarking) {
        if (this.places.containsKey(name)) {
            throw new RuntimeException("A place with name \"" + name + "\" "
                            + "already exists.");
        }
        if(this.transitions.containsKey(name)) {
            System.err.println("WARNING: You're creating a place with name \""
                    + name + "\".\n"
                    + "  A transition with that name already exists.\n"
                    + "  This may cause problems in some programs (f.e. PEP).");
        }
        
        final Place p = new Place(name, status, type, allowAnyInput, initmarking,
                        curmarking, this);
        this.places.put(p.getName(), p);
        this.addToStatusList(p, status);
        this.elemCount++;
        return p;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#createPlace(int, int, java.lang.String)
     */
    public Place createPlace(final IPlaceStatus status, final IPlaceType type,
                    final String initmarking) {
        String name;
        do {
            name = Place.freshName();
        } while (this.places.containsKey(name));
        return this.createPlace(name, status, type, initmarking);
    }
    
    public Place createPlace(final IPlaceStatus status, final IPlaceType type,
                    final String initmarking, final String curmarking) {
        String name;
        do {
            name = Place.freshName();
        } while (this.places.containsKey(name));
        return this.createPlace(name, status, type, initmarking, curmarking);
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#createTransition(String,
     *      petruchio.sim.petrinettool.MultiSet, String)
     */
    public Transition createTransition(final String name,
                    final IMultiSet actionEvents, final String guard) {
        if (this.transitions.containsKey(name)) {
            throw new RuntimeException("A transition with name \"" + name
                            + "\" already exists.");
        }
        if(this.places.containsKey(name)) {
            System.err.println("WARNING: You're creating a transition with name \""
                    + name + "\".\n"
                    + "  A place with that name already exists.\n"
                    + "  This may cause problems in some programs (f.e. PEP).");
        }
        final Transition t = new Transition(name, actionEvents, guard, this);
        this.transitions.put(name, t);
        this.elemCount++;
        return t;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#createTransition(petrinettool.MultiSet,
     *      String)
     */
    public Transition createTransition(final IMultiSet actionEvents, final String guard) {
        String name;
        do {
            name = Transition.freshName();
        } while (this.transitions.containsKey(name));
        return this.createTransition(name, actionEvents, guard);
    }

    
    private void checkArcParams(final IPlace place, final ITransition trans) {
        if(place == null) {
            throw new RuntimeException("Cannot create arc with place NULL.");
        } else if(trans == null) {
            throw new RuntimeException("Cannot create arc with transition NULL.");
        } else if(!PNModel.containsThisObject(this.places.values(), place)) {
            throw new RuntimeException("Cannot create arc with a place that is"
                            + " not part of this net.");
        } if(!PNModel.containsThisObject(this.transitions.values(), trans)) {
            throw new RuntimeException("Cannot create arc with a transition "
                            + "that is not part of this net.");
        }
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#createArc(petrinettool.Place,
     *      petruchio.sim.petrinettool.Transition, String)
     */
    public void createArc(final IPlace from, final ITransition to, final Label label) {       
        this.checkArcParams(from, to);
        this.createArcNoCheck(from, to, label);
    }
    
    
    public void createArcNoCheck(final IPlace from, final ITransition to, final Label label) {       
        final Arc arc = Arc.getArc((Place) from, (Transition) to, new Label(label));
        ((Place) from).addOutput(arc);
        ((Transition) to).addInput(arc);
        this.arcs.add(arc);
    }
    
    
    public void createArc(final ITransition trans, final IPlace place, final Label label, final Arc.Type type) {       
        this.checkArcParams(place, trans);
        this.createArcNoCheck(trans, place, label, type);
    }
    
    
    public void createArcNoCheck(final ITransition trans, final IPlace place, String label, final Arc.Type type) {
        if(((type == Arc.Type.INHIBITOR) || (type == Arc.Type.RESET))
                        && ((label == null) || (label.length() == 0))) {
            label = IPlace.STRING_ANY;
        }
        this.createArcNoCheck(trans, place, new Label(label), type);
    }
    
    
    public void createArcNoCheck(final ITransition trans, final IPlace place, final Label label, final Arc.Type type) {       
        final Arc arc = Arc.getArc((Transition) trans, (Place) place, label, type);
        if(arc.transIsSource()) {
            ((Transition) trans).addOutput(arc);
            ((Place) place).addInput(arc);
        } else {
            ((Place) place).addOutput(arc);
            ((Transition) trans).addInput(arc);
        }
        this.arcs.add(arc);
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#createArc(petrinettool.Place,
     *      petruchio.sim.petrinettool.Transition, String)
     */
    public void createArc(final IPlace from, final ITransition to, final Tuple tuple) {
        this.checkArcParams(from, to);
        this.createArcNoCheck(from, to, tuple);
    }
    
    public void createArcNoCheck(final IPlace from, final ITransition to, final Tuple tuple) {
        final Arc arc = Arc.getArc((Place) from, (Transition) to, tuple);
        ((Place) from).addOutput(arc);
        ((Transition) to).addInput(arc);
        this.arcs.add(arc);
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#createArc(petrinettool.Place,
     *      petruchio.sim.petrinettool.Transition, String)
     */
    public void createArc(final IPlace from, final ITransition to, final String label) {
        this.checkArcParams(from, to);
        this.createArcNoCheck(from, to, label);
    }
    
    
    public void createArcNoCheck(final IPlace from, final ITransition to, final String label) {
        final Arc arc = Arc.getArc((Place) from, (Transition) to, label);
        ((Place) from).addOutput(arc);
        ((Transition) to).addInput(arc);
        this.arcs.add(arc);
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#createArc(petrinettool.Transition,
     *      petruchio.sim.petrinettool.Place, String)
     */
    public void createArc(final ITransition from, final IPlace to, final Label label) {
        this.checkArcParams(to, from);
        this.createArcNoCheck(from, to, label);
    }
    
    
    public void createArcNoCheck(final ITransition from, final IPlace to, final Label label) {
        final Arc arc = Arc.getArc((Transition) from, (Place) to, new Label(label));
        ((Transition) from).addOutput(arc);
        ((Place) to).addInput(arc);
        this.arcs.add(arc);
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#createArc(petrinettool.Transition,
     *      petruchio.sim.petrinettool.Place, String)
     */
    public void createArc(final ITransition from, final IPlace to, final Tuple tuple) {
        this.checkArcParams(to, from);
        this.createArcNoCheck(from, to, tuple);
    }
    
    
    public void createArcNoCheck(ITransition from, IPlace to, Tuple tuple) {
        final Arc arc = Arc.getArc((Transition) from, (Place) to, tuple);
        ((Transition) from).addOutput(arc);
        ((Place) to).addInput(arc);
        this.arcs.add(arc);
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#createArc(petrinettool.Transition,
     *      petruchio.sim.petrinettool.Place, String)
     */
    public void createArc(final ITransition from, final IPlace to, final String label) {
        this.checkArcParams(to, from);
        this.createArcNoCheck(from, to, label);
    }
    
    
    public void createInhibitorArc(final ITransition from, final IPlace to, final String label) {
        this.checkArcParams(to, from);
        this.createArcNoCheck(from, to, label, Arc.Type.INHIBITOR);
    }
    
    
    public void createResetArc(final ITransition from, final IPlace to, final String label) {
        this.checkArcParams(to, from);
        this.createArcNoCheck(from, to, label, Arc.Type.RESET);
    }
    
    
    public void createReadArc(final ITransition from, final IPlace to, final String label) {
        this.checkArcParams(to, from);
        this.createArcNoCheck(from, to, label, Arc.Type.READ);
    }
    
    
    public void createArcNoCheck(final ITransition from, final IPlace to, final String label) {
        final Arc arc = Arc.getArc((Transition) from, (Place) to, label);
        ((Transition) from).addOutput(arc);
        ((Place) to).addInput(arc);
        this.arcs.add(arc);
    }


    public Iterator<Place> getPlaceIterator() {
        return this.places.values().iterator();
    }


    public Iterator<Transition> getTransitionIterator() {
        return this.transitions.values().iterator();
    }
    

    public Iterator<Arc> getArcIterator() {
        return this.arcs.iterator();
    }
    
    
    public Collection<Arc> getArcs() {
        return this.arcs;
    }
    

    public void mergeMultipleArcs(final ITransition trans) {
        final Transition t = (Transition) trans;
        this.mergeMultipleArcs(t.getInput());
        this.mergeMultipleArcs(t.getOutput());
    }


    public void mergeMultipleArcs(final IPlace place) {
        final Place p = (Place) place;
        this.mergeMultipleArcs(p.getInput());
        this.mergeMultipleArcs(p.getOutput());
    }


    private void mergeMultipleArcs(final List<Arc> curArcs) {
        boolean merged;
        do {
            merged = false;
            for (int i = 0; !merged && (i < curArcs.size()); i++) {
                final Arc firstArc = curArcs.get(i);
                for (int j = i + 1; !merged && (j < curArcs.size()); j++) {
                    final Arc secondArc = curArcs.get(j);
                    if((firstArc.getType() == secondArc.getType())
                           && (firstArc.getSource() == secondArc.getSource())
                           && (firstArc.getTarget() == secondArc.getTarget())) {
                        // we have to merge these two

                        // instead of creating a new arc and removing the old
                        // two, we simply change the label of one arc and
                        // remove the other one.
                        firstArc.getLabel().add(secondArc.getLabel());
                        this.removeArc(secondArc);

                        merged = true;
                    }
                }
            }
        } while (merged);
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#insert(petrinettool.PetriNet)
     */
    public void insert(final petruchio.sim.petrinettool.IPetriNet netB) {
        this.insert(netB, 0, 0);
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#insert(petrinettool.PetriNet, int, int)
     */
    public void insert(petruchio.sim.petrinettool.IPetriNet netB, final int offsetX, final int offsetY) {
        if (netB == this) {
            netB = this.clone();
        } else if(netB == null) {
            throw new RuntimeException("Cannot insert NULL.");            
        }
       
        PetriNet net = (PetriNet) netB;
        
        this.setIsLowLevelNet(this.isLowLevelNet() && netB.isLowLevelNet());

        final Map<String, String> placeNames = Pool.getPool().getMap();
        final Map<String, String> transitionNames = Pool.getPool().getMap();
        
        for(final Place p : net.getPlaces()) {
            String name = p.getName();
            int i = 1;
            while (this.containsPlace(name)) {
                name = p.getName() + ++i;
            }
            final IPlace newPlace = this.createPlace(name, p.getStatus(), p.getType(),
                            p.getInitialMarking(), p.getCurrentMarking());
            newPlace.setPos(p.getPosX() + offsetX, p.getPosY() + offsetY);
            placeNames.put(p.getName(), name);
        }
        
        for(final Transition t : net.getTransitions()) {
            String name = t.getName();
            int i = 1;
            while (this.containsTransition(name)) {
                name = t.getName() + ++i;
            }
            final ITransition newTrans = this.createTransition(name,
                            (IMultiSet) t.getActions().clone(), t.getGuard());
            
            newTrans.setPos(t.getPosX() + offsetX, t.getPosY() + offsetY);
            transitionNames.put(t.getName(), name);
        }

        for(final Arc arc : net.getArcs()) {
            final String place = placeNames.get(arc.getPlace().getName());
            final String trans = transitionNames.get(arc.getTransition().getName());
            final IPlace p = this.getPlaceByName(place);
            final ITransition t = this.getTransitionByName(trans);
            if (arc.transIsSource()) {
                this.createArcNoCheck(t, p, arc.getLabel());
            } else {
                this.createArcNoCheck(p, t, arc.getLabel());
            }
        }
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#mergePlaces(petrinettool.Place,
     *      petruchio.sim.petrinettool.Place)
     */
    public Place mergePlaces(final IPlace placeA, final IPlace placeB) {
        if(placeA == null) {
            throw new RuntimeException("Cannot merge place (A) NULL.");
        } else if(placeB == null) {
            throw new RuntimeException("Cannot merge place (B) NULL.");
        } else if(placeA == placeB) {
            throw new RuntimeException("Cannot merge places that are the same.");
        } else if(!this.places.containsValue(placeA)) {
            throw new RuntimeException("Cannot merge place (A) is not part "
                            + "of this net.");
        } else if(!this.places.containsValue(placeB)) {
            throw new RuntimeException("Cannot merge place (B) is not part "
                            + "of this net.");
        }
        
        final Place pA = (Place) placeA;
        final Place pB = (Place) placeB;

        if (!pA.getType().equals(pB.getType())) {
            throw new RuntimeException("Cannot merge places "
                            + "with different types!\n" + pA.getType() + " != "
                            + pB.getType());
        } else if (!((pA.getStatus().equals(IPlace.STATUS_ENTRY))
                        && (pB.getStatus().equals(IPlace.STATUS_EXIT)))
                        && !((pA.getStatus().equals(IPlace.STATUS_EXIT))
                        && (pB.getStatus().equals(IPlace.STATUS_ENTRY)))) {
            throw new RuntimeException(
                            "Cannot merge places. "
                          + "One place has to have status '" + IPlace.STATUS_EXIT
                          + "' and the other one has to have status '"
                          + IPlace.STATUS_ENTRY + "'.\n" + pA + "\n" + pB);
        }

        final String name = pA.getName() + "+" + pB.getName();
        String newName = name;
        while ((this.containsPlace(newName))) {
            newName = Place.freshName() + "(" + name + ")";
        }

        final IPlaceStatus status = IPlace.STATUS_INTERNAL;
        final IPlaceType typeA = pA.getType();
        final IPlaceType typeB = pB.getType();
        
        petruchio.sim.petrinettool.type.Set type = new petruchio.sim.petrinettool.type.Set();
        type.addType(typeA);
        type.addType(typeB);
        
        final String pAcM = pA.getCurrentMarking();
        final String pBcM = pB.getCurrentMarking();
        final String currentMarking;
        if (pAcM.equals("")) {
            currentMarking = pBcM;
        } else if (pBcM.equals("")) {
            currentMarking = pAcM;
        } else {
            currentMarking = pAcM + ", " + pBcM;
        }
        
        
        final String pAiM = pA.getInitialMarking();
        final String pBiM = pB.getInitialMarking();
        final String initialMarking;
        if (pAiM.equals("")) {
            initialMarking = pBiM;
        } else if (pBiM.equals("")) {
            initialMarking = pAiM;
        } else {
            initialMarking = pAiM + ", " + pBiM;
        }

        final Place newPlace = this.createPlace(newName, status, type, initialMarking, currentMarking);
        newPlace.setPosX((pA.getPosX() + pB.getPosX()) / 2);
        newPlace.setPosY((pA.getPosY() + pB.getPosY()) / 2);

        for(final Arc a : pA.getInput()) {
            this.createArc(a.getTransition(), newPlace, a.getLabel());
        }
        for(final Arc a :pB.getInput()) {
            this.createArc(a.getTransition(), newPlace, a.getLabel());
        }

        for(final Arc a : pA.getOutput()) {
            this.createArc(newPlace, a.getTransition(), a.getLabel());
        }
        for(final Arc a : pB.getOutput()) {
            this.createArc(newPlace, a.getTransition(), a.getLabel());
        }

        this.removePlace(placeB);
        this.removePlace(placeA);
        
        this.mergeMultipleArcs(newPlace);

        return newPlace;
    }


    public void restrict(String action) {
        final Transition[] trans = new Transition[this.transitions.values().size()];
        this.transitions.values().toArray(trans);
        for (int i = 0; i < trans.length; i++) {
            if (((MultiSet)trans[i].getActions()).getAction(action,
                            Action.Type.ACTION) != null) {
                this.removeTransition(trans[i]);
            }
        }
    }


    public Transition synchronize(final SynchronizationTask task) {
        final Transition transA = task.getTransitionA();
        final Transition transB = task.getTransitionB();
        final Action actionA = task.getActionA();
        final Action actionB = task.getActionB();

        final Tuple labelA = actionA.getLabel();
        final Tuple labelB = actionB.getLabel();

        // create the multiset of actions
        // newSet = oldSetA + oldSetB - { action+, action- }
        final MultiSet newActions = new MultiSet();
        newActions.addMultiSet((MultiSet)transA.getActions());
        newActions.addMultiSet((MultiSet)transB.getActions());
        newActions.removeAction(newActions.getAction(actionA));
        newActions.removeAction(newActions.getAction(actionB));

        String newGuard;
        final String guardA = transA.getGuard(); 
        final String guardB = transB.getGuard();
        if (PNModel.isTrue(guardA) || guardA.equals(guardB)) {
            newGuard = guardB;
        } else if (PNModel.isTrue(guardB)) {
            newGuard = guardA;
        } else {
            newGuard = "(" + transA.getGuard() + " and " + transB.getGuard()
                            + ")";
        }

        for (int i = 0; i < labelA.size(); i++) {
            final String tokenA = labelA.get(i);
            final String tokenB = labelB.get(i);
            if (tokenA.equals(tokenB)) {
                // skip expressions that are true
                continue;
            } else if(PNModel.isValue(tokenA) && PNModel.isValue(tokenB)) {
                // if the tokens are _different_ values the transition would
                // be dead. Do not create it!
                return null;
            }
            
            final String addition = "(" + tokenA + "=" + tokenB + ")";
            if ((i > 0) || !PNModel.isTrue(newGuard)) {
                newGuard = "(" + newGuard + " and " + addition + ")";
            } else {
                newGuard = addition;
            }
        }

        if (false) { // very neat test...
            // if the transition to create would be dead:
            // do not create it.
            return null;
        }
        

        final Transition newTrans = this.createTransition(newActions, newGuard);

        // copy all input arcs
        for(Arc arc : transA.getInput()) {
            this.createArc(arc.getPlace(), newTrans, arc.getLabel());
        }
        for(final Arc arc : transB.getInput()) {
            this.createArc(arc.getPlace(), newTrans, arc.getLabel());
        }

        // copy all output arcs
        for(final Arc arc : transA.getOutput()) {
            this.createArc(newTrans, arc.getPlace(), arc.getLabel());
        }
        for(final Arc arc : transB.getOutput()) {
            this.createArc(newTrans, arc.getPlace(), arc.getLabel());
        }

        this.mergeMultipleArcs(newTrans);
        //System.err.println(newTrans);
        
        if(this.isDuplicateTransition(newTrans)) {
            //System.err.println("Duplicate transition:\n" + newTrans);
            this.removeTransition(newTrans);
            return null;
        }
        
        // a new transition has been created
        return newTrans;
    }


    public void synchronize(final String action) {
        final LinkedList<SynchronizationTask> tasks = new LinkedList<SynchronizationTask>();
        for(final Transition t : this.getTransitions()) {
            this.appendSynchronizationTasks(t, action, tasks);
        }

        // we don't want to loop forever...
        final int boundary = PetriNet.SYNCHRONIZATION_LIMIT + tasks.size();
        int count = 0;
        while(!tasks.isEmpty() && (tasks.size() < boundary)
                        && (count < boundary)) {
            ++count;
            //System.err.println(count + " done, " + tasks.size() + " pending");
            final SynchronizationTask task = tasks.getFirst();
            tasks.removeFirst();
            final ITransition trans = this.synchronize(task);
            //System.err.println("task\n" + task + " new trans\n" + trans + "\n");
            this.appendSynchronizationTasks(trans, action, tasks);
        }
        if(count >= boundary) {
            throw new RuntimeException("It seems to be an endless "
                            + "synchronization.\n"
                            + "Already synchronized " + count + " pairs of "
                            + "transitions on symbol \"" + action + "\" but "
                            + "still " + tasks.size() + " pairs pending.");
        }
    }


    private void appendSynchronizationTasks(ITransition trans, String action,
                    LinkedList<SynchronizationTask> tasks) {
        if (trans == null) {
            return;
        }
        final Transition transition = (Transition) trans;
        final List<Action> symbols = ((MultiSet)transition.getActions()).getActionSymbols();
        for(final Action curAction : symbols) {
            // this action has to be an "receiving" "action" with the
            // requested action symbol.
            if (curAction.isSending() || !curAction.getSymbol().equals(action)) {
                // skip this action
                continue;
            }

            // we have to know how many variables/values are associated with
            // this action.
            final int variableCount = curAction.getLabel().size();

            for(final Transition curTrans : this.getTransitions()) {
                final List<Action> symbols2 = ((MultiSet)curTrans.getActions()).getActionSymbols();
                for(final Action curTransAction : symbols2) {
                    // this action has to be "sending"
                    if (!curTransAction.isSending()
                               || !curTransAction.getSymbol().equals(action)
                               || (curTransAction.getLabel().size()
                                                      != variableCount)) {
                        // skip this action
                        continue;
                    }
                    // this action is an "action", it is "sending", it has the
                    // same symbol and the same number of variables/values
                    // as the first action "curAction".
                    // ==> we can synchronize. append the task to the list

                    final SynchronizationTask task = new SynchronizationTask(
                                    transition, curAction, curTrans,
                                    curTransAction);
                    tasks.addLast(task);
                    //System.out.println(task);
                }
            }
        }
    }
    
    
    public boolean isDuplicateTransition(final Transition transition) {
        for(final Transition trans : this.getTransitions()) {
            if((trans != transition) && transition.equals(trans)) {
                return true;
            }
        }
        
        return false;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#scopeTransitions(java.lang.String)
     */
    public void scope(final String action) {
        this.synchronize(action);
        this.restrict(action);
    }

    public void tie(final String tieSymbol, final IPlaceType type) {
        final List<Transition> trans = this.getTransitions(tieSymbol, Action.Type.TIE);
        this.tie(tieSymbol, type, trans);
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#tie(java.lang.String, int)
     */
    public void tie(final String tieSymbol, IPlaceType type, final Collection<Transition> trans) {
        if((trans == null) || (trans.size() == 0)) {
            return;
        }

        // naming is optional
        final String name = tieSymbol;
        String curname = name;
        for(int i = 0; this.places.containsKey(curname); i++) {
            curname = name + "_" + i;
        }
        
        boolean typeIsSet = (type != null);
        
        final IPlace tiePlace = this.createPlace(curname, IPlace.STATUS_INTERNAL,
                (typeIsSet ? type : IPlace.TYPE_BLACK_TOKEN), "");
        int x = 0;
        int y = 0;
        
        for(final Transition curTrans : trans) {
            x += curTrans.getPosX();
            y += curTrans.getPosY();
            final Action[] actions = curTrans.getTie(tieSymbol);
 
            for(int i = 0; i < actions.length; i++) {
                final Action action = actions[i];
            
                if(type == null) {
                    type = this.getTypeOf(action.getLabel());
                    ((Place)tiePlace).setType(type);
                } else if(!typeIsSet) {
                    final IPlaceType tempType = this.getTypeOf(action.getLabel());
                    if (!tempType.equals(type)) {
                        throw new RuntimeException("The tie symbol's signature \""
                                        + action + "\" isn't compatible to symbol "
                                        + tieSymbol + "\".\n  "
                                        + "Type of signature:      " + type + "\n"
                                        + "Type of current symbol: " + tempType);
                    }
                }
                
                String label = action.getLabel().toString();
                if(label.indexOf(',') < 0) {
                    label = label.substring(1, label.length()-1);
                }
                
                if(action.isSending()) {
                    this.createArc(curTrans, tiePlace, label);
                } else {
                    this.createArc(tiePlace, curTrans, label);
                }
                
                curTrans.removeAction(action);
            }
        }
        
        x /= trans.size();
        y /= trans.size();
        tiePlace.setPos(x, y);
    }
    
    
    public IPlaceType getTypeOf(final Tuple oldTuple) {
        if(oldTuple.size() == 0) {
            return IPlace.TYPE_BLACK_TOKEN;
        }
        final String tuple = PNModel.SIMPLIFY_TUPLE.matcher(oldTuple.toString()).replaceAll("");
        return this.getTypeOf(tuple);
    }
    
    private IPlaceType getTypeOf(String tuple) {
        tuple = tuple.substring(1, tuple.length()-1);
        if(tuple.length() == 0) {
            return this.tool.getAnyType();
        }
        
        petruchio.sim.petrinettool.type.Tuple type = new petruchio.sim.petrinettool.type.Tuple();
        final char[] elems = tuple.toCharArray();
        int i = 0;
        while(i < elems.length) {
            switch(elems[i]) {
                case '(':
                    int ctr = 1;
                    int pos = i+1;
                    for(; ctr >= 1; pos++) {
                        switch(elems[pos]) {
                            case '(' : ctr++; break;
                            case ')' : ctr--; break;
                        }
                    }
                    final String part = tuple.substring(i, pos);
                    type.addType(this.getTypeOf(part));
                    i = pos;
                    break;
                default:
                    type.addType(this.tool.getAnyType());
                    i++;
                    break;
            }
        }
        if(elems[elems.length-1] == ',') {
            type.addType(this.tool.getAnyType());
        }
        
        return type;
    }
    

    public List<Transition> getTransitions(final String symbol, final Action.Type type) {
        final List<Transition> result = Pool.getPool().getList();
        for(final Transition t : this.getTransitions()) {
            if (((MultiSet)t.getActions()).getAction(symbol, type) != null) {
                result.add(t);
            }
        }

        return result;
    }
    

    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#tieAll()
     */
    public void tieAll() {
        this.tieAll(".*");
    }
    
    public void tieAll(final String regexp) {
        final Matcher matcher = Pattern.compile(regexp).matcher("");
        final Map<String, List<Transition>> trans_symbol = Pool.getPool().getMap();
        for(final Transition t : this.getTransitions()) {
            final List<Action> ties = ((MultiSet)t.getActions()).getTieSymbols();
            for(final Action a : ties) {
                final String tie = a.getSymbol();
                matcher.reset(tie);
                if(!matcher.matches()) {
                    continue;
                }
                
                List<Transition> trans = trans_symbol.get(tie);
                if(trans == null) {
                    trans = Pool.getPool().getList();
                    trans.add(t);
                    trans_symbol.put(tie, trans);
                } else if(!PNModel.containsThisObject(trans, t)) {
                    trans.add(t);
                }
            }
        }
        for(final Map.Entry<String, List<Transition>> e : trans_symbol.entrySet()) {
            final String symbol = e.getKey();
            final List<Transition> trans = e.getValue();
            final IPlaceType type = this.tool.getTypeOf(symbol);
            this.tie(symbol, type, trans);
        }
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#scopeAll()
     */
    public void scopeAll() {
        final List<String> symbols = this.getAllSymbols(Action.Type.ACTION);
        for(int i = 0; i < symbols.size(); i++) {
            this.scope(symbols.get(i));
        }
    }


    public List<String> getAllSymbols(final Action.Type type) {
        final List<String> symbols = Pool.getPool().getList();

        for(final Transition t : this.getTransitions()) {
            final List<Action> actions;
            switch(type) {
                case ACTION:
                    actions = ((MultiSet)t.getActions()).getActionSymbols();
                    break;
                case TIE:
                    actions = ((MultiSet)t.getActions()).getTieSymbols();
                    break;
                default:
                    throw new RuntimeException("Type not supported yet.");
            }
            for(final Action a : actions) {
                final String symbol = a.getSymbol();
                if (!symbols.contains(symbol)) {
                    symbols.add(symbol);
                }
            }
        }

        return symbols;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#concatenate(petrinettool.PetriNet, int)
     */
    public void concatenate(final petruchio.sim.petrinettool.IPetriNet net, final int count) {
        if(net == null) {
            throw new RuntimeException("Cannot concatenate NULL.");
        }
        
        this.setIsLowLevelNet(this.isLowLevelNet() && net.isLowLevelNet());
        
        final List<IPlace> entries = Pool.getPool().getList(this.entryPlaces);
        final List<IPlace> exits = Pool.getPool().getList(this.exitPlaces);
        
        if(count > exits.size()) {
            throw new RuntimeException("This net has less than count (" 
                            + count + ") exit places: " + exits.size());
        } else if(count > net.getEntryPlaces().length) {
            throw new RuntimeException("The concatenated net has less than count (" 
                            + count + ") entry places: " + net.getEntryPlaces().length);
        }

        final int yOffset = this.getMaxY() - net.getMinY();
        final int xOffset = this.getMinX() - net.getMinX();
        this.insert(net, xOffset, yOffset);
        
        // since the references changed because of the insert-operation
        // the entry-places of "net" are calculated as
        // (all entry-places) minus (entry-places before inserting)
        final List<IPlace> temp = Pool.getPool().getList(this.entryPlaces);
        final List<IPlace> newEntries = PNModel.setMinus(temp, entries);
        
        int entryIndex = 0;
        int exitIndex = 0;
        for (int i = 0; i < count; i++) {
            final IPlace en = newEntries.get(entryIndex++);
            final IPlace ex = exits.get(exitIndex++);
            this.mergePlaces(en, ex);
        }
        
        // OBSOLETE
        //while (entryIndex < newEntries.size()) {
        //    Place p = (Place) newEntries.get(entryIndex++);
        //    p.setStatus(Place.STATUS_INTERNAL);
        //}
        //while (exitIndex < exits.size()) {
        //    Place p = (Place) exits.get(exitIndex++);
        //    p.setStatus(Place.STATUS_INTERNAL);
        //}
    }


    public List<Place> getPlaces(final IPlaceStatus status) {
        final List<Place> list = Pool.getPool().getList();
        for(final Place p : this.getPlaces()) {
            if (p.getStatus().equals(status)) {
                list.add(p);
            }
        }
        return list;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#containsTransition(java.lang.String)
     */
    public boolean containsTransition(final String trans) {
        return this.transitions.containsKey(trans);
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#containsPlace(java.lang.String)
     */
    public boolean containsPlace(final String place) {
        return this.places.containsKey(place);
    }


    public void removeArc(final Arc arc) {
        PNModel.removeThisObject(this.arcs, arc);

        arc.getTransition().removeArc(arc);
        arc.getPlace().removeArc(arc);
        arc.clear();
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#removePlace(petrinettool.Place)
     */
    public void removePlace(final IPlace place) {
        final Place p = (Place) place;
        if(place == null) {
            throw new RuntimeException("Cannot remove place NULL.");
        } else if(!this.places.containsKey(p.getName())) {
            throw new RuntimeException("Cannot remove a place that is not "
                            + "part of this net.");
        }
        
        final List<Arc> inputArcsToRemove = Pool.getPool().getList(p.getInput());
        for(final Arc arc : inputArcsToRemove) {
            this.removeArc(arc);
        }
        
        final List<Arc> outputArcsToRemove = Pool.getPool().getList(p.getOutput());
        for(final Arc arc : outputArcsToRemove) {
            this.removeArc(arc);
        }
        
        this.removeFromStatusList(place, place.getStatus());
        this.places.remove(p.getName());

        if((place.getPosX() >= this.getMaxX())
                        || (place.getPosX() <= this.getMinX())) {
            this.setWidthIsNotCorrect();
        }
        
        if((place.getPosY() >= this.getMaxY())
                        || (place.getPosY() <= this.getMinY())) {
            this.setHeightIsNotCorrect();
        }
        
        p.clear();
        this.elemCount--;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#removeTransition(petrinettool.Transition)
     */
    public void removeTransition(final ITransition transition) {
        final Transition t = (Transition) transition;
        if(transition == null) {
            throw new RuntimeException("Cannot remove transition NULL.");            
        } else if(!this.transitions.containsKey(t.getName())) {
            throw new RuntimeException("Cannot remove a transition that is not "
                            + "part of this net.");
        }
        
        this.transitions.remove(t.getName());

        final List<Arc> inputArcsToRemove = Pool.getPool().getList(t.getInput());
        for(final Arc arc : inputArcsToRemove) {
            this.removeArc(arc);
        }
        
        final List<Arc> outputArcsToRemove = Pool.getPool().getList(t.getOutput());
        for(final Arc arc : outputArcsToRemove) {
            this.removeArc(arc);
        }

        if((transition.getPosX() >= this.getMaxX())
                        || (transition.getPosX() <= this.getMinX())) {
            this.setWidthIsNotCorrect();
        }
        
        if((transition.getPosY() >= this.getMaxY())
                        || (transition.getPosY() <= this.getMinY())) {
            this.setHeightIsNotCorrect();
        }
        
        t.clear();
        this.elemCount--;
    }
    
    
    public int getElemCount() {
        return this.elemCount;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#getPlaces()
     */
    public Place[] getPlaces() {
        final Place[] p = new Place[this.places.size()];
        this.places.values().toArray(p);
        return p;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#getPlaceByName(java.lang.String)
     */
    public Place getPlaceByName(final String name) {
        return this.places.get(name);
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#getTransitions()
     */
    public Transition[] getTransitions() {
        final Transition[] t = new Transition[this.transitions.size()];
        this.transitions.values().toArray(t);
        return t;
    }
    
    
    public Collection<Transition> getTransitionCollection() {
        return this.transitions.values();
    }
    
    
    public Collection<Place> getPlaceCollection() {
        return this.places.values();
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#getTransitionByName(java.lang.String)
     */
    public Transition getTransitionByName(final String name) {
        return this.transitions.get(name);
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#toPNML()
     */
    public String toPNML() {
        this.correctGuards();
        return (new PNMLWriter()).writeString(this);
    }


    @Override
    public String toString() {
        return this.toPNML();
    }


    protected void finalize() {
        // remove circular references
        for(final Place p : this.getPlaces()) {
            p.clear();
        }

        for(final Transition t : this.getTransitions()) {
            t.clear();
        }

        for(final Arc a : this.getArcs()) {
            a.clear();
        }

        this.arcs.clear();
        this.arcs = null;
        this.places.clear();
        this.places = null;
        this.transitions.clear();
        this.transitions = null;
        this.exitPlaces.clear();
        this.exitPlaces = null;
        this.entryPlaces.clear();
        this.entryPlaces = null;
        this.functions.clear();
        this.functions = null;
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#getEntryPlaces()
     */
    public Place[] getEntryPlaces() {
        final Place[] entries = new Place[this.entryPlaces.size()];       
        this.entryPlaces.toArray(entries);
        return entries;
    }
    
    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.PetriNet#getExitPlaces()
     */
    public Place[] getExitPlaces() {
        final Place[] exits = new Place[this.exitPlaces.size()];       
        this.exitPlaces.toArray(exits);
        return exits;
    }
    
    
    public void checkSize() {
        if(!this.isWidthCorrect()) {
            this.setMinX(Integer.MAX_VALUE);
            this.setMaxX(Integer.MIN_VALUE);
        }
        if(!this.isHeightCorrect()) {
            this.setMinY(Integer.MAX_VALUE);
            this.setMaxY(Integer.MIN_VALUE);
        }
        
        for(final Place elem : this.getPlaces()) {
            if(!this.isWidthCorrect()) {
                if(elem.getPosX() > this.maxX) {
                    this.setMaxX(elem.getPosX());
                } 
                if(elem.getPosX() < this.minX) {
                    this.setMinX(elem.getPosX());
                }
            }
            
            if(!this.isHeightCorrect()) {
                if(elem.getPosY() > this.maxY) {
                    this.setMaxY(elem.getPosY());
                }
                if(elem.getPosY() < this.minY) {
                    this.setMinY(elem.getPosY());
                }
            }
        }
        
        for(final Transition elem : this.getTransitions()) {
            if(!this.isWidthCorrect()) {
                if(elem.getPosX() > this.maxX) {
                    this.setMaxX(elem.getPosX());
                } 
                if(elem.getPosX() < this.minX) {
                    this.setMinX(elem.getPosX());
                }
            }
            
            if(!this.isHeightCorrect()) {
                if(elem.getPosY() > this.maxY) {
                    this.setMaxY(elem.getPosY());
                } 
                if(elem.getPosY() < this.minY) {
                    this.setMinY(elem.getPosY());
                }
            }
        }
        
        if(this.minX == Integer.MAX_VALUE) {
            this.setMinX(0);
        }
        if(this.maxX == Integer.MIN_VALUE) {
            this.setMaxX(0);
        }
        
        if(this.minY == Integer.MAX_VALUE) {
            this.setMinY(0);
        }
        if(this.maxY == Integer.MIN_VALUE) {
            this.setMaxY(0);
        }
        
        this.isHeightCorrect = true;
        this.isWidthCorrect = true;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.sim.petrinettool.PetriNet#translate(int, int)
     */
    public void translate(final int x, final int y) {
        for(final Place p : this.getPlaces()) {
            p.setPos(p.getPosX() + x, p.getPosY() + y);
        }
        
        for(final Transition t : this.getTransitions()) {
            t.setPos(t.getPosX() + x, t.getPosY() + y);
        }
    }
    
    
    public boolean isWidthCorrect() {
        return this.isWidthCorrect;
    }
    
    
    public boolean isHeightCorrect() {
        return this.isHeightCorrect;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.sim.petrinettool.PetriNet#getMinY()
     */
    public int getMinY() {
        if(!this.isHeightCorrect()) {
            this.checkSize();
        }
        return this.minY;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.sim.petrinettool.PetriNet#getMaxY()
     */
    public int getMaxY() {
        if(!this.isHeightCorrect()) {
            this.checkSize();
        }
        return this.maxY;
    }

    
    /* (non-Javadoc)
     * @see petruchio.sim.petrinettool.PetriNet#getMinX()
     */
    public int getMinX() {
        if(!this.isWidthCorrect()) {
            this.checkSize();
        }
        return this.minX;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.sim.petrinettool.PetriNet#getMaxX()
     */
    public int getMaxX() {
        if(!this.isWidthCorrect()) {
            this.checkSize();
        }
        return this.maxX;
    }
    
    
    public void setMinY(final int y) {
        this.minY = y;
    }
    
    
    public void setMaxY(final int y) {
        this.maxY = y;
    }
    
    
    public void setMinX(final int x) {
        this.minX = x;
    }
    
    
    public void setMaxX(final int x) {
        this.maxX = x;
    }
    
    
    public void setWidthIsNotCorrect() {
        this.isWidthCorrect = false;
    }
    
    
    public void setHeightIsNotCorrect() {
        this.isHeightCorrect = false;
    }
    
    
    public void changeName(final Place place, final String name) {
        this.places.remove(place.getName());
        if (this.places.containsKey(name)) {
            System.err.println(this.places.keySet());
            throw new RuntimeException("A place with name \"" + name + "\" "
                            + "already exists, old name was " + place.getName() + ".");
        }
        if(this.transitions.containsKey(name)) {
            System.err.println("WARNING: You're renaming a place \""
                    + place.getName() + "\" to \"" + name + "\".\n"
                    + "  A transition with that name already exists.\n"
                    + "  This may cause problems in some programs (f.e. PEP).");
        }
        
        this.places.put(name, place);
    }
    
    
    private void removeFromStatusList(final IPlace place, final IPlaceStatus status) {
        if(status.equals(IPlace.STATUS_ENTRY)) {
                PNModel.removeThisObject(this.entryPlaces, place);
        } else if(status.equals(IPlace.STATUS_EXIT)) {
                PNModel.removeThisObject(this.exitPlaces, place);
        }
    }
    
    
    private void addToStatusList(final IPlace place, final IPlaceStatus status) {
        if(status.equals(IPlace.STATUS_ENTRY)) {
            this.entryPlaces.add(place);
        } else if(status.equals(IPlace.STATUS_EXIT)) {
            this.exitPlaces.add(place);
        }
    }
    
    
    public void changeStatus(final IPlace place, final IPlaceStatus status) {
        this.removeFromStatusList(place, place.getStatus());
        this.addToStatusList(place, status);
    }
    
    
    public String toPEP() {
        this.correctGuards();
        return (new PEPWriter()).writeString(this);
    }
    
    
    public void toPEPFile(final String filename) {
        this.correctGuards();
        (new PEPWriter()).writeFile(this, filename);
    }
    
    
    public void optimize() {
        this.removeDeadParts();
    }
    
    
    public void removeDeadTransitions() {
        final List<Transition> dead = Pool.getPool().getList();
        for(final Transition t : this.getTransitions()) {
            if(Node.FALSE.equals(t.getGuardTree())) {
                dead.add(t);
            } else {
                boolean removed = false;
                final Map<String, ValueList> in = Pool.getPool().getMap();
                for(final Arc arc : t.getInput()) {
                    if(!arc.isSimpleArc() && !arc.isReadArc()) {
                        continue;
                    }
                    if(!arc.typeCheck(false)) {
                        dead.add(t);
                        removed = true;
                        break;
                    }
                    
                    // remember what is to be taken from this place
                    final String p = arc.getPlace().getName();
                    ValueList l = in.get(p);
                    if(l == null) {
                        l = Pool.getPool().getValueList();
                        in.put(p, l);
                    }
                    final Label label = arc.getLabel();
                    l.addAll(label.getAll());
                }
                if(!removed) {
                    for(final Arc arc : t.getOutput()) {
                        if(!arc.isSimpleArc()) {
                            continue;
                        }
                        if(!arc.typeCheck(false)) {
                            dead.add(t);
                            removed = true;
                            break;
                        }
                    }
                }
                
                
                // now check all the inhibitor arcs.
                // if we take or read n*X from a place, the inhibitor arc
                // may not contain k*X (k <= n) as an inscription 
                if(!removed) {
                    for(final Arc arc : t.getInput()) {
                        if(!arc.isInhibitorArc()) {
                            continue;
                        }
                        final String p = arc.getPlace().getName();
                        final ValueList l = in.get(p);
                        if(l != null) {
                            final Label label = arc.getLabel();
                            if(label.getValues().contains(Value.V_ANY)) {
                                dead.add(t);
                                removed = true;
                            }
                            if(!removed) {
                                for(final Value v : label.getAll()) {
                                    final int k = label.getAll().count(v);
                                    final int n = l.count(v);
                                    if((k != 0) && (k <= n)) {
                                        dead.add(t);
                                        removed = true;
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
                
            }
        }
        for(final Transition t : dead) {
            this.removeTransition(t);
        }
    }
    
    
    public void removeDeadParts() {
        this.removeDeadTransitions();
        
        final List<Place> placesToCheck = Pool.getPool().getList();
        placesToCheck.addAll(this.places.values());
        
        while(!placesToCheck.isEmpty()) {
            final Place p = placesToCheck.remove(placesToCheck.size()-1);
            if(this.containsPlace(p.getName()) && p.getInput().isEmpty()
                            && p.getCurrentMarkingSet().isEmpty()
                            && p.getInitialMarkingSet().isEmpty()) {
                // empty place with no input arcs
                
                // there may be dead transitions
                final List<Transition> deadTrans = Pool.getPool().getList();
                
                for(final Arc poutarc : p.getOutput()) {
                    if(!poutarc.isSimpleArc() && !poutarc.isReadArc()) {
                        continue;
                    }
                    // walk through post set of place to delete
                    final Transition t = poutarc.getTransition();
                    // this transition will be dead
                    deadTrans.add(t);
                    for(final Arc toutarc : t.getOutput()) {
                        if(!toutarc.isSimpleArc()) {
                            continue;
                        }
                        // since the transition will be dead
                        // it will not generate tokens in its post set
                        // the post set must be checked
                        placesToCheck.add(toutarc.getPlace());
                    }
                }
                
                // now delete the place
                this.removePlace(p);
                
                // delete its post set
                for(final Transition t : deadTrans) {
                    if(this.containsTransition(t.getName())) {
                        this.removeTransition(t);
                    }
                }
            } else if(this.containsPlace(p.getName()) && p.getInput().isEmpty()
                            && p.getOutput().isEmpty()) {
                // isolated place
                this.removePlace(p);
            }
        }
    }
    
    
    public void correctGuards() {
        for(Iterator<Transition> it = this.getTransitionIterator(); it.hasNext(); ) {
            Transition t = it.next();
            t.setGuard(PNModel.correctGuard(t.getGuard()));
        }
    }
    
    
    public void resetMarking() {
        for(Iterator<Place> it = this.getPlaceIterator(); it.hasNext();) {
            Place p = it.next();
            p.resetMarking();
        }
    }

    
    public void addFunction(String function) {
        this.functions.add(function);
    }
    
    
    public Collection<String> getFunctions() {
        return this.functions;
    }


    
    public boolean checkPlaceValues() {
        return this.checkPlaceValues;
    }


    
    public void setCheckPlaceValues(boolean checkPlaceValues) {
        this.checkPlaceValues = checkPlaceValues;
    }
}
