/*
 * Project:  UML
 * Filename: Place.java
 * Created:  24.04.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.net;

import petruchio.sim.petrinettool.IPlace;
import petruchio.sim.petrinettool.IPlaceType;
import petruchio.sim.petrinettool.IPlaceStatus;
import petruchio.sim.petrinettool.type.Set;

import petruchio.sim.pnmodel.PNModel;
import petruchio.sim.pnmodel.util.Pool;

import java.util.Collection;
import java.util.List;
import java.util.Iterator;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class Place implements petruchio.sim.petrinettool.IPlace {
    public static final IPlaceType defaultType = IPlace.TYPE_BLACK_TOKEN;

    private static int codeCounter = 0;
    
    private static int counter = 0;

    private String name;

    private IPlaceStatus status;

    private IPlaceType type;

    private final ValueList currentMarking;
    
    private final ValueList initialMarking;
    
    private PetriNet net;

    private final List<Arc> input;

    private final List<Arc> output;
    
    private final List<Transition> successors;
    
    private final List<Transition> predecessors;

    private String meaning;
    
    private int posX;
    
    private int posY;
    
    private int capacity;
    
    private int code;
    
    private int key;
    
    private boolean allowAnyInput;
    
    
    public Place(final String name, final IPlaceStatus status,
                    final IPlaceType type, final boolean allowAnyInput,
                    final String initmarking, final String curmarking,
                    final PetriNet net) {
        if(status == null) {
            throw new RuntimeException("Error: A place's status may not be null.");
        }
        this.capacity = IPlace.UNBOUNDED;
        this.meaning = "";
        this.name = name;
        this.status = status;
        this.setType(type);
        this.net = net;
        this.initialMarking = Pool.getPool().getValueList();
        this.currentMarking = Pool.getPool().getValueList();
        this.input = Pool.getPool().getList();
        this.output = Pool.getPool().getList();
        this.successors = Pool.getPool().getList();
        this.predecessors = Pool.getPool().getList();
        this.allowAnyInput = false;
        this.setAllowAnyInput(allowAnyInput);
        this.setPos(0, 0);
        this.setInitialMarking(initmarking);
        this.setCurrentMarking(curmarking);
        this.code = Place.codeCounter++;
    }
    
    
    public Place(final String name, final IPlaceStatus status,
                    final IPlaceType type, final String initmarking,
                    final String curmarking, final PetriNet net) {
        this(name, status, type, false, initmarking, curmarking, net);
    }


    public Place(final IPlaceStatus status, final IPlaceType type,
                    final String initmarking, final String curmarking,
                    final PetriNet net) {
        this(Place.freshName(), status, type, initmarking, curmarking, net);
    }


    public static String freshName() {
        return "_p" + (++Place.counter);
    }

    
    public boolean allowAnyInput() {
        return this.allowAnyInput;
    }
    
    
    public void setAllowAnyInput(final boolean val) {
        if(!this.allowAnyInput && val) {
            final Set buff = new Set();
            buff.addType(this.type);
            buff.addType(IPlace.TYPE_ANY);
            this.type = buff;
        } else if(this.allowAnyInput && !val) {
            if(this.type instanceof Set) {
                final Set s = (Set)this.type;
                final Set buff = new Set();
                for(final Iterator<IPlaceType> it = s.getTypes().iterator(); it.hasNext();) {
                    final IPlaceType t = it.next();
                    if(!t.equals(petruchio.sim.petrinettool.type.Any.ANY)) {
                        buff.addType(t);
                    }
                }
                this.type = buff;
            } else {
                throw new RuntimeException();
            }
        }
        this.allowAnyInput = val;
    }

    
    public String getMeaning() {
        return this.meaning;
    }
    
    
    public void setMeaning(final String m) {
        this.meaning = m;
    }
    
    
    public String getName() {
        return this.name;
    }

    
    public void setPos(final int x, final int y) {
        this.setPosX(x);
        this.setPosY(y);
    }
    
    
    /**
     * @return Returns the posX.
     */
    public int getPosX() {
        return this.posX;
    }


    /**
     * @param x The posX to set.
     */
    public void setPosX(final int x) {
        if(this.net.isWidthCorrect()) {
            if((x > this.net.getMaxX()) || (this.net.getElemCount() == 1)) {
                this.net.setMaxX(x);
            } else if((this.posX >= this.net.getMaxX()) && (x < this.posX)) {
                this.net.setWidthIsNotCorrect();
            }
        }
        if(this.net.isWidthCorrect()) {
            if((x < this.net.getMinX()) || (this.net.getElemCount() == 1)) {
                this.net.setMinX(x);
            } else if((this.posX <= this.net.getMinX()) && (x > this.posX)) {
                this.net.setWidthIsNotCorrect();
            }
        }
        
        this.posX = x;
    }


    /**
     * @return Returns the posY.
     */
    public int getPosY() {
        return this.posY;
    }
    
    
    /**
     * @param y The posY to set.
     */
    public void setPosY(final int y) {
        if(this.net.isHeightCorrect()) {
            if((y > this.net.getMaxY()) || (this.net.getElemCount() == 1)) {
                this.net.setMaxY(y);
            } else if((this.posY >= this.net.getMaxY()) && (y < this.posY)) {
                this.net.setHeightIsNotCorrect();
            }
        }
        if(this.net.isHeightCorrect()) {
            if((y < this.net.getMinY()) || (this.net.getElemCount() == 1)) {
                this.net.setMinY(y);
            } else if((this.posY <= this.net.getMinY()) && (y > this.posY)) {
                this.net.setHeightIsNotCorrect();
            }
        }
        
        this.posY = y;
    }
    

    /**
     * @return Returns the marking.
     */
    public String getCurrentMarking() {
        return Place.getMarking(this.currentMarking);
    }
    
    
    public String getInitialMarking() {
        return Place.getMarking(this.initialMarking);
    }
    
    
    private static String getMarking(final Collection<Value> m) {
        final StringBuffer buf = new StringBuffer();
        for(final Iterator<Value> it = m.iterator(); it.hasNext(); ) {
            final Value value = it.next();
            buf.append(value.getNetValue());
            if(it.hasNext()) {
                buf.append(",");
            }
        }
        return buf.toString();
    }
    
    
    public ValueList getCurrentMarkingSet() {
        return this.currentMarking;
    }
    
    
    public ValueList getInitialMarkingSet() {
        return this.initialMarking;
    }
    
    
    public void resetMarking() {
        this.setCurrentMarking(this.initialMarking);
    }
    

    /**
     * @return Returns the input.
     */
    public List<Arc> getInput() {
        return this.input;
    }

    /**
     * @return Returns the output.
     */
    public List<Arc> getOutput() {
        return this.output;
    }


    /**
     * @return Returns the status.
     */
    public IPlaceStatus getStatus() {
        return this.status;
    }


    /**
     * @return Returns the type.
     */
    public IPlaceType getType() {
        return this.type;
    }

    
    public void setStatus(final IPlaceStatus status) {
        if(status == null) {
            throw new RuntimeException("Error: A place's status may not be null.");
        }
        this.net.changeStatus(this, status);
        this.status = status;
    }
    

    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.Place#rename(java.lang.String)
     */
    public void rename(final String theName) {
        this.setName(theName);
    }
    

    /**
     * @param newMarking The marking to set.
     */
    public void setCurrentMarking(final String newMarking) {
        this.currentMarking.clear();
        final Collection<Value> tokens = Place.parseMarking(newMarking);
        for(final Value token : tokens) {
            this.addToken(token);
        }
    }
    
    
    public void setInitialMarking(final String newMarking) {
        this.initialMarking.clear();
        final Collection<Value> tokens = Place.parseMarking(newMarking);
        for(final Value token : tokens) {
            this.addToken(token, this.initialMarking);
        }
    }
    
    
    public static ValueList parseMarking(String marking) {
        if((marking != null) && (marking.trim().length() > 0)) {
            marking = PNModel.DOT.matcher(marking).replaceAll(PNModel.DOT_REPL);
            marking = PNModel.cleanup(marking);
            final ValueList tokens = Pool.getPool().getValueList();
            PNModel.parse(marking, tokens, tokens, tokens, null);
            return tokens;
        } else {
            return Pool.getPool().getValueList();
        }
    }
    
    
    public void addToken(final Value token) {
        this.addToken(token, this.currentMarking);
    }
    
    
    private void addToken(final Value token, final Collection<Value> m) {
        if(this.allowAnyInput || !this.net.checkPlaceValues()
                        || this.type.isValue(token.getNetValue())) {
            m.add(token);
        } else {
            throw new RuntimeException("\nError: Token " + token + " cannot be "
                            + "added to\n" + this + ".\nIt is "
                            + "incompatible with this place's type.");
        }
    }
    
    
    public void clearMarking() {
        this.currentMarking.clear();
    }
    
    
    public void setCurrentMarking(final Collection<Value> marking) {
        this.currentMarking.clear();
        this.currentMarking.addAll(marking);
    }
    
    
    public void addMarking(final Collection<Value> marking) {
        this.currentMarking.addAll(marking);
    }
    
    
    public boolean removeToken(final Value token) {
        return this.currentMarking.remove(token);
    }
    
    public boolean removeAllTokens(final Value token) {
        return this.currentMarking.removeAllOf(token);
    }
    
    
    public boolean containsToken(final Value token) {
        return this.currentMarking.contains(token);
    }
    
    
    /**
     * @param name The name to set.
     */
    public void setName(String name) {
        if((name == null) || (name.length() == 0)) {
            name = Place.freshName();
        }
        if(name.equals(this.name)) {
            return;
        }
        this.net.changeName(this, name);
        this.name = name;
    }
    
    
    /**
     * @param type The type to set.
     */
    public void setType(IPlaceType type) {
        if(type == null) {
            type = Place.defaultType;
        }
        if((type instanceof petruchio.sim.petrinettool.type.Tuple)
                        || (type instanceof petruchio.sim.petrinettool.type.Set)) {
            this.type = type;
        } else {
            petruchio.sim.petrinettool.type.Set types = new petruchio.sim.petrinettool.type.Set();
            types.addType(type);
            this.type = types;
        }
    }
    
    
    public void addInput(final Arc arc) {
        this.input.add(arc);
        final Transition t = arc.getTransition();
        if(!PNModel.containsThisObject(this.predecessors, t)) {
            this.predecessors.add(t);
        }
    }


    public void addOutput(final Arc arc) {
        this.output.add(arc);
        final Transition t = arc.getTransition();
        if(!PNModel.containsThisObject(this.successors, t)) {
            this.successors.add(t);
        }
    }

    
    public boolean removeArc(final Arc arc) {
        final Transition t = arc.getTransition();
        if(arc.transIsSource()) {
            PNModel.removeThisObject(this.predecessors, t);
            return PNModel.removeThisObject(this.input, arc);
        } else {
            PNModel.removeThisObject(this.successors, t);
            return PNModel.removeThisObject(this.output, arc);
        }
    }

    
    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.Place#getPredecessorTransitions()
     */
    public final Transition[] getPredecessorTransitions() {
        final List<Transition> trans = Pool.getPool().getList();

        for(final Arc arc : this.input) {
            trans.add(arc.getTransition());
        }

        final Transition[] array = new Transition[trans.size()];
        trans.toArray(array);
        return array;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.Place#getSuccessorTransitions()
     */
    public Transition[] getSuccessorTransitions() {
        final List<Transition> trans = Pool.getPool().getList();

        for(final Arc arc : this.output) {
            trans.add(arc.getTransition());
        }

        final Transition[] array = new Transition[trans.size()];
        trans.toArray(array);
        return array;
    }
    
    
    public Collection<Transition> getSuccessors() {
        return this.successors;
    }
    
    
    public Collection<Transition> getPredecessors() {
        return this.predecessors;
    }
    
    
    public void clear() {
        this.name = null;
        this.net = null;
        this.input.clear();
        this.output.clear();
        this.capacity = -1;
        this.meaning = "";
    }
    
    
    @Override
    public boolean equals(final Object o) {
        return this == o;
    }
    
    
    @Override
    public String toString() {
        return "Place(\n" 
             + "  name     = " + this.getName() + ",\n"
             + "  status   = " + this.getStatus().getName() + ",\n"
             + "  type     = " + this.getType().getName() + ",\n"
             + "  capacity = " + this.getCapacity() + ",\n"
             + "  initial  = " + this.getInitialMarking() + ",\n"
             + "  current  = " + this.getCurrentMarking() + ",\n"
             + "  pos x    = " + this.getPosX() + ",\n"
             + "  pos y    = " + this.getPosY() + "\n"
             + ")";
    }
    
    
    public void setCapacity(final int capacity) {
        this.capacity = (capacity < 1) ? -1 : capacity;
    }
    
    
    public int getCapacity() {
        return this.capacity;
    }
    
    
    public boolean infiniteCapacity() {
        return this.capacity < 1;
    }

    
    @Override
    public int hashCode() {
        return this.code;//this.name.hashCode() ^ this.net.getName().hashCode();
    }


    
    public int getKey() {
        return this.key;
    }


    
    public void setKey(int key) {
        this.key = key;
    }
    
}
