/*
 * Project:  UML
 * Filename: Transition.java
 * Created:  24.04.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.net;

import java.util.List;
import java.util.Map;
import java.util.Iterator;
import java.util.Collection;
import java.util.Set;

import petruchio.sim.petrinettool.IMultiSet;

import petruchio.sim.pnmodel.PEPWriter;
import petruchio.sim.pnmodel.PNModel;
import petruchio.sim.Simulator;
import petruchio.sim.pnmodel.util.Binding;
import petruchio.sim.pnmodel.util.Function;
import petruchio.sim.pnmodel.util.Node;
import petruchio.sim.pnmodel.util.Pool;
import petruchio.sim.pnmodel.util.TreeComputer;


/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public class Transition implements petruchio.sim.petrinettool.ITransition {
    
    private static int codeCounter = 0;
    
    private static int counter = 0;

    private final List<Arc> input;

    private final List<Arc> output;

    private String name;

    private PetriNet net;

    private final petruchio.sim.pnmodel.net.MultiSet actions;
    
    private final Map<String, Value> variables;
    
    private final List<Binding> ageMemory;
    
    private String meaning;
    
    private int posX;
    
    private int posY;
    
    private int key;
    
    private final int code;
    
    private final Node[] expressions;
    private final boolean[] hasExpression;

    
    public Transition(final String name, final IMultiSet actions,
                    final String guard, final PetriNet net) {
        this.expressions = new Node[ExpressionType.values().length]; // initialised with null
        this.hasExpression = new boolean[this.expressions.length];
        
        this.variables = Pool.getPool().getMap();
        this.ageMemory = Pool.getPool().getList();
        this.meaning = "";
        this.net = net;
        this.name = name;
        this.setTree(ExpressionType.GUARD, guard);
        this.actions = (actions == null) ? new petruchio.sim.pnmodel.net.MultiSet()
                                         : (petruchio.sim.pnmodel.net.MultiSet) actions;

        this.input = Pool.getPool().getList();
        this.output = Pool.getPool().getList();
        this.setPos(0, 0);
        this.code = Transition.codeCounter++;
    }


    public Transition(final IMultiSet actions, final String guard, final PetriNet net) {
        this(Transition.freshName(), actions, guard, net);
    }


    public static String freshName() {
        return "_t" + (++Transition.counter);
    }
    
    
    public String getMeaning() {
        return this.meaning;
    }
    
    
    public void setMeaning(final String m) {
        this.meaning = m;
    }
    

    public double enableAgeMemory(final Binding bind, final int count) {
        final int index = Simulator.indexOfSortPriority(this.ageMemory, bind, true);
        if(index <= -1) {
            final Binding binding = bind.clone();
            binding.setAge(0.0d);
            binding.setEarliestFiring(0.0d);
            binding.setLatestFiring(0.0d);
            Simulator.insertSortPriority(this.ageMemory, binding);
            return 0.0d;
        } else {
            final Binding binding = this.ageMemory.get(index);
            final double age = Math.min(binding.getAge()/count, bind.getLatestFiring());
            bind.setAge(age);
            binding.setAge(binding.getAge() - age);
            return age;
        }
    }
    
    
    public void disableAgeMemory(final Binding bind, final double time) {
        final int index = Simulator.indexOfSortPriority(this.ageMemory, bind, true);
        if(index <= -1) {
            throw new RuntimeException("This binding has not been enabled!");
        } else {
            final Binding binding = this.ageMemory.get(index);
            binding.setAge(binding.getAge() + bind.getAge() + time - bind.getCreationTime());
        }
    }
    
    
    public void setPos(final int x, final int y) {
        this.setPosX(x);
        this.setPosY(y);
    }
    
    
    /**
     * @return Returns the posX.
     */
    public int getPosX() {
        return this.posX;
    }


    /**
     * @param x The posX to set.
     */
    public void setPosX(final int x) {
        if(this.net.isWidthCorrect()) {
            if((x > this.net.getMaxX()) || (this.net.getElemCount() == 1)) {
                this.net.setMaxX(x);
            } else if((this.posX >= this.net.getMaxX()) && (x < this.posX)) {
                this.net.setWidthIsNotCorrect();
            }
        }
        if(this.net.isWidthCorrect()) {
            if((x < this.net.getMinX()) || (this.net.getElemCount() == 1)) {
                this.net.setMinX(x);
            } else if((this.posX <= this.net.getMinX()) && (x > this.posX)) {
                this.net.setWidthIsNotCorrect();
            }
        }
        
        this.posX = x;
    }


    /**
     * @return Returns the posY.
     */
    public int getPosY() {
        return this.posY;
    }
    
    
    /**
     * @param y The posY to set.
     */
    public void setPosY(final int y) {
        if(this.net.isHeightCorrect()) {
            if((y > this.net.getMaxY()) || (this.net.getElemCount() == 1)) {
                this.net.setMaxY(y);
            } else if((this.posY >= this.net.getMaxY()) && (y < this.posY)) {
                this.net.setHeightIsNotCorrect();
            }
        }
        if(this.net.isHeightCorrect()) {
            if((y < this.net.getMinY()) || (this.net.getElemCount() == 1)) {
                this.net.setMinY(y);
            } else if((this.posY <= this.net.getMinY()) && (y > this.posY)) {
                this.net.setHeightIsNotCorrect();
            }
        }
        
        this.posY = y;
    }
    
    
    /**
     * Scans both input and output arcs for variables. In the context of a
     * transition a reference to variable 'j' should reference all the
     * occurrences of that variable.
     */
    private void scanVariables() {
        for(final Arc arc : this.input) {
            final Label label = arc.getLabel();
            for(int i = 0; i < label.getVariables().size(); i++) {
                final Value var = label.getVariables().get(i);
                final Value v = this.getVariable(var.getVariable());
                if(v == null) {
                    this.addVariable(var.getVariable(), var);
                } else {
                    label.getVariables().set(i, v);
                }
            }
            for(int i = 0; i < label.getTuples().size(); i++) {
                final Value tuple = label.getTuples().get(i);
                this.scanTuple(tuple);
            }
        }
        
        for(final Arc arc : this.output) {
            final Label label = arc.getLabel();
            for(int i = 0; i < label.getVariables().size(); i++) {
                final Value var = label.getVariables().get(i);
                final Value v = this.getVariable(var.getVariable());
                if(v == null) {
                    this.addVariable(var.getVariable(), var);
                } else {
                    label.getVariables().set(i, v);
                }
            }
            for(int i = 0; i < label.getTuples().size(); i++) {
                final Value tuple = label.getTuples().get(i);
                this.scanTuple(tuple);
            }
        }
    }
    
    
    private void scanTuple(final Value tuple) {
        final Tuple t = tuple.getTuple();
        final List<Value> entries = t.getEntries();
        for(int i = 0; i < entries.size(); i++) {
            final Value val = entries.get(i);
            if(val.isTuple()) {
                this.scanTuple(val);
            } else if(val.isVariable()) {
                final Value v = this.getVariable(val.getVariable());
                if(v == null) {
                    this.addVariable(val.getVariable(), val);
                } else {
                    entries.set(i, v);
                }
            }
        }
    }
    
    
    public void addVariable(final String var, final Value variable) {
        this.variables.put(var, variable);
    }
    
    
    public Value getVariable(final String var) {
        return this.variables.get(var);
    }
    
    
    public Set<String> getVariables() {
        return this.variables.keySet();
    }
    
    
    public void setGuard(String guard) {
        this.setTree(ExpressionType.GUARD, guard);
    }
    
    
    public void setGuard(Node guard) {
        this.setTree(ExpressionType.GUARD, guard);
    }


    public Action[] getTie(final String symbol) {
        return this.actions.getTie(symbol);
    }


    public void removeAction(final Action action) {
        this.actions.removeAction(action);
    }


    public boolean removeArc(final Arc arc) {
        return (arc != null) && (arc.getTransition() == this) &&
            (arc.transIsSource() ? PNModel.removeThisObject(this.output, arc)
                                 : PNModel.removeThisObject(this.input, arc));
    }


    public String getName() {
        return this.name;
    }
    
    
    public static boolean containsArcTrans(final Collection<Arc> arcs, final Arc arc) {
        for(final Iterator<Arc> it = arcs.iterator(); it.hasNext(); ) {
            final Arc arc2 = it.next();
            if((arc.getPlace() ==  arc2.getPlace())
                            && arc.getLabel().equals(arc2.getLabel())) {
                return true;
            }
        }
        return false;
    }
    
    
    public static boolean arcSetsAreEqual(final Collection<Arc> arcs1, final Collection<Arc> arcs2) {
        if(arcs1.size() != arcs2.size()) {
            return false;
        }
        for(final Iterator<Arc> it = arcs1.iterator(); it.hasNext(); ) {
            if(!Transition.containsArcTrans(arcs2, it.next())) {
                return false;
            }
        }
        return true;
    }

    public IMultiSet getActions() {
        return this.actions;
    }

    
    public String getGuard() {
        return this.getExpression(ExpressionType.GUARD);
    }
    
    
    public Node getGuardTree() {
        return this.getTree(ExpressionType.GUARD);
    }


    public void addInput(final Arc arc) {
        this.input.add(arc);
    }


    public void addOutput(final Arc arc) {
        this.output.add(arc);
    }
    
    
    public boolean hasInputArc(final Arc arc) {
        return Transition.containsArcTrans(this.input, arc);
    }
    
    
    public boolean hasOutputArc(final Arc arc) {
        return Transition.containsArcTrans(this.input, arc);
    }


    public List<Arc> getInput() {
        return this.input;
    }


    public List<Arc> getOutput() {
        return this.output;
    }


    public void clear() {
        this.name = null;
        this.input.clear();
        this.output.clear();
        this.actions.clear();
        this.meaning = "";
        this.net = null;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.Transition#getPredecessorPlaces()
     */
    public Place[] getPredecessorPlaces() {
        final List<Place> place = Pool.getPool().getList();

        for(final Arc arc : this.input) {
            place.add(arc.getPlace());
        }

        final Place[] array = new Place[place.size()];
        place.toArray(array);
        return array;
    }


    /*
     * (non-Javadoc)
     * 
     * @see petruchio.sim.petrinettool.Transition#getSuccessorPlaces()
     */
    public Place[] getSuccessorPlaces() {
        final List<Place> place = Pool.getPool().getList();

        for(final Arc arc : this.output) {
            place.add(arc.getPlace());
        }

        final Place[] array = new Place[place.size()];
        place.toArray(array);
        return array;
    }
    
    
    @Override
    public boolean equals(final Object o) {
        return (o instanceof Transition)
                && this.equals((Transition)o);
    }
    
    
    public boolean equals(final Transition trans) {
        return (trans == this) || ((trans != null) 
                  && this.expressionsEqual(trans)
                  && Transition.arcSetsAreEqual(trans.getInput(), this.input)
                  && Transition.arcSetsAreEqual(trans.getOutput(), this.output)
                  && trans.getActions().equals(this.actions));
    }
    
    
    private boolean expressionsEqual(final Transition that) {
        for(int i = 0; i < this.expressions.length; ++i) {
            if((this.hasExpression[i] != that.hasExpression[i])
                    || (this.hasExpression[i]
                           && !this.expressions[i].equals(that.expressions[i]))) {
                return false;
            }
        }
        return true;
    }
    
    
    @Override
    public int hashCode() {
        return this.code;//this.name.hashCode() ^ this.net.getName().hashCode();
    }


    @Override
    public String toString() {
        final StringBuffer trees = new StringBuffer();
        for(final ExpressionType t : ExpressionType.values()) {
            final Node n = this.getTree(t);
            trees.append("  ");
            trees.append(t.toString());
            trees.append(" = ");
            trees.append(n);
            trees.append("\n");
        }
        return "Transition(\n" 
             + "  name     = " + this.getName() + ",\n"
             + "  meaning  = " + this.getMeaning() + ",\n"
             + "  actions  = " + this.getActions() + ",\n"
             + trees.toString()
             + "  pos x    = " + this.getPosX() + ",\n"
             + "  pos y    = " + this.getPosY() + "\n"
             + ")";
    }
    
    
    public String getLatestFiring() {
        return this.getExpression(ExpressionType.LATEST_FIRING);
    }
    
    
    public Node getLatestFiringTree() {
        return this.getTree(ExpressionType.LATEST_FIRING);
    }

    
    public void setLatestFiring(final String maxTime) {
        if(this.hasWeight()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a timing.");
        } else if(this.hasRate()) {
            throw new IllegalArgumentException("A transition cannot have a rate and another timing.");
        }
        this.setTree(ExpressionType.LATEST_FIRING, maxTime);
    }
    
    
    public void setLatestFiring(final Node maxTime) {
        if(this.hasWeight()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a timing.");
        } else if(this.hasRate()) {
            throw new IllegalArgumentException("A transition cannot have a rate and another timing.");
        }
        this.setTree(ExpressionType.LATEST_FIRING, maxTime);
    }

    
    public String getEarliestFiring() {
        return this.getExpression(ExpressionType.EARLIEST_FIRING);
    }
    
    
    public Node getEarliestFiringTree() {
        return this.getTree(ExpressionType.EARLIEST_FIRING);
    }

    
    public void setEarliestFiring(final String minTime) {
        if(this.hasWeight()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a timing.");
        } else if(this.hasRate()) {
            throw new IllegalArgumentException("A transition cannot have a rate and another timing.");
        }
        this.setTree(ExpressionType.EARLIEST_FIRING, minTime);
    }
    
    
    public void setEarliestFiring(final Node minTime) {
        if(this.hasWeight()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a timing.");
        } else if(this.hasRate()) {
            throw new IllegalArgumentException("A transition cannot have a rate and another timing.");
        }
        this.setTree(ExpressionType.EARLIEST_FIRING, minTime);
    }
    
    
    public String getPriority() {
        return this.getExpression(ExpressionType.PRIORITY);
    }
    
    
    public Node getPriorityTree() {
        return this.getTree(ExpressionType.PRIORITY);
    }
    

    public void setPriority(final String priority) {
        this.setTree(ExpressionType.PRIORITY, priority);
    }
    
    
    public void setPriority(final Node priority) {
        this.setTree(ExpressionType.PRIORITY, priority);
    }
    
    
    public String getRate() {
        return this.getExpression(ExpressionType.RATE);
    }
    
    
    public Node getRateTree() {
        return this.getTree(ExpressionType.RATE);
    }
    

    public void setRate(final String rate) {
        if(this.hasWeight()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a rate.");
        } else if(this.hasEarliestFiring() || this.hasLatestFiring()) {
            throw new IllegalArgumentException("A transition cannot have a rate and another timing.");
        }
        this.setTree(ExpressionType.RATE, rate);
    }
    
    
    public void setRate(final Node rate) {
        if(this.hasWeight()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a rate.");
        } else if(this.hasEarliestFiring() || this.hasLatestFiring()) {
            throw new IllegalArgumentException("A transition cannot have a rate and another timing.");
        }
        this.setTree(ExpressionType.RATE, rate);
    }
    
    
    public String getGeneralizedRate() {
        return this.getExpression(ExpressionType.GENERALIZED_RATE);
    }
    
    
    public Node getGeneralizedRateTree() {
        return this.getTree(ExpressionType.GENERALIZED_RATE);
    }
    

    public void setGeneralizedRate(final String rate) {
        if(this.hasWeight() || this.hasGeneralizedWeight()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a rate.");
        } else if(this.hasEarliestFiring() || this.hasLatestFiring()) {
            throw new IllegalArgumentException("A transition cannot have a rate and another timing.");
        }
        this.setTree(ExpressionType.GENERALIZED_RATE, rate);
    }
    
    
    public void setGeneralizedRate(final Node rate) {
        if(this.hasWeight() || this.hasGeneralizedWeight()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a rate.");
        } else if(this.hasEarliestFiring() || this.hasLatestFiring()) {
            throw new IllegalArgumentException("A transition cannot have a rate and another timing.");
        }
        this.setTree(ExpressionType.GENERALIZED_RATE, rate);
    }
    
    
    public String getWeight() {
        return this.getExpression(ExpressionType.WEIGHT);
    }
    
    
    public Node getWeightTree() {
        return this.getTree(ExpressionType.WEIGHT);
    }
    

    public void setWeight(final String weight) {
        if(this.hasRate() || this.hasGeneralizedRate()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a rate.");
        } else if(this.hasEarliestFiring() || this.hasLatestFiring()) {
            throw new IllegalArgumentException("A transition cannot have a weight and be timed.");
        }
        this.setTree(ExpressionType.WEIGHT, weight);
    }
    
    
    public void setWeight(final Node weight) {
        if(this.hasRate() || this.hasGeneralizedRate()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a rate.");
        } else if(this.hasEarliestFiring() || this.hasLatestFiring()) {
            throw new IllegalArgumentException("A transition cannot have a weight and be timed.");
        }
        this.setTree(ExpressionType.WEIGHT, weight);
    }
    
    
    public String getGeneralizedWeight() {
        return this.getExpression(ExpressionType.GENERALIZED_WEIGHT);
    }
    
    
    public Node getGeneralizedWeightTree() {
        return this.getTree(ExpressionType.GENERALIZED_WEIGHT);
    }
    

    public void setGeneralizedWeight(final String weight) {
        if(this.hasRate() || this.hasGeneralizedRate()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a rate.");
        } else if(this.hasEarliestFiring() || this.hasLatestFiring()) {
            throw new IllegalArgumentException("A transition cannot have a weight and be timed.");
        }
        this.setTree(ExpressionType.GENERALIZED_WEIGHT, weight);
    }
    
    
    public void setGeneralizedWeight(final Node weight) {
        if(this.hasRate() || this.hasGeneralizedRate()) {
            throw new IllegalArgumentException("A transition cannot have a weight and a rate.");
        } else if(this.hasEarliestFiring() || this.hasLatestFiring()) {
            throw new IllegalArgumentException("A transition cannot have a weight and be timed.");
        }
        this.setTree(ExpressionType.GENERALIZED_WEIGHT, weight);
    }
    
    
    public String getServers() {
        return this.getExpression(ExpressionType.SERVERS);
    }
    
    
    public Node getServersTree() {
        return this.getTree(ExpressionType.SERVERS);
    }
    

    public void setServers(final String servers) {
        this.setTree(ExpressionType.SERVERS, servers);
    }
    
    
    public void setServers(final Node servers) {
        this.setTree(ExpressionType.SERVERS, servers);
    }
    
    
    public String getAgeMemory() {
        return this.getExpression(ExpressionType.AGE_MEMORY);
    }
    
    
    public Node getAgeMemoryTree() {
        return this.getTree(ExpressionType.AGE_MEMORY);
    }
    

    public void setAgeMemory(final String ageMemory) {
        this.setTree(ExpressionType.AGE_MEMORY, ageMemory);
    }
    
    
    public void setAgeMemory(final Node ageMemory) {
        this.setTree(ExpressionType.AGE_MEMORY, ageMemory);
    }
    
    
    public boolean isImmediate() {
        return !this.hasRate() && !this.hasEarliestFiring() && !this.hasLatestFiring() && !this.hasGeneralizedRate();
    }
    
    
    public boolean hasAgeMemory() {
        return this.hasTree(ExpressionType.AGE_MEMORY);
    }
    
    
    public boolean hasRate() {
        return this.hasTree(ExpressionType.RATE);
    }
    
    
    public boolean hasWeight() {
        return this.hasTree(ExpressionType.WEIGHT);
    }
    
    
    public boolean hasGeneralizedRate() {
        return this.hasTree(ExpressionType.GENERALIZED_RATE);
    }
    
    
    public boolean hasGeneralizedWeight() {
        return this.hasTree(ExpressionType.GENERALIZED_WEIGHT);
    }
    
    
    public boolean hasPriority() {
        return this.hasTree(ExpressionType.PRIORITY);
    }
    
    
    public boolean hasServers() {
        return this.hasTree(ExpressionType.SERVERS);
    }
    
    
    public boolean hasEarliestFiring() {
        return this.hasTree(ExpressionType.EARLIEST_FIRING);
    }
    
    
    public boolean hasLatestFiring() {
        return this.hasTree(ExpressionType.LATEST_FIRING);
    }
    
    
    private boolean hasTree(final ExpressionType type) {
        return this.hasExpression[type.ordinal()];
    }
    
    
    public Map<String, Value> getVariableMapping() {
        return this.variables;
    }
    
    
    private void setTree(final ExpressionType type, final String expression) {
        if((expression == null) || (expression.length() == 0)) {
            this.setTree(type, (Node)null);
            // null isn't a real tree.
            this.hasExpression[type.ordinal()] = false;
        } else {
            this.setTree(type, TreeComputer.infixToTree(PNModel.simplifyGuard(expression)));
        }
    }
    
    
    private void setTree(final ExpressionType type, final Node expression) {
        this.expressions[type.ordinal()] = expression;
        this.hasExpression[type.ordinal()] = true;
    }
    
    
    private Node getTree(final ExpressionType type) {
        final Node expr = this.expressions[type.ordinal()];
        if(expr == null) {
            final Node clone = type.defaultTree().clone();
            this.expressions[type.ordinal()] = clone;
            return clone;
        } else {
            return expr;
        }
    }
    
    
    private String getExpression(final ExpressionType type) {
        final Node tree = this.getTree(type);
        return PEPWriter.constraintsToString(tree.getConstraints()) + tree.toString();
    }
    
    
    public void optimizeExpressions() {
        this.optimizeExpressions(null);
    }
    
    
    public void optimizeExpressions(final Map<String, Function> functions) {
        for(final ExpressionType type : ExpressionType.values()) {
            final Node tree = this.expressions[type.ordinal()];
            if((tree != null) && !tree.equals(type.defaultTree())) {
                TreeComputer.optimize(tree, functions);
            }
        }
    }
    
    
    public void identifyVariables() {
        this.variables.clear();
        this.scanVariables();
        for(final ExpressionType type : ExpressionType.values()) {
            final Node tree = this.getTree(type);
            if(!tree.equals(type.defaultTree())) {
                tree.getAndReplaceVariables(this.variables);
            }
        }
    }


    
    public int getKey() {
        return this.key;
    }


    
    public void setKey(int key) {
        this.key = key;
    }
    
    
//    public void replaceFunctions(final Map<String, Function> map) {
//        for(final ExpressionType type : ExpressionType.values()) {
//            final Node tree = this.getTree(type);
//            if(!tree.equals(type.defaultTree())) {
//                tree.replaceFunctions(map);
//            }
//        }
//    }
}