/*
 * Project:  P-UMLaut
 * Filename: Value.java
 * Created:  24.07.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.net;

import petruchio.sim.petrinettool.IPetriNetTool;
import petruchio.sim.pnmodel.PNModel;

import java.util.Collection;
import java.util.regex.Pattern;

public class Value implements Cloneable, Comparable<Value> {
    public static final Value V_NONE = new Value(null, Type.UNKNOWN);
    public static final Value V_ANY = new Value(null, Type.ANY);
    public static final Value V_DOT = new Value("dot");
    public static final Value V_TRUE = new Value(true);
    public static final Value V_FALSE = new Value(false);
    public static final Pattern POSINFTY = Pattern.compile("(?i)\\+?\\s*inf((ini)?ty)?");
    public static final Pattern NEGINFTY = Pattern.compile("(?i)-\\s*inf((ini)?ty)?");
    
    public static enum Type {UNKNOWN, VARIABLE, STRING, NUMBER, BOOLEAN, DOT, TUPLE, ANY};
    
    private static final Pattern STRING_MATCHER = Pattern.compile(IPetriNetTool.STRING_SEPARATOR+".*"+IPetriNetTool.STRING_SEPARATOR, Pattern.DOTALL);
    
    public static boolean containsAny(final Collection<?> c) {
        return c.contains(Value.V_ANY);
    }
    
    private Object value;
    private double number;
    private Type type;
    private String netValue;
    private int hash;
    private boolean isVariable = false;
    

    public boolean isAny() {
        return this.type == Type.ANY;
    }
    
    
    public static Value getValue(String value) {
        value = value.trim();
        if(value.equals("dot")) {
            return Value.V_DOT;
        } else if(value.equals("tt") || value.equals("true")) {
            return Value.V_TRUE;
        } else if(value.equals("ff") || value.equals("false")) {
            return Value.V_FALSE;
        } else if(POSINFTY.matcher(value).matches()) {
            return Value.getValue(Double.POSITIVE_INFINITY);
        } else if(NEGINFTY.matcher(value).matches()) {
            return Value.getValue(Double.NEGATIVE_INFINITY);
        } else {
            return new Value(value);
        }
    }
    
    
    public static Value getValueNoVars(String value) {
        value = value.trim();
        if(value.equals("dot")) {
            return Value.V_DOT;
        } else if(value.equals("tt")) {
            return Value.V_TRUE;
        } else if(value.equals("ff")) {
            return Value.V_FALSE;
        } else if(Value.POSINFTY.matcher(value).matches()) {
            return Value.getValue(Double.POSITIVE_INFINITY);
        } else if(Value.NEGINFTY.matcher(value).matches()) {
            return Value.getValue(Double.NEGATIVE_INFINITY);
        } else if(PNModel.isNumber(value)) {
            return Value.getValue(Double.parseDouble(value));
        } else if(value.startsWith("(")) {
            final Value val = new Value(new Tuple(value), Type.TUPLE);
            return val;
        } else {
            final Value val = new Value(value.replaceAll("\"(.*)\"", "$1"), Type.STRING);
            return val;
        }
    }
    
    
//    public static Value newValue(final Value value) {
//        final Value v = new Value(value);
//        return v;
//    }
    
    
    public static Value getValue(final Value value) {
        if(value.isHiddenVariable()) {
            final Value v = new Value(value.value, value.type, value.netValue);
            v.number = value.number;
            return v;
        } else if(value.isTuple() && value.getTuple().containsVariable()) {
            final Value v = new Value(value.getTuple().clone(), value.type);
            return v;
        } else {
            return value;
        }
    }
    
    
    public static Value getValue(final Tuple val) {
        return new Value(val);
    }
    
    
    public static Value getValue(final double val) {
        return new Value(val);
    }
    
    
//    public static Value getValue(Tuple value) {
//        if(value.containsVariable()) {
//            return new Value(value);
//        } else {
//            Value v = Value.lookup(value);
//            if(v == null) {
//                v = new Value(value);
//                Value.store(value, v);
//            }            
//            return v;
//        }
//    }
//    
//    
//    public static Value getValue(double value) {
//        Double d = new Double(value);
//        Value v = Value.lookup(d);
//        if(v == null) {
//            v = new Value(value);
//            Value.store(d, v);
//        }            
//        return v;
//    }
    
    
    public static Value getValue(final boolean value) {
        return (value ? Value.V_TRUE : Value.V_FALSE);
    }
    
    
    public static Value getStringValue(final String value) {
        final Value v = new Value(value, Type.STRING);
        return v;
    }
    
    private Value(final Object value, final Type type) {
        this.value = value;
        this.type = type;
        this.noNumber();
        this.setNetValue();
        this.setHashCode();
    }
    
    
    private Value(final Object value, final Type type, final String net) {
        this.value = value;
        this.noNumber();
        this.type = type;
        this.netValue = net;
        this.setHashCode();
    }
    
    private void noNumber() {
        this.number = Double.NaN;
    }
    
    
    private Value(final String value) {
//        System.out.println(value);
        if(PNModel.isNumber(value)) {
            this.number = Double.parseDouble(value);
//            this.value = Double.valueOf(this.number); // TODO
            this.value = null;
            this.type = Type.NUMBER;
        } else if(value.equals("dot")) {
            this.value = "dot";
            this.noNumber();
            this.type = Type.DOT;
        } else if(value.equals("tt")) {
            this.value = Boolean.TRUE;
            this.noNumber();
            this.type = Type.BOOLEAN;
        } else if(value.equals("ff")) {
            this.value = Boolean.FALSE;
            this.noNumber();
            this.type = Type.BOOLEAN;
        } else if(Value.STRING_MATCHER.matcher(value).matches()) {
            this.value = value.substring(1, value.length()-1);
            this.noNumber();
            this.type = Type.STRING;
        } else if(value.startsWith("(")) {
            this.value = new Tuple(value);
            this.noNumber();
            this.type = Type.TUPLE;
        } else if(value.length() == 0) {
            throw new RuntimeException("Cannot handle empty string as a value.");
        } else {
            this.value = value;
            this.noNumber();
            this.type = Type.VARIABLE;
            this.isVariable = true;
        }
        this.setNetValue();
        this.setHashCode();
//        System.out.println(this);
    }
    
    
    public void changeValue(final Value newValue) {
        this.netValue = newValue.netValue;
        this.value = newValue.value;
        this.number = newValue.number;
        this.type = newValue.type;
        this.hash = newValue.hash;
    }
    
    
    public void makeVariable(final String var) {
        this.netValue = var;
        this.value = var;
        this.noNumber();
        this.type = Type.VARIABLE;
        this.isVariable = true;
        this.setHashCode();
    }
    
    
    private Value(final Tuple tuple) {
        this.value = tuple;
        this.type = Type.TUPLE;
        this.noNumber();
        this.setNetValue();
        this.setHashCode();
    }
    
    
    private Value(final double value) {
//        this.value = Double.valueOf(value); // TODO
        this.value = null;
        this.number = value;
        this.type = Type.NUMBER;
        this.setNetValue();
        this.setHashCode();
    }
    

    private Value(final boolean value) {
        this.value = (value ? Boolean.TRUE : Boolean.FALSE);
        this.noNumber();
        this.type = Type.BOOLEAN;
        this.setNetValue();
        this.setHashCode();
    }
    
    
//    public Object getValue() {
//        return this.value;
//    }
    
    
    public int compareTo(final Value that) {
        // ANY < DOT < BOOLEAN < NUMBER < STRING < VARIABLE < TUPLE
        switch(this.type) {
            case ANY:
                return that.isAny() ? 0 : -1;
            case DOT:
                return that.isAny() ? 1 : (that.isDot() ? 0 : -1);
            case BOOLEAN:
                if(that.isAny() || that.isDot()) {
                    return 1;
                } else if(that.isBoolean()) {
                    // false < true
                    final boolean b = this.getBoolean();
                    return b == that.getBoolean() ? 0 : (b ? 1 : -1);
                } else {
                    return -1;
                }
            case NUMBER:
                if(that.isAny() || that.isDot() || that.isBoolean()) {
                    return 1;
                } else if(that.isNumber()) {
                    // natural order
                    return this.getNumber() == that.getNumber() ? 0
                                    : (this.getNumber() < that.getNumber() ? -1 : 1);
                } else {
                    return -1;
                }
            case STRING:
                if(that.isVariable() || that.isTuple()) {
                    return -1;
                } else if(that.isString()) {
                    // natural order
                    return this.getString().compareTo(that.getString());
                } else {
                    return 1;
                }
            case VARIABLE:
                if(that.isTuple()) {
                    return -1;
                } else if(that.isVariable()) {
                    // natural order
                    return this.getVariable().compareTo(that.getVariable());
                } else {
                    return 1;
                }
            case TUPLE:
                if(that.isTuple()) {
                    // order by size
                    // order lexicographically by elements
                    final Tuple t1 = this.getTuple();
                    final Tuple t2 = that.getTuple();
                    if(t1.size() != t2.size()) {
                        return t1.size() < t2.size() ? -1 : 1;
                    }
                    for(int i = 0, n = t1.size(); i < n; ++i) {
                        final int c = t1.getValue(i).compareTo(t2.getValue(i));
                        if(c != 0) {
                            return c;
                        }
                    }
                    return 0;
                } else {
                    return 1;
                }
            default:
                return 1;
        }
    }
    
    public Type getType() {
        return this.type;
    }
    
    
    public boolean isDot() {
        return this.type == Type.DOT;
    }
    
    
    public boolean isString() {
        return this.type == Type.STRING;
    }

    
    public boolean isNumber() {
        return this.type == Type.NUMBER;
    }
    
    
    public boolean isVariable() {
        return this.type == Type.VARIABLE;
    }
    
    
    public boolean isHiddenVariable() {
        return this.isVariable;
    }
    
    
    public boolean isBoolean() {
        return this.type == Type.BOOLEAN;
    }
    
    
    public boolean isTuple() {
        return this.type == Type.TUPLE;
    }
    
    
    public Tuple getTuple() {
        return (Tuple)this.value;
    }
    
    
    public boolean getBoolean() {
        return ((Boolean)this.value).booleanValue();
    }
    
    
    public double getNumber() {
        return this.number;
    }
    
    
    public Value copy() {
        final Value v = new Value(this.value, this.type, this.netValue);
        v.isVariable = this.isVariable;
        v.number = this.number;
        return v;
    }
    
    
    @Override
    public Value clone() {
        return Value.getValue(this);
    }
    
    
    public String getDot() {
        return "dot";
    }
    
    
    public String getVariable() {
        return (String)this.value;
    }
    
    
    public String getString() {
        return (String)this.value;
    }
    
    
    public void setType(final Type type) {
        this.type = type;
    }
    
    
    @Override
    public boolean equals(final Object o) {
        return (o instanceof Value) && this.equals((Value)o);
    }
    
    
    public boolean equals(final Value v) {
        return (v == this) || ((v != null) && ((this.type == v.type)
                        && (this.number == v.number || (Double.isNaN(this.number) && Double.isNaN(v.number)))
                        && (this.value == null ? v.value == null : this.value.equals(v.value))));
        //return this.value.equals(v.value);
    }
    
    
    @Override
    public String toString() {
        final String typeName;
        switch(this.type) {
            case BOOLEAN :
                typeName = "Bool";
                break;
            case DOT:
                typeName = "Dot";
                break;
            case NUMBER:
                typeName = "Num";
                break;
            case STRING:
                typeName = "Str";
                break;
            case VARIABLE:
                typeName = "Var";
                break;
            case TUPLE:
                typeName = "Tuple";
                break;
            case ANY:
                typeName = "Any";
                break;
            default :
                typeName = "Unknown";
                break;
        }
        return typeName + "(" + this.getNetValue() + ")";
    }
    
    
    public String getNetValue() {
        return this.netValue;
    }
    
    
    public void setNetValue() {
        switch(this.type) {
            case BOOLEAN :
                this.netValue = (this.getBoolean() ? "tt" : "ff");
                break;
            case DOT:
                this.netValue = this.getDot();
                break;
            case NUMBER:
                this.netValue = Value.number2String(this.getNumber());
                break;
            case STRING:
                this.netValue = IPetriNetTool.STRING_SEPARATOR + this.getString() + IPetriNetTool.STRING_SEPARATOR;
                break;
            case VARIABLE:
                this.netValue = this.getVariable();
                break;
            case TUPLE:
                this.netValue = this.getTuple().toString();
                break;
            case UNKNOWN:
                this.netValue = "null";
                break;
            case ANY:
                this.netValue = "*";
                break;
            default :
                this.netValue = "0";
                break;
        }
    }
    
    
    private void setHashCode() {
        switch(this.type) {
            case BOOLEAN :
                this.hash = this.getBoolean() ? 2 : 1;
                break;
            case DOT:
                this.hash = 3;
                break;
            case NUMBER:
                this.hash = (int)this.getNumber() << 2;
                break;
            case STRING:
                this.hash = this.getNetValue().hashCode();
                break;
            case VARIABLE:
                this.hash = this.getVariable().hashCode();
                break;
            case TUPLE:
                this.hash = this.getTuple().hashCode();
                break;
            case UNKNOWN:
            default :
                this.hash = 0;
                break;
        }        
    }
    
    
    @Override
    public int hashCode() {
        return this.hash;
    }
    
    
    public static String number2String(final double d) {
        if(d == Double.POSITIVE_INFINITY) {
            return "inf";
        } else if(d == Double.NEGATIVE_INFINITY) {
            return "-inf";
        } else if(d == (int)d) {
            return "" + (int)d;
        } else {
            return "" + d;
        }
    }
}
