/*
 * Project:  P-UMLaut
 * Filename: ValueList.java
 * Created:  27.06.2006
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.net;

import java.util.AbstractList;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.List;
import java.util.RandomAccess;


/*
 * @(#)ValueList.java 1.49 05/03/03
 * 
 * Copyright 2005 Sun Microsystems, Inc. All rights reserved. SUN
 * PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/**
 * Resizable-array implementation of the <tt>List</tt> interface. Implements
 * all optional list operations, and permits all elements, including
 * <tt>null</tt>. In addition to implementing the <tt>List</tt> interface,
 * this class provides methods to manipulate the size of the array that is used
 * internally to store the list. (This class is roughly equivalent to
 * <tt>Vector</tt>, except that it is unsynchronized.)
 * <p>
 * 
 * The <tt>size</tt>, <tt>isEmpty</tt>, <tt>get</tt>, <tt>set</tt>,
 * <tt>iterator</tt>, and <tt>listIterator</tt> operations run in constant
 * time. The <tt>add</tt> operation runs in <i>amortized constant time</i>,
 * that is, adding n elements requires O(n) time. All of the other operations
 * run in linear time (roughly speaking). The constant factor is low compared to
 * that for the <tt>LinkedList</tt> implementation.
 * <p>
 * 
 * Each <tt>ValueList</tt> instance has a <i>capacity</i>. The capacity is the
 * size of the array used to store the elements in the list. It is always at
 * least as large as the list size. As elements are added to an ValueList, its
 * capacity grows automatically. The details of the growth policy are not
 * specified beyond the fact that adding an element has constant amortized time
 * cost.
 * <p>
 * 
 * An application can increase the capacity of an <tt>ValueList</tt> instance
 * before adding a large number of elements using the <tt>ensureCapacity</tt>
 * operation. This may reduce the amount of incremental reallocation.
 * <p>
 * 
 * <strong>Note that this implementation is not synchronized.</strong> If
 * multiple threads access an <tt>ValueList</tt> instance concurrently, and at
 * least one of the threads modifies the list structurally, it <i>must</i> be
 * synchronized externally. (A structural modification is any operation that
 * adds or deletes one or more elements, or explicitly resizes the backing
 * array; merely setting the value of an element is not a structural
 * modification.) This is typically accomplished by synchronizing on some object
 * that naturally encapsulates the list. If no such object exists, the list
 * should be "wrapped" using the <tt>Collections.synchronizedList</tt> method.
 * This is best done at creation time, to prevent accidental unsynchronized
 * access to the list:
 * 
 * <pre>
 *   List list = Collections.synchronizedList(new ValueList(...));
 * </pre>
 * 
 * <p>
 * 
 * The iterators returned by this class's <tt>iterator</tt> and
 * <tt>listIterator</tt> methods are <i>fail-fast</i>: if list is
 * structurally modified at any time after the iterator is created, in any way
 * except through the iterator's own remove or add methods, the iterator will
 * throw a ConcurrentModificationException. Thus, in the face of concurrent
 * modification, the iterator fails quickly and cleanly, rather than risking
 * arbitrary, non-deterministic behavior at an undetermined time in the future.
 * <p>
 * 
 * Note that the fail-fast behavior of an iterator cannot be guaranteed as it
 * is, generally speaking, impossible to make any hard guarantees in the
 * presence of unsynchronized concurrent modification. Fail-fast iterators throw
 * <tt>ConcurrentModificationException</tt> on a best-effort basis. Therefore,
 * it would be wrong to write a program that depended on this exception for its
 * correctness: <i>the fail-fast behavior of iterators should be used only to
 * detect bugs.</i>
 * <p>
 * 
 * This class is a member of the <a href="{@docRoot}/../guide/collections/index.html">
 * Java Collections Framework</a>.
 * 
 * @author Josh Bloch
 * @author Neal Gafter
 * @version 1.49, 03/03/05
 * @see Collection
 * @see List
 * @see LinkedList
 * @see Vector
 * @see Collections#synchronizedList(List)
 * @since 1.2
 */

public final class ValueList extends AbstractList<Value> implements List<Value>,
                RandomAccess, Cloneable, java.io.Serializable {

    private static final long serialVersionUID = 8683452581122892189L;

    /**
     * The array buffer into which the elements of the ValueList are stored. The
     * capacity of the ValueList is the length of this array buffer.
     */
    private transient Value[] elementData;

    /**
     * The size of the ValueList (the number of elements it contains).
     * 
     * @serial
     */
    private int size;


    /**
     * Constructs an empty list with the specified initial capacity.
     * 
     * @param initialCapacity the initial capacity of the list.
     * @exception IllegalArgumentException if the specified initial capacity is
     *                negative
     */
    public ValueList(int initialCapacity) {
        super();
        if (initialCapacity < 0)
            throw new IllegalArgumentException("Illegal Capacity: "
                            + initialCapacity);
        this.elementData = new Value[initialCapacity];
    }


    /**
     * Constructs an empty list with an initial capacity of ten.
     */
    public ValueList() {
        this(1);
    }


    /**
     * Constructs a list containing the elements of the specified collection, in
     * the order they are returned by the collection's iterator. The
     * <tt>ValueList</tt> instance has an initial capacity of 110% the size of
     * the specified collection.
     * 
     * @param c the collection whose elements are to be placed into this list.
     * @throws NullPointerException if the specified collection is null.
     */
    public ValueList(Collection<? extends Value> c) {
        size = c.size();
        // Allow 10% room for growth
        elementData = new Value[(int) Math.min((size * 110L) / 100,
                        Integer.MAX_VALUE)];
        c.toArray(elementData);
    }


    /**
     * Trims the capacity of this <tt>ValueList</tt> instance to be the list's
     * current size. An application can use this operation to minimize the
     * storage of an <tt>ValueList</tt> instance.
     */
    public void trimToSize() {
        modCount++;
        int oldCapacity = elementData.length;
        if (size < oldCapacity) {
            Value[] oldData = elementData;
            elementData = new Value[size];
            System.arraycopy(oldData, 0, elementData, 0, size);
        }
    }


    /**
     * Increases the capacity of this <tt>ValueList</tt> instance, if
     * necessary, to ensure that it can hold at least the number of elements
     * specified by the minimum capacity argument.
     * 
     * @param minCapacity the desired minimum capacity.
     */
    public void ensureCapacity(int minCapacity) {
        this.modCount++;
        int oldCapacity = this.elementData.length;
        if (minCapacity > oldCapacity) {
            Value[] oldData = this.elementData;
            int newCapacity = ((oldCapacity * 3) >> 1) + 1;
            if (newCapacity < minCapacity)
                newCapacity = minCapacity;
            this.elementData = new Value[newCapacity];
            System.arraycopy(oldData, 0, this.elementData, 0, size);
        }
    }


    /**
     * Returns the number of elements in this list.
     * 
     * @return the number of elements in this list.
     */
    public int size() {
        return this.size;
    }


    /**
     * Tests if this list has no elements.
     * 
     * @return <tt>true</tt> if this list has no elements; <tt>false</tt>
     *         otherwise.
     */
    public boolean isEmpty() {
        return this.size == 0;
    }


    /**
     * Returns <tt>true</tt> if this list contains the specified element.
     * 
     * @param elem element whose presence in this List is to be tested.
     * @return <code>true</code> if the specified element is present;
     *         <code>false</code> otherwise.
     */
    public boolean contains(Object elem) {
        return this.indexOf(elem) >= 0;
    }


    /**
     * Searches for an occurrence of the given argument, testing for
     * equality using the <tt>equals</tt> method.
     * 
     * @param elem an object.
     * @return the index of an occurrence of the argument in this list;
     *         returns <tt>-1</tt> if the object is not found.
     * @see Object#equals(Object)
     */
    public int indexOf(final Object elem) {
        if(!this.isEmpty() && (elem instanceof Value)) {
            return this.indexOf((Value)elem, 0);
        } else {
            return -1;
        }
    }
    
    protected int indexOf(final Value elem, int from) {
        if(this.size() <= from) {
            return -1;
        }
        
        if(elem.isAny()) {
            return this.elementData[from].isAny() ? from : -1;
        } else {
            int to = this.size()-1;
            do {
                int middle = (from+to)>>1;
                final Value v = this.elementData[middle];
                final int comp = elem.compareTo(v);
                if(comp == 0) {
                    return middle;
                } else if(comp < 0) {
                    to = middle-1;
                } else {
                    from = middle+1;
                }
            } while(from <= to);
            return -1;
        }
    }


    /**
     * Returns the index of the last occurrence of the specified object in this
     * list.
     * 
     * @param elem the desired element.
     * @return the index of the last occurrence of the specified object in this
     *         list; returns -1 if the object is not found.
     */
    public int lastIndexOf(final Object elem, final int startIndex) {
        if(!this.isEmpty() && (elem instanceof Value)) {
            return this.lastIndexOf((Value)elem, startIndex, this.size()-1);
        }
        return -1;
    }
    
    public int lastIndexOf(final Object o) {
        return this.lastIndexOf(o, 0);
    }


    /**
     * Returns a shallow copy of this <tt>ValueList</tt> instance. (The
     * elements themselves are not copied.)
     * 
     * @return a clone of this <tt>ValueList</tt> instance.
     */
    public ValueList clone() {
        try {
            ValueList v = (ValueList) super.clone();
            v.elementData = new Value[size];
            System.arraycopy(elementData, 0, v.elementData, 0, size);
            v.modCount = 0;
            return v;
        } catch (CloneNotSupportedException e) {
            // this shouldn't happen, since we are Cloneable
            throw new InternalError();
        }
    }
    
    
    @Override
    public boolean equals(final Object o) {
        return (o != null) && ((o.getClass() == ValueList.class) ? this.equals((ValueList)o) : super.equals(o)); 
    }
    
    
    public boolean equals(final ValueList list) {
        if(this == list) {
            return true;
        }
        if(list == null) {
            return false;
        }
        if(this.size != list.size) {
            return false;
        }
        
        for(int i = 0; i < this.size; ++i) {
            final Value o1 = this.elementData[i];
            final Value o2 = list.elementData[i];
            if(!o1.equals(o2)) {
                return false;
            }
        }
        
        return true;
    }


    /**
     * Returns an array containing all of the elements in this list in the
     * correct order.
     * 
     * @return an array containing all of the elements in this list in the
     *         correct order.
     */
    public Object[] toArray() {
        Object[] result = new Object[size];
        System.arraycopy(elementData, 0, result, 0, size);
        return result;
    }


    /**
     * Returns an array containing all of the elements in this list in the
     * correct order; the runtime type of the returned array is that of the
     * specified array. If the list fits in the specified array, it is returned
     * therein. Otherwise, a new array is allocated with the runtime type of the
     * specified array and the size of this list.
     * <p>
     * 
     * If the list fits in the specified array with room to spare (i.e., the
     * array has more elements than the list), the element in the array
     * immediately following the end of the collection is set to <tt>null</tt>.
     * This is useful in determining the length of the list <i>only</i> if the
     * caller knows that the list does not contain any <tt>null</tt> elements.
     * 
     * @param a the array into which the elements of the list are to be stored,
     *            if it is big enough; otherwise, a new array of the same
     *            runtime type is allocated for this purpose.
     * @return an array containing the elements of the list.
     * @throws ArrayStoreException if the runtime type of a is not a supertype
     *             of the runtime type of every element in this list.
     */
    @SuppressWarnings("unchecked")
    public <T> T[] toArray(T[] a) {
        if (a.length < size) {
            a = (T[]) java.lang.reflect.Array.newInstance(a.getClass()
                            .getComponentType(), size);
        }
        System.arraycopy(elementData, 0, a, 0, size);
        if (a.length > size) {
            a[size] = null;
        }
        return a;
    }


    // Positional Access Operations

    /**
     * Returns the element at the specified position in this list.
     * 
     * @param index index of element to return.
     * @return the element at the specified position in this list.
     * @throws IndexOutOfBoundsException if index is out of range <tt>(index
     *        &lt; 0 || index &gt;= size())</tt>.
     */
    public Value get(int index) {
        RangeCheck(index);

        return elementData[index];
    }


    /**
     * Replaces the element at the specified position in this list with the
     * specified element.
     * 
     * @param index index of element to replace.
     * @param element element to be stored at the specified position.
     * @return the element previously at the specified position.
     * @throws IndexOutOfBoundsException if index out of range
     *             <tt>(index &lt; 0 || index &gt;= size())</tt>.
     */
    public Value set(int index, Value element) {
        RangeCheck(index);

        final Value oldValue = elementData[index];

        // check order!
        final int comp = element.compareTo(oldValue);
        if((comp != 0)
                && (((index > 0) && (comp < 0) && (this.elementData[index-1].compareTo(element) > 0))
                    || ((index + 1 < this.size()) && (comp > 0) && (this.elementData[index+1].compareTo(element) < 0)))) {
            throw new IllegalArgumentException("Replacing " + element + " at index"
                            + " " + index + " would violate the order!");
        }
        
        elementData[index] = element;
        return oldValue;
    }


    /**
     * Appends the specified element to the end of this list.
     * 
     * @param o element to be appended to this list.
     * @return <tt>true</tt> (as per the general contract of Collection.add).
     */
    public boolean add(final Value elem) {
        int from = 0;
        if(!this.isEmpty() && !elem.isAny()) {
            int to = this.size()-1;
            while(from <= to) {
                int middle = (from+to)>>1;
                final Value v = this.elementData[middle];
                final int comp = elem.compareTo(v);
                if(comp == 0) {
                    from = middle;
                    to = from-1;
                } else if(comp < 0) {
                    to = middle-1;
                } else {
                    from = middle+1;
                }
            }
        }
        
        this.addAt(from, elem);
        return true;
    }


    /**
     * Inserts the specified element at the specified position in this list.
     * Shifts the element currently at that position (if any) and any subsequent
     * elements to the right (adds one to their indices).
     * 
     * @param index index at which the specified element is to be inserted.
     * @param element element to be inserted.
     * @throws IndexOutOfBoundsException if index is out of range
     *             <tt>(index &lt; 0 || index &gt; size())</tt>.
     */
    public void add(int index, Value element) {
        if (index > size || index < 0)
            throw new IndexOutOfBoundsException("Index: " + index + ", Size: "
                            + size);

        // check order!
        if(((index > 0) && (this.elementData[index-1].compareTo(element) > 0))
                || ((index < this.size()) && (this.elementData[index].compareTo(element) < 0))) {
            throw new IllegalArgumentException("Adding " + element + " at index"
                            + " " + index + " would violate the order!");
        }
        
        ensureCapacity(size + 1); // Increments modCount!!
        System.arraycopy(elementData, index, elementData, index + 1, size
                        - index);
        elementData[index] = element;
        size++;
    }
    
    
    protected void addAt(int index, Value element) {
        ensureCapacity(size + 1); // Increments modCount!!
        System.arraycopy(elementData, index, elementData, index + 1, size
                        - index);
        elementData[index] = element;
        size++;
    }


    /**
     * Removes the element at the specified position in this list. Shifts any
     * subsequent elements to the left (subtracts one from their indices).
     * 
     * @param index the index of the element to removed.
     * @return the element that was removed from the list.
     * @throws IndexOutOfBoundsException if index out of range <tt>(index
     *        &lt; 0 || index &gt;= size())</tt>.
     */
    public Value remove(int index) {
        RangeCheck(index);

        modCount++;
        Value oldValue = elementData[index];

        int numMoved = size - index - 1;
        if (numMoved > 0)
            System.arraycopy(elementData, index + 1, elementData, index,
                            numMoved);
        elementData[--size] = null; // Let gc do its work

        return oldValue;
    }
    
    
    public boolean removeAll(Collection<?> c) {
        final boolean any = this.contains(Value.V_ANY) && c.contains(Value.V_ANY);
        if(this == c) {
            this.clear();
            if(any) {
                this.add(Value.V_ANY);
            }
            return true;
        } else if(this.isEmpty() || c.isEmpty()) {
            return false;
        }
        
        boolean modified = false;
        boolean sorted = c.getClass() == ValueList.class;
        if(sorted) {
            final Value[] elems = ((ValueList)c).elementData;
            int from = 0;
            for(int i = 0; i < c.size(); ++i) {
                final int index = this.removeAllOf(elems[i], from);
                if(index >= 0) {
                    from = index;
                    modified = true;
                }
            }
        } else {
            for(final Object v : c) {
                if(this.removeAllOf(v)) {
                    modified = true;
                }
            }
        }
        
        if(any) {
            this.add(Value.V_ANY);
        }
        
        return modified;
    }
    
    
    public boolean containsAll(Collection<?> c) {
        if(c.isEmpty() || (this == c)) {
            return true;
        } else if(this.isEmpty()) {
            return false;
        }
        
        boolean sorted = c.getClass() == ValueList.class;
        if(sorted) {
            final Value[] elems = ((ValueList)c).elementData;
            if(this.size() == c.size()) {
                return this.equals((ValueList)c);
            } else {
                int index = 0;
                for(int i = 0; i < c.size(); ++i) {
                    Value v = elems[i];
                    index = this.indexOf(v, index);
                    if(index < 0) {
                        return false;
                    }
                    i = ValueList.lastIndexOf(elems, v, i, c.size()-1);
                }
            }
        } else {
            for(final Object v : c) {
                if(!this.contains(v)) {
                    return false;
                }
            }
        }
        return true;
    }
    
    
    public boolean retainAll(final Collection<?> c) {
        if(this.isEmpty() || (this == c) || (c.contains(Value.V_ANY))) {
            return false;
        } else if(c.isEmpty()) {
            this.clear();
            return true;
        }
        boolean modified = false;
        boolean sorted = c.getClass() == ValueList.class;
        if(sorted) {
            final ValueList l = (ValueList)c;
            int from = 0;
            for(int i = 0; i < this.size(); ++i) {
                final Value v = this.elementData[i];
                final int right = this.lastIndexOf(v, i, this.size()-1);
                
                final int index = l.indexOf(v, from);
                if(index < 0) {
                    this.removeRange(i, right+1);
                    --i;
                    modified = true;
                } else {
                    i = right;
                    from = index+1;
                }
            }
        } else {
            for(int i = 0; i < this.size(); ++i) {
                final Value v = this.elementData[i];
                final int right = this.lastIndexOf(v, i, this.size()-1);
                if(!c.contains(v)) {
                    this.removeRange(i, right);
                    --i;
                    modified = true;
                } else {
                    i = right;
                }
            }
        }
        return modified;
    }

    
    public void deleteDublicates() {
        for(int i = 0; i < this.size(); ++i) {
            final Value v = this.elementData[i];
            final int to = this.lastIndexOf(v, i, this.size()-1);
            if(to > i) {
                this.removeRange(i+1, to+1);
            }
        }
    }
    
    
    public int count(final Object o) {
        if(this.isEmpty() || !(o instanceof Value)) {
            return 0;
        }
        final Value elem = (Value)o;

        int from = 0;
        int to = this.size()-1;
        if(elem.isAny()) {
            if(!this.elementData[0].isAny()) {
                return 0;
            }
            to = this.lastIndexOf(elem);
        } else {
            boolean searching = true;
            int middle;
            do {
                middle = (from+to)>>1;
                final Value v = this.elementData[middle];
                final int comp = elem.compareTo(v);
                if(comp == 0) {
                    searching = false;
                    break;
                } else if(comp < 0) {
                    to = middle-1;
                } else {
                    from = middle+1;
                }
            } while(from <= to);
            if(searching) {
                return 0;
            }
            from = this.firstIndexOf(elem, from, middle);
            to = this.lastIndexOf(elem, middle, to);
        }
        
        return to - from + 1;
    }
    

    /**
     * Removes a single instance of the specified element from this list, if it
     * is present (optional operation). More formally, removes an element
     * <tt>e</tt> such that <tt>(o==null ? e==null :
     * o.equals(e))</tt>, if
     * the list contains one or more such elements. Returns <tt>true</tt> if
     * the list contained the specified element (or equivalently, if the list
     * changed as a result of the call).
     * <p>
     * 
     * @param o element to be removed from this list, if present.
     * @return <tt>true</tt> if the list contained the specified element.
     */
    public boolean remove(final Object o) {
        if(this.isEmpty() || !(o instanceof Value)) {
            return false;
        }
        final Value v = (Value)o;
        if(v.isAny()) {
            if(this.elementData[0].isAny()) {
                this.fastRemove(0);
                return true;
            } else {
                return false;
            }
        } else {
            final int index = this.indexOf(v);
            if(index >= 0) {
                this.fastRemove(index);
                return true;
            }
            return false;
        }
    }
    
    public boolean removeAllOf(final Object o) {
        if(!(o instanceof Value) || this.isEmpty()) {
            return false;
        }
        return this.removeAllOf((Value)o, 0) >= 0;
    }
    
    protected int removeAllOf(final Value elem, int from) {
        int to = this.size()-1;
        if(from > to) {
            return -1;
        }
        
        if(elem.isAny()) {
            if(!this.elementData[0].isAny()) {
                return -1;
            }
            to = this.lastIndexOf(elem);
        } else {
            boolean searching = true;
            int middle;
            do {
                middle = (from+to)>>1;
                final Value v = this.elementData[middle];
                final int comp = v.compareTo(elem);
                if(comp == 0) {
                    searching = false;
                    break;
                } else if(comp < 0) {
                    to = middle-1;
                } else {
                    from = middle+1;
                }
            } while(from <= to);
            if(searching) {
                return -1;
            }
            from = this.firstIndexOf(elem, from, middle);
            to = this.lastIndexOf(elem, middle, to);
        }
        
        this.removeRange(from, to+1);
        return from;
    }
    
    
    protected int lastIndexOf(final Value v) {
        return this.lastIndexOf(v, 0, this.size()-1);
    }
    
    
    protected int lastIndexOf(final Value v, int from, int to) {
        return ValueList.lastIndexOf(this.elementData, v, from, to);
    }
    
    
    public static int lastIndexOf(final Value[] elements, final Value v, int from, int to) {
        while(from < to) {
            final int middle = (from + to + 1)>>1; // round up! important to terminate
            final int comp = elements[middle].compareTo(v);
            if(comp <= 0) {
                from = middle;
            } else {
                to = middle-1;
            }
        }
        return from;
    }
    
    
    public int firstIndexOf(final Object o, final int startIndex) {
        if(!this.isEmpty() && (o instanceof Value)) {
            return this.firstIndexOf((Value)o, startIndex, this.size()-1);
        }
        return -1;
    }
    
    
    public int firstIndexOf(final Object o) {
        return this.firstIndexOf(o, 0);
    }
    
    
    protected int firstIndexOf(final Value v) {
        return this.firstIndexOf(v, 0, this.size()-1);
    }
    
    
    protected int firstIndexOf(final Value v, int from, int to) {
        while(from < to) {
            final int middle = (from + to)>>1; // round down! important to terminate
            final int comp = this.elementData[middle].compareTo(v);
            if(comp >= 0) {
                to = middle;
            } else {
                from = middle+1;
            }
        }
        return from;
    }


    /*
     * Private remove method that skips bounds checking and does not return the
     * value removed.
     */
    private void fastRemove(int index) {
        modCount++;
        int numMoved = size - index - 1;
        if (numMoved > 0)
            System.arraycopy(elementData, index + 1, elementData, index,
                            numMoved);
        elementData[--size] = null; // Let gc do its work
    }


    /**
     * Removes all of the elements from this list. The list will be empty after
     * this call returns.
     */
    public void clear() {
        this.modCount++;
        for(int i = 0, n = this.size; i < n; ++i) {
            this.elementData[i] = null;
        }
        this.size = 0;
    }


    /**
     * Appends all of the elements in the specified Collection to the end of
     * this list, in the order that they are returned by the specified
     * Collection's Iterator. The behavior of this operation is undefined if the
     * specified Collection is modified while the operation is in progress.
     * (This implies that the behavior of this call is undefined if the
     * specified Collection is this list, and this list is nonempty.)
     * 
     * @param c the elements to be inserted into this list.
     * @return <tt>true</tt> if this list changed as a result of the call.
     * @throws NullPointerException if the specified collection is null.
     */
    public boolean addAll(Collection<? extends Value> c) {
        final int numNew = c.size();
        if(numNew == 0) {
            return false;
        }
        
        boolean sorted = c.getClass() == ValueList.class;
        final Value[] a = sorted && (this.isEmpty() || numNew == 1 || c != this)
                            ? ((ValueList)c).elementData : (Value[])c.toArray();
        
        ensureCapacity(size + numNew); // Increments modCount
        
        if(this.isEmpty() && sorted) {
            System.arraycopy(a, 0, this.elementData, 0, numNew);
            this.size += numNew;
            return true;
        }
        
        int from = 0;
        for(int i = 0; i < numNew; ++i) {
            if(!sorted) {
                from = 0;
            }
            final Value elem = a[i];
            final int count = ValueList.lastIndexOf(a, elem, i, numNew-1) - i + 1;
            if(!elem.isAny() && !this.isEmpty()) {
                int to = this.size()-1;
                while(from <= to) {
                    int middle = (from+to)>>1;
                    final Value v = this.elementData[middle];
                    final int comp = elem.compareTo(v);
                    if(comp == 0) {
                        from = middle;
                        to = from-1;
                    } else if(comp < 0) {
                        to = middle-1;
                    } else {
                        from = middle+1;
                    }
                }
            }
            
            System.arraycopy(this.elementData, from, this.elementData,
                            from + count, this.size - from);
            size += count;
            if(count == 1) {
                elementData[from] = elem;
            } else {
                System.arraycopy(a, i, elementData, from, count);
            }
            i += count-1;
        }
        
        
        return true;
    }


    /**
     * Inserts all of the elements in the specified Collection into this list,
     * starting at the specified position. Shifts the element currently at that
     * position (if any) and any subsequent elements to the right (increases
     * their indices). The new elements will appear in the list in the order
     * that they are returned by the specified Collection's iterator.
     * 
     * @param index index at which to insert first element from the specified
     *            collection.
     * @param c elements to be inserted into this list.
     * @return <tt>true</tt> if this list changed as a result of the call.
     * @throws IndexOutOfBoundsException if index out of range <tt>(index
     *        &lt; 0 || index &gt; size())</tt>.
     * @throws NullPointerException if the specified Collection is null.
     */
    public boolean addAll(int index, Collection<? extends Value> c) {
        if(c.isEmpty()) {
            return false;
        }
        final boolean okay;
        final Value[] elems;
        if((c.getClass() == ValueList.class)) {
            elems = ((ValueList)c).elementData;
            if(this.isEmpty()) {
                okay = true;
            } else {
                okay = ((index == 0) || (this.elementData[index-1].compareTo(elems[0]) <= 0))
                        && ((index == this.size()) || (this.elementData[index].compareTo(elems[c.size()-1]) >= 0));
            }
        } else {
            okay = false;
            elems = null;
        }
        
        if(okay) {
            this.ensureCapacity(this.size() + c.size());
            System.arraycopy(this.elementData, index, this.elementData, index + c.size(), this.size()-index);
            System.arraycopy(elems, 0, this.elementData, index, c.size());
            this.size += c.size();
            return true;
        } else {
            throw new IllegalArgumentException("Adding the collection would "
                            + "violate the lists order.");
        }
    }
    

    /**
     * Removes from this List all of the elements whose index is between
     * fromIndex, inclusive and toIndex, exclusive. Shifts any succeeding
     * elements to the left (reduces their index). This call shortens the list
     * by <tt>(toIndex - fromIndex)</tt> elements. (If
     * <tt>toIndex==fromIndex</tt>, this operation has no effect.)
     * 
     * @param fromIndex index of first element to be removed.
     * @param toIndex index after last element to be removed.
     */
    protected void removeRange(int fromIndex, int toIndex) {
        if(toIndex <= fromIndex) {
            // ignore
        } else if(fromIndex + 1 == toIndex) {
            this.fastRemove(fromIndex);
        } else {
            modCount++;
            int numMoved = size - toIndex;
            System.arraycopy(elementData, toIndex, elementData, fromIndex, numMoved);
        
            // Let gc do its work
            int newSize = size - (toIndex - fromIndex);
            while (size != newSize)
                elementData[--size] = null;
        }
    }


    /**
     * Check if the given index is in range. If not, throw an appropriate
     * runtime exception. This method does *not* check if the index is negative:
     * It is always used immediately prior to an array access, which throws an
     * ArrayIndexOutOfBoundsException if index is negative.
     */
    private void RangeCheck(int index) {
        if (index >= size)
            throw new IndexOutOfBoundsException("Index: " + index + ", Size: "
                            + size);
    }


    /**
     * Save the state of the <tt>ValueList</tt> instance to a stream (that is,
     * serialize it).
     * 
     * @serialData The length of the array backing the <tt>ValueList</tt>
     *             instance is emitted (int), followed by all of its elements
     *             (each an <tt>Object</tt>) in the proper order.
     */
    private void writeObject(java.io.ObjectOutputStream s)
                    throws java.io.IOException {
        int expectedModCount = modCount;
        // Write out element count, and any hidden stuff
        s.defaultWriteObject();

        // Write out array length
        s.writeInt(elementData.length);

        // Write out all elements in the proper order.
        for (int i = 0; i < size; i++)
            s.writeObject(elementData[i]);

        if (modCount != expectedModCount) {
            throw new ConcurrentModificationException();
        }
    }
    
}
