/*
 * Project:  P-UMLaut
 * Filename: Binding.java
 * Created:  24.07.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.util;

import petruchio.sim.pnmodel.PNModel;
import petruchio.sim.pnmodel.net.Transition;
import petruchio.sim.pnmodel.net.Value;
import petruchio.sim.pnmodel.net.ValueList;

import java.util.Iterator;
import java.util.Map;
import java.util.List;
import java.util.Set;


public final class Binding {
    private static int curObjNr = Integer.MIN_VALUE;
    
    private final int objNr;
    private double earliest;
    private double latest;
    private double priority;
    private double rate;
    private double weight;
    private double currentWeight;
    private double servers;
    private double creationTime;
    private double age;
    private boolean isImmediate;
    private boolean useAgeMemory;
    private StringList freeVariables;
    private StringList disposable;
    private Map<String, Value> binding;
    private Map<String, ValueList> constraints;
    private Transition trans;
    
    public Binding() {
        this.objNr = Binding.curObjNr++;
        this.binding = Pool.getPool().getMap();
        this.constraints = Pool.getPool().getMap();
        this.freeVariables = new StringList();//Pool.getPool().getList();
        this.disposable = new StringList();//Pool.getPool().getList();
        this.clear();
    }
    
    
    public void clear() {
        this.disposable.clear();
        this.binding.clear();
        this.clearConstraints();
        this.trans = null;
        this.isImmediate = false;
        this.priority = 0.0d;
        this.earliest = 0.0d;
        this.rate = Double.NaN;
        this.weight = Double.NaN;
        this.servers = 1.0d;
        this.currentWeight = 0.0d;
        this.useAgeMemory = false;
        this.creationTime = 0.0d;
        this.age = 0.0d;
        this.latest = Double.POSITIVE_INFINITY;
    }
    
    
    @Override
    public Binding clone() {
        final Binding copy = Pool.getPool().getBinding();
        copy.binding.putAll(this.binding);
        if(!this.constraints.isEmpty()) {
            for(final Map.Entry<String, ValueList> e : this.constraints.entrySet()) {
                String var = e.getKey();
                final ValueList values = e.getValue();
                copy.constraints.put(var, Pool.getPool().getValueList(values));
            }
        }
        copy.freeVariables.addAll(this.freeVariables);
        copy.disposable.addAll(this.disposable);
        copy.trans = this.trans;
        copy.isImmediate = this.isImmediate;
        copy.priority = this.priority;
        copy.earliest = this.earliest;
        copy.latest = this.latest;
        copy.rate = this.rate;
        copy.weight = this.weight;
        copy.currentWeight = this.currentWeight;
        copy.servers = this.servers;
        copy.useAgeMemory = this.useAgeMemory;
        copy.creationTime = this.creationTime;
        copy.age = this.age;
        return copy;
    }
    
    
    public void setCreationTime(final double time) {
        this.creationTime = time;
    }
    
    
    public double getCreationTime() {
        return this.creationTime;
    }
    
    
    public void setUseAgeMemory(final boolean useAgeMemory) {
        this.useAgeMemory = useAgeMemory;
    }
    
    
    public boolean useAgeMemory() {
        return this.useAgeMemory;
    }
    
    
    public int getObjectNumber() {
        return this.objNr;
    }
    
    
    public int hashCode() {
        return this.binding.hashCode();
    }
    
    
    public boolean isImmediate() {
        return this.isImmediate;
    }
    
    
    public void setIsImmediate(final boolean isImmediate) {
        this.isImmediate = isImmediate;
    }
    
    
    public boolean equalBindings(final Binding b) {
        return (this == b) || ((b != null) && this.binding.equals(b.binding));
    }
    
    
    public void setRate(final double rate) {
        this.rate = rate;
    }
    
    
    public double getRate() {
        return this.rate;
    }
    
    
    public void setWeight(final double weight) {
        this.weight = weight;
    }
    
    
    public double getWeight() {
        return this.weight;
    }
    
    
    public void setCurrentWeight(final double currentWeight) {
        this.currentWeight = currentWeight;
    }
    
    
    public double getCurrentWeight() {
        return this.currentWeight;
    }
    
    
    public void setServers(final double servers) {
        this.servers = servers;
    }
    
    
    public double getServers() {
        return this.servers;
    }
    
    
    public void setEarliestFiring(final double time) {
        this.earliest = time;
    }
    
    
    public double getEarliestFiring() {
        return this.earliest;
    }
    
    
    public void setLatestFiring(final double time) {
        this.latest = time;
    }
    
    
    public double getLatestFiring() {
        return this.latest;
    }
    
    
    public boolean hasFreeVariables() {
        return !this.freeVariables.isEmpty();
    }
    
    
    public Map<String, ValueList> getConstraints() {
        return this.constraints;
    }
    
    
    public boolean hasBinding(String var) {
        return this.binding.containsKey(var);
    }
    
    
    public void addConstraints(final Map<String, ValueList> map) {
        if(map != null) {
            for(final Map.Entry<String, ValueList> e : map.entrySet()) {
                this.constrain(e.getKey(), e.getValue(), true);
            }
        }
    }
    
    public Set<Map.Entry<String, Value>> getBindings() {
        return this.binding.entrySet();
    }
    
    public Set<String> getBoundVariables() {
        return this.binding.keySet();
    }
    
    
    public void setTransition(final Transition trans) {
        this.trans = trans;
    }
    
    public Transition getTransition() {
        return this.trans;
    }
    
    
    public boolean hasBindings() {
        return !this.binding.isEmpty();
    }
    
    
    public void clearConstraints() {
        this.constraints.clear();
        this.freeVariables.clear();
    }
    
    
    public ValueList getConstraint(final String var) {
        return this.constraints.get(var);
    }
    
    
    public void removePossibleValue(final String var, final Value val) {
        final List<Value> col = this.getConstraint(var);
        if(col != null) {
            col.remove(val);
        }
    }
    
    
    public ValueList removeConstraintNoVar(final String var) {
        return this.constraints.remove(var);
    }

    public ValueList removeConstraint(final String var) {
        this.freeVariables.remove(var);
        return this.removeConstraintNoVar(var);
    }
    
    
    public ValueList setConstraint(final String var, final ValueList values) {
        if(!this.freeVariables.contains(var)) {
            this.freeVariables.add(var);
        }
        final ValueList old = this.getConstraint(var);
        this.constraints.put(var, values);
        return old;
    }
    
    
    public boolean hasConstraint(final String var) {
        return this.freeVariables.contains(var);
    }
    
    
    public boolean constrain(final String var, final ValueList values, final boolean isSet) {
        ValueList c = this.getConstraint(var);
        
        if(c != null) {
            if(values.contains(Value.V_ANY)) {
                // constraining with everything changes nothing :-)
                // ignore
            } else if(c.contains(Value.V_ANY)) {
                // cutting everything down to "values"
                c.clear();
                c.addAll(values);
            } else {
                c.retainAll(values);
            }
        } else {
            this.freeVariables.add(var);
            c = Pool.getPool().getValueList();
            if(isSet) {
                c.addAll(values);
            } else {
                for(int i = 0, n = values.size(); i < n; ++i) {
                    final Value value = values.get(i);
                    if(!c.contains(value)) {
                        c.add(value);
                    }
                }
            }
            this.setConstraint(var, c);
        }
        
        return !c.isEmpty();
    }
    
    
    public boolean isEmptyConstraint() {
        return this.freeVariables.isEmpty();
    }
    
    
    public StringList getFreeVariables() {
        return this.freeVariables;
    }
    
    
    public void addBinding(final String variable, final Value value) {
        this.binding.put(variable.trim(), value.isHiddenVariable() ? Value.getValue(value) : value);
    }
    
    
    public void removeDisposability(final String var) {
        this.disposable.remove(var);
    }
    
    
    public StringList getDisposable() {
        return this.disposable;
    }
    
    
    public void makeUndisposable() {
        this.disposable.clear();
    }
    
    
    public void dispose() {
        if(!this.disposable.isEmpty()) {
            for(final Iterator<String> it = this.disposable.iterator(); it.hasNext();) {
                final String var = it.next();
                it.remove();
                this.removeBinding(var);
            }
        }
    }
    
    
    public boolean addDisposableBinding(final String variable, final Value value) {
        if(this.hasBinding(variable)) {
            return false;
        } else {
            this.addBinding(variable, value);
            this.disposable.add(variable);
            return true;
        }
    }
    
    
    public Value getBinding(final String variable) {
        return this.binding.get(variable);
    }
    
    
    public void removeBinding(final String variable) {
        this.binding.remove(variable);
    }
 
    
    @Override
    public String toString() {
        return this.toString(true, true);
    }
    
    
    public String toStringSlim() {
        return this.toString(false, false);
    }
    
    
    public double getPriority() {
        return this.priority;
    }


    public void setPriority(double priority) {
        this.priority = priority;
    }


    public String toString(final boolean time, final boolean showprecedence) {
        final StringBuffer buf = new StringBuffer();
        
        if(this.trans != null) {
            buf.append(this.trans.getName());
        }
        buf.append("{");
        if(!this.binding.isEmpty()) {
            for(final Iterator<Map.Entry<String, Value>> it = this.binding.entrySet().iterator(); it.hasNext(); ) {
                final Map.Entry<String, Value> e = it.next();
                final String var = e.getKey();
                final Value val = e.getValue();
                buf.append(var + "/" + val.getNetValue());
                if(it.hasNext()) {
                    buf.append(", ");
                }
            }
        }
        buf.append("}");
        
        if(time) {
            buf.append("[");
            buf.append(Value.number2String(this.earliest));
            if(this.earliest != this.latest) {
                buf.append(",");
                buf.append(Value.number2String(this.latest));
            }
            buf.append("]");
        }
        if(showprecedence) {
            buf.append("[");
            buf.append(Value.number2String(this.priority));
            buf.append("+");
            buf.append(Value.number2String(this.currentWeight));
            buf.append("]");
        }
        
        return buf.toString();
    }
    
    
    public String toPEPString() {
        final StringBuffer buf = new StringBuffer();
        
        if(this.trans != null) {
            buf.append(this.trans.getName());
        }
        for(final Iterator<Map.Entry<String, Value>> it = this.binding.entrySet().iterator(); it.hasNext(); ) {
            final Map.Entry<String, Value> e = it.next();
            final String var = e.getKey();
            final Value val = e.getValue();
            buf.append(var + "=" + val.getNetValue());
            if(it.hasNext()) {
                buf.append(",");
            }
        }
        
        return buf.toString();
    }
    
    
    @Override
    public boolean equals(final Object o) {
        return (o instanceof Binding) && this.equalBindings((Binding)o);
    }
    
    
    public boolean theSame(final Binding o) {
        return (o == this) || ((o != null)
                && (this.getTransition() == o.getTransition())
                && (this.isImmediate == o.isImmediate)
                && (this.creationTime == o.creationTime)
                && (this.servers == o.servers)
                && (this.useAgeMemory == o.useAgeMemory)
                && (this.rate == o.rate)
                && (this.weight == o.weight)
                && (this.currentWeight == o.currentWeight)
                && (this.earliest == o.earliest)
                && (this.latest == o.latest)
                && (this.priority == o.priority)
                && this.binding.equals(o.binding)
                && this.constraints.equals(o.constraints) // wrong
                && PNModel.setEquals(this.freeVariables, o.freeVariables)
                && PNModel.setEquals(this.disposable, o.disposable));
    }


    
    public double getAge() {
        return this.age;
    }


    
    public void setAge(double age) {
        this.age = age;
    }
}
