/*
 * Project:  Utils
 * Filename: FastStack.java
 * Created:  08.04.2006
 * Author:   Tim Strazny
 */

package petruchio.sim.pnmodel.util;

import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.RandomAccess;
import java.util.Iterator;
import java.util.Random;

/**
 * <Description>
 * 
 * @author Tim Strazny
 * @param <V> 
 */
public final class FastStack<V> implements Iterable<V>, Cloneable, RandomAccess {
    private static final Random RANDOM = new Random();
    
    private V[] speicher;
    private int top;
    private int increment;
    protected int modCount;
    
    
    @SuppressWarnings("unchecked")
    public FastStack(final FastStack<V> stack) {
        this.speicher = (V[]) new Object[stack.size() < 1 ? 1 : stack.size()];
        this.top = stack.size();
        this.increment = stack.getIncrement();
        System.arraycopy(stack.speicher, 0, this.speicher, 0, stack.size());
        this.modCount = 0;
    }
    
    
    @SuppressWarnings("unchecked")
    public FastStack(final int initialCapacity, final int increment){
        this.speicher = (V[])new Object[initialCapacity < 1 ? 1 : initialCapacity];
        this.increment = increment;
        this.top = 0;
        this.modCount = 0;
    }
    
    
    public FastStack(final int initialCapacity) {
        this(initialCapacity, 0);
    }
    
    
    public FastStack(){
        this(16);
    }
    
    
    public int getIncrement() {
        return this.increment;
    }
    
    
    public void setIncrement(final int increment) {
        this.increment = increment;
    }
    
    
    public int size() {
        return this.top;
    }
    
    
    public void clear() {
        // let gc do its work
        for(int i = 0; i < this.top; ++i) {
            this.speicher[i] = null;
        }
        this.top = 0;
        this.incModCount();
    }
    
    
    private void incModCount() {
        ++this.modCount;
    }
    
    
    public V peek() {
        return this.speicher[this.top-1];
    }
    
    
    public void moveTo(final FastStack<V> stack) {
        stack.pushAll(this);
        this.clear();
    }
    
    
    public V peek(final int i) {
        if(i >= this.top) {
            throw new IndexOutOfBoundsException("Access at index " + i + ". "
                            + "Element count is " + this.top);
        }
        return this.speicher[i];
    }
    
    
    public V pop() {
        this.incModCount();
        return this.speicher[--this.top];
    }
    
    
    public V pop(final int index) {
        this.incModCount();
        final V t = this.speicher[index];
        this.speicher[index] = this.speicher[--this.top];
        return t;
    }
    
    
    public V randomPop() {
        final int i = FastStack.RANDOM.nextInt(this.top);
        return this.pop(i);
    }
    
    
    public void push(final V t) {
        this.incModCount();
        this.ensureCapacity(this.size()+1);
        this.speicher[this.top++] = t;
    }    
    
    
    public void pushAll(final FastStack<? extends V> s) {
        this.incModCount();
        this.ensureCapacity(this.size()+s.size());
        System.arraycopy(s.speicher, 0, this.speicher, this.top, s.size());
        this.top += s.size();
    }
    
    
    public void addAll(final Collection<? extends V> s) {
        this.ensureCapacity(this.size()+s.size());
        for(final V v : s) {
            this.push(v);
        }
    }

    
    @SuppressWarnings("unchecked")
    public void ensureCapacity(final int size) {
        if(size >= this.speicher.length) {
            final V[] neu;
            if(this.increment > 0) {
                int s = 1;
                while(s*this.increment+this.speicher.length <= size) {
                    ++s;
                }
                neu = (V[])new Object[s*this.increment+this.speicher.length];
            } else {
                int s = 1;
                while(this.speicher.length << s <= size) {
                    ++s;
                }
                neu = (V[])new Object[this.speicher.length << s];
            }
            System.arraycopy(this.speicher, 0, neu, 0, this.top);
            this.speicher = neu;
        }
    }
    
    
    public boolean isEmpty() {
        return this.top == 0;
    }


    public Iterator<V> iterator() {
        return new Itr();
    }
    
    
    private class Itr implements Iterator<V> {
        private final int expectedModCount;
        private int index;
        
        public Itr() {
            this.index = 0;
            this.expectedModCount = FastStack.this.modCount;
        }
        
        private void checkForModification() {
            if(this.expectedModCount != FastStack.this.modCount) {
                throw new ConcurrentModificationException();
            }
        }
        
        public boolean hasNext() {
            this.checkForModification();
            return this.index < FastStack.this.size() - 1;
        }

        public V next() {
            this.checkForModification();
            return FastStack.this.peek(this.index++);
        }

        public void remove() {
            this.checkForModification();
            FastStack.this.pop(--this.index);
        }
        
    }
}
