/*
 * Project:  P-UMLaut
 * Filename: Function.java
 * Created:  07.10.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.util;

import java.util.List;
import java.util.Map;

import petruchio.sim.pnmodel.net.Value;

public class Function {
    private final String name;
    private final String func;
    private final List<String> formalParameters;
    private final List<Value> parameters;
    private final Node tree;
    private boolean inUse;
    
    public static String getFunctionName(final String fname, final int parameterCount) {
        return fname + "/" + parameterCount;
    }
    
    public Function(final String function) {
        this.inUse = false;
        int paramstart = function.indexOf('(');
        int paramend = function.indexOf('=');
        String params = function.substring(paramstart, paramend).trim();
        if(params.startsWith("(")) {
            params = params.substring(1);
        }
        if(params.endsWith(")")) {
            params = params.substring(0, params.length()-1);
        }
        final String[] theParams = params.split(",");
        this.formalParameters = Pool.getPool().getList();
        final List<Value> paramlist = Pool.getPool().getList();
        for(int i = 0; i < theParams.length; ++i) {
            final String param = theParams[i].trim();
            if(param.length() > 0) {
                final Value var = Value.getValue(param);
                if(!var.isVariable()) {
                    throw new RuntimeException("Parameter " + var + " of function "
                                    + function + " is no variable!");
                } else if(paramlist.contains(var)) {
                    throw new RuntimeException("Parameter " + var + " of function "
                                    + function + " is exists more than once!");
                }
                this.formalParameters.add(param);
                paramlist.add(var);
            }
        }
        
        while(function.charAt(paramend) == '=') {
            ++paramend;
        }
        
        this.func = function.substring(paramend);
        final RPN rpn = RPNComputer.infixToRPN(this.func);
        this.tree = TreeComputer.rpnToTreeWithDeletion(rpn);
        
        final List<Node> replacements = Pool.getPool().getList();
        final List<Value> replParams = Pool.getPool().getList();
        final List<Token> tokens = Pool.getPool().getList();
        for(final Value param : paramlist) {
            replParams.add(param);
            replacements.add(new Node(null, param));
            tokens.add(Token.getToken(param));
        }
        Function.replace(this.tree, tokens, replacements, false);
        this.parameters = replParams;

        this.name = Function.getFunctionName(function.substring(0, paramstart).trim(), this.parameters.size());
        
        TreeComputer.optimize(this.tree, null);
    }
    
    
    public boolean inUse() {
        return this.inUse;
    }
    
    
    public String getName() {
        return this.name;
    }
    
    
    public boolean allowAssign() {
        return this.tree.allowAssign();
    }
    
    
    public boolean canBeEvaluatedStatically() {
        return this.tree.canBeEvaluatedStatically();
    }
    
    
    public int getParameterCount() {
        return this.parameters.size();
    }
    
    
    private Node instantiateTree(final List<Node> params,
                    final Map<String, Function> functions, final Binding bind,
                    final Map<Value, Value> properties) {
        final Node clone = this.tree.clone();
        for(int i = 0; i < this.parameters.size(); i++) {
            final Token var = Token.getToken(this.parameters.get(i));
            final Token val = Token.getToken(TreeComputer.evaluate(params.get(i),
                            functions, bind, properties));
            clone.replaceValue(var, val);
        }
        return clone;
    }
    
    
    public Value evaluate(final List<Node> params,
                    final Map<String, Function> functions, final Binding bind,
                    final Map<Value, Value> properties) {
        Value result = null;
        try {
            if(this.inUse()) {
                final Node clone = this.instantiateTree(params, functions, bind,
                                properties);
                result = TreeComputer.evaluate(clone, functions, bind, properties);
                clone.clearAll();
            } else {
                try {
                    this.inUse = true;
                    this.bind(params, functions, bind, properties);
                    result = TreeComputer.evaluate(this.getTree(), functions, bind, properties);
                } finally {
                    this.unbind();
                    this.inUse = false;
                }
            }
        } catch(final Exception e) {
            result = null;;
        }
        return result;
    }
    
    private void unbind() {
        Function.unbind(this.parameters, this.formalParameters);
    }
    
    private static void unbind(final List<Value> params, final List<String> vars) {
        for(int i = 0; i < params.size(); ++i) {
            final Value val = params.get(i);
            final String var = vars.get(i);
            val.makeVariable(var);
        }
    }
    
    
    private void bind(final List<Node> params,
                    final Map<String, Function> functions, final Binding bind,
                    final Map<Value, Value> properties) {
        if(params.size() != this.getParameterCount()) {
            throw new RuntimeException("Function " + this + " cannot be "
                            + "executed with parameters " + params + ".\nThe "
                            + "function expects " + this.getParameterCount()
                            + " parameters, where " + params.size() + " where "
                            + "given.");
        }
        Function.bind(this.parameters, params, functions, bind, properties);
    }
    
    
    private static void bind(final List<Value> formalParams, final List<Node> params,
                    final Map<String, Function> functions, final Binding bind,
                    final Map<Value, Value> properties) {
        if(params.size() != formalParams.size()) {
            throw new RuntimeException("Function cannot be "
                            + "executed with parameters " + params + ".\nThe "
                            + "function expects " + formalParams.size()
                            + " parameters, where " + params.size() + " where "
                            + "given.");
        }
        for(int i = 0; i < formalParams.size(); ++i) {
            final Value var = formalParams.get(i);
            final Node n = params.get(i);
            final Value val = TreeComputer.evaluate(n, functions, bind, properties);
            if(val == null) {
                throw new RuntimeException("Error evaluating " + n);
            }
            var.changeValue(val);
        }
    }
    

    private static void replace(final Node expression, final List<Token> search,
                    final List<Node> replacements, final boolean identity) {
        for(int i = 0; i < search.size(); ++i) {
            final Token param = search.get(i);
            if((identity && (expression.getToken().getValue() == param.getValue())) || (!identity && expression.getToken().equals(param))) {
                final Node node = replacements.get(i);
                expression.clearChildren();
                expression.setToken(node.getToken());
                expression.addChildren(node.getChildren());
                break;
            }
        }
        for(final Node child : expression.getChildren()) {
            Function.replace(child, search, replacements, identity);
        }
    }
    
    
    @Override
    public String toString() {
        return this.name + this.parameters.toString().replaceFirst("\\[(.*)\\]", "($1)") + " = " + this.tree;
    }
    
    
    public Node getTree() {
        return this.tree;
    }
    
    public List<Value> getParameters() {
        return this.parameters;
    }
}
