/*
 * Project:  Utils
 * Filename: IntStack.java
 * Created:  08.04.2006
 * Author:   Tim Strazny
 */

package petruchio.sim.pnmodel.util;

import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.RandomAccess;
import java.util.Iterator;
import java.util.Random;

/**
 * <Description>
 * 
 * @author Tim Strazny
 */
public final class IntStack implements Iterable<Integer>, Cloneable, RandomAccess {
    private static final Random RANDOM = new Random();
    
    private int[] speicher;
    private int top;
    private int increment;
    protected int modCount;
    
    
    @SuppressWarnings("unchecked")
    public IntStack(final IntStack stack) {
        this.speicher = new int[stack.size() < 1 ? 1 : stack.size()];
        this.top = stack.size();
        this.increment = stack.getIncrement();
        System.arraycopy(stack.speicher, 0, this.speicher, 0, stack.size());
        this.modCount = 0;
    }
    
    
    @SuppressWarnings("unchecked")
    public IntStack(final int initialCapacity, final int increment){
        this.speicher = new int[initialCapacity < 1 ? 1 : initialCapacity];
        this.increment = increment;
        this.top = 0;
        this.modCount = 0;
    }
    
    
    public IntStack(final int initialCapacity) {
        this(initialCapacity, 0);
    }
    
    
    public IntStack(){
        this(16);
    }
    
    
    public int getIncrement() {
        return this.increment;
    }
    
    
    public void setIncrement(final int increment) {
        this.increment = increment;
    }
    
    
    public int size() {
        return this.top;
    }
    
    
    public void clear() {
        this.top = 0;
        this.incModCount();
    }
    
    
    private void incModCount() {
        ++this.modCount;
    }
    
    

    public int get(final int i) {
        return this.peek(i);
    }

    public int peek() {
        return this.speicher[this.top-1];
    }
    
    
    public void moveTo(final IntStack stack) {
        stack.pushAll(this);
        this.clear();
    }
    
    
    public int peek(final int i) {
        if(i >= this.top) {
            throw new IndexOutOfBoundsException("Access at index " + i + ". "
                            + "Element count is " + this.top);
        }
        return this.speicher[i];
    }
    
    
    public int pop() {
        this.incModCount();
        return this.speicher[--this.top];
    }
    
    
    public int pop(final int index) {
        this.incModCount();
        final int t = this.speicher[index];
        this.speicher[index] = this.speicher[--this.top];
        return t;
    }
    
    
    public int randomPop() {
        final int i = IntStack.RANDOM.nextInt(this.top);
        return this.pop(i);
    }
    
    
    public void push(final int t) {
        this.incModCount();
        this.ensureCapacity(this.size()+1);
        this.speicher[this.top++] = t;
    }    
    
    

    public void pushAll(final IntStack s) {
        this.incModCount();
        this.ensureCapacity(this.size()+s.size());
        System.arraycopy(s.speicher, 0, this.speicher, this.top, s.size());
        this.top += s.size();
    }
    
    
    public void addAll(final Collection<Integer> s) {
        this.ensureCapacity(this.size()+s.size());
        for(final int v : s) {
            this.push(v);
        }
    }

    
    @SuppressWarnings("unchecked")
    public void ensureCapacity(final int size) {
        if(size >= this.speicher.length) {
            final int[] neu;
            if(this.increment > 0) {
                int s = 1;
                while(s*this.increment+this.speicher.length <= size) {
                    ++s;
                }
                neu = new int[s*this.increment+this.speicher.length];
            } else {
                int s = 1;
                while(this.speicher.length << s <= size) {
                    ++s;
                }
                neu = new int[this.speicher.length << s];
            }
            System.arraycopy(this.speicher, 0, neu, 0, this.top);
            this.speicher = neu;
        }
    }
    
    
    public boolean isEmpty() {
        return this.top == 0;
    }


    public Iterator<Integer> iterator() {
        return new Itr();
    }

    
    private class Itr implements Iterator<Integer> {
        private final int expectedModCount;
        private int index;
        
        public Itr() {
            this.index = 0;
            this.expectedModCount = IntStack.this.modCount;
        }
        
        private void checkForModification() {
            if(this.expectedModCount != IntStack.this.modCount) {
                throw new ConcurrentModificationException();
            }
        }
        
        public boolean hasNext() {
            this.checkForModification();
            return this.index < IntStack.this.size() - 1;
        }

        public Integer next() {
            this.checkForModification();
            return IntStack.this.peek(this.index++);
        }

        public void remove() {
            this.checkForModification();
            IntStack.this.pop(--this.index);
        }
        
    }
}
