/*
 * Project:  P-UMLaut
 * Filename: Node.java
 * Created:  27.07.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.util;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import petruchio.sim.pnmodel.net.Value;
import petruchio.sim.pnmodel.net.ValueList;
import petruchio.sim.pnmodel.PNModel;

public class Node implements Cloneable {
    public static final Node FALSE = new Node(null, Value.V_FALSE);
    public static final Node TRUE = new Node(null, Value.V_TRUE);
    public static final Node ANY = new Node(null, Value.V_ANY);
    public static final Node DOT = new Node(null, Value.V_DOT);
    public static final Node ZERO = new Node(null, Value.getValue(0.0f));
    public static final Node ONE = new Node(null, Value.getValue(1.0f));
    public static final Node POS_INFTY = new Node(null, Value.getValue(Float.POSITIVE_INFINITY));
    public static final Node NEG_INFTY = new Node(null, Value.getValue(Float.NEGATIVE_INFINITY));
    
    private List<Node> children;
    private Token value;
    private Node parent;
    private Map<String, ValueList> constraints;
    private boolean canBeEvaluatedStatically;
    private boolean isOrdered;
    private boolean allowAssign;
    
    public Node(Node parent, Token value) {
        this(parent, value, null);
    }
    
    public Node(Node parent, Value value) {
        this(parent, Token.getToken(value), null);
    }
    
    public Node(Node parent, Token.Operator op) {
        this(parent, Token.getToken(op), null);
    }
    
    public Node(Node parent, String func) {
        this(parent, Token.getToken(func), null);
    }
    
    public Node(Node parent, Token value, List<Node> children) {
        this.canBeEvaluatedStatically = false;
        this.isOrdered = false;
        this.allowAssign = false;
        this.constraints = null;
        this.children = Pool.getPool().getList();
        this.setParent(parent);
        this.setToken(value);
        if(children != null) {
            this.children.addAll(children);
        }
    }
    
    
    public boolean hasConstraints() {
        return this.constraints != null;
    }
    
    
    public void addConstraints(Map<String, ValueList> m) {
        if((m != null) && !m.isEmpty()) {
            if(this.constraints == null) {
                this.constraints = Pool.getPool().getMap();
            }
            this.constraints.putAll(m);
        }
    }
    
    
    public void unifyVariables() {
        final Map<String, Value> map = Pool.getPool().getMap();
        this.unifyVariables(map);
    }
    
    
    public void unifyVariables(final Map<String, Value> map) {
        if((this.value != null) && this.value.isValue() && this.value.getValue().isVariable()) {
            String var = this.value.getValue().getVariable();
            final Value theVar = map.get(var);
//            System.out.print("looking at " + var + " ... ");
            if(theVar == null) {
                map.put(var, this.value.getValue());
//                System.out.println("created");
            } else {
                this.value = Token.getToken(theVar);
//                System.out.println("used");
            }
        } else {
            for(final Node node : this.children) {
                node.unifyVariables(map);
            }
        }
    }
    
    
    public void clearConstraints() {
        if(this.constraints != null) {
            this.constraints.clear();
            this.constraints = null;
        }
    }
    
    
    public ValueList getConstraint(final String var) {
        if(this.constraints == null) {
            return null;
        } else {
            return this.constraints.get(var);
        }
    }
    
    
    public boolean constrain(final String var, final ValueList values) {
        if(this.constraints == null) {
            this.constraints = Pool.getPool().getMap();
        }
        
        ValueList oldConstraint = this.getConstraint(var);
        if(oldConstraint == null) {
            oldConstraint = Pool.getPool().getValueList(values);
            this.constraints.put(var, oldConstraint);
        } else {
            oldConstraint.retainAll(values);
        }
        return !oldConstraint.isEmpty();
    }
    
    
    public void constrainAll(final Map<String, ValueList> map) {
        if(map != null) {
            for(final Map.Entry<String, ValueList> e : map.entrySet()) {
                this.constrain(e.getKey(), e.getValue());
            }
        }
    }
    
    
    public Map<String, ValueList> getConstraints() {
        return this.constraints;
    }
    
    
    public Map<String, ValueList> removeConstraints() {
        final Map<String, ValueList> con = this.constraints;
        this.constraints = null;
        return con;
    }
    
    
    public void setConstraints(final Map<String, ValueList> m) {
        this.constraints = m;
    }
    
    
    public boolean isOrdered() {
        return this.isOrdered;
    }
    
    
    public void setIsOrdered(final boolean isOrdered) {
        this.isOrdered = isOrdered;
    }
    
    
    public boolean allowAssign() {
        return this.allowAssign;
    }
    
    
    public void setAllowAssign(final boolean allowAssign) {
        this.allowAssign = allowAssign;
    }
    
    
    public boolean canBeEvaluatedStatically() {
        return this.canBeEvaluatedStatically;
    }
    
    
    public void setCanBeEvaluatedStatically(final boolean value) {
        this.canBeEvaluatedStatically = value;
    }
    
    
    public void setToken(final Token value) {
        this.value = value;
    }
    
    
    public void setToken(final Value value) {
        this.value = Token.getToken(value);
    }
    
    
    public void setToken(final Token.Operator op) {
        this.value = Token.getToken(op);
    }
    
    
    public void setToken(final String func) {
        this.value = Token.getToken(func);
    }
    
    
    @Override
    public boolean equals(final Object o) {
        return (o instanceof Node) && this.equals((Node)o);
    }
    
    
    public boolean equals(final Node n) {
        if(n == this) {
            return true;
        }
        if(n == null) {
            return false;
        }
        final boolean ret = this.value.equals(n.value)
                && (this.children.equals(n.children)
                        || (TreeComputer.isCommutative(this.value)
                                && PNModel.setEquals(this.children, n.children)));
        if(ret) {
            return true;
        } else if(this.value.isOperator() && TreeComputer.isReversable(this.value.getOperator())
                        && (n.value.getOperator() == TreeComputer.reversion(this.value.getOperator()))) {
            List<Node> rev = Node.reverse(n.getChildren());
            boolean retvalue = this.children.equals(rev);
            return retvalue;
        } else {
            return false;
        }
    }
    
    
    @Override
    public int hashCode() {
        assert false : "hashCode not designed";
        return 42;
    }
    
    
    public static <T> List<T> reverse(final List<T> list) {
        final List<T> result = Pool.getPool().getList();
        for(int i = list.size()-1; i >= 0; i--) {
            result.add(list.get(i));
        }
        return result;
    }
    
    
    public Node getParent() {
        return this.parent;
    }
    
    public void setParent(final Node parent) {
        this.parent = parent;
    }
    
    public List<Node> getChildren() {
        return this.children;
    }
    
    public Token getToken() {
        return this.value;
    }
    
    public void addChild(final Node n) {
        n.setParent(this);
        this.children.add(n);
    }
    
    public void addChildren(final List<Node> n) {
        for(final Node child : n) {
            this.addChild(child);
        }
    }
    
    public void clearChildren() {
        this.children.clear();
    }
    
    
    public boolean hasChildren() {
        return !this.children.isEmpty();
    }
    
    
//    public void replaceFunctions(final Map<String, Function> functions) {
//        final List<String> l = Pool.getPool().getList();
//        TreeComputer.setEvaluatability(this, functions);
//        this.replaceFunctions(functions, l);
//        Pool.getPool().returnList(l);
//    }
//    
//    
//    public void replaceFunctions(final Map<String, Function> functions, final List<String> calls) {
//        boolean ok = true;
//        for(final Node n : this.children) {
//            if(!n.canBeEvaluatedStatically()) {
//                ok = false;
//            }
//        }
//        
//        final List<String> added = Pool.getPool().getList();
//        while(ok && this.value.isFunction() && functions.containsKey(this.value.getFunction() + "/" + this.children.size())) {
//            final String key = this.value.getFunction() + "/" + this.children.size();
//            if(calls.contains(key)) {
////                 recursion
////                System.out.println("recursion: " + key + " -> " + calls);
//                break;
//            }
////            System.out.println("replacing: " + key + " -> " + calls);
//            calls.add(key);
//            added.add(key);
//            
//            final Function f = functions.get(key);
//            final Node func = f.instantiateTree(this.children, functions, null);
//            this.clearChildren();
//            this.setToken(func.getToken());
//            this.addChildren(func.getChildren());
//        }
//        
//        for(final Node child : this.children) {
//            child.replaceFunctions(functions, calls);
//        }
//        
//        calls.removeAll(added);
//        Pool.getPool().returnList(added);
//    }
    
    
    @Override
    public Node clone() {
        Node n = null;
        try {
            n = (Node)super.clone();
            n.children = Pool.getPool().getList();
            n.constraints = null;
        } catch(final CloneNotSupportedException e) {
            throw new InternalError();
        }
        n.parent = this.parent;
        n.value = this.value.clone();
        for(final Node child : this.children) {
            n.children.add(child.clone());
        }
        n.constrainAll(this.constraints);
        n.allowAssign = this.allowAssign;
        n.canBeEvaluatedStatically = this.canBeEvaluatedStatically;
        n.isOrdered = this.isOrdered;
        //n.unifyVariables();
        return n;
    }
    
    
    public void replaceValue(final Token remove, final Token newValue) {
        if(this.value.equals(remove)) {
            this.value = newValue;
        }
        for(final Node child : this.children) {
            child.replaceValue(remove, newValue);
        }
    }
    
    
    public void clearAll() {
        for(final Node n : this.children) {
            n.clearAll();
            n.setParent(null);
        }
        this.clearConstraints();
        this.clearChildren();
    }
    
    
    public void getAndReplaceVariables(final Map<String, Value> map) {
        if(this.hasConstraints()) {
            for(final String var : this.constraints.keySet()) {
                if(!map.containsKey(var)) {
                    map.put(var, Value.getValue(var));
                }
            }
        }
        if(this.value.isValue() && this.value.getValue().isVariable()) {
            final String varString = this.value.getValue().getVariable();
            final Value var = map.get(varString);
            if(var == null) {
                map.put(varString, this.value.getValue());
            } else {
                this.value = Token.getToken(var);
            }
        }
        for(final Node child : this.children) {
            child.getAndReplaceVariables(map);
        }
    }
    
    
    @Override
    public String toString() {
        final boolean isFunction = this.value.isFunction();
        if(!isFunction && (this.children.size() == 0)) {
            if(this.value.isValue()) {
                return this.value.getValue().getNetValue();
            } else {
                return this.value.toString();
            }
        } else if(!isFunction && (this.children.size() == 1)) {
            Token.Operator op = this.value.getOperator();
            if(op == Token.Operator.UNARY_MINUS) {
                op = Token.Operator.MINUS;
            } 
            return op.toString() + this.children.get(0);
        } else if(!isFunction) {
            Token.Operator op = this.value.getOperator();
            if(op == Token.Operator.STRING_CONCATENATION) {
                op = Token.Operator.PLUS;
            }
            final StringBuffer buf = new StringBuffer();
            final boolean haveParentheses = ((this.parent != null)
                        && (!this.parent.value.equals(this.value))
                        || op == Token.Operator.EQUAL || op == Token.Operator.NOT_EQUAL
                        || op == Token.Operator.LIST_START);
            if(isFunction) {
                buf.append(op);
            }
            if(haveParentheses) {
                buf.append(Token.Operator.PARENTHESES_OPEN);
            }
            for(final Iterator<Node> it = this.children.iterator(); it.hasNext();) {
                final Node child = it.next();
                buf.append(child);
                if(it.hasNext()) {
                    if(isFunction || op == Token.Operator.LIST_START) {
                        buf.append(Token.Operator.SEPARATOR);
                        buf.append(" ");
                    } else {
                        buf.append(op);
                    }
                }
            }
            if(haveParentheses) {
                buf.append(Token.Operator.PARENTHESES_CLOSE);
            }
            return buf.toString();
        } else {
            String op = this.value.getFunction();
            final StringBuffer buf = new StringBuffer();
            buf.append(op);
            buf.append(Token.Operator.PARENTHESES_OPEN);
            for(final Iterator<Node> it = this.children.iterator(); it.hasNext();) {
                final Node child = it.next();
                buf.append(child);
                if(it.hasNext()) {
                    buf.append(Token.Operator.SEPARATOR);
                    buf.append(" ");
                }
            }
            buf.append(Token.Operator.PARENTHESES_CLOSE);
            return buf.toString();
        }
    }
}
