/*
 * Project:  P-UMLaut
 * Filename: Token.java
 * Created:  28.04.2006
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.util;

import java.util.IdentityHashMap;
import java.util.Map;

import petruchio.sim.pnmodel.net.Value;

public class Token {
    public static final String PROJECTION = "proj";
    
    protected final static Map<Operator, Token> tokens = new IdentityHashMap<Operator, Token>();
    protected final static Map<String, Operator> strings = new FastHashMap<String, Operator>();
    protected final static Map<Operator, Integer> precedence = new FastHashMap<Operator, Integer>();
    
    public static enum Associativity { LEFT, RIGHT, NONASSOC };
    
    public static enum Operator {
        NO_OP("", -1, Associativity.NONASSOC, false), // NO OPERATOR (very very special)
        
        UNARY_MINUS("--", Associativity.RIGHT, 14),
        NOT("!", Associativity.RIGHT, 14),
        BITWISE_NOT("~", Associativity.RIGHT, 14),
        
        POWER("**", 13),
        
        TIMES("*", 12),
        DIVIDE("/", 12),
        MODULO("%", 12),
        
        PLUS("+", 11),
        MINUS("-", 11),
        STRING_CONCATENATION("+", 11, false),

        LESS("<", 10),
        GREATER(">", 10),
        LESS_EQUAL("<=", 10),
        GREATER_EQUAL(">=", 10),

        EQUAL("=", 9),
        NOT_EQUAL("#", 9),
        
        AND("&", 8),
        
        OR("|", 7),
        XOR("^", 7),
        
        DECIMAL_SEPARATOR(".", 0, false),
        ESCAPE("\\", 0),
        STRING_SEPARATOR("\"", 0),
        LIST_START("))", 0, false),
        LIST_END("((", 0, false),
        
        PARENTHESES_OPEN("(", 0),
        PARENTHESES_CLOSE(")", 0),
        BRACKETS_OPEN("[", 0),
        BRACKETS_CLOSE("]", 0),
        CURLY_BRACES_OPEN("{", 0),
        CURLY_BRACES_CLOSE("}", 0),

        SEPARATOR(",", -1),
        CONSTRAINT_SEPARATOR(":", -1);
        
        public static Token getToken(final Operator op) {
            return Token.tokens.get(op);
        }
        
        public static Operator getOperator(final String op) {
            return Token.strings.get(op);
        }
        
        private final String token;
        private final Associativity assoc;
        
        Operator(final String token, final Associativity assoc, final int precedence){
            this(token, precedence, assoc, true);
        }
        
        Operator(final String token, final int precedence){
            this(token, precedence, true);
        }
        
        Operator(final String token, final int precedence, boolean checkForInput){
            this(token, precedence, Associativity.LEFT, checkForInput);
        }
        
        Operator(final String token, final int precedence, final Associativity assoc, boolean checkForInput){
            if(token == null) {
                throw new IllegalArgumentException("The operator may not be 'null'.");
            } else if(checkForInput && Token.strings.containsKey(token)) {
                throw new IllegalArgumentException("This operator has already been created: " + token);
            }
            this.assoc = assoc;
            this.token = token;
            if(checkForInput) {
                Token.strings.put(token, this);
            }
            Token.tokens.put(this, new Token(this));
            Token.precedence.put(this, precedence);
        }
        
        public Associativity getAssociativity() {
            return this.assoc;
        }
        
        public boolean isLeftAssociative() {
            return this.assoc == Associativity.LEFT;
        }
        
        public String toString() {
            return this.token;
        }
    }
    
    
    public static int getPrecedence(final Operator op) {
        return Token.precedence.get(op);
    }
    
    
    public static Token getToken(final Value value) {
        return new Token(value);
    }
    
    
    public static Token getToken(final String str) {
        final Operator op = Operator.getOperator(str);
        final Token t;
        if(op == null) {
            t = new Token(str);
        } else {
            t = Operator.getToken(op);
        }
        
        if(t == null) {
            throw new RuntimeException("null: " + str); 
        }
        return t;
    }
    
    
    public static Token getToken(final Operator op) {
//        if(Operator.getToken(op) == null) {
//            throw new RuntimeException("null: " + op); 
//        }
        return Operator.getToken(op);
    }
    
    
    private final Value value;
    private final Operator operator;
    private final String function;
    private final boolean isValue;
    private final boolean isOperator;
    private final boolean isFunction;
    
    private Token(final Value value) {
        this.value = value;
        this.operator = null;
        this.function = null;
        this.isValue = true;
        this.isOperator = false;
        this.isFunction = false;
    }
    
    protected Token(final Operator op) {
        this.value = null;
        this.operator = op;
        this.function = null;
        this.isValue = false;
        this.isOperator = true;
        this.isFunction = false;
    }
    
    
    protected Token(final String func) {
        this.value = null;
        this.operator = null;
        this.function = func;
        this.isValue = false;
        this.isOperator = false;
        this.isFunction = true;
    }
    
    
    public Operator getOperator() {
        return this.operator;
    }
    
    
    public Value getValue() {
        return this.value;
    }
    
    
    public String getFunction() {
        return this.function;
    }
    
    
    public boolean isValue() {
        return this.isValue;
    }
    
    
    public boolean isOperator() {
        return this.isOperator;
    }
    
    
    public boolean isFunction() {
        return this.isFunction;
    }
    
    
    @Override
    public boolean equals(final Object o) {
        return (o instanceof Token) && this.equals((Token)o);
    }
    
    
    public boolean equals(final Token o) {
        return (o == this) || ((o != null) && (this.isFunction == o.isFunction)
                        && (this.isValue == o.isValue)
                        && (this.isOperator == o.isOperator)
                        && (this.operator == o.operator)
                        && (this.value == null ? o.value == null : this.value.equals(o.value))
                        && (this.function == null ? o.function == null : this.function.equals(o.function)));
    }
    
    
    @Override
    public Token clone() {
        return this.isValue ? new Token(Value.getValue(this.value)) : this;
    }
    
    
    @Override
    public int hashCode() {
        assert false : "hashCode not designed";
        return 42;
    }
    
    @Override
    public String toString() {
        //return this.isValue ? "v:" + this.value : this.isOperator ? "o:" + this.operator : "f:" + this.function;
        return this.isValue ? this.value.toString() : this.isOperator ? this.operator.toString() : this.function;
    }
}
