/*
 * Project:  P-UMLaut
 * Filename: TreeComputer.java
 * Created:  27.07.2005
 * Author:   Tim Strazny
 */
package petruchio.sim.pnmodel.util;

import petruchio.sim.pnmodel.net.Tuple;
import petruchio.sim.pnmodel.net.Value;
import petruchio.sim.pnmodel.net.ValueList;
import petruchio.sim.pnmodel.util.Token.Operator;

import java.util.List;
import java.util.Iterator;
import java.util.Map;

public class TreeComputer {
    public static Node infixToTree(final String expr) {
        return TreeComputer.rpnToTreeWithDeletion(RPNComputer.infixToRPN(expr));
    }
    
    
    public static Node rpnToTree(final RPN originalRPN) {
        final RPN rpn = originalRPN.clone();
        final Node n = TreeComputer.rpnToTreeWithDeletion(rpn);
        return n;
    }
    
    
    public static Node rpnToTreeWithDeletion(final RPN rpn) {
        if(rpn.isEmpty()) {
            return new Node(null, Token.Operator.NO_OP);
        }
        final Token token = rpn.removeLast();
        final Node node = new Node(null, token);
        node.addConstraints(rpn.getConstraints());
        if(token.isFunction() || (token.getOperator() == Token.Operator.LIST_START)) {
            if(token.getOperator() != Token.Operator.LIST_START) {
                // prefix-operators were encoded as functions
                node.setToken(Token.getToken(token.getFunction()));
            }
            
            if(token.getOperator() != Token.Operator.LIST_START) {
                rpn.removeLast(); // remove LIST_START
            }
            // special handling of tuples
            int arity = 0;
            final List<Node> operands = Pool.getPool().getList();
            for(; rpn.getLast().getOperator() != Token.Operator.LIST_END; ++arity) {
                final Node expr = TreeComputer.rpnToTreeWithDeletion(rpn);
                operands.add(expr);
            }
            final Token end = rpn.removeLast(); // remove LIST_END
            if(end.getOperator() != Token.Operator.LIST_END) {
                throw new RuntimeException("Bad syntax. Expected '(', found '" + end + "'.");
            }
            for(int i = arity - 1; i >= 0; i--) {
                final Node expr = operands.get(i);
                node.addChild(expr);
            }
            
        } else if(token.isOperator()) {
            final int arity = RPNComputer.getArityOfOperator(token.getOperator());
            final List<Node> operands = Pool.getPool().getList();
            for(int i = 0; i < arity; i++) {
                final Node expr = TreeComputer.rpnToTreeWithDeletion(rpn);
                operands.add(expr);
            }
            for(int i = arity - 1; i >= 0; i--) {
                final Node expr = operands.get(i);
                node.addChild(expr);
            }
            
//            System.out.println(token + " ==> " + node);
        }
        return node;
    }
    
    public static boolean optimizeHard(final Node root, final Map<String, Function> functions) {
        boolean changed;
        boolean hasChanged = false;
        do {
            changed = TreeComputer.optimizeTree(root, functions);
//            Binding bind = Pool.getPool().getBinding();
//            bind.addConstraints(root.getConstraints());
//            changed = TreeComputer.staticEvaluateWithBinding(root, bind);
//            root.setConstraints(bind.getConstraints());
//            Pool.getPool().returnBinding(bind);
//            System.out.println("opt: " + root + "\n     " + root.getConstraints());
            changed |= TreeComputer.shortenTree(root, functions);
//            System.out.println("shr: " + root + "\n     " + root.getConstraints());
            hasChanged = hasChanged || changed;
        } while(changed);
        
        return hasChanged;
    }
    
    public static void optimize(final Node root, final Map<String, Function> functions) {
        TreeComputer.optimizeTree(root, functions);
    }
    
    public static boolean shortenTree(final Node root, final Map<String, Function> functions) {
        final Binding bind = Pool.getPool().getBinding();
        bind.addConstraints(root.getConstraints());
        
        boolean found;
        boolean changed = false;
        do {
            TreeComputer.setEvaluatability(root, functions);
            found = TreeComputer.extractBindingWithAlteration(root, bind, functions, false);
            changed |= found;
        } while(found);
        
        for(final Map.Entry<String, Value> e : bind.getBindings()) {
            final String var = e.getKey();
            final Value val = e.getValue();
            TreeComputer.replace(root, Token.getToken(Value.getValue(var)), Token.getToken(val));
        }

        root.clearConstraints();
        root.constrainAll(bind.getConstraints());
        
        return changed;
    }
    
    
    public static void replace(final Node node, final Token val, final Token replacement) {
        if(val.equals(node.getToken())) {
            node.setToken(replacement);
        }
        
        final List<Node> children = node.getChildren();
        for(int i = 0, n = children.size(); i < n; ++i) {
            final Node child = children.get(i);
            TreeComputer.replace(child, val, replacement);
        }
    }
    
    
    public static boolean optimizeTree(final Node root, final Map<String, Function> functions) {
        final Map<String, ValueList> constraints = root.removeConstraints();
        if(constraints != null) {
            for(final Map.Entry<String, ValueList> e : constraints.entrySet()) {
                final ValueList values = e.getValue();
                if(values.size() == 1) {
                    final Value var = Value.getValue(e.getKey());
                    final Value val = values.get(0);
                    if(!val.isAny()) {
                        root.replaceValue(Token.getToken(var), Token.getToken(val));
                    }
                }
            }
        }
        boolean changed;
        boolean changes = false;
        final boolean debug = false;
        if(debug) {
            System.out.println("init: " + root);
        }
        do {
            changed = TreeComputer.simplify(root);
            if(debug) {
                System.out.println("simp: " + root);
            }
            changed = TreeComputer.staticEvaluate(root, functions) || changed;
            if(debug) {
                System.out.println("eva1: " + root);
            }
            TreeComputer.setEvaluatability(root, functions);
            changed = TreeComputer.staticEvaluateWithBinding(root, functions, null) || changed;
            if(debug) {
                System.out.println("eva2: " + root);
            }
            changed = TreeComputer.flatten(root, functions) || changed;
            if(debug) {
                System.out.println("flat: " + root);
            }
            changed = TreeComputer.mergeVars(root) || changed;
            if(debug) {
                System.out.println("merg: " + root);
            }
            changed = TreeComputer.sawn(root) || changed;
            if(debug) {
                System.out.println("sawn: " + root);
            }
            changed = TreeComputer.transform(root, null) || changed;
            if(debug) {
                System.out.println("tran: " + root);
            }
            
            changes = changed || changes;
        } while(changed);
        TreeComputer.setEvaluatability(root, functions);
        TreeComputer.order(root, null);
        if(debug) {
            System.out.println("orde: " + root);
        }

        root.setConstraints(constraints);
        return changes;
    }
    
    
    public static boolean mergeVars(final Node root) {
        final Token val = root.getToken();
        if(val.isValue()) {
            return false;
        }

        boolean changed = false;
        
        final List<Node> children = root.getChildren();
        for(int i = 0, n = children.size(); i < n; ++i) {
            final Node child = children.get(i);
            changed = TreeComputer.mergeVars(child) || changed;
        }
        
        Token.Operator op = val.getOperator(); 
        
        if((op == Token.Operator.PLUS) || (op == Token.Operator.TIMES)) {
            for(int i = 0; i < root.getChildren().size()-1; i++) {
                final Node a = root.getChildren().get(i);
                int ctr = 1;
                for(int j = i+1; j < root.getChildren().size();) {
                    final Node b = root.getChildren().get(j);
                    if(a.equals(b)) {
                        root.getChildren().remove(j);
                        ++ctr;
                    } else {
                        ++j;
                    }
                }
                if(ctr > 1) {
                    final Node nu = new Node(root, Token.getToken(op == Token.Operator.PLUS ? Token.Operator.TIMES : Token.Operator.POWER));
                    nu.addChild(a);
                    nu.addChild(new Node(nu, Token.getToken(Value.getValue(ctr))));
                    root.getChildren().set(i, nu);
                    changed = true;
                }
            }
        } else if((op == Token.Operator.MINUS) || (op == Token.Operator.DIVIDE)) {
            for(int i = 0; i < root.getChildren().size()-1; i++) {
                final Node a = root.getChildren().get(i);
                int ctr = 1;
                for(int j = i+1; j < root.getChildren().size();) {
                    final Node b = root.getChildren().get(j);
                    if(a.equals(b)) {
                        root.getChildren().remove(j);
                        if(i == 0) {
                            a.setToken(Token.getToken(Value.getValue(op == Token.Operator.MINUS ? 0.0d : 1.0d)));
                            break;
                        } else {
                            ++ctr;
                        }
                    } else {
                        ++j;
                    }
                }
                if(ctr > 1) {
                    final Node nu = new Node(root, Token.getToken(op == Token.Operator.PLUS ? Token.Operator.TIMES : Token.Operator.POWER));
                    nu.addChild(a);
                    nu.addChild(new Node(nu, Token.getToken(Value.getValue(ctr))));
                    root.getChildren().set(i, nu);
                    changed = true;
                }
            }
        } else if(op == Token.Operator.POWER) {
            for(int i = 1; i < root.getChildren().size()-1; i++) {
                final Node a = root.getChildren().get(i);
                int ctr = 1;
                for(int j = i+1; j < root.getChildren().size();) {
                    final Node b = root.getChildren().get(j);
                    if(a.equals(b)) {
                        root.getChildren().remove(j);
                        ++ctr;
                    } else {
                        ++j;
                    }
                }
                if(ctr > 1) {
                    final Node nu = new Node(root, Token.getToken(Token.Operator.POWER));
                    nu.addChild(a);
                    nu.addChild(new Node(nu, Token.getToken(Value.getValue(ctr))));
                    root.getChildren().set(i, nu);
                    changed = true;
                }
            }
        }
        
        if(changed && (root.getChildren().size() == 1)) {
            final Node child = root.getChildren().remove(0);
            root.setToken(child.getToken());
            root.addChildren(child.getChildren());
            changed = true;
        }
        
        return changed;
    }
    
    
    public static boolean transform(final Node root, final Binding bind) {
        final Token val = root.getToken();
        if(!val.isOperator()) {
            return false;
        }
        
        if(RPNComputer.LOGICAL_OPERATORS.contains(val.getOperator())) {
            boolean changed = false;
            final List<Node> children = root.getChildren();
            for(int i = 0, n = children.size(); i < n; ++i) {
                final Node child = children.get(i);
                changed = TreeComputer.transform(child, bind) || changed;
            }
            return changed;
            
        } else if(RPNComputer.COMPARISON_OPERATORS.contains(val.getOperator())) {
            final Node left = root.getChildren().get(0);
            final Node right = root.getChildren().get(1);
            
            final int lcount = TreeComputer.countVars(left, bind);
            final int rcount = TreeComputer.countVars(right, bind);
            if((lcount == 1) && (rcount == 0)) {
                return TreeComputer.performShift(root, left, right, bind);
            } else if((lcount == 0) && (rcount == 1)) {
                return TreeComputer.performShift(root, right, left, bind);
            }
        }
        
        return false;
    }
    
    
    public static boolean performShift(final Node root, Node from,
                    Node to, final Binding bind) {
        boolean changed = false;
        while(from.hasChildren()) {
            //System.out.println("(" + from + "=" + to + ")");
            final Token op = from.getToken();
            final Token invOp = TreeComputer.inverseOp(op);
            if(invOp == null) {
                break;
            }            
            
            if(invOp.getOperator() == Token.Operator.POWER) {
                // special code for x^y
                if(TreeComputer.countVars(from.getChildren().get(0), bind) != 1) {
                    final Node base = from.getChildren().remove(0);
                    if(from.getChildren().size() == 1) {
                        from = from.getChildren().remove(0);
                    } else {
                        from.setToken(Token.Operator.TIMES);
                    }
                    final Node func = new Node(root, "log");
                    func.addChild(base);
                    func.addChild(to);
                    to = func;
                    changed = true;
                    
                } else {
                    root.setToken(Token.getToken(Token.Operator.OR));
                    from.setToken(Token.getToken(Token.Operator.DIVIDE));
                    final Node child = from.getChildren().get(0);
                    from.getChildren().set(0, new Node(from, Value.getValue(1.0d)));
                    
                    final Node nuEq = new Node(root, Token.Operator.EQUAL);
                    final Node nu = new Node(nuEq, Token.Operator.POWER);
                    nuEq.addChild(nu);
                    nuEq.addChild(child);
                    nu.addChild(to);
                    nu.addChild(from);
                    
                    final Node and = new Node(root, Token.Operator.AND);
                    final Node andEq = new Node(and, Token.Operator.EQUAL);
                    final Node andEq2 = new Node(and, Token.Operator.EQUAL);
                    and.addChild(andEq);
                    and.addChild(andEq2);
                    final Node minus = new Node(andEq2, Token.Operator.UNARY_MINUS);
                    andEq2.addChild(minus);
                    andEq2.addChild(child.clone());
                    final Node modulo = new Node(andEq, Token.Operator.MODULO);
                    final Node base = new Node(modulo, Token.Operator.TIMES);
                    final List<Node> children = from.getChildren();
                    for(int i = 0, n = children.size(); i < n; ++i) {
                        final Node theChild = children.get(i);
                        base.addChild(theChild.clone());
                    }
                    if(base.getChildren().size() == 1) {
                        final Node child2 = base.getChildren().remove(0);
                        base.setToken(child2.getToken());
                        base.addChildren(child2.getChildren());
                    }
                    modulo.addChild(base);
                    modulo.addChild(new Node(modulo, Value.getValue(2.0d)));
                    andEq.addChild(modulo);
                    andEq.addChild(new Node(modulo, Value.getValue(0.0d)));
                    minus.addChild(nu.clone());
                    
                    to = and;
                    from = nuEq;
                    
                    changed = true;
                    break;
                }
                
            } else if(op.isFunction()) {
                final Node child = from.getChildren().get(0);
                from.clearChildren();
                from.setToken(invOp);
                from.addChild(to);
                to = from;
                from = child;
                changed = true;
                
            } else if(TreeComputer.isCommutative(op)) {
                final List<Node> children = Pool.getPool().getList();
                children.add(to);
                for(final Iterator<Node> it = from.getChildren().iterator(); it.hasNext(); ) {
                    final Node child = it.next(); 
                    if(TreeComputer.countVars(child, bind) != 1) {
                        children.add(child);
                        it.remove();
                    }
                }
                final Node child = from.getChildren().get(0);
                from.setToken(invOp);
                from.clearChildren();
                from.addChildren(children);
                if(from.getChildren().size() == 1) {
                    final Node child2 = from.getChildren().remove(0);
                    from.setToken(child2.getToken());
                    from.addChildren(child2.getChildren());
                }
                to = from;
                from = child;
                changed = true;
                
            } else if(TreeComputer.countVars(from.getChildren().get(0), bind) != 1){
                for(int i = 0; i < from.getChildren().size(); i++) {
                    final Node child = from.getChildren().get(i);
                    if(TreeComputer.countVars(child, bind) == 1) {
                        from.getChildren().set(i, to);
                        to.setParent(from);
                        to = child;
                        break;
                    }
                }
                final Node oldTo = to;
                to = from;
                from = oldTo;
                changed = true;
                
            } else {
                Node theChild = null;
                for(final Iterator<Node> it = from.getChildren().iterator(); it.hasNext();) {
                    final Node child = it.next();
                    if(TreeComputer.countVars(child, bind) == 1) {
                        theChild = child;
                        it.remove();
                    }
                }
                from.setToken(invOp);
                if(from.getChildren().size() == 1) {
                    final Node child2 = from.getChildren().remove(0);
                    from.setToken(child2.getToken());
                    from.addChildren(child2.getChildren());
                }
                final Node par = new Node(root, invOp);
                par.addChild(from);
                par.addChild(to);
                
                from = theChild;
                to = par;
                changed = true;
            }
        }
        
        root.clearChildren();
        root.addChild(from);
        root.addChild(to);
        //System.out.println(root);
        return changed;
    }
    
    
    public static Token inverseOp(final Token t) {
        if(t.isOperator()) {
            final Token.Operator op = t.getOperator();
            if(op == Token.Operator.PLUS) {
                return Token.getToken(Token.Operator.MINUS);
            } else if(op == Token.Operator.MINUS) {
                return Token.getToken(Token.Operator.PLUS);
            } else if(op == Token.Operator.TIMES) {
                return Token.getToken(Token.Operator.DIVIDE);
            } else if(op == Token.Operator.DIVIDE) {
                return Token.getToken(Token.Operator.TIMES);
            } else if(op == Token.Operator.POWER) {
                return Token.getToken(Token.Operator.POWER);
            } else {
                return null;
            }
        } else if(t.isFunction()) {
            return Token.getToken(Functions.inverse(t.getFunction()));
        } else {
            return null;
        }
    }
    
    
    public static int countVars(final Node root, final Binding bind) {
        if(root.getToken().isValue()) {
            final Value v = root.getToken().getValue();
            if(!v.isVariable()) {
                return 0;
            } else if(bind == null) {
                return 1;
            } else {
                return bind.hasBinding(v.getVariable()) ? 0 : 1;
            }
        } else {
            int count = 0;
            final List<Node> children = root.getChildren();
            for(int i = 0, n = children.size(); i < n; ++i) {
                final Node child = children.get(i);
                count += TreeComputer.countVars(child, bind);
            }
            return count;
        }
    }
    
    
    public static boolean setEvaluatability(final Node root, final Map<String, Function> functions) {
        return TreeComputer.setEvaluatability(root, functions, true);
    }
    
    
    public static boolean allowAssign(final Token op) {
        return op.isValue() ||
                (!(op.getOperator() == Token.Operator.OR
                                || op.getOperator() == Token.Operator.NOT
                                || op.getOperator() == Token.Operator.NOT_EQUAL)
                                && Functions.allowAssign(op.getFunction()));
    }
    
    
//    public static void cutByConstraints(Node root, Transition trans, Map functions) {
//        if(root.getValue() instanceof String) {
//            TreeComputer.testConstraints(root, null, functions, true);
//            TreeComputer.tryConstraints(root, trans, functions);
//            TreeComputer.cutByConstraints(root);
//            TreeComputer.setEvaluatability(root);
//        }
//    }
    
    
//    private static void cutByConstraints(Node root) {
//        if(root.allowAssign()) {
//            root.setValue(Value.getValue(true));
//            root.clearChildren();
//        } else {
//            for(Iterator it = root.getChildren().iterator(); it.hasNext();) {
//                Node child = (Node)it.next();
//                TreeComputer.cutByConstraints(child);
//            }
//        }
//    }
    
    
    public static void testConstraints(final Node root, final Binding bind,
                    final Map<String, Function> functions, final boolean first) {
        if(!root.getToken().isValue()) {
            if(first) {
                root.setAllowAssign(true);
            } else if(root.allowAssign()) {
                final Value result = TreeComputer.evaluate(root, functions, bind);
                root.setAllowAssign(result.isBoolean() && result.getBoolean());
            }
            
            if(first || !root.allowAssign()) {
                final List<Node> children = root.getChildren();
                for(int i = 0, n = children.size(); i < n; ++i) {
                    final Node child = children.get(i);
                    TreeComputer.testConstraints(child, bind, functions, first);
                }
            }
        } else if(first) {
            root.setAllowAssign(false);
        }
    }
    
    
    public static boolean setEvaluatability(final Node root, final Map<String, Function> functions,
                    boolean allowAssign) {
        boolean can = true;
        
        // this can be assigned if the father could be assigned AND the father's
        // value is not |, #, !. Thus if this is # but the father could be
        // assigned, this node can be assigned as well - but not its children.
        root.setAllowAssign(allowAssign);
        Function f = null;
        if(root.getToken().isFunction() && (functions != null)) {
             f = functions.get(Function.getFunctionName(root.getToken().getFunction(),
                                root.getChildren().size()));
            if(f != null) {
                allowAssign &= f.allowAssign();
            }
        }
        if(f == null) {
            allowAssign &= TreeComputer.allowAssign(root.getToken());
        }
        final List<Node> children = root.getChildren();
        for(int i = 0, n = children.size(); i < n; ++i) {
            final Node child = children.get(i);
            can &= TreeComputer.setEvaluatability(child, functions, allowAssign);
        }
        
        if(root.getToken().isFunction()) {
            if(f != null) {
                can &= f.canBeEvaluatedStatically();
            } else {
                can &= Functions.canBeEvaluatedStatically(root.getToken().getFunction());
            }
        }
        
        root.setCanBeEvaluatedStatically(can);
        
        return can;
    }
    
    
    public static boolean extractConstraints(final Node root, final Map<String, Function> functions) {
        boolean changed = false;
        if(root.hasConstraints() && root.getToken().isOperator()) {
//            System.out.println(root.getConstraints());
            for(final Map.Entry<String, ValueList> e : root.getConstraints().entrySet()) {
                final String var = e.getKey();
                final ValueList values = e.getValue();
                final ValueList constraint = TreeComputer.extractConstraints(root,
                                Value.getValue(var), values, functions, root.getConstraints());
//                System.out.println("  " + var + " --> " + values);
//                System.out.println("  " + var + " ==> " + constraint);
                if(values.size() != constraint.size()) {
                    changed = true;
                }
                root.constrain(var, constraint);
            }
        }
        return changed;
    }
    
    
    public static void extractConstraints(final Node root,
                    final Map<String, Function> functions,
                    final Binding bind) {
        if(bind.hasFreeVariables() && root.getToken().isOperator()) {
            //System.out.println(root);
            final List<String> vars = bind.getFreeVariables();
            for(int i = 0, n = vars.size(); i < n; ++i) {
                final String var = vars.get(i);
                final ValueList values = bind.getConstraint(var);
                final ValueList constraint = TreeComputer.extractConstraints(root,
                                Value.getValue(var), values, functions, bind.getConstraints());
                //System.out.println("  " + var + " --> " + constraint);
                bind.constrain(var, constraint, true);
            }
        }
    }
    
    
    public static boolean contains(final Node node, final Value value) {
        if(!node.getToken().isValue()) {
            final List<Node> children = node.getChildren();
            for(int i = 0, n = children.size(); i < n; ++i) {
                final Node child = children.get(i);
                if(TreeComputer.contains(child, value)) {
                    return true;
                }
            }
            return false;
        } else {
            return node.getToken().getValue().equals(value);
        }
    }
    
    
    public static Value getValueOfInTuple(final Node tVar, final Tuple tVal,
                    final Value var) {
        if(tVar.getChildren().size() != tVal.size()) {
            return null;
        }
        for(int i = 0; i < tVar.getChildren().size(); i++) {
            final Node vVar = tVar.getChildren().get(i);
            final Value vVal = tVal.getEntries().get(i);
            if(vVar.getToken().getOperator() == Token.Operator.LIST_START && vVal.isTuple()) {
                return TreeComputer.getValueOfInTuple(vVar, vVal.getTuple(), var);
            } else if(vVar.getToken().getValue() == var) {
                return vVal;
            }
        }
        return null;
    }
    
    
    public static ValueList extractConstraints(final Node node, final Value var,
                    final ValueList values, final Map<String, Function> functions,
                    final Map<String, ValueList> constraints) {
        if(node.getToken().isOperator()) {
            final Token.Operator op = node.getToken().getOperator();
            ValueList newvalues = null;
            
            if((op == Token.Operator.EQUAL) || (op == Token.Operator.NOT_EQUAL)) {
                final List<Node> children = node.getChildren();
                final Token left = children.get(0).getToken();
                final Token right = children.get(1).getToken();
//                System.out.println(children.get(1));
                if(left.isOperator() && left.getOperator() == Token.Operator.LIST_START && TreeComputer.contains(node, var)) {
                    if((!right.isValue() || !right.getValue().isVariable())
                                    && node.canBeEvaluatedStatically()) {
                        Value val = TreeComputer.evaluate(children.get(1), functions, null);
//                        System.out.println("-> " + children.get(0) + " = " + val + " => " + var);
                        if(val == null) {
                            newvalues = values;
                        } else if(!val.isTuple()) {
                            newvalues = Pool.getPool().getValueList();
                        } else if(op == Token.Operator.EQUAL) {
                            newvalues = Pool.getPool().getValueList();
                            val = TreeComputer.getValueOfInTuple(children.get(0), val.getTuple(), var);
//                            System.out.println(children.get(0) + " = " + val + " => " + var);
                            if((val != null) && values.contains(val) || Value.containsAny(values)) {
                                newvalues.add(val);
                            }
                        } else if(op == Token.Operator.NOT_EQUAL) {
                            newvalues = Pool.getPool().getValueList(values);
                            val = TreeComputer.getValueOfInTuple(children.get(0), val.getTuple(), var);
//                            System.out.println(children.get(0) + " = " + val + " => " + var);
                            if(val != null) {
                                newvalues.remove(val);
                            } else {
                                newvalues.clear();
                            }
                        }
                    }
                } else if(right.getOperator() == Token.Operator.LIST_START && TreeComputer.contains(node, var)) {
                    if((!left.isValue() || !left.getValue().isVariable())
                                    && node.canBeEvaluatedStatically()) {
                        Value val = TreeComputer.evaluate(children.get(0), functions, null);
//                        System.out.println("-> " + val + " = " + children.get(1) + " => " + var);
                        if(val == null) {
                            newvalues = values;
                        } else if(!val.isTuple()) {
                            newvalues = Pool.getPool().getValueList();
                        } else if(op == Token.Operator.EQUAL) {
                            newvalues = Pool.getPool().getValueList();
                            val = TreeComputer.getValueOfInTuple(children.get(1), val.getTuple(), var);
//                            System.out.println(val + " = " + children.get(1) + " => " + var);
                            if((val != null) && values.contains(val) || Value.containsAny(values)) {
                                newvalues.add(val);
                            }
                        } else if(op == Token.Operator.NOT_EQUAL) {
                            newvalues = Pool.getPool().getValueList(values);
                            val = TreeComputer.getValueOfInTuple(children.get(1), val.getTuple(), var);
//                            System.out.println(val + " = " + children.get(1) + " => " + var);
                            if(val != null) {
                                newvalues.remove(val);
                            } else {
                                newvalues.clear();
                            }
                        }
                    }
                } else if(var.equals(left)) {
                    if(right.isValue() && right.getValue().isVariable()) {
                        final String rvar = right.getValue().getVariable();
                        final ValueList curvalues = constraints.get(rvar);
                        if(curvalues == null) {
                            newvalues = values;
                        } else if(op == Token.Operator.EQUAL) {
                            if(values != null) {
                                newvalues = Pool.getPool().getValueList(values);
                                newvalues.retainAll(curvalues);
                            } else {
                                newvalues = Pool.getPool().getValueList(curvalues);
                            }
                        } else if((op == Token.Operator.NOT_EQUAL) && (curvalues.size() == 1)) {
                            newvalues = Pool.getPool().getValueList(values);
                            newvalues.removeAll(curvalues);
                        }
                        
                    } else if(node.canBeEvaluatedStatically()) {
                        final Value val = TreeComputer.evaluate(children.get(1), functions, null);
                        if(val == null) {
                            newvalues = values;
                        } else if(op == Token.Operator.EQUAL) {
                            newvalues = Pool.getPool().getValueList();
                            if(values.contains(val)) {
                                newvalues.add(val);
                            }
                        } else if(op == Token.Operator.NOT_EQUAL) {
                            newvalues = Pool.getPool().getValueList(values);
                            newvalues.remove(val);
                        }
                    }
                    
                } else if(var.equals(right)) {
                    if(left.isValue() && left.getValue().isVariable()) {
                        final String lvar = left.getValue().getVariable();
                        final ValueList curvalues = constraints.get(lvar);
                        if(curvalues == null) {
                            newvalues = values;
                            if(values != null) {
                                newvalues = Pool.getPool().getValueList(values);
                                newvalues.retainAll(curvalues);
                            } else {
                                newvalues = Pool.getPool().getValueList(curvalues);
                            }
                        } else if((op == Token.Operator.NOT_EQUAL) && (curvalues.size() == 1)) {
                            newvalues = Pool.getPool().getValueList(values);
                            newvalues.removeAll(curvalues);
                        }
                    } else if(node.canBeEvaluatedStatically()) {
                        final Value val = TreeComputer.evaluate(children.get(0), functions, null);
                        if(val == null) {
                            newvalues = values;
                        } else if(op == Token.Operator.EQUAL) {
                            newvalues = Pool.getPool().getValueList();
                            if(values.contains(val)) {
                                newvalues.add(val);
                            }
                        } else if(op == Token.Operator.NOT_EQUAL) {
                            newvalues = Pool.getPool().getValueList(values);
                            newvalues.remove(val);
                        }
                    }
                }
                
            } else if(op == Token.Operator.NOT) {
                final Node child = node.getChildren().get(0);
                if(TreeComputer.contains(child, var)) {
                    newvalues = Pool.getPool().getValueList(values);
                    final ValueList curvals = TreeComputer.extractConstraints(child, var,
                                    values, functions, constraints);
                    newvalues.removeAll(curvals);
                } else {
                    return values; 
                }
                
            } else if(op == Token.Operator.AND) {
                final List<Node> children = node.getChildren();
                for(int i = 0, n = children.size(); i < n; ++i) {
                    final Node child = children.get(i);
                    final ValueList curvals = TreeComputer.extractConstraints(child, var,
                                    values, functions, constraints);
                    if(newvalues == null) {
                        newvalues = Pool.getPool().getValueList(curvals);
                    } else {
                        newvalues.retainAll(curvals);
                    }
                }
                
            } else if(op == Token.Operator.OR) {
                final List<Node> children = node.getChildren();
                for(int i = 0, n = children.size(); i < n; ++i) {
                    final Node child = children.get(i);
                    final ValueList curvals = TreeComputer.extractConstraints(child, var,
                                    values, functions, constraints);
                    if(Value.containsAny(values)) {
                        if(newvalues == null) {
                            newvalues = Pool.getPool().getValueList();
                            newvalues.add(Value.V_ANY);
                        } else {
                            newvalues.clear();
                            newvalues.add(Value.V_ANY);
                        }
                    } else {
                        for(int ii = 0, in = curvals.size(); ii < in; ++ii) {
                            final Value val = curvals.get(ii);
                            if(newvalues == null) {
                                newvalues = Pool.getPool().getValueList();
                                newvalues.add(val);
                            } else if(!newvalues.contains(val)) {
                                newvalues.add(val);
                            }
                        }
                    }
                }
                
            } else if(node.canBeEvaluatedStatically() 
                    && ((op == Token.Operator.LESS) || (op == Token.Operator.GREATER)
                        || (op == Token.Operator.LESS_EQUAL) || (op == Token.Operator.GREATER_EQUAL))) {
                final Token left = node.getChildren().get(0).getToken();
                final Token right = node.getChildren().get(1).getToken();
                if(var.equals(left)) {
                    final Value val = TreeComputer.evaluate(node.getChildren().get(1), functions, null);
                    if(val != null) {
                        newvalues = Pool.getPool().getValueList(values);
                        final double value = val.getNumber();
                        for(final Iterator<Value> it = newvalues.iterator(); it.hasNext(); ) {
                            final Value v = it.next();
                            if(!v.isAny()) {
                                if(!v.isNumber() || (((op == Token.Operator.GREATER)  && (v.getNumber() <= value))
                                                 ||  ((op == Token.Operator.LESS)  && (v.getNumber() >= value))
                                                 ||  ((op == Token.Operator.GREATER_EQUAL) && (v.getNumber() <  value))
                                                 ||  ((op == Token.Operator.LESS_EQUAL) && (v.getNumber() >  value)))) {
                                    it.remove();
                                }
                            }
                        }
                    }
                } else if(var.equals(right)) {
                    final Value val = TreeComputer.evaluate(node.getChildren().get(0), functions, null);
                    if(val != null) {
                        newvalues = Pool.getPool().getValueList(values);
                        final double value = val.getNumber();
                        for(final Iterator<Value> it = newvalues.iterator(); it.hasNext(); ) {
                            final Value v = it.next();
                            if(!v.isAny()) {
                                if(!v.isNumber() || (((op == Token.Operator.GREATER)  && (v.getNumber() <= value))
                                        ||  ((op == Token.Operator.LESS)  && (v.getNumber() >= value))
                                        ||  ((op == Token.Operator.GREATER_EQUAL) && (v.getNumber() <  value))
                                        ||  ((op == Token.Operator.LESS_EQUAL) && (v.getNumber() >  value)))) {
                                   it.remove();
                               }
                            }
                        }
                    }
                }
            }
            
            if(newvalues == null) {
                newvalues = Pool.getPool().getValueList(values);
            }
            return newvalues;
            
        } else if(var.equals(node.getToken())) {
            final Node parent = node.getParent();
            final ValueList vals = Pool.getPool().getValueList(values);
            if((parent != null) && parent.getToken().getOperator() == Token.Operator.NOT) {
                vals.remove(Value.getValue(false));
            } else if((parent == null) || TreeComputer.logicalOp(parent.getToken().getOperator())) {
                if(vals.contains(Value.getValue(true)) || Value.containsAny(vals)) {
                    vals.clear();
                    vals.add(Value.getValue(true));
                } else {
                    vals.clear();
                }
            }
            return vals;
        } else {
            return values;
        }
    }
    
    
    // special algorithm without name
    public static boolean sawn(final Node root) {
        final Token.Operator op = root.getToken().getOperator();
        if(op != Token.Operator.AND && op != Token.Operator.OR) {
            return false;
        }
        
        boolean changed = false;
        final List<Node> children = root.getChildren();
        for(int i = 0, n = children.size(); i < n; ++i) {
            final Node child = children.get(i);
            changed = TreeComputer.sawn(child) || changed;
        }
        
        boolean delete = false;
        for(int i = 0, n = children.size(); i < n; ++i) {
            final Node child = children.get(i);
            for(int ii = 0; ii < n; ++ii) {
                final Node child2 = children.get(ii);
                if((child != child2)
                        && (((op == Token.Operator.AND) && TreeComputer.excluding(child, child2))
                         || ((op == Token.Operator.OR) && TreeComputer.satisfying(child, child2)))) {
                    delete = true;
                    break;
                }
            }
            if(delete) {
                break;
            }
        }
        if(delete) {
            root.clearChildren();
            root.setToken(Value.getValue(op == Token.Operator.OR));
        }
        
        
        changed = changed || delete;
        
        return changed;
    }
    
    
//    public static void removeSimpleParts(final Node root, final Binding bind) {
//        if (root.getValue() instanceof String) {
//            final List<Node> children = root.getChildren();
//            for(int i = 0, n = children.size(); i < n; ++i) {
//                final Node child = children.get(i);
//                TreeComputer.removeSimpleParts(child, bind);
//            }
//        }
//    }
    
    
    public static int order(final Node root, final Binding bind) {
        if(root == null) {
            return 0;
        }
        try {
            if(!root.getToken().isValue()) {
                final Token op = root.getToken();
                int variables = 0;
                final double[] count = new double[root.getChildren().size()];
                final List<Node> children = root.getChildren();
                for(int i = 0; i < count.length; i++) {
                    final Node child = children.get(i);
                    count[i] = TreeComputer.order(child, bind);
                    if(child.getToken().getOperator() == Token.Operator.EQUAL) {
                        count[i] -= 0.5;
                    }
                    variables += count[i];
                }
                if(TreeComputer.isCommutative(op)) {
                    for(int i = 0; i < count.length - 1; i++) {
                        int min = i;
                        for(int j = i+1; j < count.length; j++) {
                            if(count[j] < count[min]) {
                                min = j;
                            }
                        }
                        if(min != i) {
                            final Node o = children.get(i);
                            children.set(i, children.get(min));
                            children.set(min, o);
                            final double buf = count[i];
                            count[i] = count[min];
                            count[min] = buf;
                        }
                    }
                }
                
                return variables;
            } else {
                final Value v = root.getToken().getValue();
                if(!v.isVariable()) {
                    return 0;
                } else if(bind == null) {
                    return 1;
                } else {
                    return bind.hasBinding(v.getVariable()) ? 0 : 1;
                }
            }
        } catch(Exception e) {
            throw new RuntimeException("ERROR ordering " + root, e);
        }
    }
    
    
    private static boolean assigns(final Node root, final Binding bind) {
        if((root.getToken().getOperator() == Token.Operator.EQUAL)
                        || (root.getToken().getOperator() == Token.Operator.NOT_EQUAL)) {
            final Token left = root.getChildren().get(0).getToken();
            final Token right = root.getChildren().get(1).getToken();
            if((left.isValue() && left.getValue().isVariable()
                          && bind.hasBinding(left.getValue().getVariable()))
                    || (right.isValue() && right.getValue().isVariable()
                          && bind.hasBinding(right.getValue().getVariable()))) {
                return true;
            }
        } else if(root.getToken().isValue() && root.getToken().getValue().isVariable()
                        && bind.hasBinding(root.getToken().getValue().getVariable())
                        && (root.getParent() != null) && root.getParent().getToken().isOperator()) {
            final Operator par = root.getParent().getToken().getOperator();
            if((par == Token.Operator.EQUAL) || (par == Token.Operator.AND)
                            || (par == Token.Operator.OR) || (par == Token.Operator.NOT)) {
                return true;
            }
        }
        final List<Node> children = root.getChildren();
        for(int i = 0, n = children.size(); i < n; ++i) {
            final Node child = children.get(i);
            if(TreeComputer.assigns(child, bind)) {
                return true;
            }
        }
        return false;
    }
    
    
    public static boolean staticEvaluateWithBinding(final Node root,
                    final Map<String, Function> functions, Binding bind) {
        if((root != null) && !root.getToken().isValue()) {
            final Node parent = root.getParent();
            if(bind == null) {
                bind = Pool.getPool().getBinding();
            }
            
            if((parent != null) && !parent.getToken().isValue()
                            && ((parent.getToken().getOperator() == Token.Operator.OR)
                            || parent.getToken().isFunction())) {
                root.setParent(null);
                bind = bind.clone();
//                bind.clearConstraints();
            }
            
            boolean changed = false;
            if(TreeComputer.canBeEvaluatedStatically(root)) {
                try {
                    TreeComputer.extractBinding(root, bind, functions, false);
                    final Value v = TreeComputer.evaluate(root, functions, bind);
                    if(v != null) {
                        final boolean assigns = TreeComputer.assigns(root, bind);
                        if(!assigns) {
                            root.setToken(v);
                            root.clearChildren();
                            changed = true;
                        }
//                        if(!bind.hasBindings() || (v.isBoolean() && !v.getBoolean())) {
//                            root.setValue(v);
//                            root.clearChildren();
//                            changed = true;
//                        }
                    }
                } catch(Exception e) {
                    // ignore;
                }
            }
            
            for(final Iterator<Node> it = root.getChildren().iterator(); it.hasNext(); ) {
                final Node child = it.next();
                changed = TreeComputer.staticEvaluateWithBinding(child, functions, bind)
                            || changed;
            }
            
            root.setParent(parent);
            return changed;
        } else {
            return false;
        }
    }
    
    
    public static boolean staticEvaluate(final Node root, final Map<String, Function> functions) {
        if((root != null) && !root.getToken().isValue()) {
            boolean changed = false;
            
            final List<Node> children = root.getChildren();
            for(int i = 0, n = children.size(); i < n; ++i) {
                final Node child = children.get(i);
                changed = TreeComputer.staticEvaluate(child, functions) || changed;
            }
            
            if(TreeComputer.canBeEvaluatedStatically(root)) {
                try {
                    final Value v = TreeComputer.evaluate(root, functions, null);
                    if(v != null) {
//                        System.out.println("replacing: "+ root + " => " + v);
                        root.setToken(v);
                        root.clearChildren();
                        changed = true;
                    }
                } catch(Exception e) {
                    // ignore;
                }
            }
            
            return changed;
        } else {
            return false;
        }
    }
    
    
    public static boolean canBeEvaluatedStatically(final Node node) {
        if(!node.getToken().isValue()
                && !Functions.canBeEvaluatedStatically(node.getToken().getFunction())) {
            return false;
        }
        final List<Node> children = node.getChildren();
        for(int i = 0, n = children.size(); i < n; ++i) {
            final Node child = children.get(i);
            if(!TreeComputer.canBeEvaluatedStatically(child)) {
                return false;
            }
        }
        return true;
    }
    
    
    public static boolean isCommutative(final Token op) {
        return op.isOperator() && ((op.getOperator() == Token.Operator.AND)
                || (op.getOperator() == Token.Operator.OR)
                || (op.getOperator() == Token.Operator.PLUS)
                || (op.getOperator() == Token.Operator.TIMES)
                || (op.getOperator() == Token.Operator.EQUAL)
                || (op.getOperator() == Token.Operator.NOT_EQUAL));
    }
    
    
    public static boolean isReversable(final Token.Operator op) {
        return (op == Token.Operator.LESS)
                || (op == Token.Operator.GREATER)
                || (op == Token.Operator.LESS_EQUAL)
                || (op == Token.Operator.GREATER_EQUAL);
    }
    
    
    public static Token.Operator reversion(final Token.Operator op) {
        if(op == Token.Operator.LESS) {
            return Token.Operator.GREATER;
        } else if(op == Token.Operator.GREATER) {
            return Token.Operator.LESS;
        } else if(op == Token.Operator.LESS_EQUAL) {
            return Token.Operator.GREATER_EQUAL;
        } else if(op == Token.Operator.GREATER_EQUAL) {
            return Token.Operator.LESS_EQUAL;
        } else {
            throw new RuntimeException(op + " cannot be reversed.");
        }
    }
    
    
    public static boolean simplify(final Node root) {
        if((root != null) && !root.getToken().isValue() && root.hasChildren()) {
            boolean changed = false;
            
            final Token op = root.getToken();
            final Node firstChild = root.getChildren().get(0);
            if(op.getOperator() == Token.Operator.NOT) {
                // replace !(X <= Y) by (X > Y) and so on
                changed = true;
                if(firstChild.getToken().getOperator() == Token.Operator.NOT) {
                    final Node childChild = firstChild.getChildren().get(0);
                    root.setToken(childChild.getToken());
                    root.clearChildren();
                    firstChild.clearChildren();
                    root.addChildren(childChild.getChildren());
                    //System.out.println("replaced !( ! )");
                } else if(firstChild.getToken().getOperator() == Token.Operator.NOT_EQUAL) {
                    root.setToken(Token.Operator.EQUAL);
                    root.clearChildren();
                    root.addChildren(firstChild.getChildren());
                    //System.out.println("replaced !( # )");
                } else if(firstChild.getToken().getOperator() == Token.Operator.EQUAL) {
                    root.setToken(Token.Operator.NOT_EQUAL);
                    root.clearChildren();
                    root.addChildren(firstChild.getChildren());
                    //System.out.println("replaced !( = )");
                } else if(firstChild.getToken().getOperator() == Token.Operator.LESS) {
                    root.setToken(Token.Operator.GREATER_EQUAL);
                    root.clearChildren();
                    root.addChildren(firstChild.getChildren());
                    //System.out.println("replaced !( = )");
                } else if(firstChild.getToken().getOperator() == Token.Operator.LESS_EQUAL) {
                    root.setToken(Token.Operator.GREATER);
                    root.clearChildren();
                    root.addChildren(firstChild.getChildren());
                    //System.out.println("replaced !( = )");
                } else if(firstChild.getToken().getOperator() == Token.Operator.GREATER) {
                    root.setToken(Token.Operator.LESS_EQUAL);
                    root.clearChildren();
                    root.addChildren(firstChild.getChildren());
                    //System.out.println("replaced !( = )");
                } else if(firstChild.getToken().getOperator() == Token.Operator.GREATER_EQUAL) {
                    root.setToken(Token.Operator.LESS);
                    root.clearChildren();
                    root.addChildren(firstChild.getChildren());
                    //System.out.println("replaced !( = )");
                } else {
                    changed = false;
                }
            } else if(op.getOperator() == Token.Operator.EQUAL
                            || op.getOperator() == Token.Operator.NOT_EQUAL) {
                // replace (X = true) by (X)
                // replace (X = false) by (!X)
                // replace (X # true) by (!X)
                // replace (X # false) by (X)
                final Node left = root.getChildren().get(0);
                final Node right = root.getChildren().get(1);
                final boolean tt = op.getOperator() == Token.Operator.EQUAL;
                if(Value.getValue(tt).equals(left.getToken().getValue())) {
                    root.setToken(right.getToken());
                    root.clearChildren();
                    root.addChildren(right.getChildren());
                    changed = true;
                } else if(Value.getValue(tt).equals(right.getToken().getValue())) {
                    root.setToken(left.getToken());
                    root.clearChildren();
                    root.addChildren(left.getChildren());
                    changed = true;
                } else if(Value.getValue(!tt).equals(left.getToken().getValue())) {
                    root.setToken(Token.Operator.NOT);
                    root.clearChildren();
                    root.addChild(right);
                    changed = true;
                } else if(Value.getValue(!tt).equals(right.getToken().getValue())) {
                    root.setToken(Token.Operator.NOT);
                    root.clearChildren();
                    root.addChild(left);
                    changed = true;
                }
            } else if(op.getOperator() == Token.Operator.MINUS && (root.getChildren().size() == 2)) {
                final Node n = root.getChildren().get(1);
                if(n.getToken().getOperator() == Token.Operator.UNARY_MINUS) {
                    // (X - (-Y)) ==> (X + Y)
                    root.setToken(Token.Operator.PLUS);
                    final Node n2 = n.getChildren().get(0);
                    root.getChildren().set(1, n2);
                } else if(n.getToken().isValue()
                                && n.getToken().getValue().isNumber()
                                && (n.getToken().getValue().getNumber() < 0)) {
                    // (X - (-d)) ==> (X + d)
                    root.setToken(Token.Operator.PLUS);
                    n.setToken(Value.getValue(-n.getToken().getValue().getNumber()));
                }
            }
            
//            else if(op.equals(RPNComputer.MINUS)) {
//                root.setValue(RPNComputer.PLUS);
//                op = RPNComputer.PLUS;
//                Node son = new Node(root, RPNComputer.UNARY_MINUS);
//                Node left = (Node)root.getChildren().get(0);
//                Node right = (Node)root.getChildren().get(1);
//                son.addChild(right);
//                root.clearChildren();
//                root.addChild(left);
//                root.addChild(son);
//                changed = true;
//            }
            
            boolean stringConc = false;
            final boolean isAdd = op.getOperator() == Token.Operator.PLUS;
            final List<Node> children = root.getChildren();
            for(int i = 0, n = children.size(); i < n; ++i) {
                final Node child = children.get(i);
                changed = TreeComputer.simplify(child) || changed;
                if(isAdd && !stringConc
                         && (child.getToken().getOperator() == Token.Operator.STRING_CONCATENATION
                             || (child.getToken().isValue()
                                  && child.getToken().getValue().isString()))) {
                    stringConc = true;
                }
            }
//            if(!stringConc && isAdd && (root.getParent() != null)
//                            && root.getParent().getValue().equals(RPNComputer.STRING_CONCATENATION)) {
//                stringConc = true;
//            }
                
            if(stringConc) {
                root.setToken(Token.Operator.STRING_CONCATENATION);
                changed = true;
            } else if(isAdd) {
                if(children.size() == 2) {
                    final Node leftchild = children.get(0);
                    final Node rightchild = children.get(1);
                    if(leftchild.getToken().isValue() && leftchild.getToken().getValue().isNumber()
                                    && (leftchild.getToken().getValue().getNumber() == 0.0)) {
                        changed = true;
                        root.setToken(rightchild.getToken());
                        root.clearChildren();
                        root.addChildren(rightchild.getChildren());
                        rightchild.clearChildren();
                    } else if(rightchild.getToken().isValue() && rightchild.getToken().getValue().isNumber()
                                    && (rightchild.getToken().getValue().getNumber() == 0.0)) {
                        changed = true;
                        root.setToken(leftchild.getToken());
                        root.clearChildren();
                        root.addChildren(leftchild.getChildren());
                        leftchild.clearChildren();
                    }
                }
            }
            
            return changed;
        } else {
            return false;
        }
    }
    
    
    public static boolean excluding(final Node a, final Node b) {
        if(a.getToken().isValue() || b.getToken().isValue()) {
            return false;
        }
        if(b.getToken().getOperator() == Token.Operator.NOT) {
            final Node c = b.getChildren().get(0);
            return a.equals(c);
        } else if(a.getToken().getOperator() == Token.Operator.NOT) {
            final Node c = a.getChildren().get(0);
            return b.equals(c);
        }
        final Token.Operator opa = a.getToken().getOperator();
        final Token.Operator opb = b.getToken().getOperator();
        if(TreeComputer.isReversable(opa)
                  && TreeComputer.excluding(TreeComputer.reversion(opa), opb)) {
            final List<Node> rev = Node.reverse(b.getChildren());
            final Node buf = new Node(null, Token.getToken(opa), rev);
            final boolean ret = a.equals(buf);
            return ret;
        } else if(TreeComputer.excluding(opa, opb)) {
            b.setToken(opa);
            final boolean ex = a.equals(b);
            b.setToken(opb);
            return ex;
        }
        return false;
    }
    
    
    public static boolean excluding(final Token.Operator op1, final Operator op2) {
        return (op1 == Token.Operator.GREATER && (op2 == Token.Operator.LESS || op2 == Token.Operator.LESS_EQUAL))
            || (op1 == Token.Operator.LESS && (op2 == Token.Operator.GREATER || op2 == Token.Operator.GREATER_EQUAL))
            || (op2 == Token.Operator.GREATER && (op1 == Token.Operator.LESS || op1 == Token.Operator.LESS_EQUAL))
            || (op2 == Token.Operator.LESS && (op1 == Token.Operator.GREATER || op1 == Token.Operator.GREATER_EQUAL));
    }
    
    
    public static boolean satisfying(final Node a, final Node b) {
        if(a.getToken().isValue() || b.getToken().isValue()) {
            return false;
        }
        final Token.Operator opa = a.getToken().getOperator();
        final Token.Operator opb = b.getToken().getOperator();
        if(opb == Token.Operator.NOT) {
            final Node c = b.getChildren().get(0);
            return a.equals(c);
        } else if(opa == Token.Operator.NOT) {
            final Node c = a.getChildren().get(0);
            return b.equals(c);
        }
        if(TreeComputer.isReversable(opa)
                 && TreeComputer.satisfying(TreeComputer.reversion(opa), opb)) {
            final List<Node> rev = Node.reverse(b.getChildren());
            final Node buf = new Node(null, Token.getToken(opa), rev);
            final boolean ret = a.equals(buf);
            return ret;
        } else if(TreeComputer.satisfying(opa, opb)) {
            b.setToken(opa);
            final boolean ex = a.equals(b);
            b.setToken(opb);
            return ex;
        }
        
        return false;
    }
    
    
    public static boolean satisfying(final Token.Operator op1, final Token.Operator op2) {
        return ((op1 == Token.Operator.GREATER_EQUAL) && ((op2 == Token.Operator.LESS) || (op2 == Token.Operator.LESS_EQUAL)))
            || ((op1 == Token.Operator.LESS_EQUAL) && ((op2 == Token.Operator.GREATER) || (op2 == Token.Operator.GREATER_EQUAL)))
            || ((op2 == Token.Operator.GREATER_EQUAL) && ((op1 == Token.Operator.LESS) || (op1 == Token.Operator.LESS_EQUAL)))
            || ((op2 == Token.Operator.LESS_EQUAL) && ((op1 == Token.Operator.GREATER) || (op1 == Token.Operator.GREATER_EQUAL)));
    }
    
    
    public static boolean flatten(final Node root, final Map<String, Function> functions) {
        if(root == null) {
            return false;
        }
        
        if(root.getToken().isOperator()
                        && TreeComputer.canBeEvaluatedStatically(root)) {
            boolean changed = false;
            final Token.Operator op = root.getToken().getOperator();
            final int arity = root.getChildren().size();
            final List<Node> newChildren = Pool.getPool().getList();
            boolean mayBeTrue = false;
            boolean evaluated = false;
            final List<Node> childrenX = root.getChildren();
            for(int i = 0, n = childrenX.size(); i < n; ++i) {
                final Node child = childrenX.get(i);
                changed = TreeComputer.flatten(child, functions) || changed;
                
                final boolean contains = newChildren.contains(child);
                if((op == Token.Operator.AND) && Value.V_TRUE.equals(child.getToken().getValue())) {
                    // remove
                    mayBeTrue = true;
                    //System.out.println("removed " + child);
                } else if((op == Token.Operator.OR)&& Value.V_FALSE.equals(child.getToken().getValue())) {
                    // remove
                    mayBeTrue = false;
                    //System.out.println("removed " + child);
                } else if((op == Token.Operator.AND) && Value.V_FALSE.equals(child.getToken().getValue())) {
                    // remove
                    newChildren.clear();
                    newChildren.add(new Node(null, Value.getValue(false)));
                    evaluated = true;
                    //System.out.println("removed " + child);
                    break;
                } else if((op == Token.Operator.OR) && Value.V_TRUE.equals(child.getToken().getValue())) {
                    // remove
                    newChildren.clear();
                    newChildren.add(new Node(null, Value.getValue(true)));
                    evaluated = true;
                    //System.out.println("removed " + child);
                    break;
                } else if(contains && ((op == Token.Operator.AND) || (op == Token.Operator.OR))) {
                    // remove dublicate operands of &, |.
                    //System.out.println("removed " + child);
                } else if(contains && ((op == Token.Operator.EQUAL)
                                || (op == Token.Operator.LESS_EQUAL) || (op == Token.Operator.GREATER_EQUAL))) {
                    newChildren.clear();
                    newChildren.add(new Node(null, Value.getValue(true)));
                    evaluated = true;
                    //System.out.println("removed " + child);
                    break;
                } else if(contains && ((op == Token.Operator.NOT_EQUAL)
                                || (op == Token.Operator.LESS) || (op == Token.Operator.GREATER))) {
                    newChildren.clear();
                    newChildren.add(new Node(null, Value.getValue(false)));
                    evaluated = true;
                    //System.out.println("removed " + child);
                    break;
                } else if((op == child.getToken().getOperator()) &&
                                ((op == Token.Operator.AND) || (op == Token.Operator.OR)
                                 || (op == Token.Operator.STRING_CONCATENATION) || (op == Token.Operator.PLUS)
                                 || (op == Token.Operator.MINUS) || (op == Token.Operator.TIMES)
                                 || (op == Token.Operator.DIVIDE) || (op == Token.Operator.MODULO)
                                 || (op == Token.Operator.POWER))) {
                    newChildren.addAll(child.getChildren());
                    changed = true;
                } else {
                    newChildren.add(child);
                }
            }
            
            if(!evaluated) {
                if(op == Token.Operator.UNARY_MINUS) {
                    try {
                        final Value v = TreeComputer.evaluate(newChildren.get(0), functions, null);
                        if(v != null) {
                            final Value v2 = Value.getValue(-v.getNumber());
                            root.setToken(v2);
                            root.clearChildren();
                            return true;
                        }
                    } catch(final RuntimeException e) {
                        // ignore;
                    }
                }
                
                if((op == Token.Operator.PLUS) || (op == Token.Operator.TIMES)
                                || (op == Token.Operator.STRING_CONCATENATION)
                                || (op == Token.Operator.MINUS)
                                || (op == Token.Operator.DIVIDE)
                                || (op == Token.Operator.POWER)) {
                    final List<Node> children = Pool.getPool().getList(newChildren);
                    newChildren.clear();
                    boolean useBuffer = false;
                    Value buffer = null;
                    for(int i = 0, n = children.size(); i < n; ++i) {
                        final Node child = children.get(i);
                        if((op == Token.Operator.PLUS) || ((op == Token.Operator.MINUS) && (i > 0))) {
                            try {
                                final Value v = TreeComputer.evaluate(child, functions, null);
                                if(v != null) {
                                    if(useBuffer) {
                                        buffer = Value.getValue(buffer.getNumber() + v.getNumber());
                                    } else {
                                        buffer = v;
                                        useBuffer = true;
                                    }
                                } else {
                                    newChildren.add(child);
                                }
                            } catch(Exception e) {
                                newChildren.add(child);
                            }
                        } else if((op == Token.Operator.TIMES) || (((op == Token.Operator.DIVIDE) || (op == Token.Operator.POWER)) && (i > 0))) {
                            try {
                                final Value v = TreeComputer.evaluate(child, functions, null);
                                if(v != null) {
                                    if(useBuffer) {
                                        buffer = Value.getValue(buffer.getNumber() * v.getNumber());
                                    } else {
                                        buffer = v;
                                        useBuffer = true;
                                    }
                                } else {
                                    newChildren.add(child);
                                }
                            } catch(Exception e) {
                                newChildren.add(child);
                            }
                        } else if(op == Token.Operator.STRING_CONCATENATION) {
                            boolean error = false;
                            try {
                                final Value v = TreeComputer.evaluate(child, functions, null);
                                if(v != null) {
                                    if(useBuffer) {
                                        buffer = Value.getValue(buffer.getNumber() + v.getNumber());
                                    } else {
                                        buffer = v;
                                        useBuffer = true;
                                    }
                                } else {
                                    error = true;
                                }
                            } catch(Exception e) {
                                error = true;
                            }
                            if(error) {
                                if(useBuffer) {
                                    newChildren.add(new Node(null, buffer));
                                }
                                newChildren.add(child);
                                useBuffer = false;
                            }
                        } else {
                            newChildren.add(child);
                        }
                    }
                    if(useBuffer) {
                        newChildren.add(new Node(null, buffer));
                    }
                }
            }
            
            if((newChildren.size() == 1) && (arity > 1)) {
                root.clearChildren();
                final Node n = newChildren.get(0);
                root.setToken(n.getToken());
                root.addChildren(n.getChildren());
                changed = true;
            } else if(newChildren.isEmpty() && root.getToken().isOperator()) {
                root.clearChildren();
                root.setToken(Value.getValue(mayBeTrue));
                changed = true;
            } else {
                root.clearChildren();
                root.addChildren(newChildren);
            }
            
            
            return changed;
        } else {
            return false;
        }
    }
    
    
    public static boolean extractBinding(final Node root, final Binding bind,
                    final Map<String, Function> functions, final boolean isStatic) {
        final Token token = root.getToken();
        
        if(token.isValue()) {
            final Node parent = root.getParent();
            if(token.getValue().isVariable()
                            && ((parent == null) || (parent.getToken().getOperator() == Token.Operator.AND))) {
                final String var = token.getValue().getVariable();
                if(!bind.hasBinding(var)) {
                    final ValueList l = Pool.getPool().getValueList();
                    l.add(Value.getValue(true));
                    bind.constrain(var, l, true);
                    bind.addDisposableBinding(var, Value.getValue(true));
                    return true;
                } else {
                    return false;
                }
            } else {
                return false;
            }
        }
        
        boolean result = false;
        
        final List<Node> childrenX = root.getChildren();
        for(int i = 0, n = childrenX.size(); i < n; ++i) {
            final Node child = childrenX.get(i);
            result = TreeComputer.extractBinding(child, bind, functions, isStatic) || result;
        }
        
        if((isStatic || root.canBeEvaluatedStatically()) && root.allowAssign()) {
            try {
                final Token.Operator op = token.getOperator();
                if(op == Token.Operator.NOT) {
                    final Node parent = root.getParent();
                    if((parent == null) || TreeComputer.logicalOp(parent.getToken().getOperator())) {
                        final Node child = root.getChildren().get(0);
                        if(child.getToken().isValue() && child.getToken().getValue().isVariable()) {
                            final String var = child.getToken().getValue().getVariable();
                            if(!bind.hasBinding(var)) {
                                final ValueList l = Pool.getPool().getValueList();
                                l.add(Value.getValue(false));
                                bind.constrain(var, l, true);
                                bind.addDisposableBinding(var, Value.getValue(false));
                                result = true;
                            }
                        }
                    }
                    
                } else if((op == Token.Operator.EQUAL) || (op == Token.Operator.NOT_EQUAL)) {
                    final List<Node> children = root.getChildren();
                    final Token left = children.get(0).getToken();
                    final Token right = children.get(1).getToken();
                    if(!left.isValue() && !right.isValue()) {
                        return result;
                    }
                    if(left.isValue() && left.getValue().isVariable()
                                    && (bind.getBinding(left.getValue().getVariable()) == null)) {
                        final String var = left.getValue().getVariable();
                        final Value val = TreeComputer.evaluate(children.get(1), functions, bind);
                        if(val == null) {
                            return result;
                        }
                        if(!bind.hasBinding(var) && (op == Token.Operator.EQUAL)) {
                            final ValueList l = Pool.getPool().getValueList();
                            l.add(val);
                            bind.constrain(var, l, true);
                            bind.addDisposableBinding(var, val);
                            result = true;
                        }
                        if(op == Token.Operator.NOT_EQUAL) {
                            final ValueList values = bind.getConstraint(var);
                            if(values != null) {
                                values.remove(val);
                                if((values.size() == 1) && !bind.hasBinding(var)) {
                                    final Value v = values.get(0);
                                    if(!v.isAny()) {
                                        bind.addDisposableBinding(var, v);
                                        result = true;
                                    }
                                }
                            }
                        }
                        
                    } else if(right.isValue() && right.getValue().isVariable()
                                    && (bind.getBinding(right.getValue().getVariable()) == null)) {
                        final Value val = TreeComputer.evaluate(children.get(0), functions, bind);
                        if(val == null) {
                            return result;
                        }
                        final String var = right.getValue().getVariable();
                        if(!bind.hasBinding(var) && (op == Token.Operator.EQUAL)) {
                            final ValueList l = Pool.getPool().getValueList();
                            l.add(val);
                            bind.constrain(var, l, true);
                            bind.addDisposableBinding(var, val);
                            result = true;
                        }
                        if(op == Token.Operator.NOT_EQUAL) {
                            final ValueList values = bind.getConstraint(var);
                            if(values != null) {
                                values.remove(val);
                                if((values.size() == 1) && !bind.hasBinding(var)) {
                                    final Value v = values.get(0);
                                    if(!v.isAny()) {
                                        bind.addDisposableBinding(var, v);
                                        result = true;
                                    }
                                }
                            }
                        }
                    } 
                } else if((op == Token.Operator.GREATER) || (op == Token.Operator.LESS)
                                || (op == Token.Operator.GREATER_EQUAL)
                                || (op == Token.Operator.LESS_EQUAL)) {
                    final List<Node> children = root.getChildren();
                    
                    final Token left = children.get(0).getToken();
                    final Token right = children.get(1).getToken();
                    if(!left.isValue() && !right.isValue()) {
                        return result;
                    }
                    if(left.isValue() && left.getValue().isVariable()
                                    && (bind.getBinding(left.getValue().getVariable()) == null)) {
                        final String var = left.getValue().getVariable();
                        final ValueList values = bind.getConstraint(var);
                        if(values != null) {
                            final Value val = TreeComputer.evaluate(children.get(1), functions, bind);
                            if(val == null) {
                                return result;
                            }
                            final double value = val.getNumber(); 
                            for(final Iterator<Value> it = values.iterator(); it.hasNext();) {
                                final Value v = it.next();
                                if(!v.isAny()) {
                                    continue;
                                }
                                if(!v.isNumber() || (((op == Token.Operator.GREATER)  && (v.getNumber() <= value))
                                                 ||  ((op == Token.Operator.LESS)  && (v.getNumber() >= value))
                                                 ||  ((op == Token.Operator.GREATER_EQUAL) && (v.getNumber() <  value))
                                                 ||  ((op == Token.Operator.LESS_EQUAL) && (v.getNumber() >  value)))) {
                                    it.remove();
                                }
                            }
                            if((values.size() == 1) && !bind.hasBinding(var)) {
                                final Value v = values.get(0);
                                if(!v.isAny()) {
                                    bind.addDisposableBinding(var, v);
                                    result = true;
                                }
                            }
                        }
                    } else if(right.isValue() && right.getValue().isVariable()
                                    && (bind.getBinding(right.getValue().getVariable()) == null)) {
                        final String var = right.getValue().getVariable();
                        final ValueList values = bind.getConstraint(var);
                        if(values != null) {
                            final Value val = TreeComputer.evaluate(children.get(0), functions, bind);
                            if(val == null) {
                                return result;
                            }
                            final double value = val.getNumber(); 
                            for(final Iterator<Value> it = values.iterator(); it.hasNext();) {
                                final Value v = it.next();
                                if(v.isAny()) {
                                    continue;
                                }
                                if(!v.isNumber() || (((op == Token.Operator.GREATER)  && (value <= v.getNumber()))
                                                 ||  ((op == Token.Operator.LESS)  && (value >= v.getNumber()))
                                                 ||  ((op == Token.Operator.GREATER_EQUAL) && (value <  v.getNumber()))
                                                 ||  ((op == Token.Operator.LESS_EQUAL) && (value >  v.getNumber())))) {
                                    it.remove();
                                }
                            }
                            
                            if((values.size() == 1) && !bind.hasBinding(var)) {
                                final Value v = values.get(0);
                                if(!v.isAny()) {
                                    bind.addDisposableBinding(var, v);
                                    result = true;
                                }
                            }
                        }
                    }
                }
            } catch(final RuntimeException e) {
                // ignore
            }
        }
        
        return result;
    }
    
    
    public static boolean logicalOp(final Token.Operator o) {
        return RPNComputer.LOGICAL_OPERATORS.contains(o);
    }
    
    
    public static boolean extractBindingWithAlteration(final Node root,
                    final Binding bind, final Map<String, Function> functions,
                    final boolean isStatic) {
        final Token token = root.getToken();
        
        if(token.isValue()) {
            final Node parent = root.getParent();
            if(token.getValue().isVariable()
                            && ((parent == null) || (parent.getToken().getOperator() == Token.Operator.AND))) {
                final String var = token.getValue().getVariable();
                if((bind == null) || (bind.getConstraint(var) == null) || Value.containsAny(bind.getConstraint(var))) {
                //if(bind.getConstraint(var).contains(IPlace.STRING_ANY)) {
                    return false;
                }
                final ValueList l = Pool.getPool().getValueList();
                l.add(Value.getValue(true));
                bind.constrain(var, l, true);
                bind.addDisposableBinding(var, Value.getValue(true));
                
                root.setToken(Value.getValue(true));
                return true;
            } else {
                return false;
            }
        } else if(!root.allowAssign()) {
            return false;
        }
        
        boolean result = false;
        
        if(root.hasChildren()) {
            final List<Node> children = root.getChildren();
            for(int i = 0, n = children.size(); i < n; ++i) {
                final Node child = children.get(i);
                result = TreeComputer.extractBindingWithAlteration(child, bind,
                                functions, isStatic) || result;
            }
        }

        if(isStatic || root.canBeEvaluatedStatically()) {
            try {
                final Token.Operator op = token.getOperator();
                if(op == Token.Operator.NOT) {
                    final Node parent = root.getParent();
                    if((parent == null) || TreeComputer.logicalOp(parent.getToken().getOperator())) {
                        final Node child = root.getChildren().get(0);
                        if(child.getToken().isValue() && child.getToken().getValue().isVariable()) {
                            final String var = child.getToken().getValue().getVariable();
                            if((bind.getConstraint(var) != null)
                                            && Value.containsAny(bind.getConstraint(var))) {
                                bind.removeConstraint(var);
                            }
                            final ValueList l = Pool.getPool().getValueList();
                            l.add(Value.getValue(false));
                            bind.constrain(var, l, true);
                            bind.addDisposableBinding(var, Value.getValue(false));
                            result = true;
                            
                            root.clearChildren();
                            root.setToken(Value.getValue(true));
                        }
                    }
                    
                } else if((op == Token.Operator.EQUAL) || (op == Token.Operator.NOT_EQUAL)) {
                    final List<Node> children = root.getChildren();
                    final Token left = children.get(0).getToken();
                    final Token right = children.get(1).getToken();
                    if(!left.isValue() && !right.isValue()) {
                        return result;
                    }
                    if(left.isValue() && left.getValue().isVariable()
                                    && !bind.hasBinding(left.getValue().getVariable())) {
                        final String var = left.getValue().getVariable();
                        final Value val = TreeComputer.evaluate(children.get(1), functions, bind);
                        if(val == null) {
                            return result;
                        }
                        if(bind.hasConstraint(var)
                                        && Value.containsAny(bind.getConstraint(var))) {
                            bind.removeConstraint(var);
                        }
                        if(op == Token.Operator.EQUAL) {
                            final ValueList l = Pool.getPool().getValueList();
                            l.add(val);
                            bind.constrain(var, l, true);
                            bind.addDisposableBinding(var, val);
                            result = true;
                            
                            root.clearChildren();
                            root.setToken(Value.getValue(true));
                        }
                        if(op == Token.Operator.NOT_EQUAL) {
                            final ValueList values = bind.getConstraint(var);
                            if(values != null) {
                                values.remove(val);
                                if((values.size() == 1) && !Value.containsAny(values)) {
                                    bind.addDisposableBinding(var, values.get(0));
                                }
                                
                                result = true;
                                
                                root.clearChildren();
                                root.setToken(Value.getValue(true));
                            }
                        }
                        
                    } else if(right.isValue() && right.getValue().isVariable()
                                    && (bind.getBinding(right.getValue().getVariable()) == null)) {
                        final Value val = TreeComputer.evaluate(children.get(0), functions, bind);
                        if(val == null) {
                            return result;
                        }
                        final String var = right.getValue().getVariable();
                        if((bind.getConstraint(var) != null)
                                        && Value.containsAny(bind.getConstraint(var))) {
                            bind.removeConstraint(var);
                        }

                        if((bind.getBinding(var) == null) && (op == Token.Operator.EQUAL)) {
                            final ValueList l = Pool.getPool().getValueList();
                            l.add(val);
                            bind.constrain(var, l, true);
                            bind.addDisposableBinding(var, val);
                            result = true;
                            
                            root.clearChildren();
                            root.setToken(Value.getValue(true));
                        }
                        if(op == Token.Operator.NOT_EQUAL) {
                            final ValueList values = bind.getConstraint(var);
                            if(values != null) {
                                values.remove(val);
                                if((values.size()) == 1 && !Value.containsAny(values)) {
                                    bind.addDisposableBinding(var, values.get(0));
                                }
                                result = true;
                                
                                root.clearChildren();
                                root.setToken(Value.getValue(true));
                            }
                        }
                    } 
                } else if((op == Token.Operator.GREATER) || (op == Token.Operator.LESS)
                                || (op == Token.Operator.GREATER_EQUAL)
                                || (op == Token.Operator.LESS_EQUAL)) {
                    final List<Node> children = root.getChildren();
                    
                    final Token left = children.get(0).getToken();
                    final Token right = children.get(1).getToken();
                    if(!left.isValue() && !right.isValue()) {
                        return result;
                    }
                    if(left.isValue() && left.getValue().isVariable()
                                    && (bind.getBinding(left.getValue().getVariable()) == null)) {
                        final String var = left.getValue().getVariable();
                        if(!Value.containsAny(bind.getConstraint(var))) {
                            final ValueList values = bind.getConstraint(var);
                            if(values != null) {
                                final Value val = TreeComputer.evaluate(children.get(1), functions, bind);
                                if(val == null) {
                                    return result;
                                }
                                final double value = val.getNumber();
                                for(final Iterator<Value> it = values.iterator(); it.hasNext();) {
                                    final Value v = it.next();
                                    if(!v.isNumber() || (((op == Token.Operator.GREATER)  && (v.getNumber() <= value))
                                                     ||  ((op == Token.Operator.LESS)  && (v.getNumber() >= value))
                                                     ||  ((op == Token.Operator.GREATER_EQUAL) && (v.getNumber() <  value))
                                                     ||  ((op == Token.Operator.LESS_EQUAL) && (v.getNumber() >  value)))) {
                                        it.remove();
                                    }
                                }
                                
                                if(values.size() == 1) {
                                    bind.addDisposableBinding(var, values.get(0));
                                }
                                result = true;
                                
                                root.clearChildren();
                                root.setToken(Value.getValue(true));
                            }
                        }
                    } else if(right.isValue() && right.getValue().isVariable()
                                    && (bind.getBinding(right.getValue().getVariable()) == null)) {
                        final String var = right.getValue().getVariable();
                        if(!Value.containsAny(bind.getConstraint(var))) {
                            final ValueList values = bind.getConstraint(var);
                            if(values != null) {
                                final Value val = TreeComputer.evaluate(children.get(0), functions, bind);
                                if(val == null) {
                                    return result;
                                }
                                final double value = val.getNumber(); 
                                for(final Iterator<Value> it = values.iterator(); it.hasNext();) {
                                    final Value v = it.next();
                                    if(!v.isNumber() || (((op == Token.Operator.GREATER)  && (value <= v.getNumber()))
                                                     ||  ((op == Token.Operator.LESS)  && (value >= v.getNumber()))
                                                     ||  ((op == Token.Operator.GREATER_EQUAL) && (value <  v.getNumber()))
                                                     ||  ((op == Token.Operator.LESS_EQUAL) && (value >  v.getNumber())))) {
                                        it.remove();
                                    }
                                }
                                
                                if(values.size() == 1) {
                                    bind.addDisposableBinding(var, values.get(0));
                                }
                                result = true;
                                
                                root.clearChildren();
                                root.setToken(Value.getValue(true));
                            }
                        }
                    }
                } else {
                    try {
                        final Value val = TreeComputer.evaluate(root, functions, null);
                        if(val != null) {
                            root.setToken(val);
                            root.clearChildren();
                            result = true;
                        }
                    } catch(Exception e) {
                        // ignore
                    }
                }
            } catch(final RuntimeException e) {
                // ignore
                //e.printStackTrace();
            }
        }
        
        return result;
    }
    
    
    public static Value evaluate(final Node root,
                    final Map<String, Function> functions, final Binding bind) {
        return TreeComputer.evaluate(root, functions, bind, null);
    }
    
    
    public static Value evaluate(final Node root,
                    final Map<String, Function> functions, final Binding bind,
                    final Map<Value, Value> properties) {
        final Token token = root.getToken();
        if(token.isValue()) {
            if(root.allowAssign() && ((root.getParent() == null) || root.getParent().getToken().getOperator() == Token.Operator.AND)
                            && token.getValue().isVariable()
                            && (bind.getBinding(token.getValue().getVariable()) == null)) {
                final String var = token.getValue().getVariable();
                final Value val = Value.getValue(true);
                bind.addDisposableBinding(var, val);
                return val;
            }
            
            final Value val = token.getValue();
            if(val.isVariable()) {
                if(bind == null) {
                    return null;
                }
                return bind.getBinding(val.getVariable());
            } else {
                return val;
            }
        }
        
        
        final Token.Operator op = token.getOperator();
        Value value = null;
        if(op == Token.Operator.NOT) {
            final Node child = root.getChildren().get(0);
            final Token val = child.getToken();
            if(root.allowAssign() && ((root.getParent() == null) || root.getParent().getToken().getOperator() == Token.Operator.AND)
                            && val.isValue() && val.getValue().isVariable()
                            && (bind.getBinding(val.getValue().getVariable()) == null)) {
                final String var = val.getValue().getVariable();
                bind.addDisposableBinding(var, Value.getValue(false));
                value = Value.getValue(true);
            } else {
                final Value v = TreeComputer.evaluate(root.getChildren().get(0),
                                functions, bind, properties);
                if(v == null) {
                    return null;
                }
                value = Value.getValue(!v.getBoolean());
            }
        } else if(op == Token.Operator.LIST_START) {
            final Tuple result = new Tuple();
            final List<Node> children = root.getChildren();
            for(int i = 0; i < children.size(); i++) {
                final Value v = TreeComputer.evaluate(children.get(i),
                                functions, bind, properties);
                if(v == null) {
                    return null;
                }
                result.addEntry(v);
            }
            value = Value.getValue(result);
        } else if(op == Token.Operator.STRING_CONCATENATION) {
            String result = "";
            final List<Node> children = root.getChildren();
            for(int i = 0; i < children.size(); i++) {
                final Value v = TreeComputer.evaluate(children.get(i),
                                functions, bind, properties);
                if(v == null) {
                    return null;
                }
                result += (v.isString() ? v.getString() : v.getNetValue());
            }
            value = Value.getStringValue(result);
        } else if(op == Token.Operator.UNARY_MINUS) {
            final Value v = TreeComputer.evaluate(root.getChildren().get(0),
                            functions, bind, properties);
            if(v == null) {
                return null;
            }
            value = TreeComputer.performOp(Token.Operator.MINUS,
                            Value.getValue(0.0d), v);
        } else if(op == Token.Operator.PLUS) {
            boolean stringResult = false;
            final ValueList operands = Pool.getPool().getValueList();
            final List<Node> children = root.getChildren();
            for(int i = 0; i < children.size(); i++) {
                final Value v = TreeComputer.evaluate(children.get(i),
                                functions, bind, properties);
                if(v == null) {
                    return null;
                }
                operands.add(v);
                if(v.isString()) {
                    stringResult = true;
                }
            }
            if(!stringResult) {
                value = operands.get(0);
                boolean isTuple = false;
                double result = isTuple ? 0.0d : value.getNumber();
                
                for(int i = 1; i < operands.size(); i++) {
                    final Value v = operands.get(i);
                    if(!isTuple && v.isTuple()) {
                        isTuple = true;
                        value = Value.getValue(result);
                    }
                    
                    if(isTuple) {
                        value = TreeComputer.performOp(op, value, v);
                    } else {
                        result += v.getNumber();
                    }
                }
                
                if(!isTuple) {
                    value = Value.getValue(result); 
                }
            } else {
                final StringBuffer result = new StringBuffer();
                for(int i = 0, n = operands.size(); i < n; ++i) {
                    final Value v = operands.get(i); 
                    result.append(v.isString() ? v.getString() : v.getNetValue());
                }
                value = Value.getStringValue(result.toString());
            }
            
        } else if(op == Token.Operator.MINUS || op == Token.Operator.TIMES
                        || op == Token.Operator.DIVIDE || op == Token.Operator.MODULO
                        || op == Token.Operator.POWER) {
            final List<Node> children = root.getChildren();
            value = TreeComputer.evaluate(children.get(0), functions, bind, properties);
            if(value == null) {
                return null;
            }
            boolean isTuple = value.isTuple();
            double result = isTuple ? 0.0d : value.getNumber();
            
            boolean first = true;
            for(int i = 0, n = children.size(); i < n; ++i) {
                final Node child = children.get(i);
                if(first) {
                    first = false;
                    continue;
                }
                final Value v = TreeComputer.evaluate(child, functions, bind, properties);
                if(v == null) {
                    return null;
                }
                if(!isTuple && v.isTuple()) {
                    isTuple = true;
                    value = Value.getValue(result);
                }
                
                if(isTuple) {
                    value = TreeComputer.performOp(op, value, v);
                } else {
//                    if(op.equals(RPNComputer.PLUS)) {
//                        result += v.getNumber();
//                    } else 
                    if(op == Token.Operator.MINUS) {
                        result -= v.getNumber();
                    } else if(op == Token.Operator.TIMES) {
                        result *= v.getNumber();
                    } else if(op == Token.Operator.DIVIDE) {
                        result /= v.getNumber();
                    } else if(op == Token.Operator.MODULO) {
                        result %= v.getNumber();
                    } else if(op == Token.Operator.POWER) {
                        result = Math.pow(result, v.getNumber());
                    } else {
                        return null;
                    }
                }
            }
            
            if(!isTuple) {
                value = Value.getValue(result); 
            }
            
        } else if(op == Token.Operator.LESS_EQUAL || op == Token.Operator.LESS
                        || op == Token.Operator.GREATER_EQUAL
                        || op == Token.Operator.GREATER
                        || op == Token.Operator.NOT_EQUAL) {
            Value last = null;
            final List<Node> children = root.getChildren();
            boolean result = true;
            for(int i = 0, n = children.size(); i < n; ++i) {
                final Node child = children.get(i);
                final Value v = TreeComputer.evaluate(child, functions, bind, properties);
                if(v == null) {
                    return null;
                }
                if(last != null) {
                    if(((op == Token.Operator.LESS_EQUAL && (last.getNumber() > v.getNumber())))
                                    || (((op == Token.Operator.LESS) && (last.getNumber() >= v.getNumber())))
                                    || (((op == Token.Operator.GREATER_EQUAL) && (last.getNumber() < v.getNumber())))
                                    || (((op == Token.Operator.GREATER) && (last.getNumber() <= v.getNumber())))
                                    || (((op == Token.Operator.NOT_EQUAL) && last.equals(v)))) {
                        result = false;
                        break;
                    }
                }
                last = v;
            }
            value = Value.getValue(result);
        
        } else if(op == Token.Operator.EQUAL) { // special
            final List<Node> children = root.getChildren();
            
            final Token left = children.get(0).getToken();
            final Token right = children.get(1).getToken();
            final boolean allowAssign = root.allowAssign();
            if(allowAssign && left.isValue() && left.getValue().isVariable()
                            && (bind.getBinding(left.getValue().getVariable()) == null)) {
                final String var = left.getValue().getVariable();
                final Value val = TreeComputer.evaluate(children.get(1),
                                functions, bind, properties);
                if(val == null) {
                    return null;
                }
                if(bind.getBinding(var) == null) {
                    bind.addDisposableBinding(var, val);
                    //System.out.println(var + " => " + val);
                    value = Value.getValue(true);
                }
            } else if(allowAssign && right.isValue() && right.getValue().isVariable()
                            && (bind.getBinding(right.getValue().getVariable()) == null)) {
                final Value val = TreeComputer.evaluate(children.get(0),
                                functions, bind, properties);
                if(val == null) {
                    return null;
                }
                final String var = right.getValue().getVariable();
                if(bind.getBinding(var) == null) {
                    bind.addDisposableBinding(var, val);
                    //System.out.println(var + " => " + val);
                    value = Value.getValue(true);
                }
            } 
            
            if(value == null) {
                boolean result = true;
                Value last = null;
                for(int i = 0, n = children.size(); i < n; ++i) {
                    final Node child = children.get(i);
                    final Value v = TreeComputer.evaluate(child, functions,
                                    bind, properties);
                    if(v == null) {
                        return null;
                    }
                    if((last != null) && (!last.equals(v))) {
                        result = false;
                        break;
                    }
                    last = v;
                }
                value = Value.getValue(result);
            }
        
        } else if((op == Token.Operator.AND) || (op == Token.Operator.OR)) {
            final List<Node> children = root.getChildren();
            boolean result = op == Token.Operator.AND;
            for(int i = 0, n = children.size(); i < n; ++i) {
                final Node child = children.get(i);
                final Value v = TreeComputer.evaluate(child, functions, bind,
                                properties);
                if(v == null) {
                    return null;
                }
                if(((op == Token.Operator.AND) && !v.getBoolean()) || 
                                ((op == Token.Operator.OR) && v.getBoolean())) {
                    result = v.getBoolean();
                    break;
                }
            }
            value = Value.getValue(result);
            
        } else if(op == Token.Operator.XOR) {
            final List<Node> children = root.getChildren();
            boolean result = false;
            for(int i = 0; i < children.size(); i++) {
                final Node child = children.get(i);
                final Value v = TreeComputer.evaluate(child, functions, bind, properties);
                if(v == null) {
                    return null;
                }
                result ^= v.getBoolean();
            }
            value = Value.getValue(result);
            
        } else {
            value = Functions.evaluate(root.getToken().getFunction(),
                            root.getChildren(), functions, bind, properties);
        }
        
        return value;
    }
    
    
    public static Value performOp(final Token.Operator op, final Value a, final Value b) {
        if(!(a.isNumber() || a.isTuple()) || !(b.isNumber() || b.isTuple())) {
            return null;
        } else if(a.isNumber() && b.isNumber()) {
            if(op == Token.Operator.PLUS) {
                return Value.getValue(a.getNumber() + b.getNumber());
            } else if(op == Token.Operator.MINUS) {
                return Value.getValue(a.getNumber() - b.getNumber());
            } else if(op == Token.Operator.TIMES) {
                return Value.getValue(a.getNumber() * b.getNumber());
            } else if(op == Token.Operator.DIVIDE) {
                return Value.getValue(a.getNumber() / b.getNumber());
            } else if(op == Token.Operator.MODULO) {
                return Value.getValue(a.getNumber() % b.getNumber());
            } else if(op == Token.Operator.POWER) {
                return Value.getValue(Math.pow(a.getNumber(), b.getNumber()));
            } else if(op == Token.Operator.AND) {
                return Value.getValue(((long)a.getNumber()) & ((long)b.getNumber()));
            } else if(op == Token.Operator.OR) {
                return Value.getValue(((long)a.getNumber()) | ((long)b.getNumber()));
            } else if(op == Token.Operator.XOR) {
                return Value.getValue(((long)a.getNumber()) ^ ((long)b.getNumber()));
            } else {
                return null;
            }
        } else if(a.isNumber() || b.isNumber()) {
            final Value num = a.isNumber() ? a : b;
            final Value tup = a.isNumber() ? b : a;
            final Tuple result = new Tuple();
            for(final Iterator<Value> it = tup.getTuple().getEntries().iterator(); it.hasNext();) {
                final Value v = it.next();
                final Value val = TreeComputer.performOp(op, a.isNumber() ? num : v, a.isNumber() ? v : num);
                if(val == null) {
                    return null;
                }
                result.addEntry(val);
            }
            return Value.getValue(result);
        } else {
            final Tuple tupA = a.getTuple();
            final Tuple tupB = b.getTuple();
            if(tupA.size() != tupB.size()) {
                return null;
            }
            final Tuple result = new Tuple();
            for(int i = 0; i < tupA.size(); i++) {
                final Value valA = tupA.getValue(i);
                final Value valB = tupB.getValue(i);
                final Value val = TreeComputer.performOp(op, valA, valB);
                if(val == null) {
                    return null;
                }
                result.addEntry(val);
            }
            return Value.getValue(result);
        }
    }
    
    
    public static void main(final String[] args) {
        final StringBuffer buf = new StringBuffer();
        for(int i = 0; i < args.length; i++) {
            buf.append(args[i]);
            if(i < args.length - 1) {
                buf.append(" ");
            }
        }
        String guard = "(dot=A) &  !X &(-0>  -4)&(Y=\"ff\")&!!!  !(\"hallo\"#\"welt\")&(7-9<16%8)&!!(2**3 <= 8) & (Z=10+73/73-1)";
        guard = "(X<4&X>X)&X<Y|(X>=X&X=4)";
        guard = "Y=max(50,X)&floor(rand()*100+1)=X&Z=-abs(min(20,10))";
        guard = "(tt=(X=Y))";
        guard = "r'{0..1} & w'{-1.. 2,dot}:dot=dot&dot=dot&0='w&0=w'&0='r&1=r'&r'='r+1&w'='w";
        guard = "5=2*(-3+logn((e()*e())**(x+x+x+1)))";
//        guard = "X = (0-(2*4**2, 23-3)) * 2 & Y=proj(X,1)&Z=3*(3)&(3,3)=(3,3*3/3)";
        guard = "((2,4)=((1+X),4))";
//        guard = "X=map(\"floor\", map(\"cos\", (0, 60, 90, pi()))) & FUNC=\"sqr\" & Y=map(FUNC, 2*X)";
//        guard = "map(\"sin\", 4)";
//        guard = "f(g((2,3)))";
        guard = "X=(1,2,3,4) & forall(i, 0, dim(X)-2, 1, proj(X, i) <= proj(X, i+1))";
//        guard = "X{-1..10} : (X=5|X=4) & exists(X, X=5)";
//        guard = "X{0..5,tt,ff}&Y{8..10}:A=5&((X*4=16&X+A<A+4*A&Y=40/X)|X)";
//        guard = "X{\"min\", \"max\", \"sum\", \"prod\"}:create(X, evaluate(X, asTuple(A)))&A=create(X, 1, 10, 1, X)";
//        guard = "X{0..10}:create(X, fak(X))";
//        guard = "X=5&fak(X)";
//        guard = "exists1(X, (1, 2, 5, 6), X=5) & X=6";
//        guard = "X{0..8}:exists1(X,  X=5)";
//        guard = "+(3,4,/(10,5),*(10,2))";
//        guard = "((5,4),(3,1))[1][0]";
//        guard = "level{0..4}&_tmp@3{0..4}&inside{tt,ff}&_fvar_autoElevator_f__hasPassengers{tt,ff}:(level=_tmp@3)&(inside=_fvar_autoElevator_f__hasPassengers)";
//        guard = "!!X";
        guard = "2.5";
        if(buf.length() > 0) {
            guard = buf.toString();
        }
        final Map<String, Function> functions = Pool.getPool().getMap();
        functions.put("fak/1", new Function("f ( X ) = ifthen(X<=1, 1, prod(create(i,1,X,1,i)))"));
        functions.put("fak/2", new Function("f ( X, Y ) = ifthen(X<=1, 1, X*fak(X-1, Y))"));
//        functions.put("fak/2", new Function("f ( X, Y ) = ifthen(X<=1, 1, X*fak(X-1))"));
        functions.put("fa/1", new Function("f ( X ) = map           ( \"sqr\" , X )   "));
        functions.put("f/1", new Function("f ( X ) = map           ( \"sqr\" , X )   "));
        functions.put("g/1", new Function("g(X)=ifthen(pi()>3, proj(h(X), dim(h(X))-1), 0)"));
        functions.put("h/1", new Function("h(x)=append(x,     8)"));
        functions.put("sim/1", new Function("sim(X)=sin(0)+X*X"));
        Binding bind = Pool.getPool().getBinding();
//        bind.addBinding("X", Value.getValue("(1, 2, 3, 4)"));
        
//        for(Iterator it = functions.values().iterator(); it.hasNext(); ) {
//            System.out.println(it.next());
//        }
        
        System.out.println("EXPRES:   " + guard);
//        final ExpressionLexer lexer = new ExpressionLexer(new java.io.StringReader(guard));
////        final ExpressionLexer lexer = new ExpressionLexer(System.in);
//        final ExpressionParser parser = new ExpressionParser(lexer);
//        try {
//            Node n = parser.logic_with_constraints();
////            Node n = parser.logic();
//            System.out.println(n.toString());
//        } catch(Exception e) {
//            System.err.println(e.getClass());
//            e.printStackTrace();
//        }
//        final Scanner scanner = new Scanner(new java.io.StringReader(guard));
//        final Parser parser = new Parser(scanner);
//        try {
//            final Node n = (Node)parser.parse().value;
//            System.out.println(n);
//        } catch(final Exception e) {
//            System.err.println(e.getClass());
//            e.printStackTrace();
//        }
        System.out.println("==============");
        final List<Token> tokens = RPNComputer.tokenize(guard);
        System.out.println("TOKENS:   " + tokens);
        final RPN rpn = RPNComputer.infixToRPN(tokens);
        System.out.println("RPN   :   " + rpn);
        final Node node = TreeComputer.rpnToTree(rpn);
        System.out.println("TREE  :   " + node);
        System.out.println("CONSTR:   " + node.getConstraints());
        TreeComputer.optimize(node, functions);
        System.out.println("SIMPLE:   " + node);
        TreeComputer.optimizeHard(node, functions);
        System.out.println("OPTIMU:   " + node);
        TreeComputer.extractConstraints(node, functions);
        System.out.println("CONSTR:   " + node.getConstraints());

        final long start = System.currentTimeMillis();
        bind.addConstraints(node.getConstraints());
        final Value result = TreeComputer.evaluate(node, functions, bind);
        System.out.println("RESULT:   " + result + "\nBINDING:  " + bind);
        System.out.println("DURATION: " + (System.currentTimeMillis() - start) + " ms");
    }
}
