/*
 * File:    Backward.java
 * Project: petruchio
 * Author:  Tim Strazny
 * Created: 16.03.2009
 */
package petruchio.cov;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.lang.ref.SoftReference;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import net.jcip.annotations.GuardedBy;

import petruchio.common.IdentityHashSet;
import petruchio.interfaces.SelfDescribing;
import petruchio.interfaces.petrinet.ExtendedTransition;
import petruchio.interfaces.petrinet.InhibitorArc;
import petruchio.interfaces.petrinet.IntMatrix;
import petruchio.interfaces.petrinet.PTArc;
import petruchio.interfaces.petrinet.PetriNet;
import petruchio.interfaces.petrinet.Place;
import petruchio.interfaces.petrinet.PlacePair;
import petruchio.interfaces.petrinet.ResetArc;
import petruchio.interfaces.petrinet.TPArc;
import petruchio.interfaces.petrinet.TransferArc;
import petruchio.interfaces.petrinet.Transition;
import petruchio.pn.Converter;
import petruchio.pn.InvariantAnalysis;
import petruchio.pn.IntPNMatrix;

/**
 * @author Tim Strazny
 */
public class Backward extends AbstractMultiThreadedCoverability implements SelfDescribing {

    /**
     * When generating predecessors, the constructed mCreated are checked for certain properties
     * -- these are the results.
     */
    private static enum Result {
        
        /**
         * The predecessor marking is coverable.
         * This decided by {@link Backward#isCoveredByGoodState(Node)}.
         * The search is done.
         */
        COVERABLE,
        
        /**
         * The predecessor marking is found not to be coverable {@link Backward#notCoverable(Node)}.
         * This results in throwing away this predecessor marking.
         */
        NOT_COVERABLE,
        
        /**
         * The predecessor marking is neither coverable nor a cut-off occurs, nor is it known not
         * to be coverable. It has to be inspected.
         */
        UNKNOWN
    }
    
    static final Node ACCELERATED_COVERED = Node.newNode();
    
    static final Random RANDOM = new Random();
    
    private static final int MIN_TRANSITION_DISTANCE = -1;

    /**
     * Place order to ease finding usable invariants
     * TODO find good ordering
     */
    static final Comparator<Place> PLACE_COMPARATOR = new Comparator<Place>() {
        @Override
        public int compare(final Place a, final Place b) {
            return -(a.getInput().size() + a.getOutput().size() - b.getInput().size() - b.getOutput().size());
        }
    };
    
    /**
     * Transition order to ease finding usable invariants
     * TODO find good ordering
     */
    static final Comparator<Transition> TRANSITION_COMPARATOR = new Comparator<Transition>() {
        @Override
        public int compare(final Transition a, final Transition b) {
            return -(a.getInput().size() + a.getOutput().size() - b.getInput().size() - b.getOutput().size());
        }
    };
    
    /**
     * Guide search by processing nodes first which contain less tokens (always better)
     */
    static final Comparator<Node> TOKEN_COMPARATOR = new Comparator<Node>() {
        @Override
        public int compare(final Node a, final Node b) {
            return a.getTokens() - b.getTokens();
        }
    };
    
    
//    /**
//     * Guide search by processing nodes first which have a lower max token count on one place
//     */
//    static final Comparator<Node> MAX_TOKEN_COMPARATOR = new Comparator<Node>() {
//        @Override
//        public int compare(final Node a, final Node b) {
//            return a.getMaxTokens() - b.getMaxTokens();
//        }
//    };
    
    /**
     * Guide search by processing nodes first that cover less places with known bound.
     */
    static final Comparator<Node> BOUND_COMPARATOR = new Comparator<Node>() {
        @Override
        public int compare(final Node a, final Node b) {
            return a.getBoundedPlaces() - b.getBoundedPlaces();
        }
    };
    
    /**
     * Guide search by processing nodes first that are somewhat further away from the bounds.
     */
    static final Comparator<Node> BOUND_DIST_COMPARATOR = new Comparator<Node>() {
        @Override
        public int compare(final Node a, final Node b) {
            return b.getDistanceToBounds() - a.getDistanceToBounds();
        }
    };
    
    /**
     * Guide search by processing nodes first which are "nearer" to the initial state. 
     */
    static final Comparator<Node> DISTANCE_COMPARATOR = new Comparator<Node>() {
        @Override
        public int compare(final Node a, final Node b) {
            return a.getTransitionDistance() - b.getTransitionDistance();
        }
    };
    
    /**
     * Guide search by processing nodes first which are "nearer" to the initial state. 
     */
    static final Comparator<Node> ACCELERATED_COMPARATOR = new Comparator<Node>() {
        @Override
        public int compare(final Node a, final Node b) {
            if(a.accelerated) {
                return b.accelerated ? 0 : -1;
//                if(b.traceGuessSatisfied) {
//                    return (b.traceGuessHead == null ? 0 : b.traceGuessHead.getLength()) - (a.traceGuessHead == null ? 0 : a.traceGuessHead.getLength());
//                }
//                return -1;
            }
            return b.accelerated ? 1 : 0;
//            return a.traceGuessSatisfied - b.traceGuessSatisfied;
        }
    };
    
    /**
     * Guide search by processing nodes first which might have less pre transitions 
     */
    static final Comparator<Node> PRECOUNT_COMPARATOR = new Comparator<Node>() {
        @Override
        public int compare(final Node a, final Node b) {
            return a.preTrans - b.preTrans;
//            return a.preTrans.size() - b.preTrans.size();
        }
    };
    
    /**
     * Guide search by processing nodes first which are "nearer" to the target state. 
     */
    static final Comparator<Node> PATH_LENGTH_COMPARATOR = new Comparator<Node>() {
        @Override
        public int compare(final Node a, final Node b) {
            return (a.path == null ? 0 : a.path.getLength()) - (b.path == null ? 0 : b.path.getLength());
        }
    };
    
    /**
     * Guide search by processing nodes first which mark less places. 
     */
    static final Comparator<Node> PLACE_COUNT_COMPARATOR = new Comparator<Node>() {
        @Override
        public int compare(final Node a, final Node b) {
            return a.getMarking().size() - b.getMarking().size();
        }
    };
    
    /**
     * The order of comparators used for directed search in {@link Backward#searchGuide}. 
     */
    @SuppressWarnings("unchecked")
    static final Comparator<Node>[] SEARCH_GUIDE = new Comparator[] {
        ACCELERATED_COMPARATOR,
        TOKEN_COMPARATOR,
        BOUND_COMPARATOR,
        BOUND_DIST_COMPARATOR,
        DISTANCE_COMPARATOR,
        PATH_LENGTH_COMPARATOR,
        PLACE_COUNT_COMPARATOR,
        PRECOUNT_COMPARATOR,
//        MAX_TOKEN_COMPARATOR,
    };
    
    /**
     * The actual guide for directed search in {@link Backward#addPredecessor(List, Node, Node)}.
     * It uses the comparators {@link Backward#SEARCH_GUIDE} in order.
     */
    final Comparator<Node> searchGuide = new Comparator<Node>() {
        @Override
        public int compare(final Node a, final Node b) {
            // walk through array of comparators
            // return first "different" comparison result
            // if all comparisons return "equal" then return "equal"
            for(int i = 0; i < Backward.SEARCH_GUIDE.length; ++i) {
                final int r = Backward.SEARCH_GUIDE[i].compare(a, b);
                if(r != 0) {
                    return -r;
                }
            }
            return 0;
        }
    };
    
    
    /**
     * Constant for "transition can fire unboundedly often".
     */
    private static final int UNBOUNDED_FIRING = -1;
    /**
     * Constant for "it is unknown whether transition can fire unboundedly often".
     */
    private static final int UNKNOWN_FIRING_BOUND = -2;
    
    private static final boolean DEBUG = false;
    private static final boolean FINE_DEBUG = false;
    
    static int NODE_CLASS_TREE_MAX_HEIGHT = 0;
    static boolean GET_MAX_TREE_HEIGHT = false;
    int maxTreeHeight = 0;
    
    /**
     * Switch: Should verbose messages be printed to the console?
     */
    boolean verbose = false;
    
    /**
     * If {@link Backward#verbose} is <code>true</code> be verbose every X milliseconds.
     */
    private static final long VERBOSE_OUTPUT_DELAY = 10000L;
    
    /**
     * Set to <code>false</code> to turn off computations needed by {@link Backward#DISTANCE_COMPARATOR}.
     */
    private static final boolean USE_DISTANCE_HEURISTICS = true;
    
    /**
     * When this is <code>true</code> and the query takes longer than a certain amount
     * of milliseconds {@link #DYNAMIC_RECHECK_THRESHOLD} (and a re-check will be started!) then
     * postpone this query.
     */
    private static final boolean DYNAMIC_RECHECK = true;
    
    /**
     * When {@link #DYNAMIC_RECHECK} is <code>true</code> and the query takes longer than this amount
     * of milliseconds (and a re-check will be started!) then postpone this query.
     */
    private static final int DYNAMIC_RECHECK_THRESHOLD = 100;
    
    /**
     * If a search takes longer than this amount of time (milliseconds), and a certain
     * number {@link Backward#PLACE_DIFFERENCE_BEFORE_INV} of new places have been
     * added since the last invariant calculation, invariants will be computed to generate
     * some new bounds.
     */
    private static final long TIME_BEFORE_INV = 200;
    
    /**
     * If a search takes longer than a certain amount of time {@link Backward#TIME_BEFORE_INV},
     * and a this number of new places has been added since the last invariant calculation,
     * invariants will be computed to generate some new bounds.
     */
    private static final int PLACE_DIFFERENCE_BEFORE_INV = 1;
    
    /**
     * The time a search takes isn't checked every iteration of the main loop, but every multiple
     * of this number of iterations.
     */
    private static final int CHECK_TIME_EVERY = 200;
    
    /**
     * The initial marking.
     */
    final Node initialState = Node.newNode();
    
    private boolean containsParameters = false;
    
    /**
     * Number of transitions in the current net.
     */
    int transitionCount = 0;
    
    /**
     * Whether the current net is a (multi) transfer net (with instances of ExtendedTransition).
     * Only then the <code>instanceof</code> operator will be used to find the respective transitions.
     */
    boolean containsExtendedTransition = false;
    
    /**
     * The maximum number of output arcs over all transitions in the current net.
     */
    int maxTransOutBranching = 0;
    
    /**
     * The current number of places for which bounds are known.
     */
    volatile int boundedPlacesCtr = 0;
    
    /**
     * Counter for number mCreated created (for benchmarking).
     */
    long markingsCreated = 0;
    
    /**
     * Counter for number mCreated examined (for benchmarking).
     */
    long markingsExamined = 0;
    
    /**
     * Switch: Should search space be pruned using place bounds?
     */
    volatile boolean useBounds = true;
    
    /**
     * Switch: Should partial order methods be used?
     */
    volatile boolean usePO = true;
    
    volatile boolean useOrdering = true;
    
    volatile boolean useAcceleration = true;
    
    /**
     * Depth for bounded model checking: Set to <code>0</code> to turn off
     * or a larger value to enable. Caution: if it is enabled, certain cut-offs
     * from the regular backward search are not applicable and will been disabled.
     * This might lead to a heavily increased running time.
     */
    volatile int bmcDepth = 0;
    
    /**
     * Stores the number of places of the net when the last call to
     * {@link Backward#computeInvariants()} has been performed.
     * When working for the Petruchio compiler and a search takes too long
     * (see {@link Backward#TIME_BEFORE_INV}), invariants are generated via
     * {@link Backward#computeInvariants()} -- but only if the net has grown by
     * a certain number of places (see {@link Backward#PLACE_DIFFERENCE_BEFORE_INV})
     * since the last computation.
     */
    int invComputedAt = 0;
    
    
    /**
     * The (ordered) set of inspected (minimal) mCreated.
     */
    @GuardedBy("itself")
    final List<NodeClass> inspectedMarkings = new LinkedList<NodeClass>();
    
    /**
     * The (ordered) set of mCreated that still have to be processed in the search.
     */
    @GuardedBy("itself")
    final List<Node> markingsToProcess = new ArrayList<Node>();
    
    private SimpleList<petruchio.pn.Transition> friendlyTransitions = null;
    
    private SimpleList<petruchio.pn.Place> friendlyPlaces = null;
    
    private List<NodeClass> notCov = new LinkedList<NodeClass>();
    
    @GuardedBy("itself")
    final Map<Place, Pair<Collection<Place>, Integer>> clusterBounds = new IdentityHashMap<Place, Pair<Collection<Place>, Integer>>();
    @GuardedBy("itself")
    final Map<Place, Collection<Pair<Map<Place, Integer>, Integer>>> invariants = new IdentityHashMap<Place, Collection<Pair<Map<Place, Integer>, Integer>>>();
    
    private Collection<PlacePair> checked = new HashSet<PlacePair>();
    
    /**
     * Lock to make access to {@link Backward#transitionData} thread safe.
     */
    private final Lock transitionDataLock = new ReentrantLock();
    
    /**
     * Condition representing that {@link Backward#transitionData} is being read.
     */
    private final Condition transitionDataIsBeingRead = this.transitionDataLock.newCondition();
    
    /**
     * Number of readers currently accessing {@link Backward#transitionData}.
     */
    private int transitionDataReaders = 0;
    
    /**
     * Stores an {@link Backward#TransitionData} object for each transition in the net. 
     */
    @GuardedBy("transitionDataLock")
    private final Map<Transition, TransitionData> transitionData = new IdentityHashMap<Transition, TransitionData>();
    
    /**
     * Lock to ensure thread safe access to {@link Backward#placeData}.
     */
    private final Lock placeDataLock = new ReentrantLock();
    
    /**
     * Condition representing that {@link Backward#placeData} is being read.
     */
    private final Condition placeDataIsBeingRead = this.placeDataLock.newCondition();
    
    /**
     * Number of readers accessing {@link Backward#placeData}.
     */
    private int placeDataReaders = 0;
    
    /**
     * Stores an {@link Backward#PlaceData} object for each place in the net.
     */
    @GuardedBy("placeDataLock")
    private final Map<Place, PlaceData> placeData = new IdentityHashMap<Place, PlaceData>();
    
    /**
     * Collection of transitions which have to be processed when updating place bounds and the like.
     */
    private final Deque<Transition> updating = new LinkedList<Transition>();
    
    /**
     * Switch: Is this supposed to run interleaved with the Petruchio compiler. 
     */
    private boolean usedByPetruchioCompiler = true;
    
    /**
     * Returns an instance of the backward search.
     * Not to be used by Petruchio compiler.
     * 
     * @return A new instance of the backward search where "interleaving
     * with Petruchio compiler" has been turned off.
     */
    public static Backward newInstance() {
        final Backward b = new Backward();
        b.usedByPetruchioCompiler = false;
        return b;
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.cov.AbstractMultiThreadedCoverability#setThreadNumber(int)
     */
    @Override
    public void setThreadNumber(final int newThreads) {
        // only one thread
        // the bottle neck has not yet been parallelized/fully identified :)
        super.setThreadNumber(1); // TODO
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.cov.AbstractMultiThreadedCoverability#terminationCleanUp()
     */
    @Override
    void terminationCleanUp() {
        // nothing to do
    }
    
    
    /* (non-Javadoc)
     * @see petruchio.cov.AbstractMultiThreadedCoverability#supportsThreading()
     */
    @Override
    public boolean supportsThreading() {
        return this.usedByPetruchioCompiler;
    }
    
    
    /**
     * If this method returns <code>true</code> the search is stopped and "target marking
     * is coverable" is returned as the result. By it's default implementation
     * it always returns <code>false</code>.
     * 
     * @param marking The current marking to inspect.
     * @return Whether the search should be stopped and "coverable" be returned.
     */
    boolean shortcutIsCoverable(final Node marking) {
        return false;
    }
    
    /**
     * If this method returns <code>true</code> the search is stopped and "target marking
     * isn't coverable" is returned as the result. By it's default implementation
     * it always returns <code>false</code>.
     * 
     * @param marking The current marking to inspect.
     * @return Whether the search should be stopped and "not coverable" be returned.
     */
    boolean shortcutIsNotCoverable(final Node marking) {
        return false;
    }
    
    /**
     * If this method returns <code>true</code> no more predecessor mCreated
     * will be generated but the {@link Backward#markingsToProcess} collection
     * will be processed. By it's default implementation it always returns
     * <code>false</code>.
     * 
     * @param marking The current marking to inspect.
     * @return Whether the search should not generate any more predecessor mCreated.
     */
    boolean shortcutDoNotGenerateMorePredecessors(final Node marking) {
        return false;
    }
  
    /* (non-Javadoc)
     * @see petruchio.interfaces.SelfDescribing#getDescription()
     */
    @Override
    public String getDescription() {
        return "Searches backward from possible communication pairs if a covering marking "
            + "can be reached from the intial marking.";
    }
    
    /**
     * @param t
     */
    private void updateInvariants(final Transition t) {
        this.updating.add(t);
        this.computeBounds();
        this.updating.clear();
        this.updating.add(t);
        this.computeDistance();
    }
    
    public void setUseBounds(final boolean value) {
        this.useBounds = value;
    }
    
    public boolean usesBounds() {
        return this.useBounds;
    }
    
    public void setUsePO(final boolean value) {
        this.usePO = value;
    }
    
    public boolean usesPO() {
        return this.usePO;
    }

    public void setUseOrdering(final boolean value) {
        this.useOrdering = value;
    }
    
    public boolean usesOrdering() {
        return this.useOrdering;
    }
    
    
    public void setUseAcceleration(final boolean value) {
        this.useAcceleration = value;
    }
    
    public boolean usesAcceleration() {
        return this.useAcceleration;
    }
    
    private void computeDistance() {
        while(!this.updating.isEmpty() && !this.stopRequested()) {
            final Transition t = this.updating.remove();

            int minDistance = -1;
//            int maxDistance = -1;
//            int sumDist = 0;
//            int countDist = 0;
            // minimize for transitions
            for(final PTArc arc : t.getInput()) {
                final Place p = arc.getSource();
                final PlaceData pd = this.getPlaceData(p);
                final int dist = pd.fixed ? -2 : pd.distance;
                if(dist >= 0 && (minDistance < 0 || dist < minDistance)) {
                    minDistance = dist;
                }
//                if(dist >= 0 && (maxDistance < 0 || dist > maxDistance)) {
//                    maxDistance = dist;
//                }
//                if(dist >= 0) {
//                    sumDist += dist;
//                    ++countDist;
//                }
            }

            final TransitionData td = this.getTransitionData(t);
//            final int oldTDistance = td.distance;
            
            if(USE_DISTANCE_HEURISTICS && minDistance >= 0) {
                td.distance = minDistance + 1 + (t instanceof ExtendedTransition ? 10 : 0);
            }
//            if(USE_DISTANCE_HEURISTICS && maxDistance >= 0) {
//                td.distance = maxDistance + 1;
//            }
//            if(USE_DISTANCE_HEURISTICS && countDist > 0) {
//                td.distance = (sumDist / countDist)+1;
//            }
            
//            if(td.distance == oldTDistance) {
//                continue;
//            }
            
            // minimize for places
            for(final TPArc arc : t.getOutput()) {
                final Place p = arc.getTarget();
                if(!this.placeAccessible(p) || !this.addSeen(p) || p.getMarking() != 0) {
                    continue;
                }
                minDistance = -1;
//                maxDistance = -1;
//                sumDist = 0;
//                countDist = 0;
                for(final TPArc a : p.getInput()) {
                    final Transition tr = a.getSource();
                    if(!this.transitionAccessible(tr)) {
                        continue;
                    }
                    final int dist = this.getTransitionData(tr).distance;
                    if(dist >= 0 && (minDistance < 0 || dist < minDistance)) {
                        minDistance = dist;
                    }
//                    if(dist >= 0 && (maxDistance < 0 || dist > maxDistance)) {
//                        maxDistance = dist;
//                    }
//                    if(dist >= 0) {
//                        sumDist += dist;
//                        ++countDist;
//                    }
                }
                
                final PlaceData pd = this.getPlaceData(p);
                final int oldPDistance = pd.distance;
                if(USE_DISTANCE_HEURISTICS && minDistance >= 0) {
                    pd.distance = minDistance + 1;
                }
//                if(USE_DISTANCE_HEURISTICS && maxDistance >= 0) {
//                    pd.distance = maxDistance + 1;
//                }
//                if(USE_DISTANCE_HEURISTICS && countDist > 0) {
//                    pd.distance = (sumDist / countDist)+1;
//                }
                
                if(oldPDistance == pd.distance) {
                    continue;
                }
                
                for(final PTArc a : p.getOutput()) {
                    final Transition next = a.getTarget();
                    if(this.transitionAccessible(next)) {
                        this.updating.add(next);
                    }
                }
            }
        }
        this.clearSeenPlaces();
    }
    
    /**
     * Sets the depth for bounded model checking.
     * 
     * @param depth The depth for bounded model checking. Set to <= 0 to disable.
     */
    public void setBMCDepth(final int depth) {
        this.bmcDepth = depth > 0 ? depth : 0;
    }
    
    /**
     * @return The depth for bounded model checking. <= 0 if disabled. 
     */
    public int getBMCDepth() {
        return this.bmcDepth;
    }
    
    /**
     * @return
     */
    public int getMaxTreeHeight() {
        return this.maxTreeHeight;
    }
    
    /**
     * 
     */
    private void computeBounds() {
//        if(true) return; // TODO
        if(this.containsExtendedTransition) {
            this.updating.clear();
        }
        while(!this.updating.isEmpty() && !this.stopRequested()) {
            final Transition t = this.updating.remove();
            if(t instanceof ExtendedTransition) {
                continue;
            }
            final TransitionData td = this.getTransitionData(t);
            final int bound = td.firingBound;
            if(bound == UNBOUNDED_FIRING) {
                continue;
            }
            
            int maxFiringBound = UNBOUNDED_FIRING;
            for(final PTArc arc : t.getInput()) {
                final Place p = arc.getSource();
                final int change = Math.max(this.getChange(t, p), 0);
                
                if(change != 0 && p.boundKnown()) {
                    int impliedBound = p.getBound() / change;
                    if(maxFiringBound == UNBOUNDED_FIRING || maxFiringBound == UNKNOWN_FIRING_BOUND || impliedBound < maxFiringBound) {
                        maxFiringBound = impliedBound;
                    }
                }
            }
            
            if(bound == UNKNOWN_FIRING_BOUND || (maxFiringBound != UNBOUNDED_FIRING && !td.onGeneratingCycle)) {
                td.firingBound = maxFiringBound;
            } else if(bound != maxFiringBound) {
                maxFiringBound = UNBOUNDED_FIRING;
                td.firingBound = maxFiringBound;
            } else {
                continue;
            }
            
            for(final TPArc arc : t.getOutput()) {
                final Place p = arc.getTarget();
                if(p.isParametric() || !this.placeAccessible(p)) {
                    continue;
                }
                int maxOut = 2;
                if(this.hasGivenBound(p)) {
                    if(this.addSeen(p)) {
                        for(final PTArc a : p.getOutput()) {
                            final Transition next = a.getTarget();
                            if(this.transitionAccessible(next) && !(t instanceof ExtendedTransition)) {
                                int outWeight = this.getMaxOutWeight(next);
                                if(outWeight > maxOut) {
                                    maxOut = outWeight;
                                    this.updating.addFirst(next);
                                } else {
                                    this.updating.add(next);
                                }
                            }
                        }
                    }
                    continue;
                }
                
                int b = p.getMarking();
                for(final TPArc a : p.getInput()) {
                    final Transition tr = a.getSource();
                    if(!this.transitionAccessible(tr)) {
                        continue;
                    }
                    if(t instanceof ExtendedTransition) {
                        b = Place.BOUND_UNKNOWN;
                        break;
                    }
                    int out = a.getWeight();
                    for(final PTArc outArc : tr.getInput()) {
                        if(outArc.getSource() == p) {
                            out -= outArc.getWeight();
                            break;
                        }
                    }
                    if(out > 0) {
                        final int fb;
                        if(tr == t) {
                            fb = maxFiringBound;
                        } else {
                            final int fbi = this.getFiringBound(tr).intValue();
                            if(fbi == UNKNOWN_FIRING_BOUND) {
                                continue;
                            }
                            fb = fbi;
                        }
                        if(fb == UNBOUNDED_FIRING) {
                            b = Place.BOUND_UNKNOWN;
                            break;
                        }
                        
                        b += fb*out;
                    }
                }
                
                if(p.getBound() != b) {
                    if(!p.boundKnown()){
                        if(b != Place.BOUND_UNBOUNDED && b != Place.BOUND_UNKNOWN) {
                            ++this.boundedPlacesCtr;
                        }
                    } else if(b == Place.BOUND_UNBOUNDED || b == Place.BOUND_UNKNOWN) {
                        --this.boundedPlacesCtr;
                        final Pair<Collection<Place>, Integer> cluster = this.clusterBounds.get(p);
                        if(cluster != null) {
                            for(final Place p2 : cluster.getA()) {
                                this.clusterBounds.remove(p2);
                            }
                        }
                    }
                    p.setBound(b);
                    for(final PTArc a : p.getOutput()) {
                        final Transition next = a.getTarget();
                        if(this.transitionAccessible(next) && !(t instanceof ExtendedTransition)) {
                            int outWeight = this.getMaxOutWeight(next);
                            if(outWeight > maxOut) {
                                maxOut = outWeight;
                                this.updating.addFirst(next);
                            } else {
                                this.updating.add(next);
                            }
                        }
                    }
                }
            }
        }
        this.clearSeenPlaces();
    }
    
    
    /**
     * @param t
     * @return
     */
    private int getMaxOutWeight(final Transition t) {
        int max = 0;
        for(final TPArc a : t.getOutput()) {
            if(a.getWeight() > max) {
                max = a.getWeight();
            }
        }
        return max;
    }
    
    
    private final Collection<Place> newPlacesToProcess = new IdentityHashSet<Place>();

    
    
    /* (non-Javadoc)
     * @see petruchio.cov.AbstractMultiThreadedCoverability#processTransition(petruchio.interfaces.petrinet.Transition)
     */
    @Override
    synchronized void processTransition(final Transition t) {
        ++this.transitionCount;
        if(!this.containsExtendedTransition && t instanceof ExtendedTransition) {
            this.containsExtendedTransition = true;
            throw new RuntimeException("Not yet supported.");
        }
        
        this.updateInvariants(t);
        // collect new places and compute output branching
        int outW = 0;
        for(final TPArc arc : t.getOutput()) {
            final Place p = arc.getTarget();
            // not needed as t just has been added and the post-set is accessible.
//            if(!this.placeAccessible(p)) {
//                continue;
//            }
            
            outW += arc.getWeight();
            
            if(this.addPlace(p)) {
                this.newPlacesToProcess.add(p);
            }
        }
        
        // update "maxTransOutBranching"
        if(outW > this.maxTransOutBranching) {
            this.maxTransOutBranching = outW;
        }
        
//        this.checkNeeded = true;
        this.checkNewTasks();
//        if(!this.hasMoreTransitionsToProcess()) {
//            this.processCollectedTransitions();
//        }
    }
    
    private int queries_coverable = 0;
    private int queries = 0;
    
    private final Collection<PlacePair> doNotCheck = new IdentityHashSet<PlacePair>();
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Coverability#doNotCheck(petruchio.interfaces.petrinet.Place, petruchio.interfaces.petrinet.Place)
     */
    @Override
    public void doNotCheck(final Place p1, final Place p2) {
        this.doNotCheck.add(new petruchio.cov.PlacePair(p1, p2));
    }
    
//    private boolean checkNeeded = false;
    
    /**
     * Coverability checks are expensive and need to be re-performed when the net has changed,
     * therefore, we accumulate all open coverability tasks and check them as late as possible.
     */
    @Override
    void processCollectedTransitions() {
        this.recheckNotCoverable();
//        this.checkNewTransitions();
    }
    
    
    private void recheckNotCoverable() {
        final List<NodeClass> lastNotCov;
        synchronized(this.notCov) {
            if(this.notCov.isEmpty()) {
                return;
            }
            
            // re-check previous tasks that failed
            lastNotCov = this.notCov;
            this.notCov = new LinkedList<NodeClass>();
        }

        // traverse in direction of increasing token count
        for(final ListIterator<NodeClass> it = lastNotCov.listIterator(lastNotCov.size()); it.hasPrevious(); ) {
            if(this.stopRequested()) {
                return;
            }
            final NodeClass nodeclass = it.previous();
            for(final Node node : nodeclass) {
                if(this.isCoverable(node, false)) {
                    ++this.queries_coverable;
                    final Marking m = node.getMarking().get(0);
                    
                    final Place p1 = m.getPlace();
                    final Place p2 = m.getMarking() == 2 ? p1 : node.getMarking().get(1).getPlace();
                    final PlacePair pair = new petruchio.cov.PlacePair(p1, p2);
                    this.addPlacePair(pair);
                    if(FINE_DEBUG) {
                        System.out.println("revised decision on " + p1 + " + " + p2);
                    }
                }
            }
        }
        
//        this.checkNeeded = false;
    }
    
    private void checkNewTasks() {
        // check new tasks
        for(final Place p1 : this.newPlacesToProcess) {
            final Collection<Place> partners = this.getCommunicators(p1);
            for(final Place p2 : partners) {
                final PlacePair pair = new petruchio.cov.PlacePair(p1, p2);
                if(this.doNotCheck.contains(pair)
                        || !this.checked.add(pair)) {
                    continue;
                }
                
                final Node target = Node.newNode();
                if(p1 == p2) {
                    target.add(p1, 2);
                } else {
                    target.add(p1, 1);
                    target.add(p2, 1);
                }
                
                ++this.queries;
                if(this.isCoverable(target, false)) {
                    this.addPlacePair(pair);
                    ++this.queries_coverable;
                } else {
                    if(FINE_DEBUG){
                        System.out.println(p1 + " + " + p2 + " not coverable");
                    }
                }
                
                if(this.stopRequested()) {
                    return;
                }
            }
        }
        this.newPlacesToProcess.clear();
    }
    
    
//    private static boolean isComm(final Transition t) {
//        if(t.getInput().size() > 1) return true;
//        if(t.getInput().size() == 0) return false;
//        for(final PTArc a : t.getInput()) {
//            return a.getWeight() > 1;
//        }
//        return false;
//    }
    
    
    public boolean isVerbose() {
        return this.verbose;
    }
    
    public void setVerbose(final boolean verbose) {
        this.verbose = verbose;
    }
    
    
    public boolean isCoverable(final Iterable<Node> markings) {
        return this.isCoverable(markings, false);
    }
    
    /**
     * @param marking
     * @param shortestTrace
     * @return
     */
    public boolean isCoverable(final Iterable<Node> targets, final boolean shortestTrace) {
        this.startReadingTransitionData();
        for(final TransitionData d : this.transitionData.values()) {
            d.firingBound = UNKNOWN_FIRING_BOUND;
        }
        this.finishReadingTransitionData();
        if(this.friendlyTransitions != null) {
            for(final petruchio.pn.Transition t : this.friendlyTransitions) {
                ((TransitionData)t.data).firingBound = UNKNOWN_FIRING_BOUND;
            }
        }
        synchronized(this) {
            final PrintWriter pw = this.dotOutput;
            File temp = null;
            if(pw != null && shortestTrace) {
                try {
                    temp = File.createTempFile("backward_search_space", null);
                    temp.deleteOnExit();
                    this.dotOutput = new PrintWriter(temp);
                } catch(final IOException e) {
                    System.out.println("Problems creating temporary DOT file: " + e);
                }
            }
            this.maxTreeHeight = 0;
            boolean result = this.isMarkingCoverable(targets, true);
            if(result && shortestTrace) {
                final int depth = this.getBMCDepth();
                SimpleList<Transition> path;
                final AtomicBoolean pleaseStopHelperThread = new AtomicBoolean();
                try {
                    this.startShutdownKeyThread(pleaseStopHelperThread);
                    Node initial;
                    do {
                        path = this.getResultTrace();
                        initial = this.getResultTraceInitial();
                        System.out.println("> Found trace of length " + path.getLength() + ": " + Backward.formatTrace(path) + ", looking for shorter trace.");
                        this.setBMCDepth(path.getLength()-1);
                        File curTemp = null;
                        if(pw != null && shortestTrace) {
                            try {
                                curTemp = File.createTempFile("backward_search_space", null);
                                curTemp.deleteOnExit();
                                this.dotOutput = new PrintWriter(curTemp);
                            } catch(final IOException e) {
                                System.out.println("Problems creating temporary DOT file: " + e);
                            }
                        }
                        result = this.isMarkingCoverable(targets, true);
                        if(result) {
                            path = this.getResultTrace();
                            if(pw != null) {
                                temp = curTemp;
                            }
                        }
                    } while(result && path.getLength() > 0);
                    System.out.println("> No traces of length <= " + (path.getLength()-1) + " found.");
                    this.resultPath = path;
                    this.resultPathInitial = initial;
                } finally {
                    pleaseStopHelperThread.set(true);
                }
                this.setBMCDepth(depth);
                result = true;
            }
            
            if(pw != null && shortestTrace) {
                try {
                    final FileReader reader = new FileReader(temp);
                    final char[] buff = new char[1024];
                    while(reader.ready()) {
                        final int len = reader.read(buff);
                        pw.write(buff, 0, len);
                    }
                    pw.flush();
                    pw.close();
                } catch(final IOException e) {
                    System.out.println("Problems copying DOT file: " + e);
                }
            }
            return result;
        }
    }
    
    private void startShutdownKeyThread(final AtomicBoolean pleaseStopHelperThread) {
        System.out.println("> Press 'x' and 'return' to abort search for shorter traces and use shortest known so far.");
        (new Thread(new Runnable() {
            @Override
            public void run() {
                final Reader in = new InputStreamReader(System.in);
                try {
                    while(!pleaseStopHelperThread.get()) {
                        if(!in.ready()) {
                            try {
                                Thread.sleep(100);
                                continue;
                            } catch (final InterruptedException e) {
                                Thread.currentThread().interrupt();
                            }
                        }
                        final int character = in.read();
                        if(character < 0) {
                            break;
                        }
                        if(character == 'x' || character == 'X') {
                            System.out.println("> Sending abort message to coverability checker.");
                            Backward.this.pleaseStopComputation();
                        }
                    }
                } catch (final IOException e) {
                    e.printStackTrace();
                } finally {
                    try {
                        in.close();
                    } catch (final IOException e) {
                        // ignore
                    }
                }
            }
            
        })).start();
    }
    
    void verbose(final String s) {
        if(this.verbose) {
            System.out.println("> " + s.replaceAll("\n", "\n> "));
        }
    }
   
    private boolean isCoverable(final Node target, boolean exact) {
        final List<Node> alibi = new ArrayList<Node>(1);
        alibi.add(target);
        return this.isCoverableClean(alibi, exact);
    }
    
    
    private boolean isMarkingCoverable(final Iterable<Node> markings,
            final boolean exact) {
        return this.isCoverableClean(markings, exact);
    }
    
    
    void dotTransition(final Node pre, final Node post, final Transition t) {
        this.dotTransition(pre, post, t, 1);
    }
    
    void dotTransition(final Node pre, final Node post, final Transition t, final int mult) {
        if(this.dotOutput != null) {
            this.dotOutput.println("  " + this.toDotName(pre) + " -> " + this.toDotName(post) + " [label=\"" + (mult == 1 ? "" : mult + "*") + t.toString() + "\"];");
        }
    }
    
    void dotTransition(final Node pre, final Node post, final SimpleList<Transition> t, final int loops) {
        if(this.dotOutput != null) {
            final String trace = (loops > 1 ? "(" : "") + t.toString() + (loops > 1 ? ")^" + loops : "") ;
            this.dotOutput.println("  " + this.toDotName(pre) + " -> " + this.toDotName(post) + " [style=\"bold\", label=\"" + trace + "\"];");
        }
    }
    
    void dotCoversExamined(final Node less, final Node greater) {
        final String a, b;
        if(this.dotOutput != null && this.dotPrintCutoffs && !(a = this.toDotName(less)).equals(b = this.toDotName(greater))) {
            this.dotOutput.println("  " + b + " -> " + a + " [style=\"dashed\", label=\"covers examined\"];");
        }
    }
    
    void dotCoversNotCoverable(final Node less, final Node greater) {
        final String a, b;
        if(this.dotOutput != null && this.dotPrintCutoffs && !(a = this.toDotName(less)).equals(b = this.toDotName(greater))) {
            this.dotOutput.println("  " + b + " -> " + a + " [style=\"dashed\", label=\"covers not coverable\"];");
        }
    }
    
    void dotCoveredByInitial(final Node less, final Node greater) {
        final String a, b;
        if(this.dotOutput != null && !(a = this.toDotName(less)).equals(b = this.toDotName(greater))) {
            this.dotOutput.println("  " + b + " -> " + a + " [style=\"dashed\", label=\"covered by initial\"];");
        }
    }
    
    void dotCutoff(final Node a, final String reason) {
        if(this.dotOutput != null && this.dotPrintCutoffs) {
            this.dotOutput.println("  " + System.identityHashCode(a) + " -> " + this.toDotName(a) + " [style=\"dashed\"];");
            this.dotOutput.println("  " + System.identityHashCode(a) + " [shape=\"plaintext\", label=\"" + reason + "\"];");
        }
    }
    
    private String toDotName(final Node n) {
        return "\"" + n.toString() + "\"";
    }
    
    
    /**
     * @param target
     * @param exact
     * @return
     */
    private boolean isCoverableClean(final Iterable<Node> target, final boolean exact) {
        if(this.dotOutput != null) {
            this.dotOutput.println("digraph BackwardCoverable {");
            this.dotOutput.println("  node [shape=\"rectangle\", margin=\"0,0\"];");
            this.dotOutput.println("  nil  [shape=\"point\"];");
            this.dotOutput.println("  nil  -> " + this.toDotName(this.initialState) + ";");
            for(final Node n : target) {
                this.dotOutput.println("  " + this.toDotName(n) + " [style=\"bold\"];");
            }
        }
        final boolean result = this.isCoverableUnclean(target, exact);
        if(!result) {
            for(final Node n : target) {
                this.addNotCoverable(n);
            }
        }
        if(this.dotOutput != null) {
            if(result && this.resultPath.getValue() != null) {
                final Deque<Transition> reversed = new LinkedList<Transition>();
                for(final Transition t : this.resultPath) {
                    reversed.addFirst(t);
                }
                final Node reached = this.fire(this.initialState, this.resultPath);
                Node next = null;
                for(final Node n : target) {
                    if(reached.greaterEquals(n)) {
                        next = n;
                        break;
                    }
                }
                if(next != null) {
                    String nextname = this.toDotName(next);
                    for(final Transition t : reversed) {
                        final Node prev = Node.newNode(next);
                        final TransitionData d = this.getTransitionData(t);
                        prev.subtract(this.getOutNode(t, d));
                        prev.add(this.getInNode(t, d));
                        final String prevname = this.toDotName(prev);
                        this.dotOutput.println("  " + prevname + " -> " + nextname
                                + " [style=\"bold\", color=\"red\", fontcolor=\"red\", label=\"" + t.toString() + "\"];");
                        next = prev;
                        nextname = prevname;
                    }
                }
            }
            this.dotOutput.println("}");
            this.dotOutput.flush();
            this.dotOutput.close();
            this.dotOutput = null;
        }
        synchronized(this.markingsToProcess) {
            this.markingsToProcess.clear();
        }
        this.clearNodes();
        this.clearSeenPlaces();
        this.clearSeenTransitions();
        return result;
    }

    
    /**
     * 
     */
    private synchronized void clearNodes() {
        this.inspectedMarkings.clear();
    }
    
    
    /**
     * @param mark
     * @return
     */
    Node notCoverable(final Node mark) {
        return this.coversNode(mark, this.notCov);
    }
    
    public long getMarkingsCreated() {
        return this.markingsCreated;
    }
    
    public long getMarkingsExamined() {
        return this.markingsExamined;
    }
    
    
    /**
     * @param target
     * @param exact setting this parameter to <code>false</code> instructs the algorithm to speed up the search by assuming that there are no dead transitions. 
     * @return
     */
    private synchronized boolean isCoverableUnclean(final Iterable<Node> target, final boolean exact) {
        synchronized(this.predecessorResultLock) {
            this.resultPath = null;
            this.resultPathInitial = null;
        }
        
        this.markingsCreated = 0;
        this.markingsExamined = 0;
        this.markingsToProcess.clear();
        boolean allKnowinglyNotCoverable = true;
        for(final Node n : target) {
            if(this.boundedPlacesCtr != 0 && this.useBounds && this.isViolator(n)) {
                continue;
            }
            
            Node node;
            if(!exact && (node = this.notCoverable(n)) != null) {
                if(FINE_DEBUG) { System.out.println("covers node known to be uncoverable."); }
                this.dotCoversNotCoverable(node, n);
                continue;
            }
            
            if((node = this.isCoveredByGoodState(n)) != null) {
                synchronized(this.predecessorResultLock) {
                    this.resultPath = new SimpleList<Transition>(null);
                    this.resultPathInitial = Node.newNode(node);
                }
                this.dotCoveredByInitial(n, node);
                return true;
            }
           
            Backward.addSortedRandomAccess(this.markingsToProcess, n, this.searchGuide);
            ++this.markingsCreated;
            allKnowinglyNotCoverable = false;
        }
        if(allKnowinglyNotCoverable) {
            return false;
        }
        
        long lastVerboseOutput = 0;
        if(this.verbose) {
            lastVerboseOutput = System.currentTimeMillis();
        }
        
        final long startTime = System.currentTimeMillis();
        int checkTime = Backward.CHECK_TIME_EVERY;

        while(!this.markingsToProcess.isEmpty() && !this.stopRequested()) {
            if(this.verbose) {
                final long now = System.currentTimeMillis();
                if(now - lastVerboseOutput > Backward.VERBOSE_OUTPUT_DELAY) {
                    lastVerboseOutput = now;
                    int min = 0;
                    for(final NodeClass n : this.inspectedMarkings) {
                        min += n.size();
                    }
                    this.verbose("... searching. " + this.markingsCreated + "/" + this.markingsExamined
                            + " created/examined. " + this.markingsToProcess.size() + " queued. " + min + " minimal examined in " + this.inspectedMarkings.size() + " classes."
                            + "\n    next: " + this.markingsToProcess.get(this.markingsToProcess.size()-1));
                }
            }
            --checkTime;
            if(this.usedByPetruchioCompiler && checkTime <= 0) {
                checkTime = Backward.CHECK_TIME_EVERY;
                if(this.countKnownPlaces() - (this.newPlacesToProcess.isEmpty() ? 1 : Backward.PLACE_DIFFERENCE_BEFORE_INV) >= this.invComputedAt
                        && System.currentTimeMillis()-startTime >= Backward.TIME_BEFORE_INV) {
                    boolean change = this.computeInvariants();
                    change |= this.computeClusterBounds(this.getAccessiblePlaces());
                    if(change) {
                        for(final Iterator<Node> it = this.markingsToProcess.iterator(); it.hasNext();) {
                            final Node n = it.next();
                            if(this.isViolator(n)) {
                                it.remove();
                            }
                        }
                        for(final Iterator<NodeClass> it = this.inspectedMarkings.iterator(); it.hasNext();) {
                            final NodeClass n = it.next();
                            n.removeViolators();
                            if(n.isEmpty()) {
                                it.remove();
                            }
                        }
                    }
                    continue;
                }
                if(Backward.DYNAMIC_RECHECK && this.usedByPetruchioCompiler && System.currentTimeMillis() - startTime > Backward.DYNAMIC_RECHECK_THRESHOLD
                        && !this.newPlacesToProcess.isEmpty()) {
                    this.verbose("This query takes too long. Postponing...");
                    return false;
                }
            }
            
            boolean keepGeneratingPredecessors = true;
            
            final Node mark;
            synchronized(this.markingsToProcess) {
//                mark = this.stack.removeFirst();
                mark = this.markingsToProcess.remove(this.markingsToProcess.size()-1);
            }
            if(this.bmcDepth != 0
                    && mark.path != null
                    && mark.path.getLength() >= this.bmcDepth) {
                // current path too long
                this.dotCutoff(mark, "bmc depth reached");
                continue;
            }
            Node node;
            if((node = this.coversExaminedNode(mark)) != null) {
                // node has already been checked
                this.dotCoversExamined(node, mark);
                continue;
            }
            if((node = this.addExaminedNode(mark)) != null) {
                if(node == Backward.ACCELERATED_COVERED) {
                    return true;
                }
                // node has already been checked
                this.dotCoversExamined(node, mark);
                continue;
            }
            
            if(this.shortcutIsCoverable(mark)) {
                return true;
            }
            if(this.shortcutIsNotCoverable(mark)) {
                return false;
            }
            if(keepGeneratingPredecessors && this.shortcutDoNotGenerateMorePredecessors(mark)) {
                keepGeneratingPredecessors = false;
            }
            
            ++this.markingsExamined;
            
            boolean checkForCommonPreTransition = !exact && (mark.getTokens() <= this.maxTransOutBranching);
            
            boolean isPOPlace = false;
            boolean markingSize1 = mark.getMarking().size() == 1 && mark.getTokens() == 1; 
            final Collection<Transition> preTransitions = new IdentityHashSet<Transition>();
            final Collection<Transition> tempTransitions = new IdentityHashSet<Transition>();
            collect_transitions:
            for(final Marking m : mark.getMarking()) {
                if(this.stopRequested()) {
                    break;
                }
                final Place p = m.getPlace();
                isPOPlace = this.usePO && m.getMarking() > p.getMarking() && !markingSize1 && !p.isParametric();
                for(final TPArc a : p.getInput()) {
                    final Transition t = a.getSource();
                    if(!this.transitionAccessible(t)) {
                        continue;
                    }
                    if(isPOPlace) {
                        if(t.getOutput().size() > 1 || a.getWeight() > m.getMarking() || (this.containsExtendedTransition && t instanceof ExtendedTransition)) {
                            isPOPlace = false;
                            preTransitions.add(t);
                        } else {
                            tempTransitions.add(t);
                        }
                    } else {
                        preTransitions.add(t);
                    }
                }
                if(isPOPlace) {
                    preTransitions.clear();
                    preTransitions.addAll(tempTransitions);
                    tempTransitions.clear();
                    break collect_transitions;
                }
                preTransitions.addAll(tempTransitions);
                tempTransitions.clear();
                
                if(checkForCommonPreTransition) {
                    final Collection<Transition> commonPreTransitions = new IdentityHashSet<Transition>();
                    if(commonPreTransitions.isEmpty()) { // first loop iteration
                        if(m.getMarking() > 1) {
                            for(final Transition t : preTransitions) {
                                for(final TPArc a : t.getOutput()) {
                                    if(a.getTarget() == p && a.getWeight() >= m.getMarking() && this.placeAccessible(a.getTarget())) {
                                        commonPreTransitions.add(t);
                                        break;
                                    }
                                }
                            }
                        } else {
                            commonPreTransitions.addAll(preTransitions);
                        }
                    } else {
                        if(m.getMarking() > 1) {
                            for(final Iterator<Transition> it2 = commonPreTransitions.iterator(); it2.hasNext(); ) {
                                final Transition t = it2.next();
                                boolean removed = false;
                                for(final TPArc a : t.getOutput()) {
                                    if(a.getTarget() == p && a.getWeight() < m.getMarking() && this.placeAccessible(a.getTarget())) {
                                        it2.remove();
                                        removed = true;
                                        break;
                                    }
                                }
                                if(!removed && !preTransitions.contains(t)) {
                                    it2.remove();
                                }
                            }
                        } else {
                            commonPreTransitions.retainAll(preTransitions);
                        }
                    }
                    if(commonPreTransitions.isEmpty()) {
                        checkForCommonPreTransition = false;
                    }
                }
            }
            
            if(checkForCommonPreTransition) {
                if(!isPOPlace) {
                    // if all marked places share an input transition
                    // this marking can be reached as the transition can fire
                    return true;
                }
            }
            
            final Result result;
            if(keepGeneratingPredecessors) {
                result = this.generatePredecessors(isPOPlace, exact, preTransitions, mark);
            } else {
                result = Result.UNKNOWN;
            }
            switch(result) {
                case COVERABLE:
                    return true;
                case NOT_COVERABLE:
                case UNKNOWN:
                default: // UNKNOWN
                    // do nothing
                    break;
            }
            
        }
        
        return this.resultPath != null;
    }
    
    
    /**
     * @param mark
     */
    private synchronized void addNotCoverable(final Node mark) {
        this.addNode(mark, this.notCov);
    }
    
    private boolean computeInvariants() {
        this.verbose("search takes more than " + Backward.TIME_BEFORE_INV + "ms. generating invariants.");
        this.invComputedAt = this.countKnownPlaces();
        final Collection<Map<Place, Integer>> invs = this.generateInvariants(
                this.getAccessibleTransitions(), this.getAccessiblePlaces(), false, 0);
        final int pl = this.adjustBoundsByPlaceInvariants(invs);
        for(final Iterator<Map<Place, Integer>> it = invs.iterator(); it.hasNext(); ) {
            for(final Integer c : it.next().values()) {
                if(c.intValue() < 0) {
                    it.remove();
                    break;
                }
            }
        }
        if(this.verbose) {
            this.verbose(this.boundedPlacesCtr + " bounds known (" + pl + " new from " + invs.size() + " invariants).");
        }
        return pl > 0;
    }
    
    
    /**
     * @param <T>
     * @param l
     * @param key
     * @param c
     */
    static <T> void addSortedRandomAccess(final List<T> l,
            final T key, final Comparator<? super T> c) {
        synchronized(l) {
            int low = 0;
            int high = l.size() - 1;
    
            while(low <= high) {
                int mid = (low + high) >>> 1;
                T midVal = l.get(mid);
                int cmp = c.compare(midVal, key);
    
                if(cmp < 0) {
                    low = mid + 1;
                } else if(cmp > 0) {
                    high = mid - 1;
                } else {
                    low = mid;
                    break; // key found
                }
            }
            l.add(low, key);  // key not found
        }
    }
    
    static <T> int findSortedRandomAccess(final List<T> l,
        final T key, final Comparator<? super T> c) {
        int low = 0;
        int high = l.size() - 1;

        while(low <= high) {
            int mid = (low + high) >>> 1;
            T midVal = l.get(mid);
            int cmp = c.compare(midVal, key);

            if(cmp < 0) {
                low = mid + 1;
            } else if(cmp > 0) {
                high = mid - 1;
            } else {
                return low;
            }
        }
        return -1;
    }
    
    /**
     * @param <T>
     * @param l
     * @param key
     * @param c
     */
    static <T> void addSortedIteratorBased(final List<T> l,
            final T key, final Comparator<? super T> c) {
        synchronized(l) {
            int low = 0;
            int high = l.size()-1;
            final ListIterator<T> i = l.listIterator();
            while (low <= high) {
                int mid = (low + high) >>> 1;
                T midVal = null;
                int pos = i.nextIndex();
                if (pos <= mid) {
                    do {
                        midVal = i.next();
                    } while (pos++ < mid);
                } else {
                    do {
                        midVal = i.previous();
                    } while (--pos > mid);
                }
                final int comparison = c.compare(midVal, key);
                if(comparison < 0) { // mid < pre
                    low = mid + 1;
                } else if (comparison != 0) { // mid > pre
                    high = mid - 1;
                } else { // mid == pre
                    break;
                }
            }
            i.add(key);
        }
    }
    
    
    @GuardedBy("predecessorResultLock")
    volatile Result predecessorResult;
    @GuardedBy("predecessorResultLock")
    volatile SimpleList<Transition> resultPath;
    @GuardedBy("predecessorResultLock")
    volatile Node resultPathInitial;
    final Object predecessorResultLock = new Object();

    boolean dotPrintCutoffs = false;
    PrintWriter dotOutput = null;
    
    public SimpleList<Transition> getResultTrace() {
        synchronized(this.predecessorResultLock) {
            return this.resultPath;
        }
    }
    
    public Node getResultTraceInitial() {
        synchronized(this.predecessorResultLock) {
            return this.resultPathInitial;
        }
    }
    
    
    private Pair<Boolean, SimpleList<Pair<Place, Map<Place, Integer>>>> getMultipliers(final ExtendedTransition t, final TransitionData d) {
        synchronized(t) {
            Pair<Boolean, SimpleList<Pair<Place, Map<Place, Integer>>>> multipliers = d.multipliers == null ? null : d.multipliers.get();
            if(multipliers == null) {
                boolean simple = true;
                final Collection<Place> adressed = new IdentityHashSet<Place>();
                final Map<Place, Map<Place, Integer>> tmp = new IdentityHashMap<Place, Map<Place, Integer>>();
                for(final TPArc a : t.getOutput()) {
                    if(a instanceof TransferArc) {
                        final Place targetPlace = a.getTarget();
                        for(final Map.Entry<Place, Integer> e : ((TransferArc)a).getTransfer().entrySet()) {
                            final Place thePlace = e.getKey();
                            if(thePlace == targetPlace) {
                                continue;
                            }
                            Map<Place, Integer> mults = tmp.get(thePlace);
                            if(mults == null) {
                                mults = new IdentityHashMap<Place, Integer>();
                                tmp.put(thePlace, mults);
                            }
                            mults.put(targetPlace, e.getValue());
                            if(simple && !adressed.add(targetPlace)) {
                                simple = false;
                            }
                        }
                    }
                }
                SimpleList<Pair<Place, Map<Place, Integer>>> mults = null;
                for(final Map.Entry<Place, Map<Place, Integer>> e : tmp.entrySet()) {
                    mults = new SimpleList<Pair<Place, Map<Place, Integer>>>(new Pair<Place, Map<Place, Integer>>(e.getKey(), e.getValue()), mults);
                }
                multipliers = new Pair<Boolean, SimpleList<Pair<Place, Map<Place, Integer>>>>(Boolean.valueOf(simple), mults);
                d.multipliers = new SoftReference<Pair<Boolean, SimpleList<Pair<Place, Map<Place, Integer>>>>>(multipliers);
            }
            return multipliers;
        }
    }
    
    SimpleList<Place> getTransferTargets(final ExtendedTransition t, final TransitionData d) {
        synchronized(t) {
            SimpleList<Place> targets = d.transferTargets == null ? null : d.transferTargets.get();
            if(targets == null) {
                for(final TPArc a : t.getOutput()) {
                    if(a instanceof TransferArc) {
                        targets = new SimpleList<Place>(a.getTarget(), targets);
                    }
                }
                d.transferTargets = new SoftReference<SimpleList<Place>>(targets);
            }
            return targets;
        }
    }
    
    SimpleList<Place> getResetSources(final ExtendedTransition t, final TransitionData d) {
        synchronized(t) {
            SimpleList<Place> sources = d.resetSources == null ? null : d.resetSources.get();
            if(sources == null) {
                for(final PTArc a : t.getInput()) {
                    if(a instanceof ResetArc) {
                        sources = new SimpleList<Place>(a.getSource(), sources);
                    }
                }
                d.resetSources = new SoftReference<SimpleList<Place>>(sources);
            }
            return sources;
        }
    }
    
    SimpleList<InhibitorArc> getInhibitorArcs(final ExtendedTransition t, final TransitionData d) {
        synchronized(t) {
            SimpleList<InhibitorArc> sources = d.inhibitors == null ? null : d.inhibitors.get();
            if(sources == null) {
                for(final PTArc a : t.getInput()) {
                    if(a instanceof InhibitorArc) {
                        sources = new SimpleList<InhibitorArc>((InhibitorArc)a, sources);
                    }
                }
                d.inhibitors = new SoftReference<SimpleList<InhibitorArc>>(sources);
            }
            return sources;
        }
    }
    
    
    static int checkedADD(final int a, final int b) {
        final int r = a + b;
        // ((a^r)&(b^r))<0) is a shorthand for:
        // (a<0 && b<0 && r>=0) ||
        // (a>0 && b>0 && r<=0) )
        if (((a ^ r) & (b ^ r)) < 0) {
            throw new ArithmeticException("Overflow: " + a + " + " + b + " != " + r);
        }
        return r;
    }
    
    static long checkedADD(final long a, final long b) {
        if (a > b) {
            return checkedADD(b, a);
        }
        final long r;
        if (a < 0) {
            if (b < 0) {
                if (Long.MIN_VALUE - b <= a) {
                    r = a + b;
                } else {
                    throw new ArithmeticException();
                }
            } else {
                r = a + b;
            }
        } else {
            if (a <= Long.MAX_VALUE - b) {
                r = a + b;
            } else {
                throw new ArithmeticException();
            }
        }
        return r;
    }
    
    static int gcd(final int valueA, final int valueB) {
        int a = Math.abs(valueA);
        int b = Math.abs(valueB);

        // ensure b > a
        if (b <= a) {
            final int tmp = b;
            b = a;
            a = tmp;
        }

        if (a != 0) {
            for (int tmp; (b %= a) != 0;) {
                tmp = b;
                b = a;
                a = tmp;
            }
            return a;
        } else if (b != 0) {
            return b;
        } else {
            return 0;
        }
    }
    
    static int checkedMUL(final int a, final int b) {
        if(a == 0 || b == 0) {
            return 0;
        } else if(a == 1) {
            return b;
        } else if(b == 1) {
            return a;
        }
//        // check and do
//        if(((a > 0 && b > 0) || (a < 0 && b < 0))
//                ? Math.abs(a) > Math.abs(Integer.MAX_VALUE/b) || Math.abs(b) > Math.abs(Integer.MAX_VALUE/a)
//                : (a < Integer.MIN_VALUE/b || b < Integer.MAX_VALUE/a)) {
//            throw new ArithmeticException("Overflow: " + a + " * " + b);
//        }
//        return a*b;
        // do and check
        final long r = (long)a * b;
        if(r > Integer.MAX_VALUE || r < Integer.MIN_VALUE) {
            throw new ArithmeticException("Overflow: " + a + " * " + b + " != " + (int)r);
        }
        return (int)r;
    }
    
    static long checkedMUL(final long a, final long b) {
        if (a > b) {
            return checkedMUL(b, a);
        }
        final long r;
        if (a < 0) {
            if (b < 0) {
                if (a >= Long.MAX_VALUE / b) {
                    r = a * b;
                } else {
                    throw new ArithmeticException();
                }
            } else if (b > 0) {
                if (Long.MIN_VALUE / b <= a) {
                    r = a * b;
                } else {
                    throw new ArithmeticException();

                }
            } else {
                // assert b == 0
                r = 0;
            }
        } else if (a > 0) {
            if (a <= Long.MAX_VALUE / b) {
                r = a * b;
            } else {
                throw new ArithmeticException();
            }
        } else {
            r = 0;
        }
        return r;
    }
    
    
    Iterable<Node> getTakenByTransfer(final ExtendedTransition t, final TransitionData d, final Node mark) {
        final Pair<Boolean, SimpleList<Pair<Place, Map<Place, Integer>>>> multipliers = this.getMultipliers(t, d);
        if(multipliers.getB() == null) {
            return new SimpleList<Node>(null);
        }
        final Node copy = Node.newNode(mark);
        copy.subtract(this.getOutNode(t, d));
        copy.add(this.getInNode(t, d));
        if(multipliers.getA().booleanValue()) { // simple transfer
            final Node taken = Node.newNode();
            for(final Pair<Place, Map<Place, Integer>> pair : multipliers.getB()) {
                final Place place = pair.getA();
                final Map<Place, Integer> mult = pair.getB();
                int max = 0;
                for(final Map.Entry<Place, Integer> e : mult.entrySet()) {
                    final int coefficient = e.getValue().intValue();
                    final Marking m = copy.getMarking(e.getKey());
                    final int targMarking = m == null ? 0 : m.getMarking();
                    // max{0, target marking - constant part - initial marking} = tokens available on target
                    // ceil(tokens available / current coefficient) = tokens taken from source for this target
                    final int times = Math.max(0, targMarking+coefficient-1) / coefficient; // ceil
                    if(times > max) {
                        max = times;
                    }
                }
                taken.add(place, max);
            }
            return new SimpleList<Node>(taken);
        }
        
//        System.out.println("--- " + t + " --- " + mark);
        int remainingTokens = 0;
        for(final Place p : this.getTransferTargets(t, d)) {
            final Marking m = copy.getMarking(p);
            if(m != null) {
                remainingTokens += m.getMarking() - m.getPlace().getMarking();
            }
        }
        final List<Node> result = new LinkedList<Node>();
        this.getTakenByTransfer(multipliers.getB(), copy, Node.newNode(), result, remainingTokens);
        return result;
    }
    
    private void getTakenByTransfer(final SimpleList<Pair<Place, Map<Place, Integer>>> mult,
            final Node mark, final Node cur, final List<Node> bindings, final int remainingTokens) {
        if(remainingTokens == 0) {
            // no more tokens to handle
            boolean add = true;
            for(final Iterator<Node> it = bindings.iterator(); it.hasNext(); ) {
                final Node n = it.next();
                if(n.greaterEquals(cur)) {
                    it.remove();
                } else if(cur.greaterEquals(n)) {
                    add = false;
                    break;
                }
            }
            if(add) {
                bindings.add(Node.newNode(cur));
            }
            return;
        }
        if(mult == null) {
            return;
        }
//        System.out.println("looking at " + mult.getValue() + " on " + mark + " with " + remainingTokens);
        
        final Pair<Place, Map<Place, Integer>> currentPair = mult.getValue();
        final Place curPlace = currentPair.getA();
        final Map<Place, Integer> multipliers = currentPair.getB();
        int max = 0;
        for(final Map.Entry<Place, Integer> e : multipliers.entrySet()) {
            final int coefficient = e.getValue().intValue();
            final Marking m = mark.getMarking(e.getKey());
            final int targMarking = m == null ? 0 : m.getMarking();
            // max{0, target marking - constant part} = tokens available on target
            // ceil(tokens available / current coefficient) = tokens taken from source for this target
            final int times = Math.max(0, targMarking+coefficient-1 - (m == null ? 0 : m.getPlace().getMarking())) / coefficient; // ceil
            if(times > max) {
                max = times;
            }
        }
//        System.out.println(" max = " + max);

        final SimpleList<Pair<Place, Map<Place, Integer>>> next = mult.getNext();
        if(next == null && max == 0) {
            return;
        }
        
        this.getTakenByTransfer(next, mark, cur, bindings, remainingTokens); // take 0 tokens
        if(max == 0) {
            return;
        }
        
        final Node copy = Node.newNode(mark);
        int remaining = remainingTokens;
        if(next == null) {
            for(final Map.Entry<Place, Integer> e : multipliers.entrySet()) {
                final int coefficient = e.getValue().intValue();
                remaining -= copy.semisubtract(e.getKey(), checkedMUL(max, coefficient));
            }
            cur.add(curPlace, max);
            this.getTakenByTransfer(next, copy, cur, bindings, remaining);
            cur.remove(curPlace);
            return;
        }
        
        for(int tokens = 1; tokens <= max; ++tokens) { // take 1 ... max tokens
            for(final Map.Entry<Place, Integer> e : multipliers.entrySet()) {
                final int coefficient = e.getValue().intValue();
                remaining -= copy.semisubtract(e.getKey(), coefficient);
            }
            cur.add(curPlace);
            this.getTakenByTransfer(next, copy, cur, bindings, remaining);
        }
        cur.remove(curPlace);
        
        return;
    }
    
    
    /**
     * @param markingsToProcess
     * @param transitions
     * @param mark
     * @return
     */
    private synchronized Result generatePredecessors(final boolean isPOPlace, final boolean exact,
            final Collection<Transition> transitions, final Node mark) {
        this.predecessorResult = Result.UNKNOWN;

        for(final Iterator<Transition> it = transitions.iterator(); it.hasNext(); ) {
            if(this.stopRequested()) {
                transitions.clear();
                break;
            }
            final Transition t = it.next();
            it.remove();
            
            final Runnable command = new Runnable() {
                @Override
                public void run() {
                    ++Backward.this.markingsCreated;
                    
                    if(Backward.this.containsExtendedTransition && t instanceof ExtendedTransition) { // extended transition
                        final TransitionData d = Backward.this.getTransitionData(t);
                        final Node out = Backward.this.getOutNode(t, d);
                        final Iterable<Place> resets = Backward.this.getResetSources((ExtendedTransition)t, d);
                        if(resets != null) {
                            for(final Place p : resets) {
                                final Marking mm = mark.getMarking(p);
                                final Marking mm2 = out.getMarking(p);
                                if((mm == null ? 0 : mm.getMarking()) != (mm2 == null ? 0 : mm2.getMarking())) {
                                    boolean stillOkayBecauseOfTransfer = false;
                                    final Iterable<Place> targ = Backward.this.getTransferTargets((ExtendedTransition)t, d);
                                    if(targ != null) {
                                        for(final Place p2 : targ) {
                                            if(p2 == p) {
                                                stillOkayBecauseOfTransfer = true;
                                                break;
                                            }
                                        }
                                    }
                                    if(!stillOkayBecauseOfTransfer) {
                                        return;
                                    }
                                }
                            }
                        }
                        
                        final Iterable<Node> takenByTransfer = Backward.this.getTakenByTransfer((ExtendedTransition)t, d, mark);
//                        System.out.println(mark + " << " + t + " trans: " + takenByTransfer);
                        if(takenByTransfer == null) {
                            return;
                        }
                        final Node in = Backward.this.getInNode(t, d);
                        final Node preCand = Node.newNode(mark);
                        preCand.accelerated = false;
//                        pre.traceGuess = null;
//                        pre.traceGuessHead = null;
                        preCand.subtract(out);
//                        pre.add(in);
                        if(!preCand.addWithBoundsChecking(in)) {
                            return;
                        }
                        final SimpleList<Place> transferTargets = Backward.this.getTransferTargets((ExtendedTransition)t, d);
                        // floor((m - out)/(in - out)), *t1+*t2-(t1*-*t1)+...
                        transfer_configurations:
                        for(final Iterator<Node> it = takenByTransfer.iterator(); it.hasNext(); ) {
                            final Node taken = it.next();
                            final Node pre = it.hasNext() ? Node.newNode(preCand) : preCand;
                            if(taken != null) {
                                for(final Place p : transferTargets) {
                                    pre.remove(p);
                                }
                                for(final TPArc a : t.getOutput()) {
                                    if(a instanceof TransferArc) {
                                        final TransferArc transfer = (TransferArc)a;
                                        for(final Place p : transfer.getTransfer().keySet()) {
                                            if(transfer.getTarget() == p) {
                                                continue;
                                            }
                                            final Marking tokens = taken.getMarking(p);
                                            final Marking mm = in.getMarking(p);
                                            if(p.boundKnown()
                                                    && ((mm != null && p.getBound() < mm.getMarking())
                                                            || (tokens != null && p.getBound() < tokens.getMarking()))) {
                                                continue transfer_configurations;
                                            }
                                            pre.set(p, Math.max(tokens == null ? 0 : tokens.getMarking(), mm == null ? 0 : mm.getMarking()));
                                        }
                                    }
                                }
                            }
                            
                            final Iterable<InhibitorArc> inhi = Backward.this.getInhibitorArcs((ExtendedTransition)t, d);
                            if(inhi != null) {
                                boolean okayWithInhibitors = true;
                                for(final InhibitorArc a : inhi) {
                                    final Marking m = pre.getMarking(a.getSource());
                                    if(m != null && m.getMarking() > a.getBound()) {
                                        okayWithInhibitors = false;
                                        break;
                                    }
                                }
                                if(!okayWithInhibitors) {
                                    continue;
                                }
                            }
                            
                            
//                            // check if correct
//                            final Node copy = Node.newNode(pre);
//                            Exception exception = null;
//                            try {
//                                Backward.this.fire(copy, t);
//                            } catch(final RuntimeException e) {
//                                exception = e;
//                            }
//                            if(exception != null || !copy.greaterEquals(mark)) {
//                                throw new RuntimeException("ERROR: result succ does not cover true successor."
//                                      + "\n Successor:    " + mark
//                                      + "\n Computed pre: " + pre
//                                      + "\n Firing:       " + t + " " + taken
//                                      + "\n Result succ:  " + (exception != null ? "unavailable" : String.valueOf(copy))
//                                      + "\n Transfers:    " + takenByTransfer
//                                      + (exception == null ? "" : "\n Exception:    " + exception.getCause()));
//                            }
//                            // VERBOOOOOOOSE
//                            System.out.println("---"
//                                    + "\n Successor:    " + mark
//                                    + "\n Computed pre: " + pre
//                                    + "\n Firing:       " + t + " " + taken
//                                    + "\n Result succ:  " + copy
//                                    + "\n Transfers:    " + takenByTransfer);
                            
                            if(Backward.this.predecessorResult == Result.UNKNOWN
                                    && Backward.this.boundedPlacesCtr != 0
                                    && Backward.this.useBounds
                                    && Backward.this.isViolator(pre)) {
                                if(Backward.this.dotOutput == null) {
                                    Backward.this.dotTransition(pre, mark, t);
                                }
                                if(FINE_DEBUG) { System.out.println("violates bound."); }
                                continue;
                            }
                            
                            if(Backward.this.dotOutput != null) {
                                Backward.this.dotTransition(pre, mark, t);
                            }
                            
                            Node n;
                            if(!exact
                                    && Backward.this.predecessorResult == Result.UNKNOWN
                                    && (n = Backward.this.notCoverable(pre)) != null) {
                                if(FINE_DEBUG) { System.out.println("covers node known to be uncoverable."); }
                                Backward.this.dotCoversNotCoverable(pre, n);
                                return;
                            }
                                                
                            if(Backward.this.stopRequested()) {
                                return;
                            }
                            
                            if(Backward.this.predecessorResult == Result.UNKNOWN
                                    && (n = Backward.this.coversExaminedNode(pre)) != null) {
                                if(FINE_DEBUG) { System.out.println("covers explored node."); }
                                Backward.this.dotCoversExamined(n, pre);
                                continue;
                            }
                            
                            if(Backward.this.predecessorResult == Result.UNKNOWN
                                    && (n = Backward.this.isCoveredByGoodState(pre)) != null) {
                                synchronized(Backward.this.predecessorResultLock) {
                                    if(Backward.this.predecessorResult == Result.UNKNOWN) {
                                        Backward.this.predecessorResult = Result.COVERABLE;
                                        
                                        Backward.this.setMetrics(mark, t, d, pre, 1);
                                        Backward.this.resultPath = pre.path;
                                        Backward.this.resultPathInitial = Node.newNode(pre);
                                    }
                                }
                                Backward.this.dotCoveredByInitial(pre, n);
                                continue;
                            }
                            
                            pre.setTransitionDistance(d.distance);
                            Backward.this.setMetrics(mark, t, d, pre, 1);
                            
                            Backward.this.addPredecessor(Backward.this.markingsToProcess, mark, pre);
                        }
                        
                    } else { // normal transition
                        final Node pre = Node.newNode(mark);
                        pre.accelerated = false;
//                        pre.traceGuess = null;
//                        pre.traceGuessHead = null;
                        final TransitionData d = Backward.this.getTransitionData(t);
                        final Node outNode = Backward.this.getOutNode(t, d);
                        final int mult;
                        if(isPOPlace && outNode.getMarking().size() == 1) {
                            final Place p = outNode.getMarking().get(0).getPlace();
                            if(mark.getMarking(p).getMarking() > 1 && p.getInput().size() == 1) {
                                final Node tmp = Node.newNode(pre);
                                tmp.subtract(Backward.this.initialState);
                                final int tmpMult = tmp.divide(outNode);
                                if(Backward.this.bmcDepth != 0 && mark.path != null && mark.path.getLength() + tmpMult > Backward.this.bmcDepth) {
                                    mult = Backward.this.bmcDepth - mark.path.getLength();
                                } else {
                                    mult = tmpMult;
                                }
                            } else {
                                mult = 1;
                            }
                        } else {
                            mult = 1;
                        }
                        
                        if(mult > 1) {
                            pre.subtract(outNode, mult);
                        } else {
                            pre.subtract(outNode);
                        }
                        
                        final Node inNode = Backward.this.getInNode(t, d);
                        if(Backward.this.predecessorResult == Result.UNKNOWN
                                && Backward.this.boundedPlacesCtr != 0
                                && Backward.this.useBounds) {
                            if(Backward.this.dotOutput == null) {
                                if(mult > 1 ? !pre.addWithBoundsChecking(inNode, mult) : !pre.addWithBoundsChecking(inNode)) {
                                    if(FINE_DEBUG) { System.out.println("violates bound."); }
                                    return;
                                }
                                if(Backward.this.isViolatorNoPlaceBounds(pre)) {
                                    return;
                                }
                            } else {
                                if(mult > 1) {
                                    pre.add(inNode, mult);
                                } else {
                                    pre.add(inNode);
                                }
                                if(Backward.this.isViolator(pre)) {
                                    Backward.this.dotTransition(pre, mark, t, mult);
                                    return;
                                }
                            }
                        } else {
                            if(mult > 1) {
                                pre.add(inNode, mult);
                            } else {
                                pre.add(inNode);
                            }
                        }
                        
                        if(Backward.this.dotOutput != null) {
                            Backward.this.dotTransition(pre, mark, t, mult);
                        }
                        
                        Node n;
                        if(!exact
                                && Backward.this.predecessorResult == Result.UNKNOWN
                                && (n = Backward.this.notCoverable(pre)) != null) {
                            if(FINE_DEBUG) { System.out.println("covers node known to be uncoverable."); }
                            Backward.this.dotCoversNotCoverable(pre, n);
                            return;
                        }
                                            
                        if(Backward.this.stopRequested()) {
                            return;
                        }
                        
                        if(Backward.this.predecessorResult == Result.UNKNOWN
                                && (n = Backward.this.coversExaminedNode(pre)) != null) {
                            if(FINE_DEBUG) { System.out.println("covers explored node."); }
                            Backward.this.dotCoversExamined(n, pre);
                            return;
                        }
                        
                        if(Backward.this.predecessorResult == Result.UNKNOWN
                                && (n = Backward.this.isCoveredByGoodState(pre)) != null) {
                            synchronized(Backward.this.predecessorResultLock) {
                                if(Backward.this.predecessorResult == Result.UNKNOWN) {
                                    Backward.this.predecessorResult = Result.COVERABLE;
                                    
                                    Backward.this.setMetrics(mark, t, d, pre, mult);
                                    Backward.this.resultPath = pre.path;
                                    Backward.this.resultPathInitial = Node.newNode(pre);
                                }
                            }
                            Backward.this.dotCoveredByInitial(pre, n);
                            return;
                        }
                        
                        pre.setTransitionDistance(d.distance);
                        if(Backward.this.verbose && mult > 1) {
                            Backward.this.verbose("accelerated POR transition: " + mark + " == " + mult + "*" + t + " ==> " + pre);
                        }
                        Backward.this.setMetrics(mark, t, d, pre, mult);
                        
                        Backward.this.addPredecessor(Backward.this.markingsToProcess, mark, pre);
                    }
                }
            };
            
            if(Backward.this.predecessorResult == Result.UNKNOWN) {
                this.dispatchRunnableTask(command);
            }
        }
        
        this.awaitNoOpenTasks();
        
//        mark.traceGuess = null;
//        mark.traceGuessHead = null;
        
        return this.predecessorResult;
    }
    
    void setMetrics(final Node post, final Transition t, final TransitionData d, final Node pre, final int mult) {
        // construct trace
        pre.path = new SimpleList<Transition>(t, post.path);
        if(mult > 1) {
            for(int i = mult; i > 1; --i) {
                pre.path = new SimpleList<Transition>(t, pre.path);
            }
        }
//        if(post.traceGuessSatisfied && post.traceGuess == null) {
//            post.traceGuess = post.traceGuessHead;
//        }
//        if(post.traceGuess != null && post.traceGuess.getValue() == t) {
//            pre.traceGuessSatisfied = true;
//            pre.traceGuess = post.traceGuess.getNext();
//            pre.traceGuessHead = post.traceGuessHead;
//        }
    }
    
    
    /**
     * @param queue
     * @param post 
     * @param pre 
     */
    void addPredecessor(final List<Node> queue, final Node post, final Node pre) {
        this.addPredecessor(queue, pre);
    }
    
    void addPredecessor(final List<Node> queue, final Node pre) {
        if(this.useOrdering) {
            Backward.addSortedRandomAccess(queue, pre, this.searchGuide);
        } else {
            queue.add(0, pre);
        }
    }
    
    private static interface NodeClass extends Iterable<Node> {
        public int getTokenSum();
//        public int getMaxTokens();
        public Node getCoveredNode(Node n);
        public Node addNode(Node n, Node acc);
        public int size();
        public boolean isEmpty();
        public void removeViolators();
        public boolean isClassFor(Node n);
    }
    
    class StructuredNodeClass implements NodeClass {
        private final Split top;
        private int minTokenSum;
//        private int maxTokens;
        int size;
        final Collection<Collection<Node>> nodeCollection = new ArrayList<Collection<Node>>();
        
        class Split {
            Place pivot;
            Split with;
            Split without;
            List<Node> nodes;
            
            @Override
            public String toString() {
                return (this.nodes == null ? "" : this.nodes.toString())
                    + (this.nodes != null && this.pivot != null ? "+" : "")
                    + (this.with != null && this.without != null ? "("+ this.pivot.toString() + "=0 ? " + String.valueOf(this.without) + " : " + String.valueOf(this.with) + ")" : "");
            }
        }
        
        public StructuredNodeClass(final Node initial) {
            this.size = 1;
            this.top = new Split();
            this.minTokenSum = initial.getTokens();
//            this.maxTokens = initial.getMaxTokens();
            if(initial.getTokens() == 0) {
                this.top.nodes = new ArrayList<Node>();
                this.top.nodes.add(initial);
                this.nodeCollection.add(this.top.nodes);
            } else {
                Split cur = this.top;
                int depth = 0;
                for(final Iterator<Marking> it = initial.getMarking().iterator(); it.hasNext() && !(NODE_CLASS_TREE_MAX_HEIGHT > 0 && depth+1 >= NODE_CLASS_TREE_MAX_HEIGHT); ) {
                    ++depth;
                    final Marking m = it.next();
                    cur.pivot = m.getPlace();
                    cur.without = new Split();
                    cur.with = new Split();
                    cur = cur.with;
                }
                if(GET_MAX_TREE_HEIGHT && depth+1 > Backward.this.maxTreeHeight) {
                    Backward.this.maxTreeHeight = depth+1;
                }
                cur.nodes = new ArrayList<Node>();
                this.nodeCollection.add(cur.nodes);
                cur.nodes.add(initial);
            }
        }

        @Override
        public int getTokenSum() {
            return this.minTokenSum;
        }
        
//        @Override
//        public int getMaxTokens() {
//            return this.maxTokens;
//        }
        
        @Override
        public int size() {
            return this.size;
        }
        
        @Override
        public boolean isEmpty() {
            return this.size == 0;
        }
        
        @Override
        public boolean isClassFor(final Node n) {
            return true;
        }
        
        @Override
        public void removeViolators() {
            for(final Iterator<Node> it = this.iterator(); it.hasNext(); ) {
                final Node n = it.next();
                if(Backward.this.isViolator(n)) {
                    it.remove();
                }
            }
        }
        
        @Override
        public Iterator<Node> iterator() {
            return new Iterator<Node>() {
                Iterator<Collection<Node>> it = StructuredNodeClass.this.nodeCollection.iterator();
                Iterator<Node> current = null;
                
                @Override
                public boolean hasNext() {
                    if(this.current != null && this.current.hasNext()) {
                        return true;
                    }
                    while(this.it.hasNext()) {
                        this.current = this.it.next().iterator();
                        if(this.current.hasNext()) {
                            return true;
                        }
                    }
                    return false;
                }

                @Override
                public Node next() {
                    return this.current.next();
                }

                @Override
                public void remove() {
                    this.current.remove();
                    --StructuredNodeClass.this.size;
                }
            };
        }
        

        @Override
        public Node getCoveredNode(final Node n) {
            final int ntokens = n.getTokens();
            if(ntokens < this.getTokenSum()) {
                return null;
            }
            
            final boolean noBMC = Backward.this.bmcDepth == 0;
            
            final LinkedList<Split> splits = new LinkedList<Split>();
            splits.add(this.top);
            while(!splits.isEmpty()) {
                final Split cur = splits.remove();
                if(cur.pivot != null) {
                    if(n.getMarking(cur.pivot) != null) {
                        splits.add(cur.with);
                    }
                    splits.add(cur.without);
                }
                if(cur.nodes != null) {
                    for(final Node m : cur.nodes) {
//                      ++checks;
                      if((noBMC || m.path == null || (n.path != null && m.path.getLength() <= n.path.getLength()))
                              && n.greaterEquals(m)) {
                          return m;
                      }
                  }
                }
            }
            return null;
        }
        
        @Override
        public Node addNode(final Node n, final Node acc) {
            final boolean noBMC = Backward.this.bmcDepth == 0;
            final boolean extended = Backward.this.containsExtendedTransition && (n.path != null && n.path.getValue() instanceof ExtendedTransition);
            
            Node theAcc = null;
            final LinkedList<Split> splits = new LinkedList<Split>();
            splits.add(this.top);
            while(!splits.isEmpty()) {
                final Split cur = splits.remove();
                if(cur.pivot != null) {
                    if(n.getMarking(cur.pivot) == null) {
                        splits.add(cur.without);
                    }
                    splits.add(cur.with);
                }
                if(cur.nodes != null) {
                    for(final ListIterator<Node> it = cur.nodes.listIterator(cur.nodes.size()); it.hasPrevious(); ) {
                        final Node m = it.previous();
                        if((noBMC || (m.path != null && n.path.getLength() <= m.path.getLength()))
                                && m.greaterEquals(n)) {
                            if(acc == null && !extended) {
                                theAcc = Backward.this.analysePath(n, m);
                            }
                            // replace by smaller node
                            it.remove();
                            --this.size;
                            if(theAcc == Backward.ACCELERATED_COVERED) {
                                return theAcc;
                            }
                        }
                    }
                }
            }
            final Collection<Place> freeplaces = new IdentityHashSet<Place>();
            for(final Marking m : n.getMarking()) {
                freeplaces.add(m.getPlace());
            }
            
            Split cur = this.top;
            while(cur.pivot != null) {
                if(n.getMarking(cur.pivot) != null) {
                    cur = cur.with;
                } else {
                    cur = cur.without;
                }
            }
            int depth = 0;
            cur = this.top;
            while(cur.pivot != null) {
                ++depth;
                if(freeplaces.remove(cur.pivot)) {
                    cur = cur.with;
                } else {
                    cur = cur.without;
                }
            }
            
            for(final Iterator<Place> it = freeplaces.iterator(); it.hasNext() && !(NODE_CLASS_TREE_MAX_HEIGHT > 0 && depth+1 >= NODE_CLASS_TREE_MAX_HEIGHT); ) {
                ++depth;
                cur.pivot = it.next();
                cur.without = new Split();
                cur.with = new Split();
                cur = cur.with;
            }
            if(GET_MAX_TREE_HEIGHT && depth+1 > Backward.this.maxTreeHeight) {
                Backward.this.maxTreeHeight = depth+1;
            }
            cur.nodes = new ArrayList<Node>();
            this.nodeCollection.add(cur.nodes);
            Backward.addSortedRandomAccess(cur.nodes, n, Backward.TOKEN_COMPARATOR);
            ++this.size;
            if(n.getTokens() < this.minTokenSum) {
                this.minTokenSum = n.getTokens();
            }
//            if(n.getMaxTokens() < this.maxTokens) {
//                this.maxTokens = n.getMaxTokens();
//            }
            return theAcc;
        }
        
        @Override
        public String toString() {
            return String.valueOf(this.top);
        }
        
    }
    
    class EqualSumStructuredNodeClass implements NodeClass {
        private final int tokenSum;
//        private final int maxTokens;
        private final Split top;
        int size;
        final Collection<Collection<Node>> nodeCollection = new ArrayList<Collection<Node>>();
        
        class Split {
            Place pivot;
            Split with;
            Split without;
            Collection<Node> nodes;
            
            @Override
            public String toString() {
                return (this.nodes == null ? "" : this.nodes.toString())
                    + (this.nodes != null && this.pivot != null ? "+" : "")
                    + (this.with != null && this.without != null ? "("+ this.pivot.toString() + "=0 ? " + String.valueOf(this.without) + " : " + String.valueOf(this.with) + ")" : "");
            }
        }
        
        public EqualSumStructuredNodeClass(final Node initial) {
            this.tokenSum = initial.getTokens();
//            this.maxTokens = initial.getMaxTokens();
            this.size = 1;
            this.top = new Split();
            if(this.tokenSum == 0) {
                this.top.nodes = new ArrayList<Node>();
                this.top.nodes.add(initial);
                this.nodeCollection.add(this.top.nodes);
            } else {
                Split cur = this.top;
                int depth = 0;
                for(final Iterator<Marking> it = initial.getMarking().iterator(); it.hasNext() && !(NODE_CLASS_TREE_MAX_HEIGHT > 0 && depth+1 >= NODE_CLASS_TREE_MAX_HEIGHT); ) {
                    ++depth;
                    final Marking m = it.next();
                    cur.pivot = m.getPlace();
                    cur.without = new Split();
                    cur.with = new Split();
                    cur = cur.with;
                }
                if(GET_MAX_TREE_HEIGHT && depth+1 > Backward.this.maxTreeHeight) {
                    Backward.this.maxTreeHeight = depth+1;
                }
                cur.nodes = new ArrayList<Node>();
                this.nodeCollection.add(cur.nodes);
                cur.nodes.add(initial);
            }
        }

        @Override
        public int getTokenSum() {
            return this.tokenSum;
        }
        
//        @Override
//        public int getMaxTokens() {
//            return this.maxTokens;
//        }
        
        @Override
        public int size() {
            return this.size;
        }
        
        @Override
        public boolean isEmpty() {
            return this.size == 0;
        }
        
        @Override
        public boolean isClassFor(final Node n) {
            return this.tokenSum == n.getTokens(); // && this.maxTokens == n.getMaxTokens();
        }
        
        @Override
        public void removeViolators() {
            for(final Iterator<Node> it = this.iterator(); it.hasNext(); ) {
                final Node n = it.next();
                if(Backward.this.isViolator(n)) {
                    it.remove();
                }
            }
        }
        
        @Override
        public Iterator<Node> iterator() {
            return new Iterator<Node>() {
                Iterator<Collection<Node>> it = EqualSumStructuredNodeClass.this.nodeCollection.iterator();
                Iterator<Node> current = null;
                
                @Override
                public boolean hasNext() {
                    if(this.current != null && this.current.hasNext()) {
                        return true;
                    }
                    while(this.it.hasNext()) {
                        this.current = this.it.next().iterator();
                        if(this.current.hasNext()) {
                            return true;
                        }
                    }
                    return false;
                }

                @Override
                public Node next() {
                    return this.current.next();
                }

                @Override
                public void remove() {
                    this.current.remove();
                    --EqualSumStructuredNodeClass.this.size;
                }
            };
        }
        

        @Override
        public Node getCoveredNode(final Node n) {
            final int ntokens = n.getTokens();
            if(ntokens < this.getTokenSum()) {
                return null;
            }
            
            final boolean noBMC = Backward.this.bmcDepth == 0;
            
            if(ntokens == this.tokenSum) {
                Split cur = this.top;
                while(cur.pivot != null) {
                    if(n.getMarking(cur.pivot) != null) {
                        cur = cur.with;
                    } else {
                        cur = cur.without;
                    }
                }
                if(cur.nodes != null) {
                    for(final Node m : cur.nodes) {
    //                    ++checks;
                        if((noBMC || m.path == null || (n.path != null && m.path.getLength() <= n.path.getLength()))
                                && n.equals(m)) {
                            return m;
                        }
                    }
                }
                return null;
            }
            
            // assert ntokens > this.getTokens
            
            final LinkedList<Split> splits = new LinkedList<Split>();
            splits.add(this.top);
            while(!splits.isEmpty()) {
                final Split cur = splits.remove();
                if(cur.pivot != null) {
                    if(n.getMarking(cur.pivot) != null) {
                        splits.add(cur.with);
                    }
                    splits.add(cur.without);
                }
                if(cur.nodes != null) {
                    for(final Node m : cur.nodes) {
//                      ++checks;
                      if((noBMC || m.path == null || (n.path != null && m.path.getLength() <= n.path.getLength()))
                              && n.greaterEquals(m)) {
                          return m;
                      }
                  }
                }
            }
            return null;
        }
        
        @Override
        public Node addNode(final Node n, final Node acc) {
            final int ntokens = n.getTokens();
            if(ntokens > this.tokenSum) {
                return null;
            }
            final boolean noBMC = Backward.this.bmcDepth == 0;
            final boolean extended = Backward.this.containsExtendedTransition && (n.path != null && n.path.getValue() instanceof ExtendedTransition);
            
            if(ntokens < this.getTokenSum()) {
                Node theAcc = null;
//                final Collection<Node> nodes = this.getSplit(n);
//                if(nodes == null) {
//                    return null;
//                }
                final LinkedList<Split> splits = new LinkedList<Split>();
                splits.add(this.top);
                while(!splits.isEmpty()) {
                    final Split cur = splits.remove();
                    if(cur.pivot != null) {
                        if(n.getMarking(cur.pivot) == null) {
                            splits.add(cur.without);
                        }
                        splits.add(cur.with);
                    }
                    if(cur.nodes != null) {
                        for(final Iterator<Node> it = cur.nodes.iterator(); it.hasNext(); ) {
                            final Node m = it.next();
                            if((noBMC || (m.path != null && n.path.getLength() <= m.path.getLength()))
                                    && m.greaterEquals(n)) {
                                if(acc == null && !extended) {
                                    theAcc = Backward.this.analysePath(n, m);
                                }
                                // replace by smaller node
                                it.remove();
                                --this.size;
                                if(theAcc == Backward.ACCELERATED_COVERED) {
                                    return theAcc;
                                }
                            }
                        }
                    }
                }
                return theAcc;
            }
            
            // assert ntokens == this.getTokens
            
            Split cur = this.top;
            while(cur.pivot != null) {
                if(n.getMarking(cur.pivot) != null) {
                    cur = cur.with;
                } else {
                    cur = cur.without;
                }
            }
            if(cur.nodes != null) {
                for(final Iterator<Node> it = cur.nodes.iterator(); it.hasNext(); ) {
                    final Node m = it.next();
                    if(n.equals(m)) {
                        if(!extended && (n.path == null || (m.path != null && m.path.getLength() > n.path.getLength()))) {
                            m.path = n.path;
                        }
                        return m;
                    }
                }
            } else {
                final Collection<Place> freeplaces = new IdentityHashSet<Place>();
                for(final Marking m : n.getMarking()) {
                    freeplaces.add(m.getPlace());
                }
                int depth = 0;
                cur = this.top;
                while(cur.pivot != null) {
                    ++depth;
                    if(freeplaces.remove(cur.pivot)) {
                        cur = cur.with;
                    } else {
                        cur = cur.without;
                    }
                }
                
                for(final Iterator<Place> it = freeplaces.iterator(); it.hasNext() && !(NODE_CLASS_TREE_MAX_HEIGHT > 0 && depth+1 >= NODE_CLASS_TREE_MAX_HEIGHT); ) {
                    ++depth;
                    cur.pivot = it.next();
                    cur.without = new Split();
                    cur.with = new Split();
                    cur = cur.with;
                }
                if(GET_MAX_TREE_HEIGHT && depth+1 > Backward.this.maxTreeHeight) {
                    Backward.this.maxTreeHeight = depth+1;
                }
                cur.nodes = new ArrayList<Node>();
                this.nodeCollection.add(cur.nodes);
            }
            cur.nodes.add(n);
            ++this.size;
            return null;
        }
        
        @Override
        public String toString() {
            return String.valueOf(this.top);
        }
        
    }
    
        
    /**
     * @param pre
     * @return <code>true</code> iff node should not be discovered
     */
    Node coversExaminedNode(final Node pre) {
        return this.coversNode(pre, this.inspectedMarkings);
    }
    
    
    Node coversNode(final Node pre, final List<NodeClass> list) {
        // increasing token sum and max tokens order
        final int tokens = pre.getTokens();
        for(final ListIterator<NodeClass> it = list.listIterator(list.size()); it.hasPrevious(); ) {
            final NodeClass n = it.previous();
            if(tokens >= n.getTokenSum()) {
                final Node result = n.getCoveredNode(pre);
                if(result != null) {
                    return result;
                }
            } else {
                return null;
            }
        }
        return null;
    }
    
    private Node addExaminedNode(final Node pre) {
        return this.addNode(pre, this.inspectedMarkings);
    }
    
    Node addNode(final Node pre, final List<NodeClass> list) {
        return this.addNode(pre, list, true);
    }
    
    private Node addNode(final Node pre, final List<NodeClass> list, final boolean minimal) {
        // decreasing token sum and max tokens order
        final int tokens = pre.getTokens();
        final ListIterator<NodeClass> it = list.listIterator();
        NodeClass addTo = null;
        Node acc = null;
        while(it.hasNext()) {
            final NodeClass n = it.next();
            if(n.isClassFor(pre)) {
                addTo = n;
                final Node result = n.addNode(pre, acc);
                if(result != null) {
                    return result;
                }
            } else if(n.getTokenSum() > tokens) {
                if(!minimal) {
                    break;
                }
                final Node result = n.addNode(pre, acc);
                if(result == Backward.ACCELERATED_COVERED) {
                    return result;
                }
                if(acc == null) {
                    acc = result;
                }
            } else {
                // one step back
                it.previous();
                break;
            }
        }
        if(addTo == null) {
            addTo = this.newNodeClass(pre);
            it.add(addTo);
        }
        return acc;
    }
    
    private NodeClass newNodeClass(final Node n) {
        return new EqualSumStructuredNodeClass(n);
//        return new StructuredNodeClass(n);
    }
    
    Node analysePath(final Node less, final Node greater) {
//        if(less.traceGuessSatisfied && less.traceGuessHead != null) {
//            // there already exists a guess for less
//            return null;
//        }
        if(this.bmcDepth != 0 || !this.useAcceleration) {
            // do not accelerate during bounded mc
            return null;
        }
        if(less.getMarking().size() != greater.getMarking().size()) {
            // some places in less aren't marked that were marked in greater
            // path will probably not be backward enabled
            return null;
        }
        if(less.path == null || (greater.path != null && less.path.getLength() <= greater.path.getLength())) {
            // less cannot be in the backward reachable set of greater
            return null;
        }
        // if less has path <- l1 <- l2 <- l3 <- l4 <- l5 <- l6 <- l7 and
        // greater has path <- g1 <- g2 <- g3 <- g4 <- g5
        // then l1 -> l2 is the guess and l3 = g1, l4 = g2, l5 = g3, l6 = g4, l7 = g5 has to be satisfied
        
        final Node effect = Node.newNode(greater);
        effect.subtract(less);
        if(!less.greaterEquals(effect)) {
            // there may be a path from greater to less, but it cannot be accelerated
            return null;
        }
        
        SimpleList<Transition> lessTrace = less.path;
        SimpleList<Transition> guess = null;
//        boolean normalTransition = false;
        for(int i = less.path.getLength() - (greater.path == null ? 0 : greater.path.getLength()); i > 0; --i) {
//            if(!normalTransition && !lessTrace.getValue().getInput().isEmpty()) {
//                normalTransition = true;
//            }
            final Transition t = lessTrace.getValue();
            if(this.containsExtendedTransition && t instanceof ExtendedTransition) {
                return null;
            }
            guess = new SimpleList<Transition>(t, guess);
            lessTrace = lessTrace.getNext();
        }
        
//        if(!normalTransition) {
//            return null;
//        }
        SimpleList<Transition> greaterTrace = greater.path;
        while(lessTrace != null) {
            if(lessTrace.getValue() != greaterTrace.getValue()) {
                return null;
            }
            lessTrace = lessTrace.getNext();
            greaterTrace = greaterTrace.getNext();
        }
        assert guess != null;
        if(this.bmcDepth != 0 && (less.path == null ? 0 : less.path.getLength()) + guess.getLength() > this.bmcDepth) {
            return null;
        }
//        less.traceGuess = guess;
//        less.traceGuessHead = guess;
        if(this.verbose) {
            this.verbose("learned path: " + guess + " from " + greater + " to " + less);
        }
        final Node accelerated = this.accelerate(less, greater, effect, guess);
        if(accelerated != null && this.verbose) {
            this.verbose("... accelerated: " + accelerated);
        }
        return accelerated;
    }
    
    private Node accelerate(final Node less, final Node greater, final Node effect, final SimpleList<Transition> trace) {
        final Node maxEff = Node.newNode(less);
        int loops = 0;
        
        final Node accelerated = Node.newNode(less);
        for(final Transition t : trace) {
            final TransitionData d = this.getTransitionData(t);
            final Node tOut = this.getOutNode(t, d);
            final Node tIn = this.getInNode(t, d);
            accelerated.subtract(tOut);
            maxEff.minimize(accelerated);
            accelerated.add(tIn);
            accelerated.path = new SimpleList<Transition>(t, accelerated.path);
        }
        if(!less.greaterEquals(accelerated)) {
            return null;
        }
        ++loops;
        
        final Node req = Node.newNode(less);
        req.subtract(maxEff);
        if(req.getTokens() != 0) {
            int mult = accelerated.divide(req);
            if(this.bmcDepth != 0 && mult*trace.getLength()+accelerated.path.getLength() > this.bmcDepth) {
                mult = Math.max(0, (this.bmcDepth-accelerated.path.getLength())/trace.getLength());
            }
            if(mult > 0) {
                accelerated.subtract(effect, mult);
                loops += mult;
                for(int i = 0; i < mult; ++i) {
                    for(final Transition t : trace) {
                        accelerated.path = new SimpleList<Transition>(t, accelerated.path);
                    }
                }
            }
        }
        
        this.dotTransition(accelerated, less, trace, loops);
        
        final Node n;
        if((n = Backward.this.isCoveredByGoodState(accelerated)) != null) {
            this.resultPath = accelerated.path;
            this.resultPathInitial = Node.newNode(accelerated);
            Backward.this.dotCoveredByInitial(accelerated, n);
            return Backward.ACCELERATED_COVERED;
        }
        
//        acc.traceGuess = trace;
//        acc.traceGuessHead = trace;
        accelerated.accelerated = true;
//        less.accelerated = true;
        
        accelerated.setTransitionDistance(greater.getTransitionDistance());
        this.addPredecessor(this.markingsToProcess, accelerated);
        return accelerated;
    }
    
    /**
     * @param mark
     * @return
     */
    Node isCoveredByGoodState(final Node mark) {
        if(this.containsParameters) {
            return this.initialState.greaterEqualsIgnoreParametricPlaces(mark) ? this.initialState : null;
        }
        return this.initialState.greaterEquals(mark) ? this.initialState : null;
    }
   

    /* (non-Javadoc)
     * @see petruchio.cov.AbstractMultiThreadedCoverability#doClearCaches()
     */
    @Override
    void doClearCaches() {
        this.transitionDataLock.lock();
        try {
            while(this.transitionDataReaders != 0) {
                this.transitionDataIsBeingRead.await();
            }
            this.transitionData.clear();
        } catch(final InterruptedException e) {
            Thread.currentThread().interrupt();
        } finally {
            this.transitionDataLock.unlock();
        }
        if(this.friendlyTransitions != null) {
            for(final petruchio.pn.Transition t : this.friendlyTransitions) {
                t.data = null;
            }
            this.friendlyTransitions = null;
        }
        
        this.placeDataLock.lock();
        try {
            while(this.placeDataReaders != 0) {
                this.placeDataIsBeingRead.await();
            }
            for(final Map.Entry<Place, PlaceData> e : this.placeData.entrySet()) {
                if(!e.getValue().givenBound && e.getKey().getBound() == Place.BOUND_UNBOUNDED) {
                    e.getKey().setBound(Place.BOUND_UNKNOWN);
                }
            }
            this.placeData.clear();
        } catch(final InterruptedException e) {
            Thread.currentThread().interrupt();
        } finally {
            this.placeDataLock.unlock();
        }
        
        if(this.friendlyPlaces != null) {
            for(final petruchio.pn.Place p : this.friendlyPlaces) {
                p.data = null;
            }
            this.friendlyPlaces = null;
        }
        
        if(DEBUG && this.queries > 0) {
            System.out.println(this.queries + " queries. " + this.queries_coverable + " coverable.");
        }
        this.checked.clear();
        this.doNotCheck.clear();
        this.notCov = new LinkedList<NodeClass>();
        this.queries = 0;
        this.queries_coverable = 0;
        this.newPlacesToProcess.clear();
        this.clearNodes();
        synchronized(this.markingsToProcess) {
            this.markingsToProcess.clear();
        }
        this.containsExtendedTransition = false;
        this.updating.clear();
        this.boundedPlacesCtr = 0;
        this.transitionCount = 0;
        this.maxTransOutBranching = 0;
        this.invComputedAt = 0;
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#compute(java.util.Collection, java.util.Collection)
     */
    @Override
    public synchronized void compute(final Collection<Place> root, final Collection<Transition> transitions) throws UnsupportedOperationException {
        this.clearCaches();
        if(this.useBounds) {
            for(final Transition t : transitions) {
                if(!this.containsExtendedTransition && t instanceof ExtendedTransition) {
                    this.containsExtendedTransition = true;
                }
                for(final PTArc in : t.getInput()) {
                    if(in.getSource().getBound() != Place.BOUND_UNKNOWN) {
                        final PlaceData d = this.getPlaceData(in.getSource());
                        if(!d.givenBound) {
                            if(in.getSource().boundKnown()) {
                                ++this.boundedPlacesCtr;
                            }
                            d.givenBound = true;
                        }
                    }
                }
                for(final TPArc out : t.getOutput()) {
                    if(out.getTarget().getBound() != Place.BOUND_UNKNOWN) {
                        final PlaceData d = this.getPlaceData(out.getTarget());
                        if(!d.givenBound) {
                            if(out.getTarget().boundKnown()) {
                                ++this.boundedPlacesCtr;
                            }
                            d.givenBound = true;
                        }
                    }
                }
            }
        } else {
            for(final Transition t : transitions) {
                if(t instanceof ExtendedTransition) {
                    this.containsExtendedTransition = true;
                    break;
                }
            }
        }
        this.setStartState(root);
        this.analyzeDependencies(root);
//        this.computeSCCs(root);
        if(this.useBounds) {
            for(final Place p : root) {
                for(final PTArc a : p.getOutput()) {
                    this.updating.add(a.getTarget());
                }
            }
            for(final Transition t : transitions) {
                if(t.getInput().isEmpty()) {
                    this.updating.add(t);
                }
            }
            this.computeBounds();
        }
        if(USE_DISTANCE_HEURISTICS) {
            for(final Place p : root) {
                final PlaceData pd = this.getPlaceData(p);
                pd.distance = 0;
                pd.fixed = true;
                for(final PTArc a : p.getOutput()) {
                    this.updating.add(a.getTarget());
                    if(pd.fixed) {
                        pd.fixed = false;
                        if(!(a.getTarget() instanceof ExtendedTransition)) {
                            for(final TPArc b : a.getTarget().getOutput()) {
                                if(b.getTarget() == p && b.getWeight() >= a.getWeight()) {
                                    pd.fixed = true;
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            for(final Transition t : transitions) {
                if(t.getInput().isEmpty()) {
                    this.getTransitionData(t).distance = MIN_TRANSITION_DISTANCE;
                }
            }
            this.computeDistance();
        }
        this.invComputedAt = Integer.MAX_VALUE;
    }
    
    /**
     * @param init
     */
    private void analyzeDependencies(final Collection<Place> init) {
        // nothing?
    }
    
//    /**
//     * @param root
//     */
//    private void computeSCCs(final Collection<Place> init) {
//        final Collection<Collection<NetElement>> sccs = Tarjan.tarjan(init.iterator().next());
////        System.out.println(sccs.size());
//        
//        for(final Collection<NetElement> scc : sccs) {
//            boolean generator = false;
//            for(final NetElement e : scc) {
//                if(e instanceof Transition) {
//                    if(this.getChangeSum((Transition)e) > 0) {
//                        generator = true;
//                        break;
//                    }
//                }
//            }
//            if(generator) {
//                for(final NetElement e : scc) {
//                    if(e instanceof Transition) {
//                        this.setOnGeneratingCycle((Transition)e, true);
//                    }
//                }
//            }
//        }
//    }
    
    
    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#getBounds(java.util.Collection)
     */
    @Override
    public synchronized Map<Place, Integer> getBounds(final Collection<Place> places) throws UnsupportedOperationException {
        final Map<Place, Integer> map = new IdentityHashMap<Place, Integer>();
        for(final Place p : places) {
            if(p.boundKnown()) {
                map.put(p, Integer.valueOf(p.getBound()));
            } else {
                map.put(p, Integer.valueOf(Place.BOUND_UNKNOWN));
            }
        }
        return map;
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#getDeadlocks(boolean)
     */
    @Override
    public synchronized Collection<String> getDeadlocks(final boolean assumeBoundedness) throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#getFiringBounds(java.util.Collection)
     */
    @Override
    public synchronized Map<Transition, Integer> getFiringBounds(final Collection<Transition> transitions) {
        final Map<Transition, Integer> bounds = new IdentityHashMap<Transition, Integer>(transitions.size());
        for(final Transition t : transitions) {
            int hi = this.getFiringBound(t).intValue();
            final Node target = Node.newNode();
            if(hi > 0) {
                do {
                    hi <<= 1;
                    target.clear();
                    for(final PTArc in : t.getInput()) {
                        target.add(in.getSource(), in.getWeight()*hi);
                    }
                } while(this.isCoverable(target, true));
            }
            
            int lo = 1;
            while(lo < hi) {
                final int mid = (lo+hi)/2;
                
                target.clear();
                for(final PTArc in : t.getInput()) {
                    target.add(in.getSource(), in.getWeight()*mid);
                }
                
                if(this.isCoverable(target, true)) {
                    lo = mid+1;
                } else {
                    hi = mid-1;
                }
            }
            bounds.put(t, Integer.valueOf(hi));
        }
        return bounds;
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#keepDeadElements(java.util.Collection, java.util.Collection)
     */
    @Override
    public synchronized void keepDeadElements(final Collection<Place> p, final Collection<Transition> t) throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#markUnused(petruchio.interfaces.petrinet.Place)
     */
    @Override
    public synchronized void markUnused(final Place p) throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#setCheckMarkable(boolean)
     */
    @Override
    public void setCheckMarkable(final boolean b) {
        // not applicable
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#setMaxTokenCount(int)
     */
    @Override
    public synchronized void setMaxTokenCount(final int tokens) throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#setMinTokenCount(int)
     */
    @Override
    public synchronized void setMinTokenCount(final int tokens) throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#doSetStartState(java.util.Collection)
     */
    @Override
    public synchronized void doSetStartState(final Collection<Place> state) {
        this.initialState.clear();
        this.containsParameters = false;
//        this.addGoodState(n);
        final Collection<PlacePair> seenPairs = new HashSet<PlacePair>();
        for(final Place p1 : state) {
            final PlaceData dataP1 = this.getPlaceData(p1);
            dataP1.distance = 0;
            dataP1.fixed = true;
            for(final PTArc a : p1.getOutput()) {
                this.updating.add(a.getTarget());
                if(dataP1.fixed) {
                    dataP1.fixed = false;
                    for(final TPArc b : a.getTarget().getOutput()) {
                        if(b.getTarget() == p1 && b.getWeight() >= a.getWeight()) {
                            dataP1.fixed = true;
                            break;
                        }
                    }
                }
            }
            
            if(p1.isParametric()) {
                this.containsParameters = true;
                p1.setMarking(0); // remove tokens as the place is parametric.
            } else {
                this.initialState.add(p1, p1.getMarking());
            }
            if(this.useBounds && !dataP1.givenBound && !this.containsExtendedTransition) { // TODO
                if(p1.isParametric()) {
                    continue;
                }
//                boolean noExtendedInput = true;
//                for(final TPArc a : p1.getInput()) {
//                    if(a.getSource() instanceof ExtendedTransition) {
//                        noExtendedInput = false;
//                        break;
//                    }
//                }
//                if(noExtendedInput) {
                    p1.setBound(p1.getMarking());
                    ++this.boundedPlacesCtr;
//                }
            }
            
            
            if(this.usedByPetruchioCompiler) {
                final Collection<Place> partners = this.getCommunicators(p1);
    //            System.out.println(p1 + ":" + p1.getMarking() + " " + partners);
                if(!partners.isEmpty() && !this.stopRequested()) {
                    for(final Place p2 : partners) {
                        if((p1 != p2) || (p1.getMarking() >= 2)) {
                            if(dataP1.commPartners == null) {
                                dataP1.commPartners = new IdentityHashSet<Place>();
                            }
                            dataP1.commPartners.add(p2);
                            
                            final PlacePair pair = new petruchio.cov.PlacePair(p1, p2);
                            if(seenPairs.add(pair)) {
                                this.addPlacePair(pair);
                            }
                        }
                    }
                }
            }
        }
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.algorithms.Reachability#size()
     */
    @Override
    public int size() {
        return this.inspectedMarkings.size(); // not applicable
    }

    /* (non-Javadoc)
     * @see petruchio.interfaces.Resettable#reset()
     */
    @Override
    void doReset() {
        this.clearCaches();
        this.invComputedAt = 0;
        this.clusterBounds.clear();
        this.invariants.clear();
    }

    /**
     * @param t
     * @return
     */
    Node getInNode(final Transition t) {
        return this.getInNode(t, this.getTransitionData(t));
    }
    
    Node getInNode(final Transition t, final TransitionData d) {
        synchronized(t) {
            Node inNode = (d.inNode == null) ? null : d.inNode.get();
            if(inNode == null) {
                // create pre set node of transition
                inNode = Node.newNode();
                for(final PTArc arc : t.getInput()) {
                    inNode.add(arc.getSource(), arc.getWeight());
                }
                d.inNode = new SoftReference<Node>(inNode);
            }
            return inNode;
        }
    }
    
    
    /**
     * @param t
     * @return
     */
    Node getOutNode(final Transition t) {
        return this.getOutNode(t, this.getTransitionData(t));
    }
    
    Node getOutNode(final Transition t, final TransitionData d) {
        synchronized(t) {
            Node outNode = (d.outNode == null) ? null : d.outNode.get();
            if(outNode == null) {
                // create post set node of transition
                outNode = Node.newNode();
                for(final TPArc arc : t.getOutput()) {
                    outNode.add(arc.getTarget(), arc.getWeight());
                }
                d.outNode = new SoftReference<Node>(outNode);
            }
            return outNode;
        }
    }

    static class PlaceData {
        boolean added = false;
        boolean seen = false;
        boolean givenBound = false;
        boolean fixed = false;
        int index = -1;
        int distance = -1;
        Collection<Place> commPartners = null;
    }
    
    static class TransitionData {
        SoftReference<Node> inNode = null;
        SoftReference<Node> outNode = null;
        SoftReference<Map<Place, Integer>> change = null;
        SoftReference<Pair<Boolean, SimpleList<Pair<Place, Map<Place, Integer>>>>> multipliers = null;
        SoftReference<SimpleList<Place>> transferTargets = null;
        SoftReference<SimpleList<Place>> resetSources = null;
        SoftReference<SimpleList<InhibitorArc>> inhibitors = null;
        int firingBound = UNKNOWN_FIRING_BOUND;
        int index = -1;
        int distance = -1;
        boolean onGeneratingCycle = false;
        final ThreadLocal<Boolean> seen = new ThreadLocal<Boolean>() {
            @Override
            protected Boolean initialValue() {
                return Boolean.FALSE;
            }
        };
    }
    
    private boolean seen(final Transition t) {
        final TransitionData d = this.transitionData.get(t);
        return d == null ? false : d.seen.get().booleanValue();
    }
    
    private boolean addPlace(final Place p) {
        final PlaceData d = this.getPlaceData(p);
        if(d.added) {
            return false;
        }
        d.added = true;
        return true;
    }
    
    private PlaceData getPlaceData(final Place p) {
        if(p instanceof petruchio.pn.Place) {
            PlaceData d = (PlaceData)((petruchio.pn.Place)p).data;
            if(d == null) {
                synchronized(p) {
                    d = (PlaceData)((petruchio.pn.Place)p).data;
                    if(d == null) {
                        d = new PlaceData();
                        ((petruchio.pn.Place)p).data = d;
                        this.friendlyPlaces = new SimpleList<petruchio.pn.Place>((petruchio.pn.Place)p, this.friendlyPlaces);
                    }
                }
            }
            return d;
        }
        final Lock l = this.placeDataLock;
        l.lock();
        PlaceData d = null;
        try {
            d = this.placeData.get(p);
            if(d == null) {
                while(this.placeDataReaders != 0) {
                    this.placeDataIsBeingRead.await();
                }
                d = new PlaceData();
                this.placeData.put(p, d);
            }
        } catch(final InterruptedException e) {
            Thread.currentThread().interrupt();
        } finally {
            l.unlock();
        }
        return d;
    }
    
    private boolean hasGivenBound(final Place p) {
        final PlaceData d;
        if(p instanceof petruchio.pn.Place) {
            d = (PlaceData)((petruchio.pn.Place)p).data;
        } else {
            this.startReadingPlaceData();
            d = this.placeData.get(p);
            this.finishReadingPlaceData();
        }
        return d == null ? false : d.givenBound;
    }
    
//    private void setOnGeneratingCycle(final Transition p, final boolean value) {
//        this.getTransitionData(p).onGeneratingCycle = value;
//    }
    
    private int getChange(final Transition t, final Place p) {
        final TransitionData d = this.getTransitionData(t);
        Map<Place, Integer> m = (d.change == null) ? null : d.change.get();
        if(m == null) {
            m = new IdentityHashMap<Place, Integer>(t.getInput().size() + t.getOutput().size());
            int c = 0;
            for(final PTArc in : t.getInput()) {
                m.put(p, Integer.valueOf(-in.getWeight()));
                if(in.getSource() == p) {
                    c -= in.getWeight();
                }
            }
            for(final TPArc out : t.getOutput()) {
                final Integer old = m.put(p, Integer.valueOf(out.getWeight()));
                if(old != null) {
                    m.put(p, Integer.valueOf(old.intValue()+out.getWeight()));
                }
                if(out.getTarget() == p) {
                    c += out.getWeight();
                }
            }
            d.change = new SoftReference<Map<Place, Integer>>(m);
            return c;
        }
        final Integer c = m.get(p);
        return (c == null) ? 0 : c.intValue();
    }
    
//    private int getChangeSum(final Transition t) {
//        return this.getOutNode(t).getTokens() - this.getInNode(t).getTokens();
//    }
    
    private boolean addSeen(final Transition t) {
        final TransitionData d = this.getTransitionData(t);
        if(d.seen.get().booleanValue()) {
            return false;
        }
        d.seen.set(Boolean.TRUE);
        return true;
    }
    
    private boolean addSeen(final Place p) {
        final PlaceData d = this.getPlaceData(p);
        if(d.seen) {
            return false;
        }
        d.seen = true;
        return true;
    }
    
    private void startReadingTransitionData() {
        this.transitionDataLock.lock();
        try {
            ++this.transitionDataReaders;
        } finally {
            this.transitionDataLock.unlock();
        }
    }
    
    private void finishReadingTransitionData() {
        this.transitionDataLock.lock();
        try {
            --this.transitionDataReaders;
            this.transitionDataIsBeingRead.signalAll();
        } finally {
            this.transitionDataLock.unlock();
        }
    }
    
    private void clearSeenTransitions() {
    	this.startReadingTransitionData();
    	for(final TransitionData d : this.transitionData.values()) {
            d.seen.set(Boolean.FALSE);
        }
    	this.finishReadingTransitionData();
    	if(this.friendlyTransitions != null) {
        	for(final petruchio.pn.Transition t : this.friendlyTransitions) {
        	    final TransitionData data = (TransitionData)t.data;
        	    if(data != null) {
        	        data.seen.set(Boolean.FALSE);
        	    }
            }
    	}
    }
    
    private void startReadingPlaceData() {
        this.placeDataLock.lock();
        try {
            ++this.placeDataReaders;
        } finally {
            this.placeDataLock.unlock();
        }
    }
    
    private void finishReadingPlaceData() {
        this.placeDataLock.lock();
        try {
            --this.placeDataReaders;
            this.placeDataIsBeingRead.signalAll();
        } finally {
            this.placeDataLock.unlock();
        }
    }
   
    private void clearSeenPlaces() {
        this.startReadingPlaceData();
		for(final PlaceData d : this.placeData.values()) {
			d.seen = false;
		}
		this.finishReadingPlaceData();
		if(this.friendlyPlaces != null) {
    		for(final petruchio.pn.Place p : this.friendlyPlaces) {
    		    final PlaceData data = (PlaceData)p.data;
    		    if(data != null) {
    		        data.seen = false;
    		    }
    		}
		}
    }
    
    TransitionData getTransitionData(final Transition t) {
        if(t instanceof petruchio.pn.Transition) {
            TransitionData d = (TransitionData)((petruchio.pn.Transition)t).data;
            if(d == null) {
                synchronized(t) {
                    d = (TransitionData)((petruchio.pn.Transition)t).data;
                    if(d == null) {
                        d = new TransitionData();
                        ((petruchio.pn.Transition)t).data = d;
                    }
                    this.friendlyTransitions = new SimpleList<petruchio.pn.Transition>((petruchio.pn.Transition)t, this.friendlyTransitions);
                }
            }
            return d;
        }
        final Lock l = this.transitionDataLock;
        l.lock();
        TransitionData d = null;
        try {
            d = this.transitionData.get(t);
            if(d == null) {
                while(this.transitionDataReaders != 0) {
                    this.transitionDataIsBeingRead.await();
                }
                d = new TransitionData();
                this.transitionData.put(t, d);
            }
        } catch(final InterruptedException e) {
            Thread.currentThread().interrupt();
        } finally {
            l.unlock();
        }
        return d;
    }
    
    private Integer getFiringBound(final Transition t) {
        final TransitionData d;
        if(t instanceof petruchio.pn.Transition) {
            d = (TransitionData)((petruchio.pn.Transition)t).data;
        } else {
            this.startReadingTransitionData();
            d = this.transitionData.get(t);
            this.finishReadingTransitionData();
        }
        return (d == null) ? Integer.valueOf(UNKNOWN_FIRING_BOUND) : Integer.valueOf(d.firingBound);
    }
    
    @SuppressWarnings("unused")
    private void setFiringBound(final Transition t, Integer fb) {
        final Lock l = this.transitionDataLock;
        l.lock();
        try {
            TransitionData d = this.transitionData.get(t);
            if(d == null) {
                while(this.transitionDataReaders != 0) {
                    this.transitionDataIsBeingRead.await();
                }
                d = new TransitionData();
                this.transitionData.put(t, d);
            }
            d.firingBound = fb;
        } catch(final InterruptedException e) {
            Thread.currentThread().interrupt();
        } finally {
            l.unlock();
        }
    }
    
    /**
     * @param invs
     * @return
     */
    public int adjustBoundsByPlaceInvariants(final Collection<Map<Place, Integer>> invs) {
        if(invs.isEmpty()) {
            return 0;
        }
        final Collection<Map<Place, Integer>> todo = new ArrayList<Map<Place, Integer>>(invs);
        final Map<Transition, Integer> transitionBounds = new IdentityHashMap<Transition, Integer>();
        int change = 0;
        boolean updated;
        do {
            updated = false;
            for(final Iterator<Map<Place, Integer>> it = todo.iterator(); it.hasNext(); ) {
                final Map<Place, Integer> inv = it.next();
                boolean poscovered = true;
                boolean negcovered = true;
                boolean containsNeg = false;
                for(final Map.Entry<Place, Integer> e : inv.entrySet()) {
                    final Place p = e.getKey();
                    final int v = e.getValue().intValue();
                    containsNeg |= v < 0;
                    if(!p.boundKnown()) {
                        if(v < 0) {
                            negcovered = false;
                            if(!poscovered) {
                                break;
                            }
                        } else if(v != 0) {
                            poscovered = false;
                            if(!negcovered) {
                                break;
                            }
                        }
                    }
                }
                if(poscovered || negcovered) {
                    final Collection<Place> boundImproved = this.adjustBoundsByPlaceInvariant(inv, containsNeg && !poscovered);
                    if(!boundImproved.isEmpty()) {
                        change += boundImproved.size();
                        change += this.propagateBounds(boundImproved, transitionBounds);
                    }
                    it.remove();
                    updated = true;
                }
            }
        } while(updated);
        return change;
    }
    
    /**
     * @param inv
     * @param invert
     * @return
     */
    public Collection<Place> adjustBoundsByPlaceInvariant(final Map<Place, Integer> inv, final boolean invert) {
        long sum = 0;
        long negativeCoefficientSub = 0;
        for(final Map.Entry<Place, Integer> e : inv.entrySet()) {
            final Place p = e.getKey();
            if(p.isParametric()) {
                continue;
            }
            final int v = invert ? -e.getValue().intValue() : e.getValue().intValue();
            sum = checkedADD(sum, checkedMUL(v, p.getMarking()));
            if(v < 0) {
                if(p.boundKnown()) {
                    negativeCoefficientSub = checkedADD(negativeCoefficientSub, checkedMUL(-v, p.getBound()));
                } else {
                    return Collections.emptyList();
                }
            }
        }
        if(sum < 0) {
            return Collections.emptyList();
        }
        final long total = checkedADD(sum, negativeCoefficientSub);
        final Collection<Place> boundImproved = new IdentityHashSet<Place>();
        for(final Map.Entry<Place, Integer> e : inv.entrySet()) {
            final Place p = e.getKey();
            if(p.isParametric()) {
                continue;
            }
            final long v = invert ? -e.getValue().intValue() : e.getValue().intValue();
            final long b;
            if(v > 0) {
                b = total / v;
            } else {
                b = checkedADD(-total, checkedMUL(v, p.getBound()))/v; // floor
            }
            if(b <= Integer.MAX_VALUE) {
                final int ib = (int)b;
                if(!p.boundKnown() && ib != Place.BOUND_UNBOUNDED) {
                    p.setBound(ib);
                    boundImproved.add(p);
                    ++this.boundedPlacesCtr;
                } else if(p.getBound() > ib) {
                    p.setBound(ib);
                    boundImproved.add(p);
                }
            }
        }
        return boundImproved;
    }
    
    
    private int propagateBounds(final Collection<Place> someBoundedPlaces,
            final Map<Transition, Integer> transitionBounds) {
//        if(true) return 0; // TODO
        int change = 0;
        final LinkedList<Place> todo = new LinkedList<Place>(someBoundedPlaces);
        final LinkedList<Transition> checkOutput = new LinkedList<Transition>();
        while(!todo.isEmpty()) {
            final Place p = todo.remove();
            if(p.isParametric()) {
                continue;
            }
            for(final TPArc a : p.getInput()) {
                final Transition t = a.getSource();
                
                if(a instanceof TransferArc) {
                    continue;
                }
                boolean takesTooMuch = false;
                for(final PTArc b : p.getOutput()) {
                    if(b.getTarget() == a.getSource() && b.getWeight() >= a.getWeight()) {
                        takesTooMuch = true;
                        break;
                    }
                }
                if(takesTooMuch) {
                    continue;
                }
                
                final int transitionFiringBound;
                final int newTransitionFiringBound = p.getBound() / a.getWeight();
                final Integer oldTransitionFiringBound = transitionBounds.put(t, Integer.valueOf(newTransitionFiringBound));
                if(oldTransitionFiringBound == null || newTransitionFiringBound < oldTransitionFiringBound.intValue()) {
                    transitionFiringBound = newTransitionFiringBound;
                    checkOutput.add(t);
                } else {
                    transitionBounds.put(t, oldTransitionFiringBound);
                    transitionFiringBound = oldTransitionFiringBound.intValue();
                }
                
                int inputsize = t.getInput().size();
                if(a instanceof TransferArc) {
                    boolean containsinh = false;
                    for(final PTArc b : t.getInput()) {
                        if(b instanceof ResetArc) {
                            --inputsize;
                        } else if(b instanceof InhibitorArc) {
                            containsinh = true;
                            break;
                        }
                    }
                    if(containsinh) {
                        continue;
                    }
                }
                
                if(inputsize == 1) {
                    for(final PTArc b : t.getInput()) {
                        if(b instanceof ResetArc) {
                            continue;
                        }
                        final int bound = (transitionFiringBound+1) * b.getWeight() - 1;
                        final Place p2 = b.getSource();
                        if(!p2.boundKnown()) {
                            p2.setBound(bound);
                            todo.add(p2);
                            ++change;
                            ++this.boundedPlacesCtr;
                        } else if(p2.getBound() > bound) {
                            p2.setBound(bound);
                            todo.add(p2);
                            ++change;
                        }
                    }
                }
            }
            
            out_arcs:
            for(final PTArc a : p.getOutput()) {
                final Transition t = a.getTarget();
                final boolean extended = t instanceof ExtendedTransition;
                if(extended && a instanceof InhibitorArc) {
                    continue;
                }
                int putBack = 0;
                for(final TPArc b : t.getOutput()) {
                    if(b.getTarget() == p) {
                        if(extended && b instanceof TransferArc) {
                            continue out_arcs;
                        }
                        putBack = b.getWeight();
                        break;
                    }
                }
                if(putBack >= a.getWeight()) {
                    break;
                }
                final int newTransitionFiringBound = firingBound(p.getBound(), a.getWeight(), putBack);
                final Integer oldTransitionFiringBound = transitionBounds.put(t, Integer.valueOf(newTransitionFiringBound));
                if(oldTransitionFiringBound == null || newTransitionFiringBound < oldTransitionFiringBound.intValue()) {
                    checkOutput.add(t);
                } else {
                    transitionBounds.put(t, oldTransitionFiringBound);
                }
            }
            
            while(!checkOutput.isEmpty()) {
                final Transition t = checkOutput.remove();
                final boolean extended = t instanceof ExtendedTransition;
                final int transitionFiringBound = transitionBounds.get(t).intValue();
                for(final TPArc b : t.getOutput()) {
                    final Place p2 = b.getTarget();
                    if(p2.isParametric() || p2 == p || p2.getInput().size() != 1) {
                        continue;
                    }
                    int transferb = 0;
                    if(extended && b instanceof TransferArc) {
                        for(final Map.Entry<Place, Integer> e : ((TransferArc)b).getTransfer().entrySet()) {
                            final Place p3 = e.getKey();
                            if(!p3.boundKnown()) {
                                transferb = p3.getBound();
                                break;
                            }
                            transferb += p3.getBound()*e.getValue().intValue();
                        }
                    }
                    if(transferb >= 0) {
                        final int bound = transferb + transitionFiringBound * b.getWeight();
                        if(!p2.boundKnown()) {
                            p2.setBound(bound);
                            todo.add(p2);
                            ++change;
                            ++this.boundedPlacesCtr;
                        } else {
                            p2.setBound(p2.getBound() + bound);
                            todo.add(p2);
                            ++change;
                        }
                    }
                }
            }
        }
        return change;
    }
    
    
    private void propagateNonBounds(final Collection<Place> unboundedPlaces) {
//        if(true) return 0; // TODO
        final LinkedList<Place> todo = new LinkedList<Place>(unboundedPlaces);
        while(!todo.isEmpty()) {
            final Place p = todo.remove();
            
            for(final PTArc a : p.getOutput()) {
                final Transition t = a.getTarget();
                final boolean extended = t instanceof ExtendedTransition;
                if(extended && (a instanceof ResetArc || a instanceof InhibitorArc)) {
                    continue;
                }
                boolean unboundedFiring = true;
                for(final PTArc b : t.getInput()) {
                    if(b.getSource().getBound() != Place.BOUND_UNBOUNDED) {
                        unboundedFiring = false;
                        break;
                    }
                }
                if(unboundedFiring) {
                    for(final TPArc b : t.getOutput()) {
                        if(b.getTarget().getBound() != Place.BOUND_UNBOUNDED) {
                            b.getTarget().setBound(Place.BOUND_UNBOUNDED);
                            todo.add(b.getTarget());
                        }
                    }
                }
            }
        }
    }
    
    
    private static int firingBound(final int m, final int in, final int out) {
        
        return (m-out)/(in-out);
        
//        if(out == 0) {
//            return m/in;
//        }
//        
////        return m/(in-out); // not tight
//        
//        int c = m;
//        int f = 0;
//        while(c >= in) {
//            ++f;
//            c = c-in+out;
//        }
//        return f;
//        
//        // TODO: direct computation?
    }
    
    public int getBoundedPlaces() {
        return this.boundedPlacesCtr;
    }
    
    
    public Collection<Collection<Place>> findSiphonTraps(final PetriNet net, final Collection<Place> target) {
        final Collection<Collection<Place>> siphonTraps = new ArrayList<Collection<Place>>();
        Collection<Place> places = new IdentityHashSet<Place>();
        final LinkedList<Transition> border = new LinkedList<Transition>(net.getTransitions());
        
        for(final Place p : net.getPlaces()) {
            if((p.getMarking() > 0 || target.contains(p)) && this.addSeen(p)) {
                places.add(p);
                boolean closed = true;
                for(final TPArc in : p.getInput()) {
                    this.addSeen(in.getSource());
                }
                for(final PTArc out : p.getOutput()) {
                    if(!this.seen(out.getTarget())) {
                        border.add(out.getTarget());
                    }
                }
                
                while(!border.isEmpty()) {
                    final Transition t = border.remove();
                    if(this.addSeen(t)) {
                        closed = false;
                        for(final TPArc out : t.getOutput()) {
                            final Place p2 = out.getTarget();
                            if(this.addSeen(p2)) {
                                places.add(p2);
                                for(final TPArc in : p2.getInput()) {
                                    this.addSeen(in.getSource());
                                }
                                for(final PTArc out2 : p2.getOutput()) {
                                    if(!this.seen(out2.getTarget())) {
                                        border.add(out2.getTarget());
                                    }
                                }
                                closed = true;
                                break;
                            }
                        }
                        if(!closed) {
                            break;
                        }
                    }
                }
                this.clearSeenTransitions();
                
                if(closed && places.size() != net.getPlaces().size()) {
    //                System.out.println(places);
                    if(!siphonTraps.contains(places)) {
                        siphonTraps.add(places);
                        places = new IdentityHashSet<Place>();
                    } else {
                        places.clear();
                    }
                } else {
                    places.clear();
                }
            }
        }
        this.clearSeenPlaces();
        
        return siphonTraps;
    }
    
    /**
     * 
     */
//    public Collection<Map<Place, Integer>> invariantPropagation(final PetriNet net, final Map<Place, Integer> target,
//            final boolean useSiphonTraps, final boolean semipositive, final int count) {
//        this.invariantPropagation(net.getTransitions().size(), net.getPlaces(), target, useSiphonTraps, semipositive, count)
//    }
    
    private boolean computeClusterBounds(final Iterable<? extends Place> places) {
//        if(true) return false; // TODO
        final int sizeBefore = this.clusterBounds.size();
        this.clusterBounds.clear();
        for(final Place p : places) {
            if(!p.boundKnown()) {
                continue;
            }
            
            int b = p.getBound();
            Pair<Collection<Place>, Integer> pair = this.clusterBounds.get(p);
            
            final Collection<Place> cluster;
            if(pair == null) {
                cluster = new IdentityHashSet<Place>();
                cluster.add(p);
                pair = new Pair<Collection<Place>, Integer>(cluster, Integer.valueOf(b));
            } else {
                cluster = pair.getA();
            }
            
            for(final PTArc pt : p.getOutput()) {
                if(pt.getWeight() != 1) {
                    continue;
                }
                final Transition t = pt.getTarget(); 
                if(t.getOutput().size() == 1 && t.getInput().size() == 1 && !(t instanceof ExtendedTransition)) {
                    for(final TPArc tp : t.getOutput()) {
                        if(tp.getWeight() != 1) {
                            continue;
                        }
                        final Place p2 = tp.getTarget();
                        if(p2.boundKnown() && p2.getBound() == b) {
                            final Pair<Collection<Place>, Integer> pair2 = this.clusterBounds.get(p2);
                            if(pair2 != null) {
                                cluster.addAll(pair2.getA());
                            } else {
                                cluster.add(p2);
                            }
                        }
                    }
                }
            }
            
            if(cluster.size() > 1) {
                for(final Place curP : cluster) {
                    this.clusterBounds.put(curP, pair);
                }
            }
        }
        
//        final Collection<Pair<Collection<Place>, Integer>> seen = new IdentityHashSet<Pair<Collection<Place>, Integer>>();
//        for(final Pair<Collection<Place>, Integer> pair : combinedBounds.values()) {
//            if(seen.add(pair)) {
//                System.out.println(pair);
//            }
//        }
        
        return this.clusterBounds.size() > sizeBefore;
    }
    
    boolean isViolator(final Node n) {
        {
            final Marking m;
            if((m = n.violatesBound()) != null) {
                if(FINE_DEBUG) { System.out.println("violates bound."); }
                Backward.this.dotCutoff(n, "violates bound " + m.getPlace() + "<=" + m.getPlace().getBound());
                return true;
            }
        }
        return this.isViolatorNoPlaceBounds(n);
    }
    
    
    boolean isViolatorNoPlaceBounds(final Node n) {
        {
            final Pair<Collection<Place>, Integer> cluster;
            if((cluster = Backward.this.violatesClusterBound(n)) != null) {
                if(FINE_DEBUG) { System.out.println("violates invariant "); }
                if(Backward.this.dotOutput != null && Backward.this.dotPrintCutoffs) {
                    final StringBuilder s = new StringBuilder();
                    for(final Iterator<Place> theCluster = cluster.getA().iterator(); theCluster.hasNext(); ) {
                        s.append(theCluster.next());
                        if(theCluster.hasNext()) {
                            s.append(" + ");
                        }
                    }
                    s.append(" <= ");
                    s.append(cluster.getB());
                    Backward.this.dotCutoff(n, "violated cluster bound " + s.toString());
                }
                return true;
            }
        }
        {
            final Pair<Map<Place, Integer>, Integer> inv;
            if((inv = Backward.this.violatesInvariant(n)) != null) {
                if(FINE_DEBUG) { System.out.println("violates invariant"); }
                if(Backward.this.dotOutput != null && Backward.this.dotPrintCutoffs) {
                    final StringBuilder s = new StringBuilder();
                    for(final Iterator<Map.Entry<Place, Integer>> theInv = inv.getA().entrySet().iterator(); theInv.hasNext(); ) {
                        final Map.Entry<Place, Integer> e = theInv.next();
                        s.append(e.getValue());
                        s.append('*');
                        s.append(e.getKey());
                        if(theInv.hasNext()) {
                            s.append(" + ");
                        }
                    }
                    s.append(" <= ");
                    s.append(inv.getB());
                    Backward.this.dotCutoff(n, "violated invariant " + s.toString());
                }
                return true;
            }
        }
        return false;
    }
    
    
    Pair<Collection<Place>, Integer> violatesClusterBound(final Node n) {
        if(this.clusterBounds.isEmpty()) {
            return null;
        }
        Map<Pair<Collection<Place>, Integer>, Integer> mapping = null;
        int remaining = n.getMarking().size();
        for(final Marking m : n.getMarking()) {
            --remaining;
            final Pair<Collection<Place>, Integer> cluster = this.clusterBounds.get(m.getPlace());
            if(cluster != null) {
                if(mapping == null) {
                    if(remaining <= 0) {
                        return null;
                    }
                    mapping = new IdentityHashMap<Pair<Collection<Place>, Integer>, Integer>();
                }
                final Integer oldWeight = mapping.put(cluster, Integer.valueOf(m.getMarking()));
                if(oldWeight != null) {
                    int newWeight = checkedADD(oldWeight.intValue(), m.getMarking());
                    if(newWeight > cluster.getB().intValue()) {
                        return cluster;
                    }
                    mapping.put(cluster, Integer.valueOf(newWeight));
                }
            }
        }
        return null;
    }
    
    
    Pair<Map<Place, Integer>, Integer> violatesInvariant(final Node n) {
        if(this.invariants.isEmpty()) {
            return null;
        }
        Map<Pair<Map<Place, Integer>, Integer>, Integer> mapping = null;
        int remaining = n.getMarking().size();
        for(final Marking m : n.getMarking()) {
            --remaining;
            final Place p = m.getPlace();
            final Collection<Pair<Map<Place, Integer>, Integer>> invariantsWithP = this.invariants.get(p);
            if(invariantsWithP != null) {
                if(mapping == null) {
                    if(remaining <= 0) {
                        return null;
                    }
                    mapping = new IdentityHashMap<Pair<Map<Place, Integer>, Integer>, Integer>();
                }
                for(final Pair<Map<Place, Integer>, Integer> inv : invariantsWithP) {
                    final Integer coefficient = inv.getA().get(p);
                    final int w = checkedMUL(m.getMarking(), coefficient.intValue());
                    final Integer oldWeight = mapping.put(inv, w);
                    if(oldWeight != null) {
                        int newWeight = checkedADD(oldWeight.intValue(), w);
                        if(newWeight > inv.getB().intValue()) {
                            return inv;
                        }
                        mapping.put(inv, Integer.valueOf(newWeight));
                    }
                }
            }
        }
        return null;
    }
    
    
    private void removeInconsistent(final Iterable<Node> markings,
            final Collection<Map<Place, Integer>> invs) {
        int removed = 0;
        for(final Map<Place, Integer> inv : invs) {
            int initialWeightedSum = 0;
            for(final Map.Entry<Place, Integer> e : inv.entrySet()) {
                initialWeightedSum = checkedADD(initialWeightedSum, checkedMUL(e.getKey().getMarking(), e.getValue().intValue()));
            }
            for(final Iterator<Node> it = markings.iterator(); it.hasNext(); ) {
                final Node targ = it.next();
                int weightedSum = 0;
                boolean weightsUsable = true;
                Place placeBoundViolated = null;
                Pair<Collection<Place>, Integer> clusterBoundViolated = null;
                Map<Pair<Collection<Place>, Integer>, Integer> mapping = null;
                int remaining = targ.getMarking().size();
                for(final Marking m : targ.getMarking()) {
                    --remaining;
                    final Place p = m.getPlace();
                    final int tokens = m.getMarking();
                    if(p.boundKnown() && tokens > p.getBound()) {
                        placeBoundViolated = p;
                        break;
                    }

                    final Pair<Collection<Place>, Integer> cluster = this.clusterBounds.get(p);
                    if(cluster != null) {
                        if(mapping == null && remaining >= 1) {
                            mapping = new IdentityHashMap<Pair<Collection<Place>, Integer>, Integer>();
                        }
                        if(mapping != null) {
                            final Integer oldWeight = mapping.put(cluster, Integer.valueOf(m.getMarking()));
                            if(oldWeight != null) {
                                int newWeight = checkedADD(oldWeight.intValue(), tokens);
                                if(newWeight > cluster.getB().intValue()) {
                                    clusterBoundViolated = cluster;
                                    break;
                                }
                                mapping.put(cluster, Integer.valueOf(newWeight));
                            }
                        }
                    }
                    final Integer weight = inv.get(p);
                    if(weight != null) {
                        if(weight.intValue() < 0) {
                            weightsUsable = false;
                        }
                        weightedSum = checkedADD(weightedSum, checkedMUL(tokens, weight.intValue()));
                        if(weightedSum > initialWeightedSum) {
                            break;
                        }
                    }
                }
                if(placeBoundViolated != null || clusterBoundViolated != null || (weightsUsable && weightedSum > initialWeightedSum)) {
//                    this.dotCutoff(convert(targ), "violates invariant");
                    if(this.verbose) {
                        final String reason;
                        if(placeBoundViolated != null) {
                            reason = "place bound: " + placeBoundViolated + " <= " + placeBoundViolated.getBound();
                        } else if(clusterBoundViolated != null) {
                            final StringBuilder r = new StringBuilder("cluster bound: ");
                            for(final Iterator<Place> it2 = clusterBoundViolated.getA().iterator(); it2.hasNext(); ) {
                                r.append(it2.next());
                                if(it2.hasNext()) {
                                    r.append(" + ");
                                }
                            }
                            r.append(" <= ");
                            r.append(clusterBoundViolated.getB());
                            reason = r.toString();
                        } else {
                            final StringBuilder r = new StringBuilder("invariant: ");
                            for(final Iterator<Map.Entry<Place, Integer>> it2 = inv.entrySet().iterator(); it2.hasNext(); ) {
                                final Map.Entry<Place, Integer> e = it2.next();
                                if(e.getValue().intValue() > 1) {
                                    r.append(e.getValue());
                                    r.append("*");
                                }
                                r.append(e.getKey());
                                if(it2.hasNext()) {
                                    r.append(" + ");
                                }
                            }
                            r.append(" <= ");
                            r.append(initialWeightedSum);
                            reason = r.toString();
                        }
                        
                        this.verbose("Removing " + targ + " as it violates " + reason);
                        
                    }
                    it.remove();
                    ++removed;
                }
            }
        }
        
        if(removed > 0) {
            System.out.println("> Removed " + removed + " target markings which where inconsistent with invariants or bounds.");
        }
    }
    
    /**
     * Decides for an input file containing Petri net, initial marking and target markings, if a marking
     * is reachable which covers one of the target markings.
     * If so, a sequence of transitions is returned which leads to such a covering marking.
     * 
     * @param inputFile The name of the Petri net file to read which contains also the target markings.
     * @return <code>null</code> iff none of the target markings is coverable in <code>net</code>,
     *  else the trace of transitions to be fired (from the <code>initial</code> marking) which lead
     *  to some marking covering some of the <code>target</code> markings.
     *  Note that a <code>SimpleList</code> consisting of one <code>null</code>-value is returned if
     *  the <code>initial</code> marking already covers one of the <code>target</code> markings.
     */
    public static SimpleList<Transition> checkCoverability(final String inputFile) {
        final PetriNet net = Converter.loadNet(inputFile);
        
        final Collection<Place> initial = new ArrayList<Place>();
        final Map<Place, Integer> givenBounds = new IdentityHashMap<Place, Integer>();
        for(final Place p : net.getPlaces()) {
            if(p.getBound() != Place.BOUND_UNKNOWN) {
                givenBounds.put(p, Integer.valueOf(p.getBound()));
            }
            if(p.getMarking() > 0) {
                initial.add(p);
            }
        }
        
        boolean containsExtendedTransition = false;
        for(final Transition t : net.getTransitions()) {
            if(t instanceof ExtendedTransition) {
                containsExtendedTransition = true;
                break;
            }
        }
        
        final Backward bw = Backward.newInstance();
        final Nodes targets = bw.newNodes();
        Backward.readTarget(inputFile, net, targets);
        bw.containsExtendedTransition = containsExtendedTransition;
        bw.setThreadNumber(Runtime.getRuntime().availableProcessors()+1);
        
        bw.invariantPropagation(net, targets, false, false, 0);
        bw.compute(initial, net.getTransitions());
        bw.isCoverable(targets, false);
        
        return bw.getResultTrace();
    }
    
    
    /**
     * Decides for a Petri net and target markings, if a marking
     * is reachable which covers one of the target markings.
     * If so, a sequence of transitions is returned which leads to such a covering marking.
     * 
     * @param net The Petri net file to check.
     * @param targets The target markings to check.
     * @param invariants The known place invariants to help the checker.
     *  Given a net with places p1, ..., p5 and an invariant 1*p1+2*p3, the corresponding map in the
     *  <code>invariants</code> collection assigns 1 to p1 and 2 to p3. The places with coefficient 0 may be left out.  
     * @return <code>null</code> iff none of the target markings is coverable in <code>net</code>,
     *  else the trace of transitions to be fired (from the <code>initial</code> marking) which lead
     *  to some marking covering some of the <code>target</code> markings.
     *  Note that a <code>SimpleList</code> consisting of one <code>null</code>-value is returned if
     *  the <code>initial</code> marking already covers one of the <code>target</code> markings.
     */
    public static SimpleList<Transition> checkCoverability(final PetriNet net, final Collection<Map<Place, Integer>> targets, final Collection<Map<Place, Integer>> invariants) {
        final Collection<Place> initial = new ArrayList<Place>();
        for(final Place p : net.getPlaces()) {
            if(p.getMarking() > 0) {
                initial.add(p);
            }
        }
        
        boolean containsExtendedTransition = false;
        for(final Transition t : net.getTransitions()) {
            if(t instanceof ExtendedTransition) {
                containsExtendedTransition = true;
                break;
            }
        }
        
        final Backward bw = Backward.newInstance();
        final Nodes theTargets = bw.newNodes();
        for(final Map<Place, Integer> m : targets ) {
            theTargets.add(Backward.convert(m));
        }
        bw.containsExtendedTransition = containsExtendedTransition;
//        bw.setThreadNumber(Runtime.getRuntime().availableProcessors()+1);
        
        bw.invariantPropagation(net, theTargets, false, false, 0, invariants);
        bw.compute(initial, net.getTransitions());
        bw.isCoverable(theTargets, false);
        
        final SimpleList<Transition> trace = bw.getResultTrace();
        bw.pleaseTerminate();
        bw.reset();
        return trace;
    }
    
    /**
     * @param net
     * @param targets
     * @return
     */
    public static SimpleList<Transition> checkCoverability(final PetriNet net, final Collection<Map<Place, Integer>> targets) {
        return Backward.checkCoverability(net, targets, null);
    }
    
    
    public Nodes newNodes() {
        return new Nodes();
    }
    
    private class Nodes implements Iterable<Node> {
        final List<NodeClass> classes;
        
        public Nodes() {
            this(new ArrayList<NodeClass>());
        }
        
        public Nodes(final List<NodeClass> classes) {
            this.classes = classes;
        }
        
        public void add(final Node n) {
            if(this.getCoveredBy(n) == null) {
                Backward.this.addNode(n, this.classes);
            }
        }
        
        public void addAll(final Iterable<Node> nodes) {
            for(final Node n : nodes) {
                this.add(n);
            }
        }
        
        public void clear() {
            this.classes.clear();
        }
        
        public Node getCoveredBy(final Node n) {
            return Backward.this.coversNode(n, this.classes);
        }
        
        public boolean isEmpty() {
            return this.classes.isEmpty();
        }
        
        public int size() {
            int size = 0;
            for(final NodeClass nc : this.classes) {
                size += nc.size();
            }
            return size;
        }
        
        @Override
        public String toString() {
            return Backward.formatTarget(this, true);
        }
        
        @Override
        public Iterator<Node> iterator() {
            return new Iterator<Node>() {
                Iterator<NodeClass> it = Nodes.this.classes.iterator();
                NodeClass currentClass = null;
                Iterator<Node> current = null;
                
                @Override
                public boolean hasNext() {
                    if(this.current != null && this.current.hasNext()) {
                        return true;
                    }
                    while(this.it.hasNext()) {
                        this.currentClass = this.it.next();
                        this.current = this.currentClass.iterator();
                        if(this.current.hasNext()) {
                            return true;
                        }
                    }
                    return false;
                }

                @Override
                public Node next() {
                    return this.current.next();
                }

                @Override
                public void remove() {
                    this.current.remove();
                    if(this.currentClass.isEmpty()) {
                        this.it.remove();
                    }
                }
            };
        }
        
    }
    
    public Iterable<Node> convertAndMinimize(final Collection<Map<Place, Integer>> m) {
        final Nodes nodes = new Nodes();
        for(final Map<Place, Integer> n : m) {
            nodes.add(convert(n));
        }
        return nodes;
    }
    
    private static class Pair<A, B> {
        private final A a;
        private final B b;
        Pair(final A a, final B b) {
            this.a = a;
            this.b = b;
        }
        A getA() {
            return this.a;
        }
        B getB() {
            return this.b;
        }
        @Override
        public boolean equals(final Object o) {
            return (o instanceof Pair<?, ?>) && this.equals((Pair<?, ?>)o);
        }
        public boolean equals(final Pair<?, ?> p) {
            return (p != null)
                && (this.a == null ? p.getA() == null : this.a.equals(p.getA()))
                && (this.b == null ? p.getB() == null : this.b.equals(p.getB()));
        }
        @Override
        public int hashCode() {
            int h = 7 + (this.a == null ? 0 : this.a.hashCode());
            h = 31*h + (this.b == null ? 0 : this.b.hashCode());
            return h;
        }
        @Override
        public String toString() {
            return "<" + String.valueOf(this.a) + ", " + String.valueOf(this.b) + ">";
        }
    }
    
    public Collection<Map<Place, Integer>> invariantPropagation(final PetriNet net,
            final Iterable<Node> target,
            final boolean useSiphonTraps, final boolean semipositive, final int count) {
        return this.invariantPropagation(net, target, useSiphonTraps, semipositive, count, null);
    }
    
    public Collection<Map<Place, Integer>> invariantPropagation(final PetriNet net,
            final Iterable<Node> target,
            final boolean useSiphonTraps, final boolean semipositive, final int count, final Collection<Map<Place, Integer>> knownInvs) {
        final Set<Map<Place, Integer>> invs;
        if(!useSiphonTraps) {
            invs = this.generateInvariants(net, semipositive, count);
            if(knownInvs != null) {
                for(final Map<Place, Integer> inv : knownInvs) {
                    if(this.checkInvariant(inv)) {
                        invs.add(inv);
                    }
                }
            }
            this.adjustBoundsByPlaceInvariants(invs);
        } else {
            invs = new HashSet<Map<Place, Integer>>();
            final Collection<Place> pl = new IdentityHashSet<Place>();
            for(final Node n : target) {
                for(final Marking m : n.getMarking()) {
                    pl.add(m.getPlace());
                }
            }
            final Collection<Collection<Place>> siphonTraps = this.findSiphonTraps(net, pl);
            if(!siphonTraps.isEmpty()) {
                for(final Iterator<Collection<Place>> it = siphonTraps.iterator(); it.hasNext(); ) {
                    final Collection<Place> siphonTrap = it.next();
                    it.remove();
                    
                    final Set<Map<Place, Integer>> inv = this.generateInvariants(net, siphonTrap, semipositive, count);
                    if(knownInvs != null) {
                        for(final Map<Place, Integer> i : knownInvs) {
                            if(this.checkInvariant(i)) {
                                inv.add(i);
                            }
                        }
                    }
                    this.adjustBoundsByPlaceInvariants(inv);
                    
                    invs.addAll(inv);
                }
            } else {
                final Collection<Map<Place, Integer>> inv = this.generateInvariants(net, semipositive, count);
                if(knownInvs != null) {
                    for(final Map<Place, Integer> i : knownInvs) {
                        if(this.checkInvariant(i) && !inv.contains(i)) {
                            inv.add(i);
                        }
                    }
                }
                this.adjustBoundsByPlaceInvariants(inv);
                
                invs.addAll(inv);
            }
        }
        for(final Iterator<Map<Place, Integer>> it = invs.iterator(); it.hasNext(); ) {
            for(final Integer c : it.next().values()) {
                if(c.intValue() < 0) {
                    it.remove();
                    break;
                }
            }
        }
        this.computeClusterBounds(net.getPlaces());
        this.removeInconsistent(target, invs);
        final Collection<Place> unbounded = new ArrayList<Place>();
        for(final Place p : net.getPlaces()) {
            if(p.getBound() == Place.BOUND_UNBOUNDED) {
                unbounded.add(p);
            }
        }
        this.propagateNonBounds(unbounded);
        return invs;
    }
    
    public IntMatrix generateTransposedIncidenceMatrix(final Collection<Transition> theTransitions,
            final Collection<Place> thePlaces, final Collection<Place> siphonTrap, final Place[] indices) {
        final Collection<Transition> transitions; 
        final Collection<Place> places;
        if(siphonTrap == null) {
            transitions = theTransitions;
            places = thePlaces;
        } else {
            places = siphonTrap;
            transitions = new IdentityHashSet<Transition>();
            for(final Place p : places) {
                for(final PTArc out : p.getOutput()) {
                    transitions.add(out.getTarget());
                }
                for(final TPArc in : p.getInput()) {
                    transitions.add(in.getSource());
                }
            }
        }
        return this.generateTransposedIncidenceMatrix(places, transitions, indices);
    }
    
//    private void clearIndices() {
//        this.startReadingTransitionData();
//        for(final TransitionData d : this.transitionData.values()) {
//            d.index = -1;
//        }
//        this.finishReadingTransitionData();
//        this.startReadingPlaceData();
//        for(final PlaceData d : this.placeData.values()) {
//            d.index = -1;
//        }
//        this.finishReadingPlaceData();
//        if(this.friendlyTransitions != null) {
//            for(final petruchio.pn.Transition t : this.friendlyTransitions) {
//                ((TransitionData)t.data).index = -1;
//            }
//        }
//        if(this.friendlyPlaces != null) {
//            for(final petruchio.pn.Place p : this.friendlyPlaces) {
//                ((PlaceData)p.data).index = -1;
//            }
//        }
//    }
    
    public IntMatrix generateTransposedIncidenceMatrix(final Collection<Place> places,
            final Collection<Transition> transitions, final Place[] indices) {
        if(this.containsExtendedTransition) {
            return this.generateExtendedTransposedIncidenceMatrix(places, transitions, indices);
        }
        final IntMatrix incidenceTr = IntPNMatrix.getInstance(transitions.size(), places.size());
        
        int pctr = 0;
        final List<Place> thePlaces = new ArrayList<Place>(places);
        Collections.sort(thePlaces, Backward.PLACE_COMPARATOR);
        for(final Place p : thePlaces) {
            final PlaceData pd = this.getPlaceData(p);
            pd.index = pctr++;
            indices[pd.index] = p;
        }
        
        
        int tctr = 0;
        final List<Transition> theTransitions = new ArrayList<Transition>(transitions);
        Collections.sort(theTransitions, Backward.TRANSITION_COMPARATOR);
        for(final Transition t : theTransitions) {
            final TransitionData td = this.getTransitionData(t);
            td.index = tctr++;
        }
        
        for(final Place p : places) {
            final PlaceData pd = this.getPlaceData(p);

            for(final TPArc in : p.getInput()) {
                if(!transitions.contains(in.getSource())) {
                    continue;
                }
                final TransitionData td = this.getTransitionData(in.getSource());
                incidenceTr.set(td.index, pd.index, in.getWeight());
            }
            for(final PTArc out : p.getOutput()) {
                if(!transitions.contains(out.getTarget())) {
                    continue;
                }
                final TransitionData td = this.getTransitionData(out.getTarget());
                incidenceTr.subtract(td.index, pd.index, out.getWeight());
            }
        }
        return incidenceTr;
    }
    
    
    private IntMatrix generateExtendedTransposedIncidenceMatrix(final Collection<Place> places,
            final Collection<Transition> transitions, final Place[] indices) {
        final IntMatrix incidenceTr = IntPNMatrix.getInstance(places.size(), transitions.size());
        
        int pctr = 0;
        final List<Place> thePlaces = new ArrayList<Place>(places);
        Collections.sort(thePlaces, Backward.PLACE_COMPARATOR);
        for(final Place p : thePlaces) {
            final PlaceData pd = this.getPlaceData(p);
            pd.index = pctr++;
            indices[pd.index] = p;
        }
        
        int tctr = 0;
        final List<Transition> theTransitions = new ArrayList<Transition>(transitions);
        Collections.sort(theTransitions, Backward.TRANSITION_COMPARATOR);
        for(final Transition t : theTransitions) {
            final TransitionData td = this.getTransitionData(t);
            td.index = tctr++;
        }
        
        final Map<Pair<Pair<Place, Integer>, Place>, Integer> transIndices = new HashMap<Pair<Pair<Place, Integer>, Place>, Integer>();
        
        for(final Place p : places) {
            final PlaceData pd = this.getPlaceData(p);

            for(final TPArc in : p.getInput()) {
                if(!transitions.contains(in.getSource())) {
                    continue;
                }
                final TransitionData td = this.getTransitionData(in.getSource());
                incidenceTr.set(pd.index, td.index, in.getWeight());
                if(in instanceof TransferArc) {
                    for(final Map.Entry<Place, Integer> e : ((TransferArc)in).getTransfer().entrySet()) {
                        if(e.getKey() == p) {
                            continue;
                        }
                        final Pair<Pair<Place, Integer>, Place> pair = new Pair<Pair<Place, Integer>, Place>(new Pair<Place, Integer>(e.getKey(), e.getValue()), p);
                        Integer idx = transIndices.get(pair);
                        if(idx == null) {
                            idx = Integer.valueOf(incidenceTr.addColumn());
                            transIndices.put(pair, idx);
                        }
                        incidenceTr.set(pd.index, idx.intValue(), e.getValue().intValue());
                    }
                }
            }
            for(final PTArc out : p.getOutput()) {
                if(!transitions.contains(out.getTarget())) {
                    continue;
                }
                final TransitionData td = this.getTransitionData(out.getTarget());
                incidenceTr.subtract(pd.index, td.index, out.getWeight());
                if(out instanceof ResetArc) {
                    boolean matchingTransferFound = false;
                    for(final TPArc in : out.getTarget().getOutput()) {
                        if(in instanceof TransferArc) {
                            final Integer trans = ((TransferArc)in).getTransfer().get(p);
                            if(trans != null) {
                                final Pair<Pair<Place, Integer>, Place> pair = new Pair<Pair<Place, Integer>, Place>(new Pair<Place, Integer>(p, trans), in.getTarget());
                                Integer idx = transIndices.get(pair);
                                if(idx == null) {
                                    idx = Integer.valueOf(incidenceTr.addColumn());
                                    transIndices.put(pair, idx);
                                }
                                incidenceTr.subtract(pd.index, idx.intValue(), 1);
                                matchingTransferFound = true;
                            }
                        }                        
                    }
                    if(!matchingTransferFound) {
                        final int idx = incidenceTr.addColumn();
                        incidenceTr.subtract(pd.index, idx, 1);
                    }
                }
            }
        }
        
        incidenceTr.transpose();
        return incidenceTr;
    }
    
    
    public Set<Map<Place, Integer>> generateInvariants(final PetriNet net,
            final boolean semipositive, final int count) {
        return this.generateInvariants(net.getTransitions(), net.getPlaces(), semipositive, count);
    }
    
    public Set<Map<Place, Integer>> generateInvariants(final Collection<Transition> transitions,
            final Collection<Place> places, final boolean semipositive, final int count) {
        return this.generateInvariants(transitions, places, null, semipositive, count);
    }
    
    public Set<Map<Place, Integer>> generateInvariants(final PetriNet net,
            final Collection<Place> siphonTrap, final boolean semipositive, final int count) {
        return this.generateInvariants(net.getTransitions(), net.getPlaces(), siphonTrap, semipositive, count);
    }
    
    public Set<Map<Place, Integer>> generateInvariants(final Collection<Transition> transitions,
            final Collection<Place> places, final Collection<Place> siphonTrap,
            final boolean semipositive, final int count) {
        try {
            this.invariants.clear();
            final Place[] indices = new Place[places.size()];
            final IntMatrix incidence = this.generateTransposedIncidenceMatrix(transitions, places, siphonTrap, indices);
            final IntMatrix invariantMatrix = InvariantAnalysis.findInvariants(incidence, semipositive, count, indices);
            final Set<Map<Place, Integer>> invs = new HashSet<Map<Place, Integer>>(invariantMatrix.getColumns());
            for(int i = invariantMatrix.getColumns()-1; i >= 0; --i) {
                if(!invariantMatrix.columnExists(i)) {
                    continue;
                }
                final Map<Place, Integer> inv = new IdentityHashMap<Place, Integer>();
                for(int j = invariantMatrix.getRows()-1; j >= 0; --j) {
                    final int v = invariantMatrix.get(j, i);
                    if(v != 0) {
                        inv.put(indices[j], Integer.valueOf(v));
                    }
                }
                invariantMatrix.killColumn(i);
                if(!inv.isEmpty() && this.checkInvariant(inv)) {
                    invs.add(inv);
                }
            }
            return invs;
        } catch(final ArithmeticException e) {
            System.out.println(">> Error: Arithmetic exception during invariant generation: " + e.getMessage()
                    + "\n>> Will not use invariants.");
            this.invariants.clear();
        }
        return new HashSet<Map<Place, Integer>>();
    }
    
    
    private boolean checkInvariant(final Map<Place, Integer> inv) {
        boolean isSemipositive = true;
        int initialWeightedSum = 0;
        final boolean isSemipositiveChecked;
        if(this.containsExtendedTransition || this.containsParameters) {
            isSemipositiveChecked = true;
            for(final Map.Entry<Place, Integer> e : inv.entrySet()) {
                final Place p = e.getKey();
                if(p.getBound() == Place.BOUND_UNBOUNDED) {
                    return false;
                }
                if(isSemipositive) {
                    if(e.getValue().intValue() < 0) {
                        isSemipositive = false;
                    } else if(p.getMarking() > 0) {
                        initialWeightedSum = checkedADD(initialWeightedSum, checkedMUL(e.getValue().intValue(), p.getMarking()));
                    }
                }
                if(p.isParametric()) {
                    return false;
                }
                
                for(final PTArc a : p.getOutput()) {
                    if(a instanceof InhibitorArc) {
                        return false;
                    }
                }
                for(final TPArc a : p.getInput()) {
                    for(final PTArc b : a.getSource().getInput()) {
                        if(b.getSource().isParametric()) {
                            return false;
                        }
                    }
                }
            }
        } else {
            isSemipositiveChecked = false;
        }
        
        if(!isSemipositiveChecked) {
            for(final Map.Entry<Place, Integer> e : inv.entrySet()) {
                final Place p = e.getKey();
                if(p.getBound() == Place.BOUND_UNBOUNDED) {
                    return false;
                }
                if(e.getValue().intValue() < 0) {
                    isSemipositive = false;
                    break;
                } else if(p.getMarking() > 0) {
                    initialWeightedSum = checkedADD(initialWeightedSum, checkedMUL(e.getValue().intValue(), p.getMarking()));
                }
            }
        }
        if(isSemipositive) {
            final Pair<Map<Place, Integer>, Integer> invWithSum = new Pair<Map<Place, Integer>, Integer>(inv, Integer.valueOf(initialWeightedSum));
            for(final Place p : inv.keySet()) {
                Collection<Pair<Map<Place, Integer>, Integer>> invs = this.invariants.get(p);
                if(invs == null) {
                    invs = new ArrayList<Pair<Map<Place, Integer>, Integer>>();
                    this.invariants.put(p, invs);
                }
                invs.add(invWithSum);
            }
        }
        return true;
    }
    
    
    private static void formatTrace(final List<Transition> trace, final int mult, final StringBuilder s) {
        if(trace.isEmpty()) {
            return;
        }
        if(s.length() != 0) {
            s.append(", ");
        }
        if(mult > 1) {
            s.append('(');
        }
//        for(final Iterator<Transition> it = trace.iterator(); it.hasNext();) {
//            s.append(it.next());
//            if(it.hasNext()) {
//                s.append(", ");
//            }
//        }
        s.append(trace.get(0));
        Backward.formatTrace(trace.subList(1, trace.size()), s);
        if(mult > 1) {
            s.append(")^");
            s.append(mult);
        }
    }
    
    
    private static void formatTrace(final List<Transition> trace, final StringBuilder s) {
        final List<Transition> sequence = new ArrayList<Transition>();
        int multiplier = 1;
        int matched = 0;
        for(final Transition t : trace) {
            if(sequence.isEmpty() || (multiplier == 1 && sequence.get(matched) != t)) {
                for(int i = 0; i < matched; ++i) {
                    sequence.add(sequence.get(i));
                }
                sequence.add(t);
                matched = 0;
            } else if(sequence.get(matched) == t) {
                ++matched;
            } else {
                Backward.formatTrace(sequence, multiplier, s);
                
                for(int i = sequence.size()-1; i >= matched; --i) {
                    sequence.remove(i);
                }
                
                multiplier = 1;
                matched = 0;
                
                if(!sequence.isEmpty() && sequence.get(0) == t) {
                    ++matched;
                } else {
                    sequence.add(t);
                }
            }
            
            if(matched == sequence.size()) {
                ++multiplier;
                matched = 0;
            }
        }
        
        if(multiplier > 1) {
            Backward.formatTrace(sequence, multiplier, s);
            if(matched != 0) {
                Backward.formatTrace(sequence.subList(0, matched), s);
            }
        } else {
            for(int i = 0; i < matched; ++i) {
                sequence.add(sequence.get(i));
            }
            matched = 0;
            Backward.formatTrace(sequence, 1, s);
        }
    }
    
    
    /**
     * @param trace
     * @return
     */
    public static String formatTrace(final SimpleList<Transition> trace) {
        final StringBuilder s = new StringBuilder();
        final Map<String, Integer> independentTransitions = new TreeMap<String, Integer>();
        
        final List<Transition> sequence = new ArrayList<Transition>();
        int multiplier = 1;
        int matched = 0;
        for(SimpleList<Transition> cur = trace; cur != null && cur.getValue() != null; cur = cur.getNext()) {
            if(cur.getValue().getInput().isEmpty()) {
                final Integer counter = independentTransitions.put(cur.getValue().toString(), Integer.valueOf(1));
                if(counter != null) {
                    independentTransitions.put(cur.getValue().toString(), Integer.valueOf(counter.intValue() + 1));
                }
            } else {
                if(sequence.isEmpty() || (multiplier == 1 && sequence.get(matched) != cur.getValue())) {
                    for(int i = 0; i < matched; ++i) {
                        sequence.add(sequence.get(i));
                    }
                    sequence.add(cur.getValue());
                    matched = 0;
                } else if(sequence.get(matched) == cur.getValue()) {
                    ++matched;
                } else {
                    Backward.formatTrace(sequence, multiplier, s);
                    
                    for(int i = sequence.size()-1; i >= matched; --i) {
                        sequence.remove(i);
                    }
                    
                    multiplier = 1;
                    matched = 0;
                    
                    if(!sequence.isEmpty() && sequence.get(0) == cur.getValue()) {
                        ++matched;
                    } else {
                        sequence.add(cur.getValue());
                    }
                }
                
                if(matched == sequence.size()) {
                    ++multiplier;
                    matched = 0;
                }
            }
        }
        
        if(multiplier > 1) {
            Backward.formatTrace(sequence, multiplier, s);
            if(matched != 0) {
                Backward.formatTrace(sequence.subList(0, matched), s);
            }
        } else {
            for(int i = 0; i < matched; ++i) {
                sequence.add(sequence.get(i));
            }
            matched = 0;
            Backward.formatTrace(sequence, 1, s);
        }
        
        final StringBuilder indep = new StringBuilder();
        for(final Map.Entry<String, Integer> e : independentTransitions.entrySet()) {
            final int mult = e.getValue().intValue();
            if(mult > 1) indep.append('(');
            indep.append(e.getKey());
            if(mult > 1) {
                indep.append(")^");
                indep.append(e.getValue());
            }
            indep.append(", ");
        }
        
        return indep.toString() + s.toString();
    }
    
    
    private static Node convert(final Map<Place, Integer> m) {
        final Node n = Node.newNode();
        for(final Map.Entry<Place, Integer> e : m.entrySet()) {
            n.add(e.getKey(), e.getValue());
        }
        return n;
    }
    
    private static Node convert(final Collection<Place> m) {
        final Node n = Node.newNode();
        for(final Place p : m) {
            n.add(p, p.getMarking());
        }
        return n;
    }
    
    private static Map<Place, Integer> convert(final Node n) {
        final Map<Place, Integer> m = new IdentityHashMap<Place, Integer>();
        for(final Marking mark : n.getMarking()) {
            m.put(mark.getPlace(), mark.getMarking());
        }
        return m;
    }
    
    public Map<Place, Integer> fire(final Collection<Place> initialMarking,
            final SimpleList<Transition> trace) {
        return convert(this.fire(convert(initialMarking), trace));
    }
    
    public Map<Place, Integer> fire(final Map<Place, Integer> initialMarking,
            final SimpleList<Transition> trace) {
        return convert(this.fire(convert(initialMarking), trace));
    }
    
    public Node fire(final Node initialMarking, final SimpleList<Transition> trace) {
        final Node current = Node.newNode(initialMarking);
        if(trace == null) {
            return current;
        }
        if(this.verbose) {
            this.verbose("---- Firing trace ----");
            this.verbose(current.toString());
        }
        for(final Transition t : trace) {
            this.fire(current, t);
            if(this.verbose) {
                this.verbose("-- " + t.toString() + " -->");
                this.verbose(current.toString());
            }
        }
        this.verbose("----------------------");
        return current;
    }
    
    public void fire(final Node m, final Transition t) {
        if(t == null) {
            return;
        }
//        System.out.println("    in: " + t.getInput());
//        System.out.println("   out: " + t.getOutput());
        
        final boolean extended = t instanceof ExtendedTransition;
        Map<Place, Integer> addLater = null;
        if(extended) {
            for(final TPArc out : t.getOutput()) {
                if(out instanceof TransferArc) {
                    if(addLater == null) {
                        addLater = new IdentityHashMap<Place, Integer>();
                    }
                    int tokensum = 0;
                    final TransferArc trans = (TransferArc)out;
                    for(final Map.Entry<Place, Integer> e : trans.getTransfer().entrySet()) {
                        final Marking mark = m.getMarking(e.getKey());
                        if(mark != null) {
                            if(trans.getBound() <= 0) {
                                tokensum += mark.getMarking() * e.getValue();
                            } else {
                                tokensum += Math.min(mark.getMarking(), trans.getBound()) * e.getValue();
                            }
                        }
                    }
                    addLater.put(out.getTarget(), Integer.valueOf(tokensum));
                }
            }
        }
        final TransitionData d = this.getTransitionData(t);
        if(extended) {
            final Iterable<InhibitorArc> inhi = this.getInhibitorArcs((ExtendedTransition)t, d);
            if(inhi != null) {
                for(final InhibitorArc a : inhi) {
                    final Marking c = m.getMarking(a.getSource());
                    if(c != null && c.getMarking() > a.getBound()) {
                        throw new RuntimeException("Marking " + m + " does not enable " + t + " due to inhibitor arc " + a);
                    }
                }
            }
        }
        final Node in = this.getInNode(t, d);
        try {
            if(extended) {
                final Iterable<Place> trans = this.getTransferTargets((ExtendedTransition)t, d);
                for(final Marking c : in.getMarking()) {
                    boolean okay = true;
                    for(final Place p : trans) {
                        if(p == c.getPlace()) {
                            okay = false;
                            break;
                        }
                    }
                    if(okay) {
                        m.subtract(c.getPlace(), c.getMarking());
                    }
                }
            } else {
                m.subtractWithSanityCheck(in);
            }
        } catch(final RuntimeException e) {
            throw new RuntimeException("Unable to fire " + t, e);
        }
        if(addLater != null) {
            final Iterable<Place> transfers = this.getTransferTargets((ExtendedTransition)t, d);
            if(transfers != null) {
                for(final Place p : transfers) {
                    m.set(p, 0);
                }
            }
            for(final Map.Entry<Place, Integer> e : addLater.entrySet()) {
                m.add(e.getKey(), e.getValue().intValue());
            }
            // System.out.println("addLater of " + t + " is " + addLater);
        }
        if(extended) {
            final Iterable<Place> resets = this.getResetSources((ExtendedTransition)t, d);
            if(resets != null) {
                for(final Place p : resets) {
                    m.set(p, 0);
                }
            }
        }
        m.add(this.getOutNode(t, d));
        if(extended) {
            final Iterable<Place> trans = this.getTransferTargets((ExtendedTransition)t, d);
            for(final Marking c : in.getMarking()) {
                boolean okay = false;
                for(final Place p : trans) {
                    if(p == c.getPlace()) {
                        okay = true;
                        break;
                    }
                }
                if(okay) {
                    m.subtract(c.getPlace(), c.getMarking());
                }
            }
        }
    }
    
    public static String formatMarking(final Node m) {
        final Map<String, Integer> s = new TreeMap<String, Integer>(); 
        for(final Marking e : m.getMarking()) {
            s.put(e.getPlace().toString(), Integer.valueOf(e.getMarking()));
        }
        return s.toString();
    }
    
    public static String formatMarking(final Node m, final Collection<Place> ignore) {
        final Map<String, Integer> s = new TreeMap<String, Integer>(); 
        for(final Marking e : m.getMarking()) {
            if(!ignore.contains(e.getPlace())) {
                s.put(e.getPlace().toString(), Integer.valueOf(e.getMarking()));
            }
        }
        return s.toString();
    }
    
    private static Node sanityCheck(final Node r, final Nodes markings) {
        final Node covered = markings.getCoveredBy(r);
        if(covered == null) {
            System.out.flush();
            System.err.println("ERROR: Trace does not lead to a covering marking!");
            System.err.flush();
        }
        return covered;
    }
    
    public static String formatMarking(final Map<Place, Integer> m) {
        final Map<String, Integer> s = new TreeMap<String, Integer>(); 
        for(final Map.Entry<Place, Integer> e : m.entrySet()) {
            s.put(e.getKey().toString(), e.getValue());
        }
        return s.toString();
    }

    public static String formatTarget(final Iterable<Node> m) {
        return formatTarget(m, false);
    }
    
    public static String formatTarget(final Iterable<Node> m, final boolean all) {
        final StringBuilder s = new StringBuilder();
        final int NUMBER_TO_PRINT = 2;
        int printed = 0;
        s.append("{ ");
        for(final Iterator<Node> markIt = m.iterator(); markIt.hasNext(); ) {
            if(!all && printed >= NUMBER_TO_PRINT) {
                s.append("...");
                break;
            }
            ++printed;
            final Map<String, Integer> cur = new TreeMap<String, Integer>(); 
            for(final Marking ma : markIt.next().getMarking()) {
                cur.put(ma.getPlace().toString(), ma.getMarking());
            }
            s.append('(');
            for(final Iterator<Map.Entry<String, Integer>> it = cur.entrySet().iterator(); it.hasNext(); ) {
                final Map.Entry<String, Integer> e = it.next();
                s.append(e.getKey());
                s.append(">=");
                s.append(e.getValue());
                if(it.hasNext()) {
                    s.append(", ");
                }
            }
            s.append(')');
            if(markIt.hasNext()) {
                s.append(", ");
            }
        }
        s.append(" }");
        return s.toString();
    }
    
    public static String formatTargetLegacy(final Iterable<Map<Place, Integer>> m) {
        final StringBuilder s = new StringBuilder();
        final int NUMBER_TO_PRINT = 2;
        int printed = 0;
        s.append("{ ");
        for(final Iterator<Map<Place, Integer>> markIt = m.iterator(); markIt.hasNext(); ) {
            if(printed >= NUMBER_TO_PRINT) {
                s.append("...");
                break;
            }
            ++printed;
            final Map<String, Integer> cur = new TreeMap<String, Integer>(); 
            for(final Map.Entry<Place, Integer> e : markIt.next().entrySet()) {
                cur.put(e.getKey().toString(), e.getValue());
            }
            s.append('(');
            for(final Iterator<Map.Entry<String, Integer>> it = cur.entrySet().iterator(); it.hasNext(); ) {
                final Map.Entry<String, Integer> e = it.next();
                s.append(e.getKey());
                s.append(">=");
                s.append(e.getValue());
                if(it.hasNext()) {
                    s.append(", ");
                }
            }
            s.append(')');
            if(markIt.hasNext()) {
                s.append(", ");
            }
        }
        s.append(" }");
        return s.toString();
    }
    
    
    public static long getMeasuringTime() {
        return System.nanoTime();
        
//        final ThreadMXBean bean = ManagementFactory.getThreadMXBean();
//        if(!bean.isThreadCpuTimeSupported()) {
//            return 0L;
//        }
//        return bean.getCurrentThreadCpuTime();
    }
    
    private static int distribute(final int tokens, final Place[] pl, final int nextPlace, final Map<Place, Integer> m, final Nodes markings) {
        if(tokens == 0 || nextPlace >= pl.length) {
            return 0;
        }
        int created = 0;
        for(int curPlace = nextPlace; curPlace < pl.length; ++curPlace) {
            final Place place = pl[curPlace];
            m.put(place, tokens);
            markings.add(convert(m));
            ++created;
            
            for(int x = tokens-1; x > 0; --x) {
                m.put(place, x);
                created += distribute(tokens - x, pl, curPlace+1, m, markings);
            }
            m.remove(place);
        }
        return created;
    }
    
    public static long choose(final int n, final int k) {
        if(n < 0 || n < k) {
            throw new IllegalArgumentException("0 <= k <= n does not hold.");
        }
        if((n == k) || (k == 0)) {
            return 1;
        }
        if((k == 1) || (k == n - 1)) {
            return n;
        }
        if(k > n / 2) {
            return choose(n, n - k);
        }

        long result = 1;
        int i = n - k + 1;
        if(n <= 61) { // all is safe here
            for (int j = 1; j <= k; ++j, ++i) {
                result = result * i / j;
            }
        } else if (n <= 66) { // watch for intermediate overflows
            for (int j = 1; j <= k; ++j, ++i) {
                final long d = gcd(i, j);
                result = (result / (j / d)) * (i / d);
            }
        } else { // watch for overflow
            for (int j = 1; j <= k; ++j, ++i) {
                final long d = gcd(i, j);
                result = checkedMUL(result / (j / d), i / d);
            }
        }
        return result;
    }
    
    /**
     * @param args
     */
    public static void main(final String[] args) {
        if(args.length == 0 || Arrays.asList(args).contains("--help")) {
            System.out.println("Not enough arguments.");
            System.out.println("arguments: <filename> [options] [<place>[:<tokens>] [...] [ ; [...] ]]");
            System.out.println("example: mynet.ll_net p1:1 p2:3 p3 ; p1:1 p2:3 p4");
            System.out.println("  checks whether a marking p1 >= 1, p2 >= 3, p3 >= 3");
            System.out.println("  and/or a marking p1 >= 1, p2 >= 3, p4 >= 1 is reachable");
            System.out.println("options:");
            System.out.println(" --help                   Print this text and exit");
            System.out.println(" --loop-all               Repeat benchmark even if 'result' was 'false'");
            System.out.println(" --loop-count:<count>     Repeat benchmark <count> times");
            System.out.println(" --loop-startup:<ms>      Startup time before counting benchmarking results in milliseconds.");
            System.out.println(" --loop-accuracy:<grain>  Repeat benchmark until the average doesn't change more than <grain>");
            System.out.println(" --loop-accurate:<count>  Repeat benchmark until the average doesn't change for <count> times");
            System.out.println(" --loop-verbose           Print information on loop iterations");
            System.out.println(" --no-bounds              Do not make use of obtained place bounds");
            System.out.println(" --no-given-bounds        Do not make use of bounds stated in the input file");
            System.out.println(" --no-given-invs          Do not make use of invariants stated in the input file");
            System.out.println(" --no-po                  Do not make use of partial order technique");
            System.out.println(" --no-accel               Do not make use of backward acceleration technique");
            System.out.println(" --no-order               Do not make use of ordering predecessors");
            System.out.println(" --no-invs                Do not make use of place invariants");
            System.out.println(" --shortest-trace         <buggy>Return shortest trace (via BMC).</buggy>");
            System.out.println(" --sp-invs                Compute minimal support semipositive invariants (slow)");
            System.out.println(" --siphon-traps           Look for siphon-traps to use for invariants (large nets)");
            System.out.println(" --bmc-depth:<depth>      Set maximal depth for bounded model checking (slow).");
            System.out.println(" --inv-count:<count>      Additional loops to find invariants");
            System.out.println(" --read-targets           Attempts to read target markings from input file (currently only .spec).");
//            System.out.println(" --init:<init>            Sets the initial marking (e.g. p1=2,p2,p3>=1) - places not assigned are implicitly p=0.");
//            System.out.println(" --init-ge0:<init>        Sets the initial marking (e.g. p1=2,p2,p3>=1) - places not assigned are implicitly p>=0.");
            System.out.println(" --combinations:<count>:<place>[,(...,)?<place>]+\n"
                             + "                          Attempts to add all combinations of <count> tokens on the given places as target markings.");
            System.out.println(" --reachability:<bound>   Assumes the net to be bounded and checks reachability of specified markings");
            System.out.println("                          (Places not mentioned are set to zero. Default for bound is 1.)");
            System.out.println(" --dot-output:<filename>  Print computation tree in DOT format, filename may be empty");
            System.out.println("                          the input file's name with .dot suffix will then be used");
            System.out.println(" --dot-print-cutoffs      Print cutoff information for DOT output");
            System.out.println(" --print-invs             Print invariants");
            System.out.println(" --print-bounds           Print bounds");
            System.out.println(" --print-targets          Print all target markings");
            System.out.println(" --print-result-line      Print conclusion line for benchmarks");
            System.out.println(" --fire:<sequence>        Attempts to fire the given sequence (e.g. t1,t2,t3) and prints the resulting marking.");
            System.out.println(" --benchmark:<minT>:<minIt>:<maxIt>\n"
                             + "                          Performs benchmarking runs \"while((it <= minIt || time <= minT) && (maxIt < minIt || it <= maxIt))\".");
            System.out.println(" --tree-height:<lvl>      Restricts the height of the node class decision trees (values <= 0 do not restrict)");
            System.out.println(" --print-max-tree-height  Prints the maximal tree height.");
            return;
        }
        long time = getMeasuringTime();
        final Backward bw = Backward.newInstance();
        boolean all = false;
        boolean noBounds = false;
        boolean noPO = false;
        boolean noGivenBounds = false;
        boolean noGivenInvs = false;
        boolean noOrder = false;
        boolean noAccel = false;
        boolean noInv = false;
        boolean printInv = false;
        boolean printBounds = false;
        boolean loopVerbose = false;
        boolean useSiphonTraps = false;
        boolean makeSemiPositive = false;
        boolean shortestTrace = false;
        boolean verbose = false;
        boolean printTargets = false;
        boolean printResultLine = false;
        boolean readTargets = false;
        boolean benchmark = false;
        boolean getMaxTreeHeight = false;
        int benchmark_minT = 0;
        int benchmark_minIt = 0;
        int benchmark_maxIt = 0;
        int reachability = 0;
        String dotOutput = null;
        boolean dotPrintCutoffs = false;
        int nodeClassHeight = 0;
        boolean nodeClassHeightSet = false;
        
        int loopStartup = 10000;
        int runCount = 0;
        int accurateCount = 0;
        int additionalInvCount = 0;
        int bmcDepth = 0;
        double grain = 0.0;
        SimpleList<String> firingSequence = null;
        
        final String inputFile = args[0];
        
        System.out.println("Loading file '" + inputFile + "'");
        final PetriNet net = Converter.loadNet(inputFile);
        final Nodes markings = bw.newNodes();
        Map<Place, Integer> marking = new IdentityHashMap<Place, Integer>();
        for(int i = 1; i < args.length; ++i) {
            final int idx = args[i].indexOf(':');
            final String key;
            final String value;
            if(idx < 0) {
                key = args[i];
                value = "";
            } else {
                key = args[i].substring(0, idx);
                value = args[i].substring(idx+1);
            }
            if(key.equals("--loop-all")) {
                all = true;
            } else if(key.equals("--loop-count")) {
                runCount = value.length() == 0 ? 10 : Integer.parseInt(value);
            } else if(key.equals("--loop-startup")) {
                loopStartup = value.length() == 0 ? loopStartup : Integer.parseInt(value);
            } else if(key.equals("--loop-accuracy")) {
                grain = 1000*Double.parseDouble(value);
                if(accurateCount <= 0) {
                    accurateCount = 3;
                }
            } else if(key.equals("--loop-accurate")) {
                accurateCount = value.length() == 0 ? 3 : Integer.parseInt(value);
            } else if(key.equals("--loop-verbose")) {
                loopVerbose = true;
            } else if(key.equals("--bmc-depth")) {
                bmcDepth = Integer.parseInt(value);
            } else if(key.equals("--reachability")) {
                reachability = value.length() == 0 ? 1 : Integer.parseInt(value);
                if(reachability <= 0) {
                    reachability = 1;
                }
            } else if(key.equals("--bound")) {
                final int b = value.length() == 0 ? 1 : Integer.parseInt(value);
                for(final Place p : net.getPlaces()) {
                    p.setBound(b);
                }
            } else if(key.equals("--shortest-trace")) {
                shortestTrace = true;
            } else if(key.equals("--verbose")) {
                verbose = true;
            } else if(key.equals("--no-bounds")) {
                noBounds = true;
            } else if(key.equals("--no-po")) {
                noPO = true;
            } else if(key.equals("--no-order")) {
                noOrder = true;
            } else if(key.equals("--read-targets")) {
                readTargets = true;
            } else if(key.equals("--print-targets")) {
                printTargets = true;
            } else if(key.equals("--print-result-line")) {
                printResultLine = true;
            } else if(key.equals("--print-max-tree-height")) {
                getMaxTreeHeight = true;
            } else if(key.equals("--tree-height")) {
                nodeClassHeight = Integer.parseInt(value);
                nodeClassHeightSet = true;
            } else if(key.equals("--benchmark")) {
                benchmark = true;
                all = true;
                final int sep = value.indexOf(':');
                benchmark_minT = Integer.parseInt(value.substring(0, sep));
                if(benchmark_minT < 0) {
                    throw new IllegalArgumentException("<minT> must be >= 0.");
                }
                final int sep2 = value.indexOf(':', sep+1);
                benchmark_minIt = Integer.parseInt(value.substring(sep+1, sep2));
                if(benchmark_minIt < 0) {
                    throw new IllegalArgumentException("<minIt> must be >= 0.");
                }
                runCount = benchmark_minIt;
                benchmark_maxIt = Integer.parseInt(value.substring(sep2+1));
                System.out.println("Benchmarking at least for " + benchmark_minT + " minutes (but at most for " + benchmark_maxIt + " iterations) and at least " + benchmark_minIt + " loop iterations.");
            } else if(key.equals("--combinations")) {
                final int sep = value.indexOf(':');
                final int tokens = Integer.parseInt(value.substring(0, sep));
                if(tokens <= 0) {
                    System.out.println("No tokens to generate combinations.");
                    continue;
                }
                final String[] places = value.substring(sep+1).split(",");
                final Collection<Place> thePlaces = new IdentityHashSet<Place>(places.length);
                boolean dots = false;
                String lastPlace = null;
                for(final String place : places) {
                    if(place.equals("...")) {
                        dots = true; 
                        continue;
                    }
                    
                    thePlaces.addAll(net.getPlaces(place));
                    
                    if(dots) {
                        if(lastPlace == null) {
                            throw new IllegalArgumentException("Cannot resolve '...' for combinations. No recognizable first place.");
                        }
                        int lpcut = 0;
                        while(!Character.isDigit(lastPlace.charAt(lpcut))) {
                            ++lpcut;
                        }
                        int pcut = 0;
                        while(!Character.isDigit(place.charAt(pcut))) {
                            ++pcut;
                        }
                        final String prefix = lastPlace.substring(0, lpcut);
                        final int begin = Integer.parseInt(lastPlace.substring(lpcut))+1;
                        final int end = Integer.parseInt(place.substring(pcut));
                        for(int x = begin; x <= end; ++x) {
                            thePlaces.addAll(net.getPlaces(prefix + x));
                        }
                        dots = false;
                    }
                    
                    lastPlace = place;
                }
                
                final Place[] pl = thePlaces.toArray(new Place[thePlaces.size()]);
                System.out.print("Generating (" + thePlaces.size() + "+" + tokens + "-1 choose " + tokens + ") = " + choose(thePlaces.size()+tokens-1, tokens) + " markings... ");
                System.out.println("done. " + distribute(tokens, pl, 0, marking, markings) + " markings created.");
                marking = new IdentityHashMap<Place, Integer>();
                
            } else if(key.equals("--fire")) {
                int start = 0;
                while(true) {
                    final int end = value.indexOf(',', start);
                    if(end < 0) {
                        firingSequence = new SimpleList<String>(value.substring(start), firingSequence);
                        break;
                    }
                    firingSequence = new SimpleList<String>(value.substring(start, end), firingSequence);
                    start = end+1;
                }
            } else if(key.equals("--no-given-bounds")) {
                noGivenBounds = true;
            } else if(key.equals("--no-given-invs")) {
                noGivenInvs = true;
            } else if(key.equals("--no-invs")) {
                noInv = true;
            } else if(key.equals("--no-accel")) {
                noAccel = true;
            } else if(key.equals("--sp-invs")) {
                makeSemiPositive = true;
            } else if(key.equals("--siphon-traps")) {
                useSiphonTraps = true;
            } else if(key.equals("--inv-count")) {
                additionalInvCount = value.length() == 0 ? additionalInvCount : Integer.parseInt(value);
            } else if(key.equals("--print-invs")) {
                printInv = true;
            } else if(key.equals("--print-bounds")) {
                printBounds = true;
            } else if(key.equals("--dot-output")) {
                dotOutput = value.isEmpty() ? inputFile + ".dot" : value;
            } else if(key.equals("--dot-print-cutoffs")) {
                dotPrintCutoffs = true;
            } else if(key.equals(";")) {
                if(!marking.isEmpty()) {
                    markings.add(convert(marking));
                    marking.clear();
                }
            } else {
                final int v = value.isEmpty() ? 1 : Integer.parseInt(value);
                for(final Place p : net.getPlaces(key)) {
                    marking.put(p, Integer.valueOf(v));
                }
            }
        }
        
        if(nodeClassHeight >= 1) {
            Backward.NODE_CLASS_TREE_MAX_HEIGHT = nodeClassHeight;
            System.out.println("Using max height for node class trees: " + nodeClassHeight);
        }
        if(getMaxTreeHeight) {
            Backward.GET_MAX_TREE_HEIGHT = true;
        }
        
        
        if(firingSequence != null) {
            SimpleList<Transition> seq = null;
            for(final String name : firingSequence) {
                final Collection<Transition> trans = net.getTransitions(name);
                if(trans.size() > 1) {
                    throw new RuntimeException("Several transitions with name '" + trans + "'.");
                }
                final Transition t = trans.iterator().next();
                seq = new SimpleList<Transition>(t, seq);
            }
            
            final Map<Place, Integer> initial = new IdentityHashMap<Place, Integer>();
            for(final Place p : net.getPlaces()) {
                if(p.getMarking() > 0) {
                    initial.put(p, Integer.valueOf(p.getMarking()));
                }
            }
            System.out.println("Initial marking: " + formatMarking(initial));
            System.out.println("Firing sequence: " + formatTrace(seq));
            if(verbose) {
                bw.setVerbose(true);
            }
            System.out.println("Result marking:  " + formatMarking(bw.fire(initial, seq)));
            
            return;
        }
        
        
        if(!marking.isEmpty()) {
            markings.add(convert(marking));
        } 
        if(markings.isEmpty() && !readTargets) {
            System.out.println("No target markings given. Attempting to read target markings from file.");
            readTargets = true;
        }
        if(readTargets) {
            Backward.readTarget(inputFile, net, markings);
        }
        System.out.println(markings.size() + " target marking(s): " + (reachability > 0 ? formatTarget(markings, printTargets).replaceAll(">=", "=") : formatTarget(markings, printTargets)));
                
        boolean containsExtendedTransition = false;
        for(final Transition t : net.getTransitions()) {
            if(t instanceof ExtendedTransition) {
                containsExtendedTransition = true;
                break;
            }
        }
//        if(containsExtendedTransition && !noBounds) {
//            System.out.println("TRANSFER NETS NOT FULLY SUPPORTED YET.");
////            if(true) return; // TODO
//            noBounds = true;
//            noInv = true;
//            System.out.println("WARNING: Bounds aren't supported for extended Petri nets, yet.\n         (They have been turned off.)");
//        }
        
        final Collection<Place> initial = new ArrayList<Place>();
        final Map<Place, Integer> givenBounds = new IdentityHashMap<Place, Integer>();
        Collection<Map<Place, Integer>> givenInvs = null;
        if(verbose) {
            System.out.print("Initial marking: ");
        }
        boolean containsParameters = false;
        for(final Place p : net.getPlaces()) {
            if(p.getBound() != Place.BOUND_UNKNOWN) {
                if(noGivenBounds) {
                    p.setBound(Place.BOUND_UNKNOWN);
                } else {
                    if(reachability > 0 && (p.getBound() == Place.BOUND_UNBOUNDED || p.getBound() > reachability)) {
                        p.setBound(reachability);
                    }
                    givenBounds.put(p, Integer.valueOf(p.getBound()));
                }
            } else if(!noGivenBounds && reachability > 0) {
                p.setBound(reachability);
                givenBounds.put(p, Integer.valueOf(p.getBound()));
            }
            if(p.getMarking() > 0 || p.isParametric()) {
                if(reachability > 0 && reachability < p.getMarking()) {
                    throw new IllegalArgumentException("Petri net isn't " + reachability + " bounded. Initial marking of " + p + " is " + p.getMarking() + ".");
                }
                initial.add(p);
                if(p.isParametric()) {
                    containsParameters = true;
                    p.setBound(Place.BOUND_UNBOUNDED);
                    givenBounds.put(p, Integer.valueOf(p.getBound()));
                }
            }
            if(verbose) {
                if(p.getMarking() != 0) {
                    System.out.print(p + "=" + p.getMarking() + ", ");
                }
            }
        }
        if(verbose) {
            System.out.println();
        }
        bw.containsExtendedTransition = containsExtendedTransition;
        bw.containsParameters = containsParameters;
        
        bw.setThreadNumber(Runtime.getRuntime().availableProcessors()+1);
        
        if(dotOutput != null) {
            bw.setDotPrintCutoffs(dotPrintCutoffs);
            try {
                bw.setDotOutput(new PrintWriter(new File(dotOutput)));
            } catch (final FileNotFoundException e) {
                throw new RuntimeException(e);
            }
        }
        if(verbose) {
            System.out.println("Verbosity: on");
            bw.setVerbose(verbose);
        }
        if(bmcDepth > 0) {
            System.out.println("BMC depth: " + bmcDepth);
            bw.setBMCDepth(bmcDepth);
        }
        if(noPO) {
            bw.setUsePO(false);
        }
        if(noAccel) {
            bw.setUseAcceleration(false);
        }
        if(noOrder) {
//            throw new UnsupportedOperationException("disabled");
            bw.setUseOrdering(false);
        }
        
        final Map<Place, Place> complements = new IdentityHashMap<Place, Place>();
        if(reachability > 0) {
            // introduce complementary places
            final Collection<Place> pl = new ArrayList<Place>(net.getPlaces());
            final Nodes marks = bw.newNodes();
            for(final Node m : markings) {
                for(final Place p : pl) {
                    final Marking t = m.getMarking(p);
                    if(t == null || t.getMarking() < p.getBound()) {
                        Place c = complements.get(p);
                        if(c == null) {
                            c = net.addPlace("complement-of-" + p.getName(), p.getBound()-p.getMarking());
                            if(c.getMarking() > 0) {
                                initial.add(c);
                            }
                            for(final PTArc a : p.getOutput()) {
                                net.addArc(a.getTarget(), c, a.getWeight());
                            }
                            for(final TPArc a : p.getInput()) {
                                net.addArc(c, a.getSource(), a.getWeight());
                            }
                            complements.put(p, c);
                        }
                        m.add(c, p.getBound() - (t == null ? 0 : t.getMarking()));
                    }
                }
                marks.add(m);
            }
            markings.clear();
            markings.addAll(marks);
        }
        
        time = getMeasuringTime()-time;
        System.out.println("Time used before cov: " + formatTime(time) + " (" + time + " ns)");
        time = getMeasuringTime();
        
        if(noBounds) {
            noInv = true;
            bw.setUseBounds(false);
            bw.setStartState(initial);
        } else {
            System.out.println("Looking for bounds...");
            
//            for(final Transition t : net.getTransitions()) {
//                if(t instanceof ExtendedTransition) {
//                    bw.containsExtendedTransition = true;
//                    if(!noInv) {
//                        System.out.println("Invariants not yet supported for extended Petri nets.");
//                        noInv = true;
//                    }
//                    break;
//                }
//            }
            
            if(!noInv) {
                time = getMeasuringTime();
                System.out.println("Looking for invariants...");
                if(!noGivenInvs) {
                    try {
                        givenInvs = Converter.loadInvariants(inputFile, net);
                    } catch(final Exception e) {
                        // ignore
                        givenInvs = null;
                    }
                }
                final Collection<Map<Place, Integer>> invs = bw.invariantPropagation(net, markings, useSiphonTraps, makeSemiPositive, additionalInvCount, givenInvs);
                if(printInv) {
                    if(invs.isEmpty()) {
                        System.out.println("No invariants found.");
                    } else {
                        System.out.println("Found following " + invs.size() + " invariant(s):");
                        for(final Map<Place, Integer> inv : invs) {
                            int result = 0;
                            for(final Iterator<Map.Entry<Place, Integer>> it = inv.entrySet().iterator(); it.hasNext(); ) {
                                final Map.Entry<Place, Integer> e = it.next();
                                result += e.getValue().intValue()*e.getKey().getMarking();
                                if(e.getValue().intValue() > 1) {
                                    System.out.print(e.getValue());
                                    System.out.print('*');
                                }
                                System.out.print(e.getKey());
                                if(it.hasNext()) {
                                    System.out.print(" + ");
                                }
                            }
                            System.out.print(" = ");
                            System.out.println(result);
                        }
                    }
                } else {
                    System.out.println("Found " + invs.size() + " invariant(s).");
                }
                System.out.println("Time used for invariants: " + formatTime(getMeasuringTime()-time) + " (" + time + " ns)");
                time = getMeasuringTime();            
            }
            
            bw.compute(initial, net.getTransitions());
            
            if(printBounds) {
                System.out.println("Found following bound(s) for " + bw.getBoundedPlaces() + " of " + net.getPlaces().size() + " place(s).");
            } else {
                System.out.println("Found bound(s) for " + bw.getBoundedPlaces() + " of " + net.getPlaces().size() + " place(s).");
            }
            int max = 0;
            int ctrUnbounded = 0;
            int ctrUnknown = 0;
            for(final Place p : net.getPlaces()) {
                if(p.boundKnown()) {
                    if(printBounds) {
                        System.out.print(p + " <= " + p.getBound() + ", ");
                    }
                    if(p.getBound() > max) {
                        max = p.getBound();
                    }
                } else if(p.getBound() == Place.BOUND_UNBOUNDED) {
                    ++ctrUnbounded;
                } else {
                    ++ctrUnknown;
                }
            }
            if(printBounds && bw.getBoundedPlaces() != 0) {
                System.out.println();
            }
            System.out.println("(Maximal bound: " + max + ", " + ctrUnbounded + " unbounded, " + ctrUnknown + " unknown.)");
            System.out.println("Running backward " + (reachability > 0 ? "reachability with bound " + reachability : "coverability") + "...");
        }
        
        
        // run coverability checker!
        final boolean result = bw.isCoverable(markings, shortestTrace);
        
        time = getMeasuringTime()-time;
        final SimpleList<Transition> trace = bw.getResultTrace();
        System.out.println("Time used cov: " + formatTime(time) + " (" + time + " ns)");
        System.out.println("Markings created/examined: " + bw.getMarkingsCreated() + "/" + bw.getMarkingsExamined());
        if(getMaxTreeHeight) {
            System.out.println("Maximal tree height: " + bw.getMaxTreeHeight());
        }
        System.out.println("Given marking " + (reachability > 0 ? "reachable with " + reachability + "-bounded places" : "coverable") + ": " + result);
        if(result) {
            if(reachability > 0) {
                System.out.println("Result trace of " + trace.getLength() + " transitions: " + formatTrace(trace));
                final Node r = bw.fire(bw.getResultTraceInitial(), trace);
                System.out.println("Found trace starts in:    " + formatMarking(bw.getResultTraceInitial(), complements.values())
                               + "\nFound trace ends in:      " + formatMarking(r, complements.values()));
                final Node init = Node.newNode(bw.initialState);
                if(!bw.getResultTraceInitial().equals(init)) {
                    throw new RuntimeException("Used initial state is different from the actual initial state: " + init);
                }
                Backward.sanityCheck(r, markings);
            } else {
                System.out.println("Result trace of " + trace.getLength() + " transitions: " + formatTrace(trace));
                System.out.println("Found trace starts in:    " + formatMarking(bw.getResultTraceInitial(), complements.values())
                               + "\nFound trace ends in:      " + formatMarking(bw.fire(bw.getResultTraceInitial(), trace), complements.values()));
                final Node init = Node.newNode(bw.initialState);
                if(containsParameters) {
                    init.maximize(bw.getResultTraceInitial());
                }
                System.out.println("Actual initial marking:   " + formatMarking(init, complements.values()));
                final Node r = bw.fire(init, trace);
                System.out.println("Actual reachable marking: " + formatMarking(r, complements.values()));
                Backward.sanityCheck(r, markings);
            }
        }
        final Collection<Map<Place, Integer>> givenInvsNet = givenInvs != null ? new ArrayList<Map<Place, Integer>>(givenInvs.size()) : null;
        if((result || all) && (benchmark || runCount > 0 || grain != 0.0 || accurateCount > 0)) {
            final DataRunnable data = new DataRunnable(benchmark && benchmark_maxIt > benchmark_minIt ? benchmark_maxIt : runCount);
            final Thread datarunner = new Thread(data);
            Runtime.getRuntime().addShutdownHook(datarunner);
            final Random rand = new Random();
            
            if(loopStartup > 0) {
                System.out.println("Running startup iterations for " + Backward.formatTime(1000000L*loopStartup) + ".");
                final long sleeptime = loopStartup;
                final AtomicBoolean done = new AtomicBoolean(false);
                (new Thread() {
                    @Override
                    public void run() {
                        try {
                            Thread.sleep(sleeptime);
                        } catch(final InterruptedException e) {
                            Thread.currentThread().interrupt();
                        } finally {
                            done.set(true);
                            bw.pleaseStopComputation();
                        }
                    }
                }).start();
                while(!done.get()) {
                    final PetriNet randomized = new petruchio.pn.PetriNet();
                    final Iterable<Node> target = randomize(net, randomized, initial, markings, givenBounds, givenInvs, givenInvsNet, rand, bw);
                    
                    if(noBounds) {
                        bw.setStartState(initial);
                    } else {
                        bw.compute(initial, randomized.getTransitions());
                        if(!noInv) {
                            bw.invariantPropagation(randomized, target, useSiphonTraps, makeSemiPositive, additionalInvCount, givenInvsNet).size();
                        }
                    }
                    bw.isCoverable(target);
                }
            }
            
            System.out.println("Re-running test with randomized creation order of net elements for " + runCount + " iteration(s).");
            
            long lastInfoPrinted = System.currentTimeMillis();
            boolean useGrain = grain > 0.0 || accurateCount > 0;
            boolean initialInfo = true;
            double oldAverage = -1;
            double curGrain = 0;
            double oldGrain = 0;
            long minT = benchmark_minT*60L*1000L + System.currentTimeMillis();
            int accuracyReached = 0;
            int invs = 0;
            for(int i = 1; i <= runCount || useGrain || (benchmark && minT >= System.currentTimeMillis() && (benchmark_maxIt < benchmark_minIt || i <= benchmark_maxIt)); ++i) {
                if(loopVerbose) {
                    System.out.print("Iteration " + i + ": ");
                }
                
                final PetriNet randomized = new petruchio.pn.PetriNet();
                final Nodes target = randomize(net, randomized, initial, markings, givenBounds, givenInvs, givenInvsNet, rand, bw);
                
                bw.reset();
                
                time = getMeasuringTime();
                
                if(noBounds) {
                    bw.setStartState(initial);
                } else {
                    bw.compute(initial, randomized.getTransitions());
                    if(!noInv) {
                        invs = bw.invariantPropagation(randomized, target, useSiphonTraps, makeSemiPositive, additionalInvCount, givenInvsNet).size();
                    }
                }
                final boolean newResult = bw.isCoverable(target);
                
                time = getMeasuringTime()-time;
                data.time.add(time);
                data.mCreated.add(bw.getMarkingsCreated());
                data.mExamined.add(bw.getMarkingsExamined());
                if(getMaxTreeHeight) {
                    data.maxTreeHeight.add(bw.getMaxTreeHeight());
                }
                if(!noBounds) {
                    data.bounds.add(bw.getBoundedPlaces());
                    if(!noInv) {
                        data.invariants.add(invs);
                    }
                }
                final Node covered;
                if(newResult) {
                    data.tracelength.add(bw.getResultTrace().getLength());
                    final Node init = Node.newNode(bw.initialState);
                    if(containsParameters) {
                        init.maximize(bw.getResultTraceInitial());
                    }
                    covered = Backward.sanityCheck(bw.fire(init, bw.getResultTrace()), target);
                } else {
                    covered = null;
                }
                
                ++data.loops;
                
                if(useGrain) {
                    double avg = data.time.average();
                    oldGrain = curGrain;
                    curGrain = Math.abs(oldAverage - avg);
                    if(curGrain > grain) {
                        accuracyReached = 0;
                    } else {
                        ++accuracyReached;
                    }
                    useGrain = (i <= runCount) || (oldAverage < 0) || accuracyReached < accurateCount;
                    oldAverage = avg;
                }
                
                final boolean printInfo = (initialInfo && System.currentTimeMillis()-lastInfoPrinted > 10L*1000L)
                        || (System.currentTimeMillis()-lastInfoPrinted) > 60L*60L*1000L;
                if(loopVerbose && !printInfo) {
                    System.out.println("time: " + formatTime(time) + " (" + time + " ns), result: " + newResult + " [average. time: " + (data.time.total/(i*1000000L)) +  " s]");
                }
                if(printInfo) {
                    System.out.println("now: " + new java.util.Date());
                    lastInfoPrinted = System.currentTimeMillis();
                    initialInfo = false;
                    if(grain > 0.0) {
                        final int estLoops = (int)Math.abs((oldGrain-curGrain)/grain); 
                        
                        System.out.println("iteration " + i + " of approximately " + estLoops + " more (" + runCount + " fixed)");
                        System.out.println("estimate running time left " + formatTime((long)(1.0*estLoops*data.time.total/i)));
                        System.out.println("estimated end: " + new java.util.Date(System.currentTimeMillis()+(estLoops*data.time.total/(i*1000000L))));
                    } else {
                        System.out.println("iteration " + i + " of " + runCount);
                        System.out.println("estimate running time left " + formatTime((long)(1.0*(runCount-i)*data.time.total/i)));
                        System.out.println("estimated end: " + new java.util.Date(System.currentTimeMillis()+((runCount-i)*data.time.total/(i*1000000L))));
                    }
                    data.run();
                }
                if(result != newResult) {
                    System.out.flush();
                    System.err.println("ERROR: marking SHOULD" + (result ? " " : " NOT ") + "be found to be coverable but in the current run the result was different.");
                    if(newResult) {
                        System.err.println("       Trace:             " + formatTrace(bw.getResultTrace()));
                        System.err.println("       Resulting marking: " + formatMarking(bw.fire(initial, bw.getResultTrace())));
                        System.err.println("       " + (reachability > 0 ? "Reachable" : "Covered") + " marking:   " + formatMarking(covered));
                    }
                    System.err.flush();
                }
            }
            
            data.run();
            Runtime.getRuntime().removeShutdownHook(datarunner);
            
            if(printResultLine) {
                System.out.println(
                        (nodeClassHeightSet ? "{Tree Height} " : "")
                        + "{Iterations} "
                        + data.time.getResultLineHeaders() + " "
                        + data.mCreated.getResultLineHeaders() + " "
                        + data.mExamined.getResultLineHeaders() + " "
                        + (getMaxTreeHeight ? data.maxTreeHeight.getResultLineHeaders() + " " : "")
                        + data.invariants.getResultLineHeaders() + " "
                        + data.bounds.getResultLineHeaders() + " "
                        + data.tracelength.getResultLineHeaders());
                System.out.println(""
                        + (nodeClassHeightSet ? (nodeClassHeight <= 0 ? "{$\\infty$}" : String.valueOf(nodeClassHeight)) + " " : "")
                        + data.loops + " "
                        + data.time.getResultLineValues() + " "
                        + data.mCreated.getResultLineValues() + " "
                        + data.mExamined.getResultLineValues() + " "
                        + (getMaxTreeHeight ? data.maxTreeHeight.getResultLineValues() + " " : "")
                        + data.invariants.getResultLineValues() + " "
                        + data.bounds.getResultLineValues() + " "
                        + data.tracelength.getResultLineValues());
            }
        }
        
        bw.pleaseTerminate();
    }
    
    /**
     * @param printWriter
     */
    public void setDotOutput(final PrintWriter printWriter) {
        this.dotOutput = printWriter;
    }
    
    /**
     * @param dotPrintCutoffs
     */
    public void setDotPrintCutoffs(final boolean dotPrintCutoffs) {
        this.dotPrintCutoffs = dotPrintCutoffs;
    }

    static class DataRunnable implements Runnable {
        static class LongData {
            private static final int ENLARGE_BY = 1000000;
            long[] data;
            long min = Long.MAX_VALUE;
            long max = Long.MIN_VALUE;
            long total = 0;
            int idx = 0;
            int factor;
            String name;
            
            public LongData(final String name, final int count) {
                this(name, count, 1);
            }
            
            public LongData(final String name, final int count, final int factor) {
                this.data = new long[count];
                this.name = name;
                this.factor = factor != 0 ? factor : 1;
            }
            
            public void enlarge() {
                final long[] buf = new long[this.data.length+LongData.ENLARGE_BY];
                System.arraycopy(this.data, 0, buf, 0, this.data.length);
                this.data = buf;
            }
            
            public void add(final long date) {
                if(this.idx == this.data.length) {
                    this.enlarge();
                }
                this.data[this.idx++] = date;
                this.total += date;
                if(date > this.max) { this.max = date; }
                if(date < this.min) { this.min = date; }
            }
            
            public double standardDeviation() {
                if(this.idx == 0) {
                    return 0.0;
                }
                final double avg = this.average();
                double es = 0;
                for(int i = 0; i != this.idx; ++i) {
                    double d = this.data[i] - avg;
                    es += d*d;
                }
                return Math.sqrt(es/this.idx);
            }
            
            public double average() {
                return this.idx == 0 ? 0.0 : 1.0*this.total/this.idx;
            }

            @Override
            public String toString() {
                if(this.min == this.max) {
                    if(this.factor < 0) {
                        return this.name + "\n  min = max = " + (1.0*this.min/-this.factor);
                    }
                    return this.name + "\n  min = max = " + (this.factor*this.min);
                } else if(this.min > this.max) {
                    return this.name + "\n  no data";
                }
                final StringBuffer b = new StringBuffer();
                b.append(this.name);
                if(this.factor < 0) {
                    b.append("\n  avg     ");
                    b.append(this.average()/-this.factor);
                    b.append("\n  std dev ");
                    b.append(this.standardDeviation()/-this.factor);
                    b.append("\n  min     ");
                    b.append(1.0*this.min/-this.factor);
                    b.append("\n  max     ");
                    b.append(1.0*this.max/-this.factor);
                    b.append("\n  total   ");
                    b.append(1.0*this.total/-this.factor);
                } else {
                    b.append("\n  avg     ");
                    b.append(this.factor*this.average());
                    b.append("\n  std dev ");
                    b.append(this.factor*this.standardDeviation());
                    b.append("\n  min     ");
                    b.append(this.factor*this.min);
                    b.append("\n  max     ");
                    b.append(this.factor*this.max);
                    b.append("\n  total   ");
                    b.append(this.factor*this.total);
                }
                return b.toString();
            }
            
            public String getResultLineHeaders() {
                return "{Total "+this.name+"} {Minimum "+this.name+"} {Maximum "+this.name+"} {Average "+this.name+"} {Standard Deviation "+this.name+"}";
            }
            public String getResultLineValues() {
                if(this.min > this.max) {
                    return "nan nan nan nan nan";
                }
                if(this.factor < 0) {
                    return "" + (1.0*this.total/-this.factor) + " "
                              + (1.0*this.min/-this.factor) + " "
                              + (1.0*this.max/-this.factor) + " "
                              + (1.0*this.average()/-this.factor) + " "
                              + (1.0*this.standardDeviation()/-this.factor);
                }
                return "" + (this.total*this.factor) + " "
                          + (this.min*this.factor) + " "
                          + (this.max*this.factor) + " "
                          + (this.average()*this.factor) + " "
                          + (this.standardDeviation()*this.factor);
            }
        }
        
        static class Data {
            private static final int ENLARGE_BY = 1000000;
            int[] data;
            int min = Integer.MAX_VALUE;
            int max = Integer.MIN_VALUE;
            int total = 0;
            int idx = 0;
            int factor;
            String name;
            
            public Data(final String name, final int count) {
                this(name, count, 1);
            }
            
            public Data(final String name, final int count, final int factor) {
                this.data = new int[count];
                this.name = name;
                this.factor = factor != 0 ? factor : 1;
            }
            
            public void enlarge() {
                final int[] buf = new int[this.data.length+Data.ENLARGE_BY];
                System.arraycopy(this.data, 0, buf, 0, this.data.length);
                this.data = buf;
            }
            
            public void add(final int date) {
                if(this.idx == this.data.length) {
                    this.enlarge();
                }
                this.data[this.idx++] = date;
                this.total += date;
                if(date > this.max) { this.max = date; }
                if(date < this.min) { this.min = date; }
            }
            
            public double standardDeviation() {
                if(this.idx == 0) {
                    return 0.0;
                }
                final double avg = this.average();
                double es = 0;
                for(int i = 0; i != this.idx; ++i) {
                    double d = this.data[i] - avg;
                    es += d*d;
                }
                return Math.sqrt(es/this.idx);
            }
            
            public double average() {
                return this.idx == 0 ? 0.0 : 1.0*this.total/this.idx;
            }

            @Override
            public String toString() {
                if(this.min == this.max) {
                    if(this.factor < 0) {
                        return this.name + "\n  min = max = " + (1.0*this.min/-this.factor);
                    }
                    return this.name + "\n  min = max = " + (this.factor*this.min);
                } else if(this.min > this.max) {
                    return this.name + "\n  no data";
                }
                final StringBuffer b = new StringBuffer();
                b.append(this.name);
                if(this.factor < 0) {
                    b.append("\n  avg     ");
                    b.append(this.average()/-this.factor);
                    b.append("\n  std dev ");
                    b.append(this.standardDeviation()/-this.factor);
                    b.append("\n  min     ");
                    b.append(1.0*this.min/-this.factor);
                    b.append("\n  max     ");
                    b.append(1.0*this.max/-this.factor);
                    b.append("\n  total   ");
                    b.append(1.0*this.total/-this.factor);
                } else {
                    b.append("\n  avg     ");
                    b.append(this.factor*this.average());
                    b.append("\n  std dev ");
                    b.append(this.factor*this.standardDeviation());
                    b.append("\n  min     ");
                    b.append(this.factor*this.min);
                    b.append("\n  max     ");
                    b.append(this.factor*this.max);
                    b.append("\n  total   ");
                    b.append(this.factor*this.total);
                }
                return b.toString();
            }
            
            public String getResultLineHeaders() {
                return "{Total "+this.name+"} {Minimum "+this.name+"} {Maximum "+this.name+"} {Average "+this.name+"} {Standard Deviation "+this.name+"}";
            }
            public String getResultLineValues() {
                if(this.min > this.max) {
                    return "nan nan nan nan nan";
                }
                if(this.factor < 0) {
                    return "" + (1.0*this.total/-this.factor) + " "
                              + (1.0*this.min/-this.factor) + " "
                              + (1.0*this.max/-this.factor) + " "
                              + (1.0*this.average()/-this.factor) + " "
                              + (1.0*this.standardDeviation()/-this.factor);
                }
                return "" + (this.total*this.factor) + " "
                          + (this.min*this.factor) + " "
                          + (this.max*this.factor) + " "
                          + (this.average()*this.factor) + " "
                          + (this.standardDeviation()*this.factor);
            }
        }
        
        final LongData time, mCreated, mExamined;
        final Data invariants, bounds, tracelength, maxTreeHeight;
        int loops = 0;
        
        public DataRunnable(final int runCount) {
            this.time = new LongData("Time [ms]", runCount, -1000000);
            this.mCreated = new LongData("M Created", runCount);
            this.mExamined = new LongData("M Examined", runCount);
            this.invariants = new Data("Invariants", runCount);
            this.bounds = new Data("Bounds", runCount);
            this.tracelength = new Data("Trace length", runCount);
            this.maxTreeHeight = new Data("Max tree height", runCount);
        }
        
        @Override
        public void run() {
            System.out.println("Status after " + this.loops + " iteration(s).");
            System.out.println("Now: " + new java.util.Date());
            System.out.println(this.time + "\n  total   " + formatTime(this.time.total));
            System.out.println(this.mCreated + "\n" + this.mExamined + "\n" + this.maxTreeHeight + "\n" + this.invariants + "\n" + this.bounds + "\n" + this.tracelength);
        }
    }
    
    static String formatTime(final long nanoSeconds) {
        if(nanoSeconds == 0L) {
            return "<0.001s";
        }
        long ns = nanoSeconds;
        long rest = ns/60000000000L;
        long dur = 0;
        ns -= 60000000000L*rest;
        final StringBuffer time = new StringBuffer();
        if(rest >= 365*24*60) {
            dur = rest / (365*24*60);
            time.append(dur);
            rest -= dur*365*24*60;
            time.append('y');
        }
        if(rest >= 24*60) {
            dur = rest / (24*60);
            time.append(dur);
            rest -= dur*24*60;
            time.append('d');
        }
        if(rest >= 60) {
            dur = rest / (60);
            time.append(dur);
            rest -= dur*60;
            time.append('h');
        }
        if(rest > 0) {
            dur = rest;
            time.append(dur);
            rest = 0;
            time.append('m');
        }
        if(ns > 0) {
            if(ns >= 1000000000L) {
                time.append(ns / 1000000000L);
                ns %= 1000000000L;
            } else {
                time.append('0');
            }
            if(ns != 0) {
                time.append('.');
                for(long i = 100000000L; i >= 10 && ns != 0; i /= 10L) {
                    long d = ns/i;
                    ns -= i*d;
                    time.append(d);
                }
                if(ns != 0) {
                    time.append(ns);
                }
            }
            time.append('s');
        }
        return time.toString();
    }
    
    /**
     * Randomizes the creation order of places, transitions and arcs, and delivers the support of the new initial marking.
     * Use this method if you want to shuffle your examples to rule out hitting a good/bad path by chance only.
     * 
     * @param net
     * @param targetNet
     * @param root
     * @param marking
     * @param givenBounds
     * @param givenInvs
     * @param rand
     * @return
     */
    private static Nodes randomize(final PetriNet net, final PetriNet targetNet,
            final Collection<Place> root, final Iterable<Node> markings,
            final Map<Place, Integer> givenBounds, final Collection<Map<Place, Integer>> givenInvs,
            final Collection<Map<Place, Integer>> givenInvsNet, final Random rand, final Backward bw) {
        root.clear();
        final Map<Place, Place> pl = new IdentityHashMap<Place, Place>(net.getPlaces().size());
        final LinkedList<Place> places = new LinkedList<Place>(net.getPlaces());
        while(!places.isEmpty()) {
            final Place p = places.remove(rand.nextInt(places.size()));
            
            final Place newP = targetNet.addPlace(p.getName(), p.getMarking());
            pl.put(p, newP);
            if(p.getMarking() > 0 || p.isParametric()) {
                root.add(newP);
            }
            newP.setFinal(p.isFinal());
            newP.setParametric(p.isParametric());
            final Integer b = givenBounds.get(p);
            if(b != null) {
                newP.setBound(b.intValue());
            }
        }
        if(givenInvsNet != null && givenInvs != null) {
            givenInvsNet.clear();
            for(final Map<Place, Integer> inv : givenInvs) {
                final Map<Place, Integer> netInv = new IdentityHashMap<Place, Integer>(inv.size());
                givenInvsNet.add(netInv);
                for(final Map.Entry<Place, Integer> e : inv.entrySet()) {
                    netInv.put(pl.get(e.getKey()), e.getValue());
                }
            }
        }
        
        final LinkedList<Transition> transitions = new LinkedList<Transition>(net.getTransitions());
        while(!transitions.isEmpty()) {
            final Transition t = transitions.remove(rand.nextInt(transitions.size()));
            
            if(t instanceof ExtendedTransition) {
                final petruchio.pn.PetriNet n = (petruchio.pn.PetriNet)targetNet;
                final ExtendedTransition newT = n.addExtendedTransition(t.getName());
                for(final PTArc arc : t.getInput()) {
                    if(arc instanceof ResetArc) {
                        n.addResetArc(pl.get(arc.getSource()), newT, arc.getWeight());
                    } else if(arc instanceof InhibitorArc) {
                        n.addInhibitorArc(pl.get(arc.getSource()), newT).setBound(((InhibitorArc)arc).getBound());
                    } else {
                        targetNet.addArc(pl.get(arc.getSource()), newT, arc.getWeight());
                    }
                }
                for(final TPArc arc : t.getOutput()) {
                    if(arc instanceof TransferArc) {
                        final TransferArc trans = n.addTransferArc(newT, pl.get(arc.getTarget()), arc.getWeight());
                        trans.setBound(((TransferArc)arc).getBound());
                        for(final Map.Entry<Place, Integer> e : ((TransferArc)arc).getTransfer().entrySet()) {
                            trans.addMultiplier(pl.get(e.getKey()), e.getValue().intValue());
                        }
                    } else {
                        targetNet.addArc(newT, pl.get(arc.getTarget()), arc.getWeight());
                    }
                }
            } else {
                final Transition newT = targetNet.addTransition(t.getName());
                for(final PTArc arc : t.getInput()) {
                    targetNet.addArc(pl.get(arc.getSource()), newT, arc.getWeight());
                }
                for(final TPArc arc : t.getOutput()) {
                    targetNet.addArc(newT, pl.get(arc.getTarget()), arc.getWeight());
                }
            }
        }
        Nodes newMarkings = bw.newNodes(); 
        for(final Node marking : markings) {
            final Node newMarking = Node.newNode();
            for(final Marking m : marking.getMarking()) {
                newMarking.add(pl.get(m.getPlace()), m.getMarking());
            }
            newMarkings.add(newMarking);
        }
        return newMarkings;
    }
    
    private static void readTarget(final String file, final PetriNet net, final Nodes markings) {
        final Collection<Map<Place, Integer>> targets = Converter.loadTargetMarkings(file, net);
        for(final Map<Place, Integer> m : targets) {
            markings.add(convert(m));
        }
    }

//    static class Tarjan {
//        static int curindex = 0;
//        static final LinkedList<NetElement> thestack = new LinkedList<NetElement>();
//        static final List<Collection<NetElement>> scc = new LinkedList<Collection<NetElement>>();
//        
//        static class TarjanData {
//            int index = -1;
//            int lowlink = -1;
//            boolean onStack = false;
//        }
//        
//        static final Map<NetElement, TarjanData> data = new IdentityHashMap<NetElement, TarjanData>();
//        
//        static TarjanData getData(final NetElement v) {
//            TarjanData nodeDataV = Tarjan.data.get(v);
//            if(nodeDataV == null) {
//                nodeDataV = new TarjanData();
//                Tarjan.data.put(v, nodeDataV);
//            }
//            return nodeDataV;
//        }
//        
//        static Collection<Collection<NetElement>> tarjan(final NetElement first) {
//            tarjan0(first);
//            final Collection<Collection<NetElement>> result = new ArrayList<Collection<NetElement>>(Tarjan.scc);
//            Tarjan.curindex = 0;
//            Tarjan.thestack.clear();
//            Tarjan.scc.clear();
//            return result;
//        }
//        
//        static void tarjan0(final NetElement v) {
//            final TarjanData nodeDataV = Tarjan.getData(v);
//            nodeDataV.index = Tarjan.curindex;
//            nodeDataV.lowlink = Tarjan.curindex;
//            
//            ++Tarjan.curindex;
//            
//            nodeDataV.onStack = true;
//            Tarjan.thestack.push(v);
//            
//            if(v instanceof Place) {
//                for(final PTArc a : ((Place) v).getOutput()) {
//                    final NetElement n = a.getTarget();
//                    final TarjanData nodeDataN = Tarjan.getData(n);
//                    if(nodeDataN.index == -1) {
//                        tarjan0(n);
//                        nodeDataV.lowlink = Math.min(nodeDataV.lowlink, nodeDataN.lowlink);
//                    } else if(nodeDataN.onStack) {
//                        nodeDataV.lowlink = Math.min(nodeDataV.lowlink, nodeDataN.index);
//                    }
//                }
//            } else {
//                for(final TPArc a : ((Transition) v).getOutput()) {
//                    final NetElement n = a.getTarget();
//                    final TarjanData nodeDataN = Tarjan.getData(n);
//                    if(nodeDataN.index == -1) {
//                        tarjan0(n);
//                        nodeDataV.lowlink = Math.min(nodeDataV.lowlink, nodeDataN.lowlink);
//                    } else if(nodeDataN.onStack){
//                        nodeDataV.lowlink = Math.min(nodeDataV.lowlink, nodeDataN.index);
//                    }
//                }
//            }
//            
//            if(nodeDataV.lowlink == nodeDataV.index) {
//                NetElement n;
//                final Collection<NetElement> component = new ArrayList<NetElement>();
//                do {
//                    n = Tarjan.thestack.pop();
//                    final TarjanData nodeDataN = Tarjan.getData(n);
//                    nodeDataN.onStack = false;
//                    if(n instanceof Transition) {
//                        component.add(n);
//                    }
//                } while(n != v);
//                if(component.size() > 1) {
//                    Tarjan.scc.add(component);
//                }
//            }
//        }
//    }
}
