/*
 * Project:  P-UMLaut
 * Filename: Simulator.java
 * Created:  24.07.2005
 * Author:   Tim Strazny
 */
package petruchio.sim;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.Collections;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import petruchio.sim.petrinettool.IPetriNetTool;
import petruchio.sim.petrinettool.IPlace;
import petruchio.sim.petrinettool.IPlaceType;
import petruchio.sim.petrinettool.PEPReader;
import petruchio.sim.petrinettool.PNMLReader;
import petruchio.sim.petrinettool.PetriNetReader;
import petruchio.sim.pnmodel.net.Arc;
import petruchio.sim.pnmodel.net.Label;
import petruchio.sim.pnmodel.PNModel;
import petruchio.sim.pnmodel.net.PetriNet;
import petruchio.sim.pnmodel.net.Place;
import petruchio.sim.pnmodel.net.Transition;
import petruchio.sim.pnmodel.net.Tuple;
import petruchio.sim.pnmodel.net.Value;
import petruchio.sim.pnmodel.net.ValueList;
import petruchio.sim.pnmodel.util.Binding;
import petruchio.sim.pnmodel.util.IntStack;
import petruchio.sim.pnmodel.util.Function;
import petruchio.sim.pnmodel.util.Node;
import petruchio.sim.pnmodel.util.Pool;
import petruchio.sim.pnmodel.util.RPN;
import petruchio.sim.pnmodel.util.RPNComputer;
import petruchio.sim.pnmodel.util.Token;
import petruchio.sim.pnmodel.util.TreeComputer;
import petruchio.sim.pnmodel.util.Values;


public final class Simulator {
    
    public static enum Status {
        IDLE("idle"),
        RUNNING("running"),
        DEADLOCK("deadlock"),
        COUNT_REACHED("fire count reached"),
        TRANSITION_FIRED("the chosen transition has been fired"),
        MARKING_REACHED("the chosen marking has been reached"),
        INCREMENT_CLOCK("no transitions can be fired at this time - incrementing clock");
        
        private final String msg;
        
        Status(final String msg) {
            this.msg = msg;
        }
        
        @Override
        public String toString() {
            return this.msg;
        }
    };
        
    public static enum Option {
        DEBUG("debug", false),
        SHOW_FIRING("show_firing", false),
        SHOW_SIMPLIFICATIONS("verbose", false),
        GLOBAL_PRIORITY("global_priority", true),
        USE_STRICT_TIME("use_strict_time", true),
        USE_TIME("use_time", false),
        USE_PRIORITY("use_priority", false),
        USE_WEIGHTS("use_weights", false),
        USE_INHIBITOR("use_inhibitors", false),
        USE_READ("use_read_arcs", false),
        USE_RESET("use_reset_arcs", false),
        USE_CAPACITY("use_capacities", false),
        DISTRIBUTE_AGE_MEM("distribute_age_mem", true),
        CHECK_FIRE("check_fire", false),
        FIRE_PARALLEL("use_parallel_firing", false),
        USE_STOP_CONDITION("use_stop_condition", false),
        EXTENDED_HL_LL_REF("extended_hl_ll_ref", false),
        MEASURE_ALL_PLACES_EPOCHS("measure_all_places_epochs", false),
        MEASURE_ALL_PLACES_TIME("measure_all_places_time", false),
        MINIMIZE_NET("minimize_net", true);

        private static Map<String, Option> strings;
        
        public static Option getOption(final String key) {
            return Option.strings.get(key);
        }
        
        private final String key;
        private final boolean defaultValue;
        
        Option(final String key, final boolean defaultValue) {
            this.ensureUnique(key);
            this.key = key;
            this.defaultValue = defaultValue;
        }
        
        private void ensureUnique(final String theKey) {
            if(Option.strings == null) {
                Option.strings =  Pool.getPool().getMap();
            }
            if(Option.strings.containsKey(theKey)) {
                throw new IllegalArgumentException("There already exists an "
                                + "option with key '" + theKey + "'.");
            }
            Option.strings.put(theKey, this);
        }
        
        public String getKey() {
            return this.key;
        }
        
        public boolean getDefault() {
            return this.defaultValue;
        }
        
        @Override
        public String toString() {
            return this.key;
        }
    }

    public static final int OFFSET_X_TRANS = 2;
    public static final int OFFSET_Y_TRANS = 2;
    
    public static final int OFFSET_X_PLACE = 2;
    public static final int OFFSET_Y_PLACE = 2;
    
    public static final Value PROPERTY_TIME = Value.getStringValue("time");
    
    public static final boolean DFLT_ALL_BINDINGS = true;
    
    private static final Random random = new Random();
    
    
    private String checkTrans = "";

    private PetriNet net;
    
    private final boolean[] options;
    
    private List<Binding>[] bindings;
    
    private List<Binding>[] firing;
    
    private Map<Value, Integer>[] tokenCountEpochs;
    
    private Map<Value, Double>[] tokenCountTime;
    
    private final Map<Transition, Transition> enabledTrans;
    
    private final Map<Transition, Transition> firingTrans;
    
    private final List<Binding> putBindingsEarliest;
    
    private final List<Binding> putBindingsLatest;
    
    private final List<String> neverFired;
    
    private final List<Binding> bindingsImmediate;
    
    private final List<Binding> bindingsLatest;
    
    private final List<Binding> bindingsEarliest;
    
    private final Map<String, Function> functions;
    
    private final Map<Value, Value> properties;
    
    private ValueList[] markings;
    
    private Map<Thread, ValueList[]> threadMarkings = new IdentityHashMap<Thread, ValueList[]>();
    
    private IntStack used;
    private Map<Thread, IntStack> threadUsed = new IdentityHashMap<Thread, IntStack>();

    private PetriNetReader pepreader;
    
    private PetriNetReader pnmlreader;
    
    private PNModel tool;
    
    private String buildHL2LLRef;
    
    private int fired;
    
    private final List<Place> measurePlacesEpochs;
    
    private final List<Place> measurePlacesTime;
    
    private final Collection<String> stopAtTransName;
    
    private final Map<Place, ValueList> stopAtMarking;
    
    private final List<String> sequence;
    
    private int maxSeqLength;
    
    private int stopAtCount;
    
    private Status state;
    
    private double time;
    
    private Binder[] binders = null;
    
    public Simulator() {
        final Option[] opts = Simulator.Option.values();
        this.options = new boolean[opts.length];
        for(int i = 0; i < opts.length; ++i) {
            this.options[opts[i].ordinal()] = opts[i].getDefault();
        }
        this.fired = 0;
        this.stopAtMarking = new IdentityHashMap<Place, ValueList>();
        this.stopAtCount = 0;
        this.stopAtTransName = Pool.getPool().getList();
        this.state = Simulator.Status.IDLE;
        
        this.pepreader = new PEPReader();
        this.pnmlreader = new PNMLReader();
        this.tool = new PNModel();
        this.enabledTrans = new IdentityHashMap<Transition, Transition>();
        this.firingTrans = new IdentityHashMap<Transition, Transition>();
        this.neverFired = new petruchio.sim.pnmodel.util.StringList();
        this.putBindingsEarliest = Pool.getPool().getList();
        this.putBindingsLatest = Pool.getPool().getList();
        this.sequence = Pool.getPool().getList();
        this.measurePlacesEpochs = Pool.getPool().getList();
        this.measurePlacesTime = Pool.getPool().getList();
        this.maxSeqLength = 0;
        this.functions = Pool.getPool().getMap();
        this.properties = Pool.getPool().getMap();
        this.bindingsImmediate = Pool.getPool().getList();
        this.bindingsLatest = Pool.getPool().getList();
        this.bindingsEarliest = Pool.getPool().getList();
        this.net = null;
        this.time = 0.0f;
    }
    
    public void measurePlaceExpochs(final String place) {
        if(!this.getProperty(Simulator.Option.MEASURE_ALL_PLACES_EPOCHS)) {
            final Place p = this.net.getPlaceByName(place);
            if((p != null) && !this.measurePlacesEpochs.contains(p)) {
                this.measurePlacesEpochs.add(p);
            }
        }
    }
    
    
    public void measureAllPlacesEpochs() {
        if(!this.getProperty(Simulator.Option.MEASURE_ALL_PLACES_EPOCHS)) {
            this.setProperty(Simulator.Option.MEASURE_ALL_PLACES_EPOCHS, true);
            this.measurePlacesEpochs.clear();
            this.measurePlacesEpochs.addAll(this.net.getPlaceCollection());
        }
    }
    
    public void measurePlaceTime(final String place) {
        if(!this.getProperty(Simulator.Option.MEASURE_ALL_PLACES_TIME)) {
            final Place p = this.net.getPlaceByName(place);
            if((p != null) && !this.measurePlacesTime.contains(p)) {
                this.measurePlacesTime.add(p);
            }
        }
    }
    
    
    public void measureAllPlacesTime() {
        if(!this.getProperty(Simulator.Option.MEASURE_ALL_PLACES_TIME)) {
            this.setProperty(Simulator.Option.MEASURE_ALL_PLACES_TIME, true);
            this.measurePlacesTime.clear();
            this.measurePlacesTime.addAll(this.net.getPlaceCollection());
        }
    }
    
    public boolean getProperty(final Simulator.Option option) {
        return this.options[option.ordinal()];
    }
    
    
    public void setProperty(final Simulator.Option option, final boolean value) {
        this.options[option.ordinal()] = value;
    }
    
    
    public long getFiredCount() {
        return this.fired;
    }
    
    
    private void removeBindings(final List<Binding> binds) {
        for(int i = 0, n = binds.size(); i < n; ++i) {
            this.removeBinding(binds.get(i));
        }
    }
    
    
    private void removeBinding(final Binding bind) {
        if(bind.isImmediate()) {
            if(!this.removeFromImmediateBindings(bind, false)) {
                throw new RuntimeException("Error removing " + bind
                                + " from list of immediate bindings: "
                                + this.bindingsImmediate);
            }
        } else {
            if(!this.removeFromEarliestBindings(bind, false)) {
                throw new RuntimeException("Error removing " + bind
                                + " from list of earliest bindings: "
                                + this.bindingsEarliest);
            }
            if(!this.removeFromLatestBindings(bind, false)) {
                throw new RuntimeException("Error removing " + bind
                                + " from list of latest bindings: "
                                + this.bindingsLatest);
            }
        }
        
        final Transition t = bind.getTransition();
        final List<Binding> theBinds = this.getBindings(t);
        // to avoid recursion removeBindings(Transition) removes this
        // list before calling removeBinding(Binding)
        if(theBinds != null) {
            Simulator.removeSortEarliestFiring(theBinds, bind, false);
            if(theBinds.isEmpty()) {
                this.removeBindings(t); //also frees "theBinds"
            }
        }
    }
    
    
    public static boolean removeSortLatestFiring(final List<Binding> list,
                    final Binding bind, final boolean equal) {
        synchronized(list) {
            final int index = Simulator.indexOfSortLatest(list, bind, equal);
            
            if(index < 0) {
                return false;
            } 
            list.remove(index);
            return true;
        }
    }
    
    
    public static boolean removeSortEarliestFiring(final List<Binding> list,
                    final Binding bind, final boolean equal) {
        synchronized(list) {
            final int index = Simulator.indexOfSortEarliest(list, bind, equal);
            
            if(index < 0) {
                return false;
            } else {
                list.remove(index);
                return true;
            }
        }
    }
    
    
    public static int indexOfSortEarliest(final List<Binding> list,
                    final Binding bind, final boolean equal) {
        // binaeres suchen des zu loeschenden bindings
        final double e = bind.getEarliestFiring();
        final double p = bind.getPriority();
        final double w = bind.getCurrentWeight();
        final int h = bind.getObjectNumber();
        
        boolean found = false;
        int from = 0;
        int to = list.size()-1;
        while(from <= to) {
            int middle = (from+to)>>1;
            final Binding c = list.get(middle);
            final double cp = c.getPriority();
            final int ch = c.getObjectNumber();
            if((ch == h) || (equal && (p == cp) && bind.equals(c))) {
                found = true;
                from = middle;
                to = from-1;
            } else {
                final double ce = c.getEarliestFiring();
                if(ce == e) {
                    if(cp == p) {
                        final double cw = c.getCurrentWeight();
                        if(equal || (cw == w)) {
                            if((ch == h) || (equal && (from == to))) {
                                // zu loeschendes element gefunden
                                // von der mitte aus in beide richtungen suchen
                                if(equal && (ch != h)) {
                                    int pos;
                                    for(pos = middle+1; pos <= to; ++pos) {
                                        if(list.get(pos).equals(bind)) {
                                            found = true;
                                            break;
                                        }
                                    }
                                    if(!found) {
                                        for(pos = middle-1; pos >= from; --pos) {
                                            if(list.get(pos).equals(bind)) {
                                                found = true;
                                                break;
                                            }
                                        }
                                    }
                                    from = pos;
                                } else {
                                    from = middle;
                                }
                                // schleife beenden
                                to = from-1;
                            } else if(ch > h) {
                                // zu loeschendes element rechts von der mitte
                                from = middle+1;
                            } else {
                                // zu loeschendes element links von der mitte
                                to = middle-1;
                            }
                        } else if(cw > w) {
                            // zu loeschendes element rechts von der mitte
                            from = middle+1;
                        } else {
                            // zu loeschendes element links von der mitte
                            to = middle-1;
                        }
                    } else if(cp > p) {
                        // zu loeschendes element rechts von der mitte
                        from = middle+1;
                    } else {
                        // zu loeschendes element links von der mitte
                        to = middle-1;
                    }
                } else if(ce < e) {
                    // zu loeschendes element rechts von der mitte
                    from = middle+1;
                } else {
                    // zu loeschendes element links von der mitte
                    to = middle-1;
                }
            }
        }
        
        if(!found) {
            return -1;
        }
        return from;
    }
    
    
    public static int indexOfSortLatest(final List<Binding> list,
                    final Binding bind, final boolean equal) {
        // binaeres suchen des zu loeschenden bindings
        final double e = bind.getLatestFiring();
        final double p = bind.getPriority();
        final double w = bind.getCurrentWeight();
        final int h = bind.getObjectNumber();
        
        boolean found = false;
        int from = 0;
        int to = list.size()-1;
        while(from <= to) {
            int middle = (from+to)>>1;
            final Binding c = list.get(middle);
            final double cp = c.getPriority();
            final int ch = c.getObjectNumber();
            if((ch == h) || (equal && (p == cp) && bind.equals(c))) {
                found = true;
                from = middle;
                to = from-1;
            } else {
                final double ce = c.getLatestFiring();
                if(ce == e) {
                    if(cp == p) {
                        final double cw = c.getCurrentWeight();
                        if(equal || (cw == w)) {
                            if((ch == h) || (equal && (from == to))) {
                                // zu loeschendes element gefunden
                                // von der mitte aus in beide richtungen suchen
                                if(equal && (ch != h)) {
                                    int pos;
                                    for(pos = middle+1; pos <= to; ++pos) {
                                        if(list.get(pos).equals(bind)) {
                                            found = true;
                                            break;
                                        }
                                    }
                                    if(!found) {
                                        for(pos = middle-1; pos >= from; --pos) {
                                            if(list.get(pos).equals(bind)) {
                                                found = true;
                                                break;
                                            }
                                        }
                                    }
                                    from = pos;
                                } else {
                                    from = middle;
                                }
                                // schleife beenden
                                to = from-1;
                            } else if(ch > h) {
                                // zu loeschendes element rechts von der mitte
                                from = middle+1;
                            } else {
                                // zu loeschendes element links von der mitte
                                to = middle-1;
                            }
                        } else if(cw > w) {
                            // zu loeschendes element rechts von der mitte
                            from = middle+1;
                        } else {
                            // zu loeschendes element links von der mitte
                            to = middle-1;
                        }
                    } else if(cp > p) {
                        // zu loeschendes element rechts von der mitte
                        from = middle+1;
                    } else {
                        // zu loeschendes element links von der mitte
                        to = middle-1;
                    }
                } else if(ce < e) {
                    // zu loeschendes element rechts von der mitte
                    from = middle+1;
                } else {
                    // zu loeschendes element links von der mitte
                    to = middle-1;
                }
            }
        }
        
        if(!found) {
            return -1;
        }
        return from;
    }
    
    
    public static int indexOfSortPriority(final List<Binding> list,
                    final Binding bind, final boolean equal) {
        // binaeres suchen des zu loeschenden bindings
        final double p = bind.getPriority();
        final double w = bind.getCurrentWeight();
        final double h = bind.getObjectNumber();
        
        boolean found = false;
        int from = 0;
        int to = list.size()-1;
        while(from <= to) {
            int middle = (from+to)>>1;
            final Binding c = list.get(middle);
            final int ch = c.getObjectNumber();
            final double cp = c.getPriority();
            if((ch == h) || (equal && (p == cp) && bind.equals(c))) {
                found = true;
                from = middle;
                to = from-1;
            } else if(cp == p) {
                final double cw = c.getCurrentWeight();
                if(equal || (cw == w)) {
                    if((ch == h) || (equal && (from == to))) {
                        // zu loeschendes element gefunden
                        // von der mitte aus in beide richtungen suchen
                        if(equal) {
                            int pos;
                            for(pos = middle+1; pos <= to; ++pos) {
                                if(list.get(pos).equals(bind)) {
                                    found = true;
                                    break;
                                }
                            }
                            if(!found) {
                                for(pos = middle-1; pos >= from; --pos) {
                                    if(list.get(pos).equals(bind)) {
                                        found = true;
                                        break;
                                    }
                                }
                            }
                            from = pos;
                        } else {
                            from = middle;
                        }
                        // schleife beenden
                        to = from-1;
                    } else if(ch > h) {
                        // zu loeschendes element rechts von der mitte
                        from = middle+1;
                    } else {
                        // zu loeschendes element links von der mitte
                        to = middle-1;
                    }
                } else if(cw > w) {
                    // zu loeschendes element rechts von der mitte
                    from = middle+1;
                } else {
                    // zu loeschendes element links von der mitte
                    to = middle-1;
                }
            } else if(cp > p) {
                // zu loeschendes element rechts von der mitte
                from = middle+1;
            } else {
                // zu loeschendes element links von der mitte
                to = middle-1;
            }
        }
        
        if(!found) {
            return -1;
        }
        return from;
    }
    
    
    public static boolean removeSortPriority(final List<Binding> list,
                    final Binding bind, final boolean equal) {
        synchronized(list) {
            final int index = Simulator.indexOfSortPriority(list, bind, equal);
            
            if(index < 0) {
                return false;
            }
            list.remove(index);
            return true;
        }
    }
    
    
    private boolean removeFromLatestBindings(final Binding bind, final boolean equal) {
        return Simulator.removeSortLatestFiring(this.bindingsLatest, bind, equal);
    }
    
    
    private boolean removeFromEarliestBindings(final Binding bind, final boolean equal) {
        return Simulator.removeSortEarliestFiring(this.bindingsEarliest, bind, equal);
    }
    
    
    private boolean removeFromImmediateBindings(final Binding bind, final boolean equal) {
        return Simulator.removeSortPriority(this.bindingsImmediate, bind, equal);
    }
    
    
//    private void insertToImmediateBindings(final List<Binding> binds) {
//        this.insertToImmediateBindings(binds, false);
//    }
    
    
    private void insertToImmediateBindings(final List<Binding> binds, final boolean sorted) {
        // absteigend nach priority
        if(!binds.isEmpty()) {
            Simulator.insertSortPriority(this.bindingsImmediate, binds, sorted);
        }
    }
    
    
//    private void insertToLatestBindings(final List<Binding> binds) {
//        this.insertToLatestBindings(binds, false);
//    }
    
    
    private void insertToLatestBindings(final List<Binding> binds, final boolean sorted) {
        // aufsteigend nach latest, dann absteigend nach priority
        if(!binds.isEmpty()) {
            synchronized(this.bindingsLatest) {
                Simulator.insertSortLatestFiring(this.bindingsLatest, binds, sorted);
            }
        }
    }
    
    
//    private void insertToEarliestBindings(final List<Binding> binds) {
//        this.insertToEarliestBindings(binds, false);
//    }
    
    
    private void insertToEarliestBindings(final List<Binding> binds, final boolean sorted) {
        // aufsteigend nach earliest, dann absteigend nach priority
        if(!binds.isEmpty()) {
            synchronized(this.bindingsLatest) {
                Simulator.insertSortEarliestFiring(this.bindingsEarliest, binds, sorted);
            }
        }
    }
    
    
    public static void insertSortPriority(final List<Binding> list, final List<Binding> binds) {
        Simulator.insertSortPriority(list, binds, false);
    }
    
    
    public static void insertSortPriority(final List<Binding> list, final List<Binding> binds, final boolean sorted) {
        int lastPos = 0;
        for(int i = 0, n = binds.size(); i < n; ++i) {
            final Binding bind = binds.get(i);
            if(sorted) {
                lastPos = 1 + Simulator.insertSortPriority(list, bind, lastPos);
            } else {
                Simulator.insertSortPriority(list, bind);
            }
        }
    }
    
    
    public static int insertSortPriority(final List<Binding> list, final Binding bind) {
        return Simulator.insertSortPriority(list, bind, 0);
    }
    
    
    public static int insertSortPriority(final List<Binding> list, final Binding bind, final int start) {
        synchronized(list) {
            // sortieren durch einfuegen
            // binaeres suchen der einfuegestelle
            final double p = bind.getPriority();
            final double w = bind.getCurrentWeight();
            final int h = bind.getObjectNumber();
            
            int from = start;
            int to = list.size()-1;
            while(from <= to) {
                int middle = (from+to)>>1;
                final Binding c = list.get(middle);
                final double cp = c.getPriority();
                if(cp == p) {
                    final double cw = c.getCurrentWeight();
                    if(cw == w) {
                        final int ch = c.getObjectNumber();
                        if(ch == h) {
                            // einfuegestelle gefunden
                            // schleife beenden
                            from = middle;
                            to = from-1;
                        } else if(ch > h) {
                            // einfuegestelle rechts von der mitte
                            from = middle+1;
                        } else {
                            // einfuegestelle links von der mitte
                            to = middle-1;
                        }
                    } else if(cw > w) {
                        // einfuegestelle rechts von der mitte
                        from = middle+1;
                    } else {
                        // einfuegestelle links von der mitte
                        to = middle-1;
                    }
                } else if(cp > p) {
                    // einfuegestelle rechts von der mitte
                    from = middle+1;
                } else {
                    // einfuegestelle links von der mitte
                    to = middle-1;
                }
            }
            
            list.add(from, bind);
            return from;
        }
    }
    
    
    public static void insertSortEarliestFiring(final List<Binding> list, final List<Binding> binds) {
        Simulator.insertSortEarliestFiring(list, binds, false);
    }
    
    
    public static void insertSortEarliestFiring(final List<Binding> list, final List<Binding> binds, final boolean sorted) {
        int lastPos = 0;
        for(int i = 0, n = binds.size(); i < n; ++i) {
            final Binding bind = binds.get(i);
            if(sorted) {
                lastPos = 1 + Simulator.insertSortEarliestFiring(list, bind, lastPos);
            } else {
                Simulator.insertSortEarliestFiring(list, bind);
            }
        }
    }
    
    
    public static int insertSortEarliestFiring(final List<Binding> list, final Binding bind) {
        return Simulator.insertSortEarliestFiring(list, bind, 0);
    }
    
    
    public static int insertSortEarliestFiring(final List<Binding> list, final Binding bind, final int start) {
        synchronized(list) {
            // sortieren durch einfuegen
            // binaeres suchen der einfuegestelle
            final boolean i = bind.isImmediate();
            final double e = bind.getEarliestFiring();
            final double p = bind.getPriority();
            final double w = bind.getCurrentWeight();
            final int h = bind.getObjectNumber();
            
            int from = start;
            int to = list.size()-1;
            while(from <= to) {
                int middle = (from+to)>>1;
                final Binding c = list.get(middle);
                final double ce = c.getEarliestFiring();
                final boolean ci = c.isImmediate();
                
                if((ci && i) || (!i && (ce == e))) {
                    final double cp = c.getPriority();
                    if(cp == p) {
                        final double cw = c.getCurrentWeight();
                        if(cw == w) {
                            final int ch = c.getObjectNumber();
                            if(ch == h) {
                                // einfuegestelle gefunden
                                // schleife beenden
                                from = middle;
                                to = from-1;
                            } else if(ch > h) {
                                // einfuegestelle rechts von der mitte
                                from = middle+1;
                            } else {
                                // einfuegestelle links von der mitte
                                to = middle-1;
                            }
                        } else if(cw > w) {
                            // einfuegestelle rechts von der mitte
                            from = middle+1;
                        } else {
                            // einfuegestelle links von der mitte
                            to = middle-1;
                        }
                    } else if(cp > p) {
                        // einfuegestelle rechts von der mitte
                        from = middle+1;
                    } else {
                        // einfuegestelle links von der mitte
                        to = middle-1;
                    }
                } else if(!i && (ce < e)) {
                    // einfuegestelle rechts von der mitte
                    from = middle+1;
                } else {
                    // einfuegestelle links von der mitte
                    to = middle-1;
                }
            }
            
            list.add(from, bind);
            return from;
        }
    }
    
    
    public static void insertSortLatestFiring(final List<Binding> list, final List<Binding> binds) {
        Simulator.insertSortLatestFiring(list, binds, false);
    }
    
    
    public static void insertSortLatestFiring(final List<Binding> list, final List<Binding> binds, final boolean sorted) {
        int lastPos = 0;
        for(int i = 0, n = binds.size(); i < n; ++i) {
            final Binding bind = binds.get(i);
            if(sorted) {
                lastPos = 1 + Simulator.insertSortLatestFiring(list, bind, lastPos);
            } else {
                Simulator.insertSortLatestFiring(list, bind);
            }
        }
    }
    
    
    public static int insertSortLatestFiring(final List<Binding> list, final Binding bind) {
        return Simulator.insertSortLatestFiring(list, bind, 0);
    }
    
    
    public static int insertSortLatestFiring(final List<Binding> list, final Binding bind, final int start) {
        synchronized(list) {
            // sortieren durch einfuegen
            // binaeres suchen der einfuegestelle
            final double e = bind.getLatestFiring();
            final int h = bind.getObjectNumber();
            
            int from = start;
            int to = list.size()-1;
            while(from <= to) {
                int middle = (from+to)>>1;
                final Binding c = list.get(middle);
                final double ce = c.getLatestFiring();
                if(ce == e) {
                    final int ch = c.getObjectNumber();
                    if(ch == h) {
                        // einfuegestelle gefunden
                        // schleife beenden
                        from = middle;
                        to = from-1;
                    } else if(ch > h) {
                        // einfuegestelle rechts von der mitte
                        from = middle+1;
                    } else {
                        // einfuegestelle links von der mitte
                        to = middle-1;
                    }
                } else if(ce < e) {
                    // einfuegestelle rechts von der mitte
                    from = middle+1;
                } else {
                    // einfuegestelle links von der mitte
                    to = middle-1;
                }
            }
            
            list.add(from, bind);
            return from;
        }
    }
    
    
    public void setFiresParallel(final boolean v) {
        this.setProperty(Simulator.Option.FIRE_PARALLEL, v);
    }
    
    
    public boolean firesParallel() {
        return this.getProperty(Simulator.Option.FIRE_PARALLEL);
    }
    
    
    public void setMinimizeNet(final boolean v) {
        this.setProperty(Simulator.Option.MINIMIZE_NET, v);
    }
    
    
    public boolean minimizeNet() {
        return this.getProperty(Simulator.Option.MINIMIZE_NET);
    }
    
    
    public void setDistributeTime(final boolean v) {
        this.setProperty(Simulator.Option.DISTRIBUTE_AGE_MEM, v);
    }
    
    
    public boolean distributeTime() {
        return this.getProperty(Simulator.Option.DISTRIBUTE_AGE_MEM);
    }
    
    
    public void setUsesTime(final boolean v) {
        this.setProperty(Simulator.Option.USE_TIME, v);
    }
    
    
    public boolean usesTime() {
        return this.getProperty(Simulator.Option.USE_TIME);
    }
    
    
    public void setUsesStrictTime(final boolean v) {
        this.setProperty(Simulator.Option.USE_STRICT_TIME, v);
    }
    
    
    public boolean usesStrictTime() {
        return this.getProperty(Simulator.Option.USE_STRICT_TIME);
    }
    
    
    public void setUsesCapacity(final boolean v) {
        this.setProperty(Simulator.Option.USE_CAPACITY, v);
    }
    
    
    public boolean usesCapacity() {
        return this.getProperty(Simulator.Option.USE_CAPACITY);
    }
    
    
    public void setUsesInhibitorArcs(final boolean v) {
        this.setProperty(Simulator.Option.USE_INHIBITOR, v);
    }
    
    
    public boolean usesInhibitorArcs() {
        return this.getProperty(Simulator.Option.USE_INHIBITOR);
    }
    
    
    public void setUsesReadArcs(final boolean v) {
        this.setProperty(Simulator.Option.USE_READ, v);
    }
    
    
    public boolean usesReadArcs() {
        return this.getProperty(Simulator.Option.USE_READ);
    }
    
    
    public void setUsesResetArcs(final boolean v) {
        this.setProperty(Simulator.Option.USE_RESET, v);
    }
    
    
    public boolean usesResetArcs() {
        return this.getProperty(Simulator.Option.USE_RESET);
    }
    
    
    public void setUsesWeights(final boolean v) {
        this.setProperty(Simulator.Option.USE_WEIGHTS, v);
    }
    
    
    public boolean usesWeights() {
        return this.getProperty(Simulator.Option.USE_WEIGHTS);
    }
    
    
    public void setCheckFire(final boolean v) {
        this.setProperty(Simulator.Option.CHECK_FIRE, v);
    }
    
    
    public boolean checkFire() {
        return this.getProperty(Simulator.Option.CHECK_FIRE);
    }
    
    
    public void setShowSimplifications(final boolean v) {
        this.setProperty(Simulator.Option.SHOW_SIMPLIFICATIONS, v);
    }
    
    
    public boolean showSimplifications() {
        return this.getProperty(Simulator.Option.SHOW_SIMPLIFICATIONS);
    }
    
    
    public void setUseGlobalPriority(final boolean v) {
        this.setProperty(Simulator.Option.GLOBAL_PRIORITY, v);
    }
    
    
    public boolean usesGlobalPriority() {
        return this.getProperty(Simulator.Option.GLOBAL_PRIORITY);
    }
    
    
    public void setUsePriority(final boolean v) {
        this.setProperty(Simulator.Option.USE_PRIORITY, v);
    }
    
    
    public boolean usesPriority() {
        return this.getProperty(Simulator.Option.USE_PRIORITY);
    }
    
    
    public double getTime() {
        return this.time;
    }
    
    
    public double getMinimalEarliestPossibleFiring() {
        if(this.bindingsImmediate.isEmpty()) {
            final Binding first = this.bindingsEarliest.get(0);
            return first.getEarliestFiring();
        }
        return 0.0;
    }
    
    public double getMinimalLatestPossibleFiring() {
        if(this.bindingsImmediate.isEmpty()) {
            final Binding first = this.bindingsLatest.get(0);
            return first.getLatestFiring();
        }
        return 0.0;
    }
    
    
    public String incTime() {
        final double lo = this.getMinimalEarliestPossibleFiring();
        final double hi = this.getMinimalLatestPossibleFiring();
        final double delta;
        if(hi == Double.POSITIVE_INFINITY) {
            delta = lo - this.getTime() + Simulator.random.nextDouble();
        } else {
            delta = lo - this.getTime() + Simulator.random.nextDouble()*(hi-lo);
        }
        return this.incTime(delta);
    }
    
    
    public String incTime(final double ticks) {
        if(this.usesStrictTime()) {
            return this.incTimeStrict(ticks);
        }
        return this.incTimeWeak(ticks);
    }
    
    
    public String incTimeStrict(final double ticks) {
        final double newTime = this.getTime() + ticks;

        if(!this.bindingsImmediate.isEmpty()) {
            this.checkState(null);
            return "Will not increment clock. Some immediate transitions have to be fired.\n";
        } else if(!this.putBindingsEarliest.isEmpty()) {
            return "Will not increment clock. Some transition firings (put) have to be completed.\n";
        }
        
        final Binding first = this.bindingsLatest.get(0);
        if(newTime > first.getLatestFiring()) {
            this.checkState(null);
            return "Will not increment clock. Some timed transitions have to be fired.\n";
        }
        
        final StringBuilder buf = new StringBuilder();

        this.addTokenCountTime(ticks);
        
        if(this.firesParallel()) {
            for(int i = 0, n = this.putBindingsEarliest.size(); i < n; ++i) {
                final Binding bind = this.putBindingsEarliest.get(i);
                if(newTime > bind.getLatestFiring()) {
                    buf.append("Put: ");
                    buf.append(this.firePut(bind.getTransition(), bind));
                    buf.append("\n");
                }
            }
        }
        
        this.time = newTime;
        
        this.checkState(null);
        
        return buf.toString();
    }
    
    
    public String incTime_old(final double ticks) {
        final double newTime = this.time + ticks;
        
        final StringBuilder buf = new StringBuilder();
        final List<Binding> fire = Pool.getPool().getList();
        do {
            fire.clear();
            for(int i = 0, n = this.bindingsLatest.size(); i < n; ++i) {
                final Binding bind = this.bindingsLatest.get(i);
                if(newTime > bind.getLatestFiring()) {
                    fire.add(bind);
                }
            }
            if(!fire.isEmpty()) {
                final Binding bind = Simulator.choseRandom(fire);
                if(this.firesParallel()) {
                    buf.append("Take: ");
                    buf.append(this.fireTake(bind.getTransition(), bind));
                    buf.append("\n");
                } else {
                    Transition trans = bind.getTransition();
                    buf.append("Fire: ");
                    buf.append(this.fire(trans, bind, trans.getInput(), trans.getOutput()));
                    buf.append("\n");
                }
            }
        } while(!fire.isEmpty());
        
        if(this.firesParallel()) {
            for(int i = 0, n = this.putBindingsEarliest.size(); i < n; ++i) {
                final Binding bind = this.putBindingsEarliest.get(i);
                if(newTime > bind.getLatestFiring()) {
                    buf.append("Put:  ");
                    buf.append(this.firePut(bind.getTransition(), bind));
                    buf.append("\n");
                }
            }
        }
        
        this.time = newTime;
        
        this.checkState(null);
        
        return buf.toString();
    }
    
    
    public String incTimeWeak(final double ticks) {
        if(!this.putBindingsEarliest.isEmpty()) {
            return "Will not increment clock. Some transition firings (put) have to be completed.\n";
        }
        
        final double newTime = this.getTime() + ticks;
        
        // remove obsolete bindings
        final List<Binding> delete = Pool.getPool().getList();
        for(int i = 0, n = this.bindingsLatest.size(); i < n; ++i) {
            final Binding bind = this.bindingsLatest.get(i);
            if(newTime > bind.getLatestFiring()) {
                delete.add(bind);
            }
        }
        this.removeBindings(delete);
        
        this.addTokenCountTime(ticks);
        
        final StringBuilder buf = new StringBuilder();
        if(this.firesParallel()) {
            for(int i = 0, n = this.putBindingsEarliest.size(); i < n; ++i) {
                final Binding bind = this.putBindingsEarliest.get(i);
                if(newTime > bind.getLatestFiring()) {
                    buf.append("Put: ");
                    buf.append(this.firePut(bind.getTransition(), bind));
                    buf.append("\n");
                }
            }
        }
        
        this.time = newTime;
        
        this.checkState(null);
        
        return buf.toString();
    }
    
    
    public void load(final PetriNet theNet) {
        this.load(theNet, true);
    }
    
    
    @SuppressWarnings("unchecked")
    public void load(final PetriNet theNet, final boolean init) {
        this.putBindingsLatest.clear();
        this.putBindingsEarliest.clear();
        if(this.bindings != null) {
            for(int i = 0; i < this.bindings.length; ++i) {
                this.bindings[i] = null;
            }
        }
        this.bindingsLatest.clear();
        this.bindingsEarliest.clear();
        this.enabledTrans.clear();
        
        this.functions.clear();
        for(final Transition t : this.firingTrans.values()) {
            this.firing[t.getKey()] = null;
        }
        this.firingTrans.clear();
        this.neverFired.clear();
        this.sequence.clear();
        this.time = 0;
        
        this.net = theNet;
        this.net.correctGuards();
        this.buildFunctions();
        this.buildGuards();
        
        if(this.minimizeNet()) {
            if(this.showSimplifications()) {
                System.out.println("Net elements (before minimizing): " + this.net.getElemCount());
            }
            this.net.optimize();
            if(this.showSimplifications()) {
                System.out.println("Net elements (after minimizing): " + this.net.getElemCount());
//                System.out.println(this.net.toPEP());
            }
        }
        
        this.setUsePriority(this.net.usesPriority());
        this.setUsesTime(this.net.usesTime());
        this.setUsesInhibitorArcs(this.net.hasInhibitorArcs());
        this.setUsesReadArcs(this.net.hasReadArcs());
        this.setUsesResetArcs(this.net.hasResetArcs());
        this.setUsesCapacity(this.net.usesCapacity());
        this.setUsesWeights(this.net.usesWeights());
        this.net.setCheckPlaceValues(false);

        if(init) {
            if(this.checkFire()) {
                for(final Transition t : this.net.getTransitionCollection()) {
                    this.neverFired.add(t.getName());
                }
            }
            
            this.fired = 0;
            this.state = Simulator.Status.RUNNING;
            
            final Collection<Place> places = this.net.getPlaceCollection();
            this.markings = new ValueList[places.size()];
            this.tokenCountEpochs = new Map[places.size()];
            this.tokenCountTime = new Map[places.size()];
            this.used = new IntStack(places.size());
            int i = 0;
            for(final Place place : places) {
                place.setKey(i++);
            }
            
            
            final Collection<Transition> transitions = this.net.getTransitionCollection();
            this.bindings = new List[transitions.size()];
            this.firing = new List[transitions.size()];
            i = 0;
            for(final Transition trans : transitions) {
                trans.setKey(i++);
            }
            
            this.checkTransitions();
            this.checkState(null);
            this.addTokenCountEpochs();
        }
    }
    
    
    private void buildFunctions() {
        for(final String function : this.net.getFunctions()) {
            final Function func = new Function(function);
            if(this.functions.containsKey(func.getName())) {
                System.err.println("Warning: multiple definitions of function "
                                + func.getName() + "\nThe function will be "
                                + "overwritten.\nOld: "
                                + this.functions.get(func.getName())
                                + "\nNew: " + func);
            }
            this.functions.put(func.getName(), func);
        }
    }
    
    
    private void addTokenCountEpochs() {
        for(int i = 0, n = this.measurePlacesEpochs.size(); i < n; ++i) {
            this.addTokenCountEpochs(this.measurePlacesEpochs.get(i));
        }
    }
    
    
    private void addTokenCountEpochs(final Place p) {
        final ValueList m = p.getCurrentMarkingSet();
        Map<Value, Integer> c = this.tokenCountEpochs[p.getKey()];
        if(c == null) {
            this.tokenCountEpochs[p.getKey()] = c = Pool.getPool().getMap();
        }
        if(!m.isEmpty()) {
            for(int i = 0, n = m.size(); i < n;) {
                final Value v = m.get(i);
                final int count = m.lastIndexOf(v, i) - i + 1;
                i += count;
                
                final Integer old = c.get(v);
                if(old == null) {
                    c.put(v, count);
                } else {
                    c.put(v, old + count);
                }
            }
        }
    }
    
    
    private ValueList[] getMarkings() {
        if(this.binders == null) {
            return this.markings;
        }
        Thread t = Thread.currentThread();
        ValueList[] m = this.threadMarkings.get(t);
        if(m == null) {
            m = new ValueList[this.markings.length];
            this.threadMarkings.put(t, m);
        }
        return m;
    }
    
    
    private IntStack getUsed() {
        if(this.binders == null) {
            return this.used;
        }
        Thread t = Thread.currentThread();
        IntStack m = this.threadUsed.get(t);
        if(m == null) {
            m = new IntStack(this.used.size());
            this.threadUsed.put(t, m);
        }
        return m;
    }
    
    
    private void addTokenCountTime(final double ticks) {
        for(int i = 0, n = this.measurePlacesTime.size(); i < n; ++i) {
            this.addTokenCountTime(this.measurePlacesTime.get(i), ticks);
        }
    }
    
    
    private void addTokenCountTime(final Place p, final double ticks) {
        final ValueList m = p.getCurrentMarkingSet();
        Map<Value, Double> c = this.tokenCountTime[p.getKey()];
        if(c == null) {
            this.tokenCountTime[p.getKey()] = c = Pool.getPool().getMap();
        }
        if(!m.isEmpty()) {
            for(int i = 0, n = m.size(); i < n;) {
                final Value v = m.get(i);
                final int count = m.lastIndexOf(v, i) - i + 1;
                i += count;
                
                final Double old = c.get(v);
                if(old == null) {
                    c.put(v, ticks*count);
                } else {
                    c.put(v, old + ticks*count);
                }
            }
        }
    }
    
    
    public int getTokenCountEpochs(final String place, final String value) {
        final Place p = this.net.getPlaceByName(place);
        if(p == null) {
            return 0;
        }
        final Value v = Value.getValue(value);
        final Map<Value, Integer> m = this.tokenCountEpochs[p.getKey()];
        if(m == null) {
            return 0;
        }
        final Integer i = m.get(v);
        return i == null ? 0 : i;
    }
    
    
    public double getTokenCountTime(final String place, final String value) {
        final Place p = this.net.getPlaceByName(place);
        if(p == null) {
            return 0;
        }
        final Value v = Value.getValue(value);
        final Map<Value, Double> m = this.tokenCountTime[p.getKey()];
        if(m == null) {
            return 0;
        }
        final Double i = m.get(v);
        return i == null ? 0.0d : i;
    }
    
    
    public String getAverageTokenCountEpochs() {
        final StringBuilder buf = new StringBuilder();
        final long epochs = this.getFiredCount() + 1;
        buf.append("Average token count per firing epoch (" + epochs + " epochs):\n");
        for(int i = 0, n = this.measurePlacesEpochs.size(); i < n; ++i) {
            final Place p = this.measurePlacesEpochs.get(i);
            final Map<Value, Integer> m = this.tokenCountEpochs[p.getKey()];
            if(m != null) {
                buf.append(p.getName());
                buf.append(": ");
                long tokens = 0;
                for(final Map.Entry<Value, Integer> e : m.entrySet()) {
                    final Value v = e.getKey();
                    final int count = e.getValue();
                    tokens += count;
                    if(count != 0) {
                        buf.append(v.getNetValue());
                        buf.append("/");
                        final double f = 1.0d * count / epochs;
                        buf.append(f);
                        buf.append(", ");
                    }
                }
                
                if(tokens == 0L) {
                    buf.append("empty");
                } else {
                    buf.append("[sum: ");
                    buf.append(1.0d * tokens / epochs);
                    buf.append("]");
                }
                if(i < n-1) {
                    buf.append("\n");
                }
            }
        }
        return buf.toString();
    }
    
    
    public String getAverageTokenCountTime() {
        final StringBuilder buf = new StringBuilder();
        final double curtime = this.getTime();
        buf.append("Average token count per time unit (" + curtime + " time units):\n");
        for(int i = 0, n = this.measurePlacesTime.size(); i < n; ++i) {
            final Place p = this.measurePlacesTime.get(i);
            final Map<Value, Double> m = this.tokenCountTime[p.getKey()];
            if(m != null) {
                buf.append(p.getName());
                buf.append(": ");
                double tokens = 0;
                for(final Map.Entry<Value, Double> e : m.entrySet()) {
                    final Value v = e.getKey();
                    final double count = e.getValue();
                    tokens += count;
                    if(count != 0) {
                        buf.append(v.getNetValue());
                        buf.append("/");
                        final double f = count / curtime;
                        buf.append(f);
                        buf.append(", ");
                    }
                }
                
                if(tokens == 0L) {
                    buf.append("empty");
                } else {
                    buf.append("[sum: ");
                    buf.append(tokens / curtime);
                    buf.append("]");
                }
                if(i < n-1) {
                    buf.append("\n");
                }
            }
        }
        return buf.toString();
    }
    
    
    public PetriNet getNet() {
        return this.net;
    }
    
    
    public void setExtendedHL2LLRef(final boolean v) {
        this.setProperty(Simulator.Option.EXTENDED_HL_LL_REF, v);
    }
    
    
    public boolean useExtendedHL2LLRef() {
        return this.getProperty(Simulator.Option.EXTENDED_HL_LL_REF);
    }
    
    
    public void setBuildHL2LLRef(final String v) {
        this.buildHL2LLRef = v;
    }
    
    
    public PetriNet unfold() {
        final Map<String, String> pref = Pool.getPool().getMap();
        final Map<String, String> tref = Pool.getPool().getMap();
        
        final PetriNet newnet = this.tool.createPetriNet();
        newnet.setIsLowLevelNet(true);
        
        final Map<String, Map<String, Place>> places = Pool.getPool().getMap();
        for(final Place p : this.net.getPlaces()) {
            final String name = p.getName();
            final ValueList values = Values.getValues(p.getType());
            final Map<String, Place> newPlaces = Pool.getPool().getMap();
            places.put(name, newPlaces);
            int i = 0;
            for(int ii = 0, n = values.size(); ii < n; ++ii) {
                final Value v = values.get(ii);
                if(v.isAny()) {
                    i++;
                    continue;
                }
                
                final String mark = p.getInitialMarkingSet().contains(v) ? IPetriNetTool.DOT : "";
                final String curmark = p.getCurrentMarkingSet().contains(v) ? IPetriNetTool.DOT : "";
                final String theName;
                if(values.size() > 1) {
                    theName = name + "(" + i + ")";
                } else {
                    theName = name;
                }
                if(this.buildHL2LLRef != null) {
                    if(this.useExtendedHL2LLRef()) {
                        pref.put(theName, "\"" + name + "\" \"" + v.getNetValue() + "\"");
                    } else {
                        pref.put(theName, "\"" + name + "\"");
                    }
                }
                final Place place = newnet.createPlace(theName, p.getStatus(),
                                IPlace.TYPE_BLACK_TOKEN, mark, curmark);
                final String meaning;
                if((p.getMeaning() != null) && (p.getMeaning().length() > 0)) {
                    meaning = p.getMeaning() + " (" + v.getNetValue() + ")";
                } else {
                    meaning = v.getNetValue();
                }
                place.setMeaning(meaning);
                place.setPos(p.getPosX() + Simulator.OFFSET_X_PLACE*i,
                                p.getPosY() + Simulator.OFFSET_Y_PLACE*i);
                newPlaces.put(v.getNetValue(), place);
                
                i++;
            }
        }
        
        final boolean usesTime = this.usesTime();
        for(final Transition t : this.net.getTransitionCollection()) {
            final String name = t.getName();
            final List<Binding> binds = Simulator.getBindingsForTransition(t,
                            null, this.functions, usesTime);
            //System.out.println(t.getName() + " -> " + tree);
            int i = 0;
            for(final Iterator<Binding> it2 = binds.iterator(); it2.hasNext(); i++) {
                final Binding bind = it2.next();
                Simulator.bindOnTransition(t, bind);
                
                final String theName;
                if(binds.size() > 1) {
                    theName = name + "(" + i + ")";
                } else {
                    theName = name;
                }
                if(this.buildHL2LLRef != null) {
                    if(this.useExtendedHL2LLRef()) {
                        tref.put(theName, "\"" + name + "\" \"" + bind.toPEPString() + "\"");
                    } else {
                        tref.put(theName, "\"" + name + "\"");
                    }
                }
                final Transition trans = newnet.createTransition(theName, null,
                                null);
                final String meaning = t.getMeaning();
//                if(meaning.length() > 0) {
//                    meaning += " " + bind.toStringSlim();
//                } else {
//                    meaning = bind.toStringSlim();
//                }
                trans.setMeaning(meaning);
                if(t.hasRate()) {
                    trans.setRate(new Node(null, Value.getValue(bind.getRate())));
                }
                if(t.hasWeight()) {
                    trans.setWeight(new Node(null, Value.getValue(bind.getWeight())));
                }
                if(t.hasGeneralizedRate()) {
                    // use expression, all variables are bound
                    trans.setGeneralizedRate(trans.getGeneralizedRate());
                }
                if(t.hasGeneralizedWeight()) {
                    // use expression, all variables are bound
                    trans.setGeneralizedWeight(trans.getGeneralizedWeight());
                }
                if(t.hasEarliestFiring()) {
                    trans.setEarliestFiring(Value.number2String(bind.getEarliestFiring()));
                }
                if(t.hasLatestFiring()) {
                    trans.setLatestFiring(Value.number2String(bind.getLatestFiring()));
                }
                if(t.hasPriority()) {
                    trans.setPriority(Value.number2String(bind.getPriority()));
                }
                if(t.hasServers()) {
                    trans.setServers(Value.number2String(Math.ceil(bind.getServers())));
                }
                if(t.hasAgeMemory()) {
                    trans.setAgeMemory(Value.getValue(bind.useAgeMemory()).getNetValue());
                }
                trans.setPos(t.getPosX() + Simulator.OFFSET_X_TRANS*i,
                                t.getPosY() + Simulator.OFFSET_Y_TRANS*i);
                
                final List<Arc> input = t.getInput();
                for(int i3 = 0, n3 = input.size(); i3 < n3; ++i3) {
                    final Arc arc = input.get(i3);
                    final Place from = arc.getPlace();
                    final String plname = from.getName();
                    final Arc.Type type = arc.getType();
                    final Map<String, Place> newPl = places.get(plname);
                    final ValueList all = arc.getLabel().getAll();
                    
                    if(arc.isInhibitorArc()) {
                        final ValueList checked = Pool.getPool().getValueList();
                        
                        for(int i4 = 0, n4 = all.size(); i4 < n4; ++i4) {
                            final Value v = all.get(i4);
                            final IPlace p = newPl.get(v.getNetValue());
                            if((p == null) || checked.contains(v)) {
                                continue;
                            }
                            checked.add(v);
                            final int countI = PNModel.countOccurrences(all, v);
                            newnet.createInhibitorArc(trans, p, "" + countI);
                        }
                        
                        continue;
                    }
                    
                    for(int i4 = 0, n4 = all.size(); i4 < n4; ++i4) {
                        final Value v = all.get(i4);
                        v.setNetValue();
                        final IPlace p = newPl.get(v.getNetValue());
                        newnet.createArc(trans, p, null, type);
                    }
                }
                
                final List<Arc> output = t.getOutput();
                for(int i2 = 0, n2 = output.size(); i2 < n2; ++i2) {
                    final Arc arc = output.get(i2);
                    final Place to = arc.getPlace();
                    final String plname = to.getName();
                    final Arc.Type type = arc.getType();
                    final Map<String, Place> newPl = places.get(plname);
                    
                    final ValueList all = arc.getLabel().getAll();
                    for(int i3 = 0, n3 = all.size(); i3 < n3; ++i3) {
                        final Value v = all.get(i3);
                        v.setNetValue();
                        final IPlace p = newPl.get(v.getNetValue());
                        newnet.createArc(trans, p, null, type);
                    }
                }
                
                Simulator.unbindOnTransition(t, bind);
                bind.clear();
                it2.remove();
            }
            
        }
        
        places.clear();
        
        newnet.optimize();
        
        if(this.buildHL2LLRef != null) {
            this.writeRef(newnet, pref, tref, this.buildHL2LLRef);
        }
        
        return newnet;
    }
    
    
    private void writeRef(final PetriNet theNet, final Map<String, String> pref,
                    final Map<String, String> tref, final String reffile) {
        try {
            final File file = new File(reffile);
            final FileWriter outputstream = new FileWriter(file);
            final PrintWriter output = new PrintWriter(outputstream);
        
            output.println("PEP");
            output.println("References_HL_LL");
            output.println("FORMAT_N2");
            
            output.println("PL");
            for(final Map.Entry<String, String> e : pref.entrySet()) {
                if(theNet.containsPlace(e.getKey())) {
                    output.println(e.getKey() + e.getValue());
                }
            }
            
            output.println("TR");
            for(final Map.Entry<String, String> e : tref.entrySet()) {
                if(theNet.containsTransition(e.getKey())) {
                    output.println(e.getKey() + e.getValue());
                }
            }
            output.println("PTR");
            
            output.close();
        } catch (final IOException ioex) {
            throw new RuntimeException(ioex);
        }
    }
    
    
    public void saveSequence(final int length) {
        this.sequence.clear();
        this.maxSeqLength = length;
    }
    
    
    public String getSequence() {
        final StringBuilder result = new StringBuilder();
        for(final Iterator<String> it = this.sequence.iterator(); it.hasNext();) {
            result.append(it.next());
            if(it.hasNext()) {
                result.append(",");
            }
        }
        return result.toString();
    }
    
    
    public boolean debug() {
        return this.getProperty(Simulator.Option.DEBUG);
    }
    
    
    public void checkTransitions(final String regex) {
        this.setProperty(Simulator.Option.DEBUG, true);
        this.checkTrans = (regex == null ? "" : regex);
    }
    
    
    public void disableCheckTransitions() {
        this.setProperty(Simulator.Option.DEBUG, false);
        this.checkTrans = "";
    }
    
    
    public void loadFile(final String file) {
        this.loadFile(file, true);
    }
    
    public void loadFile(final String file, final boolean init) {
        final PetriNet theNet;
        if(file.endsWith(".pnml") || file.endsWith(".xml")) {
            theNet = (PetriNet)this.pnmlreader.readFromFile(this.tool, file);
        } else if(file.endsWith(".hl_net") || file.endsWith(".ll_net")) {
            theNet = (PetriNet)this.pepreader.readFromFile(this.tool, file);
        } else {
            System.err.println("Unknown format. I'll try to parse it like PEP.");
            theNet = (PetriNet)this.pepreader.readFromFile(this.tool, file);
        }
        this.load(theNet, init);
    }
    
    
    public void loadString(final String text) {
        this.loadString(text, true);
    }
    
    
    public void loadString(final String text, final boolean init) {
        final PetriNet theNet;
        if(text.trim().startsWith("<?xml")) {
            theNet = (PetriNet)this.pnmlreader.readFromString(this.tool, text);
        } else if(text.trim().startsWith("PEP")) {
            theNet = (PetriNet)this.pepreader.readFromString(this.tool, text);
        } else {
            System.err.println("Unknown format. I'll try to parse it like PEP.");
            theNet = (PetriNet)this.pepreader.readFromString(this.tool, text);
        }
        this.load(theNet, init);
    }
    
    
    public static String getFreshArcVar(final Transition trans) {
        final Collection<String> m = trans.getVariables();
        final String name = "aexpr";
        if(m.contains(name)) {
            int i = 0;
            while(m.contains(name + i)) {
                ++i;
            }
            return name + i;
        }
        return name;
    }
    
    
    public static void handleArcExpr(final Transition trans, final Node guard,
                    final Label label, final Map<String, Function> functions) {
        for(final Iterator<String> it = label.getExpressions().iterator(); it.hasNext();) {
            final String expr = it.next();
            final String theVar = Simulator.getFreshArcVar(trans);
            final Value var = Value.getValue(theVar);
            trans.addVariable(theVar, var);
            it.remove();
            label.getVariables().add(var);
            
            final RPN rpn = RPNComputer.infixToRPN(trans.getVariableMapping(), expr);
            final Node tree = TreeComputer.rpnToTreeWithDeletion(rpn);
//            tree.replaceFunctions(functions);
            if(guard.getToken().getOperator() != Token.Operator.AND) {
                final Node n = new Node(null, guard.getToken(), guard.getChildren());
                guard.clearChildren();
                guard.addChild(n);
                guard.setToken(Token.Operator.AND);
            }
            guard.addConstraints(tree.removeConstraints());
            final Node eq = new Node(guard, Token.getToken(Token.Operator.EQUAL));
            eq.addChild(new Node(eq, var));
            eq.addChild(tree);
            guard.addChild(eq);
        }
    }
    
    
    private void buildGuards() {
        for(final Transition t : this.net.getTransitionCollection()) {
            t.identifyVariables();
//            t.replaceFunctions(this.functions);
            
            final Node tree = t.getGuardTree();
            if(this.showSimplifications()) {
                System.out.println("   Transition: " + t.getName());
                System.out.println("Original term: " + tree);
                System.out.println("  Constraints: " + tree.getConstraints());
            }
            
            if(Value.V_FALSE.equals(tree.getToken().getValue())) {
                if(this.showSimplifications()) {
                    System.out.println("Transition " + t.getName() + " will "
                                    + "be deleted (guard==false).");
                }
                // will be deleted
                continue;
            }
            
            final List<Arc> deadArcs = Pool.getPool().getList();
            final List<Arc> input = t.getInput();
            for(int i = 0, n = input.size(); i < n; ++i) {
                final Arc arc = input.get(i);
                final Label label = arc.getLabel();
                Simulator.handleArcExpr(t, tree, label, this.functions);
                final ValueList values = Values.getValues(arc.getPlace().getType());
                if(arc.isInhibitorArc() || arc.isResetArc()) {
                    final ValueList lvalues = label.getValues();
                    if(lvalues.contains(Value.V_ANY) || lvalues.containsAll(values)) {
                        // simplify label to *
                        label.getValues().clear();
                        label.getVariables().clear();
                        label.getTuples().clear();
                        label.getExpressions().clear();
                        label.getAll().clear();
                        label.getValues().add(Value.V_ANY);
                        label.getAll().add(Value.V_ANY);
                    } else if(!values.contains(Value.V_ANY)) {
                        // remove every value from the label that cannot be on the place
                        lvalues.retainAll(values);
                        if(label.isEmpty()) {
                            deadArcs.add(arc);
                        }
                    }
                    
                    if(arc.isResetArc()) {
                        PNModel.deleteDublicates(lvalues);
                        PNModel.deleteDublicates(label.getVariables());
                        PNModel.deleteDublicates(label.getTuples());
                        PNModel.deleteDublicates(label.getAll());
                    }
                }
                
                if(!arc.isSimpleArc() && !arc.isReadArc()) {
                    continue;
                }
                
                if(values.contains(Value.V_ANY)) {
                    continue;
                }
                
                final ValueList vars = label.getVariables();
                for(int i2 = 0, n2 = vars.size(); i2 < n2; ++i2) {
                    final Value var = vars.get(i2);
                    if(!tree.constrain(var.getVariable(), values)) {
                        tree.setToken(Value.getValue(false));
                        tree.clearChildren();
                        t.setGuard(IPetriNetTool.FALSE);
                        if(this.showSimplifications()) {
                            System.out.println("Transition " + t.getName() + " will "
                                            + "be deleted (<-var).");
                        }
                    }
                }
                final ValueList tuples = label.getTuples();
                for(int i2 = 0, n2 = tuples.size(); i2 < n2; ++i2) {
                    final Value tuple = tuples.get(i2);
                    final int size = tuple.getTuple().size();
                    final IPlaceType type = arc.getPlace().getType();
                    if(type instanceof petruchio.sim.petrinettool.type.Set) {
                        petruchio.sim.petrinettool.type.Set set = (petruchio.sim.petrinettool.type.Set)type;
                        for(final IPlaceType theType : set.getTypes()) {
                            if(theType instanceof petruchio.sim.petrinettool.type.Tuple) {
                                petruchio.sim.petrinettool.type.Tuple ttype = (petruchio.sim.petrinettool.type.Tuple)theType;
                                if(ttype.size() == size) {
                                    if(!Simulator.constrainTuple(tuple.getTuple(), ttype, tree, true)) {
                                        tree.setToken(Value.getValue(false));
                                        tree.clearChildren();
                                        t.setGuard(IPetriNetTool.FALSE);
                                        if(this.showSimplifications()) {
                                            System.out.println("Transition " + t.getName() + " will "
                                                            + "be deleted (<-tuples).");
                                        }
                                    }
                                }
                            }
                        }
                    } else {
                        if(!Simulator.constrainTuple(tuple.getTuple(), type, tree, false)) {
                            tree.setToken(Value.getValue(false));
                            tree.clearChildren();
                            t.setGuard(IPetriNetTool.FALSE);
                            if(this.showSimplifications()) {
                                System.out.println("Transition " + t.getName() + " will "
                                                + "be deleted (<-tuples).");
                            }
                        }
                    }
                }
            }
            
            for(int i2 = 0, n2 = deadArcs.size(); i2 < n2; ++i2) {
                this.net.removeArc(deadArcs.get(i2));
            }

            if(this.showSimplifications()) {
                System.out.println("      In Arcs: " + tree.getConstraints());
            }
            
            final List<Arc> output = t.getOutput();
            for(int i = 0, n = output.size(); i < n; ++i) {
                final Arc arc = output.get(i);
                final Label label = arc.getLabel();
                Simulator.handleArcExpr(t, tree, label, this.functions);
                if(!arc.isSimpleArc()) {
                    continue;
                }
                
                final ValueList values = Values.getValues(arc.getPlace().getType());
                if(values.contains(Value.V_ANY)) {
                    continue;
                }
                final ValueList vars = label.getVariables();
                for(int i2 = 0, n2 = vars.size(); i2 < n2; ++i2) {
                    final Value var = vars.get(i2);
                    if(!tree.constrain(var.getVariable(), values)) {
                        tree.setToken(Value.getValue(false));
                        tree.clearChildren();
                        t.setGuard(IPetriNetTool.FALSE);
                        if(this.showSimplifications()) {
                            System.out.println("Transition " + t.getName() + " will "
                                            + "be deleted (var->).");
                        }
                    }
                }
                final ValueList tuples = label.getTuples();
                for(int i2 = 0, n2 = tuples.size(); i2 < n2; ++i2) {
                    final Value tuple = tuples.get(i2);
                    final int size = tuple.getTuple().size();
                    final IPlaceType type = arc.getPlace().getType();
                    if(type instanceof petruchio.sim.petrinettool.type.Set) {
                        petruchio.sim.petrinettool.type.Set set = (petruchio.sim.petrinettool.type.Set)type;
                        for(final IPlaceType theType : set.getTypes()) {
                            if(theType instanceof petruchio.sim.petrinettool.type.Tuple) {
                                petruchio.sim.petrinettool.type.Tuple ttype = (petruchio.sim.petrinettool.type.Tuple)theType;
                                if(ttype.size() == size) {
                                    if(!Simulator.constrainTuple(tuple.getTuple(), ttype, tree, true)) {
                                        tree.setToken(Value.getValue(false));
                                        tree.clearChildren();
                                        t.setGuard(IPetriNetTool.FALSE);
                                        if(this.showSimplifications()) {
                                            System.out.println("Transition " + t.getName() + " will "
                                                            + "be deleted (tuples->).");
                                        }
                                    }
                                }
                            }
                        }
                    } else {
                        if(!Simulator.constrainTuple(tuple.getTuple(), type, tree, false)) {
                            tree.setToken(Value.getValue(false));
                            tree.clearChildren();
                            t.setGuard(IPetriNetTool.FALSE);
                            if(this.showSimplifications()) {
                                System.out.println("Transition " + t.getName() + " will "
                                                + "be deleted (tuples->).");
                            }
                        }
                    }
                }
            }
            
            if(this.showSimplifications()) {
                System.out.println("     Out Arcs: " + tree.getConstraints());
            }
            
            for(final String var : t.getVariables()) {
                final ValueList constr = tree.getConstraint(var);
                if(constr == null) {
//                    System.out.println("da: " + var);
                    final ValueList l = Pool.getPool().getValueList();
                    l.add(Value.V_ANY);
                    tree.constrain(var, l);
                }
            }
            
            if(this.showSimplifications()) {
                System.out.println("    Free Vars: " + tree.getConstraints());
            }
            
            
            if(!tree.getToken().isValue()) {
                boolean changed;
                do {
                    TreeComputer.optimizeHard(tree, this.functions);
                    if(this.showSimplifications()) {
                        System.out.println("    Optimized: " + tree);
                        System.out.println("               " + tree.getConstraints());
                    }
                    changed = TreeComputer.extractConstraints(tree, this.functions);
                    if(this.showSimplifications()) {
                        System.out.println("    Extracted: " + tree);
                        System.out.println("               " + tree.getConstraints());
                    }
                } while(changed);
                tree.unifyVariables(t.getVariableMapping());
//                TreeComputer.cutByConstraints(tree, t);
//                TreeComputer.optimizeHard(tree);
                
//                if(this.showSimplifications()) {
//                    System.out.println("  Simple term: " + tree);
//                    System.out.println("  Constraints: " + tree.getConstraints());
//                }
            }
            
            if(this.showSimplifications()) {
                System.out.println();
            }
            
            boolean willBeDeleted = Value.V_FALSE.equals(tree.getToken().getValue());
            if(tree.hasConstraints()) {
                for(final Iterator<ValueList> it2 = tree.getConstraints().values().iterator();
                        !willBeDeleted && it2.hasNext();) {
                    final ValueList values = it2.next();
                    if(values.isEmpty()) {
                        willBeDeleted = true;
                    }
                }
            }
            
            if(willBeDeleted) {
                tree.setToken(Value.getValue(false));
                tree.clearChildren();
                tree.clearConstraints();
                t.setGuard(IPetriNetTool.FALSE);
            } else {
                t.setGuard(tree);
                t.optimizeExpressions(this.functions);
            }
        }
    }

    
    private List<Binding> getBindings(final Transition t) {
        return this.bindings[t.getKey()];
    }
    
    
    private void setBindings(final Transition t, final List<Binding> l) {
        this.bindings[t.getKey()] =  l;
        if(l != null) {
            synchronized(this.enabledTrans) {
                this.enabledTrans.put(t, t);
            }
        }
    }
    
    
    public void removeBindings(final Transition trans) {
        final List<Binding> binds = this.bindings[trans.getKey()];
        this.bindings[trans.getKey()] = null;
        if(binds != null) {
            if(!binds.isEmpty()) {
                this.removeBindings(binds);
            }
        }
        synchronized(this.enabledTrans) {
            this.enabledTrans.remove(trans);
        }
    }
    
    
    private List<Binding> getFiringBindings(final Transition t) {
        return this.firing[t.getKey()];
    }
    
    
    private void setFiringBindings(final Transition t, final List<Binding> l) {
        if(l == null) {
            this.firingTrans.remove(t);            
        } else if(this.firing[t.getKey()] == null){
            this.firingTrans.put(t, t);
        }
        this.firing[t.getKey()] = l;
    }
    
    
    private List<Binding> removeFiringBindings(final Transition t) {
        final List<Binding> l = this.getFiringBindings(t);
        this.setFiringBindings(t, null);
        return l;
    }
    
    
    public Collection<String> getNeverFiredTransitions() {
        if(!this.checkFire()) {
            throw new RuntimeException("This option has to be activated first!");
        }
        return Collections.unmodifiableCollection(this.neverFired);
    }
    
    
    public void checkTransitions() {
        this.checkTransitions(this.net.getTransitionCollection());
    }
    

    private void checkTransitions(final Collection<Transition> transitions) {
        if(this.binders == null) {
            // single thread
            for(final Transition t : transitions) {
                this.checkTransition(t);
            }
        } else {
            // multi threaded
            int i = 0;
            for(final Transition t : transitions) {
//                System.out.println(i + " " + t.hashCode());
                this.binders[i].addTransition(t);
                ++i;
                if(i == this.binders.length) {
                    i = 0;
                }
            }
            for(int j = this.binders.length-1; j >= 0; --j) {
//                System.out.println("wait for binder " + j + ": " + this.binders[j].transitions.size());
                this.binders[j].waitWhileWorking();
//                System.out.println("binder " + j + " done: " + !this.binders[j].isWorking());
            }
        }
    }
    
    
    public void pleaseStop() {
        this.setThreadCount(1);
    }
    
    
    private void changeTimeWithAgeMemory(final Binding bind, final int count) {
        final Transition trans = bind.getTransition();
        final double enabled = trans.enableAgeMemory(bind, count);
        if(enabled != 0.0d) {
            // this binding has already been enabled since its last firing
            // do not restart clock but continue
            final double newLatest = bind.getLatestFiring() - enabled;
            if(newLatest < this.getTime()) {
                bind.setLatestFiring(this.getTime()+(bind.getLatestFiring() - bind.getEarliestFiring()));
            } else {
                bind.setLatestFiring(newLatest);
            }
            bind.setEarliestFiring(Math.max(this.getTime(), bind.getEarliestFiring()-enabled));
        }
    }
    
    
    
    @SuppressWarnings("unchecked")
    private void checkTransition(final Transition t) {
        List<Binding> binds = this.getBindingForTransition(t, Simulator.DFLT_ALL_BINDINGS || this.usesTime());
        
        if(this.usesTime()) {
            if((binds == null) || binds.isEmpty()) {
                if(!t.isImmediate() && t.hasAgeMemory()) {
                    // all these bindings have to be removed
                    final List<Binding> b = this.getBindings(t);
                    if(b != null) {
                        for(int i2 = 0, n2 = b.size(); i2 < n2; ++i2) {
                            final Binding bind = b.get(i2);
                            if(bind.useAgeMemory()) {
                                t.disableAgeMemory(bind, this.getTime());
                            }
                        }
                    }
                }
                this.removeBindings(t);
            } else {
                final boolean isImmediate = t.isImmediate();
                final boolean hasAgeMemory = t.hasAgeMemory();
                final boolean distributeTime = this.distributeTime();
                final List<Binding> oldBinds = this.getBindings(t);
                
                final List<Binding> list = Pool.getPool().getList();
                Simulator.insertSortLatestFiring(list, binds);
                binds = list;
                
                if(oldBinds == null) {
                    this.setBindings(t, binds);
                } else {
                    final boolean hasServers = t.hasServers();
                    // remove disabled bindings from the old list
                    for(final Iterator<Binding> it2 = oldBinds.iterator(); it2.hasNext(); ) {
                        final Binding oldBind = it2.next();
                        boolean singleserver = oldBind.getServers() == 1.0;
                        final boolean exists;
                        // search for the binding (equality)
                        int pos = Simulator.indexOfSortEarliest(binds, oldBind, true);
                        if(pos >= 0) {
                            exists = true;
                            // remove existing bindings from the new list if servers == 1
                            if(hasServers && singleserver) {
                                binds.remove(pos);
                            }
                        } else {
                            exists = false;
                        }
                        
                        if(!exists) {
                            // the binding has been disabled
                            if(oldBind.useAgeMemory()) {
                                t.disableAgeMemory(oldBind, this.getTime());
                            }
                            if(t.isImmediate()) {
                                this.removeFromImmediateBindings(oldBind, false);
                            } else {
                                this.removeFromLatestBindings(oldBind, false);
                                this.removeFromEarliestBindings(oldBind, false);
                            }
                            it2.remove();
                        }
                    }
                    
                    if(!hasServers) {
                        Simulator.insertSortEarliestFiring(oldBinds, binds);
                    }
                }
                
                final boolean bindsAreSorted = oldBinds == null;
                
                if(!binds.isEmpty()) {
                    if(!isImmediate && hasAgeMemory && !t.hasServers()) {
                        for(int i2 = 0, n2 = binds.size(); i2 < n2; ++i2) {
                            final Binding bind = binds.get(i2);
                            if(bind.useAgeMemory()) {
                                // search for the binding (equality)
                                final int pos = Simulator.indexOfSortEarliest(oldBinds, bind, true);
                                if(pos < 0) { // this binding is new
                                    this.changeTimeWithAgeMemory(bind, 1);
                                }
                            }
                        }
                    } else if(t.hasServers() && !isImmediate) {
                        final List<Binding> additional = Pool.getPool().getList();
                        for(final Iterator<Binding> it = binds.iterator(); it.hasNext();) {
                            final Binding bind = it.next();
                            final List<Binding> newBindings = bind.useAgeMemory() ? Pool.getPool().<Binding>getList() : (List<Binding>)null;
                            final double servers = bind.getServers();
                            final int create;
                            if(servers == 1.0) {
                                // single-server semantics
                                // nothing to do
                                create = 1;
                            } else if(servers == Double.POSITIVE_INFINITY) {
                                // infinite-server semantics
                                final int degree = Simulator.getEnablingDegree(bind);
                                if(Integer.MAX_VALUE == degree) {
                                    // infinte enablings
                                    throw new RuntimeException("Binding "
                                        + bind + " uses infinite-server "
                                        + "semantics and has infinite "
                                        + "enablings. That's bad!");
                                }
                                final int enabled = oldBinds == null ? 0 : PNModel.countOccurrences(oldBinds, bind);
                                // create "some" additional bindings 
                                create = degree-enabled;
                                
                                // uh oh, the new binding shouldn't be enabled.
                                // enabled >= degree!
                                if(create <= 0) {
                                    // remove binding from list.
                                    it.remove();
                                }
                            } else {
                                // multiple-server semantics
                                final int enabled = oldBinds == null ? 0 : PNModel.countOccurrences(oldBinds, bind);
                                final int iservers = (int)servers;
                                if(enabled >= iservers) {
                                    // there are "enough" bindings enabled
                                    it.remove();
                                    create = 0;
                                } else {
                                    final int degree = Simulator.getEnablingDegree(bind);
                                    final int maximum = Math.min(iservers, degree);
                                    // create "some" additional bindings 
                                    create = maximum-enabled;
                                    
                                    // uh oh, the new binding shouldn't be enabled.
                                    // enabled >= degree!
                                    if(create <= 0) {
                                        // remove binding from list.
                                        it.remove();
                                    }
                                }
                            }
                            
                            // since the first binding is already in the list
                            // create degree-1 bindings.
                            for(int nr = 1; nr < create; ++nr) {
                                final Binding copy = bind.clone();
                                this.setRateAndWeight(copy, false);
                                if(copy.useAgeMemory()) {
                                    newBindings.add(copy);
                                }
                                additional.add(copy);
                            }
                            
                            if(bind.useAgeMemory()) {
                                if(create > 0) {
                                    newBindings.add(bind);
                                }
                                final List<Binding> buf;
                                if(distributeTime) {
                                    buf= Pool.getPool().getList();
                                    for(int i2 = 0, n2 = newBindings.size(); i2 < n2; ++i2) {
                                        final Binding cbind = newBindings.get(i2);
                                        Simulator.insertSortLatestFiring(buf, cbind);
                                    }
                                } else {
                                    buf = newBindings;
                                }
                                for(int i2 = 0, n2 = buf.size(); i2 < n2; ++i2) {
                                    final Binding cbind = buf.get(i2);
                                    final int count = distributeTime ? n2-i2 : 1;
                                    this.changeTimeWithAgeMemory(cbind, count);
                                }
                            }
                        }
                        if(oldBinds != null) {
                            binds.addAll(additional);
                            Simulator.insertSortEarliestFiring(oldBinds, binds);
                        } else {
                            Simulator.insertSortEarliestFiring(binds, additional);
                        }
                    }
                    
                    if(isImmediate) {
                        this.insertToImmediateBindings(binds, bindsAreSorted);
                    } else {
                        this.insertToLatestBindings(binds, bindsAreSorted);
                        this.insertToEarliestBindings(binds, bindsAreSorted);
                    }
                }
                
            }
        } else {
            this.removeBindings(t);
            if((binds != null) && !binds.isEmpty()) {
                final List<Binding> list = Pool.getPool().getList();
                Simulator.insertSortEarliestFiring(list, binds);
                binds = list;
                this.setBindings(t, binds);
                
                this.insertToImmediateBindings(binds, true);
            } 
        }
    }
    
    
    public static int getEnablingDegree(final Binding bind) {
        if(bind == null) {
            return 0;
        }
        final Transition trans = bind.getTransition();
        Simulator.bindOnTransition(trans, bind);
        
        int degree = Integer.MAX_VALUE;
        final List<Arc> input = trans.getInput();
        for(int i = 0, n = input.size(); i < n; ++i) {
            final Arc arc = input.get(i);
            if(!arc.isSimpleArc()) {
                continue;
            }
            final ValueList marking = arc.getPlace().getCurrentMarkingSet();
            final ValueList all = arc.getLabel().getAll();
            
            final ValueList checked = Pool.getPool().getValueList();
            for(int i2 = 0, n2 = all.size(); i2 < n2; ++i2) {
                final Value val = all.get(i2);
                if(!checked.contains(val)) {
                    checked.add(val);
                    final int arcCount = PNModel.countOccurrences(all, val);
                    final int markingCount = PNModel.countOccurrences(marking, val);
                    final int enabling = markingCount/arcCount;
                    if(enabling < degree) {
                        degree = enabling;
                    }
                }
                if(degree == 0) {
                    break;
                }
            }
        }
        
        Simulator.unbindOnTransition(trans, bind);
        
        return degree;
    }
    
    
    public String getBinding(final String trans) {
        final Transition t = this.net.getTransitionByName(trans);
        if(t == null) {
            return "";
        }
        return this.getBindings(t).toString();
    }
    
    
    private static ValueList getMarking(final ValueList[] markings, final IntStack used, final Place p) {
        ValueList marking = markings[p.getKey()];
        if(marking == null) {
            markings[p.getKey()] = marking = Pool.getPool().getValueList(p.getCurrentMarkingSet());
            used.push(p.getKey());
        }
        return marking;
    }
    
    
    private static void returnMarking(final ValueList[] markings, final IntStack used) {
        final int n = used.size();
        for(int i = 0; i < n; ++i) {
            final int index = used.get(i);
            markings[index] = null;
        }
        used.clear();
        //Pool.getPool().returnList(used);
    }
    
    
//    private static ValueList getMarking(final Map<Place, ValueList> markings, final Place p) {
//        ValueList marking = markings.get(p);
//        if(marking == null) {
//            markings.put(p, marking = Pool.getPool().getValueList(p.getCurrentMarkingSet()));
//        }
//        return marking;
//    }
//    
//    
//    private static void returnMarking(final Map<?, ValueList> markings) {
//        for(final ValueList list : markings.values()) {
//            Pool.getPool().returnValueList(list);
//        }
//        Pool.getPool().returnMap(markings);
//    }


    private List<Binding> getBindingForTransition(final Transition trans, final boolean allBindings) {
        return this.getBindingForTransition(trans, null, allBindings);
    }
    
    
    private List<Binding> getBindingForTransition(final Transition trans,
                    final Binding bind, final boolean allBindings) {
        final boolean output = this.debug() && (trans.getName().matches(this.checkTrans));
        return this.getBindingForTransition(trans, bind,
                        trans.getInput(), trans.getOutput(), allBindings, output);
    }
    
    
    private List<Binding> getBindingForTransition(final Transition trans,
                    Binding bind, final List<Arc> inputArcs,
                    final List<Arc> outputArcs, final boolean allBindings,
                    final boolean debug) {
        
        if(debug) {
            System.out.println("Checking transition: " + trans.getName());
        }
        
        if(this.usesCapacity()) {
            for(int i = 0, n = outputArcs.size(); i < n; ++i) {
                final Arc outarc = outputArcs.get(i);
                if(!outarc.isSimpleArc()) {
                    continue;
                }
                
                final Place p = outarc.getPlace();
                if(!p.infiniteCapacity()
                        && (p.getCurrentMarkingSet().size()
                                        + outarc.getLabel().size() > p.getCapacity())) {
                    // the capacity could be exceeded
                    boolean disabled = true; 
                    for(int i2 = 0, n2 = inputArcs.size(); i2 < n2; ++i2) {
                        final Arc inarc = inputArcs.get(i2);
                        if(!inarc.isSimpleArc()) {
                            continue;
                        }
                        if(p == inarc.getPlace()) {
                            // check if the capacity will be exceeded
                            disabled = p.getCurrentMarkingSet().size()
                                        + outarc.getLabel().size()
                                        - inarc.getLabel().size() > p.getCapacity();
                            break;
                        }
                    }
                    if(disabled) {
                        if(debug) {
                            System.out.println("disabled: capacity " + p.getCapacity()
                                            + " of place " + p.getName() + " would be "
                                            + "exceeded.");
                        }
                        return null;
                    }
                }
            }
        }
        
        final boolean freeBind = (bind == null);
        if(freeBind) {
            bind = Pool.getPool().getBinding();
        }
        assert bind != null;
        if(debug) {
            System.out.println("Constraints so far: " + bind.getConstraints());
        }
        
        final ValueList[] curmarkings = this.getMarkings();
        final IntStack usedInts = this.getUsed();
//        final Map<Place, ValueList> markings = Pool.getPool().getMap();
        
        for(int i = 0, n = inputArcs.size(); i < n; ++i) {
            final Arc arc = inputArcs.get(i);
            if(arc.isInhibitorArc()) {
                // check inhibitor arcs: test the values and tuples (w/o variables)
                final Place p = arc.getPlace();
                final ValueList marking = Simulator.getMarking(curmarkings, usedInts, p);
                final Label label = arc.getLabel();
                boolean enabled = true;
                
                if(label.getValues().contains(Value.V_ANY) && !marking.isEmpty()) {
                    if(debug) {
                        System.out.println("inhibitor arc not satisfied with place "
                                        + p.getName() + ": it's not empty!");
                    }
                    enabled = false;
                } 
                if(!enabled) {
                    final ValueList checked = Pool.getPool().getValueList();
                    final ValueList vals = label.getValues();
                    for(int i2 = 0, n2 = vals.size(); i2 < n2; ++i2) {
                        final Value val = vals.get(i2);
                        if(checked.contains(val)) {
                            continue;
                        }
                        checked.add(val);
                        final int countI = PNModel.countOccurrences(label.getValues(), val);
                        final int countM = PNModel.countOccurrences(marking, val);
                        if(countM >= countI) {
                            if(debug) {
                                System.out.println("inhibitor arc not satisfied with "
                                                + val + " on place " + p + ": "
                                                + countM + " >= " + countI);
                            }
                            enabled = false;
                            break;
                        }
                    }
                }
                if(!enabled) {
                    final ValueList checked = Pool.getPool().getValueList();
                    final ValueList vals = label.getTuples();
                    for(int i2 = 0, n2 = vals.size(); i2 < n2; ++i2) {
                        final Value val = vals.get(i2);
                        if(val.getTuple().containsVariable() || checked.contains(val)) {
                            continue;
                        }
                        checked.add(val);
                        final int countI = PNModel.countOccurrences(label.getTuples(), val);
                        final int countM = PNModel.countOccurrences(marking, val);
                        if(countM >= countI) {
                            if(debug) {
                                System.out.println("inhibitor arc not satisfied with "
                                                + val + " on place " + p + ": "
                                                + countM + " >= " + countI);
                            }
                            enabled = false;
                            break;
                        }
                    }
                }
                
                if(!enabled) {
                    // this transition is not enabled. exit immediately.
                    Simulator.returnMarking(curmarkings, usedInts);
                    return null;
                }
            }
            
            
            if(!arc.isSimpleArc()) {
                continue;
            }
            
            final Place p = arc.getPlace();
            if(p.getCurrentMarkingSet().isEmpty()) {
                if(debug) {
                    System.out.println("disabled: place " + p.getName()
                            + " is empty.");
                }
                // nice try
                Simulator.returnMarking(curmarkings, usedInts);
                return null;
            }
            
            final Label label = arc.getLabel();
            if(p.getCurrentMarkingSet().size() < label.getAll().size()) {
                // still not enough tokens
                if(debug) {
                    System.out.println("disabled: place " + p.getName()
                            + " has not enough tokens.");
                }
                Simulator.returnMarking(curmarkings, usedInts);
                return null;
            }
            
            ValueList marking = null;
            
            // let's delete the values first
            if(!label.getValues().isEmpty()) {
                final ValueList vals = label.getValues();
                for(int i2 = 0, n2 = vals.size(); i2 < n2; ++i2) {
                    final Value value = vals.get(i2);
                    if(marking == null) {
                        marking = Simulator.getMarking(curmarkings, usedInts, p);
                    }
                    if(!marking.remove(value)) {
                        if(debug) {
                            System.out.println("disabled: place " + p.getName()
                                    + " " + p.getCurrentMarkingSet() + " => "
                                    + marking + " does not contain value "
                                    + value + ". LABEL: " + label);
                        }
                        Simulator.returnMarking(curmarkings, usedInts);
                        return null;
                    }
                }
            }
            
            ValueList tuples = null;
            int ctr = label.getTuples().size();
            if(ctr > 0) {
                if(marking == null) {
                    marking = Simulator.getMarking(curmarkings, usedInts, p);
                }
                for(final Iterator<Value> it2 = marking.iterator(); it2.hasNext(); ) {
                    final Value token = it2.next();
                    if(token.isTuple()) {
                        if(tuples == null) {
                            tuples = Pool.getPool().getValueList();
                        }
                        tuples.add(token);
                        it2.remove();
                        ctr--;
                    }
                }
                if(ctr > 0) {
                    if(debug) {
                        System.out.println("disabled: place " + p.getName()
                                + " does not contain enough tuples.");
                    }
                    // not enough tuples
                    Simulator.returnMarking(curmarkings, usedInts);
                    return null;
                }
                
                if(!label.getTuples().isEmpty() && tuples != null) {
                    final ValueList vals = label.getTuples();
                    for(int i2 = 0, n2 = vals.size(); i2 < n2; ++i2) {
                        final Value tuple = vals.get(i2);
                        final int idx = tuples.indexOf(tuple);
                        if(idx >= 0) {
                            tuples.remove(idx);
                        } else if(!tuple.getTuple().containsVariable()) {
                            if(debug) {
                                System.out.println("disabled: place " + p.getName()
                                        + " does not contain tuple " + tuple + ".");
                            }
                            Simulator.returnMarking(curmarkings, usedInts);
                            return null;
                        } else {
                            final Tuple t = tuple.getTuple();
                            boolean matches = false;
                            Map<String, ValueList> tupleVarBind = null;
                            for(int i3 = 0, n3 = tuples.size(); i3 < n3; ++i3) {
                                final Value v = tuples.get(i3);
                                final Tuple curt = v.getTuple();
                                if(tupleVarBind == null) {
                                    tupleVarBind = Pool.getPool().getMap();
                                }
                                if(Simulator.match(t, curt, tupleVarBind, bind)) {
                                    matches = true;
                                }
                            }
                            if(!matches) {
                                if(debug) {
                                    System.out.println("disabled: place "
                                            + p.getName() + " doesn't contain a "
                                            + "matching tuple for " + tuple
                                            + ": " + marking);
                                }
                                Simulator.returnMarking(curmarkings, usedInts);
                                return null;
                            }
                            if(tupleVarBind != null) {
                                for(final Iterator<Map.Entry<String, ValueList>> it2 = tupleVarBind.entrySet().iterator();
                                            it2.hasNext();) {
                                    final Map.Entry<String, ValueList> entry = it2.next();
                                    final String var = entry.getKey();
                                    final ValueList values = entry.getValue();
                                    it2.remove();
                                    
                                    final boolean stop = !bind.constrain(var, values, true);
                                    
                                    if(stop) {
                                        if(debug) {
                                            System.out.println("disabled: " + var + " has no possible values!");
                                        }
                                        Simulator.returnMarking(curmarkings, usedInts);
                                        return null;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            
            if(tuples != null && marking != null) {
                marking.addAll(tuples);
                tuples = null;
            } else if(marking == null) {
                marking = Simulator.getMarking(curmarkings, usedInts, p);
            }
            
            if(!label.getVariables().isEmpty()) {
                final ValueList vars = label.getVariables();
                for(int i2 = 0, n2 = vars.size(); i2 < n2; ++i2) {
                    final Value var = vars.get(i2);
                    if(!bind.hasBinding(var.getVariable())) {
                        if(marking.isEmpty()) {
                            if(debug) {
                                System.out.println("disabled: place " + p.getName()
                                        + " does not contain enough values.");
                            }
                            Simulator.returnMarking(curmarkings, usedInts);
                            return null;
                        } else if(marking.size() == 1) {
                            final Value val = marking.remove(0);
                            final Value oldBinding = bind.getBinding(var.getVariable());
                            if((oldBinding != null) && !val.equals(oldBinding)) {
                                // d'oh
                                if(debug) {
                                    System.out.println("(0.0)disabled: variable "
                                                    + var
                                                    + " has no possible values. "
                                                    + "Place " + p.getName()
                                                    + " contains only "
                                                    + marking + " but " + var
                                                    + " is bound to " + val);
                                }
                                Simulator.returnMarking(curmarkings, usedInts);
                                return null;
                            }
                            bind.addBinding(var.getVariable(), val);
                            final ValueList l = Pool.getPool().getValueList();
                            l.add(val);
                            boolean b = bind.constrain(var.getVariable(), l, true);
                            if(!b && debug) {
                                System.out.println("(0.1)disabled: variable "
                                                + var
                                                + " has no possible values. "
                                                + "Place " + p.getName()
                                                + " contains only "
                                                + marking);
                            }
                            
                        } else {
                            if(!bind.constrain(var.getVariable(), marking, false)) {
                                if(debug) {
                                    System.out.println("(0.2)disabled: variable "
                                                    + var
                                                    + " has no possible values. "
                                                    + "Place " + p.getName()
                                                    + " contains only "
                                                    + marking);
                                }
                                Simulator.returnMarking(curmarkings, usedInts);
                                return null;
                            }
                        }
                    } else {
                        final Value val = bind.getBinding(var.getVariable());
                        if(!marking.remove(val)) {
                            if(debug) {
                                System.out.println("disabled: place " + p.getName()
                                        + " does not contain value " + val
                                        + " (bound to " + var + ").");
                            }
                            // wrong binding;
                            Simulator.returnMarking(curmarkings, usedInts);
                            return null;
                        }
                    }
                }
            }
        }
        
        Node guard = trans.getGuardTree();
        if(guard.hasConstraints()) {
            for(final Map.Entry<String, ValueList> e : guard.getConstraints().entrySet()) {
                final String var = e.getKey();
                final ValueList values = e.getValue();
//                System.out.println("Constraining: " + var + " in " + values);
//                 && !values.contains(IPlace.STRING_ANY)
                if(!bind.constrain(var, values, true)) {
                    if(debug) {
                        System.out.println("(1)disabled: variable " + var
                                        + " has no possible values.\nTerm: "
                                        + guard + "\nConstraints: "
                                        + guard.getConstraints()
                                        + "\nBindConstraints: "
                                        + bind.getConstraints());
                    }
                    Simulator.returnMarking(curmarkings, usedInts);
                    return null;
                }
            }
        }
        
        if(!Value.V_TRUE.equals(guard.getToken().getValue())) {
            if(debug) {
                System.out.println("extracting binding from term.\n"
                                + "binding: " + bind + "\nterm: " + guard);
            }
            
            if(bind.hasFreeVariables()) {
                final List<String> vars = bind.getFreeVariables();
                for(int i2 = 0, n2 = vars.size(); i2 < n2; ++i2) {
                    final String var = vars.get(i2);
                    final ValueList values = bind.getConstraint(var);
                    if(!bind.hasBinding(var) && (values.size() == 1)) {
                        final Value val = values.get(0);
                        if(val.isAny()) {
                            continue;
                        }
                        bind.addDisposableBinding(var, val);
                    }
                }
            }
            
            if(!guard.canBeEvaluatedStatically()) {
                guard = guard.clone();
                guard.unifyVariables(trans.getVariableMapping());
            }
            
            boolean mayBindMore = false;
            do {
                if(debug) {
                    System.out.println("binding: " + bind);
                }
                
                if(!guard.isOrdered()) {
                    TreeComputer.transform(guard, bind);
                    TreeComputer.order(guard, bind);
                    if(debug) {
                        System.out.println("re-ordered term: " + guard);
                    }
                }
                
                if(!guard.canBeEvaluatedStatically()) {
                    mayBindMore = TreeComputer.extractBindingWithAlteration(guard,
                                    bind, this.functions, true);
                } else {
                    mayBindMore = TreeComputer.extractBinding(guard, bind,
                                    this.functions, true);
                }
                
                mayBindMore = mayBindMore
                    && !bind.getBoundVariables().containsAll(bind.getFreeVariables());
            } while(mayBindMore);
            
            if(!guard.isOrdered()) {
                TreeComputer.transform(guard, bind);
                guard.unifyVariables(trans.getVariableMapping());
                TreeComputer.order(guard, bind);
                if(debug) {
                    System.out.println("re-ordered term: " + guard);
                }
                guard.setIsOrdered(true);
            }
            
            if(debug) {
                System.out.println("extracted: " + bind);
            }
            
            bind.dispose();
            if(debug) {
                final List<String> vars = bind.getFreeVariables();
                for(int i2 = 0, n2 = vars.size(); i2 < n2; ++i2) {
                    final String var = vars.get(i2);
                    final Collection<Value> vals = bind.getConstraint(var);
                    System.out.println(var + " in " + vals);
                }
            }
        }
        
        final List<String> toDel = Pool.getPool().getList();
        if(bind.hasFreeVariables()) {
            final List<String> vars = bind.getFreeVariables();
            for(int i2 = 0, n2 = vars.size(); i2 < n2; ++i2) {
                final String var = vars.get(i2);
                final ValueList values = bind.getConstraint(var);
                if(values.isEmpty()) {
                    if(debug) {
                        System.out.println("(2)disabled: variable " + var + " has no possible values.");
                    }
                    Simulator.returnMarking(curmarkings, usedInts);
                    return null;
                } else {
                    final Value bound = bind.getBinding(var);
                    if((bound != null) && !values.contains(bound)) {
                        if(debug) {
                            System.out.println("(3)disabled: variable " + var + " has no possible values.");
                        }
                        Simulator.returnMarking(curmarkings, usedInts);
                        return null;
                        
                    } else if((values.size() == 1) && !values.get(0).isAny()) {
                        final Value val = values.get(0);
                        toDel.add(var);
                        
                        if(bound == null) {
                            final Value theVariable = trans.getVariable(var);
                            bind.addBinding(var, val);
                            
                            final List<Arc> input = trans.getInput();
                            for(int i3 = 0, n3 = input.size(); i3 < n3; ++i3) {
                                final Arc arc = input.get(i3);
                                if(!arc.isSimpleArc()) {
                                    continue;
                                }
                                final ValueList marking = Simulator.getMarking(curmarkings, usedInts, arc.getPlace());
                                final ValueList curvars = arc.getLabel().getVariables();
                                for(int i4 = 0, n4 = curvars.size(); i4 < n4; ++i4) {
                                    final Value curvar = curvars.get(i4);
                                    if(curvar == theVariable) {
                                        if(!marking.remove(val)) {
                                            if(debug) {
                                                System.out.println("disabled: variable "
                                                        + var + " was bound to " + val
                                                        + " (" + bind + ") but place " + arc.getPlace().getName()
                                                        + " didn't contain it: " + marking);
                                            }
                                            Simulator.returnMarking(curmarkings, usedInts);
                                            return null;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        
        if(!toDel.isEmpty()) {
            for(int i2 = 0, n2 = toDel.size(); i2 < n2; ++i2) {
                bind.removeConstraint(toDel.get(i2));
            }
        }
        
        try {
            if(debug) {
                System.out.println("binding so far: " + bind);
                final List<String> vars = bind.getFreeVariables();
                for(int i2 = 0, n2 = vars.size(); i2 < n2; ++i2) {
                    final String var = vars.get(i2);
                    final Collection<Value> vals = bind.getConstraint(var);
                    System.out.println(var + " in " + vals);
                }
                System.out.println("term: " + guard);
            }
            
            bind.setTransition(trans);
            final List<Binding> binds = Pool.getPool().getList();
            
            Simulator.bindOnTransition(trans, bind);
            this.findBindings(trans, guard, inputArcs, bind, curmarkings, usedInts,
                            binds, allBindings, true, debug);
            Simulator.unbindOnTransition(trans, bind);
            
            Simulator.returnMarking(curmarkings, usedInts);
            
            if(!binds.isEmpty()) {
                if(debug) {
                    System.out.println("enabled: " + binds);
                }
                return binds;
            } else {
                if(debug) {
                    System.out.println("disabled: could not find binding.");
                }
                return null;
            }
        } catch(final Exception e) {
            throw new RuntimeException("\nError: " + e + "\n"
                            + trans.getGuard() + "\nBinding: " + bind, e);
        }
    }
    
    
    
    private static List<Binding> getBindingsForTransition(final Transition trans,
                    Binding bind, final Map<String, Function> functions,
                    final boolean useTime) {
        if(bind == null) {
            bind = Pool.getPool().getBinding();
        }

        Node guard = trans.getGuardTree();
        bind.addConstraints(guard.getConstraints());
        
        if(!Value.V_TRUE.equals(guard.getToken().getValue())) {
            if(bind.hasFreeVariables()) {
                final List<String> vars = bind.getFreeVariables();
                for(int i2 = 0, n2 = vars.size(); i2 < n2; ++i2) {
                    final String var = vars.get(i2);
                    final ValueList values = bind.getConstraint(var);
                    if(!bind.hasBinding(var) && (values.size() == 1)) {
                        final Value val = values.get(0);
                        if(val.isAny()) {
                            continue;
                        }
                        bind.addDisposableBinding(var, val);
                    }
                }
            }
            
            if(!guard.canBeEvaluatedStatically()) {
                guard = guard.clone();
                guard.unifyVariables(trans.getVariableMapping());
            }
            
            boolean mayBindMore = false;
            do {
                if(!guard.isOrdered()) {
                    TreeComputer.order(guard, bind);
                }
            
                if(!guard.canBeEvaluatedStatically()) {
                    mayBindMore = TreeComputer.extractBindingWithAlteration(guard, bind, functions, true);
                } else {
                    mayBindMore = TreeComputer.extractBinding(guard, bind, functions, true);
                }
                
                mayBindMore = mayBindMore
                    && !bind.getBoundVariables().containsAll(bind.getFreeVariables());
            } while(mayBindMore);
            
            if(!guard.isOrdered()) {
                guard.setIsOrdered(true);
            }
            
            bind.dispose();
        }
        
        
        final List<String> toDel = Pool.getPool().getList();
        if(bind.hasFreeVariables()) {
            final List<String> vars = bind.getFreeVariables();
            for(int i2 = 0, n2 = vars.size(); i2 < n2; ++i2) {
                final String var = vars.get(i2);
                final ValueList values = bind.getConstraint(var);
                if(values.isEmpty()) {
                    return null;
                }
                final Value bound = bind.getBinding(var);
                if((bound != null) && !values.contains(bound)) {
                    return Pool.getPool().getList();
                    
                } else if((values.size() == 1) && !values.get(0).isAny()) {
                    final Value val = values.get(0);
                    toDel.add(var);
                    
                    if(bound == null) {
                        bind.addBinding(var, val);
                    }
                }
            }
        }
        
        if(!toDel.isEmpty()) {
            for(int i2 = 0, n2 = toDel.size(); i2 < n2; ++i2) {
                bind.removeConstraint(toDel.get(i2));
            }
        }
        
        try {
            final List<Binding> list = Pool.getPool().getList();
            
            Simulator.bindOnTransition(trans, bind);
            Simulator.findBindings(trans, guard, functions, bind, true, list, useTime);
            //System.out.println("found: " + trans.getName() + " -> " + list);
            Simulator.unbindOnTransition(trans, bind);
            
            return list;
            
        } catch(final Exception e) {
            e.printStackTrace();
            throw new RuntimeException("\nError: " + e + "\n"
                            + trans.getGuard() + "\nBinding: " + bind);
        }
    }
    
    
    public static boolean match(final Tuple t, final Tuple t2, final Map<String, ValueList> bind,
                    final Binding binding) {
        if(t.size() != t2.size()) {
            return false;
        }
        for(int i = 0; i < t.size(); i++) {
            final Value v = t.getEntries().get(i);
            final Value v2 = t2.getEntries().get(i);
            if(v.isTuple() && !(v2.isTuple() && Simulator.match(v.getTuple(),
                                               v2.getTuple(), bind, binding))) {
                return false;
            } else if(!v.isVariable() && !v.equals(v2)) {
                return false;
            } else if(v.isVariable() && !binding.hasBinding(v.getVariable())) {
                ValueList c = bind.get(v.getVariable());
                if(c == null) {
                    c = Pool.getPool().getValueList();
                    bind.put(v.getVariable(), c);
                }
                c.add(v2);
            }
        }
        return true;
    }
    
    
    public static boolean constrainTuple(final Tuple t, final IPlaceType type,
                    final Node tree, final boolean union) {
        if(!(type instanceof petruchio.sim.petrinettool.type.Tuple)) {
            return false;
        }
        petruchio.sim.petrinettool.type.Tuple t2 = (petruchio.sim.petrinettool.type.Tuple)type;
        if(t.size() != t2.size()) {
            return false;
        }
        for(int i = 0; i < t.size(); i++) {
            final Value v = t.getEntries().get(i);
            final IPlaceType tt = t2.getType(i);
            if(v.isTuple()) {
                if(!Simulator.constrainTuple(v.getTuple(), tt, tree, union)) {
                    return false;
                }
            } else if(v.isVariable()) {
                final ValueList values = Values.getValues(tt);
                final ValueList theValues = tree.getConstraint(v.getVariable());
                if(union && (theValues != null)) {
                    for(int i2 = 0, n2 = values.size(); i2 < n2; ++i2) {
                        final Value val = values.get(i2);
                        if(val.isAny()) {
                            continue;
                        }
                        if(!theValues.contains(val)) {
                            theValues.add(val);
                        }
                    }
                } else if(!tree.constrain(v.getVariable(), values)) {
                    return false;
                }
            }
        }
        return true;
    }
    
    
    public static void bindOnTransition(final Transition trans, final Binding bind) {
        for(final Map.Entry<String, Value> e : bind.getBindings()) {
            final String var = e.getKey();
            final Value val = e.getValue();
            final Value theVariable = trans.getVariable(var);
            if(theVariable == null) {
                throw new RuntimeException("Variable \"" + var + "\" not found "
                                + "in context of transition \"" + trans.getName()
                                + "\"! Variables are " + trans.getVariables());
            }
            theVariable.changeValue(val);
        }
    }
    
    
    public static void unbindOnTransition(final Transition trans, final Binding bind) {
        for(final String var : bind.getBoundVariables()) {
            final Value theVariable = trans.getVariable(var);
            theVariable.makeVariable(var);
        }
    }
    

//    public static <T> void randomizeOrder(final Collection<T> coll) {
//        final List<T> list = Pool.getPool().getList(coll);
//        coll.clear();
//        
//        int count = list.size();
//        while(count > 0) {
//            final int idx = Simulator.random.nextInt(count);
//            coll.add(list.get(idx));
//            list.remove(idx);
//            count--;
//        }
//        
//        Pool.getPool().returnList(list);
//    }
    
    
    public boolean findBindings(final Transition trans, final Node guard,
                    final List<Arc> in, final Binding bind,
                    final ValueList[] markings, final IntStack used,
//                    final Map<Place, ValueList> markings,
                    final List<Binding> binds, final boolean allBindings,
                    final boolean firstCall, final boolean debug) {
        
        if(bind.isEmptyConstraint()) {
            if(debug) {
                System.out.println("binding: " + bind);
                System.out.println("term: " + guard);
            }
            Value result = null;
            try {
                result = TreeComputer.evaluate(guard, this.functions, bind);
            } catch(Exception e) {
                // ignore, result == null
//                e.printStackTrace();
            }
            if(result == null) {
                if(debug) {
                    System.out.println("error evaluating term");
                }
                bind.dispose();
                return false;
            } else if(!(result.isBoolean() && result.getBoolean())) {
                if(debug) {
                    System.out.println("term is false or no boolean: " + result);
                }
                bind.dispose();
                return false;
            }
            if(debug) {
                System.out.println("term evaluated to true");
            }
            
            // only take tuples when all the variables are bound
            for(int i2 = 0, n2 = in.size(); i2 < n2; ++i2) {
                final Arc arc = in.get(i2);
                final Place p = arc.getPlace();
                if(arc.isInhibitorArc()) {
                    // check inhibitor arcs: test the values and tuples (w/o variables)
                    final ValueList marking = Simulator.getMarking(markings, used, p);
                    final Label label = arc.getLabel();
                    
                    final ValueList checked = Pool.getPool().getValueList();
                    final ValueList vals = label.getTuples();
                    for(int i3 = 0, n3 = vals.size(); i3 < n3; ++i3) {
                        final Value val = vals.get(i3);
                        if(!val.getTuple().containsVariable() || checked.contains(val)) {
                            continue;
                        }
                        checked.add(val);
                        final int countI = PNModel.countOccurrences(label.getTuples(), val);
                        final int countM = PNModel.countOccurrences(marking, val);
                        if(countM >= countI) {
                            if(debug) {
                                System.out.println("inhibitor arc not satisfied with "
                                                + val + " on place " + p + ": "
                                                + countM + " >= " + countI);
                            }
                            return false;
                        }
                    }
                }
                
                if(!arc.isSimpleArc()) {
                    continue;
                }
                
                final ValueList marking = Simulator.getMarking(markings, used, p);

                final Label label = arc.getLabel();
                ValueList taken = null;
                boolean stop = false;
                final ValueList vals = label.getTuples();
                for(int i3 = 0, n3 = vals.size(); i3 < n3; ++i3) {
                    final Value val = vals.get(i3);
                    final Tuple tuple = val.getTuple();
                    if(tuple.containsVariable()) {
                        final int idx = marking.indexOf(val);
                        if(idx >= 0) {
                            if(taken == null) {
                                taken = Pool.getPool().getValueList();
                            }
                            taken.add(marking.remove(idx));
                        } else {
                            if(debug) {
                                System.out.println("cannot take " + val + " = "
                                                + val.getTuple() + " from place "
                                                + p.getName()
                                                + ": " + marking);
                            }
                            stop = true;
                        }
                    }
                }
                
                if(taken != null) {
                    for(int i = 0; i < taken.size(); i++) {
                        marking.add(taken.get(i));
                    }
                }
                
                if(stop) {
                    return false;
                }
            }
            
            // handle read arcs
            // the _remaining_ tokens have to satisfy the arcs label
            for(int i2 = 0, n2 = in.size(); i2 < n2; ++i2) {
                final Arc arc = in.get(i2);
                if(arc.isReadArc()) {
                    final ValueList marking = Simulator.getMarking(markings, used, arc.getPlace());
                    
                    final Label label = arc.getLabel();
                    final ValueList vals = label.getAll();
                    for(int i3 = 0, n3 = vals.size(); i3 < n3; ++i3) {
                        final Value val = vals.get(i3);
                        if(!marking.contains(val)) {
                            return false;
                        }
                    }
                }
            }
            
            // now check inhibitor arcs ("again") on original marking
            // all variables are bound, now we can count correctly!
            for(int i2 = 0, n2 = in.size(); i2 < n2; ++i2) {
                final Arc arc = in.get(i2);
                if(arc.isInhibitorArc()) {
                    // check inhibitor arcs: test the values and tuples (w/o variables)
                    final String p = arc.getPlace().getName();
                    final ValueList marking = this.net.getPlaceByName(p).getCurrentMarkingSet();
                    final ValueList all = arc.getLabel().getAll();
                    final ValueList checked = Pool.getPool().getValueList();
                    
                    for(int i3 = 0, n3 = all.size(); i3 < n3; ++i3) {
                        final Value val = all.get(i3);
                        if(checked.contains(val)) {
                            continue;
                        }
                        checked.add(val);
                        final int countI = PNModel.countOccurrences(all, val);
                        final int countM = PNModel.countOccurrences(marking, val);
                        if(countM >= countI) {
                            if(debug) {
                                System.out.println("inhibitor arc not satisfied with "
                                                + val + " on place " + p + ": "
                                                + countM + " >= " + countI);
                            }
                            return false;
                        }
                    }
                }
            }
            
            final Binding copy = bind.clone();
            if(!this.usesTime() || trans.isImmediate()) {
                copy.setIsImmediate(true);
                if(trans.hasWeight()) {
                    try {
                        final Node weight = trans.getWeightTree();
                        final Value v = TreeComputer.evaluate(weight, this.functions, bind);
                        if(!v.isNumber()) {
                            throw new RuntimeException("Result is not a number but "
                                            + v);
                        }
                        final double val = v.getNumber();
                        if(Double.isNaN(val)) {
                            throw new RuntimeException("Result is not a number: " + val);
                        }
                        copy.setWeight(val);
                        
                    } catch (final Exception e) {
                        System.err.println("Error evaluating lambda for "
                                        + "transition " + trans + "\nat clock "
                                        + this.getTime() + " with binding " + copy
                                        + "\nError: " + e + "\nThis binding "
                                        + "will be ignored.");
                        return false;
                    }
                } else if(trans.hasGeneralizedWeight()) {
                    // will be set later by setRateAndWeight
                }
                
            } else if(trans.hasRate()) {
                try {
                    final Node rate = trans.getRateTree();
                    final Value v = TreeComputer.evaluate(rate, this.functions, bind);
                    if(!v.isNumber()) {
                        throw new RuntimeException("Result is not a number but "
                                        + v);
                    }
                    final double val = v.getNumber();
                    if(Double.isNaN(val)) {
                        throw new RuntimeException("Result is not a number: " + val);
                    }
                    copy.setRate(val);
                    
                } catch (final Exception e) {
                    System.err.println("Error evaluating rate for "
                                    + "transition " + trans + "\nat clock "
                                    + this.getTime() + " with binding " + copy
                                    + "\nError: " + e + "\nThis binding "
                                    + "will be ignored.");
                    return false;
                }
                
            } else if(trans.hasGeneralizedRate()) {
                // will be set later by setRateAndWeight
                
            } else {
                try {
                    final Node earliest = trans.getEarliestFiringTree();
                    final Value v = TreeComputer.evaluate(earliest, this.functions, bind);
                    if(!v.isNumber()) {
                        throw new RuntimeException("Result is not a number but " + v);
                    }
                    double val = v.getNumber();
                    if(Double.isNaN(val)) {
                        throw new RuntimeException("Result is not a number: " + val);
                    } else if(val == Double.POSITIVE_INFINITY) {
                        throw new RuntimeException("Lower bound is positive infinity.");
                    } else if(val < 0) {
                        val = 0.0f;
                    }
                    copy.setEarliestFiring(this.getTime() + val);
                } catch (final Exception e) {
                    System.err.println("Error evaluating lower time bound for "
                                    + "transition " + trans + "\nwith binding "
                                    +  bind + "\nError: " + e + "\nThis binding "
                                    + "will be ignored.");
                    return false;
                }
                
                try {
                    final Node latest = trans.getLatestFiringTree();
                    final Value v = TreeComputer.evaluate(latest, this.functions, bind);
                    if(!v.isNumber()) {
                        throw new RuntimeException("Result is not a number but "
                                        + v);
                    }
                    final double val = v.getNumber();
                    if(Double.isNaN(val)) {
                        throw new RuntimeException("Result is not a number: " + val);
                    }
                    copy.setLatestFiring(this.getTime() + val);
                    
                    if(copy.getLatestFiring() < copy.getEarliestFiring()) {
                        throw new RuntimeException("Upper bound is less than lower bound.");
                    }
                } catch (final Exception e) {
                    System.err.println("Error evaluating upper time bound for "
                                    + "transition " + trans + "\nat clock "
                                    + this.getTime() + " with binding " + copy
                                    + "\nError: " + e + "\nThis binding "
                                    + "will be ignored.");
                    return false;
                }
            }
            
            try {
                final Node priority = trans.getPriorityTree();
                final Value v = TreeComputer.evaluate(priority, this.functions, bind);
                if(!v.isNumber()) {
                    throw new RuntimeException("Result is not a number but "
                                    + v);
                }
                final double val = v.getNumber();
                if(Double.isNaN(val)) {
                    throw new RuntimeException("Result is not a number: " + val);
                }
                copy.setPriority(val);
            } catch (final Exception e) {
                System.err.println("Error evaluating precedence for "
                                + "transition " + trans + "\nwith binding "
                                + copy + "\nError: " + e + "\nThis binding "
                                + "will be ignored.");
                //e.printStackTrace();
                return false;
            }
            
            
            if(!copy.isImmediate()) {
                try {
                    final Node servers = trans.getServersTree();
                    final Value v = TreeComputer.evaluate(servers, this.functions, bind);
                    if(!v.isNumber()) {
                        throw new RuntimeException("Result is not a number but "
                                        + v);
                    }
                    final double val = Math.ceil(v.getNumber());
                    if(Double.isNaN(val)) {
                        throw new RuntimeException("Result is not a number: " + val);
                    }
                    copy.setServers(val);
                } catch (final Exception e) {
                    System.err.println("Error evaluating server semantics for "
                                    + "transition " + trans + "\nwith binding "
                                    + copy + "\nError: " + e + "\nThis binding "
                                    + "will be ignored.");
                    //e.printStackTrace();
                    return false;
                }
                
                
                try {
                    final Node useAgeMemory = trans.getAgeMemoryTree();
                    final Value v = TreeComputer.evaluate(useAgeMemory, this.functions, bind);
                    if(!v.isBoolean()) {
                        throw new RuntimeException("Result is not a boolean but "
                                        + v);
                    }
                    final boolean val = v.getBoolean();
                    copy.setUseAgeMemory(val);
                } catch (final Exception e) {
                    System.err.println("Error evaluating memory policy for "
                                    + "transition " + trans + "\nwith binding " + copy
                                    + "\nError: " + e + "\nThis binding "
                                    + "will be ignored.");
                    //e.printStackTrace();
                    return false;
                }
            }
            
            this.setRateAndWeight(copy, true);
            copy.setCreationTime(this.getTime());
            copy.makeUndisposable();
            
            binds.add(copy);
            
            return !allBindings;
        } else { ///////////////////////////////////////////////////////////////
            final List<String> vars = bind.getFreeVariables();
            final String var = vars.remove(vars.size()-1);
            final Value theVariable = trans.getVariable(var);
            
            final ValueList values = bind.removeConstraintNoVar(var);
            final ValueList theValues;
            
            theValues = TreeComputer.extractConstraints(guard,
                            theVariable, values, this.functions, bind.getConstraints());

            if((theValues.size() == 1) && theValues.contains(Value.V_ANY)) {
                final boolean done = this.findBindings(trans, guard, in, bind,
                                markings, used, binds, allBindings, false, debug);
                bind.setConstraint(var, values);
                return done;
                
            } else {
                if(theValues.isEmpty()) {
                    bind.setConstraint(var, values);
                    return false;
                }
                
                final boolean take = !bind.hasBinding(var);
                
                for(int i = 0, n = theValues.size(); i < n; ++i) {
                    final Value v = theValues.get(i);
                    if(v.isAny()) {
                        continue;
                    }
                    theVariable.changeValue(v);
                    Arc stopAt = null;
                    
                    if(take) {
                        for(int i2 = 0, n2 = in.size(); i2 < n2; ++i2) {
                            final Arc arc = in.get(i2);
                            if(stopAt != null) {
                                break;
                            }

                            final Place p = arc.getPlace();
                            if(arc.isInhibitorArc()) {
                                final ValueList marking = Simulator.getMarking(markings, used, p);
                                final Label label = arc.getLabel();
                                final ValueList checked = Pool.getPool().getValueList();
                                final ValueList curvars = label.getVariables();
                                for(int i3 = 0, n3 = curvars.size(); i3 < n3; ++i3) {
                                    final Value val = curvars.get(i3);
                                    if((val != theVariable) || checked.contains(val)) {
                                        continue;
                                    }
                                    checked.add(val);
                                    final int countI = PNModel.countOccurrences(label.getVariables(), val);
                                    final int countM = PNModel.countOccurrences(marking, val);
                                    if(countM >= countI) {
                                        return false;
                                    }
                                }
                            }
                            
                            if(!arc.isSimpleArc()) {
                                continue;
                            }
                            final ValueList marking = Simulator.getMarking(markings, used, p);
                            
                            final Label label = arc.getLabel();
                            int count = 0;
                            final ValueList curvars = label.getVariables();
                            for(int i3 = 0, n3 = curvars.size(); (i3 < n3) && (stopAt == null); ++i3) {
                                final Value curvar = curvars.get(i3);
                                if(curvar == theVariable) {
                                    if(!marking.remove(v)) {
                                        stopAt = arc;
                                        for(; count > 0; --count) {
                                            marking.add(v);
                                        }
                                    } else {
                                        ++count;
                                    }
                                }
                            }
                        }
                    }
                    
                    if(stopAt == null) {
                        bind.addBinding(var, v);
                        final boolean done = this.findBindings(trans, guard,
                                        in, bind, markings, used, binds,
                                        allBindings, false, debug);
                        if(done) {
                            if(take) {
                                for(int i2 = 0, n2 = in.size(); i2 < n2; ++i2) {
                                    final Arc arc = in.get(i2);
                                    if(arc == stopAt) {
                                        break;
                                    }
                                    if(!arc.isSimpleArc()) {
                                        continue;
                                    }
                                    final ValueList marking = Simulator.getMarking(markings, used, arc.getPlace());
                                    if(marking == null) {
                                        throw new RuntimeException("ERROR");
                                    }
                                    
                                    final Label label = arc.getLabel();
                                    for(final Value curvar : label.getVariables()) {
                                        if(take && (curvar == theVariable)) {
                                            marking.add(v);
                                        }
                                    }
                                }
                            }
                            theVariable.makeVariable(var);
                            return done;
                        }
                    }
                    
                    if(take) {
                        for(final Arc arc : in) {
                            if(arc == stopAt) {
                                break;
                            }
                            if(!arc.isSimpleArc()) {
                                continue;
                            }
                            final ValueList marking = Simulator.getMarking(markings, used, arc.getPlace());
                            if(marking == null) {
                                throw new RuntimeException("ERROR");
                            }
                            
                            final Label label = arc.getLabel();
                            final ValueList curvars = label.getVariables();
                            for(int i3 = 0, n3 = curvars.size(); i3 < n3; ++i3) {
                                final Value curvar = curvars.get(i3);
                                if(take && (curvar == theVariable)) {
                                    marking.add(v);
                                }
                            }
                        }
                    }
                    
                }
            }
            bind.removeBinding(var);
            theVariable.makeVariable(var);
            bind.setConstraint(var, values);
            
            return false;
        }
    }
    
    
    // for unfolding
    public static void findBindings(final Transition trans, final Node guard,
                    final Map<String, Function> functions, final Binding bind,
                    final boolean firstCall, final List<Binding> list,
                    final boolean useTime) {
        if(bind.isEmptyConstraint()) {
            try {
                final Value result = TreeComputer.evaluate(guard, functions, bind);
                if((result != null) && (result.isBoolean() && result.getBoolean())) {
                    final Binding copy = bind.clone();
                    
                    if(useTime) {
                        if(trans.isImmediate()) {
                            copy.setIsImmediate(true);
                            if(trans.hasWeight()) {
                                try {
                                    final Node weight = trans.getWeightTree();
                                    final Value v = TreeComputer.evaluate(weight, functions, bind);
                                    if(!v.isNumber()) {
                                        throw new RuntimeException("Result is not a number but "
                                                        + v);
                                    }
                                    final double val = v.getNumber();
                                    if(Double.isNaN(val)) {
                                        throw new RuntimeException("Result is not a number: " + val);
                                    }
                                    copy.setWeight(val);
                                    
                                } catch (final RuntimeException e) {
                                    System.err.println("Error evaluating lambda for "
                                                    + "transition " + trans + "\n"
                                                    + " with binding " + copy
                                                    + "\nError: " + e + "\nThis binding "
                                                    + "will be ignored.");
                                    throw e;
                                }
                            }
                            
                        } else if(trans.hasRate()) {
                            try {
                                final Node rate = trans.getRateTree();
                                final Value v = TreeComputer.evaluate(rate, functions, bind);
                                if(!v.isNumber()) {
                                    throw new RuntimeException("Result is not a number but "
                                                    + v);
                                }
                                final double val = v.getNumber();
                                if(Double.isNaN(val)) {
                                    throw new RuntimeException("Result is not a number: " + val);
                                }
                                copy.setRate(val);
                                
                            } catch (final RuntimeException e) {
                                System.err.println("Error evaluating rate for "
                                                + "transition " + trans + "\n"
                                                + " with binding " + copy
                                                + "\nError: " + e + "\nThis binding "
                                                + "will be ignored.");
                                throw e;
                            }
                            
                        } else {
                            try {
                                final Node earliest = trans.getEarliestFiringTree();
                                final Value v = TreeComputer.evaluate(earliest, functions, bind);
                                if(!v.isNumber()) {
                                    throw new RuntimeException("Result is not a number but " + v);
                                }
                                double val = v.getNumber();
                                if(Double.isNaN(val)) {
                                    throw new RuntimeException("Result is not a number: " + val);
                                } else if(val == Double.POSITIVE_INFINITY) {
                                    throw new RuntimeException("Lower bound is positive infinity.");
                                } else if(val < 0) {
                                    val = 0.0f;
                                }
                                copy.setEarliestFiring(val);
                            } catch (final RuntimeException e) {
                                System.err.println("Error evaluating lower time bound for "
                                                + "transition " + trans + "\nwith binding "
                                                +  bind + "\nError: " + e + "\nThis binding "
                                                + "will be ignored.");
                                throw e;
                            }
                            
                            try {
                                final Node latest = trans.getLatestFiringTree();
                                final Value v = TreeComputer.evaluate(latest, functions, bind);
                                if(!v.isNumber()) {
                                    throw new RuntimeException("Result is not a number but "
                                                    + v);
                                }
                                final double val = v.getNumber();
                                if(Double.isNaN(val)) {
                                    throw new RuntimeException("Result is not a number: " + val);
                                }
                                copy.setLatestFiring(val);
                                
                                if(copy.getLatestFiring() < copy.getEarliestFiring()) {
                                    throw new RuntimeException("Upper bound is less than lower bound.");
                                }
                            } catch (final RuntimeException e) {
                                System.err.println("Error evaluating upper time bound for "
                                                + "transition " + trans + "\n"
                                                + " with binding " + copy
                                                + "\nError: " + e + "\nThis binding "
                                                + "will be ignored.");
                                throw e;
                            }
                        }
                    }
                    
                    try {
                        final Node priority = trans.getPriorityTree();
                        final Value v = TreeComputer.evaluate(priority, functions, bind);
                        if(!v.isNumber()) {
                            throw new RuntimeException("Result is not a number but "
                                            + v);
                        }
                        final double val = v.getNumber();
                        if(Double.isNaN(val)) {
                            throw new RuntimeException("Result is not a number: " + val);
                        }
                        copy.setPriority(val);
                    } catch (final RuntimeException e) {
                        System.err.println("Error evaluating precedence for "
                                        + "transition " + trans + "\nwith binding "
                                        +  copy + "\nError: " + e + "\nThis binding "
                                        + "will be ignored.");
                        throw e;
                    }
                    
                    
                    if(!copy.isImmediate()) {
                        try {
                            final Node servers = trans.getServersTree();
                            final Value v = TreeComputer.evaluate(servers, functions, bind);
                            if(!v.isNumber()) {
                                throw new RuntimeException("Result is not a number but "
                                                + v);
                            }
                            final double val = Math.ceil(v.getNumber());
                            if(Double.isNaN(val)) {
                                throw new RuntimeException("Result is not a number: " + val);
                            }
                            copy.setServers(val);
                        } catch (final RuntimeException e) {
                            System.err.println("Error evaluating server semantics for "
                                            + "transition " + trans + "\nwith binding " + copy
                                            + "\nError: " + e + "\nThis binding "
                                            + "will be ignored.");
                            //e.printStackTrace();
                            throw e;
                        }
                        
                        
                        try {
                            final Node useAgeMemory = trans.getAgeMemoryTree();
                            final Value v = TreeComputer.evaluate(useAgeMemory, functions, bind);
                            if(!v.isBoolean()) {
                                throw new RuntimeException("Result is not a boolean but "
                                                + v);
                            }
                            final boolean val = v.getBoolean();
                            copy.setUseAgeMemory(val);
                        } catch (final RuntimeException e) {
                            System.err.println("Error evaluating memory policy for "
                                            + "transition " + trans + "\nwith binding " + copy
                                            + "\nError: " + e + "\nThis binding "
                                            + "will be ignored.");
                            //e.printStackTrace();
                            throw e;
                        }
                    }
                    
                    copy.makeUndisposable();
                    
                    list.add(copy);
                }
            } catch(final RuntimeException e) {
                // ignore
            }
            
            bind.dispose();
            
        } else {
            final List<String> vars = bind.getFreeVariables();
            final String var = vars.remove(vars.size()-1);
            final Value theVariable = trans.getVariable(var);
            
            final ValueList values = bind.removeConstraintNoVar(var);
            final ValueList theValues;
            if(firstCall) {
                theValues = values;
            } else {
                theValues = TreeComputer.extractConstraints(guard, theVariable,
                                values, functions, bind.getConstraints());
            }

            if(!theValues.isEmpty()) {
                for(int i = 0, n = theValues.size(); i < n; ++i) {
                    final Value v = theValues.get(i);
                    if(v.isAny()) {
                        continue;
                    }
                    theVariable.changeValue(v);
                    bind.addBinding(var, v);
                    Simulator.findBindings(trans, guard, functions, bind, false,
                                    list, useTime);
                }
                bind.removeBinding(var);
                theVariable.makeVariable(var);
            }
            
            bind.setConstraint(var, values);
        }
    }
    
    
    public String fire(final String trans) {
        final Transition t = this.net.getTransitionByName(trans);
        if(trans != null) {
            final List<Binding> binds = this.getBindings(t);
            if(binds == null) {
                return null;
            }
            final Binding bind;
            if(this.usesTime()) {
                final int max = this.getLastFireableBinding(binds);
                bind = Simulator.choseRandom(binds, 0, max);
            } else {
                bind = Simulator.choseRandom(binds);
            }
            
            if(bind == null) {
                return null;
            }
            
            if(this.firesParallel()) {
                return this.fireTake(t, bind);
            } else {
                return this.fire(t, bind, t.getInput(), t.getOutput());
            }
        } else {
            return null;
        }
    }
    
    
//    public String undoFire(final String trans) {
//        final Transition t = this.net.getTransitionByName(trans);
//        if(trans != null) {
//            final List<Binding> binds = this.getBindings(t);
//            if(binds == null) {
//                return null;
//            }
//            final Binding bind;
//            if(this.usesTime()) {
//                int max = this.getLastFireableBinding(binds);
//                bind = Simulator.choseRandom(binds, 0, max);
//            } else {
//                bind = Simulator.choseRandom(binds);
//            }
//            
//            if(bind == null) {
//                return null;
//            }
//            
//            if(this.firesParallel()) {
//                return this.undoFireTake(t, bind);
//            } else {
//                if(this.usesResetArcs()) {
//                    for(final Arc arc : t.getInput()) {
//                        if(arc.isResetArc()) {
//                            throw new UnsupportedOperationException("Cannot undo firing when reset arcs are involved.");
//                        }
//                    }
//                }
//                return this.fire(t, bind, t.getOutput(), t.getInput());
//            }
//        } else {
//            return null;
//        }
//    }

    
    public String fireTake(final String trans) {
        return this.fire(trans);
    }
    
    
//    public String undoFireTake(final String trans) {
//        return this.undoFire(trans);
//    }
    
    
    public String firePutWithBinding(final String complete) {
        final int start = complete.indexOf('{');
        final String transName = complete.substring(0, start).trim();
        final Transition trans = this.net.getTransitionByName(transName);
        final Binding bind = Simulator.parseBinding(complete.substring(start+1,
                        complete.lastIndexOf("}")));
        bind.setTransition(trans);
        final List<Binding> theBindings = this.getFiringBindings(trans);
        Binding theBinding = null;
        if(theBindings != null) {
            for(int i = 0, n = theBindings.size(); i < n; ++i) {
                final Binding b = theBindings.get(i);
                if(b.equalBindings(bind)) {
                    theBinding = bind;
                    break;
                }
            }
        }
        
        if(theBinding != null) {
            String resultingBinding = this.firePut(trans, theBinding);
            return resultingBinding;
        } else {
            return null;
        }
    }
    
    
//    public String undoFirePutWithBinding(final String complete) {
//        final int start = complete.indexOf('{');
//        final String transName = complete.substring(0, start).trim();
//        final Transition trans = this.net.getTransitionByName(transName);
//        if(trans == null) {
//            return null;
//        }
//        final Binding bind = Simulator.parseBinding(complete.substring(start+1,
//                        complete.lastIndexOf("}")));
//        bind.setTransition(trans);
//        final List<Binding> theBindings = this.getFiringBindings(trans);
//        Binding theBinding = null;
//        if(theBindings != null) {
//            for(final Binding b : theBindings) {
//                if(b.equalBindings(bind)) {
//                    theBinding = bind;
//                    break;
//                }
//            }
//        }
//        Pool.getPool().returnBinding(bind);
//        
//        if(theBinding != null) {
//            String resultingBinding = this.undoFirePut(trans, theBinding);
//            return resultingBinding;
//        } else {
//            return null;
//        }
//    }
    
    
    public String firePut(final String trans) {
        final Transition t = this.net.getTransitionByName(trans);
        if(trans != null) {
            final List<Binding> theBindings = this.getFiringBindings(t);
            final Binding bind = Simulator.choseRandom(theBindings);
            if(bind != null) {
                return this.firePut(t, bind);
            } else {
                return null;
            }
        } else {
            return null;
        }
    }
    
    
//    public String undoFirePut(final String trans) {
//        final Transition t = this.net.getTransitionByName(trans);
//        if(trans != null) {
//            final List<Binding> theBindings = this.getFiringBindings(t);
//            final Binding bind = Simulator.choseRandom(theBindings);
//            if(bind != null) {
//                return this.undoFirePut(t, bind);
//            } else {
//                return null;
//            }
//        } else {
//            return null;
//        }
//    }
    
    
    public static <T> T choseRandom(final List<T> col, final int from, final int to) {
        if((col == null) || col.isEmpty() || (from > to)) {
            return null;
        }
        if(from == to) {
            return col.get(from);
        }
        final int i = from + Simulator.random.nextInt(to - from + 1);
//        System.out.println(from + ".." + to + " -> " + i);
        return col.get(i);
    }
    
    
    public static <T> T choseRandom(final List<T> col) {
        return Simulator.choseRandom(col, 0, col.size()-1);
    }
    
    
    public static <T> T choseRandom(final Collection<T> col) {
        if((col == null) || col.isEmpty()) {
            return null;
        }
        if(col.size() == 1) {
            final Iterator<T> it = col.iterator();
            return it.next();
        }
        int i = Simulator.random.nextInt(col.size());
        final Iterator<T> it = col.iterator();
        while(i > 0) {
            --i;
            it.next();
        }
        return it.next();
    }
    
    
    public String fireNext() {
        final Binding bind;
        if(this.usesGlobalPriority() && this.usesPriority()) {
            if(!this.bindingsImmediate.isEmpty()) {
                final int max = this.getLastBindingWithMaximumPriority(this.bindingsImmediate);
                bind = Simulator.choseRandom(this.bindingsImmediate, 0, max);
            } else {
                final int max = this.getLastFireableBindingWithMaximumPriorityAndWeight(this.bindingsEarliest);
                bind = Simulator.choseRandom(this.bindingsEarliest, 0, max);
            }
        } else {
            if(!this.bindingsImmediate.isEmpty()) {
                final int max = this.bindingsImmediate.size()-1;
                bind = Simulator.choseRandom(this.bindingsImmediate, 0, max);
            } else {
                final int max = this.getLastFireableBinding(this.bindingsEarliest);
                bind = Simulator.choseRandom(this.bindingsEarliest, 0, max);
            }
        }
        
        final Binding maxBinding;
        if(this.usesGlobalPriority() || !this.usesPriority()) {
            if(bind.isImmediate() && this.usesWeights()) {
                maxBinding = this.chooseBindingByLocalPriorityAndWeight(bind);
            } else {
                maxBinding = bind;
            }
        } else {
            maxBinding = this.chooseBindingByLocalPriorityAndWeight(bind);
        }
        final Transition maxTransition = maxBinding.getTransition();
        return this.fire(maxTransition, maxBinding,
                        maxTransition.getInput(), maxTransition.getOutput());
    }

    
    private Binding chooseBindingByLocalPriorityAndWeight(final Binding bind) {
        final Transition trans = bind.getTransition();
        Binding maxPrioBinding = bind;
        // iterate over pre set of the chosen transition
        final List<Arc> in = trans.getInput();
        for(int i = 0, n = in.size(); i < n; ++i) {
            final Arc inarc = in.get(i);
            if(!inarc.isSimpleArc()) {
                continue;
            }
            final Place place = inarc.getPlace();
            // iterate over post set of the pre set the chosen transition
            // -> all possibly conflicting transitions
            final List<Arc> out = place.getOutput();
            for(int i2 = 0, n2 = out.size(); i2 < n2; ++i2) {
                final Arc outarc = out.get(i2);
                if(!outarc.isSimpleArc()) {
                    continue;
                }                    
                final Transition conflictTransition = outarc.getTransition();
                if(conflictTransition == trans) {
                    continue;
                }
                
                // now see if the conflicted transition is enabled with a higher binding;
                final List<Binding> binds = this.getBindings(conflictTransition);
                if((binds == null) || binds.isEmpty()) {
                    // conflictTransition isn't enabled
                    continue;
                }
                
                // this list is ordered ascending with earliest time and descending with priority
                final Binding first = binds.get(0);
                if((first.getEarliestFiring() >= this.getTime())
                                && (first.getPriority() >= maxPrioBinding.getPriority())
                                && (first.getCurrentWeight() >= maxPrioBinding.getCurrentWeight())) {
                    final int max = this.getLastFireableBindingWithMaximumPriorityAndWeight(binds);
                    maxPrioBinding = Simulator.choseRandom(binds, 0, max);
                }
            }
        }
        return maxPrioBinding;
    }
    
    
    private int getLastFireableBindingWithMaximumPriorityAndWeight(final List<Binding> binds) {
        if((binds == null) || binds.isEmpty()) {
            return -1;
        }
        
        final Binding first = binds.get(0);
        final double e = this.getTime();
        final double p = first.getPriority();
        final double w = first.getCurrentWeight();
        final boolean i = first.isImmediate();
        
        // binary search
        // at the end it should be:
        // binds[0].earliest == binds[from].earliest < binds[from+1].earliest
        // binds[0].priority == binds[from].priority > binds[from+1].priority
        int from = 0;
        int to = binds.size()-1;
        while(from < to) {
            final int middle = (from + to + 1)>>1; // round up! important to terminate
            final Binding c = binds.get(middle);
            final double ce = c.getEarliestFiring();
            final double cp = c.getPriority();
            final double cw = c.getCurrentWeight();
            final boolean ci = c.isImmediate();
            
            if((cp == p) && (cw == w) && (ci || (!i && (ce <= e)))) {
                from = middle;
            } else {
                to = middle-1;
            }
        }
        
        return from;
    }
    
    
    private int getLastFireableBinding(final List<Binding> binds) {
        if((binds == null) || binds.isEmpty() || binds.get(0).getEarliestFiring() > this.getTime()) {
            return -1;
        }
        
        final Binding first = binds.get(0);
        final double e = this.getTime();
        final boolean i = first.isImmediate();
        
        // binary search
        // at the end it should be:
        // binds[0].earliest == binds[from].earliest < binds[from+1].earliest
        // binds[0].priority == binds[from].priority > binds[from+1].priority
        int from = 0;
        int to = binds.size()-1;
        while(from < to) {
            final int middle = (from + to + 1)>>1; // round up! important to terminate
            final Binding c = binds.get(middle);
            final double ce = c.getEarliestFiring();
            final boolean ci = c.isImmediate();
            
            if(ci || (!i && (ce <= e))) {
                from = middle;
            } else {
                to = middle-1;
            }
        }
        
        return from;
    }
    
    
    private int getLastBindingWithMaximumPriority(final List<Binding> binds) {
        if((binds == null) || binds.isEmpty() || binds.get(0).getEarliestFiring() > this.getTime()) {
            return -1;
        }

        final double p = binds.get(0).getPriority();
        
        // binary search
        // at the end it should be:
        // binds[0].earliest == binds[from].earliest < binds[from+1].earliest
        // binds[0].priority == binds[from].priority > binds[from+1].priority
        int from = 0;
        int to = binds.size()-1;
        while(from < to) {
            final int middle = (from + to + 1)>>1; // round up! important to terminate
            final Binding c = binds.get(middle);
            final double cp = c.getPriority();
            
            if(cp == p) {
                from = middle;
            } else {
                to = middle-1;
            }
        }
        
        return from;
    }
    
    
    public String fireNextStep() {
        final List<Binding> binds;
        final int max;
        if(this.usesGlobalPriority() && this.usesPriority()) {
            if(this.usesTime()) {
                // immediate transitions have a higher priority than timed transitions
                if(this.bindingsImmediate.isEmpty()) {
                    binds = this.bindingsEarliest;
                    max = this.getLastFireableBindingWithMaximumPriorityAndWeight(binds);
                } else {
                    binds = this.bindingsImmediate;
                    max = this.getLastBindingWithMaximumPriority(binds);
                }
            } else {
                binds = this.bindingsImmediate;
                max = this.getLastBindingWithMaximumPriority(binds);
            }
        } else {
            if(this.usesTime()) {
                // immediate transitions have a higher priority than timed transitions
                if(this.bindingsImmediate.isEmpty()) {
                    binds = this.bindingsEarliest;
                    max = this.getLastFireableBinding(binds);
                } else {
                    binds = this.bindingsImmediate;
                    max = binds.size();
                }
            } else {
                binds = this.bindingsImmediate;
                max = binds.size();
            }
        }
        final List<Binding> put = this.putBindingsEarliest;
        final int count = max + put.size();
        if(count == 0) {
            throw new RuntimeException("Nothing to fire");
        }
        
        
        final int i = Simulator.random.nextInt(count);
        if(i >= put.size()) {
            final Binding bind = Simulator.choseRandom(binds, 0, max);
            
            final Binding maxBinding;
            if(this.usesGlobalPriority() || !this.usesPriority()) {
                if(bind.isImmediate() && this.usesWeights()) {
                    maxBinding = this.chooseBindingByLocalPriorityAndWeight(bind);
                } else {
                    maxBinding = bind;
                }
            } else {
                maxBinding = this.chooseBindingByLocalPriorityAndWeight(bind);
            }
            final Transition maxTransition = maxBinding.getTransition();
            return this.fireTake(maxTransition, maxBinding);
            
        } else {
            final Binding bind = put.get(i);
            return this.firePut(bind.getTransition(), bind);
        }
    }
    
    
    public static Tuple bindTuple(final Tuple tuple, final Binding bind) {
        final Tuple t = new Tuple();
        final List<Value> entries = tuple.getEntries();
        for(int i = 0, n = entries.size(); i < n; ++i) {
            final Value v = entries.get(i);
            if(v.isVariable()) {
                Value value = bind.getBinding(v.getVariable()); 
                if(value == null) {
                    final Collection<Value> col = bind.getConstraint(v.getVariable());
                    final Iterator<Value> it2 = col.iterator();
                    Value o;
                    while((o = it2.next()).isAny());
                    value = o;
                    bind.addBinding(v.getVariable(), value);
                }
                t.addEntry(value);
            } else {
                t.addEntry(v);
            }
        }
        return t;
    }

    
    public static Tuple bindTuple(final Tuple tuple, final Value var, final Value val) {
        final Tuple t = new Tuple();
        final List<Value> entries = tuple.getEntries();
        for(int i = 0, n = entries.size(); i < n; ++i) {
            final Value v = entries.get(i);
            if(v.equals(var)) {
                t.addEntry(val);
            } else {
                t.addEntry(v);
            }
        }
        return t;
    }
    
    
    private String getMarking(final Iterator<Place> places) {
        final StringBuilder buf = new StringBuilder();
        while(places.hasNext()) {
            final Place p = places.next();
            buf.append(p.getName());
            buf.append("{");
            buf.append(p.getCurrentMarking());
            buf.append("}");
            if(places.hasNext()) {
                buf.append("; ");
            }
        }
        return buf.toString();
    }
    
    
    public String getMarking(final Collection<String> placeNames) {
        final StringBuilder buf = new StringBuilder();
        for(final Iterator<String> it = placeNames.iterator(); it.hasNext();) {
            final String name = it.next();
            final Place p = this.net.getPlaceByName(name);
            if(p != null) {
                buf.append(name);
                buf.append("{");
                buf.append(p.getCurrentMarking());
                buf.append("}");
            } else {
                buf.append("\"");
                buf.append(name);
                buf.append("\" not found");
            }
            if(it.hasNext()) {
                buf.append("; ");
            }
        }
        return buf.toString();
    }
    
    
    public String getMarking(final Map<String, String> placeNames) {
        final StringBuilder buf = new StringBuilder();
        for(final Iterator<Map.Entry<String, String>> it = placeNames.entrySet().iterator(); it.hasNext();) {
            final Map.Entry<String, String> entry = it.next();
            final String name = entry.getKey();
            final String out = entry.getValue();
            final Place p = this.net.getPlaceByName(name);
            if(p != null) {
                buf.append(out);
                buf.append("{");
                buf.append(p.getCurrentMarking());
                buf.append("}");
            } else {
                buf.append("\"");
                buf.append(name);
                buf.append("\" not found");
            }
            if(it.hasNext()) {
                buf.append("; ");
            }
        }
        return buf.toString();
    }
    
    
    public String getMarking() {
        return this.getMarking(this.net.getPlaceIterator());
    }
    
    
    private String fire(final Transition trans, final Binding bind,
                    final List<Arc> in, final List<Arc> out) {
        if((trans == null) || (bind == null)) {
            throw new RuntimeException("Cannot fire " + trans + " with binding "
                            + bind);
        }
        
        if(this.checkFire() && !this.neverFired.isEmpty()) {
            this.neverFired.remove(trans.getName());
        }
        
        Simulator.bindOnTransition(trans, bind);
        
        final Map<Transition, Transition> transitionsToCheck = Pool.getPool().getMap();
        if(in.isEmpty()) {
            transitionsToCheck.put(trans, trans);
        }
        final Map<Place, ValueList> inmarking = Pool.getPool().getMap();
        final List<Place> outmarking = Pool.getPool().getList();
        for(int i = 0, n = in.size(); i < n; ++i) {
            final Arc arc = in.get(i);
            if(arc.isReadArc()) {
                transitionsToCheck.put(trans, trans);
            } 
            if(!arc.isSimpleArc()) {
                continue;
            }
            final Place p = arc.getPlace();
            inmarking.put(p, Pool.getPool().getValueList(p.getCurrentMarkingSet()));
            
            final Label label = arc.getLabel();
            final ValueList all = label.getAll();
            for(int i2 = 0, n2 = all.size(); i2 < n2; ++i2) {
                final Value value = all.get(i2);
                if(!p.removeToken(value)) {
                    value.setNetValue();
                    throw new RuntimeException("ERROR firing " + trans + "\nBind: "
                                    + bind + "\nArc: " + arc
                                    + "\nCannot take " + value
                                    + " from place " + p);
                }
            }
        }
        
        // now handle reset arcs
        for(int i = 0, n = in.size(); i < n; ++i) {
            final Arc arc = in.get(i);
            if(!arc.isResetArc()) {
                continue;
            }
            final Place p = arc.getPlace();
            if(!inmarking.containsKey(p)) {
                inmarking.put(p, Pool.getPool().getValueList(p.getCurrentMarkingSet()));
            }
            final Label label = arc.getLabel();
            final ValueList all = label.getAll();
            for(int i2 = 0, n2 = all.size(); i2 < n2; ++i2) {
                final Value value = all.get(i2);
                if(value.isAny()) {
                    p.clearMarking();
                } else {
                    p.removeAllTokens(value);
                }
            }
        }
        
        for(int i = 0, n = out.size(); i < n; ++i) {
            final Arc arc = out.get(i);
            if(!arc.isSimpleArc()) {
                continue;
            }
            final Place p = arc.getPlace();
            outmarking.add(p);
            
            final Label label = arc.getLabel();
            final ValueList all = label.getAll();
            for(int i2 = 0, n2 = all.size(); i2 < n2; ++i2) {
                final Value value = all.get(i2);
                p.addToken(value.clone());
            }
        }
        
        Simulator.unbindOnTransition(trans, bind);
        final String bindingString;
        if(this.getProperty(Option.SHOW_FIRING)) {
            bindingString = bind.toStringSlim();
        } else {
            bindingString = null;
        }
//        if(bind.useAgeMemory()) {
//            trans.resetEnabledTime(bind, this.getTime());
//        }
        this.removeBinding(bind);
        
        // iterate over pre set of transition
        for(final Map.Entry<Place, ValueList> entry : inmarking.entrySet()) {
            final Place p = entry.getKey();
            final ValueList oldMarking = entry.getValue();
            
            if(outmarking.remove(p)) {
                // this works for inhibitor and read arcs too
                if(!PNModel.setEquals(p.getCurrentMarkingSet(), oldMarking)) {
                    // something changed on these places -> check all trans
                    final List<Arc> pout = p.getOutput();
                    for(int i2 = 0, n2 = pout.size(); i2 < n2; ++i2) {
                        final Arc arc = pout.get(i2);
                        final Transition t = arc.getTransition();
                        //System.out.println("x?x: " + t.getName());
                        transitionsToCheck.put(t, t);
                    }
                } else {
                    transitionsToCheck.put(trans, trans);
                }
                
            } else {
                // tokens were removed
                // iterate over post set of place
                final List<Arc> pout = p.getOutput();
                for(int i = 0, n = pout.size(); i < n; ++i) {
                    final Arc arc = pout.get(i);
                    final Transition t = arc.getTransition();
                    
                    if(arc.isInhibitorArc()) {
                        // check enabled and disabled transitions
                        //System.out.println("x?x_i: " + t.getName());
                        transitionsToCheck.put(t, t);
                    }
                    
                    if(!arc.isSimpleArc() && !arc.isReadArc() && !arc.isResetArc()) {
                        continue;
                    }
                    
                    // check enabled transitions
                    if((t == trans) || (this.getBindings(t) != null)) {
                        if(!p.getCurrentMarkingSet().isEmpty() || ((t == trans) && arc.isResetArc())) {
//                            System.out.println("e?d: " + t.getName());
                            transitionsToCheck.put(t, t);
                        } else { // shortcut
                            if(t.hasAgeMemory()) {
                                transitionsToCheck.put(t, t);
                            } else {
//                              System.out.println("shortcut: " + t.getName());
                                this.removeBindings(t);
                            }
                        }
                    }
                }
            }
            
            if(!p.infiniteCapacity()) {
                if(p.getCurrentMarkingSet().size() < oldMarking.size()) {
                    // iterate over pre set of place
                    final List<Arc> pin = p.getInput();
                    for(int i = 0, n = pin.size(); i < n; ++i) {
                        final Arc arc = pin.get(i);
                        final Transition t = arc.getTransition();
                        if(this.getBindings(t) == null) {
                            //System.out.println("d?e_c: " + t.getName());
                            transitionsToCheck.put(t, t);
                        }
                    }
                } else if(p.getCurrentMarkingSet().size() > oldMarking.size()) {
                    // iterate over pre set of place
                    final List<Arc> pin = p.getInput();
                    for(int i = 0, n = pin.size(); i < n; ++i) {
                        final Arc arc = pin.get(i);
                        final Transition t = arc.getTransition();
                        if(this.getBindings(t) != null) {
                            //System.out.println("e?d_c: " + t.getName());
                            transitionsToCheck.put(t, t);
                        }
                    }
                }
            }
            
        }
        
        // iterate over post set of transition
        for(int i = 0, n = outmarking.size(); i < n; ++i) {
            final Place p = outmarking.get(i);
            // tokens were added
            // iterate over post set of place
            final List<Arc> pout = p.getOutput();
            for(int i2 = 0, n2 = pout.size(); i2 < n2; ++i2) {
                final Arc arc = pout.get(i2);
                final Transition t = arc.getTransition();
                
                if(arc.isInhibitorArc() && (this.getBindings(t) != null)) {
                    // check enabled transitions
                    //System.out.println("e?d_i: " + t.getName());
                    transitionsToCheck.put(t, t);
                } 
                
                if(!arc.isSimpleArc() && !arc.isReadArc()) {
                    continue;
                }
                // check disabled transitions
                //System.out.println("d?e: " + t.getName());
                transitionsToCheck.put(t, t);
            }
            
            // check capacity
            // transitions that would output to that place
            if(!p.infiniteCapacity()) {
                final List<Arc> pin = p.getInput();
                for(int i3 = 0, n3 = pin.size(); i3 < n3; ++i3) {
                    final Arc arc = pin.get(i3);
                    final Transition t = arc.getTransition();
                    if(this.getBindings(t) != null) {
                        //System.out.println("x?x: " + t.getName());
                        transitionsToCheck.put(t, t);
                    }
                }
            }
        }
        
        this.checkTransitions(transitionsToCheck.values());
        
        
        
        ++this.fired;
        this.checkState(trans);
        this.addToSequence(trans);
        
        
        this.addTokenCountEpochs();
        return bindingString;
    }
    
    
    public void doTake(final Binding bind) {
        final Transition trans = bind.getTransition();
        Simulator.bindOnTransition(trans, bind);
        final List<Arc> in = trans.getInput();
        for(int i = 0, n = in.size(); i < n; ++i) {
            final Arc arc = in.get(i);
            if(!arc.isSimpleArc()) {
                continue;
            }
            try {
                final Place p = arc.getPlace();
                
                final Label label = arc.getLabel();
                final ValueList all = label.getAll();
                for(int i2 = 0, n2 = all.size(); i2 < n2; ++i2) {
                    final Value value = all.get(i2);
                    if(!p.removeToken(value)) {
                        value.setNetValue();
                        throw new RuntimeException("Cannot take " + value
                                        + " from place " + p);
                    }
                }
            } catch(final Exception e) {
                throw new RuntimeException("ERROR firing " + trans + "\nBind: "
                              + bind + "\nArc: " + arc + "\n" + e.getMessage());
            }
        }
        
        // now handle reset arcs
        for(int i = 0, n = in.size(); i < n; ++i) {
            final Arc arc = in.get(i);
            if(!arc.isResetArc()) {
                continue;
            }
            final Place p = arc.getPlace();
            final Label label = arc.getLabel();
            final ValueList all = label.getAll();
            for(int i2 = 0, n2 = all.size(); i2 < n2; ++i2) {
                p.removeAllTokens(all.get(i2));
            }
        }
        Simulator.unbindOnTransition(trans, bind);
    }
    
    
//    public void undoTake(final Binding bind) {
//        final Transition trans = bind.getTransition();
//        Simulator.bindOnTransition(trans, bind);
//        for(final Arc arc : trans.getInput()) {
//            if(arc.isResetArc()) {
//                throw new UnsupportedOperationException("Cannot undo firing "
//                                + "when reset arcs where involvend.");
//            }
//            if(!arc.isSimpleArc()) {
//                continue;
//            }
//            try {
//                final Place p = arc.getPlace();
//                
//                final Label label = arc.getLabel();
//                for(final Value value : label.getValues()) {
//                    p.addToken(value);
//                }
//                for(final Value tuple : label.getTuples()) {
//                    final Value addThis;
//                    if(tuple.getTuple().containsVariable()) {
//                        addThis = Value.getValue(tuple.getTuple().clone());
//                    } else {
//                        addThis = tuple;
//                    }
//                    p.addToken(addThis);
//                }
//                for(final Value var : label.getVariables()) {
//                    p.addToken(var);
//                }
//            } catch(final Exception e) {
//                throw new RuntimeException("ERROR firing " + trans + "\nBind: "
//                              + bind + "\nArc: " + arc + "\n" + e.getMessage());
//            }
//        }
//        Simulator.unbindOnTransition(trans, bind);
//    }
    
    
    public void doPut(final Binding bind) {
        final Transition trans = bind.getTransition();
        Simulator.bindOnTransition(trans, bind);
        final List<Arc> out = trans.getOutput();
        for(int i = 0, n = out.size(); i < n; ++i) {
            final Arc arc = out.get(i);
            if(!arc.isSimpleArc()) {
                continue;
            }
            final Place p = arc.getPlace();
            
            final Label label = arc.getLabel();
            final ValueList all = label.getAll();
            for(int i2 = 0, n2 = all.size(); i2 < n2; ++i2) {
                p.addToken(all.get(i2).clone());
            }
        }
        Simulator.unbindOnTransition(trans, bind);
    }
    
//    public void undoPut(final Binding bind) {
//        final Transition trans = bind.getTransition();
//        Simulator.bindOnTransition(trans, bind);
//        for(final Arc arc : trans.getOutput()) {
//            if(!arc.isSimpleArc()) {
//                continue;
//            }
//            final Place p = arc.getPlace();
//            
//            final Label label = arc.getLabel();
//            for(final Value value : label.getValues()) {
//                if(!p.removeToken(value)) {
//                    throw new RuntimeException("Cannot take " + value
//                                    + " from place " + p);
//                }
//            }
//            for(final Value tuple : label.getTuples()) {
//                if(!p.removeToken(tuple)) {
//                    throw new RuntimeException("Cannot take " + tuple
//                                    + " from place " + p);
//                }
//            }
//            for(final Value var : label.getVariables()) {
//                if(!p.removeToken(var)) {
//                    throw new RuntimeException("Cannot take " + var
//                                    + " from place " + p);
//                }
//            }
//        }
//        Simulator.unbindOnTransition(trans, bind);
//    }
    
    
    public String fireTake(final Transition trans, final Binding bind) {
        return this.fireTake(trans, bind, trans.getInput());
    }
    
    
    private String fireTake(final Transition trans, final Binding bind, final List<Arc> arcs) {
        if((trans == null) || (bind == null)) {
            throw new RuntimeException("Cannot fire (taking)" + trans + " with "
                            + "binding " + bind);
        }
        
        if(this.checkFire() && !this.neverFired.isEmpty()) {
            this.neverFired.remove(trans.getName());
        }
        
        final Map<Transition, Transition> transitionsToCheck = Pool.getPool().getMap();
        if(arcs.isEmpty()) {
            transitionsToCheck.put(trans, trans);
        }
        
        final Map<Place, ValueList> inmarking = Pool.getPool().getMap();
        Simulator.bindOnTransition(trans, bind);
        for(int i = 0, n = arcs.size(); i < n; ++i) {
            final Arc arc = arcs.get(i);
            if(arc.isReadArc()) {
                transitionsToCheck.put(trans, trans);
            }
            if(!arc.isSimpleArc()) {
                continue;
            }
            try {
                final Place p = arc.getPlace();
                inmarking.put(p, Pool.getPool().getValueList(p.getCurrentMarkingSet()));
                
                final Label label = arc.getLabel();
                final ValueList all = label.getAll();
                for(int i2 = 0, n2 = all.size(); i2 < n2; ++i2) {
                    final Value value = all.get(i2);
                    if(!p.removeToken(value)) {
                        value.setNetValue();
                        throw new RuntimeException("Cannot take " + value
                                        + " from place " + p);
                    }
                }
            } catch(final Exception e) {
                throw new RuntimeException("ERROR firing " + trans + "\nBind: "
                              + bind + "\nArc: " + arc + "\n" + e.getMessage());
            }
        }
        // now handle reset arcs
        final List<Arc> in = trans.getInput();
        for(int i = 0, n = in.size(); i < n; ++i) {
            final Arc arc = in.get(i);
            if(!arc.isResetArc()) {
                continue;
            }
            final Place p = arc.getPlace();
            if(!inmarking.containsKey(p)) {
                inmarking.put(p, Pool.getPool().getValueList(p.getCurrentMarkingSet()));
            }
            final Label label = arc.getLabel();
            final ValueList all = label.getAll();
            for(int i2 = 0, n2 = all.size(); i2 < n2; ++i2) {
                final Value value = all.get(i2);
                if(value.isAny()) {
                    p.clearMarking();
                } else {
                    p.removeAllTokens(value);
                }
            }
        }
        Simulator.unbindOnTransition(trans, bind);
        
        
        List<Binding> theBindings = this.getFiringBindings(trans);
        if(theBindings == null) {
            theBindings = Pool.getPool().getList();
            this.setFiringBindings(trans, theBindings);
        }
        
        final String bindingString;
        if(this.getProperty(Option.SHOW_FIRING)) {
            bindingString = bind.toStringSlim();
        } else {
            bindingString = null;
        }

//        if(bind.useAgeMemory()) {
//            trans.resetEnabledTime(bind, this.getTime());
//        }
        
        final Binding clone = bind.clone();
        this.removeBinding(bind);
        Simulator.insertSortEarliestFiring(theBindings, clone);
        Simulator.insertSortEarliestFiring(this.putBindingsEarliest, clone);
        
        // iterate over pre set of transition
        for(final Map.Entry<Place, ValueList> entry : inmarking.entrySet()) {
            final Place p = entry.getKey();
            
            // tokens were removed
            // iterate over post set of place
            final List<Arc> out = p.getOutput();
            for(int i = 0, n = out.size(); i < n; ++i) {
                final Arc arc = out.get(i);
                final Transition t = arc.getTransition();
                if(arc.isInhibitorArc()) {
                    // check enabled and disabled transitions
                    //System.out.println("x?x_i: " + t.getName());
                    transitionsToCheck.put(t, t);
                }
                
                if(!arc.isSimpleArc() && !arc.isReadArc() && !arc.isResetArc()) {
                    continue;
                }
                
                if((t == trans) || (this.getBindings(t) != null)) {
                    if(!p.getCurrentMarkingSet().isEmpty() || ((t == trans) && arc.isResetArc())) {
                        //System.out.println("e?d: " + t.getName());
                        transitionsToCheck.put(t, t);
                    } else { // shortcut
                        if(t.hasAgeMemory()) {
                            transitionsToCheck.put(t, t);
                        } else {
//                          System.out.println("shortcut: " + t.getName());
                            this.removeBindings(t);
                        }
                    }
                }
            }
            
            // check capacity
            // transitions that would output to that place
            if(!p.infiniteCapacity()) {
                // iterate over pre set of place
                final List<Arc> pin = p.getInput();
                for(int i = 0, n = pin.size(); i < n; ++i) {
                    final Arc arc = pin.get(i);
                    final Transition t = arc.getTransition();
                    if(this.getBindings(t) == null) {
                        // System.out.println("d?e_c: " + t.getName());
                        transitionsToCheck.put(t, t);
                    }
                }
            }
        }
        
        this.checkTransitions(transitionsToCheck.values());
        
        ++this.fired;
        this.checkState(trans);
        this.addToSequence(trans);
        
        return bindingString;
    }
    
    
//    public String undoFirePut(final Transition trans, final Binding bind) {
//        return this.fireTake(trans, bind, trans.getOutput());
//    }
    
    
    public String firePut(final Transition trans, final Binding bind) {
        return this.firePut(trans, bind, trans.getOutput());
    }
    
    
    private String firePut(final Transition trans, final Binding bind, final List<Arc> arcs) {
        if((trans == null) || (bind == null)) {
            throw new RuntimeException("Cannot fire (putting) " + trans
                            + " with binding " + bind);
        }
        
        final List<Place> outmarking = Pool.getPool().getList();

        Simulator.bindOnTransition(trans, bind);
        for(int i = 0, n = arcs.size(); i < n; ++i) {
            final Arc arc = arcs.get(i);
            if(!arc.isSimpleArc()) {
                continue;
            }
            final Place p = arc.getPlace();
            outmarking.add(p);
            
            final Label label = arc.getLabel();
            final ValueList all = label.getAll();
            for(int i2 = 0, n2 = all.size(); i2 < n2; ++i2) {
                p.addToken(all.get(i2).clone());
            }
        }
        
        Simulator.unbindOnTransition(trans, bind);
        final String bindingString;
        if(this.getProperty(Option.SHOW_FIRING)) {
            bindingString = bind.toStringSlim();
        } else {
            bindingString = null;
        }
        final List<Binding> theBindings = this.getFiringBindings(trans);
        
        Simulator.removeSortEarliestFiring(theBindings, bind, false);
        Simulator.removeSortEarliestFiring(this.putBindingsEarliest, bind, false);
        
        if(theBindings.isEmpty()) {
            this.removeFiringBindings(trans);
        }
        
        
        final Map<Transition, Transition> transitionsToCheck = Pool.getPool().getMap();
        // iterate over post set of transition
        for(int i = 0, n = outmarking.size(); i < n; ++i) {
            final Place p = outmarking.get(i);
            // tokens were added
            // iterate over post set of place
            final List<Arc> out = p.getOutput();
            for(int i2 = 0, n2 = out.size(); i2 < n2; ++i2) {
                final Arc arc = out.get(i2);
                final Transition t = arc.getTransition();
                
                if(arc.isInhibitorArc() && (this.getBindings(t) != null)) {
                    // check enabled transitions
                    //System.out.println("e?d_i: " + t.getName());
                    transitionsToCheck.put(t, t);
                } 
                
                if(!arc.isSimpleArc() && !arc.isReadArc()) {
                    continue;
                }
                // check disabled transitions
                //System.out.println("d?e: " + t.getName());
                transitionsToCheck.put(t, t);
            }
            
            // check capacity
            // transitions that would output to that place
            if(!p.infiniteCapacity()) {
                final List<Arc> in = p.getInput();
                for(int i2 = 0, n2 = in.size(); i2 < n2; ++i2) {
                    final Arc arc = in.get(i2);
                    final Transition t = arc.getTransition();
                    if(this.getBindings(t) != null) {
                        //System.out.println("e?d_c: " + t.getName());
                        transitionsToCheck.put(t, t);
                    }
                }
            }
        }
        
        this.checkTransitions(transitionsToCheck.values());

        this.checkState(trans);
        
        this.addTokenCountEpochs();
        return bindingString;
    }
    
    
//    public String undoFireTake(final Transition trans, final Binding bind) {
//        if(this.usesResetArcs()) {
//            for(final Arc arc  trans.getInput()) {
//                if(arc.isResetArc()) {
//                    throw new UnsupportedOperationException("Cannot undo firing when reset arcs are involved.");
//                }
//            }
//        }
//        return this.firePut(trans, bind, trans.getInput());
//    }
        
        
    public boolean isDeadlock() {
        return this.enabledTrans.isEmpty() && (this.firingTrans.isEmpty() || !this.firesParallel());
    }
    
    
    public Collection<String> getEnabledTransitionsPut() {
        final Collection<String> transNames = Pool.getPool().getList();
        for(final Transition t : this.firingTrans.values()) {
            transNames.add(t.getName());
        }
        return transNames;
    }
    
    
    public Collection<String> getEnabledTransitionsTake() {
        return this.getEnabledTransitions();
    }
    
    
    public Collection<String> getEnabledTransitions() {
        final Collection<String> transNames = Pool.getPool().getList();
        for(final Transition t : this.enabledTrans.values()) {
            transNames.add(t.getName());
        }
        return transNames;
    }
    
    
    public Collection<String> getEnabledBindings() {
        final Collection<String> transNames = Pool.getPool().getList();
        for(int i = 0; i < this.bindings.length; ++i) {
            final List<Binding> l = this.bindings[i];
            if(l != null) {
                for(final Binding b : l) {
                    transNames.add(b.toString());
                }
            }
        }
        return transNames;
    }
    
    
    public Collection<String> getFireableTransitions() {
        final Collection<String> transNames = Pool.getPool().getList();
        for(final Binding b : this.bindingsImmediate) {
            transNames.add(b.toString());
        }
        final int max = this.getLastFireableBinding(this.bindingsEarliest);
        if(max >= 0) {
            for(int i = 0; i <= max; ++i) {
                transNames.add(this.bindingsEarliest.get(i).toString());
            }
        }
        return transNames;
    }
    
    
    public String fireWithBinding(final String complete) {
        final int start = complete.indexOf('{');
        final String transName = complete.substring(0, start).trim();
        final Transition trans = this.net.getTransitionByName(transName);
        if(trans == null) {
            return null;
        }
        final Binding bind = Simulator.parseBinding(complete.substring(start+1,
                complete.lastIndexOf("}")));
        bind.setTransition(trans);
        final List<Binding> binds = this.getBindingForTransition(trans, bind, Simulator.DFLT_ALL_BINDINGS);
        if((binds != null) && !binds.isEmpty()) {
            final Binding b = Simulator.choseRandom(binds);
            final int index = this.getBindings(trans).indexOf(b);
            if(index <= -1) {
                throw new RuntimeException("Binding\n" + b + "\nnot found in\n"
                                + this.getBindings(trans));
            }
            final Binding theBind = this.getBindings(trans).get(index);
            final String resultingBinding = this.fire(trans, theBind, trans.getInput(), trans.getOutput());
            return resultingBinding;
        } else {
            return null;
        }
    }
    
    
//    public String undoFireWithBinding(final String complete) {
//        final int start = complete.indexOf('{');
//        final String transName = complete.substring(0, start).trim();
//        final Transition trans = this.net.getTransitionByName(transName);
//        if(trans == null) {
//            return null;
//        }
//        final Binding bind = Simulator.parseBinding(complete.substring(start+1,
//                complete.lastIndexOf("}")));
//        bind.setTransition(trans);
//        final List<Binding> binds = this.getBindingForTransition(trans, bind, this.functions, trans.getOutput(), trans.getInput(), Simulator.DFLT_ALL_BINDINGS, this.getTime(), );// TODO
//        if((binds != null) && !binds.isEmpty()) {
//            final Binding b = Simulator.choseRandom(binds);
//            final String resultingBinding = this.fire(trans, b, trans.getOutput(), trans.getInput());
//            Pool.getPool().returnBinding(bind);
//            for(final Binding binding : binds) {
//                Pool.getPool().returnBinding(binding);
//            }
//            Pool.getPool().returnValueList(binds);
//            return resultingBinding;
//        } else {
//            return null;
//        }
//    }
    
    
    public String fireTakeWithBinding(final String complete) {
        final int start = complete.indexOf('{');
        final String transName = complete.substring(0, start).trim();
        final Transition trans = this.net.getTransitionByName(transName);
        if(trans == null) {
            return null;
        }
        final Binding bind = Simulator.parseBinding(complete.substring(start+1,
                        complete.lastIndexOf("}")));
        bind.setTransition(trans);
        final List<Binding> binds = this.getBindingForTransition(trans, bind, Simulator.DFLT_ALL_BINDINGS);
        if((binds != null) && !binds.isEmpty()) {
            binds.retainAll(this.getBindings(trans));
            final Binding b = Simulator.choseRandom(binds);
            // search for the binding (equality)
            final int index = Simulator.indexOfSortEarliest(this.getBindings(trans), b, true);
            final Binding theBind = this.getBindings(trans).get(index);
            final String resultingBinding = this.fireTake(trans, theBind);
            
            return resultingBinding;
        } else {
            return null;
        }
    }
    

//    public String undoFireTakeWithBinding(final String complete) {
//        final int start = complete.indexOf('{');
//        final String transName = complete.substring(0, start).trim();
//        final Transition trans = this.net.getTransitionByName(transName);
//        if(trans == null) {
//            return null;
//        }
//        final Binding bind = Simulator.parseBinding(complete.substring(start+1,
//                        complete.lastIndexOf("}")));
//        bind.setTransition(trans);
//        final List<Binding> binds = this.getBindingForTransition(trans, bind, Simulator.DFLT_ALL_BINDINGS);
//        if((binds != null) && !binds.isEmpty()) {
//            final Binding b = Simulator.choseRandom(binds);
//            final String resultingBinding = this.undoFireTake(trans, b);
//            Pool.getPool().returnBinding(bind);
//            Pool.getPool().returnBinding(b);
//            Pool.getPool().returnValueList(binds);
//            return resultingBinding;
//        } else {
//            return null;
//        }
//    }
    
    
    private static double getNegativeExponentialRandomValue(final double mu) {
        final double y = Simulator.random.nextDouble();
        final double delta = -Math.log(y)*mu;
        return delta;
    }
    

    private void setRateAndWeight(final Binding bind, boolean bound) {
        final Transition t = bind.getTransition();
        if(t.hasWeight()) {
            bind.setCurrentWeight(Simulator.getNegativeExponentialRandomValue(bind.getWeight()));
            
        } else if(t.hasRate()) {
            final double delta = Simulator.getNegativeExponentialRandomValue(bind.getRate());
            final double fireAt = this.getTime() + delta;
            bind.setEarliestFiring(fireAt);
            bind.setLatestFiring(fireAt);
            
        } else if(t.hasGeneralizedWeight()) {
            try {
                if(!bound) {
                    Simulator.bindOnTransition(t, bind);
                }
                synchronized(this.properties) {
                    this.properties.put(Simulator.PROPERTY_TIME, Value.getValue(this.getTime()));
                }
                
                final Node rate = t.getGeneralizedWeightTree();
                final Value v = TreeComputer.evaluate(rate, this.functions,
                                bind, this.properties);
                if(!v.isNumber()) {
                    throw new RuntimeException("Result is not a number but " + v);
                }
                final double val = v.getNumber();
                if(Double.isNaN(val)) {
                    throw new RuntimeException("Result is not a number: " + val);
                }
                bind.setWeight(val);
            } finally {
                synchronized(this.properties) {
                    this.properties.remove(Simulator.PROPERTY_TIME);
                }
                if(!bound) {
                    Simulator.unbindOnTransition(t, bind);
                }
            }
            bind.setCurrentWeight(bind.getWeight());
            
        } else if(t.hasGeneralizedRate()) {
            try {
                if(!bound) {
                    Simulator.bindOnTransition(t, bind);
                }
                synchronized(this.properties) {
                    this.properties.put(Simulator.PROPERTY_TIME, Value.getValue(this.getTime()));
                }
                
                final Node rate = t.getGeneralizedRateTree();
                final Value v = TreeComputer.evaluate(rate, this.functions,
                                bind, this.properties);
                if(!v.isNumber()) {
                    throw new RuntimeException("Result is not a number but " + v);
                }
                final double val = v.getNumber();
                if(Double.isNaN(val)) {
                    throw new RuntimeException("Result is not a number: " + val);
                }
                bind.setRate(val);
            } finally {
                synchronized(this.properties) {
                    this.properties.remove(Simulator.PROPERTY_TIME);
                }
                if(!bound) {
                    Simulator.unbindOnTransition(t, bind);
                }
            }
            final double fireAt = this.getTime() + bind.getRate();
            bind.setEarliestFiring(fireAt);
            bind.setLatestFiring(fireAt);
        }
    }
    
    
    private static Binding parseBinding(String binding) {
        final Binding bind = Pool.getPool().getBinding();
        if(binding.startsWith("{")) {
            binding = binding.substring(1, binding.length()); 
        }
        if(binding.endsWith("}")) {
            binding = binding.substring(0, binding.length()-1); 
        }
        binding = binding.trim();
        if(binding.length() != 0) {
            final String[] tokens = binding.split("\\s*,\\s*");
            for(int i = 0; i < tokens.length; i++) {
                int idx = tokens[i].indexOf("/");
                if(idx < 0) {
                    idx = tokens[i].indexOf("=");
                }
                if(idx < 0) {
                    continue; // bogus...
                }
                final String var = tokens[i].substring(0, idx);
                final String value = tokens[i].substring(idx+1);
                final ValueList l = Pool.getPool().getValueList();
                l.add(Value.getValue(value));
                bind.constrain(var, l, true);
            }
        }
        return bind;
    }
    
    
    public Simulator.Status getState() {
        return this.state;
    }
    
    
    public void setUseStopCondition(final boolean v) {
        this.setProperty(Simulator.Option.USE_STOP_CONDITION, v);
    }
    
    
    public boolean useStopCondition() {
        return this.getProperty(Simulator.Option.USE_STOP_CONDITION);
    }
    
    
    public void stopAtCount(final int count) {
        if(count > 0) {
            this.stopAtCount = count;
            this.setUseStopCondition(true);
        }
    }
    
    
    public void stopAfterFiring(final String transition) {
        this.stopAtTransName.add(transition);
        this.setUseStopCondition(true);
    }
    
    
    public void stopAtMarking(final String place, final String marking) {
        this.stopAtMarking.put(this.net.getPlaceByName(place),
                    Place.parseMarking(marking));
        this.setUseStopCondition(true);
    }
    
    
    public void removeStopConditions() {
        this.setUseStopCondition(false);
        this.stopAtMarking.clear();
        this.stopAtTransName.clear();
        this.stopAtCount = 0;
    }
    
    
    public void addToSequence(final Transition trans) {
        if(this.maxSeqLength > 0) {
            if(this.sequence.size() == this.maxSeqLength) {
                this.sequence.remove(0);
            }
            this.sequence.add(trans.getName());
        }
    }
    
    
    private boolean cannotFire() {
        return this.isDeadlock() || (this.bindingsImmediate.isEmpty()
                        && (this.bindingsEarliest.get(0).getEarliestFiring() > this.getTime()));
    }
    
    
    private void checkState(final Transition trans) {
        if(this.isDeadlock()) {
            this.stop();
            this.state = Simulator.Status.DEADLOCK;
            return;
        } else if(this.usesTime() && this.cannotFire()) {
            //this.stop();
            this.state = Simulator.Status.INCREMENT_CLOCK;
            return;
        } else if(this.useStopCondition()) {
            if((this.stopAtCount > 0) && (this.fired >= this.stopAtCount)) {
                this.stop();
                this.state = Simulator.Status.COUNT_REACHED;
                return;
            }
            
            if((trans != null) && this.stopAtTransName.contains(trans.getName())) {
                this.stop();
                this.state = Simulator.Status.TRANSITION_FIRED;
                return;
            }
            
            for(final Map.Entry<Place, ValueList> e : this.stopAtMarking.entrySet()) {
                final Place p = e.getKey();
                final ValueList marking = e.getValue();
                if((p != null) && PNModel.setEquals(p.getCurrentMarkingSet(), marking)) {
                    this.stop();
                    this.state = Simulator.Status.MARKING_REACHED;
                    return;
                }
            }
        }
        
        this.state = Simulator.Status.RUNNING;
    }
    
    
    
    public void resetMarking() {
        if(this.net != null) {
            this.net.resetMarking();
        }
    }

    
    public void exit() {
        this.stop();
    }
    
    
    public void stop() {
        if(this.bindings != null) {
            for(int i = 0; i < this.bindings.length; ++i) {
                this.bindings[i] = null;
            }
        }
        this.putBindingsEarliest.clear();
    }
    
    
    public void enableCheckFire() {
        this.setCheckFire(true);
    }
    
    
    public void disableCheckFire() {
        this.setCheckFire(false);
    }
    
    
    public void setThreadCount(final int count) {
        if(count < 2) {
            if(this.binders != null) {
                for(int i = this.binders.length-1; i >= 0; --i) {
                    this.binders[i].pleaseStop();
                }
                for(int i = this.binders.length-1; i >= 0; --i) {
                    try {
                        this.binders[i].join();
                    } catch (InterruptedException e) {
                        // ignore
                    }
                }
            }
            this.binders = null;
            this.threadMarkings.clear();
            this.threadUsed.clear();
        } else {
            this.binders = new Binder[count];
            for(int i = 0; i < count; ++i) {
                this.binders[i] = new Binder(this);
                this.binders[i].start();
            }
        }
    }
    
    
    private class Binder extends Thread {
        private LinkedList<Transition> transitions;
        private Simulator sim;
        private Object mon = new Object();
        private volatile boolean pleaseStop = false;
        private volatile boolean isWorking = false;
        
        public Binder(final Simulator sim) {
            this.sim = sim;
            this.transitions = new LinkedList<Transition>();
        }
        
        
        public void addTransition(Transition trans) {
            synchronized(this.transitions) {
                this.isWorking = true;
                this.transitions.add(trans);
                this.transitions.notifyAll();
            }
        }
        
        
        public void pleaseStop() {
            this.pleaseStop = true;
            synchronized(this.transitions) {
                this.transitions.notifyAll();
            }
        }
        
        
        public boolean isWorking() {
            return this.isWorking;
        }
        
        
        public void waitWhileWorking() {
            synchronized(this.mon) {
                while(!this.pleaseStop && this.isWorking()) {
                    try {
                        this.mon.wait();
                    } catch(final InterruptedException e) {
                        //
                    }
                }
            }
        }
        
        
        public void run() {
            while(!this.pleaseStop) {
                synchronized(this.transitions) {
                    while(!this.pleaseStop && this.transitions.isEmpty()) {
                        try {
                            this.isWorking = false;
                            synchronized(this.mon) {
                                this.mon.notifyAll();
                            }
                            this.transitions.wait();
                        } catch(final InterruptedException e) {
                            //
                        }
                    }
                }
                if(!this.pleaseStop) {
                    final Transition t;
                    synchronized(this.transitions) {
                        t = this.transitions.removeFirst();
                    }
                    try {
                        this.sim.checkTransition(t);
                    } catch(final RuntimeException e) {
                        throw new RuntimeException("Error checking " + t, e);
                    }
                }
            }
            this.isWorking = false;
        }
    }
    
    
    
    public static void writeToFile(final String net, final String filename) {
        try {
            final File file = new File(filename);
            final FileWriter outputstream = new FileWriter(file);
            final PrintWriter output = new PrintWriter(outputstream);
            output.print(net);
            output.close();
        } catch (final IOException ioex) {
            throw new RuntimeException(ioex);
        }
    }
    
    
    public static void main(final String[] args) {
        final Simulator sim = new Simulator();
        
        String file = null;
        String unfold = null;
        String output = null;
        String fireSequence = null;
        boolean monitorAll = false;
        boolean measureAllEpochs = false;
        boolean measureAllTime = false;
        boolean debug = false;
        boolean quiet = false;
        boolean parallel = false;
        boolean active = false;
        boolean sequence = false;
        boolean reset = false;
        double probability = 0;
        int delay = 0;
        int measureCount = 0;
        int checkFireCount = 0;
        final List<String> measureEpochs = Pool.getPool().getList();
        final List<String> measureTime = Pool.getPool().getList();
        final Map<String, String> places = Pool.getPool().getMap();
        
        for(int i = 0; i < args.length; i++) {
            final String arg = args[i];
            if(arg.equals("-h")) {
                System.out.print("Usage: java petruchio.sim.pnmodel.simulator.Simulator <options>\n"
                         + "  -h         print this help\n"
                         + "  -f<NET>    load and possibly simulate the file\n"
                         + "  -u<LL_NET> unfolds, minimizes the loaded net and writes the ll_net\n"
                         + "  -o<NET>    minimizes the loaded net and writes it to the file\n"
                         + "  -m         monitor all the place's markings\n"
                         + "  -m<PLACE>  monitor the place's marking\n"
                         + "  -m<P>:<N>  monitor the place's marking but write N instead of the name\n"
                         + "  -ee        calculate average markings for all places (firing epochs)\n"
                         + "  -ee<PLACE> calculate average marking for this place (firing epochs)\n"
                         + "  -et        calculate average markings for all places (time units)\n"
                         + "  -et<PLACE> calculate average marking for this place (time units)\n"
                         + "  -d<DELAY>  before firing, wait this long [ms]\n"
                         + "  -F<COUNT>  tell which transitions weren't fired yet\n"
                         + "  -bT<TRANS> stop after firing the transition\n"
                         + "  -bC<COUNT> stop after firing that many transitions\n"
                         + "  -bM<P>:<M> stop when place P is marked with M\n"
                         + "  -v<REGEX>  verbose binding checking for the transitions that match the regex\n"
                         + "  -s<LENGTH> print the fired sequence (maximum length)\n"
                         + "  -s<SEQ>    try to fire the sequence f.e. T1,T2,T3{X/2},T10{A/0},...\n"
                         + "  -r<FILE>   write the PEP .hl_ll_ref file\n"
                         + "  -R         write the PEP .hl_ll_ref file with infos on bindings and values\n"
                         + "  -lp        use local priority instead of global priority\n"
                         + "  -t<COUNT>  show speed every COUNT transitions\n"
//                         + "  -T<COUNT>  use multithreading with count threads (experimental)\n"
                         + "  -td        (age mem + multiple servers) don't distribute time equally but first come first serve\n"
                         + "  -tt<PROB>  simulate a timed transition net - strict latest firing\n"
                         + "  -tT<PROB>  simulate a timed transition net - weak latest firing\n"
                         + "               PROB is the probability (f.e. 0.8) to increment the clock instead\n"
                         + "               of firing a transition\n"
                         + "  -r         reset to initial marking\n"
                         + "  -v         tell which transitions are fired\n"
                         + "  -vv        tell how terms are simplified\n"
                         + "  -p         enable parallel firing\n"
                         + "  -a         show active transitions\n"
                         + "  -q         don't output 'normal' stuff\n");
                System.exit(0);
                
            } else if(arg.startsWith("-f")) {
                file = arg.substring(2);
                if(!quiet) {
                    System.out.println("FILE: " + file);
                }
                
            } else if(arg.startsWith("-u")) {
                unfold = arg.substring(2);
                if(!quiet) {
                    System.out.println("UNFOLD and MINIMIZE: " + unfold);
                }
                
            } else if(arg.startsWith("-o")) {
                output = arg.substring(2);
                if(!quiet) {
                    System.out.println("WRITE NET TO: " + output);
                }
                
            } else if(arg.startsWith("-F")) {
                checkFireCount = Integer.parseInt(arg.substring(2));
                if(checkFireCount < 0) {
                    System.err.println("error reading option -F<COUNT>: count has to be >= 0.");
                    System.exit(1);
                }
                if(!quiet) {
                    System.out.println("SHOW NOT FIRED: " + checkFireCount);
                }
                sim.enableCheckFire();
                
            } else if(arg.equals("-lp")) {
                sim.setUseGlobalPriority(false);
                if(!quiet) {
                    System.out.println("GLOBAL PRIORITY: off");
                }
                
            } else if(arg.equals("-m")) {
                monitorAll = true;
                if(!quiet) {
                    System.out.println("MONITOR ALL PLACES: on");
                }
                
            } else if(arg.equals("-ee")) {
                measureAllEpochs = true;
                if(!quiet) {
                    System.out.println("MEASURE ALL PLACES (EPOCHS): on");
                }
                
            } else if(arg.equals("-et")) {
                measureAllTime = true;
                if(!quiet) {
                    System.out.println("MEASURE ALL PLACES (TIME): on");
                }
                
            } else if(arg.equals("-td")) {
                sim.setDistributeTime(false);
                if(!quiet) {
                    System.out.println("EQUALLY DISTRIBUTE ACUMULATED AGE MEM TIME: off");
                }
                
            } else if(arg.equals("-r")) {
                reset = true;
                if(!quiet) {
                    System.out.println("RESET: on");
                }
                
            } else if(arg.startsWith("-r")) {
                sim.setBuildHL2LLRef(arg.substring(2));
                if(!quiet) {
                    System.out.println("WRITE hl_ll_ref TO: " + arg.substring(2));
                }
                
            } else if(arg.startsWith("-ee")) {
                measureEpochs.add(arg.substring(2));
                if(!quiet) {
                    System.out.println("MEASURING PLACE (EPOCHS): " + arg.substring(2));
                }
                
            } else if(arg.startsWith("-et")) {
                measureTime.add(arg.substring(2));
                if(!quiet) {
                    System.out.println("MEASURING PLACE (TIME): " + arg.substring(2));
                }
                
            } else if(arg.equals("-R")) {
                sim.setExtendedHL2LLRef(true);
                if(!quiet) {
                    System.out.println("MORE INFOS IN hl_ll_ref: on");
                }
                
            } else if(arg.equals("-q")) {
                quiet = true;
                
            } else if(arg.startsWith("-m")) {
                String place = arg.substring(2);
                if(place.length() > 0) {
                    final int idx = place.indexOf(":");
                    final String name;
                    final String alias;
                    if(idx >= 0) {
                        name = place.substring(0, idx);
                        alias = place.substring(idx+1);
                        places.put(name, alias);
                    } else {
                        places.put(place, place);
                        name = alias = null;
                    }
                    if(!quiet) {
                        System.out.println("MONITOR PLACE: "
                                    + (idx >= 0 ? name + " as " + alias : place));
                    }
                }
                
            } else if(arg.startsWith("-d")) {
                delay = Integer.parseInt(arg.substring(2));
                if(delay < 0) {
                    System.err.println("error reading option -d<DELAY>: delay has to be >= 0.");
                    System.exit(1);
                }
                if(!quiet) {
                    System.out.println("DELAY: " + delay + " ms");
                }
                
            } else if(arg.startsWith("-s")) {
                try {
                    int length = Integer.parseInt(arg.substring(2));
                    if(length <= 0) {
                        System.err.println("error reading option -s<LENGTH>: length has to be > 0.");
                        System.exit(1);
                    }
                    sim.saveSequence(length);
                    sequence = true;
                    if(!quiet) {
                        System.out.println("SEQUENCE: " + length + " max length");
                    }
                } catch(NumberFormatException e) {
                    fireSequence = arg.substring(2);
                    if(!quiet) {
                        System.out.println("SEQUENCE: " + fireSequence);
                    }
                }
                
            } else if(arg.startsWith("-tT")) {
                final String val = arg.substring(3);
                if(val.length() == 0) {
                    probability = 0;
                } else {
                    probability = Double.parseDouble(val);
                }
                if(probability < 0) {
                    probability = 0;
                } else if(probability > 1) {
                    probability = 1;
                }
                sim.setUsesTime(true);
                sim.setUsesStrictTime(false);
                if(!quiet) {
                    System.out.println("TIMED TRANSITION NET: weak (" + probability + ")");
                }
                
            } else if(arg.startsWith("-tt")) {
                final String val = arg.substring(3);
                if(val.length() == 0) {
                    probability = 0;
                } else {
                    probability = Double.parseDouble(val);
                }
                if(probability < 0) {
                    probability = 0;
                } else if(probability > 1) {
                    probability = 1;
                }
                sim.setUsesTime(true);
                sim.setUsesStrictTime(true);
                if(!quiet) {
                    System.out.println("TIMED TRANSITION NET: strict (" + probability + ")");
                }
                
            } else if(arg.startsWith("-T")) {
                int count = Integer.parseInt(arg.substring(2));
                if(count < 0) {
                    System.err.println("error reading option -T<COUNT>: count has to be >= 0.");
                    System.exit(1);
                }
                System.out.println("THREAD COUNT: " + count);
                sim.setThreadCount(count);
                
            } else if(arg.equals("-v")) {
                if(!quiet) {
                    System.out.println("DEBUG: on");
                }
                sim.setProperty(Option.SHOW_FIRING, true);
                debug = true;
                
            } else if(arg.equals("-vv")) {
                if(!quiet) {
                    System.out.println("SHOW SIMPLIFICATIONS: on");
                }
                sim.setShowSimplifications(true);
                
            } else if(arg.startsWith("-v")) {
                final String regex = arg.substring(2);
                sim.checkTransitions(regex);
                if(!quiet) {
                    System.out.println("VERBOSE CHECK: " + regex );
                }
                
            } else if(arg.startsWith("-t")) {
                measureCount = Integer.parseInt(arg.substring(2));
                if(measureCount < 0) {
                    System.err.println("error reading option -t<COUNT>: count has to be >= 0.");
                    System.exit(1);
                }
                if(!quiet) {
                    System.out.println("SHOW SPEED: " + measureCount);
                }
                
            } else if(arg.startsWith("-bT")) {
                final String stopAtTrans = arg.substring(3);
                if(!quiet) {
                    System.out.println("STOP AFTER FIRING: " + stopAtTrans);
                }
                sim.stopAfterFiring(stopAtTrans);
                
            } else if(arg.startsWith("-bM")) {
                final int idx = arg.indexOf(":");
                final String stopAtPlace = arg.substring(3, idx);
                final String stopAtMarking = arg.substring(idx+1);
                if(!quiet) {
                    System.out.println("STOP AT MARKING: " + stopAtPlace
                            + " with marking " + stopAtMarking);
                }
                sim.stopAtMarking(stopAtPlace, stopAtMarking);
                
            } else if(arg.startsWith("-bC")) {
                final int stopAtCount = Integer.parseInt(arg.substring(3));
                if(stopAtCount <= 0) {
                    System.err.println("error reading option -F<COUNT>: count has to be > 0.");
                    System.exit(1);
                }
                if(!quiet) {
                    System.out.println("STOP AFTER FIRING: " + stopAtCount + " transitions");
                }
                sim.stopAtCount(stopAtCount);
                
            } else if(arg.equals("-a")) {
                if(!quiet) {
                    System.out.println("SHOW ACTIVE: on");
                }
                active = true;
                
            } else if(arg.equals("-p")) {
                if(!quiet) {
                    System.out.println("PARALLEL: on");
                }
                sim.setFiresParallel(true);
                parallel = true;
                
            } else {
                if(!quiet) {
                    System.out.println("Ignoring unknown option: " + arg);
                }
            }
        }
        
        if((file == null) || (file.length() == 0)) {
            System.err.println("No file specified. Start \'java petruchio.sim.pnmodel."
                            + "simulator.Simulator -h\' for help.");
            System.exit(1);
        }
        
        File f = new File(file);
        if(!f.isAbsolute()) {
            f = new File(System.getProperty("user.dir") + File.separator + file);
        }
        
        try {
            file = f.getCanonicalPath();
        } catch(IOException e) {
            file = f.getAbsolutePath();
        }
        
        try {
            final long start = System.currentTimeMillis();
            sim.loadFile(file, unfold == null);
            final long dur = System.currentTimeMillis() - start;
            if(!quiet) {
                System.out.println("Loaded " + file + " in " + dur + " ms");
            }
        } catch(final Exception e) {
            e.printStackTrace();
            System.err.println("Unable to load " + file + "\n" + e.getMessage());
        }
        
        if(reset) {
            sim.resetMarking();
        }
        
        if(unfold != null) {
            if(!quiet) {
                System.out.println("== UNFOLDING ==");
            }
            final long start = System.currentTimeMillis();
            final PetriNet p = sim.unfold();
            final long dur = System.currentTimeMillis() - start;
            p.toPEPFile(unfold);
            if(!quiet) {
                System.out.println("== DONE after " + dur + " ms");
            }
            return;
        } else if(output != null) {
            if(!quiet) {
                System.out.println("== WRITE NET ==");
            }
            final PetriNet p = sim.getNet();
            p.toPEPFile(output);
            if(!quiet) {
                System.out.println("== DONE ==");
            }
            return;
        }
        
        
        if(measureAllEpochs) {
            sim.measureAllPlacesEpochs();
        } else {
            for(final String s : measureEpochs) {
                sim.measurePlaceExpochs(s);
            }
        }
        
        if(measureAllTime) {
            sim.measureAllPlacesTime();
        } else {
            for(final String s : measureTime) {
                sim.measurePlaceTime(s);
            }
        }
        
        if(!quiet) {
            System.out.println("== STARTING SIMULATION ==");
        }
        
        final List<String> seq;
        final Iterator<String> it;
        if(fireSequence != null) {
            seq = Pool.getPool().getList();
            final Matcher m = Pattern.compile("[^\\{,]+?(?=,|$)|[^,]+?\\}(?=,|$)").matcher(fireSequence);
            while(m.find()) {
                String step = fireSequence.substring(m.start(), m.end());
                if(!step.endsWith("}")) {
                    step += "{}";
                }
                seq.add(step);
//                System.out.println(step);
            }
            it = seq.iterator();
        } else {
            seq = null;
            it = null;
        }
        
        final long start = System.currentTimeMillis();
        if(!places.isEmpty()) {
            System.out.println("marking: " + sim.getMarking(places));
        }
        if(monitorAll) {
            System.out.println("marking: " + sim.getMarking());
        }
        if(active) {
            System.out.println("enabled:  " + sim.getEnabledBindings());
            if(sim.usesTime()) {
                System.out.println("fireable: " + sim.getFireableTransitions());
            }
        }
        Simulator.Status state = Simulator.Status.IDLE;
        
        while(((state = sim.getState()) == Simulator.Status.RUNNING)
                        || (state == Simulator.Status.INCREMENT_CLOCK)) {
            if(delay > 0) {
                try {
                    Thread.sleep(delay);
                } catch(InterruptedException e) {
                    // ignore
                }
            }
            
            final boolean clock = sim.usesTime() && (Math.random() < probability);
            
            if(sim.usesTime() && (clock || (sim.getState() == Simulator.Status.INCREMENT_CLOCK))) {
                if(debug) {
                    if(clock) {
                        System.out.println("Rolled < " + probability + " => Increment clock.");
                    } else {
                        System.out.println("No transitions fireable => Increment clock.");
                    }
                }
                final String fired;
                fired = sim.incTime();
                if(debug && (fired.length() > 0)) {
                    System.out.print(fired);
                }
                if(debug) {
                    System.out.println("Clock: " + sim.getTime());
                }
                
            } else {
                if(debug && sim.usesTime()) {
                    System.out.println("Rolled >= " + probability + " => Fire transition.");
                }
                
                final String binding;
                if((it != null) && it.hasNext()) {
                    String next = it.next();
                    System.out.println("firing: " + next);
                    binding = sim.fireWithBinding(next);
                    if(binding == null) {
                        throw new RuntimeException("Unable to fire " + next);
                    }
                    if(!it.hasNext()) {
                        System.out.println("SEQUENCE FIRED.");
                    }
                    
                } else if(parallel) {
                    binding = sim.fireNextStep();
                } else {
                    binding = sim.fireNext();
                }
                
                if(debug) {
                    System.out.println("Fired: " + binding);
                }
                
                if(!places.isEmpty()) {
                    System.out.println("marking: " + sim.getMarking(places));
                }
                if(monitorAll) {
                    System.out.println("marking: " + sim.getMarking());
                }
                if((checkFireCount > 0) && (sim.getFiredCount() % checkFireCount == 0)) {
                    if(!sim.getNeverFiredTransitions().isEmpty()) {
                        System.out.println(sim.getFiredCount() + " trans fired. Not fired yet: "
                                        + sim.getNeverFiredTransitions());
                    }
                }
                if((measureCount > 0) && (sim.getFiredCount() % measureCount == 0)) {
                    long dur = System.currentTimeMillis() - start;
                    dur = dur > 0 ? dur : 1;
                    System.out.println((1000*sim.getFiredCount()/dur) + " trans/s (" + sim.getFiredCount()
                                    + " trans fired in " + dur + " ms)");
                }
            }
            
            if(active) {
                System.out.println("enabled: " + sim.getEnabledBindings());
                if(sim.usesTime()) {
                    System.out.println("fireable: " + sim.getFireableTransitions());
                }
            }
        }
        sim.pleaseStop();
        if(sequence) {
            System.out.println("Fired sequence:\n" + sim.getSequence());
        }
        if((measureCount > 0) || debug || (checkFireCount > 0)) {
            long dur = System.currentTimeMillis() - start;
            dur = dur > 0 ? dur : 1;
            System.out.println((1000*sim.getFiredCount()/dur) + " trans/s (" + sim.getFiredCount()
                            + " trans fired in " + dur + " ms)");
        }
        
        if(measureAllEpochs || !measureEpochs.isEmpty()) {
            System.out.println(sim.getAverageTokenCountEpochs());
        }
        
        if(measureAllTime || !measureTime.isEmpty()) {
            System.out.println(sim.getAverageTokenCountTime());
        }
        
        final String msg = state.toString();
        
        if(!quiet) {
            System.out.println("The simulator stopped: " + msg);
        }
    }
}
